/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 7001
#define LARGE_STATE_COUNT 1094
#define SYMBOL_COUNT 234
#define ALIAS_COUNT 1
#define TOKEN_COUNT 124
#define EXTERNAL_TOKEN_COUNT 26
#define FIELD_COUNT 9
#define MAX_ALIAS_SEQUENCE_LENGTH 7
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 28
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  aux_sym__terminator_token1 = 1,
  anon_sym_SEMI = 2,
  anon_sym_LPAREN = 3,
  anon_sym_RPAREN = 4,
  aux_sym_identifier_token1 = 5,
  anon_sym_DOT_DOT_DOT = 6,
  sym_alias = 7,
  sym_integer = 8,
  sym_float = 9,
  sym_char = 10,
  anon_sym_true = 11,
  anon_sym_false = 12,
  anon_sym_nil = 13,
  sym_atom = 14,
  anon_sym_DQUOTE = 15,
  anon_sym_SQUOTE = 16,
  anon_sym_SQUOTE_SQUOTE_SQUOTE = 17,
  anon_sym_DQUOTE_DQUOTE_DQUOTE = 18,
  anon_sym_LBRACE = 19,
  anon_sym_RBRACE = 20,
  anon_sym_LBRACK = 21,
  anon_sym_RBRACK = 22,
  anon_sym_LT = 23,
  anon_sym_GT = 24,
  anon_sym_PIPE = 25,
  anon_sym_SLASH = 26,
  anon_sym_POUND_LBRACE = 27,
  sym_escape_sequence = 28,
  anon_sym_TILDE = 29,
  aux_sym_sigil_token1 = 30,
  aux_sym_sigil_token2 = 31,
  aux_sym_sigil_token3 = 32,
  anon_sym_COMMA = 33,
  sym_keyword = 34,
  aux_sym_quoted_keyword_token1 = 35,
  anon_sym_LT_LT = 36,
  anon_sym_GT_GT = 37,
  anon_sym_PERCENT = 38,
  anon_sym_DOT_DOT = 39,
  anon_sym_AMP = 40,
  anon_sym_PLUS = 41,
  anon_sym_DASH = 42,
  anon_sym_BANG = 43,
  anon_sym_CARET = 44,
  anon_sym_TILDE_TILDE_TILDE = 45,
  anon_sym_not = 46,
  anon_sym_AT = 47,
  anon_sym_LT_DASH = 48,
  anon_sym_BSLASH_BSLASH = 49,
  anon_sym_when = 50,
  anon_sym_COLON_COLON = 51,
  anon_sym_EQ_GT = 52,
  anon_sym_EQ = 53,
  anon_sym_PIPE_PIPE = 54,
  anon_sym_PIPE_PIPE_PIPE = 55,
  anon_sym_or = 56,
  anon_sym_AMP_AMP = 57,
  anon_sym_AMP_AMP_AMP = 58,
  anon_sym_and = 59,
  anon_sym_EQ_EQ = 60,
  anon_sym_BANG_EQ = 61,
  anon_sym_EQ_TILDE = 62,
  anon_sym_EQ_EQ_EQ = 63,
  anon_sym_BANG_EQ_EQ = 64,
  anon_sym_LT_EQ = 65,
  anon_sym_GT_EQ = 66,
  anon_sym_PIPE_GT = 67,
  anon_sym_LT_LT_LT = 68,
  anon_sym_GT_GT_GT = 69,
  anon_sym_LT_LT_TILDE = 70,
  anon_sym_TILDE_GT_GT = 71,
  anon_sym_LT_TILDE = 72,
  anon_sym_TILDE_GT = 73,
  anon_sym_LT_TILDE_GT = 74,
  anon_sym_LT_PIPE_GT = 75,
  anon_sym_in = 76,
  anon_sym_CARET_CARET_CARET = 77,
  anon_sym_SLASH_SLASH = 78,
  anon_sym_PLUS_PLUS = 79,
  anon_sym_DASH_DASH = 80,
  anon_sym_PLUS_PLUS_PLUS = 81,
  anon_sym_DASH_DASH_DASH = 82,
  anon_sym_LT_GT = 83,
  anon_sym_STAR = 84,
  anon_sym_STAR_STAR = 85,
  anon_sym_DASH_GT = 86,
  anon_sym_DOT = 87,
  anon_sym_after = 88,
  anon_sym_catch = 89,
  anon_sym_do = 90,
  anon_sym_else = 91,
  anon_sym_end = 92,
  anon_sym_fn = 93,
  anon_sym_rescue = 94,
  anon_sym_LPAREN2 = 95,
  anon_sym_LBRACK2 = 96,
  sym_comment = 97,
  sym__quoted_content_i_single = 98,
  sym__quoted_content_i_double = 99,
  sym__quoted_content_i_heredoc_single = 100,
  sym__quoted_content_i_heredoc_double = 101,
  sym__quoted_content_i_parenthesis = 102,
  sym__quoted_content_i_curly = 103,
  sym__quoted_content_i_square = 104,
  sym__quoted_content_i_angle = 105,
  sym__quoted_content_i_bar = 106,
  sym__quoted_content_i_slash = 107,
  sym__quoted_content_single = 108,
  sym__quoted_content_double = 109,
  sym__quoted_content_heredoc_single = 110,
  sym__quoted_content_heredoc_double = 111,
  sym__quoted_content_parenthesis = 112,
  sym__quoted_content_curly = 113,
  sym__quoted_content_square = 114,
  sym__quoted_content_angle = 115,
  sym__quoted_content_bar = 116,
  sym__quoted_content_slash = 117,
  sym__newline_before_do = 118,
  sym__newline_before_binary_operator = 119,
  sym__newline_before_comment = 120,
  sym__before_unary_op = 121,
  sym__not_in = 122,
  sym__quoted_atom_start = 123,
  sym_source = 124,
  sym__terminator = 125,
  sym__expression = 126,
  sym_block = 127,
  sym_identifier = 128,
  sym_boolean = 129,
  sym_nil = 130,
  sym__atom = 131,
  sym_quoted_atom = 132,
  sym__quoted_i_double = 133,
  sym__quoted_double = 134,
  sym__quoted_i_single = 135,
  sym__quoted_single = 136,
  sym__quoted_i_heredoc_single = 137,
  sym__quoted_heredoc_single = 138,
  sym__quoted_i_heredoc_double = 139,
  sym__quoted_heredoc_double = 140,
  sym__quoted_i_parenthesis = 141,
  sym__quoted_parenthesis = 142,
  sym__quoted_i_curly = 143,
  sym__quoted_curly = 144,
  sym__quoted_i_square = 145,
  sym__quoted_square = 146,
  sym__quoted_i_angle = 147,
  sym__quoted_angle = 148,
  sym__quoted_i_bar = 149,
  sym__quoted_bar = 150,
  sym__quoted_i_slash = 151,
  sym__quoted_slash = 152,
  sym_string = 153,
  sym_charlist = 154,
  sym_interpolation = 155,
  sym_sigil = 156,
  sym_keywords = 157,
  sym__keywords_with_trailing_separator = 158,
  sym_pair = 159,
  sym__keyword = 160,
  sym_quoted_keyword = 161,
  sym_list = 162,
  sym_tuple = 163,
  sym_bitstring = 164,
  sym_map = 165,
  sym_struct = 166,
  sym__items_with_trailing_separator = 167,
  sym__nullary_operator = 168,
  sym_unary_operator = 169,
  sym__capture_expression = 170,
  sym_binary_operator = 171,
  sym_operator_identifier = 172,
  sym_dot = 173,
  sym_call = 174,
  sym__call_without_parentheses = 175,
  sym__call_with_parentheses = 176,
  sym__local_call_without_parentheses = 177,
  sym__local_call_with_parentheses = 178,
  sym__local_call_just_do_block = 179,
  sym__remote_call_without_parentheses = 180,
  sym__remote_call_with_parentheses = 181,
  sym__remote_dot = 182,
  sym__anonymous_call = 183,
  sym__anonymous_dot = 184,
  sym__double_call = 185,
  sym__call_arguments_with_parentheses = 186,
  sym__call_arguments_with_parentheses_immediate = 187,
  sym__call_arguments_with_trailing_separator = 188,
  sym__call_arguments_without_parentheses = 189,
  sym_do_block = 190,
  sym_after_block = 191,
  sym_rescue_block = 192,
  sym_catch_block = 193,
  sym_else_block = 194,
  sym_access_call = 195,
  sym_stab_clause = 196,
  sym__stab_clause_left = 197,
  sym__stab_clause_arguments_with_parentheses = 198,
  sym__stab_clause_arguments_without_parentheses = 199,
  sym__stab_clause_arguments_with_parentheses_with_guard = 200,
  sym__stab_clause_arguments_without_parentheses_with_guard = 201,
  sym_body = 202,
  sym_anonymous_function = 203,
  aux_sym_source_repeat1 = 204,
  aux_sym__terminator_repeat1 = 205,
  aux_sym_block_repeat1 = 206,
  aux_sym_block_repeat2 = 207,
  aux_sym__quoted_i_double_repeat1 = 208,
  aux_sym__quoted_double_repeat1 = 209,
  aux_sym__quoted_i_single_repeat1 = 210,
  aux_sym__quoted_single_repeat1 = 211,
  aux_sym__quoted_i_heredoc_single_repeat1 = 212,
  aux_sym__quoted_heredoc_single_repeat1 = 213,
  aux_sym__quoted_i_heredoc_double_repeat1 = 214,
  aux_sym__quoted_heredoc_double_repeat1 = 215,
  aux_sym__quoted_i_parenthesis_repeat1 = 216,
  aux_sym__quoted_parenthesis_repeat1 = 217,
  aux_sym__quoted_i_curly_repeat1 = 218,
  aux_sym__quoted_curly_repeat1 = 219,
  aux_sym__quoted_i_square_repeat1 = 220,
  aux_sym__quoted_square_repeat1 = 221,
  aux_sym__quoted_i_angle_repeat1 = 222,
  aux_sym__quoted_angle_repeat1 = 223,
  aux_sym__quoted_i_bar_repeat1 = 224,
  aux_sym__quoted_bar_repeat1 = 225,
  aux_sym__quoted_i_slash_repeat1 = 226,
  aux_sym__quoted_slash_repeat1 = 227,
  aux_sym_keywords_repeat1 = 228,
  aux_sym__items_with_trailing_separator_repeat1 = 229,
  aux_sym_do_block_repeat1 = 230,
  aux_sym__stab_clause_arguments_with_parentheses_repeat1 = 231,
  aux_sym__stab_clause_arguments_without_parentheses_repeat1 = 232,
  aux_sym_anonymous_function_repeat1 = 233,
  alias_sym_map_content = 234,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [aux_sym__terminator_token1] = "_terminator_token1",
  [anon_sym_SEMI] = ";",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [aux_sym_identifier_token1] = "identifier_token1",
  [anon_sym_DOT_DOT_DOT] = "...",
  [sym_alias] = "alias",
  [sym_integer] = "integer",
  [sym_float] = "float",
  [sym_char] = "char",
  [anon_sym_true] = "true",
  [anon_sym_false] = "false",
  [anon_sym_nil] = "nil",
  [sym_atom] = "atom",
  [anon_sym_DQUOTE] = "\"",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_SQUOTE_SQUOTE_SQUOTE] = "'''",
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = "\"\"\"",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_LT] = "<",
  [anon_sym_GT] = ">",
  [anon_sym_PIPE] = "|",
  [anon_sym_SLASH] = "/",
  [anon_sym_POUND_LBRACE] = "#{",
  [sym_escape_sequence] = "escape_sequence",
  [anon_sym_TILDE] = "~",
  [aux_sym_sigil_token1] = "sigil_name",
  [aux_sym_sigil_token2] = "sigil_name",
  [aux_sym_sigil_token3] = "sigil_modifiers",
  [anon_sym_COMMA] = ",",
  [sym_keyword] = "keyword",
  [aux_sym_quoted_keyword_token1] = "quoted_keyword_token1",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_PERCENT] = "%",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_AMP] = "&",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_BANG] = "!",
  [anon_sym_CARET] = "^",
  [anon_sym_TILDE_TILDE_TILDE] = "~~~",
  [anon_sym_not] = "not",
  [anon_sym_AT] = "@",
  [anon_sym_LT_DASH] = "<-",
  [anon_sym_BSLASH_BSLASH] = "\\\\",
  [anon_sym_when] = "when",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_EQ_GT] = "=>",
  [anon_sym_EQ] = "=",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_PIPE_PIPE_PIPE] = "|||",
  [anon_sym_or] = "or",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_AMP_AMP_AMP] = "&&&",
  [anon_sym_and] = "and",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_EQ_TILDE] = "=~",
  [anon_sym_EQ_EQ_EQ] = "===",
  [anon_sym_BANG_EQ_EQ] = "!==",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_PIPE_GT] = "|>",
  [anon_sym_LT_LT_LT] = "<<<",
  [anon_sym_GT_GT_GT] = ">>>",
  [anon_sym_LT_LT_TILDE] = "<<~",
  [anon_sym_TILDE_GT_GT] = "~>>",
  [anon_sym_LT_TILDE] = "<~",
  [anon_sym_TILDE_GT] = "~>",
  [anon_sym_LT_TILDE_GT] = "<~>",
  [anon_sym_LT_PIPE_GT] = "<|>",
  [anon_sym_in] = "in",
  [anon_sym_CARET_CARET_CARET] = "^^^",
  [anon_sym_SLASH_SLASH] = "//",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_PLUS_PLUS_PLUS] = "+++",
  [anon_sym_DASH_DASH_DASH] = "---",
  [anon_sym_LT_GT] = "<>",
  [anon_sym_STAR] = "*",
  [anon_sym_STAR_STAR] = "**",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_DOT] = ".",
  [anon_sym_after] = "after",
  [anon_sym_catch] = "catch",
  [anon_sym_do] = "do",
  [anon_sym_else] = "else",
  [anon_sym_end] = "end",
  [anon_sym_fn] = "fn",
  [anon_sym_rescue] = "rescue",
  [anon_sym_LPAREN2] = "(",
  [anon_sym_LBRACK2] = "[",
  [sym_comment] = "comment",
  [sym__quoted_content_i_single] = "quoted_content",
  [sym__quoted_content_i_double] = "quoted_content",
  [sym__quoted_content_i_heredoc_single] = "quoted_content",
  [sym__quoted_content_i_heredoc_double] = "quoted_content",
  [sym__quoted_content_i_parenthesis] = "quoted_content",
  [sym__quoted_content_i_curly] = "quoted_content",
  [sym__quoted_content_i_square] = "quoted_content",
  [sym__quoted_content_i_angle] = "quoted_content",
  [sym__quoted_content_i_bar] = "quoted_content",
  [sym__quoted_content_i_slash] = "quoted_content",
  [sym__quoted_content_single] = "quoted_content",
  [sym__quoted_content_double] = "quoted_content",
  [sym__quoted_content_heredoc_single] = "quoted_content",
  [sym__quoted_content_heredoc_double] = "quoted_content",
  [sym__quoted_content_parenthesis] = "quoted_content",
  [sym__quoted_content_curly] = "quoted_content",
  [sym__quoted_content_square] = "quoted_content",
  [sym__quoted_content_angle] = "quoted_content",
  [sym__quoted_content_bar] = "quoted_content",
  [sym__quoted_content_slash] = "quoted_content",
  [sym__newline_before_do] = "_newline_before_do",
  [sym__newline_before_binary_operator] = "_newline_before_binary_operator",
  [sym__newline_before_comment] = "_newline_before_comment",
  [sym__before_unary_op] = "_before_unary_op",
  [sym__not_in] = "not in",
  [sym__quoted_atom_start] = ":",
  [sym_source] = "source",
  [sym__terminator] = "_terminator",
  [sym__expression] = "_expression",
  [sym_block] = "block",
  [sym_identifier] = "identifier",
  [sym_boolean] = "boolean",
  [sym_nil] = "nil",
  [sym__atom] = "_atom",
  [sym_quoted_atom] = "quoted_atom",
  [sym__quoted_i_double] = "_quoted_i_double",
  [sym__quoted_double] = "_quoted_double",
  [sym__quoted_i_single] = "_quoted_i_single",
  [sym__quoted_single] = "_quoted_single",
  [sym__quoted_i_heredoc_single] = "_quoted_i_heredoc_single",
  [sym__quoted_heredoc_single] = "_quoted_heredoc_single",
  [sym__quoted_i_heredoc_double] = "_quoted_i_heredoc_double",
  [sym__quoted_heredoc_double] = "_quoted_heredoc_double",
  [sym__quoted_i_parenthesis] = "_quoted_i_parenthesis",
  [sym__quoted_parenthesis] = "_quoted_parenthesis",
  [sym__quoted_i_curly] = "_quoted_i_curly",
  [sym__quoted_curly] = "_quoted_curly",
  [sym__quoted_i_square] = "_quoted_i_square",
  [sym__quoted_square] = "_quoted_square",
  [sym__quoted_i_angle] = "_quoted_i_angle",
  [sym__quoted_angle] = "_quoted_angle",
  [sym__quoted_i_bar] = "_quoted_i_bar",
  [sym__quoted_bar] = "_quoted_bar",
  [sym__quoted_i_slash] = "_quoted_i_slash",
  [sym__quoted_slash] = "_quoted_slash",
  [sym_string] = "string",
  [sym_charlist] = "charlist",
  [sym_interpolation] = "interpolation",
  [sym_sigil] = "sigil",
  [sym_keywords] = "keywords",
  [sym__keywords_with_trailing_separator] = "keywords",
  [sym_pair] = "pair",
  [sym__keyword] = "_keyword",
  [sym_quoted_keyword] = "quoted_keyword",
  [sym_list] = "list",
  [sym_tuple] = "tuple",
  [sym_bitstring] = "bitstring",
  [sym_map] = "map",
  [sym_struct] = "struct",
  [sym__items_with_trailing_separator] = "_items_with_trailing_separator",
  [sym__nullary_operator] = "_nullary_operator",
  [sym_unary_operator] = "unary_operator",
  [sym__capture_expression] = "_capture_expression",
  [sym_binary_operator] = "binary_operator",
  [sym_operator_identifier] = "operator_identifier",
  [sym_dot] = "dot",
  [sym_call] = "call",
  [sym__call_without_parentheses] = "_call_without_parentheses",
  [sym__call_with_parentheses] = "_call_with_parentheses",
  [sym__local_call_without_parentheses] = "_local_call_without_parentheses",
  [sym__local_call_with_parentheses] = "_local_call_with_parentheses",
  [sym__local_call_just_do_block] = "_local_call_just_do_block",
  [sym__remote_call_without_parentheses] = "_remote_call_without_parentheses",
  [sym__remote_call_with_parentheses] = "_remote_call_with_parentheses",
  [sym__remote_dot] = "dot",
  [sym__anonymous_call] = "_anonymous_call",
  [sym__anonymous_dot] = "dot",
  [sym__double_call] = "_double_call",
  [sym__call_arguments_with_parentheses] = "arguments",
  [sym__call_arguments_with_parentheses_immediate] = "arguments",
  [sym__call_arguments_with_trailing_separator] = "_call_arguments_with_trailing_separator",
  [sym__call_arguments_without_parentheses] = "arguments",
  [sym_do_block] = "do_block",
  [sym_after_block] = "after_block",
  [sym_rescue_block] = "rescue_block",
  [sym_catch_block] = "catch_block",
  [sym_else_block] = "else_block",
  [sym_access_call] = "access_call",
  [sym_stab_clause] = "stab_clause",
  [sym__stab_clause_left] = "_stab_clause_left",
  [sym__stab_clause_arguments_with_parentheses] = "arguments",
  [sym__stab_clause_arguments_without_parentheses] = "arguments",
  [sym__stab_clause_arguments_with_parentheses_with_guard] = "binary_operator",
  [sym__stab_clause_arguments_without_parentheses_with_guard] = "binary_operator",
  [sym_body] = "body",
  [sym_anonymous_function] = "anonymous_function",
  [aux_sym_source_repeat1] = "source_repeat1",
  [aux_sym__terminator_repeat1] = "_terminator_repeat1",
  [aux_sym_block_repeat1] = "block_repeat1",
  [aux_sym_block_repeat2] = "block_repeat2",
  [aux_sym__quoted_i_double_repeat1] = "_quoted_i_double_repeat1",
  [aux_sym__quoted_double_repeat1] = "_quoted_double_repeat1",
  [aux_sym__quoted_i_single_repeat1] = "_quoted_i_single_repeat1",
  [aux_sym__quoted_single_repeat1] = "_quoted_single_repeat1",
  [aux_sym__quoted_i_heredoc_single_repeat1] = "_quoted_i_heredoc_single_repeat1",
  [aux_sym__quoted_heredoc_single_repeat1] = "_quoted_heredoc_single_repeat1",
  [aux_sym__quoted_i_heredoc_double_repeat1] = "_quoted_i_heredoc_double_repeat1",
  [aux_sym__quoted_heredoc_double_repeat1] = "_quoted_heredoc_double_repeat1",
  [aux_sym__quoted_i_parenthesis_repeat1] = "_quoted_i_parenthesis_repeat1",
  [aux_sym__quoted_parenthesis_repeat1] = "_quoted_parenthesis_repeat1",
  [aux_sym__quoted_i_curly_repeat1] = "_quoted_i_curly_repeat1",
  [aux_sym__quoted_curly_repeat1] = "_quoted_curly_repeat1",
  [aux_sym__quoted_i_square_repeat1] = "_quoted_i_square_repeat1",
  [aux_sym__quoted_square_repeat1] = "_quoted_square_repeat1",
  [aux_sym__quoted_i_angle_repeat1] = "_quoted_i_angle_repeat1",
  [aux_sym__quoted_angle_repeat1] = "_quoted_angle_repeat1",
  [aux_sym__quoted_i_bar_repeat1] = "_quoted_i_bar_repeat1",
  [aux_sym__quoted_bar_repeat1] = "_quoted_bar_repeat1",
  [aux_sym__quoted_i_slash_repeat1] = "_quoted_i_slash_repeat1",
  [aux_sym__quoted_slash_repeat1] = "_quoted_slash_repeat1",
  [aux_sym_keywords_repeat1] = "keywords_repeat1",
  [aux_sym__items_with_trailing_separator_repeat1] = "_items_with_trailing_separator_repeat1",
  [aux_sym_do_block_repeat1] = "do_block_repeat1",
  [aux_sym__stab_clause_arguments_with_parentheses_repeat1] = "_stab_clause_arguments_with_parentheses_repeat1",
  [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = "_stab_clause_arguments_without_parentheses_repeat1",
  [aux_sym_anonymous_function_repeat1] = "anonymous_function_repeat1",
  [alias_sym_map_content] = "map_content",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [aux_sym__terminator_token1] = aux_sym__terminator_token1,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [aux_sym_identifier_token1] = aux_sym_identifier_token1,
  [anon_sym_DOT_DOT_DOT] = anon_sym_DOT_DOT_DOT,
  [sym_alias] = sym_alias,
  [sym_integer] = sym_integer,
  [sym_float] = sym_float,
  [sym_char] = sym_char,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_nil] = anon_sym_nil,
  [sym_atom] = sym_atom,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_SQUOTE_SQUOTE_SQUOTE] = anon_sym_SQUOTE_SQUOTE_SQUOTE,
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = anon_sym_DQUOTE_DQUOTE_DQUOTE,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_POUND_LBRACE] = anon_sym_POUND_LBRACE,
  [sym_escape_sequence] = sym_escape_sequence,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [aux_sym_sigil_token1] = aux_sym_sigil_token1,
  [aux_sym_sigil_token2] = aux_sym_sigil_token1,
  [aux_sym_sigil_token3] = aux_sym_sigil_token3,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [sym_keyword] = sym_keyword,
  [aux_sym_quoted_keyword_token1] = aux_sym_quoted_keyword_token1,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_TILDE_TILDE_TILDE] = anon_sym_TILDE_TILDE_TILDE,
  [anon_sym_not] = anon_sym_not,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_LT_DASH] = anon_sym_LT_DASH,
  [anon_sym_BSLASH_BSLASH] = anon_sym_BSLASH_BSLASH,
  [anon_sym_when] = anon_sym_when,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_EQ_GT] = anon_sym_EQ_GT,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_PIPE_PIPE_PIPE] = anon_sym_PIPE_PIPE_PIPE,
  [anon_sym_or] = anon_sym_or,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_AMP_AMP_AMP] = anon_sym_AMP_AMP_AMP,
  [anon_sym_and] = anon_sym_and,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_EQ_TILDE] = anon_sym_EQ_TILDE,
  [anon_sym_EQ_EQ_EQ] = anon_sym_EQ_EQ_EQ,
  [anon_sym_BANG_EQ_EQ] = anon_sym_BANG_EQ_EQ,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_PIPE_GT] = anon_sym_PIPE_GT,
  [anon_sym_LT_LT_LT] = anon_sym_LT_LT_LT,
  [anon_sym_GT_GT_GT] = anon_sym_GT_GT_GT,
  [anon_sym_LT_LT_TILDE] = anon_sym_LT_LT_TILDE,
  [anon_sym_TILDE_GT_GT] = anon_sym_TILDE_GT_GT,
  [anon_sym_LT_TILDE] = anon_sym_LT_TILDE,
  [anon_sym_TILDE_GT] = anon_sym_TILDE_GT,
  [anon_sym_LT_TILDE_GT] = anon_sym_LT_TILDE_GT,
  [anon_sym_LT_PIPE_GT] = anon_sym_LT_PIPE_GT,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_CARET_CARET_CARET] = anon_sym_CARET_CARET_CARET,
  [anon_sym_SLASH_SLASH] = anon_sym_SLASH_SLASH,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_PLUS_PLUS_PLUS] = anon_sym_PLUS_PLUS_PLUS,
  [anon_sym_DASH_DASH_DASH] = anon_sym_DASH_DASH_DASH,
  [anon_sym_LT_GT] = anon_sym_LT_GT,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_STAR_STAR] = anon_sym_STAR_STAR,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_after] = anon_sym_after,
  [anon_sym_catch] = anon_sym_catch,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_end] = anon_sym_end,
  [anon_sym_fn] = anon_sym_fn,
  [anon_sym_rescue] = anon_sym_rescue,
  [anon_sym_LPAREN2] = anon_sym_LPAREN,
  [anon_sym_LBRACK2] = anon_sym_LBRACK,
  [sym_comment] = sym_comment,
  [sym__quoted_content_i_single] = sym__quoted_content_i_single,
  [sym__quoted_content_i_double] = sym__quoted_content_i_single,
  [sym__quoted_content_i_heredoc_single] = sym__quoted_content_i_single,
  [sym__quoted_content_i_heredoc_double] = sym__quoted_content_i_single,
  [sym__quoted_content_i_parenthesis] = sym__quoted_content_i_single,
  [sym__quoted_content_i_curly] = sym__quoted_content_i_single,
  [sym__quoted_content_i_square] = sym__quoted_content_i_single,
  [sym__quoted_content_i_angle] = sym__quoted_content_i_single,
  [sym__quoted_content_i_bar] = sym__quoted_content_i_single,
  [sym__quoted_content_i_slash] = sym__quoted_content_i_single,
  [sym__quoted_content_single] = sym__quoted_content_i_single,
  [sym__quoted_content_double] = sym__quoted_content_i_single,
  [sym__quoted_content_heredoc_single] = sym__quoted_content_i_single,
  [sym__quoted_content_heredoc_double] = sym__quoted_content_i_single,
  [sym__quoted_content_parenthesis] = sym__quoted_content_i_single,
  [sym__quoted_content_curly] = sym__quoted_content_i_single,
  [sym__quoted_content_square] = sym__quoted_content_i_single,
  [sym__quoted_content_angle] = sym__quoted_content_i_single,
  [sym__quoted_content_bar] = sym__quoted_content_i_single,
  [sym__quoted_content_slash] = sym__quoted_content_i_single,
  [sym__newline_before_do] = sym__newline_before_do,
  [sym__newline_before_binary_operator] = sym__newline_before_binary_operator,
  [sym__newline_before_comment] = sym__newline_before_comment,
  [sym__before_unary_op] = sym__before_unary_op,
  [sym__not_in] = sym__not_in,
  [sym__quoted_atom_start] = sym__quoted_atom_start,
  [sym_source] = sym_source,
  [sym__terminator] = sym__terminator,
  [sym__expression] = sym__expression,
  [sym_block] = sym_block,
  [sym_identifier] = sym_identifier,
  [sym_boolean] = sym_boolean,
  [sym_nil] = sym_nil,
  [sym__atom] = sym__atom,
  [sym_quoted_atom] = sym_quoted_atom,
  [sym__quoted_i_double] = sym__quoted_i_double,
  [sym__quoted_double] = sym__quoted_double,
  [sym__quoted_i_single] = sym__quoted_i_single,
  [sym__quoted_single] = sym__quoted_single,
  [sym__quoted_i_heredoc_single] = sym__quoted_i_heredoc_single,
  [sym__quoted_heredoc_single] = sym__quoted_heredoc_single,
  [sym__quoted_i_heredoc_double] = sym__quoted_i_heredoc_double,
  [sym__quoted_heredoc_double] = sym__quoted_heredoc_double,
  [sym__quoted_i_parenthesis] = sym__quoted_i_parenthesis,
  [sym__quoted_parenthesis] = sym__quoted_parenthesis,
  [sym__quoted_i_curly] = sym__quoted_i_curly,
  [sym__quoted_curly] = sym__quoted_curly,
  [sym__quoted_i_square] = sym__quoted_i_square,
  [sym__quoted_square] = sym__quoted_square,
  [sym__quoted_i_angle] = sym__quoted_i_angle,
  [sym__quoted_angle] = sym__quoted_angle,
  [sym__quoted_i_bar] = sym__quoted_i_bar,
  [sym__quoted_bar] = sym__quoted_bar,
  [sym__quoted_i_slash] = sym__quoted_i_slash,
  [sym__quoted_slash] = sym__quoted_slash,
  [sym_string] = sym_string,
  [sym_charlist] = sym_charlist,
  [sym_interpolation] = sym_interpolation,
  [sym_sigil] = sym_sigil,
  [sym_keywords] = sym_keywords,
  [sym__keywords_with_trailing_separator] = sym_keywords,
  [sym_pair] = sym_pair,
  [sym__keyword] = sym__keyword,
  [sym_quoted_keyword] = sym_quoted_keyword,
  [sym_list] = sym_list,
  [sym_tuple] = sym_tuple,
  [sym_bitstring] = sym_bitstring,
  [sym_map] = sym_map,
  [sym_struct] = sym_struct,
  [sym__items_with_trailing_separator] = sym__items_with_trailing_separator,
  [sym__nullary_operator] = sym__nullary_operator,
  [sym_unary_operator] = sym_unary_operator,
  [sym__capture_expression] = sym__capture_expression,
  [sym_binary_operator] = sym_binary_operator,
  [sym_operator_identifier] = sym_operator_identifier,
  [sym_dot] = sym_dot,
  [sym_call] = sym_call,
  [sym__call_without_parentheses] = sym__call_without_parentheses,
  [sym__call_with_parentheses] = sym__call_with_parentheses,
  [sym__local_call_without_parentheses] = sym__local_call_without_parentheses,
  [sym__local_call_with_parentheses] = sym__local_call_with_parentheses,
  [sym__local_call_just_do_block] = sym__local_call_just_do_block,
  [sym__remote_call_without_parentheses] = sym__remote_call_without_parentheses,
  [sym__remote_call_with_parentheses] = sym__remote_call_with_parentheses,
  [sym__remote_dot] = sym_dot,
  [sym__anonymous_call] = sym__anonymous_call,
  [sym__anonymous_dot] = sym_dot,
  [sym__double_call] = sym__double_call,
  [sym__call_arguments_with_parentheses] = sym__call_arguments_with_parentheses,
  [sym__call_arguments_with_parentheses_immediate] = sym__call_arguments_with_parentheses,
  [sym__call_arguments_with_trailing_separator] = sym__call_arguments_with_trailing_separator,
  [sym__call_arguments_without_parentheses] = sym__call_arguments_with_parentheses,
  [sym_do_block] = sym_do_block,
  [sym_after_block] = sym_after_block,
  [sym_rescue_block] = sym_rescue_block,
  [sym_catch_block] = sym_catch_block,
  [sym_else_block] = sym_else_block,
  [sym_access_call] = sym_access_call,
  [sym_stab_clause] = sym_stab_clause,
  [sym__stab_clause_left] = sym__stab_clause_left,
  [sym__stab_clause_arguments_with_parentheses] = sym__call_arguments_with_parentheses,
  [sym__stab_clause_arguments_without_parentheses] = sym__call_arguments_with_parentheses,
  [sym__stab_clause_arguments_with_parentheses_with_guard] = sym_binary_operator,
  [sym__stab_clause_arguments_without_parentheses_with_guard] = sym_binary_operator,
  [sym_body] = sym_body,
  [sym_anonymous_function] = sym_anonymous_function,
  [aux_sym_source_repeat1] = aux_sym_source_repeat1,
  [aux_sym__terminator_repeat1] = aux_sym__terminator_repeat1,
  [aux_sym_block_repeat1] = aux_sym_block_repeat1,
  [aux_sym_block_repeat2] = aux_sym_block_repeat2,
  [aux_sym__quoted_i_double_repeat1] = aux_sym__quoted_i_double_repeat1,
  [aux_sym__quoted_double_repeat1] = aux_sym__quoted_double_repeat1,
  [aux_sym__quoted_i_single_repeat1] = aux_sym__quoted_i_single_repeat1,
  [aux_sym__quoted_single_repeat1] = aux_sym__quoted_single_repeat1,
  [aux_sym__quoted_i_heredoc_single_repeat1] = aux_sym__quoted_i_heredoc_single_repeat1,
  [aux_sym__quoted_heredoc_single_repeat1] = aux_sym__quoted_heredoc_single_repeat1,
  [aux_sym__quoted_i_heredoc_double_repeat1] = aux_sym__quoted_i_heredoc_double_repeat1,
  [aux_sym__quoted_heredoc_double_repeat1] = aux_sym__quoted_heredoc_double_repeat1,
  [aux_sym__quoted_i_parenthesis_repeat1] = aux_sym__quoted_i_parenthesis_repeat1,
  [aux_sym__quoted_parenthesis_repeat1] = aux_sym__quoted_parenthesis_repeat1,
  [aux_sym__quoted_i_curly_repeat1] = aux_sym__quoted_i_curly_repeat1,
  [aux_sym__quoted_curly_repeat1] = aux_sym__quoted_curly_repeat1,
  [aux_sym__quoted_i_square_repeat1] = aux_sym__quoted_i_square_repeat1,
  [aux_sym__quoted_square_repeat1] = aux_sym__quoted_square_repeat1,
  [aux_sym__quoted_i_angle_repeat1] = aux_sym__quoted_i_angle_repeat1,
  [aux_sym__quoted_angle_repeat1] = aux_sym__quoted_angle_repeat1,
  [aux_sym__quoted_i_bar_repeat1] = aux_sym__quoted_i_bar_repeat1,
  [aux_sym__quoted_bar_repeat1] = aux_sym__quoted_bar_repeat1,
  [aux_sym__quoted_i_slash_repeat1] = aux_sym__quoted_i_slash_repeat1,
  [aux_sym__quoted_slash_repeat1] = aux_sym__quoted_slash_repeat1,
  [aux_sym_keywords_repeat1] = aux_sym_keywords_repeat1,
  [aux_sym__items_with_trailing_separator_repeat1] = aux_sym__items_with_trailing_separator_repeat1,
  [aux_sym_do_block_repeat1] = aux_sym_do_block_repeat1,
  [aux_sym__stab_clause_arguments_with_parentheses_repeat1] = aux_sym__stab_clause_arguments_with_parentheses_repeat1,
  [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = aux_sym__stab_clause_arguments_without_parentheses_repeat1,
  [aux_sym_anonymous_function_repeat1] = aux_sym_anonymous_function_repeat1,
  [alias_sym_map_content] = alias_sym_map_content,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__terminator_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_identifier_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DOT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [sym_alias] = {
    .visible = true,
    .named = true,
  },
  [sym_integer] = {
    .visible = true,
    .named = true,
  },
  [sym_float] = {
    .visible = true,
    .named = true,
  },
  [sym_char] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nil] = {
    .visible = true,
    .named = false,
  },
  [sym_atom] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE_SQUOTE_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_sigil_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_sigil_token2] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_sigil_token3] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [sym_keyword] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_quoted_keyword_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE_TILDE_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_not] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_when] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_or] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_and] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_TILDE_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_PIPE_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_CARET_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_after] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_catch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_end] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fn] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rescue] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK2] = {
    .visible = true,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_single] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_double] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_heredoc_single] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_heredoc_double] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_parenthesis] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_curly] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_square] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_angle] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_bar] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_i_slash] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_single] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_double] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_heredoc_single] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_heredoc_double] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_parenthesis] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_curly] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_square] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_angle] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_bar] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_content_slash] = {
    .visible = true,
    .named = true,
  },
  [sym__newline_before_do] = {
    .visible = false,
    .named = true,
  },
  [sym__newline_before_binary_operator] = {
    .visible = false,
    .named = true,
  },
  [sym__newline_before_comment] = {
    .visible = false,
    .named = true,
  },
  [sym__before_unary_op] = {
    .visible = false,
    .named = true,
  },
  [sym__not_in] = {
    .visible = true,
    .named = false,
  },
  [sym__quoted_atom_start] = {
    .visible = true,
    .named = false,
  },
  [sym_source] = {
    .visible = true,
    .named = true,
  },
  [sym__terminator] = {
    .visible = false,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
  },
  [sym_block] = {
    .visible = true,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym_nil] = {
    .visible = true,
    .named = true,
  },
  [sym__atom] = {
    .visible = false,
    .named = true,
  },
  [sym_quoted_atom] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_i_double] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_double] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_i_single] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_single] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_i_heredoc_single] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_heredoc_single] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_i_heredoc_double] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_heredoc_double] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_i_parenthesis] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_parenthesis] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_i_curly] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_curly] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_i_square] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_square] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_i_angle] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_angle] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_i_bar] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_bar] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_i_slash] = {
    .visible = false,
    .named = true,
  },
  [sym__quoted_slash] = {
    .visible = false,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_charlist] = {
    .visible = true,
    .named = true,
  },
  [sym_interpolation] = {
    .visible = true,
    .named = true,
  },
  [sym_sigil] = {
    .visible = true,
    .named = true,
  },
  [sym_keywords] = {
    .visible = true,
    .named = true,
  },
  [sym__keywords_with_trailing_separator] = {
    .visible = true,
    .named = true,
  },
  [sym_pair] = {
    .visible = true,
    .named = true,
  },
  [sym__keyword] = {
    .visible = false,
    .named = true,
  },
  [sym_quoted_keyword] = {
    .visible = true,
    .named = true,
  },
  [sym_list] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple] = {
    .visible = true,
    .named = true,
  },
  [sym_bitstring] = {
    .visible = true,
    .named = true,
  },
  [sym_map] = {
    .visible = true,
    .named = true,
  },
  [sym_struct] = {
    .visible = true,
    .named = true,
  },
  [sym__items_with_trailing_separator] = {
    .visible = false,
    .named = true,
  },
  [sym__nullary_operator] = {
    .visible = false,
    .named = true,
  },
  [sym_unary_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__capture_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_binary_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_operator_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_dot] = {
    .visible = true,
    .named = true,
  },
  [sym_call] = {
    .visible = true,
    .named = true,
  },
  [sym__call_without_parentheses] = {
    .visible = false,
    .named = true,
  },
  [sym__call_with_parentheses] = {
    .visible = false,
    .named = true,
  },
  [sym__local_call_without_parentheses] = {
    .visible = false,
    .named = true,
  },
  [sym__local_call_with_parentheses] = {
    .visible = false,
    .named = true,
  },
  [sym__local_call_just_do_block] = {
    .visible = false,
    .named = true,
  },
  [sym__remote_call_without_parentheses] = {
    .visible = false,
    .named = true,
  },
  [sym__remote_call_with_parentheses] = {
    .visible = false,
    .named = true,
  },
  [sym__remote_dot] = {
    .visible = true,
    .named = true,
  },
  [sym__anonymous_call] = {
    .visible = false,
    .named = true,
  },
  [sym__anonymous_dot] = {
    .visible = true,
    .named = true,
  },
  [sym__double_call] = {
    .visible = false,
    .named = true,
  },
  [sym__call_arguments_with_parentheses] = {
    .visible = true,
    .named = true,
  },
  [sym__call_arguments_with_parentheses_immediate] = {
    .visible = true,
    .named = true,
  },
  [sym__call_arguments_with_trailing_separator] = {
    .visible = false,
    .named = true,
  },
  [sym__call_arguments_without_parentheses] = {
    .visible = true,
    .named = true,
  },
  [sym_do_block] = {
    .visible = true,
    .named = true,
  },
  [sym_after_block] = {
    .visible = true,
    .named = true,
  },
  [sym_rescue_block] = {
    .visible = true,
    .named = true,
  },
  [sym_catch_block] = {
    .visible = true,
    .named = true,
  },
  [sym_else_block] = {
    .visible = true,
    .named = true,
  },
  [sym_access_call] = {
    .visible = true,
    .named = true,
  },
  [sym_stab_clause] = {
    .visible = true,
    .named = true,
  },
  [sym__stab_clause_left] = {
    .visible = false,
    .named = true,
  },
  [sym__stab_clause_arguments_with_parentheses] = {
    .visible = true,
    .named = true,
  },
  [sym__stab_clause_arguments_without_parentheses] = {
    .visible = true,
    .named = true,
  },
  [sym__stab_clause_arguments_with_parentheses_with_guard] = {
    .visible = true,
    .named = true,
  },
  [sym__stab_clause_arguments_without_parentheses_with_guard] = {
    .visible = true,
    .named = true,
  },
  [sym_body] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_function] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_source_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__terminator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_block_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_double_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_double_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_single_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_single_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_heredoc_single_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_heredoc_single_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_heredoc_double_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_heredoc_double_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_parenthesis_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_parenthesis_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_curly_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_curly_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_square_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_square_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_angle_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_angle_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_bar_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_bar_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_i_slash_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__quoted_slash_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_keywords_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__items_with_trailing_separator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_do_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__stab_clause_arguments_with_parentheses_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_anonymous_function_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_map_content] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_key = 1,
  field_left = 2,
  field_operand = 3,
  field_operator = 4,
  field_quoted_end = 5,
  field_quoted_start = 6,
  field_right = 7,
  field_target = 8,
  field_value = 9,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_key] = "key",
  [field_left] = "left",
  [field_operand] = "operand",
  [field_operator] = "operator",
  [field_quoted_end] = "quoted_end",
  [field_quoted_start] = "quoted_start",
  [field_right] = "right",
  [field_target] = "target",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 2},
  [3] = {.index = 2, .length = 1},
  [4] = {.index = 3, .length = 4},
  [5] = {.index = 7, .length = 2},
  [6] = {.index = 9, .length = 1},
  [7] = {.index = 10, .length = 3},
  [8] = {.index = 13, .length = 2},
  [9] = {.index = 2, .length = 1},
  [10] = {.index = 15, .length = 2},
  [11] = {.index = 17, .length = 2},
  [12] = {.index = 19, .length = 1},
  [13] = {.index = 20, .length = 2},
  [14] = {.index = 22, .length = 3},
  [15] = {.index = 25, .length = 2},
  [16] = {.index = 27, .length = 2},
  [17] = {.index = 29, .length = 2},
  [18] = {.index = 31, .length = 2},
  [19] = {.index = 33, .length = 2},
  [20] = {.index = 35, .length = 3},
  [21] = {.index = 35, .length = 3},
  [22] = {.index = 38, .length = 5},
  [23] = {.index = 38, .length = 5},
  [24] = {.index = 43, .length = 2},
  [26] = {.index = 45, .length = 2},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_quoted_end, 0, .inherited = true},
    {field_quoted_start, 0, .inherited = true},
  [2] =
    {field_target, 0, .inherited = true},
  [3] =
    {field_left, 0, .inherited = true},
    {field_operator, 0, .inherited = true},
    {field_right, 0, .inherited = true},
    {field_target, 0},
  [7] =
    {field_quoted_end, 1, .inherited = true},
    {field_quoted_start, 1, .inherited = true},
  [9] =
    {field_operator, 0},
  [10] =
    {field_left, 0, .inherited = true},
    {field_operator, 0, .inherited = true},
    {field_right, 0, .inherited = true},
  [13] =
    {field_quoted_end, 1},
    {field_quoted_start, 0},
  [15] =
    {field_operand, 1},
    {field_operator, 0},
  [17] =
    {field_left, 0},
    {field_operator, 1},
  [19] =
    {field_target, 0},
  [20] =
    {field_target, 0},
    {field_target, 0, .inherited = true},
  [22] =
    {field_left, 0, .inherited = true},
    {field_operator, 0, .inherited = true},
    {field_target, 0},
  [25] =
    {field_operand, 2},
    {field_operator, 1},
  [27] =
    {field_operator, 0},
    {field_right, 1},
  [29] =
    {field_key, 0},
    {field_value, 1},
  [31] =
    {field_quoted_end, 2},
    {field_quoted_start, 0},
  [33] =
    {field_quoted_end, 2, .inherited = true},
    {field_quoted_start, 2, .inherited = true},
  [35] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [38] =
    {field_left, 0},
    {field_operator, 1},
    {field_quoted_end, 2, .inherited = true},
    {field_quoted_start, 2, .inherited = true},
    {field_right, 2},
  [43] =
    {field_quoted_end, 3},
    {field_quoted_start, 0},
  [45] =
    {field_key, 2},
    {field_target, 0},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = sym_operator_identifier,
  },
  [9] = {
    [0] = sym_call,
  },
  [13] = {
    [0] = sym_call,
  },
  [21] = {
    [2] = sym_identifier,
  },
  [22] = {
    [2] = sym_string,
  },
  [23] = {
    [2] = sym_charlist,
  },
  [25] = {
    [2] = alias_sym_map_content,
  },
  [27] = {
    [3] = alias_sym_map_content,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__quoted_i_double, 2,
    sym__quoted_i_double,
    sym_string,
  sym__quoted_i_single, 2,
    sym__quoted_i_single,
    sym_charlist,
  sym__items_with_trailing_separator, 2,
    sym__items_with_trailing_separator,
    alias_sym_map_content,
  sym__call_with_parentheses, 2,
    sym__call_with_parentheses,
    sym_call,
  sym__local_call_with_parentheses, 2,
    sym__local_call_with_parentheses,
    sym_call,
  sym__remote_call_with_parentheses, 2,
    sym__remote_call_with_parentheses,
    sym_call,
  sym__anonymous_call, 2,
    sym__anonymous_call,
    sym_call,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 2,
  [4] = 2,
  [5] = 2,
  [6] = 2,
  [7] = 2,
  [8] = 2,
  [9] = 2,
  [10] = 2,
  [11] = 2,
  [12] = 2,
  [13] = 2,
  [14] = 14,
  [15] = 15,
  [16] = 14,
  [17] = 14,
  [18] = 14,
  [19] = 15,
  [20] = 20,
  [21] = 15,
  [22] = 20,
  [23] = 15,
  [24] = 20,
  [25] = 20,
  [26] = 20,
  [27] = 20,
  [28] = 20,
  [29] = 20,
  [30] = 20,
  [31] = 20,
  [32] = 20,
  [33] = 20,
  [34] = 34,
  [35] = 14,
  [36] = 15,
  [37] = 37,
  [38] = 14,
  [39] = 39,
  [40] = 40,
  [41] = 14,
  [42] = 14,
  [43] = 14,
  [44] = 15,
  [45] = 14,
  [46] = 15,
  [47] = 14,
  [48] = 14,
  [49] = 14,
  [50] = 14,
  [51] = 15,
  [52] = 15,
  [53] = 15,
  [54] = 15,
  [55] = 14,
  [56] = 15,
  [57] = 15,
  [58] = 14,
  [59] = 15,
  [60] = 15,
  [61] = 14,
  [62] = 14,
  [63] = 15,
  [64] = 14,
  [65] = 15,
  [66] = 14,
  [67] = 67,
  [68] = 68,
  [69] = 69,
  [70] = 70,
  [71] = 15,
  [72] = 15,
  [73] = 73,
  [74] = 73,
  [75] = 75,
  [76] = 76,
  [77] = 73,
  [78] = 75,
  [79] = 73,
  [80] = 75,
  [81] = 76,
  [82] = 82,
  [83] = 75,
  [84] = 73,
  [85] = 75,
  [86] = 76,
  [87] = 73,
  [88] = 15,
  [89] = 73,
  [90] = 76,
  [91] = 73,
  [92] = 76,
  [93] = 73,
  [94] = 75,
  [95] = 76,
  [96] = 73,
  [97] = 76,
  [98] = 73,
  [99] = 75,
  [100] = 76,
  [101] = 82,
  [102] = 76,
  [103] = 73,
  [104] = 75,
  [105] = 76,
  [106] = 76,
  [107] = 75,
  [108] = 75,
  [109] = 76,
  [110] = 75,
  [111] = 75,
  [112] = 112,
  [113] = 113,
  [114] = 113,
  [115] = 115,
  [116] = 113,
  [117] = 115,
  [118] = 113,
  [119] = 113,
  [120] = 115,
  [121] = 113,
  [122] = 115,
  [123] = 115,
  [124] = 115,
  [125] = 113,
  [126] = 113,
  [127] = 115,
  [128] = 115,
  [129] = 113,
  [130] = 115,
  [131] = 113,
  [132] = 113,
  [133] = 115,
  [134] = 113,
  [135] = 115,
  [136] = 115,
  [137] = 137,
  [138] = 137,
  [139] = 139,
  [140] = 140,
  [141] = 141,
  [142] = 140,
  [143] = 143,
  [144] = 141,
  [145] = 140,
  [146] = 143,
  [147] = 141,
  [148] = 140,
  [149] = 143,
  [150] = 141,
  [151] = 140,
  [152] = 143,
  [153] = 141,
  [154] = 140,
  [155] = 143,
  [156] = 141,
  [157] = 140,
  [158] = 143,
  [159] = 141,
  [160] = 140,
  [161] = 143,
  [162] = 141,
  [163] = 143,
  [164] = 141,
  [165] = 140,
  [166] = 143,
  [167] = 141,
  [168] = 140,
  [169] = 143,
  [170] = 141,
  [171] = 140,
  [172] = 143,
  [173] = 141,
  [174] = 140,
  [175] = 143,
  [176] = 176,
  [177] = 177,
  [178] = 178,
  [179] = 179,
  [180] = 180,
  [181] = 180,
  [182] = 182,
  [183] = 183,
  [184] = 178,
  [185] = 182,
  [186] = 183,
  [187] = 187,
  [188] = 188,
  [189] = 187,
  [190] = 180,
  [191] = 178,
  [192] = 182,
  [193] = 179,
  [194] = 182,
  [195] = 179,
  [196] = 188,
  [197] = 188,
  [198] = 179,
  [199] = 199,
  [200] = 178,
  [201] = 182,
  [202] = 183,
  [203] = 187,
  [204] = 188,
  [205] = 179,
  [206] = 187,
  [207] = 180,
  [208] = 183,
  [209] = 182,
  [210] = 179,
  [211] = 187,
  [212] = 182,
  [213] = 179,
  [214] = 179,
  [215] = 188,
  [216] = 183,
  [217] = 187,
  [218] = 188,
  [219] = 178,
  [220] = 178,
  [221] = 183,
  [222] = 182,
  [223] = 179,
  [224] = 180,
  [225] = 178,
  [226] = 183,
  [227] = 227,
  [228] = 182,
  [229] = 179,
  [230] = 183,
  [231] = 187,
  [232] = 188,
  [233] = 187,
  [234] = 180,
  [235] = 178,
  [236] = 178,
  [237] = 182,
  [238] = 179,
  [239] = 188,
  [240] = 183,
  [241] = 187,
  [242] = 188,
  [243] = 183,
  [244] = 183,
  [245] = 187,
  [246] = 188,
  [247] = 183,
  [248] = 187,
  [249] = 178,
  [250] = 188,
  [251] = 187,
  [252] = 182,
  [253] = 179,
  [254] = 178,
  [255] = 188,
  [256] = 182,
  [257] = 178,
  [258] = 258,
  [259] = 259,
  [260] = 260,
  [261] = 261,
  [262] = 199,
  [263] = 227,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 267,
  [268] = 268,
  [269] = 269,
  [270] = 270,
  [271] = 265,
  [272] = 270,
  [273] = 273,
  [274] = 270,
  [275] = 265,
  [276] = 276,
  [277] = 277,
  [278] = 266,
  [279] = 270,
  [280] = 267,
  [281] = 265,
  [282] = 282,
  [283] = 283,
  [284] = 270,
  [285] = 265,
  [286] = 286,
  [287] = 270,
  [288] = 265,
  [289] = 289,
  [290] = 290,
  [291] = 266,
  [292] = 267,
  [293] = 270,
  [294] = 265,
  [295] = 295,
  [296] = 270,
  [297] = 265,
  [298] = 265,
  [299] = 270,
  [300] = 265,
  [301] = 266,
  [302] = 267,
  [303] = 176,
  [304] = 270,
  [305] = 265,
  [306] = 270,
  [307] = 270,
  [308] = 270,
  [309] = 265,
  [310] = 177,
  [311] = 266,
  [312] = 267,
  [313] = 266,
  [314] = 314,
  [315] = 270,
  [316] = 265,
  [317] = 317,
  [318] = 266,
  [319] = 266,
  [320] = 267,
  [321] = 267,
  [322] = 270,
  [323] = 265,
  [324] = 270,
  [325] = 265,
  [326] = 266,
  [327] = 176,
  [328] = 267,
  [329] = 177,
  [330] = 270,
  [331] = 265,
  [332] = 332,
  [333] = 270,
  [334] = 265,
  [335] = 266,
  [336] = 267,
  [337] = 267,
  [338] = 338,
  [339] = 270,
  [340] = 265,
  [341] = 270,
  [342] = 265,
  [343] = 343,
  [344] = 344,
  [345] = 265,
  [346] = 261,
  [347] = 258,
  [348] = 260,
  [349] = 259,
  [350] = 258,
  [351] = 259,
  [352] = 260,
  [353] = 261,
  [354] = 354,
  [355] = 355,
  [356] = 355,
  [357] = 354,
  [358] = 354,
  [359] = 359,
  [360] = 355,
  [361] = 361,
  [362] = 362,
  [363] = 363,
  [364] = 364,
  [365] = 363,
  [366] = 354,
  [367] = 359,
  [368] = 355,
  [369] = 355,
  [370] = 364,
  [371] = 359,
  [372] = 354,
  [373] = 359,
  [374] = 364,
  [375] = 361,
  [376] = 364,
  [377] = 361,
  [378] = 361,
  [379] = 363,
  [380] = 354,
  [381] = 354,
  [382] = 363,
  [383] = 361,
  [384] = 355,
  [385] = 363,
  [386] = 361,
  [387] = 387,
  [388] = 388,
  [389] = 364,
  [390] = 363,
  [391] = 355,
  [392] = 355,
  [393] = 359,
  [394] = 363,
  [395] = 363,
  [396] = 355,
  [397] = 388,
  [398] = 364,
  [399] = 363,
  [400] = 363,
  [401] = 359,
  [402] = 354,
  [403] = 403,
  [404] = 355,
  [405] = 361,
  [406] = 355,
  [407] = 355,
  [408] = 363,
  [409] = 364,
  [410] = 359,
  [411] = 363,
  [412] = 363,
  [413] = 363,
  [414] = 355,
  [415] = 363,
  [416] = 364,
  [417] = 363,
  [418] = 363,
  [419] = 364,
  [420] = 354,
  [421] = 354,
  [422] = 354,
  [423] = 355,
  [424] = 361,
  [425] = 359,
  [426] = 361,
  [427] = 361,
  [428] = 355,
  [429] = 359,
  [430] = 355,
  [431] = 363,
  [432] = 364,
  [433] = 361,
  [434] = 355,
  [435] = 364,
  [436] = 359,
  [437] = 364,
  [438] = 355,
  [439] = 363,
  [440] = 354,
  [441] = 363,
  [442] = 361,
  [443] = 443,
  [444] = 355,
  [445] = 388,
  [446] = 388,
  [447] = 388,
  [448] = 388,
  [449] = 388,
  [450] = 388,
  [451] = 388,
  [452] = 388,
  [453] = 388,
  [454] = 388,
  [455] = 355,
  [456] = 456,
  [457] = 457,
  [458] = 458,
  [459] = 459,
  [460] = 460,
  [461] = 461,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 465,
  [466] = 466,
  [467] = 467,
  [468] = 468,
  [469] = 456,
  [470] = 470,
  [471] = 471,
  [472] = 472,
  [473] = 473,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 457,
  [480] = 460,
  [481] = 462,
  [482] = 458,
  [483] = 459,
  [484] = 463,
  [485] = 462,
  [486] = 463,
  [487] = 464,
  [488] = 465,
  [489] = 459,
  [490] = 467,
  [491] = 468,
  [492] = 456,
  [493] = 470,
  [494] = 471,
  [495] = 472,
  [496] = 473,
  [497] = 474,
  [498] = 475,
  [499] = 476,
  [500] = 477,
  [501] = 478,
  [502] = 457,
  [503] = 460,
  [504] = 458,
  [505] = 459,
  [506] = 462,
  [507] = 463,
  [508] = 464,
  [509] = 465,
  [510] = 467,
  [511] = 468,
  [512] = 456,
  [513] = 470,
  [514] = 471,
  [515] = 472,
  [516] = 473,
  [517] = 474,
  [518] = 475,
  [519] = 476,
  [520] = 477,
  [521] = 478,
  [522] = 457,
  [523] = 460,
  [524] = 458,
  [525] = 459,
  [526] = 526,
  [527] = 462,
  [528] = 463,
  [529] = 464,
  [530] = 465,
  [531] = 531,
  [532] = 467,
  [533] = 468,
  [534] = 456,
  [535] = 470,
  [536] = 536,
  [537] = 471,
  [538] = 472,
  [539] = 473,
  [540] = 474,
  [541] = 475,
  [542] = 476,
  [543] = 477,
  [544] = 478,
  [545] = 457,
  [546] = 460,
  [547] = 458,
  [548] = 459,
  [549] = 462,
  [550] = 463,
  [551] = 464,
  [552] = 465,
  [553] = 467,
  [554] = 468,
  [555] = 456,
  [556] = 470,
  [557] = 471,
  [558] = 472,
  [559] = 473,
  [560] = 474,
  [561] = 475,
  [562] = 476,
  [563] = 477,
  [564] = 478,
  [565] = 457,
  [566] = 460,
  [567] = 458,
  [568] = 459,
  [569] = 462,
  [570] = 463,
  [571] = 464,
  [572] = 465,
  [573] = 573,
  [574] = 467,
  [575] = 468,
  [576] = 456,
  [577] = 470,
  [578] = 471,
  [579] = 472,
  [580] = 473,
  [581] = 474,
  [582] = 475,
  [583] = 476,
  [584] = 477,
  [585] = 478,
  [586] = 457,
  [587] = 460,
  [588] = 458,
  [589] = 458,
  [590] = 459,
  [591] = 464,
  [592] = 462,
  [593] = 463,
  [594] = 464,
  [595] = 465,
  [596] = 467,
  [597] = 468,
  [598] = 456,
  [599] = 470,
  [600] = 471,
  [601] = 472,
  [602] = 473,
  [603] = 474,
  [604] = 475,
  [605] = 476,
  [606] = 477,
  [607] = 478,
  [608] = 457,
  [609] = 460,
  [610] = 465,
  [611] = 458,
  [612] = 459,
  [613] = 613,
  [614] = 462,
  [615] = 463,
  [616] = 464,
  [617] = 465,
  [618] = 467,
  [619] = 468,
  [620] = 456,
  [621] = 470,
  [622] = 471,
  [623] = 472,
  [624] = 473,
  [625] = 474,
  [626] = 475,
  [627] = 476,
  [628] = 477,
  [629] = 478,
  [630] = 457,
  [631] = 460,
  [632] = 458,
  [633] = 459,
  [634] = 634,
  [635] = 458,
  [636] = 459,
  [637] = 462,
  [638] = 463,
  [639] = 464,
  [640] = 465,
  [641] = 467,
  [642] = 467,
  [643] = 468,
  [644] = 456,
  [645] = 470,
  [646] = 471,
  [647] = 472,
  [648] = 473,
  [649] = 474,
  [650] = 475,
  [651] = 476,
  [652] = 477,
  [653] = 478,
  [654] = 457,
  [655] = 460,
  [656] = 458,
  [657] = 459,
  [658] = 462,
  [659] = 463,
  [660] = 464,
  [661] = 465,
  [662] = 467,
  [663] = 468,
  [664] = 456,
  [665] = 470,
  [666] = 471,
  [667] = 472,
  [668] = 473,
  [669] = 474,
  [670] = 475,
  [671] = 476,
  [672] = 477,
  [673] = 478,
  [674] = 457,
  [675] = 460,
  [676] = 458,
  [677] = 459,
  [678] = 468,
  [679] = 462,
  [680] = 463,
  [681] = 464,
  [682] = 465,
  [683] = 467,
  [684] = 468,
  [685] = 456,
  [686] = 470,
  [687] = 471,
  [688] = 472,
  [689] = 473,
  [690] = 474,
  [691] = 475,
  [692] = 476,
  [693] = 477,
  [694] = 478,
  [695] = 457,
  [696] = 460,
  [697] = 458,
  [698] = 459,
  [699] = 462,
  [700] = 463,
  [701] = 462,
  [702] = 463,
  [703] = 464,
  [704] = 478,
  [705] = 467,
  [706] = 468,
  [707] = 456,
  [708] = 470,
  [709] = 471,
  [710] = 472,
  [711] = 473,
  [712] = 474,
  [713] = 475,
  [714] = 476,
  [715] = 477,
  [716] = 478,
  [717] = 457,
  [718] = 460,
  [719] = 458,
  [720] = 459,
  [721] = 721,
  [722] = 462,
  [723] = 463,
  [724] = 464,
  [725] = 465,
  [726] = 467,
  [727] = 468,
  [728] = 456,
  [729] = 470,
  [730] = 471,
  [731] = 472,
  [732] = 473,
  [733] = 474,
  [734] = 475,
  [735] = 476,
  [736] = 477,
  [737] = 478,
  [738] = 457,
  [739] = 460,
  [740] = 458,
  [741] = 459,
  [742] = 462,
  [743] = 463,
  [744] = 464,
  [745] = 465,
  [746] = 467,
  [747] = 468,
  [748] = 456,
  [749] = 470,
  [750] = 470,
  [751] = 471,
  [752] = 472,
  [753] = 473,
  [754] = 474,
  [755] = 475,
  [756] = 476,
  [757] = 477,
  [758] = 478,
  [759] = 457,
  [760] = 460,
  [761] = 458,
  [762] = 459,
  [763] = 462,
  [764] = 463,
  [765] = 464,
  [766] = 465,
  [767] = 464,
  [768] = 467,
  [769] = 468,
  [770] = 456,
  [771] = 470,
  [772] = 465,
  [773] = 471,
  [774] = 472,
  [775] = 473,
  [776] = 474,
  [777] = 475,
  [778] = 476,
  [779] = 477,
  [780] = 478,
  [781] = 457,
  [782] = 460,
  [783] = 467,
  [784] = 468,
  [785] = 456,
  [786] = 470,
  [787] = 471,
  [788] = 472,
  [789] = 473,
  [790] = 474,
  [791] = 475,
  [792] = 721,
  [793] = 476,
  [794] = 477,
  [795] = 478,
  [796] = 457,
  [797] = 460,
  [798] = 526,
  [799] = 536,
  [800] = 458,
  [801] = 459,
  [802] = 462,
  [803] = 463,
  [804] = 464,
  [805] = 465,
  [806] = 467,
  [807] = 468,
  [808] = 456,
  [809] = 470,
  [810] = 471,
  [811] = 472,
  [812] = 473,
  [813] = 474,
  [814] = 475,
  [815] = 476,
  [816] = 477,
  [817] = 478,
  [818] = 457,
  [819] = 460,
  [820] = 461,
  [821] = 458,
  [822] = 459,
  [823] = 462,
  [824] = 463,
  [825] = 464,
  [826] = 465,
  [827] = 467,
  [828] = 468,
  [829] = 456,
  [830] = 470,
  [831] = 471,
  [832] = 472,
  [833] = 473,
  [834] = 474,
  [835] = 475,
  [836] = 476,
  [837] = 477,
  [838] = 478,
  [839] = 457,
  [840] = 460,
  [841] = 721,
  [842] = 536,
  [843] = 721,
  [844] = 536,
  [845] = 721,
  [846] = 536,
  [847] = 461,
  [848] = 721,
  [849] = 721,
  [850] = 536,
  [851] = 471,
  [852] = 458,
  [853] = 459,
  [854] = 472,
  [855] = 721,
  [856] = 473,
  [857] = 474,
  [858] = 462,
  [859] = 463,
  [860] = 536,
  [861] = 464,
  [862] = 465,
  [863] = 475,
  [864] = 467,
  [865] = 468,
  [866] = 456,
  [867] = 470,
  [868] = 721,
  [869] = 471,
  [870] = 472,
  [871] = 473,
  [872] = 474,
  [873] = 475,
  [874] = 476,
  [875] = 477,
  [876] = 478,
  [877] = 457,
  [878] = 460,
  [879] = 536,
  [880] = 721,
  [881] = 536,
  [882] = 721,
  [883] = 536,
  [884] = 476,
  [885] = 721,
  [886] = 536,
  [887] = 721,
  [888] = 461,
  [889] = 477,
  [890] = 536,
  [891] = 721,
  [892] = 721,
  [893] = 721,
  [894] = 721,
  [895] = 721,
  [896] = 721,
  [897] = 721,
  [898] = 466,
  [899] = 466,
  [900] = 466,
  [901] = 466,
  [902] = 466,
  [903] = 466,
  [904] = 466,
  [905] = 466,
  [906] = 466,
  [907] = 466,
  [908] = 466,
  [909] = 465,
  [910] = 910,
  [911] = 911,
  [912] = 912,
  [913] = 913,
  [914] = 914,
  [915] = 915,
  [916] = 916,
  [917] = 917,
  [918] = 918,
  [919] = 919,
  [920] = 920,
  [921] = 921,
  [922] = 922,
  [923] = 923,
  [924] = 924,
  [925] = 925,
  [926] = 926,
  [927] = 914,
  [928] = 921,
  [929] = 918,
  [930] = 919,
  [931] = 914,
  [932] = 924,
  [933] = 925,
  [934] = 926,
  [935] = 910,
  [936] = 911,
  [937] = 912,
  [938] = 920,
  [939] = 922,
  [940] = 923,
  [941] = 915,
  [942] = 916,
  [943] = 917,
  [944] = 913,
  [945] = 926,
  [946] = 923,
  [947] = 925,
  [948] = 915,
  [949] = 916,
  [950] = 917,
  [951] = 913,
  [952] = 918,
  [953] = 919,
  [954] = 914,
  [955] = 920,
  [956] = 921,
  [957] = 922,
  [958] = 924,
  [959] = 910,
  [960] = 911,
  [961] = 924,
  [962] = 925,
  [963] = 926,
  [964] = 910,
  [965] = 911,
  [966] = 912,
  [967] = 920,
  [968] = 914,
  [969] = 919,
  [970] = 922,
  [971] = 923,
  [972] = 915,
  [973] = 916,
  [974] = 920,
  [975] = 921,
  [976] = 922,
  [977] = 923,
  [978] = 915,
  [979] = 916,
  [980] = 917,
  [981] = 913,
  [982] = 918,
  [983] = 917,
  [984] = 913,
  [985] = 918,
  [986] = 919,
  [987] = 914,
  [988] = 914,
  [989] = 912,
  [990] = 924,
  [991] = 925,
  [992] = 926,
  [993] = 910,
  [994] = 911,
  [995] = 912,
  [996] = 921,
  [997] = 912,
  [998] = 922,
  [999] = 923,
  [1000] = 920,
  [1001] = 914,
  [1002] = 921,
  [1003] = 924,
  [1004] = 915,
  [1005] = 914,
  [1006] = 925,
  [1007] = 914,
  [1008] = 918,
  [1009] = 919,
  [1010] = 910,
  [1011] = 916,
  [1012] = 917,
  [1013] = 914,
  [1014] = 926,
  [1015] = 913,
  [1016] = 911,
  [1017] = 914,
  [1018] = 1018,
  [1019] = 1019,
  [1020] = 1019,
  [1021] = 1018,
  [1022] = 1018,
  [1023] = 1023,
  [1024] = 1023,
  [1025] = 1023,
  [1026] = 1019,
  [1027] = 1018,
  [1028] = 1019,
  [1029] = 1018,
  [1030] = 1018,
  [1031] = 1019,
  [1032] = 1018,
  [1033] = 1019,
  [1034] = 1019,
  [1035] = 1018,
  [1036] = 1018,
  [1037] = 1023,
  [1038] = 1023,
  [1039] = 1023,
  [1040] = 1023,
  [1041] = 1023,
  [1042] = 1023,
  [1043] = 1043,
  [1044] = 1044,
  [1045] = 1045,
  [1046] = 1045,
  [1047] = 1045,
  [1048] = 1045,
  [1049] = 1045,
  [1050] = 1045,
  [1051] = 1045,
  [1052] = 1045,
  [1053] = 1045,
  [1054] = 1045,
  [1055] = 1045,
  [1056] = 1045,
  [1057] = 1045,
  [1058] = 1045,
  [1059] = 1045,
  [1060] = 1045,
  [1061] = 1045,
  [1062] = 1045,
  [1063] = 1063,
  [1064] = 1063,
  [1065] = 1065,
  [1066] = 1063,
  [1067] = 1063,
  [1068] = 1063,
  [1069] = 1065,
  [1070] = 1065,
  [1071] = 1065,
  [1072] = 1065,
  [1073] = 1065,
  [1074] = 1063,
  [1075] = 1065,
  [1076] = 1065,
  [1077] = 1063,
  [1078] = 1065,
  [1079] = 1063,
  [1080] = 1065,
  [1081] = 1065,
  [1082] = 1063,
  [1083] = 1063,
  [1084] = 1063,
  [1085] = 1065,
  [1086] = 1063,
  [1087] = 1087,
  [1088] = 1088,
  [1089] = 1089,
  [1090] = 1087,
  [1091] = 1091,
  [1092] = 1089,
  [1093] = 1093,
  [1094] = 1094,
  [1095] = 1095,
  [1096] = 1096,
  [1097] = 1097,
  [1098] = 910,
  [1099] = 1099,
  [1100] = 1100,
  [1101] = 1101,
  [1102] = 1102,
  [1103] = 1103,
  [1104] = 1104,
  [1105] = 1105,
  [1106] = 1106,
  [1107] = 1087,
  [1108] = 1100,
  [1109] = 1089,
  [1110] = 1110,
  [1111] = 1111,
  [1112] = 1112,
  [1113] = 1113,
  [1114] = 926,
  [1115] = 1115,
  [1116] = 1116,
  [1117] = 1117,
  [1118] = 911,
  [1119] = 1087,
  [1120] = 1120,
  [1121] = 1089,
  [1122] = 912,
  [1123] = 1123,
  [1124] = 1124,
  [1125] = 1123,
  [1126] = 1126,
  [1127] = 1127,
  [1128] = 1128,
  [1129] = 1129,
  [1130] = 1096,
  [1131] = 924,
  [1132] = 1132,
  [1133] = 1133,
  [1134] = 1091,
  [1135] = 1097,
  [1136] = 1093,
  [1137] = 925,
  [1138] = 1128,
  [1139] = 1088,
  [1140] = 1140,
  [1141] = 1141,
  [1142] = 1142,
  [1143] = 1143,
  [1144] = 1144,
  [1145] = 1145,
  [1146] = 1146,
  [1147] = 1147,
  [1148] = 1148,
  [1149] = 1149,
  [1150] = 1150,
  [1151] = 1151,
  [1152] = 1152,
  [1153] = 1153,
  [1154] = 1154,
  [1155] = 1155,
  [1156] = 1156,
  [1157] = 1157,
  [1158] = 1158,
  [1159] = 1159,
  [1160] = 1160,
  [1161] = 1161,
  [1162] = 1162,
  [1163] = 1163,
  [1164] = 1164,
  [1165] = 1165,
  [1166] = 1166,
  [1167] = 1167,
  [1168] = 1168,
  [1169] = 1169,
  [1170] = 1170,
  [1171] = 1171,
  [1172] = 1172,
  [1173] = 1173,
  [1174] = 1174,
  [1175] = 1175,
  [1176] = 1176,
  [1177] = 1177,
  [1178] = 1178,
  [1179] = 1179,
  [1180] = 1180,
  [1181] = 1181,
  [1182] = 1182,
  [1183] = 1183,
  [1184] = 1184,
  [1185] = 1185,
  [1186] = 1186,
  [1187] = 1187,
  [1188] = 1188,
  [1189] = 1189,
  [1190] = 1115,
  [1191] = 1140,
  [1192] = 1192,
  [1193] = 1093,
  [1194] = 1088,
  [1195] = 1195,
  [1196] = 1196,
  [1197] = 1197,
  [1198] = 1198,
  [1199] = 1199,
  [1200] = 1200,
  [1201] = 1201,
  [1202] = 1202,
  [1203] = 1203,
  [1204] = 1095,
  [1205] = 1205,
  [1206] = 1206,
  [1207] = 1207,
  [1208] = 1208,
  [1209] = 1209,
  [1210] = 1210,
  [1211] = 1211,
  [1212] = 1195,
  [1213] = 1196,
  [1214] = 1214,
  [1215] = 1215,
  [1216] = 1216,
  [1217] = 1217,
  [1218] = 1218,
  [1219] = 1097,
  [1220] = 1128,
  [1221] = 1100,
  [1222] = 924,
  [1223] = 925,
  [1224] = 926,
  [1225] = 910,
  [1226] = 911,
  [1227] = 912,
  [1228] = 1123,
  [1229] = 1096,
  [1230] = 924,
  [1231] = 925,
  [1232] = 1215,
  [1233] = 1216,
  [1234] = 1234,
  [1235] = 1235,
  [1236] = 1236,
  [1237] = 1237,
  [1238] = 1238,
  [1239] = 1239,
  [1240] = 1240,
  [1241] = 1241,
  [1242] = 1242,
  [1243] = 1243,
  [1244] = 1244,
  [1245] = 1245,
  [1246] = 1246,
  [1247] = 1247,
  [1248] = 1248,
  [1249] = 1249,
  [1250] = 1250,
  [1251] = 1251,
  [1252] = 1252,
  [1253] = 1105,
  [1254] = 926,
  [1255] = 1124,
  [1256] = 910,
  [1257] = 1257,
  [1258] = 1258,
  [1259] = 1259,
  [1260] = 1260,
  [1261] = 1261,
  [1262] = 1262,
  [1263] = 1263,
  [1264] = 911,
  [1265] = 1265,
  [1266] = 1266,
  [1267] = 912,
  [1268] = 1268,
  [1269] = 1269,
  [1270] = 1270,
  [1271] = 1271,
  [1272] = 1127,
  [1273] = 1132,
  [1274] = 1274,
  [1275] = 1275,
  [1276] = 1110,
  [1277] = 1277,
  [1278] = 1278,
  [1279] = 1279,
  [1280] = 1280,
  [1281] = 1281,
  [1282] = 1282,
  [1283] = 1283,
  [1284] = 1284,
  [1285] = 1111,
  [1286] = 1286,
  [1287] = 1287,
  [1288] = 1288,
  [1289] = 1113,
  [1290] = 1290,
  [1291] = 1291,
  [1292] = 1210,
  [1293] = 1293,
  [1294] = 1294,
  [1295] = 1141,
  [1296] = 1211,
  [1297] = 1297,
  [1298] = 1298,
  [1299] = 1299,
  [1300] = 1091,
  [1301] = 1301,
  [1302] = 1117,
  [1303] = 1133,
  [1304] = 1304,
  [1305] = 1305,
  [1306] = 1306,
  [1307] = 1307,
  [1308] = 1308,
  [1309] = 1309,
  [1310] = 1310,
  [1311] = 1311,
  [1312] = 1312,
  [1313] = 1313,
  [1314] = 1314,
  [1315] = 1315,
  [1316] = 1316,
  [1317] = 1317,
  [1318] = 1318,
  [1319] = 1319,
  [1320] = 1097,
  [1321] = 1128,
  [1322] = 1322,
  [1323] = 1100,
  [1324] = 1324,
  [1325] = 1325,
  [1326] = 1123,
  [1327] = 1327,
  [1328] = 1096,
  [1329] = 1329,
  [1330] = 1330,
  [1331] = 1331,
  [1332] = 1332,
  [1333] = 1104,
  [1334] = 1334,
  [1335] = 1335,
  [1336] = 1336,
  [1337] = 1337,
  [1338] = 1338,
  [1339] = 1339,
  [1340] = 1102,
  [1341] = 1341,
  [1342] = 1103,
  [1343] = 1343,
  [1344] = 1344,
  [1345] = 1345,
  [1346] = 1346,
  [1347] = 1112,
  [1348] = 1348,
  [1349] = 1349,
  [1350] = 1350,
  [1351] = 1351,
  [1352] = 1352,
  [1353] = 1215,
  [1354] = 1104,
  [1355] = 1160,
  [1356] = 1161,
  [1357] = 1162,
  [1358] = 1163,
  [1359] = 1164,
  [1360] = 1165,
  [1361] = 1352,
  [1362] = 1166,
  [1363] = 1167,
  [1364] = 1168,
  [1365] = 1169,
  [1366] = 1170,
  [1367] = 1171,
  [1368] = 1172,
  [1369] = 1173,
  [1370] = 1174,
  [1371] = 1175,
  [1372] = 1176,
  [1373] = 1177,
  [1374] = 1178,
  [1375] = 1179,
  [1376] = 1180,
  [1377] = 1181,
  [1378] = 1182,
  [1379] = 1183,
  [1380] = 1380,
  [1381] = 1381,
  [1382] = 1184,
  [1383] = 1383,
  [1384] = 1297,
  [1385] = 1234,
  [1386] = 1282,
  [1387] = 1235,
  [1388] = 1237,
  [1389] = 1242,
  [1390] = 1088,
  [1391] = 1243,
  [1392] = 1257,
  [1393] = 1258,
  [1394] = 1259,
  [1395] = 1260,
  [1396] = 1261,
  [1397] = 1262,
  [1398] = 1263,
  [1399] = 1286,
  [1400] = 1265,
  [1401] = 1266,
  [1402] = 924,
  [1403] = 1280,
  [1404] = 1288,
  [1405] = 1287,
  [1406] = 1281,
  [1407] = 926,
  [1408] = 910,
  [1409] = 1270,
  [1410] = 925,
  [1411] = 1298,
  [1412] = 1283,
  [1413] = 1290,
  [1414] = 1284,
  [1415] = 1210,
  [1416] = 1211,
  [1417] = 1312,
  [1418] = 1313,
  [1419] = 1291,
  [1420] = 1195,
  [1421] = 1314,
  [1422] = 1315,
  [1423] = 1250,
  [1424] = 1244,
  [1425] = 1102,
  [1426] = 1245,
  [1427] = 1103,
  [1428] = 1196,
  [1429] = 1246,
  [1430] = 1112,
  [1431] = 1215,
  [1432] = 1247,
  [1433] = 911,
  [1434] = 1316,
  [1435] = 912,
  [1436] = 1317,
  [1437] = 1248,
  [1438] = 1351,
  [1439] = 1277,
  [1440] = 1249,
  [1441] = 1251,
  [1442] = 1252,
  [1443] = 1278,
  [1444] = 1186,
  [1445] = 1115,
  [1446] = 1216,
  [1447] = 1318,
  [1448] = 1140,
  [1449] = 1299,
  [1450] = 1319,
  [1451] = 1095,
  [1452] = 1322,
  [1453] = 1268,
  [1454] = 1110,
  [1455] = 1269,
  [1456] = 1210,
  [1457] = 1305,
  [1458] = 1307,
  [1459] = 1111,
  [1460] = 1113,
  [1461] = 1141,
  [1462] = 1301,
  [1463] = 1304,
  [1464] = 1308,
  [1465] = 1309,
  [1466] = 1310,
  [1467] = 1311,
  [1468] = 1211,
  [1469] = 1195,
  [1470] = 1196,
  [1471] = 1271,
  [1472] = 1158,
  [1473] = 1159,
  [1474] = 1216,
  [1475] = 1324,
  [1476] = 1275,
  [1477] = 1279,
  [1478] = 1325,
  [1479] = 1091,
  [1480] = 1127,
  [1481] = 1093,
  [1482] = 1132,
  [1483] = 1338,
  [1484] = 1239,
  [1485] = 1327,
  [1486] = 1329,
  [1487] = 1330,
  [1488] = 1331,
  [1489] = 1332,
  [1490] = 1334,
  [1491] = 1335,
  [1492] = 1336,
  [1493] = 1306,
  [1494] = 1337,
  [1495] = 924,
  [1496] = 925,
  [1497] = 926,
  [1498] = 910,
  [1499] = 911,
  [1500] = 1185,
  [1501] = 912,
  [1502] = 1293,
  [1503] = 1294,
  [1504] = 1339,
  [1505] = 1341,
  [1506] = 1343,
  [1507] = 1344,
  [1508] = 1187,
  [1509] = 1348,
  [1510] = 1143,
  [1511] = 1350,
  [1512] = 1105,
  [1513] = 1144,
  [1514] = 1514,
  [1515] = 1515,
  [1516] = 1516,
  [1517] = 1517,
  [1518] = 1518,
  [1519] = 1519,
  [1520] = 1520,
  [1521] = 1188,
  [1522] = 1522,
  [1523] = 1189,
  [1524] = 1145,
  [1525] = 1192,
  [1526] = 1526,
  [1527] = 1527,
  [1528] = 1528,
  [1529] = 1197,
  [1530] = 1142,
  [1531] = 1198,
  [1532] = 1532,
  [1533] = 1533,
  [1534] = 1534,
  [1535] = 1535,
  [1536] = 1199,
  [1537] = 1200,
  [1538] = 1201,
  [1539] = 1539,
  [1540] = 1202,
  [1541] = 1541,
  [1542] = 1203,
  [1543] = 1146,
  [1544] = 1147,
  [1545] = 1148,
  [1546] = 1149,
  [1547] = 1150,
  [1548] = 1205,
  [1549] = 1151,
  [1550] = 1206,
  [1551] = 1207,
  [1552] = 1124,
  [1553] = 1152,
  [1554] = 1554,
  [1555] = 1555,
  [1556] = 1556,
  [1557] = 1208,
  [1558] = 1345,
  [1559] = 1346,
  [1560] = 1209,
  [1561] = 1153,
  [1562] = 1214,
  [1563] = 1217,
  [1564] = 1218,
  [1565] = 1154,
  [1566] = 1236,
  [1567] = 1238,
  [1568] = 1240,
  [1569] = 1241,
  [1570] = 1155,
  [1571] = 1156,
  [1572] = 1157,
  [1573] = 1349,
  [1574] = 1274,
  [1575] = 1241,
  [1576] = 1210,
  [1577] = 1211,
  [1578] = 1299,
  [1579] = 1305,
  [1580] = 1307,
  [1581] = 1104,
  [1582] = 926,
  [1583] = 910,
  [1584] = 1102,
  [1585] = 1103,
  [1586] = 1308,
  [1587] = 1309,
  [1588] = 1310,
  [1589] = 1311,
  [1590] = 1112,
  [1591] = 1234,
  [1592] = 1235,
  [1593] = 911,
  [1594] = 912,
  [1595] = 1514,
  [1596] = 1515,
  [1597] = 1516,
  [1598] = 1517,
  [1599] = 1518,
  [1600] = 1519,
  [1601] = 1520,
  [1602] = 1522,
  [1603] = 1526,
  [1604] = 1527,
  [1605] = 1528,
  [1606] = 1532,
  [1607] = 1533,
  [1608] = 1534,
  [1609] = 1535,
  [1610] = 1539,
  [1611] = 1541,
  [1612] = 1237,
  [1613] = 1105,
  [1614] = 1124,
  [1615] = 1115,
  [1616] = 1140,
  [1617] = 1259,
  [1618] = 1095,
  [1619] = 1110,
  [1620] = 1111,
  [1621] = 1113,
  [1622] = 1141,
  [1623] = 1186,
  [1624] = 1210,
  [1625] = 1211,
  [1626] = 1195,
  [1627] = 1196,
  [1628] = 1215,
  [1629] = 1216,
  [1630] = 924,
  [1631] = 925,
  [1632] = 1127,
  [1633] = 924,
  [1634] = 925,
  [1635] = 926,
  [1636] = 910,
  [1637] = 911,
  [1638] = 912,
  [1639] = 1132,
  [1640] = 1640,
  [1641] = 1306,
  [1642] = 1185,
  [1643] = 1091,
  [1644] = 1260,
  [1645] = 1093,
  [1646] = 1261,
  [1647] = 1262,
  [1648] = 1263,
  [1649] = 1088,
  [1650] = 1187,
  [1651] = 1651,
  [1652] = 1188,
  [1653] = 1265,
  [1654] = 1189,
  [1655] = 1312,
  [1656] = 1313,
  [1657] = 1314,
  [1658] = 1315,
  [1659] = 1316,
  [1660] = 1317,
  [1661] = 1192,
  [1662] = 1318,
  [1663] = 1319,
  [1664] = 1322,
  [1665] = 1324,
  [1666] = 1325,
  [1667] = 1327,
  [1668] = 1329,
  [1669] = 1330,
  [1670] = 1331,
  [1671] = 1332,
  [1672] = 1334,
  [1673] = 1335,
  [1674] = 1336,
  [1675] = 1337,
  [1676] = 1195,
  [1677] = 1196,
  [1678] = 1197,
  [1679] = 1142,
  [1680] = 1198,
  [1681] = 1199,
  [1682] = 1200,
  [1683] = 1683,
  [1684] = 1201,
  [1685] = 1202,
  [1686] = 1203,
  [1687] = 1687,
  [1688] = 1380,
  [1689] = 1381,
  [1690] = 1383,
  [1691] = 1339,
  [1692] = 1341,
  [1693] = 1343,
  [1694] = 1205,
  [1695] = 1344,
  [1696] = 1206,
  [1697] = 1207,
  [1698] = 1208,
  [1699] = 1351,
  [1700] = 1209,
  [1701] = 1143,
  [1702] = 1214,
  [1703] = 1217,
  [1704] = 1144,
  [1705] = 1218,
  [1706] = 1145,
  [1707] = 1146,
  [1708] = 1147,
  [1709] = 1148,
  [1710] = 1215,
  [1711] = 1216,
  [1712] = 1149,
  [1713] = 1150,
  [1714] = 1236,
  [1715] = 1238,
  [1716] = 1240,
  [1717] = 1151,
  [1718] = 1242,
  [1719] = 1243,
  [1720] = 1244,
  [1721] = 1152,
  [1722] = 1245,
  [1723] = 1246,
  [1724] = 1247,
  [1725] = 1153,
  [1726] = 1248,
  [1727] = 1249,
  [1728] = 1251,
  [1729] = 1252,
  [1730] = 1268,
  [1731] = 1269,
  [1732] = 1154,
  [1733] = 1271,
  [1734] = 1274,
  [1735] = 1155,
  [1736] = 1275,
  [1737] = 1279,
  [1738] = 1280,
  [1739] = 1156,
  [1740] = 1281,
  [1741] = 1282,
  [1742] = 1283,
  [1743] = 1284,
  [1744] = 1157,
  [1745] = 1158,
  [1746] = 1159,
  [1747] = 1160,
  [1748] = 1161,
  [1749] = 1162,
  [1750] = 1286,
  [1751] = 1287,
  [1752] = 1288,
  [1753] = 1163,
  [1754] = 1164,
  [1755] = 1290,
  [1756] = 1291,
  [1757] = 1165,
  [1758] = 1293,
  [1759] = 1294,
  [1760] = 1352,
  [1761] = 1297,
  [1762] = 1166,
  [1763] = 1167,
  [1764] = 1168,
  [1765] = 1169,
  [1766] = 1298,
  [1767] = 1170,
  [1768] = 1171,
  [1769] = 1172,
  [1770] = 1173,
  [1771] = 1174,
  [1772] = 1175,
  [1773] = 1176,
  [1774] = 1177,
  [1775] = 1178,
  [1776] = 1179,
  [1777] = 1180,
  [1778] = 1257,
  [1779] = 1258,
  [1780] = 1181,
  [1781] = 1182,
  [1782] = 1183,
  [1783] = 1184,
  [1784] = 1266,
  [1785] = 1640,
  [1786] = 1270,
  [1787] = 1087,
  [1788] = 1089,
  [1789] = 1087,
  [1790] = 1089,
  [1791] = 1791,
  [1792] = 1277,
  [1793] = 1278,
  [1794] = 1338,
  [1795] = 1554,
  [1796] = 1555,
  [1797] = 1556,
  [1798] = 1349,
  [1799] = 1350,
  [1800] = 1348,
  [1801] = 1159,
  [1802] = 1218,
  [1803] = 1215,
  [1804] = 1216,
  [1805] = 1236,
  [1806] = 1238,
  [1807] = 1339,
  [1808] = 1341,
  [1809] = 1343,
  [1810] = 1344,
  [1811] = 1348,
  [1812] = 1143,
  [1813] = 1144,
  [1814] = 1145,
  [1815] = 1240,
  [1816] = 1241,
  [1817] = 1242,
  [1818] = 1243,
  [1819] = 1244,
  [1820] = 1245,
  [1821] = 1246,
  [1822] = 1247,
  [1823] = 1248,
  [1824] = 1249,
  [1825] = 1251,
  [1826] = 1252,
  [1827] = 1268,
  [1828] = 1269,
  [1829] = 1271,
  [1830] = 1274,
  [1831] = 1275,
  [1832] = 1146,
  [1833] = 1147,
  [1834] = 1279,
  [1835] = 1280,
  [1836] = 1281,
  [1837] = 1282,
  [1838] = 1283,
  [1839] = 1284,
  [1840] = 1148,
  [1841] = 1149,
  [1842] = 1150,
  [1843] = 1286,
  [1844] = 1287,
  [1845] = 1288,
  [1846] = 1151,
  [1847] = 1290,
  [1848] = 1291,
  [1849] = 1293,
  [1850] = 1294,
  [1851] = 1297,
  [1852] = 1152,
  [1853] = 1298,
  [1854] = 1153,
  [1855] = 1154,
  [1856] = 1155,
  [1857] = 1156,
  [1858] = 1210,
  [1859] = 1211,
  [1860] = 1277,
  [1861] = 1195,
  [1862] = 1196,
  [1863] = 1278,
  [1864] = 1215,
  [1865] = 1216,
  [1866] = 1157,
  [1867] = 1158,
  [1868] = 1160,
  [1869] = 1161,
  [1870] = 1162,
  [1871] = 924,
  [1872] = 925,
  [1873] = 926,
  [1874] = 910,
  [1875] = 911,
  [1876] = 912,
  [1877] = 1210,
  [1878] = 1211,
  [1879] = 1195,
  [1880] = 1196,
  [1881] = 1215,
  [1882] = 1216,
  [1883] = 1105,
  [1884] = 1124,
  [1885] = 1163,
  [1886] = 1164,
  [1887] = 1165,
  [1888] = 1352,
  [1889] = 1166,
  [1890] = 1167,
  [1891] = 1168,
  [1892] = 1169,
  [1893] = 1170,
  [1894] = 1210,
  [1895] = 1211,
  [1896] = 924,
  [1897] = 925,
  [1898] = 926,
  [1899] = 910,
  [1900] = 911,
  [1901] = 912,
  [1902] = 1299,
  [1903] = 1305,
  [1904] = 1307,
  [1905] = 1905,
  [1906] = 1308,
  [1907] = 1309,
  [1908] = 1310,
  [1909] = 1171,
  [1910] = 1172,
  [1911] = 1911,
  [1912] = 1311,
  [1913] = 1913,
  [1914] = 1173,
  [1915] = 1174,
  [1916] = 1175,
  [1917] = 1383,
  [1918] = 1176,
  [1919] = 1177,
  [1920] = 1178,
  [1921] = 1179,
  [1922] = 1180,
  [1923] = 1181,
  [1924] = 1182,
  [1925] = 1183,
  [1926] = 1186,
  [1927] = 1184,
  [1928] = 924,
  [1929] = 925,
  [1930] = 1325,
  [1931] = 1257,
  [1932] = 1327,
  [1933] = 1329,
  [1934] = 1330,
  [1935] = 1331,
  [1936] = 1144,
  [1937] = 1145,
  [1938] = 1146,
  [1939] = 1147,
  [1940] = 1148,
  [1941] = 1149,
  [1942] = 1150,
  [1943] = 1151,
  [1944] = 1152,
  [1945] = 1153,
  [1946] = 1514,
  [1947] = 1515,
  [1948] = 1516,
  [1949] = 1517,
  [1950] = 1905,
  [1951] = 1518,
  [1952] = 1519,
  [1953] = 1520,
  [1954] = 1911,
  [1955] = 1913,
  [1956] = 1522,
  [1957] = 1526,
  [1958] = 1527,
  [1959] = 1528,
  [1960] = 1532,
  [1961] = 1533,
  [1962] = 1534,
  [1963] = 1535,
  [1964] = 1539,
  [1965] = 1541,
  [1966] = 1338,
  [1967] = 1234,
  [1968] = 1115,
  [1969] = 1235,
  [1970] = 1140,
  [1971] = 1905,
  [1972] = 1095,
  [1973] = 1154,
  [1974] = 1155,
  [1975] = 1156,
  [1976] = 1324,
  [1977] = 1911,
  [1978] = 1158,
  [1979] = 1159,
  [1980] = 1913,
  [1981] = 1160,
  [1982] = 1161,
  [1983] = 1162,
  [1984] = 1163,
  [1985] = 1164,
  [1986] = 1165,
  [1987] = 1352,
  [1988] = 1166,
  [1989] = 1167,
  [1990] = 1168,
  [1991] = 1110,
  [1992] = 1169,
  [1993] = 1170,
  [1994] = 1905,
  [1995] = 1171,
  [1996] = 1172,
  [1997] = 1173,
  [1998] = 1174,
  [1999] = 1175,
  [2000] = 1176,
  [2001] = 1177,
  [2002] = 1178,
  [2003] = 1911,
  [2004] = 1179,
  [2005] = 1180,
  [2006] = 1913,
  [2007] = 1181,
  [2008] = 1182,
  [2009] = 1183,
  [2010] = 1184,
  [2011] = 1111,
  [2012] = 1113,
  [2013] = 1141,
  [2014] = 1185,
  [2015] = 1270,
  [2016] = 1332,
  [2017] = 1911,
  [2018] = 1905,
  [2019] = 1258,
  [2020] = 1187,
  [2021] = 1188,
  [2022] = 1189,
  [2023] = 1905,
  [2024] = 1192,
  [2025] = 1334,
  [2026] = 1335,
  [2027] = 1195,
  [2028] = 1196,
  [2029] = 1197,
  [2030] = 1142,
  [2031] = 1198,
  [2032] = 1911,
  [2033] = 1199,
  [2034] = 1200,
  [2035] = 1913,
  [2036] = 1201,
  [2037] = 1202,
  [2038] = 1203,
  [2039] = 1336,
  [2040] = 1337,
  [2041] = 1205,
  [2042] = 1206,
  [2043] = 1207,
  [2044] = 1208,
  [2045] = 1209,
  [2046] = 1214,
  [2047] = 1905,
  [2048] = 1217,
  [2049] = 1911,
  [2050] = 1913,
  [2051] = 1237,
  [2052] = 1127,
  [2053] = 1259,
  [2054] = 1260,
  [2055] = 1261,
  [2056] = 1262,
  [2057] = 1263,
  [2058] = 1132,
  [2059] = 1905,
  [2060] = 1306,
  [2061] = 1265,
  [2062] = 1266,
  [2063] = 1097,
  [2064] = 1911,
  [2065] = 1128,
  [2066] = 1100,
  [2067] = 1913,
  [2068] = 1123,
  [2069] = 1096,
  [2070] = 1104,
  [2071] = 1097,
  [2072] = 1128,
  [2073] = 1100,
  [2074] = 1123,
  [2075] = 1096,
  [2076] = 1381,
  [2077] = 1905,
  [2078] = 1911,
  [2079] = 1913,
  [2080] = 1913,
  [2081] = 926,
  [2082] = 1905,
  [2083] = 910,
  [2084] = 1911,
  [2085] = 1913,
  [2086] = 1312,
  [2087] = 1313,
  [2088] = 1314,
  [2089] = 1315,
  [2090] = 1316,
  [2091] = 1317,
  [2092] = 1318,
  [2093] = 1117,
  [2094] = 1319,
  [2095] = 1322,
  [2096] = 1324,
  [2097] = 1905,
  [2098] = 1325,
  [2099] = 1327,
  [2100] = 1911,
  [2101] = 1913,
  [2102] = 1329,
  [2103] = 1330,
  [2104] = 1331,
  [2105] = 1905,
  [2106] = 1332,
  [2107] = 1334,
  [2108] = 1335,
  [2109] = 1133,
  [2110] = 1911,
  [2111] = 1913,
  [2112] = 1336,
  [2113] = 1337,
  [2114] = 1102,
  [2115] = 1103,
  [2116] = 1112,
  [2117] = 1554,
  [2118] = 1555,
  [2119] = 1556,
  [2120] = 911,
  [2121] = 912,
  [2122] = 1339,
  [2123] = 1349,
  [2124] = 1350,
  [2125] = 1351,
  [2126] = 1341,
  [2127] = 1343,
  [2128] = 1344,
  [2129] = 1348,
  [2130] = 1143,
  [2131] = 1312,
  [2132] = 1313,
  [2133] = 1314,
  [2134] = 1315,
  [2135] = 1316,
  [2136] = 1317,
  [2137] = 1318,
  [2138] = 1319,
  [2139] = 1322,
  [2140] = 1157,
  [2141] = 1520,
  [2142] = 1351,
  [2143] = 1150,
  [2144] = 1257,
  [2145] = 1258,
  [2146] = 1259,
  [2147] = 1260,
  [2148] = 1261,
  [2149] = 1262,
  [2150] = 1263,
  [2151] = 1151,
  [2152] = 1152,
  [2153] = 1265,
  [2154] = 1153,
  [2155] = 1534,
  [2156] = 1266,
  [2157] = 1339,
  [2158] = 1349,
  [2159] = 1341,
  [2160] = 1343,
  [2161] = 1344,
  [2162] = 1346,
  [2163] = 1348,
  [2164] = 1270,
  [2165] = 1338,
  [2166] = 1143,
  [2167] = 1535,
  [2168] = 1087,
  [2169] = 1539,
  [2170] = 1277,
  [2171] = 1541,
  [2172] = 1278,
  [2173] = 1186,
  [2174] = 1238,
  [2175] = 1240,
  [2176] = 1089,
  [2177] = 1210,
  [2178] = 1211,
  [2179] = 1299,
  [2180] = 1305,
  [2181] = 1307,
  [2182] = 1308,
  [2183] = 1309,
  [2184] = 1554,
  [2185] = 1310,
  [2186] = 1311,
  [2187] = 1514,
  [2188] = 1089,
  [2189] = 1241,
  [2190] = 1515,
  [2191] = 1516,
  [2192] = 1517,
  [2193] = 1087,
  [2194] = 1242,
  [2195] = 1210,
  [2196] = 1518,
  [2197] = 1154,
  [2198] = 1091,
  [2199] = 1093,
  [2200] = 1155,
  [2201] = 1243,
  [2202] = 1088,
  [2203] = 1312,
  [2204] = 1091,
  [2205] = 1313,
  [2206] = 1093,
  [2207] = 1314,
  [2208] = 1244,
  [2209] = 1088,
  [2210] = 1315,
  [2211] = 1245,
  [2212] = 1316,
  [2213] = 1317,
  [2214] = 1246,
  [2215] = 1087,
  [2216] = 1089,
  [2217] = 1318,
  [2218] = 1349,
  [2219] = 1156,
  [2220] = 1247,
  [2221] = 1157,
  [2222] = 1158,
  [2223] = 1159,
  [2224] = 1160,
  [2225] = 1161,
  [2226] = 1519,
  [2227] = 1211,
  [2228] = 1162,
  [2229] = 1163,
  [2230] = 1164,
  [2231] = 1319,
  [2232] = 1165,
  [2233] = 1306,
  [2234] = 1248,
  [2235] = 1195,
  [2236] = 1196,
  [2237] = 1215,
  [2238] = 1322,
  [2239] = 1216,
  [2240] = 1185,
  [2241] = 1249,
  [2242] = 1251,
  [2243] = 1350,
  [2244] = 1250,
  [2245] = 1239,
  [2246] = 1187,
  [2247] = 1324,
  [2248] = 1089,
  [2249] = 1522,
  [2250] = 1352,
  [2251] = 1325,
  [2252] = 1252,
  [2253] = 1268,
  [2254] = 1166,
  [2255] = 1269,
  [2256] = 1188,
  [2257] = 1555,
  [2258] = 1350,
  [2259] = 1189,
  [2260] = 912,
  [2261] = 1088,
  [2262] = 1274,
  [2263] = 1192,
  [2264] = 1275,
  [2265] = 1167,
  [2266] = 1279,
  [2267] = 1280,
  [2268] = 1195,
  [2269] = 1196,
  [2270] = 1168,
  [2271] = 1327,
  [2272] = 1197,
  [2273] = 1281,
  [2274] = 1142,
  [2275] = 1526,
  [2276] = 1527,
  [2277] = 1282,
  [2278] = 1091,
  [2279] = 1329,
  [2280] = 1528,
  [2281] = 1532,
  [2282] = 1144,
  [2283] = 1145,
  [2284] = 1345,
  [2285] = 1093,
  [2286] = 1146,
  [2287] = 1147,
  [2288] = 1351,
  [2289] = 1169,
  [2290] = 1170,
  [2291] = 1383,
  [2292] = 1346,
  [2293] = 1283,
  [2294] = 1284,
  [2295] = 1171,
  [2296] = 1172,
  [2297] = 1173,
  [2298] = 1286,
  [2299] = 1345,
  [2300] = 1287,
  [2301] = 1288,
  [2302] = 1174,
  [2303] = 1175,
  [2304] = 1301,
  [2305] = 1290,
  [2306] = 1291,
  [2307] = 1381,
  [2308] = 1293,
  [2309] = 1198,
  [2310] = 1199,
  [2311] = 1200,
  [2312] = 1201,
  [2313] = 1202,
  [2314] = 1556,
  [2315] = 1203,
  [2316] = 1176,
  [2317] = 1177,
  [2318] = 1205,
  [2319] = 1206,
  [2320] = 1207,
  [2321] = 1208,
  [2322] = 1209,
  [2323] = 1178,
  [2324] = 1214,
  [2325] = 1179,
  [2326] = 1180,
  [2327] = 1181,
  [2328] = 1217,
  [2329] = 1218,
  [2330] = 1330,
  [2331] = 1182,
  [2332] = 1183,
  [2333] = 1215,
  [2334] = 1216,
  [2335] = 1554,
  [2336] = 1555,
  [2337] = 1556,
  [2338] = 1184,
  [2339] = 1331,
  [2340] = 1332,
  [2341] = 1334,
  [2342] = 924,
  [2343] = 925,
  [2344] = 1148,
  [2345] = 1149,
  [2346] = 926,
  [2347] = 1234,
  [2348] = 1235,
  [2349] = 1237,
  [2350] = 1236,
  [2351] = 910,
  [2352] = 1294,
  [2353] = 1297,
  [2354] = 1250,
  [2355] = 1298,
  [2356] = 1304,
  [2357] = 1533,
  [2358] = 1335,
  [2359] = 1336,
  [2360] = 1337,
  [2361] = 911,
  [2362] = 1087,
  [2363] = 1271,
  [2364] = 1164,
  [2365] = 1330,
  [2366] = 1331,
  [2367] = 1332,
  [2368] = 1334,
  [2369] = 1335,
  [2370] = 1336,
  [2371] = 1337,
  [2372] = 1324,
  [2373] = 1325,
  [2374] = 1210,
  [2375] = 1211,
  [2376] = 1215,
  [2377] = 1216,
  [2378] = 1195,
  [2379] = 1196,
  [2380] = 1339,
  [2381] = 1341,
  [2382] = 1215,
  [2383] = 1216,
  [2384] = 1210,
  [2385] = 1211,
  [2386] = 1195,
  [2387] = 1196,
  [2388] = 1215,
  [2389] = 1216,
  [2390] = 1343,
  [2391] = 1344,
  [2392] = 1348,
  [2393] = 1143,
  [2394] = 1144,
  [2395] = 1145,
  [2396] = 1146,
  [2397] = 1147,
  [2398] = 1148,
  [2399] = 1149,
  [2400] = 1150,
  [2401] = 1151,
  [2402] = 1152,
  [2403] = 1153,
  [2404] = 1154,
  [2405] = 1155,
  [2406] = 1156,
  [2407] = 1157,
  [2408] = 924,
  [2409] = 925,
  [2410] = 1158,
  [2411] = 1159,
  [2412] = 926,
  [2413] = 910,
  [2414] = 1160,
  [2415] = 1161,
  [2416] = 911,
  [2417] = 912,
  [2418] = 1162,
  [2419] = 1163,
  [2420] = 1164,
  [2421] = 1165,
  [2422] = 1352,
  [2423] = 1166,
  [2424] = 1167,
  [2425] = 1168,
  [2426] = 1169,
  [2427] = 1170,
  [2428] = 1171,
  [2429] = 1172,
  [2430] = 1173,
  [2431] = 1174,
  [2432] = 924,
  [2433] = 925,
  [2434] = 926,
  [2435] = 910,
  [2436] = 911,
  [2437] = 912,
  [2438] = 924,
  [2439] = 925,
  [2440] = 926,
  [2441] = 910,
  [2442] = 911,
  [2443] = 912,
  [2444] = 1175,
  [2445] = 1176,
  [2446] = 1177,
  [2447] = 1178,
  [2448] = 1179,
  [2449] = 1180,
  [2450] = 1181,
  [2451] = 1182,
  [2452] = 1183,
  [2453] = 1184,
  [2454] = 1105,
  [2455] = 1327,
  [2456] = 1329,
  [2457] = 1124,
  [2458] = 1349,
  [2459] = 1132,
  [2460] = 1330,
  [2461] = 1331,
  [2462] = 1332,
  [2463] = 1334,
  [2464] = 1335,
  [2465] = 1336,
  [2466] = 1337,
  [2467] = 911,
  [2468] = 912,
  [2469] = 1339,
  [2470] = 1341,
  [2471] = 1343,
  [2472] = 1344,
  [2473] = 1348,
  [2474] = 1143,
  [2475] = 1169,
  [2476] = 1170,
  [2477] = 1171,
  [2478] = 1380,
  [2479] = 1555,
  [2480] = 1172,
  [2481] = 1173,
  [2482] = 1350,
  [2483] = 1250,
  [2484] = 1174,
  [2485] = 1144,
  [2486] = 1145,
  [2487] = 2487,
  [2488] = 1146,
  [2489] = 1147,
  [2490] = 1148,
  [2491] = 1149,
  [2492] = 1150,
  [2493] = 1151,
  [2494] = 1152,
  [2495] = 1153,
  [2496] = 1175,
  [2497] = 1102,
  [2498] = 1176,
  [2499] = 1345,
  [2500] = 1103,
  [2501] = 1127,
  [2502] = 1112,
  [2503] = 1351,
  [2504] = 911,
  [2505] = 912,
  [2506] = 1346,
  [2507] = 1339,
  [2508] = 1341,
  [2509] = 1105,
  [2510] = 1124,
  [2511] = 1343,
  [2512] = 1344,
  [2513] = 1348,
  [2514] = 1143,
  [2515] = 1144,
  [2516] = 1145,
  [2517] = 1146,
  [2518] = 1147,
  [2519] = 1383,
  [2520] = 1148,
  [2521] = 1154,
  [2522] = 1155,
  [2523] = 1156,
  [2524] = 1157,
  [2525] = 1158,
  [2526] = 1159,
  [2527] = 1250,
  [2528] = 1160,
  [2529] = 1161,
  [2530] = 1162,
  [2531] = 1163,
  [2532] = 1164,
  [2533] = 1165,
  [2534] = 1352,
  [2535] = 1166,
  [2536] = 1167,
  [2537] = 1168,
  [2538] = 1149,
  [2539] = 1150,
  [2540] = 1151,
  [2541] = 1152,
  [2542] = 1169,
  [2543] = 1350,
  [2544] = 1170,
  [2545] = 1171,
  [2546] = 1172,
  [2547] = 1173,
  [2548] = 1174,
  [2549] = 1175,
  [2550] = 1176,
  [2551] = 1177,
  [2552] = 1178,
  [2553] = 1179,
  [2554] = 1180,
  [2555] = 1181,
  [2556] = 1182,
  [2557] = 1183,
  [2558] = 1184,
  [2559] = 1153,
  [2560] = 1514,
  [2561] = 1515,
  [2562] = 1516,
  [2563] = 1091,
  [2564] = 1093,
  [2565] = 1115,
  [2566] = 1088,
  [2567] = 1517,
  [2568] = 1518,
  [2569] = 1519,
  [2570] = 1520,
  [2571] = 1522,
  [2572] = 1177,
  [2573] = 1526,
  [2574] = 1527,
  [2575] = 1528,
  [2576] = 1532,
  [2577] = 1533,
  [2578] = 1534,
  [2579] = 1535,
  [2580] = 1539,
  [2581] = 1541,
  [2582] = 1556,
  [2583] = 1087,
  [2584] = 1346,
  [2585] = 1140,
  [2586] = 1132,
  [2587] = 2587,
  [2588] = 1095,
  [2589] = 1087,
  [2590] = 1089,
  [2591] = 1154,
  [2592] = 1155,
  [2593] = 1156,
  [2594] = 1157,
  [2595] = 1158,
  [2596] = 1159,
  [2597] = 1160,
  [2598] = 1161,
  [2599] = 1351,
  [2600] = 1162,
  [2601] = 1163,
  [2602] = 1164,
  [2603] = 1165,
  [2604] = 1352,
  [2605] = 1166,
  [2606] = 1167,
  [2607] = 1168,
  [2608] = 1110,
  [2609] = 1133,
  [2610] = 1169,
  [2611] = 1170,
  [2612] = 1171,
  [2613] = 1172,
  [2614] = 1351,
  [2615] = 1173,
  [2616] = 1174,
  [2617] = 1175,
  [2618] = 1176,
  [2619] = 1177,
  [2620] = 1178,
  [2621] = 1179,
  [2622] = 1346,
  [2623] = 1180,
  [2624] = 1181,
  [2625] = 1182,
  [2626] = 1183,
  [2627] = 1097,
  [2628] = 1128,
  [2629] = 1100,
  [2630] = 1184,
  [2631] = 1123,
  [2632] = 1096,
  [2633] = 1104,
  [2634] = 1102,
  [2635] = 1103,
  [2636] = 1111,
  [2637] = 1112,
  [2638] = 1115,
  [2639] = 1140,
  [2640] = 1514,
  [2641] = 1095,
  [2642] = 1110,
  [2643] = 1111,
  [2644] = 1113,
  [2645] = 1141,
  [2646] = 1515,
  [2647] = 1516,
  [2648] = 1113,
  [2649] = 1517,
  [2650] = 1518,
  [2651] = 1519,
  [2652] = 1520,
  [2653] = 1522,
  [2654] = 1141,
  [2655] = 1526,
  [2656] = 1527,
  [2657] = 1528,
  [2658] = 1532,
  [2659] = 1533,
  [2660] = 1534,
  [2661] = 1535,
  [2662] = 1539,
  [2663] = 1541,
  [2664] = 1089,
  [2665] = 1250,
  [2666] = 1178,
  [2667] = 1097,
  [2668] = 1128,
  [2669] = 1349,
  [2670] = 1127,
  [2671] = 924,
  [2672] = 925,
  [2673] = 1128,
  [2674] = 1100,
  [2675] = 1179,
  [2676] = 924,
  [2677] = 925,
  [2678] = 1097,
  [2679] = 1123,
  [2680] = 1167,
  [2681] = 1180,
  [2682] = 1181,
  [2683] = 1132,
  [2684] = 1381,
  [2685] = 1100,
  [2686] = 1182,
  [2687] = 1183,
  [2688] = 1105,
  [2689] = 1124,
  [2690] = 1123,
  [2691] = 1184,
  [2692] = 1111,
  [2693] = 1113,
  [2694] = 1141,
  [2695] = 1349,
  [2696] = 924,
  [2697] = 925,
  [2698] = 1210,
  [2699] = 1211,
  [2700] = 1168,
  [2701] = 1250,
  [2702] = 1117,
  [2703] = 1345,
  [2704] = 1097,
  [2705] = 1128,
  [2706] = 1100,
  [2707] = 1110,
  [2708] = 1383,
  [2709] = 1123,
  [2710] = 1554,
  [2711] = 924,
  [2712] = 1096,
  [2713] = 925,
  [2714] = 1117,
  [2715] = 1104,
  [2716] = 1380,
  [2717] = 1133,
  [2718] = 1350,
  [2719] = 1381,
  [2720] = 926,
  [2721] = 926,
  [2722] = 910,
  [2723] = 1195,
  [2724] = 1345,
  [2725] = 1196,
  [2726] = 1312,
  [2727] = 1313,
  [2728] = 1314,
  [2729] = 1315,
  [2730] = 1316,
  [2731] = 1317,
  [2732] = 1318,
  [2733] = 1319,
  [2734] = 1322,
  [2735] = 1324,
  [2736] = 1325,
  [2737] = 1327,
  [2738] = 1329,
  [2739] = 1330,
  [2740] = 1331,
  [2741] = 1332,
  [2742] = 1334,
  [2743] = 1335,
  [2744] = 1336,
  [2745] = 1337,
  [2746] = 910,
  [2747] = 1312,
  [2748] = 1339,
  [2749] = 1341,
  [2750] = 1104,
  [2751] = 1343,
  [2752] = 1344,
  [2753] = 1348,
  [2754] = 1143,
  [2755] = 1127,
  [2756] = 1144,
  [2757] = 1145,
  [2758] = 1146,
  [2759] = 926,
  [2760] = 1147,
  [2761] = 910,
  [2762] = 1312,
  [2763] = 1148,
  [2764] = 1313,
  [2765] = 1149,
  [2766] = 1314,
  [2767] = 1315,
  [2768] = 1150,
  [2769] = 1316,
  [2770] = 1317,
  [2771] = 1318,
  [2772] = 1151,
  [2773] = 1319,
  [2774] = 1322,
  [2775] = 1324,
  [2776] = 1325,
  [2777] = 1327,
  [2778] = 1329,
  [2779] = 1330,
  [2780] = 1331,
  [2781] = 1332,
  [2782] = 1152,
  [2783] = 911,
  [2784] = 1334,
  [2785] = 1335,
  [2786] = 1336,
  [2787] = 912,
  [2788] = 1337,
  [2789] = 1215,
  [2790] = 1216,
  [2791] = 1153,
  [2792] = 1154,
  [2793] = 1155,
  [2794] = 1156,
  [2795] = 1157,
  [2796] = 1158,
  [2797] = 1159,
  [2798] = 1160,
  [2799] = 1161,
  [2800] = 1162,
  [2801] = 1163,
  [2802] = 1164,
  [2803] = 1102,
  [2804] = 1165,
  [2805] = 1352,
  [2806] = 1103,
  [2807] = 1166,
  [2808] = 1112,
  [2809] = 1167,
  [2810] = 1168,
  [2811] = 1169,
  [2812] = 1170,
  [2813] = 1171,
  [2814] = 911,
  [2815] = 912,
  [2816] = 1172,
  [2817] = 1173,
  [2818] = 1339,
  [2819] = 1174,
  [2820] = 1341,
  [2821] = 1343,
  [2822] = 1175,
  [2823] = 1344,
  [2824] = 1176,
  [2825] = 1348,
  [2826] = 1177,
  [2827] = 1143,
  [2828] = 1144,
  [2829] = 1178,
  [2830] = 1145,
  [2831] = 1146,
  [2832] = 1147,
  [2833] = 1148,
  [2834] = 1149,
  [2835] = 1150,
  [2836] = 1151,
  [2837] = 1152,
  [2838] = 1179,
  [2839] = 1180,
  [2840] = 1153,
  [2841] = 1181,
  [2842] = 1182,
  [2843] = 1115,
  [2844] = 1183,
  [2845] = 1184,
  [2846] = 1096,
  [2847] = 1140,
  [2848] = 926,
  [2849] = 910,
  [2850] = 1313,
  [2851] = 1095,
  [2852] = 1314,
  [2853] = 1312,
  [2854] = 1154,
  [2855] = 1313,
  [2856] = 1155,
  [2857] = 1156,
  [2858] = 1314,
  [2859] = 1315,
  [2860] = 1316,
  [2861] = 1210,
  [2862] = 1211,
  [2863] = 1317,
  [2864] = 1318,
  [2865] = 1319,
  [2866] = 1322,
  [2867] = 1324,
  [2868] = 1325,
  [2869] = 1327,
  [2870] = 1329,
  [2871] = 1157,
  [2872] = 1330,
  [2873] = 1331,
  [2874] = 1332,
  [2875] = 1334,
  [2876] = 1335,
  [2877] = 1336,
  [2878] = 1345,
  [2879] = 1346,
  [2880] = 1337,
  [2881] = 1315,
  [2882] = 1316,
  [2883] = 1158,
  [2884] = 1317,
  [2885] = 1318,
  [2886] = 1319,
  [2887] = 1322,
  [2888] = 1195,
  [2889] = 1196,
  [2890] = 1312,
  [2891] = 1313,
  [2892] = 1314,
  [2893] = 1315,
  [2894] = 1316,
  [2895] = 1317,
  [2896] = 1159,
  [2897] = 1318,
  [2898] = 1160,
  [2899] = 1319,
  [2900] = 1161,
  [2901] = 924,
  [2902] = 1162,
  [2903] = 925,
  [2904] = 1163,
  [2905] = 926,
  [2906] = 910,
  [2907] = 911,
  [2908] = 1165,
  [2909] = 912,
  [2910] = 1352,
  [2911] = 1210,
  [2912] = 1166,
  [2913] = 1211,
  [2914] = 1195,
  [2915] = 1196,
  [2916] = 1215,
  [2917] = 1349,
  [2918] = 1350,
  [2919] = 1351,
  [2920] = 1216,
  [2921] = 1322,
  [2922] = 1324,
  [2923] = 1325,
  [2924] = 1327,
  [2925] = 1329,
  [2926] = 1096,
  [2927] = 1160,
  [2928] = 1234,
  [2929] = 1185,
  [2930] = 1235,
  [2931] = 1237,
  [2932] = 1187,
  [2933] = 1188,
  [2934] = 1555,
  [2935] = 1189,
  [2936] = 1192,
  [2937] = 1304,
  [2938] = 1195,
  [2939] = 1555,
  [2940] = 1196,
  [2941] = 1197,
  [2942] = 1142,
  [2943] = 1259,
  [2944] = 1195,
  [2945] = 1196,
  [2946] = 1260,
  [2947] = 1261,
  [2948] = 1262,
  [2949] = 1263,
  [2950] = 1265,
  [2951] = 1266,
  [2952] = 1270,
  [2953] = 1277,
  [2954] = 1278,
  [2955] = 1198,
  [2956] = 1199,
  [2957] = 1200,
  [2958] = 1514,
  [2959] = 1515,
  [2960] = 1516,
  [2961] = 1514,
  [2962] = 1515,
  [2963] = 1516,
  [2964] = 1201,
  [2965] = 1517,
  [2966] = 1518,
  [2967] = 1519,
  [2968] = 1520,
  [2969] = 1522,
  [2970] = 1517,
  [2971] = 1526,
  [2972] = 1527,
  [2973] = 1528,
  [2974] = 1532,
  [2975] = 1533,
  [2976] = 1534,
  [2977] = 1535,
  [2978] = 1539,
  [2979] = 1541,
  [2980] = 1556,
  [2981] = 1518,
  [2982] = 1519,
  [2983] = 1520,
  [2984] = 1522,
  [2985] = 1202,
  [2986] = 1526,
  [2987] = 1527,
  [2988] = 1528,
  [2989] = 1532,
  [2990] = 1533,
  [2991] = 1534,
  [2992] = 1535,
  [2993] = 1539,
  [2994] = 1541,
  [2995] = 1556,
  [2996] = 1215,
  [2997] = 1216,
  [2998] = 1203,
  [2999] = 1210,
  [3000] = 1211,
  [3001] = 1299,
  [3002] = 1305,
  [3003] = 1307,
  [3004] = 1308,
  [3005] = 1205,
  [3006] = 1206,
  [3007] = 1309,
  [3008] = 1207,
  [3009] = 1208,
  [3010] = 1209,
  [3011] = 1554,
  [3012] = 1214,
  [3013] = 1217,
  [3014] = 1218,
  [3015] = 1338,
  [3016] = 1310,
  [3017] = 1311,
  [3018] = 1215,
  [3019] = 1216,
  [3020] = 1236,
  [3021] = 1238,
  [3022] = 1240,
  [3023] = 1241,
  [3024] = 1242,
  [3025] = 1243,
  [3026] = 1244,
  [3027] = 1245,
  [3028] = 1246,
  [3029] = 1247,
  [3030] = 1248,
  [3031] = 1249,
  [3032] = 1251,
  [3033] = 1252,
  [3034] = 1268,
  [3035] = 1269,
  [3036] = 1271,
  [3037] = 1274,
  [3038] = 1275,
  [3039] = 1279,
  [3040] = 1280,
  [3041] = 1281,
  [3042] = 1282,
  [3043] = 1283,
  [3044] = 1284,
  [3045] = 1286,
  [3046] = 1306,
  [3047] = 1287,
  [3048] = 1288,
  [3049] = 1091,
  [3050] = 1290,
  [3051] = 1291,
  [3052] = 1293,
  [3053] = 1257,
  [3054] = 1185,
  [3055] = 1258,
  [3056] = 1187,
  [3057] = 1188,
  [3058] = 1555,
  [3059] = 1189,
  [3060] = 1192,
  [3061] = 1210,
  [3062] = 1211,
  [3063] = 1294,
  [3064] = 1297,
  [3065] = 1298,
  [3066] = 1195,
  [3067] = 1196,
  [3068] = 1197,
  [3069] = 1142,
  [3070] = 1195,
  [3071] = 1196,
  [3072] = 1186,
  [3073] = 1198,
  [3074] = 1199,
  [3075] = 1200,
  [3076] = 1257,
  [3077] = 1258,
  [3078] = 1201,
  [3079] = 1202,
  [3080] = 1556,
  [3081] = 1203,
  [3082] = 1215,
  [3083] = 1216,
  [3084] = 1239,
  [3085] = 1093,
  [3086] = 1088,
  [3087] = 1205,
  [3088] = 1206,
  [3089] = 1207,
  [3090] = 1208,
  [3091] = 1209,
  [3092] = 1214,
  [3093] = 1217,
  [3094] = 1218,
  [3095] = 1215,
  [3096] = 1216,
  [3097] = 1380,
  [3098] = 1236,
  [3099] = 1238,
  [3100] = 1240,
  [3101] = 1241,
  [3102] = 1242,
  [3103] = 1243,
  [3104] = 1244,
  [3105] = 1245,
  [3106] = 1246,
  [3107] = 1247,
  [3108] = 1687,
  [3109] = 924,
  [3110] = 925,
  [3111] = 926,
  [3112] = 910,
  [3113] = 1514,
  [3114] = 1515,
  [3115] = 1516,
  [3116] = 1517,
  [3117] = 1518,
  [3118] = 1519,
  [3119] = 1520,
  [3120] = 1522,
  [3121] = 1526,
  [3122] = 1527,
  [3123] = 1528,
  [3124] = 1532,
  [3125] = 1533,
  [3126] = 1534,
  [3127] = 1535,
  [3128] = 1539,
  [3129] = 1541,
  [3130] = 911,
  [3131] = 912,
  [3132] = 1248,
  [3133] = 1249,
  [3134] = 1251,
  [3135] = 1127,
  [3136] = 1252,
  [3137] = 1268,
  [3138] = 1269,
  [3139] = 1271,
  [3140] = 1274,
  [3141] = 1275,
  [3142] = 1279,
  [3143] = 1280,
  [3144] = 1281,
  [3145] = 1282,
  [3146] = 1283,
  [3147] = 1380,
  [3148] = 1284,
  [3149] = 1286,
  [3150] = 1287,
  [3151] = 1288,
  [3152] = 1290,
  [3153] = 1291,
  [3154] = 1293,
  [3155] = 1132,
  [3156] = 1294,
  [3157] = 1297,
  [3158] = 1312,
  [3159] = 1313,
  [3160] = 1314,
  [3161] = 1298,
  [3162] = 1315,
  [3163] = 1316,
  [3164] = 1514,
  [3165] = 1515,
  [3166] = 1516,
  [3167] = 1517,
  [3168] = 1518,
  [3169] = 1519,
  [3170] = 1520,
  [3171] = 1522,
  [3172] = 1526,
  [3173] = 1527,
  [3174] = 1528,
  [3175] = 1532,
  [3176] = 1533,
  [3177] = 1534,
  [3178] = 1535,
  [3179] = 1539,
  [3180] = 1541,
  [3181] = 1317,
  [3182] = 1318,
  [3183] = 1319,
  [3184] = 1322,
  [3185] = 1324,
  [3186] = 1325,
  [3187] = 1327,
  [3188] = 1329,
  [3189] = 1330,
  [3190] = 1331,
  [3191] = 1332,
  [3192] = 1334,
  [3193] = 1335,
  [3194] = 1336,
  [3195] = 1337,
  [3196] = 1105,
  [3197] = 1124,
  [3198] = 1186,
  [3199] = 1381,
  [3200] = 1250,
  [3201] = 1097,
  [3202] = 1128,
  [3203] = 1100,
  [3204] = 1123,
  [3205] = 1096,
  [3206] = 1381,
  [3207] = 1380,
  [3208] = 1117,
  [3209] = 1345,
  [3210] = 3210,
  [3211] = 1338,
  [3212] = 1683,
  [3213] = 1133,
  [3214] = 1349,
  [3215] = 3215,
  [3216] = 1339,
  [3217] = 1091,
  [3218] = 1093,
  [3219] = 1341,
  [3220] = 1088,
  [3221] = 1343,
  [3222] = 1344,
  [3223] = 1348,
  [3224] = 1143,
  [3225] = 1349,
  [3226] = 1350,
  [3227] = 1301,
  [3228] = 1144,
  [3229] = 1145,
  [3230] = 1146,
  [3231] = 1147,
  [3232] = 1148,
  [3233] = 1149,
  [3234] = 1150,
  [3235] = 1151,
  [3236] = 1152,
  [3237] = 1153,
  [3238] = 1239,
  [3239] = 1350,
  [3240] = 1351,
  [3241] = 1210,
  [3242] = 1211,
  [3243] = 1195,
  [3244] = 1196,
  [3245] = 1215,
  [3246] = 1216,
  [3247] = 1351,
  [3248] = 1346,
  [3249] = 924,
  [3250] = 925,
  [3251] = 926,
  [3252] = 910,
  [3253] = 911,
  [3254] = 912,
  [3255] = 1184,
  [3256] = 1301,
  [3257] = 1097,
  [3258] = 1128,
  [3259] = 1100,
  [3260] = 1123,
  [3261] = 1096,
  [3262] = 1104,
  [3263] = 1350,
  [3264] = 1381,
  [3265] = 1683,
  [3266] = 1312,
  [3267] = 1313,
  [3268] = 1314,
  [3269] = 1687,
  [3270] = 1315,
  [3271] = 1316,
  [3272] = 1317,
  [3273] = 1318,
  [3274] = 1319,
  [3275] = 1322,
  [3276] = 1324,
  [3277] = 1325,
  [3278] = 1327,
  [3279] = 1329,
  [3280] = 1330,
  [3281] = 1331,
  [3282] = 1332,
  [3283] = 1334,
  [3284] = 1335,
  [3285] = 1336,
  [3286] = 1337,
  [3287] = 1102,
  [3288] = 1103,
  [3289] = 1380,
  [3290] = 1112,
  [3291] = 1338,
  [3292] = 1351,
  [3293] = 1339,
  [3294] = 1341,
  [3295] = 1343,
  [3296] = 1344,
  [3297] = 1348,
  [3298] = 1143,
  [3299] = 1381,
  [3300] = 1383,
  [3301] = 1144,
  [3302] = 1145,
  [3303] = 1146,
  [3304] = 1147,
  [3305] = 1148,
  [3306] = 1149,
  [3307] = 1150,
  [3308] = 1151,
  [3309] = 1152,
  [3310] = 1153,
  [3311] = 1115,
  [3312] = 1140,
  [3313] = 1095,
  [3314] = 1154,
  [3315] = 1155,
  [3316] = 1156,
  [3317] = 1157,
  [3318] = 1158,
  [3319] = 1159,
  [3320] = 1161,
  [3321] = 1162,
  [3322] = 1163,
  [3323] = 1164,
  [3324] = 1165,
  [3325] = 1352,
  [3326] = 1166,
  [3327] = 1167,
  [3328] = 1168,
  [3329] = 1110,
  [3330] = 1169,
  [3331] = 1170,
  [3332] = 1171,
  [3333] = 1172,
  [3334] = 1173,
  [3335] = 1174,
  [3336] = 1175,
  [3337] = 1176,
  [3338] = 1177,
  [3339] = 1178,
  [3340] = 1179,
  [3341] = 1180,
  [3342] = 1181,
  [3343] = 1182,
  [3344] = 1183,
  [3345] = 1184,
  [3346] = 1111,
  [3347] = 1113,
  [3348] = 1141,
  [3349] = 1234,
  [3350] = 1235,
  [3351] = 1237,
  [3352] = 1257,
  [3353] = 1258,
  [3354] = 1259,
  [3355] = 1260,
  [3356] = 1261,
  [3357] = 1262,
  [3358] = 1263,
  [3359] = 1265,
  [3360] = 1266,
  [3361] = 1270,
  [3362] = 1277,
  [3363] = 1278,
  [3364] = 1186,
  [3365] = 1304,
  [3366] = 924,
  [3367] = 925,
  [3368] = 1299,
  [3369] = 1305,
  [3370] = 1307,
  [3371] = 1308,
  [3372] = 1383,
  [3373] = 1309,
  [3374] = 1554,
  [3375] = 1310,
  [3376] = 1311,
  [3377] = 1306,
  [3378] = 1234,
  [3379] = 1235,
  [3380] = 1237,
  [3381] = 1185,
  [3382] = 1154,
  [3383] = 1155,
  [3384] = 1187,
  [3385] = 1188,
  [3386] = 1555,
  [3387] = 1189,
  [3388] = 1192,
  [3389] = 926,
  [3390] = 910,
  [3391] = 1197,
  [3392] = 1142,
  [3393] = 1198,
  [3394] = 1199,
  [3395] = 1200,
  [3396] = 1201,
  [3397] = 1202,
  [3398] = 1556,
  [3399] = 1203,
  [3400] = 1156,
  [3401] = 1157,
  [3402] = 1205,
  [3403] = 1206,
  [3404] = 1207,
  [3405] = 1208,
  [3406] = 1209,
  [3407] = 1214,
  [3408] = 1217,
  [3409] = 1218,
  [3410] = 911,
  [3411] = 912,
  [3412] = 1236,
  [3413] = 1238,
  [3414] = 1240,
  [3415] = 1241,
  [3416] = 1242,
  [3417] = 1243,
  [3418] = 1244,
  [3419] = 1245,
  [3420] = 1246,
  [3421] = 1247,
  [3422] = 1248,
  [3423] = 1249,
  [3424] = 1251,
  [3425] = 1252,
  [3426] = 1268,
  [3427] = 1269,
  [3428] = 1271,
  [3429] = 1274,
  [3430] = 1275,
  [3431] = 1279,
  [3432] = 1280,
  [3433] = 1281,
  [3434] = 1282,
  [3435] = 1283,
  [3436] = 1284,
  [3437] = 1259,
  [3438] = 1260,
  [3439] = 1286,
  [3440] = 1261,
  [3441] = 1287,
  [3442] = 1288,
  [3443] = 1262,
  [3444] = 1290,
  [3445] = 1291,
  [3446] = 1293,
  [3447] = 1294,
  [3448] = 1297,
  [3449] = 1263,
  [3450] = 1298,
  [3451] = 1306,
  [3452] = 1158,
  [3453] = 1265,
  [3454] = 1266,
  [3455] = 1159,
  [3456] = 1160,
  [3457] = 1161,
  [3458] = 1270,
  [3459] = 1162,
  [3460] = 1250,
  [3461] = 1345,
  [3462] = 1163,
  [3463] = 1346,
  [3464] = 1164,
  [3465] = 1165,
  [3466] = 1277,
  [3467] = 1278,
  [3468] = 1352,
  [3469] = 1554,
  [3470] = 1555,
  [3471] = 1556,
  [3472] = 1166,
  [3473] = 1167,
  [3474] = 1168,
  [3475] = 1210,
  [3476] = 1211,
  [3477] = 1299,
  [3478] = 1305,
  [3479] = 1307,
  [3480] = 1308,
  [3481] = 1383,
  [3482] = 1309,
  [3483] = 1554,
  [3484] = 1310,
  [3485] = 1311,
  [3486] = 1169,
  [3487] = 1210,
  [3488] = 1211,
  [3489] = 1170,
  [3490] = 1171,
  [3491] = 1172,
  [3492] = 1173,
  [3493] = 1383,
  [3494] = 1174,
  [3495] = 1554,
  [3496] = 1175,
  [3497] = 1176,
  [3498] = 1177,
  [3499] = 1349,
  [3500] = 1350,
  [3501] = 1351,
  [3502] = 1178,
  [3503] = 1179,
  [3504] = 1180,
  [3505] = 1181,
  [3506] = 1182,
  [3507] = 1183,
  [3508] = 1349,
  [3509] = 1172,
  [3510] = 1149,
  [3511] = 1150,
  [3512] = 1151,
  [3513] = 1152,
  [3514] = 1153,
  [3515] = 1176,
  [3516] = 1239,
  [3517] = 1319,
  [3518] = 1351,
  [3519] = 1317,
  [3520] = 1110,
  [3521] = 1127,
  [3522] = 1210,
  [3523] = 1211,
  [3524] = 1195,
  [3525] = 1196,
  [3526] = 1215,
  [3527] = 1216,
  [3528] = 1301,
  [3529] = 1304,
  [3530] = 1103,
  [3531] = 1115,
  [3532] = 1350,
  [3533] = 1181,
  [3534] = 1112,
  [3535] = 924,
  [3536] = 925,
  [3537] = 1140,
  [3538] = 1380,
  [3539] = 1182,
  [3540] = 1183,
  [3541] = 1184,
  [3542] = 924,
  [3543] = 1350,
  [3544] = 1381,
  [3545] = 1095,
  [3546] = 1325,
  [3547] = 926,
  [3548] = 910,
  [3549] = 1104,
  [3550] = 1327,
  [3551] = 926,
  [3552] = 910,
  [3553] = 1312,
  [3554] = 1313,
  [3555] = 1314,
  [3556] = 1315,
  [3557] = 1316,
  [3558] = 1317,
  [3559] = 1318,
  [3560] = 1319,
  [3561] = 1322,
  [3562] = 1324,
  [3563] = 1325,
  [3564] = 1327,
  [3565] = 1329,
  [3566] = 1330,
  [3567] = 1331,
  [3568] = 1332,
  [3569] = 1334,
  [3570] = 1335,
  [3571] = 1336,
  [3572] = 1337,
  [3573] = 1210,
  [3574] = 1211,
  [3575] = 1195,
  [3576] = 1196,
  [3577] = 1215,
  [3578] = 1216,
  [3579] = 1154,
  [3580] = 1155,
  [3581] = 1156,
  [3582] = 1157,
  [3583] = 1158,
  [3584] = 1159,
  [3585] = 1160,
  [3586] = 1161,
  [3587] = 924,
  [3588] = 925,
  [3589] = 1104,
  [3590] = 1329,
  [3591] = 926,
  [3592] = 910,
  [3593] = 1312,
  [3594] = 1313,
  [3595] = 911,
  [3596] = 912,
  [3597] = 1162,
  [3598] = 1102,
  [3599] = 1210,
  [3600] = 1211,
  [3601] = 1132,
  [3602] = 1163,
  [3603] = 1195,
  [3604] = 1196,
  [3605] = 1103,
  [3606] = 1215,
  [3607] = 1216,
  [3608] = 924,
  [3609] = 925,
  [3610] = 1164,
  [3611] = 926,
  [3612] = 910,
  [3613] = 911,
  [3614] = 912,
  [3615] = 1112,
  [3616] = 1338,
  [3617] = 1314,
  [3618] = 911,
  [3619] = 912,
  [3620] = 1165,
  [3621] = 1315,
  [3622] = 1339,
  [3623] = 1341,
  [3624] = 1343,
  [3625] = 1344,
  [3626] = 1348,
  [3627] = 1143,
  [3628] = 1316,
  [3629] = 1317,
  [3630] = 1318,
  [3631] = 1319,
  [3632] = 1144,
  [3633] = 1322,
  [3634] = 1324,
  [3635] = 1145,
  [3636] = 1146,
  [3637] = 924,
  [3638] = 925,
  [3639] = 1147,
  [3640] = 1148,
  [3641] = 926,
  [3642] = 910,
  [3643] = 1149,
  [3644] = 1150,
  [3645] = 911,
  [3646] = 912,
  [3647] = 1151,
  [3648] = 1152,
  [3649] = 1153,
  [3650] = 1234,
  [3651] = 1235,
  [3652] = 1237,
  [3653] = 1257,
  [3654] = 1258,
  [3655] = 1259,
  [3656] = 1260,
  [3657] = 1261,
  [3658] = 1262,
  [3659] = 1263,
  [3660] = 1265,
  [3661] = 1266,
  [3662] = 1270,
  [3663] = 1277,
  [3664] = 1278,
  [3665] = 1186,
  [3666] = 1105,
  [3667] = 1124,
  [3668] = 1325,
  [3669] = 924,
  [3670] = 925,
  [3671] = 1210,
  [3672] = 1211,
  [3673] = 1299,
  [3674] = 1305,
  [3675] = 1307,
  [3676] = 1308,
  [3677] = 1383,
  [3678] = 1309,
  [3679] = 1554,
  [3680] = 1310,
  [3681] = 1311,
  [3682] = 1115,
  [3683] = 1352,
  [3684] = 1327,
  [3685] = 1140,
  [3686] = 1166,
  [3687] = 1095,
  [3688] = 1154,
  [3689] = 1155,
  [3690] = 1156,
  [3691] = 1157,
  [3692] = 1158,
  [3693] = 1329,
  [3694] = 1306,
  [3695] = 1330,
  [3696] = 1331,
  [3697] = 1159,
  [3698] = 1332,
  [3699] = 1185,
  [3700] = 1316,
  [3701] = 1335,
  [3702] = 1187,
  [3703] = 1188,
  [3704] = 1555,
  [3705] = 1189,
  [3706] = 1192,
  [3707] = 1336,
  [3708] = 1337,
  [3709] = 926,
  [3710] = 910,
  [3711] = 1195,
  [3712] = 1196,
  [3713] = 1197,
  [3714] = 1142,
  [3715] = 1330,
  [3716] = 1198,
  [3717] = 1199,
  [3718] = 1200,
  [3719] = 1514,
  [3720] = 1515,
  [3721] = 1516,
  [3722] = 1201,
  [3723] = 1517,
  [3724] = 1518,
  [3725] = 1519,
  [3726] = 1520,
  [3727] = 1522,
  [3728] = 1202,
  [3729] = 1526,
  [3730] = 1527,
  [3731] = 1528,
  [3732] = 1532,
  [3733] = 1533,
  [3734] = 1534,
  [3735] = 1535,
  [3736] = 1539,
  [3737] = 1541,
  [3738] = 1556,
  [3739] = 1203,
  [3740] = 1205,
  [3741] = 1206,
  [3742] = 1127,
  [3743] = 1207,
  [3744] = 1208,
  [3745] = 1209,
  [3746] = 1214,
  [3747] = 1217,
  [3748] = 1218,
  [3749] = 911,
  [3750] = 911,
  [3751] = 912,
  [3752] = 1215,
  [3753] = 1216,
  [3754] = 1177,
  [3755] = 1349,
  [3756] = 1178,
  [3757] = 1236,
  [3758] = 1238,
  [3759] = 1240,
  [3760] = 1514,
  [3761] = 1515,
  [3762] = 1516,
  [3763] = 1517,
  [3764] = 1518,
  [3765] = 1519,
  [3766] = 1520,
  [3767] = 1522,
  [3768] = 1526,
  [3769] = 1527,
  [3770] = 1528,
  [3771] = 1532,
  [3772] = 1533,
  [3773] = 1534,
  [3774] = 1535,
  [3775] = 1539,
  [3776] = 1541,
  [3777] = 1179,
  [3778] = 1241,
  [3779] = 1242,
  [3780] = 1243,
  [3781] = 1244,
  [3782] = 1245,
  [3783] = 1246,
  [3784] = 1247,
  [3785] = 1169,
  [3786] = 1170,
  [3787] = 1171,
  [3788] = 1322,
  [3789] = 1111,
  [3790] = 1331,
  [3791] = 1248,
  [3792] = 1249,
  [3793] = 1251,
  [3794] = 1252,
  [3795] = 1268,
  [3796] = 1269,
  [3797] = 1271,
  [3798] = 1274,
  [3799] = 1275,
  [3800] = 1279,
  [3801] = 1280,
  [3802] = 1281,
  [3803] = 1282,
  [3804] = 1283,
  [3805] = 1284,
  [3806] = 1332,
  [3807] = 1286,
  [3808] = 1132,
  [3809] = 1287,
  [3810] = 1288,
  [3811] = 1290,
  [3812] = 1291,
  [3813] = 1293,
  [3814] = 1350,
  [3815] = 1160,
  [3816] = 1351,
  [3817] = 1334,
  [3818] = 1335,
  [3819] = 1336,
  [3820] = 1294,
  [3821] = 1297,
  [3822] = 1298,
  [3823] = 1339,
  [3824] = 1341,
  [3825] = 1343,
  [3826] = 1344,
  [3827] = 1348,
  [3828] = 1143,
  [3829] = 912,
  [3830] = 1380,
  [3831] = 1339,
  [3832] = 925,
  [3833] = 1341,
  [3834] = 1144,
  [3835] = 1145,
  [3836] = 1146,
  [3837] = 1161,
  [3838] = 1147,
  [3839] = 1148,
  [3840] = 1149,
  [3841] = 1162,
  [3842] = 1163,
  [3843] = 1164,
  [3844] = 1165,
  [3845] = 1150,
  [3846] = 1151,
  [3847] = 1152,
  [3848] = 1153,
  [3849] = 1343,
  [3850] = 1352,
  [3851] = 1555,
  [3852] = 1166,
  [3853] = 1167,
  [3854] = 1168,
  [3855] = 1344,
  [3856] = 1110,
  [3857] = 1348,
  [3858] = 1143,
  [3859] = 1381,
  [3860] = 1210,
  [3861] = 1211,
  [3862] = 3862,
  [3863] = 1169,
  [3864] = 1170,
  [3865] = 1171,
  [3866] = 1172,
  [3867] = 1173,
  [3868] = 1174,
  [3869] = 1175,
  [3870] = 1176,
  [3871] = 1177,
  [3872] = 1178,
  [3873] = 1179,
  [3874] = 1180,
  [3875] = 1181,
  [3876] = 1182,
  [3877] = 1183,
  [3878] = 1184,
  [3879] = 1111,
  [3880] = 1113,
  [3881] = 1141,
  [3882] = 1167,
  [3883] = 1168,
  [3884] = 1113,
  [3885] = 1141,
  [3886] = 1337,
  [3887] = 1154,
  [3888] = 1155,
  [3889] = 1156,
  [3890] = 1157,
  [3891] = 1158,
  [3892] = 1159,
  [3893] = 1160,
  [3894] = 1161,
  [3895] = 1381,
  [3896] = 1162,
  [3897] = 1163,
  [3898] = 1195,
  [3899] = 1196,
  [3900] = 1164,
  [3901] = 1165,
  [3902] = 1312,
  [3903] = 1313,
  [3904] = 1314,
  [3905] = 1315,
  [3906] = 1316,
  [3907] = 1317,
  [3908] = 1318,
  [3909] = 1319,
  [3910] = 1322,
  [3911] = 1324,
  [3912] = 1325,
  [3913] = 1327,
  [3914] = 1329,
  [3915] = 1330,
  [3916] = 1331,
  [3917] = 1332,
  [3918] = 1334,
  [3919] = 1335,
  [3920] = 1336,
  [3921] = 1337,
  [3922] = 1352,
  [3923] = 1166,
  [3924] = 1167,
  [3925] = 1168,
  [3926] = 1180,
  [3927] = 1514,
  [3928] = 1515,
  [3929] = 1169,
  [3930] = 1170,
  [3931] = 1171,
  [3932] = 1172,
  [3933] = 1173,
  [3934] = 1174,
  [3935] = 1175,
  [3936] = 1176,
  [3937] = 1177,
  [3938] = 1178,
  [3939] = 1179,
  [3940] = 1215,
  [3941] = 1216,
  [3942] = 1339,
  [3943] = 1341,
  [3944] = 1343,
  [3945] = 1344,
  [3946] = 1348,
  [3947] = 1143,
  [3948] = 1180,
  [3949] = 1181,
  [3950] = 1182,
  [3951] = 1183,
  [3952] = 1516,
  [3953] = 1172,
  [3954] = 1517,
  [3955] = 1518,
  [3956] = 1519,
  [3957] = 1520,
  [3958] = 1144,
  [3959] = 1145,
  [3960] = 1146,
  [3961] = 1147,
  [3962] = 1148,
  [3963] = 1149,
  [3964] = 1150,
  [3965] = 1151,
  [3966] = 1152,
  [3967] = 1153,
  [3968] = 1522,
  [3969] = 1383,
  [3970] = 1526,
  [3971] = 1527,
  [3972] = 1528,
  [3973] = 1532,
  [3974] = 1533,
  [3975] = 1534,
  [3976] = 1535,
  [3977] = 1539,
  [3978] = 1541,
  [3979] = 1184,
  [3980] = 1173,
  [3981] = 1174,
  [3982] = 1349,
  [3983] = 1351,
  [3984] = 1175,
  [3985] = 1324,
  [3986] = 1318,
  [3987] = 1383,
  [3988] = 1154,
  [3989] = 1155,
  [3990] = 1156,
  [3991] = 1157,
  [3992] = 1158,
  [3993] = 1159,
  [3994] = 1160,
  [3995] = 1161,
  [3996] = 1162,
  [3997] = 1163,
  [3998] = 1164,
  [3999] = 1165,
  [4000] = 1352,
  [4001] = 1166,
  [4002] = 1167,
  [4003] = 1168,
  [4004] = 1144,
  [4005] = 1145,
  [4006] = 1169,
  [4007] = 1170,
  [4008] = 1171,
  [4009] = 1146,
  [4010] = 1173,
  [4011] = 1174,
  [4012] = 1175,
  [4013] = 1176,
  [4014] = 1177,
  [4015] = 1178,
  [4016] = 1179,
  [4017] = 1180,
  [4018] = 1181,
  [4019] = 1182,
  [4020] = 1183,
  [4021] = 1184,
  [4022] = 1514,
  [4023] = 1515,
  [4024] = 1516,
  [4025] = 1147,
  [4026] = 1517,
  [4027] = 1518,
  [4028] = 1519,
  [4029] = 1380,
  [4030] = 1381,
  [4031] = 1554,
  [4032] = 1383,
  [4033] = 1091,
  [4034] = 1349,
  [4035] = 1350,
  [4036] = 1351,
  [4037] = 1093,
  [4038] = 1520,
  [4039] = 1088,
  [4040] = 1522,
  [4041] = 1148,
  [4042] = 1526,
  [4043] = 1527,
  [4044] = 1102,
  [4045] = 1105,
  [4046] = 1124,
  [4047] = 1312,
  [4048] = 1349,
  [4049] = 1313,
  [4050] = 1528,
  [4051] = 1532,
  [4052] = 1533,
  [4053] = 1534,
  [4054] = 1535,
  [4055] = 1539,
  [4056] = 1541,
  [4057] = 1556,
  [4058] = 1314,
  [4059] = 1315,
  [4060] = 1334,
  [4061] = 1172,
  [4062] = 1270,
  [4063] = 1201,
  [4064] = 1173,
  [4065] = 1514,
  [4066] = 1515,
  [4067] = 1516,
  [4068] = 1517,
  [4069] = 1518,
  [4070] = 1519,
  [4071] = 1520,
  [4072] = 1522,
  [4073] = 1526,
  [4074] = 1527,
  [4075] = 1528,
  [4076] = 1532,
  [4077] = 1533,
  [4078] = 1534,
  [4079] = 1535,
  [4080] = 1174,
  [4081] = 1175,
  [4082] = 1176,
  [4083] = 1338,
  [4084] = 1177,
  [4085] = 1178,
  [4086] = 1179,
  [4087] = 1180,
  [4088] = 1181,
  [4089] = 1182,
  [4090] = 1183,
  [4091] = 1184,
  [4092] = 1111,
  [4093] = 1113,
  [4094] = 1141,
  [4095] = 1205,
  [4096] = 1237,
  [4097] = 1234,
  [4098] = 1235,
  [4099] = 1237,
  [4100] = 1257,
  [4101] = 1271,
  [4102] = 1206,
  [4103] = 1539,
  [4104] = 1541,
  [4105] = 1383,
  [4106] = 1257,
  [4107] = 1556,
  [4108] = 1258,
  [4109] = 1259,
  [4110] = 1260,
  [4111] = 1261,
  [4112] = 1262,
  [4113] = 1187,
  [4114] = 1263,
  [4115] = 1200,
  [4116] = 1265,
  [4117] = 1266,
  [4118] = 1105,
  [4119] = 1383,
  [4120] = 1185,
  [4121] = 1270,
  [4122] = 1124,
  [4123] = 1207,
  [4124] = 1517,
  [4125] = 1208,
  [4126] = 1209,
  [4127] = 1518,
  [4128] = 1195,
  [4129] = 1277,
  [4130] = 1278,
  [4131] = 1277,
  [4132] = 1278,
  [4133] = 1381,
  [4134] = 1214,
  [4135] = 1104,
  [4136] = 1274,
  [4137] = 1210,
  [4138] = 1217,
  [4139] = 1381,
  [4140] = 1211,
  [4141] = 1299,
  [4142] = 1305,
  [4143] = 1307,
  [4144] = 1308,
  [4145] = 1383,
  [4146] = 1309,
  [4147] = 1554,
  [4148] = 1310,
  [4149] = 1311,
  [4150] = 1258,
  [4151] = 1298,
  [4152] = 1186,
  [4153] = 1275,
  [4154] = 1286,
  [4155] = 1259,
  [4156] = 1287,
  [4157] = 1381,
  [4158] = 1519,
  [4159] = 1520,
  [4160] = 1522,
  [4161] = 1555,
  [4162] = 1279,
  [4163] = 1196,
  [4164] = 1514,
  [4165] = 1288,
  [4166] = 926,
  [4167] = 910,
  [4168] = 1290,
  [4169] = 1291,
  [4170] = 1306,
  [4171] = 4171,
  [4172] = 1260,
  [4173] = 1293,
  [4174] = 1185,
  [4175] = 1202,
  [4176] = 1195,
  [4177] = 1196,
  [4178] = 1526,
  [4179] = 1187,
  [4180] = 1188,
  [4181] = 1555,
  [4182] = 1527,
  [4183] = 1528,
  [4184] = 1532,
  [4185] = 1533,
  [4186] = 1534,
  [4187] = 1189,
  [4188] = 1192,
  [4189] = 1381,
  [4190] = 1535,
  [4191] = 1338,
  [4192] = 1539,
  [4193] = 1261,
  [4194] = 1312,
  [4195] = 1541,
  [4196] = 1313,
  [4197] = 1195,
  [4198] = 1196,
  [4199] = 1556,
  [4200] = 1197,
  [4201] = 1142,
  [4202] = 1314,
  [4203] = 1315,
  [4204] = 1316,
  [4205] = 1317,
  [4206] = 1318,
  [4207] = 1319,
  [4208] = 1215,
  [4209] = 1216,
  [4210] = 1215,
  [4211] = 1188,
  [4212] = 1203,
  [4213] = 1216,
  [4214] = 1265,
  [4215] = 1310,
  [4216] = 1322,
  [4217] = 1324,
  [4218] = 1325,
  [4219] = 1327,
  [4220] = 1329,
  [4221] = 1330,
  [4222] = 1331,
  [4223] = 1332,
  [4224] = 1334,
  [4225] = 1335,
  [4226] = 1336,
  [4227] = 1337,
  [4228] = 1198,
  [4229] = 1199,
  [4230] = 1200,
  [4231] = 1514,
  [4232] = 1210,
  [4233] = 1211,
  [4234] = 1515,
  [4235] = 1516,
  [4236] = 1299,
  [4237] = 1201,
  [4238] = 1517,
  [4239] = 1518,
  [4240] = 1294,
  [4241] = 1297,
  [4242] = 1519,
  [4243] = 1236,
  [4244] = 1298,
  [4245] = 1349,
  [4246] = 1520,
  [4247] = 1238,
  [4248] = 1522,
  [4249] = 1240,
  [4250] = 1241,
  [4251] = 1242,
  [4252] = 1243,
  [4253] = 1244,
  [4254] = 1245,
  [4255] = 1127,
  [4256] = 1246,
  [4257] = 1247,
  [4258] = 1202,
  [4259] = 1526,
  [4260] = 924,
  [4261] = 925,
  [4262] = 1527,
  [4263] = 1528,
  [4264] = 926,
  [4265] = 910,
  [4266] = 1532,
  [4267] = 1533,
  [4268] = 911,
  [4269] = 912,
  [4270] = 1534,
  [4271] = 1535,
  [4272] = 1539,
  [4273] = 1541,
  [4274] = 1556,
  [4275] = 1203,
  [4276] = 1305,
  [4277] = 1252,
  [4278] = 1350,
  [4279] = 1515,
  [4280] = 1516,
  [4281] = 1280,
  [4282] = 1306,
  [4283] = 1218,
  [4284] = 1282,
  [4285] = 1283,
  [4286] = 1555,
  [4287] = 1284,
  [4288] = 1205,
  [4289] = 1189,
  [4290] = 1102,
  [4291] = 4291,
  [4292] = 1206,
  [4293] = 1199,
  [4294] = 1103,
  [4295] = 1207,
  [4296] = 1208,
  [4297] = 4297,
  [4298] = 1209,
  [4299] = 1514,
  [4300] = 1515,
  [4301] = 1516,
  [4302] = 1214,
  [4303] = 1517,
  [4304] = 1518,
  [4305] = 1519,
  [4306] = 1520,
  [4307] = 1522,
  [4308] = 1198,
  [4309] = 1526,
  [4310] = 1527,
  [4311] = 1528,
  [4312] = 1532,
  [4313] = 1533,
  [4314] = 1534,
  [4315] = 1535,
  [4316] = 1539,
  [4317] = 1541,
  [4318] = 1217,
  [4319] = 1218,
  [4320] = 1268,
  [4321] = 1215,
  [4322] = 1216,
  [4323] = 1112,
  [4324] = 1351,
  [4325] = 1262,
  [4326] = 1309,
  [4327] = 911,
  [4328] = 912,
  [4329] = 1236,
  [4330] = 1238,
  [4331] = 1240,
  [4332] = 1241,
  [4333] = 1242,
  [4334] = 1243,
  [4335] = 1244,
  [4336] = 1245,
  [4337] = 1246,
  [4338] = 1235,
  [4339] = 1263,
  [4340] = 1247,
  [4341] = 1339,
  [4342] = 1341,
  [4343] = 1343,
  [4344] = 1344,
  [4345] = 1348,
  [4346] = 1248,
  [4347] = 1249,
  [4348] = 1143,
  [4349] = 1132,
  [4350] = 1248,
  [4351] = 1249,
  [4352] = 1144,
  [4353] = 1145,
  [4354] = 1251,
  [4355] = 1252,
  [4356] = 1268,
  [4357] = 1269,
  [4358] = 1146,
  [4359] = 1307,
  [4360] = 3862,
  [4361] = 1147,
  [4362] = 1148,
  [4363] = 1266,
  [4364] = 1149,
  [4365] = 1150,
  [4366] = 1151,
  [4367] = 1152,
  [4368] = 1271,
  [4369] = 1153,
  [4370] = 1274,
  [4371] = 1275,
  [4372] = 1279,
  [4373] = 1280,
  [4374] = 1281,
  [4375] = 1282,
  [4376] = 1234,
  [4377] = 1383,
  [4378] = 1269,
  [4379] = 1308,
  [4380] = 1115,
  [4381] = 1554,
  [4382] = 1140,
  [4383] = 1283,
  [4384] = 1284,
  [4385] = 1192,
  [4386] = 1286,
  [4387] = 1095,
  [4388] = 1287,
  [4389] = 1288,
  [4390] = 1290,
  [4391] = 1186,
  [4392] = 1554,
  [4393] = 1555,
  [4394] = 1556,
  [4395] = 1291,
  [4396] = 1554,
  [4397] = 1293,
  [4398] = 1349,
  [4399] = 1350,
  [4400] = 1351,
  [4401] = 1154,
  [4402] = 1155,
  [4403] = 1156,
  [4404] = 1157,
  [4405] = 1294,
  [4406] = 1297,
  [4407] = 1158,
  [4408] = 1159,
  [4409] = 1160,
  [4410] = 1161,
  [4411] = 1162,
  [4412] = 1163,
  [4413] = 1164,
  [4414] = 1165,
  [4415] = 1197,
  [4416] = 1142,
  [4417] = 1352,
  [4418] = 1166,
  [4419] = 1167,
  [4420] = 1168,
  [4421] = 924,
  [4422] = 925,
  [4423] = 1210,
  [4424] = 1211,
  [4425] = 1110,
  [4426] = 1251,
  [4427] = 1311,
  [4428] = 1169,
  [4429] = 1170,
  [4430] = 1171,
  [4431] = 1281,
  [4432] = 1526,
  [4433] = 1539,
  [4434] = 1247,
  [4435] = 1189,
  [4436] = 1192,
  [4437] = 1205,
  [4438] = 1541,
  [4439] = 1142,
  [4440] = 1206,
  [4441] = 1185,
  [4442] = 1380,
  [4443] = 1207,
  [4444] = 2487,
  [4445] = 4445,
  [4446] = 2587,
  [4447] = 2587,
  [4448] = 1514,
  [4449] = 1186,
  [4450] = 1515,
  [4451] = 1516,
  [4452] = 4452,
  [4453] = 1517,
  [4454] = 1195,
  [4455] = 1248,
  [4456] = 1518,
  [4457] = 1519,
  [4458] = 1520,
  [4459] = 1249,
  [4460] = 1198,
  [4461] = 1199,
  [4462] = 1196,
  [4463] = 1251,
  [4464] = 1252,
  [4465] = 1268,
  [4466] = 1269,
  [4467] = 1197,
  [4468] = 1516,
  [4469] = 1271,
  [4470] = 1274,
  [4471] = 1275,
  [4472] = 1279,
  [4473] = 1280,
  [4474] = 1281,
  [4475] = 1282,
  [4476] = 1283,
  [4477] = 1284,
  [4478] = 1201,
  [4479] = 1522,
  [4480] = 1517,
  [4481] = 1518,
  [4482] = 1306,
  [4483] = 1208,
  [4484] = 1519,
  [4485] = 2587,
  [4486] = 1286,
  [4487] = 1209,
  [4488] = 1234,
  [4489] = 1235,
  [4490] = 1237,
  [4491] = 1287,
  [4492] = 1288,
  [4493] = 1381,
  [4494] = 1257,
  [4495] = 1520,
  [4496] = 1383,
  [4497] = 1522,
  [4498] = 1202,
  [4499] = 1259,
  [4500] = 1260,
  [4501] = 1290,
  [4502] = 1258,
  [4503] = 1214,
  [4504] = 1555,
  [4505] = 1218,
  [4506] = 1338,
  [4507] = 1246,
  [4508] = 1527,
  [4509] = 1261,
  [4510] = 1262,
  [4511] = 1528,
  [4512] = 1532,
  [4513] = 1533,
  [4514] = 1526,
  [4515] = 1263,
  [4516] = 1291,
  [4517] = 1265,
  [4518] = 1293,
  [4519] = 1651,
  [4520] = 1534,
  [4521] = 1266,
  [4522] = 1535,
  [4523] = 1539,
  [4524] = 1541,
  [4525] = 1200,
  [4526] = 1215,
  [4527] = 1270,
  [4528] = 1216,
  [4529] = 1514,
  [4530] = 1556,
  [4531] = 1203,
  [4532] = 1277,
  [4533] = 1278,
  [4534] = 1527,
  [4535] = 1381,
  [4536] = 1528,
  [4537] = 1532,
  [4538] = 1533,
  [4539] = 1534,
  [4540] = 1515,
  [4541] = 1187,
  [4542] = 1535,
  [4543] = 1791,
  [4544] = 1236,
  [4545] = 1210,
  [4546] = 1211,
  [4547] = 1299,
  [4548] = 1305,
  [4549] = 1307,
  [4550] = 1308,
  [4551] = 1188,
  [4552] = 1238,
  [4553] = 1383,
  [4554] = 1309,
  [4555] = 1554,
  [4556] = 1310,
  [4557] = 1311,
  [4558] = 1294,
  [4559] = 1297,
  [4560] = 1240,
  [4561] = 1241,
  [4562] = 1298,
  [4563] = 1242,
  [4564] = 1243,
  [4565] = 1244,
  [4566] = 1245,
  [4567] = 1217,
  [4568] = 4568,
  [4569] = 4569,
  [4570] = 4570,
  [4571] = 4569,
  [4572] = 4570,
  [4573] = 4569,
  [4574] = 4574,
  [4575] = 4570,
  [4576] = 4569,
  [4577] = 4570,
  [4578] = 4569,
  [4579] = 4579,
  [4580] = 4580,
  [4581] = 4570,
  [4582] = 4569,
  [4583] = 4570,
  [4584] = 4569,
  [4585] = 4570,
  [4586] = 4569,
  [4587] = 4570,
  [4588] = 4570,
  [4589] = 4569,
  [4590] = 4569,
  [4591] = 4569,
  [4592] = 4569,
  [4593] = 4570,
  [4594] = 4594,
  [4595] = 4595,
  [4596] = 4595,
  [4597] = 4597,
  [4598] = 4597,
  [4599] = 4595,
  [4600] = 4595,
  [4601] = 4597,
  [4602] = 4595,
  [4603] = 4597,
  [4604] = 4597,
  [4605] = 4595,
  [4606] = 4597,
  [4607] = 4595,
  [4608] = 4597,
  [4609] = 4597,
  [4610] = 4597,
  [4611] = 4595,
  [4612] = 4595,
  [4613] = 4597,
  [4614] = 4595,
  [4615] = 4595,
  [4616] = 4597,
  [4617] = 4595,
  [4618] = 4595,
  [4619] = 4597,
  [4620] = 4595,
  [4621] = 4597,
  [4622] = 4595,
  [4623] = 4597,
  [4624] = 4595,
  [4625] = 4597,
  [4626] = 4595,
  [4627] = 4597,
  [4628] = 4595,
  [4629] = 4597,
  [4630] = 4597,
  [4631] = 4631,
  [4632] = 4632,
  [4633] = 4631,
  [4634] = 4634,
  [4635] = 4635,
  [4636] = 4636,
  [4637] = 4636,
  [4638] = 4638,
  [4639] = 4631,
  [4640] = 4634,
  [4641] = 4635,
  [4642] = 4632,
  [4643] = 4636,
  [4644] = 4631,
  [4645] = 4634,
  [4646] = 4635,
  [4647] = 4636,
  [4648] = 4638,
  [4649] = 4638,
  [4650] = 4632,
  [4651] = 4631,
  [4652] = 4634,
  [4653] = 4635,
  [4654] = 4636,
  [4655] = 4638,
  [4656] = 4632,
  [4657] = 4631,
  [4658] = 4634,
  [4659] = 4635,
  [4660] = 4636,
  [4661] = 4638,
  [4662] = 4634,
  [4663] = 4635,
  [4664] = 4632,
  [4665] = 4632,
  [4666] = 4631,
  [4667] = 4634,
  [4668] = 4635,
  [4669] = 4636,
  [4670] = 4638,
  [4671] = 4631,
  [4672] = 4634,
  [4673] = 4635,
  [4674] = 4636,
  [4675] = 4632,
  [4676] = 4632,
  [4677] = 4631,
  [4678] = 4634,
  [4679] = 4635,
  [4680] = 4636,
  [4681] = 4638,
  [4682] = 4638,
  [4683] = 4632,
  [4684] = 4631,
  [4685] = 4634,
  [4686] = 4635,
  [4687] = 4636,
  [4688] = 4638,
  [4689] = 4632,
  [4690] = 4631,
  [4691] = 4631,
  [4692] = 4634,
  [4693] = 4635,
  [4694] = 4634,
  [4695] = 4635,
  [4696] = 4636,
  [4697] = 4638,
  [4698] = 4632,
  [4699] = 4632,
  [4700] = 4636,
  [4701] = 4638,
  [4702] = 4638,
  [4703] = 4703,
  [4704] = 4704,
  [4705] = 4705,
  [4706] = 4705,
  [4707] = 4707,
  [4708] = 4708,
  [4709] = 4708,
  [4710] = 4705,
  [4711] = 4711,
  [4712] = 4712,
  [4713] = 4707,
  [4714] = 4711,
  [4715] = 4707,
  [4716] = 4705,
  [4717] = 4705,
  [4718] = 4712,
  [4719] = 4707,
  [4720] = 4708,
  [4721] = 4708,
  [4722] = 4711,
  [4723] = 4712,
  [4724] = 4711,
  [4725] = 4707,
  [4726] = 4712,
  [4727] = 4705,
  [4728] = 4705,
  [4729] = 4707,
  [4730] = 4708,
  [4731] = 4707,
  [4732] = 4708,
  [4733] = 4705,
  [4734] = 4712,
  [4735] = 4708,
  [4736] = 4711,
  [4737] = 4705,
  [4738] = 4707,
  [4739] = 4711,
  [4740] = 4711,
  [4741] = 4708,
  [4742] = 4712,
  [4743] = 4711,
  [4744] = 4712,
  [4745] = 4712,
  [4746] = 4708,
  [4747] = 4705,
  [4748] = 4705,
  [4749] = 4707,
  [4750] = 4707,
  [4751] = 4707,
  [4752] = 4707,
  [4753] = 4712,
  [4754] = 4708,
  [4755] = 4708,
  [4756] = 4711,
  [4757] = 4712,
  [4758] = 4758,
  [4759] = 4711,
  [4760] = 4705,
  [4761] = 4712,
  [4762] = 4708,
  [4763] = 4711,
  [4764] = 4712,
  [4765] = 4711,
  [4766] = 4766,
  [4767] = 4767,
  [4768] = 4768,
  [4769] = 4769,
  [4770] = 4770,
  [4771] = 4771,
  [4772] = 4772,
  [4773] = 4773,
  [4774] = 4774,
  [4775] = 4775,
  [4776] = 4776,
  [4777] = 4777,
  [4778] = 4778,
  [4779] = 4779,
  [4780] = 4780,
  [4781] = 4781,
  [4782] = 4782,
  [4783] = 4783,
  [4784] = 4784,
  [4785] = 4785,
  [4786] = 4786,
  [4787] = 4787,
  [4788] = 4788,
  [4789] = 4789,
  [4790] = 4704,
  [4791] = 4791,
  [4792] = 4703,
  [4793] = 4793,
  [4794] = 4794,
  [4795] = 4795,
  [4796] = 4796,
  [4797] = 4797,
  [4798] = 4797,
  [4799] = 4797,
  [4800] = 4800,
  [4801] = 4800,
  [4802] = 4797,
  [4803] = 4800,
  [4804] = 4800,
  [4805] = 4800,
  [4806] = 4797,
  [4807] = 4800,
  [4808] = 4800,
  [4809] = 4800,
  [4810] = 4797,
  [4811] = 4800,
  [4812] = 4800,
  [4813] = 4797,
  [4814] = 4797,
  [4815] = 4797,
  [4816] = 4797,
  [4817] = 4797,
  [4818] = 4797,
  [4819] = 4800,
  [4820] = 4800,
  [4821] = 4797,
  [4822] = 4797,
  [4823] = 4797,
  [4824] = 4800,
  [4825] = 4800,
  [4826] = 4797,
  [4827] = 4800,
  [4828] = 4797,
  [4829] = 4800,
  [4830] = 4800,
  [4831] = 4800,
  [4832] = 4800,
  [4833] = 4800,
  [4834] = 4800,
  [4835] = 4800,
  [4836] = 4797,
  [4837] = 4797,
  [4838] = 4797,
  [4839] = 4800,
  [4840] = 1338,
  [4841] = 4841,
  [4842] = 4842,
  [4843] = 4843,
  [4844] = 4844,
  [4845] = 4845,
  [4846] = 4846,
  [4847] = 4847,
  [4848] = 4848,
  [4849] = 4849,
  [4850] = 4850,
  [4851] = 4848,
  [4852] = 4849,
  [4853] = 4850,
  [4854] = 4854,
  [4855] = 4855,
  [4856] = 4856,
  [4857] = 4844,
  [4858] = 4845,
  [4859] = 4846,
  [4860] = 4844,
  [4861] = 4845,
  [4862] = 4846,
  [4863] = 4847,
  [4864] = 4847,
  [4865] = 4844,
  [4866] = 4845,
  [4867] = 4846,
  [4868] = 4848,
  [4869] = 4849,
  [4870] = 4850,
  [4871] = 4854,
  [4872] = 4855,
  [4873] = 4856,
  [4874] = 4847,
  [4875] = 4854,
  [4876] = 4848,
  [4877] = 4849,
  [4878] = 4850,
  [4879] = 4848,
  [4880] = 4849,
  [4881] = 4850,
  [4882] = 4854,
  [4883] = 4855,
  [4884] = 4856,
  [4885] = 4854,
  [4886] = 4855,
  [4887] = 4844,
  [4888] = 4845,
  [4889] = 4846,
  [4890] = 4847,
  [4891] = 4856,
  [4892] = 4855,
  [4893] = 4856,
  [4894] = 4848,
  [4895] = 4849,
  [4896] = 4850,
  [4897] = 4854,
  [4898] = 4855,
  [4899] = 4856,
  [4900] = 4848,
  [4901] = 4849,
  [4902] = 4850,
  [4903] = 4854,
  [4904] = 4855,
  [4905] = 4856,
  [4906] = 4848,
  [4907] = 4855,
  [4908] = 4844,
  [4909] = 4845,
  [4910] = 4844,
  [4911] = 4845,
  [4912] = 4846,
  [4913] = 4847,
  [4914] = 4846,
  [4915] = 4847,
  [4916] = 4854,
  [4917] = 4844,
  [4918] = 4845,
  [4919] = 4846,
  [4920] = 4847,
  [4921] = 4844,
  [4922] = 4856,
  [4923] = 4849,
  [4924] = 4848,
  [4925] = 4849,
  [4926] = 4850,
  [4927] = 4854,
  [4928] = 4855,
  [4929] = 4856,
  [4930] = 4848,
  [4931] = 4844,
  [4932] = 4845,
  [4933] = 4846,
  [4934] = 4847,
  [4935] = 4849,
  [4936] = 4850,
  [4937] = 4854,
  [4938] = 4848,
  [4939] = 4849,
  [4940] = 4850,
  [4941] = 4854,
  [4942] = 4855,
  [4943] = 4856,
  [4944] = 4855,
  [4945] = 4844,
  [4946] = 4845,
  [4947] = 4846,
  [4948] = 4847,
  [4949] = 4856,
  [4950] = 4844,
  [4951] = 4845,
  [4952] = 4848,
  [4953] = 4849,
  [4954] = 4850,
  [4955] = 4854,
  [4956] = 4855,
  [4957] = 4856,
  [4958] = 4846,
  [4959] = 4844,
  [4960] = 4845,
  [4961] = 4846,
  [4962] = 4847,
  [4963] = 4844,
  [4964] = 4845,
  [4965] = 4846,
  [4966] = 4848,
  [4967] = 4849,
  [4968] = 4850,
  [4969] = 4854,
  [4970] = 4855,
  [4971] = 4856,
  [4972] = 4847,
  [4973] = 4844,
  [4974] = 4845,
  [4975] = 4846,
  [4976] = 4844,
  [4977] = 4844,
  [4978] = 4845,
  [4979] = 4848,
  [4980] = 4848,
  [4981] = 4849,
  [4982] = 4850,
  [4983] = 4854,
  [4984] = 4855,
  [4985] = 4856,
  [4986] = 4849,
  [4987] = 4844,
  [4988] = 4845,
  [4989] = 4846,
  [4990] = 4847,
  [4991] = 4846,
  [4992] = 4848,
  [4993] = 4849,
  [4994] = 4848,
  [4995] = 4849,
  [4996] = 4850,
  [4997] = 4854,
  [4998] = 4855,
  [4999] = 4856,
  [5000] = 4850,
  [5001] = 4844,
  [5002] = 4845,
  [5003] = 4846,
  [5004] = 4847,
  [5005] = 4850,
  [5006] = 4844,
  [5007] = 4845,
  [5008] = 4846,
  [5009] = 4847,
  [5010] = 4844,
  [5011] = 4845,
  [5012] = 4846,
  [5013] = 4847,
  [5014] = 4844,
  [5015] = 4845,
  [5016] = 4846,
  [5017] = 4847,
  [5018] = 4844,
  [5019] = 4845,
  [5020] = 4846,
  [5021] = 4847,
  [5022] = 4844,
  [5023] = 4845,
  [5024] = 4846,
  [5025] = 4847,
  [5026] = 4844,
  [5027] = 4845,
  [5028] = 4846,
  [5029] = 4847,
  [5030] = 4844,
  [5031] = 4845,
  [5032] = 4846,
  [5033] = 4847,
  [5034] = 4844,
  [5035] = 4845,
  [5036] = 4846,
  [5037] = 4847,
  [5038] = 4844,
  [5039] = 4845,
  [5040] = 4846,
  [5041] = 4847,
  [5042] = 4844,
  [5043] = 4845,
  [5044] = 4846,
  [5045] = 4847,
  [5046] = 4844,
  [5047] = 4845,
  [5048] = 4846,
  [5049] = 4847,
  [5050] = 4844,
  [5051] = 4845,
  [5052] = 4854,
  [5053] = 4845,
  [5054] = 4844,
  [5055] = 4845,
  [5056] = 4844,
  [5057] = 4845,
  [5058] = 4844,
  [5059] = 4845,
  [5060] = 4844,
  [5061] = 4845,
  [5062] = 4844,
  [5063] = 4845,
  [5064] = 4854,
  [5065] = 4855,
  [5066] = 4855,
  [5067] = 4844,
  [5068] = 4845,
  [5069] = 4846,
  [5070] = 4847,
  [5071] = 4856,
  [5072] = 4845,
  [5073] = 4856,
  [5074] = 4847,
  [5075] = 4847,
  [5076] = 4850,
  [5077] = 4848,
  [5078] = 4846,
  [5079] = 4849,
  [5080] = 4850,
  [5081] = 4847,
  [5082] = 4854,
  [5083] = 4855,
  [5084] = 4856,
  [5085] = 4847,
  [5086] = 5086,
  [5087] = 5087,
  [5088] = 5088,
  [5089] = 5089,
  [5090] = 5090,
  [5091] = 5091,
  [5092] = 5092,
  [5093] = 5093,
  [5094] = 5094,
  [5095] = 5095,
  [5096] = 5096,
  [5097] = 5097,
  [5098] = 5098,
  [5099] = 5099,
  [5100] = 5100,
  [5101] = 5091,
  [5102] = 5089,
  [5103] = 5103,
  [5104] = 5089,
  [5105] = 5105,
  [5106] = 5105,
  [5107] = 5107,
  [5108] = 5108,
  [5109] = 5109,
  [5110] = 5110,
  [5111] = 5111,
  [5112] = 5112,
  [5113] = 5113,
  [5114] = 5114,
  [5115] = 5086,
  [5116] = 5116,
  [5117] = 5117,
  [5118] = 5118,
  [5119] = 5119,
  [5120] = 4773,
  [5121] = 5121,
  [5122] = 5108,
  [5123] = 5123,
  [5124] = 5124,
  [5125] = 5125,
  [5126] = 5126,
  [5127] = 5110,
  [5128] = 5105,
  [5129] = 5108,
  [5130] = 5111,
  [5131] = 4773,
  [5132] = 5123,
  [5133] = 5133,
  [5134] = 5134,
  [5135] = 5087,
  [5136] = 5088,
  [5137] = 5099,
  [5138] = 5100,
  [5139] = 5103,
  [5140] = 5107,
  [5141] = 5109,
  [5142] = 5112,
  [5143] = 5116,
  [5144] = 5117,
  [5145] = 5118,
  [5146] = 5119,
  [5147] = 5113,
  [5148] = 5133,
  [5149] = 5149,
  [5150] = 5114,
  [5151] = 5151,
  [5152] = 5121,
  [5153] = 5086,
  [5154] = 5154,
  [5155] = 5149,
  [5156] = 5151,
  [5157] = 5121,
  [5158] = 5154,
  [5159] = 5159,
  [5160] = 5160,
  [5161] = 5159,
  [5162] = 5160,
  [5163] = 5110,
  [5164] = 5111,
  [5165] = 5165,
  [5166] = 5113,
  [5167] = 5114,
  [5168] = 5090,
  [5169] = 5091,
  [5170] = 5092,
  [5171] = 5093,
  [5172] = 5094,
  [5173] = 5095,
  [5174] = 5096,
  [5175] = 5097,
  [5176] = 5090,
  [5177] = 5091,
  [5178] = 5092,
  [5179] = 5093,
  [5180] = 5094,
  [5181] = 5095,
  [5182] = 5096,
  [5183] = 5097,
  [5184] = 5098,
  [5185] = 5098,
  [5186] = 5086,
  [5187] = 5089,
  [5188] = 5105,
  [5189] = 5124,
  [5190] = 5108,
  [5191] = 5191,
  [5192] = 5110,
  [5193] = 5111,
  [5194] = 5090,
  [5195] = 5113,
  [5196] = 5114,
  [5197] = 5086,
  [5198] = 5091,
  [5199] = 4791,
  [5200] = 5092,
  [5201] = 5093,
  [5202] = 5094,
  [5203] = 5095,
  [5204] = 5125,
  [5205] = 5096,
  [5206] = 5124,
  [5207] = 5125,
  [5208] = 5097,
  [5209] = 5098,
  [5210] = 5089,
  [5211] = 5105,
  [5212] = 5089,
  [5213] = 5123,
  [5214] = 5133,
  [5215] = 5134,
  [5216] = 5087,
  [5217] = 5088,
  [5218] = 5099,
  [5219] = 5100,
  [5220] = 5103,
  [5221] = 5107,
  [5222] = 5109,
  [5223] = 5112,
  [5224] = 5116,
  [5225] = 5117,
  [5226] = 5118,
  [5227] = 5119,
  [5228] = 5105,
  [5229] = 5108,
  [5230] = 5108,
  [5231] = 5086,
  [5232] = 5149,
  [5233] = 5151,
  [5234] = 5121,
  [5235] = 5154,
  [5236] = 5159,
  [5237] = 5160,
  [5238] = 5110,
  [5239] = 5111,
  [5240] = 5110,
  [5241] = 5113,
  [5242] = 5114,
  [5243] = 5086,
  [5244] = 5111,
  [5245] = 4770,
  [5246] = 5113,
  [5247] = 5114,
  [5248] = 5092,
  [5249] = 4787,
  [5250] = 5123,
  [5251] = 5133,
  [5252] = 5090,
  [5253] = 5091,
  [5254] = 5092,
  [5255] = 5093,
  [5256] = 5094,
  [5257] = 5095,
  [5258] = 5096,
  [5259] = 5097,
  [5260] = 5098,
  [5261] = 5086,
  [5262] = 5149,
  [5263] = 1351,
  [5264] = 1351,
  [5265] = 5089,
  [5266] = 5134,
  [5267] = 5087,
  [5268] = 5088,
  [5269] = 5099,
  [5270] = 5105,
  [5271] = 5100,
  [5272] = 5108,
  [5273] = 5273,
  [5274] = 5110,
  [5275] = 5111,
  [5276] = 5113,
  [5277] = 5114,
  [5278] = 5086,
  [5279] = 5124,
  [5280] = 5125,
  [5281] = 5090,
  [5282] = 4794,
  [5283] = 5154,
  [5284] = 5124,
  [5285] = 5103,
  [5286] = 5107,
  [5287] = 5124,
  [5288] = 5125,
  [5289] = 5125,
  [5290] = 5109,
  [5291] = 5112,
  [5292] = 5124,
  [5293] = 5125,
  [5294] = 5116,
  [5295] = 5123,
  [5296] = 5133,
  [5297] = 5134,
  [5298] = 5087,
  [5299] = 5088,
  [5300] = 5099,
  [5301] = 5100,
  [5302] = 5103,
  [5303] = 5107,
  [5304] = 5109,
  [5305] = 5112,
  [5306] = 5116,
  [5307] = 5117,
  [5308] = 5118,
  [5309] = 5119,
  [5310] = 5117,
  [5311] = 5123,
  [5312] = 5133,
  [5313] = 5118,
  [5314] = 5134,
  [5315] = 5087,
  [5316] = 5149,
  [5317] = 5151,
  [5318] = 5121,
  [5319] = 5154,
  [5320] = 5159,
  [5321] = 5160,
  [5322] = 5088,
  [5323] = 5099,
  [5324] = 5100,
  [5325] = 5103,
  [5326] = 5107,
  [5327] = 5109,
  [5328] = 5112,
  [5329] = 5116,
  [5330] = 5117,
  [5331] = 5118,
  [5332] = 5090,
  [5333] = 5091,
  [5334] = 5092,
  [5335] = 5093,
  [5336] = 5094,
  [5337] = 5095,
  [5338] = 5096,
  [5339] = 5097,
  [5340] = 5098,
  [5341] = 5119,
  [5342] = 5091,
  [5343] = 5089,
  [5344] = 5105,
  [5345] = 5108,
  [5346] = 5346,
  [5347] = 5110,
  [5348] = 5111,
  [5349] = 5113,
  [5350] = 5114,
  [5351] = 5086,
  [5352] = 5352,
  [5353] = 5092,
  [5354] = 5149,
  [5355] = 5151,
  [5356] = 5121,
  [5357] = 5154,
  [5358] = 5124,
  [5359] = 5125,
  [5360] = 5159,
  [5361] = 5119,
  [5362] = 5160,
  [5363] = 5123,
  [5364] = 5133,
  [5365] = 5123,
  [5366] = 5133,
  [5367] = 5134,
  [5368] = 5087,
  [5369] = 5088,
  [5370] = 5099,
  [5371] = 5100,
  [5372] = 5103,
  [5373] = 5107,
  [5374] = 5109,
  [5375] = 5112,
  [5376] = 5116,
  [5377] = 5117,
  [5378] = 5118,
  [5379] = 5119,
  [5380] = 5134,
  [5381] = 5087,
  [5382] = 5088,
  [5383] = 5099,
  [5384] = 1346,
  [5385] = 5100,
  [5386] = 5103,
  [5387] = 5149,
  [5388] = 5151,
  [5389] = 5121,
  [5390] = 5154,
  [5391] = 5159,
  [5392] = 5160,
  [5393] = 5124,
  [5394] = 5107,
  [5395] = 5109,
  [5396] = 5112,
  [5397] = 5134,
  [5398] = 4770,
  [5399] = 5116,
  [5400] = 5117,
  [5401] = 5118,
  [5402] = 5119,
  [5403] = 5090,
  [5404] = 5091,
  [5405] = 5091,
  [5406] = 5092,
  [5407] = 5093,
  [5408] = 5094,
  [5409] = 5095,
  [5410] = 5096,
  [5411] = 5097,
  [5412] = 5098,
  [5413] = 5092,
  [5414] = 5093,
  [5415] = 5094,
  [5416] = 5111,
  [5417] = 5113,
  [5418] = 5114,
  [5419] = 5419,
  [5420] = 5095,
  [5421] = 5096,
  [5422] = 5097,
  [5423] = 5098,
  [5424] = 5093,
  [5425] = 5159,
  [5426] = 5134,
  [5427] = 5087,
  [5428] = 5088,
  [5429] = 5099,
  [5430] = 5100,
  [5431] = 5103,
  [5432] = 5107,
  [5433] = 5109,
  [5434] = 5112,
  [5435] = 5116,
  [5436] = 5117,
  [5437] = 5118,
  [5438] = 5149,
  [5439] = 5151,
  [5440] = 5121,
  [5441] = 5154,
  [5442] = 5159,
  [5443] = 5160,
  [5444] = 5094,
  [5445] = 5445,
  [5446] = 5089,
  [5447] = 5087,
  [5448] = 5149,
  [5449] = 5091,
  [5450] = 5092,
  [5451] = 5093,
  [5452] = 5094,
  [5453] = 5095,
  [5454] = 5096,
  [5455] = 5097,
  [5456] = 5098,
  [5457] = 5151,
  [5458] = 5105,
  [5459] = 5149,
  [5460] = 5088,
  [5461] = 5111,
  [5462] = 5113,
  [5463] = 5114,
  [5464] = 5086,
  [5465] = 5108,
  [5466] = 5121,
  [5467] = 5110,
  [5468] = 5111,
  [5469] = 5154,
  [5470] = 5113,
  [5471] = 5134,
  [5472] = 5087,
  [5473] = 5088,
  [5474] = 5099,
  [5475] = 5100,
  [5476] = 5103,
  [5477] = 5107,
  [5478] = 5109,
  [5479] = 5112,
  [5480] = 5116,
  [5481] = 5117,
  [5482] = 5118,
  [5483] = 5149,
  [5484] = 5151,
  [5485] = 5121,
  [5486] = 5154,
  [5487] = 5159,
  [5488] = 5160,
  [5489] = 5114,
  [5490] = 5086,
  [5491] = 5159,
  [5492] = 5160,
  [5493] = 5123,
  [5494] = 5151,
  [5495] = 5091,
  [5496] = 5092,
  [5497] = 5093,
  [5498] = 5094,
  [5499] = 5095,
  [5500] = 5096,
  [5501] = 5097,
  [5502] = 5098,
  [5503] = 5133,
  [5504] = 5134,
  [5505] = 5121,
  [5506] = 5087,
  [5507] = 5111,
  [5508] = 5113,
  [5509] = 5114,
  [5510] = 5086,
  [5511] = 5088,
  [5512] = 5124,
  [5513] = 5125,
  [5514] = 5099,
  [5515] = 5100,
  [5516] = 5103,
  [5517] = 5134,
  [5518] = 5087,
  [5519] = 5088,
  [5520] = 5099,
  [5521] = 5100,
  [5522] = 5103,
  [5523] = 5107,
  [5524] = 5109,
  [5525] = 5112,
  [5526] = 5116,
  [5527] = 5117,
  [5528] = 5118,
  [5529] = 5149,
  [5530] = 5151,
  [5531] = 5121,
  [5532] = 5154,
  [5533] = 5159,
  [5534] = 5160,
  [5535] = 5099,
  [5536] = 5123,
  [5537] = 5133,
  [5538] = 5134,
  [5539] = 5087,
  [5540] = 5154,
  [5541] = 5091,
  [5542] = 5092,
  [5543] = 5093,
  [5544] = 5094,
  [5545] = 5095,
  [5546] = 5096,
  [5547] = 5097,
  [5548] = 5098,
  [5549] = 5088,
  [5550] = 5099,
  [5551] = 5159,
  [5552] = 5100,
  [5553] = 5111,
  [5554] = 5113,
  [5555] = 5114,
  [5556] = 5086,
  [5557] = 5103,
  [5558] = 5107,
  [5559] = 5109,
  [5560] = 5112,
  [5561] = 5116,
  [5562] = 5117,
  [5563] = 5134,
  [5564] = 5087,
  [5565] = 5088,
  [5566] = 5099,
  [5567] = 5100,
  [5568] = 5103,
  [5569] = 5107,
  [5570] = 5109,
  [5571] = 5112,
  [5572] = 5116,
  [5573] = 5117,
  [5574] = 5118,
  [5575] = 5149,
  [5576] = 5151,
  [5577] = 5121,
  [5578] = 5578,
  [5579] = 5159,
  [5580] = 5160,
  [5581] = 5118,
  [5582] = 5119,
  [5583] = 5100,
  [5584] = 5107,
  [5585] = 5109,
  [5586] = 5160,
  [5587] = 5091,
  [5588] = 5092,
  [5589] = 5093,
  [5590] = 5094,
  [5591] = 5095,
  [5592] = 5096,
  [5593] = 5097,
  [5594] = 5098,
  [5595] = 1349,
  [5596] = 5112,
  [5597] = 5090,
  [5598] = 5111,
  [5599] = 5113,
  [5600] = 5114,
  [5601] = 5086,
  [5602] = 5091,
  [5603] = 5103,
  [5604] = 5149,
  [5605] = 5151,
  [5606] = 5121,
  [5607] = 5154,
  [5608] = 5134,
  [5609] = 5087,
  [5610] = 5088,
  [5611] = 5099,
  [5612] = 5100,
  [5613] = 5103,
  [5614] = 5107,
  [5615] = 5109,
  [5616] = 5112,
  [5617] = 5116,
  [5618] = 5117,
  [5619] = 5118,
  [5620] = 5149,
  [5621] = 5151,
  [5622] = 5121,
  [5623] = 5154,
  [5624] = 5159,
  [5625] = 5160,
  [5626] = 5159,
  [5627] = 5160,
  [5628] = 5107,
  [5629] = 5092,
  [5630] = 5093,
  [5631] = 5091,
  [5632] = 5092,
  [5633] = 5093,
  [5634] = 5094,
  [5635] = 5095,
  [5636] = 5096,
  [5637] = 5097,
  [5638] = 5098,
  [5639] = 5094,
  [5640] = 5095,
  [5641] = 5096,
  [5642] = 5111,
  [5643] = 5113,
  [5644] = 5114,
  [5645] = 5086,
  [5646] = 5097,
  [5647] = 5098,
  [5648] = 5116,
  [5649] = 5117,
  [5650] = 5089,
  [5651] = 5118,
  [5652] = 5134,
  [5653] = 5087,
  [5654] = 5088,
  [5655] = 5099,
  [5656] = 5100,
  [5657] = 5103,
  [5658] = 5107,
  [5659] = 5109,
  [5660] = 5112,
  [5661] = 5116,
  [5662] = 5117,
  [5663] = 5118,
  [5664] = 5149,
  [5665] = 5151,
  [5666] = 5121,
  [5667] = 5154,
  [5668] = 5159,
  [5669] = 5160,
  [5670] = 5109,
  [5671] = 5119,
  [5672] = 5095,
  [5673] = 5096,
  [5674] = 4787,
  [5675] = 5091,
  [5676] = 5092,
  [5677] = 5093,
  [5678] = 5094,
  [5679] = 5095,
  [5680] = 5096,
  [5681] = 5097,
  [5682] = 5098,
  [5683] = 5111,
  [5684] = 5113,
  [5685] = 5114,
  [5686] = 5086,
  [5687] = 5112,
  [5688] = 5090,
  [5689] = 5091,
  [5690] = 5092,
  [5691] = 5093,
  [5692] = 5091,
  [5693] = 5092,
  [5694] = 5093,
  [5695] = 5094,
  [5696] = 5095,
  [5697] = 5096,
  [5698] = 5097,
  [5699] = 5098,
  [5700] = 5111,
  [5701] = 5113,
  [5702] = 5114,
  [5703] = 5086,
  [5704] = 5094,
  [5705] = 5095,
  [5706] = 5096,
  [5707] = 5097,
  [5708] = 5091,
  [5709] = 5092,
  [5710] = 5093,
  [5711] = 5094,
  [5712] = 5095,
  [5713] = 5096,
  [5714] = 5097,
  [5715] = 5098,
  [5716] = 5111,
  [5717] = 5113,
  [5718] = 5114,
  [5719] = 5086,
  [5720] = 5098,
  [5721] = 5105,
  [5722] = 5160,
  [5723] = 5089,
  [5724] = 5091,
  [5725] = 5092,
  [5726] = 5093,
  [5727] = 5094,
  [5728] = 5095,
  [5729] = 5096,
  [5730] = 5097,
  [5731] = 5098,
  [5732] = 5111,
  [5733] = 5113,
  [5734] = 5114,
  [5735] = 5086,
  [5736] = 5116,
  [5737] = 5117,
  [5738] = 5090,
  [5739] = 5105,
  [5740] = 5091,
  [5741] = 5092,
  [5742] = 5093,
  [5743] = 5094,
  [5744] = 5095,
  [5745] = 5096,
  [5746] = 5097,
  [5747] = 5098,
  [5748] = 5111,
  [5749] = 5113,
  [5750] = 5114,
  [5751] = 5086,
  [5752] = 5108,
  [5753] = 5108,
  [5754] = 5110,
  [5755] = 5111,
  [5756] = 5091,
  [5757] = 5092,
  [5758] = 5093,
  [5759] = 5094,
  [5760] = 5095,
  [5761] = 5096,
  [5762] = 5097,
  [5763] = 5098,
  [5764] = 5111,
  [5765] = 5113,
  [5766] = 5114,
  [5767] = 5086,
  [5768] = 5097,
  [5769] = 5113,
  [5770] = 5114,
  [5771] = 5086,
  [5772] = 5091,
  [5773] = 5092,
  [5774] = 5093,
  [5775] = 5094,
  [5776] = 5095,
  [5777] = 5096,
  [5778] = 5097,
  [5779] = 5098,
  [5780] = 5111,
  [5781] = 5113,
  [5782] = 5114,
  [5783] = 5086,
  [5784] = 5110,
  [5785] = 5111,
  [5786] = 5098,
  [5787] = 5113,
  [5788] = 5091,
  [5789] = 5092,
  [5790] = 5093,
  [5791] = 5094,
  [5792] = 5095,
  [5793] = 5096,
  [5794] = 5097,
  [5795] = 5098,
  [5796] = 5111,
  [5797] = 5113,
  [5798] = 5114,
  [5799] = 5086,
  [5800] = 5114,
  [5801] = 5086,
  [5802] = 5090,
  [5803] = 5124,
  [5804] = 5091,
  [5805] = 5092,
  [5806] = 5093,
  [5807] = 5094,
  [5808] = 5095,
  [5809] = 5096,
  [5810] = 5097,
  [5811] = 5098,
  [5812] = 5111,
  [5813] = 5113,
  [5814] = 5114,
  [5815] = 5086,
  [5816] = 5125,
  [5817] = 5817,
  [5818] = 5091,
  [5819] = 5118,
  [5820] = 5091,
  [5821] = 5092,
  [5822] = 5093,
  [5823] = 5094,
  [5824] = 5095,
  [5825] = 5096,
  [5826] = 5097,
  [5827] = 5098,
  [5828] = 5111,
  [5829] = 5113,
  [5830] = 5114,
  [5831] = 5086,
  [5832] = 5832,
  [5833] = 5149,
  [5834] = 5092,
  [5835] = 5093,
  [5836] = 5091,
  [5837] = 5092,
  [5838] = 5093,
  [5839] = 5094,
  [5840] = 5095,
  [5841] = 5096,
  [5842] = 5097,
  [5843] = 5098,
  [5844] = 5111,
  [5845] = 5113,
  [5846] = 5114,
  [5847] = 5086,
  [5848] = 5123,
  [5849] = 5133,
  [5850] = 5134,
  [5851] = 5087,
  [5852] = 5091,
  [5853] = 5092,
  [5854] = 5093,
  [5855] = 5094,
  [5856] = 5095,
  [5857] = 5096,
  [5858] = 5097,
  [5859] = 5098,
  [5860] = 5111,
  [5861] = 5113,
  [5862] = 5114,
  [5863] = 5086,
  [5864] = 5088,
  [5865] = 5099,
  [5866] = 5091,
  [5867] = 5092,
  [5868] = 5093,
  [5869] = 5094,
  [5870] = 5111,
  [5871] = 5113,
  [5872] = 5100,
  [5873] = 5103,
  [5874] = 5091,
  [5875] = 5092,
  [5876] = 5093,
  [5877] = 5094,
  [5878] = 5111,
  [5879] = 5113,
  [5880] = 5107,
  [5881] = 5109,
  [5882] = 5091,
  [5883] = 5092,
  [5884] = 5093,
  [5885] = 5094,
  [5886] = 5111,
  [5887] = 5113,
  [5888] = 5112,
  [5889] = 5116,
  [5890] = 5091,
  [5891] = 5092,
  [5892] = 5093,
  [5893] = 5094,
  [5894] = 5111,
  [5895] = 5113,
  [5896] = 5117,
  [5897] = 5118,
  [5898] = 5091,
  [5899] = 5092,
  [5900] = 5093,
  [5901] = 5094,
  [5902] = 5111,
  [5903] = 5113,
  [5904] = 5119,
  [5905] = 5094,
  [5906] = 5091,
  [5907] = 5092,
  [5908] = 5093,
  [5909] = 5094,
  [5910] = 5111,
  [5911] = 5113,
  [5912] = 5151,
  [5913] = 5121,
  [5914] = 5091,
  [5915] = 5092,
  [5916] = 5093,
  [5917] = 5094,
  [5918] = 5111,
  [5919] = 5113,
  [5920] = 5095,
  [5921] = 5154,
  [5922] = 5159,
  [5923] = 5096,
  [5924] = 5097,
  [5925] = 5098,
  [5926] = 5149,
  [5927] = 5151,
  [5928] = 5125,
  [5929] = 4773,
  [5930] = 5093,
  [5931] = 5931,
  [5932] = 5121,
  [5933] = 5154,
  [5934] = 5159,
  [5935] = 5160,
  [5936] = 5124,
  [5937] = 5125,
  [5938] = 5094,
  [5939] = 5105,
  [5940] = 5095,
  [5941] = 5160,
  [5942] = 5114,
  [5943] = 5108,
  [5944] = 5096,
  [5945] = 5151,
  [5946] = 1350,
  [5947] = 5097,
  [5948] = 5089,
  [5949] = 5110,
  [5950] = 5119,
  [5951] = 5951,
  [5952] = 5098,
  [5953] = 5111,
  [5954] = 5954,
  [5955] = 5123,
  [5956] = 5133,
  [5957] = 5957,
  [5958] = 5134,
  [5959] = 5113,
  [5960] = 5154,
  [5961] = 5961,
  [5962] = 5962,
  [5963] = 5963,
  [5964] = 5964,
  [5965] = 5965,
  [5966] = 5966,
  [5967] = 5963,
  [5968] = 5965,
  [5969] = 5966,
  [5970] = 5970,
  [5971] = 5971,
  [5972] = 5972,
  [5973] = 5973,
  [5974] = 5974,
  [5975] = 5975,
  [5976] = 5976,
  [5977] = 5970,
  [5978] = 5971,
  [5979] = 5972,
  [5980] = 5973,
  [5981] = 5974,
  [5982] = 5975,
  [5983] = 5976,
  [5984] = 5984,
  [5985] = 5985,
  [5986] = 5962,
  [5987] = 5962,
  [5988] = 5961,
  [5989] = 5961,
  [5990] = 5990,
  [5991] = 5961,
  [5992] = 5992,
  [5993] = 5963,
  [5994] = 5965,
  [5995] = 5966,
  [5996] = 5970,
  [5997] = 5971,
  [5998] = 5972,
  [5999] = 5973,
  [6000] = 5974,
  [6001] = 5975,
  [6002] = 5976,
  [6003] = 5963,
  [6004] = 5965,
  [6005] = 5966,
  [6006] = 5970,
  [6007] = 5971,
  [6008] = 5972,
  [6009] = 5973,
  [6010] = 5974,
  [6011] = 5975,
  [6012] = 5976,
  [6013] = 6013,
  [6014] = 5962,
  [6015] = 5963,
  [6016] = 5965,
  [6017] = 5966,
  [6018] = 5970,
  [6019] = 5971,
  [6020] = 5972,
  [6021] = 5973,
  [6022] = 5974,
  [6023] = 5975,
  [6024] = 5976,
  [6025] = 5963,
  [6026] = 5965,
  [6027] = 5966,
  [6028] = 5963,
  [6029] = 5965,
  [6030] = 5966,
  [6031] = 5970,
  [6032] = 5971,
  [6033] = 5972,
  [6034] = 5973,
  [6035] = 5974,
  [6036] = 5975,
  [6037] = 5976,
  [6038] = 5970,
  [6039] = 5984,
  [6040] = 5971,
  [6041] = 5962,
  [6042] = 5972,
  [6043] = 5973,
  [6044] = 5962,
  [6045] = 5974,
  [6046] = 5975,
  [6047] = 5976,
  [6048] = 5963,
  [6049] = 5961,
  [6050] = 5963,
  [6051] = 5965,
  [6052] = 5966,
  [6053] = 5970,
  [6054] = 5971,
  [6055] = 5972,
  [6056] = 5973,
  [6057] = 5974,
  [6058] = 5975,
  [6059] = 5976,
  [6060] = 5965,
  [6061] = 6061,
  [6062] = 5961,
  [6063] = 5963,
  [6064] = 5965,
  [6065] = 5966,
  [6066] = 5970,
  [6067] = 5971,
  [6068] = 5972,
  [6069] = 5973,
  [6070] = 5974,
  [6071] = 5975,
  [6072] = 5976,
  [6073] = 5963,
  [6074] = 5965,
  [6075] = 5966,
  [6076] = 5970,
  [6077] = 5971,
  [6078] = 5972,
  [6079] = 5973,
  [6080] = 5974,
  [6081] = 5975,
  [6082] = 5976,
  [6083] = 5963,
  [6084] = 5965,
  [6085] = 5984,
  [6086] = 5970,
  [6087] = 5971,
  [6088] = 5972,
  [6089] = 5973,
  [6090] = 5974,
  [6091] = 5975,
  [6092] = 5976,
  [6093] = 5963,
  [6094] = 5965,
  [6095] = 5966,
  [6096] = 5970,
  [6097] = 5971,
  [6098] = 5972,
  [6099] = 5973,
  [6100] = 5974,
  [6101] = 5975,
  [6102] = 5976,
  [6103] = 5962,
  [6104] = 5963,
  [6105] = 5965,
  [6106] = 5966,
  [6107] = 5970,
  [6108] = 5971,
  [6109] = 5972,
  [6110] = 5973,
  [6111] = 5974,
  [6112] = 5975,
  [6113] = 5976,
  [6114] = 6114,
  [6115] = 5963,
  [6116] = 5966,
  [6117] = 5965,
  [6118] = 5961,
  [6119] = 5966,
  [6120] = 5970,
  [6121] = 5971,
  [6122] = 5972,
  [6123] = 5963,
  [6124] = 5965,
  [6125] = 5966,
  [6126] = 5970,
  [6127] = 5971,
  [6128] = 5972,
  [6129] = 5973,
  [6130] = 5974,
  [6131] = 5975,
  [6132] = 5976,
  [6133] = 5962,
  [6134] = 5970,
  [6135] = 5971,
  [6136] = 5973,
  [6137] = 5972,
  [6138] = 5973,
  [6139] = 5974,
  [6140] = 5974,
  [6141] = 5962,
  [6142] = 5975,
  [6143] = 5975,
  [6144] = 5976,
  [6145] = 5962,
  [6146] = 6146,
  [6147] = 5962,
  [6148] = 5984,
  [6149] = 5984,
  [6150] = 5963,
  [6151] = 5965,
  [6152] = 5963,
  [6153] = 5965,
  [6154] = 5966,
  [6155] = 5970,
  [6156] = 5971,
  [6157] = 5972,
  [6158] = 5973,
  [6159] = 5974,
  [6160] = 5975,
  [6161] = 5976,
  [6162] = 5966,
  [6163] = 5970,
  [6164] = 5971,
  [6165] = 5972,
  [6166] = 5973,
  [6167] = 5974,
  [6168] = 5975,
  [6169] = 5976,
  [6170] = 5976,
  [6171] = 5961,
  [6172] = 6172,
  [6173] = 5961,
  [6174] = 5984,
  [6175] = 6175,
  [6176] = 5962,
  [6177] = 6177,
  [6178] = 6178,
  [6179] = 5984,
  [6180] = 6180,
  [6181] = 5984,
  [6182] = 5984,
  [6183] = 5961,
  [6184] = 5984,
  [6185] = 5966,
  [6186] = 6186,
  [6187] = 6187,
  [6188] = 6188,
  [6189] = 6189,
  [6190] = 6190,
  [6191] = 6191,
  [6192] = 6192,
  [6193] = 6193,
  [6194] = 6186,
  [6195] = 6195,
  [6196] = 6196,
  [6197] = 6197,
  [6198] = 6198,
  [6199] = 6199,
  [6200] = 6200,
  [6201] = 6201,
  [6202] = 6202,
  [6203] = 6203,
  [6204] = 6204,
  [6205] = 6205,
  [6206] = 6192,
  [6207] = 6207,
  [6208] = 6208,
  [6209] = 6209,
  [6210] = 6210,
  [6211] = 6211,
  [6212] = 6203,
  [6213] = 6199,
  [6214] = 6214,
  [6215] = 6215,
  [6216] = 6214,
  [6217] = 6215,
  [6218] = 6189,
  [6219] = 6190,
  [6220] = 6202,
  [6221] = 6203,
  [6222] = 6192,
  [6223] = 6207,
  [6224] = 6208,
  [6225] = 6209,
  [6226] = 6214,
  [6227] = 6215,
  [6228] = 6189,
  [6229] = 6190,
  [6230] = 6202,
  [6231] = 6203,
  [6232] = 6192,
  [6233] = 6207,
  [6234] = 6208,
  [6235] = 6209,
  [6236] = 6210,
  [6237] = 6211,
  [6238] = 6191,
  [6239] = 6193,
  [6240] = 6186,
  [6241] = 6195,
  [6242] = 6196,
  [6243] = 6197,
  [6244] = 6198,
  [6245] = 6199,
  [6246] = 6210,
  [6247] = 6211,
  [6248] = 6191,
  [6249] = 6193,
  [6250] = 6186,
  [6251] = 6195,
  [6252] = 6252,
  [6253] = 6253,
  [6254] = 6254,
  [6255] = 6255,
  [6256] = 6200,
  [6257] = 6257,
  [6258] = 6258,
  [6259] = 6259,
  [6260] = 6187,
  [6261] = 6188,
  [6262] = 6196,
  [6263] = 6197,
  [6264] = 6198,
  [6265] = 6199,
  [6266] = 6189,
  [6267] = 6252,
  [6268] = 6253,
  [6269] = 6254,
  [6270] = 6255,
  [6271] = 6200,
  [6272] = 6257,
  [6273] = 6258,
  [6274] = 6259,
  [6275] = 6187,
  [6276] = 6188,
  [6277] = 6190,
  [6278] = 6202,
  [6279] = 6192,
  [6280] = 6207,
  [6281] = 6252,
  [6282] = 6253,
  [6283] = 6254,
  [6284] = 6255,
  [6285] = 6200,
  [6286] = 6257,
  [6287] = 6258,
  [6288] = 6208,
  [6289] = 6209,
  [6290] = 6190,
  [6291] = 6210,
  [6292] = 6211,
  [6293] = 6252,
  [6294] = 6253,
  [6295] = 6254,
  [6296] = 6255,
  [6297] = 6259,
  [6298] = 6187,
  [6299] = 6188,
  [6300] = 6200,
  [6301] = 6257,
  [6302] = 6258,
  [6303] = 6259,
  [6304] = 6187,
  [6305] = 6188,
  [6306] = 6214,
  [6307] = 6215,
  [6308] = 6189,
  [6309] = 6190,
  [6310] = 6202,
  [6311] = 6203,
  [6312] = 6192,
  [6313] = 6207,
  [6314] = 6208,
  [6315] = 6209,
  [6316] = 6210,
  [6317] = 6211,
  [6318] = 6191,
  [6319] = 6193,
  [6320] = 6186,
  [6321] = 6195,
  [6322] = 6196,
  [6323] = 6197,
  [6324] = 6198,
  [6325] = 6199,
  [6326] = 6191,
  [6327] = 6193,
  [6328] = 6252,
  [6329] = 6253,
  [6330] = 6254,
  [6331] = 6255,
  [6332] = 6200,
  [6333] = 6257,
  [6334] = 6258,
  [6335] = 6259,
  [6336] = 6187,
  [6337] = 6188,
  [6338] = 6214,
  [6339] = 6339,
  [6340] = 6203,
  [6341] = 6192,
  [6342] = 6207,
  [6343] = 6208,
  [6344] = 6257,
  [6345] = 6209,
  [6346] = 6210,
  [6347] = 6215,
  [6348] = 6189,
  [6349] = 6190,
  [6350] = 6202,
  [6351] = 6203,
  [6352] = 6192,
  [6353] = 6207,
  [6354] = 6208,
  [6355] = 6209,
  [6356] = 6210,
  [6357] = 6211,
  [6358] = 6191,
  [6359] = 6214,
  [6360] = 6215,
  [6361] = 6189,
  [6362] = 6190,
  [6363] = 6202,
  [6364] = 6203,
  [6365] = 6192,
  [6366] = 6207,
  [6367] = 6208,
  [6368] = 6209,
  [6369] = 6210,
  [6370] = 6211,
  [6371] = 6191,
  [6372] = 6193,
  [6373] = 6186,
  [6374] = 6195,
  [6375] = 6196,
  [6376] = 6197,
  [6377] = 6198,
  [6378] = 6199,
  [6379] = 6193,
  [6380] = 6252,
  [6381] = 6253,
  [6382] = 6254,
  [6383] = 6255,
  [6384] = 6200,
  [6385] = 6257,
  [6386] = 6258,
  [6387] = 6259,
  [6388] = 6187,
  [6389] = 6188,
  [6390] = 6186,
  [6391] = 6195,
  [6392] = 6196,
  [6393] = 6197,
  [6394] = 6198,
  [6395] = 6199,
  [6396] = 6396,
  [6397] = 6211,
  [6398] = 6191,
  [6399] = 6193,
  [6400] = 6252,
  [6401] = 6186,
  [6402] = 1350,
  [6403] = 6195,
  [6404] = 6196,
  [6405] = 6197,
  [6406] = 6198,
  [6407] = 6199,
  [6408] = 6202,
  [6409] = 6409,
  [6410] = 6410,
  [6411] = 6186,
  [6412] = 6195,
  [6413] = 6252,
  [6414] = 6253,
  [6415] = 6254,
  [6416] = 6255,
  [6417] = 6200,
  [6418] = 6257,
  [6419] = 6258,
  [6420] = 6259,
  [6421] = 6214,
  [6422] = 6214,
  [6423] = 6215,
  [6424] = 6189,
  [6425] = 6190,
  [6426] = 6202,
  [6427] = 6203,
  [6428] = 6192,
  [6429] = 6207,
  [6430] = 6208,
  [6431] = 6209,
  [6432] = 6210,
  [6433] = 6211,
  [6434] = 6191,
  [6435] = 6193,
  [6436] = 6186,
  [6437] = 6195,
  [6438] = 6196,
  [6439] = 6197,
  [6440] = 6198,
  [6441] = 6199,
  [6442] = 6215,
  [6443] = 6189,
  [6444] = 6190,
  [6445] = 6202,
  [6446] = 6203,
  [6447] = 6192,
  [6448] = 6252,
  [6449] = 6253,
  [6450] = 6254,
  [6451] = 6255,
  [6452] = 6200,
  [6453] = 6257,
  [6454] = 6258,
  [6455] = 6259,
  [6456] = 6187,
  [6457] = 6188,
  [6458] = 6207,
  [6459] = 6208,
  [6460] = 6209,
  [6461] = 6210,
  [6462] = 6211,
  [6463] = 6253,
  [6464] = 6191,
  [6465] = 6193,
  [6466] = 6186,
  [6467] = 6195,
  [6468] = 6196,
  [6469] = 6197,
  [6470] = 6198,
  [6471] = 6199,
  [6472] = 6254,
  [6473] = 6214,
  [6474] = 6215,
  [6475] = 6189,
  [6476] = 6190,
  [6477] = 6202,
  [6478] = 6203,
  [6479] = 6013,
  [6480] = 6192,
  [6481] = 6207,
  [6482] = 6187,
  [6483] = 6188,
  [6484] = 6208,
  [6485] = 6209,
  [6486] = 6210,
  [6487] = 6252,
  [6488] = 6253,
  [6489] = 6254,
  [6490] = 6255,
  [6491] = 6200,
  [6492] = 6257,
  [6493] = 6258,
  [6494] = 6259,
  [6495] = 6214,
  [6496] = 6215,
  [6497] = 6189,
  [6498] = 6190,
  [6499] = 6202,
  [6500] = 6203,
  [6501] = 6192,
  [6502] = 6214,
  [6503] = 6208,
  [6504] = 6209,
  [6505] = 6210,
  [6506] = 6211,
  [6507] = 6191,
  [6508] = 6193,
  [6509] = 6195,
  [6510] = 6196,
  [6511] = 6197,
  [6512] = 6198,
  [6513] = 6199,
  [6514] = 6187,
  [6515] = 6188,
  [6516] = 6211,
  [6517] = 6191,
  [6518] = 6193,
  [6519] = 6186,
  [6520] = 6252,
  [6521] = 6253,
  [6522] = 6254,
  [6523] = 6255,
  [6524] = 6200,
  [6525] = 6257,
  [6526] = 6258,
  [6527] = 6259,
  [6528] = 6187,
  [6529] = 6188,
  [6530] = 6195,
  [6531] = 6196,
  [6532] = 6197,
  [6533] = 6198,
  [6534] = 6199,
  [6535] = 6255,
  [6536] = 6214,
  [6537] = 6215,
  [6538] = 6189,
  [6539] = 6539,
  [6540] = 6540,
  [6541] = 6190,
  [6542] = 6202,
  [6543] = 6203,
  [6544] = 6192,
  [6545] = 6207,
  [6546] = 6208,
  [6547] = 6209,
  [6548] = 6210,
  [6549] = 6211,
  [6550] = 6191,
  [6551] = 6193,
  [6552] = 6186,
  [6553] = 6214,
  [6554] = 6195,
  [6555] = 6196,
  [6556] = 6197,
  [6557] = 6198,
  [6558] = 6199,
  [6559] = 6191,
  [6560] = 6193,
  [6561] = 6258,
  [6562] = 6203,
  [6563] = 6563,
  [6564] = 6564,
  [6565] = 6196,
  [6566] = 6214,
  [6567] = 6215,
  [6568] = 6186,
  [6569] = 6195,
  [6570] = 6570,
  [6571] = 6196,
  [6572] = 6197,
  [6573] = 6573,
  [6574] = 6215,
  [6575] = 6197,
  [6576] = 6252,
  [6577] = 6253,
  [6578] = 6254,
  [6579] = 6255,
  [6580] = 6200,
  [6581] = 6257,
  [6582] = 6258,
  [6583] = 6259,
  [6584] = 6187,
  [6585] = 6188,
  [6586] = 6198,
  [6587] = 6198,
  [6588] = 6199,
  [6589] = 6252,
  [6590] = 6590,
  [6591] = 6253,
  [6592] = 6214,
  [6593] = 6215,
  [6594] = 6189,
  [6595] = 6190,
  [6596] = 6202,
  [6597] = 6203,
  [6598] = 6192,
  [6599] = 6207,
  [6600] = 6208,
  [6601] = 6209,
  [6602] = 6210,
  [6603] = 6211,
  [6604] = 6191,
  [6605] = 6193,
  [6606] = 6186,
  [6607] = 6195,
  [6608] = 6196,
  [6609] = 6197,
  [6610] = 6198,
  [6611] = 6199,
  [6612] = 6189,
  [6613] = 6254,
  [6614] = 6190,
  [6615] = 6189,
  [6616] = 6255,
  [6617] = 6200,
  [6618] = 6257,
  [6619] = 6258,
  [6620] = 6259,
  [6621] = 6187,
  [6622] = 6188,
  [6623] = 6259,
  [6624] = 6252,
  [6625] = 6253,
  [6626] = 6254,
  [6627] = 6255,
  [6628] = 6200,
  [6629] = 6257,
  [6630] = 6258,
  [6631] = 6259,
  [6632] = 6187,
  [6633] = 6188,
  [6634] = 6202,
  [6635] = 6203,
  [6636] = 6192,
  [6637] = 6252,
  [6638] = 6253,
  [6639] = 6207,
  [6640] = 6208,
  [6641] = 6254,
  [6642] = 6642,
  [6643] = 6255,
  [6644] = 6209,
  [6645] = 6645,
  [6646] = 6200,
  [6647] = 6210,
  [6648] = 6257,
  [6649] = 6258,
  [6650] = 6259,
  [6651] = 6187,
  [6652] = 6652,
  [6653] = 6187,
  [6654] = 6188,
  [6655] = 6211,
  [6656] = 6656,
  [6657] = 6188,
  [6658] = 6658,
  [6659] = 6659,
  [6660] = 6660,
  [6661] = 6661,
  [6662] = 6662,
  [6663] = 6663,
  [6664] = 6252,
  [6665] = 6665,
  [6666] = 6214,
  [6667] = 6215,
  [6668] = 6189,
  [6669] = 6190,
  [6670] = 6202,
  [6671] = 6203,
  [6672] = 6192,
  [6673] = 6207,
  [6674] = 6208,
  [6675] = 6209,
  [6676] = 6210,
  [6677] = 6211,
  [6678] = 6191,
  [6679] = 6193,
  [6680] = 6186,
  [6681] = 6195,
  [6682] = 6196,
  [6683] = 6197,
  [6684] = 6198,
  [6685] = 6199,
  [6686] = 6686,
  [6687] = 6214,
  [6688] = 6215,
  [6689] = 6189,
  [6690] = 6190,
  [6691] = 6202,
  [6692] = 6203,
  [6693] = 6192,
  [6694] = 6207,
  [6695] = 6208,
  [6696] = 6209,
  [6697] = 6210,
  [6698] = 6211,
  [6699] = 6200,
  [6700] = 1349,
  [6701] = 6257,
  [6702] = 6191,
  [6703] = 6193,
  [6704] = 1346,
  [6705] = 6186,
  [6706] = 6258,
  [6707] = 6259,
  [6708] = 6187,
  [6709] = 6188,
  [6710] = 6710,
  [6711] = 6195,
  [6712] = 6196,
  [6713] = 6197,
  [6714] = 6198,
  [6715] = 6715,
  [6716] = 6199,
  [6717] = 6207,
  [6718] = 6208,
  [6719] = 6719,
  [6720] = 6209,
  [6721] = 6210,
  [6722] = 6722,
  [6723] = 6211,
  [6724] = 6191,
  [6725] = 6725,
  [6726] = 6193,
  [6727] = 6186,
  [6728] = 6255,
  [6729] = 6729,
  [6730] = 1351,
  [6731] = 6253,
  [6732] = 6214,
  [6733] = 6195,
  [6734] = 6196,
  [6735] = 6215,
  [6736] = 6736,
  [6737] = 6197,
  [6738] = 6198,
  [6739] = 6739,
  [6740] = 6199,
  [6741] = 6202,
  [6742] = 6215,
  [6743] = 6252,
  [6744] = 6253,
  [6745] = 6252,
  [6746] = 6253,
  [6747] = 6254,
  [6748] = 6255,
  [6749] = 6200,
  [6750] = 6257,
  [6751] = 6254,
  [6752] = 6258,
  [6753] = 6259,
  [6754] = 6187,
  [6755] = 5578,
  [6756] = 5191,
  [6757] = 6188,
  [6758] = 6255,
  [6759] = 6200,
  [6760] = 6189,
  [6761] = 6761,
  [6762] = 6762,
  [6763] = 6257,
  [6764] = 6764,
  [6765] = 6254,
  [6766] = 6190,
  [6767] = 6767,
  [6768] = 6258,
  [6769] = 6259,
  [6770] = 6207,
  [6771] = 6771,
  [6772] = 6772,
  [6773] = 6773,
  [6774] = 6774,
  [6775] = 6775,
  [6776] = 6776,
  [6777] = 6777,
  [6778] = 6778,
  [6779] = 6777,
  [6780] = 6780,
  [6781] = 6781,
  [6782] = 6782,
  [6783] = 6783,
  [6784] = 6774,
  [6785] = 6783,
  [6786] = 6777,
  [6787] = 6777,
  [6788] = 6788,
  [6789] = 6780,
  [6790] = 6773,
  [6791] = 6777,
  [6792] = 6777,
  [6793] = 6773,
  [6794] = 6794,
  [6795] = 6780,
  [6796] = 6783,
  [6797] = 6771,
  [6798] = 6798,
  [6799] = 6799,
  [6800] = 6800,
  [6801] = 6794,
  [6802] = 6802,
  [6803] = 6803,
  [6804] = 6804,
  [6805] = 6773,
  [6806] = 6783,
  [6807] = 6807,
  [6808] = 6774,
  [6809] = 6803,
  [6810] = 6804,
  [6811] = 6811,
  [6812] = 6802,
  [6813] = 6771,
  [6814] = 6777,
  [6815] = 6815,
  [6816] = 6777,
  [6817] = 6777,
  [6818] = 6803,
  [6819] = 6771,
  [6820] = 6820,
  [6821] = 6794,
  [6822] = 6771,
  [6823] = 6794,
  [6824] = 6774,
  [6825] = 6825,
  [6826] = 6773,
  [6827] = 6780,
  [6828] = 6828,
  [6829] = 6803,
  [6830] = 6780,
  [6831] = 6804,
  [6832] = 6802,
  [6833] = 6774,
  [6834] = 6783,
  [6835] = 6835,
  [6836] = 6836,
  [6837] = 6802,
  [6838] = 6803,
  [6839] = 6802,
  [6840] = 6803,
  [6841] = 6804,
  [6842] = 6802,
  [6843] = 6802,
  [6844] = 6780,
  [6845] = 6794,
  [6846] = 6802,
  [6847] = 6802,
  [6848] = 6802,
  [6849] = 6849,
  [6850] = 6850,
  [6851] = 6802,
  [6852] = 6802,
  [6853] = 6777,
  [6854] = 6802,
  [6855] = 6802,
  [6856] = 6802,
  [6857] = 6802,
  [6858] = 6802,
  [6859] = 6802,
  [6860] = 6804,
  [6861] = 6774,
  [6862] = 6804,
  [6863] = 6773,
  [6864] = 6771,
  [6865] = 6777,
  [6866] = 6777,
  [6867] = 6783,
  [6868] = 6794,
  [6869] = 6869,
  [6870] = 6870,
  [6871] = 6871,
  [6872] = 6872,
  [6873] = 6873,
  [6874] = 6874,
  [6875] = 6875,
  [6876] = 6873,
  [6877] = 6877,
  [6878] = 6878,
  [6879] = 6879,
  [6880] = 6875,
  [6881] = 6873,
  [6882] = 6882,
  [6883] = 6875,
  [6884] = 6875,
  [6885] = 6885,
  [6886] = 6886,
  [6887] = 6887,
  [6888] = 6885,
  [6889] = 6871,
  [6890] = 6879,
  [6891] = 6891,
  [6892] = 6882,
  [6893] = 6885,
  [6894] = 6870,
  [6895] = 6873,
  [6896] = 6875,
  [6897] = 6873,
  [6898] = 6898,
  [6899] = 6885,
  [6900] = 6900,
  [6901] = 6875,
  [6902] = 6870,
  [6903] = 6875,
  [6904] = 6879,
  [6905] = 6870,
  [6906] = 6900,
  [6907] = 6900,
  [6908] = 6871,
  [6909] = 6874,
  [6910] = 6871,
  [6911] = 6882,
  [6912] = 6900,
  [6913] = 6871,
  [6914] = 6879,
  [6915] = 6882,
  [6916] = 6882,
  [6917] = 6872,
  [6918] = 6874,
  [6919] = 926,
  [6920] = 6885,
  [6921] = 6871,
  [6922] = 6900,
  [6923] = 6879,
  [6924] = 910,
  [6925] = 6882,
  [6926] = 6885,
  [6927] = 6872,
  [6928] = 6885,
  [6929] = 6873,
  [6930] = 6900,
  [6931] = 6882,
  [6932] = 6875,
  [6933] = 6870,
  [6934] = 6872,
  [6935] = 6874,
  [6936] = 6879,
  [6937] = 6873,
  [6938] = 6879,
  [6939] = 6872,
  [6940] = 6874,
  [6941] = 6871,
  [6942] = 6870,
  [6943] = 6871,
  [6944] = 6872,
  [6945] = 6874,
  [6946] = 6885,
  [6947] = 6870,
  [6948] = 6882,
  [6949] = 6872,
  [6950] = 6874,
  [6951] = 6951,
  [6952] = 6875,
  [6953] = 6873,
  [6954] = 6872,
  [6955] = 6874,
  [6956] = 6885,
  [6957] = 6957,
  [6958] = 6879,
  [6959] = 6872,
  [6960] = 6874,
  [6961] = 912,
  [6962] = 6870,
  [6963] = 6870,
  [6964] = 6872,
  [6965] = 6874,
  [6966] = 6885,
  [6967] = 6871,
  [6968] = 911,
  [6969] = 6872,
  [6970] = 6874,
  [6971] = 6885,
  [6972] = 6882,
  [6973] = 919,
  [6974] = 6872,
  [6975] = 6874,
  [6976] = 6873,
  [6977] = 6871,
  [6978] = 6870,
  [6979] = 925,
  [6980] = 6882,
  [6981] = 6879,
  [6982] = 6879,
  [6983] = 6879,
  [6984] = 6871,
  [6985] = 6875,
  [6986] = 6882,
  [6987] = 6879,
  [6988] = 6873,
  [6989] = 6887,
  [6990] = 6875,
  [6991] = 924,
  [6992] = 6882,
  [6993] = 6873,
  [6994] = 6885,
  [6995] = 6870,
  [6996] = 6875,
  [6997] = 6887,
  [6998] = 6873,
  [6999] = 6870,
  [7000] = 6871,
};

static const TSCharacterRange aux_sym_identifier_token1_character_set_1[] = {
  {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xba, 0xba}, {0xdf, 0xf6}, {0xf8, 0xff}, {0x101, 0x101},
  {0x103, 0x103}, {0x105, 0x105}, {0x107, 0x107}, {0x109, 0x109}, {0x10b, 0x10b}, {0x10d, 0x10d}, {0x10f, 0x10f}, {0x111, 0x111},
  {0x113, 0x113}, {0x115, 0x115}, {0x117, 0x117}, {0x119, 0x119}, {0x11b, 0x11b}, {0x11d, 0x11d}, {0x11f, 0x11f}, {0x121, 0x121},
  {0x123, 0x123}, {0x125, 0x125}, {0x127, 0x127}, {0x129, 0x129}, {0x12b, 0x12b}, {0x12d, 0x12d}, {0x12f, 0x12f}, {0x131, 0x131},
  {0x133, 0x133}, {0x135, 0x135}, {0x137, 0x138}, {0x13a, 0x13a}, {0x13c, 0x13c}, {0x13e, 0x13e}, {0x140, 0x140}, {0x142, 0x142},
  {0x144, 0x144}, {0x146, 0x146}, {0x148, 0x149}, {0x14b, 0x14b}, {0x14d, 0x14d}, {0x14f, 0x14f}, {0x151, 0x151}, {0x153, 0x153},
  {0x155, 0x155}, {0x157, 0x157}, {0x159, 0x159}, {0x15b, 0x15b}, {0x15d, 0x15d}, {0x15f, 0x15f}, {0x161, 0x161}, {0x163, 0x163},
  {0x165, 0x165}, {0x167, 0x167}, {0x169, 0x169}, {0x16b, 0x16b}, {0x16d, 0x16d}, {0x16f, 0x16f}, {0x171, 0x171}, {0x173, 0x173},
  {0x175, 0x175}, {0x177, 0x177}, {0x17a, 0x17a}, {0x17c, 0x17c}, {0x17e, 0x180}, {0x183, 0x183}, {0x185, 0x185}, {0x188, 0x188},
  {0x18c, 0x18d}, {0x192, 0x192}, {0x195, 0x195}, {0x199, 0x19b}, {0x19e, 0x19e}, {0x1a1, 0x1a1}, {0x1a3, 0x1a3}, {0x1a5, 0x1a5},
  {0x1a8, 0x1a8}, {0x1aa, 0x1ab}, {0x1ad, 0x1ad}, {0x1b0, 0x1b0}, {0x1b4, 0x1b4}, {0x1b6, 0x1b6}, {0x1b9, 0x1bb}, {0x1bd, 0x1c3},
  {0x1c6, 0x1c6}, {0x1c9, 0x1c9}, {0x1cc, 0x1cc}, {0x1ce, 0x1ce}, {0x1d0, 0x1d0}, {0x1d2, 0x1d2}, {0x1d4, 0x1d4}, {0x1d6, 0x1d6},
  {0x1d8, 0x1d8}, {0x1da, 0x1da}, {0x1dc, 0x1dd}, {0x1df, 0x1df}, {0x1e1, 0x1e1}, {0x1e3, 0x1e3}, {0x1e5, 0x1e5}, {0x1e7, 0x1e7},
  {0x1e9, 0x1e9}, {0x1eb, 0x1eb}, {0x1ed, 0x1ed}, {0x1ef, 0x1f0}, {0x1f3, 0x1f3}, {0x1f5, 0x1f5}, {0x1f9, 0x1f9}, {0x1fb, 0x1fb},
  {0x1fd, 0x1fd}, {0x1ff, 0x1ff}, {0x201, 0x201}, {0x203, 0x203}, {0x205, 0x205}, {0x207, 0x207}, {0x209, 0x209}, {0x20b, 0x20b},
  {0x20d, 0x20d}, {0x20f, 0x20f}, {0x211, 0x211}, {0x213, 0x213}, {0x215, 0x215}, {0x217, 0x217}, {0x219, 0x219}, {0x21b, 0x21b},
  {0x21d, 0x21d}, {0x21f, 0x21f}, {0x221, 0x221}, {0x223, 0x223}, {0x225, 0x225}, {0x227, 0x227}, {0x229, 0x229}, {0x22b, 0x22b},
  {0x22d, 0x22d}, {0x22f, 0x22f}, {0x231, 0x231}, {0x233, 0x239}, {0x23c, 0x23c}, {0x23f, 0x240}, {0x242, 0x242}, {0x247, 0x247},
  {0x249, 0x249}, {0x24b, 0x24b}, {0x24d, 0x24d}, {0x24f, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee},
  {0x371, 0x371}, {0x373, 0x374}, {0x377, 0x377}, {0x37a, 0x37d}, {0x390, 0x390}, {0x3ac, 0x3ce}, {0x3d0, 0x3d1}, {0x3d5, 0x3d7},
  {0x3d9, 0x3d9}, {0x3db, 0x3db}, {0x3dd, 0x3dd}, {0x3df, 0x3df}, {0x3e1, 0x3e1}, {0x3e3, 0x3e3}, {0x3e5, 0x3e5}, {0x3e7, 0x3e7},
  {0x3e9, 0x3e9}, {0x3eb, 0x3eb}, {0x3ed, 0x3ed}, {0x3ef, 0x3f3}, {0x3f5, 0x3f5}, {0x3f8, 0x3f8}, {0x3fb, 0x3fc}, {0x430, 0x45f},
  {0x461, 0x461}, {0x463, 0x463}, {0x465, 0x465}, {0x467, 0x467}, {0x469, 0x469}, {0x46b, 0x46b}, {0x46d, 0x46d}, {0x46f, 0x46f},
  {0x471, 0x471}, {0x473, 0x473}, {0x475, 0x475}, {0x477, 0x477}, {0x479, 0x479}, {0x47b, 0x47b}, {0x47d, 0x47d}, {0x47f, 0x47f},
  {0x481, 0x481}, {0x48b, 0x48b}, {0x48d, 0x48d}, {0x48f, 0x48f}, {0x491, 0x491}, {0x493, 0x493}, {0x495, 0x495}, {0x497, 0x497},
  {0x499, 0x499}, {0x49b, 0x49b}, {0x49d, 0x49d}, {0x49f, 0x49f}, {0x4a1, 0x4a1}, {0x4a3, 0x4a3}, {0x4a5, 0x4a5}, {0x4a7, 0x4a7},
  {0x4a9, 0x4a9}, {0x4ab, 0x4ab}, {0x4ad, 0x4ad}, {0x4af, 0x4af}, {0x4b1, 0x4b1}, {0x4b3, 0x4b3}, {0x4b5, 0x4b5}, {0x4b7, 0x4b7},
  {0x4b9, 0x4b9}, {0x4bb, 0x4bb}, {0x4bd, 0x4bd}, {0x4bf, 0x4bf}, {0x4c2, 0x4c2}, {0x4c4, 0x4c4}, {0x4c6, 0x4c6}, {0x4c8, 0x4c8},
  {0x4ca, 0x4ca}, {0x4cc, 0x4cc}, {0x4ce, 0x4cf}, {0x4d1, 0x4d1}, {0x4d3, 0x4d3}, {0x4d5, 0x4d5}, {0x4d7, 0x4d7}, {0x4d9, 0x4d9},
  {0x4db, 0x4db}, {0x4dd, 0x4dd}, {0x4df, 0x4df}, {0x4e1, 0x4e1}, {0x4e3, 0x4e3}, {0x4e5, 0x4e5}, {0x4e7, 0x4e7}, {0x4e9, 0x4e9},
  {0x4eb, 0x4eb}, {0x4ed, 0x4ed}, {0x4ef, 0x4ef}, {0x4f1, 0x4f1}, {0x4f3, 0x4f3}, {0x4f5, 0x4f5}, {0x4f7, 0x4f7}, {0x4f9, 0x4f9},
  {0x4fb, 0x4fb}, {0x4fd, 0x4fd}, {0x4ff, 0x4ff}, {0x501, 0x501}, {0x503, 0x503}, {0x505, 0x505}, {0x507, 0x507}, {0x509, 0x509},
  {0x50b, 0x50b}, {0x50d, 0x50d}, {0x50f, 0x50f}, {0x511, 0x511}, {0x513, 0x513}, {0x515, 0x515}, {0x517, 0x517}, {0x519, 0x519},
  {0x51b, 0x51b}, {0x51d, 0x51d}, {0x51f, 0x51f}, {0x521, 0x521}, {0x523, 0x523}, {0x525, 0x525}, {0x527, 0x527}, {0x529, 0x529},
  {0x52b, 0x52b}, {0x52d, 0x52d}, {0x52f, 0x52f}, {0x559, 0x559}, {0x560, 0x588}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x620, 0x64a},
  {0x66e, 0x66f}, {0x671, 0x6d3}, {0x6d5, 0x6d5}, {0x6e5, 0x6e6}, {0x6ee, 0x6ef}, {0x6fa, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x710},
  {0x712, 0x72f}, {0x74d, 0x7a5}, {0x7b1, 0x7b1}, {0x7ca, 0x7ea}, {0x7f4, 0x7f5}, {0x7fa, 0x7fa}, {0x800, 0x815}, {0x81a, 0x81a},
  {0x824, 0x824}, {0x828, 0x828}, {0x840, 0x858}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x8a0, 0x8c9}, {0x904, 0x939},
  {0x93d, 0x93d}, {0x950, 0x950}, {0x958, 0x961}, {0x971, 0x980}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0},
  {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bd, 0x9bd}, {0x9ce, 0x9ce}, {0x9dc, 0x9dd}, {0x9df, 0x9e1}, {0x9f0, 0x9f1}, {0x9fc, 0x9fc},
  {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa59, 0xa5c},
  {0xa5e, 0xa5e}, {0xa72, 0xa74}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabd, 0xabd}, {0xad0, 0xad0}, {0xae0, 0xae1}, {0xaf9, 0xaf9}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30},
  {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3d, 0xb3d}, {0xb5c, 0xb5d}, {0xb5f, 0xb61}, {0xb71, 0xb71}, {0xb83, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbd0, 0xbd0}, {0xc05, 0xc0c}, {0xc0e, 0xc10}, {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3d, 0xc3d}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc61}, {0xc80, 0xc80}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9}, {0xcbd, 0xcbd},
  {0xcdd, 0xcde}, {0xce0, 0xce1}, {0xcf1, 0xcf2}, {0xd04, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd3a}, {0xd3d, 0xd3d}, {0xd4e, 0xd4e},
  {0xd54, 0xd56}, {0xd5f, 0xd61}, {0xd7a, 0xd7f}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6},
  {0xe01, 0xe30}, {0xe32, 0xe33}, {0xe40, 0xe46}, {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5},
  {0xea7, 0xeb0}, {0xeb2, 0xeb3}, {0xebd, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf40, 0xf47},
  {0xf49, 0xf6c}, {0xf88, 0xf8c}, {0x1000, 0x102a}, {0x103f, 0x103f}, {0x1050, 0x1055}, {0x105a, 0x105d}, {0x1061, 0x1061}, {0x1065, 0x1066},
  {0x106e, 0x1070}, {0x1075, 0x1081}, {0x108e, 0x108e}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256}, {0x1258, 0x1258},
  {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0}, {0x12c2, 0x12c5},
  {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x1380, 0x138f}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f},
  {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1711}, {0x171f, 0x1731}, {0x1740, 0x1751}, {0x1760, 0x176c}, {0x176e, 0x1770},
  {0x1780, 0x17b3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dc}, {0x1820, 0x1878}, {0x1880, 0x18a8}, {0x18aa, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e},
  {0x1950, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x1a00, 0x1a16}, {0x1a20, 0x1a54}, {0x1aa7, 0x1aa7}, {0x1b05, 0x1b33},
  {0x1b45, 0x1b4c}, {0x1b83, 0x1ba0}, {0x1bae, 0x1baf}, {0x1bba, 0x1be5}, {0x1c00, 0x1c23}, {0x1c4d, 0x1c4f}, {0x1c5a, 0x1c7d}, {0x1c80, 0x1c88},
  {0x1c8a, 0x1c8a}, {0x1ce9, 0x1cec}, {0x1cee, 0x1cf3}, {0x1cf5, 0x1cf6}, {0x1cfa, 0x1cfa}, {0x1d00, 0x1dbf}, {0x1e01, 0x1e01}, {0x1e03, 0x1e03},
  {0x1e05, 0x1e05}, {0x1e07, 0x1e07}, {0x1e09, 0x1e09}, {0x1e0b, 0x1e0b}, {0x1e0d, 0x1e0d}, {0x1e0f, 0x1e0f}, {0x1e11, 0x1e11}, {0x1e13, 0x1e13},
  {0x1e15, 0x1e15}, {0x1e17, 0x1e17}, {0x1e19, 0x1e19}, {0x1e1b, 0x1e1b}, {0x1e1d, 0x1e1d}, {0x1e1f, 0x1e1f}, {0x1e21, 0x1e21}, {0x1e23, 0x1e23},
  {0x1e25, 0x1e25}, {0x1e27, 0x1e27}, {0x1e29, 0x1e29}, {0x1e2b, 0x1e2b}, {0x1e2d, 0x1e2d}, {0x1e2f, 0x1e2f}, {0x1e31, 0x1e31}, {0x1e33, 0x1e33},
  {0x1e35, 0x1e35}, {0x1e37, 0x1e37}, {0x1e39, 0x1e39}, {0x1e3b, 0x1e3b}, {0x1e3d, 0x1e3d}, {0x1e3f, 0x1e3f}, {0x1e41, 0x1e41}, {0x1e43, 0x1e43},
  {0x1e45, 0x1e45}, {0x1e47, 0x1e47}, {0x1e49, 0x1e49}, {0x1e4b, 0x1e4b}, {0x1e4d, 0x1e4d}, {0x1e4f, 0x1e4f}, {0x1e51, 0x1e51}, {0x1e53, 0x1e53},
  {0x1e55, 0x1e55}, {0x1e57, 0x1e57}, {0x1e59, 0x1e59}, {0x1e5b, 0x1e5b}, {0x1e5d, 0x1e5d}, {0x1e5f, 0x1e5f}, {0x1e61, 0x1e61}, {0x1e63, 0x1e63},
  {0x1e65, 0x1e65}, {0x1e67, 0x1e67}, {0x1e69, 0x1e69}, {0x1e6b, 0x1e6b}, {0x1e6d, 0x1e6d}, {0x1e6f, 0x1e6f}, {0x1e71, 0x1e71}, {0x1e73, 0x1e73},
  {0x1e75, 0x1e75}, {0x1e77, 0x1e77}, {0x1e79, 0x1e79}, {0x1e7b, 0x1e7b}, {0x1e7d, 0x1e7d}, {0x1e7f, 0x1e7f}, {0x1e81, 0x1e81}, {0x1e83, 0x1e83},
  {0x1e85, 0x1e85}, {0x1e87, 0x1e87}, {0x1e89, 0x1e89}, {0x1e8b, 0x1e8b}, {0x1e8d, 0x1e8d}, {0x1e8f, 0x1e8f}, {0x1e91, 0x1e91}, {0x1e93, 0x1e93},
  {0x1e95, 0x1e9d}, {0x1e9f, 0x1e9f}, {0x1ea1, 0x1ea1}, {0x1ea3, 0x1ea3}, {0x1ea5, 0x1ea5}, {0x1ea7, 0x1ea7}, {0x1ea9, 0x1ea9}, {0x1eab, 0x1eab},
  {0x1ead, 0x1ead}, {0x1eaf, 0x1eaf}, {0x1eb1, 0x1eb1}, {0x1eb3, 0x1eb3}, {0x1eb5, 0x1eb5}, {0x1eb7, 0x1eb7}, {0x1eb9, 0x1eb9}, {0x1ebb, 0x1ebb},
  {0x1ebd, 0x1ebd}, {0x1ebf, 0x1ebf}, {0x1ec1, 0x1ec1}, {0x1ec3, 0x1ec3}, {0x1ec5, 0x1ec5}, {0x1ec7, 0x1ec7}, {0x1ec9, 0x1ec9}, {0x1ecb, 0x1ecb},
  {0x1ecd, 0x1ecd}, {0x1ecf, 0x1ecf}, {0x1ed1, 0x1ed1}, {0x1ed3, 0x1ed3}, {0x1ed5, 0x1ed5}, {0x1ed7, 0x1ed7}, {0x1ed9, 0x1ed9}, {0x1edb, 0x1edb},
  {0x1edd, 0x1edd}, {0x1edf, 0x1edf}, {0x1ee1, 0x1ee1}, {0x1ee3, 0x1ee3}, {0x1ee5, 0x1ee5}, {0x1ee7, 0x1ee7}, {0x1ee9, 0x1ee9}, {0x1eeb, 0x1eeb},
  {0x1eed, 0x1eed}, {0x1eef, 0x1eef}, {0x1ef1, 0x1ef1}, {0x1ef3, 0x1ef3}, {0x1ef5, 0x1ef5}, {0x1ef7, 0x1ef7}, {0x1ef9, 0x1ef9}, {0x1efb, 0x1efb},
  {0x1efd, 0x1efd}, {0x1eff, 0x1f07}, {0x1f10, 0x1f15}, {0x1f20, 0x1f27}, {0x1f30, 0x1f37}, {0x1f40, 0x1f45}, {0x1f50, 0x1f57}, {0x1f60, 0x1f67},
  {0x1f70, 0x1f7d}, {0x1f80, 0x1f87}, {0x1f90, 0x1f97}, {0x1fa0, 0x1fa7}, {0x1fb0, 0x1fb4}, {0x1fb6, 0x1fb7}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4},
  {0x1fc6, 0x1fc7}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fd7}, {0x1fe0, 0x1fe7}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ff7}, {0x2071, 0x2071}, {0x207f, 0x207f},
  {0x2090, 0x209c}, {0x210a, 0x210a}, {0x210e, 0x210f}, {0x2113, 0x2113}, {0x2118, 0x2118}, {0x212e, 0x212f}, {0x2134, 0x2139}, {0x213c, 0x213d},
  {0x2146, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2182}, {0x2184, 0x2188}, {0x2c30, 0x2c5f}, {0x2c61, 0x2c61}, {0x2c65, 0x2c66}, {0x2c68, 0x2c68},
  {0x2c6a, 0x2c6a}, {0x2c6c, 0x2c6c}, {0x2c71, 0x2c71}, {0x2c73, 0x2c74}, {0x2c76, 0x2c7d}, {0x2c81, 0x2c81}, {0x2c83, 0x2c83}, {0x2c85, 0x2c85},
  {0x2c87, 0x2c87}, {0x2c89, 0x2c89}, {0x2c8b, 0x2c8b}, {0x2c8d, 0x2c8d}, {0x2c8f, 0x2c8f}, {0x2c91, 0x2c91}, {0x2c93, 0x2c93}, {0x2c95, 0x2c95},
  {0x2c97, 0x2c97}, {0x2c99, 0x2c99}, {0x2c9b, 0x2c9b}, {0x2c9d, 0x2c9d}, {0x2c9f, 0x2c9f}, {0x2ca1, 0x2ca1}, {0x2ca3, 0x2ca3}, {0x2ca5, 0x2ca5},
  {0x2ca7, 0x2ca7}, {0x2ca9, 0x2ca9}, {0x2cab, 0x2cab}, {0x2cad, 0x2cad}, {0x2caf, 0x2caf}, {0x2cb1, 0x2cb1}, {0x2cb3, 0x2cb3}, {0x2cb5, 0x2cb5},
  {0x2cb7, 0x2cb7}, {0x2cb9, 0x2cb9}, {0x2cbb, 0x2cbb}, {0x2cbd, 0x2cbd}, {0x2cbf, 0x2cbf}, {0x2cc1, 0x2cc1}, {0x2cc3, 0x2cc3}, {0x2cc5, 0x2cc5},
  {0x2cc7, 0x2cc7}, {0x2cc9, 0x2cc9}, {0x2ccb, 0x2ccb}, {0x2ccd, 0x2ccd}, {0x2ccf, 0x2ccf}, {0x2cd1, 0x2cd1}, {0x2cd3, 0x2cd3}, {0x2cd5, 0x2cd5},
  {0x2cd7, 0x2cd7}, {0x2cd9, 0x2cd9}, {0x2cdb, 0x2cdb}, {0x2cdd, 0x2cdd}, {0x2cdf, 0x2cdf}, {0x2ce1, 0x2ce1}, {0x2ce3, 0x2ce4}, {0x2cec, 0x2cec},
  {0x2cee, 0x2cee}, {0x2cf3, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d80, 0x2d96},
  {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde},
  {0x2e2f, 0x2e2f}, {0x3005, 0x3007}, {0x3021, 0x3029}, {0x3031, 0x3035}, {0x3038, 0x303c}, {0x3041, 0x3096}, {0x309b, 0x309f}, {0x30a1, 0x30fa},
  {0x30fc, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd},
  {0xa500, 0xa60c}, {0xa610, 0xa61f}, {0xa62a, 0xa62b}, {0xa641, 0xa641}, {0xa643, 0xa643}, {0xa645, 0xa645}, {0xa647, 0xa647}, {0xa649, 0xa649},
  {0xa64b, 0xa64b}, {0xa64d, 0xa64d}, {0xa64f, 0xa64f}, {0xa651, 0xa651}, {0xa653, 0xa653}, {0xa655, 0xa655}, {0xa657, 0xa657}, {0xa659, 0xa659},
  {0xa65b, 0xa65b}, {0xa65d, 0xa65d}, {0xa65f, 0xa65f}, {0xa661, 0xa661}, {0xa663, 0xa663}, {0xa665, 0xa665}, {0xa667, 0xa667}, {0xa669, 0xa669},
  {0xa66b, 0xa66b}, {0xa66d, 0xa66e}, {0xa67f, 0xa67f}, {0xa681, 0xa681}, {0xa683, 0xa683}, {0xa685, 0xa685}, {0xa687, 0xa687}, {0xa689, 0xa689},
  {0xa68b, 0xa68b}, {0xa68d, 0xa68d}, {0xa68f, 0xa68f}, {0xa691, 0xa691}, {0xa693, 0xa693}, {0xa695, 0xa695}, {0xa697, 0xa697}, {0xa699, 0xa699},
  {0xa69b, 0xa69d}, {0xa6a0, 0xa6ef}, {0xa717, 0xa71f}, {0xa723, 0xa723}, {0xa725, 0xa725}, {0xa727, 0xa727}, {0xa729, 0xa729}, {0xa72b, 0xa72b},
  {0xa72d, 0xa72d}, {0xa72f, 0xa731}, {0xa733, 0xa733}, {0xa735, 0xa735}, {0xa737, 0xa737}, {0xa739, 0xa739}, {0xa73b, 0xa73b}, {0xa73d, 0xa73d},
  {0xa73f, 0xa73f}, {0xa741, 0xa741}, {0xa743, 0xa743}, {0xa745, 0xa745}, {0xa747, 0xa747}, {0xa749, 0xa749}, {0xa74b, 0xa74b}, {0xa74d, 0xa74d},
  {0xa74f, 0xa74f}, {0xa751, 0xa751}, {0xa753, 0xa753}, {0xa755, 0xa755}, {0xa757, 0xa757}, {0xa759, 0xa759}, {0xa75b, 0xa75b}, {0xa75d, 0xa75d},
  {0xa75f, 0xa75f}, {0xa761, 0xa761}, {0xa763, 0xa763}, {0xa765, 0xa765}, {0xa767, 0xa767}, {0xa769, 0xa769}, {0xa76b, 0xa76b}, {0xa76d, 0xa76d},
  {0xa76f, 0xa778}, {0xa77a, 0xa77a}, {0xa77c, 0xa77c}, {0xa77f, 0xa77f}, {0xa781, 0xa781}, {0xa783, 0xa783}, {0xa785, 0xa785}, {0xa787, 0xa788},
  {0xa78c, 0xa78c}, {0xa78e, 0xa78f}, {0xa791, 0xa791}, {0xa793, 0xa795}, {0xa797, 0xa797}, {0xa799, 0xa799}, {0xa79b, 0xa79b}, {0xa79d, 0xa79d},
  {0xa79f, 0xa79f}, {0xa7a1, 0xa7a1}, {0xa7a3, 0xa7a3}, {0xa7a5, 0xa7a5}, {0xa7a7, 0xa7a7}, {0xa7a9, 0xa7a9}, {0xa7af, 0xa7af}, {0xa7b5, 0xa7b5},
  {0xa7b7, 0xa7b7}, {0xa7b9, 0xa7b9}, {0xa7bb, 0xa7bb}, {0xa7bd, 0xa7bd}, {0xa7bf, 0xa7bf}, {0xa7c1, 0xa7c1}, {0xa7c3, 0xa7c3}, {0xa7c8, 0xa7c8},
  {0xa7ca, 0xa7ca}, {0xa7cd, 0xa7cd}, {0xa7d1, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7d5}, {0xa7d7, 0xa7d7}, {0xa7d9, 0xa7d9}, {0xa7db, 0xa7db},
  {0xa7f2, 0xa7f4}, {0xa7f6, 0xa801}, {0xa803, 0xa805}, {0xa807, 0xa80a}, {0xa80c, 0xa822}, {0xa840, 0xa873}, {0xa882, 0xa8b3}, {0xa8f2, 0xa8f7},
  {0xa8fb, 0xa8fb}, {0xa8fd, 0xa8fe}, {0xa90a, 0xa925}, {0xa930, 0xa946}, {0xa960, 0xa97c}, {0xa984, 0xa9b2}, {0xa9cf, 0xa9cf}, {0xa9e0, 0xa9e4},
  {0xa9e6, 0xa9ef}, {0xa9fa, 0xa9fe}, {0xaa00, 0xaa28}, {0xaa40, 0xaa42}, {0xaa44, 0xaa4b}, {0xaa60, 0xaa76}, {0xaa7a, 0xaa7a}, {0xaa7e, 0xaaaf},
  {0xaab1, 0xaab1}, {0xaab5, 0xaab6}, {0xaab9, 0xaabd}, {0xaac0, 0xaac0}, {0xaac2, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaea}, {0xaaf2, 0xaaf4},
  {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabe2},
  {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb1d},
  {0xfb1f, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfd3d},
  {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdfb}, {0xfe70, 0xfe74}, {0xfe76, 0xfefc}, {0xff41, 0xff5a}, {0xff66, 0xffbe}, {0xffc2, 0xffc7},
  {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d}, {0x1003f, 0x1004d},
  {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x10375},
  {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10428, 0x1049d}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563},
  {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767},
  {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c},
  {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7},
  {0x109be, 0x109bf}, {0x10a00, 0x10a00}, {0x10a10, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7},
  {0x10ac9, 0x10ae4}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72}, {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d23},
  {0x10d4a, 0x10d4f}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10f00, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f45},
  {0x10f70, 0x10f81}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11003, 0x11037}, {0x11071, 0x11072}, {0x11075, 0x11075}, {0x11083, 0x110af}, {0x110d0, 0x110e8},
  {0x11103, 0x11126}, {0x11144, 0x11144}, {0x11147, 0x11147}, {0x11150, 0x11172}, {0x11176, 0x11176}, {0x11183, 0x111b2}, {0x111c1, 0x111c4}, {0x111da, 0x111da},
  {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x1122b}, {0x1123f, 0x11240}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d},
  {0x1129f, 0x112a8}, {0x112b0, 0x112de}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339},
  {0x1133d, 0x1133d}, {0x11350, 0x11350}, {0x1135d, 0x11361}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113b7},
  {0x113d1, 0x113d1}, {0x113d3, 0x113d3}, {0x11400, 0x11434}, {0x11447, 0x1144a}, {0x1145f, 0x11461}, {0x11480, 0x114af}, {0x114c4, 0x114c5}, {0x114c7, 0x114c7},
  {0x11580, 0x115ae}, {0x115d8, 0x115db}, {0x11600, 0x1162f}, {0x11644, 0x11644}, {0x11680, 0x116aa}, {0x116b8, 0x116b8}, {0x11700, 0x1171a}, {0x11740, 0x11746},
  {0x11800, 0x1182b}, {0x118c0, 0x118df}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x1192f}, {0x1193f, 0x1193f},
  {0x11941, 0x11941}, {0x119a0, 0x119a7}, {0x119aa, 0x119d0}, {0x119e1, 0x119e1}, {0x119e3, 0x119e3}, {0x11a00, 0x11a00}, {0x11a0b, 0x11a32}, {0x11a3a, 0x11a3a},
  {0x11a50, 0x11a50}, {0x11a5c, 0x11a89}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c2e}, {0x11c40, 0x11c40},
  {0x11c72, 0x11c8f}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d30}, {0x11d46, 0x11d46}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d89},
  {0x11d98, 0x11d98}, {0x11ee0, 0x11ef2}, {0x11f02, 0x11f02}, {0x11f04, 0x11f10}, {0x11f12, 0x11f33}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e},
  {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13441, 0x13446}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x1611d}, {0x16800, 0x16a38},
  {0x16a40, 0x16a5e}, {0x16a70, 0x16abe}, {0x16ad0, 0x16aed}, {0x16b00, 0x16b2f}, {0x16b40, 0x16b43}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c},
  {0x16e60, 0x16e7f}, {0x16f00, 0x16f4a}, {0x16f50, 0x16f50}, {0x16f93, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe3}, {0x17000, 0x187f7}, {0x18800, 0x18cd5},
  {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155},
  {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1d41a, 0x1d433}, {0x1d44e, 0x1d454},
  {0x1d456, 0x1d467}, {0x1d482, 0x1d49b}, {0x1d4b6, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d4cf}, {0x1d4ea, 0x1d503}, {0x1d51e, 0x1d537},
  {0x1d552, 0x1d56b}, {0x1d586, 0x1d59f}, {0x1d5ba, 0x1d5d3}, {0x1d5ee, 0x1d607}, {0x1d622, 0x1d63b}, {0x1d656, 0x1d66f}, {0x1d68a, 0x1d6a5}, {0x1d6c2, 0x1d6da},
  {0x1d6dc, 0x1d6e1}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d71b}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d755}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d78f}, {0x1d7aa, 0x1d7c2},
  {0x1d7c4, 0x1d7c9}, {0x1d7cb, 0x1d7cb}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e030, 0x1e06d}, {0x1e100, 0x1e12c}, {0x1e137, 0x1e13d}, {0x1e14e, 0x1e14e},
  {0x1e290, 0x1e2ad}, {0x1e2c0, 0x1e2eb}, {0x1e4d0, 0x1e4eb}, {0x1e5d0, 0x1e5ed}, {0x1e5f0, 0x1e5f0}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee},
  {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e922, 0x1e943}, {0x1e94b, 0x1e94b}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24},
  {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49},
  {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d},
  {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e},
  {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d},
  {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
};

static const TSCharacterRange aux_sym_identifier_token1_character_set_2[] = {
  {'!', '!'}, {'0', '9'}, {'?', '?'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5},
  {0xb7, 0xb7}, {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec},
  {0x2ee, 0x2ee}, {0x300, 0x374}, {0x376, 0x377}, {0x37a, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1},
  {0x3a3, 0x3f5}, {0x3f7, 0x481}, {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd},
  {0x5bf, 0x5bf}, {0x5c1, 0x5c2}, {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669},
  {0x66e, 0x6d3}, {0x6d5, 0x6dc}, {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5},
  {0x7fa, 0x7fa}, {0x7fd, 0x7fd}, {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1},
  {0x8e3, 0x963}, {0x966, 0x96f}, {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2},
  {0x9b6, 0x9b9}, {0x9bc, 0x9c4}, {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1},
  {0x9fc, 0x9fc}, {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33},
  {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c},
  {0xa5e, 0xa5e}, {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3},
  {0xab5, 0xab9}, {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff},
  {0xb01, 0xb03}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44},
  {0xb47, 0xb48}, {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83},
  {0xb85, 0xb8a}, {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa},
  {0xbae, 0xbb9}, {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c},
  {0xc0e, 0xc10}, {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a},
  {0xc5d, 0xc5d}, {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3},
  {0xcb5, 0xcb9}, {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef},
  {0xcf1, 0xcf3}, {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63},
  {0xd66, 0xd6f}, {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6},
  {0xdca, 0xdca}, {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e},
  {0xe50, 0xe59}, {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4},
  {0xec6, 0xec6}, {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35},
  {0xf37, 0xf37}, {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6},
  {0x1000, 0x1049}, {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d},
  {0x1250, 0x1256}, {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be},
  {0x12c0, 0x12c0}, {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371},
  {0x1380, 0x138f}, {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8},
  {0x1700, 0x1715}, {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7},
  {0x17dc, 0x17dd}, {0x17e0, 0x17e9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e},
  {0x1920, 0x192b}, {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b},
  {0x1a20, 0x1a5e}, {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c},
  {0x1b50, 0x1b59}, {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba},
  {0x1cbd, 0x1cbf}, {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57},
  {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4},
  {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040},
  {0x2054, 0x2054}, {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102},
  {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139},
  {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27},
  {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe},
  {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035},
  {0x3038, 0x303c}, {0x3041, 0x3096}, {0x3099, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff},
  {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1},
  {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c},
  {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c},
  {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2},
  {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e},
  {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb},
  {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e},
  {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdfb}, {0xfe00, 0xfe0f},
  {0xfe20, 0xfe2f}, {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe70, 0xfe74}, {0xfe76, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f},
  {0xff41, 0xff5a}, {0xff65, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026},
  {0x10028, 0x1003a}, {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c},
  {0x102a0, 0x102d0}, {0x102e0, 0x102e0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf},
  {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a},
  {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3},
  {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808},
  {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5},
  {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17},
  {0x10a19, 0x10a35}, {0x10a38, 0x10a3a}, {0x10a3f, 0x10a3f}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35},
  {0x10b40, 0x10b55}, {0x10b60, 0x10b72}, {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39},
  {0x10d40, 0x10d65}, {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c},
  {0x10f27, 0x10f27}, {0x10f30, 0x10f50}, {0x10f70, 0x10f85}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba},
  {0x110c2, 0x110c2}, {0x110d0, 0x110e8}, {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176},
  {0x11180, 0x111c4}, {0x111c9, 0x111cc}, {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286},
  {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c},
  {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d},
  {0x11350, 0x11350}, {0x11357, 0x11357}, {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e},
  {0x11390, 0x113b5}, {0x113b7, 0x113c0}, {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a},
  {0x11450, 0x11459}, {0x1145e, 0x11461}, {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd},
  {0x11600, 0x11640}, {0x11644, 0x11644}, {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b},
  {0x11730, 0x11739}, {0x11740, 0x11746}, {0x11800, 0x1183a}, {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916},
  {0x11918, 0x11935}, {0x11937, 0x11938}, {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4},
  {0x11a00, 0x11a3e}, {0x11a47, 0x11a47}, {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08},
  {0x11c0a, 0x11c36}, {0x11c38, 0x11c40}, {0x11c50, 0x11c59}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09},
  {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a}, {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e},
  {0x11d90, 0x11d91}, {0x11d93, 0x11d98}, {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a},
  {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa},
  {0x14400, 0x14646}, {0x16100, 0x16139}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed},
  {0x16af0, 0x16af4}, {0x16b00, 0x16b36}, {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79},
  {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a}, {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7},
  {0x18800, 0x18cd5}, {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152},
  {0x1b155, 0x1b155}, {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e},
  {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d}, {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad},
  {0x1d242, 0x1d244}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9},
  {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e},
  {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714},
  {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff},
  {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a},
  {0x1e000, 0x1e006}, {0x1e008, 0x1e018}, {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c},
  {0x1e130, 0x1e13d}, {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6},
  {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03},
  {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b},
  {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57},
  {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72},
  {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb},
  {0x1fbf0, 0x1fbf9}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d},
  {0x30000, 0x3134a}, {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static const TSCharacterRange sym_atom_character_set_2[] = {
  {'!', '!'}, {'0', '9'}, {'?', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xb7, 0xb7},
  {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee},
  {0x300, 0x374}, {0x376, 0x377}, {0x37a, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf},
  {0x5c1, 0x5c2}, {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3},
  {0x6d5, 0x6dc}, {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa},
  {0x7fd, 0x7fd}, {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963},
  {0x966, 0x96f}, {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9},
  {0x9bc, 0x9c4}, {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9fc, 0x9fc},
  {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36},
  {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e},
  {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03},
  {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48},
  {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c}, {0xc0e, 0xc10},
  {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9},
  {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3},
  {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63}, {0xd66, 0xd6f},
  {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca},
  {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59},
  {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6},
  {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35}, {0xf37, 0xf37},
  {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049},
  {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256},
  {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0},
  {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371}, {0x1380, 0x138f},
  {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715},
  {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd},
  {0x17e0, 0x17e9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b},
  {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e},
  {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59},
  {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf},
  {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59},
  {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc},
  {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040}, {0x2054, 0x2054},
  {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107},
  {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f},
  {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d},
  {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6},
  {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035}, {0x3038, 0x303c},
  {0x3041, 0x3096}, {0x3099, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3400, 0x4dbf},
  {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1}, {0xa717, 0xa71f},
  {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c}, {0xa840, 0xa873},
  {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c}, {0xa980, 0xa9c0},
  {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2}, {0xaadb, 0xaadd},
  {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a},
  {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d},
  {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41},
  {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdfb}, {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f},
  {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe70, 0xfe74}, {0xfe76, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a},
  {0xff65, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a},
  {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0},
  {0x102e0, 0x102e0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5},
  {0x10400, 0x1049d}, {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a},
  {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736},
  {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835},
  {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915},
  {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35},
  {0x10a38, 0x10a3a}, {0x10a3f, 0x10a3f}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55},
  {0x10b60, 0x10b72}, {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65},
  {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27},
  {0x10f30, 0x10f50}, {0x10f70, 0x10f85}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2},
  {0x110d0, 0x110e8}, {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4},
  {0x111c9, 0x111cc}, {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288},
  {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310},
  {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350},
  {0x11357, 0x11357}, {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5},
  {0x113b7, 0x113c0}, {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459},
  {0x1145e, 0x11461}, {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640},
  {0x11644, 0x11644}, {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739},
  {0x11740, 0x11746}, {0x11800, 0x1183a}, {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935},
  {0x11937, 0x11938}, {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e},
  {0x11a47, 0x11a47}, {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36},
  {0x11c38, 0x11c40}, {0x11c50, 0x11c59}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36},
  {0x11d3a, 0x11d3a}, {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91},
  {0x11d93, 0x11d98}, {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0},
  {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646},
  {0x16100, 0x16139}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4},
  {0x16b00, 0x16b36}, {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f},
  {0x16f00, 0x16f4a}, {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5},
  {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155},
  {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9},
  {0x1cf00, 0x1cf2d}, {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244},
  {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb},
  {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544},
  {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734},
  {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36},
  {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006},
  {0x1e008, 0x1e018}, {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d},
  {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb},
  {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f},
  {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42},
  {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59},
  {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77},
  {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9},
  {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a},
  {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static const TSCharacterRange sym_keyword_character_set_1[] = {
  {'!', '!'}, {'%', '&'}, {'*', '+'}, {'-', '/'}, {'<', '>'}, {'@', 'Z'}, {'\\', '\\'}, {'^', '_'},
  {'a', '|'}, {'~', '~'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1},
  {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x370, 0x374}, {0x376, 0x377}, {0x37a, 0x37d}, {0x37f, 0x37f},
  {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5}, {0x3f7, 0x481}, {0x48a, 0x52f}, {0x531, 0x556},
  {0x559, 0x559}, {0x560, 0x588}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x620, 0x64a}, {0x66e, 0x66f}, {0x671, 0x6d3}, {0x6d5, 0x6d5},
  {0x6e5, 0x6e6}, {0x6ee, 0x6ef}, {0x6fa, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x710}, {0x712, 0x72f}, {0x74d, 0x7a5}, {0x7b1, 0x7b1},
  {0x7ca, 0x7ea}, {0x7f4, 0x7f5}, {0x7fa, 0x7fa}, {0x800, 0x815}, {0x81a, 0x81a}, {0x824, 0x824}, {0x828, 0x828}, {0x840, 0x858},
  {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x8a0, 0x8c9}, {0x904, 0x939}, {0x93d, 0x93d}, {0x950, 0x950}, {0x958, 0x961},
  {0x971, 0x980}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bd, 0x9bd},
  {0x9ce, 0x9ce}, {0x9dc, 0x9dd}, {0x9df, 0x9e1}, {0x9f0, 0x9f1}, {0x9fc, 0x9fc}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28},
  {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa59, 0xa5c}, {0xa5e, 0xa5e}, {0xa72, 0xa74}, {0xa85, 0xa8d},
  {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9}, {0xabd, 0xabd}, {0xad0, 0xad0}, {0xae0, 0xae1},
  {0xaf9, 0xaf9}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3d, 0xb3d},
  {0xb5c, 0xb5d}, {0xb5f, 0xb61}, {0xb71, 0xb71}, {0xb83, 0xb83}, {0xb85, 0xb8a}, {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a},
  {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9}, {0xbd0, 0xbd0}, {0xc05, 0xc0c}, {0xc0e, 0xc10},
  {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3d, 0xc3d}, {0xc58, 0xc5a}, {0xc5d, 0xc5d}, {0xc60, 0xc61}, {0xc80, 0xc80}, {0xc85, 0xc8c},
  {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9}, {0xcbd, 0xcbd}, {0xcdd, 0xcde}, {0xce0, 0xce1}, {0xcf1, 0xcf2},
  {0xd04, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd3a}, {0xd3d, 0xd3d}, {0xd4e, 0xd4e}, {0xd54, 0xd56}, {0xd5f, 0xd61}, {0xd7a, 0xd7f},
  {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xe01, 0xe30}, {0xe32, 0xe33}, {0xe40, 0xe46},
  {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xeb0}, {0xeb2, 0xeb3}, {0xebd, 0xebd},
  {0xec0, 0xec4}, {0xec6, 0xec6}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf40, 0xf47}, {0xf49, 0xf6c}, {0xf88, 0xf8c}, {0x1000, 0x102a},
  {0x103f, 0x103f}, {0x1050, 0x1055}, {0x105a, 0x105d}, {0x1061, 0x1061}, {0x1065, 0x1066}, {0x106e, 0x1070}, {0x1075, 0x1081}, {0x108e, 0x108e},
  {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256}, {0x1258, 0x1258},
  {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0}, {0x12c2, 0x12c5},
  {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x1380, 0x138f}, {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c},
  {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1711}, {0x171f, 0x1731}, {0x1740, 0x1751}, {0x1760, 0x176c},
  {0x176e, 0x1770}, {0x1780, 0x17b3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dc}, {0x1820, 0x1878}, {0x1880, 0x18a8}, {0x18aa, 0x18aa}, {0x18b0, 0x18f5},
  {0x1900, 0x191e}, {0x1950, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x1a00, 0x1a16}, {0x1a20, 0x1a54}, {0x1aa7, 0x1aa7},
  {0x1b05, 0x1b33}, {0x1b45, 0x1b4c}, {0x1b83, 0x1ba0}, {0x1bae, 0x1baf}, {0x1bba, 0x1be5}, {0x1c00, 0x1c23}, {0x1c4d, 0x1c4f}, {0x1c5a, 0x1c7d},
  {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1ce9, 0x1cec}, {0x1cee, 0x1cf3}, {0x1cf5, 0x1cf6}, {0x1cfa, 0x1cfa}, {0x1d00, 0x1dbf},
  {0x1e00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d},
  {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb},
  {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x2102, 0x2102}, {0x2107, 0x2107},
  {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f},
  {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cee}, {0x2cf2, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27},
  {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d80, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe},
  {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x3005, 0x3007}, {0x3021, 0x3029}, {0x3031, 0x3035}, {0x3038, 0x303c},
  {0x3041, 0x3096}, {0x309b, 0x309f}, {0x30a1, 0x30fa}, {0x30fc, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff},
  {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa61f}, {0xa62a, 0xa62b}, {0xa640, 0xa66e}, {0xa67f, 0xa69d},
  {0xa6a0, 0xa6ef}, {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa801},
  {0xa803, 0xa805}, {0xa807, 0xa80a}, {0xa80c, 0xa822}, {0xa840, 0xa873}, {0xa882, 0xa8b3}, {0xa8f2, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa8fe},
  {0xa90a, 0xa925}, {0xa930, 0xa946}, {0xa960, 0xa97c}, {0xa984, 0xa9b2}, {0xa9cf, 0xa9cf}, {0xa9e0, 0xa9e4}, {0xa9e6, 0xa9ef}, {0xa9fa, 0xa9fe},
  {0xaa00, 0xaa28}, {0xaa40, 0xaa42}, {0xaa44, 0xaa4b}, {0xaa60, 0xaa76}, {0xaa7a, 0xaa7a}, {0xaa7e, 0xaaaf}, {0xaab1, 0xaab1}, {0xaab5, 0xaab6},
  {0xaab9, 0xaabd}, {0xaac0, 0xaac0}, {0xaac2, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaea}, {0xaaf2, 0xaaf4}, {0xab01, 0xab06}, {0xab09, 0xab0e},
  {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabe2}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6},
  {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb1d}, {0xfb1f, 0xfb28}, {0xfb2a, 0xfb36},
  {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7},
  {0xfdf0, 0xfdfb}, {0xfe70, 0xfe74}, {0xfe76, 0xfefc}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0xff66, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf},
  {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d},
  {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x10375}, {0x10380, 0x1039d},
  {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563},
  {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc},
  {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805},
  {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2},
  {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a00}, {0x10a10, 0x10a13}, {0x10a15, 0x10a17},
  {0x10a19, 0x10a35}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae4}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72},
  {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d23}, {0x10d4a, 0x10d65}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9},
  {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10f00, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f45}, {0x10f70, 0x10f81}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6},
  {0x11003, 0x11037}, {0x11071, 0x11072}, {0x11075, 0x11075}, {0x11083, 0x110af}, {0x110d0, 0x110e8}, {0x11103, 0x11126}, {0x11144, 0x11144}, {0x11147, 0x11147},
  {0x11150, 0x11172}, {0x11176, 0x11176}, {0x11183, 0x111b2}, {0x111c1, 0x111c4}, {0x111da, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x1122b},
  {0x1123f, 0x11240}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112de}, {0x11305, 0x1130c},
  {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133d, 0x1133d}, {0x11350, 0x11350}, {0x1135d, 0x11361},
  {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113b7}, {0x113d1, 0x113d1}, {0x113d3, 0x113d3}, {0x11400, 0x11434},
  {0x11447, 0x1144a}, {0x1145f, 0x11461}, {0x11480, 0x114af}, {0x114c4, 0x114c5}, {0x114c7, 0x114c7}, {0x11580, 0x115ae}, {0x115d8, 0x115db}, {0x11600, 0x1162f},
  {0x11644, 0x11644}, {0x11680, 0x116aa}, {0x116b8, 0x116b8}, {0x11700, 0x1171a}, {0x11740, 0x11746}, {0x11800, 0x1182b}, {0x118a0, 0x118df}, {0x118ff, 0x11906},
  {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x1192f}, {0x1193f, 0x1193f}, {0x11941, 0x11941}, {0x119a0, 0x119a7}, {0x119aa, 0x119d0},
  {0x119e1, 0x119e1}, {0x119e3, 0x119e3}, {0x11a00, 0x11a00}, {0x11a0b, 0x11a32}, {0x11a3a, 0x11a3a}, {0x11a50, 0x11a50}, {0x11a5c, 0x11a89}, {0x11a9d, 0x11a9d},
  {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c2e}, {0x11c40, 0x11c40}, {0x11c72, 0x11c8f}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09},
  {0x11d0b, 0x11d30}, {0x11d46, 0x11d46}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d89}, {0x11d98, 0x11d98}, {0x11ee0, 0x11ef2}, {0x11f02, 0x11f02},
  {0x11f04, 0x11f10}, {0x11f12, 0x11f33}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f},
  {0x13441, 0x13446}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x1611d}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a70, 0x16abe}, {0x16ad0, 0x16aed},
  {0x16b00, 0x16b2f}, {0x16b40, 0x16b43}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a}, {0x16f50, 0x16f50},
  {0x16f93, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe3}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb},
  {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a},
  {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6},
  {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c},
  {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da},
  {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2},
  {0x1d7c4, 0x1d7cb}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e030, 0x1e06d}, {0x1e100, 0x1e12c}, {0x1e137, 0x1e13d}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ad},
  {0x1e2c0, 0x1e2eb}, {0x1e4d0, 0x1e4eb}, {0x1e5d0, 0x1e5ed}, {0x1e5f0, 0x1e5f0}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe},
  {0x1e800, 0x1e8c4}, {0x1e900, 0x1e943}, {0x1e94b, 0x1e94b}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27},
  {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b},
  {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f},
  {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89},
  {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1},
  {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
};

static const TSCharacterRange sym_keyword_character_set_2[] = {
  {'!', '!'}, {'0', ':'}, {'?', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xb7, 0xb7},
  {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee},
  {0x300, 0x374}, {0x376, 0x377}, {0x37a, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf},
  {0x5c1, 0x5c2}, {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3},
  {0x6d5, 0x6dc}, {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa},
  {0x7fd, 0x7fd}, {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963},
  {0x966, 0x96f}, {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9},
  {0x9bc, 0x9c4}, {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9fc, 0x9fc},
  {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36},
  {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e},
  {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03},
  {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48},
  {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c}, {0xc0e, 0xc10},
  {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9},
  {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3},
  {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63}, {0xd66, 0xd6f},
  {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca},
  {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59},
  {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6},
  {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35}, {0xf37, 0xf37},
  {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049},
  {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256},
  {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0},
  {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371}, {0x1380, 0x138f},
  {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715},
  {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd},
  {0x17e0, 0x17e9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b},
  {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e},
  {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59},
  {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf},
  {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59},
  {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc},
  {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040}, {0x2054, 0x2054},
  {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107},
  {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f},
  {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d},
  {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6},
  {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035}, {0x3038, 0x303c},
  {0x3041, 0x3096}, {0x3099, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3400, 0x4dbf},
  {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1}, {0xa717, 0xa71f},
  {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c}, {0xa840, 0xa873},
  {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c}, {0xa980, 0xa9c0},
  {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2}, {0xaadb, 0xaadd},
  {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a},
  {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d},
  {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41},
  {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdfb}, {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f},
  {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe70, 0xfe74}, {0xfe76, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a},
  {0xff65, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a},
  {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0},
  {0x102e0, 0x102e0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5},
  {0x10400, 0x1049d}, {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a},
  {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736},
  {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835},
  {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915},
  {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35},
  {0x10a38, 0x10a3a}, {0x10a3f, 0x10a3f}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55},
  {0x10b60, 0x10b72}, {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65},
  {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27},
  {0x10f30, 0x10f50}, {0x10f70, 0x10f85}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2},
  {0x110d0, 0x110e8}, {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4},
  {0x111c9, 0x111cc}, {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288},
  {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310},
  {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350},
  {0x11357, 0x11357}, {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5},
  {0x113b7, 0x113c0}, {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459},
  {0x1145e, 0x11461}, {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640},
  {0x11644, 0x11644}, {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739},
  {0x11740, 0x11746}, {0x11800, 0x1183a}, {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935},
  {0x11937, 0x11938}, {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e},
  {0x11a47, 0x11a47}, {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36},
  {0x11c38, 0x11c40}, {0x11c50, 0x11c59}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36},
  {0x11d3a, 0x11d3a}, {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91},
  {0x11d93, 0x11d98}, {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0},
  {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646},
  {0x16100, 0x16139}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4},
  {0x16b00, 0x16b36}, {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f},
  {0x16f00, 0x16f4a}, {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5},
  {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155},
  {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9},
  {0x1cf00, 0x1cf2d}, {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244},
  {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb},
  {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544},
  {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734},
  {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36},
  {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006},
  {0x1e008, 0x1e018}, {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d},
  {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb},
  {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f},
  {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42},
  {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59},
  {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77},
  {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9},
  {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a},
  {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 278,
        '\r', 1,
        '!', 535,
        '"', 436,
        '#', 664,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 662,
        ')', 311,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 149,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 451,
        '?', 206,
        '@', 546,
        '[', 663,
        '\\', 4,
        ']', 445,
        '^', 538,
        'a', 472,
        'c', 470,
        'd', 477,
        'e', 475,
        'f', 469,
        'i', 476,
        'n', 474,
        'o', 478,
        'r', 471,
        't', 479,
        'w', 473,
        '{', 442,
        '|', 456,
        '}', 443,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(266);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(480);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(481);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(278);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        '\n', 278,
        '\r', 1,
        '!', 535,
        '"', 436,
        '#', 664,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        ')', 311,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 451,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 4,
        ']', 445,
        '^', 538,
        'a', 327,
        'c', 315,
        'd', 338,
        'e', 331,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        'r', 324,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '}', 443,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(2);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 3:
      if (lookahead == '\n') SKIP(2);
      END_STATE();
    case 4:
      if (lookahead == '\n') SKIP(2);
      if (lookahead == '\r') SKIP(3);
      if (lookahead == '\\') ADVANCE(550);
      END_STATE();
    case 5:
      if (lookahead == '\n') SKIP(79);
      END_STATE();
    case 6:
      if (lookahead == '\n') SKIP(79);
      if (lookahead == '\r') SKIP(5);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 7:
      if (lookahead == '\n') SKIP(81);
      END_STATE();
    case 8:
      if (lookahead == '\n') SKIP(81);
      if (lookahead == '\r') SKIP(7);
      if (lookahead == '\\') ADVANCE(550);
      END_STATE();
    case 9:
      if (lookahead == '\n') SKIP(86);
      END_STATE();
    case 10:
      if (lookahead == '\n') SKIP(86);
      if (lookahead == '\r') SKIP(9);
      if (lookahead == '\\') ADVANCE(550);
      END_STATE();
    case 11:
      if (lookahead == '\n') SKIP(94);
      END_STATE();
    case 12:
      if (lookahead == '\n') SKIP(94);
      if (lookahead == '\r') SKIP(11);
      if (lookahead == '\\') ADVANCE(550);
      END_STATE();
    case 13:
      if (lookahead == '\n') SKIP(96);
      END_STATE();
    case 14:
      if (lookahead == '\n') SKIP(96);
      if (lookahead == '\r') SKIP(13);
      if (lookahead == '\\') ADVANCE(550);
      END_STATE();
    case 15:
      if (lookahead == '\n') SKIP(98);
      END_STATE();
    case 16:
      if (lookahead == '\n') SKIP(98);
      if (lookahead == '\r') SKIP(15);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 17:
      if (lookahead == '\n') SKIP(100);
      END_STATE();
    case 18:
      if (lookahead == '\n') SKIP(100);
      if (lookahead == '\r') SKIP(17);
      if (lookahead == '\\') ADVANCE(550);
      END_STATE();
    case 19:
      if (lookahead == '\n') SKIP(102);
      END_STATE();
    case 20:
      if (lookahead == '\n') SKIP(102);
      if (lookahead == '\r') SKIP(19);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 21:
      if (lookahead == '\n') SKIP(104);
      END_STATE();
    case 22:
      if (lookahead == '\n') SKIP(104);
      if (lookahead == '\r') SKIP(21);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 23:
      if (lookahead == '\n') SKIP(107);
      END_STATE();
    case 24:
      if (lookahead == '\n') SKIP(107);
      if (lookahead == '\r') SKIP(23);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 25:
      if (lookahead == '\n') SKIP(109);
      END_STATE();
    case 26:
      if (lookahead == '\n') SKIP(109);
      if (lookahead == '\r') SKIP(25);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 27:
      if (lookahead == '\n') SKIP(112);
      END_STATE();
    case 28:
      if (lookahead == '\n') SKIP(112);
      if (lookahead == '\r') SKIP(27);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 29:
      if (lookahead == '\n') SKIP(118);
      END_STATE();
    case 30:
      if (lookahead == '\n') SKIP(118);
      if (lookahead == '\r') SKIP(29);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 31:
      if (lookahead == '\n') SKIP(123);
      END_STATE();
    case 32:
      if (lookahead == '\n') SKIP(123);
      if (lookahead == '\r') SKIP(31);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 33:
      if (lookahead == '\n') SKIP(126);
      END_STATE();
    case 34:
      if (lookahead == '\n') SKIP(126);
      if (lookahead == '\r') SKIP(33);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 35:
      if (lookahead == '\n') SKIP(128);
      END_STATE();
    case 36:
      if (lookahead == '\n') SKIP(128);
      if (lookahead == '\r') SKIP(35);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 37:
      if (lookahead == '\n') SKIP(131);
      END_STATE();
    case 38:
      if (lookahead == '\n') SKIP(131);
      if (lookahead == '\r') SKIP(37);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 39:
      if (lookahead == '\n') SKIP(134);
      END_STATE();
    case 40:
      if (lookahead == '\n') SKIP(134);
      if (lookahead == '\r') SKIP(39);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 41:
      if (lookahead == '\n') SKIP(137);
      END_STATE();
    case 42:
      if (lookahead == '\n') SKIP(137);
      if (lookahead == '\r') SKIP(41);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 43:
      if (lookahead == '\n') SKIP(140);
      END_STATE();
    case 44:
      if (lookahead == '\n') SKIP(140);
      if (lookahead == '\r') SKIP(43);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 45:
      if (lookahead == '\n') SKIP(143);
      END_STATE();
    case 46:
      if (lookahead == '\n') SKIP(143);
      if (lookahead == '\r') SKIP(45);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 47:
      if (lookahead == '\n') SKIP(146);
      END_STATE();
    case 48:
      if (lookahead == '\n') SKIP(146);
      if (lookahead == '\r') SKIP(47);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 49:
      if (lookahead == '\n') SKIP(52);
      END_STATE();
    case 50:
      if (lookahead == '\n') SKIP(52);
      if (lookahead == '\r') SKIP(49);
      END_STATE();
    case 51:
      if (lookahead == '\n') ADVANCE(279);
      END_STATE();
    case 52:
      ADVANCE_MAP(
        '\n', 279,
        '\r', 51,
        '!', 534,
        '"', 436,
        '#', 665,
        '&', 524,
        '\'', 438,
        '(', 310,
        '+', 527,
        ',', 510,
        '-', 530,
        '/', 458,
        '<', 446,
        '>', 201,
        '@', 545,
        '[', 444,
      );
      if (lookahead == '\\') SKIP(50);
      if (lookahead == ']') ADVANCE(445);
      if (lookahead == '^') ADVANCE(537);
      if (lookahead == 'n') ADVANCE(231);
      if (lookahead == '{') ADVANCE(441);
      if (lookahead == '|') ADVANCE(455);
      if (lookahead == '}') ADVANCE(443);
      if (lookahead == '~') ADVANCE(249);
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(52);
      END_STATE();
    case 53:
      if (lookahead == '\n') ADVANCE(279);
      if (lookahead == '\r') ADVANCE(51);
      if (lookahead == '#') ADVANCE(665);
      if (lookahead == ')') ADVANCE(311);
      if (lookahead == ',') ADVANCE(510);
      if (lookahead == '-') ADVANCE(196);
      if (lookahead == '>') ADVANCE(201);
      if (lookahead == '\\') SKIP(59);
      if (lookahead == ']') ADVANCE(445);
      if (lookahead == 'w') ADVANCE(224);
      if (lookahead == '}') ADVANCE(443);
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(53);
      END_STATE();
    case 54:
      if (lookahead == '\n') ADVANCE(279);
      if (lookahead == '\r') ADVANCE(51);
      if (lookahead == '#') ADVANCE(665);
      if (lookahead == '\\') SKIP(62);
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(54);
      END_STATE();
    case 55:
      if (lookahead == '\n') ADVANCE(279);
      if (lookahead == '\r') ADVANCE(51);
      if (lookahead == '#') ADVANCE(665);
      if (lookahead == '\\') SKIP(62);
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(54);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(482);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(468);
      END_STATE();
    case 56:
      if (lookahead == '\n') SKIP(148);
      END_STATE();
    case 57:
      if (lookahead == '\n') SKIP(148);
      if (lookahead == '\r') SKIP(56);
      if (lookahead == '\\') ADVANCE(175);
      END_STATE();
    case 58:
      if (lookahead == '\n') SKIP(53);
      END_STATE();
    case 59:
      if (lookahead == '\n') SKIP(53);
      if (lookahead == '\r') SKIP(58);
      END_STATE();
    case 60:
      if (lookahead == '\n') ADVANCE(464);
      if (lookahead == '\r') ADVANCE(463);
      if (lookahead == 'u') ADVANCE(241);
      if (lookahead == 'x') ADVANCE(240);
      if (lookahead != 0) ADVANCE(463);
      END_STATE();
    case 61:
      if (lookahead == '\n') SKIP(54);
      END_STATE();
    case 62:
      if (lookahead == '\n') SKIP(54);
      if (lookahead == '\r') SKIP(61);
      END_STATE();
    case 63:
      if (lookahead == '\n') ADVANCE(280);
      END_STATE();
    case 64:
      ADVANCE_MAP(
        '\n', 280,
        '\r', 63,
        '"', 435,
        '#', 664,
        '\'', 437,
        ')', 311,
        '/', 458,
        '>', 450,
        '\\', 60,
        ']', 445,
        '|', 455,
        '}', 443,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(64);
      END_STATE();
    case 65:
      ADVANCE_MAP(
        '\n', 280,
        '\r', 63,
        '"', 435,
        '#', 665,
        '\'', 437,
        ')', 311,
        '/', 458,
        '>', 450,
        '\\', 60,
        ']', 445,
        '|', 455,
        '}', 443,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(65);
      END_STATE();
    case 66:
      if (lookahead == '\n') ADVANCE(280);
      if (lookahead == '\r') ADVANCE(63);
      if (lookahead == '"') ADVANCE(151);
      if (lookahead == '#') ADVANCE(664);
      if (lookahead == '\'') ADVANCE(156);
      if (lookahead == '\\') ADVANCE(60);
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(66);
      END_STATE();
    case 67:
      if (lookahead == '\n') ADVANCE(280);
      if (lookahead == '\r') ADVANCE(63);
      if (lookahead == '"') ADVANCE(151);
      if (lookahead == '#') ADVANCE(665);
      if (lookahead == '\'') ADVANCE(156);
      if (lookahead == '\\') ADVANCE(60);
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(67);
      END_STATE();
    case 68:
      if (lookahead == '\n') SKIP(84);
      END_STATE();
    case 69:
      if (lookahead == '\n') SKIP(84);
      if (lookahead == '\r') SKIP(68);
      if (lookahead == '\\') ADVANCE(550);
      END_STATE();
    case 70:
      if (lookahead == '\n') SKIP(89);
      END_STATE();
    case 71:
      if (lookahead == '\n') SKIP(89);
      if (lookahead == '\r') SKIP(70);
      if (lookahead == '\\') ADVANCE(550);
      END_STATE();
    case 72:
      if (lookahead == '\n') SKIP(92);
      END_STATE();
    case 73:
      if (lookahead == '\n') SKIP(92);
      if (lookahead == '\r') SKIP(72);
      if (lookahead == '\\') ADVANCE(550);
      END_STATE();
    case 74:
      if (lookahead == '\n') SKIP(115);
      END_STATE();
    case 75:
      if (lookahead == '\n') SKIP(115);
      if (lookahead == '\r') SKIP(74);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 76:
      if (lookahead == '\n') SKIP(121);
      END_STATE();
    case 77:
      if (lookahead == '\n') SKIP(121);
      if (lookahead == '\r') SKIP(76);
      if (lookahead == '\\') ADVANCE(549);
      END_STATE();
    case 78:
      if (lookahead == '\n') ADVANCE(281);
      END_STATE();
    case 79:
      ADVANCE_MAP(
        '\n', 281,
        '\r', 78,
        '!', 536,
        '"', 436,
        '#', 665,
        '%', 519,
        '&', 526,
        '\'', 438,
        '(', 310,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 165,
        '/', 458,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 448,
        '=', 561,
        '>', 454,
        '?', 206,
        '@', 545,
        '[', 444,
        '\\', 6,
        ']', 445,
        '^', 539,
        'a', 371,
        'f', 351,
        'i', 372,
        'n', 367,
        'o', 376,
        't', 377,
        'w', 365,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 467,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(79);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(391);
      if (set_contains(aux_sym_identifier_token1_character_set_1, 1206, lookahead)) ADVANCE(387);
      END_STATE();
    case 80:
      if (lookahead == '\n') ADVANCE(282);
      END_STATE();
    case 81:
      ADVANCE_MAP(
        '\n', 282,
        '\r', 80,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        '*', 628,
        '+', 528,
        '-', 531,
        '.', 163,
        '/', 460,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 559,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 8,
        '^', 538,
        'a', 327,
        'c', 315,
        'e', 331,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        'r', 324,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(81);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 82:
      if (lookahead == '\n') ADVANCE(283);
      END_STATE();
    case 83:
      ADVANCE_MAP(
        '\n', 283,
        '\r', 82,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 662,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 663,
        '\\', 69,
        '^', 538,
        'a', 327,
        'c', 315,
        'd', 338,
        'e', 331,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        'r', 324,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(84);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 84:
      ADVANCE_MAP(
        '\n', 283,
        '\r', 82,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 69,
        '^', 538,
        'a', 327,
        'c', 315,
        'd', 338,
        'e', 331,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        'r', 324,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(84);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 85:
      if (lookahead == '\n') ADVANCE(284);
      END_STATE();
    case 86:
      ADVANCE_MAP(
        '\n', 284,
        '\r', 85,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        ')', 311,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 10,
        ']', 445,
        '^', 538,
        'a', 334,
        'd', 338,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '}', 443,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(86);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 87:
      if (lookahead == '\n') ADVANCE(285);
      END_STATE();
    case 88:
      ADVANCE_MAP(
        '\n', 285,
        '\r', 87,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 662,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 663,
        '\\', 71,
        '^', 538,
        'a', 334,
        'd', 338,
        'e', 337,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(89);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 89:
      ADVANCE_MAP(
        '\n', 285,
        '\r', 87,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 71,
        '^', 538,
        'a', 334,
        'd', 338,
        'e', 337,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(89);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 90:
      if (lookahead == '\n') ADVANCE(286);
      END_STATE();
    case 91:
      ADVANCE_MAP(
        '\n', 286,
        '\r', 90,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 662,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        '<', 447,
        '=', 558,
        '>', 451,
        '?', 206,
        '@', 546,
        '[', 663,
        '\\', 73,
        '^', 538,
        'a', 334,
        'd', 338,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(92);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 92:
      ADVANCE_MAP(
        '\n', 286,
        '\r', 90,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        '<', 447,
        '=', 558,
        '>', 451,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 73,
        '^', 538,
        'a', 334,
        'd', 338,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(92);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 93:
      if (lookahead == '\n') ADVANCE(287);
      END_STATE();
    case 94:
      ADVANCE_MAP(
        '\n', 287,
        '\r', 93,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        ')', 311,
        '*', 628,
        '+', 528,
        '-', 531,
        '.', 163,
        '/', 460,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 559,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 12,
        ']', 445,
        '^', 538,
        'a', 334,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '}', 443,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(94);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 95:
      if (lookahead == '\n') ADVANCE(288);
      END_STATE();
    case 96:
      ADVANCE_MAP(
        '\n', 288,
        '\r', 95,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        '*', 628,
        '+', 528,
        '-', 531,
        '.', 163,
        '/', 460,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 559,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 14,
        '^', 538,
        'a', 334,
        'e', 337,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(96);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 97:
      if (lookahead == '\n') ADVANCE(289);
      END_STATE();
    case 98:
      ADVANCE_MAP(
        '\n', 289,
        '\r', 97,
        '!', 536,
        '"', 436,
        '#', 665,
        '%', 519,
        '&', 526,
        '\'', 438,
        '(', 310,
        '*', 629,
        '+', 529,
        '-', 533,
        '.', 165,
        '/', 458,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 448,
        '=', 561,
        '>', 454,
        '?', 206,
        '@', 545,
        '[', 444,
        '\\', 16,
        '^', 539,
        'a', 364,
        'c', 352,
        'e', 370,
        'f', 351,
        'i', 372,
        'n', 367,
        'o', 376,
        'r', 361,
        't', 377,
        'w', 365,
        '{', 441,
        '|', 457,
        '~', 467,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(98);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(391);
      if (set_contains(aux_sym_identifier_token1_character_set_1, 1206, lookahead)) ADVANCE(387);
      END_STATE();
    case 99:
      if (lookahead == '\n') ADVANCE(290);
      END_STATE();
    case 100:
      ADVANCE_MAP(
        '\n', 290,
        '\r', 99,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        '*', 628,
        '+', 528,
        '-', 531,
        '.', 163,
        '/', 460,
        '0', 392,
        ':', 150,
        '<', 447,
        '=', 559,
        '>', 451,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 18,
        '^', 538,
        'a', 334,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(100);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 101:
      if (lookahead == '\n') ADVANCE(291);
      END_STATE();
    case 102:
      ADVANCE_MAP(
        '\n', 291,
        '\r', 101,
        '!', 536,
        '"', 436,
        '#', 665,
        '%', 519,
        '&', 526,
        '\'', 438,
        '(', 310,
        '*', 629,
        '+', 529,
        '-', 533,
        '.', 165,
        '/', 458,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 448,
        '=', 561,
        '>', 454,
        '?', 206,
        '@', 545,
        '[', 444,
        '\\', 20,
        '^', 539,
        'a', 371,
        'e', 374,
        'f', 351,
        'i', 372,
        'n', 367,
        'o', 376,
        't', 377,
        'w', 365,
        '{', 441,
        '|', 457,
        '~', 467,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(102);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(391);
      if (set_contains(aux_sym_identifier_token1_character_set_1, 1206, lookahead)) ADVANCE(387);
      END_STATE();
    case 103:
      if (lookahead == '\n') ADVANCE(292);
      END_STATE();
    case 104:
      ADVANCE_MAP(
        '\n', 292,
        '\r', 103,
        '!', 536,
        '"', 435,
        '#', 665,
        '&', 526,
        '\'', 437,
        '(', 310,
        '*', 629,
        '+', 529,
        '-', 533,
        '.', 169,
        '/', 458,
        '0', 397,
        ':', 187,
        '<', 449,
        '=', 561,
        '>', 454,
        '@', 545,
        '\\', 22,
        '^', 539,
        'a', 364,
        'c', 352,
        'd', 375,
        'e', 370,
        'f', 351,
        'i', 372,
        'n', 367,
        'o', 376,
        'r', 361,
        't', 377,
        'w', 365,
        '{', 441,
        '|', 457,
        '~', 198,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(104);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(398);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(391);
      if (set_contains(aux_sym_identifier_token1_character_set_1, 1206, lookahead)) ADVANCE(387);
      END_STATE();
    case 105:
      if (lookahead == '\n') ADVANCE(293);
      END_STATE();
    case 106:
      ADVANCE_MAP(
        '\n', 293,
        '\r', 105,
        '!', 190,
        '#', 665,
        '&', 154,
        '(', 310,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 188,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 24,
        '^', 211,
        'a', 222,
        'c', 212,
        'd', 230,
        'e', 225,
        'i', 227,
        'o', 232,
        'r', 221,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(107);
      END_STATE();
    case 107:
      ADVANCE_MAP(
        '\n', 293,
        '\r', 105,
        '!', 190,
        '#', 665,
        '&', 154,
        '(', 310,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 24,
        '^', 211,
        'a', 222,
        'c', 212,
        'd', 230,
        'e', 225,
        'i', 227,
        'o', 232,
        'r', 221,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(107);
      END_STATE();
    case 108:
      if (lookahead == '\n') ADVANCE(294);
      END_STATE();
    case 109:
      ADVANCE_MAP(
        '\n', 294,
        '\r', 108,
        '!', 190,
        '#', 665,
        '&', 154,
        '(', 310,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 26,
        ']', 445,
        '^', 211,
        'a', 222,
        'c', 212,
        'd', 230,
        'e', 225,
        'i', 227,
        'o', 232,
        'r', 221,
        'w', 224,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(109);
      END_STATE();
    case 110:
      if (lookahead == '\n') ADVANCE(295);
      END_STATE();
    case 111:
      ADVANCE_MAP(
        '\n', 295,
        '\r', 110,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 28,
        '^', 211,
        'a', 493,
        'c', 483,
        'd', 501,
        'e', 496,
        'i', 498,
        'o', 502,
        'r', 492,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(112);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 112:
      ADVANCE_MAP(
        '\n', 295,
        '\r', 110,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 28,
        '^', 211,
        'a', 222,
        'c', 212,
        'd', 230,
        'e', 225,
        'i', 227,
        'o', 232,
        'r', 221,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(112);
      END_STATE();
    case 113:
      if (lookahead == '\n') ADVANCE(296);
      END_STATE();
    case 114:
      ADVANCE_MAP(
        '\n', 296,
        '\r', 113,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 75,
        '^', 211,
        'a', 493,
        'c', 483,
        'd', 501,
        'e', 496,
        'i', 498,
        'o', 502,
        'r', 492,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(115);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 115:
      ADVANCE_MAP(
        '\n', 296,
        '\r', 113,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 75,
        '^', 211,
        'a', 222,
        'c', 212,
        'd', 230,
        'e', 225,
        'i', 227,
        'o', 232,
        'r', 221,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(115);
      END_STATE();
    case 116:
      if (lookahead == '\n') ADVANCE(297);
      END_STATE();
    case 117:
      ADVANCE_MAP(
        '\n', 297,
        '\r', 116,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 30,
        '^', 211,
        'a', 493,
        'c', 483,
        'e', 496,
        'i', 498,
        'o', 502,
        'r', 492,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(118);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 118:
      ADVANCE_MAP(
        '\n', 297,
        '\r', 116,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 30,
        '^', 211,
        'a', 222,
        'c', 212,
        'e', 225,
        'i', 227,
        'o', 232,
        'r', 221,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(118);
      END_STATE();
    case 119:
      if (lookahead == '\n') ADVANCE(298);
      END_STATE();
    case 120:
      ADVANCE_MAP(
        '\n', 298,
        '\r', 119,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 77,
        '^', 211,
        'a', 493,
        'c', 483,
        'e', 496,
        'i', 498,
        'o', 502,
        'r', 492,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(121);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 121:
      ADVANCE_MAP(
        '\n', 298,
        '\r', 119,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 77,
        '^', 211,
        'a', 222,
        'c', 212,
        'e', 225,
        'i', 227,
        'o', 232,
        'r', 221,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(121);
      END_STATE();
    case 122:
      if (lookahead == '\n') ADVANCE(299);
      END_STATE();
    case 123:
      ADVANCE_MAP(
        '\n', 299,
        '\r', 122,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 32,
        ']', 445,
        '^', 211,
        'a', 226,
        'd', 230,
        'i', 227,
        'o', 232,
        'w', 224,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(123);
      END_STATE();
    case 124:
      if (lookahead == '\n') ADVANCE(300);
      END_STATE();
    case 125:
      ADVANCE_MAP(
        '\n', 300,
        '\r', 124,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 34,
        '^', 211,
        'a', 497,
        'd', 501,
        'i', 498,
        'o', 502,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(126);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 126:
      ADVANCE_MAP(
        '\n', 300,
        '\r', 124,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 34,
        '^', 211,
        'a', 226,
        'd', 230,
        'i', 227,
        'o', 232,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(126);
      END_STATE();
    case 127:
      if (lookahead == '\n') ADVANCE(301);
      END_STATE();
    case 128:
      ADVANCE_MAP(
        '\n', 301,
        '\r', 127,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 36,
        ']', 445,
        '^', 211,
        'a', 226,
        'i', 227,
        'o', 232,
        'w', 224,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(128);
      END_STATE();
    case 129:
      if (lookahead == '\n') ADVANCE(302);
      END_STATE();
    case 130:
      ADVANCE_MAP(
        '\n', 302,
        '\r', 129,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 38,
        '^', 211,
        'a', 497,
        'd', 501,
        'e', 500,
        'i', 498,
        'o', 502,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(131);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 131:
      ADVANCE_MAP(
        '\n', 302,
        '\r', 129,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 38,
        '^', 211,
        'a', 226,
        'd', 230,
        'e', 229,
        'i', 227,
        'o', 232,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(131);
      END_STATE();
    case 132:
      if (lookahead == '\n') ADVANCE(303);
      END_STATE();
    case 133:
      ADVANCE_MAP(
        '\n', 303,
        '\r', 132,
        '!', 190,
        '#', 665,
        '&', 154,
        '(', 310,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 188,
        '<', 449,
        '=', 560,
        '>', 453,
        '[', 663,
        '\\', 40,
        '^', 211,
        'a', 226,
        'd', 230,
        'i', 227,
        'o', 232,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(134);
      END_STATE();
    case 134:
      ADVANCE_MAP(
        '\n', 303,
        '\r', 132,
        '!', 190,
        '#', 665,
        '&', 154,
        '(', 310,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        '<', 449,
        '=', 560,
        '>', 453,
        '\\', 40,
        '^', 211,
        'a', 226,
        'd', 230,
        'i', 227,
        'o', 232,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(134);
      END_STATE();
    case 135:
      if (lookahead == '\n') ADVANCE(304);
      END_STATE();
    case 136:
      ADVANCE_MAP(
        '\n', 304,
        '\r', 135,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        '<', 449,
        '=', 560,
        '>', 453,
        '[', 663,
        '\\', 42,
        '^', 211,
        'a', 497,
        'd', 501,
        'i', 498,
        'o', 502,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(137);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 137:
      ADVANCE_MAP(
        '\n', 304,
        '\r', 135,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        '<', 449,
        '=', 560,
        '>', 453,
        '\\', 42,
        '^', 211,
        'a', 226,
        'd', 230,
        'i', 227,
        'o', 232,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(137);
      END_STATE();
    case 138:
      if (lookahead == '\n') ADVANCE(305);
      END_STATE();
    case 139:
      ADVANCE_MAP(
        '\n', 305,
        '\r', 138,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 44,
        '^', 211,
        'a', 497,
        'i', 498,
        'o', 502,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 140:
      ADVANCE_MAP(
        '\n', 305,
        '\r', 138,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 44,
        '^', 211,
        'a', 226,
        'i', 227,
        'o', 232,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(140);
      END_STATE();
    case 141:
      if (lookahead == '\n') ADVANCE(306);
      END_STATE();
    case 142:
      ADVANCE_MAP(
        '\n', 306,
        '\r', 141,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 46,
        '^', 211,
        'a', 497,
        'e', 500,
        'i', 498,
        'o', 502,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(143);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 143:
      ADVANCE_MAP(
        '\n', 306,
        '\r', 141,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 46,
        '^', 211,
        'a', 226,
        'e', 229,
        'i', 227,
        'o', 232,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(143);
      END_STATE();
    case 144:
      if (lookahead == '\n') ADVANCE(307);
      END_STATE();
    case 145:
      ADVANCE_MAP(
        '\n', 307,
        '\r', 144,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        '<', 449,
        '=', 560,
        '>', 453,
        '[', 663,
        '\\', 48,
        '^', 211,
        'a', 497,
        'i', 498,
        'o', 502,
        'w', 495,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(146);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 146:
      ADVANCE_MAP(
        '\n', 307,
        '\r', 144,
        '!', 190,
        '#', 665,
        '&', 154,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        '<', 449,
        '=', 560,
        '>', 453,
        '\\', 48,
        '^', 211,
        'a', 226,
        'i', 227,
        'o', 232,
        'w', 224,
        '|', 457,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(146);
      END_STATE();
    case 147:
      if (lookahead == '\n') ADVANCE(308);
      END_STATE();
    case 148:
      ADVANCE_MAP(
        '\n', 308,
        '\r', 147,
        '!', 179,
        '"', 435,
        '#', 665,
        '%', 184,
        '&', 155,
        '\'', 437,
        ')', 311,
        '*', 158,
        '+', 160,
        '-', 162,
        '.', 168,
        '/', 172,
        '<', 176,
        '=', 180,
        '>', 178,
        '@', 175,
        '\\', 57,
        ']', 445,
        '^', 183,
        '{', 243,
        '|', 182,
        '}', 443,
        '~', 202,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(148);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(186);
      END_STATE();
    case 149:
      ADVANCE_MAP(
        '!', 427,
        '%', 432,
        '&', 415,
        '*', 416,
        '+', 418,
        '-', 420,
        '.', 422,
        '/', 423,
        ':', 556,
        '<', 424,
        '=', 428,
        '>', 426,
        '@', 413,
        '\\', 205,
        '^', 431,
        '{', 242,
        '|', 430,
        '~', 193,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(512);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(434);
      END_STATE();
    case 150:
      ADVANCE_MAP(
        '!', 427,
        '%', 432,
        '&', 415,
        '*', 416,
        '+', 418,
        '-', 420,
        '.', 422,
        '/', 423,
        ':', 556,
        '<', 424,
        '=', 428,
        '>', 426,
        '@', 413,
        '\\', 205,
        '^', 431,
        '{', 242,
        '|', 430,
        '~', 193,
      );
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(434);
      END_STATE();
    case 151:
      if (lookahead == '"') ADVANCE(152);
      END_STATE();
    case 152:
      if (lookahead == '"') ADVANCE(440);
      END_STATE();
    case 153:
      if (lookahead == '&') ADVANCE(175);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 154:
      if (lookahead == '&') ADVANCE(571);
      END_STATE();
    case 155:
      if (lookahead == '&') ADVANCE(153);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 156:
      if (lookahead == '\'') ADVANCE(157);
      END_STATE();
    case 157:
      if (lookahead == '\'') ADVANCE(439);
      END_STATE();
    case 158:
      if (lookahead == '*') ADVANCE(175);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 159:
      if (lookahead == '+') ADVANCE(175);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 160:
      if (lookahead == '+') ADVANCE(159);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 161:
      if (lookahead == '-') ADVANCE(175);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 162:
      if (lookahead == '-') ADVANCE(161);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '>') ADVANCE(175);
      END_STATE();
    case 163:
      if (lookahead == '.') ADVANCE(522);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 164:
      if (lookahead == '.') ADVANCE(175);
      if (lookahead == '/') ADVANCE(171);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 165:
      if (lookahead == '.') ADVANCE(523);
      END_STATE();
    case 166:
      if (lookahead == '.') ADVANCE(253);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(166);
      END_STATE();
    case 167:
      if (lookahead == '.') ADVANCE(388);
      END_STATE();
    case 168:
      if (lookahead == '.') ADVANCE(164);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 169:
      if (lookahead == '.') ADVANCE(167);
      END_STATE();
    case 170:
      if (lookahead == '/') ADVANCE(413);
      END_STATE();
    case 171:
      if (lookahead == '/') ADVANCE(175);
      END_STATE();
    case 172:
      if (lookahead == '/') ADVANCE(175);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 173:
      if (lookahead == '0') ADVANCE(403);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(174);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(404);
      END_STATE();
    case 174:
      if (lookahead == '0') ADVANCE(403);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(404);
      END_STATE();
    case 175:
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 176:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '<') ADVANCE(204);
      if (lookahead == '|') ADVANCE(192);
      if (lookahead == '~') ADVANCE(181);
      if (lookahead == '-' ||
          lookahead == '=' ||
          lookahead == '>') ADVANCE(175);
      END_STATE();
    case 177:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(175);
      END_STATE();
    case 178:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(175);
      if (lookahead == '>') ADVANCE(516);
      END_STATE();
    case 179:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(177);
      END_STATE();
    case 180:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(177);
      if (lookahead == '~') ADVANCE(175);
      END_STATE();
    case 181:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '>') ADVANCE(175);
      END_STATE();
    case 182:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '>') ADVANCE(175);
      if (lookahead == '|') ADVANCE(185);
      END_STATE();
    case 183:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '^') ADVANCE(208);
      END_STATE();
    case 184:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '{') ADVANCE(243);
      END_STATE();
    case 185:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '|') ADVANCE(175);
      END_STATE();
    case 186:
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(175);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(186);
      END_STATE();
    case 187:
      if (lookahead == ':') ADVANCE(555);
      END_STATE();
    case 188:
      if (lookahead == ':') ADVANCE(555);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(512);
      END_STATE();
    case 189:
      if (lookahead == '<') ADVANCE(594);
      if (lookahead == '~') ADVANCE(598);
      END_STATE();
    case 190:
      if (lookahead == '=') ADVANCE(581);
      END_STATE();
    case 191:
      if (lookahead == '>') ADVANCE(413);
      END_STATE();
    case 192:
      if (lookahead == '>') ADVANCE(175);
      END_STATE();
    case 193:
      if (lookahead == '>') ADVANCE(429);
      if (lookahead == '~') ADVANCE(245);
      END_STATE();
    case 194:
      if (lookahead == '>') ADVANCE(609);
      END_STATE();
    case 195:
      if (lookahead == '>') ADVANCE(597);
      END_STATE();
    case 196:
      if (lookahead == '>') ADVANCE(632);
      END_STATE();
    case 197:
      if (lookahead == '>') ADVANCE(605);
      END_STATE();
    case 198:
      if (lookahead == '>') ADVANCE(605);
      if (lookahead == '~') ADVANCE(248);
      END_STATE();
    case 199:
      if (lookahead == '>') ADVANCE(608);
      END_STATE();
    case 200:
      if (lookahead == '>') ADVANCE(596);
      END_STATE();
    case 201:
      if (lookahead == '>') ADVANCE(515);
      END_STATE();
    case 202:
      if (lookahead == '>') ADVANCE(181);
      if (lookahead == '~') ADVANCE(246);
      END_STATE();
    case 203:
      if (lookahead == '>') ADVANCE(191);
      if (lookahead == '<' ||
          lookahead == '~') ADVANCE(413);
      END_STATE();
    case 204:
      if (lookahead == '>') ADVANCE(192);
      if (lookahead == '<' ||
          lookahead == '~') ADVANCE(175);
      END_STATE();
    case 205:
      if (lookahead == '\\') ADVANCE(413);
      END_STATE();
    case 206:
      if (lookahead == '\\') ADVANCE(406);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(405);
      END_STATE();
    case 207:
      if (lookahead == '^') ADVANCE(413);
      END_STATE();
    case 208:
      if (lookahead == '^') ADVANCE(175);
      END_STATE();
    case 209:
      if (lookahead == '^') ADVANCE(615);
      END_STATE();
    case 210:
      if (lookahead == '^') ADVANCE(614);
      END_STATE();
    case 211:
      if (lookahead == '^') ADVANCE(210);
      END_STATE();
    case 212:
      if (lookahead == 'a') ADVANCE(236);
      END_STATE();
    case 213:
      if (lookahead == 'c') ADVANCE(239);
      END_STATE();
    case 214:
      if (lookahead == 'c') ADVANCE(223);
      END_STATE();
    case 215:
      if (lookahead == 'd') ADVANCE(574);
      END_STATE();
    case 216:
      if (lookahead == 'd') ADVANCE(652);
      END_STATE();
    case 217:
      if (lookahead == 'e') ADVANCE(648);
      END_STATE();
    case 218:
      if (lookahead == 'e') ADVANCE(658);
      END_STATE();
    case 219:
      if (lookahead == 'e') ADVANCE(228);
      END_STATE();
    case 220:
      if (lookahead == 'e') ADVANCE(233);
      END_STATE();
    case 221:
      if (lookahead == 'e') ADVANCE(234);
      END_STATE();
    case 222:
      if (lookahead == 'f') ADVANCE(238);
      if (lookahead == 'n') ADVANCE(215);
      END_STATE();
    case 223:
      if (lookahead == 'h') ADVANCE(640);
      END_STATE();
    case 224:
      if (lookahead == 'h') ADVANCE(219);
      END_STATE();
    case 225:
      if (lookahead == 'l') ADVANCE(235);
      if (lookahead == 'n') ADVANCE(216);
      END_STATE();
    case 226:
      if (lookahead == 'n') ADVANCE(215);
      END_STATE();
    case 227:
      if (lookahead == 'n') ADVANCE(610);
      END_STATE();
    case 228:
      if (lookahead == 'n') ADVANCE(551);
      END_STATE();
    case 229:
      if (lookahead == 'n') ADVANCE(216);
      END_STATE();
    case 230:
      if (lookahead == 'o') ADVANCE(644);
      END_STATE();
    case 231:
      if (lookahead == 'o') ADVANCE(237);
      END_STATE();
    case 232:
      if (lookahead == 'r') ADVANCE(566);
      END_STATE();
    case 233:
      if (lookahead == 'r') ADVANCE(636);
      END_STATE();
    case 234:
      if (lookahead == 's') ADVANCE(213);
      END_STATE();
    case 235:
      if (lookahead == 's') ADVANCE(217);
      END_STATE();
    case 236:
      if (lookahead == 't') ADVANCE(214);
      END_STATE();
    case 237:
      if (lookahead == 't') ADVANCE(542);
      END_STATE();
    case 238:
      if (lookahead == 't') ADVANCE(220);
      END_STATE();
    case 239:
      if (lookahead == 'u') ADVANCE(218);
      END_STATE();
    case 240:
      if (lookahead == '{') ADVANCE(263);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(465);
      END_STATE();
    case 241:
      if (lookahead == '{') ADVANCE(263);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(265);
      END_STATE();
    case 242:
      if (lookahead == '}') ADVANCE(413);
      END_STATE();
    case 243:
      if (lookahead == '}') ADVANCE(175);
      END_STATE();
    case 244:
      if (lookahead == '}') ADVANCE(463);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(244);
      END_STATE();
    case 245:
      if (lookahead == '~') ADVANCE(413);
      END_STATE();
    case 246:
      if (lookahead == '~') ADVANCE(175);
      END_STATE();
    case 247:
      if (lookahead == '~') ADVANCE(541);
      END_STATE();
    case 248:
      if (lookahead == '~') ADVANCE(540);
      END_STATE();
    case 249:
      if (lookahead == '~') ADVANCE(248);
      END_STATE();
    case 250:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(394);
      END_STATE();
    case 251:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(400);
      END_STATE();
    case 252:
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(511);
      END_STATE();
    case 253:
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(253);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(391);
      END_STATE();
    case 254:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(395);
      END_STATE();
    case 255:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(401);
      END_STATE();
    case 256:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(393);
      END_STATE();
    case 257:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(399);
      END_STATE();
    case 258:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(404);
      END_STATE();
    case 259:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(398);
      END_STATE();
    case 260:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(396);
      END_STATE();
    case 261:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(402);
      END_STATE();
    case 262:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(463);
      END_STATE();
    case 263:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(244);
      END_STATE();
    case 264:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(262);
      END_STATE();
    case 265:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(264);
      END_STATE();
    case 266:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 278,
        '\r', 1,
        '!', 535,
        '"', 436,
        '#', 664,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        ')', 311,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 451,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 4,
        ']', 445,
        '^', 538,
        'a', 327,
        'c', 315,
        'd', 338,
        'e', 331,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        'r', 324,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '}', 443,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(266);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 267:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 281,
        '\r', 78,
        '!', 536,
        '"', 436,
        '#', 665,
        '%', 519,
        '&', 526,
        '\'', 438,
        '(', 310,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 165,
        '/', 458,
        '0', 392,
        ':', 149,
        ';', 309,
        '<', 448,
        '=', 561,
        '>', 454,
        '?', 206,
        '@', 545,
        '[', 444,
        '\\', 6,
        ']', 445,
        '^', 539,
        'a', 371,
        'f', 351,
        'i', 372,
        'n', 367,
        'o', 376,
        't', 377,
        'w', 365,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 467,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(268);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(391);
      if (set_contains(aux_sym_identifier_token1_character_set_1, 1206, lookahead)) ADVANCE(387);
      END_STATE();
    case 268:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 281,
        '\r', 78,
        '!', 536,
        '"', 436,
        '#', 665,
        '%', 519,
        '&', 526,
        '\'', 438,
        '(', 310,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 533,
        '.', 165,
        '/', 458,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 448,
        '=', 561,
        '>', 454,
        '?', 206,
        '@', 545,
        '[', 444,
        '\\', 6,
        ']', 445,
        '^', 539,
        'a', 371,
        'f', 351,
        'i', 372,
        'n', 367,
        'o', 376,
        't', 377,
        'w', 365,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 467,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(268);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(391);
      if (set_contains(aux_sym_identifier_token1_character_set_1, 1206, lookahead)) ADVANCE(387);
      END_STATE();
    case 269:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 284,
        '\r', 85,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 662,
        ')', 311,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 663,
        '\\', 10,
        ']', 445,
        '^', 538,
        'a', 334,
        'd', 338,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '}', 443,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(270);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 270:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 284,
        '\r', 85,
        '!', 535,
        '"', 436,
        '#', 665,
        '%', 520,
        '&', 525,
        '\'', 438,
        '(', 310,
        ')', 311,
        '*', 628,
        '+', 528,
        ',', 510,
        '-', 531,
        '.', 634,
        '/', 459,
        '0', 392,
        ':', 150,
        ';', 309,
        '<', 447,
        '=', 558,
        '>', 452,
        '?', 206,
        '@', 546,
        '[', 444,
        '\\', 10,
        ']', 445,
        '^', 538,
        'a', 334,
        'd', 338,
        'f', 314,
        'i', 335,
        'n', 330,
        'o', 339,
        't', 340,
        'w', 328,
        '{', 442,
        '|', 456,
        '}', 443,
        '~', 466,
        0x2e2f, 387,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(270);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(393);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(390);
      if ((0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xde) ||
          lookahead == 0x100 ||
          lookahead == 0x102 ||
          lookahead == 0x104 ||
          lookahead == 0x106 ||
          lookahead == 0x108 ||
          lookahead == 0x10a ||
          lookahead == 0x10c ||
          lookahead == 0x10e ||
          lookahead == 0x110 ||
          lookahead == 0x112 ||
          lookahead == 0x114 ||
          lookahead == 0x116 ||
          lookahead == 0x118 ||
          lookahead == 0x11a ||
          lookahead == 0x11c ||
          lookahead == 0x11e ||
          lookahead == 0x120 ||
          lookahead == 0x122 ||
          lookahead == 0x124 ||
          lookahead == 0x126 ||
          lookahead == 0x128 ||
          lookahead == 0x12a ||
          lookahead == 0x12c ||
          lookahead == 0x12e ||
          lookahead == 0x130 ||
          lookahead == 0x132 ||
          lookahead == 0x134 ||
          lookahead == 0x136 ||
          lookahead == 0x139 ||
          lookahead == 0x13b ||
          lookahead == 0x13d ||
          lookahead == 0x13f ||
          lookahead == 0x141 ||
          lookahead == 0x143 ||
          lookahead == 0x145 ||
          lookahead == 0x147 ||
          lookahead == 0x14a ||
          lookahead == 0x14c ||
          lookahead == 0x14e ||
          lookahead == 0x150 ||
          lookahead == 0x152 ||
          lookahead == 0x154 ||
          lookahead == 0x156 ||
          lookahead == 0x158 ||
          lookahead == 0x15a ||
          lookahead == 0x15c ||
          lookahead == 0x15e ||
          lookahead == 0x160 ||
          lookahead == 0x162 ||
          lookahead == 0x164 ||
          lookahead == 0x166 ||
          lookahead == 0x168 ||
          lookahead == 0x16a ||
          lookahead == 0x16c ||
          lookahead == 0x16e ||
          lookahead == 0x170 ||
          lookahead == 0x172 ||
          lookahead == 0x174 ||
          lookahead == 0x176 ||
          lookahead == 0x178 ||
          lookahead == 0x179 ||
          lookahead == 0x17b ||
          lookahead == 0x17d ||
          lookahead == 0x181 ||
          lookahead == 0x182 ||
          lookahead == 0x184 ||
          lookahead == 0x186 ||
          lookahead == 0x187 ||
          (0x189 <= lookahead && lookahead <= 0x18b) ||
          (0x18e <= lookahead && lookahead <= 0x191) ||
          lookahead == 0x193 ||
          lookahead == 0x194 ||
          (0x196 <= lookahead && lookahead <= 0x198) ||
          lookahead == 0x19c ||
          lookahead == 0x19d ||
          lookahead == 0x19f ||
          lookahead == 0x1a0 ||
          lookahead == 0x1a2 ||
          lookahead == 0x1a4 ||
          lookahead == 0x1a6 ||
          lookahead == 0x1a7 ||
          lookahead == 0x1a9 ||
          lookahead == 0x1ac ||
          lookahead == 0x1ae ||
          lookahead == 0x1af ||
          (0x1b1 <= lookahead && lookahead <= 0x1b3) ||
          lookahead == 0x1b5 ||
          lookahead == 0x1b7 ||
          lookahead == 0x1b8 ||
          lookahead == 0x1bc ||
          lookahead == 0x1c4 ||
          lookahead == 0x1c5 ||
          lookahead == 0x1c7 ||
          lookahead == 0x1c8 ||
          lookahead == 0x1ca ||
          lookahead == 0x1cb ||
          lookahead == 0x1cd ||
          lookahead == 0x1cf ||
          lookahead == 0x1d1 ||
          lookahead == 0x1d3 ||
          lookahead == 0x1d5 ||
          lookahead == 0x1d7 ||
          lookahead == 0x1d9 ||
          lookahead == 0x1db ||
          lookahead == 0x1de ||
          lookahead == 0x1e0 ||
          lookahead == 0x1e2 ||
          lookahead == 0x1e4 ||
          lookahead == 0x1e6 ||
          lookahead == 0x1e8 ||
          lookahead == 0x1ea ||
          lookahead == 0x1ec ||
          lookahead == 0x1ee ||
          lookahead == 0x1f1 ||
          lookahead == 0x1f2 ||
          lookahead == 0x1f4 ||
          (0x1f6 <= lookahead && lookahead <= 0x1f8) ||
          lookahead == 0x1fa ||
          lookahead == 0x1fc ||
          lookahead == 0x1fe ||
          lookahead == 0x200 ||
          lookahead == 0x202 ||
          lookahead == 0x204 ||
          lookahead == 0x206 ||
          lookahead == 0x208 ||
          lookahead == 0x20a ||
          lookahead == 0x20c ||
          lookahead == 0x20e ||
          lookahead == 0x210 ||
          lookahead == 0x212 ||
          lookahead == 0x214 ||
          lookahead == 0x216 ||
          lookahead == 0x218 ||
          lookahead == 0x21a ||
          lookahead == 0x21c ||
          lookahead == 0x21e ||
          lookahead == 0x220 ||
          lookahead == 0x222 ||
          lookahead == 0x224 ||
          lookahead == 0x226 ||
          lookahead == 0x228 ||
          lookahead == 0x22a ||
          lookahead == 0x22c ||
          lookahead == 0x22e ||
          lookahead == 0x230 ||
          lookahead == 0x232 ||
          lookahead == 0x23a ||
          lookahead == 0x23b ||
          lookahead == 0x23d ||
          lookahead == 0x23e ||
          lookahead == 0x241 ||
          (0x243 <= lookahead && lookahead <= 0x246) ||
          lookahead == 0x248 ||
          lookahead == 0x24a ||
          lookahead == 0x24c ||
          lookahead == 0x24e ||
          lookahead == 0x370 ||
          lookahead == 0x372 ||
          lookahead == 0x376 ||
          lookahead == 0x37f ||
          lookahead == 0x386 ||
          (0x388 <= lookahead && lookahead <= 0x38a) ||
          lookahead == 0x38c ||
          lookahead == 0x38e ||
          lookahead == 0x38f ||
          (0x391 <= lookahead && lookahead <= 0x3a1) ||
          (0x3a3 <= lookahead && lookahead <= 0x3ab) ||
          lookahead == 0x3cf ||
          (0x3d2 <= lookahead && lookahead <= 0x3d4) ||
          lookahead == 0x3d8 ||
          lookahead == 0x3da ||
          lookahead == 0x3dc ||
          lookahead == 0x3de ||
          lookahead == 0x3e0 ||
          lookahead == 0x3e2 ||
          lookahead == 0x3e4 ||
          lookahead == 0x3e6 ||
          lookahead == 0x3e8 ||
          lookahead == 0x3ea ||
          lookahead == 0x3ec ||
          lookahead == 0x3ee ||
          lookahead == 0x3f4 ||
          lookahead == 0x3f7 ||
          lookahead == 0x3f9 ||
          lookahead == 0x3fa ||
          (0x3fd <= lookahead && lookahead <= 0x42f) ||
          lookahead == 0x460 ||
          lookahead == 0x462 ||
          lookahead == 0x464 ||
          lookahead == 0x466 ||
          lookahead == 0x468 ||
          lookahead == 0x46a ||
          lookahead == 0x46c ||
          lookahead == 0x46e ||
          lookahead == 0x470 ||
          lookahead == 0x472 ||
          lookahead == 0x474 ||
          lookahead == 0x476 ||
          lookahead == 0x478 ||
          lookahead == 0x47a ||
          lookahead == 0x47c ||
          lookahead == 0x47e ||
          lookahead == 0x480 ||
          lookahead == 0x48a ||
          lookahead == 0x48c ||
          lookahead == 0x48e ||
          lookahead == 0x490 ||
          lookahead == 0x492 ||
          lookahead == 0x494 ||
          lookahead == 0x496 ||
          lookahead == 0x498 ||
          lookahead == 0x49a ||
          lookahead == 0x49c ||
          lookahead == 0x49e ||
          lookahead == 0x4a0 ||
          lookahead == 0x4a2 ||
          lookahead == 0x4a4 ||
          lookahead == 0x4a6 ||
          lookahead == 0x4a8 ||
          lookahead == 0x4aa ||
          lookahead == 0x4ac ||
          lookahead == 0x4ae ||
          lookahead == 0x4b0 ||
          lookahead == 0x4b2 ||
          lookahead == 0x4b4 ||
          lookahead == 0x4b6 ||
          lookahead == 0x4b8 ||
          lookahead == 0x4ba ||
          lookahead == 0x4bc ||
          lookahead == 0x4be ||
          lookahead == 0x4c0 ||
          lookahead == 0x4c1 ||
          lookahead == 0x4c3 ||
          lookahead == 0x4c5 ||
          lookahead == 0x4c7 ||
          lookahead == 0x4c9 ||
          lookahead == 0x4cb ||
          lookahead == 0x4cd ||
          lookahead == 0x4d0 ||
          lookahead == 0x4d2 ||
          lookahead == 0x4d4 ||
          lookahead == 0x4d6 ||
          lookahead == 0x4d8 ||
          lookahead == 0x4da ||
          lookahead == 0x4dc ||
          lookahead == 0x4de ||
          lookahead == 0x4e0 ||
          lookahead == 0x4e2 ||
          lookahead == 0x4e4 ||
          lookahead == 0x4e6 ||
          lookahead == 0x4e8 ||
          lookahead == 0x4ea ||
          lookahead == 0x4ec ||
          lookahead == 0x4ee ||
          lookahead == 0x4f0 ||
          lookahead == 0x4f2 ||
          lookahead == 0x4f4 ||
          lookahead == 0x4f6 ||
          lookahead == 0x4f8 ||
          lookahead == 0x4fa ||
          lookahead == 0x4fc ||
          lookahead == 0x4fe ||
          lookahead == 0x500 ||
          lookahead == 0x502 ||
          lookahead == 0x504 ||
          lookahead == 0x506 ||
          lookahead == 0x508 ||
          lookahead == 0x50a ||
          lookahead == 0x50c ||
          lookahead == 0x50e ||
          lookahead == 0x510 ||
          lookahead == 0x512 ||
          lookahead == 0x514 ||
          lookahead == 0x516 ||
          lookahead == 0x518 ||
          lookahead == 0x51a ||
          lookahead == 0x51c ||
          lookahead == 0x51e ||
          lookahead == 0x520 ||
          lookahead == 0x522 ||
          lookahead == 0x524 ||
          lookahead == 0x526 ||
          lookahead == 0x528 ||
          lookahead == 0x52a ||
          lookahead == 0x52c ||
          lookahead == 0x52e ||
          (0x531 <= lookahead && lookahead <= 0x556) ||
          (0x10a0 <= lookahead && lookahead <= 0x10c5) ||
          lookahead == 0x10c7 ||
          lookahead == 0x10cd ||
          (0x13a0 <= lookahead && lookahead <= 0x13f5) ||
          lookahead == 0x1c89 ||
          (0x1c90 <= lookahead && lookahead <= 0x1cba) ||
          (0x1cbd <= lookahead && lookahead <= 0x1cbf) ||
          lookahead == 0x1e00 ||
          lookahead == 0x1e02 ||
          lookahead == 0x1e04 ||
          lookahead == 0x1e06 ||
          lookahead == 0x1e08 ||
          lookahead == 0x1e0a ||
          lookahead == 0x1e0c ||
          lookahead == 0x1e0e ||
          lookahead == 0x1e10 ||
          lookahead == 0x1e12 ||
          lookahead == 0x1e14 ||
          lookahead == 0x1e16 ||
          lookahead == 0x1e18 ||
          lookahead == 0x1e1a ||
          lookahead == 0x1e1c ||
          lookahead == 0x1e1e ||
          lookahead == 0x1e20 ||
          lookahead == 0x1e22 ||
          lookahead == 0x1e24 ||
          lookahead == 0x1e26 ||
          lookahead == 0x1e28 ||
          lookahead == 0x1e2a ||
          lookahead == 0x1e2c ||
          lookahead == 0x1e2e ||
          lookahead == 0x1e30 ||
          lookahead == 0x1e32 ||
          lookahead == 0x1e34 ||
          lookahead == 0x1e36 ||
          lookahead == 0x1e38 ||
          lookahead == 0x1e3a ||
          lookahead == 0x1e3c ||
          lookahead == 0x1e3e ||
          lookahead == 0x1e40 ||
          lookahead == 0x1e42 ||
          lookahead == 0x1e44 ||
          lookahead == 0x1e46 ||
          lookahead == 0x1e48 ||
          lookahead == 0x1e4a ||
          lookahead == 0x1e4c ||
          lookahead == 0x1e4e ||
          lookahead == 0x1e50 ||
          lookahead == 0x1e52 ||
          lookahead == 0x1e54 ||
          lookahead == 0x1e56 ||
          lookahead == 0x1e58 ||
          lookahead == 0x1e5a ||
          lookahead == 0x1e5c ||
          lookahead == 0x1e5e ||
          lookahead == 0x1e60 ||
          lookahead == 0x1e62 ||
          lookahead == 0x1e64 ||
          lookahead == 0x1e66 ||
          lookahead == 0x1e68 ||
          lookahead == 0x1e6a ||
          lookahead == 0x1e6c ||
          lookahead == 0x1e6e ||
          lookahead == 0x1e70 ||
          lookahead == 0x1e72 ||
          lookahead == 0x1e74 ||
          lookahead == 0x1e76 ||
          lookahead == 0x1e78 ||
          lookahead == 0x1e7a ||
          lookahead == 0x1e7c ||
          lookahead == 0x1e7e ||
          lookahead == 0x1e80 ||
          lookahead == 0x1e82 ||
          lookahead == 0x1e84 ||
          lookahead == 0x1e86 ||
          lookahead == 0x1e88 ||
          lookahead == 0x1e8a ||
          lookahead == 0x1e8c ||
          lookahead == 0x1e8e ||
          lookahead == 0x1e90 ||
          lookahead == 0x1e92 ||
          lookahead == 0x1e94 ||
          lookahead == 0x1e9e ||
          lookahead == 0x1ea0 ||
          lookahead == 0x1ea2 ||
          lookahead == 0x1ea4 ||
          lookahead == 0x1ea6 ||
          lookahead == 0x1ea8 ||
          lookahead == 0x1eaa ||
          lookahead == 0x1eac ||
          lookahead == 0x1eae ||
          lookahead == 0x1eb0 ||
          lookahead == 0x1eb2 ||
          lookahead == 0x1eb4 ||
          lookahead == 0x1eb6 ||
          lookahead == 0x1eb8 ||
          lookahead == 0x1eba ||
          lookahead == 0x1ebc ||
          lookahead == 0x1ebe ||
          lookahead == 0x1ec0 ||
          lookahead == 0x1ec2 ||
          lookahead == 0x1ec4 ||
          lookahead == 0x1ec6 ||
          lookahead == 0x1ec8 ||
          lookahead == 0x1eca ||
          lookahead == 0x1ecc ||
          lookahead == 0x1ece ||
          lookahead == 0x1ed0 ||
          lookahead == 0x1ed2 ||
          lookahead == 0x1ed4 ||
          lookahead == 0x1ed6 ||
          lookahead == 0x1ed8 ||
          lookahead == 0x1eda ||
          lookahead == 0x1edc ||
          lookahead == 0x1ede ||
          lookahead == 0x1ee0 ||
          lookahead == 0x1ee2 ||
          lookahead == 0x1ee4 ||
          lookahead == 0x1ee6 ||
          lookahead == 0x1ee8 ||
          lookahead == 0x1eea ||
          lookahead == 0x1eec ||
          lookahead == 0x1eee ||
          lookahead == 0x1ef0 ||
          lookahead == 0x1ef2 ||
          lookahead == 0x1ef4 ||
          lookahead == 0x1ef6 ||
          lookahead == 0x1ef8 ||
          lookahead == 0x1efa ||
          lookahead == 0x1efc ||
          lookahead == 0x1efe ||
          (0x1f08 <= lookahead && lookahead <= 0x1f0f) ||
          (0x1f18 <= lookahead && lookahead <= 0x1f1d) ||
          (0x1f28 <= lookahead && lookahead <= 0x1f2f) ||
          (0x1f38 <= lookahead && lookahead <= 0x1f3f) ||
          (0x1f48 <= lookahead && lookahead <= 0x1f4d) ||
          lookahead == 0x1f59 ||
          lookahead == 0x1f5b ||
          lookahead == 0x1f5d ||
          lookahead == 0x1f5f ||
          (0x1f68 <= lookahead && lookahead <= 0x1f6f) ||
          (0x1f88 <= lookahead && lookahead <= 0x1f8f) ||
          (0x1f98 <= lookahead && lookahead <= 0x1f9f) ||
          (0x1fa8 <= lookahead && lookahead <= 0x1faf) ||
          (0x1fb8 <= lookahead && lookahead <= 0x1fbc) ||
          (0x1fc8 <= lookahead && lookahead <= 0x1fcc) ||
          (0x1fd8 <= lookahead && lookahead <= 0x1fdb) ||
          (0x1fe8 <= lookahead && lookahead <= 0x1fec) ||
          (0x1ff8 <= lookahead && lookahead <= 0x1ffc) ||
          lookahead == 0x2102 ||
          lookahead == 0x2107 ||
          (0x210b <= lookahead && lookahead <= 0x210d) ||
          (0x2110 <= lookahead && lookahead <= 0x2112) ||
          lookahead == 0x2115 ||
          (0x2119 <= lookahead && lookahead <= 0x211d) ||
          lookahead == 0x2124 ||
          lookahead == 0x2126 ||
          lookahead == 0x2128 ||
          (0x212a <= lookahead && lookahead <= 0x212d) ||
          (0x2130 <= lookahead && lookahead <= 0x2133) ||
          lookahead == 0x213e ||
          lookahead == 0x213f ||
          lookahead == 0x2145 ||
          lookahead == 0x2183 ||
          (0x2c00 <= lookahead && lookahead <= 0x2c2f) ||
          lookahead == 0x2c60 ||
          (0x2c62 <= lookahead && lookahead <= 0x2c64) ||
          lookahead == 0x2c67 ||
          lookahead == 0x2c69 ||
          lookahead == 0x2c6b ||
          (0x2c6d <= lookahead && lookahead <= 0x2c70) ||
          lookahead == 0x2c72 ||
          lookahead == 0x2c75 ||
          (0x2c7e <= lookahead && lookahead <= 0x2c80) ||
          lookahead == 0x2c82 ||
          lookahead == 0x2c84 ||
          lookahead == 0x2c86 ||
          lookahead == 0x2c88 ||
          lookahead == 0x2c8a ||
          lookahead == 0x2c8c ||
          lookahead == 0x2c8e ||
          lookahead == 0x2c90 ||
          lookahead == 0x2c92 ||
          lookahead == 0x2c94 ||
          lookahead == 0x2c96 ||
          lookahead == 0x2c98 ||
          lookahead == 0x2c9a ||
          lookahead == 0x2c9c ||
          lookahead == 0x2c9e ||
          lookahead == 0x2ca0 ||
          lookahead == 0x2ca2 ||
          lookahead == 0x2ca4 ||
          lookahead == 0x2ca6 ||
          lookahead == 0x2ca8 ||
          lookahead == 0x2caa ||
          lookahead == 0x2cac ||
          lookahead == 0x2cae ||
          lookahead == 0x2cb0 ||
          lookahead == 0x2cb2 ||
          lookahead == 0x2cb4 ||
          lookahead == 0x2cb6 ||
          lookahead == 0x2cb8 ||
          lookahead == 0x2cba ||
          lookahead == 0x2cbc ||
          lookahead == 0x2cbe ||
          lookahead == 0x2cc0 ||
          lookahead == 0x2cc2 ||
          lookahead == 0x2cc4 ||
          lookahead == 0x2cc6 ||
          lookahead == 0x2cc8 ||
          lookahead == 0x2cca ||
          lookahead == 0x2ccc ||
          lookahead == 0x2cce ||
          lookahead == 0x2cd0 ||
          lookahead == 0x2cd2 ||
          lookahead == 0x2cd4 ||
          lookahead == 0x2cd6 ||
          lookahead == 0x2cd8 ||
          lookahead == 0x2cda ||
          lookahead == 0x2cdc ||
          lookahead == 0x2cde ||
          lookahead == 0x2ce0 ||
          lookahead == 0x2ce2 ||
          lookahead == 0x2ceb ||
          lookahead == 0x2ced ||
          lookahead == 0x2cf2 ||
          lookahead == 0xa640 ||
          lookahead == 0xa642 ||
          lookahead == 0xa644 ||
          lookahead == 0xa646 ||
          lookahead == 0xa648 ||
          lookahead == 0xa64a ||
          lookahead == 0xa64c ||
          lookahead == 0xa64e ||
          lookahead == 0xa650 ||
          lookahead == 0xa652 ||
          lookahead == 0xa654 ||
          lookahead == 0xa656 ||
          lookahead == 0xa658 ||
          lookahead == 0xa65a ||
          lookahead == 0xa65c ||
          lookahead == 0xa65e ||
          lookahead == 0xa660 ||
          lookahead == 0xa662 ||
          lookahead == 0xa664 ||
          lookahead == 0xa666 ||
          lookahead == 0xa668 ||
          lookahead == 0xa66a ||
          lookahead == 0xa66c ||
          lookahead == 0xa680 ||
          lookahead == 0xa682 ||
          lookahead == 0xa684 ||
          lookahead == 0xa686 ||
          lookahead == 0xa688 ||
          lookahead == 0xa68a ||
          lookahead == 0xa68c ||
          lookahead == 0xa68e ||
          lookahead == 0xa690 ||
          lookahead == 0xa692 ||
          lookahead == 0xa694 ||
          lookahead == 0xa696 ||
          lookahead == 0xa698 ||
          lookahead == 0xa69a ||
          lookahead == 0xa722 ||
          lookahead == 0xa724 ||
          lookahead == 0xa726 ||
          lookahead == 0xa728 ||
          lookahead == 0xa72a ||
          lookahead == 0xa72c ||
          lookahead == 0xa72e ||
          lookahead == 0xa732 ||
          lookahead == 0xa734 ||
          lookahead == 0xa736 ||
          lookahead == 0xa738 ||
          lookahead == 0xa73a ||
          lookahead == 0xa73c ||
          lookahead == 0xa73e ||
          lookahead == 0xa740 ||
          lookahead == 0xa742 ||
          lookahead == 0xa744 ||
          lookahead == 0xa746 ||
          lookahead == 0xa748 ||
          lookahead == 0xa74a ||
          lookahead == 0xa74c ||
          lookahead == 0xa74e ||
          lookahead == 0xa750 ||
          lookahead == 0xa752 ||
          lookahead == 0xa754 ||
          lookahead == 0xa756 ||
          lookahead == 0xa758 ||
          lookahead == 0xa75a ||
          lookahead == 0xa75c ||
          lookahead == 0xa75e ||
          lookahead == 0xa760 ||
          lookahead == 0xa762 ||
          lookahead == 0xa764 ||
          lookahead == 0xa766 ||
          lookahead == 0xa768 ||
          lookahead == 0xa76a ||
          lookahead == 0xa76c ||
          lookahead == 0xa76e ||
          lookahead == 0xa779 ||
          lookahead == 0xa77b ||
          lookahead == 0xa77d ||
          lookahead == 0xa77e ||
          lookahead == 0xa780 ||
          lookahead == 0xa782 ||
          lookahead == 0xa784 ||
          lookahead == 0xa786 ||
          lookahead == 0xa78b ||
          lookahead == 0xa78d ||
          lookahead == 0xa790 ||
          lookahead == 0xa792 ||
          lookahead == 0xa796 ||
          lookahead == 0xa798 ||
          lookahead == 0xa79a ||
          lookahead == 0xa79c ||
          lookahead == 0xa79e ||
          lookahead == 0xa7a0 ||
          lookahead == 0xa7a2 ||
          lookahead == 0xa7a4 ||
          lookahead == 0xa7a6 ||
          lookahead == 0xa7a8 ||
          (0xa7aa <= lookahead && lookahead <= 0xa7ae) ||
          (0xa7b0 <= lookahead && lookahead <= 0xa7b4) ||
          lookahead == 0xa7b6 ||
          lookahead == 0xa7b8 ||
          lookahead == 0xa7ba ||
          lookahead == 0xa7bc ||
          lookahead == 0xa7be ||
          lookahead == 0xa7c0 ||
          lookahead == 0xa7c2 ||
          (0xa7c4 <= lookahead && lookahead <= 0xa7c7) ||
          lookahead == 0xa7c9 ||
          lookahead == 0xa7cb ||
          lookahead == 0xa7cc ||
          lookahead == 0xa7d0 ||
          lookahead == 0xa7d6 ||
          lookahead == 0xa7d8 ||
          lookahead == 0xa7da ||
          lookahead == 0xa7dc ||
          lookahead == 0xa7f5 ||
          (0xff21 <= lookahead && lookahead <= 0xff3a) ||
          (0x10400 <= lookahead && lookahead <= 0x10427) ||
          (0x104b0 <= lookahead && lookahead <= 0x104d3) ||
          (0x10570 <= lookahead && lookahead <= 0x1057a) ||
          (0x1057c <= lookahead && lookahead <= 0x1058a) ||
          (0x1058c <= lookahead && lookahead <= 0x10592) ||
          lookahead == 0x10594 ||
          lookahead == 0x10595 ||
          (0x10c80 <= lookahead && lookahead <= 0x10cb2) ||
          (0x10d50 <= lookahead && lookahead <= 0x10d65) ||
          (0x118a0 <= lookahead && lookahead <= 0x118bf) ||
          (0x16e40 <= lookahead && lookahead <= 0x16e5f) ||
          (0x1d400 <= lookahead && lookahead <= 0x1d419) ||
          (0x1d434 <= lookahead && lookahead <= 0x1d44d) ||
          (0x1d468 <= lookahead && lookahead <= 0x1d481) ||
          lookahead == 0x1d49c ||
          lookahead == 0x1d49e ||
          lookahead == 0x1d49f ||
          lookahead == 0x1d4a2 ||
          lookahead == 0x1d4a5 ||
          lookahead == 0x1d4a6 ||
          (0x1d4a9 <= lookahead && lookahead <= 0x1d4ac) ||
          (0x1d4ae <= lookahead && lookahead <= 0x1d4b5) ||
          (0x1d4d0 <= lookahead && lookahead <= 0x1d4e9) ||
          lookahead == 0x1d504 ||
          lookahead == 0x1d505 ||
          (0x1d507 <= lookahead && lookahead <= 0x1d50a) ||
          (0x1d50d <= lookahead && lookahead <= 0x1d514) ||
          (0x1d516 <= lookahead && lookahead <= 0x1d51c) ||
          lookahead == 0x1d538 ||
          lookahead == 0x1d539 ||
          (0x1d53b <= lookahead && lookahead <= 0x1d53e) ||
          (0x1d540 <= lookahead && lookahead <= 0x1d544) ||
          lookahead == 0x1d546 ||
          (0x1d54a <= lookahead && lookahead <= 0x1d550) ||
          (0x1d56c <= lookahead && lookahead <= 0x1d585) ||
          (0x1d5a0 <= lookahead && lookahead <= 0x1d5b9) ||
          (0x1d5d4 <= lookahead && lookahead <= 0x1d5ed) ||
          (0x1d608 <= lookahead && lookahead <= 0x1d621) ||
          (0x1d63c <= lookahead && lookahead <= 0x1d655) ||
          (0x1d670 <= lookahead && lookahead <= 0x1d689) ||
          (0x1d6a8 <= lookahead && lookahead <= 0x1d6c0) ||
          (0x1d6e2 <= lookahead && lookahead <= 0x1d6fa) ||
          (0x1d71c <= lookahead && lookahead <= 0x1d734) ||
          (0x1d756 <= lookahead && lookahead <= 0x1d76e) ||
          (0x1d790 <= lookahead && lookahead <= 0x1d7a8) ||
          lookahead == 0x1d7ca ||
          (0x1e900 <= lookahead && lookahead <= 0x1e921)) ADVANCE(186);
      if (set_contains(sym_keyword_character_set_1, 685, lookahead)) ADVANCE(350);
      END_STATE();
    case 271:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 294,
        '\r', 108,
        '!', 190,
        '#', 665,
        '&', 154,
        '(', 310,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 188,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 26,
        ']', 445,
        '^', 211,
        'a', 222,
        'c', 212,
        'd', 230,
        'e', 225,
        'i', 227,
        'o', 232,
        'r', 221,
        'w', 224,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(272);
      END_STATE();
    case 272:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 294,
        '\r', 108,
        '!', 190,
        '#', 665,
        '&', 154,
        '(', 310,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 26,
        ']', 445,
        '^', 211,
        'a', 222,
        'c', 212,
        'd', 230,
        'e', 225,
        'i', 227,
        'o', 232,
        'r', 221,
        'w', 224,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(272);
      END_STATE();
    case 273:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 299,
        '\r', 122,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 32,
        ']', 445,
        '^', 211,
        'a', 497,
        'd', 501,
        'i', 498,
        'o', 502,
        'w', 495,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(274);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 274:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 299,
        '\r', 122,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 32,
        ']', 445,
        '^', 211,
        'a', 226,
        'd', 230,
        'i', 227,
        'o', 232,
        'w', 224,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(274);
      END_STATE();
    case 275:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 301,
        '\r', 127,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '[', 663,
        '\\', 36,
        ']', 445,
        '^', 211,
        'a', 497,
        'i', 498,
        'o', 502,
        'w', 495,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(276);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 276:
      if (eof) ADVANCE(277);
      ADVANCE_MAP(
        '\n', 301,
        '\r', 127,
        '!', 190,
        '#', 665,
        '&', 154,
        ')', 311,
        '*', 629,
        '+', 529,
        ',', 510,
        '-', 532,
        '.', 635,
        '/', 461,
        ':', 187,
        ';', 309,
        '<', 449,
        '=', 560,
        '>', 454,
        '\\', 36,
        ']', 445,
        '^', 211,
        'a', 226,
        'i', 227,
        'o', 232,
        'w', 224,
        '{', 441,
        '|', 457,
        '}', 443,
        '~', 197,
      );
      if (lookahead == '\t' ||
          lookahead == ' ') SKIP(276);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(278);
      if (lookahead == '\r') ADVANCE(1);
      if (lookahead == '\\') ADVANCE(4);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(279);
      if (lookahead == '\r') ADVANCE(51);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(280);
      if (lookahead == '\r') ADVANCE(63);
      if (lookahead == '\\') ADVANCE(60);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(281);
      if (lookahead == '\r') ADVANCE(78);
      if (lookahead == '\\') ADVANCE(6);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(282);
      if (lookahead == '\r') ADVANCE(80);
      if (lookahead == '\\') ADVANCE(8);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(283);
      if (lookahead == '\r') ADVANCE(82);
      if (lookahead == '\\') ADVANCE(69);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(284);
      if (lookahead == '\r') ADVANCE(85);
      if (lookahead == '\\') ADVANCE(10);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(285);
      if (lookahead == '\r') ADVANCE(87);
      if (lookahead == '\\') ADVANCE(71);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(286);
      if (lookahead == '\r') ADVANCE(90);
      if (lookahead == '\\') ADVANCE(73);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(287);
      if (lookahead == '\r') ADVANCE(93);
      if (lookahead == '\\') ADVANCE(12);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(288);
      if (lookahead == '\r') ADVANCE(95);
      if (lookahead == '\\') ADVANCE(14);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(289);
      if (lookahead == '\r') ADVANCE(97);
      if (lookahead == '\\') ADVANCE(16);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(290);
      if (lookahead == '\r') ADVANCE(99);
      if (lookahead == '\\') ADVANCE(18);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(291);
      if (lookahead == '\r') ADVANCE(101);
      if (lookahead == '\\') ADVANCE(20);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(292);
      if (lookahead == '\r') ADVANCE(103);
      if (lookahead == '\\') ADVANCE(22);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(293);
      if (lookahead == '\r') ADVANCE(105);
      if (lookahead == '\\') ADVANCE(24);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(294);
      if (lookahead == '\r') ADVANCE(108);
      if (lookahead == '\\') ADVANCE(26);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(295);
      if (lookahead == '\r') ADVANCE(110);
      if (lookahead == '\\') ADVANCE(28);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(296);
      if (lookahead == '\r') ADVANCE(113);
      if (lookahead == '\\') ADVANCE(75);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(297);
      if (lookahead == '\r') ADVANCE(116);
      if (lookahead == '\\') ADVANCE(30);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(298);
      if (lookahead == '\r') ADVANCE(119);
      if (lookahead == '\\') ADVANCE(77);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(299);
      if (lookahead == '\r') ADVANCE(122);
      if (lookahead == '\\') ADVANCE(32);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(300);
      if (lookahead == '\r') ADVANCE(124);
      if (lookahead == '\\') ADVANCE(34);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(301);
      if (lookahead == '\r') ADVANCE(127);
      if (lookahead == '\\') ADVANCE(36);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(302);
      if (lookahead == '\r') ADVANCE(129);
      if (lookahead == '\\') ADVANCE(38);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(303);
      if (lookahead == '\r') ADVANCE(132);
      if (lookahead == '\\') ADVANCE(40);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(304);
      if (lookahead == '\r') ADVANCE(135);
      if (lookahead == '\\') ADVANCE(42);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(305);
      if (lookahead == '\r') ADVANCE(138);
      if (lookahead == '\\') ADVANCE(44);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(306);
      if (lookahead == '\r') ADVANCE(141);
      if (lookahead == '\\') ADVANCE(46);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(307);
      if (lookahead == '\r') ADVANCE(144);
      if (lookahead == '\\') ADVANCE(48);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(aux_sym__terminator_token1);
      if (lookahead == '\n') ADVANCE(308);
      if (lookahead == '\r') ADVANCE(147);
      if (lookahead == '\\') ADVANCE(57);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(333);
      if (lookahead == 'n') ADVANCE(656);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(329);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'c') ADVANCE(349);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'd') ADVANCE(575);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'd') ADVANCE(653);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'e') ADVANCE(649);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'e') ADVANCE(407);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'e') ADVANCE(409);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'e') ADVANCE(659);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'e') ADVANCE(342);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'e') ADVANCE(336);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'e') ADVANCE(341);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'f') ADVANCE(347);
      if (lookahead == 'n') ADVANCE(318);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'h') ADVANCE(325);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'h') ADVANCE(641);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'i') ADVANCE(332);
      if (lookahead == 'o') ADVANCE(346);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'l') ADVANCE(343);
      if (lookahead == 'n') ADVANCE(319);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'l') ADVANCE(411);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'l') ADVANCE(344);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'n') ADVANCE(318);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'n') ADVANCE(611);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'n') ADVANCE(552);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'n') ADVANCE(319);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'o') ADVANCE(645);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'r') ADVANCE(567);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'r') ADVANCE(348);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'r') ADVANCE(637);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 's') ADVANCE(317);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 's') ADVANCE(320);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 's') ADVANCE(322);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 't') ADVANCE(316);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 't') ADVANCE(543);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 't') ADVANCE(326);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'u') ADVANCE(321);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'u') ADVANCE(323);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'a') ADVANCE(368);
      if (lookahead == 'n') ADVANCE(657);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'a') ADVANCE(383);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'c') ADVANCE(366);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'c') ADVANCE(386);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'd') ADVANCE(576);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 356:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'd') ADVANCE(654);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 357:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'e') ADVANCE(408);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 358:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'e') ADVANCE(410);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'e') ADVANCE(650);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 360:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'e') ADVANCE(660);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 361:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'e') ADVANCE(379);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 362:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'e') ADVANCE(373);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 363:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'e') ADVANCE(378);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 364:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'f') ADVANCE(384);
      if (lookahead == 'n') ADVANCE(355);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 365:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'h') ADVANCE(362);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 366:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'h') ADVANCE(642);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 367:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'i') ADVANCE(369);
      if (lookahead == 'o') ADVANCE(382);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 368:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'l') ADVANCE(380);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 369:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'l') ADVANCE(412);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 370:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'l') ADVANCE(381);
      if (lookahead == 'n') ADVANCE(356);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 371:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'n') ADVANCE(355);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 372:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'n') ADVANCE(612);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 373:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'n') ADVANCE(553);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 374:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'n') ADVANCE(356);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 375:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'o') ADVANCE(646);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 376:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'r') ADVANCE(568);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 377:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'r') ADVANCE(385);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 378:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'r') ADVANCE(638);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 379:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 's') ADVANCE(354);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 380:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 's') ADVANCE(358);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 381:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 's') ADVANCE(359);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 382:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 't') ADVANCE(544);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 383:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 't') ADVANCE(353);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 384:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 't') ADVANCE(363);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 385:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'u') ADVANCE(357);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 386:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == 'u') ADVANCE(360);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 387:
      ACCEPT_TOKEN(aux_sym_identifier_token1);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 388:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      END_STATE();
    case 389:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 390:
      ACCEPT_TOKEN(sym_alias);
      if (lookahead == '.') ADVANCE(253);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(175);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(166);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(390);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(186);
      END_STATE();
    case 391:
      ACCEPT_TOKEN(sym_alias);
      if (lookahead == '.') ADVANCE(253);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(166);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(391);
      END_STATE();
    case 392:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '.') ADVANCE(257);
      if (lookahead == '_') ADVANCE(256);
      if (lookahead == 'b') ADVANCE(250);
      if (lookahead == 'o') ADVANCE(254);
      if (lookahead == 'x') ADVANCE(260);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(393);
      END_STATE();
    case 393:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '.') ADVANCE(257);
      if (lookahead == '_') ADVANCE(256);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(393);
      END_STATE();
    case 394:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(250);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(394);
      END_STATE();
    case 395:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(254);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(395);
      END_STATE();
    case 396:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(260);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(396);
      END_STATE();
    case 397:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(259);
      if (lookahead == 'b') ADVANCE(250);
      if (lookahead == 'o') ADVANCE(254);
      if (lookahead == 'x') ADVANCE(260);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(398);
      END_STATE();
    case 398:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(259);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(398);
      END_STATE();
    case 399:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(257);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(173);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(399);
      END_STATE();
    case 400:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(251);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(400);
      END_STATE();
    case 401:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(255);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(401);
      END_STATE();
    case 402:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(402);
      END_STATE();
    case 403:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(258);
      if (lookahead == 'b') ADVANCE(251);
      if (lookahead == 'o') ADVANCE(255);
      if (lookahead == 'x') ADVANCE(261);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(404);
      END_STATE();
    case 404:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(258);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(404);
      END_STATE();
    case 405:
      ACCEPT_TOKEN(sym_char);
      END_STATE();
    case 406:
      ACCEPT_TOKEN(sym_char);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(405);
      END_STATE();
    case 407:
      ACCEPT_TOKEN(anon_sym_true);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 408:
      ACCEPT_TOKEN(anon_sym_true);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 409:
      ACCEPT_TOKEN(anon_sym_false);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 410:
      ACCEPT_TOKEN(anon_sym_false);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 411:
      ACCEPT_TOKEN(anon_sym_nil);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 412:
      ACCEPT_TOKEN(anon_sym_nil);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 413:
      ACCEPT_TOKEN(sym_atom);
      END_STATE();
    case 414:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '&') ADVANCE(413);
      END_STATE();
    case 415:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '&') ADVANCE(414);
      END_STATE();
    case 416:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '*') ADVANCE(413);
      END_STATE();
    case 417:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '+') ADVANCE(413);
      END_STATE();
    case 418:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '+') ADVANCE(417);
      END_STATE();
    case 419:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '-') ADVANCE(413);
      END_STATE();
    case 420:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '-') ADVANCE(419);
      if (lookahead == '>') ADVANCE(413);
      END_STATE();
    case 421:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '.') ADVANCE(413);
      if (lookahead == '/') ADVANCE(170);
      END_STATE();
    case 422:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '.') ADVANCE(421);
      END_STATE();
    case 423:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '/') ADVANCE(413);
      END_STATE();
    case 424:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '<') ADVANCE(203);
      if (lookahead == '|') ADVANCE(191);
      if (lookahead == '~') ADVANCE(429);
      if (lookahead == '-' ||
          lookahead == '=' ||
          lookahead == '>') ADVANCE(413);
      END_STATE();
    case 425:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '=') ADVANCE(413);
      END_STATE();
    case 426:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '=') ADVANCE(413);
      if (lookahead == '>') ADVANCE(191);
      END_STATE();
    case 427:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '=') ADVANCE(425);
      END_STATE();
    case 428:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '=') ADVANCE(425);
      if (lookahead == '~') ADVANCE(413);
      END_STATE();
    case 429:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '>') ADVANCE(413);
      END_STATE();
    case 430:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '>') ADVANCE(413);
      if (lookahead == '|') ADVANCE(433);
      END_STATE();
    case 431:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '^') ADVANCE(207);
      END_STATE();
    case 432:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '{') ADVANCE(242);
      END_STATE();
    case 433:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '|') ADVANCE(413);
      END_STATE();
    case 434:
      ACCEPT_TOKEN(sym_atom);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(413);
      if (set_contains(sym_atom_character_set_2, 794, lookahead)) ADVANCE(434);
      END_STATE();
    case 435:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 436:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      if (lookahead == '"') ADVANCE(152);
      END_STATE();
    case 437:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 438:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      if (lookahead == '\'') ADVANCE(157);
      END_STATE();
    case 439:
      ACCEPT_TOKEN(anon_sym_SQUOTE_SQUOTE_SQUOTE);
      END_STATE();
    case 440:
      ACCEPT_TOKEN(anon_sym_DQUOTE_DQUOTE_DQUOTE);
      END_STATE();
    case 441:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 442:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '}') ADVANCE(175);
      END_STATE();
    case 443:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 444:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 445:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 446:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 447:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '-') ADVANCE(548);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '<') ADVANCE(513);
      if (lookahead == '=') ADVANCE(589);
      if (lookahead == '>') ADVANCE(627);
      if (lookahead == '|') ADVANCE(194);
      if (lookahead == '~') ADVANCE(602);
      END_STATE();
    case 448:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '-') ADVANCE(547);
      if (lookahead == '<') ADVANCE(514);
      if (lookahead == '=') ADVANCE(588);
      if (lookahead == '>') ADVANCE(626);
      if (lookahead == '|') ADVANCE(199);
      if (lookahead == '~') ADVANCE(603);
      END_STATE();
    case 449:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '-') ADVANCE(547);
      if (lookahead == '<') ADVANCE(189);
      if (lookahead == '=') ADVANCE(588);
      if (lookahead == '>') ADVANCE(626);
      if (lookahead == '|') ADVANCE(199);
      if (lookahead == '~') ADVANCE(603);
      END_STATE();
    case 450:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 451:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(591);
      if (lookahead == '>') ADVANCE(517);
      END_STATE();
    case 452:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(591);
      if (lookahead == '>') ADVANCE(195);
      END_STATE();
    case 453:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(590);
      if (lookahead == '>') ADVANCE(518);
      END_STATE();
    case 454:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(590);
      if (lookahead == '>') ADVANCE(200);
      END_STATE();
    case 455:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 456:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '>') ADVANCE(593);
      if (lookahead == '|') ADVANCE(562);
      END_STATE();
    case 457:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '>') ADVANCE(592);
      if (lookahead == '|') ADVANCE(563);
      END_STATE();
    case 458:
      ACCEPT_TOKEN(anon_sym_SLASH);
      END_STATE();
    case 459:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(617);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 460:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(175);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 461:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(616);
      END_STATE();
    case 462:
      ACCEPT_TOKEN(anon_sym_POUND_LBRACE);
      END_STATE();
    case 463:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 464:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (lookahead == '\n') ADVANCE(280);
      if (lookahead == '\r') ADVANCE(63);
      if (lookahead == '\\') ADVANCE(60);
      END_STATE();
    case 465:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(463);
      END_STATE();
    case 466:
      ACCEPT_TOKEN(anon_sym_TILDE);
      if (lookahead == '>') ADVANCE(604);
      if (lookahead == '~') ADVANCE(247);
      END_STATE();
    case 467:
      ACCEPT_TOKEN(anon_sym_TILDE);
      if (lookahead == '>') ADVANCE(605);
      if (lookahead == '~') ADVANCE(248);
      END_STATE();
    case 468:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      END_STATE();
    case 469:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(333);
      if (lookahead == 'n') ADVANCE(656);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 470:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 471:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'e') ADVANCE(342);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 472:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'f') ADVANCE(347);
      if (lookahead == 'n') ADVANCE(318);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 473:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'h') ADVANCE(325);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 474:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'i') ADVANCE(332);
      if (lookahead == 'o') ADVANCE(346);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 475:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'l') ADVANCE(343);
      if (lookahead == 'n') ADVANCE(319);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 476:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'n') ADVANCE(611);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 477:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'o') ADVANCE(645);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 478:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'r') ADVANCE(567);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 479:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == 'r') ADVANCE(348);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 480:
      ACCEPT_TOKEN(aux_sym_sigil_token1);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 481:
      ACCEPT_TOKEN(aux_sym_sigil_token2);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(175);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z')) ADVANCE(481);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(186);
      END_STATE();
    case 482:
      ACCEPT_TOKEN(aux_sym_sigil_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z')) ADVANCE(482);
      END_STATE();
    case 483:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'a') ADVANCE(506);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 484:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'c') ADVANCE(508);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 485:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'c') ADVANCE(494);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 486:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'd') ADVANCE(577);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 487:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'd') ADVANCE(655);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 488:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'e') ADVANCE(651);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 489:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'e') ADVANCE(661);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 490:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'e') ADVANCE(499);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 491:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'e') ADVANCE(503);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 492:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'e') ADVANCE(504);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 493:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'f') ADVANCE(507);
      if (lookahead == 'n') ADVANCE(486);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 494:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'h') ADVANCE(643);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 495:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'h') ADVANCE(490);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 496:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'l') ADVANCE(505);
      if (lookahead == 'n') ADVANCE(487);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 497:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'n') ADVANCE(486);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 498:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'n') ADVANCE(613);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 499:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'n') ADVANCE(554);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 500:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'n') ADVANCE(487);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 501:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'o') ADVANCE(647);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 502:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'r') ADVANCE(569);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 503:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'r') ADVANCE(639);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 504:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 's') ADVANCE(484);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 505:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 's') ADVANCE(488);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 506:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 't') ADVANCE(485);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 507:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 't') ADVANCE(491);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 508:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (lookahead == 'u') ADVANCE(489);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 509:
      ACCEPT_TOKEN(aux_sym_sigil_token3);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 510:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 511:
      ACCEPT_TOKEN(sym_keyword);
      END_STATE();
    case 512:
      ACCEPT_TOKEN(aux_sym_quoted_keyword_token1);
      END_STATE();
    case 513:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '<') ADVANCE(595);
      if (lookahead == '>') ADVANCE(192);
      if (lookahead == '~') ADVANCE(599);
      END_STATE();
    case 514:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '<') ADVANCE(594);
      if (lookahead == '~') ADVANCE(598);
      END_STATE();
    case 515:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 516:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '>') ADVANCE(175);
      END_STATE();
    case 517:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '>') ADVANCE(597);
      END_STATE();
    case 518:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '>') ADVANCE(596);
      END_STATE();
    case 519:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 520:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '{') ADVANCE(243);
      END_STATE();
    case 521:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 522:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      if (lookahead == '.') ADVANCE(389);
      if (lookahead == '/') ADVANCE(171);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 523:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      if (lookahead == '.') ADVANCE(388);
      END_STATE();
    case 524:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 525:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(570);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 526:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(571);
      END_STATE();
    case 527:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 528:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(618);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 529:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(619);
      END_STATE();
    case 530:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 531:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(620);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '>') ADVANCE(633);
      END_STATE();
    case 532:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(621);
      END_STATE();
    case 533:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(621);
      if (lookahead == '>') ADVANCE(632);
      END_STATE();
    case 534:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 535:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(580);
      END_STATE();
    case 536:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(581);
      END_STATE();
    case 537:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 538:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '^') ADVANCE(209);
      END_STATE();
    case 539:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '^') ADVANCE(210);
      END_STATE();
    case 540:
      ACCEPT_TOKEN(anon_sym_TILDE_TILDE_TILDE);
      END_STATE();
    case 541:
      ACCEPT_TOKEN(anon_sym_TILDE_TILDE_TILDE);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 542:
      ACCEPT_TOKEN(anon_sym_not);
      END_STATE();
    case 543:
      ACCEPT_TOKEN(anon_sym_not);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 544:
      ACCEPT_TOKEN(anon_sym_not);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 545:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 546:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 547:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      END_STATE();
    case 548:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 549:
      ACCEPT_TOKEN(anon_sym_BSLASH_BSLASH);
      END_STATE();
    case 550:
      ACCEPT_TOKEN(anon_sym_BSLASH_BSLASH);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 551:
      ACCEPT_TOKEN(anon_sym_when);
      END_STATE();
    case 552:
      ACCEPT_TOKEN(anon_sym_when);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 553:
      ACCEPT_TOKEN(anon_sym_when);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 554:
      ACCEPT_TOKEN(anon_sym_when);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 555:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 556:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      if (lookahead == ':') ADVANCE(413);
      END_STATE();
    case 557:
      ACCEPT_TOKEN(anon_sym_EQ_GT);
      END_STATE();
    case 558:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(578);
      if (lookahead == '>') ADVANCE(557);
      if (lookahead == '~') ADVANCE(583);
      END_STATE();
    case 559:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(578);
      if (lookahead == '~') ADVANCE(583);
      END_STATE();
    case 560:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(579);
      if (lookahead == '>') ADVANCE(557);
      if (lookahead == '~') ADVANCE(582);
      END_STATE();
    case 561:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(579);
      if (lookahead == '~') ADVANCE(582);
      END_STATE();
    case 562:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '|') ADVANCE(565);
      END_STATE();
    case 563:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      if (lookahead == '|') ADVANCE(564);
      END_STATE();
    case 564:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE_PIPE);
      END_STATE();
    case 565:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE_PIPE);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 566:
      ACCEPT_TOKEN(anon_sym_or);
      END_STATE();
    case 567:
      ACCEPT_TOKEN(anon_sym_or);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 568:
      ACCEPT_TOKEN(anon_sym_or);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 569:
      ACCEPT_TOKEN(anon_sym_or);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 570:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      if (lookahead == '&') ADVANCE(573);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 571:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      if (lookahead == '&') ADVANCE(572);
      END_STATE();
    case 572:
      ACCEPT_TOKEN(anon_sym_AMP_AMP_AMP);
      END_STATE();
    case 573:
      ACCEPT_TOKEN(anon_sym_AMP_AMP_AMP);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 574:
      ACCEPT_TOKEN(anon_sym_and);
      END_STATE();
    case 575:
      ACCEPT_TOKEN(anon_sym_and);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 576:
      ACCEPT_TOKEN(anon_sym_and);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 577:
      ACCEPT_TOKEN(anon_sym_and);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 578:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(585);
      END_STATE();
    case 579:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      if (lookahead == '=') ADVANCE(584);
      END_STATE();
    case 580:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '=') ADVANCE(587);
      END_STATE();
    case 581:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      if (lookahead == '=') ADVANCE(586);
      END_STATE();
    case 582:
      ACCEPT_TOKEN(anon_sym_EQ_TILDE);
      END_STATE();
    case 583:
      ACCEPT_TOKEN(anon_sym_EQ_TILDE);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 584:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_EQ);
      END_STATE();
    case 585:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_EQ);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 586:
      ACCEPT_TOKEN(anon_sym_BANG_EQ_EQ);
      END_STATE();
    case 587:
      ACCEPT_TOKEN(anon_sym_BANG_EQ_EQ);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 588:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 589:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 590:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 591:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 592:
      ACCEPT_TOKEN(anon_sym_PIPE_GT);
      END_STATE();
    case 593:
      ACCEPT_TOKEN(anon_sym_PIPE_GT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 594:
      ACCEPT_TOKEN(anon_sym_LT_LT_LT);
      END_STATE();
    case 595:
      ACCEPT_TOKEN(anon_sym_LT_LT_LT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 596:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT);
      END_STATE();
    case 597:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 598:
      ACCEPT_TOKEN(anon_sym_LT_LT_TILDE);
      END_STATE();
    case 599:
      ACCEPT_TOKEN(anon_sym_LT_LT_TILDE);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 600:
      ACCEPT_TOKEN(anon_sym_TILDE_GT_GT);
      END_STATE();
    case 601:
      ACCEPT_TOKEN(anon_sym_TILDE_GT_GT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 602:
      ACCEPT_TOKEN(anon_sym_LT_TILDE);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '>') ADVANCE(607);
      END_STATE();
    case 603:
      ACCEPT_TOKEN(anon_sym_LT_TILDE);
      if (lookahead == '>') ADVANCE(606);
      END_STATE();
    case 604:
      ACCEPT_TOKEN(anon_sym_TILDE_GT);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '>') ADVANCE(601);
      END_STATE();
    case 605:
      ACCEPT_TOKEN(anon_sym_TILDE_GT);
      if (lookahead == '>') ADVANCE(600);
      END_STATE();
    case 606:
      ACCEPT_TOKEN(anon_sym_LT_TILDE_GT);
      END_STATE();
    case 607:
      ACCEPT_TOKEN(anon_sym_LT_TILDE_GT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 608:
      ACCEPT_TOKEN(anon_sym_LT_PIPE_GT);
      END_STATE();
    case 609:
      ACCEPT_TOKEN(anon_sym_LT_PIPE_GT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 610:
      ACCEPT_TOKEN(anon_sym_in);
      END_STATE();
    case 611:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 612:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 613:
      ACCEPT_TOKEN(anon_sym_in);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 614:
      ACCEPT_TOKEN(anon_sym_CARET_CARET_CARET);
      END_STATE();
    case 615:
      ACCEPT_TOKEN(anon_sym_CARET_CARET_CARET);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 616:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH);
      END_STATE();
    case 617:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 618:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      if (lookahead == '+') ADVANCE(623);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 619:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      if (lookahead == '+') ADVANCE(622);
      END_STATE();
    case 620:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '-') ADVANCE(625);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 621:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead == '-') ADVANCE(624);
      END_STATE();
    case 622:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS_PLUS);
      END_STATE();
    case 623:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS_PLUS);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 624:
      ACCEPT_TOKEN(anon_sym_DASH_DASH_DASH);
      END_STATE();
    case 625:
      ACCEPT_TOKEN(anon_sym_DASH_DASH_DASH);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 626:
      ACCEPT_TOKEN(anon_sym_LT_GT);
      END_STATE();
    case 627:
      ACCEPT_TOKEN(anon_sym_LT_GT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 628:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(631);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 629:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(630);
      END_STATE();
    case 630:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      END_STATE();
    case 631:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 632:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 633:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 634:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(522);
      if (lookahead == ':') ADVANCE(252);
      END_STATE();
    case 635:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(521);
      END_STATE();
    case 636:
      ACCEPT_TOKEN(anon_sym_after);
      END_STATE();
    case 637:
      ACCEPT_TOKEN(anon_sym_after);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 638:
      ACCEPT_TOKEN(anon_sym_after);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 639:
      ACCEPT_TOKEN(anon_sym_after);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 640:
      ACCEPT_TOKEN(anon_sym_catch);
      END_STATE();
    case 641:
      ACCEPT_TOKEN(anon_sym_catch);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 642:
      ACCEPT_TOKEN(anon_sym_catch);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 643:
      ACCEPT_TOKEN(anon_sym_catch);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 644:
      ACCEPT_TOKEN(anon_sym_do);
      END_STATE();
    case 645:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 646:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 647:
      ACCEPT_TOKEN(anon_sym_do);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 648:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 649:
      ACCEPT_TOKEN(anon_sym_else);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 650:
      ACCEPT_TOKEN(anon_sym_else);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 651:
      ACCEPT_TOKEN(anon_sym_else);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 652:
      ACCEPT_TOKEN(anon_sym_end);
      END_STATE();
    case 653:
      ACCEPT_TOKEN(anon_sym_end);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 654:
      ACCEPT_TOKEN(anon_sym_end);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 655:
      ACCEPT_TOKEN(anon_sym_end);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 656:
      ACCEPT_TOKEN(anon_sym_fn);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 657:
      ACCEPT_TOKEN(anon_sym_fn);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 658:
      ACCEPT_TOKEN(anon_sym_rescue);
      END_STATE();
    case 659:
      ACCEPT_TOKEN(anon_sym_rescue);
      if (lookahead == ':') ADVANCE(252);
      if (lookahead == '@') ADVANCE(186);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(313);
      if (set_contains(sym_keyword_character_set_2, 794, lookahead)) ADVANCE(350);
      END_STATE();
    case 660:
      ACCEPT_TOKEN(anon_sym_rescue);
      if (lookahead == '!' ||
          lookahead == '?') ADVANCE(312);
      if (set_contains(aux_sym_identifier_token1_character_set_2, 795, lookahead)) ADVANCE(387);
      END_STATE();
    case 661:
      ACCEPT_TOKEN(anon_sym_rescue);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(509);
      END_STATE();
    case 662:
      ACCEPT_TOKEN(anon_sym_LPAREN2);
      END_STATE();
    case 663:
      ACCEPT_TOKEN(anon_sym_LBRACK2);
      END_STATE();
    case 664:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '{') ADVANCE(462);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(665);
      END_STATE();
    case 665:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(665);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 267, .external_lex_state = 2},
  [2] = {.lex_state = 81, .external_lex_state = 2},
  [3] = {.lex_state = 81, .external_lex_state = 2},
  [4] = {.lex_state = 81, .external_lex_state = 2},
  [5] = {.lex_state = 81, .external_lex_state = 2},
  [6] = {.lex_state = 81, .external_lex_state = 2},
  [7] = {.lex_state = 81, .external_lex_state = 2},
  [8] = {.lex_state = 81, .external_lex_state = 2},
  [9] = {.lex_state = 81, .external_lex_state = 2},
  [10] = {.lex_state = 81, .external_lex_state = 2},
  [11] = {.lex_state = 81, .external_lex_state = 2},
  [12] = {.lex_state = 81, .external_lex_state = 2},
  [13] = {.lex_state = 81, .external_lex_state = 2},
  [14] = {.lex_state = 83, .external_lex_state = 3},
  [15] = {.lex_state = 83, .external_lex_state = 3},
  [16] = {.lex_state = 83, .external_lex_state = 3},
  [17] = {.lex_state = 83, .external_lex_state = 3},
  [18] = {.lex_state = 83, .external_lex_state = 3},
  [19] = {.lex_state = 83, .external_lex_state = 3},
  [20] = {.lex_state = 81, .external_lex_state = 2},
  [21] = {.lex_state = 83, .external_lex_state = 2},
  [22] = {.lex_state = 81, .external_lex_state = 2},
  [23] = {.lex_state = 83, .external_lex_state = 2},
  [24] = {.lex_state = 81, .external_lex_state = 2},
  [25] = {.lex_state = 81, .external_lex_state = 2},
  [26] = {.lex_state = 81, .external_lex_state = 2},
  [27] = {.lex_state = 81, .external_lex_state = 2},
  [28] = {.lex_state = 81, .external_lex_state = 2},
  [29] = {.lex_state = 81, .external_lex_state = 2},
  [30] = {.lex_state = 81, .external_lex_state = 2},
  [31] = {.lex_state = 81, .external_lex_state = 2},
  [32] = {.lex_state = 81, .external_lex_state = 2},
  [33] = {.lex_state = 81, .external_lex_state = 2},
  [34] = {.lex_state = 81, .external_lex_state = 2},
  [35] = {.lex_state = 269, .external_lex_state = 3},
  [36] = {.lex_state = 269, .external_lex_state = 3},
  [37] = {.lex_state = 81, .external_lex_state = 2},
  [38] = {.lex_state = 269, .external_lex_state = 3},
  [39] = {.lex_state = 81, .external_lex_state = 2},
  [40] = {.lex_state = 81, .external_lex_state = 2},
  [41] = {.lex_state = 269, .external_lex_state = 3},
  [42] = {.lex_state = 269, .external_lex_state = 3},
  [43] = {.lex_state = 269, .external_lex_state = 3},
  [44] = {.lex_state = 269, .external_lex_state = 3},
  [45] = {.lex_state = 88, .external_lex_state = 3},
  [46] = {.lex_state = 269, .external_lex_state = 3},
  [47] = {.lex_state = 269, .external_lex_state = 3},
  [48] = {.lex_state = 88, .external_lex_state = 3},
  [49] = {.lex_state = 269, .external_lex_state = 3},
  [50] = {.lex_state = 269, .external_lex_state = 3},
  [51] = {.lex_state = 269, .external_lex_state = 2},
  [52] = {.lex_state = 88, .external_lex_state = 3},
  [53] = {.lex_state = 269, .external_lex_state = 3},
  [54] = {.lex_state = 91, .external_lex_state = 3},
  [55] = {.lex_state = 269, .external_lex_state = 3},
  [56] = {.lex_state = 88, .external_lex_state = 2},
  [57] = {.lex_state = 269, .external_lex_state = 3},
  [58] = {.lex_state = 269, .external_lex_state = 3},
  [59] = {.lex_state = 269, .external_lex_state = 2},
  [60] = {.lex_state = 269, .external_lex_state = 2},
  [61] = {.lex_state = 91, .external_lex_state = 3},
  [62] = {.lex_state = 91, .external_lex_state = 3},
  [63] = {.lex_state = 269, .external_lex_state = 2},
  [64] = {.lex_state = 269, .external_lex_state = 3},
  [65] = {.lex_state = 269, .external_lex_state = 3},
  [66] = {.lex_state = 269, .external_lex_state = 3},
  [67] = {.lex_state = 81, .external_lex_state = 2},
  [68] = {.lex_state = 81, .external_lex_state = 2},
  [69] = {.lex_state = 81, .external_lex_state = 2},
  [70] = {.lex_state = 81, .external_lex_state = 2},
  [71] = {.lex_state = 91, .external_lex_state = 2},
  [72] = {.lex_state = 269, .external_lex_state = 2},
  [73] = {.lex_state = 94, .external_lex_state = 2},
  [74] = {.lex_state = 94, .external_lex_state = 2},
  [75] = {.lex_state = 94, .external_lex_state = 2},
  [76] = {.lex_state = 96, .external_lex_state = 2},
  [77] = {.lex_state = 94, .external_lex_state = 2},
  [78] = {.lex_state = 94, .external_lex_state = 2},
  [79] = {.lex_state = 94, .external_lex_state = 2},
  [80] = {.lex_state = 94, .external_lex_state = 2},
  [81] = {.lex_state = 96, .external_lex_state = 2},
  [82] = {.lex_state = 94, .external_lex_state = 2},
  [83] = {.lex_state = 94, .external_lex_state = 2},
  [84] = {.lex_state = 94, .external_lex_state = 2},
  [85] = {.lex_state = 94, .external_lex_state = 2},
  [86] = {.lex_state = 96, .external_lex_state = 2},
  [87] = {.lex_state = 94, .external_lex_state = 2},
  [88] = {.lex_state = 269, .external_lex_state = 2},
  [89] = {.lex_state = 94, .external_lex_state = 2},
  [90] = {.lex_state = 96, .external_lex_state = 2},
  [91] = {.lex_state = 94, .external_lex_state = 2},
  [92] = {.lex_state = 96, .external_lex_state = 2},
  [93] = {.lex_state = 94, .external_lex_state = 2},
  [94] = {.lex_state = 94, .external_lex_state = 2},
  [95] = {.lex_state = 96, .external_lex_state = 2},
  [96] = {.lex_state = 94, .external_lex_state = 2},
  [97] = {.lex_state = 96, .external_lex_state = 2},
  [98] = {.lex_state = 94, .external_lex_state = 2},
  [99] = {.lex_state = 94, .external_lex_state = 2},
  [100] = {.lex_state = 96, .external_lex_state = 2},
  [101] = {.lex_state = 94, .external_lex_state = 2},
  [102] = {.lex_state = 96, .external_lex_state = 2},
  [103] = {.lex_state = 94, .external_lex_state = 2},
  [104] = {.lex_state = 94, .external_lex_state = 2},
  [105] = {.lex_state = 96, .external_lex_state = 2},
  [106] = {.lex_state = 96, .external_lex_state = 2},
  [107] = {.lex_state = 94, .external_lex_state = 2},
  [108] = {.lex_state = 94, .external_lex_state = 2},
  [109] = {.lex_state = 96, .external_lex_state = 2},
  [110] = {.lex_state = 94, .external_lex_state = 2},
  [111] = {.lex_state = 94, .external_lex_state = 2},
  [112] = {.lex_state = 269, .external_lex_state = 2},
  [113] = {.lex_state = 96, .external_lex_state = 2},
  [114] = {.lex_state = 96, .external_lex_state = 2},
  [115] = {.lex_state = 94, .external_lex_state = 2},
  [116] = {.lex_state = 96, .external_lex_state = 2},
  [117] = {.lex_state = 94, .external_lex_state = 2},
  [118] = {.lex_state = 96, .external_lex_state = 2},
  [119] = {.lex_state = 96, .external_lex_state = 2},
  [120] = {.lex_state = 94, .external_lex_state = 2},
  [121] = {.lex_state = 96, .external_lex_state = 2},
  [122] = {.lex_state = 94, .external_lex_state = 2},
  [123] = {.lex_state = 94, .external_lex_state = 2},
  [124] = {.lex_state = 94, .external_lex_state = 2},
  [125] = {.lex_state = 96, .external_lex_state = 2},
  [126] = {.lex_state = 96, .external_lex_state = 2},
  [127] = {.lex_state = 94, .external_lex_state = 2},
  [128] = {.lex_state = 94, .external_lex_state = 2},
  [129] = {.lex_state = 96, .external_lex_state = 2},
  [130] = {.lex_state = 94, .external_lex_state = 2},
  [131] = {.lex_state = 96, .external_lex_state = 2},
  [132] = {.lex_state = 96, .external_lex_state = 2},
  [133] = {.lex_state = 94, .external_lex_state = 2},
  [134] = {.lex_state = 96, .external_lex_state = 2},
  [135] = {.lex_state = 94, .external_lex_state = 2},
  [136] = {.lex_state = 94, .external_lex_state = 2},
  [137] = {.lex_state = 94, .external_lex_state = 2},
  [138] = {.lex_state = 94, .external_lex_state = 2},
  [139] = {.lex_state = 94, .external_lex_state = 2},
  [140] = {.lex_state = 98, .external_lex_state = 2},
  [141] = {.lex_state = 98, .external_lex_state = 2},
  [142] = {.lex_state = 98, .external_lex_state = 2},
  [143] = {.lex_state = 98, .external_lex_state = 2},
  [144] = {.lex_state = 98, .external_lex_state = 2},
  [145] = {.lex_state = 98, .external_lex_state = 2},
  [146] = {.lex_state = 98, .external_lex_state = 2},
  [147] = {.lex_state = 98, .external_lex_state = 2},
  [148] = {.lex_state = 98, .external_lex_state = 2},
  [149] = {.lex_state = 98, .external_lex_state = 2},
  [150] = {.lex_state = 98, .external_lex_state = 2},
  [151] = {.lex_state = 98, .external_lex_state = 2},
  [152] = {.lex_state = 98, .external_lex_state = 2},
  [153] = {.lex_state = 98, .external_lex_state = 2},
  [154] = {.lex_state = 98, .external_lex_state = 2},
  [155] = {.lex_state = 98, .external_lex_state = 2},
  [156] = {.lex_state = 98, .external_lex_state = 2},
  [157] = {.lex_state = 98, .external_lex_state = 2},
  [158] = {.lex_state = 98, .external_lex_state = 2},
  [159] = {.lex_state = 98, .external_lex_state = 2},
  [160] = {.lex_state = 98, .external_lex_state = 2},
  [161] = {.lex_state = 98, .external_lex_state = 2},
  [162] = {.lex_state = 98, .external_lex_state = 2},
  [163] = {.lex_state = 98, .external_lex_state = 2},
  [164] = {.lex_state = 98, .external_lex_state = 2},
  [165] = {.lex_state = 98, .external_lex_state = 2},
  [166] = {.lex_state = 98, .external_lex_state = 2},
  [167] = {.lex_state = 98, .external_lex_state = 2},
  [168] = {.lex_state = 98, .external_lex_state = 2},
  [169] = {.lex_state = 98, .external_lex_state = 2},
  [170] = {.lex_state = 98, .external_lex_state = 2},
  [171] = {.lex_state = 98, .external_lex_state = 2},
  [172] = {.lex_state = 98, .external_lex_state = 2},
  [173] = {.lex_state = 98, .external_lex_state = 2},
  [174] = {.lex_state = 98, .external_lex_state = 2},
  [175] = {.lex_state = 98, .external_lex_state = 2},
  [176] = {.lex_state = 98, .external_lex_state = 2},
  [177] = {.lex_state = 98, .external_lex_state = 2},
  [178] = {.lex_state = 94, .external_lex_state = 2},
  [179] = {.lex_state = 94, .external_lex_state = 2},
  [180] = {.lex_state = 94, .external_lex_state = 2},
  [181] = {.lex_state = 94, .external_lex_state = 2},
  [182] = {.lex_state = 94, .external_lex_state = 2},
  [183] = {.lex_state = 94, .external_lex_state = 2},
  [184] = {.lex_state = 94, .external_lex_state = 2},
  [185] = {.lex_state = 94, .external_lex_state = 2},
  [186] = {.lex_state = 94, .external_lex_state = 2},
  [187] = {.lex_state = 94, .external_lex_state = 2},
  [188] = {.lex_state = 100, .external_lex_state = 2},
  [189] = {.lex_state = 94, .external_lex_state = 2},
  [190] = {.lex_state = 94, .external_lex_state = 2},
  [191] = {.lex_state = 94, .external_lex_state = 2},
  [192] = {.lex_state = 94, .external_lex_state = 2},
  [193] = {.lex_state = 94, .external_lex_state = 2},
  [194] = {.lex_state = 94, .external_lex_state = 2},
  [195] = {.lex_state = 94, .external_lex_state = 2},
  [196] = {.lex_state = 100, .external_lex_state = 2},
  [197] = {.lex_state = 100, .external_lex_state = 2},
  [198] = {.lex_state = 94, .external_lex_state = 2},
  [199] = {.lex_state = 94, .external_lex_state = 2},
  [200] = {.lex_state = 94, .external_lex_state = 2},
  [201] = {.lex_state = 94, .external_lex_state = 2},
  [202] = {.lex_state = 94, .external_lex_state = 2},
  [203] = {.lex_state = 94, .external_lex_state = 2},
  [204] = {.lex_state = 100, .external_lex_state = 2},
  [205] = {.lex_state = 94, .external_lex_state = 2},
  [206] = {.lex_state = 94, .external_lex_state = 2},
  [207] = {.lex_state = 94, .external_lex_state = 2},
  [208] = {.lex_state = 94, .external_lex_state = 2},
  [209] = {.lex_state = 94, .external_lex_state = 2},
  [210] = {.lex_state = 94, .external_lex_state = 2},
  [211] = {.lex_state = 94, .external_lex_state = 2},
  [212] = {.lex_state = 94, .external_lex_state = 2},
  [213] = {.lex_state = 94, .external_lex_state = 2},
  [214] = {.lex_state = 94, .external_lex_state = 2},
  [215] = {.lex_state = 100, .external_lex_state = 2},
  [216] = {.lex_state = 94, .external_lex_state = 2},
  [217] = {.lex_state = 94, .external_lex_state = 2},
  [218] = {.lex_state = 100, .external_lex_state = 2},
  [219] = {.lex_state = 94, .external_lex_state = 2},
  [220] = {.lex_state = 94, .external_lex_state = 2},
  [221] = {.lex_state = 94, .external_lex_state = 2},
  [222] = {.lex_state = 94, .external_lex_state = 2},
  [223] = {.lex_state = 94, .external_lex_state = 2},
  [224] = {.lex_state = 94, .external_lex_state = 2},
  [225] = {.lex_state = 94, .external_lex_state = 2},
  [226] = {.lex_state = 94, .external_lex_state = 2},
  [227] = {.lex_state = 94, .external_lex_state = 2},
  [228] = {.lex_state = 94, .external_lex_state = 2},
  [229] = {.lex_state = 94, .external_lex_state = 2},
  [230] = {.lex_state = 94, .external_lex_state = 2},
  [231] = {.lex_state = 94, .external_lex_state = 2},
  [232] = {.lex_state = 100, .external_lex_state = 2},
  [233] = {.lex_state = 94, .external_lex_state = 2},
  [234] = {.lex_state = 94, .external_lex_state = 2},
  [235] = {.lex_state = 94, .external_lex_state = 2},
  [236] = {.lex_state = 94, .external_lex_state = 2},
  [237] = {.lex_state = 94, .external_lex_state = 2},
  [238] = {.lex_state = 94, .external_lex_state = 2},
  [239] = {.lex_state = 100, .external_lex_state = 2},
  [240] = {.lex_state = 94, .external_lex_state = 2},
  [241] = {.lex_state = 94, .external_lex_state = 2},
  [242] = {.lex_state = 100, .external_lex_state = 2},
  [243] = {.lex_state = 94, .external_lex_state = 2},
  [244] = {.lex_state = 94, .external_lex_state = 2},
  [245] = {.lex_state = 94, .external_lex_state = 2},
  [246] = {.lex_state = 100, .external_lex_state = 2},
  [247] = {.lex_state = 94, .external_lex_state = 2},
  [248] = {.lex_state = 94, .external_lex_state = 2},
  [249] = {.lex_state = 94, .external_lex_state = 2},
  [250] = {.lex_state = 100, .external_lex_state = 2},
  [251] = {.lex_state = 94, .external_lex_state = 2},
  [252] = {.lex_state = 94, .external_lex_state = 2},
  [253] = {.lex_state = 94, .external_lex_state = 2},
  [254] = {.lex_state = 94, .external_lex_state = 2},
  [255] = {.lex_state = 100, .external_lex_state = 2},
  [256] = {.lex_state = 94, .external_lex_state = 2},
  [257] = {.lex_state = 94, .external_lex_state = 2},
  [258] = {.lex_state = 98, .external_lex_state = 2},
  [259] = {.lex_state = 98, .external_lex_state = 2},
  [260] = {.lex_state = 98, .external_lex_state = 2},
  [261] = {.lex_state = 98, .external_lex_state = 2},
  [262] = {.lex_state = 100, .external_lex_state = 2},
  [263] = {.lex_state = 100, .external_lex_state = 2},
  [264] = {.lex_state = 94, .external_lex_state = 2},
  [265] = {.lex_state = 94, .external_lex_state = 2},
  [266] = {.lex_state = 94, .external_lex_state = 2},
  [267] = {.lex_state = 94, .external_lex_state = 2},
  [268] = {.lex_state = 98, .external_lex_state = 2},
  [269] = {.lex_state = 98, .external_lex_state = 2},
  [270] = {.lex_state = 94, .external_lex_state = 2},
  [271] = {.lex_state = 94, .external_lex_state = 2},
  [272] = {.lex_state = 94, .external_lex_state = 2},
  [273] = {.lex_state = 94, .external_lex_state = 2},
  [274] = {.lex_state = 94, .external_lex_state = 2},
  [275] = {.lex_state = 94, .external_lex_state = 2},
  [276] = {.lex_state = 98, .external_lex_state = 2},
  [277] = {.lex_state = 98, .external_lex_state = 2},
  [278] = {.lex_state = 94, .external_lex_state = 2},
  [279] = {.lex_state = 94, .external_lex_state = 2},
  [280] = {.lex_state = 94, .external_lex_state = 2},
  [281] = {.lex_state = 94, .external_lex_state = 2},
  [282] = {.lex_state = 98, .external_lex_state = 2},
  [283] = {.lex_state = 98, .external_lex_state = 2},
  [284] = {.lex_state = 94, .external_lex_state = 2},
  [285] = {.lex_state = 94, .external_lex_state = 2},
  [286] = {.lex_state = 98, .external_lex_state = 2},
  [287] = {.lex_state = 94, .external_lex_state = 2},
  [288] = {.lex_state = 94, .external_lex_state = 2},
  [289] = {.lex_state = 98, .external_lex_state = 2},
  [290] = {.lex_state = 98, .external_lex_state = 2},
  [291] = {.lex_state = 94, .external_lex_state = 2},
  [292] = {.lex_state = 94, .external_lex_state = 2},
  [293] = {.lex_state = 94, .external_lex_state = 2},
  [294] = {.lex_state = 94, .external_lex_state = 2},
  [295] = {.lex_state = 98, .external_lex_state = 2},
  [296] = {.lex_state = 94, .external_lex_state = 2},
  [297] = {.lex_state = 94, .external_lex_state = 2},
  [298] = {.lex_state = 94, .external_lex_state = 2},
  [299] = {.lex_state = 94, .external_lex_state = 2},
  [300] = {.lex_state = 94, .external_lex_state = 2},
  [301] = {.lex_state = 94, .external_lex_state = 2},
  [302] = {.lex_state = 94, .external_lex_state = 2},
  [303] = {.lex_state = 102, .external_lex_state = 2},
  [304] = {.lex_state = 94, .external_lex_state = 2},
  [305] = {.lex_state = 94, .external_lex_state = 2},
  [306] = {.lex_state = 94, .external_lex_state = 2},
  [307] = {.lex_state = 94, .external_lex_state = 2},
  [308] = {.lex_state = 94, .external_lex_state = 2},
  [309] = {.lex_state = 94, .external_lex_state = 2},
  [310] = {.lex_state = 102, .external_lex_state = 2},
  [311] = {.lex_state = 94, .external_lex_state = 2},
  [312] = {.lex_state = 94, .external_lex_state = 2},
  [313] = {.lex_state = 94, .external_lex_state = 2},
  [314] = {.lex_state = 94, .external_lex_state = 2},
  [315] = {.lex_state = 94, .external_lex_state = 2},
  [316] = {.lex_state = 94, .external_lex_state = 2},
  [317] = {.lex_state = 94, .external_lex_state = 2},
  [318] = {.lex_state = 94, .external_lex_state = 2},
  [319] = {.lex_state = 94, .external_lex_state = 2},
  [320] = {.lex_state = 94, .external_lex_state = 2},
  [321] = {.lex_state = 94, .external_lex_state = 2},
  [322] = {.lex_state = 94, .external_lex_state = 2},
  [323] = {.lex_state = 94, .external_lex_state = 2},
  [324] = {.lex_state = 94, .external_lex_state = 2},
  [325] = {.lex_state = 94, .external_lex_state = 2},
  [326] = {.lex_state = 94, .external_lex_state = 2},
  [327] = {.lex_state = 267, .external_lex_state = 2},
  [328] = {.lex_state = 94, .external_lex_state = 2},
  [329] = {.lex_state = 267, .external_lex_state = 2},
  [330] = {.lex_state = 94, .external_lex_state = 2},
  [331] = {.lex_state = 94, .external_lex_state = 2},
  [332] = {.lex_state = 94, .external_lex_state = 2},
  [333] = {.lex_state = 94, .external_lex_state = 2},
  [334] = {.lex_state = 94, .external_lex_state = 2},
  [335] = {.lex_state = 94, .external_lex_state = 2},
  [336] = {.lex_state = 94, .external_lex_state = 2},
  [337] = {.lex_state = 94, .external_lex_state = 2},
  [338] = {.lex_state = 94, .external_lex_state = 2},
  [339] = {.lex_state = 94, .external_lex_state = 2},
  [340] = {.lex_state = 94, .external_lex_state = 2},
  [341] = {.lex_state = 94, .external_lex_state = 2},
  [342] = {.lex_state = 94, .external_lex_state = 2},
  [343] = {.lex_state = 98, .external_lex_state = 2},
  [344] = {.lex_state = 98, .external_lex_state = 2},
  [345] = {.lex_state = 94, .external_lex_state = 2},
  [346] = {.lex_state = 102, .external_lex_state = 2},
  [347] = {.lex_state = 267, .external_lex_state = 2},
  [348] = {.lex_state = 102, .external_lex_state = 2},
  [349] = {.lex_state = 102, .external_lex_state = 2},
  [350] = {.lex_state = 102, .external_lex_state = 2},
  [351] = {.lex_state = 267, .external_lex_state = 2},
  [352] = {.lex_state = 267, .external_lex_state = 2},
  [353] = {.lex_state = 267, .external_lex_state = 2},
  [354] = {.lex_state = 267, .external_lex_state = 2},
  [355] = {.lex_state = 267, .external_lex_state = 2},
  [356] = {.lex_state = 267, .external_lex_state = 2},
  [357] = {.lex_state = 267, .external_lex_state = 2},
  [358] = {.lex_state = 267, .external_lex_state = 2},
  [359] = {.lex_state = 267, .external_lex_state = 2},
  [360] = {.lex_state = 267, .external_lex_state = 2},
  [361] = {.lex_state = 267, .external_lex_state = 2},
  [362] = {.lex_state = 267, .external_lex_state = 2},
  [363] = {.lex_state = 267, .external_lex_state = 2},
  [364] = {.lex_state = 267, .external_lex_state = 2},
  [365] = {.lex_state = 267, .external_lex_state = 2},
  [366] = {.lex_state = 267, .external_lex_state = 2},
  [367] = {.lex_state = 267, .external_lex_state = 2},
  [368] = {.lex_state = 267, .external_lex_state = 2},
  [369] = {.lex_state = 267, .external_lex_state = 2},
  [370] = {.lex_state = 267, .external_lex_state = 2},
  [371] = {.lex_state = 267, .external_lex_state = 2},
  [372] = {.lex_state = 267, .external_lex_state = 2},
  [373] = {.lex_state = 267, .external_lex_state = 2},
  [374] = {.lex_state = 267, .external_lex_state = 2},
  [375] = {.lex_state = 267, .external_lex_state = 2},
  [376] = {.lex_state = 267, .external_lex_state = 2},
  [377] = {.lex_state = 267, .external_lex_state = 2},
  [378] = {.lex_state = 267, .external_lex_state = 2},
  [379] = {.lex_state = 267, .external_lex_state = 2},
  [380] = {.lex_state = 267, .external_lex_state = 2},
  [381] = {.lex_state = 267, .external_lex_state = 2},
  [382] = {.lex_state = 267, .external_lex_state = 2},
  [383] = {.lex_state = 267, .external_lex_state = 2},
  [384] = {.lex_state = 267, .external_lex_state = 2},
  [385] = {.lex_state = 267, .external_lex_state = 2},
  [386] = {.lex_state = 267, .external_lex_state = 2},
  [387] = {.lex_state = 267, .external_lex_state = 2},
  [388] = {.lex_state = 267, .external_lex_state = 2},
  [389] = {.lex_state = 267, .external_lex_state = 2},
  [390] = {.lex_state = 267, .external_lex_state = 2},
  [391] = {.lex_state = 267, .external_lex_state = 2},
  [392] = {.lex_state = 267, .external_lex_state = 2},
  [393] = {.lex_state = 267, .external_lex_state = 2},
  [394] = {.lex_state = 267, .external_lex_state = 2},
  [395] = {.lex_state = 267, .external_lex_state = 2},
  [396] = {.lex_state = 267, .external_lex_state = 2},
  [397] = {.lex_state = 267, .external_lex_state = 2},
  [398] = {.lex_state = 267, .external_lex_state = 2},
  [399] = {.lex_state = 267, .external_lex_state = 2},
  [400] = {.lex_state = 267, .external_lex_state = 2},
  [401] = {.lex_state = 267, .external_lex_state = 2},
  [402] = {.lex_state = 267, .external_lex_state = 2},
  [403] = {.lex_state = 267, .external_lex_state = 2},
  [404] = {.lex_state = 267, .external_lex_state = 2},
  [405] = {.lex_state = 267, .external_lex_state = 2},
  [406] = {.lex_state = 267, .external_lex_state = 2},
  [407] = {.lex_state = 267, .external_lex_state = 2},
  [408] = {.lex_state = 267, .external_lex_state = 2},
  [409] = {.lex_state = 267, .external_lex_state = 2},
  [410] = {.lex_state = 267, .external_lex_state = 2},
  [411] = {.lex_state = 267, .external_lex_state = 2},
  [412] = {.lex_state = 267, .external_lex_state = 2},
  [413] = {.lex_state = 267, .external_lex_state = 2},
  [414] = {.lex_state = 267, .external_lex_state = 2},
  [415] = {.lex_state = 267, .external_lex_state = 2},
  [416] = {.lex_state = 267, .external_lex_state = 2},
  [417] = {.lex_state = 267, .external_lex_state = 2},
  [418] = {.lex_state = 267, .external_lex_state = 2},
  [419] = {.lex_state = 267, .external_lex_state = 2},
  [420] = {.lex_state = 267, .external_lex_state = 2},
  [421] = {.lex_state = 267, .external_lex_state = 2},
  [422] = {.lex_state = 267, .external_lex_state = 2},
  [423] = {.lex_state = 267, .external_lex_state = 2},
  [424] = {.lex_state = 267, .external_lex_state = 2},
  [425] = {.lex_state = 267, .external_lex_state = 2},
  [426] = {.lex_state = 267, .external_lex_state = 2},
  [427] = {.lex_state = 267, .external_lex_state = 2},
  [428] = {.lex_state = 267, .external_lex_state = 2},
  [429] = {.lex_state = 267, .external_lex_state = 2},
  [430] = {.lex_state = 267, .external_lex_state = 2},
  [431] = {.lex_state = 267, .external_lex_state = 2},
  [432] = {.lex_state = 267, .external_lex_state = 2},
  [433] = {.lex_state = 267, .external_lex_state = 2},
  [434] = {.lex_state = 267, .external_lex_state = 2},
  [435] = {.lex_state = 267, .external_lex_state = 2},
  [436] = {.lex_state = 267, .external_lex_state = 2},
  [437] = {.lex_state = 267, .external_lex_state = 2},
  [438] = {.lex_state = 267, .external_lex_state = 2},
  [439] = {.lex_state = 267, .external_lex_state = 2},
  [440] = {.lex_state = 267, .external_lex_state = 2},
  [441] = {.lex_state = 267, .external_lex_state = 2},
  [442] = {.lex_state = 267, .external_lex_state = 2},
  [443] = {.lex_state = 267, .external_lex_state = 2},
  [444] = {.lex_state = 267, .external_lex_state = 2},
  [445] = {.lex_state = 267, .external_lex_state = 2},
  [446] = {.lex_state = 267, .external_lex_state = 2},
  [447] = {.lex_state = 267, .external_lex_state = 2},
  [448] = {.lex_state = 267, .external_lex_state = 2},
  [449] = {.lex_state = 267, .external_lex_state = 2},
  [450] = {.lex_state = 267, .external_lex_state = 2},
  [451] = {.lex_state = 267, .external_lex_state = 2},
  [452] = {.lex_state = 267, .external_lex_state = 2},
  [453] = {.lex_state = 267, .external_lex_state = 2},
  [454] = {.lex_state = 267, .external_lex_state = 2},
  [455] = {.lex_state = 267, .external_lex_state = 2},
  [456] = {.lex_state = 267, .external_lex_state = 2},
  [457] = {.lex_state = 267, .external_lex_state = 2},
  [458] = {.lex_state = 267, .external_lex_state = 2},
  [459] = {.lex_state = 267, .external_lex_state = 2},
  [460] = {.lex_state = 267, .external_lex_state = 2},
  [461] = {.lex_state = 267, .external_lex_state = 2},
  [462] = {.lex_state = 267, .external_lex_state = 2},
  [463] = {.lex_state = 267, .external_lex_state = 2},
  [464] = {.lex_state = 267, .external_lex_state = 2},
  [465] = {.lex_state = 267, .external_lex_state = 2},
  [466] = {.lex_state = 267, .external_lex_state = 2},
  [467] = {.lex_state = 267, .external_lex_state = 2},
  [468] = {.lex_state = 267, .external_lex_state = 2},
  [469] = {.lex_state = 267, .external_lex_state = 2},
  [470] = {.lex_state = 267, .external_lex_state = 2},
  [471] = {.lex_state = 267, .external_lex_state = 2},
  [472] = {.lex_state = 267, .external_lex_state = 2},
  [473] = {.lex_state = 267, .external_lex_state = 2},
  [474] = {.lex_state = 267, .external_lex_state = 2},
  [475] = {.lex_state = 267, .external_lex_state = 2},
  [476] = {.lex_state = 267, .external_lex_state = 2},
  [477] = {.lex_state = 267, .external_lex_state = 2},
  [478] = {.lex_state = 267, .external_lex_state = 2},
  [479] = {.lex_state = 267, .external_lex_state = 2},
  [480] = {.lex_state = 267, .external_lex_state = 2},
  [481] = {.lex_state = 267, .external_lex_state = 2},
  [482] = {.lex_state = 267, .external_lex_state = 2},
  [483] = {.lex_state = 267, .external_lex_state = 2},
  [484] = {.lex_state = 267, .external_lex_state = 2},
  [485] = {.lex_state = 267, .external_lex_state = 2},
  [486] = {.lex_state = 267, .external_lex_state = 2},
  [487] = {.lex_state = 267, .external_lex_state = 2},
  [488] = {.lex_state = 267, .external_lex_state = 2},
  [489] = {.lex_state = 267, .external_lex_state = 2},
  [490] = {.lex_state = 267, .external_lex_state = 2},
  [491] = {.lex_state = 267, .external_lex_state = 2},
  [492] = {.lex_state = 267, .external_lex_state = 2},
  [493] = {.lex_state = 267, .external_lex_state = 2},
  [494] = {.lex_state = 267, .external_lex_state = 2},
  [495] = {.lex_state = 267, .external_lex_state = 2},
  [496] = {.lex_state = 267, .external_lex_state = 2},
  [497] = {.lex_state = 267, .external_lex_state = 2},
  [498] = {.lex_state = 267, .external_lex_state = 2},
  [499] = {.lex_state = 267, .external_lex_state = 2},
  [500] = {.lex_state = 267, .external_lex_state = 2},
  [501] = {.lex_state = 267, .external_lex_state = 2},
  [502] = {.lex_state = 267, .external_lex_state = 2},
  [503] = {.lex_state = 267, .external_lex_state = 2},
  [504] = {.lex_state = 267, .external_lex_state = 2},
  [505] = {.lex_state = 267, .external_lex_state = 2},
  [506] = {.lex_state = 267, .external_lex_state = 2},
  [507] = {.lex_state = 267, .external_lex_state = 2},
  [508] = {.lex_state = 267, .external_lex_state = 2},
  [509] = {.lex_state = 267, .external_lex_state = 2},
  [510] = {.lex_state = 267, .external_lex_state = 2},
  [511] = {.lex_state = 267, .external_lex_state = 2},
  [512] = {.lex_state = 267, .external_lex_state = 2},
  [513] = {.lex_state = 267, .external_lex_state = 2},
  [514] = {.lex_state = 267, .external_lex_state = 2},
  [515] = {.lex_state = 267, .external_lex_state = 2},
  [516] = {.lex_state = 267, .external_lex_state = 2},
  [517] = {.lex_state = 267, .external_lex_state = 2},
  [518] = {.lex_state = 267, .external_lex_state = 2},
  [519] = {.lex_state = 267, .external_lex_state = 2},
  [520] = {.lex_state = 267, .external_lex_state = 2},
  [521] = {.lex_state = 267, .external_lex_state = 2},
  [522] = {.lex_state = 267, .external_lex_state = 2},
  [523] = {.lex_state = 267, .external_lex_state = 2},
  [524] = {.lex_state = 267, .external_lex_state = 2},
  [525] = {.lex_state = 267, .external_lex_state = 2},
  [526] = {.lex_state = 267, .external_lex_state = 2},
  [527] = {.lex_state = 267, .external_lex_state = 2},
  [528] = {.lex_state = 267, .external_lex_state = 2},
  [529] = {.lex_state = 267, .external_lex_state = 2},
  [530] = {.lex_state = 267, .external_lex_state = 2},
  [531] = {.lex_state = 267, .external_lex_state = 2},
  [532] = {.lex_state = 267, .external_lex_state = 2},
  [533] = {.lex_state = 267, .external_lex_state = 2},
  [534] = {.lex_state = 267, .external_lex_state = 2},
  [535] = {.lex_state = 267, .external_lex_state = 2},
  [536] = {.lex_state = 267, .external_lex_state = 2},
  [537] = {.lex_state = 267, .external_lex_state = 2},
  [538] = {.lex_state = 267, .external_lex_state = 2},
  [539] = {.lex_state = 267, .external_lex_state = 2},
  [540] = {.lex_state = 267, .external_lex_state = 2},
  [541] = {.lex_state = 267, .external_lex_state = 2},
  [542] = {.lex_state = 267, .external_lex_state = 2},
  [543] = {.lex_state = 267, .external_lex_state = 2},
  [544] = {.lex_state = 267, .external_lex_state = 2},
  [545] = {.lex_state = 267, .external_lex_state = 2},
  [546] = {.lex_state = 267, .external_lex_state = 2},
  [547] = {.lex_state = 267, .external_lex_state = 2},
  [548] = {.lex_state = 267, .external_lex_state = 2},
  [549] = {.lex_state = 267, .external_lex_state = 2},
  [550] = {.lex_state = 267, .external_lex_state = 2},
  [551] = {.lex_state = 267, .external_lex_state = 2},
  [552] = {.lex_state = 267, .external_lex_state = 2},
  [553] = {.lex_state = 267, .external_lex_state = 2},
  [554] = {.lex_state = 267, .external_lex_state = 2},
  [555] = {.lex_state = 267, .external_lex_state = 2},
  [556] = {.lex_state = 267, .external_lex_state = 2},
  [557] = {.lex_state = 267, .external_lex_state = 2},
  [558] = {.lex_state = 267, .external_lex_state = 2},
  [559] = {.lex_state = 267, .external_lex_state = 2},
  [560] = {.lex_state = 267, .external_lex_state = 2},
  [561] = {.lex_state = 267, .external_lex_state = 2},
  [562] = {.lex_state = 267, .external_lex_state = 2},
  [563] = {.lex_state = 267, .external_lex_state = 2},
  [564] = {.lex_state = 267, .external_lex_state = 2},
  [565] = {.lex_state = 267, .external_lex_state = 2},
  [566] = {.lex_state = 267, .external_lex_state = 2},
  [567] = {.lex_state = 267, .external_lex_state = 2},
  [568] = {.lex_state = 267, .external_lex_state = 2},
  [569] = {.lex_state = 267, .external_lex_state = 2},
  [570] = {.lex_state = 267, .external_lex_state = 2},
  [571] = {.lex_state = 267, .external_lex_state = 2},
  [572] = {.lex_state = 267, .external_lex_state = 2},
  [573] = {.lex_state = 267, .external_lex_state = 2},
  [574] = {.lex_state = 267, .external_lex_state = 2},
  [575] = {.lex_state = 267, .external_lex_state = 2},
  [576] = {.lex_state = 267, .external_lex_state = 2},
  [577] = {.lex_state = 267, .external_lex_state = 2},
  [578] = {.lex_state = 267, .external_lex_state = 2},
  [579] = {.lex_state = 267, .external_lex_state = 2},
  [580] = {.lex_state = 267, .external_lex_state = 2},
  [581] = {.lex_state = 267, .external_lex_state = 2},
  [582] = {.lex_state = 267, .external_lex_state = 2},
  [583] = {.lex_state = 267, .external_lex_state = 2},
  [584] = {.lex_state = 267, .external_lex_state = 2},
  [585] = {.lex_state = 267, .external_lex_state = 2},
  [586] = {.lex_state = 267, .external_lex_state = 2},
  [587] = {.lex_state = 267, .external_lex_state = 2},
  [588] = {.lex_state = 267, .external_lex_state = 2},
  [589] = {.lex_state = 267, .external_lex_state = 2},
  [590] = {.lex_state = 267, .external_lex_state = 2},
  [591] = {.lex_state = 267, .external_lex_state = 2},
  [592] = {.lex_state = 267, .external_lex_state = 2},
  [593] = {.lex_state = 267, .external_lex_state = 2},
  [594] = {.lex_state = 267, .external_lex_state = 2},
  [595] = {.lex_state = 267, .external_lex_state = 2},
  [596] = {.lex_state = 267, .external_lex_state = 2},
  [597] = {.lex_state = 267, .external_lex_state = 2},
  [598] = {.lex_state = 267, .external_lex_state = 2},
  [599] = {.lex_state = 267, .external_lex_state = 2},
  [600] = {.lex_state = 267, .external_lex_state = 2},
  [601] = {.lex_state = 267, .external_lex_state = 2},
  [602] = {.lex_state = 267, .external_lex_state = 2},
  [603] = {.lex_state = 267, .external_lex_state = 2},
  [604] = {.lex_state = 267, .external_lex_state = 2},
  [605] = {.lex_state = 267, .external_lex_state = 2},
  [606] = {.lex_state = 267, .external_lex_state = 2},
  [607] = {.lex_state = 267, .external_lex_state = 2},
  [608] = {.lex_state = 267, .external_lex_state = 2},
  [609] = {.lex_state = 267, .external_lex_state = 2},
  [610] = {.lex_state = 267, .external_lex_state = 2},
  [611] = {.lex_state = 267, .external_lex_state = 2},
  [612] = {.lex_state = 267, .external_lex_state = 2},
  [613] = {.lex_state = 267, .external_lex_state = 2},
  [614] = {.lex_state = 267, .external_lex_state = 2},
  [615] = {.lex_state = 267, .external_lex_state = 2},
  [616] = {.lex_state = 267, .external_lex_state = 2},
  [617] = {.lex_state = 267, .external_lex_state = 2},
  [618] = {.lex_state = 267, .external_lex_state = 2},
  [619] = {.lex_state = 267, .external_lex_state = 2},
  [620] = {.lex_state = 267, .external_lex_state = 2},
  [621] = {.lex_state = 267, .external_lex_state = 2},
  [622] = {.lex_state = 267, .external_lex_state = 2},
  [623] = {.lex_state = 267, .external_lex_state = 2},
  [624] = {.lex_state = 267, .external_lex_state = 2},
  [625] = {.lex_state = 267, .external_lex_state = 2},
  [626] = {.lex_state = 267, .external_lex_state = 2},
  [627] = {.lex_state = 267, .external_lex_state = 2},
  [628] = {.lex_state = 267, .external_lex_state = 2},
  [629] = {.lex_state = 267, .external_lex_state = 2},
  [630] = {.lex_state = 267, .external_lex_state = 2},
  [631] = {.lex_state = 267, .external_lex_state = 2},
  [632] = {.lex_state = 267, .external_lex_state = 2},
  [633] = {.lex_state = 267, .external_lex_state = 2},
  [634] = {.lex_state = 267, .external_lex_state = 2},
  [635] = {.lex_state = 267, .external_lex_state = 2},
  [636] = {.lex_state = 267, .external_lex_state = 2},
  [637] = {.lex_state = 267, .external_lex_state = 2},
  [638] = {.lex_state = 267, .external_lex_state = 2},
  [639] = {.lex_state = 267, .external_lex_state = 2},
  [640] = {.lex_state = 267, .external_lex_state = 2},
  [641] = {.lex_state = 267, .external_lex_state = 2},
  [642] = {.lex_state = 267, .external_lex_state = 2},
  [643] = {.lex_state = 267, .external_lex_state = 2},
  [644] = {.lex_state = 267, .external_lex_state = 2},
  [645] = {.lex_state = 267, .external_lex_state = 2},
  [646] = {.lex_state = 267, .external_lex_state = 2},
  [647] = {.lex_state = 267, .external_lex_state = 2},
  [648] = {.lex_state = 267, .external_lex_state = 2},
  [649] = {.lex_state = 267, .external_lex_state = 2},
  [650] = {.lex_state = 267, .external_lex_state = 2},
  [651] = {.lex_state = 267, .external_lex_state = 2},
  [652] = {.lex_state = 267, .external_lex_state = 2},
  [653] = {.lex_state = 267, .external_lex_state = 2},
  [654] = {.lex_state = 267, .external_lex_state = 2},
  [655] = {.lex_state = 267, .external_lex_state = 2},
  [656] = {.lex_state = 267, .external_lex_state = 2},
  [657] = {.lex_state = 267, .external_lex_state = 2},
  [658] = {.lex_state = 267, .external_lex_state = 2},
  [659] = {.lex_state = 267, .external_lex_state = 2},
  [660] = {.lex_state = 267, .external_lex_state = 2},
  [661] = {.lex_state = 267, .external_lex_state = 2},
  [662] = {.lex_state = 267, .external_lex_state = 2},
  [663] = {.lex_state = 267, .external_lex_state = 2},
  [664] = {.lex_state = 267, .external_lex_state = 2},
  [665] = {.lex_state = 267, .external_lex_state = 2},
  [666] = {.lex_state = 267, .external_lex_state = 2},
  [667] = {.lex_state = 267, .external_lex_state = 2},
  [668] = {.lex_state = 267, .external_lex_state = 2},
  [669] = {.lex_state = 267, .external_lex_state = 2},
  [670] = {.lex_state = 267, .external_lex_state = 2},
  [671] = {.lex_state = 267, .external_lex_state = 2},
  [672] = {.lex_state = 267, .external_lex_state = 2},
  [673] = {.lex_state = 267, .external_lex_state = 2},
  [674] = {.lex_state = 267, .external_lex_state = 2},
  [675] = {.lex_state = 267, .external_lex_state = 2},
  [676] = {.lex_state = 267, .external_lex_state = 2},
  [677] = {.lex_state = 267, .external_lex_state = 2},
  [678] = {.lex_state = 267, .external_lex_state = 2},
  [679] = {.lex_state = 267, .external_lex_state = 2},
  [680] = {.lex_state = 267, .external_lex_state = 2},
  [681] = {.lex_state = 267, .external_lex_state = 2},
  [682] = {.lex_state = 267, .external_lex_state = 2},
  [683] = {.lex_state = 267, .external_lex_state = 2},
  [684] = {.lex_state = 267, .external_lex_state = 2},
  [685] = {.lex_state = 267, .external_lex_state = 2},
  [686] = {.lex_state = 267, .external_lex_state = 2},
  [687] = {.lex_state = 267, .external_lex_state = 2},
  [688] = {.lex_state = 267, .external_lex_state = 2},
  [689] = {.lex_state = 267, .external_lex_state = 2},
  [690] = {.lex_state = 267, .external_lex_state = 2},
  [691] = {.lex_state = 267, .external_lex_state = 2},
  [692] = {.lex_state = 267, .external_lex_state = 2},
  [693] = {.lex_state = 267, .external_lex_state = 2},
  [694] = {.lex_state = 267, .external_lex_state = 2},
  [695] = {.lex_state = 267, .external_lex_state = 2},
  [696] = {.lex_state = 267, .external_lex_state = 2},
  [697] = {.lex_state = 267, .external_lex_state = 2},
  [698] = {.lex_state = 267, .external_lex_state = 2},
  [699] = {.lex_state = 267, .external_lex_state = 2},
  [700] = {.lex_state = 267, .external_lex_state = 2},
  [701] = {.lex_state = 267, .external_lex_state = 2},
  [702] = {.lex_state = 267, .external_lex_state = 2},
  [703] = {.lex_state = 267, .external_lex_state = 2},
  [704] = {.lex_state = 267, .external_lex_state = 2},
  [705] = {.lex_state = 267, .external_lex_state = 2},
  [706] = {.lex_state = 267, .external_lex_state = 2},
  [707] = {.lex_state = 267, .external_lex_state = 2},
  [708] = {.lex_state = 267, .external_lex_state = 2},
  [709] = {.lex_state = 267, .external_lex_state = 2},
  [710] = {.lex_state = 267, .external_lex_state = 2},
  [711] = {.lex_state = 267, .external_lex_state = 2},
  [712] = {.lex_state = 267, .external_lex_state = 2},
  [713] = {.lex_state = 267, .external_lex_state = 2},
  [714] = {.lex_state = 267, .external_lex_state = 2},
  [715] = {.lex_state = 267, .external_lex_state = 2},
  [716] = {.lex_state = 267, .external_lex_state = 2},
  [717] = {.lex_state = 267, .external_lex_state = 2},
  [718] = {.lex_state = 267, .external_lex_state = 2},
  [719] = {.lex_state = 267, .external_lex_state = 2},
  [720] = {.lex_state = 267, .external_lex_state = 2},
  [721] = {.lex_state = 267, .external_lex_state = 2},
  [722] = {.lex_state = 267, .external_lex_state = 2},
  [723] = {.lex_state = 267, .external_lex_state = 2},
  [724] = {.lex_state = 267, .external_lex_state = 2},
  [725] = {.lex_state = 267, .external_lex_state = 2},
  [726] = {.lex_state = 267, .external_lex_state = 2},
  [727] = {.lex_state = 267, .external_lex_state = 2},
  [728] = {.lex_state = 267, .external_lex_state = 2},
  [729] = {.lex_state = 267, .external_lex_state = 2},
  [730] = {.lex_state = 267, .external_lex_state = 2},
  [731] = {.lex_state = 267, .external_lex_state = 2},
  [732] = {.lex_state = 267, .external_lex_state = 2},
  [733] = {.lex_state = 267, .external_lex_state = 2},
  [734] = {.lex_state = 267, .external_lex_state = 2},
  [735] = {.lex_state = 267, .external_lex_state = 2},
  [736] = {.lex_state = 267, .external_lex_state = 2},
  [737] = {.lex_state = 267, .external_lex_state = 2},
  [738] = {.lex_state = 267, .external_lex_state = 2},
  [739] = {.lex_state = 267, .external_lex_state = 2},
  [740] = {.lex_state = 267, .external_lex_state = 2},
  [741] = {.lex_state = 267, .external_lex_state = 2},
  [742] = {.lex_state = 267, .external_lex_state = 2},
  [743] = {.lex_state = 267, .external_lex_state = 2},
  [744] = {.lex_state = 267, .external_lex_state = 2},
  [745] = {.lex_state = 267, .external_lex_state = 2},
  [746] = {.lex_state = 267, .external_lex_state = 2},
  [747] = {.lex_state = 267, .external_lex_state = 2},
  [748] = {.lex_state = 267, .external_lex_state = 2},
  [749] = {.lex_state = 267, .external_lex_state = 2},
  [750] = {.lex_state = 267, .external_lex_state = 2},
  [751] = {.lex_state = 267, .external_lex_state = 2},
  [752] = {.lex_state = 267, .external_lex_state = 2},
  [753] = {.lex_state = 267, .external_lex_state = 2},
  [754] = {.lex_state = 267, .external_lex_state = 2},
  [755] = {.lex_state = 267, .external_lex_state = 2},
  [756] = {.lex_state = 267, .external_lex_state = 2},
  [757] = {.lex_state = 267, .external_lex_state = 2},
  [758] = {.lex_state = 267, .external_lex_state = 2},
  [759] = {.lex_state = 267, .external_lex_state = 2},
  [760] = {.lex_state = 267, .external_lex_state = 2},
  [761] = {.lex_state = 267, .external_lex_state = 2},
  [762] = {.lex_state = 267, .external_lex_state = 2},
  [763] = {.lex_state = 267, .external_lex_state = 2},
  [764] = {.lex_state = 267, .external_lex_state = 2},
  [765] = {.lex_state = 267, .external_lex_state = 2},
  [766] = {.lex_state = 267, .external_lex_state = 2},
  [767] = {.lex_state = 267, .external_lex_state = 2},
  [768] = {.lex_state = 267, .external_lex_state = 2},
  [769] = {.lex_state = 267, .external_lex_state = 2},
  [770] = {.lex_state = 267, .external_lex_state = 2},
  [771] = {.lex_state = 267, .external_lex_state = 2},
  [772] = {.lex_state = 267, .external_lex_state = 2},
  [773] = {.lex_state = 267, .external_lex_state = 2},
  [774] = {.lex_state = 267, .external_lex_state = 2},
  [775] = {.lex_state = 267, .external_lex_state = 2},
  [776] = {.lex_state = 267, .external_lex_state = 2},
  [777] = {.lex_state = 267, .external_lex_state = 2},
  [778] = {.lex_state = 267, .external_lex_state = 2},
  [779] = {.lex_state = 267, .external_lex_state = 2},
  [780] = {.lex_state = 267, .external_lex_state = 2},
  [781] = {.lex_state = 267, .external_lex_state = 2},
  [782] = {.lex_state = 267, .external_lex_state = 2},
  [783] = {.lex_state = 267, .external_lex_state = 2},
  [784] = {.lex_state = 267, .external_lex_state = 2},
  [785] = {.lex_state = 267, .external_lex_state = 2},
  [786] = {.lex_state = 267, .external_lex_state = 2},
  [787] = {.lex_state = 267, .external_lex_state = 2},
  [788] = {.lex_state = 267, .external_lex_state = 2},
  [789] = {.lex_state = 267, .external_lex_state = 2},
  [790] = {.lex_state = 267, .external_lex_state = 2},
  [791] = {.lex_state = 267, .external_lex_state = 2},
  [792] = {.lex_state = 267, .external_lex_state = 2},
  [793] = {.lex_state = 267, .external_lex_state = 2},
  [794] = {.lex_state = 267, .external_lex_state = 2},
  [795] = {.lex_state = 267, .external_lex_state = 2},
  [796] = {.lex_state = 267, .external_lex_state = 2},
  [797] = {.lex_state = 267, .external_lex_state = 2},
  [798] = {.lex_state = 267, .external_lex_state = 2},
  [799] = {.lex_state = 267, .external_lex_state = 2},
  [800] = {.lex_state = 267, .external_lex_state = 2},
  [801] = {.lex_state = 267, .external_lex_state = 2},
  [802] = {.lex_state = 267, .external_lex_state = 2},
  [803] = {.lex_state = 267, .external_lex_state = 2},
  [804] = {.lex_state = 267, .external_lex_state = 2},
  [805] = {.lex_state = 267, .external_lex_state = 2},
  [806] = {.lex_state = 267, .external_lex_state = 2},
  [807] = {.lex_state = 267, .external_lex_state = 2},
  [808] = {.lex_state = 267, .external_lex_state = 2},
  [809] = {.lex_state = 267, .external_lex_state = 2},
  [810] = {.lex_state = 267, .external_lex_state = 2},
  [811] = {.lex_state = 267, .external_lex_state = 2},
  [812] = {.lex_state = 267, .external_lex_state = 2},
  [813] = {.lex_state = 267, .external_lex_state = 2},
  [814] = {.lex_state = 267, .external_lex_state = 2},
  [815] = {.lex_state = 267, .external_lex_state = 2},
  [816] = {.lex_state = 267, .external_lex_state = 2},
  [817] = {.lex_state = 267, .external_lex_state = 2},
  [818] = {.lex_state = 267, .external_lex_state = 2},
  [819] = {.lex_state = 267, .external_lex_state = 2},
  [820] = {.lex_state = 267, .external_lex_state = 2},
  [821] = {.lex_state = 267, .external_lex_state = 2},
  [822] = {.lex_state = 267, .external_lex_state = 2},
  [823] = {.lex_state = 267, .external_lex_state = 2},
  [824] = {.lex_state = 267, .external_lex_state = 2},
  [825] = {.lex_state = 267, .external_lex_state = 2},
  [826] = {.lex_state = 267, .external_lex_state = 2},
  [827] = {.lex_state = 267, .external_lex_state = 2},
  [828] = {.lex_state = 267, .external_lex_state = 2},
  [829] = {.lex_state = 267, .external_lex_state = 2},
  [830] = {.lex_state = 267, .external_lex_state = 2},
  [831] = {.lex_state = 267, .external_lex_state = 2},
  [832] = {.lex_state = 267, .external_lex_state = 2},
  [833] = {.lex_state = 267, .external_lex_state = 2},
  [834] = {.lex_state = 267, .external_lex_state = 2},
  [835] = {.lex_state = 267, .external_lex_state = 2},
  [836] = {.lex_state = 267, .external_lex_state = 2},
  [837] = {.lex_state = 267, .external_lex_state = 2},
  [838] = {.lex_state = 267, .external_lex_state = 2},
  [839] = {.lex_state = 267, .external_lex_state = 2},
  [840] = {.lex_state = 267, .external_lex_state = 2},
  [841] = {.lex_state = 267, .external_lex_state = 2},
  [842] = {.lex_state = 267, .external_lex_state = 2},
  [843] = {.lex_state = 267, .external_lex_state = 2},
  [844] = {.lex_state = 267, .external_lex_state = 2},
  [845] = {.lex_state = 267, .external_lex_state = 2},
  [846] = {.lex_state = 267, .external_lex_state = 2},
  [847] = {.lex_state = 267, .external_lex_state = 2},
  [848] = {.lex_state = 267, .external_lex_state = 2},
  [849] = {.lex_state = 267, .external_lex_state = 2},
  [850] = {.lex_state = 267, .external_lex_state = 2},
  [851] = {.lex_state = 267, .external_lex_state = 2},
  [852] = {.lex_state = 267, .external_lex_state = 2},
  [853] = {.lex_state = 267, .external_lex_state = 2},
  [854] = {.lex_state = 267, .external_lex_state = 2},
  [855] = {.lex_state = 267, .external_lex_state = 2},
  [856] = {.lex_state = 267, .external_lex_state = 2},
  [857] = {.lex_state = 267, .external_lex_state = 2},
  [858] = {.lex_state = 267, .external_lex_state = 2},
  [859] = {.lex_state = 267, .external_lex_state = 2},
  [860] = {.lex_state = 267, .external_lex_state = 2},
  [861] = {.lex_state = 267, .external_lex_state = 2},
  [862] = {.lex_state = 267, .external_lex_state = 2},
  [863] = {.lex_state = 267, .external_lex_state = 2},
  [864] = {.lex_state = 267, .external_lex_state = 2},
  [865] = {.lex_state = 267, .external_lex_state = 2},
  [866] = {.lex_state = 267, .external_lex_state = 2},
  [867] = {.lex_state = 267, .external_lex_state = 2},
  [868] = {.lex_state = 267, .external_lex_state = 2},
  [869] = {.lex_state = 267, .external_lex_state = 2},
  [870] = {.lex_state = 267, .external_lex_state = 2},
  [871] = {.lex_state = 267, .external_lex_state = 2},
  [872] = {.lex_state = 267, .external_lex_state = 2},
  [873] = {.lex_state = 267, .external_lex_state = 2},
  [874] = {.lex_state = 267, .external_lex_state = 2},
  [875] = {.lex_state = 267, .external_lex_state = 2},
  [876] = {.lex_state = 267, .external_lex_state = 2},
  [877] = {.lex_state = 267, .external_lex_state = 2},
  [878] = {.lex_state = 267, .external_lex_state = 2},
  [879] = {.lex_state = 267, .external_lex_state = 2},
  [880] = {.lex_state = 267, .external_lex_state = 2},
  [881] = {.lex_state = 267, .external_lex_state = 2},
  [882] = {.lex_state = 267, .external_lex_state = 2},
  [883] = {.lex_state = 267, .external_lex_state = 2},
  [884] = {.lex_state = 267, .external_lex_state = 2},
  [885] = {.lex_state = 267, .external_lex_state = 2},
  [886] = {.lex_state = 267, .external_lex_state = 2},
  [887] = {.lex_state = 267, .external_lex_state = 2},
  [888] = {.lex_state = 267, .external_lex_state = 2},
  [889] = {.lex_state = 267, .external_lex_state = 2},
  [890] = {.lex_state = 267, .external_lex_state = 2},
  [891] = {.lex_state = 267, .external_lex_state = 2},
  [892] = {.lex_state = 267, .external_lex_state = 2},
  [893] = {.lex_state = 267, .external_lex_state = 2},
  [894] = {.lex_state = 267, .external_lex_state = 2},
  [895] = {.lex_state = 267, .external_lex_state = 2},
  [896] = {.lex_state = 267, .external_lex_state = 2},
  [897] = {.lex_state = 267, .external_lex_state = 2},
  [898] = {.lex_state = 267, .external_lex_state = 2},
  [899] = {.lex_state = 267, .external_lex_state = 2},
  [900] = {.lex_state = 267, .external_lex_state = 2},
  [901] = {.lex_state = 267, .external_lex_state = 2},
  [902] = {.lex_state = 267, .external_lex_state = 2},
  [903] = {.lex_state = 267, .external_lex_state = 2},
  [904] = {.lex_state = 267, .external_lex_state = 2},
  [905] = {.lex_state = 267, .external_lex_state = 2},
  [906] = {.lex_state = 267, .external_lex_state = 2},
  [907] = {.lex_state = 267, .external_lex_state = 2},
  [908] = {.lex_state = 267, .external_lex_state = 2},
  [909] = {.lex_state = 267, .external_lex_state = 2},
  [910] = {.lex_state = 83, .external_lex_state = 3},
  [911] = {.lex_state = 83, .external_lex_state = 3},
  [912] = {.lex_state = 83, .external_lex_state = 3},
  [913] = {.lex_state = 83, .external_lex_state = 3},
  [914] = {.lex_state = 83, .external_lex_state = 3},
  [915] = {.lex_state = 83, .external_lex_state = 3},
  [916] = {.lex_state = 83, .external_lex_state = 3},
  [917] = {.lex_state = 83, .external_lex_state = 3},
  [918] = {.lex_state = 83, .external_lex_state = 3},
  [919] = {.lex_state = 83, .external_lex_state = 3},
  [920] = {.lex_state = 83, .external_lex_state = 3},
  [921] = {.lex_state = 83, .external_lex_state = 3},
  [922] = {.lex_state = 83, .external_lex_state = 3},
  [923] = {.lex_state = 83, .external_lex_state = 3},
  [924] = {.lex_state = 83, .external_lex_state = 3},
  [925] = {.lex_state = 83, .external_lex_state = 3},
  [926] = {.lex_state = 83, .external_lex_state = 3},
  [927] = {.lex_state = 83, .external_lex_state = 2},
  [928] = {.lex_state = 269, .external_lex_state = 3},
  [929] = {.lex_state = 269, .external_lex_state = 3},
  [930] = {.lex_state = 269, .external_lex_state = 3},
  [931] = {.lex_state = 269, .external_lex_state = 3},
  [932] = {.lex_state = 269, .external_lex_state = 3},
  [933] = {.lex_state = 269, .external_lex_state = 3},
  [934] = {.lex_state = 269, .external_lex_state = 3},
  [935] = {.lex_state = 269, .external_lex_state = 3},
  [936] = {.lex_state = 269, .external_lex_state = 3},
  [937] = {.lex_state = 269, .external_lex_state = 3},
  [938] = {.lex_state = 269, .external_lex_state = 3},
  [939] = {.lex_state = 269, .external_lex_state = 3},
  [940] = {.lex_state = 269, .external_lex_state = 3},
  [941] = {.lex_state = 269, .external_lex_state = 3},
  [942] = {.lex_state = 269, .external_lex_state = 3},
  [943] = {.lex_state = 269, .external_lex_state = 3},
  [944] = {.lex_state = 269, .external_lex_state = 3},
  [945] = {.lex_state = 269, .external_lex_state = 3},
  [946] = {.lex_state = 88, .external_lex_state = 3},
  [947] = {.lex_state = 269, .external_lex_state = 3},
  [948] = {.lex_state = 88, .external_lex_state = 3},
  [949] = {.lex_state = 88, .external_lex_state = 3},
  [950] = {.lex_state = 88, .external_lex_state = 3},
  [951] = {.lex_state = 88, .external_lex_state = 3},
  [952] = {.lex_state = 88, .external_lex_state = 3},
  [953] = {.lex_state = 88, .external_lex_state = 3},
  [954] = {.lex_state = 269, .external_lex_state = 3},
  [955] = {.lex_state = 269, .external_lex_state = 3},
  [956] = {.lex_state = 88, .external_lex_state = 3},
  [957] = {.lex_state = 88, .external_lex_state = 3},
  [958] = {.lex_state = 269, .external_lex_state = 3},
  [959] = {.lex_state = 269, .external_lex_state = 3},
  [960] = {.lex_state = 269, .external_lex_state = 3},
  [961] = {.lex_state = 88, .external_lex_state = 3},
  [962] = {.lex_state = 88, .external_lex_state = 3},
  [963] = {.lex_state = 88, .external_lex_state = 3},
  [964] = {.lex_state = 88, .external_lex_state = 3},
  [965] = {.lex_state = 88, .external_lex_state = 3},
  [966] = {.lex_state = 88, .external_lex_state = 3},
  [967] = {.lex_state = 88, .external_lex_state = 3},
  [968] = {.lex_state = 88, .external_lex_state = 3},
  [969] = {.lex_state = 269, .external_lex_state = 3},
  [970] = {.lex_state = 269, .external_lex_state = 3},
  [971] = {.lex_state = 269, .external_lex_state = 3},
  [972] = {.lex_state = 269, .external_lex_state = 3},
  [973] = {.lex_state = 269, .external_lex_state = 3},
  [974] = {.lex_state = 269, .external_lex_state = 3},
  [975] = {.lex_state = 269, .external_lex_state = 3},
  [976] = {.lex_state = 269, .external_lex_state = 3},
  [977] = {.lex_state = 269, .external_lex_state = 3},
  [978] = {.lex_state = 269, .external_lex_state = 3},
  [979] = {.lex_state = 269, .external_lex_state = 3},
  [980] = {.lex_state = 269, .external_lex_state = 3},
  [981] = {.lex_state = 269, .external_lex_state = 3},
  [982] = {.lex_state = 269, .external_lex_state = 3},
  [983] = {.lex_state = 269, .external_lex_state = 3},
  [984] = {.lex_state = 269, .external_lex_state = 3},
  [985] = {.lex_state = 269, .external_lex_state = 3},
  [986] = {.lex_state = 269, .external_lex_state = 3},
  [987] = {.lex_state = 269, .external_lex_state = 2},
  [988] = {.lex_state = 269, .external_lex_state = 3},
  [989] = {.lex_state = 269, .external_lex_state = 3},
  [990] = {.lex_state = 269, .external_lex_state = 3},
  [991] = {.lex_state = 269, .external_lex_state = 3},
  [992] = {.lex_state = 269, .external_lex_state = 3},
  [993] = {.lex_state = 269, .external_lex_state = 3},
  [994] = {.lex_state = 269, .external_lex_state = 3},
  [995] = {.lex_state = 269, .external_lex_state = 3},
  [996] = {.lex_state = 269, .external_lex_state = 3},
  [997] = {.lex_state = 91, .external_lex_state = 3},
  [998] = {.lex_state = 91, .external_lex_state = 3},
  [999] = {.lex_state = 91, .external_lex_state = 3},
  [1000] = {.lex_state = 91, .external_lex_state = 3},
  [1001] = {.lex_state = 91, .external_lex_state = 3},
  [1002] = {.lex_state = 91, .external_lex_state = 3},
  [1003] = {.lex_state = 91, .external_lex_state = 3},
  [1004] = {.lex_state = 91, .external_lex_state = 3},
  [1005] = {.lex_state = 88, .external_lex_state = 2},
  [1006] = {.lex_state = 91, .external_lex_state = 3},
  [1007] = {.lex_state = 269, .external_lex_state = 2},
  [1008] = {.lex_state = 91, .external_lex_state = 3},
  [1009] = {.lex_state = 91, .external_lex_state = 3},
  [1010] = {.lex_state = 91, .external_lex_state = 3},
  [1011] = {.lex_state = 91, .external_lex_state = 3},
  [1012] = {.lex_state = 91, .external_lex_state = 3},
  [1013] = {.lex_state = 269, .external_lex_state = 2},
  [1014] = {.lex_state = 91, .external_lex_state = 3},
  [1015] = {.lex_state = 91, .external_lex_state = 3},
  [1016] = {.lex_state = 91, .external_lex_state = 3},
  [1017] = {.lex_state = 91, .external_lex_state = 2},
  [1018] = {.lex_state = 81, .external_lex_state = 2},
  [1019] = {.lex_state = 81, .external_lex_state = 2},
  [1020] = {.lex_state = 98, .external_lex_state = 2},
  [1021] = {.lex_state = 98, .external_lex_state = 2},
  [1022] = {.lex_state = 98, .external_lex_state = 2},
  [1023] = {.lex_state = 81, .external_lex_state = 2},
  [1024] = {.lex_state = 98, .external_lex_state = 2},
  [1025] = {.lex_state = 98, .external_lex_state = 2},
  [1026] = {.lex_state = 94, .external_lex_state = 2},
  [1027] = {.lex_state = 96, .external_lex_state = 2},
  [1028] = {.lex_state = 96, .external_lex_state = 2},
  [1029] = {.lex_state = 94, .external_lex_state = 2},
  [1030] = {.lex_state = 267, .external_lex_state = 2},
  [1031] = {.lex_state = 267, .external_lex_state = 2},
  [1032] = {.lex_state = 102, .external_lex_state = 2},
  [1033] = {.lex_state = 102, .external_lex_state = 2},
  [1034] = {.lex_state = 267, .external_lex_state = 2},
  [1035] = {.lex_state = 267, .external_lex_state = 2},
  [1036] = {.lex_state = 267, .external_lex_state = 2},
  [1037] = {.lex_state = 94, .external_lex_state = 2},
  [1038] = {.lex_state = 267, .external_lex_state = 2},
  [1039] = {.lex_state = 102, .external_lex_state = 2},
  [1040] = {.lex_state = 96, .external_lex_state = 2},
  [1041] = {.lex_state = 267, .external_lex_state = 2},
  [1042] = {.lex_state = 267, .external_lex_state = 2},
  [1043] = {.lex_state = 267, .external_lex_state = 2},
  [1044] = {.lex_state = 267, .external_lex_state = 2},
  [1045] = {.lex_state = 104, .external_lex_state = 4},
  [1046] = {.lex_state = 104, .external_lex_state = 4},
  [1047] = {.lex_state = 104, .external_lex_state = 4},
  [1048] = {.lex_state = 104, .external_lex_state = 4},
  [1049] = {.lex_state = 104, .external_lex_state = 4},
  [1050] = {.lex_state = 104, .external_lex_state = 4},
  [1051] = {.lex_state = 104, .external_lex_state = 4},
  [1052] = {.lex_state = 104, .external_lex_state = 4},
  [1053] = {.lex_state = 104, .external_lex_state = 4},
  [1054] = {.lex_state = 104, .external_lex_state = 4},
  [1055] = {.lex_state = 104, .external_lex_state = 4},
  [1056] = {.lex_state = 104, .external_lex_state = 4},
  [1057] = {.lex_state = 104, .external_lex_state = 4},
  [1058] = {.lex_state = 104, .external_lex_state = 4},
  [1059] = {.lex_state = 104, .external_lex_state = 4},
  [1060] = {.lex_state = 104, .external_lex_state = 4},
  [1061] = {.lex_state = 104, .external_lex_state = 4},
  [1062] = {.lex_state = 104, .external_lex_state = 4},
  [1063] = {.lex_state = 106, .external_lex_state = 4},
  [1064] = {.lex_state = 106, .external_lex_state = 4},
  [1065] = {.lex_state = 106, .external_lex_state = 4},
  [1066] = {.lex_state = 106, .external_lex_state = 4},
  [1067] = {.lex_state = 106, .external_lex_state = 4},
  [1068] = {.lex_state = 106, .external_lex_state = 4},
  [1069] = {.lex_state = 106, .external_lex_state = 4},
  [1070] = {.lex_state = 106, .external_lex_state = 4},
  [1071] = {.lex_state = 106, .external_lex_state = 4},
  [1072] = {.lex_state = 106, .external_lex_state = 4},
  [1073] = {.lex_state = 106, .external_lex_state = 4},
  [1074] = {.lex_state = 106, .external_lex_state = 4},
  [1075] = {.lex_state = 106, .external_lex_state = 4},
  [1076] = {.lex_state = 106, .external_lex_state = 4},
  [1077] = {.lex_state = 106, .external_lex_state = 4},
  [1078] = {.lex_state = 106, .external_lex_state = 4},
  [1079] = {.lex_state = 106, .external_lex_state = 4},
  [1080] = {.lex_state = 106, .external_lex_state = 4},
  [1081] = {.lex_state = 106, .external_lex_state = 4},
  [1082] = {.lex_state = 106, .external_lex_state = 4},
  [1083] = {.lex_state = 106, .external_lex_state = 4},
  [1084] = {.lex_state = 106, .external_lex_state = 4},
  [1085] = {.lex_state = 106, .external_lex_state = 4},
  [1086] = {.lex_state = 106, .external_lex_state = 4},
  [1087] = {.lex_state = 106, .external_lex_state = 5},
  [1088] = {.lex_state = 106, .external_lex_state = 5},
  [1089] = {.lex_state = 106, .external_lex_state = 5},
  [1090] = {.lex_state = 106, .external_lex_state = 5},
  [1091] = {.lex_state = 106, .external_lex_state = 5},
  [1092] = {.lex_state = 106, .external_lex_state = 5},
  [1093] = {.lex_state = 106, .external_lex_state = 5},
  [1094] = {.lex_state = 106, .external_lex_state = 4},
  [1095] = {.lex_state = 106, .external_lex_state = 5},
  [1096] = {.lex_state = 106, .external_lex_state = 5},
  [1097] = {.lex_state = 106, .external_lex_state = 5},
  [1098] = {.lex_state = 106, .external_lex_state = 5},
  [1099] = {.lex_state = 106, .external_lex_state = 4},
  [1100] = {.lex_state = 106, .external_lex_state = 5},
  [1101] = {.lex_state = 106, .external_lex_state = 4},
  [1102] = {.lex_state = 106, .external_lex_state = 5},
  [1103] = {.lex_state = 106, .external_lex_state = 5},
  [1104] = {.lex_state = 106, .external_lex_state = 5},
  [1105] = {.lex_state = 106, .external_lex_state = 5},
  [1106] = {.lex_state = 106, .external_lex_state = 4},
  [1107] = {.lex_state = 271, .external_lex_state = 5},
  [1108] = {.lex_state = 106, .external_lex_state = 5},
  [1109] = {.lex_state = 271, .external_lex_state = 5},
  [1110] = {.lex_state = 106, .external_lex_state = 5},
  [1111] = {.lex_state = 106, .external_lex_state = 5},
  [1112] = {.lex_state = 106, .external_lex_state = 5},
  [1113] = {.lex_state = 106, .external_lex_state = 5},
  [1114] = {.lex_state = 106, .external_lex_state = 5},
  [1115] = {.lex_state = 106, .external_lex_state = 5},
  [1116] = {.lex_state = 106, .external_lex_state = 4},
  [1117] = {.lex_state = 106, .external_lex_state = 5},
  [1118] = {.lex_state = 106, .external_lex_state = 5},
  [1119] = {.lex_state = 271, .external_lex_state = 5},
  [1120] = {.lex_state = 106, .external_lex_state = 4},
  [1121] = {.lex_state = 271, .external_lex_state = 5},
  [1122] = {.lex_state = 106, .external_lex_state = 5},
  [1123] = {.lex_state = 106, .external_lex_state = 5},
  [1124] = {.lex_state = 106, .external_lex_state = 5},
  [1125] = {.lex_state = 106, .external_lex_state = 5},
  [1126] = {.lex_state = 106, .external_lex_state = 4},
  [1127] = {.lex_state = 106, .external_lex_state = 5},
  [1128] = {.lex_state = 106, .external_lex_state = 5},
  [1129] = {.lex_state = 106, .external_lex_state = 4},
  [1130] = {.lex_state = 106, .external_lex_state = 5},
  [1131] = {.lex_state = 106, .external_lex_state = 5},
  [1132] = {.lex_state = 106, .external_lex_state = 5},
  [1133] = {.lex_state = 106, .external_lex_state = 5},
  [1134] = {.lex_state = 271, .external_lex_state = 5},
  [1135] = {.lex_state = 106, .external_lex_state = 5},
  [1136] = {.lex_state = 271, .external_lex_state = 5},
  [1137] = {.lex_state = 106, .external_lex_state = 5},
  [1138] = {.lex_state = 106, .external_lex_state = 5},
  [1139] = {.lex_state = 271, .external_lex_state = 5},
  [1140] = {.lex_state = 106, .external_lex_state = 5},
  [1141] = {.lex_state = 106, .external_lex_state = 5},
  [1142] = {.lex_state = 106, .external_lex_state = 5},
  [1143] = {.lex_state = 111, .external_lex_state = 5},
  [1144] = {.lex_state = 111, .external_lex_state = 5},
  [1145] = {.lex_state = 111, .external_lex_state = 5},
  [1146] = {.lex_state = 111, .external_lex_state = 5},
  [1147] = {.lex_state = 111, .external_lex_state = 5},
  [1148] = {.lex_state = 111, .external_lex_state = 5},
  [1149] = {.lex_state = 111, .external_lex_state = 5},
  [1150] = {.lex_state = 111, .external_lex_state = 5},
  [1151] = {.lex_state = 111, .external_lex_state = 5},
  [1152] = {.lex_state = 111, .external_lex_state = 5},
  [1153] = {.lex_state = 111, .external_lex_state = 5},
  [1154] = {.lex_state = 111, .external_lex_state = 5},
  [1155] = {.lex_state = 111, .external_lex_state = 5},
  [1156] = {.lex_state = 111, .external_lex_state = 5},
  [1157] = {.lex_state = 111, .external_lex_state = 5},
  [1158] = {.lex_state = 111, .external_lex_state = 5},
  [1159] = {.lex_state = 111, .external_lex_state = 5},
  [1160] = {.lex_state = 111, .external_lex_state = 5},
  [1161] = {.lex_state = 111, .external_lex_state = 5},
  [1162] = {.lex_state = 111, .external_lex_state = 5},
  [1163] = {.lex_state = 111, .external_lex_state = 5},
  [1164] = {.lex_state = 111, .external_lex_state = 5},
  [1165] = {.lex_state = 111, .external_lex_state = 5},
  [1166] = {.lex_state = 111, .external_lex_state = 5},
  [1167] = {.lex_state = 111, .external_lex_state = 5},
  [1168] = {.lex_state = 111, .external_lex_state = 5},
  [1169] = {.lex_state = 111, .external_lex_state = 5},
  [1170] = {.lex_state = 111, .external_lex_state = 5},
  [1171] = {.lex_state = 111, .external_lex_state = 5},
  [1172] = {.lex_state = 111, .external_lex_state = 5},
  [1173] = {.lex_state = 111, .external_lex_state = 5},
  [1174] = {.lex_state = 111, .external_lex_state = 5},
  [1175] = {.lex_state = 111, .external_lex_state = 5},
  [1176] = {.lex_state = 111, .external_lex_state = 5},
  [1177] = {.lex_state = 111, .external_lex_state = 5},
  [1178] = {.lex_state = 111, .external_lex_state = 5},
  [1179] = {.lex_state = 111, .external_lex_state = 5},
  [1180] = {.lex_state = 111, .external_lex_state = 5},
  [1181] = {.lex_state = 111, .external_lex_state = 5},
  [1182] = {.lex_state = 111, .external_lex_state = 5},
  [1183] = {.lex_state = 111, .external_lex_state = 5},
  [1184] = {.lex_state = 111, .external_lex_state = 5},
  [1185] = {.lex_state = 106, .external_lex_state = 5},
  [1186] = {.lex_state = 106, .external_lex_state = 5},
  [1187] = {.lex_state = 106, .external_lex_state = 5},
  [1188] = {.lex_state = 106, .external_lex_state = 5},
  [1189] = {.lex_state = 106, .external_lex_state = 5},
  [1190] = {.lex_state = 271, .external_lex_state = 5},
  [1191] = {.lex_state = 271, .external_lex_state = 5},
  [1192] = {.lex_state = 106, .external_lex_state = 5},
  [1193] = {.lex_state = 106, .external_lex_state = 4},
  [1194] = {.lex_state = 106, .external_lex_state = 4},
  [1195] = {.lex_state = 106, .external_lex_state = 5},
  [1196] = {.lex_state = 106, .external_lex_state = 5},
  [1197] = {.lex_state = 106, .external_lex_state = 5},
  [1198] = {.lex_state = 106, .external_lex_state = 5},
  [1199] = {.lex_state = 106, .external_lex_state = 5},
  [1200] = {.lex_state = 106, .external_lex_state = 5},
  [1201] = {.lex_state = 106, .external_lex_state = 5},
  [1202] = {.lex_state = 106, .external_lex_state = 5},
  [1203] = {.lex_state = 106, .external_lex_state = 5},
  [1204] = {.lex_state = 271, .external_lex_state = 5},
  [1205] = {.lex_state = 106, .external_lex_state = 5},
  [1206] = {.lex_state = 106, .external_lex_state = 5},
  [1207] = {.lex_state = 106, .external_lex_state = 5},
  [1208] = {.lex_state = 106, .external_lex_state = 5},
  [1209] = {.lex_state = 106, .external_lex_state = 5},
  [1210] = {.lex_state = 111, .external_lex_state = 5},
  [1211] = {.lex_state = 111, .external_lex_state = 5},
  [1212] = {.lex_state = 111, .external_lex_state = 5},
  [1213] = {.lex_state = 111, .external_lex_state = 5},
  [1214] = {.lex_state = 106, .external_lex_state = 5},
  [1215] = {.lex_state = 111, .external_lex_state = 5},
  [1216] = {.lex_state = 111, .external_lex_state = 5},
  [1217] = {.lex_state = 106, .external_lex_state = 5},
  [1218] = {.lex_state = 106, .external_lex_state = 5},
  [1219] = {.lex_state = 271, .external_lex_state = 5},
  [1220] = {.lex_state = 271, .external_lex_state = 5},
  [1221] = {.lex_state = 271, .external_lex_state = 5},
  [1222] = {.lex_state = 111, .external_lex_state = 5},
  [1223] = {.lex_state = 111, .external_lex_state = 5},
  [1224] = {.lex_state = 111, .external_lex_state = 5},
  [1225] = {.lex_state = 111, .external_lex_state = 5},
  [1226] = {.lex_state = 111, .external_lex_state = 5},
  [1227] = {.lex_state = 111, .external_lex_state = 5},
  [1228] = {.lex_state = 271, .external_lex_state = 5},
  [1229] = {.lex_state = 271, .external_lex_state = 5},
  [1230] = {.lex_state = 271, .external_lex_state = 5},
  [1231] = {.lex_state = 271, .external_lex_state = 5},
  [1232] = {.lex_state = 106, .external_lex_state = 5},
  [1233] = {.lex_state = 106, .external_lex_state = 5},
  [1234] = {.lex_state = 106, .external_lex_state = 5},
  [1235] = {.lex_state = 106, .external_lex_state = 5},
  [1236] = {.lex_state = 106, .external_lex_state = 5},
  [1237] = {.lex_state = 106, .external_lex_state = 5},
  [1238] = {.lex_state = 106, .external_lex_state = 5},
  [1239] = {.lex_state = 106, .external_lex_state = 5},
  [1240] = {.lex_state = 106, .external_lex_state = 5},
  [1241] = {.lex_state = 106, .external_lex_state = 5},
  [1242] = {.lex_state = 106, .external_lex_state = 5},
  [1243] = {.lex_state = 106, .external_lex_state = 5},
  [1244] = {.lex_state = 106, .external_lex_state = 5},
  [1245] = {.lex_state = 106, .external_lex_state = 5},
  [1246] = {.lex_state = 106, .external_lex_state = 5},
  [1247] = {.lex_state = 106, .external_lex_state = 5},
  [1248] = {.lex_state = 106, .external_lex_state = 5},
  [1249] = {.lex_state = 106, .external_lex_state = 5},
  [1250] = {.lex_state = 106, .external_lex_state = 5},
  [1251] = {.lex_state = 106, .external_lex_state = 5},
  [1252] = {.lex_state = 106, .external_lex_state = 5},
  [1253] = {.lex_state = 271, .external_lex_state = 5},
  [1254] = {.lex_state = 271, .external_lex_state = 5},
  [1255] = {.lex_state = 271, .external_lex_state = 5},
  [1256] = {.lex_state = 271, .external_lex_state = 5},
  [1257] = {.lex_state = 106, .external_lex_state = 5},
  [1258] = {.lex_state = 106, .external_lex_state = 5},
  [1259] = {.lex_state = 106, .external_lex_state = 5},
  [1260] = {.lex_state = 106, .external_lex_state = 5},
  [1261] = {.lex_state = 106, .external_lex_state = 5},
  [1262] = {.lex_state = 106, .external_lex_state = 5},
  [1263] = {.lex_state = 106, .external_lex_state = 5},
  [1264] = {.lex_state = 271, .external_lex_state = 5},
  [1265] = {.lex_state = 106, .external_lex_state = 5},
  [1266] = {.lex_state = 106, .external_lex_state = 5},
  [1267] = {.lex_state = 271, .external_lex_state = 5},
  [1268] = {.lex_state = 106, .external_lex_state = 5},
  [1269] = {.lex_state = 106, .external_lex_state = 5},
  [1270] = {.lex_state = 106, .external_lex_state = 5},
  [1271] = {.lex_state = 106, .external_lex_state = 5},
  [1272] = {.lex_state = 271, .external_lex_state = 5},
  [1273] = {.lex_state = 271, .external_lex_state = 5},
  [1274] = {.lex_state = 106, .external_lex_state = 5},
  [1275] = {.lex_state = 106, .external_lex_state = 5},
  [1276] = {.lex_state = 271, .external_lex_state = 5},
  [1277] = {.lex_state = 106, .external_lex_state = 5},
  [1278] = {.lex_state = 106, .external_lex_state = 5},
  [1279] = {.lex_state = 106, .external_lex_state = 5},
  [1280] = {.lex_state = 106, .external_lex_state = 5},
  [1281] = {.lex_state = 106, .external_lex_state = 5},
  [1282] = {.lex_state = 106, .external_lex_state = 5},
  [1283] = {.lex_state = 106, .external_lex_state = 5},
  [1284] = {.lex_state = 106, .external_lex_state = 5},
  [1285] = {.lex_state = 271, .external_lex_state = 5},
  [1286] = {.lex_state = 106, .external_lex_state = 5},
  [1287] = {.lex_state = 106, .external_lex_state = 5},
  [1288] = {.lex_state = 106, .external_lex_state = 5},
  [1289] = {.lex_state = 271, .external_lex_state = 5},
  [1290] = {.lex_state = 106, .external_lex_state = 5},
  [1291] = {.lex_state = 106, .external_lex_state = 5},
  [1292] = {.lex_state = 106, .external_lex_state = 5},
  [1293] = {.lex_state = 106, .external_lex_state = 5},
  [1294] = {.lex_state = 106, .external_lex_state = 5},
  [1295] = {.lex_state = 271, .external_lex_state = 5},
  [1296] = {.lex_state = 106, .external_lex_state = 5},
  [1297] = {.lex_state = 106, .external_lex_state = 5},
  [1298] = {.lex_state = 106, .external_lex_state = 5},
  [1299] = {.lex_state = 106, .external_lex_state = 5},
  [1300] = {.lex_state = 106, .external_lex_state = 4},
  [1301] = {.lex_state = 106, .external_lex_state = 5},
  [1302] = {.lex_state = 271, .external_lex_state = 5},
  [1303] = {.lex_state = 271, .external_lex_state = 5},
  [1304] = {.lex_state = 106, .external_lex_state = 5},
  [1305] = {.lex_state = 106, .external_lex_state = 5},
  [1306] = {.lex_state = 106, .external_lex_state = 5},
  [1307] = {.lex_state = 106, .external_lex_state = 5},
  [1308] = {.lex_state = 106, .external_lex_state = 5},
  [1309] = {.lex_state = 106, .external_lex_state = 5},
  [1310] = {.lex_state = 106, .external_lex_state = 5},
  [1311] = {.lex_state = 106, .external_lex_state = 5},
  [1312] = {.lex_state = 111, .external_lex_state = 5},
  [1313] = {.lex_state = 111, .external_lex_state = 5},
  [1314] = {.lex_state = 111, .external_lex_state = 5},
  [1315] = {.lex_state = 111, .external_lex_state = 5},
  [1316] = {.lex_state = 111, .external_lex_state = 5},
  [1317] = {.lex_state = 111, .external_lex_state = 5},
  [1318] = {.lex_state = 111, .external_lex_state = 5},
  [1319] = {.lex_state = 111, .external_lex_state = 5},
  [1320] = {.lex_state = 271, .external_lex_state = 5},
  [1321] = {.lex_state = 271, .external_lex_state = 5},
  [1322] = {.lex_state = 111, .external_lex_state = 5},
  [1323] = {.lex_state = 271, .external_lex_state = 5},
  [1324] = {.lex_state = 111, .external_lex_state = 5},
  [1325] = {.lex_state = 111, .external_lex_state = 5},
  [1326] = {.lex_state = 271, .external_lex_state = 5},
  [1327] = {.lex_state = 111, .external_lex_state = 5},
  [1328] = {.lex_state = 271, .external_lex_state = 5},
  [1329] = {.lex_state = 111, .external_lex_state = 5},
  [1330] = {.lex_state = 111, .external_lex_state = 5},
  [1331] = {.lex_state = 111, .external_lex_state = 5},
  [1332] = {.lex_state = 111, .external_lex_state = 5},
  [1333] = {.lex_state = 271, .external_lex_state = 5},
  [1334] = {.lex_state = 111, .external_lex_state = 5},
  [1335] = {.lex_state = 111, .external_lex_state = 5},
  [1336] = {.lex_state = 111, .external_lex_state = 5},
  [1337] = {.lex_state = 111, .external_lex_state = 5},
  [1338] = {.lex_state = 106, .external_lex_state = 5},
  [1339] = {.lex_state = 111, .external_lex_state = 5},
  [1340] = {.lex_state = 271, .external_lex_state = 5},
  [1341] = {.lex_state = 111, .external_lex_state = 5},
  [1342] = {.lex_state = 271, .external_lex_state = 5},
  [1343] = {.lex_state = 111, .external_lex_state = 5},
  [1344] = {.lex_state = 111, .external_lex_state = 5},
  [1345] = {.lex_state = 106, .external_lex_state = 5},
  [1346] = {.lex_state = 106, .external_lex_state = 5},
  [1347] = {.lex_state = 271, .external_lex_state = 5},
  [1348] = {.lex_state = 111, .external_lex_state = 5},
  [1349] = {.lex_state = 106, .external_lex_state = 5},
  [1350] = {.lex_state = 106, .external_lex_state = 5},
  [1351] = {.lex_state = 106, .external_lex_state = 5},
  [1352] = {.lex_state = 111, .external_lex_state = 5},
  [1353] = {.lex_state = 114, .external_lex_state = 5},
  [1354] = {.lex_state = 106, .external_lex_state = 4},
  [1355] = {.lex_state = 114, .external_lex_state = 5},
  [1356] = {.lex_state = 114, .external_lex_state = 5},
  [1357] = {.lex_state = 114, .external_lex_state = 5},
  [1358] = {.lex_state = 114, .external_lex_state = 5},
  [1359] = {.lex_state = 114, .external_lex_state = 5},
  [1360] = {.lex_state = 114, .external_lex_state = 5},
  [1361] = {.lex_state = 114, .external_lex_state = 5},
  [1362] = {.lex_state = 114, .external_lex_state = 5},
  [1363] = {.lex_state = 114, .external_lex_state = 5},
  [1364] = {.lex_state = 114, .external_lex_state = 5},
  [1365] = {.lex_state = 114, .external_lex_state = 5},
  [1366] = {.lex_state = 114, .external_lex_state = 5},
  [1367] = {.lex_state = 114, .external_lex_state = 5},
  [1368] = {.lex_state = 114, .external_lex_state = 5},
  [1369] = {.lex_state = 114, .external_lex_state = 5},
  [1370] = {.lex_state = 114, .external_lex_state = 5},
  [1371] = {.lex_state = 114, .external_lex_state = 5},
  [1372] = {.lex_state = 114, .external_lex_state = 5},
  [1373] = {.lex_state = 114, .external_lex_state = 5},
  [1374] = {.lex_state = 114, .external_lex_state = 5},
  [1375] = {.lex_state = 114, .external_lex_state = 5},
  [1376] = {.lex_state = 114, .external_lex_state = 5},
  [1377] = {.lex_state = 114, .external_lex_state = 5},
  [1378] = {.lex_state = 114, .external_lex_state = 5},
  [1379] = {.lex_state = 114, .external_lex_state = 5},
  [1380] = {.lex_state = 106, .external_lex_state = 5},
  [1381] = {.lex_state = 106, .external_lex_state = 5},
  [1382] = {.lex_state = 114, .external_lex_state = 5},
  [1383] = {.lex_state = 106, .external_lex_state = 5},
  [1384] = {.lex_state = 271, .external_lex_state = 5},
  [1385] = {.lex_state = 271, .external_lex_state = 5},
  [1386] = {.lex_state = 271, .external_lex_state = 5},
  [1387] = {.lex_state = 271, .external_lex_state = 5},
  [1388] = {.lex_state = 271, .external_lex_state = 5},
  [1389] = {.lex_state = 271, .external_lex_state = 5},
  [1390] = {.lex_state = 271, .external_lex_state = 4},
  [1391] = {.lex_state = 271, .external_lex_state = 5},
  [1392] = {.lex_state = 271, .external_lex_state = 5},
  [1393] = {.lex_state = 271, .external_lex_state = 5},
  [1394] = {.lex_state = 271, .external_lex_state = 5},
  [1395] = {.lex_state = 271, .external_lex_state = 5},
  [1396] = {.lex_state = 271, .external_lex_state = 5},
  [1397] = {.lex_state = 271, .external_lex_state = 5},
  [1398] = {.lex_state = 271, .external_lex_state = 5},
  [1399] = {.lex_state = 271, .external_lex_state = 5},
  [1400] = {.lex_state = 271, .external_lex_state = 5},
  [1401] = {.lex_state = 271, .external_lex_state = 5},
  [1402] = {.lex_state = 106, .external_lex_state = 4},
  [1403] = {.lex_state = 271, .external_lex_state = 5},
  [1404] = {.lex_state = 271, .external_lex_state = 5},
  [1405] = {.lex_state = 271, .external_lex_state = 5},
  [1406] = {.lex_state = 271, .external_lex_state = 5},
  [1407] = {.lex_state = 106, .external_lex_state = 4},
  [1408] = {.lex_state = 106, .external_lex_state = 4},
  [1409] = {.lex_state = 271, .external_lex_state = 5},
  [1410] = {.lex_state = 106, .external_lex_state = 4},
  [1411] = {.lex_state = 271, .external_lex_state = 5},
  [1412] = {.lex_state = 271, .external_lex_state = 5},
  [1413] = {.lex_state = 271, .external_lex_state = 5},
  [1414] = {.lex_state = 271, .external_lex_state = 5},
  [1415] = {.lex_state = 271, .external_lex_state = 5},
  [1416] = {.lex_state = 271, .external_lex_state = 5},
  [1417] = {.lex_state = 114, .external_lex_state = 5},
  [1418] = {.lex_state = 114, .external_lex_state = 5},
  [1419] = {.lex_state = 271, .external_lex_state = 5},
  [1420] = {.lex_state = 271, .external_lex_state = 5},
  [1421] = {.lex_state = 114, .external_lex_state = 5},
  [1422] = {.lex_state = 114, .external_lex_state = 5},
  [1423] = {.lex_state = 271, .external_lex_state = 5},
  [1424] = {.lex_state = 271, .external_lex_state = 5},
  [1425] = {.lex_state = 106, .external_lex_state = 4},
  [1426] = {.lex_state = 271, .external_lex_state = 5},
  [1427] = {.lex_state = 106, .external_lex_state = 4},
  [1428] = {.lex_state = 271, .external_lex_state = 5},
  [1429] = {.lex_state = 271, .external_lex_state = 5},
  [1430] = {.lex_state = 106, .external_lex_state = 4},
  [1431] = {.lex_state = 271, .external_lex_state = 5},
  [1432] = {.lex_state = 271, .external_lex_state = 5},
  [1433] = {.lex_state = 106, .external_lex_state = 4},
  [1434] = {.lex_state = 114, .external_lex_state = 5},
  [1435] = {.lex_state = 106, .external_lex_state = 4},
  [1436] = {.lex_state = 114, .external_lex_state = 5},
  [1437] = {.lex_state = 271, .external_lex_state = 5},
  [1438] = {.lex_state = 271, .external_lex_state = 5},
  [1439] = {.lex_state = 271, .external_lex_state = 5},
  [1440] = {.lex_state = 271, .external_lex_state = 5},
  [1441] = {.lex_state = 271, .external_lex_state = 5},
  [1442] = {.lex_state = 271, .external_lex_state = 5},
  [1443] = {.lex_state = 271, .external_lex_state = 5},
  [1444] = {.lex_state = 271, .external_lex_state = 5},
  [1445] = {.lex_state = 106, .external_lex_state = 4},
  [1446] = {.lex_state = 271, .external_lex_state = 5},
  [1447] = {.lex_state = 114, .external_lex_state = 5},
  [1448] = {.lex_state = 106, .external_lex_state = 4},
  [1449] = {.lex_state = 271, .external_lex_state = 5},
  [1450] = {.lex_state = 114, .external_lex_state = 5},
  [1451] = {.lex_state = 106, .external_lex_state = 4},
  [1452] = {.lex_state = 114, .external_lex_state = 5},
  [1453] = {.lex_state = 271, .external_lex_state = 5},
  [1454] = {.lex_state = 106, .external_lex_state = 4},
  [1455] = {.lex_state = 271, .external_lex_state = 5},
  [1456] = {.lex_state = 114, .external_lex_state = 5},
  [1457] = {.lex_state = 271, .external_lex_state = 5},
  [1458] = {.lex_state = 271, .external_lex_state = 5},
  [1459] = {.lex_state = 106, .external_lex_state = 4},
  [1460] = {.lex_state = 106, .external_lex_state = 4},
  [1461] = {.lex_state = 106, .external_lex_state = 4},
  [1462] = {.lex_state = 271, .external_lex_state = 5},
  [1463] = {.lex_state = 271, .external_lex_state = 5},
  [1464] = {.lex_state = 271, .external_lex_state = 5},
  [1465] = {.lex_state = 271, .external_lex_state = 5},
  [1466] = {.lex_state = 271, .external_lex_state = 5},
  [1467] = {.lex_state = 271, .external_lex_state = 5},
  [1468] = {.lex_state = 114, .external_lex_state = 5},
  [1469] = {.lex_state = 114, .external_lex_state = 5},
  [1470] = {.lex_state = 114, .external_lex_state = 5},
  [1471] = {.lex_state = 271, .external_lex_state = 5},
  [1472] = {.lex_state = 114, .external_lex_state = 5},
  [1473] = {.lex_state = 114, .external_lex_state = 5},
  [1474] = {.lex_state = 114, .external_lex_state = 5},
  [1475] = {.lex_state = 114, .external_lex_state = 5},
  [1476] = {.lex_state = 271, .external_lex_state = 5},
  [1477] = {.lex_state = 271, .external_lex_state = 5},
  [1478] = {.lex_state = 114, .external_lex_state = 5},
  [1479] = {.lex_state = 271, .external_lex_state = 4},
  [1480] = {.lex_state = 106, .external_lex_state = 4},
  [1481] = {.lex_state = 271, .external_lex_state = 4},
  [1482] = {.lex_state = 106, .external_lex_state = 4},
  [1483] = {.lex_state = 271, .external_lex_state = 5},
  [1484] = {.lex_state = 271, .external_lex_state = 5},
  [1485] = {.lex_state = 114, .external_lex_state = 5},
  [1486] = {.lex_state = 114, .external_lex_state = 5},
  [1487] = {.lex_state = 114, .external_lex_state = 5},
  [1488] = {.lex_state = 114, .external_lex_state = 5},
  [1489] = {.lex_state = 114, .external_lex_state = 5},
  [1490] = {.lex_state = 114, .external_lex_state = 5},
  [1491] = {.lex_state = 114, .external_lex_state = 5},
  [1492] = {.lex_state = 114, .external_lex_state = 5},
  [1493] = {.lex_state = 271, .external_lex_state = 5},
  [1494] = {.lex_state = 114, .external_lex_state = 5},
  [1495] = {.lex_state = 114, .external_lex_state = 5},
  [1496] = {.lex_state = 114, .external_lex_state = 5},
  [1497] = {.lex_state = 114, .external_lex_state = 5},
  [1498] = {.lex_state = 114, .external_lex_state = 5},
  [1499] = {.lex_state = 114, .external_lex_state = 5},
  [1500] = {.lex_state = 271, .external_lex_state = 5},
  [1501] = {.lex_state = 114, .external_lex_state = 5},
  [1502] = {.lex_state = 271, .external_lex_state = 5},
  [1503] = {.lex_state = 271, .external_lex_state = 5},
  [1504] = {.lex_state = 114, .external_lex_state = 5},
  [1505] = {.lex_state = 114, .external_lex_state = 5},
  [1506] = {.lex_state = 114, .external_lex_state = 5},
  [1507] = {.lex_state = 114, .external_lex_state = 5},
  [1508] = {.lex_state = 271, .external_lex_state = 5},
  [1509] = {.lex_state = 114, .external_lex_state = 5},
  [1510] = {.lex_state = 114, .external_lex_state = 5},
  [1511] = {.lex_state = 271, .external_lex_state = 5},
  [1512] = {.lex_state = 106, .external_lex_state = 4},
  [1513] = {.lex_state = 114, .external_lex_state = 5},
  [1514] = {.lex_state = 106, .external_lex_state = 5},
  [1515] = {.lex_state = 106, .external_lex_state = 5},
  [1516] = {.lex_state = 106, .external_lex_state = 5},
  [1517] = {.lex_state = 106, .external_lex_state = 5},
  [1518] = {.lex_state = 106, .external_lex_state = 5},
  [1519] = {.lex_state = 106, .external_lex_state = 5},
  [1520] = {.lex_state = 106, .external_lex_state = 5},
  [1521] = {.lex_state = 271, .external_lex_state = 5},
  [1522] = {.lex_state = 106, .external_lex_state = 5},
  [1523] = {.lex_state = 271, .external_lex_state = 5},
  [1524] = {.lex_state = 114, .external_lex_state = 5},
  [1525] = {.lex_state = 271, .external_lex_state = 5},
  [1526] = {.lex_state = 106, .external_lex_state = 5},
  [1527] = {.lex_state = 106, .external_lex_state = 5},
  [1528] = {.lex_state = 106, .external_lex_state = 5},
  [1529] = {.lex_state = 271, .external_lex_state = 5},
  [1530] = {.lex_state = 271, .external_lex_state = 5},
  [1531] = {.lex_state = 271, .external_lex_state = 5},
  [1532] = {.lex_state = 106, .external_lex_state = 5},
  [1533] = {.lex_state = 106, .external_lex_state = 5},
  [1534] = {.lex_state = 106, .external_lex_state = 5},
  [1535] = {.lex_state = 106, .external_lex_state = 5},
  [1536] = {.lex_state = 271, .external_lex_state = 5},
  [1537] = {.lex_state = 271, .external_lex_state = 5},
  [1538] = {.lex_state = 271, .external_lex_state = 5},
  [1539] = {.lex_state = 106, .external_lex_state = 5},
  [1540] = {.lex_state = 271, .external_lex_state = 5},
  [1541] = {.lex_state = 106, .external_lex_state = 5},
  [1542] = {.lex_state = 271, .external_lex_state = 5},
  [1543] = {.lex_state = 114, .external_lex_state = 5},
  [1544] = {.lex_state = 114, .external_lex_state = 5},
  [1545] = {.lex_state = 114, .external_lex_state = 5},
  [1546] = {.lex_state = 114, .external_lex_state = 5},
  [1547] = {.lex_state = 114, .external_lex_state = 5},
  [1548] = {.lex_state = 271, .external_lex_state = 5},
  [1549] = {.lex_state = 114, .external_lex_state = 5},
  [1550] = {.lex_state = 271, .external_lex_state = 5},
  [1551] = {.lex_state = 271, .external_lex_state = 5},
  [1552] = {.lex_state = 106, .external_lex_state = 4},
  [1553] = {.lex_state = 114, .external_lex_state = 5},
  [1554] = {.lex_state = 106, .external_lex_state = 5},
  [1555] = {.lex_state = 106, .external_lex_state = 5},
  [1556] = {.lex_state = 106, .external_lex_state = 5},
  [1557] = {.lex_state = 271, .external_lex_state = 5},
  [1558] = {.lex_state = 271, .external_lex_state = 5},
  [1559] = {.lex_state = 271, .external_lex_state = 5},
  [1560] = {.lex_state = 271, .external_lex_state = 5},
  [1561] = {.lex_state = 114, .external_lex_state = 5},
  [1562] = {.lex_state = 271, .external_lex_state = 5},
  [1563] = {.lex_state = 271, .external_lex_state = 5},
  [1564] = {.lex_state = 271, .external_lex_state = 5},
  [1565] = {.lex_state = 114, .external_lex_state = 5},
  [1566] = {.lex_state = 271, .external_lex_state = 5},
  [1567] = {.lex_state = 271, .external_lex_state = 5},
  [1568] = {.lex_state = 271, .external_lex_state = 5},
  [1569] = {.lex_state = 271, .external_lex_state = 5},
  [1570] = {.lex_state = 114, .external_lex_state = 5},
  [1571] = {.lex_state = 114, .external_lex_state = 5},
  [1572] = {.lex_state = 114, .external_lex_state = 5},
  [1573] = {.lex_state = 271, .external_lex_state = 5},
  [1574] = {.lex_state = 271, .external_lex_state = 5},
  [1575] = {.lex_state = 106, .external_lex_state = 4},
  [1576] = {.lex_state = 106, .external_lex_state = 4},
  [1577] = {.lex_state = 106, .external_lex_state = 4},
  [1578] = {.lex_state = 106, .external_lex_state = 4},
  [1579] = {.lex_state = 106, .external_lex_state = 4},
  [1580] = {.lex_state = 106, .external_lex_state = 4},
  [1581] = {.lex_state = 271, .external_lex_state = 4},
  [1582] = {.lex_state = 271, .external_lex_state = 4},
  [1583] = {.lex_state = 271, .external_lex_state = 4},
  [1584] = {.lex_state = 271, .external_lex_state = 4},
  [1585] = {.lex_state = 271, .external_lex_state = 4},
  [1586] = {.lex_state = 106, .external_lex_state = 4},
  [1587] = {.lex_state = 106, .external_lex_state = 4},
  [1588] = {.lex_state = 106, .external_lex_state = 4},
  [1589] = {.lex_state = 106, .external_lex_state = 4},
  [1590] = {.lex_state = 271, .external_lex_state = 4},
  [1591] = {.lex_state = 106, .external_lex_state = 4},
  [1592] = {.lex_state = 106, .external_lex_state = 4},
  [1593] = {.lex_state = 271, .external_lex_state = 4},
  [1594] = {.lex_state = 271, .external_lex_state = 4},
  [1595] = {.lex_state = 271, .external_lex_state = 5},
  [1596] = {.lex_state = 271, .external_lex_state = 5},
  [1597] = {.lex_state = 271, .external_lex_state = 5},
  [1598] = {.lex_state = 271, .external_lex_state = 5},
  [1599] = {.lex_state = 271, .external_lex_state = 5},
  [1600] = {.lex_state = 271, .external_lex_state = 5},
  [1601] = {.lex_state = 271, .external_lex_state = 5},
  [1602] = {.lex_state = 271, .external_lex_state = 5},
  [1603] = {.lex_state = 271, .external_lex_state = 5},
  [1604] = {.lex_state = 271, .external_lex_state = 5},
  [1605] = {.lex_state = 271, .external_lex_state = 5},
  [1606] = {.lex_state = 271, .external_lex_state = 5},
  [1607] = {.lex_state = 271, .external_lex_state = 5},
  [1608] = {.lex_state = 271, .external_lex_state = 5},
  [1609] = {.lex_state = 271, .external_lex_state = 5},
  [1610] = {.lex_state = 271, .external_lex_state = 5},
  [1611] = {.lex_state = 271, .external_lex_state = 5},
  [1612] = {.lex_state = 106, .external_lex_state = 4},
  [1613] = {.lex_state = 271, .external_lex_state = 4},
  [1614] = {.lex_state = 271, .external_lex_state = 4},
  [1615] = {.lex_state = 271, .external_lex_state = 4},
  [1616] = {.lex_state = 271, .external_lex_state = 4},
  [1617] = {.lex_state = 106, .external_lex_state = 4},
  [1618] = {.lex_state = 271, .external_lex_state = 4},
  [1619] = {.lex_state = 271, .external_lex_state = 4},
  [1620] = {.lex_state = 271, .external_lex_state = 4},
  [1621] = {.lex_state = 271, .external_lex_state = 4},
  [1622] = {.lex_state = 271, .external_lex_state = 4},
  [1623] = {.lex_state = 106, .external_lex_state = 4},
  [1624] = {.lex_state = 117, .external_lex_state = 4},
  [1625] = {.lex_state = 117, .external_lex_state = 4},
  [1626] = {.lex_state = 117, .external_lex_state = 4},
  [1627] = {.lex_state = 117, .external_lex_state = 4},
  [1628] = {.lex_state = 117, .external_lex_state = 4},
  [1629] = {.lex_state = 117, .external_lex_state = 4},
  [1630] = {.lex_state = 271, .external_lex_state = 4},
  [1631] = {.lex_state = 271, .external_lex_state = 4},
  [1632] = {.lex_state = 271, .external_lex_state = 4},
  [1633] = {.lex_state = 117, .external_lex_state = 4},
  [1634] = {.lex_state = 117, .external_lex_state = 4},
  [1635] = {.lex_state = 117, .external_lex_state = 4},
  [1636] = {.lex_state = 117, .external_lex_state = 4},
  [1637] = {.lex_state = 117, .external_lex_state = 4},
  [1638] = {.lex_state = 117, .external_lex_state = 4},
  [1639] = {.lex_state = 271, .external_lex_state = 4},
  [1640] = {.lex_state = 106, .external_lex_state = 4},
  [1641] = {.lex_state = 106, .external_lex_state = 4},
  [1642] = {.lex_state = 106, .external_lex_state = 4},
  [1643] = {.lex_state = 271, .external_lex_state = 5},
  [1644] = {.lex_state = 106, .external_lex_state = 4},
  [1645] = {.lex_state = 271, .external_lex_state = 5},
  [1646] = {.lex_state = 106, .external_lex_state = 4},
  [1647] = {.lex_state = 106, .external_lex_state = 4},
  [1648] = {.lex_state = 106, .external_lex_state = 4},
  [1649] = {.lex_state = 271, .external_lex_state = 5},
  [1650] = {.lex_state = 106, .external_lex_state = 4},
  [1651] = {.lex_state = 106, .external_lex_state = 4},
  [1652] = {.lex_state = 106, .external_lex_state = 4},
  [1653] = {.lex_state = 106, .external_lex_state = 4},
  [1654] = {.lex_state = 106, .external_lex_state = 4},
  [1655] = {.lex_state = 117, .external_lex_state = 4},
  [1656] = {.lex_state = 117, .external_lex_state = 4},
  [1657] = {.lex_state = 117, .external_lex_state = 4},
  [1658] = {.lex_state = 117, .external_lex_state = 4},
  [1659] = {.lex_state = 117, .external_lex_state = 4},
  [1660] = {.lex_state = 117, .external_lex_state = 4},
  [1661] = {.lex_state = 106, .external_lex_state = 4},
  [1662] = {.lex_state = 117, .external_lex_state = 4},
  [1663] = {.lex_state = 117, .external_lex_state = 4},
  [1664] = {.lex_state = 117, .external_lex_state = 4},
  [1665] = {.lex_state = 117, .external_lex_state = 4},
  [1666] = {.lex_state = 117, .external_lex_state = 4},
  [1667] = {.lex_state = 117, .external_lex_state = 4},
  [1668] = {.lex_state = 117, .external_lex_state = 4},
  [1669] = {.lex_state = 117, .external_lex_state = 4},
  [1670] = {.lex_state = 117, .external_lex_state = 4},
  [1671] = {.lex_state = 117, .external_lex_state = 4},
  [1672] = {.lex_state = 117, .external_lex_state = 4},
  [1673] = {.lex_state = 117, .external_lex_state = 4},
  [1674] = {.lex_state = 117, .external_lex_state = 4},
  [1675] = {.lex_state = 117, .external_lex_state = 4},
  [1676] = {.lex_state = 106, .external_lex_state = 4},
  [1677] = {.lex_state = 106, .external_lex_state = 4},
  [1678] = {.lex_state = 106, .external_lex_state = 4},
  [1679] = {.lex_state = 106, .external_lex_state = 4},
  [1680] = {.lex_state = 106, .external_lex_state = 4},
  [1681] = {.lex_state = 106, .external_lex_state = 4},
  [1682] = {.lex_state = 106, .external_lex_state = 4},
  [1683] = {.lex_state = 271, .external_lex_state = 4},
  [1684] = {.lex_state = 106, .external_lex_state = 4},
  [1685] = {.lex_state = 106, .external_lex_state = 4},
  [1686] = {.lex_state = 106, .external_lex_state = 4},
  [1687] = {.lex_state = 271, .external_lex_state = 4},
  [1688] = {.lex_state = 271, .external_lex_state = 5},
  [1689] = {.lex_state = 271, .external_lex_state = 5},
  [1690] = {.lex_state = 271, .external_lex_state = 5},
  [1691] = {.lex_state = 117, .external_lex_state = 4},
  [1692] = {.lex_state = 117, .external_lex_state = 4},
  [1693] = {.lex_state = 117, .external_lex_state = 4},
  [1694] = {.lex_state = 106, .external_lex_state = 4},
  [1695] = {.lex_state = 117, .external_lex_state = 4},
  [1696] = {.lex_state = 106, .external_lex_state = 4},
  [1697] = {.lex_state = 106, .external_lex_state = 4},
  [1698] = {.lex_state = 106, .external_lex_state = 4},
  [1699] = {.lex_state = 106, .external_lex_state = 4},
  [1700] = {.lex_state = 106, .external_lex_state = 4},
  [1701] = {.lex_state = 117, .external_lex_state = 4},
  [1702] = {.lex_state = 106, .external_lex_state = 4},
  [1703] = {.lex_state = 106, .external_lex_state = 4},
  [1704] = {.lex_state = 117, .external_lex_state = 4},
  [1705] = {.lex_state = 106, .external_lex_state = 4},
  [1706] = {.lex_state = 117, .external_lex_state = 4},
  [1707] = {.lex_state = 117, .external_lex_state = 4},
  [1708] = {.lex_state = 117, .external_lex_state = 4},
  [1709] = {.lex_state = 117, .external_lex_state = 4},
  [1710] = {.lex_state = 106, .external_lex_state = 4},
  [1711] = {.lex_state = 106, .external_lex_state = 4},
  [1712] = {.lex_state = 117, .external_lex_state = 4},
  [1713] = {.lex_state = 117, .external_lex_state = 4},
  [1714] = {.lex_state = 106, .external_lex_state = 4},
  [1715] = {.lex_state = 106, .external_lex_state = 4},
  [1716] = {.lex_state = 106, .external_lex_state = 4},
  [1717] = {.lex_state = 117, .external_lex_state = 4},
  [1718] = {.lex_state = 106, .external_lex_state = 4},
  [1719] = {.lex_state = 106, .external_lex_state = 4},
  [1720] = {.lex_state = 106, .external_lex_state = 4},
  [1721] = {.lex_state = 117, .external_lex_state = 4},
  [1722] = {.lex_state = 106, .external_lex_state = 4},
  [1723] = {.lex_state = 106, .external_lex_state = 4},
  [1724] = {.lex_state = 106, .external_lex_state = 4},
  [1725] = {.lex_state = 117, .external_lex_state = 4},
  [1726] = {.lex_state = 106, .external_lex_state = 4},
  [1727] = {.lex_state = 106, .external_lex_state = 4},
  [1728] = {.lex_state = 106, .external_lex_state = 4},
  [1729] = {.lex_state = 106, .external_lex_state = 4},
  [1730] = {.lex_state = 106, .external_lex_state = 4},
  [1731] = {.lex_state = 106, .external_lex_state = 4},
  [1732] = {.lex_state = 117, .external_lex_state = 4},
  [1733] = {.lex_state = 106, .external_lex_state = 4},
  [1734] = {.lex_state = 106, .external_lex_state = 4},
  [1735] = {.lex_state = 117, .external_lex_state = 4},
  [1736] = {.lex_state = 106, .external_lex_state = 4},
  [1737] = {.lex_state = 106, .external_lex_state = 4},
  [1738] = {.lex_state = 106, .external_lex_state = 4},
  [1739] = {.lex_state = 117, .external_lex_state = 4},
  [1740] = {.lex_state = 106, .external_lex_state = 4},
  [1741] = {.lex_state = 106, .external_lex_state = 4},
  [1742] = {.lex_state = 106, .external_lex_state = 4},
  [1743] = {.lex_state = 106, .external_lex_state = 4},
  [1744] = {.lex_state = 117, .external_lex_state = 4},
  [1745] = {.lex_state = 117, .external_lex_state = 4},
  [1746] = {.lex_state = 117, .external_lex_state = 4},
  [1747] = {.lex_state = 117, .external_lex_state = 4},
  [1748] = {.lex_state = 117, .external_lex_state = 4},
  [1749] = {.lex_state = 117, .external_lex_state = 4},
  [1750] = {.lex_state = 106, .external_lex_state = 4},
  [1751] = {.lex_state = 106, .external_lex_state = 4},
  [1752] = {.lex_state = 106, .external_lex_state = 4},
  [1753] = {.lex_state = 117, .external_lex_state = 4},
  [1754] = {.lex_state = 117, .external_lex_state = 4},
  [1755] = {.lex_state = 106, .external_lex_state = 4},
  [1756] = {.lex_state = 106, .external_lex_state = 4},
  [1757] = {.lex_state = 117, .external_lex_state = 4},
  [1758] = {.lex_state = 106, .external_lex_state = 4},
  [1759] = {.lex_state = 106, .external_lex_state = 4},
  [1760] = {.lex_state = 117, .external_lex_state = 4},
  [1761] = {.lex_state = 106, .external_lex_state = 4},
  [1762] = {.lex_state = 117, .external_lex_state = 4},
  [1763] = {.lex_state = 117, .external_lex_state = 4},
  [1764] = {.lex_state = 117, .external_lex_state = 4},
  [1765] = {.lex_state = 117, .external_lex_state = 4},
  [1766] = {.lex_state = 106, .external_lex_state = 4},
  [1767] = {.lex_state = 117, .external_lex_state = 4},
  [1768] = {.lex_state = 117, .external_lex_state = 4},
  [1769] = {.lex_state = 117, .external_lex_state = 4},
  [1770] = {.lex_state = 117, .external_lex_state = 4},
  [1771] = {.lex_state = 117, .external_lex_state = 4},
  [1772] = {.lex_state = 117, .external_lex_state = 4},
  [1773] = {.lex_state = 117, .external_lex_state = 4},
  [1774] = {.lex_state = 117, .external_lex_state = 4},
  [1775] = {.lex_state = 117, .external_lex_state = 4},
  [1776] = {.lex_state = 117, .external_lex_state = 4},
  [1777] = {.lex_state = 117, .external_lex_state = 4},
  [1778] = {.lex_state = 106, .external_lex_state = 4},
  [1779] = {.lex_state = 106, .external_lex_state = 4},
  [1780] = {.lex_state = 117, .external_lex_state = 4},
  [1781] = {.lex_state = 117, .external_lex_state = 4},
  [1782] = {.lex_state = 117, .external_lex_state = 4},
  [1783] = {.lex_state = 117, .external_lex_state = 4},
  [1784] = {.lex_state = 106, .external_lex_state = 4},
  [1785] = {.lex_state = 106, .external_lex_state = 4},
  [1786] = {.lex_state = 106, .external_lex_state = 4},
  [1787] = {.lex_state = 271, .external_lex_state = 5},
  [1788] = {.lex_state = 271, .external_lex_state = 5},
  [1789] = {.lex_state = 271, .external_lex_state = 5},
  [1790] = {.lex_state = 271, .external_lex_state = 5},
  [1791] = {.lex_state = 106, .external_lex_state = 4},
  [1792] = {.lex_state = 106, .external_lex_state = 4},
  [1793] = {.lex_state = 106, .external_lex_state = 4},
  [1794] = {.lex_state = 106, .external_lex_state = 4},
  [1795] = {.lex_state = 271, .external_lex_state = 5},
  [1796] = {.lex_state = 271, .external_lex_state = 5},
  [1797] = {.lex_state = 271, .external_lex_state = 5},
  [1798] = {.lex_state = 106, .external_lex_state = 4},
  [1799] = {.lex_state = 106, .external_lex_state = 4},
  [1800] = {.lex_state = 117, .external_lex_state = 4},
  [1801] = {.lex_state = 120, .external_lex_state = 4},
  [1802] = {.lex_state = 271, .external_lex_state = 4},
  [1803] = {.lex_state = 271, .external_lex_state = 4},
  [1804] = {.lex_state = 271, .external_lex_state = 4},
  [1805] = {.lex_state = 271, .external_lex_state = 4},
  [1806] = {.lex_state = 271, .external_lex_state = 4},
  [1807] = {.lex_state = 120, .external_lex_state = 4},
  [1808] = {.lex_state = 120, .external_lex_state = 4},
  [1809] = {.lex_state = 120, .external_lex_state = 4},
  [1810] = {.lex_state = 120, .external_lex_state = 4},
  [1811] = {.lex_state = 120, .external_lex_state = 4},
  [1812] = {.lex_state = 120, .external_lex_state = 4},
  [1813] = {.lex_state = 120, .external_lex_state = 4},
  [1814] = {.lex_state = 120, .external_lex_state = 4},
  [1815] = {.lex_state = 271, .external_lex_state = 4},
  [1816] = {.lex_state = 271, .external_lex_state = 4},
  [1817] = {.lex_state = 271, .external_lex_state = 4},
  [1818] = {.lex_state = 271, .external_lex_state = 4},
  [1819] = {.lex_state = 271, .external_lex_state = 4},
  [1820] = {.lex_state = 271, .external_lex_state = 4},
  [1821] = {.lex_state = 271, .external_lex_state = 4},
  [1822] = {.lex_state = 271, .external_lex_state = 4},
  [1823] = {.lex_state = 271, .external_lex_state = 4},
  [1824] = {.lex_state = 271, .external_lex_state = 4},
  [1825] = {.lex_state = 271, .external_lex_state = 4},
  [1826] = {.lex_state = 271, .external_lex_state = 4},
  [1827] = {.lex_state = 271, .external_lex_state = 4},
  [1828] = {.lex_state = 271, .external_lex_state = 4},
  [1829] = {.lex_state = 271, .external_lex_state = 4},
  [1830] = {.lex_state = 271, .external_lex_state = 4},
  [1831] = {.lex_state = 271, .external_lex_state = 4},
  [1832] = {.lex_state = 120, .external_lex_state = 4},
  [1833] = {.lex_state = 120, .external_lex_state = 4},
  [1834] = {.lex_state = 271, .external_lex_state = 4},
  [1835] = {.lex_state = 271, .external_lex_state = 4},
  [1836] = {.lex_state = 271, .external_lex_state = 4},
  [1837] = {.lex_state = 271, .external_lex_state = 4},
  [1838] = {.lex_state = 271, .external_lex_state = 4},
  [1839] = {.lex_state = 271, .external_lex_state = 4},
  [1840] = {.lex_state = 120, .external_lex_state = 4},
  [1841] = {.lex_state = 120, .external_lex_state = 4},
  [1842] = {.lex_state = 120, .external_lex_state = 4},
  [1843] = {.lex_state = 271, .external_lex_state = 4},
  [1844] = {.lex_state = 271, .external_lex_state = 4},
  [1845] = {.lex_state = 271, .external_lex_state = 4},
  [1846] = {.lex_state = 120, .external_lex_state = 4},
  [1847] = {.lex_state = 271, .external_lex_state = 4},
  [1848] = {.lex_state = 271, .external_lex_state = 4},
  [1849] = {.lex_state = 271, .external_lex_state = 4},
  [1850] = {.lex_state = 271, .external_lex_state = 4},
  [1851] = {.lex_state = 271, .external_lex_state = 4},
  [1852] = {.lex_state = 120, .external_lex_state = 4},
  [1853] = {.lex_state = 271, .external_lex_state = 4},
  [1854] = {.lex_state = 120, .external_lex_state = 4},
  [1855] = {.lex_state = 120, .external_lex_state = 4},
  [1856] = {.lex_state = 120, .external_lex_state = 4},
  [1857] = {.lex_state = 120, .external_lex_state = 4},
  [1858] = {.lex_state = 273, .external_lex_state = 5},
  [1859] = {.lex_state = 273, .external_lex_state = 5},
  [1860] = {.lex_state = 271, .external_lex_state = 4},
  [1861] = {.lex_state = 273, .external_lex_state = 5},
  [1862] = {.lex_state = 273, .external_lex_state = 5},
  [1863] = {.lex_state = 271, .external_lex_state = 4},
  [1864] = {.lex_state = 273, .external_lex_state = 5},
  [1865] = {.lex_state = 273, .external_lex_state = 5},
  [1866] = {.lex_state = 120, .external_lex_state = 4},
  [1867] = {.lex_state = 120, .external_lex_state = 4},
  [1868] = {.lex_state = 120, .external_lex_state = 4},
  [1869] = {.lex_state = 120, .external_lex_state = 4},
  [1870] = {.lex_state = 120, .external_lex_state = 4},
  [1871] = {.lex_state = 273, .external_lex_state = 5},
  [1872] = {.lex_state = 273, .external_lex_state = 5},
  [1873] = {.lex_state = 273, .external_lex_state = 5},
  [1874] = {.lex_state = 273, .external_lex_state = 5},
  [1875] = {.lex_state = 273, .external_lex_state = 5},
  [1876] = {.lex_state = 273, .external_lex_state = 5},
  [1877] = {.lex_state = 120, .external_lex_state = 4},
  [1878] = {.lex_state = 120, .external_lex_state = 4},
  [1879] = {.lex_state = 120, .external_lex_state = 4},
  [1880] = {.lex_state = 120, .external_lex_state = 4},
  [1881] = {.lex_state = 120, .external_lex_state = 4},
  [1882] = {.lex_state = 120, .external_lex_state = 4},
  [1883] = {.lex_state = 271, .external_lex_state = 5},
  [1884] = {.lex_state = 271, .external_lex_state = 5},
  [1885] = {.lex_state = 120, .external_lex_state = 4},
  [1886] = {.lex_state = 120, .external_lex_state = 4},
  [1887] = {.lex_state = 120, .external_lex_state = 4},
  [1888] = {.lex_state = 120, .external_lex_state = 4},
  [1889] = {.lex_state = 120, .external_lex_state = 4},
  [1890] = {.lex_state = 120, .external_lex_state = 4},
  [1891] = {.lex_state = 120, .external_lex_state = 4},
  [1892] = {.lex_state = 120, .external_lex_state = 4},
  [1893] = {.lex_state = 120, .external_lex_state = 4},
  [1894] = {.lex_state = 271, .external_lex_state = 4},
  [1895] = {.lex_state = 271, .external_lex_state = 4},
  [1896] = {.lex_state = 120, .external_lex_state = 4},
  [1897] = {.lex_state = 120, .external_lex_state = 4},
  [1898] = {.lex_state = 120, .external_lex_state = 4},
  [1899] = {.lex_state = 120, .external_lex_state = 4},
  [1900] = {.lex_state = 120, .external_lex_state = 4},
  [1901] = {.lex_state = 120, .external_lex_state = 4},
  [1902] = {.lex_state = 271, .external_lex_state = 4},
  [1903] = {.lex_state = 271, .external_lex_state = 4},
  [1904] = {.lex_state = 271, .external_lex_state = 4},
  [1905] = {.lex_state = 106, .external_lex_state = 4},
  [1906] = {.lex_state = 271, .external_lex_state = 4},
  [1907] = {.lex_state = 271, .external_lex_state = 4},
  [1908] = {.lex_state = 271, .external_lex_state = 4},
  [1909] = {.lex_state = 120, .external_lex_state = 4},
  [1910] = {.lex_state = 120, .external_lex_state = 4},
  [1911] = {.lex_state = 106, .external_lex_state = 4},
  [1912] = {.lex_state = 271, .external_lex_state = 4},
  [1913] = {.lex_state = 106, .external_lex_state = 4},
  [1914] = {.lex_state = 120, .external_lex_state = 4},
  [1915] = {.lex_state = 120, .external_lex_state = 4},
  [1916] = {.lex_state = 120, .external_lex_state = 4},
  [1917] = {.lex_state = 106, .external_lex_state = 4},
  [1918] = {.lex_state = 120, .external_lex_state = 4},
  [1919] = {.lex_state = 120, .external_lex_state = 4},
  [1920] = {.lex_state = 120, .external_lex_state = 4},
  [1921] = {.lex_state = 120, .external_lex_state = 4},
  [1922] = {.lex_state = 120, .external_lex_state = 4},
  [1923] = {.lex_state = 120, .external_lex_state = 4},
  [1924] = {.lex_state = 120, .external_lex_state = 4},
  [1925] = {.lex_state = 120, .external_lex_state = 4},
  [1926] = {.lex_state = 271, .external_lex_state = 4},
  [1927] = {.lex_state = 120, .external_lex_state = 4},
  [1928] = {.lex_state = 271, .external_lex_state = 5},
  [1929] = {.lex_state = 271, .external_lex_state = 5},
  [1930] = {.lex_state = 120, .external_lex_state = 4},
  [1931] = {.lex_state = 271, .external_lex_state = 4},
  [1932] = {.lex_state = 120, .external_lex_state = 4},
  [1933] = {.lex_state = 120, .external_lex_state = 4},
  [1934] = {.lex_state = 120, .external_lex_state = 4},
  [1935] = {.lex_state = 120, .external_lex_state = 4},
  [1936] = {.lex_state = 273, .external_lex_state = 5},
  [1937] = {.lex_state = 273, .external_lex_state = 5},
  [1938] = {.lex_state = 273, .external_lex_state = 5},
  [1939] = {.lex_state = 273, .external_lex_state = 5},
  [1940] = {.lex_state = 273, .external_lex_state = 5},
  [1941] = {.lex_state = 273, .external_lex_state = 5},
  [1942] = {.lex_state = 273, .external_lex_state = 5},
  [1943] = {.lex_state = 273, .external_lex_state = 5},
  [1944] = {.lex_state = 273, .external_lex_state = 5},
  [1945] = {.lex_state = 273, .external_lex_state = 5},
  [1946] = {.lex_state = 106, .external_lex_state = 4},
  [1947] = {.lex_state = 106, .external_lex_state = 4},
  [1948] = {.lex_state = 106, .external_lex_state = 4},
  [1949] = {.lex_state = 106, .external_lex_state = 4},
  [1950] = {.lex_state = 106, .external_lex_state = 4},
  [1951] = {.lex_state = 106, .external_lex_state = 4},
  [1952] = {.lex_state = 106, .external_lex_state = 4},
  [1953] = {.lex_state = 106, .external_lex_state = 4},
  [1954] = {.lex_state = 106, .external_lex_state = 4},
  [1955] = {.lex_state = 106, .external_lex_state = 4},
  [1956] = {.lex_state = 106, .external_lex_state = 4},
  [1957] = {.lex_state = 106, .external_lex_state = 4},
  [1958] = {.lex_state = 106, .external_lex_state = 4},
  [1959] = {.lex_state = 106, .external_lex_state = 4},
  [1960] = {.lex_state = 106, .external_lex_state = 4},
  [1961] = {.lex_state = 106, .external_lex_state = 4},
  [1962] = {.lex_state = 106, .external_lex_state = 4},
  [1963] = {.lex_state = 106, .external_lex_state = 4},
  [1964] = {.lex_state = 106, .external_lex_state = 4},
  [1965] = {.lex_state = 106, .external_lex_state = 4},
  [1966] = {.lex_state = 271, .external_lex_state = 4},
  [1967] = {.lex_state = 271, .external_lex_state = 4},
  [1968] = {.lex_state = 271, .external_lex_state = 5},
  [1969] = {.lex_state = 271, .external_lex_state = 4},
  [1970] = {.lex_state = 271, .external_lex_state = 5},
  [1971] = {.lex_state = 106, .external_lex_state = 4},
  [1972] = {.lex_state = 271, .external_lex_state = 5},
  [1973] = {.lex_state = 273, .external_lex_state = 5},
  [1974] = {.lex_state = 273, .external_lex_state = 5},
  [1975] = {.lex_state = 273, .external_lex_state = 5},
  [1976] = {.lex_state = 120, .external_lex_state = 4},
  [1977] = {.lex_state = 106, .external_lex_state = 4},
  [1978] = {.lex_state = 273, .external_lex_state = 5},
  [1979] = {.lex_state = 273, .external_lex_state = 5},
  [1980] = {.lex_state = 106, .external_lex_state = 4},
  [1981] = {.lex_state = 273, .external_lex_state = 5},
  [1982] = {.lex_state = 273, .external_lex_state = 5},
  [1983] = {.lex_state = 273, .external_lex_state = 5},
  [1984] = {.lex_state = 273, .external_lex_state = 5},
  [1985] = {.lex_state = 273, .external_lex_state = 5},
  [1986] = {.lex_state = 273, .external_lex_state = 5},
  [1987] = {.lex_state = 273, .external_lex_state = 5},
  [1988] = {.lex_state = 273, .external_lex_state = 5},
  [1989] = {.lex_state = 273, .external_lex_state = 5},
  [1990] = {.lex_state = 273, .external_lex_state = 5},
  [1991] = {.lex_state = 271, .external_lex_state = 5},
  [1992] = {.lex_state = 273, .external_lex_state = 5},
  [1993] = {.lex_state = 273, .external_lex_state = 5},
  [1994] = {.lex_state = 106, .external_lex_state = 4},
  [1995] = {.lex_state = 273, .external_lex_state = 5},
  [1996] = {.lex_state = 273, .external_lex_state = 5},
  [1997] = {.lex_state = 273, .external_lex_state = 5},
  [1998] = {.lex_state = 273, .external_lex_state = 5},
  [1999] = {.lex_state = 273, .external_lex_state = 5},
  [2000] = {.lex_state = 273, .external_lex_state = 5},
  [2001] = {.lex_state = 273, .external_lex_state = 5},
  [2002] = {.lex_state = 273, .external_lex_state = 5},
  [2003] = {.lex_state = 106, .external_lex_state = 4},
  [2004] = {.lex_state = 273, .external_lex_state = 5},
  [2005] = {.lex_state = 273, .external_lex_state = 5},
  [2006] = {.lex_state = 106, .external_lex_state = 4},
  [2007] = {.lex_state = 273, .external_lex_state = 5},
  [2008] = {.lex_state = 273, .external_lex_state = 5},
  [2009] = {.lex_state = 273, .external_lex_state = 5},
  [2010] = {.lex_state = 273, .external_lex_state = 5},
  [2011] = {.lex_state = 271, .external_lex_state = 5},
  [2012] = {.lex_state = 271, .external_lex_state = 5},
  [2013] = {.lex_state = 271, .external_lex_state = 5},
  [2014] = {.lex_state = 271, .external_lex_state = 4},
  [2015] = {.lex_state = 271, .external_lex_state = 4},
  [2016] = {.lex_state = 120, .external_lex_state = 4},
  [2017] = {.lex_state = 106, .external_lex_state = 4},
  [2018] = {.lex_state = 106, .external_lex_state = 4},
  [2019] = {.lex_state = 271, .external_lex_state = 4},
  [2020] = {.lex_state = 271, .external_lex_state = 4},
  [2021] = {.lex_state = 271, .external_lex_state = 4},
  [2022] = {.lex_state = 271, .external_lex_state = 4},
  [2023] = {.lex_state = 106, .external_lex_state = 4},
  [2024] = {.lex_state = 271, .external_lex_state = 4},
  [2025] = {.lex_state = 120, .external_lex_state = 4},
  [2026] = {.lex_state = 120, .external_lex_state = 4},
  [2027] = {.lex_state = 271, .external_lex_state = 4},
  [2028] = {.lex_state = 271, .external_lex_state = 4},
  [2029] = {.lex_state = 271, .external_lex_state = 4},
  [2030] = {.lex_state = 271, .external_lex_state = 4},
  [2031] = {.lex_state = 271, .external_lex_state = 4},
  [2032] = {.lex_state = 106, .external_lex_state = 4},
  [2033] = {.lex_state = 271, .external_lex_state = 4},
  [2034] = {.lex_state = 271, .external_lex_state = 4},
  [2035] = {.lex_state = 106, .external_lex_state = 4},
  [2036] = {.lex_state = 271, .external_lex_state = 4},
  [2037] = {.lex_state = 271, .external_lex_state = 4},
  [2038] = {.lex_state = 271, .external_lex_state = 4},
  [2039] = {.lex_state = 120, .external_lex_state = 4},
  [2040] = {.lex_state = 120, .external_lex_state = 4},
  [2041] = {.lex_state = 271, .external_lex_state = 4},
  [2042] = {.lex_state = 271, .external_lex_state = 4},
  [2043] = {.lex_state = 271, .external_lex_state = 4},
  [2044] = {.lex_state = 271, .external_lex_state = 4},
  [2045] = {.lex_state = 271, .external_lex_state = 4},
  [2046] = {.lex_state = 271, .external_lex_state = 4},
  [2047] = {.lex_state = 106, .external_lex_state = 4},
  [2048] = {.lex_state = 271, .external_lex_state = 4},
  [2049] = {.lex_state = 106, .external_lex_state = 4},
  [2050] = {.lex_state = 106, .external_lex_state = 4},
  [2051] = {.lex_state = 271, .external_lex_state = 4},
  [2052] = {.lex_state = 271, .external_lex_state = 5},
  [2053] = {.lex_state = 271, .external_lex_state = 4},
  [2054] = {.lex_state = 271, .external_lex_state = 4},
  [2055] = {.lex_state = 271, .external_lex_state = 4},
  [2056] = {.lex_state = 271, .external_lex_state = 4},
  [2057] = {.lex_state = 271, .external_lex_state = 4},
  [2058] = {.lex_state = 271, .external_lex_state = 5},
  [2059] = {.lex_state = 106, .external_lex_state = 4},
  [2060] = {.lex_state = 271, .external_lex_state = 4},
  [2061] = {.lex_state = 271, .external_lex_state = 4},
  [2062] = {.lex_state = 271, .external_lex_state = 4},
  [2063] = {.lex_state = 271, .external_lex_state = 5},
  [2064] = {.lex_state = 106, .external_lex_state = 4},
  [2065] = {.lex_state = 271, .external_lex_state = 5},
  [2066] = {.lex_state = 271, .external_lex_state = 5},
  [2067] = {.lex_state = 106, .external_lex_state = 4},
  [2068] = {.lex_state = 271, .external_lex_state = 5},
  [2069] = {.lex_state = 271, .external_lex_state = 5},
  [2070] = {.lex_state = 271, .external_lex_state = 5},
  [2071] = {.lex_state = 271, .external_lex_state = 5},
  [2072] = {.lex_state = 271, .external_lex_state = 5},
  [2073] = {.lex_state = 271, .external_lex_state = 5},
  [2074] = {.lex_state = 271, .external_lex_state = 5},
  [2075] = {.lex_state = 271, .external_lex_state = 5},
  [2076] = {.lex_state = 106, .external_lex_state = 4},
  [2077] = {.lex_state = 106, .external_lex_state = 4},
  [2078] = {.lex_state = 106, .external_lex_state = 4},
  [2079] = {.lex_state = 106, .external_lex_state = 4},
  [2080] = {.lex_state = 106, .external_lex_state = 4},
  [2081] = {.lex_state = 271, .external_lex_state = 5},
  [2082] = {.lex_state = 106, .external_lex_state = 4},
  [2083] = {.lex_state = 271, .external_lex_state = 5},
  [2084] = {.lex_state = 106, .external_lex_state = 4},
  [2085] = {.lex_state = 106, .external_lex_state = 4},
  [2086] = {.lex_state = 273, .external_lex_state = 5},
  [2087] = {.lex_state = 273, .external_lex_state = 5},
  [2088] = {.lex_state = 273, .external_lex_state = 5},
  [2089] = {.lex_state = 273, .external_lex_state = 5},
  [2090] = {.lex_state = 273, .external_lex_state = 5},
  [2091] = {.lex_state = 273, .external_lex_state = 5},
  [2092] = {.lex_state = 273, .external_lex_state = 5},
  [2093] = {.lex_state = 271, .external_lex_state = 5},
  [2094] = {.lex_state = 273, .external_lex_state = 5},
  [2095] = {.lex_state = 273, .external_lex_state = 5},
  [2096] = {.lex_state = 273, .external_lex_state = 5},
  [2097] = {.lex_state = 106, .external_lex_state = 4},
  [2098] = {.lex_state = 273, .external_lex_state = 5},
  [2099] = {.lex_state = 273, .external_lex_state = 5},
  [2100] = {.lex_state = 106, .external_lex_state = 4},
  [2101] = {.lex_state = 106, .external_lex_state = 4},
  [2102] = {.lex_state = 273, .external_lex_state = 5},
  [2103] = {.lex_state = 273, .external_lex_state = 5},
  [2104] = {.lex_state = 273, .external_lex_state = 5},
  [2105] = {.lex_state = 106, .external_lex_state = 4},
  [2106] = {.lex_state = 273, .external_lex_state = 5},
  [2107] = {.lex_state = 273, .external_lex_state = 5},
  [2108] = {.lex_state = 273, .external_lex_state = 5},
  [2109] = {.lex_state = 271, .external_lex_state = 5},
  [2110] = {.lex_state = 106, .external_lex_state = 4},
  [2111] = {.lex_state = 106, .external_lex_state = 4},
  [2112] = {.lex_state = 273, .external_lex_state = 5},
  [2113] = {.lex_state = 273, .external_lex_state = 5},
  [2114] = {.lex_state = 271, .external_lex_state = 5},
  [2115] = {.lex_state = 271, .external_lex_state = 5},
  [2116] = {.lex_state = 271, .external_lex_state = 5},
  [2117] = {.lex_state = 106, .external_lex_state = 4},
  [2118] = {.lex_state = 106, .external_lex_state = 4},
  [2119] = {.lex_state = 106, .external_lex_state = 4},
  [2120] = {.lex_state = 271, .external_lex_state = 5},
  [2121] = {.lex_state = 271, .external_lex_state = 5},
  [2122] = {.lex_state = 273, .external_lex_state = 5},
  [2123] = {.lex_state = 271, .external_lex_state = 4},
  [2124] = {.lex_state = 271, .external_lex_state = 4},
  [2125] = {.lex_state = 271, .external_lex_state = 4},
  [2126] = {.lex_state = 273, .external_lex_state = 5},
  [2127] = {.lex_state = 273, .external_lex_state = 5},
  [2128] = {.lex_state = 273, .external_lex_state = 5},
  [2129] = {.lex_state = 273, .external_lex_state = 5},
  [2130] = {.lex_state = 273, .external_lex_state = 5},
  [2131] = {.lex_state = 120, .external_lex_state = 4},
  [2132] = {.lex_state = 120, .external_lex_state = 4},
  [2133] = {.lex_state = 120, .external_lex_state = 4},
  [2134] = {.lex_state = 120, .external_lex_state = 4},
  [2135] = {.lex_state = 120, .external_lex_state = 4},
  [2136] = {.lex_state = 120, .external_lex_state = 4},
  [2137] = {.lex_state = 120, .external_lex_state = 4},
  [2138] = {.lex_state = 120, .external_lex_state = 4},
  [2139] = {.lex_state = 120, .external_lex_state = 4},
  [2140] = {.lex_state = 273, .external_lex_state = 5},
  [2141] = {.lex_state = 271, .external_lex_state = 4},
  [2142] = {.lex_state = 271, .external_lex_state = 5},
  [2143] = {.lex_state = 125, .external_lex_state = 5},
  [2144] = {.lex_state = 271, .external_lex_state = 5},
  [2145] = {.lex_state = 271, .external_lex_state = 5},
  [2146] = {.lex_state = 271, .external_lex_state = 5},
  [2147] = {.lex_state = 271, .external_lex_state = 5},
  [2148] = {.lex_state = 271, .external_lex_state = 5},
  [2149] = {.lex_state = 271, .external_lex_state = 5},
  [2150] = {.lex_state = 271, .external_lex_state = 5},
  [2151] = {.lex_state = 125, .external_lex_state = 5},
  [2152] = {.lex_state = 125, .external_lex_state = 5},
  [2153] = {.lex_state = 271, .external_lex_state = 5},
  [2154] = {.lex_state = 125, .external_lex_state = 5},
  [2155] = {.lex_state = 271, .external_lex_state = 4},
  [2156] = {.lex_state = 271, .external_lex_state = 5},
  [2157] = {.lex_state = 125, .external_lex_state = 5},
  [2158] = {.lex_state = 271, .external_lex_state = 5},
  [2159] = {.lex_state = 125, .external_lex_state = 5},
  [2160] = {.lex_state = 125, .external_lex_state = 5},
  [2161] = {.lex_state = 125, .external_lex_state = 5},
  [2162] = {.lex_state = 271, .external_lex_state = 5},
  [2163] = {.lex_state = 125, .external_lex_state = 5},
  [2164] = {.lex_state = 271, .external_lex_state = 5},
  [2165] = {.lex_state = 271, .external_lex_state = 5},
  [2166] = {.lex_state = 125, .external_lex_state = 5},
  [2167] = {.lex_state = 271, .external_lex_state = 4},
  [2168] = {.lex_state = 106, .external_lex_state = 5},
  [2169] = {.lex_state = 271, .external_lex_state = 4},
  [2170] = {.lex_state = 271, .external_lex_state = 5},
  [2171] = {.lex_state = 271, .external_lex_state = 4},
  [2172] = {.lex_state = 271, .external_lex_state = 5},
  [2173] = {.lex_state = 271, .external_lex_state = 5},
  [2174] = {.lex_state = 271, .external_lex_state = 5},
  [2175] = {.lex_state = 271, .external_lex_state = 5},
  [2176] = {.lex_state = 271, .external_lex_state = 5},
  [2177] = {.lex_state = 271, .external_lex_state = 5},
  [2178] = {.lex_state = 271, .external_lex_state = 5},
  [2179] = {.lex_state = 271, .external_lex_state = 5},
  [2180] = {.lex_state = 271, .external_lex_state = 5},
  [2181] = {.lex_state = 271, .external_lex_state = 5},
  [2182] = {.lex_state = 271, .external_lex_state = 5},
  [2183] = {.lex_state = 271, .external_lex_state = 5},
  [2184] = {.lex_state = 271, .external_lex_state = 5},
  [2185] = {.lex_state = 271, .external_lex_state = 5},
  [2186] = {.lex_state = 271, .external_lex_state = 5},
  [2187] = {.lex_state = 271, .external_lex_state = 4},
  [2188] = {.lex_state = 106, .external_lex_state = 5},
  [2189] = {.lex_state = 271, .external_lex_state = 5},
  [2190] = {.lex_state = 271, .external_lex_state = 4},
  [2191] = {.lex_state = 271, .external_lex_state = 4},
  [2192] = {.lex_state = 271, .external_lex_state = 4},
  [2193] = {.lex_state = 106, .external_lex_state = 5},
  [2194] = {.lex_state = 271, .external_lex_state = 5},
  [2195] = {.lex_state = 125, .external_lex_state = 5},
  [2196] = {.lex_state = 271, .external_lex_state = 4},
  [2197] = {.lex_state = 125, .external_lex_state = 5},
  [2198] = {.lex_state = 106, .external_lex_state = 5},
  [2199] = {.lex_state = 106, .external_lex_state = 5},
  [2200] = {.lex_state = 125, .external_lex_state = 5},
  [2201] = {.lex_state = 271, .external_lex_state = 5},
  [2202] = {.lex_state = 106, .external_lex_state = 5},
  [2203] = {.lex_state = 125, .external_lex_state = 5},
  [2204] = {.lex_state = 271, .external_lex_state = 4},
  [2205] = {.lex_state = 125, .external_lex_state = 5},
  [2206] = {.lex_state = 271, .external_lex_state = 4},
  [2207] = {.lex_state = 125, .external_lex_state = 5},
  [2208] = {.lex_state = 271, .external_lex_state = 5},
  [2209] = {.lex_state = 271, .external_lex_state = 4},
  [2210] = {.lex_state = 125, .external_lex_state = 5},
  [2211] = {.lex_state = 271, .external_lex_state = 5},
  [2212] = {.lex_state = 125, .external_lex_state = 5},
  [2213] = {.lex_state = 125, .external_lex_state = 5},
  [2214] = {.lex_state = 271, .external_lex_state = 5},
  [2215] = {.lex_state = 271, .external_lex_state = 5},
  [2216] = {.lex_state = 271, .external_lex_state = 5},
  [2217] = {.lex_state = 125, .external_lex_state = 5},
  [2218] = {.lex_state = 106, .external_lex_state = 5},
  [2219] = {.lex_state = 125, .external_lex_state = 5},
  [2220] = {.lex_state = 271, .external_lex_state = 5},
  [2221] = {.lex_state = 125, .external_lex_state = 5},
  [2222] = {.lex_state = 125, .external_lex_state = 5},
  [2223] = {.lex_state = 125, .external_lex_state = 5},
  [2224] = {.lex_state = 125, .external_lex_state = 5},
  [2225] = {.lex_state = 125, .external_lex_state = 5},
  [2226] = {.lex_state = 271, .external_lex_state = 4},
  [2227] = {.lex_state = 125, .external_lex_state = 5},
  [2228] = {.lex_state = 125, .external_lex_state = 5},
  [2229] = {.lex_state = 125, .external_lex_state = 5},
  [2230] = {.lex_state = 125, .external_lex_state = 5},
  [2231] = {.lex_state = 125, .external_lex_state = 5},
  [2232] = {.lex_state = 125, .external_lex_state = 5},
  [2233] = {.lex_state = 271, .external_lex_state = 5},
  [2234] = {.lex_state = 271, .external_lex_state = 5},
  [2235] = {.lex_state = 125, .external_lex_state = 5},
  [2236] = {.lex_state = 125, .external_lex_state = 5},
  [2237] = {.lex_state = 125, .external_lex_state = 5},
  [2238] = {.lex_state = 125, .external_lex_state = 5},
  [2239] = {.lex_state = 125, .external_lex_state = 5},
  [2240] = {.lex_state = 271, .external_lex_state = 5},
  [2241] = {.lex_state = 271, .external_lex_state = 5},
  [2242] = {.lex_state = 271, .external_lex_state = 5},
  [2243] = {.lex_state = 271, .external_lex_state = 5},
  [2244] = {.lex_state = 271, .external_lex_state = 5},
  [2245] = {.lex_state = 271, .external_lex_state = 5},
  [2246] = {.lex_state = 271, .external_lex_state = 5},
  [2247] = {.lex_state = 125, .external_lex_state = 5},
  [2248] = {.lex_state = 106, .external_lex_state = 5},
  [2249] = {.lex_state = 271, .external_lex_state = 4},
  [2250] = {.lex_state = 125, .external_lex_state = 5},
  [2251] = {.lex_state = 125, .external_lex_state = 5},
  [2252] = {.lex_state = 271, .external_lex_state = 5},
  [2253] = {.lex_state = 271, .external_lex_state = 5},
  [2254] = {.lex_state = 125, .external_lex_state = 5},
  [2255] = {.lex_state = 271, .external_lex_state = 5},
  [2256] = {.lex_state = 271, .external_lex_state = 5},
  [2257] = {.lex_state = 271, .external_lex_state = 5},
  [2258] = {.lex_state = 106, .external_lex_state = 5},
  [2259] = {.lex_state = 271, .external_lex_state = 5},
  [2260] = {.lex_state = 125, .external_lex_state = 5},
  [2261] = {.lex_state = 271, .external_lex_state = 5},
  [2262] = {.lex_state = 271, .external_lex_state = 5},
  [2263] = {.lex_state = 271, .external_lex_state = 5},
  [2264] = {.lex_state = 271, .external_lex_state = 5},
  [2265] = {.lex_state = 125, .external_lex_state = 5},
  [2266] = {.lex_state = 271, .external_lex_state = 5},
  [2267] = {.lex_state = 271, .external_lex_state = 5},
  [2268] = {.lex_state = 271, .external_lex_state = 5},
  [2269] = {.lex_state = 271, .external_lex_state = 5},
  [2270] = {.lex_state = 125, .external_lex_state = 5},
  [2271] = {.lex_state = 125, .external_lex_state = 5},
  [2272] = {.lex_state = 271, .external_lex_state = 5},
  [2273] = {.lex_state = 271, .external_lex_state = 5},
  [2274] = {.lex_state = 271, .external_lex_state = 5},
  [2275] = {.lex_state = 271, .external_lex_state = 4},
  [2276] = {.lex_state = 271, .external_lex_state = 4},
  [2277] = {.lex_state = 271, .external_lex_state = 5},
  [2278] = {.lex_state = 271, .external_lex_state = 5},
  [2279] = {.lex_state = 125, .external_lex_state = 5},
  [2280] = {.lex_state = 271, .external_lex_state = 4},
  [2281] = {.lex_state = 271, .external_lex_state = 4},
  [2282] = {.lex_state = 125, .external_lex_state = 5},
  [2283] = {.lex_state = 125, .external_lex_state = 5},
  [2284] = {.lex_state = 271, .external_lex_state = 5},
  [2285] = {.lex_state = 271, .external_lex_state = 5},
  [2286] = {.lex_state = 125, .external_lex_state = 5},
  [2287] = {.lex_state = 125, .external_lex_state = 5},
  [2288] = {.lex_state = 106, .external_lex_state = 5},
  [2289] = {.lex_state = 125, .external_lex_state = 5},
  [2290] = {.lex_state = 125, .external_lex_state = 5},
  [2291] = {.lex_state = 271, .external_lex_state = 4},
  [2292] = {.lex_state = 106, .external_lex_state = 5},
  [2293] = {.lex_state = 271, .external_lex_state = 5},
  [2294] = {.lex_state = 271, .external_lex_state = 5},
  [2295] = {.lex_state = 125, .external_lex_state = 5},
  [2296] = {.lex_state = 125, .external_lex_state = 5},
  [2297] = {.lex_state = 125, .external_lex_state = 5},
  [2298] = {.lex_state = 271, .external_lex_state = 5},
  [2299] = {.lex_state = 106, .external_lex_state = 5},
  [2300] = {.lex_state = 271, .external_lex_state = 5},
  [2301] = {.lex_state = 271, .external_lex_state = 5},
  [2302] = {.lex_state = 125, .external_lex_state = 5},
  [2303] = {.lex_state = 125, .external_lex_state = 5},
  [2304] = {.lex_state = 271, .external_lex_state = 5},
  [2305] = {.lex_state = 271, .external_lex_state = 5},
  [2306] = {.lex_state = 271, .external_lex_state = 5},
  [2307] = {.lex_state = 271, .external_lex_state = 4},
  [2308] = {.lex_state = 271, .external_lex_state = 5},
  [2309] = {.lex_state = 271, .external_lex_state = 5},
  [2310] = {.lex_state = 271, .external_lex_state = 5},
  [2311] = {.lex_state = 271, .external_lex_state = 5},
  [2312] = {.lex_state = 271, .external_lex_state = 5},
  [2313] = {.lex_state = 271, .external_lex_state = 5},
  [2314] = {.lex_state = 271, .external_lex_state = 5},
  [2315] = {.lex_state = 271, .external_lex_state = 5},
  [2316] = {.lex_state = 125, .external_lex_state = 5},
  [2317] = {.lex_state = 125, .external_lex_state = 5},
  [2318] = {.lex_state = 271, .external_lex_state = 5},
  [2319] = {.lex_state = 271, .external_lex_state = 5},
  [2320] = {.lex_state = 271, .external_lex_state = 5},
  [2321] = {.lex_state = 271, .external_lex_state = 5},
  [2322] = {.lex_state = 271, .external_lex_state = 5},
  [2323] = {.lex_state = 125, .external_lex_state = 5},
  [2324] = {.lex_state = 271, .external_lex_state = 5},
  [2325] = {.lex_state = 125, .external_lex_state = 5},
  [2326] = {.lex_state = 125, .external_lex_state = 5},
  [2327] = {.lex_state = 125, .external_lex_state = 5},
  [2328] = {.lex_state = 271, .external_lex_state = 5},
  [2329] = {.lex_state = 271, .external_lex_state = 5},
  [2330] = {.lex_state = 125, .external_lex_state = 5},
  [2331] = {.lex_state = 125, .external_lex_state = 5},
  [2332] = {.lex_state = 125, .external_lex_state = 5},
  [2333] = {.lex_state = 271, .external_lex_state = 5},
  [2334] = {.lex_state = 271, .external_lex_state = 5},
  [2335] = {.lex_state = 271, .external_lex_state = 4},
  [2336] = {.lex_state = 271, .external_lex_state = 4},
  [2337] = {.lex_state = 271, .external_lex_state = 4},
  [2338] = {.lex_state = 125, .external_lex_state = 5},
  [2339] = {.lex_state = 125, .external_lex_state = 5},
  [2340] = {.lex_state = 125, .external_lex_state = 5},
  [2341] = {.lex_state = 125, .external_lex_state = 5},
  [2342] = {.lex_state = 125, .external_lex_state = 5},
  [2343] = {.lex_state = 125, .external_lex_state = 5},
  [2344] = {.lex_state = 125, .external_lex_state = 5},
  [2345] = {.lex_state = 125, .external_lex_state = 5},
  [2346] = {.lex_state = 125, .external_lex_state = 5},
  [2347] = {.lex_state = 271, .external_lex_state = 5},
  [2348] = {.lex_state = 271, .external_lex_state = 5},
  [2349] = {.lex_state = 271, .external_lex_state = 5},
  [2350] = {.lex_state = 271, .external_lex_state = 5},
  [2351] = {.lex_state = 125, .external_lex_state = 5},
  [2352] = {.lex_state = 271, .external_lex_state = 5},
  [2353] = {.lex_state = 271, .external_lex_state = 5},
  [2354] = {.lex_state = 106, .external_lex_state = 5},
  [2355] = {.lex_state = 271, .external_lex_state = 5},
  [2356] = {.lex_state = 271, .external_lex_state = 5},
  [2357] = {.lex_state = 271, .external_lex_state = 4},
  [2358] = {.lex_state = 125, .external_lex_state = 5},
  [2359] = {.lex_state = 125, .external_lex_state = 5},
  [2360] = {.lex_state = 125, .external_lex_state = 5},
  [2361] = {.lex_state = 125, .external_lex_state = 5},
  [2362] = {.lex_state = 271, .external_lex_state = 5},
  [2363] = {.lex_state = 271, .external_lex_state = 5},
  [2364] = {.lex_state = 275, .external_lex_state = 4},
  [2365] = {.lex_state = 130, .external_lex_state = 5},
  [2366] = {.lex_state = 130, .external_lex_state = 5},
  [2367] = {.lex_state = 130, .external_lex_state = 5},
  [2368] = {.lex_state = 130, .external_lex_state = 5},
  [2369] = {.lex_state = 130, .external_lex_state = 5},
  [2370] = {.lex_state = 130, .external_lex_state = 5},
  [2371] = {.lex_state = 130, .external_lex_state = 5},
  [2372] = {.lex_state = 273, .external_lex_state = 5},
  [2373] = {.lex_state = 273, .external_lex_state = 5},
  [2374] = {.lex_state = 273, .external_lex_state = 5},
  [2375] = {.lex_state = 273, .external_lex_state = 5},
  [2376] = {.lex_state = 275, .external_lex_state = 4},
  [2377] = {.lex_state = 275, .external_lex_state = 4},
  [2378] = {.lex_state = 273, .external_lex_state = 5},
  [2379] = {.lex_state = 273, .external_lex_state = 5},
  [2380] = {.lex_state = 130, .external_lex_state = 5},
  [2381] = {.lex_state = 130, .external_lex_state = 5},
  [2382] = {.lex_state = 273, .external_lex_state = 5},
  [2383] = {.lex_state = 273, .external_lex_state = 5},
  [2384] = {.lex_state = 130, .external_lex_state = 5},
  [2385] = {.lex_state = 130, .external_lex_state = 5},
  [2386] = {.lex_state = 130, .external_lex_state = 5},
  [2387] = {.lex_state = 130, .external_lex_state = 5},
  [2388] = {.lex_state = 130, .external_lex_state = 5},
  [2389] = {.lex_state = 130, .external_lex_state = 5},
  [2390] = {.lex_state = 130, .external_lex_state = 5},
  [2391] = {.lex_state = 130, .external_lex_state = 5},
  [2392] = {.lex_state = 130, .external_lex_state = 5},
  [2393] = {.lex_state = 130, .external_lex_state = 5},
  [2394] = {.lex_state = 130, .external_lex_state = 5},
  [2395] = {.lex_state = 130, .external_lex_state = 5},
  [2396] = {.lex_state = 130, .external_lex_state = 5},
  [2397] = {.lex_state = 130, .external_lex_state = 5},
  [2398] = {.lex_state = 130, .external_lex_state = 5},
  [2399] = {.lex_state = 130, .external_lex_state = 5},
  [2400] = {.lex_state = 130, .external_lex_state = 5},
  [2401] = {.lex_state = 130, .external_lex_state = 5},
  [2402] = {.lex_state = 130, .external_lex_state = 5},
  [2403] = {.lex_state = 130, .external_lex_state = 5},
  [2404] = {.lex_state = 130, .external_lex_state = 5},
  [2405] = {.lex_state = 130, .external_lex_state = 5},
  [2406] = {.lex_state = 130, .external_lex_state = 5},
  [2407] = {.lex_state = 130, .external_lex_state = 5},
  [2408] = {.lex_state = 125, .external_lex_state = 5},
  [2409] = {.lex_state = 125, .external_lex_state = 5},
  [2410] = {.lex_state = 130, .external_lex_state = 5},
  [2411] = {.lex_state = 130, .external_lex_state = 5},
  [2412] = {.lex_state = 125, .external_lex_state = 5},
  [2413] = {.lex_state = 125, .external_lex_state = 5},
  [2414] = {.lex_state = 130, .external_lex_state = 5},
  [2415] = {.lex_state = 130, .external_lex_state = 5},
  [2416] = {.lex_state = 125, .external_lex_state = 5},
  [2417] = {.lex_state = 125, .external_lex_state = 5},
  [2418] = {.lex_state = 130, .external_lex_state = 5},
  [2419] = {.lex_state = 130, .external_lex_state = 5},
  [2420] = {.lex_state = 130, .external_lex_state = 5},
  [2421] = {.lex_state = 130, .external_lex_state = 5},
  [2422] = {.lex_state = 130, .external_lex_state = 5},
  [2423] = {.lex_state = 130, .external_lex_state = 5},
  [2424] = {.lex_state = 130, .external_lex_state = 5},
  [2425] = {.lex_state = 130, .external_lex_state = 5},
  [2426] = {.lex_state = 130, .external_lex_state = 5},
  [2427] = {.lex_state = 130, .external_lex_state = 5},
  [2428] = {.lex_state = 130, .external_lex_state = 5},
  [2429] = {.lex_state = 130, .external_lex_state = 5},
  [2430] = {.lex_state = 130, .external_lex_state = 5},
  [2431] = {.lex_state = 130, .external_lex_state = 5},
  [2432] = {.lex_state = 273, .external_lex_state = 5},
  [2433] = {.lex_state = 273, .external_lex_state = 5},
  [2434] = {.lex_state = 273, .external_lex_state = 5},
  [2435] = {.lex_state = 273, .external_lex_state = 5},
  [2436] = {.lex_state = 273, .external_lex_state = 5},
  [2437] = {.lex_state = 273, .external_lex_state = 5},
  [2438] = {.lex_state = 130, .external_lex_state = 5},
  [2439] = {.lex_state = 130, .external_lex_state = 5},
  [2440] = {.lex_state = 130, .external_lex_state = 5},
  [2441] = {.lex_state = 130, .external_lex_state = 5},
  [2442] = {.lex_state = 130, .external_lex_state = 5},
  [2443] = {.lex_state = 130, .external_lex_state = 5},
  [2444] = {.lex_state = 130, .external_lex_state = 5},
  [2445] = {.lex_state = 130, .external_lex_state = 5},
  [2446] = {.lex_state = 130, .external_lex_state = 5},
  [2447] = {.lex_state = 130, .external_lex_state = 5},
  [2448] = {.lex_state = 130, .external_lex_state = 5},
  [2449] = {.lex_state = 130, .external_lex_state = 5},
  [2450] = {.lex_state = 130, .external_lex_state = 5},
  [2451] = {.lex_state = 130, .external_lex_state = 5},
  [2452] = {.lex_state = 130, .external_lex_state = 5},
  [2453] = {.lex_state = 130, .external_lex_state = 5},
  [2454] = {.lex_state = 271, .external_lex_state = 4},
  [2455] = {.lex_state = 273, .external_lex_state = 5},
  [2456] = {.lex_state = 273, .external_lex_state = 5},
  [2457] = {.lex_state = 271, .external_lex_state = 4},
  [2458] = {.lex_state = 271, .external_lex_state = 5},
  [2459] = {.lex_state = 106, .external_lex_state = 5},
  [2460] = {.lex_state = 273, .external_lex_state = 5},
  [2461] = {.lex_state = 273, .external_lex_state = 5},
  [2462] = {.lex_state = 273, .external_lex_state = 5},
  [2463] = {.lex_state = 273, .external_lex_state = 5},
  [2464] = {.lex_state = 273, .external_lex_state = 5},
  [2465] = {.lex_state = 273, .external_lex_state = 5},
  [2466] = {.lex_state = 273, .external_lex_state = 5},
  [2467] = {.lex_state = 106, .external_lex_state = 5},
  [2468] = {.lex_state = 106, .external_lex_state = 5},
  [2469] = {.lex_state = 125, .external_lex_state = 5},
  [2470] = {.lex_state = 125, .external_lex_state = 5},
  [2471] = {.lex_state = 125, .external_lex_state = 5},
  [2472] = {.lex_state = 125, .external_lex_state = 5},
  [2473] = {.lex_state = 125, .external_lex_state = 5},
  [2474] = {.lex_state = 125, .external_lex_state = 5},
  [2475] = {.lex_state = 275, .external_lex_state = 4},
  [2476] = {.lex_state = 275, .external_lex_state = 4},
  [2477] = {.lex_state = 275, .external_lex_state = 4},
  [2478] = {.lex_state = 106, .external_lex_state = 5},
  [2479] = {.lex_state = 106, .external_lex_state = 5},
  [2480] = {.lex_state = 275, .external_lex_state = 4},
  [2481] = {.lex_state = 275, .external_lex_state = 4},
  [2482] = {.lex_state = 271, .external_lex_state = 5},
  [2483] = {.lex_state = 271, .external_lex_state = 5},
  [2484] = {.lex_state = 275, .external_lex_state = 4},
  [2485] = {.lex_state = 125, .external_lex_state = 5},
  [2486] = {.lex_state = 125, .external_lex_state = 5},
  [2487] = {.lex_state = 271, .external_lex_state = 4},
  [2488] = {.lex_state = 125, .external_lex_state = 5},
  [2489] = {.lex_state = 125, .external_lex_state = 5},
  [2490] = {.lex_state = 125, .external_lex_state = 5},
  [2491] = {.lex_state = 125, .external_lex_state = 5},
  [2492] = {.lex_state = 125, .external_lex_state = 5},
  [2493] = {.lex_state = 125, .external_lex_state = 5},
  [2494] = {.lex_state = 125, .external_lex_state = 5},
  [2495] = {.lex_state = 125, .external_lex_state = 5},
  [2496] = {.lex_state = 275, .external_lex_state = 4},
  [2497] = {.lex_state = 271, .external_lex_state = 5},
  [2498] = {.lex_state = 275, .external_lex_state = 4},
  [2499] = {.lex_state = 271, .external_lex_state = 5},
  [2500] = {.lex_state = 271, .external_lex_state = 5},
  [2501] = {.lex_state = 271, .external_lex_state = 4},
  [2502] = {.lex_state = 271, .external_lex_state = 5},
  [2503] = {.lex_state = 106, .external_lex_state = 5},
  [2504] = {.lex_state = 271, .external_lex_state = 5},
  [2505] = {.lex_state = 271, .external_lex_state = 5},
  [2506] = {.lex_state = 106, .external_lex_state = 5},
  [2507] = {.lex_state = 273, .external_lex_state = 5},
  [2508] = {.lex_state = 273, .external_lex_state = 5},
  [2509] = {.lex_state = 106, .external_lex_state = 5},
  [2510] = {.lex_state = 106, .external_lex_state = 5},
  [2511] = {.lex_state = 273, .external_lex_state = 5},
  [2512] = {.lex_state = 273, .external_lex_state = 5},
  [2513] = {.lex_state = 273, .external_lex_state = 5},
  [2514] = {.lex_state = 273, .external_lex_state = 5},
  [2515] = {.lex_state = 273, .external_lex_state = 5},
  [2516] = {.lex_state = 273, .external_lex_state = 5},
  [2517] = {.lex_state = 273, .external_lex_state = 5},
  [2518] = {.lex_state = 273, .external_lex_state = 5},
  [2519] = {.lex_state = 271, .external_lex_state = 5},
  [2520] = {.lex_state = 273, .external_lex_state = 5},
  [2521] = {.lex_state = 125, .external_lex_state = 5},
  [2522] = {.lex_state = 125, .external_lex_state = 5},
  [2523] = {.lex_state = 125, .external_lex_state = 5},
  [2524] = {.lex_state = 125, .external_lex_state = 5},
  [2525] = {.lex_state = 125, .external_lex_state = 5},
  [2526] = {.lex_state = 125, .external_lex_state = 5},
  [2527] = {.lex_state = 271, .external_lex_state = 5},
  [2528] = {.lex_state = 125, .external_lex_state = 5},
  [2529] = {.lex_state = 125, .external_lex_state = 5},
  [2530] = {.lex_state = 125, .external_lex_state = 5},
  [2531] = {.lex_state = 125, .external_lex_state = 5},
  [2532] = {.lex_state = 125, .external_lex_state = 5},
  [2533] = {.lex_state = 125, .external_lex_state = 5},
  [2534] = {.lex_state = 125, .external_lex_state = 5},
  [2535] = {.lex_state = 125, .external_lex_state = 5},
  [2536] = {.lex_state = 125, .external_lex_state = 5},
  [2537] = {.lex_state = 125, .external_lex_state = 5},
  [2538] = {.lex_state = 273, .external_lex_state = 5},
  [2539] = {.lex_state = 273, .external_lex_state = 5},
  [2540] = {.lex_state = 273, .external_lex_state = 5},
  [2541] = {.lex_state = 273, .external_lex_state = 5},
  [2542] = {.lex_state = 125, .external_lex_state = 5},
  [2543] = {.lex_state = 271, .external_lex_state = 5},
  [2544] = {.lex_state = 125, .external_lex_state = 5},
  [2545] = {.lex_state = 125, .external_lex_state = 5},
  [2546] = {.lex_state = 125, .external_lex_state = 5},
  [2547] = {.lex_state = 125, .external_lex_state = 5},
  [2548] = {.lex_state = 125, .external_lex_state = 5},
  [2549] = {.lex_state = 125, .external_lex_state = 5},
  [2550] = {.lex_state = 125, .external_lex_state = 5},
  [2551] = {.lex_state = 125, .external_lex_state = 5},
  [2552] = {.lex_state = 125, .external_lex_state = 5},
  [2553] = {.lex_state = 125, .external_lex_state = 5},
  [2554] = {.lex_state = 125, .external_lex_state = 5},
  [2555] = {.lex_state = 125, .external_lex_state = 5},
  [2556] = {.lex_state = 125, .external_lex_state = 5},
  [2557] = {.lex_state = 125, .external_lex_state = 5},
  [2558] = {.lex_state = 125, .external_lex_state = 5},
  [2559] = {.lex_state = 273, .external_lex_state = 5},
  [2560] = {.lex_state = 106, .external_lex_state = 5},
  [2561] = {.lex_state = 106, .external_lex_state = 5},
  [2562] = {.lex_state = 106, .external_lex_state = 5},
  [2563] = {.lex_state = 133, .external_lex_state = 5},
  [2564] = {.lex_state = 133, .external_lex_state = 5},
  [2565] = {.lex_state = 271, .external_lex_state = 5},
  [2566] = {.lex_state = 133, .external_lex_state = 5},
  [2567] = {.lex_state = 106, .external_lex_state = 5},
  [2568] = {.lex_state = 106, .external_lex_state = 5},
  [2569] = {.lex_state = 106, .external_lex_state = 5},
  [2570] = {.lex_state = 106, .external_lex_state = 5},
  [2571] = {.lex_state = 106, .external_lex_state = 5},
  [2572] = {.lex_state = 275, .external_lex_state = 4},
  [2573] = {.lex_state = 106, .external_lex_state = 5},
  [2574] = {.lex_state = 106, .external_lex_state = 5},
  [2575] = {.lex_state = 106, .external_lex_state = 5},
  [2576] = {.lex_state = 106, .external_lex_state = 5},
  [2577] = {.lex_state = 106, .external_lex_state = 5},
  [2578] = {.lex_state = 106, .external_lex_state = 5},
  [2579] = {.lex_state = 106, .external_lex_state = 5},
  [2580] = {.lex_state = 106, .external_lex_state = 5},
  [2581] = {.lex_state = 106, .external_lex_state = 5},
  [2582] = {.lex_state = 106, .external_lex_state = 5},
  [2583] = {.lex_state = 133, .external_lex_state = 5},
  [2584] = {.lex_state = 271, .external_lex_state = 5},
  [2585] = {.lex_state = 271, .external_lex_state = 5},
  [2586] = {.lex_state = 271, .external_lex_state = 4},
  [2587] = {.lex_state = 271, .external_lex_state = 4},
  [2588] = {.lex_state = 271, .external_lex_state = 5},
  [2589] = {.lex_state = 133, .external_lex_state = 5},
  [2590] = {.lex_state = 133, .external_lex_state = 5},
  [2591] = {.lex_state = 273, .external_lex_state = 5},
  [2592] = {.lex_state = 273, .external_lex_state = 5},
  [2593] = {.lex_state = 273, .external_lex_state = 5},
  [2594] = {.lex_state = 273, .external_lex_state = 5},
  [2595] = {.lex_state = 273, .external_lex_state = 5},
  [2596] = {.lex_state = 273, .external_lex_state = 5},
  [2597] = {.lex_state = 273, .external_lex_state = 5},
  [2598] = {.lex_state = 273, .external_lex_state = 5},
  [2599] = {.lex_state = 271, .external_lex_state = 5},
  [2600] = {.lex_state = 273, .external_lex_state = 5},
  [2601] = {.lex_state = 273, .external_lex_state = 5},
  [2602] = {.lex_state = 273, .external_lex_state = 5},
  [2603] = {.lex_state = 273, .external_lex_state = 5},
  [2604] = {.lex_state = 273, .external_lex_state = 5},
  [2605] = {.lex_state = 273, .external_lex_state = 5},
  [2606] = {.lex_state = 273, .external_lex_state = 5},
  [2607] = {.lex_state = 273, .external_lex_state = 5},
  [2608] = {.lex_state = 271, .external_lex_state = 5},
  [2609] = {.lex_state = 271, .external_lex_state = 5},
  [2610] = {.lex_state = 273, .external_lex_state = 5},
  [2611] = {.lex_state = 273, .external_lex_state = 5},
  [2612] = {.lex_state = 273, .external_lex_state = 5},
  [2613] = {.lex_state = 273, .external_lex_state = 5},
  [2614] = {.lex_state = 271, .external_lex_state = 5},
  [2615] = {.lex_state = 273, .external_lex_state = 5},
  [2616] = {.lex_state = 273, .external_lex_state = 5},
  [2617] = {.lex_state = 273, .external_lex_state = 5},
  [2618] = {.lex_state = 273, .external_lex_state = 5},
  [2619] = {.lex_state = 273, .external_lex_state = 5},
  [2620] = {.lex_state = 273, .external_lex_state = 5},
  [2621] = {.lex_state = 273, .external_lex_state = 5},
  [2622] = {.lex_state = 271, .external_lex_state = 5},
  [2623] = {.lex_state = 273, .external_lex_state = 5},
  [2624] = {.lex_state = 273, .external_lex_state = 5},
  [2625] = {.lex_state = 273, .external_lex_state = 5},
  [2626] = {.lex_state = 273, .external_lex_state = 5},
  [2627] = {.lex_state = 106, .external_lex_state = 5},
  [2628] = {.lex_state = 106, .external_lex_state = 5},
  [2629] = {.lex_state = 106, .external_lex_state = 5},
  [2630] = {.lex_state = 273, .external_lex_state = 5},
  [2631] = {.lex_state = 106, .external_lex_state = 5},
  [2632] = {.lex_state = 106, .external_lex_state = 5},
  [2633] = {.lex_state = 106, .external_lex_state = 5},
  [2634] = {.lex_state = 106, .external_lex_state = 5},
  [2635] = {.lex_state = 106, .external_lex_state = 5},
  [2636] = {.lex_state = 271, .external_lex_state = 5},
  [2637] = {.lex_state = 106, .external_lex_state = 5},
  [2638] = {.lex_state = 106, .external_lex_state = 5},
  [2639] = {.lex_state = 106, .external_lex_state = 5},
  [2640] = {.lex_state = 271, .external_lex_state = 5},
  [2641] = {.lex_state = 106, .external_lex_state = 5},
  [2642] = {.lex_state = 106, .external_lex_state = 5},
  [2643] = {.lex_state = 106, .external_lex_state = 5},
  [2644] = {.lex_state = 106, .external_lex_state = 5},
  [2645] = {.lex_state = 106, .external_lex_state = 5},
  [2646] = {.lex_state = 271, .external_lex_state = 5},
  [2647] = {.lex_state = 271, .external_lex_state = 5},
  [2648] = {.lex_state = 271, .external_lex_state = 5},
  [2649] = {.lex_state = 271, .external_lex_state = 5},
  [2650] = {.lex_state = 271, .external_lex_state = 5},
  [2651] = {.lex_state = 271, .external_lex_state = 5},
  [2652] = {.lex_state = 271, .external_lex_state = 5},
  [2653] = {.lex_state = 271, .external_lex_state = 5},
  [2654] = {.lex_state = 271, .external_lex_state = 5},
  [2655] = {.lex_state = 271, .external_lex_state = 5},
  [2656] = {.lex_state = 271, .external_lex_state = 5},
  [2657] = {.lex_state = 271, .external_lex_state = 5},
  [2658] = {.lex_state = 271, .external_lex_state = 5},
  [2659] = {.lex_state = 271, .external_lex_state = 5},
  [2660] = {.lex_state = 271, .external_lex_state = 5},
  [2661] = {.lex_state = 271, .external_lex_state = 5},
  [2662] = {.lex_state = 271, .external_lex_state = 5},
  [2663] = {.lex_state = 271, .external_lex_state = 5},
  [2664] = {.lex_state = 133, .external_lex_state = 5},
  [2665] = {.lex_state = 106, .external_lex_state = 5},
  [2666] = {.lex_state = 275, .external_lex_state = 4},
  [2667] = {.lex_state = 106, .external_lex_state = 5},
  [2668] = {.lex_state = 106, .external_lex_state = 5},
  [2669] = {.lex_state = 106, .external_lex_state = 5},
  [2670] = {.lex_state = 106, .external_lex_state = 5},
  [2671] = {.lex_state = 271, .external_lex_state = 5},
  [2672] = {.lex_state = 271, .external_lex_state = 5},
  [2673] = {.lex_state = 271, .external_lex_state = 5},
  [2674] = {.lex_state = 106, .external_lex_state = 5},
  [2675] = {.lex_state = 275, .external_lex_state = 4},
  [2676] = {.lex_state = 106, .external_lex_state = 5},
  [2677] = {.lex_state = 106, .external_lex_state = 5},
  [2678] = {.lex_state = 271, .external_lex_state = 5},
  [2679] = {.lex_state = 106, .external_lex_state = 5},
  [2680] = {.lex_state = 275, .external_lex_state = 4},
  [2681] = {.lex_state = 275, .external_lex_state = 4},
  [2682] = {.lex_state = 275, .external_lex_state = 4},
  [2683] = {.lex_state = 271, .external_lex_state = 5},
  [2684] = {.lex_state = 106, .external_lex_state = 5},
  [2685] = {.lex_state = 271, .external_lex_state = 5},
  [2686] = {.lex_state = 275, .external_lex_state = 4},
  [2687] = {.lex_state = 275, .external_lex_state = 4},
  [2688] = {.lex_state = 271, .external_lex_state = 5},
  [2689] = {.lex_state = 271, .external_lex_state = 5},
  [2690] = {.lex_state = 271, .external_lex_state = 5},
  [2691] = {.lex_state = 275, .external_lex_state = 4},
  [2692] = {.lex_state = 271, .external_lex_state = 4},
  [2693] = {.lex_state = 271, .external_lex_state = 4},
  [2694] = {.lex_state = 271, .external_lex_state = 4},
  [2695] = {.lex_state = 271, .external_lex_state = 5},
  [2696] = {.lex_state = 275, .external_lex_state = 4},
  [2697] = {.lex_state = 275, .external_lex_state = 4},
  [2698] = {.lex_state = 273, .external_lex_state = 5},
  [2699] = {.lex_state = 273, .external_lex_state = 5},
  [2700] = {.lex_state = 275, .external_lex_state = 4},
  [2701] = {.lex_state = 271, .external_lex_state = 5},
  [2702] = {.lex_state = 106, .external_lex_state = 5},
  [2703] = {.lex_state = 106, .external_lex_state = 5},
  [2704] = {.lex_state = 271, .external_lex_state = 5},
  [2705] = {.lex_state = 271, .external_lex_state = 5},
  [2706] = {.lex_state = 271, .external_lex_state = 5},
  [2707] = {.lex_state = 271, .external_lex_state = 4},
  [2708] = {.lex_state = 106, .external_lex_state = 5},
  [2709] = {.lex_state = 271, .external_lex_state = 5},
  [2710] = {.lex_state = 106, .external_lex_state = 5},
  [2711] = {.lex_state = 271, .external_lex_state = 4},
  [2712] = {.lex_state = 271, .external_lex_state = 5},
  [2713] = {.lex_state = 271, .external_lex_state = 4},
  [2714] = {.lex_state = 271, .external_lex_state = 5},
  [2715] = {.lex_state = 271, .external_lex_state = 5},
  [2716] = {.lex_state = 271, .external_lex_state = 5},
  [2717] = {.lex_state = 106, .external_lex_state = 5},
  [2718] = {.lex_state = 106, .external_lex_state = 5},
  [2719] = {.lex_state = 271, .external_lex_state = 5},
  [2720] = {.lex_state = 271, .external_lex_state = 5},
  [2721] = {.lex_state = 275, .external_lex_state = 4},
  [2722] = {.lex_state = 275, .external_lex_state = 4},
  [2723] = {.lex_state = 273, .external_lex_state = 5},
  [2724] = {.lex_state = 271, .external_lex_state = 5},
  [2725] = {.lex_state = 273, .external_lex_state = 5},
  [2726] = {.lex_state = 273, .external_lex_state = 5},
  [2727] = {.lex_state = 273, .external_lex_state = 5},
  [2728] = {.lex_state = 273, .external_lex_state = 5},
  [2729] = {.lex_state = 273, .external_lex_state = 5},
  [2730] = {.lex_state = 273, .external_lex_state = 5},
  [2731] = {.lex_state = 273, .external_lex_state = 5},
  [2732] = {.lex_state = 273, .external_lex_state = 5},
  [2733] = {.lex_state = 273, .external_lex_state = 5},
  [2734] = {.lex_state = 273, .external_lex_state = 5},
  [2735] = {.lex_state = 273, .external_lex_state = 5},
  [2736] = {.lex_state = 273, .external_lex_state = 5},
  [2737] = {.lex_state = 273, .external_lex_state = 5},
  [2738] = {.lex_state = 273, .external_lex_state = 5},
  [2739] = {.lex_state = 273, .external_lex_state = 5},
  [2740] = {.lex_state = 273, .external_lex_state = 5},
  [2741] = {.lex_state = 273, .external_lex_state = 5},
  [2742] = {.lex_state = 273, .external_lex_state = 5},
  [2743] = {.lex_state = 273, .external_lex_state = 5},
  [2744] = {.lex_state = 273, .external_lex_state = 5},
  [2745] = {.lex_state = 273, .external_lex_state = 5},
  [2746] = {.lex_state = 271, .external_lex_state = 5},
  [2747] = {.lex_state = 273, .external_lex_state = 5},
  [2748] = {.lex_state = 273, .external_lex_state = 5},
  [2749] = {.lex_state = 273, .external_lex_state = 5},
  [2750] = {.lex_state = 271, .external_lex_state = 4},
  [2751] = {.lex_state = 273, .external_lex_state = 5},
  [2752] = {.lex_state = 273, .external_lex_state = 5},
  [2753] = {.lex_state = 273, .external_lex_state = 5},
  [2754] = {.lex_state = 273, .external_lex_state = 5},
  [2755] = {.lex_state = 271, .external_lex_state = 5},
  [2756] = {.lex_state = 273, .external_lex_state = 5},
  [2757] = {.lex_state = 273, .external_lex_state = 5},
  [2758] = {.lex_state = 273, .external_lex_state = 5},
  [2759] = {.lex_state = 271, .external_lex_state = 4},
  [2760] = {.lex_state = 273, .external_lex_state = 5},
  [2761] = {.lex_state = 271, .external_lex_state = 4},
  [2762] = {.lex_state = 275, .external_lex_state = 4},
  [2763] = {.lex_state = 273, .external_lex_state = 5},
  [2764] = {.lex_state = 275, .external_lex_state = 4},
  [2765] = {.lex_state = 273, .external_lex_state = 5},
  [2766] = {.lex_state = 275, .external_lex_state = 4},
  [2767] = {.lex_state = 275, .external_lex_state = 4},
  [2768] = {.lex_state = 273, .external_lex_state = 5},
  [2769] = {.lex_state = 275, .external_lex_state = 4},
  [2770] = {.lex_state = 275, .external_lex_state = 4},
  [2771] = {.lex_state = 275, .external_lex_state = 4},
  [2772] = {.lex_state = 273, .external_lex_state = 5},
  [2773] = {.lex_state = 275, .external_lex_state = 4},
  [2774] = {.lex_state = 275, .external_lex_state = 4},
  [2775] = {.lex_state = 275, .external_lex_state = 4},
  [2776] = {.lex_state = 275, .external_lex_state = 4},
  [2777] = {.lex_state = 275, .external_lex_state = 4},
  [2778] = {.lex_state = 275, .external_lex_state = 4},
  [2779] = {.lex_state = 275, .external_lex_state = 4},
  [2780] = {.lex_state = 275, .external_lex_state = 4},
  [2781] = {.lex_state = 275, .external_lex_state = 4},
  [2782] = {.lex_state = 273, .external_lex_state = 5},
  [2783] = {.lex_state = 275, .external_lex_state = 4},
  [2784] = {.lex_state = 275, .external_lex_state = 4},
  [2785] = {.lex_state = 275, .external_lex_state = 4},
  [2786] = {.lex_state = 275, .external_lex_state = 4},
  [2787] = {.lex_state = 275, .external_lex_state = 4},
  [2788] = {.lex_state = 275, .external_lex_state = 4},
  [2789] = {.lex_state = 273, .external_lex_state = 5},
  [2790] = {.lex_state = 273, .external_lex_state = 5},
  [2791] = {.lex_state = 273, .external_lex_state = 5},
  [2792] = {.lex_state = 273, .external_lex_state = 5},
  [2793] = {.lex_state = 273, .external_lex_state = 5},
  [2794] = {.lex_state = 273, .external_lex_state = 5},
  [2795] = {.lex_state = 273, .external_lex_state = 5},
  [2796] = {.lex_state = 273, .external_lex_state = 5},
  [2797] = {.lex_state = 273, .external_lex_state = 5},
  [2798] = {.lex_state = 273, .external_lex_state = 5},
  [2799] = {.lex_state = 273, .external_lex_state = 5},
  [2800] = {.lex_state = 273, .external_lex_state = 5},
  [2801] = {.lex_state = 273, .external_lex_state = 5},
  [2802] = {.lex_state = 273, .external_lex_state = 5},
  [2803] = {.lex_state = 271, .external_lex_state = 4},
  [2804] = {.lex_state = 273, .external_lex_state = 5},
  [2805] = {.lex_state = 273, .external_lex_state = 5},
  [2806] = {.lex_state = 271, .external_lex_state = 4},
  [2807] = {.lex_state = 273, .external_lex_state = 5},
  [2808] = {.lex_state = 271, .external_lex_state = 4},
  [2809] = {.lex_state = 273, .external_lex_state = 5},
  [2810] = {.lex_state = 273, .external_lex_state = 5},
  [2811] = {.lex_state = 273, .external_lex_state = 5},
  [2812] = {.lex_state = 273, .external_lex_state = 5},
  [2813] = {.lex_state = 273, .external_lex_state = 5},
  [2814] = {.lex_state = 271, .external_lex_state = 4},
  [2815] = {.lex_state = 271, .external_lex_state = 4},
  [2816] = {.lex_state = 273, .external_lex_state = 5},
  [2817] = {.lex_state = 273, .external_lex_state = 5},
  [2818] = {.lex_state = 275, .external_lex_state = 4},
  [2819] = {.lex_state = 273, .external_lex_state = 5},
  [2820] = {.lex_state = 275, .external_lex_state = 4},
  [2821] = {.lex_state = 275, .external_lex_state = 4},
  [2822] = {.lex_state = 273, .external_lex_state = 5},
  [2823] = {.lex_state = 275, .external_lex_state = 4},
  [2824] = {.lex_state = 273, .external_lex_state = 5},
  [2825] = {.lex_state = 275, .external_lex_state = 4},
  [2826] = {.lex_state = 273, .external_lex_state = 5},
  [2827] = {.lex_state = 275, .external_lex_state = 4},
  [2828] = {.lex_state = 275, .external_lex_state = 4},
  [2829] = {.lex_state = 273, .external_lex_state = 5},
  [2830] = {.lex_state = 275, .external_lex_state = 4},
  [2831] = {.lex_state = 275, .external_lex_state = 4},
  [2832] = {.lex_state = 275, .external_lex_state = 4},
  [2833] = {.lex_state = 275, .external_lex_state = 4},
  [2834] = {.lex_state = 275, .external_lex_state = 4},
  [2835] = {.lex_state = 275, .external_lex_state = 4},
  [2836] = {.lex_state = 275, .external_lex_state = 4},
  [2837] = {.lex_state = 275, .external_lex_state = 4},
  [2838] = {.lex_state = 273, .external_lex_state = 5},
  [2839] = {.lex_state = 273, .external_lex_state = 5},
  [2840] = {.lex_state = 275, .external_lex_state = 4},
  [2841] = {.lex_state = 273, .external_lex_state = 5},
  [2842] = {.lex_state = 273, .external_lex_state = 5},
  [2843] = {.lex_state = 271, .external_lex_state = 4},
  [2844] = {.lex_state = 273, .external_lex_state = 5},
  [2845] = {.lex_state = 273, .external_lex_state = 5},
  [2846] = {.lex_state = 271, .external_lex_state = 5},
  [2847] = {.lex_state = 271, .external_lex_state = 4},
  [2848] = {.lex_state = 106, .external_lex_state = 5},
  [2849] = {.lex_state = 106, .external_lex_state = 5},
  [2850] = {.lex_state = 273, .external_lex_state = 5},
  [2851] = {.lex_state = 271, .external_lex_state = 4},
  [2852] = {.lex_state = 273, .external_lex_state = 5},
  [2853] = {.lex_state = 125, .external_lex_state = 5},
  [2854] = {.lex_state = 275, .external_lex_state = 4},
  [2855] = {.lex_state = 125, .external_lex_state = 5},
  [2856] = {.lex_state = 275, .external_lex_state = 4},
  [2857] = {.lex_state = 275, .external_lex_state = 4},
  [2858] = {.lex_state = 125, .external_lex_state = 5},
  [2859] = {.lex_state = 125, .external_lex_state = 5},
  [2860] = {.lex_state = 125, .external_lex_state = 5},
  [2861] = {.lex_state = 275, .external_lex_state = 4},
  [2862] = {.lex_state = 275, .external_lex_state = 4},
  [2863] = {.lex_state = 125, .external_lex_state = 5},
  [2864] = {.lex_state = 125, .external_lex_state = 5},
  [2865] = {.lex_state = 125, .external_lex_state = 5},
  [2866] = {.lex_state = 125, .external_lex_state = 5},
  [2867] = {.lex_state = 125, .external_lex_state = 5},
  [2868] = {.lex_state = 125, .external_lex_state = 5},
  [2869] = {.lex_state = 125, .external_lex_state = 5},
  [2870] = {.lex_state = 125, .external_lex_state = 5},
  [2871] = {.lex_state = 275, .external_lex_state = 4},
  [2872] = {.lex_state = 125, .external_lex_state = 5},
  [2873] = {.lex_state = 125, .external_lex_state = 5},
  [2874] = {.lex_state = 125, .external_lex_state = 5},
  [2875] = {.lex_state = 125, .external_lex_state = 5},
  [2876] = {.lex_state = 125, .external_lex_state = 5},
  [2877] = {.lex_state = 125, .external_lex_state = 5},
  [2878] = {.lex_state = 271, .external_lex_state = 5},
  [2879] = {.lex_state = 271, .external_lex_state = 5},
  [2880] = {.lex_state = 125, .external_lex_state = 5},
  [2881] = {.lex_state = 273, .external_lex_state = 5},
  [2882] = {.lex_state = 273, .external_lex_state = 5},
  [2883] = {.lex_state = 275, .external_lex_state = 4},
  [2884] = {.lex_state = 273, .external_lex_state = 5},
  [2885] = {.lex_state = 273, .external_lex_state = 5},
  [2886] = {.lex_state = 273, .external_lex_state = 5},
  [2887] = {.lex_state = 273, .external_lex_state = 5},
  [2888] = {.lex_state = 275, .external_lex_state = 4},
  [2889] = {.lex_state = 275, .external_lex_state = 4},
  [2890] = {.lex_state = 130, .external_lex_state = 5},
  [2891] = {.lex_state = 130, .external_lex_state = 5},
  [2892] = {.lex_state = 130, .external_lex_state = 5},
  [2893] = {.lex_state = 130, .external_lex_state = 5},
  [2894] = {.lex_state = 130, .external_lex_state = 5},
  [2895] = {.lex_state = 130, .external_lex_state = 5},
  [2896] = {.lex_state = 275, .external_lex_state = 4},
  [2897] = {.lex_state = 130, .external_lex_state = 5},
  [2898] = {.lex_state = 275, .external_lex_state = 4},
  [2899] = {.lex_state = 130, .external_lex_state = 5},
  [2900] = {.lex_state = 275, .external_lex_state = 4},
  [2901] = {.lex_state = 273, .external_lex_state = 5},
  [2902] = {.lex_state = 275, .external_lex_state = 4},
  [2903] = {.lex_state = 273, .external_lex_state = 5},
  [2904] = {.lex_state = 275, .external_lex_state = 4},
  [2905] = {.lex_state = 273, .external_lex_state = 5},
  [2906] = {.lex_state = 273, .external_lex_state = 5},
  [2907] = {.lex_state = 273, .external_lex_state = 5},
  [2908] = {.lex_state = 275, .external_lex_state = 4},
  [2909] = {.lex_state = 273, .external_lex_state = 5},
  [2910] = {.lex_state = 275, .external_lex_state = 4},
  [2911] = {.lex_state = 125, .external_lex_state = 5},
  [2912] = {.lex_state = 275, .external_lex_state = 4},
  [2913] = {.lex_state = 125, .external_lex_state = 5},
  [2914] = {.lex_state = 125, .external_lex_state = 5},
  [2915] = {.lex_state = 125, .external_lex_state = 5},
  [2916] = {.lex_state = 125, .external_lex_state = 5},
  [2917] = {.lex_state = 271, .external_lex_state = 5},
  [2918] = {.lex_state = 271, .external_lex_state = 5},
  [2919] = {.lex_state = 271, .external_lex_state = 5},
  [2920] = {.lex_state = 125, .external_lex_state = 5},
  [2921] = {.lex_state = 130, .external_lex_state = 5},
  [2922] = {.lex_state = 130, .external_lex_state = 5},
  [2923] = {.lex_state = 130, .external_lex_state = 5},
  [2924] = {.lex_state = 130, .external_lex_state = 5},
  [2925] = {.lex_state = 130, .external_lex_state = 5},
  [2926] = {.lex_state = 106, .external_lex_state = 5},
  [2927] = {.lex_state = 136, .external_lex_state = 5},
  [2928] = {.lex_state = 271, .external_lex_state = 4},
  [2929] = {.lex_state = 271, .external_lex_state = 5},
  [2930] = {.lex_state = 271, .external_lex_state = 4},
  [2931] = {.lex_state = 271, .external_lex_state = 4},
  [2932] = {.lex_state = 271, .external_lex_state = 5},
  [2933] = {.lex_state = 271, .external_lex_state = 5},
  [2934] = {.lex_state = 271, .external_lex_state = 5},
  [2935] = {.lex_state = 271, .external_lex_state = 5},
  [2936] = {.lex_state = 271, .external_lex_state = 5},
  [2937] = {.lex_state = 271, .external_lex_state = 5},
  [2938] = {.lex_state = 271, .external_lex_state = 5},
  [2939] = {.lex_state = 271, .external_lex_state = 5},
  [2940] = {.lex_state = 271, .external_lex_state = 5},
  [2941] = {.lex_state = 271, .external_lex_state = 5},
  [2942] = {.lex_state = 271, .external_lex_state = 5},
  [2943] = {.lex_state = 271, .external_lex_state = 4},
  [2944] = {.lex_state = 106, .external_lex_state = 5},
  [2945] = {.lex_state = 106, .external_lex_state = 5},
  [2946] = {.lex_state = 271, .external_lex_state = 4},
  [2947] = {.lex_state = 271, .external_lex_state = 4},
  [2948] = {.lex_state = 271, .external_lex_state = 4},
  [2949] = {.lex_state = 271, .external_lex_state = 4},
  [2950] = {.lex_state = 271, .external_lex_state = 4},
  [2951] = {.lex_state = 271, .external_lex_state = 4},
  [2952] = {.lex_state = 271, .external_lex_state = 4},
  [2953] = {.lex_state = 271, .external_lex_state = 4},
  [2954] = {.lex_state = 271, .external_lex_state = 4},
  [2955] = {.lex_state = 271, .external_lex_state = 5},
  [2956] = {.lex_state = 271, .external_lex_state = 5},
  [2957] = {.lex_state = 271, .external_lex_state = 5},
  [2958] = {.lex_state = 271, .external_lex_state = 5},
  [2959] = {.lex_state = 271, .external_lex_state = 5},
  [2960] = {.lex_state = 271, .external_lex_state = 5},
  [2961] = {.lex_state = 106, .external_lex_state = 5},
  [2962] = {.lex_state = 106, .external_lex_state = 5},
  [2963] = {.lex_state = 106, .external_lex_state = 5},
  [2964] = {.lex_state = 271, .external_lex_state = 5},
  [2965] = {.lex_state = 106, .external_lex_state = 5},
  [2966] = {.lex_state = 106, .external_lex_state = 5},
  [2967] = {.lex_state = 106, .external_lex_state = 5},
  [2968] = {.lex_state = 106, .external_lex_state = 5},
  [2969] = {.lex_state = 106, .external_lex_state = 5},
  [2970] = {.lex_state = 271, .external_lex_state = 5},
  [2971] = {.lex_state = 106, .external_lex_state = 5},
  [2972] = {.lex_state = 106, .external_lex_state = 5},
  [2973] = {.lex_state = 106, .external_lex_state = 5},
  [2974] = {.lex_state = 106, .external_lex_state = 5},
  [2975] = {.lex_state = 106, .external_lex_state = 5},
  [2976] = {.lex_state = 106, .external_lex_state = 5},
  [2977] = {.lex_state = 106, .external_lex_state = 5},
  [2978] = {.lex_state = 106, .external_lex_state = 5},
  [2979] = {.lex_state = 106, .external_lex_state = 5},
  [2980] = {.lex_state = 271, .external_lex_state = 5},
  [2981] = {.lex_state = 271, .external_lex_state = 5},
  [2982] = {.lex_state = 271, .external_lex_state = 5},
  [2983] = {.lex_state = 271, .external_lex_state = 5},
  [2984] = {.lex_state = 271, .external_lex_state = 5},
  [2985] = {.lex_state = 271, .external_lex_state = 5},
  [2986] = {.lex_state = 271, .external_lex_state = 5},
  [2987] = {.lex_state = 271, .external_lex_state = 5},
  [2988] = {.lex_state = 271, .external_lex_state = 5},
  [2989] = {.lex_state = 271, .external_lex_state = 5},
  [2990] = {.lex_state = 271, .external_lex_state = 5},
  [2991] = {.lex_state = 271, .external_lex_state = 5},
  [2992] = {.lex_state = 271, .external_lex_state = 5},
  [2993] = {.lex_state = 271, .external_lex_state = 5},
  [2994] = {.lex_state = 271, .external_lex_state = 5},
  [2995] = {.lex_state = 271, .external_lex_state = 5},
  [2996] = {.lex_state = 106, .external_lex_state = 5},
  [2997] = {.lex_state = 106, .external_lex_state = 5},
  [2998] = {.lex_state = 271, .external_lex_state = 5},
  [2999] = {.lex_state = 271, .external_lex_state = 4},
  [3000] = {.lex_state = 271, .external_lex_state = 4},
  [3001] = {.lex_state = 271, .external_lex_state = 4},
  [3002] = {.lex_state = 271, .external_lex_state = 4},
  [3003] = {.lex_state = 271, .external_lex_state = 4},
  [3004] = {.lex_state = 271, .external_lex_state = 4},
  [3005] = {.lex_state = 271, .external_lex_state = 5},
  [3006] = {.lex_state = 271, .external_lex_state = 5},
  [3007] = {.lex_state = 271, .external_lex_state = 4},
  [3008] = {.lex_state = 271, .external_lex_state = 5},
  [3009] = {.lex_state = 271, .external_lex_state = 5},
  [3010] = {.lex_state = 271, .external_lex_state = 5},
  [3011] = {.lex_state = 271, .external_lex_state = 4},
  [3012] = {.lex_state = 271, .external_lex_state = 5},
  [3013] = {.lex_state = 271, .external_lex_state = 5},
  [3014] = {.lex_state = 271, .external_lex_state = 5},
  [3015] = {.lex_state = 106, .external_lex_state = 5},
  [3016] = {.lex_state = 271, .external_lex_state = 4},
  [3017] = {.lex_state = 271, .external_lex_state = 4},
  [3018] = {.lex_state = 271, .external_lex_state = 5},
  [3019] = {.lex_state = 271, .external_lex_state = 5},
  [3020] = {.lex_state = 271, .external_lex_state = 5},
  [3021] = {.lex_state = 271, .external_lex_state = 5},
  [3022] = {.lex_state = 271, .external_lex_state = 5},
  [3023] = {.lex_state = 271, .external_lex_state = 5},
  [3024] = {.lex_state = 271, .external_lex_state = 5},
  [3025] = {.lex_state = 271, .external_lex_state = 5},
  [3026] = {.lex_state = 271, .external_lex_state = 5},
  [3027] = {.lex_state = 271, .external_lex_state = 5},
  [3028] = {.lex_state = 271, .external_lex_state = 5},
  [3029] = {.lex_state = 271, .external_lex_state = 5},
  [3030] = {.lex_state = 271, .external_lex_state = 5},
  [3031] = {.lex_state = 271, .external_lex_state = 5},
  [3032] = {.lex_state = 271, .external_lex_state = 5},
  [3033] = {.lex_state = 271, .external_lex_state = 5},
  [3034] = {.lex_state = 271, .external_lex_state = 5},
  [3035] = {.lex_state = 271, .external_lex_state = 5},
  [3036] = {.lex_state = 271, .external_lex_state = 5},
  [3037] = {.lex_state = 271, .external_lex_state = 5},
  [3038] = {.lex_state = 271, .external_lex_state = 5},
  [3039] = {.lex_state = 271, .external_lex_state = 5},
  [3040] = {.lex_state = 271, .external_lex_state = 5},
  [3041] = {.lex_state = 271, .external_lex_state = 5},
  [3042] = {.lex_state = 271, .external_lex_state = 5},
  [3043] = {.lex_state = 271, .external_lex_state = 5},
  [3044] = {.lex_state = 271, .external_lex_state = 5},
  [3045] = {.lex_state = 271, .external_lex_state = 5},
  [3046] = {.lex_state = 271, .external_lex_state = 4},
  [3047] = {.lex_state = 271, .external_lex_state = 5},
  [3048] = {.lex_state = 271, .external_lex_state = 5},
  [3049] = {.lex_state = 271, .external_lex_state = 4},
  [3050] = {.lex_state = 271, .external_lex_state = 5},
  [3051] = {.lex_state = 271, .external_lex_state = 5},
  [3052] = {.lex_state = 271, .external_lex_state = 5},
  [3053] = {.lex_state = 271, .external_lex_state = 4},
  [3054] = {.lex_state = 271, .external_lex_state = 4},
  [3055] = {.lex_state = 271, .external_lex_state = 4},
  [3056] = {.lex_state = 271, .external_lex_state = 4},
  [3057] = {.lex_state = 271, .external_lex_state = 4},
  [3058] = {.lex_state = 271, .external_lex_state = 4},
  [3059] = {.lex_state = 271, .external_lex_state = 4},
  [3060] = {.lex_state = 271, .external_lex_state = 4},
  [3061] = {.lex_state = 139, .external_lex_state = 4},
  [3062] = {.lex_state = 139, .external_lex_state = 4},
  [3063] = {.lex_state = 271, .external_lex_state = 5},
  [3064] = {.lex_state = 271, .external_lex_state = 5},
  [3065] = {.lex_state = 271, .external_lex_state = 5},
  [3066] = {.lex_state = 271, .external_lex_state = 4},
  [3067] = {.lex_state = 271, .external_lex_state = 4},
  [3068] = {.lex_state = 271, .external_lex_state = 4},
  [3069] = {.lex_state = 271, .external_lex_state = 4},
  [3070] = {.lex_state = 139, .external_lex_state = 4},
  [3071] = {.lex_state = 139, .external_lex_state = 4},
  [3072] = {.lex_state = 271, .external_lex_state = 4},
  [3073] = {.lex_state = 271, .external_lex_state = 4},
  [3074] = {.lex_state = 271, .external_lex_state = 4},
  [3075] = {.lex_state = 271, .external_lex_state = 4},
  [3076] = {.lex_state = 271, .external_lex_state = 5},
  [3077] = {.lex_state = 271, .external_lex_state = 5},
  [3078] = {.lex_state = 271, .external_lex_state = 4},
  [3079] = {.lex_state = 271, .external_lex_state = 4},
  [3080] = {.lex_state = 271, .external_lex_state = 4},
  [3081] = {.lex_state = 271, .external_lex_state = 4},
  [3082] = {.lex_state = 139, .external_lex_state = 4},
  [3083] = {.lex_state = 139, .external_lex_state = 4},
  [3084] = {.lex_state = 271, .external_lex_state = 5},
  [3085] = {.lex_state = 271, .external_lex_state = 4},
  [3086] = {.lex_state = 271, .external_lex_state = 4},
  [3087] = {.lex_state = 271, .external_lex_state = 4},
  [3088] = {.lex_state = 271, .external_lex_state = 4},
  [3089] = {.lex_state = 271, .external_lex_state = 4},
  [3090] = {.lex_state = 271, .external_lex_state = 4},
  [3091] = {.lex_state = 271, .external_lex_state = 4},
  [3092] = {.lex_state = 271, .external_lex_state = 4},
  [3093] = {.lex_state = 271, .external_lex_state = 4},
  [3094] = {.lex_state = 271, .external_lex_state = 4},
  [3095] = {.lex_state = 271, .external_lex_state = 4},
  [3096] = {.lex_state = 271, .external_lex_state = 4},
  [3097] = {.lex_state = 271, .external_lex_state = 5},
  [3098] = {.lex_state = 271, .external_lex_state = 4},
  [3099] = {.lex_state = 271, .external_lex_state = 4},
  [3100] = {.lex_state = 271, .external_lex_state = 4},
  [3101] = {.lex_state = 271, .external_lex_state = 4},
  [3102] = {.lex_state = 271, .external_lex_state = 4},
  [3103] = {.lex_state = 271, .external_lex_state = 4},
  [3104] = {.lex_state = 271, .external_lex_state = 4},
  [3105] = {.lex_state = 271, .external_lex_state = 4},
  [3106] = {.lex_state = 271, .external_lex_state = 4},
  [3107] = {.lex_state = 271, .external_lex_state = 4},
  [3108] = {.lex_state = 271, .external_lex_state = 4},
  [3109] = {.lex_state = 139, .external_lex_state = 4},
  [3110] = {.lex_state = 139, .external_lex_state = 4},
  [3111] = {.lex_state = 139, .external_lex_state = 4},
  [3112] = {.lex_state = 139, .external_lex_state = 4},
  [3113] = {.lex_state = 271, .external_lex_state = 5},
  [3114] = {.lex_state = 271, .external_lex_state = 5},
  [3115] = {.lex_state = 271, .external_lex_state = 5},
  [3116] = {.lex_state = 271, .external_lex_state = 5},
  [3117] = {.lex_state = 271, .external_lex_state = 5},
  [3118] = {.lex_state = 271, .external_lex_state = 5},
  [3119] = {.lex_state = 271, .external_lex_state = 5},
  [3120] = {.lex_state = 271, .external_lex_state = 5},
  [3121] = {.lex_state = 271, .external_lex_state = 5},
  [3122] = {.lex_state = 271, .external_lex_state = 5},
  [3123] = {.lex_state = 271, .external_lex_state = 5},
  [3124] = {.lex_state = 271, .external_lex_state = 5},
  [3125] = {.lex_state = 271, .external_lex_state = 5},
  [3126] = {.lex_state = 271, .external_lex_state = 5},
  [3127] = {.lex_state = 271, .external_lex_state = 5},
  [3128] = {.lex_state = 271, .external_lex_state = 5},
  [3129] = {.lex_state = 271, .external_lex_state = 5},
  [3130] = {.lex_state = 139, .external_lex_state = 4},
  [3131] = {.lex_state = 139, .external_lex_state = 4},
  [3132] = {.lex_state = 271, .external_lex_state = 4},
  [3133] = {.lex_state = 271, .external_lex_state = 4},
  [3134] = {.lex_state = 271, .external_lex_state = 4},
  [3135] = {.lex_state = 133, .external_lex_state = 5},
  [3136] = {.lex_state = 271, .external_lex_state = 4},
  [3137] = {.lex_state = 271, .external_lex_state = 4},
  [3138] = {.lex_state = 271, .external_lex_state = 4},
  [3139] = {.lex_state = 271, .external_lex_state = 4},
  [3140] = {.lex_state = 271, .external_lex_state = 4},
  [3141] = {.lex_state = 271, .external_lex_state = 4},
  [3142] = {.lex_state = 271, .external_lex_state = 4},
  [3143] = {.lex_state = 271, .external_lex_state = 4},
  [3144] = {.lex_state = 271, .external_lex_state = 4},
  [3145] = {.lex_state = 271, .external_lex_state = 4},
  [3146] = {.lex_state = 271, .external_lex_state = 4},
  [3147] = {.lex_state = 106, .external_lex_state = 5},
  [3148] = {.lex_state = 271, .external_lex_state = 4},
  [3149] = {.lex_state = 271, .external_lex_state = 4},
  [3150] = {.lex_state = 271, .external_lex_state = 4},
  [3151] = {.lex_state = 271, .external_lex_state = 4},
  [3152] = {.lex_state = 271, .external_lex_state = 4},
  [3153] = {.lex_state = 271, .external_lex_state = 4},
  [3154] = {.lex_state = 271, .external_lex_state = 4},
  [3155] = {.lex_state = 133, .external_lex_state = 5},
  [3156] = {.lex_state = 271, .external_lex_state = 4},
  [3157] = {.lex_state = 271, .external_lex_state = 4},
  [3158] = {.lex_state = 139, .external_lex_state = 4},
  [3159] = {.lex_state = 139, .external_lex_state = 4},
  [3160] = {.lex_state = 139, .external_lex_state = 4},
  [3161] = {.lex_state = 271, .external_lex_state = 4},
  [3162] = {.lex_state = 139, .external_lex_state = 4},
  [3163] = {.lex_state = 139, .external_lex_state = 4},
  [3164] = {.lex_state = 271, .external_lex_state = 5},
  [3165] = {.lex_state = 271, .external_lex_state = 5},
  [3166] = {.lex_state = 271, .external_lex_state = 5},
  [3167] = {.lex_state = 271, .external_lex_state = 5},
  [3168] = {.lex_state = 271, .external_lex_state = 5},
  [3169] = {.lex_state = 271, .external_lex_state = 5},
  [3170] = {.lex_state = 271, .external_lex_state = 5},
  [3171] = {.lex_state = 271, .external_lex_state = 5},
  [3172] = {.lex_state = 271, .external_lex_state = 5},
  [3173] = {.lex_state = 271, .external_lex_state = 5},
  [3174] = {.lex_state = 271, .external_lex_state = 5},
  [3175] = {.lex_state = 271, .external_lex_state = 5},
  [3176] = {.lex_state = 271, .external_lex_state = 5},
  [3177] = {.lex_state = 271, .external_lex_state = 5},
  [3178] = {.lex_state = 271, .external_lex_state = 5},
  [3179] = {.lex_state = 271, .external_lex_state = 5},
  [3180] = {.lex_state = 271, .external_lex_state = 5},
  [3181] = {.lex_state = 139, .external_lex_state = 4},
  [3182] = {.lex_state = 139, .external_lex_state = 4},
  [3183] = {.lex_state = 139, .external_lex_state = 4},
  [3184] = {.lex_state = 139, .external_lex_state = 4},
  [3185] = {.lex_state = 139, .external_lex_state = 4},
  [3186] = {.lex_state = 139, .external_lex_state = 4},
  [3187] = {.lex_state = 139, .external_lex_state = 4},
  [3188] = {.lex_state = 139, .external_lex_state = 4},
  [3189] = {.lex_state = 139, .external_lex_state = 4},
  [3190] = {.lex_state = 139, .external_lex_state = 4},
  [3191] = {.lex_state = 139, .external_lex_state = 4},
  [3192] = {.lex_state = 139, .external_lex_state = 4},
  [3193] = {.lex_state = 139, .external_lex_state = 4},
  [3194] = {.lex_state = 139, .external_lex_state = 4},
  [3195] = {.lex_state = 139, .external_lex_state = 4},
  [3196] = {.lex_state = 133, .external_lex_state = 5},
  [3197] = {.lex_state = 133, .external_lex_state = 5},
  [3198] = {.lex_state = 271, .external_lex_state = 5},
  [3199] = {.lex_state = 106, .external_lex_state = 5},
  [3200] = {.lex_state = 133, .external_lex_state = 5},
  [3201] = {.lex_state = 133, .external_lex_state = 5},
  [3202] = {.lex_state = 133, .external_lex_state = 5},
  [3203] = {.lex_state = 133, .external_lex_state = 5},
  [3204] = {.lex_state = 133, .external_lex_state = 5},
  [3205] = {.lex_state = 133, .external_lex_state = 5},
  [3206] = {.lex_state = 271, .external_lex_state = 5},
  [3207] = {.lex_state = 271, .external_lex_state = 5},
  [3208] = {.lex_state = 133, .external_lex_state = 5},
  [3209] = {.lex_state = 133, .external_lex_state = 5},
  [3210] = {.lex_state = 271, .external_lex_state = 4},
  [3211] = {.lex_state = 271, .external_lex_state = 5},
  [3212] = {.lex_state = 271, .external_lex_state = 4},
  [3213] = {.lex_state = 133, .external_lex_state = 5},
  [3214] = {.lex_state = 106, .external_lex_state = 4},
  [3215] = {.lex_state = 271, .external_lex_state = 4},
  [3216] = {.lex_state = 139, .external_lex_state = 4},
  [3217] = {.lex_state = 106, .external_lex_state = 4},
  [3218] = {.lex_state = 106, .external_lex_state = 4},
  [3219] = {.lex_state = 139, .external_lex_state = 4},
  [3220] = {.lex_state = 106, .external_lex_state = 4},
  [3221] = {.lex_state = 139, .external_lex_state = 4},
  [3222] = {.lex_state = 139, .external_lex_state = 4},
  [3223] = {.lex_state = 139, .external_lex_state = 4},
  [3224] = {.lex_state = 139, .external_lex_state = 4},
  [3225] = {.lex_state = 133, .external_lex_state = 5},
  [3226] = {.lex_state = 106, .external_lex_state = 4},
  [3227] = {.lex_state = 271, .external_lex_state = 5},
  [3228] = {.lex_state = 139, .external_lex_state = 4},
  [3229] = {.lex_state = 139, .external_lex_state = 4},
  [3230] = {.lex_state = 139, .external_lex_state = 4},
  [3231] = {.lex_state = 139, .external_lex_state = 4},
  [3232] = {.lex_state = 139, .external_lex_state = 4},
  [3233] = {.lex_state = 139, .external_lex_state = 4},
  [3234] = {.lex_state = 139, .external_lex_state = 4},
  [3235] = {.lex_state = 139, .external_lex_state = 4},
  [3236] = {.lex_state = 139, .external_lex_state = 4},
  [3237] = {.lex_state = 139, .external_lex_state = 4},
  [3238] = {.lex_state = 106, .external_lex_state = 5},
  [3239] = {.lex_state = 133, .external_lex_state = 5},
  [3240] = {.lex_state = 106, .external_lex_state = 4},
  [3241] = {.lex_state = 136, .external_lex_state = 5},
  [3242] = {.lex_state = 136, .external_lex_state = 5},
  [3243] = {.lex_state = 136, .external_lex_state = 5},
  [3244] = {.lex_state = 136, .external_lex_state = 5},
  [3245] = {.lex_state = 136, .external_lex_state = 5},
  [3246] = {.lex_state = 136, .external_lex_state = 5},
  [3247] = {.lex_state = 133, .external_lex_state = 5},
  [3248] = {.lex_state = 133, .external_lex_state = 5},
  [3249] = {.lex_state = 136, .external_lex_state = 5},
  [3250] = {.lex_state = 136, .external_lex_state = 5},
  [3251] = {.lex_state = 136, .external_lex_state = 5},
  [3252] = {.lex_state = 136, .external_lex_state = 5},
  [3253] = {.lex_state = 136, .external_lex_state = 5},
  [3254] = {.lex_state = 136, .external_lex_state = 5},
  [3255] = {.lex_state = 139, .external_lex_state = 4},
  [3256] = {.lex_state = 106, .external_lex_state = 5},
  [3257] = {.lex_state = 133, .external_lex_state = 5},
  [3258] = {.lex_state = 133, .external_lex_state = 5},
  [3259] = {.lex_state = 133, .external_lex_state = 5},
  [3260] = {.lex_state = 133, .external_lex_state = 5},
  [3261] = {.lex_state = 133, .external_lex_state = 5},
  [3262] = {.lex_state = 133, .external_lex_state = 5},
  [3263] = {.lex_state = 271, .external_lex_state = 4},
  [3264] = {.lex_state = 271, .external_lex_state = 5},
  [3265] = {.lex_state = 271, .external_lex_state = 4},
  [3266] = {.lex_state = 136, .external_lex_state = 5},
  [3267] = {.lex_state = 136, .external_lex_state = 5},
  [3268] = {.lex_state = 136, .external_lex_state = 5},
  [3269] = {.lex_state = 271, .external_lex_state = 4},
  [3270] = {.lex_state = 136, .external_lex_state = 5},
  [3271] = {.lex_state = 136, .external_lex_state = 5},
  [3272] = {.lex_state = 136, .external_lex_state = 5},
  [3273] = {.lex_state = 136, .external_lex_state = 5},
  [3274] = {.lex_state = 136, .external_lex_state = 5},
  [3275] = {.lex_state = 136, .external_lex_state = 5},
  [3276] = {.lex_state = 136, .external_lex_state = 5},
  [3277] = {.lex_state = 136, .external_lex_state = 5},
  [3278] = {.lex_state = 136, .external_lex_state = 5},
  [3279] = {.lex_state = 136, .external_lex_state = 5},
  [3280] = {.lex_state = 136, .external_lex_state = 5},
  [3281] = {.lex_state = 136, .external_lex_state = 5},
  [3282] = {.lex_state = 136, .external_lex_state = 5},
  [3283] = {.lex_state = 136, .external_lex_state = 5},
  [3284] = {.lex_state = 136, .external_lex_state = 5},
  [3285] = {.lex_state = 136, .external_lex_state = 5},
  [3286] = {.lex_state = 136, .external_lex_state = 5},
  [3287] = {.lex_state = 133, .external_lex_state = 5},
  [3288] = {.lex_state = 133, .external_lex_state = 5},
  [3289] = {.lex_state = 271, .external_lex_state = 5},
  [3290] = {.lex_state = 133, .external_lex_state = 5},
  [3291] = {.lex_state = 271, .external_lex_state = 4},
  [3292] = {.lex_state = 271, .external_lex_state = 4},
  [3293] = {.lex_state = 136, .external_lex_state = 5},
  [3294] = {.lex_state = 136, .external_lex_state = 5},
  [3295] = {.lex_state = 136, .external_lex_state = 5},
  [3296] = {.lex_state = 136, .external_lex_state = 5},
  [3297] = {.lex_state = 136, .external_lex_state = 5},
  [3298] = {.lex_state = 136, .external_lex_state = 5},
  [3299] = {.lex_state = 271, .external_lex_state = 5},
  [3300] = {.lex_state = 271, .external_lex_state = 5},
  [3301] = {.lex_state = 136, .external_lex_state = 5},
  [3302] = {.lex_state = 136, .external_lex_state = 5},
  [3303] = {.lex_state = 136, .external_lex_state = 5},
  [3304] = {.lex_state = 136, .external_lex_state = 5},
  [3305] = {.lex_state = 136, .external_lex_state = 5},
  [3306] = {.lex_state = 136, .external_lex_state = 5},
  [3307] = {.lex_state = 136, .external_lex_state = 5},
  [3308] = {.lex_state = 136, .external_lex_state = 5},
  [3309] = {.lex_state = 136, .external_lex_state = 5},
  [3310] = {.lex_state = 136, .external_lex_state = 5},
  [3311] = {.lex_state = 133, .external_lex_state = 5},
  [3312] = {.lex_state = 133, .external_lex_state = 5},
  [3313] = {.lex_state = 133, .external_lex_state = 5},
  [3314] = {.lex_state = 136, .external_lex_state = 5},
  [3315] = {.lex_state = 136, .external_lex_state = 5},
  [3316] = {.lex_state = 136, .external_lex_state = 5},
  [3317] = {.lex_state = 136, .external_lex_state = 5},
  [3318] = {.lex_state = 136, .external_lex_state = 5},
  [3319] = {.lex_state = 136, .external_lex_state = 5},
  [3320] = {.lex_state = 136, .external_lex_state = 5},
  [3321] = {.lex_state = 136, .external_lex_state = 5},
  [3322] = {.lex_state = 136, .external_lex_state = 5},
  [3323] = {.lex_state = 136, .external_lex_state = 5},
  [3324] = {.lex_state = 136, .external_lex_state = 5},
  [3325] = {.lex_state = 136, .external_lex_state = 5},
  [3326] = {.lex_state = 136, .external_lex_state = 5},
  [3327] = {.lex_state = 136, .external_lex_state = 5},
  [3328] = {.lex_state = 136, .external_lex_state = 5},
  [3329] = {.lex_state = 133, .external_lex_state = 5},
  [3330] = {.lex_state = 136, .external_lex_state = 5},
  [3331] = {.lex_state = 136, .external_lex_state = 5},
  [3332] = {.lex_state = 136, .external_lex_state = 5},
  [3333] = {.lex_state = 136, .external_lex_state = 5},
  [3334] = {.lex_state = 136, .external_lex_state = 5},
  [3335] = {.lex_state = 136, .external_lex_state = 5},
  [3336] = {.lex_state = 136, .external_lex_state = 5},
  [3337] = {.lex_state = 136, .external_lex_state = 5},
  [3338] = {.lex_state = 136, .external_lex_state = 5},
  [3339] = {.lex_state = 136, .external_lex_state = 5},
  [3340] = {.lex_state = 136, .external_lex_state = 5},
  [3341] = {.lex_state = 136, .external_lex_state = 5},
  [3342] = {.lex_state = 136, .external_lex_state = 5},
  [3343] = {.lex_state = 136, .external_lex_state = 5},
  [3344] = {.lex_state = 136, .external_lex_state = 5},
  [3345] = {.lex_state = 136, .external_lex_state = 5},
  [3346] = {.lex_state = 133, .external_lex_state = 5},
  [3347] = {.lex_state = 133, .external_lex_state = 5},
  [3348] = {.lex_state = 133, .external_lex_state = 5},
  [3349] = {.lex_state = 106, .external_lex_state = 5},
  [3350] = {.lex_state = 106, .external_lex_state = 5},
  [3351] = {.lex_state = 106, .external_lex_state = 5},
  [3352] = {.lex_state = 106, .external_lex_state = 5},
  [3353] = {.lex_state = 106, .external_lex_state = 5},
  [3354] = {.lex_state = 106, .external_lex_state = 5},
  [3355] = {.lex_state = 106, .external_lex_state = 5},
  [3356] = {.lex_state = 106, .external_lex_state = 5},
  [3357] = {.lex_state = 106, .external_lex_state = 5},
  [3358] = {.lex_state = 106, .external_lex_state = 5},
  [3359] = {.lex_state = 106, .external_lex_state = 5},
  [3360] = {.lex_state = 106, .external_lex_state = 5},
  [3361] = {.lex_state = 106, .external_lex_state = 5},
  [3362] = {.lex_state = 106, .external_lex_state = 5},
  [3363] = {.lex_state = 106, .external_lex_state = 5},
  [3364] = {.lex_state = 106, .external_lex_state = 5},
  [3365] = {.lex_state = 106, .external_lex_state = 5},
  [3366] = {.lex_state = 133, .external_lex_state = 5},
  [3367] = {.lex_state = 133, .external_lex_state = 5},
  [3368] = {.lex_state = 106, .external_lex_state = 5},
  [3369] = {.lex_state = 106, .external_lex_state = 5},
  [3370] = {.lex_state = 106, .external_lex_state = 5},
  [3371] = {.lex_state = 106, .external_lex_state = 5},
  [3372] = {.lex_state = 106, .external_lex_state = 5},
  [3373] = {.lex_state = 106, .external_lex_state = 5},
  [3374] = {.lex_state = 106, .external_lex_state = 5},
  [3375] = {.lex_state = 106, .external_lex_state = 5},
  [3376] = {.lex_state = 106, .external_lex_state = 5},
  [3377] = {.lex_state = 106, .external_lex_state = 5},
  [3378] = {.lex_state = 271, .external_lex_state = 5},
  [3379] = {.lex_state = 271, .external_lex_state = 5},
  [3380] = {.lex_state = 271, .external_lex_state = 5},
  [3381] = {.lex_state = 106, .external_lex_state = 5},
  [3382] = {.lex_state = 139, .external_lex_state = 4},
  [3383] = {.lex_state = 139, .external_lex_state = 4},
  [3384] = {.lex_state = 106, .external_lex_state = 5},
  [3385] = {.lex_state = 106, .external_lex_state = 5},
  [3386] = {.lex_state = 106, .external_lex_state = 5},
  [3387] = {.lex_state = 106, .external_lex_state = 5},
  [3388] = {.lex_state = 106, .external_lex_state = 5},
  [3389] = {.lex_state = 133, .external_lex_state = 5},
  [3390] = {.lex_state = 133, .external_lex_state = 5},
  [3391] = {.lex_state = 106, .external_lex_state = 5},
  [3392] = {.lex_state = 106, .external_lex_state = 5},
  [3393] = {.lex_state = 106, .external_lex_state = 5},
  [3394] = {.lex_state = 106, .external_lex_state = 5},
  [3395] = {.lex_state = 106, .external_lex_state = 5},
  [3396] = {.lex_state = 106, .external_lex_state = 5},
  [3397] = {.lex_state = 106, .external_lex_state = 5},
  [3398] = {.lex_state = 106, .external_lex_state = 5},
  [3399] = {.lex_state = 106, .external_lex_state = 5},
  [3400] = {.lex_state = 139, .external_lex_state = 4},
  [3401] = {.lex_state = 139, .external_lex_state = 4},
  [3402] = {.lex_state = 106, .external_lex_state = 5},
  [3403] = {.lex_state = 106, .external_lex_state = 5},
  [3404] = {.lex_state = 106, .external_lex_state = 5},
  [3405] = {.lex_state = 106, .external_lex_state = 5},
  [3406] = {.lex_state = 106, .external_lex_state = 5},
  [3407] = {.lex_state = 106, .external_lex_state = 5},
  [3408] = {.lex_state = 106, .external_lex_state = 5},
  [3409] = {.lex_state = 106, .external_lex_state = 5},
  [3410] = {.lex_state = 133, .external_lex_state = 5},
  [3411] = {.lex_state = 133, .external_lex_state = 5},
  [3412] = {.lex_state = 106, .external_lex_state = 5},
  [3413] = {.lex_state = 106, .external_lex_state = 5},
  [3414] = {.lex_state = 106, .external_lex_state = 5},
  [3415] = {.lex_state = 106, .external_lex_state = 5},
  [3416] = {.lex_state = 106, .external_lex_state = 5},
  [3417] = {.lex_state = 106, .external_lex_state = 5},
  [3418] = {.lex_state = 106, .external_lex_state = 5},
  [3419] = {.lex_state = 106, .external_lex_state = 5},
  [3420] = {.lex_state = 106, .external_lex_state = 5},
  [3421] = {.lex_state = 106, .external_lex_state = 5},
  [3422] = {.lex_state = 106, .external_lex_state = 5},
  [3423] = {.lex_state = 106, .external_lex_state = 5},
  [3424] = {.lex_state = 106, .external_lex_state = 5},
  [3425] = {.lex_state = 106, .external_lex_state = 5},
  [3426] = {.lex_state = 106, .external_lex_state = 5},
  [3427] = {.lex_state = 106, .external_lex_state = 5},
  [3428] = {.lex_state = 106, .external_lex_state = 5},
  [3429] = {.lex_state = 106, .external_lex_state = 5},
  [3430] = {.lex_state = 106, .external_lex_state = 5},
  [3431] = {.lex_state = 106, .external_lex_state = 5},
  [3432] = {.lex_state = 106, .external_lex_state = 5},
  [3433] = {.lex_state = 106, .external_lex_state = 5},
  [3434] = {.lex_state = 106, .external_lex_state = 5},
  [3435] = {.lex_state = 106, .external_lex_state = 5},
  [3436] = {.lex_state = 106, .external_lex_state = 5},
  [3437] = {.lex_state = 271, .external_lex_state = 5},
  [3438] = {.lex_state = 271, .external_lex_state = 5},
  [3439] = {.lex_state = 106, .external_lex_state = 5},
  [3440] = {.lex_state = 271, .external_lex_state = 5},
  [3441] = {.lex_state = 106, .external_lex_state = 5},
  [3442] = {.lex_state = 106, .external_lex_state = 5},
  [3443] = {.lex_state = 271, .external_lex_state = 5},
  [3444] = {.lex_state = 106, .external_lex_state = 5},
  [3445] = {.lex_state = 106, .external_lex_state = 5},
  [3446] = {.lex_state = 106, .external_lex_state = 5},
  [3447] = {.lex_state = 106, .external_lex_state = 5},
  [3448] = {.lex_state = 106, .external_lex_state = 5},
  [3449] = {.lex_state = 271, .external_lex_state = 5},
  [3450] = {.lex_state = 106, .external_lex_state = 5},
  [3451] = {.lex_state = 271, .external_lex_state = 5},
  [3452] = {.lex_state = 139, .external_lex_state = 4},
  [3453] = {.lex_state = 271, .external_lex_state = 5},
  [3454] = {.lex_state = 271, .external_lex_state = 5},
  [3455] = {.lex_state = 139, .external_lex_state = 4},
  [3456] = {.lex_state = 139, .external_lex_state = 4},
  [3457] = {.lex_state = 139, .external_lex_state = 4},
  [3458] = {.lex_state = 271, .external_lex_state = 5},
  [3459] = {.lex_state = 139, .external_lex_state = 4},
  [3460] = {.lex_state = 271, .external_lex_state = 5},
  [3461] = {.lex_state = 271, .external_lex_state = 5},
  [3462] = {.lex_state = 139, .external_lex_state = 4},
  [3463] = {.lex_state = 271, .external_lex_state = 5},
  [3464] = {.lex_state = 139, .external_lex_state = 4},
  [3465] = {.lex_state = 139, .external_lex_state = 4},
  [3466] = {.lex_state = 271, .external_lex_state = 5},
  [3467] = {.lex_state = 271, .external_lex_state = 5},
  [3468] = {.lex_state = 139, .external_lex_state = 4},
  [3469] = {.lex_state = 271, .external_lex_state = 5},
  [3470] = {.lex_state = 271, .external_lex_state = 5},
  [3471] = {.lex_state = 271, .external_lex_state = 5},
  [3472] = {.lex_state = 139, .external_lex_state = 4},
  [3473] = {.lex_state = 139, .external_lex_state = 4},
  [3474] = {.lex_state = 139, .external_lex_state = 4},
  [3475] = {.lex_state = 271, .external_lex_state = 5},
  [3476] = {.lex_state = 271, .external_lex_state = 5},
  [3477] = {.lex_state = 271, .external_lex_state = 5},
  [3478] = {.lex_state = 271, .external_lex_state = 5},
  [3479] = {.lex_state = 271, .external_lex_state = 5},
  [3480] = {.lex_state = 271, .external_lex_state = 5},
  [3481] = {.lex_state = 271, .external_lex_state = 5},
  [3482] = {.lex_state = 271, .external_lex_state = 5},
  [3483] = {.lex_state = 271, .external_lex_state = 5},
  [3484] = {.lex_state = 271, .external_lex_state = 5},
  [3485] = {.lex_state = 271, .external_lex_state = 5},
  [3486] = {.lex_state = 139, .external_lex_state = 4},
  [3487] = {.lex_state = 106, .external_lex_state = 5},
  [3488] = {.lex_state = 106, .external_lex_state = 5},
  [3489] = {.lex_state = 139, .external_lex_state = 4},
  [3490] = {.lex_state = 139, .external_lex_state = 4},
  [3491] = {.lex_state = 139, .external_lex_state = 4},
  [3492] = {.lex_state = 139, .external_lex_state = 4},
  [3493] = {.lex_state = 271, .external_lex_state = 5},
  [3494] = {.lex_state = 139, .external_lex_state = 4},
  [3495] = {.lex_state = 271, .external_lex_state = 5},
  [3496] = {.lex_state = 139, .external_lex_state = 4},
  [3497] = {.lex_state = 139, .external_lex_state = 4},
  [3498] = {.lex_state = 139, .external_lex_state = 4},
  [3499] = {.lex_state = 271, .external_lex_state = 5},
  [3500] = {.lex_state = 271, .external_lex_state = 5},
  [3501] = {.lex_state = 271, .external_lex_state = 5},
  [3502] = {.lex_state = 139, .external_lex_state = 4},
  [3503] = {.lex_state = 139, .external_lex_state = 4},
  [3504] = {.lex_state = 139, .external_lex_state = 4},
  [3505] = {.lex_state = 139, .external_lex_state = 4},
  [3506] = {.lex_state = 139, .external_lex_state = 4},
  [3507] = {.lex_state = 139, .external_lex_state = 4},
  [3508] = {.lex_state = 271, .external_lex_state = 4},
  [3509] = {.lex_state = 142, .external_lex_state = 4},
  [3510] = {.lex_state = 275, .external_lex_state = 4},
  [3511] = {.lex_state = 275, .external_lex_state = 4},
  [3512] = {.lex_state = 275, .external_lex_state = 4},
  [3513] = {.lex_state = 275, .external_lex_state = 4},
  [3514] = {.lex_state = 275, .external_lex_state = 4},
  [3515] = {.lex_state = 275, .external_lex_state = 4},
  [3516] = {.lex_state = 133, .external_lex_state = 5},
  [3517] = {.lex_state = 275, .external_lex_state = 4},
  [3518] = {.lex_state = 106, .external_lex_state = 4},
  [3519] = {.lex_state = 275, .external_lex_state = 4},
  [3520] = {.lex_state = 271, .external_lex_state = 4},
  [3521] = {.lex_state = 271, .external_lex_state = 4},
  [3522] = {.lex_state = 139, .external_lex_state = 4},
  [3523] = {.lex_state = 139, .external_lex_state = 4},
  [3524] = {.lex_state = 139, .external_lex_state = 4},
  [3525] = {.lex_state = 139, .external_lex_state = 4},
  [3526] = {.lex_state = 139, .external_lex_state = 4},
  [3527] = {.lex_state = 139, .external_lex_state = 4},
  [3528] = {.lex_state = 133, .external_lex_state = 5},
  [3529] = {.lex_state = 133, .external_lex_state = 5},
  [3530] = {.lex_state = 271, .external_lex_state = 4},
  [3531] = {.lex_state = 271, .external_lex_state = 4},
  [3532] = {.lex_state = 106, .external_lex_state = 4},
  [3533] = {.lex_state = 275, .external_lex_state = 4},
  [3534] = {.lex_state = 271, .external_lex_state = 4},
  [3535] = {.lex_state = 106, .external_lex_state = 4},
  [3536] = {.lex_state = 106, .external_lex_state = 4},
  [3537] = {.lex_state = 271, .external_lex_state = 4},
  [3538] = {.lex_state = 133, .external_lex_state = 5},
  [3539] = {.lex_state = 275, .external_lex_state = 4},
  [3540] = {.lex_state = 275, .external_lex_state = 4},
  [3541] = {.lex_state = 275, .external_lex_state = 4},
  [3542] = {.lex_state = 271, .external_lex_state = 4},
  [3543] = {.lex_state = 271, .external_lex_state = 4},
  [3544] = {.lex_state = 133, .external_lex_state = 5},
  [3545] = {.lex_state = 271, .external_lex_state = 4},
  [3546] = {.lex_state = 275, .external_lex_state = 4},
  [3547] = {.lex_state = 271, .external_lex_state = 4},
  [3548] = {.lex_state = 271, .external_lex_state = 4},
  [3549] = {.lex_state = 106, .external_lex_state = 4},
  [3550] = {.lex_state = 275, .external_lex_state = 4},
  [3551] = {.lex_state = 106, .external_lex_state = 4},
  [3552] = {.lex_state = 106, .external_lex_state = 4},
  [3553] = {.lex_state = 139, .external_lex_state = 4},
  [3554] = {.lex_state = 139, .external_lex_state = 4},
  [3555] = {.lex_state = 139, .external_lex_state = 4},
  [3556] = {.lex_state = 139, .external_lex_state = 4},
  [3557] = {.lex_state = 139, .external_lex_state = 4},
  [3558] = {.lex_state = 139, .external_lex_state = 4},
  [3559] = {.lex_state = 139, .external_lex_state = 4},
  [3560] = {.lex_state = 139, .external_lex_state = 4},
  [3561] = {.lex_state = 139, .external_lex_state = 4},
  [3562] = {.lex_state = 139, .external_lex_state = 4},
  [3563] = {.lex_state = 139, .external_lex_state = 4},
  [3564] = {.lex_state = 139, .external_lex_state = 4},
  [3565] = {.lex_state = 139, .external_lex_state = 4},
  [3566] = {.lex_state = 139, .external_lex_state = 4},
  [3567] = {.lex_state = 139, .external_lex_state = 4},
  [3568] = {.lex_state = 139, .external_lex_state = 4},
  [3569] = {.lex_state = 139, .external_lex_state = 4},
  [3570] = {.lex_state = 139, .external_lex_state = 4},
  [3571] = {.lex_state = 139, .external_lex_state = 4},
  [3572] = {.lex_state = 139, .external_lex_state = 4},
  [3573] = {.lex_state = 275, .external_lex_state = 4},
  [3574] = {.lex_state = 275, .external_lex_state = 4},
  [3575] = {.lex_state = 275, .external_lex_state = 4},
  [3576] = {.lex_state = 275, .external_lex_state = 4},
  [3577] = {.lex_state = 275, .external_lex_state = 4},
  [3578] = {.lex_state = 275, .external_lex_state = 4},
  [3579] = {.lex_state = 275, .external_lex_state = 4},
  [3580] = {.lex_state = 275, .external_lex_state = 4},
  [3581] = {.lex_state = 275, .external_lex_state = 4},
  [3582] = {.lex_state = 275, .external_lex_state = 4},
  [3583] = {.lex_state = 275, .external_lex_state = 4},
  [3584] = {.lex_state = 275, .external_lex_state = 4},
  [3585] = {.lex_state = 275, .external_lex_state = 4},
  [3586] = {.lex_state = 275, .external_lex_state = 4},
  [3587] = {.lex_state = 139, .external_lex_state = 4},
  [3588] = {.lex_state = 139, .external_lex_state = 4},
  [3589] = {.lex_state = 271, .external_lex_state = 4},
  [3590] = {.lex_state = 275, .external_lex_state = 4},
  [3591] = {.lex_state = 139, .external_lex_state = 4},
  [3592] = {.lex_state = 139, .external_lex_state = 4},
  [3593] = {.lex_state = 275, .external_lex_state = 4},
  [3594] = {.lex_state = 275, .external_lex_state = 4},
  [3595] = {.lex_state = 139, .external_lex_state = 4},
  [3596] = {.lex_state = 139, .external_lex_state = 4},
  [3597] = {.lex_state = 275, .external_lex_state = 4},
  [3598] = {.lex_state = 106, .external_lex_state = 4},
  [3599] = {.lex_state = 142, .external_lex_state = 4},
  [3600] = {.lex_state = 142, .external_lex_state = 4},
  [3601] = {.lex_state = 271, .external_lex_state = 4},
  [3602] = {.lex_state = 275, .external_lex_state = 4},
  [3603] = {.lex_state = 142, .external_lex_state = 4},
  [3604] = {.lex_state = 142, .external_lex_state = 4},
  [3605] = {.lex_state = 106, .external_lex_state = 4},
  [3606] = {.lex_state = 142, .external_lex_state = 4},
  [3607] = {.lex_state = 142, .external_lex_state = 4},
  [3608] = {.lex_state = 275, .external_lex_state = 4},
  [3609] = {.lex_state = 275, .external_lex_state = 4},
  [3610] = {.lex_state = 275, .external_lex_state = 4},
  [3611] = {.lex_state = 275, .external_lex_state = 4},
  [3612] = {.lex_state = 275, .external_lex_state = 4},
  [3613] = {.lex_state = 275, .external_lex_state = 4},
  [3614] = {.lex_state = 275, .external_lex_state = 4},
  [3615] = {.lex_state = 106, .external_lex_state = 4},
  [3616] = {.lex_state = 133, .external_lex_state = 5},
  [3617] = {.lex_state = 275, .external_lex_state = 4},
  [3618] = {.lex_state = 106, .external_lex_state = 4},
  [3619] = {.lex_state = 106, .external_lex_state = 4},
  [3620] = {.lex_state = 275, .external_lex_state = 4},
  [3621] = {.lex_state = 275, .external_lex_state = 4},
  [3622] = {.lex_state = 139, .external_lex_state = 4},
  [3623] = {.lex_state = 139, .external_lex_state = 4},
  [3624] = {.lex_state = 139, .external_lex_state = 4},
  [3625] = {.lex_state = 139, .external_lex_state = 4},
  [3626] = {.lex_state = 139, .external_lex_state = 4},
  [3627] = {.lex_state = 139, .external_lex_state = 4},
  [3628] = {.lex_state = 275, .external_lex_state = 4},
  [3629] = {.lex_state = 275, .external_lex_state = 4},
  [3630] = {.lex_state = 275, .external_lex_state = 4},
  [3631] = {.lex_state = 275, .external_lex_state = 4},
  [3632] = {.lex_state = 139, .external_lex_state = 4},
  [3633] = {.lex_state = 275, .external_lex_state = 4},
  [3634] = {.lex_state = 275, .external_lex_state = 4},
  [3635] = {.lex_state = 139, .external_lex_state = 4},
  [3636] = {.lex_state = 139, .external_lex_state = 4},
  [3637] = {.lex_state = 142, .external_lex_state = 4},
  [3638] = {.lex_state = 142, .external_lex_state = 4},
  [3639] = {.lex_state = 139, .external_lex_state = 4},
  [3640] = {.lex_state = 139, .external_lex_state = 4},
  [3641] = {.lex_state = 142, .external_lex_state = 4},
  [3642] = {.lex_state = 142, .external_lex_state = 4},
  [3643] = {.lex_state = 139, .external_lex_state = 4},
  [3644] = {.lex_state = 139, .external_lex_state = 4},
  [3645] = {.lex_state = 142, .external_lex_state = 4},
  [3646] = {.lex_state = 142, .external_lex_state = 4},
  [3647] = {.lex_state = 139, .external_lex_state = 4},
  [3648] = {.lex_state = 139, .external_lex_state = 4},
  [3649] = {.lex_state = 139, .external_lex_state = 4},
  [3650] = {.lex_state = 133, .external_lex_state = 5},
  [3651] = {.lex_state = 133, .external_lex_state = 5},
  [3652] = {.lex_state = 133, .external_lex_state = 5},
  [3653] = {.lex_state = 133, .external_lex_state = 5},
  [3654] = {.lex_state = 133, .external_lex_state = 5},
  [3655] = {.lex_state = 133, .external_lex_state = 5},
  [3656] = {.lex_state = 133, .external_lex_state = 5},
  [3657] = {.lex_state = 133, .external_lex_state = 5},
  [3658] = {.lex_state = 133, .external_lex_state = 5},
  [3659] = {.lex_state = 133, .external_lex_state = 5},
  [3660] = {.lex_state = 133, .external_lex_state = 5},
  [3661] = {.lex_state = 133, .external_lex_state = 5},
  [3662] = {.lex_state = 133, .external_lex_state = 5},
  [3663] = {.lex_state = 133, .external_lex_state = 5},
  [3664] = {.lex_state = 133, .external_lex_state = 5},
  [3665] = {.lex_state = 133, .external_lex_state = 5},
  [3666] = {.lex_state = 271, .external_lex_state = 4},
  [3667] = {.lex_state = 271, .external_lex_state = 4},
  [3668] = {.lex_state = 275, .external_lex_state = 4},
  [3669] = {.lex_state = 275, .external_lex_state = 4},
  [3670] = {.lex_state = 275, .external_lex_state = 4},
  [3671] = {.lex_state = 275, .external_lex_state = 4},
  [3672] = {.lex_state = 275, .external_lex_state = 4},
  [3673] = {.lex_state = 133, .external_lex_state = 5},
  [3674] = {.lex_state = 133, .external_lex_state = 5},
  [3675] = {.lex_state = 133, .external_lex_state = 5},
  [3676] = {.lex_state = 133, .external_lex_state = 5},
  [3677] = {.lex_state = 133, .external_lex_state = 5},
  [3678] = {.lex_state = 133, .external_lex_state = 5},
  [3679] = {.lex_state = 133, .external_lex_state = 5},
  [3680] = {.lex_state = 133, .external_lex_state = 5},
  [3681] = {.lex_state = 133, .external_lex_state = 5},
  [3682] = {.lex_state = 106, .external_lex_state = 4},
  [3683] = {.lex_state = 275, .external_lex_state = 4},
  [3684] = {.lex_state = 275, .external_lex_state = 4},
  [3685] = {.lex_state = 106, .external_lex_state = 4},
  [3686] = {.lex_state = 275, .external_lex_state = 4},
  [3687] = {.lex_state = 106, .external_lex_state = 4},
  [3688] = {.lex_state = 139, .external_lex_state = 4},
  [3689] = {.lex_state = 139, .external_lex_state = 4},
  [3690] = {.lex_state = 139, .external_lex_state = 4},
  [3691] = {.lex_state = 139, .external_lex_state = 4},
  [3692] = {.lex_state = 139, .external_lex_state = 4},
  [3693] = {.lex_state = 275, .external_lex_state = 4},
  [3694] = {.lex_state = 133, .external_lex_state = 5},
  [3695] = {.lex_state = 275, .external_lex_state = 4},
  [3696] = {.lex_state = 275, .external_lex_state = 4},
  [3697] = {.lex_state = 139, .external_lex_state = 4},
  [3698] = {.lex_state = 275, .external_lex_state = 4},
  [3699] = {.lex_state = 133, .external_lex_state = 5},
  [3700] = {.lex_state = 275, .external_lex_state = 4},
  [3701] = {.lex_state = 275, .external_lex_state = 4},
  [3702] = {.lex_state = 133, .external_lex_state = 5},
  [3703] = {.lex_state = 133, .external_lex_state = 5},
  [3704] = {.lex_state = 133, .external_lex_state = 5},
  [3705] = {.lex_state = 133, .external_lex_state = 5},
  [3706] = {.lex_state = 133, .external_lex_state = 5},
  [3707] = {.lex_state = 275, .external_lex_state = 4},
  [3708] = {.lex_state = 275, .external_lex_state = 4},
  [3709] = {.lex_state = 275, .external_lex_state = 4},
  [3710] = {.lex_state = 275, .external_lex_state = 4},
  [3711] = {.lex_state = 275, .external_lex_state = 4},
  [3712] = {.lex_state = 275, .external_lex_state = 4},
  [3713] = {.lex_state = 133, .external_lex_state = 5},
  [3714] = {.lex_state = 133, .external_lex_state = 5},
  [3715] = {.lex_state = 275, .external_lex_state = 4},
  [3716] = {.lex_state = 133, .external_lex_state = 5},
  [3717] = {.lex_state = 133, .external_lex_state = 5},
  [3718] = {.lex_state = 133, .external_lex_state = 5},
  [3719] = {.lex_state = 133, .external_lex_state = 5},
  [3720] = {.lex_state = 133, .external_lex_state = 5},
  [3721] = {.lex_state = 133, .external_lex_state = 5},
  [3722] = {.lex_state = 133, .external_lex_state = 5},
  [3723] = {.lex_state = 133, .external_lex_state = 5},
  [3724] = {.lex_state = 133, .external_lex_state = 5},
  [3725] = {.lex_state = 133, .external_lex_state = 5},
  [3726] = {.lex_state = 133, .external_lex_state = 5},
  [3727] = {.lex_state = 133, .external_lex_state = 5},
  [3728] = {.lex_state = 133, .external_lex_state = 5},
  [3729] = {.lex_state = 133, .external_lex_state = 5},
  [3730] = {.lex_state = 133, .external_lex_state = 5},
  [3731] = {.lex_state = 133, .external_lex_state = 5},
  [3732] = {.lex_state = 133, .external_lex_state = 5},
  [3733] = {.lex_state = 133, .external_lex_state = 5},
  [3734] = {.lex_state = 133, .external_lex_state = 5},
  [3735] = {.lex_state = 133, .external_lex_state = 5},
  [3736] = {.lex_state = 133, .external_lex_state = 5},
  [3737] = {.lex_state = 133, .external_lex_state = 5},
  [3738] = {.lex_state = 133, .external_lex_state = 5},
  [3739] = {.lex_state = 133, .external_lex_state = 5},
  [3740] = {.lex_state = 133, .external_lex_state = 5},
  [3741] = {.lex_state = 133, .external_lex_state = 5},
  [3742] = {.lex_state = 106, .external_lex_state = 4},
  [3743] = {.lex_state = 133, .external_lex_state = 5},
  [3744] = {.lex_state = 133, .external_lex_state = 5},
  [3745] = {.lex_state = 133, .external_lex_state = 5},
  [3746] = {.lex_state = 133, .external_lex_state = 5},
  [3747] = {.lex_state = 133, .external_lex_state = 5},
  [3748] = {.lex_state = 133, .external_lex_state = 5},
  [3749] = {.lex_state = 271, .external_lex_state = 4},
  [3750] = {.lex_state = 275, .external_lex_state = 4},
  [3751] = {.lex_state = 275, .external_lex_state = 4},
  [3752] = {.lex_state = 275, .external_lex_state = 4},
  [3753] = {.lex_state = 275, .external_lex_state = 4},
  [3754] = {.lex_state = 275, .external_lex_state = 4},
  [3755] = {.lex_state = 271, .external_lex_state = 4},
  [3756] = {.lex_state = 275, .external_lex_state = 4},
  [3757] = {.lex_state = 133, .external_lex_state = 5},
  [3758] = {.lex_state = 133, .external_lex_state = 5},
  [3759] = {.lex_state = 133, .external_lex_state = 5},
  [3760] = {.lex_state = 271, .external_lex_state = 5},
  [3761] = {.lex_state = 271, .external_lex_state = 5},
  [3762] = {.lex_state = 271, .external_lex_state = 5},
  [3763] = {.lex_state = 271, .external_lex_state = 5},
  [3764] = {.lex_state = 271, .external_lex_state = 5},
  [3765] = {.lex_state = 271, .external_lex_state = 5},
  [3766] = {.lex_state = 271, .external_lex_state = 5},
  [3767] = {.lex_state = 271, .external_lex_state = 5},
  [3768] = {.lex_state = 271, .external_lex_state = 5},
  [3769] = {.lex_state = 271, .external_lex_state = 5},
  [3770] = {.lex_state = 271, .external_lex_state = 5},
  [3771] = {.lex_state = 271, .external_lex_state = 5},
  [3772] = {.lex_state = 271, .external_lex_state = 5},
  [3773] = {.lex_state = 271, .external_lex_state = 5},
  [3774] = {.lex_state = 271, .external_lex_state = 5},
  [3775] = {.lex_state = 271, .external_lex_state = 5},
  [3776] = {.lex_state = 271, .external_lex_state = 5},
  [3777] = {.lex_state = 275, .external_lex_state = 4},
  [3778] = {.lex_state = 133, .external_lex_state = 5},
  [3779] = {.lex_state = 133, .external_lex_state = 5},
  [3780] = {.lex_state = 133, .external_lex_state = 5},
  [3781] = {.lex_state = 133, .external_lex_state = 5},
  [3782] = {.lex_state = 133, .external_lex_state = 5},
  [3783] = {.lex_state = 133, .external_lex_state = 5},
  [3784] = {.lex_state = 133, .external_lex_state = 5},
  [3785] = {.lex_state = 275, .external_lex_state = 4},
  [3786] = {.lex_state = 275, .external_lex_state = 4},
  [3787] = {.lex_state = 275, .external_lex_state = 4},
  [3788] = {.lex_state = 275, .external_lex_state = 4},
  [3789] = {.lex_state = 271, .external_lex_state = 4},
  [3790] = {.lex_state = 275, .external_lex_state = 4},
  [3791] = {.lex_state = 133, .external_lex_state = 5},
  [3792] = {.lex_state = 133, .external_lex_state = 5},
  [3793] = {.lex_state = 133, .external_lex_state = 5},
  [3794] = {.lex_state = 133, .external_lex_state = 5},
  [3795] = {.lex_state = 133, .external_lex_state = 5},
  [3796] = {.lex_state = 133, .external_lex_state = 5},
  [3797] = {.lex_state = 133, .external_lex_state = 5},
  [3798] = {.lex_state = 133, .external_lex_state = 5},
  [3799] = {.lex_state = 133, .external_lex_state = 5},
  [3800] = {.lex_state = 133, .external_lex_state = 5},
  [3801] = {.lex_state = 133, .external_lex_state = 5},
  [3802] = {.lex_state = 133, .external_lex_state = 5},
  [3803] = {.lex_state = 133, .external_lex_state = 5},
  [3804] = {.lex_state = 133, .external_lex_state = 5},
  [3805] = {.lex_state = 133, .external_lex_state = 5},
  [3806] = {.lex_state = 275, .external_lex_state = 4},
  [3807] = {.lex_state = 133, .external_lex_state = 5},
  [3808] = {.lex_state = 106, .external_lex_state = 4},
  [3809] = {.lex_state = 133, .external_lex_state = 5},
  [3810] = {.lex_state = 133, .external_lex_state = 5},
  [3811] = {.lex_state = 133, .external_lex_state = 5},
  [3812] = {.lex_state = 133, .external_lex_state = 5},
  [3813] = {.lex_state = 133, .external_lex_state = 5},
  [3814] = {.lex_state = 271, .external_lex_state = 4},
  [3815] = {.lex_state = 139, .external_lex_state = 4},
  [3816] = {.lex_state = 271, .external_lex_state = 4},
  [3817] = {.lex_state = 275, .external_lex_state = 4},
  [3818] = {.lex_state = 275, .external_lex_state = 4},
  [3819] = {.lex_state = 275, .external_lex_state = 4},
  [3820] = {.lex_state = 133, .external_lex_state = 5},
  [3821] = {.lex_state = 133, .external_lex_state = 5},
  [3822] = {.lex_state = 133, .external_lex_state = 5},
  [3823] = {.lex_state = 275, .external_lex_state = 4},
  [3824] = {.lex_state = 275, .external_lex_state = 4},
  [3825] = {.lex_state = 275, .external_lex_state = 4},
  [3826] = {.lex_state = 275, .external_lex_state = 4},
  [3827] = {.lex_state = 275, .external_lex_state = 4},
  [3828] = {.lex_state = 275, .external_lex_state = 4},
  [3829] = {.lex_state = 271, .external_lex_state = 4},
  [3830] = {.lex_state = 271, .external_lex_state = 4},
  [3831] = {.lex_state = 275, .external_lex_state = 4},
  [3832] = {.lex_state = 271, .external_lex_state = 4},
  [3833] = {.lex_state = 275, .external_lex_state = 4},
  [3834] = {.lex_state = 275, .external_lex_state = 4},
  [3835] = {.lex_state = 275, .external_lex_state = 4},
  [3836] = {.lex_state = 275, .external_lex_state = 4},
  [3837] = {.lex_state = 139, .external_lex_state = 4},
  [3838] = {.lex_state = 275, .external_lex_state = 4},
  [3839] = {.lex_state = 275, .external_lex_state = 4},
  [3840] = {.lex_state = 275, .external_lex_state = 4},
  [3841] = {.lex_state = 139, .external_lex_state = 4},
  [3842] = {.lex_state = 139, .external_lex_state = 4},
  [3843] = {.lex_state = 139, .external_lex_state = 4},
  [3844] = {.lex_state = 139, .external_lex_state = 4},
  [3845] = {.lex_state = 275, .external_lex_state = 4},
  [3846] = {.lex_state = 275, .external_lex_state = 4},
  [3847] = {.lex_state = 275, .external_lex_state = 4},
  [3848] = {.lex_state = 275, .external_lex_state = 4},
  [3849] = {.lex_state = 275, .external_lex_state = 4},
  [3850] = {.lex_state = 139, .external_lex_state = 4},
  [3851] = {.lex_state = 106, .external_lex_state = 4},
  [3852] = {.lex_state = 139, .external_lex_state = 4},
  [3853] = {.lex_state = 139, .external_lex_state = 4},
  [3854] = {.lex_state = 139, .external_lex_state = 4},
  [3855] = {.lex_state = 275, .external_lex_state = 4},
  [3856] = {.lex_state = 106, .external_lex_state = 4},
  [3857] = {.lex_state = 275, .external_lex_state = 4},
  [3858] = {.lex_state = 275, .external_lex_state = 4},
  [3859] = {.lex_state = 271, .external_lex_state = 4},
  [3860] = {.lex_state = 133, .external_lex_state = 5},
  [3861] = {.lex_state = 133, .external_lex_state = 5},
  [3862] = {.lex_state = 271, .external_lex_state = 4},
  [3863] = {.lex_state = 139, .external_lex_state = 4},
  [3864] = {.lex_state = 139, .external_lex_state = 4},
  [3865] = {.lex_state = 139, .external_lex_state = 4},
  [3866] = {.lex_state = 139, .external_lex_state = 4},
  [3867] = {.lex_state = 139, .external_lex_state = 4},
  [3868] = {.lex_state = 139, .external_lex_state = 4},
  [3869] = {.lex_state = 139, .external_lex_state = 4},
  [3870] = {.lex_state = 139, .external_lex_state = 4},
  [3871] = {.lex_state = 139, .external_lex_state = 4},
  [3872] = {.lex_state = 139, .external_lex_state = 4},
  [3873] = {.lex_state = 139, .external_lex_state = 4},
  [3874] = {.lex_state = 139, .external_lex_state = 4},
  [3875] = {.lex_state = 139, .external_lex_state = 4},
  [3876] = {.lex_state = 139, .external_lex_state = 4},
  [3877] = {.lex_state = 139, .external_lex_state = 4},
  [3878] = {.lex_state = 139, .external_lex_state = 4},
  [3879] = {.lex_state = 106, .external_lex_state = 4},
  [3880] = {.lex_state = 106, .external_lex_state = 4},
  [3881] = {.lex_state = 106, .external_lex_state = 4},
  [3882] = {.lex_state = 275, .external_lex_state = 4},
  [3883] = {.lex_state = 275, .external_lex_state = 4},
  [3884] = {.lex_state = 271, .external_lex_state = 4},
  [3885] = {.lex_state = 271, .external_lex_state = 4},
  [3886] = {.lex_state = 275, .external_lex_state = 4},
  [3887] = {.lex_state = 275, .external_lex_state = 4},
  [3888] = {.lex_state = 275, .external_lex_state = 4},
  [3889] = {.lex_state = 275, .external_lex_state = 4},
  [3890] = {.lex_state = 275, .external_lex_state = 4},
  [3891] = {.lex_state = 275, .external_lex_state = 4},
  [3892] = {.lex_state = 275, .external_lex_state = 4},
  [3893] = {.lex_state = 275, .external_lex_state = 4},
  [3894] = {.lex_state = 275, .external_lex_state = 4},
  [3895] = {.lex_state = 106, .external_lex_state = 4},
  [3896] = {.lex_state = 275, .external_lex_state = 4},
  [3897] = {.lex_state = 275, .external_lex_state = 4},
  [3898] = {.lex_state = 133, .external_lex_state = 5},
  [3899] = {.lex_state = 133, .external_lex_state = 5},
  [3900] = {.lex_state = 275, .external_lex_state = 4},
  [3901] = {.lex_state = 275, .external_lex_state = 4},
  [3902] = {.lex_state = 142, .external_lex_state = 4},
  [3903] = {.lex_state = 142, .external_lex_state = 4},
  [3904] = {.lex_state = 142, .external_lex_state = 4},
  [3905] = {.lex_state = 142, .external_lex_state = 4},
  [3906] = {.lex_state = 142, .external_lex_state = 4},
  [3907] = {.lex_state = 142, .external_lex_state = 4},
  [3908] = {.lex_state = 142, .external_lex_state = 4},
  [3909] = {.lex_state = 142, .external_lex_state = 4},
  [3910] = {.lex_state = 142, .external_lex_state = 4},
  [3911] = {.lex_state = 142, .external_lex_state = 4},
  [3912] = {.lex_state = 142, .external_lex_state = 4},
  [3913] = {.lex_state = 142, .external_lex_state = 4},
  [3914] = {.lex_state = 142, .external_lex_state = 4},
  [3915] = {.lex_state = 142, .external_lex_state = 4},
  [3916] = {.lex_state = 142, .external_lex_state = 4},
  [3917] = {.lex_state = 142, .external_lex_state = 4},
  [3918] = {.lex_state = 142, .external_lex_state = 4},
  [3919] = {.lex_state = 142, .external_lex_state = 4},
  [3920] = {.lex_state = 142, .external_lex_state = 4},
  [3921] = {.lex_state = 142, .external_lex_state = 4},
  [3922] = {.lex_state = 275, .external_lex_state = 4},
  [3923] = {.lex_state = 275, .external_lex_state = 4},
  [3924] = {.lex_state = 275, .external_lex_state = 4},
  [3925] = {.lex_state = 275, .external_lex_state = 4},
  [3926] = {.lex_state = 275, .external_lex_state = 4},
  [3927] = {.lex_state = 271, .external_lex_state = 4},
  [3928] = {.lex_state = 271, .external_lex_state = 4},
  [3929] = {.lex_state = 275, .external_lex_state = 4},
  [3930] = {.lex_state = 275, .external_lex_state = 4},
  [3931] = {.lex_state = 275, .external_lex_state = 4},
  [3932] = {.lex_state = 275, .external_lex_state = 4},
  [3933] = {.lex_state = 275, .external_lex_state = 4},
  [3934] = {.lex_state = 275, .external_lex_state = 4},
  [3935] = {.lex_state = 275, .external_lex_state = 4},
  [3936] = {.lex_state = 275, .external_lex_state = 4},
  [3937] = {.lex_state = 275, .external_lex_state = 4},
  [3938] = {.lex_state = 275, .external_lex_state = 4},
  [3939] = {.lex_state = 275, .external_lex_state = 4},
  [3940] = {.lex_state = 133, .external_lex_state = 5},
  [3941] = {.lex_state = 133, .external_lex_state = 5},
  [3942] = {.lex_state = 142, .external_lex_state = 4},
  [3943] = {.lex_state = 142, .external_lex_state = 4},
  [3944] = {.lex_state = 142, .external_lex_state = 4},
  [3945] = {.lex_state = 142, .external_lex_state = 4},
  [3946] = {.lex_state = 142, .external_lex_state = 4},
  [3947] = {.lex_state = 142, .external_lex_state = 4},
  [3948] = {.lex_state = 275, .external_lex_state = 4},
  [3949] = {.lex_state = 275, .external_lex_state = 4},
  [3950] = {.lex_state = 275, .external_lex_state = 4},
  [3951] = {.lex_state = 275, .external_lex_state = 4},
  [3952] = {.lex_state = 271, .external_lex_state = 4},
  [3953] = {.lex_state = 275, .external_lex_state = 4},
  [3954] = {.lex_state = 271, .external_lex_state = 4},
  [3955] = {.lex_state = 271, .external_lex_state = 4},
  [3956] = {.lex_state = 271, .external_lex_state = 4},
  [3957] = {.lex_state = 271, .external_lex_state = 4},
  [3958] = {.lex_state = 142, .external_lex_state = 4},
  [3959] = {.lex_state = 142, .external_lex_state = 4},
  [3960] = {.lex_state = 142, .external_lex_state = 4},
  [3961] = {.lex_state = 142, .external_lex_state = 4},
  [3962] = {.lex_state = 142, .external_lex_state = 4},
  [3963] = {.lex_state = 142, .external_lex_state = 4},
  [3964] = {.lex_state = 142, .external_lex_state = 4},
  [3965] = {.lex_state = 142, .external_lex_state = 4},
  [3966] = {.lex_state = 142, .external_lex_state = 4},
  [3967] = {.lex_state = 142, .external_lex_state = 4},
  [3968] = {.lex_state = 271, .external_lex_state = 4},
  [3969] = {.lex_state = 271, .external_lex_state = 4},
  [3970] = {.lex_state = 271, .external_lex_state = 4},
  [3971] = {.lex_state = 271, .external_lex_state = 4},
  [3972] = {.lex_state = 271, .external_lex_state = 4},
  [3973] = {.lex_state = 271, .external_lex_state = 4},
  [3974] = {.lex_state = 271, .external_lex_state = 4},
  [3975] = {.lex_state = 271, .external_lex_state = 4},
  [3976] = {.lex_state = 271, .external_lex_state = 4},
  [3977] = {.lex_state = 271, .external_lex_state = 4},
  [3978] = {.lex_state = 271, .external_lex_state = 4},
  [3979] = {.lex_state = 275, .external_lex_state = 4},
  [3980] = {.lex_state = 275, .external_lex_state = 4},
  [3981] = {.lex_state = 275, .external_lex_state = 4},
  [3982] = {.lex_state = 271, .external_lex_state = 4},
  [3983] = {.lex_state = 271, .external_lex_state = 4},
  [3984] = {.lex_state = 275, .external_lex_state = 4},
  [3985] = {.lex_state = 275, .external_lex_state = 4},
  [3986] = {.lex_state = 275, .external_lex_state = 4},
  [3987] = {.lex_state = 106, .external_lex_state = 4},
  [3988] = {.lex_state = 142, .external_lex_state = 4},
  [3989] = {.lex_state = 142, .external_lex_state = 4},
  [3990] = {.lex_state = 142, .external_lex_state = 4},
  [3991] = {.lex_state = 142, .external_lex_state = 4},
  [3992] = {.lex_state = 142, .external_lex_state = 4},
  [3993] = {.lex_state = 142, .external_lex_state = 4},
  [3994] = {.lex_state = 142, .external_lex_state = 4},
  [3995] = {.lex_state = 142, .external_lex_state = 4},
  [3996] = {.lex_state = 142, .external_lex_state = 4},
  [3997] = {.lex_state = 142, .external_lex_state = 4},
  [3998] = {.lex_state = 142, .external_lex_state = 4},
  [3999] = {.lex_state = 142, .external_lex_state = 4},
  [4000] = {.lex_state = 142, .external_lex_state = 4},
  [4001] = {.lex_state = 142, .external_lex_state = 4},
  [4002] = {.lex_state = 142, .external_lex_state = 4},
  [4003] = {.lex_state = 142, .external_lex_state = 4},
  [4004] = {.lex_state = 275, .external_lex_state = 4},
  [4005] = {.lex_state = 275, .external_lex_state = 4},
  [4006] = {.lex_state = 142, .external_lex_state = 4},
  [4007] = {.lex_state = 142, .external_lex_state = 4},
  [4008] = {.lex_state = 142, .external_lex_state = 4},
  [4009] = {.lex_state = 275, .external_lex_state = 4},
  [4010] = {.lex_state = 142, .external_lex_state = 4},
  [4011] = {.lex_state = 142, .external_lex_state = 4},
  [4012] = {.lex_state = 142, .external_lex_state = 4},
  [4013] = {.lex_state = 142, .external_lex_state = 4},
  [4014] = {.lex_state = 142, .external_lex_state = 4},
  [4015] = {.lex_state = 142, .external_lex_state = 4},
  [4016] = {.lex_state = 142, .external_lex_state = 4},
  [4017] = {.lex_state = 142, .external_lex_state = 4},
  [4018] = {.lex_state = 142, .external_lex_state = 4},
  [4019] = {.lex_state = 142, .external_lex_state = 4},
  [4020] = {.lex_state = 142, .external_lex_state = 4},
  [4021] = {.lex_state = 142, .external_lex_state = 4},
  [4022] = {.lex_state = 106, .external_lex_state = 4},
  [4023] = {.lex_state = 106, .external_lex_state = 4},
  [4024] = {.lex_state = 106, .external_lex_state = 4},
  [4025] = {.lex_state = 275, .external_lex_state = 4},
  [4026] = {.lex_state = 106, .external_lex_state = 4},
  [4027] = {.lex_state = 106, .external_lex_state = 4},
  [4028] = {.lex_state = 106, .external_lex_state = 4},
  [4029] = {.lex_state = 271, .external_lex_state = 5},
  [4030] = {.lex_state = 271, .external_lex_state = 5},
  [4031] = {.lex_state = 106, .external_lex_state = 4},
  [4032] = {.lex_state = 271, .external_lex_state = 5},
  [4033] = {.lex_state = 133, .external_lex_state = 4},
  [4034] = {.lex_state = 271, .external_lex_state = 4},
  [4035] = {.lex_state = 271, .external_lex_state = 4},
  [4036] = {.lex_state = 271, .external_lex_state = 4},
  [4037] = {.lex_state = 133, .external_lex_state = 4},
  [4038] = {.lex_state = 106, .external_lex_state = 4},
  [4039] = {.lex_state = 133, .external_lex_state = 4},
  [4040] = {.lex_state = 106, .external_lex_state = 4},
  [4041] = {.lex_state = 275, .external_lex_state = 4},
  [4042] = {.lex_state = 106, .external_lex_state = 4},
  [4043] = {.lex_state = 106, .external_lex_state = 4},
  [4044] = {.lex_state = 271, .external_lex_state = 4},
  [4045] = {.lex_state = 106, .external_lex_state = 4},
  [4046] = {.lex_state = 106, .external_lex_state = 4},
  [4047] = {.lex_state = 275, .external_lex_state = 4},
  [4048] = {.lex_state = 106, .external_lex_state = 4},
  [4049] = {.lex_state = 275, .external_lex_state = 4},
  [4050] = {.lex_state = 106, .external_lex_state = 4},
  [4051] = {.lex_state = 106, .external_lex_state = 4},
  [4052] = {.lex_state = 106, .external_lex_state = 4},
  [4053] = {.lex_state = 106, .external_lex_state = 4},
  [4054] = {.lex_state = 106, .external_lex_state = 4},
  [4055] = {.lex_state = 106, .external_lex_state = 4},
  [4056] = {.lex_state = 106, .external_lex_state = 4},
  [4057] = {.lex_state = 106, .external_lex_state = 4},
  [4058] = {.lex_state = 275, .external_lex_state = 4},
  [4059] = {.lex_state = 275, .external_lex_state = 4},
  [4060] = {.lex_state = 275, .external_lex_state = 4},
  [4061] = {.lex_state = 145, .external_lex_state = 4},
  [4062] = {.lex_state = 271, .external_lex_state = 4},
  [4063] = {.lex_state = 271, .external_lex_state = 4},
  [4064] = {.lex_state = 145, .external_lex_state = 4},
  [4065] = {.lex_state = 271, .external_lex_state = 4},
  [4066] = {.lex_state = 271, .external_lex_state = 4},
  [4067] = {.lex_state = 271, .external_lex_state = 4},
  [4068] = {.lex_state = 271, .external_lex_state = 4},
  [4069] = {.lex_state = 271, .external_lex_state = 4},
  [4070] = {.lex_state = 271, .external_lex_state = 4},
  [4071] = {.lex_state = 271, .external_lex_state = 4},
  [4072] = {.lex_state = 271, .external_lex_state = 4},
  [4073] = {.lex_state = 271, .external_lex_state = 4},
  [4074] = {.lex_state = 271, .external_lex_state = 4},
  [4075] = {.lex_state = 271, .external_lex_state = 4},
  [4076] = {.lex_state = 271, .external_lex_state = 4},
  [4077] = {.lex_state = 271, .external_lex_state = 4},
  [4078] = {.lex_state = 271, .external_lex_state = 4},
  [4079] = {.lex_state = 271, .external_lex_state = 4},
  [4080] = {.lex_state = 145, .external_lex_state = 4},
  [4081] = {.lex_state = 145, .external_lex_state = 4},
  [4082] = {.lex_state = 145, .external_lex_state = 4},
  [4083] = {.lex_state = 106, .external_lex_state = 4},
  [4084] = {.lex_state = 145, .external_lex_state = 4},
  [4085] = {.lex_state = 145, .external_lex_state = 4},
  [4086] = {.lex_state = 145, .external_lex_state = 4},
  [4087] = {.lex_state = 145, .external_lex_state = 4},
  [4088] = {.lex_state = 145, .external_lex_state = 4},
  [4089] = {.lex_state = 145, .external_lex_state = 4},
  [4090] = {.lex_state = 145, .external_lex_state = 4},
  [4091] = {.lex_state = 145, .external_lex_state = 4},
  [4092] = {.lex_state = 133, .external_lex_state = 4},
  [4093] = {.lex_state = 133, .external_lex_state = 4},
  [4094] = {.lex_state = 133, .external_lex_state = 4},
  [4095] = {.lex_state = 271, .external_lex_state = 4},
  [4096] = {.lex_state = 271, .external_lex_state = 4},
  [4097] = {.lex_state = 106, .external_lex_state = 4},
  [4098] = {.lex_state = 106, .external_lex_state = 4},
  [4099] = {.lex_state = 106, .external_lex_state = 4},
  [4100] = {.lex_state = 271, .external_lex_state = 4},
  [4101] = {.lex_state = 271, .external_lex_state = 4},
  [4102] = {.lex_state = 271, .external_lex_state = 4},
  [4103] = {.lex_state = 271, .external_lex_state = 4},
  [4104] = {.lex_state = 271, .external_lex_state = 4},
  [4105] = {.lex_state = 271, .external_lex_state = 4},
  [4106] = {.lex_state = 106, .external_lex_state = 4},
  [4107] = {.lex_state = 271, .external_lex_state = 4},
  [4108] = {.lex_state = 106, .external_lex_state = 4},
  [4109] = {.lex_state = 106, .external_lex_state = 4},
  [4110] = {.lex_state = 106, .external_lex_state = 4},
  [4111] = {.lex_state = 106, .external_lex_state = 4},
  [4112] = {.lex_state = 106, .external_lex_state = 4},
  [4113] = {.lex_state = 271, .external_lex_state = 4},
  [4114] = {.lex_state = 106, .external_lex_state = 4},
  [4115] = {.lex_state = 271, .external_lex_state = 4},
  [4116] = {.lex_state = 106, .external_lex_state = 4},
  [4117] = {.lex_state = 106, .external_lex_state = 4},
  [4118] = {.lex_state = 133, .external_lex_state = 4},
  [4119] = {.lex_state = 271, .external_lex_state = 4},
  [4120] = {.lex_state = 271, .external_lex_state = 4},
  [4121] = {.lex_state = 106, .external_lex_state = 4},
  [4122] = {.lex_state = 133, .external_lex_state = 4},
  [4123] = {.lex_state = 271, .external_lex_state = 4},
  [4124] = {.lex_state = 271, .external_lex_state = 4},
  [4125] = {.lex_state = 271, .external_lex_state = 4},
  [4126] = {.lex_state = 271, .external_lex_state = 4},
  [4127] = {.lex_state = 271, .external_lex_state = 4},
  [4128] = {.lex_state = 145, .external_lex_state = 4},
  [4129] = {.lex_state = 106, .external_lex_state = 4},
  [4130] = {.lex_state = 106, .external_lex_state = 4},
  [4131] = {.lex_state = 271, .external_lex_state = 4},
  [4132] = {.lex_state = 271, .external_lex_state = 4},
  [4133] = {.lex_state = 106, .external_lex_state = 4},
  [4134] = {.lex_state = 271, .external_lex_state = 4},
  [4135] = {.lex_state = 133, .external_lex_state = 4},
  [4136] = {.lex_state = 271, .external_lex_state = 4},
  [4137] = {.lex_state = 106, .external_lex_state = 4},
  [4138] = {.lex_state = 271, .external_lex_state = 4},
  [4139] = {.lex_state = 271, .external_lex_state = 4},
  [4140] = {.lex_state = 106, .external_lex_state = 4},
  [4141] = {.lex_state = 106, .external_lex_state = 4},
  [4142] = {.lex_state = 106, .external_lex_state = 4},
  [4143] = {.lex_state = 106, .external_lex_state = 4},
  [4144] = {.lex_state = 106, .external_lex_state = 4},
  [4145] = {.lex_state = 106, .external_lex_state = 4},
  [4146] = {.lex_state = 106, .external_lex_state = 4},
  [4147] = {.lex_state = 106, .external_lex_state = 4},
  [4148] = {.lex_state = 106, .external_lex_state = 4},
  [4149] = {.lex_state = 106, .external_lex_state = 4},
  [4150] = {.lex_state = 271, .external_lex_state = 4},
  [4151] = {.lex_state = 271, .external_lex_state = 4},
  [4152] = {.lex_state = 271, .external_lex_state = 4},
  [4153] = {.lex_state = 271, .external_lex_state = 4},
  [4154] = {.lex_state = 271, .external_lex_state = 4},
  [4155] = {.lex_state = 271, .external_lex_state = 4},
  [4156] = {.lex_state = 271, .external_lex_state = 4},
  [4157] = {.lex_state = 271, .external_lex_state = 4},
  [4158] = {.lex_state = 271, .external_lex_state = 4},
  [4159] = {.lex_state = 271, .external_lex_state = 4},
  [4160] = {.lex_state = 271, .external_lex_state = 4},
  [4161] = {.lex_state = 271, .external_lex_state = 4},
  [4162] = {.lex_state = 271, .external_lex_state = 4},
  [4163] = {.lex_state = 145, .external_lex_state = 4},
  [4164] = {.lex_state = 271, .external_lex_state = 4},
  [4165] = {.lex_state = 271, .external_lex_state = 4},
  [4166] = {.lex_state = 133, .external_lex_state = 4},
  [4167] = {.lex_state = 133, .external_lex_state = 4},
  [4168] = {.lex_state = 271, .external_lex_state = 4},
  [4169] = {.lex_state = 271, .external_lex_state = 4},
  [4170] = {.lex_state = 106, .external_lex_state = 4},
  [4171] = {.lex_state = 106, .external_lex_state = 4},
  [4172] = {.lex_state = 271, .external_lex_state = 4},
  [4173] = {.lex_state = 271, .external_lex_state = 4},
  [4174] = {.lex_state = 106, .external_lex_state = 4},
  [4175] = {.lex_state = 271, .external_lex_state = 4},
  [4176] = {.lex_state = 271, .external_lex_state = 4},
  [4177] = {.lex_state = 271, .external_lex_state = 4},
  [4178] = {.lex_state = 271, .external_lex_state = 4},
  [4179] = {.lex_state = 106, .external_lex_state = 4},
  [4180] = {.lex_state = 106, .external_lex_state = 4},
  [4181] = {.lex_state = 106, .external_lex_state = 4},
  [4182] = {.lex_state = 271, .external_lex_state = 4},
  [4183] = {.lex_state = 271, .external_lex_state = 4},
  [4184] = {.lex_state = 271, .external_lex_state = 4},
  [4185] = {.lex_state = 271, .external_lex_state = 4},
  [4186] = {.lex_state = 271, .external_lex_state = 4},
  [4187] = {.lex_state = 106, .external_lex_state = 4},
  [4188] = {.lex_state = 106, .external_lex_state = 4},
  [4189] = {.lex_state = 271, .external_lex_state = 4},
  [4190] = {.lex_state = 271, .external_lex_state = 4},
  [4191] = {.lex_state = 271, .external_lex_state = 4},
  [4192] = {.lex_state = 271, .external_lex_state = 4},
  [4193] = {.lex_state = 271, .external_lex_state = 4},
  [4194] = {.lex_state = 145, .external_lex_state = 4},
  [4195] = {.lex_state = 271, .external_lex_state = 4},
  [4196] = {.lex_state = 145, .external_lex_state = 4},
  [4197] = {.lex_state = 106, .external_lex_state = 4},
  [4198] = {.lex_state = 106, .external_lex_state = 4},
  [4199] = {.lex_state = 271, .external_lex_state = 4},
  [4200] = {.lex_state = 106, .external_lex_state = 4},
  [4201] = {.lex_state = 106, .external_lex_state = 4},
  [4202] = {.lex_state = 145, .external_lex_state = 4},
  [4203] = {.lex_state = 145, .external_lex_state = 4},
  [4204] = {.lex_state = 145, .external_lex_state = 4},
  [4205] = {.lex_state = 145, .external_lex_state = 4},
  [4206] = {.lex_state = 145, .external_lex_state = 4},
  [4207] = {.lex_state = 145, .external_lex_state = 4},
  [4208] = {.lex_state = 271, .external_lex_state = 4},
  [4209] = {.lex_state = 271, .external_lex_state = 4},
  [4210] = {.lex_state = 145, .external_lex_state = 4},
  [4211] = {.lex_state = 271, .external_lex_state = 4},
  [4212] = {.lex_state = 271, .external_lex_state = 4},
  [4213] = {.lex_state = 145, .external_lex_state = 4},
  [4214] = {.lex_state = 271, .external_lex_state = 4},
  [4215] = {.lex_state = 271, .external_lex_state = 4},
  [4216] = {.lex_state = 145, .external_lex_state = 4},
  [4217] = {.lex_state = 145, .external_lex_state = 4},
  [4218] = {.lex_state = 145, .external_lex_state = 4},
  [4219] = {.lex_state = 145, .external_lex_state = 4},
  [4220] = {.lex_state = 145, .external_lex_state = 4},
  [4221] = {.lex_state = 145, .external_lex_state = 4},
  [4222] = {.lex_state = 145, .external_lex_state = 4},
  [4223] = {.lex_state = 145, .external_lex_state = 4},
  [4224] = {.lex_state = 145, .external_lex_state = 4},
  [4225] = {.lex_state = 145, .external_lex_state = 4},
  [4226] = {.lex_state = 145, .external_lex_state = 4},
  [4227] = {.lex_state = 145, .external_lex_state = 4},
  [4228] = {.lex_state = 106, .external_lex_state = 4},
  [4229] = {.lex_state = 106, .external_lex_state = 4},
  [4230] = {.lex_state = 106, .external_lex_state = 4},
  [4231] = {.lex_state = 106, .external_lex_state = 4},
  [4232] = {.lex_state = 271, .external_lex_state = 4},
  [4233] = {.lex_state = 271, .external_lex_state = 4},
  [4234] = {.lex_state = 106, .external_lex_state = 4},
  [4235] = {.lex_state = 106, .external_lex_state = 4},
  [4236] = {.lex_state = 271, .external_lex_state = 4},
  [4237] = {.lex_state = 106, .external_lex_state = 4},
  [4238] = {.lex_state = 106, .external_lex_state = 4},
  [4239] = {.lex_state = 106, .external_lex_state = 4},
  [4240] = {.lex_state = 106, .external_lex_state = 4},
  [4241] = {.lex_state = 106, .external_lex_state = 4},
  [4242] = {.lex_state = 106, .external_lex_state = 4},
  [4243] = {.lex_state = 271, .external_lex_state = 4},
  [4244] = {.lex_state = 106, .external_lex_state = 4},
  [4245] = {.lex_state = 133, .external_lex_state = 4},
  [4246] = {.lex_state = 106, .external_lex_state = 4},
  [4247] = {.lex_state = 271, .external_lex_state = 4},
  [4248] = {.lex_state = 106, .external_lex_state = 4},
  [4249] = {.lex_state = 271, .external_lex_state = 4},
  [4250] = {.lex_state = 271, .external_lex_state = 4},
  [4251] = {.lex_state = 271, .external_lex_state = 4},
  [4252] = {.lex_state = 271, .external_lex_state = 4},
  [4253] = {.lex_state = 271, .external_lex_state = 4},
  [4254] = {.lex_state = 271, .external_lex_state = 4},
  [4255] = {.lex_state = 133, .external_lex_state = 4},
  [4256] = {.lex_state = 271, .external_lex_state = 4},
  [4257] = {.lex_state = 271, .external_lex_state = 4},
  [4258] = {.lex_state = 106, .external_lex_state = 4},
  [4259] = {.lex_state = 106, .external_lex_state = 4},
  [4260] = {.lex_state = 145, .external_lex_state = 4},
  [4261] = {.lex_state = 145, .external_lex_state = 4},
  [4262] = {.lex_state = 106, .external_lex_state = 4},
  [4263] = {.lex_state = 106, .external_lex_state = 4},
  [4264] = {.lex_state = 145, .external_lex_state = 4},
  [4265] = {.lex_state = 145, .external_lex_state = 4},
  [4266] = {.lex_state = 106, .external_lex_state = 4},
  [4267] = {.lex_state = 106, .external_lex_state = 4},
  [4268] = {.lex_state = 145, .external_lex_state = 4},
  [4269] = {.lex_state = 145, .external_lex_state = 4},
  [4270] = {.lex_state = 106, .external_lex_state = 4},
  [4271] = {.lex_state = 106, .external_lex_state = 4},
  [4272] = {.lex_state = 106, .external_lex_state = 4},
  [4273] = {.lex_state = 106, .external_lex_state = 4},
  [4274] = {.lex_state = 106, .external_lex_state = 4},
  [4275] = {.lex_state = 106, .external_lex_state = 4},
  [4276] = {.lex_state = 271, .external_lex_state = 4},
  [4277] = {.lex_state = 271, .external_lex_state = 4},
  [4278] = {.lex_state = 133, .external_lex_state = 4},
  [4279] = {.lex_state = 271, .external_lex_state = 4},
  [4280] = {.lex_state = 271, .external_lex_state = 4},
  [4281] = {.lex_state = 271, .external_lex_state = 4},
  [4282] = {.lex_state = 271, .external_lex_state = 4},
  [4283] = {.lex_state = 271, .external_lex_state = 4},
  [4284] = {.lex_state = 271, .external_lex_state = 4},
  [4285] = {.lex_state = 271, .external_lex_state = 4},
  [4286] = {.lex_state = 271, .external_lex_state = 4},
  [4287] = {.lex_state = 271, .external_lex_state = 4},
  [4288] = {.lex_state = 106, .external_lex_state = 4},
  [4289] = {.lex_state = 271, .external_lex_state = 4},
  [4290] = {.lex_state = 133, .external_lex_state = 4},
  [4291] = {.lex_state = 271, .external_lex_state = 4},
  [4292] = {.lex_state = 106, .external_lex_state = 4},
  [4293] = {.lex_state = 271, .external_lex_state = 4},
  [4294] = {.lex_state = 133, .external_lex_state = 4},
  [4295] = {.lex_state = 106, .external_lex_state = 4},
  [4296] = {.lex_state = 106, .external_lex_state = 4},
  [4297] = {.lex_state = 106, .external_lex_state = 4},
  [4298] = {.lex_state = 106, .external_lex_state = 4},
  [4299] = {.lex_state = 271, .external_lex_state = 4},
  [4300] = {.lex_state = 271, .external_lex_state = 4},
  [4301] = {.lex_state = 271, .external_lex_state = 4},
  [4302] = {.lex_state = 106, .external_lex_state = 4},
  [4303] = {.lex_state = 271, .external_lex_state = 4},
  [4304] = {.lex_state = 271, .external_lex_state = 4},
  [4305] = {.lex_state = 271, .external_lex_state = 4},
  [4306] = {.lex_state = 271, .external_lex_state = 4},
  [4307] = {.lex_state = 271, .external_lex_state = 4},
  [4308] = {.lex_state = 271, .external_lex_state = 4},
  [4309] = {.lex_state = 271, .external_lex_state = 4},
  [4310] = {.lex_state = 271, .external_lex_state = 4},
  [4311] = {.lex_state = 271, .external_lex_state = 4},
  [4312] = {.lex_state = 271, .external_lex_state = 4},
  [4313] = {.lex_state = 271, .external_lex_state = 4},
  [4314] = {.lex_state = 271, .external_lex_state = 4},
  [4315] = {.lex_state = 271, .external_lex_state = 4},
  [4316] = {.lex_state = 271, .external_lex_state = 4},
  [4317] = {.lex_state = 271, .external_lex_state = 4},
  [4318] = {.lex_state = 106, .external_lex_state = 4},
  [4319] = {.lex_state = 106, .external_lex_state = 4},
  [4320] = {.lex_state = 271, .external_lex_state = 4},
  [4321] = {.lex_state = 106, .external_lex_state = 4},
  [4322] = {.lex_state = 106, .external_lex_state = 4},
  [4323] = {.lex_state = 133, .external_lex_state = 4},
  [4324] = {.lex_state = 133, .external_lex_state = 4},
  [4325] = {.lex_state = 271, .external_lex_state = 4},
  [4326] = {.lex_state = 271, .external_lex_state = 4},
  [4327] = {.lex_state = 133, .external_lex_state = 4},
  [4328] = {.lex_state = 133, .external_lex_state = 4},
  [4329] = {.lex_state = 106, .external_lex_state = 4},
  [4330] = {.lex_state = 106, .external_lex_state = 4},
  [4331] = {.lex_state = 106, .external_lex_state = 4},
  [4332] = {.lex_state = 106, .external_lex_state = 4},
  [4333] = {.lex_state = 106, .external_lex_state = 4},
  [4334] = {.lex_state = 106, .external_lex_state = 4},
  [4335] = {.lex_state = 106, .external_lex_state = 4},
  [4336] = {.lex_state = 106, .external_lex_state = 4},
  [4337] = {.lex_state = 106, .external_lex_state = 4},
  [4338] = {.lex_state = 271, .external_lex_state = 4},
  [4339] = {.lex_state = 271, .external_lex_state = 4},
  [4340] = {.lex_state = 106, .external_lex_state = 4},
  [4341] = {.lex_state = 145, .external_lex_state = 4},
  [4342] = {.lex_state = 145, .external_lex_state = 4},
  [4343] = {.lex_state = 145, .external_lex_state = 4},
  [4344] = {.lex_state = 145, .external_lex_state = 4},
  [4345] = {.lex_state = 145, .external_lex_state = 4},
  [4346] = {.lex_state = 271, .external_lex_state = 4},
  [4347] = {.lex_state = 271, .external_lex_state = 4},
  [4348] = {.lex_state = 145, .external_lex_state = 4},
  [4349] = {.lex_state = 133, .external_lex_state = 4},
  [4350] = {.lex_state = 106, .external_lex_state = 4},
  [4351] = {.lex_state = 106, .external_lex_state = 4},
  [4352] = {.lex_state = 145, .external_lex_state = 4},
  [4353] = {.lex_state = 145, .external_lex_state = 4},
  [4354] = {.lex_state = 106, .external_lex_state = 4},
  [4355] = {.lex_state = 106, .external_lex_state = 4},
  [4356] = {.lex_state = 106, .external_lex_state = 4},
  [4357] = {.lex_state = 106, .external_lex_state = 4},
  [4358] = {.lex_state = 145, .external_lex_state = 4},
  [4359] = {.lex_state = 271, .external_lex_state = 4},
  [4360] = {.lex_state = 133, .external_lex_state = 4},
  [4361] = {.lex_state = 145, .external_lex_state = 4},
  [4362] = {.lex_state = 145, .external_lex_state = 4},
  [4363] = {.lex_state = 271, .external_lex_state = 4},
  [4364] = {.lex_state = 145, .external_lex_state = 4},
  [4365] = {.lex_state = 145, .external_lex_state = 4},
  [4366] = {.lex_state = 145, .external_lex_state = 4},
  [4367] = {.lex_state = 145, .external_lex_state = 4},
  [4368] = {.lex_state = 106, .external_lex_state = 4},
  [4369] = {.lex_state = 145, .external_lex_state = 4},
  [4370] = {.lex_state = 106, .external_lex_state = 4},
  [4371] = {.lex_state = 106, .external_lex_state = 4},
  [4372] = {.lex_state = 106, .external_lex_state = 4},
  [4373] = {.lex_state = 106, .external_lex_state = 4},
  [4374] = {.lex_state = 106, .external_lex_state = 4},
  [4375] = {.lex_state = 106, .external_lex_state = 4},
  [4376] = {.lex_state = 271, .external_lex_state = 4},
  [4377] = {.lex_state = 271, .external_lex_state = 4},
  [4378] = {.lex_state = 271, .external_lex_state = 4},
  [4379] = {.lex_state = 271, .external_lex_state = 4},
  [4380] = {.lex_state = 133, .external_lex_state = 4},
  [4381] = {.lex_state = 271, .external_lex_state = 4},
  [4382] = {.lex_state = 133, .external_lex_state = 4},
  [4383] = {.lex_state = 106, .external_lex_state = 4},
  [4384] = {.lex_state = 106, .external_lex_state = 4},
  [4385] = {.lex_state = 271, .external_lex_state = 4},
  [4386] = {.lex_state = 106, .external_lex_state = 4},
  [4387] = {.lex_state = 133, .external_lex_state = 4},
  [4388] = {.lex_state = 106, .external_lex_state = 4},
  [4389] = {.lex_state = 106, .external_lex_state = 4},
  [4390] = {.lex_state = 106, .external_lex_state = 4},
  [4391] = {.lex_state = 106, .external_lex_state = 4},
  [4392] = {.lex_state = 271, .external_lex_state = 4},
  [4393] = {.lex_state = 271, .external_lex_state = 4},
  [4394] = {.lex_state = 271, .external_lex_state = 4},
  [4395] = {.lex_state = 106, .external_lex_state = 4},
  [4396] = {.lex_state = 271, .external_lex_state = 4},
  [4397] = {.lex_state = 106, .external_lex_state = 4},
  [4398] = {.lex_state = 271, .external_lex_state = 4},
  [4399] = {.lex_state = 271, .external_lex_state = 4},
  [4400] = {.lex_state = 271, .external_lex_state = 4},
  [4401] = {.lex_state = 145, .external_lex_state = 4},
  [4402] = {.lex_state = 145, .external_lex_state = 4},
  [4403] = {.lex_state = 145, .external_lex_state = 4},
  [4404] = {.lex_state = 145, .external_lex_state = 4},
  [4405] = {.lex_state = 271, .external_lex_state = 4},
  [4406] = {.lex_state = 271, .external_lex_state = 4},
  [4407] = {.lex_state = 145, .external_lex_state = 4},
  [4408] = {.lex_state = 145, .external_lex_state = 4},
  [4409] = {.lex_state = 145, .external_lex_state = 4},
  [4410] = {.lex_state = 145, .external_lex_state = 4},
  [4411] = {.lex_state = 145, .external_lex_state = 4},
  [4412] = {.lex_state = 145, .external_lex_state = 4},
  [4413] = {.lex_state = 145, .external_lex_state = 4},
  [4414] = {.lex_state = 145, .external_lex_state = 4},
  [4415] = {.lex_state = 271, .external_lex_state = 4},
  [4416] = {.lex_state = 271, .external_lex_state = 4},
  [4417] = {.lex_state = 145, .external_lex_state = 4},
  [4418] = {.lex_state = 145, .external_lex_state = 4},
  [4419] = {.lex_state = 145, .external_lex_state = 4},
  [4420] = {.lex_state = 145, .external_lex_state = 4},
  [4421] = {.lex_state = 133, .external_lex_state = 4},
  [4422] = {.lex_state = 133, .external_lex_state = 4},
  [4423] = {.lex_state = 145, .external_lex_state = 4},
  [4424] = {.lex_state = 145, .external_lex_state = 4},
  [4425] = {.lex_state = 133, .external_lex_state = 4},
  [4426] = {.lex_state = 271, .external_lex_state = 4},
  [4427] = {.lex_state = 271, .external_lex_state = 4},
  [4428] = {.lex_state = 145, .external_lex_state = 4},
  [4429] = {.lex_state = 145, .external_lex_state = 4},
  [4430] = {.lex_state = 145, .external_lex_state = 4},
  [4431] = {.lex_state = 271, .external_lex_state = 4},
  [4432] = {.lex_state = 133, .external_lex_state = 4},
  [4433] = {.lex_state = 271, .external_lex_state = 4},
  [4434] = {.lex_state = 133, .external_lex_state = 4},
  [4435] = {.lex_state = 133, .external_lex_state = 4},
  [4436] = {.lex_state = 133, .external_lex_state = 4},
  [4437] = {.lex_state = 133, .external_lex_state = 4},
  [4438] = {.lex_state = 271, .external_lex_state = 4},
  [4439] = {.lex_state = 133, .external_lex_state = 4},
  [4440] = {.lex_state = 133, .external_lex_state = 4},
  [4441] = {.lex_state = 133, .external_lex_state = 4},
  [4442] = {.lex_state = 133, .external_lex_state = 4},
  [4443] = {.lex_state = 133, .external_lex_state = 4},
  [4444] = {.lex_state = 271, .external_lex_state = 4},
  [4445] = {.lex_state = 106, .external_lex_state = 4},
  [4446] = {.lex_state = 271, .external_lex_state = 4},
  [4447] = {.lex_state = 271, .external_lex_state = 4},
  [4448] = {.lex_state = 271, .external_lex_state = 4},
  [4449] = {.lex_state = 133, .external_lex_state = 4},
  [4450] = {.lex_state = 271, .external_lex_state = 4},
  [4451] = {.lex_state = 271, .external_lex_state = 4},
  [4452] = {.lex_state = 271, .external_lex_state = 4},
  [4453] = {.lex_state = 271, .external_lex_state = 4},
  [4454] = {.lex_state = 133, .external_lex_state = 4},
  [4455] = {.lex_state = 133, .external_lex_state = 4},
  [4456] = {.lex_state = 271, .external_lex_state = 4},
  [4457] = {.lex_state = 271, .external_lex_state = 4},
  [4458] = {.lex_state = 271, .external_lex_state = 4},
  [4459] = {.lex_state = 133, .external_lex_state = 4},
  [4460] = {.lex_state = 133, .external_lex_state = 4},
  [4461] = {.lex_state = 133, .external_lex_state = 4},
  [4462] = {.lex_state = 133, .external_lex_state = 4},
  [4463] = {.lex_state = 133, .external_lex_state = 4},
  [4464] = {.lex_state = 133, .external_lex_state = 4},
  [4465] = {.lex_state = 133, .external_lex_state = 4},
  [4466] = {.lex_state = 133, .external_lex_state = 4},
  [4467] = {.lex_state = 133, .external_lex_state = 4},
  [4468] = {.lex_state = 133, .external_lex_state = 4},
  [4469] = {.lex_state = 133, .external_lex_state = 4},
  [4470] = {.lex_state = 133, .external_lex_state = 4},
  [4471] = {.lex_state = 133, .external_lex_state = 4},
  [4472] = {.lex_state = 133, .external_lex_state = 4},
  [4473] = {.lex_state = 133, .external_lex_state = 4},
  [4474] = {.lex_state = 133, .external_lex_state = 4},
  [4475] = {.lex_state = 133, .external_lex_state = 4},
  [4476] = {.lex_state = 133, .external_lex_state = 4},
  [4477] = {.lex_state = 133, .external_lex_state = 4},
  [4478] = {.lex_state = 133, .external_lex_state = 4},
  [4479] = {.lex_state = 271, .external_lex_state = 4},
  [4480] = {.lex_state = 133, .external_lex_state = 4},
  [4481] = {.lex_state = 133, .external_lex_state = 4},
  [4482] = {.lex_state = 133, .external_lex_state = 4},
  [4483] = {.lex_state = 133, .external_lex_state = 4},
  [4484] = {.lex_state = 133, .external_lex_state = 4},
  [4485] = {.lex_state = 271, .external_lex_state = 4},
  [4486] = {.lex_state = 133, .external_lex_state = 4},
  [4487] = {.lex_state = 133, .external_lex_state = 4},
  [4488] = {.lex_state = 133, .external_lex_state = 4},
  [4489] = {.lex_state = 133, .external_lex_state = 4},
  [4490] = {.lex_state = 133, .external_lex_state = 4},
  [4491] = {.lex_state = 133, .external_lex_state = 4},
  [4492] = {.lex_state = 133, .external_lex_state = 4},
  [4493] = {.lex_state = 271, .external_lex_state = 4},
  [4494] = {.lex_state = 133, .external_lex_state = 4},
  [4495] = {.lex_state = 133, .external_lex_state = 4},
  [4496] = {.lex_state = 271, .external_lex_state = 4},
  [4497] = {.lex_state = 133, .external_lex_state = 4},
  [4498] = {.lex_state = 133, .external_lex_state = 4},
  [4499] = {.lex_state = 133, .external_lex_state = 4},
  [4500] = {.lex_state = 133, .external_lex_state = 4},
  [4501] = {.lex_state = 133, .external_lex_state = 4},
  [4502] = {.lex_state = 133, .external_lex_state = 4},
  [4503] = {.lex_state = 133, .external_lex_state = 4},
  [4504] = {.lex_state = 133, .external_lex_state = 4},
  [4505] = {.lex_state = 133, .external_lex_state = 4},
  [4506] = {.lex_state = 133, .external_lex_state = 4},
  [4507] = {.lex_state = 133, .external_lex_state = 4},
  [4508] = {.lex_state = 133, .external_lex_state = 4},
  [4509] = {.lex_state = 133, .external_lex_state = 4},
  [4510] = {.lex_state = 133, .external_lex_state = 4},
  [4511] = {.lex_state = 133, .external_lex_state = 4},
  [4512] = {.lex_state = 133, .external_lex_state = 4},
  [4513] = {.lex_state = 133, .external_lex_state = 4},
  [4514] = {.lex_state = 271, .external_lex_state = 4},
  [4515] = {.lex_state = 133, .external_lex_state = 4},
  [4516] = {.lex_state = 133, .external_lex_state = 4},
  [4517] = {.lex_state = 133, .external_lex_state = 4},
  [4518] = {.lex_state = 133, .external_lex_state = 4},
  [4519] = {.lex_state = 106, .external_lex_state = 4},
  [4520] = {.lex_state = 133, .external_lex_state = 4},
  [4521] = {.lex_state = 133, .external_lex_state = 4},
  [4522] = {.lex_state = 133, .external_lex_state = 4},
  [4523] = {.lex_state = 133, .external_lex_state = 4},
  [4524] = {.lex_state = 133, .external_lex_state = 4},
  [4525] = {.lex_state = 133, .external_lex_state = 4},
  [4526] = {.lex_state = 133, .external_lex_state = 4},
  [4527] = {.lex_state = 133, .external_lex_state = 4},
  [4528] = {.lex_state = 133, .external_lex_state = 4},
  [4529] = {.lex_state = 133, .external_lex_state = 4},
  [4530] = {.lex_state = 133, .external_lex_state = 4},
  [4531] = {.lex_state = 133, .external_lex_state = 4},
  [4532] = {.lex_state = 133, .external_lex_state = 4},
  [4533] = {.lex_state = 133, .external_lex_state = 4},
  [4534] = {.lex_state = 271, .external_lex_state = 4},
  [4535] = {.lex_state = 133, .external_lex_state = 4},
  [4536] = {.lex_state = 271, .external_lex_state = 4},
  [4537] = {.lex_state = 271, .external_lex_state = 4},
  [4538] = {.lex_state = 271, .external_lex_state = 4},
  [4539] = {.lex_state = 271, .external_lex_state = 4},
  [4540] = {.lex_state = 133, .external_lex_state = 4},
  [4541] = {.lex_state = 133, .external_lex_state = 4},
  [4542] = {.lex_state = 271, .external_lex_state = 4},
  [4543] = {.lex_state = 106, .external_lex_state = 4},
  [4544] = {.lex_state = 133, .external_lex_state = 4},
  [4545] = {.lex_state = 133, .external_lex_state = 4},
  [4546] = {.lex_state = 133, .external_lex_state = 4},
  [4547] = {.lex_state = 133, .external_lex_state = 4},
  [4548] = {.lex_state = 133, .external_lex_state = 4},
  [4549] = {.lex_state = 133, .external_lex_state = 4},
  [4550] = {.lex_state = 133, .external_lex_state = 4},
  [4551] = {.lex_state = 133, .external_lex_state = 4},
  [4552] = {.lex_state = 133, .external_lex_state = 4},
  [4553] = {.lex_state = 133, .external_lex_state = 4},
  [4554] = {.lex_state = 133, .external_lex_state = 4},
  [4555] = {.lex_state = 133, .external_lex_state = 4},
  [4556] = {.lex_state = 133, .external_lex_state = 4},
  [4557] = {.lex_state = 133, .external_lex_state = 4},
  [4558] = {.lex_state = 133, .external_lex_state = 4},
  [4559] = {.lex_state = 133, .external_lex_state = 4},
  [4560] = {.lex_state = 133, .external_lex_state = 4},
  [4561] = {.lex_state = 133, .external_lex_state = 4},
  [4562] = {.lex_state = 133, .external_lex_state = 4},
  [4563] = {.lex_state = 133, .external_lex_state = 4},
  [4564] = {.lex_state = 133, .external_lex_state = 4},
  [4565] = {.lex_state = 133, .external_lex_state = 4},
  [4566] = {.lex_state = 133, .external_lex_state = 4},
  [4567] = {.lex_state = 133, .external_lex_state = 4},
  [4568] = {.lex_state = 271, .external_lex_state = 4},
  [4569] = {.lex_state = 271, .external_lex_state = 4},
  [4570] = {.lex_state = 271, .external_lex_state = 4},
  [4571] = {.lex_state = 271, .external_lex_state = 4},
  [4572] = {.lex_state = 271, .external_lex_state = 4},
  [4573] = {.lex_state = 271, .external_lex_state = 4},
  [4574] = {.lex_state = 271, .external_lex_state = 4},
  [4575] = {.lex_state = 271, .external_lex_state = 4},
  [4576] = {.lex_state = 271, .external_lex_state = 4},
  [4577] = {.lex_state = 271, .external_lex_state = 4},
  [4578] = {.lex_state = 271, .external_lex_state = 4},
  [4579] = {.lex_state = 106, .external_lex_state = 4},
  [4580] = {.lex_state = 106, .external_lex_state = 4},
  [4581] = {.lex_state = 271, .external_lex_state = 4},
  [4582] = {.lex_state = 271, .external_lex_state = 4},
  [4583] = {.lex_state = 271, .external_lex_state = 4},
  [4584] = {.lex_state = 271, .external_lex_state = 4},
  [4585] = {.lex_state = 271, .external_lex_state = 4},
  [4586] = {.lex_state = 271, .external_lex_state = 4},
  [4587] = {.lex_state = 271, .external_lex_state = 4},
  [4588] = {.lex_state = 271, .external_lex_state = 4},
  [4589] = {.lex_state = 271, .external_lex_state = 4},
  [4590] = {.lex_state = 271, .external_lex_state = 4},
  [4591] = {.lex_state = 271, .external_lex_state = 4},
  [4592] = {.lex_state = 271, .external_lex_state = 4},
  [4593] = {.lex_state = 271, .external_lex_state = 4},
  [4594] = {.lex_state = 271, .external_lex_state = 4},
  [4595] = {.lex_state = 52, .external_lex_state = 6},
  [4596] = {.lex_state = 52, .external_lex_state = 6},
  [4597] = {.lex_state = 52, .external_lex_state = 6},
  [4598] = {.lex_state = 52, .external_lex_state = 6},
  [4599] = {.lex_state = 52, .external_lex_state = 6},
  [4600] = {.lex_state = 52, .external_lex_state = 6},
  [4601] = {.lex_state = 52, .external_lex_state = 6},
  [4602] = {.lex_state = 52, .external_lex_state = 6},
  [4603] = {.lex_state = 52, .external_lex_state = 6},
  [4604] = {.lex_state = 52, .external_lex_state = 6},
  [4605] = {.lex_state = 52, .external_lex_state = 6},
  [4606] = {.lex_state = 52, .external_lex_state = 6},
  [4607] = {.lex_state = 52, .external_lex_state = 6},
  [4608] = {.lex_state = 52, .external_lex_state = 6},
  [4609] = {.lex_state = 52, .external_lex_state = 6},
  [4610] = {.lex_state = 52, .external_lex_state = 6},
  [4611] = {.lex_state = 52, .external_lex_state = 6},
  [4612] = {.lex_state = 52, .external_lex_state = 6},
  [4613] = {.lex_state = 52, .external_lex_state = 6},
  [4614] = {.lex_state = 52, .external_lex_state = 6},
  [4615] = {.lex_state = 52, .external_lex_state = 6},
  [4616] = {.lex_state = 52, .external_lex_state = 6},
  [4617] = {.lex_state = 52, .external_lex_state = 6},
  [4618] = {.lex_state = 52, .external_lex_state = 6},
  [4619] = {.lex_state = 52, .external_lex_state = 6},
  [4620] = {.lex_state = 52, .external_lex_state = 6},
  [4621] = {.lex_state = 52, .external_lex_state = 6},
  [4622] = {.lex_state = 52, .external_lex_state = 6},
  [4623] = {.lex_state = 52, .external_lex_state = 6},
  [4624] = {.lex_state = 52, .external_lex_state = 6},
  [4625] = {.lex_state = 52, .external_lex_state = 6},
  [4626] = {.lex_state = 52, .external_lex_state = 6},
  [4627] = {.lex_state = 52, .external_lex_state = 6},
  [4628] = {.lex_state = 52, .external_lex_state = 6},
  [4629] = {.lex_state = 52, .external_lex_state = 6},
  [4630] = {.lex_state = 52, .external_lex_state = 6},
  [4631] = {.lex_state = 106, .external_lex_state = 6},
  [4632] = {.lex_state = 106, .external_lex_state = 6},
  [4633] = {.lex_state = 106, .external_lex_state = 6},
  [4634] = {.lex_state = 106, .external_lex_state = 6},
  [4635] = {.lex_state = 106, .external_lex_state = 6},
  [4636] = {.lex_state = 106, .external_lex_state = 6},
  [4637] = {.lex_state = 106, .external_lex_state = 6},
  [4638] = {.lex_state = 106, .external_lex_state = 6},
  [4639] = {.lex_state = 106, .external_lex_state = 6},
  [4640] = {.lex_state = 106, .external_lex_state = 6},
  [4641] = {.lex_state = 106, .external_lex_state = 6},
  [4642] = {.lex_state = 106, .external_lex_state = 6},
  [4643] = {.lex_state = 106, .external_lex_state = 6},
  [4644] = {.lex_state = 106, .external_lex_state = 6},
  [4645] = {.lex_state = 106, .external_lex_state = 6},
  [4646] = {.lex_state = 106, .external_lex_state = 6},
  [4647] = {.lex_state = 106, .external_lex_state = 6},
  [4648] = {.lex_state = 106, .external_lex_state = 6},
  [4649] = {.lex_state = 106, .external_lex_state = 6},
  [4650] = {.lex_state = 106, .external_lex_state = 6},
  [4651] = {.lex_state = 106, .external_lex_state = 6},
  [4652] = {.lex_state = 106, .external_lex_state = 6},
  [4653] = {.lex_state = 106, .external_lex_state = 6},
  [4654] = {.lex_state = 106, .external_lex_state = 6},
  [4655] = {.lex_state = 106, .external_lex_state = 6},
  [4656] = {.lex_state = 106, .external_lex_state = 6},
  [4657] = {.lex_state = 106, .external_lex_state = 6},
  [4658] = {.lex_state = 106, .external_lex_state = 6},
  [4659] = {.lex_state = 106, .external_lex_state = 6},
  [4660] = {.lex_state = 106, .external_lex_state = 6},
  [4661] = {.lex_state = 106, .external_lex_state = 6},
  [4662] = {.lex_state = 106, .external_lex_state = 6},
  [4663] = {.lex_state = 106, .external_lex_state = 6},
  [4664] = {.lex_state = 106, .external_lex_state = 6},
  [4665] = {.lex_state = 106, .external_lex_state = 6},
  [4666] = {.lex_state = 106, .external_lex_state = 6},
  [4667] = {.lex_state = 106, .external_lex_state = 6},
  [4668] = {.lex_state = 106, .external_lex_state = 6},
  [4669] = {.lex_state = 106, .external_lex_state = 6},
  [4670] = {.lex_state = 106, .external_lex_state = 6},
  [4671] = {.lex_state = 106, .external_lex_state = 6},
  [4672] = {.lex_state = 106, .external_lex_state = 6},
  [4673] = {.lex_state = 106, .external_lex_state = 6},
  [4674] = {.lex_state = 106, .external_lex_state = 6},
  [4675] = {.lex_state = 106, .external_lex_state = 6},
  [4676] = {.lex_state = 106, .external_lex_state = 6},
  [4677] = {.lex_state = 106, .external_lex_state = 6},
  [4678] = {.lex_state = 106, .external_lex_state = 6},
  [4679] = {.lex_state = 106, .external_lex_state = 6},
  [4680] = {.lex_state = 106, .external_lex_state = 6},
  [4681] = {.lex_state = 106, .external_lex_state = 6},
  [4682] = {.lex_state = 106, .external_lex_state = 6},
  [4683] = {.lex_state = 106, .external_lex_state = 6},
  [4684] = {.lex_state = 106, .external_lex_state = 6},
  [4685] = {.lex_state = 106, .external_lex_state = 6},
  [4686] = {.lex_state = 106, .external_lex_state = 6},
  [4687] = {.lex_state = 106, .external_lex_state = 6},
  [4688] = {.lex_state = 106, .external_lex_state = 6},
  [4689] = {.lex_state = 106, .external_lex_state = 6},
  [4690] = {.lex_state = 106, .external_lex_state = 6},
  [4691] = {.lex_state = 106, .external_lex_state = 6},
  [4692] = {.lex_state = 106, .external_lex_state = 6},
  [4693] = {.lex_state = 106, .external_lex_state = 6},
  [4694] = {.lex_state = 106, .external_lex_state = 6},
  [4695] = {.lex_state = 106, .external_lex_state = 6},
  [4696] = {.lex_state = 106, .external_lex_state = 6},
  [4697] = {.lex_state = 106, .external_lex_state = 6},
  [4698] = {.lex_state = 106, .external_lex_state = 6},
  [4699] = {.lex_state = 106, .external_lex_state = 6},
  [4700] = {.lex_state = 106, .external_lex_state = 6},
  [4701] = {.lex_state = 106, .external_lex_state = 6},
  [4702] = {.lex_state = 106, .external_lex_state = 6},
  [4703] = {.lex_state = 148, .external_lex_state = 6},
  [4704] = {.lex_state = 148, .external_lex_state = 6},
  [4705] = {.lex_state = 106, .external_lex_state = 6},
  [4706] = {.lex_state = 106, .external_lex_state = 6},
  [4707] = {.lex_state = 106, .external_lex_state = 6},
  [4708] = {.lex_state = 106, .external_lex_state = 6},
  [4709] = {.lex_state = 106, .external_lex_state = 6},
  [4710] = {.lex_state = 106, .external_lex_state = 6},
  [4711] = {.lex_state = 106, .external_lex_state = 6},
  [4712] = {.lex_state = 106, .external_lex_state = 6},
  [4713] = {.lex_state = 106, .external_lex_state = 6},
  [4714] = {.lex_state = 106, .external_lex_state = 6},
  [4715] = {.lex_state = 106, .external_lex_state = 6},
  [4716] = {.lex_state = 106, .external_lex_state = 6},
  [4717] = {.lex_state = 106, .external_lex_state = 6},
  [4718] = {.lex_state = 106, .external_lex_state = 6},
  [4719] = {.lex_state = 106, .external_lex_state = 6},
  [4720] = {.lex_state = 106, .external_lex_state = 6},
  [4721] = {.lex_state = 106, .external_lex_state = 6},
  [4722] = {.lex_state = 106, .external_lex_state = 6},
  [4723] = {.lex_state = 106, .external_lex_state = 6},
  [4724] = {.lex_state = 106, .external_lex_state = 6},
  [4725] = {.lex_state = 106, .external_lex_state = 6},
  [4726] = {.lex_state = 106, .external_lex_state = 6},
  [4727] = {.lex_state = 106, .external_lex_state = 6},
  [4728] = {.lex_state = 106, .external_lex_state = 6},
  [4729] = {.lex_state = 106, .external_lex_state = 6},
  [4730] = {.lex_state = 106, .external_lex_state = 6},
  [4731] = {.lex_state = 106, .external_lex_state = 6},
  [4732] = {.lex_state = 106, .external_lex_state = 6},
  [4733] = {.lex_state = 106, .external_lex_state = 6},
  [4734] = {.lex_state = 106, .external_lex_state = 6},
  [4735] = {.lex_state = 106, .external_lex_state = 6},
  [4736] = {.lex_state = 106, .external_lex_state = 6},
  [4737] = {.lex_state = 106, .external_lex_state = 6},
  [4738] = {.lex_state = 106, .external_lex_state = 6},
  [4739] = {.lex_state = 106, .external_lex_state = 6},
  [4740] = {.lex_state = 106, .external_lex_state = 6},
  [4741] = {.lex_state = 106, .external_lex_state = 6},
  [4742] = {.lex_state = 106, .external_lex_state = 6},
  [4743] = {.lex_state = 106, .external_lex_state = 6},
  [4744] = {.lex_state = 106, .external_lex_state = 6},
  [4745] = {.lex_state = 106, .external_lex_state = 6},
  [4746] = {.lex_state = 106, .external_lex_state = 6},
  [4747] = {.lex_state = 106, .external_lex_state = 6},
  [4748] = {.lex_state = 106, .external_lex_state = 6},
  [4749] = {.lex_state = 106, .external_lex_state = 6},
  [4750] = {.lex_state = 106, .external_lex_state = 6},
  [4751] = {.lex_state = 106, .external_lex_state = 6},
  [4752] = {.lex_state = 106, .external_lex_state = 6},
  [4753] = {.lex_state = 106, .external_lex_state = 6},
  [4754] = {.lex_state = 106, .external_lex_state = 6},
  [4755] = {.lex_state = 106, .external_lex_state = 6},
  [4756] = {.lex_state = 106, .external_lex_state = 6},
  [4757] = {.lex_state = 106, .external_lex_state = 6},
  [4758] = {.lex_state = 106, .external_lex_state = 6},
  [4759] = {.lex_state = 106, .external_lex_state = 6},
  [4760] = {.lex_state = 106, .external_lex_state = 6},
  [4761] = {.lex_state = 106, .external_lex_state = 6},
  [4762] = {.lex_state = 106, .external_lex_state = 6},
  [4763] = {.lex_state = 106, .external_lex_state = 6},
  [4764] = {.lex_state = 106, .external_lex_state = 6},
  [4765] = {.lex_state = 106, .external_lex_state = 6},
  [4766] = {.lex_state = 106, .external_lex_state = 6},
  [4767] = {.lex_state = 106, .external_lex_state = 6},
  [4768] = {.lex_state = 106, .external_lex_state = 6},
  [4769] = {.lex_state = 106, .external_lex_state = 6},
  [4770] = {.lex_state = 106, .external_lex_state = 6},
  [4771] = {.lex_state = 106, .external_lex_state = 6},
  [4772] = {.lex_state = 106, .external_lex_state = 6},
  [4773] = {.lex_state = 106, .external_lex_state = 6},
  [4774] = {.lex_state = 106, .external_lex_state = 6},
  [4775] = {.lex_state = 106, .external_lex_state = 6},
  [4776] = {.lex_state = 106, .external_lex_state = 6},
  [4777] = {.lex_state = 106, .external_lex_state = 6},
  [4778] = {.lex_state = 106, .external_lex_state = 6},
  [4779] = {.lex_state = 106, .external_lex_state = 6},
  [4780] = {.lex_state = 106, .external_lex_state = 6},
  [4781] = {.lex_state = 106, .external_lex_state = 6},
  [4782] = {.lex_state = 106, .external_lex_state = 6},
  [4783] = {.lex_state = 106, .external_lex_state = 6},
  [4784] = {.lex_state = 106, .external_lex_state = 6},
  [4785] = {.lex_state = 106, .external_lex_state = 6},
  [4786] = {.lex_state = 106, .external_lex_state = 6},
  [4787] = {.lex_state = 106, .external_lex_state = 6},
  [4788] = {.lex_state = 106, .external_lex_state = 6},
  [4789] = {.lex_state = 106, .external_lex_state = 6},
  [4790] = {.lex_state = 148, .external_lex_state = 6},
  [4791] = {.lex_state = 106, .external_lex_state = 6},
  [4792] = {.lex_state = 148, .external_lex_state = 6},
  [4793] = {.lex_state = 106, .external_lex_state = 6},
  [4794] = {.lex_state = 106, .external_lex_state = 6},
  [4795] = {.lex_state = 106, .external_lex_state = 6},
  [4796] = {.lex_state = 106, .external_lex_state = 6},
  [4797] = {.lex_state = 52, .external_lex_state = 6},
  [4798] = {.lex_state = 52, .external_lex_state = 6},
  [4799] = {.lex_state = 52, .external_lex_state = 6},
  [4800] = {.lex_state = 148, .external_lex_state = 6},
  [4801] = {.lex_state = 148, .external_lex_state = 6},
  [4802] = {.lex_state = 52, .external_lex_state = 6},
  [4803] = {.lex_state = 148, .external_lex_state = 6},
  [4804] = {.lex_state = 148, .external_lex_state = 6},
  [4805] = {.lex_state = 148, .external_lex_state = 6},
  [4806] = {.lex_state = 52, .external_lex_state = 6},
  [4807] = {.lex_state = 148, .external_lex_state = 6},
  [4808] = {.lex_state = 148, .external_lex_state = 6},
  [4809] = {.lex_state = 148, .external_lex_state = 6},
  [4810] = {.lex_state = 52, .external_lex_state = 6},
  [4811] = {.lex_state = 148, .external_lex_state = 6},
  [4812] = {.lex_state = 148, .external_lex_state = 6},
  [4813] = {.lex_state = 52, .external_lex_state = 6},
  [4814] = {.lex_state = 52, .external_lex_state = 6},
  [4815] = {.lex_state = 52, .external_lex_state = 6},
  [4816] = {.lex_state = 52, .external_lex_state = 6},
  [4817] = {.lex_state = 52, .external_lex_state = 6},
  [4818] = {.lex_state = 52, .external_lex_state = 6},
  [4819] = {.lex_state = 148, .external_lex_state = 6},
  [4820] = {.lex_state = 148, .external_lex_state = 6},
  [4821] = {.lex_state = 52, .external_lex_state = 6},
  [4822] = {.lex_state = 52, .external_lex_state = 6},
  [4823] = {.lex_state = 52, .external_lex_state = 6},
  [4824] = {.lex_state = 148, .external_lex_state = 6},
  [4825] = {.lex_state = 148, .external_lex_state = 6},
  [4826] = {.lex_state = 52, .external_lex_state = 6},
  [4827] = {.lex_state = 148, .external_lex_state = 6},
  [4828] = {.lex_state = 52, .external_lex_state = 6},
  [4829] = {.lex_state = 148, .external_lex_state = 6},
  [4830] = {.lex_state = 148, .external_lex_state = 6},
  [4831] = {.lex_state = 148, .external_lex_state = 6},
  [4832] = {.lex_state = 148, .external_lex_state = 6},
  [4833] = {.lex_state = 148, .external_lex_state = 6},
  [4834] = {.lex_state = 148, .external_lex_state = 6},
  [4835] = {.lex_state = 148, .external_lex_state = 6},
  [4836] = {.lex_state = 52, .external_lex_state = 6},
  [4837] = {.lex_state = 52, .external_lex_state = 6},
  [4838] = {.lex_state = 52, .external_lex_state = 6},
  [4839] = {.lex_state = 148, .external_lex_state = 6},
  [4840] = {.lex_state = 53, .external_lex_state = 6},
  [4841] = {.lex_state = 106, .external_lex_state = 6},
  [4842] = {.lex_state = 106, .external_lex_state = 6},
  [4843] = {.lex_state = 106, .external_lex_state = 6},
  [4844] = {.lex_state = 64, .external_lex_state = 7},
  [4845] = {.lex_state = 64, .external_lex_state = 8},
  [4846] = {.lex_state = 66, .external_lex_state = 9},
  [4847] = {.lex_state = 66, .external_lex_state = 10},
  [4848] = {.lex_state = 64, .external_lex_state = 11},
  [4849] = {.lex_state = 64, .external_lex_state = 12},
  [4850] = {.lex_state = 64, .external_lex_state = 13},
  [4851] = {.lex_state = 64, .external_lex_state = 11},
  [4852] = {.lex_state = 64, .external_lex_state = 12},
  [4853] = {.lex_state = 64, .external_lex_state = 13},
  [4854] = {.lex_state = 64, .external_lex_state = 14},
  [4855] = {.lex_state = 64, .external_lex_state = 15},
  [4856] = {.lex_state = 64, .external_lex_state = 16},
  [4857] = {.lex_state = 64, .external_lex_state = 7},
  [4858] = {.lex_state = 64, .external_lex_state = 8},
  [4859] = {.lex_state = 66, .external_lex_state = 9},
  [4860] = {.lex_state = 64, .external_lex_state = 7},
  [4861] = {.lex_state = 64, .external_lex_state = 8},
  [4862] = {.lex_state = 66, .external_lex_state = 9},
  [4863] = {.lex_state = 66, .external_lex_state = 10},
  [4864] = {.lex_state = 66, .external_lex_state = 10},
  [4865] = {.lex_state = 64, .external_lex_state = 7},
  [4866] = {.lex_state = 64, .external_lex_state = 8},
  [4867] = {.lex_state = 66, .external_lex_state = 9},
  [4868] = {.lex_state = 64, .external_lex_state = 11},
  [4869] = {.lex_state = 64, .external_lex_state = 12},
  [4870] = {.lex_state = 64, .external_lex_state = 13},
  [4871] = {.lex_state = 64, .external_lex_state = 14},
  [4872] = {.lex_state = 64, .external_lex_state = 15},
  [4873] = {.lex_state = 64, .external_lex_state = 16},
  [4874] = {.lex_state = 66, .external_lex_state = 10},
  [4875] = {.lex_state = 64, .external_lex_state = 14},
  [4876] = {.lex_state = 64, .external_lex_state = 11},
  [4877] = {.lex_state = 64, .external_lex_state = 12},
  [4878] = {.lex_state = 64, .external_lex_state = 13},
  [4879] = {.lex_state = 64, .external_lex_state = 11},
  [4880] = {.lex_state = 64, .external_lex_state = 12},
  [4881] = {.lex_state = 64, .external_lex_state = 13},
  [4882] = {.lex_state = 64, .external_lex_state = 14},
  [4883] = {.lex_state = 64, .external_lex_state = 15},
  [4884] = {.lex_state = 64, .external_lex_state = 16},
  [4885] = {.lex_state = 64, .external_lex_state = 14},
  [4886] = {.lex_state = 64, .external_lex_state = 15},
  [4887] = {.lex_state = 64, .external_lex_state = 7},
  [4888] = {.lex_state = 64, .external_lex_state = 8},
  [4889] = {.lex_state = 66, .external_lex_state = 9},
  [4890] = {.lex_state = 66, .external_lex_state = 10},
  [4891] = {.lex_state = 64, .external_lex_state = 16},
  [4892] = {.lex_state = 64, .external_lex_state = 15},
  [4893] = {.lex_state = 64, .external_lex_state = 16},
  [4894] = {.lex_state = 64, .external_lex_state = 11},
  [4895] = {.lex_state = 64, .external_lex_state = 12},
  [4896] = {.lex_state = 64, .external_lex_state = 13},
  [4897] = {.lex_state = 64, .external_lex_state = 14},
  [4898] = {.lex_state = 64, .external_lex_state = 15},
  [4899] = {.lex_state = 64, .external_lex_state = 16},
  [4900] = {.lex_state = 64, .external_lex_state = 11},
  [4901] = {.lex_state = 64, .external_lex_state = 12},
  [4902] = {.lex_state = 64, .external_lex_state = 13},
  [4903] = {.lex_state = 64, .external_lex_state = 14},
  [4904] = {.lex_state = 64, .external_lex_state = 15},
  [4905] = {.lex_state = 64, .external_lex_state = 16},
  [4906] = {.lex_state = 64, .external_lex_state = 11},
  [4907] = {.lex_state = 64, .external_lex_state = 15},
  [4908] = {.lex_state = 64, .external_lex_state = 7},
  [4909] = {.lex_state = 64, .external_lex_state = 8},
  [4910] = {.lex_state = 64, .external_lex_state = 7},
  [4911] = {.lex_state = 64, .external_lex_state = 8},
  [4912] = {.lex_state = 66, .external_lex_state = 9},
  [4913] = {.lex_state = 66, .external_lex_state = 10},
  [4914] = {.lex_state = 66, .external_lex_state = 9},
  [4915] = {.lex_state = 66, .external_lex_state = 10},
  [4916] = {.lex_state = 64, .external_lex_state = 14},
  [4917] = {.lex_state = 64, .external_lex_state = 7},
  [4918] = {.lex_state = 64, .external_lex_state = 8},
  [4919] = {.lex_state = 66, .external_lex_state = 9},
  [4920] = {.lex_state = 66, .external_lex_state = 10},
  [4921] = {.lex_state = 64, .external_lex_state = 7},
  [4922] = {.lex_state = 64, .external_lex_state = 16},
  [4923] = {.lex_state = 64, .external_lex_state = 12},
  [4924] = {.lex_state = 64, .external_lex_state = 11},
  [4925] = {.lex_state = 64, .external_lex_state = 12},
  [4926] = {.lex_state = 64, .external_lex_state = 13},
  [4927] = {.lex_state = 64, .external_lex_state = 14},
  [4928] = {.lex_state = 64, .external_lex_state = 15},
  [4929] = {.lex_state = 64, .external_lex_state = 16},
  [4930] = {.lex_state = 64, .external_lex_state = 11},
  [4931] = {.lex_state = 64, .external_lex_state = 7},
  [4932] = {.lex_state = 64, .external_lex_state = 8},
  [4933] = {.lex_state = 66, .external_lex_state = 9},
  [4934] = {.lex_state = 66, .external_lex_state = 10},
  [4935] = {.lex_state = 64, .external_lex_state = 12},
  [4936] = {.lex_state = 64, .external_lex_state = 13},
  [4937] = {.lex_state = 64, .external_lex_state = 14},
  [4938] = {.lex_state = 64, .external_lex_state = 11},
  [4939] = {.lex_state = 64, .external_lex_state = 12},
  [4940] = {.lex_state = 64, .external_lex_state = 13},
  [4941] = {.lex_state = 64, .external_lex_state = 14},
  [4942] = {.lex_state = 64, .external_lex_state = 15},
  [4943] = {.lex_state = 64, .external_lex_state = 16},
  [4944] = {.lex_state = 64, .external_lex_state = 15},
  [4945] = {.lex_state = 64, .external_lex_state = 7},
  [4946] = {.lex_state = 64, .external_lex_state = 8},
  [4947] = {.lex_state = 66, .external_lex_state = 9},
  [4948] = {.lex_state = 66, .external_lex_state = 10},
  [4949] = {.lex_state = 64, .external_lex_state = 16},
  [4950] = {.lex_state = 64, .external_lex_state = 7},
  [4951] = {.lex_state = 64, .external_lex_state = 8},
  [4952] = {.lex_state = 64, .external_lex_state = 11},
  [4953] = {.lex_state = 64, .external_lex_state = 12},
  [4954] = {.lex_state = 64, .external_lex_state = 13},
  [4955] = {.lex_state = 64, .external_lex_state = 14},
  [4956] = {.lex_state = 64, .external_lex_state = 15},
  [4957] = {.lex_state = 64, .external_lex_state = 16},
  [4958] = {.lex_state = 66, .external_lex_state = 9},
  [4959] = {.lex_state = 64, .external_lex_state = 7},
  [4960] = {.lex_state = 64, .external_lex_state = 8},
  [4961] = {.lex_state = 66, .external_lex_state = 9},
  [4962] = {.lex_state = 66, .external_lex_state = 10},
  [4963] = {.lex_state = 64, .external_lex_state = 7},
  [4964] = {.lex_state = 64, .external_lex_state = 8},
  [4965] = {.lex_state = 66, .external_lex_state = 9},
  [4966] = {.lex_state = 64, .external_lex_state = 11},
  [4967] = {.lex_state = 64, .external_lex_state = 12},
  [4968] = {.lex_state = 64, .external_lex_state = 13},
  [4969] = {.lex_state = 64, .external_lex_state = 14},
  [4970] = {.lex_state = 64, .external_lex_state = 15},
  [4971] = {.lex_state = 64, .external_lex_state = 16},
  [4972] = {.lex_state = 66, .external_lex_state = 10},
  [4973] = {.lex_state = 64, .external_lex_state = 7},
  [4974] = {.lex_state = 64, .external_lex_state = 8},
  [4975] = {.lex_state = 66, .external_lex_state = 9},
  [4976] = {.lex_state = 64, .external_lex_state = 7},
  [4977] = {.lex_state = 64, .external_lex_state = 7},
  [4978] = {.lex_state = 64, .external_lex_state = 8},
  [4979] = {.lex_state = 64, .external_lex_state = 11},
  [4980] = {.lex_state = 64, .external_lex_state = 11},
  [4981] = {.lex_state = 64, .external_lex_state = 12},
  [4982] = {.lex_state = 64, .external_lex_state = 13},
  [4983] = {.lex_state = 64, .external_lex_state = 14},
  [4984] = {.lex_state = 64, .external_lex_state = 15},
  [4985] = {.lex_state = 64, .external_lex_state = 16},
  [4986] = {.lex_state = 64, .external_lex_state = 12},
  [4987] = {.lex_state = 64, .external_lex_state = 7},
  [4988] = {.lex_state = 64, .external_lex_state = 8},
  [4989] = {.lex_state = 66, .external_lex_state = 9},
  [4990] = {.lex_state = 66, .external_lex_state = 10},
  [4991] = {.lex_state = 66, .external_lex_state = 9},
  [4992] = {.lex_state = 64, .external_lex_state = 11},
  [4993] = {.lex_state = 64, .external_lex_state = 12},
  [4994] = {.lex_state = 64, .external_lex_state = 11},
  [4995] = {.lex_state = 64, .external_lex_state = 12},
  [4996] = {.lex_state = 64, .external_lex_state = 13},
  [4997] = {.lex_state = 64, .external_lex_state = 14},
  [4998] = {.lex_state = 64, .external_lex_state = 15},
  [4999] = {.lex_state = 64, .external_lex_state = 16},
  [5000] = {.lex_state = 64, .external_lex_state = 13},
  [5001] = {.lex_state = 64, .external_lex_state = 7},
  [5002] = {.lex_state = 64, .external_lex_state = 8},
  [5003] = {.lex_state = 66, .external_lex_state = 9},
  [5004] = {.lex_state = 66, .external_lex_state = 10},
  [5005] = {.lex_state = 64, .external_lex_state = 13},
  [5006] = {.lex_state = 64, .external_lex_state = 7},
  [5007] = {.lex_state = 64, .external_lex_state = 8},
  [5008] = {.lex_state = 66, .external_lex_state = 9},
  [5009] = {.lex_state = 66, .external_lex_state = 10},
  [5010] = {.lex_state = 64, .external_lex_state = 7},
  [5011] = {.lex_state = 64, .external_lex_state = 8},
  [5012] = {.lex_state = 66, .external_lex_state = 9},
  [5013] = {.lex_state = 66, .external_lex_state = 10},
  [5014] = {.lex_state = 64, .external_lex_state = 7},
  [5015] = {.lex_state = 64, .external_lex_state = 8},
  [5016] = {.lex_state = 66, .external_lex_state = 9},
  [5017] = {.lex_state = 66, .external_lex_state = 10},
  [5018] = {.lex_state = 64, .external_lex_state = 7},
  [5019] = {.lex_state = 64, .external_lex_state = 8},
  [5020] = {.lex_state = 66, .external_lex_state = 9},
  [5021] = {.lex_state = 66, .external_lex_state = 10},
  [5022] = {.lex_state = 64, .external_lex_state = 7},
  [5023] = {.lex_state = 64, .external_lex_state = 8},
  [5024] = {.lex_state = 66, .external_lex_state = 9},
  [5025] = {.lex_state = 66, .external_lex_state = 10},
  [5026] = {.lex_state = 64, .external_lex_state = 7},
  [5027] = {.lex_state = 64, .external_lex_state = 8},
  [5028] = {.lex_state = 66, .external_lex_state = 9},
  [5029] = {.lex_state = 66, .external_lex_state = 10},
  [5030] = {.lex_state = 64, .external_lex_state = 7},
  [5031] = {.lex_state = 64, .external_lex_state = 8},
  [5032] = {.lex_state = 66, .external_lex_state = 9},
  [5033] = {.lex_state = 66, .external_lex_state = 10},
  [5034] = {.lex_state = 64, .external_lex_state = 7},
  [5035] = {.lex_state = 64, .external_lex_state = 8},
  [5036] = {.lex_state = 66, .external_lex_state = 9},
  [5037] = {.lex_state = 66, .external_lex_state = 10},
  [5038] = {.lex_state = 64, .external_lex_state = 7},
  [5039] = {.lex_state = 64, .external_lex_state = 8},
  [5040] = {.lex_state = 66, .external_lex_state = 9},
  [5041] = {.lex_state = 66, .external_lex_state = 10},
  [5042] = {.lex_state = 64, .external_lex_state = 7},
  [5043] = {.lex_state = 64, .external_lex_state = 8},
  [5044] = {.lex_state = 66, .external_lex_state = 9},
  [5045] = {.lex_state = 66, .external_lex_state = 10},
  [5046] = {.lex_state = 64, .external_lex_state = 7},
  [5047] = {.lex_state = 64, .external_lex_state = 8},
  [5048] = {.lex_state = 66, .external_lex_state = 9},
  [5049] = {.lex_state = 66, .external_lex_state = 10},
  [5050] = {.lex_state = 64, .external_lex_state = 7},
  [5051] = {.lex_state = 64, .external_lex_state = 8},
  [5052] = {.lex_state = 64, .external_lex_state = 14},
  [5053] = {.lex_state = 64, .external_lex_state = 8},
  [5054] = {.lex_state = 64, .external_lex_state = 7},
  [5055] = {.lex_state = 64, .external_lex_state = 8},
  [5056] = {.lex_state = 64, .external_lex_state = 7},
  [5057] = {.lex_state = 64, .external_lex_state = 8},
  [5058] = {.lex_state = 64, .external_lex_state = 7},
  [5059] = {.lex_state = 64, .external_lex_state = 8},
  [5060] = {.lex_state = 64, .external_lex_state = 7},
  [5061] = {.lex_state = 64, .external_lex_state = 8},
  [5062] = {.lex_state = 64, .external_lex_state = 7},
  [5063] = {.lex_state = 64, .external_lex_state = 8},
  [5064] = {.lex_state = 64, .external_lex_state = 14},
  [5065] = {.lex_state = 64, .external_lex_state = 15},
  [5066] = {.lex_state = 64, .external_lex_state = 15},
  [5067] = {.lex_state = 64, .external_lex_state = 7},
  [5068] = {.lex_state = 64, .external_lex_state = 8},
  [5069] = {.lex_state = 66, .external_lex_state = 9},
  [5070] = {.lex_state = 66, .external_lex_state = 10},
  [5071] = {.lex_state = 64, .external_lex_state = 16},
  [5072] = {.lex_state = 64, .external_lex_state = 8},
  [5073] = {.lex_state = 64, .external_lex_state = 16},
  [5074] = {.lex_state = 66, .external_lex_state = 10},
  [5075] = {.lex_state = 66, .external_lex_state = 10},
  [5076] = {.lex_state = 64, .external_lex_state = 13},
  [5077] = {.lex_state = 64, .external_lex_state = 11},
  [5078] = {.lex_state = 66, .external_lex_state = 9},
  [5079] = {.lex_state = 64, .external_lex_state = 12},
  [5080] = {.lex_state = 64, .external_lex_state = 13},
  [5081] = {.lex_state = 66, .external_lex_state = 10},
  [5082] = {.lex_state = 64, .external_lex_state = 14},
  [5083] = {.lex_state = 64, .external_lex_state = 15},
  [5084] = {.lex_state = 64, .external_lex_state = 16},
  [5085] = {.lex_state = 66, .external_lex_state = 10},
  [5086] = {.lex_state = 66, .external_lex_state = 6},
  [5087] = {.lex_state = 64, .external_lex_state = 6},
  [5088] = {.lex_state = 64, .external_lex_state = 6},
  [5089] = {.lex_state = 106, .external_lex_state = 6},
  [5090] = {.lex_state = 267, .external_lex_state = 6},
  [5091] = {.lex_state = 64, .external_lex_state = 6},
  [5092] = {.lex_state = 64, .external_lex_state = 6},
  [5093] = {.lex_state = 64, .external_lex_state = 6},
  [5094] = {.lex_state = 64, .external_lex_state = 6},
  [5095] = {.lex_state = 66, .external_lex_state = 6},
  [5096] = {.lex_state = 66, .external_lex_state = 6},
  [5097] = {.lex_state = 66, .external_lex_state = 6},
  [5098] = {.lex_state = 66, .external_lex_state = 6},
  [5099] = {.lex_state = 64, .external_lex_state = 6},
  [5100] = {.lex_state = 64, .external_lex_state = 6},
  [5101] = {.lex_state = 64, .external_lex_state = 6},
  [5102] = {.lex_state = 106, .external_lex_state = 6},
  [5103] = {.lex_state = 64, .external_lex_state = 6},
  [5104] = {.lex_state = 106, .external_lex_state = 6},
  [5105] = {.lex_state = 267, .external_lex_state = 6},
  [5106] = {.lex_state = 267, .external_lex_state = 6},
  [5107] = {.lex_state = 64, .external_lex_state = 6},
  [5108] = {.lex_state = 267, .external_lex_state = 6},
  [5109] = {.lex_state = 64, .external_lex_state = 6},
  [5110] = {.lex_state = 267, .external_lex_state = 6},
  [5111] = {.lex_state = 64, .external_lex_state = 6},
  [5112] = {.lex_state = 64, .external_lex_state = 6},
  [5113] = {.lex_state = 64, .external_lex_state = 6},
  [5114] = {.lex_state = 66, .external_lex_state = 6},
  [5115] = {.lex_state = 66, .external_lex_state = 6},
  [5116] = {.lex_state = 64, .external_lex_state = 6},
  [5117] = {.lex_state = 64, .external_lex_state = 6},
  [5118] = {.lex_state = 64, .external_lex_state = 6},
  [5119] = {.lex_state = 106, .external_lex_state = 6},
  [5120] = {.lex_state = 267, .external_lex_state = 6},
  [5121] = {.lex_state = 64, .external_lex_state = 6},
  [5122] = {.lex_state = 267, .external_lex_state = 6},
  [5123] = {.lex_state = 267, .external_lex_state = 6},
  [5124] = {.lex_state = 106, .external_lex_state = 6},
  [5125] = {.lex_state = 106, .external_lex_state = 6},
  [5126] = {.lex_state = 64, .external_lex_state = 6},
  [5127] = {.lex_state = 267, .external_lex_state = 6},
  [5128] = {.lex_state = 267, .external_lex_state = 6},
  [5129] = {.lex_state = 267, .external_lex_state = 6},
  [5130] = {.lex_state = 64, .external_lex_state = 6},
  [5131] = {.lex_state = 267, .external_lex_state = 6},
  [5132] = {.lex_state = 267, .external_lex_state = 6},
  [5133] = {.lex_state = 267, .external_lex_state = 6},
  [5134] = {.lex_state = 64, .external_lex_state = 6},
  [5135] = {.lex_state = 64, .external_lex_state = 6},
  [5136] = {.lex_state = 64, .external_lex_state = 6},
  [5137] = {.lex_state = 64, .external_lex_state = 6},
  [5138] = {.lex_state = 64, .external_lex_state = 6},
  [5139] = {.lex_state = 64, .external_lex_state = 6},
  [5140] = {.lex_state = 64, .external_lex_state = 6},
  [5141] = {.lex_state = 64, .external_lex_state = 6},
  [5142] = {.lex_state = 64, .external_lex_state = 6},
  [5143] = {.lex_state = 64, .external_lex_state = 6},
  [5144] = {.lex_state = 64, .external_lex_state = 6},
  [5145] = {.lex_state = 64, .external_lex_state = 6},
  [5146] = {.lex_state = 106, .external_lex_state = 6},
  [5147] = {.lex_state = 64, .external_lex_state = 6},
  [5148] = {.lex_state = 267, .external_lex_state = 6},
  [5149] = {.lex_state = 64, .external_lex_state = 6},
  [5150] = {.lex_state = 66, .external_lex_state = 6},
  [5151] = {.lex_state = 64, .external_lex_state = 6},
  [5152] = {.lex_state = 64, .external_lex_state = 6},
  [5153] = {.lex_state = 66, .external_lex_state = 6},
  [5154] = {.lex_state = 64, .external_lex_state = 6},
  [5155] = {.lex_state = 64, .external_lex_state = 6},
  [5156] = {.lex_state = 64, .external_lex_state = 6},
  [5157] = {.lex_state = 64, .external_lex_state = 6},
  [5158] = {.lex_state = 64, .external_lex_state = 6},
  [5159] = {.lex_state = 64, .external_lex_state = 6},
  [5160] = {.lex_state = 64, .external_lex_state = 6},
  [5161] = {.lex_state = 64, .external_lex_state = 6},
  [5162] = {.lex_state = 64, .external_lex_state = 6},
  [5163] = {.lex_state = 267, .external_lex_state = 6},
  [5164] = {.lex_state = 64, .external_lex_state = 6},
  [5165] = {.lex_state = 66, .external_lex_state = 6},
  [5166] = {.lex_state = 64, .external_lex_state = 6},
  [5167] = {.lex_state = 66, .external_lex_state = 6},
  [5168] = {.lex_state = 267, .external_lex_state = 6},
  [5169] = {.lex_state = 64, .external_lex_state = 6},
  [5170] = {.lex_state = 64, .external_lex_state = 6},
  [5171] = {.lex_state = 64, .external_lex_state = 6},
  [5172] = {.lex_state = 64, .external_lex_state = 6},
  [5173] = {.lex_state = 66, .external_lex_state = 6},
  [5174] = {.lex_state = 66, .external_lex_state = 6},
  [5175] = {.lex_state = 66, .external_lex_state = 6},
  [5176] = {.lex_state = 267, .external_lex_state = 6},
  [5177] = {.lex_state = 64, .external_lex_state = 6},
  [5178] = {.lex_state = 64, .external_lex_state = 6},
  [5179] = {.lex_state = 64, .external_lex_state = 6},
  [5180] = {.lex_state = 64, .external_lex_state = 6},
  [5181] = {.lex_state = 66, .external_lex_state = 6},
  [5182] = {.lex_state = 66, .external_lex_state = 6},
  [5183] = {.lex_state = 66, .external_lex_state = 6},
  [5184] = {.lex_state = 66, .external_lex_state = 6},
  [5185] = {.lex_state = 66, .external_lex_state = 6},
  [5186] = {.lex_state = 66, .external_lex_state = 6},
  [5187] = {.lex_state = 106, .external_lex_state = 6},
  [5188] = {.lex_state = 267, .external_lex_state = 6},
  [5189] = {.lex_state = 106, .external_lex_state = 6},
  [5190] = {.lex_state = 267, .external_lex_state = 6},
  [5191] = {.lex_state = 267, .external_lex_state = 6},
  [5192] = {.lex_state = 267, .external_lex_state = 6},
  [5193] = {.lex_state = 64, .external_lex_state = 6},
  [5194] = {.lex_state = 267, .external_lex_state = 6},
  [5195] = {.lex_state = 64, .external_lex_state = 6},
  [5196] = {.lex_state = 66, .external_lex_state = 6},
  [5197] = {.lex_state = 66, .external_lex_state = 6},
  [5198] = {.lex_state = 64, .external_lex_state = 6},
  [5199] = {.lex_state = 267, .external_lex_state = 6},
  [5200] = {.lex_state = 64, .external_lex_state = 6},
  [5201] = {.lex_state = 64, .external_lex_state = 6},
  [5202] = {.lex_state = 64, .external_lex_state = 6},
  [5203] = {.lex_state = 66, .external_lex_state = 6},
  [5204] = {.lex_state = 106, .external_lex_state = 6},
  [5205] = {.lex_state = 66, .external_lex_state = 6},
  [5206] = {.lex_state = 106, .external_lex_state = 6},
  [5207] = {.lex_state = 106, .external_lex_state = 6},
  [5208] = {.lex_state = 66, .external_lex_state = 6},
  [5209] = {.lex_state = 66, .external_lex_state = 6},
  [5210] = {.lex_state = 106, .external_lex_state = 6},
  [5211] = {.lex_state = 267, .external_lex_state = 6},
  [5212] = {.lex_state = 106, .external_lex_state = 6},
  [5213] = {.lex_state = 267, .external_lex_state = 6},
  [5214] = {.lex_state = 267, .external_lex_state = 6},
  [5215] = {.lex_state = 64, .external_lex_state = 6},
  [5216] = {.lex_state = 64, .external_lex_state = 6},
  [5217] = {.lex_state = 64, .external_lex_state = 6},
  [5218] = {.lex_state = 64, .external_lex_state = 6},
  [5219] = {.lex_state = 64, .external_lex_state = 6},
  [5220] = {.lex_state = 64, .external_lex_state = 6},
  [5221] = {.lex_state = 64, .external_lex_state = 6},
  [5222] = {.lex_state = 64, .external_lex_state = 6},
  [5223] = {.lex_state = 64, .external_lex_state = 6},
  [5224] = {.lex_state = 64, .external_lex_state = 6},
  [5225] = {.lex_state = 64, .external_lex_state = 6},
  [5226] = {.lex_state = 64, .external_lex_state = 6},
  [5227] = {.lex_state = 106, .external_lex_state = 6},
  [5228] = {.lex_state = 267, .external_lex_state = 6},
  [5229] = {.lex_state = 267, .external_lex_state = 6},
  [5230] = {.lex_state = 267, .external_lex_state = 6},
  [5231] = {.lex_state = 66, .external_lex_state = 6},
  [5232] = {.lex_state = 64, .external_lex_state = 6},
  [5233] = {.lex_state = 64, .external_lex_state = 6},
  [5234] = {.lex_state = 64, .external_lex_state = 6},
  [5235] = {.lex_state = 64, .external_lex_state = 6},
  [5236] = {.lex_state = 64, .external_lex_state = 6},
  [5237] = {.lex_state = 64, .external_lex_state = 6},
  [5238] = {.lex_state = 267, .external_lex_state = 6},
  [5239] = {.lex_state = 64, .external_lex_state = 6},
  [5240] = {.lex_state = 267, .external_lex_state = 6},
  [5241] = {.lex_state = 64, .external_lex_state = 6},
  [5242] = {.lex_state = 66, .external_lex_state = 6},
  [5243] = {.lex_state = 66, .external_lex_state = 6},
  [5244] = {.lex_state = 64, .external_lex_state = 6},
  [5245] = {.lex_state = 267, .external_lex_state = 6},
  [5246] = {.lex_state = 64, .external_lex_state = 6},
  [5247] = {.lex_state = 66, .external_lex_state = 6},
  [5248] = {.lex_state = 64, .external_lex_state = 6},
  [5249] = {.lex_state = 106, .external_lex_state = 6},
  [5250] = {.lex_state = 267, .external_lex_state = 6},
  [5251] = {.lex_state = 267, .external_lex_state = 6},
  [5252] = {.lex_state = 267, .external_lex_state = 6},
  [5253] = {.lex_state = 64, .external_lex_state = 6},
  [5254] = {.lex_state = 64, .external_lex_state = 6},
  [5255] = {.lex_state = 64, .external_lex_state = 6},
  [5256] = {.lex_state = 64, .external_lex_state = 6},
  [5257] = {.lex_state = 66, .external_lex_state = 6},
  [5258] = {.lex_state = 66, .external_lex_state = 6},
  [5259] = {.lex_state = 66, .external_lex_state = 6},
  [5260] = {.lex_state = 66, .external_lex_state = 6},
  [5261] = {.lex_state = 66, .external_lex_state = 6},
  [5262] = {.lex_state = 64, .external_lex_state = 6},
  [5263] = {.lex_state = 267, .external_lex_state = 6},
  [5264] = {.lex_state = 106, .external_lex_state = 6},
  [5265] = {.lex_state = 106, .external_lex_state = 6},
  [5266] = {.lex_state = 64, .external_lex_state = 6},
  [5267] = {.lex_state = 64, .external_lex_state = 6},
  [5268] = {.lex_state = 64, .external_lex_state = 6},
  [5269] = {.lex_state = 64, .external_lex_state = 6},
  [5270] = {.lex_state = 267, .external_lex_state = 6},
  [5271] = {.lex_state = 64, .external_lex_state = 6},
  [5272] = {.lex_state = 267, .external_lex_state = 6},
  [5273] = {.lex_state = 64, .external_lex_state = 6},
  [5274] = {.lex_state = 267, .external_lex_state = 6},
  [5275] = {.lex_state = 64, .external_lex_state = 6},
  [5276] = {.lex_state = 64, .external_lex_state = 6},
  [5277] = {.lex_state = 66, .external_lex_state = 6},
  [5278] = {.lex_state = 66, .external_lex_state = 6},
  [5279] = {.lex_state = 106, .external_lex_state = 6},
  [5280] = {.lex_state = 106, .external_lex_state = 6},
  [5281] = {.lex_state = 267, .external_lex_state = 6},
  [5282] = {.lex_state = 267, .external_lex_state = 6},
  [5283] = {.lex_state = 64, .external_lex_state = 6},
  [5284] = {.lex_state = 106, .external_lex_state = 6},
  [5285] = {.lex_state = 64, .external_lex_state = 6},
  [5286] = {.lex_state = 64, .external_lex_state = 6},
  [5287] = {.lex_state = 106, .external_lex_state = 6},
  [5288] = {.lex_state = 106, .external_lex_state = 6},
  [5289] = {.lex_state = 106, .external_lex_state = 6},
  [5290] = {.lex_state = 64, .external_lex_state = 6},
  [5291] = {.lex_state = 64, .external_lex_state = 6},
  [5292] = {.lex_state = 106, .external_lex_state = 6},
  [5293] = {.lex_state = 106, .external_lex_state = 6},
  [5294] = {.lex_state = 64, .external_lex_state = 6},
  [5295] = {.lex_state = 267, .external_lex_state = 6},
  [5296] = {.lex_state = 267, .external_lex_state = 6},
  [5297] = {.lex_state = 64, .external_lex_state = 6},
  [5298] = {.lex_state = 64, .external_lex_state = 6},
  [5299] = {.lex_state = 64, .external_lex_state = 6},
  [5300] = {.lex_state = 64, .external_lex_state = 6},
  [5301] = {.lex_state = 64, .external_lex_state = 6},
  [5302] = {.lex_state = 64, .external_lex_state = 6},
  [5303] = {.lex_state = 64, .external_lex_state = 6},
  [5304] = {.lex_state = 64, .external_lex_state = 6},
  [5305] = {.lex_state = 64, .external_lex_state = 6},
  [5306] = {.lex_state = 64, .external_lex_state = 6},
  [5307] = {.lex_state = 64, .external_lex_state = 6},
  [5308] = {.lex_state = 64, .external_lex_state = 6},
  [5309] = {.lex_state = 106, .external_lex_state = 6},
  [5310] = {.lex_state = 64, .external_lex_state = 6},
  [5311] = {.lex_state = 267, .external_lex_state = 6},
  [5312] = {.lex_state = 267, .external_lex_state = 6},
  [5313] = {.lex_state = 64, .external_lex_state = 6},
  [5314] = {.lex_state = 64, .external_lex_state = 6},
  [5315] = {.lex_state = 64, .external_lex_state = 6},
  [5316] = {.lex_state = 64, .external_lex_state = 6},
  [5317] = {.lex_state = 64, .external_lex_state = 6},
  [5318] = {.lex_state = 64, .external_lex_state = 6},
  [5319] = {.lex_state = 64, .external_lex_state = 6},
  [5320] = {.lex_state = 64, .external_lex_state = 6},
  [5321] = {.lex_state = 64, .external_lex_state = 6},
  [5322] = {.lex_state = 64, .external_lex_state = 6},
  [5323] = {.lex_state = 64, .external_lex_state = 6},
  [5324] = {.lex_state = 64, .external_lex_state = 6},
  [5325] = {.lex_state = 64, .external_lex_state = 6},
  [5326] = {.lex_state = 64, .external_lex_state = 6},
  [5327] = {.lex_state = 64, .external_lex_state = 6},
  [5328] = {.lex_state = 64, .external_lex_state = 6},
  [5329] = {.lex_state = 64, .external_lex_state = 6},
  [5330] = {.lex_state = 64, .external_lex_state = 6},
  [5331] = {.lex_state = 64, .external_lex_state = 6},
  [5332] = {.lex_state = 267, .external_lex_state = 6},
  [5333] = {.lex_state = 64, .external_lex_state = 6},
  [5334] = {.lex_state = 64, .external_lex_state = 6},
  [5335] = {.lex_state = 64, .external_lex_state = 6},
  [5336] = {.lex_state = 64, .external_lex_state = 6},
  [5337] = {.lex_state = 66, .external_lex_state = 6},
  [5338] = {.lex_state = 66, .external_lex_state = 6},
  [5339] = {.lex_state = 66, .external_lex_state = 6},
  [5340] = {.lex_state = 66, .external_lex_state = 6},
  [5341] = {.lex_state = 106, .external_lex_state = 6},
  [5342] = {.lex_state = 64, .external_lex_state = 6},
  [5343] = {.lex_state = 106, .external_lex_state = 6},
  [5344] = {.lex_state = 267, .external_lex_state = 6},
  [5345] = {.lex_state = 267, .external_lex_state = 6},
  [5346] = {.lex_state = 64, .external_lex_state = 6},
  [5347] = {.lex_state = 267, .external_lex_state = 6},
  [5348] = {.lex_state = 64, .external_lex_state = 6},
  [5349] = {.lex_state = 64, .external_lex_state = 6},
  [5350] = {.lex_state = 66, .external_lex_state = 6},
  [5351] = {.lex_state = 66, .external_lex_state = 6},
  [5352] = {.lex_state = 267, .external_lex_state = 6},
  [5353] = {.lex_state = 64, .external_lex_state = 6},
  [5354] = {.lex_state = 64, .external_lex_state = 6},
  [5355] = {.lex_state = 64, .external_lex_state = 6},
  [5356] = {.lex_state = 64, .external_lex_state = 6},
  [5357] = {.lex_state = 64, .external_lex_state = 6},
  [5358] = {.lex_state = 106, .external_lex_state = 6},
  [5359] = {.lex_state = 106, .external_lex_state = 6},
  [5360] = {.lex_state = 64, .external_lex_state = 6},
  [5361] = {.lex_state = 106, .external_lex_state = 6},
  [5362] = {.lex_state = 64, .external_lex_state = 6},
  [5363] = {.lex_state = 267, .external_lex_state = 6},
  [5364] = {.lex_state = 267, .external_lex_state = 6},
  [5365] = {.lex_state = 267, .external_lex_state = 6},
  [5366] = {.lex_state = 267, .external_lex_state = 6},
  [5367] = {.lex_state = 64, .external_lex_state = 6},
  [5368] = {.lex_state = 64, .external_lex_state = 6},
  [5369] = {.lex_state = 64, .external_lex_state = 6},
  [5370] = {.lex_state = 64, .external_lex_state = 6},
  [5371] = {.lex_state = 64, .external_lex_state = 6},
  [5372] = {.lex_state = 64, .external_lex_state = 6},
  [5373] = {.lex_state = 64, .external_lex_state = 6},
  [5374] = {.lex_state = 64, .external_lex_state = 6},
  [5375] = {.lex_state = 64, .external_lex_state = 6},
  [5376] = {.lex_state = 64, .external_lex_state = 6},
  [5377] = {.lex_state = 64, .external_lex_state = 6},
  [5378] = {.lex_state = 64, .external_lex_state = 6},
  [5379] = {.lex_state = 106, .external_lex_state = 6},
  [5380] = {.lex_state = 64, .external_lex_state = 6},
  [5381] = {.lex_state = 64, .external_lex_state = 6},
  [5382] = {.lex_state = 64, .external_lex_state = 6},
  [5383] = {.lex_state = 64, .external_lex_state = 6},
  [5384] = {.lex_state = 267, .external_lex_state = 6},
  [5385] = {.lex_state = 64, .external_lex_state = 6},
  [5386] = {.lex_state = 64, .external_lex_state = 6},
  [5387] = {.lex_state = 64, .external_lex_state = 6},
  [5388] = {.lex_state = 64, .external_lex_state = 6},
  [5389] = {.lex_state = 64, .external_lex_state = 6},
  [5390] = {.lex_state = 64, .external_lex_state = 6},
  [5391] = {.lex_state = 64, .external_lex_state = 6},
  [5392] = {.lex_state = 64, .external_lex_state = 6},
  [5393] = {.lex_state = 106, .external_lex_state = 6},
  [5394] = {.lex_state = 64, .external_lex_state = 6},
  [5395] = {.lex_state = 64, .external_lex_state = 6},
  [5396] = {.lex_state = 64, .external_lex_state = 6},
  [5397] = {.lex_state = 64, .external_lex_state = 6},
  [5398] = {.lex_state = 106, .external_lex_state = 6},
  [5399] = {.lex_state = 64, .external_lex_state = 6},
  [5400] = {.lex_state = 64, .external_lex_state = 6},
  [5401] = {.lex_state = 64, .external_lex_state = 6},
  [5402] = {.lex_state = 106, .external_lex_state = 6},
  [5403] = {.lex_state = 267, .external_lex_state = 6},
  [5404] = {.lex_state = 64, .external_lex_state = 6},
  [5405] = {.lex_state = 64, .external_lex_state = 6},
  [5406] = {.lex_state = 64, .external_lex_state = 6},
  [5407] = {.lex_state = 64, .external_lex_state = 6},
  [5408] = {.lex_state = 64, .external_lex_state = 6},
  [5409] = {.lex_state = 66, .external_lex_state = 6},
  [5410] = {.lex_state = 66, .external_lex_state = 6},
  [5411] = {.lex_state = 66, .external_lex_state = 6},
  [5412] = {.lex_state = 66, .external_lex_state = 6},
  [5413] = {.lex_state = 64, .external_lex_state = 6},
  [5414] = {.lex_state = 64, .external_lex_state = 6},
  [5415] = {.lex_state = 64, .external_lex_state = 6},
  [5416] = {.lex_state = 64, .external_lex_state = 6},
  [5417] = {.lex_state = 64, .external_lex_state = 6},
  [5418] = {.lex_state = 66, .external_lex_state = 6},
  [5419] = {.lex_state = 64, .external_lex_state = 6},
  [5420] = {.lex_state = 66, .external_lex_state = 6},
  [5421] = {.lex_state = 66, .external_lex_state = 6},
  [5422] = {.lex_state = 66, .external_lex_state = 6},
  [5423] = {.lex_state = 66, .external_lex_state = 6},
  [5424] = {.lex_state = 64, .external_lex_state = 6},
  [5425] = {.lex_state = 64, .external_lex_state = 6},
  [5426] = {.lex_state = 64, .external_lex_state = 6},
  [5427] = {.lex_state = 64, .external_lex_state = 6},
  [5428] = {.lex_state = 64, .external_lex_state = 6},
  [5429] = {.lex_state = 64, .external_lex_state = 6},
  [5430] = {.lex_state = 64, .external_lex_state = 6},
  [5431] = {.lex_state = 64, .external_lex_state = 6},
  [5432] = {.lex_state = 64, .external_lex_state = 6},
  [5433] = {.lex_state = 64, .external_lex_state = 6},
  [5434] = {.lex_state = 64, .external_lex_state = 6},
  [5435] = {.lex_state = 64, .external_lex_state = 6},
  [5436] = {.lex_state = 64, .external_lex_state = 6},
  [5437] = {.lex_state = 64, .external_lex_state = 6},
  [5438] = {.lex_state = 64, .external_lex_state = 6},
  [5439] = {.lex_state = 64, .external_lex_state = 6},
  [5440] = {.lex_state = 64, .external_lex_state = 6},
  [5441] = {.lex_state = 64, .external_lex_state = 6},
  [5442] = {.lex_state = 64, .external_lex_state = 6},
  [5443] = {.lex_state = 64, .external_lex_state = 6},
  [5444] = {.lex_state = 64, .external_lex_state = 6},
  [5445] = {.lex_state = 64, .external_lex_state = 6},
  [5446] = {.lex_state = 106, .external_lex_state = 6},
  [5447] = {.lex_state = 64, .external_lex_state = 6},
  [5448] = {.lex_state = 64, .external_lex_state = 6},
  [5449] = {.lex_state = 64, .external_lex_state = 6},
  [5450] = {.lex_state = 64, .external_lex_state = 6},
  [5451] = {.lex_state = 64, .external_lex_state = 6},
  [5452] = {.lex_state = 64, .external_lex_state = 6},
  [5453] = {.lex_state = 66, .external_lex_state = 6},
  [5454] = {.lex_state = 66, .external_lex_state = 6},
  [5455] = {.lex_state = 66, .external_lex_state = 6},
  [5456] = {.lex_state = 66, .external_lex_state = 6},
  [5457] = {.lex_state = 64, .external_lex_state = 6},
  [5458] = {.lex_state = 267, .external_lex_state = 6},
  [5459] = {.lex_state = 64, .external_lex_state = 6},
  [5460] = {.lex_state = 64, .external_lex_state = 6},
  [5461] = {.lex_state = 64, .external_lex_state = 6},
  [5462] = {.lex_state = 64, .external_lex_state = 6},
  [5463] = {.lex_state = 66, .external_lex_state = 6},
  [5464] = {.lex_state = 66, .external_lex_state = 6},
  [5465] = {.lex_state = 267, .external_lex_state = 6},
  [5466] = {.lex_state = 64, .external_lex_state = 6},
  [5467] = {.lex_state = 267, .external_lex_state = 6},
  [5468] = {.lex_state = 64, .external_lex_state = 6},
  [5469] = {.lex_state = 64, .external_lex_state = 6},
  [5470] = {.lex_state = 64, .external_lex_state = 6},
  [5471] = {.lex_state = 64, .external_lex_state = 6},
  [5472] = {.lex_state = 64, .external_lex_state = 6},
  [5473] = {.lex_state = 64, .external_lex_state = 6},
  [5474] = {.lex_state = 64, .external_lex_state = 6},
  [5475] = {.lex_state = 64, .external_lex_state = 6},
  [5476] = {.lex_state = 64, .external_lex_state = 6},
  [5477] = {.lex_state = 64, .external_lex_state = 6},
  [5478] = {.lex_state = 64, .external_lex_state = 6},
  [5479] = {.lex_state = 64, .external_lex_state = 6},
  [5480] = {.lex_state = 64, .external_lex_state = 6},
  [5481] = {.lex_state = 64, .external_lex_state = 6},
  [5482] = {.lex_state = 64, .external_lex_state = 6},
  [5483] = {.lex_state = 64, .external_lex_state = 6},
  [5484] = {.lex_state = 64, .external_lex_state = 6},
  [5485] = {.lex_state = 64, .external_lex_state = 6},
  [5486] = {.lex_state = 64, .external_lex_state = 6},
  [5487] = {.lex_state = 64, .external_lex_state = 6},
  [5488] = {.lex_state = 64, .external_lex_state = 6},
  [5489] = {.lex_state = 66, .external_lex_state = 6},
  [5490] = {.lex_state = 66, .external_lex_state = 6},
  [5491] = {.lex_state = 64, .external_lex_state = 6},
  [5492] = {.lex_state = 64, .external_lex_state = 6},
  [5493] = {.lex_state = 267, .external_lex_state = 6},
  [5494] = {.lex_state = 64, .external_lex_state = 6},
  [5495] = {.lex_state = 64, .external_lex_state = 6},
  [5496] = {.lex_state = 64, .external_lex_state = 6},
  [5497] = {.lex_state = 64, .external_lex_state = 6},
  [5498] = {.lex_state = 64, .external_lex_state = 6},
  [5499] = {.lex_state = 66, .external_lex_state = 6},
  [5500] = {.lex_state = 66, .external_lex_state = 6},
  [5501] = {.lex_state = 66, .external_lex_state = 6},
  [5502] = {.lex_state = 66, .external_lex_state = 6},
  [5503] = {.lex_state = 267, .external_lex_state = 6},
  [5504] = {.lex_state = 64, .external_lex_state = 6},
  [5505] = {.lex_state = 64, .external_lex_state = 6},
  [5506] = {.lex_state = 64, .external_lex_state = 6},
  [5507] = {.lex_state = 64, .external_lex_state = 6},
  [5508] = {.lex_state = 64, .external_lex_state = 6},
  [5509] = {.lex_state = 66, .external_lex_state = 6},
  [5510] = {.lex_state = 66, .external_lex_state = 6},
  [5511] = {.lex_state = 64, .external_lex_state = 6},
  [5512] = {.lex_state = 106, .external_lex_state = 6},
  [5513] = {.lex_state = 106, .external_lex_state = 6},
  [5514] = {.lex_state = 64, .external_lex_state = 6},
  [5515] = {.lex_state = 64, .external_lex_state = 6},
  [5516] = {.lex_state = 64, .external_lex_state = 6},
  [5517] = {.lex_state = 64, .external_lex_state = 6},
  [5518] = {.lex_state = 64, .external_lex_state = 6},
  [5519] = {.lex_state = 64, .external_lex_state = 6},
  [5520] = {.lex_state = 64, .external_lex_state = 6},
  [5521] = {.lex_state = 64, .external_lex_state = 6},
  [5522] = {.lex_state = 64, .external_lex_state = 6},
  [5523] = {.lex_state = 64, .external_lex_state = 6},
  [5524] = {.lex_state = 64, .external_lex_state = 6},
  [5525] = {.lex_state = 64, .external_lex_state = 6},
  [5526] = {.lex_state = 64, .external_lex_state = 6},
  [5527] = {.lex_state = 64, .external_lex_state = 6},
  [5528] = {.lex_state = 64, .external_lex_state = 6},
  [5529] = {.lex_state = 64, .external_lex_state = 6},
  [5530] = {.lex_state = 64, .external_lex_state = 6},
  [5531] = {.lex_state = 64, .external_lex_state = 6},
  [5532] = {.lex_state = 64, .external_lex_state = 6},
  [5533] = {.lex_state = 64, .external_lex_state = 6},
  [5534] = {.lex_state = 64, .external_lex_state = 6},
  [5535] = {.lex_state = 64, .external_lex_state = 6},
  [5536] = {.lex_state = 267, .external_lex_state = 6},
  [5537] = {.lex_state = 267, .external_lex_state = 6},
  [5538] = {.lex_state = 64, .external_lex_state = 6},
  [5539] = {.lex_state = 64, .external_lex_state = 6},
  [5540] = {.lex_state = 64, .external_lex_state = 6},
  [5541] = {.lex_state = 64, .external_lex_state = 6},
  [5542] = {.lex_state = 64, .external_lex_state = 6},
  [5543] = {.lex_state = 64, .external_lex_state = 6},
  [5544] = {.lex_state = 64, .external_lex_state = 6},
  [5545] = {.lex_state = 66, .external_lex_state = 6},
  [5546] = {.lex_state = 66, .external_lex_state = 6},
  [5547] = {.lex_state = 66, .external_lex_state = 6},
  [5548] = {.lex_state = 66, .external_lex_state = 6},
  [5549] = {.lex_state = 64, .external_lex_state = 6},
  [5550] = {.lex_state = 64, .external_lex_state = 6},
  [5551] = {.lex_state = 64, .external_lex_state = 6},
  [5552] = {.lex_state = 64, .external_lex_state = 6},
  [5553] = {.lex_state = 64, .external_lex_state = 6},
  [5554] = {.lex_state = 64, .external_lex_state = 6},
  [5555] = {.lex_state = 66, .external_lex_state = 6},
  [5556] = {.lex_state = 66, .external_lex_state = 6},
  [5557] = {.lex_state = 64, .external_lex_state = 6},
  [5558] = {.lex_state = 64, .external_lex_state = 6},
  [5559] = {.lex_state = 64, .external_lex_state = 6},
  [5560] = {.lex_state = 64, .external_lex_state = 6},
  [5561] = {.lex_state = 64, .external_lex_state = 6},
  [5562] = {.lex_state = 64, .external_lex_state = 6},
  [5563] = {.lex_state = 64, .external_lex_state = 6},
  [5564] = {.lex_state = 64, .external_lex_state = 6},
  [5565] = {.lex_state = 64, .external_lex_state = 6},
  [5566] = {.lex_state = 64, .external_lex_state = 6},
  [5567] = {.lex_state = 64, .external_lex_state = 6},
  [5568] = {.lex_state = 64, .external_lex_state = 6},
  [5569] = {.lex_state = 64, .external_lex_state = 6},
  [5570] = {.lex_state = 64, .external_lex_state = 6},
  [5571] = {.lex_state = 64, .external_lex_state = 6},
  [5572] = {.lex_state = 64, .external_lex_state = 6},
  [5573] = {.lex_state = 64, .external_lex_state = 6},
  [5574] = {.lex_state = 64, .external_lex_state = 6},
  [5575] = {.lex_state = 64, .external_lex_state = 6},
  [5576] = {.lex_state = 64, .external_lex_state = 6},
  [5577] = {.lex_state = 64, .external_lex_state = 6},
  [5578] = {.lex_state = 267, .external_lex_state = 6},
  [5579] = {.lex_state = 64, .external_lex_state = 6},
  [5580] = {.lex_state = 64, .external_lex_state = 6},
  [5581] = {.lex_state = 64, .external_lex_state = 6},
  [5582] = {.lex_state = 106, .external_lex_state = 6},
  [5583] = {.lex_state = 64, .external_lex_state = 6},
  [5584] = {.lex_state = 64, .external_lex_state = 6},
  [5585] = {.lex_state = 64, .external_lex_state = 6},
  [5586] = {.lex_state = 64, .external_lex_state = 6},
  [5587] = {.lex_state = 64, .external_lex_state = 6},
  [5588] = {.lex_state = 64, .external_lex_state = 6},
  [5589] = {.lex_state = 64, .external_lex_state = 6},
  [5590] = {.lex_state = 64, .external_lex_state = 6},
  [5591] = {.lex_state = 66, .external_lex_state = 6},
  [5592] = {.lex_state = 66, .external_lex_state = 6},
  [5593] = {.lex_state = 66, .external_lex_state = 6},
  [5594] = {.lex_state = 66, .external_lex_state = 6},
  [5595] = {.lex_state = 106, .external_lex_state = 6},
  [5596] = {.lex_state = 64, .external_lex_state = 6},
  [5597] = {.lex_state = 267, .external_lex_state = 6},
  [5598] = {.lex_state = 64, .external_lex_state = 6},
  [5599] = {.lex_state = 64, .external_lex_state = 6},
  [5600] = {.lex_state = 66, .external_lex_state = 6},
  [5601] = {.lex_state = 66, .external_lex_state = 6},
  [5602] = {.lex_state = 64, .external_lex_state = 6},
  [5603] = {.lex_state = 64, .external_lex_state = 6},
  [5604] = {.lex_state = 64, .external_lex_state = 6},
  [5605] = {.lex_state = 64, .external_lex_state = 6},
  [5606] = {.lex_state = 64, .external_lex_state = 6},
  [5607] = {.lex_state = 64, .external_lex_state = 6},
  [5608] = {.lex_state = 64, .external_lex_state = 6},
  [5609] = {.lex_state = 64, .external_lex_state = 6},
  [5610] = {.lex_state = 64, .external_lex_state = 6},
  [5611] = {.lex_state = 64, .external_lex_state = 6},
  [5612] = {.lex_state = 64, .external_lex_state = 6},
  [5613] = {.lex_state = 64, .external_lex_state = 6},
  [5614] = {.lex_state = 64, .external_lex_state = 6},
  [5615] = {.lex_state = 64, .external_lex_state = 6},
  [5616] = {.lex_state = 64, .external_lex_state = 6},
  [5617] = {.lex_state = 64, .external_lex_state = 6},
  [5618] = {.lex_state = 64, .external_lex_state = 6},
  [5619] = {.lex_state = 64, .external_lex_state = 6},
  [5620] = {.lex_state = 64, .external_lex_state = 6},
  [5621] = {.lex_state = 64, .external_lex_state = 6},
  [5622] = {.lex_state = 64, .external_lex_state = 6},
  [5623] = {.lex_state = 64, .external_lex_state = 6},
  [5624] = {.lex_state = 64, .external_lex_state = 6},
  [5625] = {.lex_state = 64, .external_lex_state = 6},
  [5626] = {.lex_state = 64, .external_lex_state = 6},
  [5627] = {.lex_state = 64, .external_lex_state = 6},
  [5628] = {.lex_state = 64, .external_lex_state = 6},
  [5629] = {.lex_state = 64, .external_lex_state = 6},
  [5630] = {.lex_state = 64, .external_lex_state = 6},
  [5631] = {.lex_state = 64, .external_lex_state = 6},
  [5632] = {.lex_state = 64, .external_lex_state = 6},
  [5633] = {.lex_state = 64, .external_lex_state = 6},
  [5634] = {.lex_state = 64, .external_lex_state = 6},
  [5635] = {.lex_state = 66, .external_lex_state = 6},
  [5636] = {.lex_state = 66, .external_lex_state = 6},
  [5637] = {.lex_state = 66, .external_lex_state = 6},
  [5638] = {.lex_state = 66, .external_lex_state = 6},
  [5639] = {.lex_state = 64, .external_lex_state = 6},
  [5640] = {.lex_state = 66, .external_lex_state = 6},
  [5641] = {.lex_state = 66, .external_lex_state = 6},
  [5642] = {.lex_state = 64, .external_lex_state = 6},
  [5643] = {.lex_state = 64, .external_lex_state = 6},
  [5644] = {.lex_state = 66, .external_lex_state = 6},
  [5645] = {.lex_state = 66, .external_lex_state = 6},
  [5646] = {.lex_state = 66, .external_lex_state = 6},
  [5647] = {.lex_state = 66, .external_lex_state = 6},
  [5648] = {.lex_state = 64, .external_lex_state = 6},
  [5649] = {.lex_state = 64, .external_lex_state = 6},
  [5650] = {.lex_state = 106, .external_lex_state = 6},
  [5651] = {.lex_state = 64, .external_lex_state = 6},
  [5652] = {.lex_state = 64, .external_lex_state = 6},
  [5653] = {.lex_state = 64, .external_lex_state = 6},
  [5654] = {.lex_state = 64, .external_lex_state = 6},
  [5655] = {.lex_state = 64, .external_lex_state = 6},
  [5656] = {.lex_state = 64, .external_lex_state = 6},
  [5657] = {.lex_state = 64, .external_lex_state = 6},
  [5658] = {.lex_state = 64, .external_lex_state = 6},
  [5659] = {.lex_state = 64, .external_lex_state = 6},
  [5660] = {.lex_state = 64, .external_lex_state = 6},
  [5661] = {.lex_state = 64, .external_lex_state = 6},
  [5662] = {.lex_state = 64, .external_lex_state = 6},
  [5663] = {.lex_state = 64, .external_lex_state = 6},
  [5664] = {.lex_state = 64, .external_lex_state = 6},
  [5665] = {.lex_state = 64, .external_lex_state = 6},
  [5666] = {.lex_state = 64, .external_lex_state = 6},
  [5667] = {.lex_state = 64, .external_lex_state = 6},
  [5668] = {.lex_state = 64, .external_lex_state = 6},
  [5669] = {.lex_state = 64, .external_lex_state = 6},
  [5670] = {.lex_state = 64, .external_lex_state = 6},
  [5671] = {.lex_state = 106, .external_lex_state = 6},
  [5672] = {.lex_state = 66, .external_lex_state = 6},
  [5673] = {.lex_state = 66, .external_lex_state = 6},
  [5674] = {.lex_state = 267, .external_lex_state = 6},
  [5675] = {.lex_state = 64, .external_lex_state = 6},
  [5676] = {.lex_state = 64, .external_lex_state = 6},
  [5677] = {.lex_state = 64, .external_lex_state = 6},
  [5678] = {.lex_state = 64, .external_lex_state = 6},
  [5679] = {.lex_state = 66, .external_lex_state = 6},
  [5680] = {.lex_state = 66, .external_lex_state = 6},
  [5681] = {.lex_state = 66, .external_lex_state = 6},
  [5682] = {.lex_state = 66, .external_lex_state = 6},
  [5683] = {.lex_state = 64, .external_lex_state = 6},
  [5684] = {.lex_state = 64, .external_lex_state = 6},
  [5685] = {.lex_state = 66, .external_lex_state = 6},
  [5686] = {.lex_state = 66, .external_lex_state = 6},
  [5687] = {.lex_state = 64, .external_lex_state = 6},
  [5688] = {.lex_state = 267, .external_lex_state = 6},
  [5689] = {.lex_state = 64, .external_lex_state = 6},
  [5690] = {.lex_state = 64, .external_lex_state = 6},
  [5691] = {.lex_state = 64, .external_lex_state = 6},
  [5692] = {.lex_state = 64, .external_lex_state = 6},
  [5693] = {.lex_state = 64, .external_lex_state = 6},
  [5694] = {.lex_state = 64, .external_lex_state = 6},
  [5695] = {.lex_state = 64, .external_lex_state = 6},
  [5696] = {.lex_state = 66, .external_lex_state = 6},
  [5697] = {.lex_state = 66, .external_lex_state = 6},
  [5698] = {.lex_state = 66, .external_lex_state = 6},
  [5699] = {.lex_state = 66, .external_lex_state = 6},
  [5700] = {.lex_state = 64, .external_lex_state = 6},
  [5701] = {.lex_state = 64, .external_lex_state = 6},
  [5702] = {.lex_state = 66, .external_lex_state = 6},
  [5703] = {.lex_state = 66, .external_lex_state = 6},
  [5704] = {.lex_state = 64, .external_lex_state = 6},
  [5705] = {.lex_state = 66, .external_lex_state = 6},
  [5706] = {.lex_state = 66, .external_lex_state = 6},
  [5707] = {.lex_state = 66, .external_lex_state = 6},
  [5708] = {.lex_state = 64, .external_lex_state = 6},
  [5709] = {.lex_state = 64, .external_lex_state = 6},
  [5710] = {.lex_state = 64, .external_lex_state = 6},
  [5711] = {.lex_state = 64, .external_lex_state = 6},
  [5712] = {.lex_state = 66, .external_lex_state = 6},
  [5713] = {.lex_state = 66, .external_lex_state = 6},
  [5714] = {.lex_state = 66, .external_lex_state = 6},
  [5715] = {.lex_state = 66, .external_lex_state = 6},
  [5716] = {.lex_state = 64, .external_lex_state = 6},
  [5717] = {.lex_state = 64, .external_lex_state = 6},
  [5718] = {.lex_state = 66, .external_lex_state = 6},
  [5719] = {.lex_state = 66, .external_lex_state = 6},
  [5720] = {.lex_state = 66, .external_lex_state = 6},
  [5721] = {.lex_state = 267, .external_lex_state = 6},
  [5722] = {.lex_state = 64, .external_lex_state = 6},
  [5723] = {.lex_state = 106, .external_lex_state = 6},
  [5724] = {.lex_state = 64, .external_lex_state = 6},
  [5725] = {.lex_state = 64, .external_lex_state = 6},
  [5726] = {.lex_state = 64, .external_lex_state = 6},
  [5727] = {.lex_state = 64, .external_lex_state = 6},
  [5728] = {.lex_state = 66, .external_lex_state = 6},
  [5729] = {.lex_state = 66, .external_lex_state = 6},
  [5730] = {.lex_state = 66, .external_lex_state = 6},
  [5731] = {.lex_state = 66, .external_lex_state = 6},
  [5732] = {.lex_state = 64, .external_lex_state = 6},
  [5733] = {.lex_state = 64, .external_lex_state = 6},
  [5734] = {.lex_state = 66, .external_lex_state = 6},
  [5735] = {.lex_state = 66, .external_lex_state = 6},
  [5736] = {.lex_state = 64, .external_lex_state = 6},
  [5737] = {.lex_state = 64, .external_lex_state = 6},
  [5738] = {.lex_state = 267, .external_lex_state = 6},
  [5739] = {.lex_state = 267, .external_lex_state = 6},
  [5740] = {.lex_state = 64, .external_lex_state = 6},
  [5741] = {.lex_state = 64, .external_lex_state = 6},
  [5742] = {.lex_state = 64, .external_lex_state = 6},
  [5743] = {.lex_state = 64, .external_lex_state = 6},
  [5744] = {.lex_state = 66, .external_lex_state = 6},
  [5745] = {.lex_state = 66, .external_lex_state = 6},
  [5746] = {.lex_state = 66, .external_lex_state = 6},
  [5747] = {.lex_state = 66, .external_lex_state = 6},
  [5748] = {.lex_state = 64, .external_lex_state = 6},
  [5749] = {.lex_state = 64, .external_lex_state = 6},
  [5750] = {.lex_state = 66, .external_lex_state = 6},
  [5751] = {.lex_state = 66, .external_lex_state = 6},
  [5752] = {.lex_state = 267, .external_lex_state = 6},
  [5753] = {.lex_state = 267, .external_lex_state = 6},
  [5754] = {.lex_state = 267, .external_lex_state = 6},
  [5755] = {.lex_state = 64, .external_lex_state = 6},
  [5756] = {.lex_state = 64, .external_lex_state = 6},
  [5757] = {.lex_state = 64, .external_lex_state = 6},
  [5758] = {.lex_state = 64, .external_lex_state = 6},
  [5759] = {.lex_state = 64, .external_lex_state = 6},
  [5760] = {.lex_state = 66, .external_lex_state = 6},
  [5761] = {.lex_state = 66, .external_lex_state = 6},
  [5762] = {.lex_state = 66, .external_lex_state = 6},
  [5763] = {.lex_state = 66, .external_lex_state = 6},
  [5764] = {.lex_state = 64, .external_lex_state = 6},
  [5765] = {.lex_state = 64, .external_lex_state = 6},
  [5766] = {.lex_state = 66, .external_lex_state = 6},
  [5767] = {.lex_state = 66, .external_lex_state = 6},
  [5768] = {.lex_state = 66, .external_lex_state = 6},
  [5769] = {.lex_state = 64, .external_lex_state = 6},
  [5770] = {.lex_state = 66, .external_lex_state = 6},
  [5771] = {.lex_state = 66, .external_lex_state = 6},
  [5772] = {.lex_state = 64, .external_lex_state = 6},
  [5773] = {.lex_state = 64, .external_lex_state = 6},
  [5774] = {.lex_state = 64, .external_lex_state = 6},
  [5775] = {.lex_state = 64, .external_lex_state = 6},
  [5776] = {.lex_state = 66, .external_lex_state = 6},
  [5777] = {.lex_state = 66, .external_lex_state = 6},
  [5778] = {.lex_state = 66, .external_lex_state = 6},
  [5779] = {.lex_state = 66, .external_lex_state = 6},
  [5780] = {.lex_state = 64, .external_lex_state = 6},
  [5781] = {.lex_state = 64, .external_lex_state = 6},
  [5782] = {.lex_state = 66, .external_lex_state = 6},
  [5783] = {.lex_state = 66, .external_lex_state = 6},
  [5784] = {.lex_state = 267, .external_lex_state = 6},
  [5785] = {.lex_state = 64, .external_lex_state = 6},
  [5786] = {.lex_state = 66, .external_lex_state = 6},
  [5787] = {.lex_state = 64, .external_lex_state = 6},
  [5788] = {.lex_state = 64, .external_lex_state = 6},
  [5789] = {.lex_state = 64, .external_lex_state = 6},
  [5790] = {.lex_state = 64, .external_lex_state = 6},
  [5791] = {.lex_state = 64, .external_lex_state = 6},
  [5792] = {.lex_state = 66, .external_lex_state = 6},
  [5793] = {.lex_state = 66, .external_lex_state = 6},
  [5794] = {.lex_state = 66, .external_lex_state = 6},
  [5795] = {.lex_state = 66, .external_lex_state = 6},
  [5796] = {.lex_state = 64, .external_lex_state = 6},
  [5797] = {.lex_state = 64, .external_lex_state = 6},
  [5798] = {.lex_state = 66, .external_lex_state = 6},
  [5799] = {.lex_state = 66, .external_lex_state = 6},
  [5800] = {.lex_state = 66, .external_lex_state = 6},
  [5801] = {.lex_state = 66, .external_lex_state = 6},
  [5802] = {.lex_state = 267, .external_lex_state = 6},
  [5803] = {.lex_state = 106, .external_lex_state = 6},
  [5804] = {.lex_state = 64, .external_lex_state = 6},
  [5805] = {.lex_state = 64, .external_lex_state = 6},
  [5806] = {.lex_state = 64, .external_lex_state = 6},
  [5807] = {.lex_state = 64, .external_lex_state = 6},
  [5808] = {.lex_state = 66, .external_lex_state = 6},
  [5809] = {.lex_state = 66, .external_lex_state = 6},
  [5810] = {.lex_state = 66, .external_lex_state = 6},
  [5811] = {.lex_state = 66, .external_lex_state = 6},
  [5812] = {.lex_state = 64, .external_lex_state = 6},
  [5813] = {.lex_state = 64, .external_lex_state = 6},
  [5814] = {.lex_state = 66, .external_lex_state = 6},
  [5815] = {.lex_state = 66, .external_lex_state = 6},
  [5816] = {.lex_state = 106, .external_lex_state = 6},
  [5817] = {.lex_state = 66, .external_lex_state = 6},
  [5818] = {.lex_state = 64, .external_lex_state = 6},
  [5819] = {.lex_state = 64, .external_lex_state = 6},
  [5820] = {.lex_state = 64, .external_lex_state = 6},
  [5821] = {.lex_state = 64, .external_lex_state = 6},
  [5822] = {.lex_state = 64, .external_lex_state = 6},
  [5823] = {.lex_state = 64, .external_lex_state = 6},
  [5824] = {.lex_state = 66, .external_lex_state = 6},
  [5825] = {.lex_state = 66, .external_lex_state = 6},
  [5826] = {.lex_state = 66, .external_lex_state = 6},
  [5827] = {.lex_state = 66, .external_lex_state = 6},
  [5828] = {.lex_state = 64, .external_lex_state = 6},
  [5829] = {.lex_state = 64, .external_lex_state = 6},
  [5830] = {.lex_state = 66, .external_lex_state = 6},
  [5831] = {.lex_state = 66, .external_lex_state = 6},
  [5832] = {.lex_state = 64, .external_lex_state = 6},
  [5833] = {.lex_state = 64, .external_lex_state = 6},
  [5834] = {.lex_state = 64, .external_lex_state = 6},
  [5835] = {.lex_state = 64, .external_lex_state = 6},
  [5836] = {.lex_state = 64, .external_lex_state = 6},
  [5837] = {.lex_state = 64, .external_lex_state = 6},
  [5838] = {.lex_state = 64, .external_lex_state = 6},
  [5839] = {.lex_state = 64, .external_lex_state = 6},
  [5840] = {.lex_state = 66, .external_lex_state = 6},
  [5841] = {.lex_state = 66, .external_lex_state = 6},
  [5842] = {.lex_state = 66, .external_lex_state = 6},
  [5843] = {.lex_state = 66, .external_lex_state = 6},
  [5844] = {.lex_state = 64, .external_lex_state = 6},
  [5845] = {.lex_state = 64, .external_lex_state = 6},
  [5846] = {.lex_state = 66, .external_lex_state = 6},
  [5847] = {.lex_state = 66, .external_lex_state = 6},
  [5848] = {.lex_state = 267, .external_lex_state = 6},
  [5849] = {.lex_state = 267, .external_lex_state = 6},
  [5850] = {.lex_state = 64, .external_lex_state = 6},
  [5851] = {.lex_state = 64, .external_lex_state = 6},
  [5852] = {.lex_state = 64, .external_lex_state = 6},
  [5853] = {.lex_state = 64, .external_lex_state = 6},
  [5854] = {.lex_state = 64, .external_lex_state = 6},
  [5855] = {.lex_state = 64, .external_lex_state = 6},
  [5856] = {.lex_state = 66, .external_lex_state = 6},
  [5857] = {.lex_state = 66, .external_lex_state = 6},
  [5858] = {.lex_state = 66, .external_lex_state = 6},
  [5859] = {.lex_state = 66, .external_lex_state = 6},
  [5860] = {.lex_state = 64, .external_lex_state = 6},
  [5861] = {.lex_state = 64, .external_lex_state = 6},
  [5862] = {.lex_state = 66, .external_lex_state = 6},
  [5863] = {.lex_state = 66, .external_lex_state = 6},
  [5864] = {.lex_state = 64, .external_lex_state = 6},
  [5865] = {.lex_state = 64, .external_lex_state = 6},
  [5866] = {.lex_state = 64, .external_lex_state = 6},
  [5867] = {.lex_state = 64, .external_lex_state = 6},
  [5868] = {.lex_state = 64, .external_lex_state = 6},
  [5869] = {.lex_state = 64, .external_lex_state = 6},
  [5870] = {.lex_state = 64, .external_lex_state = 6},
  [5871] = {.lex_state = 64, .external_lex_state = 6},
  [5872] = {.lex_state = 64, .external_lex_state = 6},
  [5873] = {.lex_state = 64, .external_lex_state = 6},
  [5874] = {.lex_state = 64, .external_lex_state = 6},
  [5875] = {.lex_state = 64, .external_lex_state = 6},
  [5876] = {.lex_state = 64, .external_lex_state = 6},
  [5877] = {.lex_state = 64, .external_lex_state = 6},
  [5878] = {.lex_state = 64, .external_lex_state = 6},
  [5879] = {.lex_state = 64, .external_lex_state = 6},
  [5880] = {.lex_state = 64, .external_lex_state = 6},
  [5881] = {.lex_state = 64, .external_lex_state = 6},
  [5882] = {.lex_state = 64, .external_lex_state = 6},
  [5883] = {.lex_state = 64, .external_lex_state = 6},
  [5884] = {.lex_state = 64, .external_lex_state = 6},
  [5885] = {.lex_state = 64, .external_lex_state = 6},
  [5886] = {.lex_state = 64, .external_lex_state = 6},
  [5887] = {.lex_state = 64, .external_lex_state = 6},
  [5888] = {.lex_state = 64, .external_lex_state = 6},
  [5889] = {.lex_state = 64, .external_lex_state = 6},
  [5890] = {.lex_state = 64, .external_lex_state = 6},
  [5891] = {.lex_state = 64, .external_lex_state = 6},
  [5892] = {.lex_state = 64, .external_lex_state = 6},
  [5893] = {.lex_state = 64, .external_lex_state = 6},
  [5894] = {.lex_state = 64, .external_lex_state = 6},
  [5895] = {.lex_state = 64, .external_lex_state = 6},
  [5896] = {.lex_state = 64, .external_lex_state = 6},
  [5897] = {.lex_state = 64, .external_lex_state = 6},
  [5898] = {.lex_state = 64, .external_lex_state = 6},
  [5899] = {.lex_state = 64, .external_lex_state = 6},
  [5900] = {.lex_state = 64, .external_lex_state = 6},
  [5901] = {.lex_state = 64, .external_lex_state = 6},
  [5902] = {.lex_state = 64, .external_lex_state = 6},
  [5903] = {.lex_state = 64, .external_lex_state = 6},
  [5904] = {.lex_state = 106, .external_lex_state = 6},
  [5905] = {.lex_state = 64, .external_lex_state = 6},
  [5906] = {.lex_state = 64, .external_lex_state = 6},
  [5907] = {.lex_state = 64, .external_lex_state = 6},
  [5908] = {.lex_state = 64, .external_lex_state = 6},
  [5909] = {.lex_state = 64, .external_lex_state = 6},
  [5910] = {.lex_state = 64, .external_lex_state = 6},
  [5911] = {.lex_state = 64, .external_lex_state = 6},
  [5912] = {.lex_state = 64, .external_lex_state = 6},
  [5913] = {.lex_state = 64, .external_lex_state = 6},
  [5914] = {.lex_state = 64, .external_lex_state = 6},
  [5915] = {.lex_state = 64, .external_lex_state = 6},
  [5916] = {.lex_state = 64, .external_lex_state = 6},
  [5917] = {.lex_state = 64, .external_lex_state = 6},
  [5918] = {.lex_state = 64, .external_lex_state = 6},
  [5919] = {.lex_state = 64, .external_lex_state = 6},
  [5920] = {.lex_state = 66, .external_lex_state = 6},
  [5921] = {.lex_state = 64, .external_lex_state = 6},
  [5922] = {.lex_state = 64, .external_lex_state = 6},
  [5923] = {.lex_state = 66, .external_lex_state = 6},
  [5924] = {.lex_state = 66, .external_lex_state = 6},
  [5925] = {.lex_state = 66, .external_lex_state = 6},
  [5926] = {.lex_state = 64, .external_lex_state = 6},
  [5927] = {.lex_state = 64, .external_lex_state = 6},
  [5928] = {.lex_state = 106, .external_lex_state = 6},
  [5929] = {.lex_state = 106, .external_lex_state = 6},
  [5930] = {.lex_state = 64, .external_lex_state = 6},
  [5931] = {.lex_state = 267, .external_lex_state = 6},
  [5932] = {.lex_state = 64, .external_lex_state = 6},
  [5933] = {.lex_state = 64, .external_lex_state = 6},
  [5934] = {.lex_state = 64, .external_lex_state = 6},
  [5935] = {.lex_state = 64, .external_lex_state = 6},
  [5936] = {.lex_state = 106, .external_lex_state = 6},
  [5937] = {.lex_state = 106, .external_lex_state = 6},
  [5938] = {.lex_state = 64, .external_lex_state = 6},
  [5939] = {.lex_state = 267, .external_lex_state = 6},
  [5940] = {.lex_state = 66, .external_lex_state = 6},
  [5941] = {.lex_state = 64, .external_lex_state = 6},
  [5942] = {.lex_state = 66, .external_lex_state = 6},
  [5943] = {.lex_state = 267, .external_lex_state = 6},
  [5944] = {.lex_state = 66, .external_lex_state = 6},
  [5945] = {.lex_state = 64, .external_lex_state = 6},
  [5946] = {.lex_state = 106, .external_lex_state = 6},
  [5947] = {.lex_state = 66, .external_lex_state = 6},
  [5948] = {.lex_state = 106, .external_lex_state = 6},
  [5949] = {.lex_state = 267, .external_lex_state = 6},
  [5950] = {.lex_state = 106, .external_lex_state = 6},
  [5951] = {.lex_state = 106, .external_lex_state = 6},
  [5952] = {.lex_state = 66, .external_lex_state = 6},
  [5953] = {.lex_state = 64, .external_lex_state = 6},
  [5954] = {.lex_state = 64, .external_lex_state = 6},
  [5955] = {.lex_state = 267, .external_lex_state = 6},
  [5956] = {.lex_state = 267, .external_lex_state = 6},
  [5957] = {.lex_state = 64, .external_lex_state = 6},
  [5958] = {.lex_state = 64, .external_lex_state = 6},
  [5959] = {.lex_state = 64, .external_lex_state = 6},
  [5960] = {.lex_state = 64, .external_lex_state = 6},
  [5961] = {.lex_state = 64, .external_lex_state = 12},
  [5962] = {.lex_state = 104, .external_lex_state = 6},
  [5963] = {.lex_state = 65, .external_lex_state = 17},
  [5964] = {.lex_state = 64, .external_lex_state = 13},
  [5965] = {.lex_state = 65, .external_lex_state = 18},
  [5966] = {.lex_state = 65, .external_lex_state = 19},
  [5967] = {.lex_state = 65, .external_lex_state = 17},
  [5968] = {.lex_state = 65, .external_lex_state = 18},
  [5969] = {.lex_state = 65, .external_lex_state = 19},
  [5970] = {.lex_state = 67, .external_lex_state = 20},
  [5971] = {.lex_state = 67, .external_lex_state = 21},
  [5972] = {.lex_state = 65, .external_lex_state = 22},
  [5973] = {.lex_state = 65, .external_lex_state = 23},
  [5974] = {.lex_state = 65, .external_lex_state = 24},
  [5975] = {.lex_state = 65, .external_lex_state = 25},
  [5976] = {.lex_state = 65, .external_lex_state = 26},
  [5977] = {.lex_state = 67, .external_lex_state = 20},
  [5978] = {.lex_state = 67, .external_lex_state = 21},
  [5979] = {.lex_state = 65, .external_lex_state = 22},
  [5980] = {.lex_state = 65, .external_lex_state = 23},
  [5981] = {.lex_state = 65, .external_lex_state = 24},
  [5982] = {.lex_state = 65, .external_lex_state = 25},
  [5983] = {.lex_state = 65, .external_lex_state = 26},
  [5984] = {.lex_state = 64, .external_lex_state = 16},
  [5985] = {.lex_state = 64, .external_lex_state = 14},
  [5986] = {.lex_state = 104, .external_lex_state = 6},
  [5987] = {.lex_state = 104, .external_lex_state = 6},
  [5988] = {.lex_state = 64, .external_lex_state = 11},
  [5989] = {.lex_state = 66, .external_lex_state = 9},
  [5990] = {.lex_state = 64, .external_lex_state = 15},
  [5991] = {.lex_state = 64, .external_lex_state = 15},
  [5992] = {.lex_state = 64, .external_lex_state = 16},
  [5993] = {.lex_state = 65, .external_lex_state = 17},
  [5994] = {.lex_state = 65, .external_lex_state = 18},
  [5995] = {.lex_state = 65, .external_lex_state = 19},
  [5996] = {.lex_state = 67, .external_lex_state = 20},
  [5997] = {.lex_state = 67, .external_lex_state = 21},
  [5998] = {.lex_state = 65, .external_lex_state = 22},
  [5999] = {.lex_state = 65, .external_lex_state = 23},
  [6000] = {.lex_state = 65, .external_lex_state = 24},
  [6001] = {.lex_state = 65, .external_lex_state = 25},
  [6002] = {.lex_state = 65, .external_lex_state = 26},
  [6003] = {.lex_state = 65, .external_lex_state = 17},
  [6004] = {.lex_state = 65, .external_lex_state = 18},
  [6005] = {.lex_state = 65, .external_lex_state = 19},
  [6006] = {.lex_state = 67, .external_lex_state = 20},
  [6007] = {.lex_state = 67, .external_lex_state = 21},
  [6008] = {.lex_state = 65, .external_lex_state = 22},
  [6009] = {.lex_state = 65, .external_lex_state = 23},
  [6010] = {.lex_state = 65, .external_lex_state = 24},
  [6011] = {.lex_state = 65, .external_lex_state = 25},
  [6012] = {.lex_state = 65, .external_lex_state = 26},
  [6013] = {.lex_state = 267, .external_lex_state = 6},
  [6014] = {.lex_state = 104, .external_lex_state = 6},
  [6015] = {.lex_state = 65, .external_lex_state = 17},
  [6016] = {.lex_state = 65, .external_lex_state = 18},
  [6017] = {.lex_state = 65, .external_lex_state = 19},
  [6018] = {.lex_state = 67, .external_lex_state = 20},
  [6019] = {.lex_state = 67, .external_lex_state = 21},
  [6020] = {.lex_state = 65, .external_lex_state = 22},
  [6021] = {.lex_state = 65, .external_lex_state = 23},
  [6022] = {.lex_state = 65, .external_lex_state = 24},
  [6023] = {.lex_state = 65, .external_lex_state = 25},
  [6024] = {.lex_state = 65, .external_lex_state = 26},
  [6025] = {.lex_state = 65, .external_lex_state = 17},
  [6026] = {.lex_state = 65, .external_lex_state = 18},
  [6027] = {.lex_state = 65, .external_lex_state = 19},
  [6028] = {.lex_state = 65, .external_lex_state = 17},
  [6029] = {.lex_state = 65, .external_lex_state = 18},
  [6030] = {.lex_state = 65, .external_lex_state = 19},
  [6031] = {.lex_state = 67, .external_lex_state = 20},
  [6032] = {.lex_state = 67, .external_lex_state = 21},
  [6033] = {.lex_state = 65, .external_lex_state = 22},
  [6034] = {.lex_state = 65, .external_lex_state = 23},
  [6035] = {.lex_state = 65, .external_lex_state = 24},
  [6036] = {.lex_state = 65, .external_lex_state = 25},
  [6037] = {.lex_state = 65, .external_lex_state = 26},
  [6038] = {.lex_state = 67, .external_lex_state = 20},
  [6039] = {.lex_state = 64, .external_lex_state = 15},
  [6040] = {.lex_state = 67, .external_lex_state = 21},
  [6041] = {.lex_state = 104, .external_lex_state = 6},
  [6042] = {.lex_state = 65, .external_lex_state = 22},
  [6043] = {.lex_state = 65, .external_lex_state = 23},
  [6044] = {.lex_state = 104, .external_lex_state = 6},
  [6045] = {.lex_state = 65, .external_lex_state = 24},
  [6046] = {.lex_state = 65, .external_lex_state = 25},
  [6047] = {.lex_state = 65, .external_lex_state = 26},
  [6048] = {.lex_state = 65, .external_lex_state = 17},
  [6049] = {.lex_state = 64, .external_lex_state = 8},
  [6050] = {.lex_state = 65, .external_lex_state = 17},
  [6051] = {.lex_state = 65, .external_lex_state = 18},
  [6052] = {.lex_state = 65, .external_lex_state = 19},
  [6053] = {.lex_state = 67, .external_lex_state = 20},
  [6054] = {.lex_state = 67, .external_lex_state = 21},
  [6055] = {.lex_state = 65, .external_lex_state = 22},
  [6056] = {.lex_state = 65, .external_lex_state = 23},
  [6057] = {.lex_state = 65, .external_lex_state = 24},
  [6058] = {.lex_state = 65, .external_lex_state = 25},
  [6059] = {.lex_state = 65, .external_lex_state = 26},
  [6060] = {.lex_state = 65, .external_lex_state = 18},
  [6061] = {.lex_state = 106, .external_lex_state = 6},
  [6062] = {.lex_state = 64, .external_lex_state = 16},
  [6063] = {.lex_state = 65, .external_lex_state = 17},
  [6064] = {.lex_state = 65, .external_lex_state = 18},
  [6065] = {.lex_state = 65, .external_lex_state = 19},
  [6066] = {.lex_state = 67, .external_lex_state = 20},
  [6067] = {.lex_state = 67, .external_lex_state = 21},
  [6068] = {.lex_state = 65, .external_lex_state = 22},
  [6069] = {.lex_state = 65, .external_lex_state = 23},
  [6070] = {.lex_state = 65, .external_lex_state = 24},
  [6071] = {.lex_state = 65, .external_lex_state = 25},
  [6072] = {.lex_state = 65, .external_lex_state = 26},
  [6073] = {.lex_state = 65, .external_lex_state = 17},
  [6074] = {.lex_state = 65, .external_lex_state = 18},
  [6075] = {.lex_state = 65, .external_lex_state = 19},
  [6076] = {.lex_state = 67, .external_lex_state = 20},
  [6077] = {.lex_state = 67, .external_lex_state = 21},
  [6078] = {.lex_state = 65, .external_lex_state = 22},
  [6079] = {.lex_state = 65, .external_lex_state = 23},
  [6080] = {.lex_state = 65, .external_lex_state = 24},
  [6081] = {.lex_state = 65, .external_lex_state = 25},
  [6082] = {.lex_state = 65, .external_lex_state = 26},
  [6083] = {.lex_state = 65, .external_lex_state = 17},
  [6084] = {.lex_state = 65, .external_lex_state = 18},
  [6085] = {.lex_state = 66, .external_lex_state = 9},
  [6086] = {.lex_state = 67, .external_lex_state = 20},
  [6087] = {.lex_state = 67, .external_lex_state = 21},
  [6088] = {.lex_state = 65, .external_lex_state = 22},
  [6089] = {.lex_state = 65, .external_lex_state = 23},
  [6090] = {.lex_state = 65, .external_lex_state = 24},
  [6091] = {.lex_state = 65, .external_lex_state = 25},
  [6092] = {.lex_state = 65, .external_lex_state = 26},
  [6093] = {.lex_state = 65, .external_lex_state = 17},
  [6094] = {.lex_state = 65, .external_lex_state = 18},
  [6095] = {.lex_state = 65, .external_lex_state = 19},
  [6096] = {.lex_state = 67, .external_lex_state = 20},
  [6097] = {.lex_state = 67, .external_lex_state = 21},
  [6098] = {.lex_state = 65, .external_lex_state = 22},
  [6099] = {.lex_state = 65, .external_lex_state = 23},
  [6100] = {.lex_state = 65, .external_lex_state = 24},
  [6101] = {.lex_state = 65, .external_lex_state = 25},
  [6102] = {.lex_state = 65, .external_lex_state = 26},
  [6103] = {.lex_state = 104, .external_lex_state = 6},
  [6104] = {.lex_state = 65, .external_lex_state = 17},
  [6105] = {.lex_state = 65, .external_lex_state = 18},
  [6106] = {.lex_state = 65, .external_lex_state = 19},
  [6107] = {.lex_state = 67, .external_lex_state = 20},
  [6108] = {.lex_state = 67, .external_lex_state = 21},
  [6109] = {.lex_state = 65, .external_lex_state = 22},
  [6110] = {.lex_state = 65, .external_lex_state = 23},
  [6111] = {.lex_state = 65, .external_lex_state = 24},
  [6112] = {.lex_state = 65, .external_lex_state = 25},
  [6113] = {.lex_state = 65, .external_lex_state = 26},
  [6114] = {.lex_state = 64, .external_lex_state = 7},
  [6115] = {.lex_state = 65, .external_lex_state = 17},
  [6116] = {.lex_state = 65, .external_lex_state = 19},
  [6117] = {.lex_state = 65, .external_lex_state = 18},
  [6118] = {.lex_state = 64, .external_lex_state = 7},
  [6119] = {.lex_state = 65, .external_lex_state = 19},
  [6120] = {.lex_state = 67, .external_lex_state = 20},
  [6121] = {.lex_state = 67, .external_lex_state = 21},
  [6122] = {.lex_state = 65, .external_lex_state = 22},
  [6123] = {.lex_state = 65, .external_lex_state = 17},
  [6124] = {.lex_state = 65, .external_lex_state = 18},
  [6125] = {.lex_state = 65, .external_lex_state = 19},
  [6126] = {.lex_state = 67, .external_lex_state = 20},
  [6127] = {.lex_state = 67, .external_lex_state = 21},
  [6128] = {.lex_state = 65, .external_lex_state = 22},
  [6129] = {.lex_state = 65, .external_lex_state = 23},
  [6130] = {.lex_state = 65, .external_lex_state = 24},
  [6131] = {.lex_state = 65, .external_lex_state = 25},
  [6132] = {.lex_state = 65, .external_lex_state = 26},
  [6133] = {.lex_state = 104, .external_lex_state = 6},
  [6134] = {.lex_state = 67, .external_lex_state = 20},
  [6135] = {.lex_state = 67, .external_lex_state = 21},
  [6136] = {.lex_state = 65, .external_lex_state = 23},
  [6137] = {.lex_state = 65, .external_lex_state = 22},
  [6138] = {.lex_state = 65, .external_lex_state = 23},
  [6139] = {.lex_state = 65, .external_lex_state = 24},
  [6140] = {.lex_state = 65, .external_lex_state = 24},
  [6141] = {.lex_state = 104, .external_lex_state = 6},
  [6142] = {.lex_state = 65, .external_lex_state = 25},
  [6143] = {.lex_state = 65, .external_lex_state = 25},
  [6144] = {.lex_state = 65, .external_lex_state = 26},
  [6145] = {.lex_state = 104, .external_lex_state = 6},
  [6146] = {.lex_state = 64, .external_lex_state = 11},
  [6147] = {.lex_state = 104, .external_lex_state = 6},
  [6148] = {.lex_state = 64, .external_lex_state = 13},
  [6149] = {.lex_state = 64, .external_lex_state = 14},
  [6150] = {.lex_state = 65, .external_lex_state = 17},
  [6151] = {.lex_state = 65, .external_lex_state = 18},
  [6152] = {.lex_state = 65, .external_lex_state = 17},
  [6153] = {.lex_state = 65, .external_lex_state = 18},
  [6154] = {.lex_state = 65, .external_lex_state = 19},
  [6155] = {.lex_state = 67, .external_lex_state = 20},
  [6156] = {.lex_state = 67, .external_lex_state = 21},
  [6157] = {.lex_state = 65, .external_lex_state = 22},
  [6158] = {.lex_state = 65, .external_lex_state = 23},
  [6159] = {.lex_state = 65, .external_lex_state = 24},
  [6160] = {.lex_state = 65, .external_lex_state = 25},
  [6161] = {.lex_state = 65, .external_lex_state = 26},
  [6162] = {.lex_state = 65, .external_lex_state = 19},
  [6163] = {.lex_state = 67, .external_lex_state = 20},
  [6164] = {.lex_state = 67, .external_lex_state = 21},
  [6165] = {.lex_state = 65, .external_lex_state = 22},
  [6166] = {.lex_state = 65, .external_lex_state = 23},
  [6167] = {.lex_state = 65, .external_lex_state = 24},
  [6168] = {.lex_state = 65, .external_lex_state = 25},
  [6169] = {.lex_state = 65, .external_lex_state = 26},
  [6170] = {.lex_state = 65, .external_lex_state = 26},
  [6171] = {.lex_state = 66, .external_lex_state = 10},
  [6172] = {.lex_state = 64, .external_lex_state = 12},
  [6173] = {.lex_state = 64, .external_lex_state = 14},
  [6174] = {.lex_state = 64, .external_lex_state = 12},
  [6175] = {.lex_state = 64, .external_lex_state = 8},
  [6176] = {.lex_state = 104, .external_lex_state = 6},
  [6177] = {.lex_state = 106, .external_lex_state = 6},
  [6178] = {.lex_state = 66, .external_lex_state = 9},
  [6179] = {.lex_state = 64, .external_lex_state = 11},
  [6180] = {.lex_state = 66, .external_lex_state = 10},
  [6181] = {.lex_state = 64, .external_lex_state = 7},
  [6182] = {.lex_state = 66, .external_lex_state = 10},
  [6183] = {.lex_state = 64, .external_lex_state = 13},
  [6184] = {.lex_state = 64, .external_lex_state = 8},
  [6185] = {.lex_state = 65, .external_lex_state = 19},
  [6186] = {.lex_state = 65, .external_lex_state = 6},
  [6187] = {.lex_state = 65, .external_lex_state = 6},
  [6188] = {.lex_state = 65, .external_lex_state = 6},
  [6189] = {.lex_state = 65, .external_lex_state = 6},
  [6190] = {.lex_state = 65, .external_lex_state = 6},
  [6191] = {.lex_state = 65, .external_lex_state = 6},
  [6192] = {.lex_state = 67, .external_lex_state = 6},
  [6193] = {.lex_state = 65, .external_lex_state = 6},
  [6194] = {.lex_state = 65, .external_lex_state = 6},
  [6195] = {.lex_state = 65, .external_lex_state = 6},
  [6196] = {.lex_state = 65, .external_lex_state = 6},
  [6197] = {.lex_state = 65, .external_lex_state = 6},
  [6198] = {.lex_state = 65, .external_lex_state = 6},
  [6199] = {.lex_state = 65, .external_lex_state = 6},
  [6200] = {.lex_state = 67, .external_lex_state = 6},
  [6201] = {.lex_state = 66, .external_lex_state = 6},
  [6202] = {.lex_state = 65, .external_lex_state = 6},
  [6203] = {.lex_state = 65, .external_lex_state = 6},
  [6204] = {.lex_state = 66, .external_lex_state = 6},
  [6205] = {.lex_state = 106, .external_lex_state = 6},
  [6206] = {.lex_state = 67, .external_lex_state = 6},
  [6207] = {.lex_state = 67, .external_lex_state = 6},
  [6208] = {.lex_state = 67, .external_lex_state = 6},
  [6209] = {.lex_state = 67, .external_lex_state = 6},
  [6210] = {.lex_state = 65, .external_lex_state = 6},
  [6211] = {.lex_state = 65, .external_lex_state = 6},
  [6212] = {.lex_state = 65, .external_lex_state = 6},
  [6213] = {.lex_state = 65, .external_lex_state = 6},
  [6214] = {.lex_state = 65, .external_lex_state = 6},
  [6215] = {.lex_state = 65, .external_lex_state = 6},
  [6216] = {.lex_state = 65, .external_lex_state = 6},
  [6217] = {.lex_state = 65, .external_lex_state = 6},
  [6218] = {.lex_state = 65, .external_lex_state = 6},
  [6219] = {.lex_state = 65, .external_lex_state = 6},
  [6220] = {.lex_state = 65, .external_lex_state = 6},
  [6221] = {.lex_state = 65, .external_lex_state = 6},
  [6222] = {.lex_state = 67, .external_lex_state = 6},
  [6223] = {.lex_state = 67, .external_lex_state = 6},
  [6224] = {.lex_state = 67, .external_lex_state = 6},
  [6225] = {.lex_state = 67, .external_lex_state = 6},
  [6226] = {.lex_state = 65, .external_lex_state = 6},
  [6227] = {.lex_state = 65, .external_lex_state = 6},
  [6228] = {.lex_state = 65, .external_lex_state = 6},
  [6229] = {.lex_state = 65, .external_lex_state = 6},
  [6230] = {.lex_state = 65, .external_lex_state = 6},
  [6231] = {.lex_state = 65, .external_lex_state = 6},
  [6232] = {.lex_state = 67, .external_lex_state = 6},
  [6233] = {.lex_state = 67, .external_lex_state = 6},
  [6234] = {.lex_state = 67, .external_lex_state = 6},
  [6235] = {.lex_state = 67, .external_lex_state = 6},
  [6236] = {.lex_state = 65, .external_lex_state = 6},
  [6237] = {.lex_state = 65, .external_lex_state = 6},
  [6238] = {.lex_state = 65, .external_lex_state = 6},
  [6239] = {.lex_state = 65, .external_lex_state = 6},
  [6240] = {.lex_state = 65, .external_lex_state = 6},
  [6241] = {.lex_state = 65, .external_lex_state = 6},
  [6242] = {.lex_state = 65, .external_lex_state = 6},
  [6243] = {.lex_state = 65, .external_lex_state = 6},
  [6244] = {.lex_state = 65, .external_lex_state = 6},
  [6245] = {.lex_state = 65, .external_lex_state = 6},
  [6246] = {.lex_state = 65, .external_lex_state = 6},
  [6247] = {.lex_state = 65, .external_lex_state = 6},
  [6248] = {.lex_state = 65, .external_lex_state = 6},
  [6249] = {.lex_state = 65, .external_lex_state = 6},
  [6250] = {.lex_state = 65, .external_lex_state = 6},
  [6251] = {.lex_state = 65, .external_lex_state = 6},
  [6252] = {.lex_state = 65, .external_lex_state = 6},
  [6253] = {.lex_state = 65, .external_lex_state = 6},
  [6254] = {.lex_state = 65, .external_lex_state = 6},
  [6255] = {.lex_state = 67, .external_lex_state = 6},
  [6256] = {.lex_state = 67, .external_lex_state = 6},
  [6257] = {.lex_state = 65, .external_lex_state = 6},
  [6258] = {.lex_state = 65, .external_lex_state = 6},
  [6259] = {.lex_state = 65, .external_lex_state = 6},
  [6260] = {.lex_state = 65, .external_lex_state = 6},
  [6261] = {.lex_state = 65, .external_lex_state = 6},
  [6262] = {.lex_state = 65, .external_lex_state = 6},
  [6263] = {.lex_state = 65, .external_lex_state = 6},
  [6264] = {.lex_state = 65, .external_lex_state = 6},
  [6265] = {.lex_state = 65, .external_lex_state = 6},
  [6266] = {.lex_state = 65, .external_lex_state = 6},
  [6267] = {.lex_state = 65, .external_lex_state = 6},
  [6268] = {.lex_state = 65, .external_lex_state = 6},
  [6269] = {.lex_state = 65, .external_lex_state = 6},
  [6270] = {.lex_state = 67, .external_lex_state = 6},
  [6271] = {.lex_state = 67, .external_lex_state = 6},
  [6272] = {.lex_state = 65, .external_lex_state = 6},
  [6273] = {.lex_state = 65, .external_lex_state = 6},
  [6274] = {.lex_state = 65, .external_lex_state = 6},
  [6275] = {.lex_state = 65, .external_lex_state = 6},
  [6276] = {.lex_state = 65, .external_lex_state = 6},
  [6277] = {.lex_state = 65, .external_lex_state = 6},
  [6278] = {.lex_state = 65, .external_lex_state = 6},
  [6279] = {.lex_state = 67, .external_lex_state = 6},
  [6280] = {.lex_state = 67, .external_lex_state = 6},
  [6281] = {.lex_state = 65, .external_lex_state = 6},
  [6282] = {.lex_state = 65, .external_lex_state = 6},
  [6283] = {.lex_state = 65, .external_lex_state = 6},
  [6284] = {.lex_state = 67, .external_lex_state = 6},
  [6285] = {.lex_state = 67, .external_lex_state = 6},
  [6286] = {.lex_state = 65, .external_lex_state = 6},
  [6287] = {.lex_state = 65, .external_lex_state = 6},
  [6288] = {.lex_state = 67, .external_lex_state = 6},
  [6289] = {.lex_state = 67, .external_lex_state = 6},
  [6290] = {.lex_state = 65, .external_lex_state = 6},
  [6291] = {.lex_state = 65, .external_lex_state = 6},
  [6292] = {.lex_state = 65, .external_lex_state = 6},
  [6293] = {.lex_state = 65, .external_lex_state = 6},
  [6294] = {.lex_state = 65, .external_lex_state = 6},
  [6295] = {.lex_state = 65, .external_lex_state = 6},
  [6296] = {.lex_state = 67, .external_lex_state = 6},
  [6297] = {.lex_state = 65, .external_lex_state = 6},
  [6298] = {.lex_state = 65, .external_lex_state = 6},
  [6299] = {.lex_state = 65, .external_lex_state = 6},
  [6300] = {.lex_state = 67, .external_lex_state = 6},
  [6301] = {.lex_state = 65, .external_lex_state = 6},
  [6302] = {.lex_state = 65, .external_lex_state = 6},
  [6303] = {.lex_state = 65, .external_lex_state = 6},
  [6304] = {.lex_state = 65, .external_lex_state = 6},
  [6305] = {.lex_state = 65, .external_lex_state = 6},
  [6306] = {.lex_state = 65, .external_lex_state = 6},
  [6307] = {.lex_state = 65, .external_lex_state = 6},
  [6308] = {.lex_state = 65, .external_lex_state = 6},
  [6309] = {.lex_state = 65, .external_lex_state = 6},
  [6310] = {.lex_state = 65, .external_lex_state = 6},
  [6311] = {.lex_state = 65, .external_lex_state = 6},
  [6312] = {.lex_state = 67, .external_lex_state = 6},
  [6313] = {.lex_state = 67, .external_lex_state = 6},
  [6314] = {.lex_state = 67, .external_lex_state = 6},
  [6315] = {.lex_state = 67, .external_lex_state = 6},
  [6316] = {.lex_state = 65, .external_lex_state = 6},
  [6317] = {.lex_state = 65, .external_lex_state = 6},
  [6318] = {.lex_state = 65, .external_lex_state = 6},
  [6319] = {.lex_state = 65, .external_lex_state = 6},
  [6320] = {.lex_state = 65, .external_lex_state = 6},
  [6321] = {.lex_state = 65, .external_lex_state = 6},
  [6322] = {.lex_state = 65, .external_lex_state = 6},
  [6323] = {.lex_state = 65, .external_lex_state = 6},
  [6324] = {.lex_state = 65, .external_lex_state = 6},
  [6325] = {.lex_state = 65, .external_lex_state = 6},
  [6326] = {.lex_state = 65, .external_lex_state = 6},
  [6327] = {.lex_state = 65, .external_lex_state = 6},
  [6328] = {.lex_state = 65, .external_lex_state = 6},
  [6329] = {.lex_state = 65, .external_lex_state = 6},
  [6330] = {.lex_state = 65, .external_lex_state = 6},
  [6331] = {.lex_state = 67, .external_lex_state = 6},
  [6332] = {.lex_state = 67, .external_lex_state = 6},
  [6333] = {.lex_state = 65, .external_lex_state = 6},
  [6334] = {.lex_state = 65, .external_lex_state = 6},
  [6335] = {.lex_state = 65, .external_lex_state = 6},
  [6336] = {.lex_state = 65, .external_lex_state = 6},
  [6337] = {.lex_state = 65, .external_lex_state = 6},
  [6338] = {.lex_state = 65, .external_lex_state = 6},
  [6339] = {.lex_state = 67, .external_lex_state = 6},
  [6340] = {.lex_state = 65, .external_lex_state = 6},
  [6341] = {.lex_state = 67, .external_lex_state = 6},
  [6342] = {.lex_state = 67, .external_lex_state = 6},
  [6343] = {.lex_state = 67, .external_lex_state = 6},
  [6344] = {.lex_state = 65, .external_lex_state = 6},
  [6345] = {.lex_state = 67, .external_lex_state = 6},
  [6346] = {.lex_state = 65, .external_lex_state = 6},
  [6347] = {.lex_state = 65, .external_lex_state = 6},
  [6348] = {.lex_state = 65, .external_lex_state = 6},
  [6349] = {.lex_state = 65, .external_lex_state = 6},
  [6350] = {.lex_state = 65, .external_lex_state = 6},
  [6351] = {.lex_state = 65, .external_lex_state = 6},
  [6352] = {.lex_state = 67, .external_lex_state = 6},
  [6353] = {.lex_state = 67, .external_lex_state = 6},
  [6354] = {.lex_state = 67, .external_lex_state = 6},
  [6355] = {.lex_state = 67, .external_lex_state = 6},
  [6356] = {.lex_state = 65, .external_lex_state = 6},
  [6357] = {.lex_state = 65, .external_lex_state = 6},
  [6358] = {.lex_state = 65, .external_lex_state = 6},
  [6359] = {.lex_state = 65, .external_lex_state = 6},
  [6360] = {.lex_state = 65, .external_lex_state = 6},
  [6361] = {.lex_state = 65, .external_lex_state = 6},
  [6362] = {.lex_state = 65, .external_lex_state = 6},
  [6363] = {.lex_state = 65, .external_lex_state = 6},
  [6364] = {.lex_state = 65, .external_lex_state = 6},
  [6365] = {.lex_state = 67, .external_lex_state = 6},
  [6366] = {.lex_state = 67, .external_lex_state = 6},
  [6367] = {.lex_state = 67, .external_lex_state = 6},
  [6368] = {.lex_state = 67, .external_lex_state = 6},
  [6369] = {.lex_state = 65, .external_lex_state = 6},
  [6370] = {.lex_state = 65, .external_lex_state = 6},
  [6371] = {.lex_state = 65, .external_lex_state = 6},
  [6372] = {.lex_state = 65, .external_lex_state = 6},
  [6373] = {.lex_state = 65, .external_lex_state = 6},
  [6374] = {.lex_state = 65, .external_lex_state = 6},
  [6375] = {.lex_state = 65, .external_lex_state = 6},
  [6376] = {.lex_state = 65, .external_lex_state = 6},
  [6377] = {.lex_state = 65, .external_lex_state = 6},
  [6378] = {.lex_state = 65, .external_lex_state = 6},
  [6379] = {.lex_state = 65, .external_lex_state = 6},
  [6380] = {.lex_state = 65, .external_lex_state = 6},
  [6381] = {.lex_state = 65, .external_lex_state = 6},
  [6382] = {.lex_state = 65, .external_lex_state = 6},
  [6383] = {.lex_state = 67, .external_lex_state = 6},
  [6384] = {.lex_state = 67, .external_lex_state = 6},
  [6385] = {.lex_state = 65, .external_lex_state = 6},
  [6386] = {.lex_state = 65, .external_lex_state = 6},
  [6387] = {.lex_state = 65, .external_lex_state = 6},
  [6388] = {.lex_state = 65, .external_lex_state = 6},
  [6389] = {.lex_state = 65, .external_lex_state = 6},
  [6390] = {.lex_state = 65, .external_lex_state = 6},
  [6391] = {.lex_state = 65, .external_lex_state = 6},
  [6392] = {.lex_state = 65, .external_lex_state = 6},
  [6393] = {.lex_state = 65, .external_lex_state = 6},
  [6394] = {.lex_state = 65, .external_lex_state = 6},
  [6395] = {.lex_state = 65, .external_lex_state = 6},
  [6396] = {.lex_state = 65, .external_lex_state = 6},
  [6397] = {.lex_state = 65, .external_lex_state = 6},
  [6398] = {.lex_state = 65, .external_lex_state = 6},
  [6399] = {.lex_state = 65, .external_lex_state = 6},
  [6400] = {.lex_state = 65, .external_lex_state = 6},
  [6401] = {.lex_state = 65, .external_lex_state = 6},
  [6402] = {.lex_state = 267, .external_lex_state = 6},
  [6403] = {.lex_state = 65, .external_lex_state = 6},
  [6404] = {.lex_state = 65, .external_lex_state = 6},
  [6405] = {.lex_state = 65, .external_lex_state = 6},
  [6406] = {.lex_state = 65, .external_lex_state = 6},
  [6407] = {.lex_state = 65, .external_lex_state = 6},
  [6408] = {.lex_state = 65, .external_lex_state = 6},
  [6409] = {.lex_state = 64, .external_lex_state = 6},
  [6410] = {.lex_state = 64, .external_lex_state = 6},
  [6411] = {.lex_state = 65, .external_lex_state = 6},
  [6412] = {.lex_state = 65, .external_lex_state = 6},
  [6413] = {.lex_state = 65, .external_lex_state = 6},
  [6414] = {.lex_state = 65, .external_lex_state = 6},
  [6415] = {.lex_state = 65, .external_lex_state = 6},
  [6416] = {.lex_state = 67, .external_lex_state = 6},
  [6417] = {.lex_state = 67, .external_lex_state = 6},
  [6418] = {.lex_state = 65, .external_lex_state = 6},
  [6419] = {.lex_state = 65, .external_lex_state = 6},
  [6420] = {.lex_state = 65, .external_lex_state = 6},
  [6421] = {.lex_state = 65, .external_lex_state = 6},
  [6422] = {.lex_state = 65, .external_lex_state = 6},
  [6423] = {.lex_state = 65, .external_lex_state = 6},
  [6424] = {.lex_state = 65, .external_lex_state = 6},
  [6425] = {.lex_state = 65, .external_lex_state = 6},
  [6426] = {.lex_state = 65, .external_lex_state = 6},
  [6427] = {.lex_state = 65, .external_lex_state = 6},
  [6428] = {.lex_state = 67, .external_lex_state = 6},
  [6429] = {.lex_state = 67, .external_lex_state = 6},
  [6430] = {.lex_state = 67, .external_lex_state = 6},
  [6431] = {.lex_state = 67, .external_lex_state = 6},
  [6432] = {.lex_state = 65, .external_lex_state = 6},
  [6433] = {.lex_state = 65, .external_lex_state = 6},
  [6434] = {.lex_state = 65, .external_lex_state = 6},
  [6435] = {.lex_state = 65, .external_lex_state = 6},
  [6436] = {.lex_state = 65, .external_lex_state = 6},
  [6437] = {.lex_state = 65, .external_lex_state = 6},
  [6438] = {.lex_state = 65, .external_lex_state = 6},
  [6439] = {.lex_state = 65, .external_lex_state = 6},
  [6440] = {.lex_state = 65, .external_lex_state = 6},
  [6441] = {.lex_state = 65, .external_lex_state = 6},
  [6442] = {.lex_state = 65, .external_lex_state = 6},
  [6443] = {.lex_state = 65, .external_lex_state = 6},
  [6444] = {.lex_state = 65, .external_lex_state = 6},
  [6445] = {.lex_state = 65, .external_lex_state = 6},
  [6446] = {.lex_state = 65, .external_lex_state = 6},
  [6447] = {.lex_state = 67, .external_lex_state = 6},
  [6448] = {.lex_state = 65, .external_lex_state = 6},
  [6449] = {.lex_state = 65, .external_lex_state = 6},
  [6450] = {.lex_state = 65, .external_lex_state = 6},
  [6451] = {.lex_state = 67, .external_lex_state = 6},
  [6452] = {.lex_state = 67, .external_lex_state = 6},
  [6453] = {.lex_state = 65, .external_lex_state = 6},
  [6454] = {.lex_state = 65, .external_lex_state = 6},
  [6455] = {.lex_state = 65, .external_lex_state = 6},
  [6456] = {.lex_state = 65, .external_lex_state = 6},
  [6457] = {.lex_state = 65, .external_lex_state = 6},
  [6458] = {.lex_state = 67, .external_lex_state = 6},
  [6459] = {.lex_state = 67, .external_lex_state = 6},
  [6460] = {.lex_state = 67, .external_lex_state = 6},
  [6461] = {.lex_state = 65, .external_lex_state = 6},
  [6462] = {.lex_state = 65, .external_lex_state = 6},
  [6463] = {.lex_state = 65, .external_lex_state = 6},
  [6464] = {.lex_state = 65, .external_lex_state = 6},
  [6465] = {.lex_state = 65, .external_lex_state = 6},
  [6466] = {.lex_state = 65, .external_lex_state = 6},
  [6467] = {.lex_state = 65, .external_lex_state = 6},
  [6468] = {.lex_state = 65, .external_lex_state = 6},
  [6469] = {.lex_state = 65, .external_lex_state = 6},
  [6470] = {.lex_state = 65, .external_lex_state = 6},
  [6471] = {.lex_state = 65, .external_lex_state = 6},
  [6472] = {.lex_state = 65, .external_lex_state = 6},
  [6473] = {.lex_state = 65, .external_lex_state = 6},
  [6474] = {.lex_state = 65, .external_lex_state = 6},
  [6475] = {.lex_state = 65, .external_lex_state = 6},
  [6476] = {.lex_state = 65, .external_lex_state = 6},
  [6477] = {.lex_state = 65, .external_lex_state = 6},
  [6478] = {.lex_state = 65, .external_lex_state = 6},
  [6479] = {.lex_state = 52, .external_lex_state = 6},
  [6480] = {.lex_state = 67, .external_lex_state = 6},
  [6481] = {.lex_state = 67, .external_lex_state = 6},
  [6482] = {.lex_state = 65, .external_lex_state = 6},
  [6483] = {.lex_state = 65, .external_lex_state = 6},
  [6484] = {.lex_state = 67, .external_lex_state = 6},
  [6485] = {.lex_state = 67, .external_lex_state = 6},
  [6486] = {.lex_state = 65, .external_lex_state = 6},
  [6487] = {.lex_state = 65, .external_lex_state = 6},
  [6488] = {.lex_state = 65, .external_lex_state = 6},
  [6489] = {.lex_state = 65, .external_lex_state = 6},
  [6490] = {.lex_state = 67, .external_lex_state = 6},
  [6491] = {.lex_state = 67, .external_lex_state = 6},
  [6492] = {.lex_state = 65, .external_lex_state = 6},
  [6493] = {.lex_state = 65, .external_lex_state = 6},
  [6494] = {.lex_state = 65, .external_lex_state = 6},
  [6495] = {.lex_state = 65, .external_lex_state = 6},
  [6496] = {.lex_state = 65, .external_lex_state = 6},
  [6497] = {.lex_state = 65, .external_lex_state = 6},
  [6498] = {.lex_state = 65, .external_lex_state = 6},
  [6499] = {.lex_state = 65, .external_lex_state = 6},
  [6500] = {.lex_state = 65, .external_lex_state = 6},
  [6501] = {.lex_state = 67, .external_lex_state = 6},
  [6502] = {.lex_state = 65, .external_lex_state = 6},
  [6503] = {.lex_state = 67, .external_lex_state = 6},
  [6504] = {.lex_state = 67, .external_lex_state = 6},
  [6505] = {.lex_state = 65, .external_lex_state = 6},
  [6506] = {.lex_state = 65, .external_lex_state = 6},
  [6507] = {.lex_state = 65, .external_lex_state = 6},
  [6508] = {.lex_state = 65, .external_lex_state = 6},
  [6509] = {.lex_state = 65, .external_lex_state = 6},
  [6510] = {.lex_state = 65, .external_lex_state = 6},
  [6511] = {.lex_state = 65, .external_lex_state = 6},
  [6512] = {.lex_state = 65, .external_lex_state = 6},
  [6513] = {.lex_state = 65, .external_lex_state = 6},
  [6514] = {.lex_state = 65, .external_lex_state = 6},
  [6515] = {.lex_state = 65, .external_lex_state = 6},
  [6516] = {.lex_state = 65, .external_lex_state = 6},
  [6517] = {.lex_state = 65, .external_lex_state = 6},
  [6518] = {.lex_state = 65, .external_lex_state = 6},
  [6519] = {.lex_state = 65, .external_lex_state = 6},
  [6520] = {.lex_state = 65, .external_lex_state = 6},
  [6521] = {.lex_state = 65, .external_lex_state = 6},
  [6522] = {.lex_state = 65, .external_lex_state = 6},
  [6523] = {.lex_state = 67, .external_lex_state = 6},
  [6524] = {.lex_state = 67, .external_lex_state = 6},
  [6525] = {.lex_state = 65, .external_lex_state = 6},
  [6526] = {.lex_state = 65, .external_lex_state = 6},
  [6527] = {.lex_state = 65, .external_lex_state = 6},
  [6528] = {.lex_state = 65, .external_lex_state = 6},
  [6529] = {.lex_state = 65, .external_lex_state = 6},
  [6530] = {.lex_state = 65, .external_lex_state = 6},
  [6531] = {.lex_state = 65, .external_lex_state = 6},
  [6532] = {.lex_state = 65, .external_lex_state = 6},
  [6533] = {.lex_state = 65, .external_lex_state = 6},
  [6534] = {.lex_state = 65, .external_lex_state = 6},
  [6535] = {.lex_state = 67, .external_lex_state = 6},
  [6536] = {.lex_state = 65, .external_lex_state = 6},
  [6537] = {.lex_state = 65, .external_lex_state = 6},
  [6538] = {.lex_state = 65, .external_lex_state = 6},
  [6539] = {.lex_state = 64, .external_lex_state = 6},
  [6540] = {.lex_state = 65, .external_lex_state = 19},
  [6541] = {.lex_state = 65, .external_lex_state = 6},
  [6542] = {.lex_state = 65, .external_lex_state = 6},
  [6543] = {.lex_state = 65, .external_lex_state = 6},
  [6544] = {.lex_state = 67, .external_lex_state = 6},
  [6545] = {.lex_state = 67, .external_lex_state = 6},
  [6546] = {.lex_state = 67, .external_lex_state = 6},
  [6547] = {.lex_state = 67, .external_lex_state = 6},
  [6548] = {.lex_state = 65, .external_lex_state = 6},
  [6549] = {.lex_state = 65, .external_lex_state = 6},
  [6550] = {.lex_state = 65, .external_lex_state = 6},
  [6551] = {.lex_state = 65, .external_lex_state = 6},
  [6552] = {.lex_state = 65, .external_lex_state = 6},
  [6553] = {.lex_state = 65, .external_lex_state = 6},
  [6554] = {.lex_state = 65, .external_lex_state = 6},
  [6555] = {.lex_state = 65, .external_lex_state = 6},
  [6556] = {.lex_state = 65, .external_lex_state = 6},
  [6557] = {.lex_state = 65, .external_lex_state = 6},
  [6558] = {.lex_state = 65, .external_lex_state = 6},
  [6559] = {.lex_state = 65, .external_lex_state = 6},
  [6560] = {.lex_state = 65, .external_lex_state = 6},
  [6561] = {.lex_state = 65, .external_lex_state = 6},
  [6562] = {.lex_state = 65, .external_lex_state = 6},
  [6563] = {.lex_state = 64, .external_lex_state = 6},
  [6564] = {.lex_state = 67, .external_lex_state = 6},
  [6565] = {.lex_state = 65, .external_lex_state = 6},
  [6566] = {.lex_state = 65, .external_lex_state = 6},
  [6567] = {.lex_state = 65, .external_lex_state = 6},
  [6568] = {.lex_state = 65, .external_lex_state = 6},
  [6569] = {.lex_state = 65, .external_lex_state = 6},
  [6570] = {.lex_state = 65, .external_lex_state = 6},
  [6571] = {.lex_state = 65, .external_lex_state = 6},
  [6572] = {.lex_state = 65, .external_lex_state = 6},
  [6573] = {.lex_state = 65, .external_lex_state = 17},
  [6574] = {.lex_state = 65, .external_lex_state = 6},
  [6575] = {.lex_state = 65, .external_lex_state = 6},
  [6576] = {.lex_state = 65, .external_lex_state = 6},
  [6577] = {.lex_state = 65, .external_lex_state = 6},
  [6578] = {.lex_state = 65, .external_lex_state = 6},
  [6579] = {.lex_state = 67, .external_lex_state = 6},
  [6580] = {.lex_state = 67, .external_lex_state = 6},
  [6581] = {.lex_state = 65, .external_lex_state = 6},
  [6582] = {.lex_state = 65, .external_lex_state = 6},
  [6583] = {.lex_state = 65, .external_lex_state = 6},
  [6584] = {.lex_state = 65, .external_lex_state = 6},
  [6585] = {.lex_state = 65, .external_lex_state = 6},
  [6586] = {.lex_state = 65, .external_lex_state = 6},
  [6587] = {.lex_state = 65, .external_lex_state = 6},
  [6588] = {.lex_state = 65, .external_lex_state = 6},
  [6589] = {.lex_state = 65, .external_lex_state = 6},
  [6590] = {.lex_state = 52, .external_lex_state = 6},
  [6591] = {.lex_state = 65, .external_lex_state = 6},
  [6592] = {.lex_state = 65, .external_lex_state = 6},
  [6593] = {.lex_state = 65, .external_lex_state = 6},
  [6594] = {.lex_state = 65, .external_lex_state = 6},
  [6595] = {.lex_state = 65, .external_lex_state = 6},
  [6596] = {.lex_state = 65, .external_lex_state = 6},
  [6597] = {.lex_state = 65, .external_lex_state = 6},
  [6598] = {.lex_state = 67, .external_lex_state = 6},
  [6599] = {.lex_state = 67, .external_lex_state = 6},
  [6600] = {.lex_state = 67, .external_lex_state = 6},
  [6601] = {.lex_state = 67, .external_lex_state = 6},
  [6602] = {.lex_state = 65, .external_lex_state = 6},
  [6603] = {.lex_state = 65, .external_lex_state = 6},
  [6604] = {.lex_state = 65, .external_lex_state = 6},
  [6605] = {.lex_state = 65, .external_lex_state = 6},
  [6606] = {.lex_state = 65, .external_lex_state = 6},
  [6607] = {.lex_state = 65, .external_lex_state = 6},
  [6608] = {.lex_state = 65, .external_lex_state = 6},
  [6609] = {.lex_state = 65, .external_lex_state = 6},
  [6610] = {.lex_state = 65, .external_lex_state = 6},
  [6611] = {.lex_state = 65, .external_lex_state = 6},
  [6612] = {.lex_state = 65, .external_lex_state = 6},
  [6613] = {.lex_state = 65, .external_lex_state = 6},
  [6614] = {.lex_state = 65, .external_lex_state = 6},
  [6615] = {.lex_state = 65, .external_lex_state = 6},
  [6616] = {.lex_state = 67, .external_lex_state = 6},
  [6617] = {.lex_state = 67, .external_lex_state = 6},
  [6618] = {.lex_state = 65, .external_lex_state = 6},
  [6619] = {.lex_state = 65, .external_lex_state = 6},
  [6620] = {.lex_state = 65, .external_lex_state = 6},
  [6621] = {.lex_state = 65, .external_lex_state = 6},
  [6622] = {.lex_state = 65, .external_lex_state = 6},
  [6623] = {.lex_state = 65, .external_lex_state = 6},
  [6624] = {.lex_state = 65, .external_lex_state = 6},
  [6625] = {.lex_state = 65, .external_lex_state = 6},
  [6626] = {.lex_state = 65, .external_lex_state = 6},
  [6627] = {.lex_state = 67, .external_lex_state = 6},
  [6628] = {.lex_state = 67, .external_lex_state = 6},
  [6629] = {.lex_state = 65, .external_lex_state = 6},
  [6630] = {.lex_state = 65, .external_lex_state = 6},
  [6631] = {.lex_state = 65, .external_lex_state = 6},
  [6632] = {.lex_state = 65, .external_lex_state = 6},
  [6633] = {.lex_state = 65, .external_lex_state = 6},
  [6634] = {.lex_state = 65, .external_lex_state = 6},
  [6635] = {.lex_state = 65, .external_lex_state = 6},
  [6636] = {.lex_state = 67, .external_lex_state = 6},
  [6637] = {.lex_state = 65, .external_lex_state = 6},
  [6638] = {.lex_state = 65, .external_lex_state = 6},
  [6639] = {.lex_state = 67, .external_lex_state = 6},
  [6640] = {.lex_state = 67, .external_lex_state = 6},
  [6641] = {.lex_state = 65, .external_lex_state = 6},
  [6642] = {.lex_state = 65, .external_lex_state = 6},
  [6643] = {.lex_state = 67, .external_lex_state = 6},
  [6644] = {.lex_state = 67, .external_lex_state = 6},
  [6645] = {.lex_state = 65, .external_lex_state = 18},
  [6646] = {.lex_state = 67, .external_lex_state = 6},
  [6647] = {.lex_state = 65, .external_lex_state = 6},
  [6648] = {.lex_state = 65, .external_lex_state = 6},
  [6649] = {.lex_state = 65, .external_lex_state = 6},
  [6650] = {.lex_state = 65, .external_lex_state = 6},
  [6651] = {.lex_state = 65, .external_lex_state = 6},
  [6652] = {.lex_state = 65, .external_lex_state = 6},
  [6653] = {.lex_state = 65, .external_lex_state = 6},
  [6654] = {.lex_state = 65, .external_lex_state = 6},
  [6655] = {.lex_state = 65, .external_lex_state = 6},
  [6656] = {.lex_state = 65, .external_lex_state = 6},
  [6657] = {.lex_state = 65, .external_lex_state = 6},
  [6658] = {.lex_state = 267, .external_lex_state = 6},
  [6659] = {.lex_state = 52, .external_lex_state = 6},
  [6660] = {.lex_state = 64, .external_lex_state = 6},
  [6661] = {.lex_state = 267, .external_lex_state = 6},
  [6662] = {.lex_state = 64, .external_lex_state = 6},
  [6663] = {.lex_state = 267, .external_lex_state = 6},
  [6664] = {.lex_state = 65, .external_lex_state = 6},
  [6665] = {.lex_state = 65, .external_lex_state = 6},
  [6666] = {.lex_state = 65, .external_lex_state = 6},
  [6667] = {.lex_state = 65, .external_lex_state = 6},
  [6668] = {.lex_state = 65, .external_lex_state = 6},
  [6669] = {.lex_state = 65, .external_lex_state = 6},
  [6670] = {.lex_state = 65, .external_lex_state = 6},
  [6671] = {.lex_state = 65, .external_lex_state = 6},
  [6672] = {.lex_state = 67, .external_lex_state = 6},
  [6673] = {.lex_state = 67, .external_lex_state = 6},
  [6674] = {.lex_state = 67, .external_lex_state = 6},
  [6675] = {.lex_state = 67, .external_lex_state = 6},
  [6676] = {.lex_state = 65, .external_lex_state = 6},
  [6677] = {.lex_state = 65, .external_lex_state = 6},
  [6678] = {.lex_state = 65, .external_lex_state = 6},
  [6679] = {.lex_state = 65, .external_lex_state = 6},
  [6680] = {.lex_state = 65, .external_lex_state = 6},
  [6681] = {.lex_state = 65, .external_lex_state = 6},
  [6682] = {.lex_state = 65, .external_lex_state = 6},
  [6683] = {.lex_state = 65, .external_lex_state = 6},
  [6684] = {.lex_state = 65, .external_lex_state = 6},
  [6685] = {.lex_state = 65, .external_lex_state = 6},
  [6686] = {.lex_state = 65, .external_lex_state = 6},
  [6687] = {.lex_state = 65, .external_lex_state = 6},
  [6688] = {.lex_state = 65, .external_lex_state = 6},
  [6689] = {.lex_state = 65, .external_lex_state = 6},
  [6690] = {.lex_state = 65, .external_lex_state = 6},
  [6691] = {.lex_state = 65, .external_lex_state = 6},
  [6692] = {.lex_state = 65, .external_lex_state = 6},
  [6693] = {.lex_state = 67, .external_lex_state = 6},
  [6694] = {.lex_state = 67, .external_lex_state = 6},
  [6695] = {.lex_state = 67, .external_lex_state = 6},
  [6696] = {.lex_state = 67, .external_lex_state = 6},
  [6697] = {.lex_state = 65, .external_lex_state = 6},
  [6698] = {.lex_state = 65, .external_lex_state = 6},
  [6699] = {.lex_state = 67, .external_lex_state = 6},
  [6700] = {.lex_state = 267, .external_lex_state = 6},
  [6701] = {.lex_state = 65, .external_lex_state = 6},
  [6702] = {.lex_state = 65, .external_lex_state = 6},
  [6703] = {.lex_state = 65, .external_lex_state = 6},
  [6704] = {.lex_state = 52, .external_lex_state = 6},
  [6705] = {.lex_state = 65, .external_lex_state = 6},
  [6706] = {.lex_state = 65, .external_lex_state = 6},
  [6707] = {.lex_state = 65, .external_lex_state = 6},
  [6708] = {.lex_state = 65, .external_lex_state = 6},
  [6709] = {.lex_state = 65, .external_lex_state = 6},
  [6710] = {.lex_state = 106, .external_lex_state = 6},
  [6711] = {.lex_state = 65, .external_lex_state = 6},
  [6712] = {.lex_state = 65, .external_lex_state = 6},
  [6713] = {.lex_state = 65, .external_lex_state = 6},
  [6714] = {.lex_state = 65, .external_lex_state = 6},
  [6715] = {.lex_state = 67, .external_lex_state = 20},
  [6716] = {.lex_state = 65, .external_lex_state = 6},
  [6717] = {.lex_state = 67, .external_lex_state = 6},
  [6718] = {.lex_state = 67, .external_lex_state = 6},
  [6719] = {.lex_state = 67, .external_lex_state = 21},
  [6720] = {.lex_state = 67, .external_lex_state = 6},
  [6721] = {.lex_state = 65, .external_lex_state = 6},
  [6722] = {.lex_state = 65, .external_lex_state = 22},
  [6723] = {.lex_state = 65, .external_lex_state = 6},
  [6724] = {.lex_state = 65, .external_lex_state = 6},
  [6725] = {.lex_state = 65, .external_lex_state = 23},
  [6726] = {.lex_state = 65, .external_lex_state = 6},
  [6727] = {.lex_state = 65, .external_lex_state = 6},
  [6728] = {.lex_state = 67, .external_lex_state = 6},
  [6729] = {.lex_state = 65, .external_lex_state = 24},
  [6730] = {.lex_state = 52, .external_lex_state = 6},
  [6731] = {.lex_state = 65, .external_lex_state = 6},
  [6732] = {.lex_state = 65, .external_lex_state = 6},
  [6733] = {.lex_state = 65, .external_lex_state = 6},
  [6734] = {.lex_state = 65, .external_lex_state = 6},
  [6735] = {.lex_state = 65, .external_lex_state = 6},
  [6736] = {.lex_state = 65, .external_lex_state = 25},
  [6737] = {.lex_state = 65, .external_lex_state = 6},
  [6738] = {.lex_state = 65, .external_lex_state = 6},
  [6739] = {.lex_state = 65, .external_lex_state = 26},
  [6740] = {.lex_state = 65, .external_lex_state = 6},
  [6741] = {.lex_state = 65, .external_lex_state = 6},
  [6742] = {.lex_state = 65, .external_lex_state = 6},
  [6743] = {.lex_state = 65, .external_lex_state = 6},
  [6744] = {.lex_state = 65, .external_lex_state = 6},
  [6745] = {.lex_state = 65, .external_lex_state = 6},
  [6746] = {.lex_state = 65, .external_lex_state = 6},
  [6747] = {.lex_state = 65, .external_lex_state = 6},
  [6748] = {.lex_state = 67, .external_lex_state = 6},
  [6749] = {.lex_state = 67, .external_lex_state = 6},
  [6750] = {.lex_state = 65, .external_lex_state = 6},
  [6751] = {.lex_state = 65, .external_lex_state = 6},
  [6752] = {.lex_state = 65, .external_lex_state = 6},
  [6753] = {.lex_state = 65, .external_lex_state = 6},
  [6754] = {.lex_state = 65, .external_lex_state = 6},
  [6755] = {.lex_state = 52, .external_lex_state = 6},
  [6756] = {.lex_state = 52, .external_lex_state = 6},
  [6757] = {.lex_state = 65, .external_lex_state = 6},
  [6758] = {.lex_state = 67, .external_lex_state = 6},
  [6759] = {.lex_state = 67, .external_lex_state = 6},
  [6760] = {.lex_state = 65, .external_lex_state = 6},
  [6761] = {.lex_state = 64, .external_lex_state = 6},
  [6762] = {.lex_state = 65, .external_lex_state = 6},
  [6763] = {.lex_state = 65, .external_lex_state = 6},
  [6764] = {.lex_state = 52, .external_lex_state = 6},
  [6765] = {.lex_state = 65, .external_lex_state = 6},
  [6766] = {.lex_state = 65, .external_lex_state = 6},
  [6767] = {.lex_state = 64, .external_lex_state = 6},
  [6768] = {.lex_state = 65, .external_lex_state = 6},
  [6769] = {.lex_state = 65, .external_lex_state = 6},
  [6770] = {.lex_state = 67, .external_lex_state = 6},
  [6771] = {.lex_state = 106, .external_lex_state = 6},
  [6772] = {.lex_state = 65, .external_lex_state = 6},
  [6773] = {.lex_state = 106, .external_lex_state = 6},
  [6774] = {.lex_state = 106, .external_lex_state = 6},
  [6775] = {.lex_state = 67, .external_lex_state = 6},
  [6776] = {.lex_state = 106, .external_lex_state = 6},
  [6777] = {.lex_state = 267, .external_lex_state = 6},
  [6778] = {.lex_state = 65, .external_lex_state = 6},
  [6779] = {.lex_state = 267, .external_lex_state = 6},
  [6780] = {.lex_state = 106, .external_lex_state = 6},
  [6781] = {.lex_state = 67, .external_lex_state = 6},
  [6782] = {.lex_state = 65, .external_lex_state = 6},
  [6783] = {.lex_state = 106, .external_lex_state = 6},
  [6784] = {.lex_state = 106, .external_lex_state = 6},
  [6785] = {.lex_state = 106, .external_lex_state = 6},
  [6786] = {.lex_state = 267, .external_lex_state = 6},
  [6787] = {.lex_state = 267, .external_lex_state = 6},
  [6788] = {.lex_state = 106, .external_lex_state = 6},
  [6789] = {.lex_state = 106, .external_lex_state = 6},
  [6790] = {.lex_state = 106, .external_lex_state = 6},
  [6791] = {.lex_state = 267, .external_lex_state = 6},
  [6792] = {.lex_state = 267, .external_lex_state = 6},
  [6793] = {.lex_state = 106, .external_lex_state = 6},
  [6794] = {.lex_state = 106, .external_lex_state = 6},
  [6795] = {.lex_state = 106, .external_lex_state = 6},
  [6796] = {.lex_state = 106, .external_lex_state = 6},
  [6797] = {.lex_state = 106, .external_lex_state = 6},
  [6798] = {.lex_state = 106, .external_lex_state = 6},
  [6799] = {.lex_state = 65, .external_lex_state = 6},
  [6800] = {.lex_state = 106, .external_lex_state = 6},
  [6801] = {.lex_state = 106, .external_lex_state = 6},
  [6802] = {.lex_state = 55, .external_lex_state = 6},
  [6803] = {.lex_state = 106, .external_lex_state = 6},
  [6804] = {.lex_state = 106, .external_lex_state = 6},
  [6805] = {.lex_state = 106, .external_lex_state = 6},
  [6806] = {.lex_state = 106, .external_lex_state = 6},
  [6807] = {.lex_state = 65, .external_lex_state = 6},
  [6808] = {.lex_state = 106, .external_lex_state = 6},
  [6809] = {.lex_state = 106, .external_lex_state = 6},
  [6810] = {.lex_state = 106, .external_lex_state = 6},
  [6811] = {.lex_state = 106, .external_lex_state = 6},
  [6812] = {.lex_state = 55, .external_lex_state = 6},
  [6813] = {.lex_state = 106, .external_lex_state = 6},
  [6814] = {.lex_state = 267, .external_lex_state = 6},
  [6815] = {.lex_state = 106, .external_lex_state = 6},
  [6816] = {.lex_state = 267, .external_lex_state = 6},
  [6817] = {.lex_state = 267, .external_lex_state = 6},
  [6818] = {.lex_state = 106, .external_lex_state = 6},
  [6819] = {.lex_state = 106, .external_lex_state = 6},
  [6820] = {.lex_state = 65, .external_lex_state = 6},
  [6821] = {.lex_state = 106, .external_lex_state = 6},
  [6822] = {.lex_state = 106, .external_lex_state = 6},
  [6823] = {.lex_state = 106, .external_lex_state = 6},
  [6824] = {.lex_state = 106, .external_lex_state = 6},
  [6825] = {.lex_state = 106, .external_lex_state = 6},
  [6826] = {.lex_state = 106, .external_lex_state = 6},
  [6827] = {.lex_state = 106, .external_lex_state = 6},
  [6828] = {.lex_state = 65, .external_lex_state = 6},
  [6829] = {.lex_state = 106, .external_lex_state = 6},
  [6830] = {.lex_state = 106, .external_lex_state = 6},
  [6831] = {.lex_state = 106, .external_lex_state = 6},
  [6832] = {.lex_state = 55, .external_lex_state = 6},
  [6833] = {.lex_state = 106, .external_lex_state = 6},
  [6834] = {.lex_state = 106, .external_lex_state = 6},
  [6835] = {.lex_state = 106, .external_lex_state = 6},
  [6836] = {.lex_state = 65, .external_lex_state = 6},
  [6837] = {.lex_state = 55, .external_lex_state = 6},
  [6838] = {.lex_state = 106, .external_lex_state = 6},
  [6839] = {.lex_state = 55, .external_lex_state = 6},
  [6840] = {.lex_state = 106, .external_lex_state = 6},
  [6841] = {.lex_state = 106, .external_lex_state = 6},
  [6842] = {.lex_state = 55, .external_lex_state = 6},
  [6843] = {.lex_state = 55, .external_lex_state = 6},
  [6844] = {.lex_state = 106, .external_lex_state = 6},
  [6845] = {.lex_state = 106, .external_lex_state = 6},
  [6846] = {.lex_state = 55, .external_lex_state = 6},
  [6847] = {.lex_state = 55, .external_lex_state = 6},
  [6848] = {.lex_state = 55, .external_lex_state = 6},
  [6849] = {.lex_state = 106, .external_lex_state = 6},
  [6850] = {.lex_state = 106, .external_lex_state = 6},
  [6851] = {.lex_state = 55, .external_lex_state = 6},
  [6852] = {.lex_state = 55, .external_lex_state = 6},
  [6853] = {.lex_state = 267, .external_lex_state = 6},
  [6854] = {.lex_state = 55, .external_lex_state = 6},
  [6855] = {.lex_state = 55, .external_lex_state = 6},
  [6856] = {.lex_state = 55, .external_lex_state = 6},
  [6857] = {.lex_state = 55, .external_lex_state = 6},
  [6858] = {.lex_state = 55, .external_lex_state = 6},
  [6859] = {.lex_state = 55, .external_lex_state = 6},
  [6860] = {.lex_state = 106, .external_lex_state = 6},
  [6861] = {.lex_state = 106, .external_lex_state = 6},
  [6862] = {.lex_state = 106, .external_lex_state = 6},
  [6863] = {.lex_state = 106, .external_lex_state = 6},
  [6864] = {.lex_state = 106, .external_lex_state = 6},
  [6865] = {.lex_state = 267, .external_lex_state = 6},
  [6866] = {.lex_state = 267, .external_lex_state = 6},
  [6867] = {.lex_state = 106, .external_lex_state = 6},
  [6868] = {.lex_state = 106, .external_lex_state = 6},
  [6869] = {.lex_state = 267, .external_lex_state = 6},
  [6870] = {.lex_state = 52, .external_lex_state = 6},
  [6871] = {.lex_state = 267, .external_lex_state = 6},
  [6872] = {.lex_state = 267, .external_lex_state = 6},
  [6873] = {.lex_state = 267, .external_lex_state = 6},
  [6874] = {.lex_state = 267, .external_lex_state = 6},
  [6875] = {.lex_state = 267, .external_lex_state = 6},
  [6876] = {.lex_state = 267, .external_lex_state = 6},
  [6877] = {.lex_state = 267, .external_lex_state = 6},
  [6878] = {.lex_state = 267, .external_lex_state = 6},
  [6879] = {.lex_state = 267, .external_lex_state = 6},
  [6880] = {.lex_state = 267, .external_lex_state = 6},
  [6881] = {.lex_state = 267, .external_lex_state = 6},
  [6882] = {.lex_state = 104, .external_lex_state = 6},
  [6883] = {.lex_state = 267, .external_lex_state = 6},
  [6884] = {.lex_state = 267, .external_lex_state = 6},
  [6885] = {.lex_state = 267, .external_lex_state = 6},
  [6886] = {.lex_state = 267, .external_lex_state = 6},
  [6887] = {.lex_state = 267, .external_lex_state = 6},
  [6888] = {.lex_state = 267, .external_lex_state = 6},
  [6889] = {.lex_state = 267, .external_lex_state = 6},
  [6890] = {.lex_state = 267, .external_lex_state = 6},
  [6891] = {.lex_state = 267, .external_lex_state = 6},
  [6892] = {.lex_state = 104, .external_lex_state = 6},
  [6893] = {.lex_state = 267, .external_lex_state = 6},
  [6894] = {.lex_state = 52, .external_lex_state = 6},
  [6895] = {.lex_state = 267, .external_lex_state = 6},
  [6896] = {.lex_state = 267, .external_lex_state = 6},
  [6897] = {.lex_state = 267, .external_lex_state = 6},
  [6898] = {.lex_state = 267, .external_lex_state = 6},
  [6899] = {.lex_state = 267, .external_lex_state = 6},
  [6900] = {.lex_state = 267, .external_lex_state = 6},
  [6901] = {.lex_state = 267, .external_lex_state = 6},
  [6902] = {.lex_state = 52, .external_lex_state = 6},
  [6903] = {.lex_state = 267, .external_lex_state = 6},
  [6904] = {.lex_state = 267, .external_lex_state = 6},
  [6905] = {.lex_state = 52, .external_lex_state = 6},
  [6906] = {.lex_state = 267, .external_lex_state = 6},
  [6907] = {.lex_state = 267, .external_lex_state = 6},
  [6908] = {.lex_state = 267, .external_lex_state = 6},
  [6909] = {.lex_state = 267, .external_lex_state = 6},
  [6910] = {.lex_state = 267, .external_lex_state = 6},
  [6911] = {.lex_state = 104, .external_lex_state = 6},
  [6912] = {.lex_state = 267, .external_lex_state = 6},
  [6913] = {.lex_state = 267, .external_lex_state = 6},
  [6914] = {.lex_state = 267, .external_lex_state = 6},
  [6915] = {.lex_state = 104, .external_lex_state = 6},
  [6916] = {.lex_state = 104, .external_lex_state = 6},
  [6917] = {.lex_state = 267, .external_lex_state = 6},
  [6918] = {.lex_state = 267, .external_lex_state = 6},
  [6919] = {.lex_state = 267, .external_lex_state = 6},
  [6920] = {.lex_state = 267, .external_lex_state = 6},
  [6921] = {.lex_state = 267, .external_lex_state = 6},
  [6922] = {.lex_state = 267, .external_lex_state = 6},
  [6923] = {.lex_state = 267, .external_lex_state = 6},
  [6924] = {.lex_state = 267, .external_lex_state = 6},
  [6925] = {.lex_state = 104, .external_lex_state = 6},
  [6926] = {.lex_state = 267, .external_lex_state = 6},
  [6927] = {.lex_state = 267, .external_lex_state = 6},
  [6928] = {.lex_state = 267, .external_lex_state = 6},
  [6929] = {.lex_state = 267, .external_lex_state = 6},
  [6930] = {.lex_state = 267, .external_lex_state = 6},
  [6931] = {.lex_state = 104, .external_lex_state = 6},
  [6932] = {.lex_state = 267, .external_lex_state = 6},
  [6933] = {.lex_state = 52, .external_lex_state = 6},
  [6934] = {.lex_state = 267, .external_lex_state = 6},
  [6935] = {.lex_state = 267, .external_lex_state = 6},
  [6936] = {.lex_state = 267, .external_lex_state = 6},
  [6937] = {.lex_state = 267, .external_lex_state = 6},
  [6938] = {.lex_state = 267, .external_lex_state = 6},
  [6939] = {.lex_state = 267, .external_lex_state = 6},
  [6940] = {.lex_state = 267, .external_lex_state = 6},
  [6941] = {.lex_state = 267, .external_lex_state = 6},
  [6942] = {.lex_state = 52, .external_lex_state = 6},
  [6943] = {.lex_state = 267, .external_lex_state = 6},
  [6944] = {.lex_state = 267, .external_lex_state = 6},
  [6945] = {.lex_state = 267, .external_lex_state = 6},
  [6946] = {.lex_state = 267, .external_lex_state = 6},
  [6947] = {.lex_state = 52, .external_lex_state = 6},
  [6948] = {.lex_state = 104, .external_lex_state = 6},
  [6949] = {.lex_state = 267, .external_lex_state = 6},
  [6950] = {.lex_state = 267, .external_lex_state = 6},
  [6951] = {.lex_state = 267, .external_lex_state = 6},
  [6952] = {.lex_state = 267, .external_lex_state = 6},
  [6953] = {.lex_state = 267, .external_lex_state = 6},
  [6954] = {.lex_state = 267, .external_lex_state = 6},
  [6955] = {.lex_state = 267, .external_lex_state = 6},
  [6956] = {.lex_state = 267, .external_lex_state = 6},
  [6957] = {.lex_state = 267, .external_lex_state = 6},
  [6958] = {.lex_state = 267, .external_lex_state = 6},
  [6959] = {.lex_state = 267, .external_lex_state = 6},
  [6960] = {.lex_state = 267, .external_lex_state = 6},
  [6961] = {.lex_state = 267, .external_lex_state = 6},
  [6962] = {.lex_state = 52, .external_lex_state = 6},
  [6963] = {.lex_state = 52, .external_lex_state = 6},
  [6964] = {.lex_state = 267, .external_lex_state = 6},
  [6965] = {.lex_state = 267, .external_lex_state = 6},
  [6966] = {.lex_state = 267, .external_lex_state = 6},
  [6967] = {.lex_state = 267, .external_lex_state = 6},
  [6968] = {.lex_state = 267, .external_lex_state = 6},
  [6969] = {.lex_state = 267, .external_lex_state = 6},
  [6970] = {.lex_state = 267, .external_lex_state = 6},
  [6971] = {.lex_state = 267, .external_lex_state = 6},
  [6972] = {.lex_state = 104, .external_lex_state = 6},
  [6973] = {.lex_state = 267, .external_lex_state = 6},
  [6974] = {.lex_state = 267, .external_lex_state = 6},
  [6975] = {.lex_state = 267, .external_lex_state = 6},
  [6976] = {.lex_state = 267, .external_lex_state = 6},
  [6977] = {.lex_state = 267, .external_lex_state = 6},
  [6978] = {.lex_state = 52, .external_lex_state = 6},
  [6979] = {.lex_state = 267, .external_lex_state = 6},
  [6980] = {.lex_state = 104, .external_lex_state = 6},
  [6981] = {.lex_state = 267, .external_lex_state = 6},
  [6982] = {.lex_state = 267, .external_lex_state = 6},
  [6983] = {.lex_state = 267, .external_lex_state = 6},
  [6984] = {.lex_state = 267, .external_lex_state = 6},
  [6985] = {.lex_state = 267, .external_lex_state = 6},
  [6986] = {.lex_state = 104, .external_lex_state = 6},
  [6987] = {.lex_state = 267, .external_lex_state = 6},
  [6988] = {.lex_state = 267, .external_lex_state = 6},
  [6989] = {.lex_state = 267, .external_lex_state = 6},
  [6990] = {.lex_state = 267, .external_lex_state = 6},
  [6991] = {.lex_state = 267, .external_lex_state = 6},
  [6992] = {.lex_state = 104, .external_lex_state = 6},
  [6993] = {.lex_state = 267, .external_lex_state = 6},
  [6994] = {.lex_state = 267, .external_lex_state = 6},
  [6995] = {.lex_state = 52, .external_lex_state = 6},
  [6996] = {.lex_state = 267, .external_lex_state = 6},
  [6997] = {.lex_state = 267, .external_lex_state = 6},
  [6998] = {.lex_state = 267, .external_lex_state = 6},
  [6999] = {.lex_state = 52, .external_lex_state = 6},
  [7000] = {.lex_state = 267, .external_lex_state = 6},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [aux_sym_identifier_token1] = ACTIONS(1),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1),
    [sym_integer] = ACTIONS(1),
    [sym_float] = ACTIONS(1),
    [sym_char] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_nil] = ACTIONS(1),
    [sym_atom] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_POUND_LBRACE] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [aux_sym_sigil_token1] = ACTIONS(1),
    [aux_sym_sigil_token2] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [sym_keyword] = ACTIONS(1),
    [aux_sym_quoted_keyword_token1] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1),
    [anon_sym_not] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_LT_DASH] = ACTIONS(1),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(1),
    [anon_sym_when] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym_EQ_GT] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_or] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(1),
    [anon_sym_and] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_EQ_TILDE] = ACTIONS(1),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_PIPE_GT] = ACTIONS(1),
    [anon_sym_LT_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT_GT] = ACTIONS(1),
    [anon_sym_LT_LT_TILDE] = ACTIONS(1),
    [anon_sym_TILDE_GT_GT] = ACTIONS(1),
    [anon_sym_LT_TILDE] = ACTIONS(1),
    [anon_sym_TILDE_GT] = ACTIONS(1),
    [anon_sym_LT_TILDE_GT] = ACTIONS(1),
    [anon_sym_LT_PIPE_GT] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(1),
    [anon_sym_SLASH_SLASH] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(1),
    [anon_sym_LT_GT] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_STAR_STAR] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_after] = ACTIONS(1),
    [anon_sym_catch] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_end] = ACTIONS(1),
    [anon_sym_fn] = ACTIONS(1),
    [anon_sym_rescue] = ACTIONS(1),
    [anon_sym_LPAREN2] = ACTIONS(1),
    [anon_sym_LBRACK2] = ACTIONS(1),
    [sym_comment] = ACTIONS(5),
    [sym__quoted_content_i_single] = ACTIONS(1),
    [sym__quoted_content_i_double] = ACTIONS(1),
    [sym__quoted_content_i_heredoc_single] = ACTIONS(1),
    [sym__quoted_content_i_heredoc_double] = ACTIONS(1),
    [sym__quoted_content_i_parenthesis] = ACTIONS(1),
    [sym__quoted_content_i_curly] = ACTIONS(1),
    [sym__quoted_content_i_square] = ACTIONS(1),
    [sym__quoted_content_i_angle] = ACTIONS(1),
    [sym__quoted_content_i_bar] = ACTIONS(1),
    [sym__quoted_content_i_slash] = ACTIONS(1),
    [sym__quoted_content_single] = ACTIONS(1),
    [sym__quoted_content_double] = ACTIONS(1),
    [sym__quoted_content_heredoc_single] = ACTIONS(1),
    [sym__quoted_content_heredoc_double] = ACTIONS(1),
    [sym__quoted_content_parenthesis] = ACTIONS(1),
    [sym__quoted_content_curly] = ACTIONS(1),
    [sym__quoted_content_square] = ACTIONS(1),
    [sym__quoted_content_angle] = ACTIONS(1),
    [sym__quoted_content_bar] = ACTIONS(1),
    [sym__quoted_content_slash] = ACTIONS(1),
    [sym__newline_before_do] = ACTIONS(1),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1),
    [sym__not_in] = ACTIONS(1),
    [sym__quoted_atom_start] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_source] = STATE(6951),
    [sym__terminator] = STATE(362),
    [sym__expression] = STATE(3215),
    [sym_block] = STATE(3215),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(3215),
    [sym_nil] = STATE(3215),
    [sym__atom] = STATE(3215),
    [sym_quoted_atom] = STATE(3215),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(3215),
    [sym_charlist] = STATE(3215),
    [sym_sigil] = STATE(3215),
    [sym_list] = STATE(3215),
    [sym_tuple] = STATE(3215),
    [sym_bitstring] = STATE(3215),
    [sym_map] = STATE(3215),
    [sym__nullary_operator] = STATE(3215),
    [sym_unary_operator] = STATE(3215),
    [sym_binary_operator] = STATE(3215),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(3215),
    [sym_call] = STATE(3215),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(3215),
    [sym_anonymous_function] = STATE(3215),
    [aux_sym__terminator_repeat1] = STATE(1035),
    [ts_builtin_sym_end] = ACTIONS(7),
    [aux_sym__terminator_token1] = ACTIONS(9),
    [anon_sym_SEMI] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(17),
    [sym_integer] = ACTIONS(17),
    [sym_float] = ACTIONS(17),
    [sym_char] = ACTIONS(17),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(17),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(2)] = {
    [sym__terminator] = STATE(25),
    [sym__expression] = STATE(1080),
    [sym_block] = STATE(1080),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1080),
    [sym_nil] = STATE(1080),
    [sym__atom] = STATE(1080),
    [sym_quoted_atom] = STATE(1080),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1080),
    [sym_charlist] = STATE(1080),
    [sym_sigil] = STATE(1080),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1080),
    [sym_tuple] = STATE(1080),
    [sym_bitstring] = STATE(1080),
    [sym_map] = STATE(1080),
    [sym__nullary_operator] = STATE(1080),
    [sym_unary_operator] = STATE(1080),
    [sym_binary_operator] = STATE(1080),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1080),
    [sym_call] = STATE(1080),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4760),
    [sym_rescue_block] = STATE(4760),
    [sym_catch_block] = STATE(4760),
    [sym_else_block] = STATE(4760),
    [sym_access_call] = STATE(1080),
    [sym_stab_clause] = STATE(4699),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1080),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4760),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(61),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(67),
    [sym_integer] = ACTIONS(67),
    [sym_float] = ACTIONS(67),
    [sym_char] = ACTIONS(67),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(67),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(109),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(3)] = {
    [sym__terminator] = STATE(24),
    [sym__expression] = STATE(1069),
    [sym_block] = STATE(1069),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1069),
    [sym_nil] = STATE(1069),
    [sym__atom] = STATE(1069),
    [sym_quoted_atom] = STATE(1069),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1069),
    [sym_charlist] = STATE(1069),
    [sym_sigil] = STATE(1069),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1069),
    [sym_tuple] = STATE(1069),
    [sym_bitstring] = STATE(1069),
    [sym_map] = STATE(1069),
    [sym__nullary_operator] = STATE(1069),
    [sym_unary_operator] = STATE(1069),
    [sym_binary_operator] = STATE(1069),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1069),
    [sym_call] = STATE(1069),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4728),
    [sym_rescue_block] = STATE(4728),
    [sym_catch_block] = STATE(4728),
    [sym_else_block] = STATE(4728),
    [sym_access_call] = STATE(1069),
    [sym_stab_clause] = STATE(4676),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1069),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4728),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(121),
    [sym_integer] = ACTIONS(121),
    [sym_float] = ACTIONS(121),
    [sym_char] = ACTIONS(121),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(121),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(123),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(4)] = {
    [sym__terminator] = STATE(22),
    [sym__expression] = STATE(1065),
    [sym_block] = STATE(1065),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1065),
    [sym_nil] = STATE(1065),
    [sym__atom] = STATE(1065),
    [sym_quoted_atom] = STATE(1065),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1065),
    [sym_charlist] = STATE(1065),
    [sym_sigil] = STATE(1065),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1065),
    [sym_tuple] = STATE(1065),
    [sym_bitstring] = STATE(1065),
    [sym_map] = STATE(1065),
    [sym__nullary_operator] = STATE(1065),
    [sym_unary_operator] = STATE(1065),
    [sym_binary_operator] = STATE(1065),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1065),
    [sym_call] = STATE(1065),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4733),
    [sym_rescue_block] = STATE(4733),
    [sym_catch_block] = STATE(4733),
    [sym_else_block] = STATE(4733),
    [sym_access_call] = STATE(1065),
    [sym_stab_clause] = STATE(4698),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1065),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4733),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(127),
    [sym_integer] = ACTIONS(127),
    [sym_float] = ACTIONS(127),
    [sym_char] = ACTIONS(127),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(127),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(129),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(5)] = {
    [sym__terminator] = STATE(20),
    [sym__expression] = STATE(1081),
    [sym_block] = STATE(1081),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1081),
    [sym_nil] = STATE(1081),
    [sym__atom] = STATE(1081),
    [sym_quoted_atom] = STATE(1081),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1081),
    [sym_charlist] = STATE(1081),
    [sym_sigil] = STATE(1081),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1081),
    [sym_tuple] = STATE(1081),
    [sym_bitstring] = STATE(1081),
    [sym_map] = STATE(1081),
    [sym__nullary_operator] = STATE(1081),
    [sym_unary_operator] = STATE(1081),
    [sym_binary_operator] = STATE(1081),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1081),
    [sym_call] = STATE(1081),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4747),
    [sym_rescue_block] = STATE(4747),
    [sym_catch_block] = STATE(4747),
    [sym_else_block] = STATE(4747),
    [sym_access_call] = STATE(1081),
    [sym_stab_clause] = STATE(4689),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1081),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4747),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(131),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(133),
    [sym_integer] = ACTIONS(133),
    [sym_float] = ACTIONS(133),
    [sym_char] = ACTIONS(133),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(133),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(135),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(6)] = {
    [sym__terminator] = STATE(29),
    [sym__expression] = STATE(1071),
    [sym_block] = STATE(1071),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1071),
    [sym_nil] = STATE(1071),
    [sym__atom] = STATE(1071),
    [sym_quoted_atom] = STATE(1071),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1071),
    [sym_charlist] = STATE(1071),
    [sym_sigil] = STATE(1071),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1071),
    [sym_tuple] = STATE(1071),
    [sym_bitstring] = STATE(1071),
    [sym_map] = STATE(1071),
    [sym__nullary_operator] = STATE(1071),
    [sym_unary_operator] = STATE(1071),
    [sym_binary_operator] = STATE(1071),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1071),
    [sym_call] = STATE(1071),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4717),
    [sym_rescue_block] = STATE(4717),
    [sym_catch_block] = STATE(4717),
    [sym_else_block] = STATE(4717),
    [sym_access_call] = STATE(1071),
    [sym_stab_clause] = STATE(4664),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1071),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4717),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(137),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(139),
    [sym_integer] = ACTIONS(139),
    [sym_float] = ACTIONS(139),
    [sym_char] = ACTIONS(139),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(139),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(141),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(7)] = {
    [sym__terminator] = STATE(31),
    [sym__expression] = STATE(1072),
    [sym_block] = STATE(1072),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1072),
    [sym_nil] = STATE(1072),
    [sym__atom] = STATE(1072),
    [sym_quoted_atom] = STATE(1072),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1072),
    [sym_charlist] = STATE(1072),
    [sym_sigil] = STATE(1072),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1072),
    [sym_tuple] = STATE(1072),
    [sym_bitstring] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym__nullary_operator] = STATE(1072),
    [sym_unary_operator] = STATE(1072),
    [sym_binary_operator] = STATE(1072),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1072),
    [sym_call] = STATE(1072),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4716),
    [sym_rescue_block] = STATE(4716),
    [sym_catch_block] = STATE(4716),
    [sym_else_block] = STATE(4716),
    [sym_access_call] = STATE(1072),
    [sym_stab_clause] = STATE(4665),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1072),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4716),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(143),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(145),
    [sym_integer] = ACTIONS(145),
    [sym_float] = ACTIONS(145),
    [sym_char] = ACTIONS(145),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(145),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(147),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(8)] = {
    [sym__terminator] = STATE(27),
    [sym__expression] = STATE(1075),
    [sym_block] = STATE(1075),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1075),
    [sym_nil] = STATE(1075),
    [sym__atom] = STATE(1075),
    [sym_quoted_atom] = STATE(1075),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1075),
    [sym_charlist] = STATE(1075),
    [sym_sigil] = STATE(1075),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1075),
    [sym_tuple] = STATE(1075),
    [sym_bitstring] = STATE(1075),
    [sym_map] = STATE(1075),
    [sym__nullary_operator] = STATE(1075),
    [sym_unary_operator] = STATE(1075),
    [sym_binary_operator] = STATE(1075),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1075),
    [sym_call] = STATE(1075),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4705),
    [sym_rescue_block] = STATE(4705),
    [sym_catch_block] = STATE(4705),
    [sym_else_block] = STATE(4705),
    [sym_access_call] = STATE(1075),
    [sym_stab_clause] = STATE(4632),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1075),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4705),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(149),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(151),
    [sym_integer] = ACTIONS(151),
    [sym_float] = ACTIONS(151),
    [sym_char] = ACTIONS(151),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(151),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(153),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(9)] = {
    [sym__terminator] = STATE(30),
    [sym__expression] = STATE(1085),
    [sym_block] = STATE(1085),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1085),
    [sym_nil] = STATE(1085),
    [sym__atom] = STATE(1085),
    [sym_quoted_atom] = STATE(1085),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1085),
    [sym_charlist] = STATE(1085),
    [sym_sigil] = STATE(1085),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1085),
    [sym_tuple] = STATE(1085),
    [sym_bitstring] = STATE(1085),
    [sym_map] = STATE(1085),
    [sym__nullary_operator] = STATE(1085),
    [sym_unary_operator] = STATE(1085),
    [sym_binary_operator] = STATE(1085),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1085),
    [sym_call] = STATE(1085),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4748),
    [sym_rescue_block] = STATE(4748),
    [sym_catch_block] = STATE(4748),
    [sym_else_block] = STATE(4748),
    [sym_access_call] = STATE(1085),
    [sym_stab_clause] = STATE(4650),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1085),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4748),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(155),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(157),
    [sym_integer] = ACTIONS(157),
    [sym_float] = ACTIONS(157),
    [sym_char] = ACTIONS(157),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(157),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(159),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(10)] = {
    [sym__terminator] = STATE(26),
    [sym__expression] = STATE(1076),
    [sym_block] = STATE(1076),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1076),
    [sym_nil] = STATE(1076),
    [sym__atom] = STATE(1076),
    [sym_quoted_atom] = STATE(1076),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1076),
    [sym_charlist] = STATE(1076),
    [sym_sigil] = STATE(1076),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1076),
    [sym_tuple] = STATE(1076),
    [sym_bitstring] = STATE(1076),
    [sym_map] = STATE(1076),
    [sym__nullary_operator] = STATE(1076),
    [sym_unary_operator] = STATE(1076),
    [sym_binary_operator] = STATE(1076),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1076),
    [sym_call] = STATE(1076),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4710),
    [sym_rescue_block] = STATE(4710),
    [sym_catch_block] = STATE(4710),
    [sym_else_block] = STATE(4710),
    [sym_access_call] = STATE(1076),
    [sym_stab_clause] = STATE(4642),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1076),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4710),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(161),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(163),
    [sym_integer] = ACTIONS(163),
    [sym_float] = ACTIONS(163),
    [sym_char] = ACTIONS(163),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(163),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(165),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(11)] = {
    [sym__terminator] = STATE(28),
    [sym__expression] = STATE(1070),
    [sym_block] = STATE(1070),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1070),
    [sym_nil] = STATE(1070),
    [sym__atom] = STATE(1070),
    [sym_quoted_atom] = STATE(1070),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1070),
    [sym_charlist] = STATE(1070),
    [sym_sigil] = STATE(1070),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1070),
    [sym_tuple] = STATE(1070),
    [sym_bitstring] = STATE(1070),
    [sym_map] = STATE(1070),
    [sym__nullary_operator] = STATE(1070),
    [sym_unary_operator] = STATE(1070),
    [sym_binary_operator] = STATE(1070),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1070),
    [sym_call] = STATE(1070),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4706),
    [sym_rescue_block] = STATE(4706),
    [sym_catch_block] = STATE(4706),
    [sym_else_block] = STATE(4706),
    [sym_access_call] = STATE(1070),
    [sym_stab_clause] = STATE(4656),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1070),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4706),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(167),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(169),
    [sym_integer] = ACTIONS(169),
    [sym_float] = ACTIONS(169),
    [sym_char] = ACTIONS(169),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(169),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(171),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(12)] = {
    [sym__terminator] = STATE(32),
    [sym__expression] = STATE(1073),
    [sym_block] = STATE(1073),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1073),
    [sym_nil] = STATE(1073),
    [sym__atom] = STATE(1073),
    [sym_quoted_atom] = STATE(1073),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1073),
    [sym_charlist] = STATE(1073),
    [sym_sigil] = STATE(1073),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1073),
    [sym_tuple] = STATE(1073),
    [sym_bitstring] = STATE(1073),
    [sym_map] = STATE(1073),
    [sym__nullary_operator] = STATE(1073),
    [sym_unary_operator] = STATE(1073),
    [sym_binary_operator] = STATE(1073),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1073),
    [sym_call] = STATE(1073),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4727),
    [sym_rescue_block] = STATE(4727),
    [sym_catch_block] = STATE(4727),
    [sym_else_block] = STATE(4727),
    [sym_access_call] = STATE(1073),
    [sym_stab_clause] = STATE(4675),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1073),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4727),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(175),
    [sym_integer] = ACTIONS(175),
    [sym_float] = ACTIONS(175),
    [sym_char] = ACTIONS(175),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(175),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(177),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(13)] = {
    [sym__terminator] = STATE(33),
    [sym__expression] = STATE(1078),
    [sym_block] = STATE(1078),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1078),
    [sym_nil] = STATE(1078),
    [sym__atom] = STATE(1078),
    [sym_quoted_atom] = STATE(1078),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1078),
    [sym_charlist] = STATE(1078),
    [sym_sigil] = STATE(1078),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1078),
    [sym_tuple] = STATE(1078),
    [sym_bitstring] = STATE(1078),
    [sym_map] = STATE(1078),
    [sym__nullary_operator] = STATE(1078),
    [sym_unary_operator] = STATE(1078),
    [sym_binary_operator] = STATE(1078),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1078),
    [sym_call] = STATE(1078),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4737),
    [sym_rescue_block] = STATE(4737),
    [sym_catch_block] = STATE(4737),
    [sym_else_block] = STATE(4737),
    [sym_access_call] = STATE(1078),
    [sym_stab_clause] = STATE(4683),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1078),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym_do_block_repeat1] = STATE(4737),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(181),
    [sym_integer] = ACTIONS(181),
    [sym_float] = ACTIONS(181),
    [sym_char] = ACTIONS(181),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(181),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(183),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(14)] = {
    [sym__expression] = STATE(1423),
    [sym_block] = STATE(1423),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1423),
    [sym_nil] = STATE(1423),
    [sym__atom] = STATE(1423),
    [sym_quoted_atom] = STATE(1423),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1423),
    [sym_charlist] = STATE(1423),
    [sym_sigil] = STATE(1423),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(1573),
    [sym__keyword] = STATE(895),
    [sym_quoted_keyword] = STATE(895),
    [sym_list] = STATE(1423),
    [sym_tuple] = STATE(1423),
    [sym_bitstring] = STATE(1423),
    [sym_map] = STATE(1423),
    [sym__nullary_operator] = STATE(1423),
    [sym_unary_operator] = STATE(1423),
    [sym_binary_operator] = STATE(1423),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1423),
    [sym_call] = STATE(1423),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1109),
    [sym__call_arguments_without_parentheses] = STATE(1326),
    [sym_do_block] = STATE(1500),
    [sym_access_call] = STATE(1423),
    [sym_anonymous_function] = STATE(1423),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(193),
    [sym_integer] = ACTIONS(193),
    [sym_float] = ACTIONS(193),
    [sym_char] = ACTIONS(193),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(193),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(213),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_after] = ACTIONS(187),
    [anon_sym_catch] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_else] = ACTIONS(187),
    [anon_sym_end] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_rescue] = ACTIONS(187),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(15)] = {
    [sym__expression] = STATE(1250),
    [sym_block] = STATE(1250),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1250),
    [sym_nil] = STATE(1250),
    [sym__atom] = STATE(1250),
    [sym_quoted_atom] = STATE(1250),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1250),
    [sym_charlist] = STATE(1250),
    [sym_sigil] = STATE(1250),
    [sym_keywords] = STATE(1239),
    [sym_pair] = STATE(1349),
    [sym__keyword] = STATE(891),
    [sym_quoted_keyword] = STATE(891),
    [sym_list] = STATE(1250),
    [sym_tuple] = STATE(1250),
    [sym_bitstring] = STATE(1250),
    [sym_map] = STATE(1250),
    [sym__nullary_operator] = STATE(1250),
    [sym_unary_operator] = STATE(1250),
    [sym_binary_operator] = STATE(1250),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1250),
    [sym_call] = STATE(1250),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1087),
    [sym__call_arguments_without_parentheses] = STATE(1135),
    [sym_do_block] = STATE(1306),
    [sym_access_call] = STATE(1250),
    [sym_anonymous_function] = STATE(1250),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(239),
    [sym_integer] = ACTIONS(239),
    [sym_float] = ACTIONS(239),
    [sym_char] = ACTIONS(239),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(239),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(262),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(268),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(273),
    [anon_sym_DASH] = ACTIONS(273),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(257),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_after] = ACTIONS(235),
    [anon_sym_catch] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_else] = ACTIONS(235),
    [anon_sym_end] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(280),
    [anon_sym_rescue] = ACTIONS(235),
    [anon_sym_LPAREN2] = ACTIONS(282),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(16)] = {
    [sym__expression] = STATE(1423),
    [sym_block] = STATE(1423),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1423),
    [sym_nil] = STATE(1423),
    [sym__atom] = STATE(1423),
    [sym_quoted_atom] = STATE(1423),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1423),
    [sym_charlist] = STATE(1423),
    [sym_sigil] = STATE(1423),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(1573),
    [sym__keyword] = STATE(895),
    [sym_quoted_keyword] = STATE(895),
    [sym_list] = STATE(1423),
    [sym_tuple] = STATE(1423),
    [sym_bitstring] = STATE(1423),
    [sym_map] = STATE(1423),
    [sym__nullary_operator] = STATE(1423),
    [sym_unary_operator] = STATE(1423),
    [sym_binary_operator] = STATE(1423),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1423),
    [sym_call] = STATE(1423),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1121),
    [sym__call_arguments_without_parentheses] = STATE(1228),
    [sym_do_block] = STATE(2014),
    [sym_access_call] = STATE(1423),
    [sym_anonymous_function] = STATE(1423),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(193),
    [sym_integer] = ACTIONS(193),
    [sym_float] = ACTIONS(193),
    [sym_char] = ACTIONS(193),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(193),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(213),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_after] = ACTIONS(187),
    [anon_sym_catch] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(291),
    [anon_sym_else] = ACTIONS(187),
    [anon_sym_end] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_rescue] = ACTIONS(187),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(293),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(17)] = {
    [sym__expression] = STATE(1250),
    [sym_block] = STATE(1250),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1250),
    [sym_nil] = STATE(1250),
    [sym__atom] = STATE(1250),
    [sym_quoted_atom] = STATE(1250),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1250),
    [sym_charlist] = STATE(1250),
    [sym_sigil] = STATE(1250),
    [sym_keywords] = STATE(1239),
    [sym_pair] = STATE(1349),
    [sym__keyword] = STATE(891),
    [sym_quoted_keyword] = STATE(891),
    [sym_list] = STATE(1250),
    [sym_tuple] = STATE(1250),
    [sym_bitstring] = STATE(1250),
    [sym_map] = STATE(1250),
    [sym__nullary_operator] = STATE(1250),
    [sym_unary_operator] = STATE(1250),
    [sym_binary_operator] = STATE(1250),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1250),
    [sym_call] = STATE(1250),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1089),
    [sym__call_arguments_without_parentheses] = STATE(1123),
    [sym_do_block] = STATE(1642),
    [sym_access_call] = STATE(1250),
    [sym_anonymous_function] = STATE(1250),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(239),
    [sym_integer] = ACTIONS(239),
    [sym_float] = ACTIONS(239),
    [sym_char] = ACTIONS(239),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(239),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(262),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_after] = ACTIONS(187),
    [anon_sym_catch] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(295),
    [anon_sym_else] = ACTIONS(187),
    [anon_sym_end] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(280),
    [anon_sym_rescue] = ACTIONS(187),
    [anon_sym_LPAREN2] = ACTIONS(282),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(297),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(18)] = {
    [sym__expression] = STATE(1250),
    [sym_block] = STATE(1250),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1250),
    [sym_nil] = STATE(1250),
    [sym__atom] = STATE(1250),
    [sym_quoted_atom] = STATE(1250),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1250),
    [sym_charlist] = STATE(1250),
    [sym_sigil] = STATE(1250),
    [sym_keywords] = STATE(1239),
    [sym_pair] = STATE(1349),
    [sym__keyword] = STATE(891),
    [sym_quoted_keyword] = STATE(891),
    [sym_list] = STATE(1250),
    [sym_tuple] = STATE(1250),
    [sym_bitstring] = STATE(1250),
    [sym_map] = STATE(1250),
    [sym__nullary_operator] = STATE(1250),
    [sym_unary_operator] = STATE(1250),
    [sym_binary_operator] = STATE(1250),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1250),
    [sym_call] = STATE(1250),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1092),
    [sym__call_arguments_without_parentheses] = STATE(1125),
    [sym_do_block] = STATE(1185),
    [sym_access_call] = STATE(1250),
    [sym_anonymous_function] = STATE(1250),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(239),
    [sym_integer] = ACTIONS(239),
    [sym_float] = ACTIONS(239),
    [sym_char] = ACTIONS(239),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(239),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(262),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_after] = ACTIONS(187),
    [anon_sym_catch] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_else] = ACTIONS(187),
    [anon_sym_end] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(280),
    [anon_sym_rescue] = ACTIONS(187),
    [anon_sym_LPAREN2] = ACTIONS(282),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(19)] = {
    [sym__expression] = STATE(1423),
    [sym_block] = STATE(1423),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1423),
    [sym_nil] = STATE(1423),
    [sym__atom] = STATE(1423),
    [sym_quoted_atom] = STATE(1423),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1423),
    [sym_charlist] = STATE(1423),
    [sym_sigil] = STATE(1423),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(1573),
    [sym__keyword] = STATE(895),
    [sym_quoted_keyword] = STATE(895),
    [sym_list] = STATE(1423),
    [sym_tuple] = STATE(1423),
    [sym_bitstring] = STATE(1423),
    [sym_map] = STATE(1423),
    [sym__nullary_operator] = STATE(1423),
    [sym_unary_operator] = STATE(1423),
    [sym_binary_operator] = STATE(1423),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1423),
    [sym_call] = STATE(1423),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1107),
    [sym__call_arguments_without_parentheses] = STATE(1320),
    [sym_do_block] = STATE(1493),
    [sym_access_call] = STATE(1423),
    [sym_anonymous_function] = STATE(1423),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(193),
    [sym_integer] = ACTIONS(193),
    [sym_float] = ACTIONS(193),
    [sym_char] = ACTIONS(193),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(193),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(213),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(299),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(302),
    [anon_sym_DASH] = ACTIONS(302),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_after] = ACTIONS(235),
    [anon_sym_catch] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_else] = ACTIONS(235),
    [anon_sym_end] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_rescue] = ACTIONS(235),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(20)] = {
    [sym__expression] = STATE(1082),
    [sym_block] = STATE(1082),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1082),
    [sym_nil] = STATE(1082),
    [sym__atom] = STATE(1082),
    [sym_quoted_atom] = STATE(1082),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1082),
    [sym_charlist] = STATE(1082),
    [sym_sigil] = STATE(1082),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1082),
    [sym_tuple] = STATE(1082),
    [sym_bitstring] = STATE(1082),
    [sym_map] = STATE(1082),
    [sym__nullary_operator] = STATE(1082),
    [sym_unary_operator] = STATE(1082),
    [sym_binary_operator] = STATE(1082),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1082),
    [sym_call] = STATE(1082),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4750),
    [sym_rescue_block] = STATE(4750),
    [sym_catch_block] = STATE(4750),
    [sym_else_block] = STATE(4750),
    [sym_access_call] = STATE(1082),
    [sym_stab_clause] = STATE(4691),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1082),
    [aux_sym_do_block_repeat1] = STATE(4750),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(305),
    [sym_integer] = ACTIONS(305),
    [sym_float] = ACTIONS(305),
    [sym_char] = ACTIONS(305),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(305),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(307),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(21)] = {
    [sym__expression] = STATE(1250),
    [sym_block] = STATE(1250),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1250),
    [sym_nil] = STATE(1250),
    [sym__atom] = STATE(1250),
    [sym_quoted_atom] = STATE(1250),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1250),
    [sym_charlist] = STATE(1250),
    [sym_sigil] = STATE(1250),
    [sym_keywords] = STATE(1239),
    [sym_pair] = STATE(1349),
    [sym__keyword] = STATE(891),
    [sym_quoted_keyword] = STATE(891),
    [sym_list] = STATE(1250),
    [sym_tuple] = STATE(1250),
    [sym_bitstring] = STATE(1250),
    [sym_map] = STATE(1250),
    [sym__nullary_operator] = STATE(1250),
    [sym_unary_operator] = STATE(1250),
    [sym_binary_operator] = STATE(1250),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1250),
    [sym_call] = STATE(1250),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1090),
    [sym__call_arguments_without_parentheses] = STATE(1097),
    [sym_do_block] = STATE(1641),
    [sym_access_call] = STATE(1250),
    [sym_anonymous_function] = STATE(1250),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(239),
    [sym_integer] = ACTIONS(239),
    [sym_float] = ACTIONS(239),
    [sym_char] = ACTIONS(239),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(239),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(262),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(268),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(273),
    [anon_sym_DASH] = ACTIONS(273),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(257),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_after] = ACTIONS(235),
    [anon_sym_catch] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(295),
    [anon_sym_else] = ACTIONS(235),
    [anon_sym_end] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(280),
    [anon_sym_rescue] = ACTIONS(235),
    [anon_sym_LPAREN2] = ACTIONS(282),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(22)] = {
    [sym__expression] = STATE(1063),
    [sym_block] = STATE(1063),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1063),
    [sym_nil] = STATE(1063),
    [sym__atom] = STATE(1063),
    [sym_quoted_atom] = STATE(1063),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1063),
    [sym_charlist] = STATE(1063),
    [sym_sigil] = STATE(1063),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1063),
    [sym_tuple] = STATE(1063),
    [sym_bitstring] = STATE(1063),
    [sym_map] = STATE(1063),
    [sym__nullary_operator] = STATE(1063),
    [sym_unary_operator] = STATE(1063),
    [sym_binary_operator] = STATE(1063),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1063),
    [sym_call] = STATE(1063),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4713),
    [sym_rescue_block] = STATE(4713),
    [sym_catch_block] = STATE(4713),
    [sym_else_block] = STATE(4713),
    [sym_access_call] = STATE(1063),
    [sym_stab_clause] = STATE(4631),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1063),
    [aux_sym_do_block_repeat1] = STATE(4713),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(309),
    [sym_integer] = ACTIONS(309),
    [sym_float] = ACTIONS(309),
    [sym_char] = ACTIONS(309),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(309),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(311),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(23)] = {
    [sym__expression] = STATE(1423),
    [sym_block] = STATE(1423),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1423),
    [sym_nil] = STATE(1423),
    [sym__atom] = STATE(1423),
    [sym_quoted_atom] = STATE(1423),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1423),
    [sym_charlist] = STATE(1423),
    [sym_sigil] = STATE(1423),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(1573),
    [sym__keyword] = STATE(895),
    [sym_quoted_keyword] = STATE(895),
    [sym_list] = STATE(1423),
    [sym_tuple] = STATE(1423),
    [sym_bitstring] = STATE(1423),
    [sym_map] = STATE(1423),
    [sym__nullary_operator] = STATE(1423),
    [sym_unary_operator] = STATE(1423),
    [sym_binary_operator] = STATE(1423),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1423),
    [sym_call] = STATE(1423),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1119),
    [sym__call_arguments_without_parentheses] = STATE(1219),
    [sym_do_block] = STATE(2060),
    [sym_access_call] = STATE(1423),
    [sym_anonymous_function] = STATE(1423),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(193),
    [sym_integer] = ACTIONS(193),
    [sym_float] = ACTIONS(193),
    [sym_char] = ACTIONS(193),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(193),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(213),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(299),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(302),
    [anon_sym_DASH] = ACTIONS(302),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_after] = ACTIONS(235),
    [anon_sym_catch] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(291),
    [anon_sym_else] = ACTIONS(235),
    [anon_sym_end] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_rescue] = ACTIONS(235),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(24)] = {
    [sym__expression] = STATE(1064),
    [sym_block] = STATE(1064),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1064),
    [sym_nil] = STATE(1064),
    [sym__atom] = STATE(1064),
    [sym_quoted_atom] = STATE(1064),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1064),
    [sym_charlist] = STATE(1064),
    [sym_sigil] = STATE(1064),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1064),
    [sym_tuple] = STATE(1064),
    [sym_bitstring] = STATE(1064),
    [sym_map] = STATE(1064),
    [sym__nullary_operator] = STATE(1064),
    [sym_unary_operator] = STATE(1064),
    [sym_binary_operator] = STATE(1064),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1064),
    [sym_call] = STATE(1064),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4749),
    [sym_rescue_block] = STATE(4749),
    [sym_catch_block] = STATE(4749),
    [sym_else_block] = STATE(4749),
    [sym_access_call] = STATE(1064),
    [sym_stab_clause] = STATE(4690),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1064),
    [aux_sym_do_block_repeat1] = STATE(4749),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(313),
    [sym_integer] = ACTIONS(313),
    [sym_float] = ACTIONS(313),
    [sym_char] = ACTIONS(313),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(313),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(315),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(25)] = {
    [sym__expression] = STATE(1084),
    [sym_block] = STATE(1084),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1084),
    [sym_nil] = STATE(1084),
    [sym__atom] = STATE(1084),
    [sym_quoted_atom] = STATE(1084),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1084),
    [sym_charlist] = STATE(1084),
    [sym_sigil] = STATE(1084),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1084),
    [sym_tuple] = STATE(1084),
    [sym_bitstring] = STATE(1084),
    [sym_map] = STATE(1084),
    [sym__nullary_operator] = STATE(1084),
    [sym_unary_operator] = STATE(1084),
    [sym_binary_operator] = STATE(1084),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1084),
    [sym_call] = STATE(1084),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4752),
    [sym_rescue_block] = STATE(4752),
    [sym_catch_block] = STATE(4752),
    [sym_else_block] = STATE(4752),
    [sym_access_call] = STATE(1084),
    [sym_stab_clause] = STATE(4639),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1084),
    [aux_sym_do_block_repeat1] = STATE(4752),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(317),
    [sym_integer] = ACTIONS(317),
    [sym_float] = ACTIONS(317),
    [sym_char] = ACTIONS(317),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(317),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(319),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(26)] = {
    [sym__expression] = STATE(1083),
    [sym_block] = STATE(1083),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1083),
    [sym_nil] = STATE(1083),
    [sym__atom] = STATE(1083),
    [sym_quoted_atom] = STATE(1083),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1083),
    [sym_charlist] = STATE(1083),
    [sym_sigil] = STATE(1083),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1083),
    [sym_tuple] = STATE(1083),
    [sym_bitstring] = STATE(1083),
    [sym_map] = STATE(1083),
    [sym__nullary_operator] = STATE(1083),
    [sym_unary_operator] = STATE(1083),
    [sym_binary_operator] = STATE(1083),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1083),
    [sym_call] = STATE(1083),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4715),
    [sym_rescue_block] = STATE(4715),
    [sym_catch_block] = STATE(4715),
    [sym_else_block] = STATE(4715),
    [sym_access_call] = STATE(1083),
    [sym_stab_clause] = STATE(4644),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1083),
    [aux_sym_do_block_repeat1] = STATE(4715),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(321),
    [sym_integer] = ACTIONS(321),
    [sym_float] = ACTIONS(321),
    [sym_char] = ACTIONS(321),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(321),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(323),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(27)] = {
    [sym__expression] = STATE(1077),
    [sym_block] = STATE(1077),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1077),
    [sym_nil] = STATE(1077),
    [sym__atom] = STATE(1077),
    [sym_quoted_atom] = STATE(1077),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1077),
    [sym_charlist] = STATE(1077),
    [sym_sigil] = STATE(1077),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1077),
    [sym_tuple] = STATE(1077),
    [sym_bitstring] = STATE(1077),
    [sym_map] = STATE(1077),
    [sym__nullary_operator] = STATE(1077),
    [sym_unary_operator] = STATE(1077),
    [sym_binary_operator] = STATE(1077),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1077),
    [sym_call] = STATE(1077),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4731),
    [sym_rescue_block] = STATE(4731),
    [sym_catch_block] = STATE(4731),
    [sym_else_block] = STATE(4731),
    [sym_access_call] = STATE(1077),
    [sym_stab_clause] = STATE(4633),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1077),
    [aux_sym_do_block_repeat1] = STATE(4731),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(325),
    [sym_integer] = ACTIONS(325),
    [sym_float] = ACTIONS(325),
    [sym_char] = ACTIONS(325),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(325),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(327),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(28)] = {
    [sym__expression] = STATE(1067),
    [sym_block] = STATE(1067),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1067),
    [sym_nil] = STATE(1067),
    [sym__atom] = STATE(1067),
    [sym_quoted_atom] = STATE(1067),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1067),
    [sym_charlist] = STATE(1067),
    [sym_sigil] = STATE(1067),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1067),
    [sym_tuple] = STATE(1067),
    [sym_bitstring] = STATE(1067),
    [sym_map] = STATE(1067),
    [sym__nullary_operator] = STATE(1067),
    [sym_unary_operator] = STATE(1067),
    [sym_binary_operator] = STATE(1067),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1067),
    [sym_call] = STATE(1067),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4707),
    [sym_rescue_block] = STATE(4707),
    [sym_catch_block] = STATE(4707),
    [sym_else_block] = STATE(4707),
    [sym_access_call] = STATE(1067),
    [sym_stab_clause] = STATE(4657),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1067),
    [aux_sym_do_block_repeat1] = STATE(4707),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(329),
    [sym_integer] = ACTIONS(329),
    [sym_float] = ACTIONS(329),
    [sym_char] = ACTIONS(329),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(329),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(331),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(29)] = {
    [sym__expression] = STATE(1066),
    [sym_block] = STATE(1066),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1066),
    [sym_nil] = STATE(1066),
    [sym__atom] = STATE(1066),
    [sym_quoted_atom] = STATE(1066),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1066),
    [sym_charlist] = STATE(1066),
    [sym_sigil] = STATE(1066),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1066),
    [sym_tuple] = STATE(1066),
    [sym_bitstring] = STATE(1066),
    [sym_map] = STATE(1066),
    [sym__nullary_operator] = STATE(1066),
    [sym_unary_operator] = STATE(1066),
    [sym_binary_operator] = STATE(1066),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1066),
    [sym_call] = STATE(1066),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4725),
    [sym_rescue_block] = STATE(4725),
    [sym_catch_block] = STATE(4725),
    [sym_else_block] = STATE(4725),
    [sym_access_call] = STATE(1066),
    [sym_stab_clause] = STATE(4671),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1066),
    [aux_sym_do_block_repeat1] = STATE(4725),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(333),
    [sym_integer] = ACTIONS(333),
    [sym_float] = ACTIONS(333),
    [sym_char] = ACTIONS(333),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(333),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(335),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(30)] = {
    [sym__expression] = STATE(1086),
    [sym_block] = STATE(1086),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1086),
    [sym_nil] = STATE(1086),
    [sym__atom] = STATE(1086),
    [sym_quoted_atom] = STATE(1086),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1086),
    [sym_charlist] = STATE(1086),
    [sym_sigil] = STATE(1086),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1086),
    [sym_tuple] = STATE(1086),
    [sym_bitstring] = STATE(1086),
    [sym_map] = STATE(1086),
    [sym__nullary_operator] = STATE(1086),
    [sym_unary_operator] = STATE(1086),
    [sym_binary_operator] = STATE(1086),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1086),
    [sym_call] = STATE(1086),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4751),
    [sym_rescue_block] = STATE(4751),
    [sym_catch_block] = STATE(4751),
    [sym_else_block] = STATE(4751),
    [sym_access_call] = STATE(1086),
    [sym_stab_clause] = STATE(4651),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1086),
    [aux_sym_do_block_repeat1] = STATE(4751),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(337),
    [sym_integer] = ACTIONS(337),
    [sym_float] = ACTIONS(337),
    [sym_char] = ACTIONS(337),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(337),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(339),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(31)] = {
    [sym__expression] = STATE(1068),
    [sym_block] = STATE(1068),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1068),
    [sym_nil] = STATE(1068),
    [sym__atom] = STATE(1068),
    [sym_quoted_atom] = STATE(1068),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1068),
    [sym_charlist] = STATE(1068),
    [sym_sigil] = STATE(1068),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1068),
    [sym_tuple] = STATE(1068),
    [sym_bitstring] = STATE(1068),
    [sym_map] = STATE(1068),
    [sym__nullary_operator] = STATE(1068),
    [sym_unary_operator] = STATE(1068),
    [sym_binary_operator] = STATE(1068),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1068),
    [sym_call] = STATE(1068),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4719),
    [sym_rescue_block] = STATE(4719),
    [sym_catch_block] = STATE(4719),
    [sym_else_block] = STATE(4719),
    [sym_access_call] = STATE(1068),
    [sym_stab_clause] = STATE(4666),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1068),
    [aux_sym_do_block_repeat1] = STATE(4719),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(341),
    [sym_integer] = ACTIONS(341),
    [sym_float] = ACTIONS(341),
    [sym_char] = ACTIONS(341),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(341),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(343),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(32)] = {
    [sym__expression] = STATE(1074),
    [sym_block] = STATE(1074),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1074),
    [sym_nil] = STATE(1074),
    [sym__atom] = STATE(1074),
    [sym_quoted_atom] = STATE(1074),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1074),
    [sym_charlist] = STATE(1074),
    [sym_sigil] = STATE(1074),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1074),
    [sym_tuple] = STATE(1074),
    [sym_bitstring] = STATE(1074),
    [sym_map] = STATE(1074),
    [sym__nullary_operator] = STATE(1074),
    [sym_unary_operator] = STATE(1074),
    [sym_binary_operator] = STATE(1074),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1074),
    [sym_call] = STATE(1074),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4729),
    [sym_rescue_block] = STATE(4729),
    [sym_catch_block] = STATE(4729),
    [sym_else_block] = STATE(4729),
    [sym_access_call] = STATE(1074),
    [sym_stab_clause] = STATE(4677),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1074),
    [aux_sym_do_block_repeat1] = STATE(4729),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(345),
    [sym_integer] = ACTIONS(345),
    [sym_float] = ACTIONS(345),
    [sym_char] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(345),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(347),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(33)] = {
    [sym__expression] = STATE(1079),
    [sym_block] = STATE(1079),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1079),
    [sym_nil] = STATE(1079),
    [sym__atom] = STATE(1079),
    [sym_quoted_atom] = STATE(1079),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1079),
    [sym_charlist] = STATE(1079),
    [sym_sigil] = STATE(1079),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1079),
    [sym_tuple] = STATE(1079),
    [sym_bitstring] = STATE(1079),
    [sym_map] = STATE(1079),
    [sym__nullary_operator] = STATE(1079),
    [sym_unary_operator] = STATE(1079),
    [sym_binary_operator] = STATE(1079),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1079),
    [sym_call] = STATE(1079),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_after_block] = STATE(4738),
    [sym_rescue_block] = STATE(4738),
    [sym_catch_block] = STATE(4738),
    [sym_else_block] = STATE(4738),
    [sym_access_call] = STATE(1079),
    [sym_stab_clause] = STATE(4684),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1079),
    [aux_sym_do_block_repeat1] = STATE(4738),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(349),
    [sym_integer] = ACTIONS(349),
    [sym_float] = ACTIONS(349),
    [sym_char] = ACTIONS(349),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(349),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(351),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(34)] = {
    [sym__terminator] = STATE(67),
    [sym__expression] = STATE(1126),
    [sym_block] = STATE(1126),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1126),
    [sym_nil] = STATE(1126),
    [sym__atom] = STATE(1126),
    [sym_quoted_atom] = STATE(1126),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1126),
    [sym_charlist] = STATE(1126),
    [sym_sigil] = STATE(1126),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1126),
    [sym_tuple] = STATE(1126),
    [sym_bitstring] = STATE(1126),
    [sym_map] = STATE(1126),
    [sym__nullary_operator] = STATE(1126),
    [sym_unary_operator] = STATE(1126),
    [sym_binary_operator] = STATE(1126),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1126),
    [sym_call] = STATE(1126),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1126),
    [sym_stab_clause] = STATE(4795),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1126),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(353),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(355),
    [sym_integer] = ACTIONS(355),
    [sym_float] = ACTIONS(355),
    [sym_char] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(355),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(357),
    [anon_sym_catch] = ACTIONS(357),
    [anon_sym_else] = ACTIONS(357),
    [anon_sym_end] = ACTIONS(357),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(357),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(35)] = {
    [sym__expression] = STATE(2244),
    [sym_block] = STATE(2244),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2244),
    [sym_nil] = STATE(2244),
    [sym__atom] = STATE(2244),
    [sym_quoted_atom] = STATE(2244),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2244),
    [sym_charlist] = STATE(2244),
    [sym_sigil] = STATE(2244),
    [sym_keywords] = STATE(2245),
    [sym_pair] = STATE(2158),
    [sym__keyword] = STATE(848),
    [sym_quoted_keyword] = STATE(848),
    [sym_list] = STATE(2244),
    [sym_tuple] = STATE(2244),
    [sym_bitstring] = STATE(2244),
    [sym_map] = STATE(2244),
    [sym__nullary_operator] = STATE(2244),
    [sym_unary_operator] = STATE(2244),
    [sym_binary_operator] = STATE(2244),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2244),
    [sym_call] = STATE(2244),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1788),
    [sym__call_arguments_without_parentheses] = STATE(2068),
    [sym_do_block] = STATE(2240),
    [sym_access_call] = STATE(2244),
    [sym_anonymous_function] = STATE(2244),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [anon_sym_RPAREN] = ACTIONS(187),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(363),
    [sym_integer] = ACTIONS(363),
    [sym_float] = ACTIONS(363),
    [sym_char] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_RBRACK] = ACTIONS(187),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(395),
    [anon_sym_LPAREN2] = ACTIONS(397),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(36)] = {
    [sym__expression] = STATE(2244),
    [sym_block] = STATE(2244),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2244),
    [sym_nil] = STATE(2244),
    [sym__atom] = STATE(2244),
    [sym_quoted_atom] = STATE(2244),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2244),
    [sym_charlist] = STATE(2244),
    [sym_sigil] = STATE(2244),
    [sym_keywords] = STATE(2245),
    [sym_pair] = STATE(2158),
    [sym__keyword] = STATE(848),
    [sym_quoted_keyword] = STATE(848),
    [sym_list] = STATE(2244),
    [sym_tuple] = STATE(2244),
    [sym_bitstring] = STATE(2244),
    [sym_map] = STATE(2244),
    [sym__nullary_operator] = STATE(2244),
    [sym_unary_operator] = STATE(2244),
    [sym_binary_operator] = STATE(2244),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2244),
    [sym_call] = STATE(2244),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1787),
    [sym__call_arguments_without_parentheses] = STATE(2063),
    [sym_do_block] = STATE(2233),
    [sym_access_call] = STATE(2244),
    [sym_anonymous_function] = STATE(2244),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [anon_sym_RPAREN] = ACTIONS(235),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(363),
    [sym_integer] = ACTIONS(363),
    [sym_float] = ACTIONS(363),
    [sym_char] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(235),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_RBRACK] = ACTIONS(235),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(403),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(406),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(395),
    [anon_sym_LPAREN2] = ACTIONS(397),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(37)] = {
    [sym__terminator] = STATE(68),
    [sym__expression] = STATE(1094),
    [sym_block] = STATE(1094),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1094),
    [sym_nil] = STATE(1094),
    [sym__atom] = STATE(1094),
    [sym_quoted_atom] = STATE(1094),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1094),
    [sym_charlist] = STATE(1094),
    [sym_sigil] = STATE(1094),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1094),
    [sym_tuple] = STATE(1094),
    [sym_bitstring] = STATE(1094),
    [sym_map] = STATE(1094),
    [sym__nullary_operator] = STATE(1094),
    [sym_unary_operator] = STATE(1094),
    [sym_binary_operator] = STATE(1094),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1094),
    [sym_call] = STATE(1094),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1094),
    [sym_stab_clause] = STATE(4796),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1094),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(411),
    [sym_integer] = ACTIONS(411),
    [sym_float] = ACTIONS(411),
    [sym_char] = ACTIONS(411),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(411),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(413),
    [anon_sym_catch] = ACTIONS(413),
    [anon_sym_else] = ACTIONS(413),
    [anon_sym_end] = ACTIONS(413),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(413),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(38)] = {
    [sym__expression] = STATE(2244),
    [sym_block] = STATE(2244),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2244),
    [sym_nil] = STATE(2244),
    [sym__atom] = STATE(2244),
    [sym_quoted_atom] = STATE(2244),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2244),
    [sym_charlist] = STATE(2244),
    [sym_sigil] = STATE(2244),
    [sym_keywords] = STATE(2245),
    [sym_pair] = STATE(2158),
    [sym__keyword] = STATE(848),
    [sym_quoted_keyword] = STATE(848),
    [sym_list] = STATE(2244),
    [sym_tuple] = STATE(2244),
    [sym_bitstring] = STATE(2244),
    [sym_map] = STATE(2244),
    [sym__nullary_operator] = STATE(2244),
    [sym_unary_operator] = STATE(2244),
    [sym_binary_operator] = STATE(2244),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2244),
    [sym_call] = STATE(2244),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1790),
    [sym__call_arguments_without_parentheses] = STATE(2074),
    [sym_do_block] = STATE(3054),
    [sym_access_call] = STATE(2244),
    [sym_anonymous_function] = STATE(2244),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [anon_sym_RPAREN] = ACTIONS(187),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(363),
    [sym_integer] = ACTIONS(363),
    [sym_float] = ACTIONS(363),
    [sym_char] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_RBRACK] = ACTIONS(187),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(415),
    [anon_sym_fn] = ACTIONS(395),
    [anon_sym_LPAREN2] = ACTIONS(397),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(417),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(39)] = {
    [sym__terminator] = STATE(70),
    [sym__expression] = STATE(1106),
    [sym_block] = STATE(1106),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1106),
    [sym_nil] = STATE(1106),
    [sym__atom] = STATE(1106),
    [sym_quoted_atom] = STATE(1106),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1106),
    [sym_charlist] = STATE(1106),
    [sym_sigil] = STATE(1106),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1106),
    [sym_tuple] = STATE(1106),
    [sym_bitstring] = STATE(1106),
    [sym_map] = STATE(1106),
    [sym__nullary_operator] = STATE(1106),
    [sym_unary_operator] = STATE(1106),
    [sym_binary_operator] = STATE(1106),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1106),
    [sym_call] = STATE(1106),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1106),
    [sym_stab_clause] = STATE(4779),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1106),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(419),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(421),
    [sym_integer] = ACTIONS(421),
    [sym_float] = ACTIONS(421),
    [sym_char] = ACTIONS(421),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(421),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(423),
    [anon_sym_catch] = ACTIONS(423),
    [anon_sym_else] = ACTIONS(423),
    [anon_sym_end] = ACTIONS(423),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(423),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(40)] = {
    [sym__terminator] = STATE(69),
    [sym__expression] = STATE(1101),
    [sym_block] = STATE(1101),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1101),
    [sym_nil] = STATE(1101),
    [sym__atom] = STATE(1101),
    [sym_quoted_atom] = STATE(1101),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1101),
    [sym_charlist] = STATE(1101),
    [sym_sigil] = STATE(1101),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1101),
    [sym_tuple] = STATE(1101),
    [sym_bitstring] = STATE(1101),
    [sym_map] = STATE(1101),
    [sym__nullary_operator] = STATE(1101),
    [sym_unary_operator] = STATE(1101),
    [sym_binary_operator] = STATE(1101),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1101),
    [sym_call] = STATE(1101),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1101),
    [sym_stab_clause] = STATE(4789),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1101),
    [aux_sym__terminator_repeat1] = STATE(1018),
    [aux_sym__terminator_token1] = ACTIONS(59),
    [anon_sym_SEMI] = ACTIONS(425),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(427),
    [sym_integer] = ACTIONS(427),
    [sym_float] = ACTIONS(427),
    [sym_char] = ACTIONS(427),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(427),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(429),
    [anon_sym_catch] = ACTIONS(429),
    [anon_sym_else] = ACTIONS(429),
    [anon_sym_end] = ACTIONS(429),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(429),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(41)] = {
    [sym__expression] = STATE(2483),
    [sym_block] = STATE(2483),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2483),
    [sym_nil] = STATE(2483),
    [sym__atom] = STATE(2483),
    [sym_quoted_atom] = STATE(2483),
    [sym__quoted_i_double] = STATE(2688),
    [sym__quoted_i_single] = STATE(2689),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2483),
    [sym_charlist] = STATE(2483),
    [sym_sigil] = STATE(2483),
    [sym_keywords] = STATE(3084),
    [sym_pair] = STATE(2458),
    [sym__keyword] = STATE(843),
    [sym_quoted_keyword] = STATE(843),
    [sym_list] = STATE(2483),
    [sym_tuple] = STATE(2483),
    [sym_bitstring] = STATE(2483),
    [sym_map] = STATE(2483),
    [sym__nullary_operator] = STATE(2483),
    [sym_unary_operator] = STATE(2483),
    [sym_binary_operator] = STATE(2483),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2483),
    [sym_call] = STATE(2483),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2216),
    [sym__call_arguments_without_parentheses] = STATE(2709),
    [sym_do_block] = STATE(2929),
    [sym_access_call] = STATE(2483),
    [sym_anonymous_function] = STATE(2483),
    [ts_builtin_sym_end] = ACTIONS(185),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(435),
    [sym_integer] = ACTIONS(435),
    [sym_float] = ACTIONS(435),
    [sym_char] = ACTIONS(435),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(435),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(455),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(467),
    [anon_sym_LPAREN2] = ACTIONS(469),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(42)] = {
    [sym__expression] = STATE(2354),
    [sym_block] = STATE(2354),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2354),
    [sym_nil] = STATE(2354),
    [sym__atom] = STATE(2354),
    [sym_quoted_atom] = STATE(2354),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2354),
    [sym_charlist] = STATE(2354),
    [sym_sigil] = STATE(2354),
    [sym_keywords] = STATE(1239),
    [sym_pair] = STATE(2218),
    [sym__keyword] = STATE(845),
    [sym_quoted_keyword] = STATE(845),
    [sym_list] = STATE(2354),
    [sym_tuple] = STATE(2354),
    [sym_bitstring] = STATE(2354),
    [sym_map] = STATE(2354),
    [sym__nullary_operator] = STATE(2354),
    [sym_unary_operator] = STATE(2354),
    [sym_binary_operator] = STATE(2354),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2354),
    [sym_call] = STATE(2354),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1092),
    [sym__call_arguments_without_parentheses] = STATE(1125),
    [sym_do_block] = STATE(1185),
    [sym_access_call] = STATE(2354),
    [sym_anonymous_function] = STATE(2354),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(237),
    [anon_sym_RPAREN] = ACTIONS(187),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(477),
    [sym_integer] = ACTIONS(477),
    [sym_float] = ACTIONS(477),
    [sym_char] = ACTIONS(477),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(477),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(481),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(280),
    [anon_sym_LPAREN2] = ACTIONS(282),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(43)] = {
    [sym__expression] = STATE(2701),
    [sym_block] = STATE(2701),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(2701),
    [sym_nil] = STATE(2701),
    [sym__atom] = STATE(2701),
    [sym_quoted_atom] = STATE(2701),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2701),
    [sym_charlist] = STATE(2701),
    [sym_sigil] = STATE(2701),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(2695),
    [sym__keyword] = STATE(880),
    [sym_quoted_keyword] = STATE(880),
    [sym_list] = STATE(2701),
    [sym_tuple] = STATE(2701),
    [sym_bitstring] = STATE(2701),
    [sym_map] = STATE(2701),
    [sym__nullary_operator] = STATE(2701),
    [sym_unary_operator] = STATE(2701),
    [sym_binary_operator] = STATE(2701),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2701),
    [sym_call] = STATE(2701),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1121),
    [sym__call_arguments_without_parentheses] = STATE(1228),
    [sym_do_block] = STATE(2014),
    [sym_access_call] = STATE(2701),
    [sym_anonymous_function] = STATE(2701),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(189),
    [anon_sym_RPAREN] = ACTIONS(187),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(491),
    [sym_integer] = ACTIONS(491),
    [sym_float] = ACTIONS(491),
    [sym_char] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(491),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(495),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(291),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(293),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(44)] = {
    [sym__expression] = STATE(2354),
    [sym_block] = STATE(2354),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2354),
    [sym_nil] = STATE(2354),
    [sym__atom] = STATE(2354),
    [sym_quoted_atom] = STATE(2354),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2354),
    [sym_charlist] = STATE(2354),
    [sym_sigil] = STATE(2354),
    [sym_keywords] = STATE(1239),
    [sym_pair] = STATE(2218),
    [sym__keyword] = STATE(845),
    [sym_quoted_keyword] = STATE(845),
    [sym_list] = STATE(2354),
    [sym_tuple] = STATE(2354),
    [sym_bitstring] = STATE(2354),
    [sym_map] = STATE(2354),
    [sym__nullary_operator] = STATE(2354),
    [sym_unary_operator] = STATE(2354),
    [sym_binary_operator] = STATE(2354),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2354),
    [sym_call] = STATE(2354),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1087),
    [sym__call_arguments_without_parentheses] = STATE(1135),
    [sym_do_block] = STATE(1306),
    [sym_access_call] = STATE(2354),
    [sym_anonymous_function] = STATE(2354),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(237),
    [anon_sym_RPAREN] = ACTIONS(235),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(477),
    [sym_integer] = ACTIONS(477),
    [sym_float] = ACTIONS(477),
    [sym_char] = ACTIONS(477),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(477),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(481),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(268),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(505),
    [anon_sym_DASH] = ACTIONS(505),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(257),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(280),
    [anon_sym_LPAREN2] = ACTIONS(282),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(45)] = {
    [sym__expression] = STATE(2527),
    [sym_block] = STATE(2527),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(2527),
    [sym_nil] = STATE(2527),
    [sym__atom] = STATE(2527),
    [sym_quoted_atom] = STATE(2527),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2527),
    [sym_charlist] = STATE(2527),
    [sym_sigil] = STATE(2527),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(2917),
    [sym__keyword] = STATE(887),
    [sym_quoted_keyword] = STATE(887),
    [sym_list] = STATE(2527),
    [sym_tuple] = STATE(2527),
    [sym_bitstring] = STATE(2527),
    [sym_map] = STATE(2527),
    [sym__nullary_operator] = STATE(2527),
    [sym_unary_operator] = STATE(2527),
    [sym_binary_operator] = STATE(2527),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2527),
    [sym_call] = STATE(2527),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1121),
    [sym__call_arguments_without_parentheses] = STATE(1228),
    [sym_do_block] = STATE(2014),
    [sym_access_call] = STATE(2527),
    [sym_anonymous_function] = STATE(2527),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(510),
    [sym_integer] = ACTIONS(510),
    [sym_float] = ACTIONS(510),
    [sym_char] = ACTIONS(510),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(510),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(514),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(291),
    [anon_sym_end] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(293),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(46)] = {
    [sym__expression] = STATE(2701),
    [sym_block] = STATE(2701),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(2701),
    [sym_nil] = STATE(2701),
    [sym__atom] = STATE(2701),
    [sym_quoted_atom] = STATE(2701),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2701),
    [sym_charlist] = STATE(2701),
    [sym_sigil] = STATE(2701),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(2695),
    [sym__keyword] = STATE(880),
    [sym_quoted_keyword] = STATE(880),
    [sym_list] = STATE(2701),
    [sym_tuple] = STATE(2701),
    [sym_bitstring] = STATE(2701),
    [sym_map] = STATE(2701),
    [sym__nullary_operator] = STATE(2701),
    [sym_unary_operator] = STATE(2701),
    [sym_binary_operator] = STATE(2701),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2701),
    [sym_call] = STATE(2701),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1107),
    [sym__call_arguments_without_parentheses] = STATE(1320),
    [sym_do_block] = STATE(1493),
    [sym_access_call] = STATE(2701),
    [sym_anonymous_function] = STATE(2701),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(189),
    [anon_sym_RPAREN] = ACTIONS(235),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(491),
    [sym_integer] = ACTIONS(491),
    [sym_float] = ACTIONS(491),
    [sym_char] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(491),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(495),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(299),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(524),
    [anon_sym_DASH] = ACTIONS(524),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(47)] = {
    [sym__expression] = STATE(2701),
    [sym_block] = STATE(2701),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(2701),
    [sym_nil] = STATE(2701),
    [sym__atom] = STATE(2701),
    [sym_quoted_atom] = STATE(2701),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2701),
    [sym_charlist] = STATE(2701),
    [sym_sigil] = STATE(2701),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(2695),
    [sym__keyword] = STATE(880),
    [sym_quoted_keyword] = STATE(880),
    [sym_list] = STATE(2701),
    [sym_tuple] = STATE(2701),
    [sym_bitstring] = STATE(2701),
    [sym_map] = STATE(2701),
    [sym__nullary_operator] = STATE(2701),
    [sym_unary_operator] = STATE(2701),
    [sym_binary_operator] = STATE(2701),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2701),
    [sym_call] = STATE(2701),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1109),
    [sym__call_arguments_without_parentheses] = STATE(1326),
    [sym_do_block] = STATE(1500),
    [sym_access_call] = STATE(2701),
    [sym_anonymous_function] = STATE(2701),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(189),
    [anon_sym_RPAREN] = ACTIONS(187),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(491),
    [sym_integer] = ACTIONS(491),
    [sym_float] = ACTIONS(491),
    [sym_char] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(491),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(495),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(48)] = {
    [sym__expression] = STATE(2527),
    [sym_block] = STATE(2527),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(2527),
    [sym_nil] = STATE(2527),
    [sym__atom] = STATE(2527),
    [sym_quoted_atom] = STATE(2527),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2527),
    [sym_charlist] = STATE(2527),
    [sym_sigil] = STATE(2527),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(2917),
    [sym__keyword] = STATE(887),
    [sym_quoted_keyword] = STATE(887),
    [sym_list] = STATE(2527),
    [sym_tuple] = STATE(2527),
    [sym_bitstring] = STATE(2527),
    [sym_map] = STATE(2527),
    [sym__nullary_operator] = STATE(2527),
    [sym_unary_operator] = STATE(2527),
    [sym_binary_operator] = STATE(2527),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2527),
    [sym_call] = STATE(2527),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1109),
    [sym__call_arguments_without_parentheses] = STATE(1326),
    [sym_do_block] = STATE(1500),
    [sym_access_call] = STATE(2527),
    [sym_anonymous_function] = STATE(2527),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(510),
    [sym_integer] = ACTIONS(510),
    [sym_float] = ACTIONS(510),
    [sym_char] = ACTIONS(510),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(510),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(514),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_end] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(49)] = {
    [sym__expression] = STATE(2483),
    [sym_block] = STATE(2483),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2483),
    [sym_nil] = STATE(2483),
    [sym__atom] = STATE(2483),
    [sym_quoted_atom] = STATE(2483),
    [sym__quoted_i_double] = STATE(2688),
    [sym__quoted_i_single] = STATE(2689),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2483),
    [sym_charlist] = STATE(2483),
    [sym_sigil] = STATE(2483),
    [sym_keywords] = STATE(3084),
    [sym_pair] = STATE(2458),
    [sym__keyword] = STATE(843),
    [sym_quoted_keyword] = STATE(843),
    [sym_list] = STATE(2483),
    [sym_tuple] = STATE(2483),
    [sym_bitstring] = STATE(2483),
    [sym_map] = STATE(2483),
    [sym__nullary_operator] = STATE(2483),
    [sym_unary_operator] = STATE(2483),
    [sym_binary_operator] = STATE(2483),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2483),
    [sym_call] = STATE(2483),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2176),
    [sym__call_arguments_without_parentheses] = STATE(2690),
    [sym_do_block] = STATE(4120),
    [sym_access_call] = STATE(2483),
    [sym_anonymous_function] = STATE(2483),
    [ts_builtin_sym_end] = ACTIONS(185),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(435),
    [sym_integer] = ACTIONS(435),
    [sym_float] = ACTIONS(435),
    [sym_char] = ACTIONS(435),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(435),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(455),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(527),
    [anon_sym_fn] = ACTIONS(467),
    [anon_sym_LPAREN2] = ACTIONS(469),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(529),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(50)] = {
    [sym__expression] = STATE(2354),
    [sym_block] = STATE(2354),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2354),
    [sym_nil] = STATE(2354),
    [sym__atom] = STATE(2354),
    [sym_quoted_atom] = STATE(2354),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2354),
    [sym_charlist] = STATE(2354),
    [sym_sigil] = STATE(2354),
    [sym_keywords] = STATE(1239),
    [sym_pair] = STATE(2218),
    [sym__keyword] = STATE(845),
    [sym_quoted_keyword] = STATE(845),
    [sym_list] = STATE(2354),
    [sym_tuple] = STATE(2354),
    [sym_bitstring] = STATE(2354),
    [sym_map] = STATE(2354),
    [sym__nullary_operator] = STATE(2354),
    [sym_unary_operator] = STATE(2354),
    [sym_binary_operator] = STATE(2354),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2354),
    [sym_call] = STATE(2354),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1089),
    [sym__call_arguments_without_parentheses] = STATE(1123),
    [sym_do_block] = STATE(1642),
    [sym_access_call] = STATE(2354),
    [sym_anonymous_function] = STATE(2354),
    [aux_sym__terminator_token1] = ACTIONS(185),
    [anon_sym_SEMI] = ACTIONS(187),
    [anon_sym_LPAREN] = ACTIONS(237),
    [anon_sym_RPAREN] = ACTIONS(187),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(477),
    [sym_integer] = ACTIONS(477),
    [sym_float] = ACTIONS(477),
    [sym_char] = ACTIONS(477),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(477),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(481),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(295),
    [anon_sym_fn] = ACTIONS(280),
    [anon_sym_LPAREN2] = ACTIONS(282),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(297),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(51)] = {
    [sym__expression] = STATE(2244),
    [sym_block] = STATE(2244),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2244),
    [sym_nil] = STATE(2244),
    [sym__atom] = STATE(2244),
    [sym_quoted_atom] = STATE(2244),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2244),
    [sym_charlist] = STATE(2244),
    [sym_sigil] = STATE(2244),
    [sym_keywords] = STATE(2245),
    [sym_pair] = STATE(2158),
    [sym__keyword] = STATE(848),
    [sym_quoted_keyword] = STATE(848),
    [sym_list] = STATE(2244),
    [sym_tuple] = STATE(2244),
    [sym_bitstring] = STATE(2244),
    [sym_map] = STATE(2244),
    [sym__nullary_operator] = STATE(2244),
    [sym_unary_operator] = STATE(2244),
    [sym_binary_operator] = STATE(2244),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2244),
    [sym_call] = STATE(2244),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1789),
    [sym__call_arguments_without_parentheses] = STATE(2071),
    [sym_do_block] = STATE(3046),
    [sym_access_call] = STATE(2244),
    [sym_anonymous_function] = STATE(2244),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [anon_sym_RPAREN] = ACTIONS(235),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(363),
    [sym_integer] = ACTIONS(363),
    [sym_float] = ACTIONS(363),
    [sym_char] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(235),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_RBRACK] = ACTIONS(235),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(403),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(406),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(415),
    [anon_sym_fn] = ACTIONS(395),
    [anon_sym_LPAREN2] = ACTIONS(397),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(52)] = {
    [sym__expression] = STATE(2527),
    [sym_block] = STATE(2527),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(2527),
    [sym_nil] = STATE(2527),
    [sym__atom] = STATE(2527),
    [sym_quoted_atom] = STATE(2527),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2527),
    [sym_charlist] = STATE(2527),
    [sym_sigil] = STATE(2527),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(2917),
    [sym__keyword] = STATE(887),
    [sym_quoted_keyword] = STATE(887),
    [sym_list] = STATE(2527),
    [sym_tuple] = STATE(2527),
    [sym_bitstring] = STATE(2527),
    [sym_map] = STATE(2527),
    [sym__nullary_operator] = STATE(2527),
    [sym_unary_operator] = STATE(2527),
    [sym_binary_operator] = STATE(2527),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2527),
    [sym_call] = STATE(2527),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1107),
    [sym__call_arguments_without_parentheses] = STATE(1320),
    [sym_do_block] = STATE(1493),
    [sym_access_call] = STATE(2527),
    [sym_anonymous_function] = STATE(2527),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(510),
    [sym_integer] = ACTIONS(510),
    [sym_float] = ACTIONS(510),
    [sym_char] = ACTIONS(510),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(510),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(514),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(299),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(531),
    [anon_sym_DASH] = ACTIONS(531),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_end] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(53)] = {
    [sym__expression] = STATE(2483),
    [sym_block] = STATE(2483),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2483),
    [sym_nil] = STATE(2483),
    [sym__atom] = STATE(2483),
    [sym_quoted_atom] = STATE(2483),
    [sym__quoted_i_double] = STATE(2688),
    [sym__quoted_i_single] = STATE(2689),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2483),
    [sym_charlist] = STATE(2483),
    [sym_sigil] = STATE(2483),
    [sym_keywords] = STATE(3084),
    [sym_pair] = STATE(2458),
    [sym__keyword] = STATE(843),
    [sym_quoted_keyword] = STATE(843),
    [sym_list] = STATE(2483),
    [sym_tuple] = STATE(2483),
    [sym_bitstring] = STATE(2483),
    [sym_map] = STATE(2483),
    [sym__nullary_operator] = STATE(2483),
    [sym_unary_operator] = STATE(2483),
    [sym_binary_operator] = STATE(2483),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2483),
    [sym_call] = STATE(2483),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2215),
    [sym__call_arguments_without_parentheses] = STATE(2704),
    [sym_do_block] = STATE(3451),
    [sym_access_call] = STATE(2483),
    [sym_anonymous_function] = STATE(2483),
    [ts_builtin_sym_end] = ACTIONS(233),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(435),
    [sym_integer] = ACTIONS(435),
    [sym_float] = ACTIONS(435),
    [sym_char] = ACTIONS(435),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(435),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(455),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(534),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(537),
    [anon_sym_DASH] = ACTIONS(537),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(467),
    [anon_sym_LPAREN2] = ACTIONS(469),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(54)] = {
    [sym__expression] = STATE(3200),
    [sym_block] = STATE(3200),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3200),
    [sym_nil] = STATE(3200),
    [sym__atom] = STATE(3200),
    [sym_quoted_atom] = STATE(3200),
    [sym__quoted_i_double] = STATE(3196),
    [sym__quoted_i_single] = STATE(3197),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3200),
    [sym_charlist] = STATE(3200),
    [sym_sigil] = STATE(3200),
    [sym_keywords] = STATE(3516),
    [sym_pair] = STATE(3225),
    [sym__keyword] = STATE(855),
    [sym_quoted_keyword] = STATE(855),
    [sym_list] = STATE(3200),
    [sym_tuple] = STATE(3200),
    [sym_bitstring] = STATE(3200),
    [sym_map] = STATE(3200),
    [sym__nullary_operator] = STATE(3200),
    [sym_unary_operator] = STATE(3200),
    [sym_binary_operator] = STATE(3200),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3200),
    [sym_call] = STATE(3200),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2589),
    [sym__call_arguments_without_parentheses] = STATE(3257),
    [sym_do_block] = STATE(3694),
    [sym_access_call] = STATE(3200),
    [sym_anonymous_function] = STATE(3200),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(544),
    [sym_integer] = ACTIONS(544),
    [sym_float] = ACTIONS(544),
    [sym_char] = ACTIONS(544),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(544),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(564),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_GT_GT] = ACTIONS(235),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(570),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(575),
    [anon_sym_DASH] = ACTIONS(575),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(582),
    [anon_sym_LPAREN2] = ACTIONS(584),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(55)] = {
    [sym__expression] = STATE(2665),
    [sym_block] = STATE(2665),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2665),
    [sym_nil] = STATE(2665),
    [sym__atom] = STATE(2665),
    [sym_quoted_atom] = STATE(2665),
    [sym__quoted_i_double] = STATE(2509),
    [sym__quoted_i_single] = STATE(2510),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2665),
    [sym_charlist] = STATE(2665),
    [sym_sigil] = STATE(2665),
    [sym_keywords] = STATE(3238),
    [sym_pair] = STATE(2669),
    [sym__keyword] = STATE(868),
    [sym_quoted_keyword] = STATE(868),
    [sym_list] = STATE(2665),
    [sym_tuple] = STATE(2665),
    [sym_bitstring] = STATE(2665),
    [sym_map] = STATE(2665),
    [sym__nullary_operator] = STATE(2665),
    [sym_unary_operator] = STATE(2665),
    [sym_binary_operator] = STATE(2665),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2665),
    [sym_call] = STATE(2665),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2188),
    [sym__call_arguments_without_parentheses] = STATE(2679),
    [sym_do_block] = STATE(4174),
    [sym_access_call] = STATE(2665),
    [sym_anonymous_function] = STATE(2665),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [anon_sym_RPAREN] = ACTIONS(187),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(592),
    [sym_integer] = ACTIONS(592),
    [sym_float] = ACTIONS(592),
    [sym_char] = ACTIONS(592),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(592),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(612),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(624),
    [anon_sym_fn] = ACTIONS(626),
    [anon_sym_LPAREN2] = ACTIONS(628),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(630),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(56)] = {
    [sym__expression] = STATE(2527),
    [sym_block] = STATE(2527),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(2527),
    [sym_nil] = STATE(2527),
    [sym__atom] = STATE(2527),
    [sym_quoted_atom] = STATE(2527),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2527),
    [sym_charlist] = STATE(2527),
    [sym_sigil] = STATE(2527),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(2917),
    [sym__keyword] = STATE(887),
    [sym_quoted_keyword] = STATE(887),
    [sym_list] = STATE(2527),
    [sym_tuple] = STATE(2527),
    [sym_bitstring] = STATE(2527),
    [sym_map] = STATE(2527),
    [sym__nullary_operator] = STATE(2527),
    [sym_unary_operator] = STATE(2527),
    [sym_binary_operator] = STATE(2527),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2527),
    [sym_call] = STATE(2527),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1119),
    [sym__call_arguments_without_parentheses] = STATE(1219),
    [sym_do_block] = STATE(2060),
    [sym_access_call] = STATE(2527),
    [sym_anonymous_function] = STATE(2527),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(510),
    [sym_integer] = ACTIONS(510),
    [sym_float] = ACTIONS(510),
    [sym_char] = ACTIONS(510),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(510),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(514),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(299),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(531),
    [anon_sym_DASH] = ACTIONS(531),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(291),
    [anon_sym_end] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(57)] = {
    [sym__expression] = STATE(2665),
    [sym_block] = STATE(2665),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2665),
    [sym_nil] = STATE(2665),
    [sym__atom] = STATE(2665),
    [sym_quoted_atom] = STATE(2665),
    [sym__quoted_i_double] = STATE(2509),
    [sym__quoted_i_single] = STATE(2510),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2665),
    [sym_charlist] = STATE(2665),
    [sym_sigil] = STATE(2665),
    [sym_keywords] = STATE(3238),
    [sym_pair] = STATE(2669),
    [sym__keyword] = STATE(868),
    [sym_quoted_keyword] = STATE(868),
    [sym_list] = STATE(2665),
    [sym_tuple] = STATE(2665),
    [sym_bitstring] = STATE(2665),
    [sym_map] = STATE(2665),
    [sym__nullary_operator] = STATE(2665),
    [sym_unary_operator] = STATE(2665),
    [sym_binary_operator] = STATE(2665),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2665),
    [sym_call] = STATE(2665),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2193),
    [sym__call_arguments_without_parentheses] = STATE(2627),
    [sym_do_block] = STATE(3377),
    [sym_access_call] = STATE(2665),
    [sym_anonymous_function] = STATE(2665),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [anon_sym_RPAREN] = ACTIONS(235),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(592),
    [sym_integer] = ACTIONS(592),
    [sym_float] = ACTIONS(592),
    [sym_char] = ACTIONS(592),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(592),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(612),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(636),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(639),
    [anon_sym_DASH] = ACTIONS(639),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(257),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(626),
    [anon_sym_LPAREN2] = ACTIONS(628),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(58)] = {
    [sym__expression] = STATE(2665),
    [sym_block] = STATE(2665),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2665),
    [sym_nil] = STATE(2665),
    [sym__atom] = STATE(2665),
    [sym_quoted_atom] = STATE(2665),
    [sym__quoted_i_double] = STATE(2509),
    [sym__quoted_i_single] = STATE(2510),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2665),
    [sym_charlist] = STATE(2665),
    [sym_sigil] = STATE(2665),
    [sym_keywords] = STATE(3238),
    [sym_pair] = STATE(2669),
    [sym__keyword] = STATE(868),
    [sym_quoted_keyword] = STATE(868),
    [sym_list] = STATE(2665),
    [sym_tuple] = STATE(2665),
    [sym_bitstring] = STATE(2665),
    [sym_map] = STATE(2665),
    [sym__nullary_operator] = STATE(2665),
    [sym_unary_operator] = STATE(2665),
    [sym_binary_operator] = STATE(2665),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2665),
    [sym_call] = STATE(2665),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2248),
    [sym__call_arguments_without_parentheses] = STATE(2631),
    [sym_do_block] = STATE(3381),
    [sym_access_call] = STATE(2665),
    [sym_anonymous_function] = STATE(2665),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [anon_sym_RPAREN] = ACTIONS(187),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(592),
    [sym_integer] = ACTIONS(592),
    [sym_float] = ACTIONS(592),
    [sym_char] = ACTIONS(592),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(592),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(612),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(626),
    [anon_sym_LPAREN2] = ACTIONS(628),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(59)] = {
    [sym__expression] = STATE(2354),
    [sym_block] = STATE(2354),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2354),
    [sym_nil] = STATE(2354),
    [sym__atom] = STATE(2354),
    [sym_quoted_atom] = STATE(2354),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2354),
    [sym_charlist] = STATE(2354),
    [sym_sigil] = STATE(2354),
    [sym_keywords] = STATE(1239),
    [sym_pair] = STATE(2218),
    [sym__keyword] = STATE(845),
    [sym_quoted_keyword] = STATE(845),
    [sym_list] = STATE(2354),
    [sym_tuple] = STATE(2354),
    [sym_bitstring] = STATE(2354),
    [sym_map] = STATE(2354),
    [sym__nullary_operator] = STATE(2354),
    [sym_unary_operator] = STATE(2354),
    [sym_binary_operator] = STATE(2354),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2354),
    [sym_call] = STATE(2354),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1090),
    [sym__call_arguments_without_parentheses] = STATE(1097),
    [sym_do_block] = STATE(1641),
    [sym_access_call] = STATE(2354),
    [sym_anonymous_function] = STATE(2354),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(237),
    [anon_sym_RPAREN] = ACTIONS(235),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(477),
    [sym_integer] = ACTIONS(477),
    [sym_float] = ACTIONS(477),
    [sym_char] = ACTIONS(477),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(477),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(481),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(268),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(505),
    [anon_sym_DASH] = ACTIONS(505),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(257),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(295),
    [anon_sym_fn] = ACTIONS(280),
    [anon_sym_LPAREN2] = ACTIONS(282),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(60)] = {
    [sym__expression] = STATE(2483),
    [sym_block] = STATE(2483),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2483),
    [sym_nil] = STATE(2483),
    [sym__atom] = STATE(2483),
    [sym_quoted_atom] = STATE(2483),
    [sym__quoted_i_double] = STATE(2688),
    [sym__quoted_i_single] = STATE(2689),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2483),
    [sym_charlist] = STATE(2483),
    [sym_sigil] = STATE(2483),
    [sym_keywords] = STATE(3084),
    [sym_pair] = STATE(2458),
    [sym__keyword] = STATE(843),
    [sym_quoted_keyword] = STATE(843),
    [sym_list] = STATE(2483),
    [sym_tuple] = STATE(2483),
    [sym_bitstring] = STATE(2483),
    [sym_map] = STATE(2483),
    [sym__nullary_operator] = STATE(2483),
    [sym_unary_operator] = STATE(2483),
    [sym_binary_operator] = STATE(2483),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2483),
    [sym_call] = STATE(2483),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2362),
    [sym__call_arguments_without_parentheses] = STATE(2678),
    [sym_do_block] = STATE(4282),
    [sym_access_call] = STATE(2483),
    [sym_anonymous_function] = STATE(2483),
    [ts_builtin_sym_end] = ACTIONS(233),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(435),
    [sym_integer] = ACTIONS(435),
    [sym_float] = ACTIONS(435),
    [sym_char] = ACTIONS(435),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(435),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(455),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(534),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(537),
    [anon_sym_DASH] = ACTIONS(537),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(527),
    [anon_sym_fn] = ACTIONS(467),
    [anon_sym_LPAREN2] = ACTIONS(469),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(61)] = {
    [sym__expression] = STATE(3200),
    [sym_block] = STATE(3200),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3200),
    [sym_nil] = STATE(3200),
    [sym__atom] = STATE(3200),
    [sym_quoted_atom] = STATE(3200),
    [sym__quoted_i_double] = STATE(3196),
    [sym__quoted_i_single] = STATE(3197),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3200),
    [sym_charlist] = STATE(3200),
    [sym_sigil] = STATE(3200),
    [sym_keywords] = STATE(3516),
    [sym_pair] = STATE(3225),
    [sym__keyword] = STATE(855),
    [sym_quoted_keyword] = STATE(855),
    [sym_list] = STATE(3200),
    [sym_tuple] = STATE(3200),
    [sym_bitstring] = STATE(3200),
    [sym_map] = STATE(3200),
    [sym__nullary_operator] = STATE(3200),
    [sym_unary_operator] = STATE(3200),
    [sym_binary_operator] = STATE(3200),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3200),
    [sym_call] = STATE(3200),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2590),
    [sym__call_arguments_without_parentheses] = STATE(3260),
    [sym_do_block] = STATE(3699),
    [sym_access_call] = STATE(3200),
    [sym_anonymous_function] = STATE(3200),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(544),
    [sym_integer] = ACTIONS(544),
    [sym_float] = ACTIONS(544),
    [sym_char] = ACTIONS(544),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(544),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(564),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_GT_GT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(582),
    [anon_sym_LPAREN2] = ACTIONS(584),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(62)] = {
    [sym__expression] = STATE(3200),
    [sym_block] = STATE(3200),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3200),
    [sym_nil] = STATE(3200),
    [sym__atom] = STATE(3200),
    [sym_quoted_atom] = STATE(3200),
    [sym__quoted_i_double] = STATE(3196),
    [sym__quoted_i_single] = STATE(3197),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3200),
    [sym_charlist] = STATE(3200),
    [sym_sigil] = STATE(3200),
    [sym_keywords] = STATE(3516),
    [sym_pair] = STATE(3225),
    [sym__keyword] = STATE(855),
    [sym_quoted_keyword] = STATE(855),
    [sym_list] = STATE(3200),
    [sym_tuple] = STATE(3200),
    [sym_bitstring] = STATE(3200),
    [sym_map] = STATE(3200),
    [sym__nullary_operator] = STATE(3200),
    [sym_unary_operator] = STATE(3200),
    [sym_binary_operator] = STATE(3200),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3200),
    [sym_call] = STATE(3200),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2664),
    [sym__call_arguments_without_parentheses] = STATE(3204),
    [sym_do_block] = STATE(4441),
    [sym_access_call] = STATE(3200),
    [sym_anonymous_function] = STATE(3200),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(544),
    [sym_integer] = ACTIONS(544),
    [sym_float] = ACTIONS(544),
    [sym_char] = ACTIONS(544),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(544),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(564),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_GT_GT] = ACTIONS(187),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(642),
    [anon_sym_fn] = ACTIONS(582),
    [anon_sym_LPAREN2] = ACTIONS(584),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(644),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(63)] = {
    [sym__expression] = STATE(2701),
    [sym_block] = STATE(2701),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(2701),
    [sym_nil] = STATE(2701),
    [sym__atom] = STATE(2701),
    [sym_quoted_atom] = STATE(2701),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2701),
    [sym_charlist] = STATE(2701),
    [sym_sigil] = STATE(2701),
    [sym_keywords] = STATE(1484),
    [sym_pair] = STATE(2695),
    [sym__keyword] = STATE(880),
    [sym_quoted_keyword] = STATE(880),
    [sym_list] = STATE(2701),
    [sym_tuple] = STATE(2701),
    [sym_bitstring] = STATE(2701),
    [sym_map] = STATE(2701),
    [sym__nullary_operator] = STATE(2701),
    [sym_unary_operator] = STATE(2701),
    [sym_binary_operator] = STATE(2701),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2701),
    [sym_call] = STATE(2701),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1119),
    [sym__call_arguments_without_parentheses] = STATE(1219),
    [sym_do_block] = STATE(2060),
    [sym_access_call] = STATE(2701),
    [sym_anonymous_function] = STATE(2701),
    [aux_sym__terminator_token1] = ACTIONS(233),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(189),
    [anon_sym_RPAREN] = ACTIONS(235),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(491),
    [sym_integer] = ACTIONS(491),
    [sym_float] = ACTIONS(491),
    [sym_char] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(491),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(495),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(299),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(524),
    [anon_sym_DASH] = ACTIONS(524),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(291),
    [anon_sym_fn] = ACTIONS(225),
    [anon_sym_LPAREN2] = ACTIONS(227),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(64)] = {
    [sym__expression] = STATE(3460),
    [sym_block] = STATE(3460),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3460),
    [sym_nil] = STATE(3460),
    [sym__atom] = STATE(3460),
    [sym_quoted_atom] = STATE(3460),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3460),
    [sym_charlist] = STATE(3460),
    [sym_sigil] = STATE(3460),
    [sym_keywords] = STATE(2245),
    [sym_pair] = STATE(3499),
    [sym__keyword] = STATE(885),
    [sym_quoted_keyword] = STATE(885),
    [sym_list] = STATE(3460),
    [sym_tuple] = STATE(3460),
    [sym_bitstring] = STATE(3460),
    [sym_map] = STATE(3460),
    [sym__nullary_operator] = STATE(3460),
    [sym_unary_operator] = STATE(3460),
    [sym_binary_operator] = STATE(3460),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3460),
    [sym_call] = STATE(3460),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1788),
    [sym__call_arguments_without_parentheses] = STATE(2068),
    [sym_do_block] = STATE(2240),
    [sym_access_call] = STATE(3460),
    [sym_anonymous_function] = STATE(3460),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(646),
    [sym_integer] = ACTIONS(646),
    [sym_float] = ACTIONS(646),
    [sym_char] = ACTIONS(646),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(646),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_fn] = ACTIONS(395),
    [anon_sym_LPAREN2] = ACTIONS(397),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(185),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(65)] = {
    [sym__expression] = STATE(3460),
    [sym_block] = STATE(3460),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3460),
    [sym_nil] = STATE(3460),
    [sym__atom] = STATE(3460),
    [sym_quoted_atom] = STATE(3460),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3460),
    [sym_charlist] = STATE(3460),
    [sym_sigil] = STATE(3460),
    [sym_keywords] = STATE(2245),
    [sym_pair] = STATE(3499),
    [sym__keyword] = STATE(885),
    [sym_quoted_keyword] = STATE(885),
    [sym_list] = STATE(3460),
    [sym_tuple] = STATE(3460),
    [sym_bitstring] = STATE(3460),
    [sym_map] = STATE(3460),
    [sym__nullary_operator] = STATE(3460),
    [sym_unary_operator] = STATE(3460),
    [sym_binary_operator] = STATE(3460),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3460),
    [sym_call] = STATE(3460),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1787),
    [sym__call_arguments_without_parentheses] = STATE(2063),
    [sym_do_block] = STATE(2233),
    [sym_access_call] = STATE(3460),
    [sym_anonymous_function] = STATE(3460),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(646),
    [sym_integer] = ACTIONS(646),
    [sym_float] = ACTIONS(646),
    [sym_char] = ACTIONS(646),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(646),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(658),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(403),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(661),
    [anon_sym_DASH] = ACTIONS(661),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(235),
    [anon_sym_fn] = ACTIONS(395),
    [anon_sym_LPAREN2] = ACTIONS(397),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(233),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(66)] = {
    [sym__expression] = STATE(3460),
    [sym_block] = STATE(3460),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3460),
    [sym_nil] = STATE(3460),
    [sym__atom] = STATE(3460),
    [sym_quoted_atom] = STATE(3460),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3460),
    [sym_charlist] = STATE(3460),
    [sym_sigil] = STATE(3460),
    [sym_keywords] = STATE(2245),
    [sym_pair] = STATE(3499),
    [sym__keyword] = STATE(885),
    [sym_quoted_keyword] = STATE(885),
    [sym_list] = STATE(3460),
    [sym_tuple] = STATE(3460),
    [sym_bitstring] = STATE(3460),
    [sym_map] = STATE(3460),
    [sym__nullary_operator] = STATE(3460),
    [sym_unary_operator] = STATE(3460),
    [sym_binary_operator] = STATE(3460),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3460),
    [sym_call] = STATE(3460),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1790),
    [sym__call_arguments_without_parentheses] = STATE(2074),
    [sym_do_block] = STATE(3054),
    [sym_access_call] = STATE(3460),
    [sym_anonymous_function] = STATE(3460),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(646),
    [sym_integer] = ACTIONS(646),
    [sym_float] = ACTIONS(646),
    [sym_char] = ACTIONS(646),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(646),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(187),
    [anon_sym_GT] = ACTIONS(187),
    [anon_sym_PIPE] = ACTIONS(187),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_COMMA] = ACTIONS(187),
    [sym_keyword] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(187),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(187),
    [anon_sym_DASH] = ACTIONS(187),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(187),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(187),
    [anon_sym_when] = ACTIONS(187),
    [anon_sym_COLON_COLON] = ACTIONS(187),
    [anon_sym_EQ_GT] = ACTIONS(187),
    [anon_sym_EQ] = ACTIONS(187),
    [anon_sym_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(187),
    [anon_sym_or] = ACTIONS(187),
    [anon_sym_AMP_AMP] = ACTIONS(187),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(187),
    [anon_sym_and] = ACTIONS(187),
    [anon_sym_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ] = ACTIONS(187),
    [anon_sym_EQ_TILDE] = ACTIONS(187),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(187),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(187),
    [anon_sym_LT_EQ] = ACTIONS(187),
    [anon_sym_GT_EQ] = ACTIONS(187),
    [anon_sym_PIPE_GT] = ACTIONS(187),
    [anon_sym_LT_LT_LT] = ACTIONS(187),
    [anon_sym_GT_GT_GT] = ACTIONS(187),
    [anon_sym_LT_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE] = ACTIONS(187),
    [anon_sym_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_TILDE_GT] = ACTIONS(187),
    [anon_sym_LT_PIPE_GT] = ACTIONS(187),
    [anon_sym_in] = ACTIONS(187),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(187),
    [anon_sym_SLASH_SLASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH] = ACTIONS(187),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(187),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(187),
    [anon_sym_LT_GT] = ACTIONS(187),
    [anon_sym_STAR] = ACTIONS(187),
    [anon_sym_STAR_STAR] = ACTIONS(187),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(187),
    [anon_sym_do] = ACTIONS(415),
    [anon_sym_fn] = ACTIONS(395),
    [anon_sym_LPAREN2] = ACTIONS(397),
    [anon_sym_LBRACK2] = ACTIONS(185),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(417),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(185),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(67)] = {
    [sym__expression] = STATE(1129),
    [sym_block] = STATE(1129),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1129),
    [sym_nil] = STATE(1129),
    [sym__atom] = STATE(1129),
    [sym_quoted_atom] = STATE(1129),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1129),
    [sym_charlist] = STATE(1129),
    [sym_sigil] = STATE(1129),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1129),
    [sym_tuple] = STATE(1129),
    [sym_bitstring] = STATE(1129),
    [sym_map] = STATE(1129),
    [sym__nullary_operator] = STATE(1129),
    [sym_unary_operator] = STATE(1129),
    [sym_binary_operator] = STATE(1129),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1129),
    [sym_call] = STATE(1129),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1129),
    [sym_stab_clause] = STATE(4781),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1129),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(664),
    [sym_integer] = ACTIONS(664),
    [sym_float] = ACTIONS(664),
    [sym_char] = ACTIONS(664),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(664),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(666),
    [anon_sym_catch] = ACTIONS(666),
    [anon_sym_else] = ACTIONS(666),
    [anon_sym_end] = ACTIONS(666),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(666),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(68)] = {
    [sym__expression] = STATE(1120),
    [sym_block] = STATE(1120),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1120),
    [sym_nil] = STATE(1120),
    [sym__atom] = STATE(1120),
    [sym_quoted_atom] = STATE(1120),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1120),
    [sym_charlist] = STATE(1120),
    [sym_sigil] = STATE(1120),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1120),
    [sym_tuple] = STATE(1120),
    [sym_bitstring] = STATE(1120),
    [sym_map] = STATE(1120),
    [sym__nullary_operator] = STATE(1120),
    [sym_unary_operator] = STATE(1120),
    [sym_binary_operator] = STATE(1120),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1120),
    [sym_call] = STATE(1120),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1120),
    [sym_stab_clause] = STATE(4785),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1120),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(668),
    [sym_integer] = ACTIONS(668),
    [sym_float] = ACTIONS(668),
    [sym_char] = ACTIONS(668),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(668),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(670),
    [anon_sym_catch] = ACTIONS(670),
    [anon_sym_else] = ACTIONS(670),
    [anon_sym_end] = ACTIONS(670),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(670),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(69)] = {
    [sym__expression] = STATE(1099),
    [sym_block] = STATE(1099),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1099),
    [sym_nil] = STATE(1099),
    [sym__atom] = STATE(1099),
    [sym_quoted_atom] = STATE(1099),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1099),
    [sym_charlist] = STATE(1099),
    [sym_sigil] = STATE(1099),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1099),
    [sym_tuple] = STATE(1099),
    [sym_bitstring] = STATE(1099),
    [sym_map] = STATE(1099),
    [sym__nullary_operator] = STATE(1099),
    [sym_unary_operator] = STATE(1099),
    [sym_binary_operator] = STATE(1099),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1099),
    [sym_call] = STATE(1099),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1099),
    [sym_stab_clause] = STATE(4777),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1099),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(672),
    [sym_integer] = ACTIONS(672),
    [sym_float] = ACTIONS(672),
    [sym_char] = ACTIONS(672),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(672),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(674),
    [anon_sym_catch] = ACTIONS(674),
    [anon_sym_else] = ACTIONS(674),
    [anon_sym_end] = ACTIONS(674),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(674),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(70)] = {
    [sym__expression] = STATE(1116),
    [sym_block] = STATE(1116),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1116),
    [sym_nil] = STATE(1116),
    [sym__atom] = STATE(1116),
    [sym_quoted_atom] = STATE(1116),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1116),
    [sym_charlist] = STATE(1116),
    [sym_sigil] = STATE(1116),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1116),
    [sym_tuple] = STATE(1116),
    [sym_bitstring] = STATE(1116),
    [sym_map] = STATE(1116),
    [sym__nullary_operator] = STATE(1116),
    [sym_unary_operator] = STATE(1116),
    [sym_binary_operator] = STATE(1116),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1116),
    [sym_call] = STATE(1116),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1116),
    [sym_stab_clause] = STATE(4774),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1116),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(676),
    [sym_integer] = ACTIONS(676),
    [sym_float] = ACTIONS(676),
    [sym_char] = ACTIONS(676),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(676),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_after] = ACTIONS(678),
    [anon_sym_catch] = ACTIONS(678),
    [anon_sym_else] = ACTIONS(678),
    [anon_sym_end] = ACTIONS(678),
    [anon_sym_fn] = ACTIONS(111),
    [anon_sym_rescue] = ACTIONS(678),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(71)] = {
    [sym__expression] = STATE(3200),
    [sym_block] = STATE(3200),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3200),
    [sym_nil] = STATE(3200),
    [sym__atom] = STATE(3200),
    [sym_quoted_atom] = STATE(3200),
    [sym__quoted_i_double] = STATE(3196),
    [sym__quoted_i_single] = STATE(3197),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3200),
    [sym_charlist] = STATE(3200),
    [sym_sigil] = STATE(3200),
    [sym_keywords] = STATE(3516),
    [sym_pair] = STATE(3225),
    [sym__keyword] = STATE(855),
    [sym_quoted_keyword] = STATE(855),
    [sym_list] = STATE(3200),
    [sym_tuple] = STATE(3200),
    [sym_bitstring] = STATE(3200),
    [sym_map] = STATE(3200),
    [sym__nullary_operator] = STATE(3200),
    [sym_unary_operator] = STATE(3200),
    [sym_binary_operator] = STATE(3200),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3200),
    [sym_call] = STATE(3200),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2583),
    [sym__call_arguments_without_parentheses] = STATE(3201),
    [sym_do_block] = STATE(4482),
    [sym_access_call] = STATE(3200),
    [sym_anonymous_function] = STATE(3200),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(544),
    [sym_integer] = ACTIONS(544),
    [sym_float] = ACTIONS(544),
    [sym_char] = ACTIONS(544),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(544),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(564),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_GT_GT] = ACTIONS(235),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(570),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(575),
    [anon_sym_DASH] = ACTIONS(575),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(642),
    [anon_sym_fn] = ACTIONS(582),
    [anon_sym_LPAREN2] = ACTIONS(584),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(72)] = {
    [sym__expression] = STATE(2665),
    [sym_block] = STATE(2665),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2665),
    [sym_nil] = STATE(2665),
    [sym__atom] = STATE(2665),
    [sym_quoted_atom] = STATE(2665),
    [sym__quoted_i_double] = STATE(2509),
    [sym__quoted_i_single] = STATE(2510),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2665),
    [sym_charlist] = STATE(2665),
    [sym_sigil] = STATE(2665),
    [sym_keywords] = STATE(3238),
    [sym_pair] = STATE(2669),
    [sym__keyword] = STATE(868),
    [sym_quoted_keyword] = STATE(868),
    [sym_list] = STATE(2665),
    [sym_tuple] = STATE(2665),
    [sym_bitstring] = STATE(2665),
    [sym_map] = STATE(2665),
    [sym__nullary_operator] = STATE(2665),
    [sym_unary_operator] = STATE(2665),
    [sym_binary_operator] = STATE(2665),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2665),
    [sym_call] = STATE(2665),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym__call_arguments_with_parentheses_immediate] = STATE(2168),
    [sym__call_arguments_without_parentheses] = STATE(2667),
    [sym_do_block] = STATE(4170),
    [sym_access_call] = STATE(2665),
    [sym_anonymous_function] = STATE(2665),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [anon_sym_RPAREN] = ACTIONS(235),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(592),
    [sym_integer] = ACTIONS(592),
    [sym_float] = ACTIONS(592),
    [sym_char] = ACTIONS(592),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(592),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(612),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(636),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(639),
    [anon_sym_DASH] = ACTIONS(639),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(257),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(624),
    [anon_sym_fn] = ACTIONS(626),
    [anon_sym_LPAREN2] = ACTIONS(628),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(73)] = {
    [sym__terminator] = STATE(117),
    [sym__expression] = STATE(2079),
    [sym_block] = STATE(2079),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2079),
    [sym_nil] = STATE(2079),
    [sym__atom] = STATE(2079),
    [sym_quoted_atom] = STATE(2079),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2079),
    [sym_charlist] = STATE(2079),
    [sym_sigil] = STATE(2079),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2079),
    [sym_tuple] = STATE(2079),
    [sym_bitstring] = STATE(2079),
    [sym_map] = STATE(2079),
    [sym__nullary_operator] = STATE(2079),
    [sym_unary_operator] = STATE(2079),
    [sym_binary_operator] = STATE(2079),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2079),
    [sym_call] = STATE(2079),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2079),
    [sym_stab_clause] = STATE(5090),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2079),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(682),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(684),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(688),
    [sym_integer] = ACTIONS(688),
    [sym_float] = ACTIONS(688),
    [sym_char] = ACTIONS(688),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(688),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(74)] = {
    [sym__terminator] = STATE(115),
    [sym__expression] = STATE(2035),
    [sym_block] = STATE(2035),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2035),
    [sym_nil] = STATE(2035),
    [sym__atom] = STATE(2035),
    [sym_quoted_atom] = STATE(2035),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2035),
    [sym_charlist] = STATE(2035),
    [sym_sigil] = STATE(2035),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2035),
    [sym_tuple] = STATE(2035),
    [sym_bitstring] = STATE(2035),
    [sym_map] = STATE(2035),
    [sym__nullary_operator] = STATE(2035),
    [sym_unary_operator] = STATE(2035),
    [sym_binary_operator] = STATE(2035),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2035),
    [sym_call] = STATE(2035),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2035),
    [sym_stab_clause] = STATE(5194),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2035),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(702),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(704),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(706),
    [sym_integer] = ACTIONS(706),
    [sym_float] = ACTIONS(706),
    [sym_char] = ACTIONS(706),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(706),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(75)] = {
    [sym__terminator] = STATE(128),
    [sym__expression] = STATE(1994),
    [sym_block] = STATE(1994),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1994),
    [sym_nil] = STATE(1994),
    [sym__atom] = STATE(1994),
    [sym_quoted_atom] = STATE(1994),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1994),
    [sym_charlist] = STATE(1994),
    [sym_sigil] = STATE(1994),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1994),
    [sym_tuple] = STATE(1994),
    [sym_bitstring] = STATE(1994),
    [sym_map] = STATE(1994),
    [sym__nullary_operator] = STATE(1994),
    [sym_unary_operator] = STATE(1994),
    [sym_binary_operator] = STATE(1994),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1994),
    [sym_call] = STATE(1994),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1994),
    [sym_stab_clause] = STATE(5802),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1994),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(708),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(710),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(712),
    [sym_integer] = ACTIONS(712),
    [sym_float] = ACTIONS(712),
    [sym_char] = ACTIONS(712),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(712),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(76)] = {
    [sym__terminator] = STATE(129),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5650),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(716),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(756),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(77)] = {
    [sym__terminator] = STATE(127),
    [sym__expression] = STATE(2006),
    [sym_block] = STATE(2006),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2006),
    [sym_nil] = STATE(2006),
    [sym__atom] = STATE(2006),
    [sym_quoted_atom] = STATE(2006),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2006),
    [sym_charlist] = STATE(2006),
    [sym_sigil] = STATE(2006),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2006),
    [sym_tuple] = STATE(2006),
    [sym_bitstring] = STATE(2006),
    [sym_map] = STATE(2006),
    [sym__nullary_operator] = STATE(2006),
    [sym_unary_operator] = STATE(2006),
    [sym_binary_operator] = STATE(2006),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2006),
    [sym_call] = STATE(2006),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2006),
    [sym_stab_clause] = STATE(5597),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2006),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(764),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(766),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(768),
    [sym_integer] = ACTIONS(768),
    [sym_float] = ACTIONS(768),
    [sym_char] = ACTIONS(768),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(768),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(78)] = {
    [sym__terminator] = STATE(115),
    [sym__expression] = STATE(1971),
    [sym_block] = STATE(1971),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1971),
    [sym_nil] = STATE(1971),
    [sym__atom] = STATE(1971),
    [sym_quoted_atom] = STATE(1971),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1971),
    [sym_charlist] = STATE(1971),
    [sym_sigil] = STATE(1971),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1971),
    [sym_tuple] = STATE(1971),
    [sym_bitstring] = STATE(1971),
    [sym_map] = STATE(1971),
    [sym__nullary_operator] = STATE(1971),
    [sym_unary_operator] = STATE(1971),
    [sym_binary_operator] = STATE(1971),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1971),
    [sym_call] = STATE(1971),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1971),
    [sym_stab_clause] = STATE(5194),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1971),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(702),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(704),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(770),
    [sym_integer] = ACTIONS(770),
    [sym_float] = ACTIONS(770),
    [sym_char] = ACTIONS(770),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(770),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(79)] = {
    [sym__terminator] = STATE(124),
    [sym__expression] = STATE(2050),
    [sym_block] = STATE(2050),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2050),
    [sym_nil] = STATE(2050),
    [sym__atom] = STATE(2050),
    [sym_quoted_atom] = STATE(2050),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2050),
    [sym_charlist] = STATE(2050),
    [sym_sigil] = STATE(2050),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2050),
    [sym_tuple] = STATE(2050),
    [sym_bitstring] = STATE(2050),
    [sym_map] = STATE(2050),
    [sym__nullary_operator] = STATE(2050),
    [sym_unary_operator] = STATE(2050),
    [sym_binary_operator] = STATE(2050),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2050),
    [sym_call] = STATE(2050),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2050),
    [sym_stab_clause] = STATE(5403),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2050),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(772),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(774),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(776),
    [sym_integer] = ACTIONS(776),
    [sym_float] = ACTIONS(776),
    [sym_char] = ACTIONS(776),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(776),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(80)] = {
    [sym__terminator] = STATE(124),
    [sym__expression] = STATE(2023),
    [sym_block] = STATE(2023),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2023),
    [sym_nil] = STATE(2023),
    [sym__atom] = STATE(2023),
    [sym_quoted_atom] = STATE(2023),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2023),
    [sym_charlist] = STATE(2023),
    [sym_sigil] = STATE(2023),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2023),
    [sym_tuple] = STATE(2023),
    [sym_bitstring] = STATE(2023),
    [sym_map] = STATE(2023),
    [sym__nullary_operator] = STATE(2023),
    [sym_unary_operator] = STATE(2023),
    [sym_binary_operator] = STATE(2023),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2023),
    [sym_call] = STATE(2023),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2023),
    [sym_stab_clause] = STATE(5403),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2023),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(772),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(774),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(778),
    [sym_integer] = ACTIONS(778),
    [sym_float] = ACTIONS(778),
    [sym_char] = ACTIONS(778),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(778),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(81)] = {
    [sym__terminator] = STATE(116),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5212),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(780),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(782),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(82)] = {
    [sym__terminator] = STATE(128),
    [sym__expression] = STATE(1785),
    [sym_block] = STATE(1785),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1785),
    [sym_nil] = STATE(1785),
    [sym__atom] = STATE(1785),
    [sym_quoted_atom] = STATE(1785),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1785),
    [sym_charlist] = STATE(1785),
    [sym_sigil] = STATE(1785),
    [sym_keywords] = STATE(6710),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1785),
    [sym_tuple] = STATE(1785),
    [sym_bitstring] = STATE(1785),
    [sym_map] = STATE(1785),
    [sym__nullary_operator] = STATE(1785),
    [sym_unary_operator] = STATE(1785),
    [sym_binary_operator] = STATE(1785),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1785),
    [sym_call] = STATE(1785),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1785),
    [sym_stab_clause] = STATE(5802),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1785),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(708),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(784),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(786),
    [sym_integer] = ACTIONS(786),
    [sym_float] = ACTIONS(786),
    [sym_char] = ACTIONS(786),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(786),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(83)] = {
    [sym__terminator] = STATE(120),
    [sym__expression] = STATE(1905),
    [sym_block] = STATE(1905),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1905),
    [sym_nil] = STATE(1905),
    [sym__atom] = STATE(1905),
    [sym_quoted_atom] = STATE(1905),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1905),
    [sym_charlist] = STATE(1905),
    [sym_sigil] = STATE(1905),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1905),
    [sym_tuple] = STATE(1905),
    [sym_bitstring] = STATE(1905),
    [sym_map] = STATE(1905),
    [sym__nullary_operator] = STATE(1905),
    [sym_unary_operator] = STATE(1905),
    [sym_binary_operator] = STATE(1905),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1905),
    [sym_call] = STATE(1905),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1905),
    [sym_stab_clause] = STATE(5281),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1905),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(788),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(790),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(792),
    [sym_integer] = ACTIONS(792),
    [sym_float] = ACTIONS(792),
    [sym_char] = ACTIONS(792),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(792),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(84)] = {
    [sym__terminator] = STATE(133),
    [sym__expression] = STATE(2067),
    [sym_block] = STATE(2067),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2067),
    [sym_nil] = STATE(2067),
    [sym__atom] = STATE(2067),
    [sym_quoted_atom] = STATE(2067),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2067),
    [sym_charlist] = STATE(2067),
    [sym_sigil] = STATE(2067),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2067),
    [sym_tuple] = STATE(2067),
    [sym_bitstring] = STATE(2067),
    [sym_map] = STATE(2067),
    [sym__nullary_operator] = STATE(2067),
    [sym_unary_operator] = STATE(2067),
    [sym_binary_operator] = STATE(2067),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2067),
    [sym_call] = STATE(2067),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2067),
    [sym_stab_clause] = STATE(5688),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2067),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(794),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(796),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(798),
    [sym_integer] = ACTIONS(798),
    [sym_float] = ACTIONS(798),
    [sym_char] = ACTIONS(798),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(798),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(85)] = {
    [sym__terminator] = STATE(133),
    [sym__expression] = STATE(2047),
    [sym_block] = STATE(2047),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2047),
    [sym_nil] = STATE(2047),
    [sym__atom] = STATE(2047),
    [sym_quoted_atom] = STATE(2047),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2047),
    [sym_charlist] = STATE(2047),
    [sym_sigil] = STATE(2047),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2047),
    [sym_tuple] = STATE(2047),
    [sym_bitstring] = STATE(2047),
    [sym_map] = STATE(2047),
    [sym__nullary_operator] = STATE(2047),
    [sym_unary_operator] = STATE(2047),
    [sym_binary_operator] = STATE(2047),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2047),
    [sym_call] = STATE(2047),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2047),
    [sym_stab_clause] = STATE(5688),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2047),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(794),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(796),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(800),
    [sym_integer] = ACTIONS(800),
    [sym_float] = ACTIONS(800),
    [sym_char] = ACTIONS(800),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(800),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(86)] = {
    [sym__terminator] = STATE(126),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5446),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(802),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(804),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(87)] = {
    [sym__terminator] = STATE(128),
    [sym__expression] = STATE(1913),
    [sym_block] = STATE(1913),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1913),
    [sym_nil] = STATE(1913),
    [sym__atom] = STATE(1913),
    [sym_quoted_atom] = STATE(1913),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1913),
    [sym_charlist] = STATE(1913),
    [sym_sigil] = STATE(1913),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1913),
    [sym_tuple] = STATE(1913),
    [sym_bitstring] = STATE(1913),
    [sym_map] = STATE(1913),
    [sym__nullary_operator] = STATE(1913),
    [sym_unary_operator] = STATE(1913),
    [sym_binary_operator] = STATE(1913),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1913),
    [sym_call] = STATE(1913),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1913),
    [sym_stab_clause] = STATE(5802),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1913),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(708),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(710),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(806),
    [sym_integer] = ACTIONS(806),
    [sym_float] = ACTIONS(806),
    [sym_char] = ACTIONS(806),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(806),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(88)] = {
    [sym__expression] = STATE(3460),
    [sym_block] = STATE(3460),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3460),
    [sym_nil] = STATE(3460),
    [sym__atom] = STATE(3460),
    [sym_quoted_atom] = STATE(3460),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3460),
    [sym_charlist] = STATE(3460),
    [sym_sigil] = STATE(3460),
    [sym_keywords] = STATE(2245),
    [sym_pair] = STATE(3499),
    [sym__keyword] = STATE(885),
    [sym_quoted_keyword] = STATE(885),
    [sym_list] = STATE(3460),
    [sym_tuple] = STATE(3460),
    [sym_bitstring] = STATE(3460),
    [sym_map] = STATE(3460),
    [sym__nullary_operator] = STATE(3460),
    [sym_unary_operator] = STATE(3460),
    [sym_binary_operator] = STATE(3460),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3460),
    [sym_call] = STATE(3460),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1789),
    [sym__call_arguments_without_parentheses] = STATE(2071),
    [sym_do_block] = STATE(3046),
    [sym_access_call] = STATE(3460),
    [sym_anonymous_function] = STATE(3460),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(646),
    [sym_integer] = ACTIONS(646),
    [sym_float] = ACTIONS(646),
    [sym_char] = ACTIONS(646),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(646),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(658),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_COMMA] = ACTIONS(235),
    [sym_keyword] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(403),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(661),
    [anon_sym_DASH] = ACTIONS(661),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(415),
    [anon_sym_fn] = ACTIONS(395),
    [anon_sym_LPAREN2] = ACTIONS(397),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(89)] = {
    [sym__terminator] = STATE(136),
    [sym__expression] = STATE(1980),
    [sym_block] = STATE(1980),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1980),
    [sym_nil] = STATE(1980),
    [sym__atom] = STATE(1980),
    [sym_quoted_atom] = STATE(1980),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1980),
    [sym_charlist] = STATE(1980),
    [sym_sigil] = STATE(1980),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1980),
    [sym_tuple] = STATE(1980),
    [sym_bitstring] = STATE(1980),
    [sym_map] = STATE(1980),
    [sym__nullary_operator] = STATE(1980),
    [sym_unary_operator] = STATE(1980),
    [sym_binary_operator] = STATE(1980),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1980),
    [sym_call] = STATE(1980),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1980),
    [sym_stab_clause] = STATE(5168),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1980),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(808),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(810),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(812),
    [sym_integer] = ACTIONS(812),
    [sym_float] = ACTIONS(812),
    [sym_char] = ACTIONS(812),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(812),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(90)] = {
    [sym__terminator] = STATE(134),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5723),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(814),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(816),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(91)] = {
    [sym__terminator] = STATE(123),
    [sym__expression] = STATE(2080),
    [sym_block] = STATE(2080),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2080),
    [sym_nil] = STATE(2080),
    [sym__atom] = STATE(2080),
    [sym_quoted_atom] = STATE(2080),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2080),
    [sym_charlist] = STATE(2080),
    [sym_sigil] = STATE(2080),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2080),
    [sym_tuple] = STATE(2080),
    [sym_bitstring] = STATE(2080),
    [sym_map] = STATE(2080),
    [sym__nullary_operator] = STATE(2080),
    [sym_unary_operator] = STATE(2080),
    [sym_binary_operator] = STATE(2080),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2080),
    [sym_call] = STATE(2080),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2080),
    [sym_stab_clause] = STATE(5738),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2080),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(818),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(820),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(822),
    [sym_integer] = ACTIONS(822),
    [sym_float] = ACTIONS(822),
    [sym_char] = ACTIONS(822),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(822),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(92)] = {
    [sym__terminator] = STATE(119),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5089),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(824),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(826),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(93)] = {
    [sym__terminator] = STATE(122),
    [sym__expression] = STATE(2085),
    [sym_block] = STATE(2085),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2085),
    [sym_nil] = STATE(2085),
    [sym__atom] = STATE(2085),
    [sym_quoted_atom] = STATE(2085),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2085),
    [sym_charlist] = STATE(2085),
    [sym_sigil] = STATE(2085),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2085),
    [sym_tuple] = STATE(2085),
    [sym_bitstring] = STATE(2085),
    [sym_map] = STATE(2085),
    [sym__nullary_operator] = STATE(2085),
    [sym_unary_operator] = STATE(2085),
    [sym_binary_operator] = STATE(2085),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2085),
    [sym_call] = STATE(2085),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2085),
    [sym_stab_clause] = STATE(5176),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2085),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(828),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(830),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(832),
    [sym_integer] = ACTIONS(832),
    [sym_float] = ACTIONS(832),
    [sym_char] = ACTIONS(832),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(832),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(94)] = {
    [sym__terminator] = STATE(117),
    [sym__expression] = STATE(2077),
    [sym_block] = STATE(2077),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2077),
    [sym_nil] = STATE(2077),
    [sym__atom] = STATE(2077),
    [sym_quoted_atom] = STATE(2077),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2077),
    [sym_charlist] = STATE(2077),
    [sym_sigil] = STATE(2077),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2077),
    [sym_tuple] = STATE(2077),
    [sym_bitstring] = STATE(2077),
    [sym_map] = STATE(2077),
    [sym__nullary_operator] = STATE(2077),
    [sym_unary_operator] = STATE(2077),
    [sym_binary_operator] = STATE(2077),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2077),
    [sym_call] = STATE(2077),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2077),
    [sym_stab_clause] = STATE(5090),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2077),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(682),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(684),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(834),
    [sym_integer] = ACTIONS(834),
    [sym_float] = ACTIONS(834),
    [sym_char] = ACTIONS(834),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(834),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(95)] = {
    [sym__terminator] = STATE(118),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5102),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(836),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(838),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(96)] = {
    [sym__terminator] = STATE(120),
    [sym__expression] = STATE(1955),
    [sym_block] = STATE(1955),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1955),
    [sym_nil] = STATE(1955),
    [sym__atom] = STATE(1955),
    [sym_quoted_atom] = STATE(1955),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1955),
    [sym_charlist] = STATE(1955),
    [sym_sigil] = STATE(1955),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1955),
    [sym_tuple] = STATE(1955),
    [sym_bitstring] = STATE(1955),
    [sym_map] = STATE(1955),
    [sym__nullary_operator] = STATE(1955),
    [sym_unary_operator] = STATE(1955),
    [sym_binary_operator] = STATE(1955),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1955),
    [sym_call] = STATE(1955),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1955),
    [sym_stab_clause] = STATE(5281),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1955),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(788),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(790),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(840),
    [sym_integer] = ACTIONS(840),
    [sym_float] = ACTIONS(840),
    [sym_char] = ACTIONS(840),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(840),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(97)] = {
    [sym__terminator] = STATE(114),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5210),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(842),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(844),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(98)] = {
    [sym__terminator] = STATE(130),
    [sym__expression] = STATE(2101),
    [sym_block] = STATE(2101),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2101),
    [sym_nil] = STATE(2101),
    [sym__atom] = STATE(2101),
    [sym_quoted_atom] = STATE(2101),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2101),
    [sym_charlist] = STATE(2101),
    [sym_sigil] = STATE(2101),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2101),
    [sym_tuple] = STATE(2101),
    [sym_bitstring] = STATE(2101),
    [sym_map] = STATE(2101),
    [sym__nullary_operator] = STATE(2101),
    [sym_unary_operator] = STATE(2101),
    [sym_binary_operator] = STATE(2101),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2101),
    [sym_call] = STATE(2101),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2101),
    [sym_stab_clause] = STATE(5252),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2101),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(846),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(848),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(850),
    [sym_integer] = ACTIONS(850),
    [sym_float] = ACTIONS(850),
    [sym_char] = ACTIONS(850),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(850),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(99)] = {
    [sym__terminator] = STATE(122),
    [sym__expression] = STATE(2082),
    [sym_block] = STATE(2082),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2082),
    [sym_nil] = STATE(2082),
    [sym__atom] = STATE(2082),
    [sym_quoted_atom] = STATE(2082),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2082),
    [sym_charlist] = STATE(2082),
    [sym_sigil] = STATE(2082),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2082),
    [sym_tuple] = STATE(2082),
    [sym_bitstring] = STATE(2082),
    [sym_map] = STATE(2082),
    [sym__nullary_operator] = STATE(2082),
    [sym_unary_operator] = STATE(2082),
    [sym_binary_operator] = STATE(2082),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2082),
    [sym_call] = STATE(2082),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2082),
    [sym_stab_clause] = STATE(5176),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2082),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(828),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(830),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(852),
    [sym_integer] = ACTIONS(852),
    [sym_float] = ACTIONS(852),
    [sym_char] = ACTIONS(852),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(852),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(100)] = {
    [sym__terminator] = STATE(125),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5187),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(854),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(856),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(101)] = {
    [sym__terminator] = STATE(127),
    [sym__expression] = STATE(1640),
    [sym_block] = STATE(1640),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1640),
    [sym_nil] = STATE(1640),
    [sym__atom] = STATE(1640),
    [sym_quoted_atom] = STATE(1640),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1640),
    [sym_charlist] = STATE(1640),
    [sym_sigil] = STATE(1640),
    [sym_keywords] = STATE(6710),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1640),
    [sym_tuple] = STATE(1640),
    [sym_bitstring] = STATE(1640),
    [sym_map] = STATE(1640),
    [sym__nullary_operator] = STATE(1640),
    [sym_unary_operator] = STATE(1640),
    [sym_binary_operator] = STATE(1640),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1640),
    [sym_call] = STATE(1640),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1640),
    [sym_stab_clause] = STATE(5597),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1640),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(764),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(858),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(860),
    [sym_integer] = ACTIONS(860),
    [sym_float] = ACTIONS(860),
    [sym_char] = ACTIONS(860),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(860),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(102)] = {
    [sym__terminator] = STATE(121),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5948),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(862),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(103)] = {
    [sym__terminator] = STATE(135),
    [sym__expression] = STATE(2111),
    [sym_block] = STATE(2111),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2111),
    [sym_nil] = STATE(2111),
    [sym__atom] = STATE(2111),
    [sym_quoted_atom] = STATE(2111),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2111),
    [sym_charlist] = STATE(2111),
    [sym_sigil] = STATE(2111),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2111),
    [sym_tuple] = STATE(2111),
    [sym_bitstring] = STATE(2111),
    [sym_map] = STATE(2111),
    [sym__nullary_operator] = STATE(2111),
    [sym_unary_operator] = STATE(2111),
    [sym_binary_operator] = STATE(2111),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2111),
    [sym_call] = STATE(2111),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2111),
    [sym_stab_clause] = STATE(5332),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2111),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(866),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(868),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(870),
    [sym_integer] = ACTIONS(870),
    [sym_float] = ACTIONS(870),
    [sym_char] = ACTIONS(870),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(870),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(104)] = {
    [sym__terminator] = STATE(130),
    [sym__expression] = STATE(2097),
    [sym_block] = STATE(2097),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2097),
    [sym_nil] = STATE(2097),
    [sym__atom] = STATE(2097),
    [sym_quoted_atom] = STATE(2097),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2097),
    [sym_charlist] = STATE(2097),
    [sym_sigil] = STATE(2097),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2097),
    [sym_tuple] = STATE(2097),
    [sym_bitstring] = STATE(2097),
    [sym_map] = STATE(2097),
    [sym__nullary_operator] = STATE(2097),
    [sym_unary_operator] = STATE(2097),
    [sym_binary_operator] = STATE(2097),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2097),
    [sym_call] = STATE(2097),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2097),
    [sym_stab_clause] = STATE(5252),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2097),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(846),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(848),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(872),
    [sym_integer] = ACTIONS(872),
    [sym_float] = ACTIONS(872),
    [sym_char] = ACTIONS(872),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(872),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(105)] = {
    [sym__terminator] = STATE(113),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5265),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(874),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(876),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(106)] = {
    [sym__terminator] = STATE(132),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5104),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(878),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(880),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(107)] = {
    [sym__terminator] = STATE(123),
    [sym__expression] = STATE(2018),
    [sym_block] = STATE(2018),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2018),
    [sym_nil] = STATE(2018),
    [sym__atom] = STATE(2018),
    [sym_quoted_atom] = STATE(2018),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2018),
    [sym_charlist] = STATE(2018),
    [sym_sigil] = STATE(2018),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2018),
    [sym_tuple] = STATE(2018),
    [sym_bitstring] = STATE(2018),
    [sym_map] = STATE(2018),
    [sym__nullary_operator] = STATE(2018),
    [sym_unary_operator] = STATE(2018),
    [sym_binary_operator] = STATE(2018),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2018),
    [sym_call] = STATE(2018),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2018),
    [sym_stab_clause] = STATE(5738),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2018),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(818),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(820),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(882),
    [sym_integer] = ACTIONS(882),
    [sym_float] = ACTIONS(882),
    [sym_char] = ACTIONS(882),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(882),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(108)] = {
    [sym__terminator] = STATE(135),
    [sym__expression] = STATE(2105),
    [sym_block] = STATE(2105),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2105),
    [sym_nil] = STATE(2105),
    [sym__atom] = STATE(2105),
    [sym_quoted_atom] = STATE(2105),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2105),
    [sym_charlist] = STATE(2105),
    [sym_sigil] = STATE(2105),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2105),
    [sym_tuple] = STATE(2105),
    [sym_bitstring] = STATE(2105),
    [sym_map] = STATE(2105),
    [sym__nullary_operator] = STATE(2105),
    [sym_unary_operator] = STATE(2105),
    [sym_binary_operator] = STATE(2105),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2105),
    [sym_call] = STATE(2105),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2105),
    [sym_stab_clause] = STATE(5332),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2105),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(866),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(868),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(884),
    [sym_integer] = ACTIONS(884),
    [sym_float] = ACTIONS(884),
    [sym_char] = ACTIONS(884),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(884),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(109)] = {
    [sym__terminator] = STATE(131),
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5343),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_repeat1] = STATE(1027),
    [aux_sym__terminator_token1] = ACTIONS(714),
    [anon_sym_SEMI] = ACTIONS(886),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(888),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(110)] = {
    [sym__terminator] = STATE(136),
    [sym__expression] = STATE(1950),
    [sym_block] = STATE(1950),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1950),
    [sym_nil] = STATE(1950),
    [sym__atom] = STATE(1950),
    [sym_quoted_atom] = STATE(1950),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1950),
    [sym_charlist] = STATE(1950),
    [sym_sigil] = STATE(1950),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1950),
    [sym_tuple] = STATE(1950),
    [sym_bitstring] = STATE(1950),
    [sym_map] = STATE(1950),
    [sym__nullary_operator] = STATE(1950),
    [sym_unary_operator] = STATE(1950),
    [sym_binary_operator] = STATE(1950),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1950),
    [sym_call] = STATE(1950),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1950),
    [sym_stab_clause] = STATE(5168),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1950),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(808),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(810),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(890),
    [sym_integer] = ACTIONS(890),
    [sym_float] = ACTIONS(890),
    [sym_char] = ACTIONS(890),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(890),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(111)] = {
    [sym__terminator] = STATE(127),
    [sym__expression] = STATE(2059),
    [sym_block] = STATE(2059),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2059),
    [sym_nil] = STATE(2059),
    [sym__atom] = STATE(2059),
    [sym_quoted_atom] = STATE(2059),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2059),
    [sym_charlist] = STATE(2059),
    [sym_sigil] = STATE(2059),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2059),
    [sym_tuple] = STATE(2059),
    [sym_bitstring] = STATE(2059),
    [sym_map] = STATE(2059),
    [sym__nullary_operator] = STATE(2059),
    [sym_unary_operator] = STATE(2059),
    [sym_binary_operator] = STATE(2059),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2059),
    [sym_call] = STATE(2059),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2059),
    [sym_stab_clause] = STATE(5597),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2059),
    [aux_sym__terminator_repeat1] = STATE(1029),
    [aux_sym__terminator_token1] = ACTIONS(680),
    [anon_sym_SEMI] = ACTIONS(764),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(766),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(892),
    [sym_integer] = ACTIONS(892),
    [sym_float] = ACTIONS(892),
    [sym_char] = ACTIONS(892),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(892),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(112)] = {
    [sym__expression] = STATE(2244),
    [sym_block] = STATE(2244),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2244),
    [sym_nil] = STATE(2244),
    [sym__atom] = STATE(2244),
    [sym_quoted_atom] = STATE(2244),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2244),
    [sym_charlist] = STATE(2244),
    [sym_sigil] = STATE(2244),
    [sym_keywords] = STATE(2245),
    [sym_pair] = STATE(2158),
    [sym__keyword] = STATE(848),
    [sym_quoted_keyword] = STATE(848),
    [sym_list] = STATE(2244),
    [sym_tuple] = STATE(2244),
    [sym_bitstring] = STATE(2244),
    [sym_map] = STATE(2244),
    [sym__nullary_operator] = STATE(2244),
    [sym_unary_operator] = STATE(2244),
    [sym_binary_operator] = STATE(2244),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2244),
    [sym_call] = STATE(2244),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym__call_arguments_with_parentheses_immediate] = STATE(1789),
    [sym__call_arguments_without_parentheses] = STATE(2071),
    [sym_do_block] = STATE(3046),
    [sym_access_call] = STATE(2244),
    [sym_anonymous_function] = STATE(2244),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(363),
    [sym_integer] = ACTIONS(363),
    [sym_float] = ACTIONS(363),
    [sym_char] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(894),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(257),
    [anon_sym_GT] = ACTIONS(257),
    [anon_sym_PIPE] = ACTIONS(257),
    [anon_sym_SLASH] = ACTIONS(257),
    [anon_sym_TILDE] = ACTIONS(381),
    [sym_keyword] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(403),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(406),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(257),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(257),
    [anon_sym_when] = ACTIONS(257),
    [anon_sym_COLON_COLON] = ACTIONS(257),
    [anon_sym_EQ_GT] = ACTIONS(235),
    [anon_sym_EQ] = ACTIONS(257),
    [anon_sym_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(257),
    [anon_sym_or] = ACTIONS(257),
    [anon_sym_AMP_AMP] = ACTIONS(257),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(257),
    [anon_sym_and] = ACTIONS(257),
    [anon_sym_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ] = ACTIONS(257),
    [anon_sym_EQ_TILDE] = ACTIONS(257),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(257),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(257),
    [anon_sym_LT_EQ] = ACTIONS(257),
    [anon_sym_GT_EQ] = ACTIONS(257),
    [anon_sym_PIPE_GT] = ACTIONS(257),
    [anon_sym_LT_LT_LT] = ACTIONS(257),
    [anon_sym_GT_GT_GT] = ACTIONS(257),
    [anon_sym_LT_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE] = ACTIONS(257),
    [anon_sym_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_TILDE_GT] = ACTIONS(257),
    [anon_sym_LT_PIPE_GT] = ACTIONS(257),
    [anon_sym_in] = ACTIONS(257),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(257),
    [anon_sym_SLASH_SLASH] = ACTIONS(235),
    [anon_sym_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH] = ACTIONS(257),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(257),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(257),
    [anon_sym_LT_GT] = ACTIONS(257),
    [anon_sym_STAR] = ACTIONS(257),
    [anon_sym_STAR_STAR] = ACTIONS(257),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_DOT] = ACTIONS(235),
    [anon_sym_do] = ACTIONS(415),
    [anon_sym_fn] = ACTIONS(395),
    [anon_sym_LPAREN2] = ACTIONS(397),
    [anon_sym_LBRACK2] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(286),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(113)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5287),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(896),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(114)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5292),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(898),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(115)] = {
    [sym__expression] = STATE(2032),
    [sym_block] = STATE(2032),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2032),
    [sym_nil] = STATE(2032),
    [sym__atom] = STATE(2032),
    [sym_quoted_atom] = STATE(2032),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2032),
    [sym_charlist] = STATE(2032),
    [sym_sigil] = STATE(2032),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2032),
    [sym_tuple] = STATE(2032),
    [sym_bitstring] = STATE(2032),
    [sym_map] = STATE(2032),
    [sym__nullary_operator] = STATE(2032),
    [sym_unary_operator] = STATE(2032),
    [sym_binary_operator] = STATE(2032),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2032),
    [sym_call] = STATE(2032),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2032),
    [sym_stab_clause] = STATE(5228),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2032),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(900),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(902),
    [sym_integer] = ACTIONS(902),
    [sym_float] = ACTIONS(902),
    [sym_char] = ACTIONS(902),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(902),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(116)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5279),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(904),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(117)] = {
    [sym__expression] = STATE(2078),
    [sym_block] = STATE(2078),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2078),
    [sym_nil] = STATE(2078),
    [sym__atom] = STATE(2078),
    [sym_quoted_atom] = STATE(2078),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2078),
    [sym_charlist] = STATE(2078),
    [sym_sigil] = STATE(2078),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2078),
    [sym_tuple] = STATE(2078),
    [sym_bitstring] = STATE(2078),
    [sym_map] = STATE(2078),
    [sym__nullary_operator] = STATE(2078),
    [sym_unary_operator] = STATE(2078),
    [sym_binary_operator] = STATE(2078),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2078),
    [sym_call] = STATE(2078),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2078),
    [sym_stab_clause] = STATE(5106),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2078),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(906),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(908),
    [sym_integer] = ACTIONS(908),
    [sym_float] = ACTIONS(908),
    [sym_char] = ACTIONS(908),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(908),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(118)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5124),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(910),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(119)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5189),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(912),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(120)] = {
    [sym__expression] = STATE(1954),
    [sym_block] = STATE(1954),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1954),
    [sym_nil] = STATE(1954),
    [sym__atom] = STATE(1954),
    [sym_quoted_atom] = STATE(1954),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1954),
    [sym_charlist] = STATE(1954),
    [sym_sigil] = STATE(1954),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1954),
    [sym_tuple] = STATE(1954),
    [sym_bitstring] = STATE(1954),
    [sym_map] = STATE(1954),
    [sym__nullary_operator] = STATE(1954),
    [sym_unary_operator] = STATE(1954),
    [sym_binary_operator] = STATE(1954),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1954),
    [sym_call] = STATE(1954),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1954),
    [sym_stab_clause] = STATE(5128),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1954),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(914),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(916),
    [sym_integer] = ACTIONS(916),
    [sym_float] = ACTIONS(916),
    [sym_char] = ACTIONS(916),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(916),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(121)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5284),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(918),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(122)] = {
    [sym__expression] = STATE(2084),
    [sym_block] = STATE(2084),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2084),
    [sym_nil] = STATE(2084),
    [sym__atom] = STATE(2084),
    [sym_quoted_atom] = STATE(2084),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2084),
    [sym_charlist] = STATE(2084),
    [sym_sigil] = STATE(2084),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2084),
    [sym_tuple] = STATE(2084),
    [sym_bitstring] = STATE(2084),
    [sym_map] = STATE(2084),
    [sym__nullary_operator] = STATE(2084),
    [sym_unary_operator] = STATE(2084),
    [sym_binary_operator] = STATE(2084),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2084),
    [sym_call] = STATE(2084),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2084),
    [sym_stab_clause] = STATE(5188),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2084),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(920),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(922),
    [sym_integer] = ACTIONS(922),
    [sym_float] = ACTIONS(922),
    [sym_char] = ACTIONS(922),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(922),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(123)] = {
    [sym__expression] = STATE(2017),
    [sym_block] = STATE(2017),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2017),
    [sym_nil] = STATE(2017),
    [sym__atom] = STATE(2017),
    [sym_quoted_atom] = STATE(2017),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2017),
    [sym_charlist] = STATE(2017),
    [sym_sigil] = STATE(2017),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2017),
    [sym_tuple] = STATE(2017),
    [sym_bitstring] = STATE(2017),
    [sym_map] = STATE(2017),
    [sym__nullary_operator] = STATE(2017),
    [sym_unary_operator] = STATE(2017),
    [sym_binary_operator] = STATE(2017),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2017),
    [sym_call] = STATE(2017),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2017),
    [sym_stab_clause] = STATE(5939),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2017),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(924),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(926),
    [sym_integer] = ACTIONS(926),
    [sym_float] = ACTIONS(926),
    [sym_char] = ACTIONS(926),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(926),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(124)] = {
    [sym__expression] = STATE(2049),
    [sym_block] = STATE(2049),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2049),
    [sym_nil] = STATE(2049),
    [sym__atom] = STATE(2049),
    [sym_quoted_atom] = STATE(2049),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2049),
    [sym_charlist] = STATE(2049),
    [sym_sigil] = STATE(2049),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2049),
    [sym_tuple] = STATE(2049),
    [sym_bitstring] = STATE(2049),
    [sym_map] = STATE(2049),
    [sym__nullary_operator] = STATE(2049),
    [sym_unary_operator] = STATE(2049),
    [sym_binary_operator] = STATE(2049),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2049),
    [sym_call] = STATE(2049),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2049),
    [sym_stab_clause] = STATE(5458),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2049),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(928),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(930),
    [sym_integer] = ACTIONS(930),
    [sym_float] = ACTIONS(930),
    [sym_char] = ACTIONS(930),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(930),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(125)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5206),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(932),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(126)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5512),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(934),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(127)] = {
    [sym__expression] = STATE(2003),
    [sym_block] = STATE(2003),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2003),
    [sym_nil] = STATE(2003),
    [sym__atom] = STATE(2003),
    [sym_quoted_atom] = STATE(2003),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2003),
    [sym_charlist] = STATE(2003),
    [sym_sigil] = STATE(2003),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2003),
    [sym_tuple] = STATE(2003),
    [sym_bitstring] = STATE(2003),
    [sym_map] = STATE(2003),
    [sym__nullary_operator] = STATE(2003),
    [sym_unary_operator] = STATE(2003),
    [sym_binary_operator] = STATE(2003),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2003),
    [sym_call] = STATE(2003),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2003),
    [sym_stab_clause] = STATE(5721),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2003),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(936),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(938),
    [sym_integer] = ACTIONS(938),
    [sym_float] = ACTIONS(938),
    [sym_char] = ACTIONS(938),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(938),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(128)] = {
    [sym__expression] = STATE(1911),
    [sym_block] = STATE(1911),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1911),
    [sym_nil] = STATE(1911),
    [sym__atom] = STATE(1911),
    [sym_quoted_atom] = STATE(1911),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1911),
    [sym_charlist] = STATE(1911),
    [sym_sigil] = STATE(1911),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1911),
    [sym_tuple] = STATE(1911),
    [sym_bitstring] = STATE(1911),
    [sym_map] = STATE(1911),
    [sym__nullary_operator] = STATE(1911),
    [sym_unary_operator] = STATE(1911),
    [sym_binary_operator] = STATE(1911),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1911),
    [sym_call] = STATE(1911),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1911),
    [sym_stab_clause] = STATE(5105),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1911),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(940),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(942),
    [sym_integer] = ACTIONS(942),
    [sym_float] = ACTIONS(942),
    [sym_char] = ACTIONS(942),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(942),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(129)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5936),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(944),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(130)] = {
    [sym__expression] = STATE(2100),
    [sym_block] = STATE(2100),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2100),
    [sym_nil] = STATE(2100),
    [sym__atom] = STATE(2100),
    [sym_quoted_atom] = STATE(2100),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2100),
    [sym_charlist] = STATE(2100),
    [sym_sigil] = STATE(2100),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2100),
    [sym_tuple] = STATE(2100),
    [sym_bitstring] = STATE(2100),
    [sym_map] = STATE(2100),
    [sym__nullary_operator] = STATE(2100),
    [sym_unary_operator] = STATE(2100),
    [sym_binary_operator] = STATE(2100),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2100),
    [sym_call] = STATE(2100),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2100),
    [sym_stab_clause] = STATE(5270),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2100),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(946),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(948),
    [sym_integer] = ACTIONS(948),
    [sym_float] = ACTIONS(948),
    [sym_char] = ACTIONS(948),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(948),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(131)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5358),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(950),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(132)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5393),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(952),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(133)] = {
    [sym__expression] = STATE(2064),
    [sym_block] = STATE(2064),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2064),
    [sym_nil] = STATE(2064),
    [sym__atom] = STATE(2064),
    [sym_quoted_atom] = STATE(2064),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2064),
    [sym_charlist] = STATE(2064),
    [sym_sigil] = STATE(2064),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2064),
    [sym_tuple] = STATE(2064),
    [sym_bitstring] = STATE(2064),
    [sym_map] = STATE(2064),
    [sym__nullary_operator] = STATE(2064),
    [sym_unary_operator] = STATE(2064),
    [sym_binary_operator] = STATE(2064),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2064),
    [sym_call] = STATE(2064),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2064),
    [sym_stab_clause] = STATE(5739),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2064),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(954),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(956),
    [sym_integer] = ACTIONS(956),
    [sym_float] = ACTIONS(956),
    [sym_char] = ACTIONS(956),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(956),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(134)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(5803),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_end] = ACTIONS(958),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(135)] = {
    [sym__expression] = STATE(2110),
    [sym_block] = STATE(2110),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(2110),
    [sym_nil] = STATE(2110),
    [sym__atom] = STATE(2110),
    [sym_quoted_atom] = STATE(2110),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2110),
    [sym_charlist] = STATE(2110),
    [sym_sigil] = STATE(2110),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(2110),
    [sym_tuple] = STATE(2110),
    [sym_bitstring] = STATE(2110),
    [sym_map] = STATE(2110),
    [sym__nullary_operator] = STATE(2110),
    [sym_unary_operator] = STATE(2110),
    [sym_binary_operator] = STATE(2110),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2110),
    [sym_call] = STATE(2110),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2110),
    [sym_stab_clause] = STATE(5344),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(2110),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(960),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(962),
    [sym_integer] = ACTIONS(962),
    [sym_float] = ACTIONS(962),
    [sym_char] = ACTIONS(962),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(962),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(136)] = {
    [sym__expression] = STATE(1977),
    [sym_block] = STATE(1977),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1977),
    [sym_nil] = STATE(1977),
    [sym__atom] = STATE(1977),
    [sym_quoted_atom] = STATE(1977),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1977),
    [sym_charlist] = STATE(1977),
    [sym_sigil] = STATE(1977),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(1977),
    [sym_tuple] = STATE(1977),
    [sym_bitstring] = STATE(1977),
    [sym_map] = STATE(1977),
    [sym__nullary_operator] = STATE(1977),
    [sym_unary_operator] = STATE(1977),
    [sym_binary_operator] = STATE(1977),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1977),
    [sym_call] = STATE(1977),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1977),
    [sym_stab_clause] = STATE(5211),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(1977),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(63),
    [anon_sym_RPAREN] = ACTIONS(964),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(966),
    [sym_integer] = ACTIONS(966),
    [sym_float] = ACTIONS(966),
    [sym_char] = ACTIONS(966),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(966),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(137)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(4841),
    [sym__stab_clause_left] = STATE(6997),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(698),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(138)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(4841),
    [sym__stab_clause_left] = STATE(6989),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(101),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(139)] = {
    [sym__expression] = STATE(4297),
    [sym_block] = STATE(4297),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4297),
    [sym_nil] = STATE(4297),
    [sym__atom] = STATE(4297),
    [sym_quoted_atom] = STATE(4297),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4297),
    [sym_charlist] = STATE(4297),
    [sym_sigil] = STATE(4297),
    [sym_keywords] = STATE(6788),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4297),
    [sym_tuple] = STATE(4297),
    [sym_bitstring] = STATE(4297),
    [sym_map] = STATE(4297),
    [sym__nullary_operator] = STATE(4297),
    [sym_unary_operator] = STATE(4297),
    [sym_binary_operator] = STATE(4297),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4297),
    [sym_call] = STATE(4297),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4297),
    [sym_stab_clause] = STATE(6850),
    [sym__stab_clause_left] = STATE(6887),
    [sym__stab_clause_arguments_with_parentheses] = STATE(6811),
    [sym__stab_clause_arguments_without_parentheses] = STATE(6815),
    [sym__stab_clause_arguments_with_parentheses_with_guard] = STATE(6877),
    [sym__stab_clause_arguments_without_parentheses_with_guard] = STATE(6878),
    [sym_anonymous_function] = STATE(4297),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(718),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(722),
    [sym_integer] = ACTIONS(722),
    [sym_float] = ACTIONS(722),
    [sym_char] = ACTIONS(722),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(722),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(754),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(140)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4765),
    [sym_rescue_block] = STATE(4765),
    [sym_catch_block] = STATE(4765),
    [sym_else_block] = STATE(4765),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4765),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1002),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(141)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4708),
    [sym_rescue_block] = STATE(4708),
    [sym_catch_block] = STATE(4708),
    [sym_else_block] = STATE(4708),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4708),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1010),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(142)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4714),
    [sym_rescue_block] = STATE(4714),
    [sym_catch_block] = STATE(4714),
    [sym_else_block] = STATE(4714),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4714),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1012),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(143)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4718),
    [sym_rescue_block] = STATE(4718),
    [sym_catch_block] = STATE(4718),
    [sym_else_block] = STATE(4718),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4718),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1014),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(144)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4762),
    [sym_rescue_block] = STATE(4762),
    [sym_catch_block] = STATE(4762),
    [sym_else_block] = STATE(4762),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4762),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1016),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(145)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4763),
    [sym_rescue_block] = STATE(4763),
    [sym_catch_block] = STATE(4763),
    [sym_else_block] = STATE(4763),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4763),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1018),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(146)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4764),
    [sym_rescue_block] = STATE(4764),
    [sym_catch_block] = STATE(4764),
    [sym_else_block] = STATE(4764),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4764),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1020),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(147)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4720),
    [sym_rescue_block] = STATE(4720),
    [sym_catch_block] = STATE(4720),
    [sym_else_block] = STATE(4720),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4720),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1022),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(148)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4724),
    [sym_rescue_block] = STATE(4724),
    [sym_catch_block] = STATE(4724),
    [sym_else_block] = STATE(4724),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4724),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1024),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(149)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4726),
    [sym_rescue_block] = STATE(4726),
    [sym_catch_block] = STATE(4726),
    [sym_else_block] = STATE(4726),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4726),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1026),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(150)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4755),
    [sym_rescue_block] = STATE(4755),
    [sym_catch_block] = STATE(4755),
    [sym_else_block] = STATE(4755),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4755),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1028),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(151)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4759),
    [sym_rescue_block] = STATE(4759),
    [sym_catch_block] = STATE(4759),
    [sym_else_block] = STATE(4759),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4759),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1030),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(152)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4761),
    [sym_rescue_block] = STATE(4761),
    [sym_catch_block] = STATE(4761),
    [sym_else_block] = STATE(4761),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4761),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1032),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(153)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4709),
    [sym_rescue_block] = STATE(4709),
    [sym_catch_block] = STATE(4709),
    [sym_else_block] = STATE(4709),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4709),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1034),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(154)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4711),
    [sym_rescue_block] = STATE(4711),
    [sym_catch_block] = STATE(4711),
    [sym_else_block] = STATE(4711),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4711),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1036),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(155)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4757),
    [sym_rescue_block] = STATE(4757),
    [sym_catch_block] = STATE(4757),
    [sym_else_block] = STATE(4757),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4757),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1038),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(156)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4730),
    [sym_rescue_block] = STATE(4730),
    [sym_catch_block] = STATE(4730),
    [sym_else_block] = STATE(4730),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4730),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1040),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(157)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4739),
    [sym_rescue_block] = STATE(4739),
    [sym_catch_block] = STATE(4739),
    [sym_else_block] = STATE(4739),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4739),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1042),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(158)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4742),
    [sym_rescue_block] = STATE(4742),
    [sym_catch_block] = STATE(4742),
    [sym_else_block] = STATE(4742),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4742),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1044),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(159)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4721),
    [sym_rescue_block] = STATE(4721),
    [sym_catch_block] = STATE(4721),
    [sym_else_block] = STATE(4721),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4721),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1046),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(160)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4722),
    [sym_rescue_block] = STATE(4722),
    [sym_catch_block] = STATE(4722),
    [sym_else_block] = STATE(4722),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4722),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1048),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(161)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4723),
    [sym_rescue_block] = STATE(4723),
    [sym_catch_block] = STATE(4723),
    [sym_else_block] = STATE(4723),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4723),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1050),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(162)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4732),
    [sym_rescue_block] = STATE(4732),
    [sym_catch_block] = STATE(4732),
    [sym_else_block] = STATE(4732),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4732),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1052),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(163)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4734),
    [sym_rescue_block] = STATE(4734),
    [sym_catch_block] = STATE(4734),
    [sym_else_block] = STATE(4734),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4734),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1054),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(164)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4735),
    [sym_rescue_block] = STATE(4735),
    [sym_catch_block] = STATE(4735),
    [sym_else_block] = STATE(4735),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4735),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1056),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(165)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4736),
    [sym_rescue_block] = STATE(4736),
    [sym_catch_block] = STATE(4736),
    [sym_else_block] = STATE(4736),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4736),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1058),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(166)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4745),
    [sym_rescue_block] = STATE(4745),
    [sym_catch_block] = STATE(4745),
    [sym_else_block] = STATE(4745),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4745),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1060),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(167)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4741),
    [sym_rescue_block] = STATE(4741),
    [sym_catch_block] = STATE(4741),
    [sym_else_block] = STATE(4741),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4741),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1062),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(168)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4743),
    [sym_rescue_block] = STATE(4743),
    [sym_catch_block] = STATE(4743),
    [sym_else_block] = STATE(4743),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4743),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1064),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(169)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4744),
    [sym_rescue_block] = STATE(4744),
    [sym_catch_block] = STATE(4744),
    [sym_else_block] = STATE(4744),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4744),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1066),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(170)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4746),
    [sym_rescue_block] = STATE(4746),
    [sym_catch_block] = STATE(4746),
    [sym_else_block] = STATE(4746),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4746),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1068),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(171)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4740),
    [sym_rescue_block] = STATE(4740),
    [sym_catch_block] = STATE(4740),
    [sym_else_block] = STATE(4740),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4740),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1070),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(172)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4753),
    [sym_rescue_block] = STATE(4753),
    [sym_catch_block] = STATE(4753),
    [sym_else_block] = STATE(4753),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4753),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1072),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(173)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4754),
    [sym_rescue_block] = STATE(4754),
    [sym_catch_block] = STATE(4754),
    [sym_else_block] = STATE(4754),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4754),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1074),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(174)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4756),
    [sym_rescue_block] = STATE(4756),
    [sym_catch_block] = STATE(4756),
    [sym_else_block] = STATE(4756),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4756),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1076),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(175)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_after_block] = STATE(4712),
    [sym_rescue_block] = STATE(4712),
    [sym_catch_block] = STATE(4712),
    [sym_else_block] = STATE(4712),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym_do_block_repeat1] = STATE(4712),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1078),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(113),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(176)] = {
    [sym__terminator] = STATE(260),
    [sym__expression] = STATE(1683),
    [sym_block] = STATE(1683),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(1683),
    [sym_nil] = STATE(1683),
    [sym__atom] = STATE(1683),
    [sym_quoted_atom] = STATE(1683),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(1683),
    [sym_charlist] = STATE(1683),
    [sym_sigil] = STATE(1683),
    [sym_list] = STATE(1683),
    [sym_tuple] = STATE(1683),
    [sym_bitstring] = STATE(1683),
    [sym_map] = STATE(1683),
    [sym__nullary_operator] = STATE(1683),
    [sym_unary_operator] = STATE(1683),
    [sym_binary_operator] = STATE(1683),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(1683),
    [sym_call] = STATE(1683),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(1683),
    [sym_body] = STATE(4842),
    [sym_anonymous_function] = STATE(1683),
    [aux_sym__terminator_repeat1] = STATE(1021),
    [aux_sym__terminator_token1] = ACTIONS(1080),
    [anon_sym_SEMI] = ACTIONS(1082),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1084),
    [sym_integer] = ACTIONS(1084),
    [sym_float] = ACTIONS(1084),
    [sym_char] = ACTIONS(1084),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1084),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1089),
    [anon_sym_catch] = ACTIONS(1089),
    [anon_sym_else] = ACTIONS(1089),
    [anon_sym_end] = ACTIONS(1089),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1089),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(177)] = {
    [sym__terminator] = STATE(260),
    [sym__expression] = STATE(1683),
    [sym_block] = STATE(1683),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(1683),
    [sym_nil] = STATE(1683),
    [sym__atom] = STATE(1683),
    [sym_quoted_atom] = STATE(1683),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(1683),
    [sym_charlist] = STATE(1683),
    [sym_sigil] = STATE(1683),
    [sym_list] = STATE(1683),
    [sym_tuple] = STATE(1683),
    [sym_bitstring] = STATE(1683),
    [sym_map] = STATE(1683),
    [sym__nullary_operator] = STATE(1683),
    [sym_unary_operator] = STATE(1683),
    [sym_binary_operator] = STATE(1683),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(1683),
    [sym_call] = STATE(1683),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(1683),
    [sym_body] = STATE(4843),
    [sym_anonymous_function] = STATE(1683),
    [aux_sym__terminator_repeat1] = STATE(1021),
    [aux_sym__terminator_token1] = ACTIONS(1080),
    [anon_sym_SEMI] = ACTIONS(1082),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1084),
    [sym_integer] = ACTIONS(1084),
    [sym_float] = ACTIONS(1084),
    [sym_char] = ACTIONS(1084),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1084),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1091),
    [anon_sym_catch] = ACTIONS(1091),
    [anon_sym_else] = ACTIONS(1091),
    [anon_sym_end] = ACTIONS(1091),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1091),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(178)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6958),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1111),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(179)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6885),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1137),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(180)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6912),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6912),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(181)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6922),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6922),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1143),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(182)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6943),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6943),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1145),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(183)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6993),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1147),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(184)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6983),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1149),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(185)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6908),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6908),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1151),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(186)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6937),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1153),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(187)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6952),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1155),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(188)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6870),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1185),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(189)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6996),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1203),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(190)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6930),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6930),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1205),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(191)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6936),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1207),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(192)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6967),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6967),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1209),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(193)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6994),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1211),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(194)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6889),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6889),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1213),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(195)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6928),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1215),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(196)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6963),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1217),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(197)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6962),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1219),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(198)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6926),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1221),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(199)] = {
    [sym__expression] = STATE(3830),
    [sym_block] = STATE(3830),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3830),
    [sym_nil] = STATE(3830),
    [sym__atom] = STATE(3830),
    [sym_quoted_atom] = STATE(3830),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3830),
    [sym_charlist] = STATE(3830),
    [sym_sigil] = STATE(3830),
    [sym__keywords_with_trailing_separator] = STATE(6659),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3830),
    [sym_tuple] = STATE(3830),
    [sym_bitstring] = STATE(3830),
    [sym_map] = STATE(3830),
    [sym__nullary_operator] = STATE(3830),
    [sym_unary_operator] = STATE(3830),
    [sym_binary_operator] = STATE(3830),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3830),
    [sym_call] = STATE(3830),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3830),
    [sym_anonymous_function] = STATE(3830),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1223),
    [sym_integer] = ACTIONS(1223),
    [sym_float] = ACTIONS(1223),
    [sym_char] = ACTIONS(1223),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1223),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1225),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1225),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(200)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6879),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1227),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(201)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6913),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6913),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1229),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(202)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6897),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1231),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(203)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6901),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1233),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(204)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6905),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1235),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(205)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6888),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1237),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(206)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6985),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1239),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(207)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6900),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6900),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1241),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(208)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6988),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1243),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(209)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6910),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6910),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1245),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(210)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6966),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1247),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(211)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6990),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1249),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(212)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6984),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6984),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1251),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(213)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6899),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1253),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(214)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6956),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1255),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(215)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6894),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1257),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(216)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6895),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1259),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(217)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6896),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1261),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(218)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6933),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1263),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(219)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6904),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1265),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(220)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6981),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1267),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(221)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6881),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1269),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(222)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6871),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6871),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1271),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(223)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6920),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1273),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(224)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6906),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6906),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1275),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(225)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6982),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1277),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(226)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6998),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1279),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(227)] = {
    [sym__expression] = STATE(3830),
    [sym_block] = STATE(3830),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3830),
    [sym_nil] = STATE(3830),
    [sym__atom] = STATE(3830),
    [sym_quoted_atom] = STATE(3830),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3830),
    [sym_charlist] = STATE(3830),
    [sym_sigil] = STATE(3830),
    [sym__keywords_with_trailing_separator] = STATE(6590),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3830),
    [sym_tuple] = STATE(3830),
    [sym_bitstring] = STATE(3830),
    [sym_map] = STATE(3830),
    [sym__nullary_operator] = STATE(3830),
    [sym_unary_operator] = STATE(3830),
    [sym_binary_operator] = STATE(3830),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3830),
    [sym_call] = STATE(3830),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3830),
    [sym_anonymous_function] = STATE(3830),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1223),
    [sym_integer] = ACTIONS(1223),
    [sym_float] = ACTIONS(1223),
    [sym_char] = ACTIONS(1223),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1223),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1281),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1281),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(228)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6941),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6941),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1283),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(229)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6971),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1285),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(230)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6929),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1287),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(231)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6932),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1289),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(232)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6947),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1291),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(233)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6884),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1293),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(234)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6907),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6907),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1295),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(235)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6914),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(236)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6987),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1299),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(237)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(7000),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(7000),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1301),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(238)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6893),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1303),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(239)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6995),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1305),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(240)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6873),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1307),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(241)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6883),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1309),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(242)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6902),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1311),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(243)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6976),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1313),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(244)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6953),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1315),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(245)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6875),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1317),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(246)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6978),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1319),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(247)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6876),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1321),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(248)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6880),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1323),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(249)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6890),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1325),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(250)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6942),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1327),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(251)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6903),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_RBRACK] = ACTIONS(1329),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(252)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6921),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6921),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1331),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(253)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6946),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1333),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(254)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6938),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1335),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(255)] = {
    [sym__expression] = STATE(4360),
    [sym_block] = STATE(4360),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4360),
    [sym_nil] = STATE(4360),
    [sym__atom] = STATE(4360),
    [sym_quoted_atom] = STATE(4360),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4360),
    [sym_charlist] = STATE(4360),
    [sym_sigil] = STATE(4360),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4360),
    [sym_tuple] = STATE(4360),
    [sym_bitstring] = STATE(4360),
    [sym_map] = STATE(4360),
    [sym__items_with_trailing_separator] = STATE(6999),
    [sym__nullary_operator] = STATE(4360),
    [sym_unary_operator] = STATE(4360),
    [sym_binary_operator] = STATE(4360),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4360),
    [sym_call] = STATE(4360),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4360),
    [sym_anonymous_function] = STATE(4360),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1161),
    [sym_integer] = ACTIONS(1161),
    [sym_float] = ACTIONS(1161),
    [sym_char] = ACTIONS(1161),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1161),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1337),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(256)] = {
    [sym__expression] = STATE(4291),
    [sym_block] = STATE(4291),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4291),
    [sym_nil] = STATE(4291),
    [sym__atom] = STATE(4291),
    [sym_quoted_atom] = STATE(4291),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4291),
    [sym_charlist] = STATE(4291),
    [sym_sigil] = STATE(4291),
    [sym__keywords_with_trailing_separator] = STATE(6977),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4291),
    [sym_tuple] = STATE(4291),
    [sym_bitstring] = STATE(4291),
    [sym_map] = STATE(4291),
    [sym__nullary_operator] = STATE(4291),
    [sym_unary_operator] = STATE(4291),
    [sym_binary_operator] = STATE(4291),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4291),
    [sym_call] = STATE(4291),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym__call_arguments_with_trailing_separator] = STATE(6977),
    [sym_access_call] = STATE(4291),
    [sym_anonymous_function] = STATE(4291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [anon_sym_RPAREN] = ACTIONS(1339),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1141),
    [sym_integer] = ACTIONS(1141),
    [sym_float] = ACTIONS(1141),
    [sym_char] = ACTIONS(1141),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1141),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(257)] = {
    [sym__expression] = STATE(3862),
    [sym_block] = STATE(3862),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3862),
    [sym_nil] = STATE(3862),
    [sym__atom] = STATE(3862),
    [sym_quoted_atom] = STATE(3862),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3862),
    [sym_charlist] = STATE(3862),
    [sym_sigil] = STATE(3862),
    [sym__keywords_with_trailing_separator] = STATE(6764),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3862),
    [sym_tuple] = STATE(3862),
    [sym_bitstring] = STATE(3862),
    [sym_map] = STATE(3862),
    [sym__items_with_trailing_separator] = STATE(6923),
    [sym__nullary_operator] = STATE(3862),
    [sym_unary_operator] = STATE(3862),
    [sym_binary_operator] = STATE(3862),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3862),
    [sym_call] = STATE(3862),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3862),
    [sym_anonymous_function] = STATE(3862),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1095),
    [sym_integer] = ACTIONS(1095),
    [sym_float] = ACTIONS(1095),
    [sym_char] = ACTIONS(1095),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1095),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1341),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(258)] = {
    [sym__expression] = STATE(2587),
    [sym_block] = STATE(2587),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2587),
    [sym_nil] = STATE(2587),
    [sym__atom] = STATE(2587),
    [sym_quoted_atom] = STATE(2587),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2587),
    [sym_charlist] = STATE(2587),
    [sym_sigil] = STATE(2587),
    [sym_list] = STATE(2587),
    [sym_tuple] = STATE(2587),
    [sym_bitstring] = STATE(2587),
    [sym_map] = STATE(2587),
    [sym__nullary_operator] = STATE(2587),
    [sym_unary_operator] = STATE(2587),
    [sym_binary_operator] = STATE(2587),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2587),
    [sym_call] = STATE(2587),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2587),
    [sym_anonymous_function] = STATE(2587),
    [aux_sym__terminator_token1] = ACTIONS(1343),
    [anon_sym_SEMI] = ACTIONS(1345),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1347),
    [sym_integer] = ACTIONS(1347),
    [sym_float] = ACTIONS(1347),
    [sym_char] = ACTIONS(1347),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1347),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1345),
    [anon_sym_catch] = ACTIONS(1345),
    [anon_sym_else] = ACTIONS(1345),
    [anon_sym_end] = ACTIONS(1345),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1345),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(259)] = {
    [sym__expression] = STATE(2587),
    [sym_block] = STATE(2587),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2587),
    [sym_nil] = STATE(2587),
    [sym__atom] = STATE(2587),
    [sym_quoted_atom] = STATE(2587),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2587),
    [sym_charlist] = STATE(2587),
    [sym_sigil] = STATE(2587),
    [sym_list] = STATE(2587),
    [sym_tuple] = STATE(2587),
    [sym_bitstring] = STATE(2587),
    [sym_map] = STATE(2587),
    [sym__nullary_operator] = STATE(2587),
    [sym_unary_operator] = STATE(2587),
    [sym_binary_operator] = STATE(2587),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2587),
    [sym_call] = STATE(2587),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2587),
    [sym_anonymous_function] = STATE(2587),
    [aux_sym__terminator_token1] = ACTIONS(1349),
    [anon_sym_SEMI] = ACTIONS(1351),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1347),
    [sym_integer] = ACTIONS(1347),
    [sym_float] = ACTIONS(1347),
    [sym_char] = ACTIONS(1347),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1347),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1351),
    [anon_sym_catch] = ACTIONS(1351),
    [anon_sym_else] = ACTIONS(1351),
    [anon_sym_end] = ACTIONS(1351),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1351),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(260)] = {
    [sym__expression] = STATE(1687),
    [sym_block] = STATE(1687),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(1687),
    [sym_nil] = STATE(1687),
    [sym__atom] = STATE(1687),
    [sym_quoted_atom] = STATE(1687),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(1687),
    [sym_charlist] = STATE(1687),
    [sym_sigil] = STATE(1687),
    [sym_list] = STATE(1687),
    [sym_tuple] = STATE(1687),
    [sym_bitstring] = STATE(1687),
    [sym_map] = STATE(1687),
    [sym__nullary_operator] = STATE(1687),
    [sym_unary_operator] = STATE(1687),
    [sym_binary_operator] = STATE(1687),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(1687),
    [sym_call] = STATE(1687),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(1687),
    [sym_anonymous_function] = STATE(1687),
    [aux_sym__terminator_token1] = ACTIONS(1353),
    [anon_sym_SEMI] = ACTIONS(1355),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1357),
    [sym_integer] = ACTIONS(1357),
    [sym_float] = ACTIONS(1357),
    [sym_char] = ACTIONS(1357),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1357),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1355),
    [anon_sym_catch] = ACTIONS(1355),
    [anon_sym_else] = ACTIONS(1355),
    [anon_sym_end] = ACTIONS(1355),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1355),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(261)] = {
    [sym__expression] = STATE(2587),
    [sym_block] = STATE(2587),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2587),
    [sym_nil] = STATE(2587),
    [sym__atom] = STATE(2587),
    [sym_quoted_atom] = STATE(2587),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2587),
    [sym_charlist] = STATE(2587),
    [sym_sigil] = STATE(2587),
    [sym_list] = STATE(2587),
    [sym_tuple] = STATE(2587),
    [sym_bitstring] = STATE(2587),
    [sym_map] = STATE(2587),
    [sym__nullary_operator] = STATE(2587),
    [sym_unary_operator] = STATE(2587),
    [sym_binary_operator] = STATE(2587),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2587),
    [sym_call] = STATE(2587),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2587),
    [sym_anonymous_function] = STATE(2587),
    [aux_sym__terminator_token1] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1361),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1347),
    [sym_integer] = ACTIONS(1347),
    [sym_float] = ACTIONS(1347),
    [sym_char] = ACTIONS(1347),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1347),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1361),
    [anon_sym_catch] = ACTIONS(1361),
    [anon_sym_else] = ACTIONS(1361),
    [anon_sym_end] = ACTIONS(1361),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1361),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(262)] = {
    [sym__expression] = STATE(4442),
    [sym_block] = STATE(4442),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4442),
    [sym_nil] = STATE(4442),
    [sym__atom] = STATE(4442),
    [sym_quoted_atom] = STATE(4442),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4442),
    [sym_charlist] = STATE(4442),
    [sym_sigil] = STATE(4442),
    [sym__keywords_with_trailing_separator] = STATE(6659),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4442),
    [sym_tuple] = STATE(4442),
    [sym_bitstring] = STATE(4442),
    [sym_map] = STATE(4442),
    [sym__nullary_operator] = STATE(4442),
    [sym_unary_operator] = STATE(4442),
    [sym_binary_operator] = STATE(4442),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4442),
    [sym_call] = STATE(4442),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4442),
    [sym_anonymous_function] = STATE(4442),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1363),
    [sym_integer] = ACTIONS(1363),
    [sym_float] = ACTIONS(1363),
    [sym_char] = ACTIONS(1363),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1363),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1225),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(263)] = {
    [sym__expression] = STATE(4442),
    [sym_block] = STATE(4442),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4442),
    [sym_nil] = STATE(4442),
    [sym__atom] = STATE(4442),
    [sym_quoted_atom] = STATE(4442),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4442),
    [sym_charlist] = STATE(4442),
    [sym_sigil] = STATE(4442),
    [sym__keywords_with_trailing_separator] = STATE(6590),
    [sym_pair] = STATE(6755),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4442),
    [sym_tuple] = STATE(4442),
    [sym_bitstring] = STATE(4442),
    [sym_map] = STATE(4442),
    [sym__nullary_operator] = STATE(4442),
    [sym_unary_operator] = STATE(4442),
    [sym_binary_operator] = STATE(4442),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4442),
    [sym_call] = STATE(4442),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4442),
    [sym_anonymous_function] = STATE(4442),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1363),
    [sym_integer] = ACTIONS(1363),
    [sym_float] = ACTIONS(1363),
    [sym_char] = ACTIONS(1363),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1363),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_GT_GT] = ACTIONS(1281),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(264)] = {
    [sym__expression] = STATE(4445),
    [sym_block] = STATE(4445),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4445),
    [sym_nil] = STATE(4445),
    [sym__atom] = STATE(4445),
    [sym_quoted_atom] = STATE(4445),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4445),
    [sym_charlist] = STATE(4445),
    [sym_sigil] = STATE(4445),
    [sym_keywords] = STATE(6835),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4445),
    [sym_tuple] = STATE(4445),
    [sym_bitstring] = STATE(4445),
    [sym_map] = STATE(4445),
    [sym__nullary_operator] = STATE(4445),
    [sym_unary_operator] = STATE(4445),
    [sym_binary_operator] = STATE(4445),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4445),
    [sym_call] = STATE(4445),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4445),
    [sym_anonymous_function] = STATE(4445),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1367),
    [sym_integer] = ACTIONS(1367),
    [sym_float] = ACTIONS(1367),
    [sym_char] = ACTIONS(1367),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1367),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(265)] = {
    [sym__expression] = STATE(2571),
    [sym_block] = STATE(2571),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2571),
    [sym_nil] = STATE(2571),
    [sym__atom] = STATE(2571),
    [sym_quoted_atom] = STATE(2571),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2571),
    [sym_charlist] = STATE(2571),
    [sym_sigil] = STATE(2571),
    [sym_keywords] = STATE(1202),
    [sym_pair] = STATE(2218),
    [sym__keyword] = STATE(845),
    [sym_quoted_keyword] = STATE(845),
    [sym_list] = STATE(2571),
    [sym_tuple] = STATE(2571),
    [sym_bitstring] = STATE(2571),
    [sym_map] = STATE(2571),
    [sym__nullary_operator] = STATE(2571),
    [sym_unary_operator] = STATE(2571),
    [sym_binary_operator] = STATE(2571),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2571),
    [sym_call] = STATE(2571),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2571),
    [sym_anonymous_function] = STATE(2571),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1369),
    [sym_integer] = ACTIONS(1369),
    [sym_float] = ACTIONS(1369),
    [sym_char] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1369),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [sym_keyword] = ACTIONS(481),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(266)] = {
    [sym__expression] = STATE(3538),
    [sym_block] = STATE(3538),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3538),
    [sym_nil] = STATE(3538),
    [sym__atom] = STATE(3538),
    [sym_quoted_atom] = STATE(3538),
    [sym__quoted_i_double] = STATE(3196),
    [sym__quoted_i_single] = STATE(3197),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3538),
    [sym_charlist] = STATE(3538),
    [sym_sigil] = STATE(3538),
    [sym_keywords] = STATE(3528),
    [sym_pair] = STATE(3225),
    [sym__keyword] = STATE(855),
    [sym_quoted_keyword] = STATE(855),
    [sym_list] = STATE(3538),
    [sym_tuple] = STATE(3538),
    [sym_bitstring] = STATE(3538),
    [sym_map] = STATE(3538),
    [sym__nullary_operator] = STATE(3538),
    [sym_unary_operator] = STATE(3538),
    [sym_binary_operator] = STATE(3538),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3538),
    [sym_call] = STATE(3538),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3538),
    [sym_anonymous_function] = STATE(3538),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(1373),
    [sym_integer] = ACTIONS(1373),
    [sym_float] = ACTIONS(1373),
    [sym_char] = ACTIONS(1373),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(1373),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [sym_keyword] = ACTIONS(564),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(267)] = {
    [sym__expression] = STATE(3538),
    [sym_block] = STATE(3538),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3538),
    [sym_nil] = STATE(3538),
    [sym__atom] = STATE(3538),
    [sym_quoted_atom] = STATE(3538),
    [sym__quoted_i_double] = STATE(3196),
    [sym__quoted_i_single] = STATE(3197),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3538),
    [sym_charlist] = STATE(3538),
    [sym_sigil] = STATE(3538),
    [sym_keywords] = STATE(3529),
    [sym_pair] = STATE(3225),
    [sym__keyword] = STATE(855),
    [sym_quoted_keyword] = STATE(855),
    [sym_list] = STATE(3538),
    [sym_tuple] = STATE(3538),
    [sym_bitstring] = STATE(3538),
    [sym_map] = STATE(3538),
    [sym__nullary_operator] = STATE(3538),
    [sym_unary_operator] = STATE(3538),
    [sym_binary_operator] = STATE(3538),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3538),
    [sym_call] = STATE(3538),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3538),
    [sym_anonymous_function] = STATE(3538),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(1373),
    [sym_integer] = ACTIONS(1373),
    [sym_float] = ACTIONS(1373),
    [sym_char] = ACTIONS(1373),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(1373),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [sym_keyword] = ACTIONS(564),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(268)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1377),
    [anon_sym_catch] = ACTIONS(1377),
    [anon_sym_else] = ACTIONS(1377),
    [anon_sym_end] = ACTIONS(1377),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1377),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(269)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1379),
    [anon_sym_catch] = ACTIONS(1379),
    [anon_sym_else] = ACTIONS(1379),
    [anon_sym_end] = ACTIONS(1379),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1379),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(270)] = {
    [sym__expression] = STATE(2647),
    [sym_block] = STATE(2647),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2647),
    [sym_nil] = STATE(2647),
    [sym__atom] = STATE(2647),
    [sym_quoted_atom] = STATE(2647),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2647),
    [sym_charlist] = STATE(2647),
    [sym_sigil] = STATE(2647),
    [sym_keywords] = STATE(2312),
    [sym_pair] = STATE(2158),
    [sym__keyword] = STATE(848),
    [sym_quoted_keyword] = STATE(848),
    [sym_list] = STATE(2647),
    [sym_tuple] = STATE(2647),
    [sym_bitstring] = STATE(2647),
    [sym_map] = STATE(2647),
    [sym__nullary_operator] = STATE(2647),
    [sym_unary_operator] = STATE(2647),
    [sym_binary_operator] = STATE(2647),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2647),
    [sym_call] = STATE(2647),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2647),
    [sym_anonymous_function] = STATE(2647),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1381),
    [sym_integer] = ACTIONS(1381),
    [sym_float] = ACTIONS(1381),
    [sym_char] = ACTIONS(1381),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1381),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [sym_keyword] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(271)] = {
    [sym__expression] = STATE(2653),
    [sym_block] = STATE(2653),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2653),
    [sym_nil] = STATE(2653),
    [sym__atom] = STATE(2653),
    [sym_quoted_atom] = STATE(2653),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2653),
    [sym_charlist] = STATE(2653),
    [sym_sigil] = STATE(2653),
    [sym_keywords] = STATE(2313),
    [sym_pair] = STATE(2158),
    [sym__keyword] = STATE(848),
    [sym_quoted_keyword] = STATE(848),
    [sym_list] = STATE(2653),
    [sym_tuple] = STATE(2653),
    [sym_bitstring] = STATE(2653),
    [sym_map] = STATE(2653),
    [sym__nullary_operator] = STATE(2653),
    [sym_unary_operator] = STATE(2653),
    [sym_binary_operator] = STATE(2653),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2653),
    [sym_call] = STATE(2653),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2653),
    [sym_anonymous_function] = STATE(2653),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1385),
    [sym_integer] = ACTIONS(1385),
    [sym_float] = ACTIONS(1385),
    [sym_char] = ACTIONS(1385),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1385),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [sym_keyword] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(272)] = {
    [sym__expression] = STATE(4280),
    [sym_block] = STATE(4280),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4280),
    [sym_nil] = STATE(4280),
    [sym__atom] = STATE(4280),
    [sym_quoted_atom] = STATE(4280),
    [sym__quoted_i_double] = STATE(3666),
    [sym__quoted_i_single] = STATE(3667),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4280),
    [sym_charlist] = STATE(4280),
    [sym_sigil] = STATE(4280),
    [sym_keywords] = STATE(4063),
    [sym_pair] = STATE(3982),
    [sym__keyword] = STATE(721),
    [sym_quoted_keyword] = STATE(721),
    [sym_list] = STATE(4280),
    [sym_tuple] = STATE(4280),
    [sym_bitstring] = STATE(4280),
    [sym_map] = STATE(4280),
    [sym__nullary_operator] = STATE(4280),
    [sym_unary_operator] = STATE(4280),
    [sym_binary_operator] = STATE(4280),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4280),
    [sym_call] = STATE(4280),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4280),
    [sym_anonymous_function] = STATE(4280),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1387),
    [sym_integer] = ACTIONS(1387),
    [sym_float] = ACTIONS(1387),
    [sym_char] = ACTIONS(1387),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1387),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [sym_keyword] = ACTIONS(1389),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(273)] = {
    [sym__expression] = STATE(3830),
    [sym_block] = STATE(3830),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3830),
    [sym_nil] = STATE(3830),
    [sym__atom] = STATE(3830),
    [sym_quoted_atom] = STATE(3830),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3830),
    [sym_charlist] = STATE(3830),
    [sym_sigil] = STATE(3830),
    [sym__keywords_with_trailing_separator] = STATE(6957),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3830),
    [sym_tuple] = STATE(3830),
    [sym_bitstring] = STATE(3830),
    [sym_map] = STATE(3830),
    [sym__nullary_operator] = STATE(3830),
    [sym_unary_operator] = STATE(3830),
    [sym_binary_operator] = STATE(3830),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3830),
    [sym_call] = STATE(3830),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3830),
    [sym_anonymous_function] = STATE(3830),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1223),
    [sym_integer] = ACTIONS(1223),
    [sym_float] = ACTIONS(1223),
    [sym_char] = ACTIONS(1223),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1223),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(274)] = {
    [sym__expression] = STATE(3721),
    [sym_block] = STATE(3721),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3721),
    [sym_nil] = STATE(3721),
    [sym__atom] = STATE(3721),
    [sym_quoted_atom] = STATE(3721),
    [sym__quoted_i_double] = STATE(3196),
    [sym__quoted_i_single] = STATE(3197),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3721),
    [sym_charlist] = STATE(3721),
    [sym_sigil] = STATE(3721),
    [sym_keywords] = STATE(3722),
    [sym_pair] = STATE(3225),
    [sym__keyword] = STATE(855),
    [sym_quoted_keyword] = STATE(855),
    [sym_list] = STATE(3721),
    [sym_tuple] = STATE(3721),
    [sym_bitstring] = STATE(3721),
    [sym_map] = STATE(3721),
    [sym__nullary_operator] = STATE(3721),
    [sym_unary_operator] = STATE(3721),
    [sym_binary_operator] = STATE(3721),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3721),
    [sym_call] = STATE(3721),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3721),
    [sym_anonymous_function] = STATE(3721),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(1391),
    [sym_integer] = ACTIONS(1391),
    [sym_float] = ACTIONS(1391),
    [sym_char] = ACTIONS(1391),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(1391),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [sym_keyword] = ACTIONS(564),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(275)] = {
    [sym__expression] = STATE(3727),
    [sym_block] = STATE(3727),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3727),
    [sym_nil] = STATE(3727),
    [sym__atom] = STATE(3727),
    [sym_quoted_atom] = STATE(3727),
    [sym__quoted_i_double] = STATE(3196),
    [sym__quoted_i_single] = STATE(3197),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3727),
    [sym_charlist] = STATE(3727),
    [sym_sigil] = STATE(3727),
    [sym_keywords] = STATE(3728),
    [sym_pair] = STATE(3225),
    [sym__keyword] = STATE(855),
    [sym_quoted_keyword] = STATE(855),
    [sym_list] = STATE(3727),
    [sym_tuple] = STATE(3727),
    [sym_bitstring] = STATE(3727),
    [sym_map] = STATE(3727),
    [sym__nullary_operator] = STATE(3727),
    [sym_unary_operator] = STATE(3727),
    [sym_binary_operator] = STATE(3727),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3727),
    [sym_call] = STATE(3727),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3727),
    [sym_anonymous_function] = STATE(3727),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(1393),
    [sym_integer] = ACTIONS(1393),
    [sym_float] = ACTIONS(1393),
    [sym_char] = ACTIONS(1393),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(1393),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [sym_keyword] = ACTIONS(564),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(276)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1395),
    [anon_sym_catch] = ACTIONS(1395),
    [anon_sym_else] = ACTIONS(1395),
    [anon_sym_end] = ACTIONS(1395),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(277)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1397),
    [anon_sym_catch] = ACTIONS(1397),
    [anon_sym_else] = ACTIONS(1397),
    [anon_sym_end] = ACTIONS(1397),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1397),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(278)] = {
    [sym__expression] = STATE(3147),
    [sym_block] = STATE(3147),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3147),
    [sym_nil] = STATE(3147),
    [sym__atom] = STATE(3147),
    [sym_quoted_atom] = STATE(3147),
    [sym__quoted_i_double] = STATE(2509),
    [sym__quoted_i_single] = STATE(2510),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3147),
    [sym_charlist] = STATE(3147),
    [sym_sigil] = STATE(3147),
    [sym_keywords] = STATE(3256),
    [sym_pair] = STATE(2669),
    [sym__keyword] = STATE(868),
    [sym_quoted_keyword] = STATE(868),
    [sym_list] = STATE(3147),
    [sym_tuple] = STATE(3147),
    [sym_bitstring] = STATE(3147),
    [sym_map] = STATE(3147),
    [sym__nullary_operator] = STATE(3147),
    [sym_unary_operator] = STATE(3147),
    [sym_binary_operator] = STATE(3147),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3147),
    [sym_call] = STATE(3147),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3147),
    [sym_anonymous_function] = STATE(3147),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1399),
    [sym_integer] = ACTIONS(1399),
    [sym_float] = ACTIONS(1399),
    [sym_char] = ACTIONS(1399),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [sym_keyword] = ACTIONS(612),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(279)] = {
    [sym__expression] = STATE(4301),
    [sym_block] = STATE(4301),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4301),
    [sym_nil] = STATE(4301),
    [sym__atom] = STATE(4301),
    [sym_quoted_atom] = STATE(4301),
    [sym__quoted_i_double] = STATE(1613),
    [sym__quoted_i_single] = STATE(1614),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4301),
    [sym_charlist] = STATE(4301),
    [sym_sigil] = STATE(4301),
    [sym_keywords] = STATE(2036),
    [sym_pair] = STATE(4034),
    [sym__keyword] = STATE(894),
    [sym_quoted_keyword] = STATE(894),
    [sym_list] = STATE(4301),
    [sym_tuple] = STATE(4301),
    [sym_bitstring] = STATE(4301),
    [sym_map] = STATE(4301),
    [sym__nullary_operator] = STATE(4301),
    [sym_unary_operator] = STATE(4301),
    [sym_binary_operator] = STATE(4301),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4301),
    [sym_call] = STATE(4301),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4301),
    [sym_anonymous_function] = STATE(4301),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1405),
    [sym_integer] = ACTIONS(1405),
    [sym_float] = ACTIONS(1405),
    [sym_char] = ACTIONS(1405),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1405),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [sym_keyword] = ACTIONS(1409),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(280)] = {
    [sym__expression] = STATE(3147),
    [sym_block] = STATE(3147),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3147),
    [sym_nil] = STATE(3147),
    [sym__atom] = STATE(3147),
    [sym_quoted_atom] = STATE(3147),
    [sym__quoted_i_double] = STATE(2509),
    [sym__quoted_i_single] = STATE(2510),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3147),
    [sym_charlist] = STATE(3147),
    [sym_sigil] = STATE(3147),
    [sym_keywords] = STATE(3365),
    [sym_pair] = STATE(2669),
    [sym__keyword] = STATE(868),
    [sym_quoted_keyword] = STATE(868),
    [sym_list] = STATE(3147),
    [sym_tuple] = STATE(3147),
    [sym_bitstring] = STATE(3147),
    [sym_map] = STATE(3147),
    [sym__nullary_operator] = STATE(3147),
    [sym_unary_operator] = STATE(3147),
    [sym_binary_operator] = STATE(3147),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3147),
    [sym_call] = STATE(3147),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3147),
    [sym_anonymous_function] = STATE(3147),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1399),
    [sym_integer] = ACTIONS(1399),
    [sym_float] = ACTIONS(1399),
    [sym_char] = ACTIONS(1399),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [sym_keyword] = ACTIONS(612),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(281)] = {
    [sym__expression] = STATE(4307),
    [sym_block] = STATE(4307),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4307),
    [sym_nil] = STATE(4307),
    [sym__atom] = STATE(4307),
    [sym_quoted_atom] = STATE(4307),
    [sym__quoted_i_double] = STATE(1613),
    [sym__quoted_i_single] = STATE(1614),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4307),
    [sym_charlist] = STATE(4307),
    [sym_sigil] = STATE(4307),
    [sym_keywords] = STATE(2037),
    [sym_pair] = STATE(4034),
    [sym__keyword] = STATE(894),
    [sym_quoted_keyword] = STATE(894),
    [sym_list] = STATE(4307),
    [sym_tuple] = STATE(4307),
    [sym_bitstring] = STATE(4307),
    [sym_map] = STATE(4307),
    [sym__nullary_operator] = STATE(4307),
    [sym_unary_operator] = STATE(4307),
    [sym_binary_operator] = STATE(4307),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4307),
    [sym_call] = STATE(4307),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4307),
    [sym_anonymous_function] = STATE(4307),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1419),
    [sym_integer] = ACTIONS(1419),
    [sym_float] = ACTIONS(1419),
    [sym_char] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1419),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [sym_keyword] = ACTIONS(1409),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(282)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1421),
    [anon_sym_catch] = ACTIONS(1421),
    [anon_sym_else] = ACTIONS(1421),
    [anon_sym_end] = ACTIONS(1421),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1421),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(283)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1423),
    [anon_sym_catch] = ACTIONS(1423),
    [anon_sym_else] = ACTIONS(1423),
    [anon_sym_end] = ACTIONS(1423),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1423),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(284)] = {
    [sym__expression] = STATE(2963),
    [sym_block] = STATE(2963),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2963),
    [sym_nil] = STATE(2963),
    [sym__atom] = STATE(2963),
    [sym_quoted_atom] = STATE(2963),
    [sym__quoted_i_double] = STATE(2509),
    [sym__quoted_i_single] = STATE(2510),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2963),
    [sym_charlist] = STATE(2963),
    [sym_sigil] = STATE(2963),
    [sym_keywords] = STATE(3396),
    [sym_pair] = STATE(2669),
    [sym__keyword] = STATE(868),
    [sym_quoted_keyword] = STATE(868),
    [sym_list] = STATE(2963),
    [sym_tuple] = STATE(2963),
    [sym_bitstring] = STATE(2963),
    [sym_map] = STATE(2963),
    [sym__nullary_operator] = STATE(2963),
    [sym_unary_operator] = STATE(2963),
    [sym_binary_operator] = STATE(2963),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2963),
    [sym_call] = STATE(2963),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2963),
    [sym_anonymous_function] = STATE(2963),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1425),
    [sym_integer] = ACTIONS(1425),
    [sym_float] = ACTIONS(1425),
    [sym_char] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(1425),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [sym_keyword] = ACTIONS(612),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(285)] = {
    [sym__expression] = STATE(2969),
    [sym_block] = STATE(2969),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2969),
    [sym_nil] = STATE(2969),
    [sym__atom] = STATE(2969),
    [sym_quoted_atom] = STATE(2969),
    [sym__quoted_i_double] = STATE(2509),
    [sym__quoted_i_single] = STATE(2510),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2969),
    [sym_charlist] = STATE(2969),
    [sym_sigil] = STATE(2969),
    [sym_keywords] = STATE(3397),
    [sym_pair] = STATE(2669),
    [sym__keyword] = STATE(868),
    [sym_quoted_keyword] = STATE(868),
    [sym_list] = STATE(2969),
    [sym_tuple] = STATE(2969),
    [sym_bitstring] = STATE(2969),
    [sym_map] = STATE(2969),
    [sym__nullary_operator] = STATE(2969),
    [sym_unary_operator] = STATE(2969),
    [sym_binary_operator] = STATE(2969),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2969),
    [sym_call] = STATE(2969),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2969),
    [sym_anonymous_function] = STATE(2969),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1427),
    [sym_integer] = ACTIONS(1427),
    [sym_float] = ACTIONS(1427),
    [sym_char] = ACTIONS(1427),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(1427),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [sym_keyword] = ACTIONS(612),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(286)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1429),
    [anon_sym_catch] = ACTIONS(1429),
    [anon_sym_else] = ACTIONS(1429),
    [anon_sym_end] = ACTIONS(1429),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1429),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(287)] = {
    [sym__expression] = STATE(1948),
    [sym_block] = STATE(1948),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1948),
    [sym_nil] = STATE(1948),
    [sym__atom] = STATE(1948),
    [sym_quoted_atom] = STATE(1948),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1948),
    [sym_charlist] = STATE(1948),
    [sym_sigil] = STATE(1948),
    [sym_keywords] = STATE(1684),
    [sym_pair] = STATE(1798),
    [sym__keyword] = STATE(896),
    [sym_quoted_keyword] = STATE(896),
    [sym_list] = STATE(1948),
    [sym_tuple] = STATE(1948),
    [sym_bitstring] = STATE(1948),
    [sym_map] = STATE(1948),
    [sym__nullary_operator] = STATE(1948),
    [sym_unary_operator] = STATE(1948),
    [sym_binary_operator] = STATE(1948),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1948),
    [sym_call] = STATE(1948),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1948),
    [sym_anonymous_function] = STATE(1948),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1433),
    [sym_integer] = ACTIONS(1433),
    [sym_float] = ACTIONS(1433),
    [sym_char] = ACTIONS(1433),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(1433),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(288)] = {
    [sym__expression] = STATE(1956),
    [sym_block] = STATE(1956),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1956),
    [sym_nil] = STATE(1956),
    [sym__atom] = STATE(1956),
    [sym_quoted_atom] = STATE(1956),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1956),
    [sym_charlist] = STATE(1956),
    [sym_sigil] = STATE(1956),
    [sym_keywords] = STATE(1685),
    [sym_pair] = STATE(1798),
    [sym__keyword] = STATE(896),
    [sym_quoted_keyword] = STATE(896),
    [sym_list] = STATE(1956),
    [sym_tuple] = STATE(1956),
    [sym_bitstring] = STATE(1956),
    [sym_map] = STATE(1956),
    [sym__nullary_operator] = STATE(1956),
    [sym_unary_operator] = STATE(1956),
    [sym_binary_operator] = STATE(1956),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1956),
    [sym_call] = STATE(1956),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1956),
    [sym_anonymous_function] = STATE(1956),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1437),
    [sym_integer] = ACTIONS(1437),
    [sym_float] = ACTIONS(1437),
    [sym_char] = ACTIONS(1437),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(1437),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [sym_keyword] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(289)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1439),
    [anon_sym_catch] = ACTIONS(1439),
    [anon_sym_else] = ACTIONS(1439),
    [anon_sym_end] = ACTIONS(1439),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1439),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(290)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1441),
    [anon_sym_catch] = ACTIONS(1441),
    [anon_sym_else] = ACTIONS(1441),
    [anon_sym_end] = ACTIONS(1441),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1441),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(291)] = {
    [sym__expression] = STATE(3207),
    [sym_block] = STATE(3207),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3207),
    [sym_nil] = STATE(3207),
    [sym__atom] = STATE(3207),
    [sym_quoted_atom] = STATE(3207),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3207),
    [sym_charlist] = STATE(3207),
    [sym_sigil] = STATE(3207),
    [sym_keywords] = STATE(1462),
    [sym_pair] = STATE(2695),
    [sym__keyword] = STATE(880),
    [sym_quoted_keyword] = STATE(880),
    [sym_list] = STATE(3207),
    [sym_tuple] = STATE(3207),
    [sym_bitstring] = STATE(3207),
    [sym_map] = STATE(3207),
    [sym__nullary_operator] = STATE(3207),
    [sym_unary_operator] = STATE(3207),
    [sym_binary_operator] = STATE(3207),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3207),
    [sym_call] = STATE(3207),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3207),
    [sym_anonymous_function] = STATE(3207),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1443),
    [sym_integer] = ACTIONS(1443),
    [sym_float] = ACTIONS(1443),
    [sym_char] = ACTIONS(1443),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1443),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [sym_keyword] = ACTIONS(495),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(292)] = {
    [sym__expression] = STATE(3207),
    [sym_block] = STATE(3207),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3207),
    [sym_nil] = STATE(3207),
    [sym__atom] = STATE(3207),
    [sym_quoted_atom] = STATE(3207),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3207),
    [sym_charlist] = STATE(3207),
    [sym_sigil] = STATE(3207),
    [sym_keywords] = STATE(1463),
    [sym_pair] = STATE(2695),
    [sym__keyword] = STATE(880),
    [sym_quoted_keyword] = STATE(880),
    [sym_list] = STATE(3207),
    [sym_tuple] = STATE(3207),
    [sym_bitstring] = STATE(3207),
    [sym_map] = STATE(3207),
    [sym__nullary_operator] = STATE(3207),
    [sym_unary_operator] = STATE(3207),
    [sym_binary_operator] = STATE(3207),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3207),
    [sym_call] = STATE(3207),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3207),
    [sym_anonymous_function] = STATE(3207),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1443),
    [sym_integer] = ACTIONS(1443),
    [sym_float] = ACTIONS(1443),
    [sym_char] = ACTIONS(1443),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1443),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [sym_keyword] = ACTIONS(495),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(293)] = {
    [sym__expression] = STATE(3115),
    [sym_block] = STATE(3115),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3115),
    [sym_nil] = STATE(3115),
    [sym__atom] = STATE(3115),
    [sym_quoted_atom] = STATE(3115),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3115),
    [sym_charlist] = STATE(3115),
    [sym_sigil] = STATE(3115),
    [sym_keywords] = STATE(1538),
    [sym_pair] = STATE(2695),
    [sym__keyword] = STATE(880),
    [sym_quoted_keyword] = STATE(880),
    [sym_list] = STATE(3115),
    [sym_tuple] = STATE(3115),
    [sym_bitstring] = STATE(3115),
    [sym_map] = STATE(3115),
    [sym__nullary_operator] = STATE(3115),
    [sym_unary_operator] = STATE(3115),
    [sym_binary_operator] = STATE(3115),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3115),
    [sym_call] = STATE(3115),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3115),
    [sym_anonymous_function] = STATE(3115),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1447),
    [sym_integer] = ACTIONS(1447),
    [sym_float] = ACTIONS(1447),
    [sym_char] = ACTIONS(1447),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1447),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [sym_keyword] = ACTIONS(495),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(294)] = {
    [sym__expression] = STATE(3120),
    [sym_block] = STATE(3120),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3120),
    [sym_nil] = STATE(3120),
    [sym__atom] = STATE(3120),
    [sym_quoted_atom] = STATE(3120),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3120),
    [sym_charlist] = STATE(3120),
    [sym_sigil] = STATE(3120),
    [sym_keywords] = STATE(1540),
    [sym_pair] = STATE(2695),
    [sym__keyword] = STATE(880),
    [sym_quoted_keyword] = STATE(880),
    [sym_list] = STATE(3120),
    [sym_tuple] = STATE(3120),
    [sym_bitstring] = STATE(3120),
    [sym_map] = STATE(3120),
    [sym__nullary_operator] = STATE(3120),
    [sym_unary_operator] = STATE(3120),
    [sym_binary_operator] = STATE(3120),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3120),
    [sym_call] = STATE(3120),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3120),
    [sym_anonymous_function] = STATE(3120),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1449),
    [sym_integer] = ACTIONS(1449),
    [sym_float] = ACTIONS(1449),
    [sym_char] = ACTIONS(1449),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1449),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [sym_keyword] = ACTIONS(495),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(295)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1451),
    [anon_sym_catch] = ACTIONS(1451),
    [anon_sym_else] = ACTIONS(1451),
    [anon_sym_end] = ACTIONS(1451),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1451),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(296)] = {
    [sym__expression] = STATE(3166),
    [sym_block] = STATE(3166),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3166),
    [sym_nil] = STATE(3166),
    [sym__atom] = STATE(3166),
    [sym_quoted_atom] = STATE(3166),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3166),
    [sym_charlist] = STATE(3166),
    [sym_sigil] = STATE(3166),
    [sym_keywords] = STATE(1538),
    [sym_pair] = STATE(2917),
    [sym__keyword] = STATE(887),
    [sym_quoted_keyword] = STATE(887),
    [sym_list] = STATE(3166),
    [sym_tuple] = STATE(3166),
    [sym_bitstring] = STATE(3166),
    [sym_map] = STATE(3166),
    [sym__nullary_operator] = STATE(3166),
    [sym_unary_operator] = STATE(3166),
    [sym_binary_operator] = STATE(3166),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3166),
    [sym_call] = STATE(3166),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3166),
    [sym_anonymous_function] = STATE(3166),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(1453),
    [sym_integer] = ACTIONS(1453),
    [sym_float] = ACTIONS(1453),
    [sym_char] = ACTIONS(1453),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1453),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [sym_keyword] = ACTIONS(514),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(297)] = {
    [sym__expression] = STATE(3171),
    [sym_block] = STATE(3171),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3171),
    [sym_nil] = STATE(3171),
    [sym__atom] = STATE(3171),
    [sym_quoted_atom] = STATE(3171),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3171),
    [sym_charlist] = STATE(3171),
    [sym_sigil] = STATE(3171),
    [sym_keywords] = STATE(1540),
    [sym_pair] = STATE(2917),
    [sym__keyword] = STATE(887),
    [sym_quoted_keyword] = STATE(887),
    [sym_list] = STATE(3171),
    [sym_tuple] = STATE(3171),
    [sym_bitstring] = STATE(3171),
    [sym_map] = STATE(3171),
    [sym__nullary_operator] = STATE(3171),
    [sym_unary_operator] = STATE(3171),
    [sym_binary_operator] = STATE(3171),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3171),
    [sym_call] = STATE(3171),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3171),
    [sym_anonymous_function] = STATE(3171),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(1455),
    [sym_integer] = ACTIONS(1455),
    [sym_float] = ACTIONS(1455),
    [sym_char] = ACTIONS(1455),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1455),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [sym_keyword] = ACTIONS(514),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(298)] = {
    [sym__expression] = STATE(4160),
    [sym_block] = STATE(4160),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4160),
    [sym_nil] = STATE(4160),
    [sym__atom] = STATE(4160),
    [sym_quoted_atom] = STATE(4160),
    [sym__quoted_i_double] = STATE(3666),
    [sym__quoted_i_single] = STATE(3667),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4160),
    [sym_charlist] = STATE(4160),
    [sym_sigil] = STATE(4160),
    [sym_keywords] = STATE(4175),
    [sym_pair] = STATE(3982),
    [sym__keyword] = STATE(721),
    [sym_quoted_keyword] = STATE(721),
    [sym_list] = STATE(4160),
    [sym_tuple] = STATE(4160),
    [sym_bitstring] = STATE(4160),
    [sym_map] = STATE(4160),
    [sym__nullary_operator] = STATE(4160),
    [sym_unary_operator] = STATE(4160),
    [sym_binary_operator] = STATE(4160),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4160),
    [sym_call] = STATE(4160),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4160),
    [sym_anonymous_function] = STATE(4160),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1457),
    [sym_integer] = ACTIONS(1457),
    [sym_float] = ACTIONS(1457),
    [sym_char] = ACTIONS(1457),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [sym_keyword] = ACTIONS(1389),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(299)] = {
    [sym__expression] = STATE(2191),
    [sym_block] = STATE(2191),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2191),
    [sym_nil] = STATE(2191),
    [sym__atom] = STATE(2191),
    [sym_quoted_atom] = STATE(2191),
    [sym__quoted_i_double] = STATE(1613),
    [sym__quoted_i_single] = STATE(1614),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2191),
    [sym_charlist] = STATE(2191),
    [sym_sigil] = STATE(2191),
    [sym_keywords] = STATE(2036),
    [sym_pair] = STATE(2123),
    [sym__keyword] = STATE(897),
    [sym_quoted_keyword] = STATE(897),
    [sym_list] = STATE(2191),
    [sym_tuple] = STATE(2191),
    [sym_bitstring] = STATE(2191),
    [sym_map] = STATE(2191),
    [sym__nullary_operator] = STATE(2191),
    [sym_unary_operator] = STATE(2191),
    [sym_binary_operator] = STATE(2191),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2191),
    [sym_call] = STATE(2191),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2191),
    [sym_anonymous_function] = STATE(2191),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1459),
    [sym_integer] = ACTIONS(1459),
    [sym_float] = ACTIONS(1459),
    [sym_char] = ACTIONS(1459),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1459),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [sym_keyword] = ACTIONS(1461),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(300)] = {
    [sym__expression] = STATE(2249),
    [sym_block] = STATE(2249),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2249),
    [sym_nil] = STATE(2249),
    [sym__atom] = STATE(2249),
    [sym_quoted_atom] = STATE(2249),
    [sym__quoted_i_double] = STATE(1613),
    [sym__quoted_i_single] = STATE(1614),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2249),
    [sym_charlist] = STATE(2249),
    [sym_sigil] = STATE(2249),
    [sym_keywords] = STATE(2037),
    [sym_pair] = STATE(2123),
    [sym__keyword] = STATE(897),
    [sym_quoted_keyword] = STATE(897),
    [sym_list] = STATE(2249),
    [sym_tuple] = STATE(2249),
    [sym_bitstring] = STATE(2249),
    [sym_map] = STATE(2249),
    [sym__nullary_operator] = STATE(2249),
    [sym_unary_operator] = STATE(2249),
    [sym_binary_operator] = STATE(2249),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2249),
    [sym_call] = STATE(2249),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2249),
    [sym_anonymous_function] = STATE(2249),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1463),
    [sym_integer] = ACTIONS(1463),
    [sym_float] = ACTIONS(1463),
    [sym_char] = ACTIONS(1463),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1463),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [sym_keyword] = ACTIONS(1461),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(301)] = {
    [sym__expression] = STATE(4029),
    [sym_block] = STATE(4029),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(4029),
    [sym_nil] = STATE(4029),
    [sym__atom] = STATE(4029),
    [sym_quoted_atom] = STATE(4029),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(4029),
    [sym_charlist] = STATE(4029),
    [sym_sigil] = STATE(4029),
    [sym_keywords] = STATE(2304),
    [sym_pair] = STATE(3499),
    [sym__keyword] = STATE(885),
    [sym_quoted_keyword] = STATE(885),
    [sym_list] = STATE(4029),
    [sym_tuple] = STATE(4029),
    [sym_bitstring] = STATE(4029),
    [sym_map] = STATE(4029),
    [sym__nullary_operator] = STATE(4029),
    [sym_unary_operator] = STATE(4029),
    [sym_binary_operator] = STATE(4029),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(4029),
    [sym_call] = STATE(4029),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(4029),
    [sym_anonymous_function] = STATE(4029),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1465),
    [sym_integer] = ACTIONS(1465),
    [sym_float] = ACTIONS(1465),
    [sym_char] = ACTIONS(1465),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1465),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [sym_keyword] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(302)] = {
    [sym__expression] = STATE(4029),
    [sym_block] = STATE(4029),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(4029),
    [sym_nil] = STATE(4029),
    [sym__atom] = STATE(4029),
    [sym_quoted_atom] = STATE(4029),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(4029),
    [sym_charlist] = STATE(4029),
    [sym_sigil] = STATE(4029),
    [sym_keywords] = STATE(2356),
    [sym_pair] = STATE(3499),
    [sym__keyword] = STATE(885),
    [sym_quoted_keyword] = STATE(885),
    [sym_list] = STATE(4029),
    [sym_tuple] = STATE(4029),
    [sym_bitstring] = STATE(4029),
    [sym_map] = STATE(4029),
    [sym__nullary_operator] = STATE(4029),
    [sym_unary_operator] = STATE(4029),
    [sym_binary_operator] = STATE(4029),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(4029),
    [sym_call] = STATE(4029),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(4029),
    [sym_anonymous_function] = STATE(4029),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1465),
    [sym_integer] = ACTIONS(1465),
    [sym_float] = ACTIONS(1465),
    [sym_char] = ACTIONS(1465),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1465),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [sym_keyword] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(303)] = {
    [sym__terminator] = STATE(348),
    [sym__expression] = STATE(3265),
    [sym_block] = STATE(3265),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(3265),
    [sym_nil] = STATE(3265),
    [sym__atom] = STATE(3265),
    [sym_quoted_atom] = STATE(3265),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(3265),
    [sym_charlist] = STATE(3265),
    [sym_sigil] = STATE(3265),
    [sym_list] = STATE(3265),
    [sym_tuple] = STATE(3265),
    [sym_bitstring] = STATE(3265),
    [sym_map] = STATE(3265),
    [sym__nullary_operator] = STATE(3265),
    [sym_unary_operator] = STATE(3265),
    [sym_binary_operator] = STATE(3265),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(3265),
    [sym_call] = STATE(3265),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(3265),
    [sym_body] = STATE(4842),
    [sym_anonymous_function] = STATE(3265),
    [aux_sym__terminator_repeat1] = STATE(1032),
    [aux_sym__terminator_token1] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1471),
    [sym_integer] = ACTIONS(1471),
    [sym_float] = ACTIONS(1471),
    [sym_char] = ACTIONS(1471),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1471),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_end] = ACTIONS(1089),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(304)] = {
    [sym__expression] = STATE(1516),
    [sym_block] = STATE(1516),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1516),
    [sym_nil] = STATE(1516),
    [sym__atom] = STATE(1516),
    [sym_quoted_atom] = STATE(1516),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1516),
    [sym_charlist] = STATE(1516),
    [sym_sigil] = STATE(1516),
    [sym_keywords] = STATE(1201),
    [sym_pair] = STATE(1349),
    [sym__keyword] = STATE(891),
    [sym_quoted_keyword] = STATE(891),
    [sym_list] = STATE(1516),
    [sym_tuple] = STATE(1516),
    [sym_bitstring] = STATE(1516),
    [sym_map] = STATE(1516),
    [sym__nullary_operator] = STATE(1516),
    [sym_unary_operator] = STATE(1516),
    [sym_binary_operator] = STATE(1516),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1516),
    [sym_call] = STATE(1516),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1516),
    [sym_anonymous_function] = STATE(1516),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1473),
    [sym_integer] = ACTIONS(1473),
    [sym_float] = ACTIONS(1473),
    [sym_char] = ACTIONS(1473),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1473),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [sym_keyword] = ACTIONS(262),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(305)] = {
    [sym__expression] = STATE(1522),
    [sym_block] = STATE(1522),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1522),
    [sym_nil] = STATE(1522),
    [sym__atom] = STATE(1522),
    [sym_quoted_atom] = STATE(1522),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1522),
    [sym_charlist] = STATE(1522),
    [sym_sigil] = STATE(1522),
    [sym_keywords] = STATE(1202),
    [sym_pair] = STATE(1349),
    [sym__keyword] = STATE(891),
    [sym_quoted_keyword] = STATE(891),
    [sym_list] = STATE(1522),
    [sym_tuple] = STATE(1522),
    [sym_bitstring] = STATE(1522),
    [sym_map] = STATE(1522),
    [sym__nullary_operator] = STATE(1522),
    [sym_unary_operator] = STATE(1522),
    [sym_binary_operator] = STATE(1522),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1522),
    [sym_call] = STATE(1522),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1522),
    [sym_anonymous_function] = STATE(1522),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1475),
    [sym_integer] = ACTIONS(1475),
    [sym_float] = ACTIONS(1475),
    [sym_char] = ACTIONS(1475),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1475),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [sym_keyword] = ACTIONS(262),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(306)] = {
    [sym__expression] = STATE(4024),
    [sym_block] = STATE(4024),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4024),
    [sym_nil] = STATE(4024),
    [sym__atom] = STATE(4024),
    [sym_quoted_atom] = STATE(4024),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4024),
    [sym_charlist] = STATE(4024),
    [sym_sigil] = STATE(4024),
    [sym_keywords] = STATE(1684),
    [sym_pair] = STATE(3214),
    [sym__keyword] = STATE(882),
    [sym_quoted_keyword] = STATE(882),
    [sym_list] = STATE(4024),
    [sym_tuple] = STATE(4024),
    [sym_bitstring] = STATE(4024),
    [sym_map] = STATE(4024),
    [sym__nullary_operator] = STATE(4024),
    [sym_unary_operator] = STATE(4024),
    [sym_binary_operator] = STATE(4024),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4024),
    [sym_call] = STATE(4024),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4024),
    [sym_anonymous_function] = STATE(4024),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1477),
    [sym_integer] = ACTIONS(1477),
    [sym_float] = ACTIONS(1477),
    [sym_char] = ACTIONS(1477),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(1477),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(1479),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(307)] = {
    [sym__expression] = STATE(1597),
    [sym_block] = STATE(1597),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1597),
    [sym_nil] = STATE(1597),
    [sym__atom] = STATE(1597),
    [sym_quoted_atom] = STATE(1597),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1597),
    [sym_charlist] = STATE(1597),
    [sym_sigil] = STATE(1597),
    [sym_keywords] = STATE(1538),
    [sym_pair] = STATE(1573),
    [sym__keyword] = STATE(895),
    [sym_quoted_keyword] = STATE(895),
    [sym_list] = STATE(1597),
    [sym_tuple] = STATE(1597),
    [sym_bitstring] = STATE(1597),
    [sym_map] = STATE(1597),
    [sym__nullary_operator] = STATE(1597),
    [sym_unary_operator] = STATE(1597),
    [sym_binary_operator] = STATE(1597),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1597),
    [sym_call] = STATE(1597),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1597),
    [sym_anonymous_function] = STATE(1597),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1481),
    [sym_integer] = ACTIONS(1481),
    [sym_float] = ACTIONS(1481),
    [sym_char] = ACTIONS(1481),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1481),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [sym_keyword] = ACTIONS(213),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(308)] = {
    [sym__expression] = STATE(2562),
    [sym_block] = STATE(2562),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2562),
    [sym_nil] = STATE(2562),
    [sym__atom] = STATE(2562),
    [sym_quoted_atom] = STATE(2562),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2562),
    [sym_charlist] = STATE(2562),
    [sym_sigil] = STATE(2562),
    [sym_keywords] = STATE(1201),
    [sym_pair] = STATE(2218),
    [sym__keyword] = STATE(845),
    [sym_quoted_keyword] = STATE(845),
    [sym_list] = STATE(2562),
    [sym_tuple] = STATE(2562),
    [sym_bitstring] = STATE(2562),
    [sym_map] = STATE(2562),
    [sym__nullary_operator] = STATE(2562),
    [sym_unary_operator] = STATE(2562),
    [sym_binary_operator] = STATE(2562),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2562),
    [sym_call] = STATE(2562),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2562),
    [sym_anonymous_function] = STATE(2562),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1483),
    [sym_integer] = ACTIONS(1483),
    [sym_float] = ACTIONS(1483),
    [sym_char] = ACTIONS(1483),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1483),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [sym_keyword] = ACTIONS(481),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(309)] = {
    [sym__expression] = STATE(4040),
    [sym_block] = STATE(4040),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4040),
    [sym_nil] = STATE(4040),
    [sym__atom] = STATE(4040),
    [sym_quoted_atom] = STATE(4040),
    [sym__quoted_i_double] = STATE(1512),
    [sym__quoted_i_single] = STATE(1552),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4040),
    [sym_charlist] = STATE(4040),
    [sym_sigil] = STATE(4040),
    [sym_keywords] = STATE(1685),
    [sym_pair] = STATE(3214),
    [sym__keyword] = STATE(882),
    [sym_quoted_keyword] = STATE(882),
    [sym_list] = STATE(4040),
    [sym_tuple] = STATE(4040),
    [sym_bitstring] = STATE(4040),
    [sym_map] = STATE(4040),
    [sym__nullary_operator] = STATE(4040),
    [sym_unary_operator] = STATE(4040),
    [sym_binary_operator] = STATE(4040),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4040),
    [sym_call] = STATE(4040),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4040),
    [sym_anonymous_function] = STATE(4040),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1485),
    [sym_integer] = ACTIONS(1485),
    [sym_float] = ACTIONS(1485),
    [sym_char] = ACTIONS(1485),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(1485),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [sym_keyword] = ACTIONS(1479),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(310)] = {
    [sym__terminator] = STATE(348),
    [sym__expression] = STATE(3265),
    [sym_block] = STATE(3265),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(3265),
    [sym_nil] = STATE(3265),
    [sym__atom] = STATE(3265),
    [sym_quoted_atom] = STATE(3265),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(3265),
    [sym_charlist] = STATE(3265),
    [sym_sigil] = STATE(3265),
    [sym_list] = STATE(3265),
    [sym_tuple] = STATE(3265),
    [sym_bitstring] = STATE(3265),
    [sym_map] = STATE(3265),
    [sym__nullary_operator] = STATE(3265),
    [sym_unary_operator] = STATE(3265),
    [sym_binary_operator] = STATE(3265),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(3265),
    [sym_call] = STATE(3265),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(3265),
    [sym_body] = STATE(4843),
    [sym_anonymous_function] = STATE(3265),
    [aux_sym__terminator_repeat1] = STATE(1032),
    [aux_sym__terminator_token1] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1471),
    [sym_integer] = ACTIONS(1471),
    [sym_float] = ACTIONS(1471),
    [sym_char] = ACTIONS(1471),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1471),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_end] = ACTIONS(1091),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(311)] = {
    [sym__expression] = STATE(3289),
    [sym_block] = STATE(3289),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3289),
    [sym_nil] = STATE(3289),
    [sym__atom] = STATE(3289),
    [sym_quoted_atom] = STATE(3289),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3289),
    [sym_charlist] = STATE(3289),
    [sym_sigil] = STATE(3289),
    [sym_keywords] = STATE(1462),
    [sym_pair] = STATE(2917),
    [sym__keyword] = STATE(887),
    [sym_quoted_keyword] = STATE(887),
    [sym_list] = STATE(3289),
    [sym_tuple] = STATE(3289),
    [sym_bitstring] = STATE(3289),
    [sym_map] = STATE(3289),
    [sym__nullary_operator] = STATE(3289),
    [sym_unary_operator] = STATE(3289),
    [sym_binary_operator] = STATE(3289),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3289),
    [sym_call] = STATE(3289),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3289),
    [sym_anonymous_function] = STATE(3289),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(1487),
    [sym_integer] = ACTIONS(1487),
    [sym_float] = ACTIONS(1487),
    [sym_char] = ACTIONS(1487),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1487),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [sym_keyword] = ACTIONS(514),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(312)] = {
    [sym__expression] = STATE(3289),
    [sym_block] = STATE(3289),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3289),
    [sym_nil] = STATE(3289),
    [sym__atom] = STATE(3289),
    [sym_quoted_atom] = STATE(3289),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3289),
    [sym_charlist] = STATE(3289),
    [sym_sigil] = STATE(3289),
    [sym_keywords] = STATE(1463),
    [sym_pair] = STATE(2917),
    [sym__keyword] = STATE(887),
    [sym_quoted_keyword] = STATE(887),
    [sym_list] = STATE(3289),
    [sym_tuple] = STATE(3289),
    [sym_bitstring] = STATE(3289),
    [sym_map] = STATE(3289),
    [sym__nullary_operator] = STATE(3289),
    [sym_unary_operator] = STATE(3289),
    [sym_binary_operator] = STATE(3289),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3289),
    [sym_call] = STATE(3289),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3289),
    [sym_anonymous_function] = STATE(3289),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(1487),
    [sym_integer] = ACTIONS(1487),
    [sym_float] = ACTIONS(1487),
    [sym_char] = ACTIONS(1487),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1487),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [sym_keyword] = ACTIONS(514),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(313)] = {
    [sym__expression] = STATE(3097),
    [sym_block] = STATE(3097),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(3097),
    [sym_nil] = STATE(3097),
    [sym__atom] = STATE(3097),
    [sym_quoted_atom] = STATE(3097),
    [sym__quoted_i_double] = STATE(2688),
    [sym__quoted_i_single] = STATE(2689),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(3097),
    [sym_charlist] = STATE(3097),
    [sym_sigil] = STATE(3097),
    [sym_keywords] = STATE(3227),
    [sym_pair] = STATE(2458),
    [sym__keyword] = STATE(843),
    [sym_quoted_keyword] = STATE(843),
    [sym_list] = STATE(3097),
    [sym_tuple] = STATE(3097),
    [sym_bitstring] = STATE(3097),
    [sym_map] = STATE(3097),
    [sym__nullary_operator] = STATE(3097),
    [sym_unary_operator] = STATE(3097),
    [sym_binary_operator] = STATE(3097),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(3097),
    [sym_call] = STATE(3097),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(3097),
    [sym_anonymous_function] = STATE(3097),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1489),
    [sym_integer] = ACTIONS(1489),
    [sym_float] = ACTIONS(1489),
    [sym_char] = ACTIONS(1489),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1489),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [sym_keyword] = ACTIONS(455),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(314)] = {
    [sym__expression] = STATE(4171),
    [sym_block] = STATE(4171),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4171),
    [sym_nil] = STATE(4171),
    [sym__atom] = STATE(4171),
    [sym_quoted_atom] = STATE(4171),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4171),
    [sym_charlist] = STATE(4171),
    [sym_sigil] = STATE(4171),
    [sym_keywords] = STATE(6205),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4171),
    [sym_tuple] = STATE(4171),
    [sym_bitstring] = STATE(4171),
    [sym_map] = STATE(4171),
    [sym__nullary_operator] = STATE(4171),
    [sym_unary_operator] = STATE(4171),
    [sym_binary_operator] = STATE(4171),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4171),
    [sym_call] = STATE(4171),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4171),
    [sym_anonymous_function] = STATE(4171),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1493),
    [sym_integer] = ACTIONS(1493),
    [sym_float] = ACTIONS(1493),
    [sym_char] = ACTIONS(1493),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1493),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(315)] = {
    [sym__expression] = STATE(3952),
    [sym_block] = STATE(3952),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3952),
    [sym_nil] = STATE(3952),
    [sym__atom] = STATE(3952),
    [sym_quoted_atom] = STATE(3952),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3952),
    [sym_charlist] = STATE(3952),
    [sym_sigil] = STATE(3952),
    [sym_keywords] = STATE(3078),
    [sym_pair] = STATE(3508),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3952),
    [sym_tuple] = STATE(3952),
    [sym_bitstring] = STATE(3952),
    [sym_map] = STATE(3952),
    [sym__nullary_operator] = STATE(3952),
    [sym_unary_operator] = STATE(3952),
    [sym_binary_operator] = STATE(3952),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3952),
    [sym_call] = STATE(3952),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3952),
    [sym_anonymous_function] = STATE(3952),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1495),
    [sym_integer] = ACTIONS(1495),
    [sym_float] = ACTIONS(1495),
    [sym_char] = ACTIONS(1495),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1495),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(316)] = {
    [sym__expression] = STATE(3968),
    [sym_block] = STATE(3968),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3968),
    [sym_nil] = STATE(3968),
    [sym__atom] = STATE(3968),
    [sym_quoted_atom] = STATE(3968),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3968),
    [sym_charlist] = STATE(3968),
    [sym_sigil] = STATE(3968),
    [sym_keywords] = STATE(3079),
    [sym_pair] = STATE(3508),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3968),
    [sym_tuple] = STATE(3968),
    [sym_bitstring] = STATE(3968),
    [sym_map] = STATE(3968),
    [sym__nullary_operator] = STATE(3968),
    [sym_unary_operator] = STATE(3968),
    [sym_binary_operator] = STATE(3968),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3968),
    [sym_call] = STATE(3968),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3968),
    [sym_anonymous_function] = STATE(3968),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1497),
    [sym_integer] = ACTIONS(1497),
    [sym_float] = ACTIONS(1497),
    [sym_char] = ACTIONS(1497),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1497),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(317)] = {
    [sym__expression] = STATE(4445),
    [sym_block] = STATE(4445),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4445),
    [sym_nil] = STATE(4445),
    [sym__atom] = STATE(4445),
    [sym_quoted_atom] = STATE(4445),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4445),
    [sym_charlist] = STATE(4445),
    [sym_sigil] = STATE(4445),
    [sym_keywords] = STATE(6825),
    [sym_pair] = STATE(5595),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4445),
    [sym_tuple] = STATE(4445),
    [sym_bitstring] = STATE(4445),
    [sym_map] = STATE(4445),
    [sym__nullary_operator] = STATE(4445),
    [sym_unary_operator] = STATE(4445),
    [sym_binary_operator] = STATE(4445),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4445),
    [sym_call] = STATE(4445),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4445),
    [sym_anonymous_function] = STATE(4445),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1367),
    [sym_integer] = ACTIONS(1367),
    [sym_float] = ACTIONS(1367),
    [sym_char] = ACTIONS(1367),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1367),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(318)] = {
    [sym__expression] = STATE(2478),
    [sym_block] = STATE(2478),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2478),
    [sym_nil] = STATE(2478),
    [sym__atom] = STATE(2478),
    [sym_quoted_atom] = STATE(2478),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2478),
    [sym_charlist] = STATE(2478),
    [sym_sigil] = STATE(2478),
    [sym_keywords] = STATE(1301),
    [sym_pair] = STATE(2218),
    [sym__keyword] = STATE(845),
    [sym_quoted_keyword] = STATE(845),
    [sym_list] = STATE(2478),
    [sym_tuple] = STATE(2478),
    [sym_bitstring] = STATE(2478),
    [sym_map] = STATE(2478),
    [sym__nullary_operator] = STATE(2478),
    [sym_unary_operator] = STATE(2478),
    [sym_binary_operator] = STATE(2478),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2478),
    [sym_call] = STATE(2478),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2478),
    [sym_anonymous_function] = STATE(2478),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1499),
    [sym_integer] = ACTIONS(1499),
    [sym_float] = ACTIONS(1499),
    [sym_char] = ACTIONS(1499),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1499),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [sym_keyword] = ACTIONS(481),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(319)] = {
    [sym__expression] = STATE(1380),
    [sym_block] = STATE(1380),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1380),
    [sym_nil] = STATE(1380),
    [sym__atom] = STATE(1380),
    [sym_quoted_atom] = STATE(1380),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1380),
    [sym_charlist] = STATE(1380),
    [sym_sigil] = STATE(1380),
    [sym_keywords] = STATE(1301),
    [sym_pair] = STATE(1349),
    [sym__keyword] = STATE(891),
    [sym_quoted_keyword] = STATE(891),
    [sym_list] = STATE(1380),
    [sym_tuple] = STATE(1380),
    [sym_bitstring] = STATE(1380),
    [sym_map] = STATE(1380),
    [sym__nullary_operator] = STATE(1380),
    [sym_unary_operator] = STATE(1380),
    [sym_binary_operator] = STATE(1380),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1380),
    [sym_call] = STATE(1380),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1380),
    [sym_anonymous_function] = STATE(1380),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1501),
    [sym_integer] = ACTIONS(1501),
    [sym_float] = ACTIONS(1501),
    [sym_char] = ACTIONS(1501),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1501),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [sym_keyword] = ACTIONS(262),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(320)] = {
    [sym__expression] = STATE(1380),
    [sym_block] = STATE(1380),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1380),
    [sym_nil] = STATE(1380),
    [sym__atom] = STATE(1380),
    [sym_quoted_atom] = STATE(1380),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1380),
    [sym_charlist] = STATE(1380),
    [sym_sigil] = STATE(1380),
    [sym_keywords] = STATE(1304),
    [sym_pair] = STATE(1349),
    [sym__keyword] = STATE(891),
    [sym_quoted_keyword] = STATE(891),
    [sym_list] = STATE(1380),
    [sym_tuple] = STATE(1380),
    [sym_bitstring] = STATE(1380),
    [sym_map] = STATE(1380),
    [sym__nullary_operator] = STATE(1380),
    [sym_unary_operator] = STATE(1380),
    [sym_binary_operator] = STATE(1380),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1380),
    [sym_call] = STATE(1380),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1380),
    [sym_anonymous_function] = STATE(1380),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1501),
    [sym_integer] = ACTIONS(1501),
    [sym_float] = ACTIONS(1501),
    [sym_char] = ACTIONS(1501),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1501),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [sym_keyword] = ACTIONS(262),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(321)] = {
    [sym__expression] = STATE(2478),
    [sym_block] = STATE(2478),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2478),
    [sym_nil] = STATE(2478),
    [sym__atom] = STATE(2478),
    [sym_quoted_atom] = STATE(2478),
    [sym__quoted_i_double] = STATE(1105),
    [sym__quoted_i_single] = STATE(1124),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2478),
    [sym_charlist] = STATE(2478),
    [sym_sigil] = STATE(2478),
    [sym_keywords] = STATE(1304),
    [sym_pair] = STATE(2218),
    [sym__keyword] = STATE(845),
    [sym_quoted_keyword] = STATE(845),
    [sym_list] = STATE(2478),
    [sym_tuple] = STATE(2478),
    [sym_bitstring] = STATE(2478),
    [sym_map] = STATE(2478),
    [sym__nullary_operator] = STATE(2478),
    [sym_unary_operator] = STATE(2478),
    [sym_binary_operator] = STATE(2478),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2478),
    [sym_call] = STATE(2478),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2478),
    [sym_anonymous_function] = STATE(2478),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1499),
    [sym_integer] = ACTIONS(1499),
    [sym_float] = ACTIONS(1499),
    [sym_char] = ACTIONS(1499),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1499),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [sym_keyword] = ACTIONS(481),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(322)] = {
    [sym__expression] = STATE(4451),
    [sym_block] = STATE(4451),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4451),
    [sym_nil] = STATE(4451),
    [sym__atom] = STATE(4451),
    [sym_quoted_atom] = STATE(4451),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4451),
    [sym_charlist] = STATE(4451),
    [sym_sigil] = STATE(4451),
    [sym_keywords] = STATE(3078),
    [sym_pair] = STATE(4398),
    [sym__keyword] = STATE(893),
    [sym_quoted_keyword] = STATE(893),
    [sym_list] = STATE(4451),
    [sym_tuple] = STATE(4451),
    [sym_bitstring] = STATE(4451),
    [sym_map] = STATE(4451),
    [sym__nullary_operator] = STATE(4451),
    [sym_unary_operator] = STATE(4451),
    [sym_binary_operator] = STATE(4451),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4451),
    [sym_call] = STATE(4451),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4451),
    [sym_anonymous_function] = STATE(4451),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1503),
    [sym_integer] = ACTIONS(1503),
    [sym_float] = ACTIONS(1503),
    [sym_char] = ACTIONS(1503),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1503),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1505),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(323)] = {
    [sym__expression] = STATE(4479),
    [sym_block] = STATE(4479),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4479),
    [sym_nil] = STATE(4479),
    [sym__atom] = STATE(4479),
    [sym_quoted_atom] = STATE(4479),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4479),
    [sym_charlist] = STATE(4479),
    [sym_sigil] = STATE(4479),
    [sym_keywords] = STATE(3079),
    [sym_pair] = STATE(4398),
    [sym__keyword] = STATE(893),
    [sym_quoted_keyword] = STATE(893),
    [sym_list] = STATE(4479),
    [sym_tuple] = STATE(4479),
    [sym_bitstring] = STATE(4479),
    [sym_map] = STATE(4479),
    [sym__nullary_operator] = STATE(4479),
    [sym_unary_operator] = STATE(4479),
    [sym_binary_operator] = STATE(4479),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4479),
    [sym_call] = STATE(4479),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4479),
    [sym_anonymous_function] = STATE(4479),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1515),
    [sym_integer] = ACTIONS(1515),
    [sym_float] = ACTIONS(1515),
    [sym_char] = ACTIONS(1515),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1505),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(324)] = {
    [sym__expression] = STATE(3762),
    [sym_block] = STATE(3762),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3762),
    [sym_nil] = STATE(3762),
    [sym__atom] = STATE(3762),
    [sym_quoted_atom] = STATE(3762),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3762),
    [sym_charlist] = STATE(3762),
    [sym_sigil] = STATE(3762),
    [sym_keywords] = STATE(2312),
    [sym_pair] = STATE(3499),
    [sym__keyword] = STATE(885),
    [sym_quoted_keyword] = STATE(885),
    [sym_list] = STATE(3762),
    [sym_tuple] = STATE(3762),
    [sym_bitstring] = STATE(3762),
    [sym_map] = STATE(3762),
    [sym__nullary_operator] = STATE(3762),
    [sym_unary_operator] = STATE(3762),
    [sym_binary_operator] = STATE(3762),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3762),
    [sym_call] = STATE(3762),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3762),
    [sym_anonymous_function] = STATE(3762),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1517),
    [sym_integer] = ACTIONS(1517),
    [sym_float] = ACTIONS(1517),
    [sym_char] = ACTIONS(1517),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1517),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [sym_keyword] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(325)] = {
    [sym__expression] = STATE(3767),
    [sym_block] = STATE(3767),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3767),
    [sym_nil] = STATE(3767),
    [sym__atom] = STATE(3767),
    [sym_quoted_atom] = STATE(3767),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3767),
    [sym_charlist] = STATE(3767),
    [sym_sigil] = STATE(3767),
    [sym_keywords] = STATE(2313),
    [sym_pair] = STATE(3499),
    [sym__keyword] = STATE(885),
    [sym_quoted_keyword] = STATE(885),
    [sym_list] = STATE(3767),
    [sym_tuple] = STATE(3767),
    [sym_bitstring] = STATE(3767),
    [sym_map] = STATE(3767),
    [sym__nullary_operator] = STATE(3767),
    [sym_unary_operator] = STATE(3767),
    [sym_binary_operator] = STATE(3767),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3767),
    [sym_call] = STATE(3767),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3767),
    [sym_anonymous_function] = STATE(3767),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1519),
    [sym_integer] = ACTIONS(1519),
    [sym_float] = ACTIONS(1519),
    [sym_char] = ACTIONS(1519),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [sym_keyword] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(326)] = {
    [sym__expression] = STATE(1688),
    [sym_block] = STATE(1688),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1688),
    [sym_nil] = STATE(1688),
    [sym__atom] = STATE(1688),
    [sym_quoted_atom] = STATE(1688),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1688),
    [sym_charlist] = STATE(1688),
    [sym_sigil] = STATE(1688),
    [sym_keywords] = STATE(1462),
    [sym_pair] = STATE(1573),
    [sym__keyword] = STATE(895),
    [sym_quoted_keyword] = STATE(895),
    [sym_list] = STATE(1688),
    [sym_tuple] = STATE(1688),
    [sym_bitstring] = STATE(1688),
    [sym_map] = STATE(1688),
    [sym__nullary_operator] = STATE(1688),
    [sym_unary_operator] = STATE(1688),
    [sym_binary_operator] = STATE(1688),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1688),
    [sym_call] = STATE(1688),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1688),
    [sym_anonymous_function] = STATE(1688),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1521),
    [sym_integer] = ACTIONS(1521),
    [sym_float] = ACTIONS(1521),
    [sym_char] = ACTIONS(1521),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1521),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [sym_keyword] = ACTIONS(213),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(327)] = {
    [sym__terminator] = STATE(352),
    [sym__expression] = STATE(3212),
    [sym_block] = STATE(3212),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(3212),
    [sym_nil] = STATE(3212),
    [sym__atom] = STATE(3212),
    [sym_quoted_atom] = STATE(3212),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(3212),
    [sym_charlist] = STATE(3212),
    [sym_sigil] = STATE(3212),
    [sym_list] = STATE(3212),
    [sym_tuple] = STATE(3212),
    [sym_bitstring] = STATE(3212),
    [sym_map] = STATE(3212),
    [sym__nullary_operator] = STATE(3212),
    [sym_unary_operator] = STATE(3212),
    [sym_binary_operator] = STATE(3212),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(3212),
    [sym_call] = STATE(3212),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(3212),
    [sym_body] = STATE(4842),
    [sym_anonymous_function] = STATE(3212),
    [aux_sym__terminator_repeat1] = STATE(1036),
    [aux_sym__terminator_token1] = ACTIONS(1523),
    [anon_sym_SEMI] = ACTIONS(1525),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1089),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1527),
    [sym_integer] = ACTIONS(1527),
    [sym_float] = ACTIONS(1527),
    [sym_char] = ACTIONS(1527),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1527),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(328)] = {
    [sym__expression] = STATE(1688),
    [sym_block] = STATE(1688),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1688),
    [sym_nil] = STATE(1688),
    [sym__atom] = STATE(1688),
    [sym_quoted_atom] = STATE(1688),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1688),
    [sym_charlist] = STATE(1688),
    [sym_sigil] = STATE(1688),
    [sym_keywords] = STATE(1463),
    [sym_pair] = STATE(1573),
    [sym__keyword] = STATE(895),
    [sym_quoted_keyword] = STATE(895),
    [sym_list] = STATE(1688),
    [sym_tuple] = STATE(1688),
    [sym_bitstring] = STATE(1688),
    [sym_map] = STATE(1688),
    [sym__nullary_operator] = STATE(1688),
    [sym_unary_operator] = STATE(1688),
    [sym_binary_operator] = STATE(1688),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1688),
    [sym_call] = STATE(1688),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1688),
    [sym_anonymous_function] = STATE(1688),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1521),
    [sym_integer] = ACTIONS(1521),
    [sym_float] = ACTIONS(1521),
    [sym_char] = ACTIONS(1521),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1521),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [sym_keyword] = ACTIONS(213),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(329)] = {
    [sym__terminator] = STATE(352),
    [sym__expression] = STATE(3212),
    [sym_block] = STATE(3212),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(3212),
    [sym_nil] = STATE(3212),
    [sym__atom] = STATE(3212),
    [sym_quoted_atom] = STATE(3212),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(3212),
    [sym_charlist] = STATE(3212),
    [sym_sigil] = STATE(3212),
    [sym_list] = STATE(3212),
    [sym_tuple] = STATE(3212),
    [sym_bitstring] = STATE(3212),
    [sym_map] = STATE(3212),
    [sym__nullary_operator] = STATE(3212),
    [sym_unary_operator] = STATE(3212),
    [sym_binary_operator] = STATE(3212),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(3212),
    [sym_call] = STATE(3212),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(3212),
    [sym_body] = STATE(4843),
    [sym_anonymous_function] = STATE(3212),
    [aux_sym__terminator_repeat1] = STATE(1036),
    [aux_sym__terminator_token1] = ACTIONS(1523),
    [anon_sym_SEMI] = ACTIONS(1525),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1091),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1527),
    [sym_integer] = ACTIONS(1527),
    [sym_float] = ACTIONS(1527),
    [sym_char] = ACTIONS(1527),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1527),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(330)] = {
    [sym__expression] = STATE(4468),
    [sym_block] = STATE(4468),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4468),
    [sym_nil] = STATE(4468),
    [sym__atom] = STATE(4468),
    [sym_quoted_atom] = STATE(4468),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4468),
    [sym_charlist] = STATE(4468),
    [sym_sigil] = STATE(4468),
    [sym_keywords] = STATE(4478),
    [sym_pair] = STATE(4245),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4468),
    [sym_tuple] = STATE(4468),
    [sym_bitstring] = STATE(4468),
    [sym_map] = STATE(4468),
    [sym__nullary_operator] = STATE(4468),
    [sym_unary_operator] = STATE(4468),
    [sym_binary_operator] = STATE(4468),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4468),
    [sym_call] = STATE(4468),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4468),
    [sym_anonymous_function] = STATE(4468),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1539),
    [sym_integer] = ACTIONS(1539),
    [sym_float] = ACTIONS(1539),
    [sym_char] = ACTIONS(1539),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1539),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(331)] = {
    [sym__expression] = STATE(4497),
    [sym_block] = STATE(4497),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4497),
    [sym_nil] = STATE(4497),
    [sym__atom] = STATE(4497),
    [sym_quoted_atom] = STATE(4497),
    [sym__quoted_i_double] = STATE(4118),
    [sym__quoted_i_single] = STATE(4122),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4497),
    [sym_charlist] = STATE(4497),
    [sym_sigil] = STATE(4497),
    [sym_keywords] = STATE(4498),
    [sym_pair] = STATE(4245),
    [sym__keyword] = STATE(841),
    [sym_quoted_keyword] = STATE(841),
    [sym_list] = STATE(4497),
    [sym_tuple] = STATE(4497),
    [sym_bitstring] = STATE(4497),
    [sym_map] = STATE(4497),
    [sym__nullary_operator] = STATE(4497),
    [sym_unary_operator] = STATE(4497),
    [sym_binary_operator] = STATE(4497),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4497),
    [sym_call] = STATE(4497),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4497),
    [sym_anonymous_function] = STATE(4497),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1541),
    [sym_integer] = ACTIONS(1541),
    [sym_float] = ACTIONS(1541),
    [sym_char] = ACTIONS(1541),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1541),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [sym_keyword] = ACTIONS(1181),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(332)] = {
    [sym__expression] = STATE(3830),
    [sym_block] = STATE(3830),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3830),
    [sym_nil] = STATE(3830),
    [sym__atom] = STATE(3830),
    [sym_quoted_atom] = STATE(3830),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3830),
    [sym_charlist] = STATE(3830),
    [sym_sigil] = STATE(3830),
    [sym__keywords_with_trailing_separator] = STATE(6886),
    [sym_pair] = STATE(5578),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(3830),
    [sym_tuple] = STATE(3830),
    [sym_bitstring] = STATE(3830),
    [sym_map] = STATE(3830),
    [sym__nullary_operator] = STATE(3830),
    [sym_unary_operator] = STATE(3830),
    [sym_binary_operator] = STATE(3830),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3830),
    [sym_call] = STATE(3830),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3830),
    [sym_anonymous_function] = STATE(3830),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1223),
    [sym_integer] = ACTIONS(1223),
    [sym_float] = ACTIONS(1223),
    [sym_char] = ACTIONS(1223),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1223),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(333)] = {
    [sym__expression] = STATE(4235),
    [sym_block] = STATE(4235),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4235),
    [sym_nil] = STATE(4235),
    [sym__atom] = STATE(4235),
    [sym_quoted_atom] = STATE(4235),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4235),
    [sym_charlist] = STATE(4235),
    [sym_sigil] = STATE(4235),
    [sym_keywords] = STATE(4237),
    [sym_pair] = STATE(4048),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4235),
    [sym_tuple] = STATE(4235),
    [sym_bitstring] = STATE(4235),
    [sym_map] = STATE(4235),
    [sym__nullary_operator] = STATE(4235),
    [sym_unary_operator] = STATE(4235),
    [sym_binary_operator] = STATE(4235),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4235),
    [sym_call] = STATE(4235),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4235),
    [sym_anonymous_function] = STATE(4235),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1543),
    [sym_integer] = ACTIONS(1543),
    [sym_float] = ACTIONS(1543),
    [sym_char] = ACTIONS(1543),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1543),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(334)] = {
    [sym__expression] = STATE(4248),
    [sym_block] = STATE(4248),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4248),
    [sym_nil] = STATE(4248),
    [sym__atom] = STATE(4248),
    [sym_quoted_atom] = STATE(4248),
    [sym__quoted_i_double] = STATE(4045),
    [sym__quoted_i_single] = STATE(4046),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4248),
    [sym_charlist] = STATE(4248),
    [sym_sigil] = STATE(4248),
    [sym_keywords] = STATE(4258),
    [sym_pair] = STATE(4048),
    [sym__keyword] = STATE(849),
    [sym_quoted_keyword] = STATE(849),
    [sym_list] = STATE(4248),
    [sym_tuple] = STATE(4248),
    [sym_bitstring] = STATE(4248),
    [sym_map] = STATE(4248),
    [sym__nullary_operator] = STATE(4248),
    [sym_unary_operator] = STATE(4248),
    [sym_binary_operator] = STATE(4248),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4248),
    [sym_call] = STATE(4248),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4248),
    [sym_anonymous_function] = STATE(4248),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1545),
    [sym_integer] = ACTIONS(1545),
    [sym_float] = ACTIONS(1545),
    [sym_char] = ACTIONS(1545),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1545),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [sym_keyword] = ACTIONS(87),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(335)] = {
    [sym__expression] = STATE(2716),
    [sym_block] = STATE(2716),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2716),
    [sym_nil] = STATE(2716),
    [sym__atom] = STATE(2716),
    [sym_quoted_atom] = STATE(2716),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2716),
    [sym_charlist] = STATE(2716),
    [sym_sigil] = STATE(2716),
    [sym_keywords] = STATE(2304),
    [sym_pair] = STATE(2158),
    [sym__keyword] = STATE(848),
    [sym_quoted_keyword] = STATE(848),
    [sym_list] = STATE(2716),
    [sym_tuple] = STATE(2716),
    [sym_bitstring] = STATE(2716),
    [sym_map] = STATE(2716),
    [sym__nullary_operator] = STATE(2716),
    [sym_unary_operator] = STATE(2716),
    [sym_binary_operator] = STATE(2716),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2716),
    [sym_call] = STATE(2716),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2716),
    [sym_anonymous_function] = STATE(2716),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1547),
    [sym_integer] = ACTIONS(1547),
    [sym_float] = ACTIONS(1547),
    [sym_char] = ACTIONS(1547),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1547),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [sym_keyword] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(336)] = {
    [sym__expression] = STATE(3097),
    [sym_block] = STATE(3097),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(3097),
    [sym_nil] = STATE(3097),
    [sym__atom] = STATE(3097),
    [sym_quoted_atom] = STATE(3097),
    [sym__quoted_i_double] = STATE(2688),
    [sym__quoted_i_single] = STATE(2689),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(3097),
    [sym_charlist] = STATE(3097),
    [sym_sigil] = STATE(3097),
    [sym_keywords] = STATE(2937),
    [sym_pair] = STATE(2458),
    [sym__keyword] = STATE(843),
    [sym_quoted_keyword] = STATE(843),
    [sym_list] = STATE(3097),
    [sym_tuple] = STATE(3097),
    [sym_bitstring] = STATE(3097),
    [sym_map] = STATE(3097),
    [sym__nullary_operator] = STATE(3097),
    [sym_unary_operator] = STATE(3097),
    [sym_binary_operator] = STATE(3097),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(3097),
    [sym_call] = STATE(3097),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(3097),
    [sym_anonymous_function] = STATE(3097),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1489),
    [sym_integer] = ACTIONS(1489),
    [sym_float] = ACTIONS(1489),
    [sym_char] = ACTIONS(1489),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1489),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [sym_keyword] = ACTIONS(455),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(337)] = {
    [sym__expression] = STATE(2716),
    [sym_block] = STATE(2716),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2716),
    [sym_nil] = STATE(2716),
    [sym__atom] = STATE(2716),
    [sym_quoted_atom] = STATE(2716),
    [sym__quoted_i_double] = STATE(1883),
    [sym__quoted_i_single] = STATE(1884),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2716),
    [sym_charlist] = STATE(2716),
    [sym_sigil] = STATE(2716),
    [sym_keywords] = STATE(2356),
    [sym_pair] = STATE(2158),
    [sym__keyword] = STATE(848),
    [sym_quoted_keyword] = STATE(848),
    [sym_list] = STATE(2716),
    [sym_tuple] = STATE(2716),
    [sym_bitstring] = STATE(2716),
    [sym_map] = STATE(2716),
    [sym__nullary_operator] = STATE(2716),
    [sym_unary_operator] = STATE(2716),
    [sym_binary_operator] = STATE(2716),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2716),
    [sym_call] = STATE(2716),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2716),
    [sym_anonymous_function] = STATE(2716),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1547),
    [sym_integer] = ACTIONS(1547),
    [sym_float] = ACTIONS(1547),
    [sym_char] = ACTIONS(1547),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1547),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [sym_keyword] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(338)] = {
    [sym__expression] = STATE(4452),
    [sym_block] = STATE(4452),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4452),
    [sym_nil] = STATE(4452),
    [sym__atom] = STATE(4452),
    [sym_quoted_atom] = STATE(4452),
    [sym__quoted_i_double] = STATE(2454),
    [sym__quoted_i_single] = STATE(2457),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4452),
    [sym_charlist] = STATE(4452),
    [sym_sigil] = STATE(4452),
    [sym_keywords] = STATE(6898),
    [sym_pair] = STATE(6700),
    [sym__keyword] = STATE(792),
    [sym_quoted_keyword] = STATE(792),
    [sym_list] = STATE(4452),
    [sym_tuple] = STATE(4452),
    [sym_bitstring] = STATE(4452),
    [sym_map] = STATE(4452),
    [sym__nullary_operator] = STATE(4452),
    [sym_unary_operator] = STATE(4452),
    [sym_binary_operator] = STATE(4452),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4452),
    [sym_call] = STATE(4452),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4452),
    [sym_anonymous_function] = STATE(4452),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1549),
    [sym_integer] = ACTIONS(1549),
    [sym_float] = ACTIONS(1549),
    [sym_char] = ACTIONS(1549),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1549),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [sym_keyword] = ACTIONS(1117),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(339)] = {
    [sym__expression] = STATE(2960),
    [sym_block] = STATE(2960),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2960),
    [sym_nil] = STATE(2960),
    [sym__atom] = STATE(2960),
    [sym_quoted_atom] = STATE(2960),
    [sym__quoted_i_double] = STATE(2688),
    [sym__quoted_i_single] = STATE(2689),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2960),
    [sym_charlist] = STATE(2960),
    [sym_sigil] = STATE(2960),
    [sym_keywords] = STATE(2964),
    [sym_pair] = STATE(2458),
    [sym__keyword] = STATE(843),
    [sym_quoted_keyword] = STATE(843),
    [sym_list] = STATE(2960),
    [sym_tuple] = STATE(2960),
    [sym_bitstring] = STATE(2960),
    [sym_map] = STATE(2960),
    [sym__nullary_operator] = STATE(2960),
    [sym_unary_operator] = STATE(2960),
    [sym_binary_operator] = STATE(2960),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2960),
    [sym_call] = STATE(2960),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2960),
    [sym_anonymous_function] = STATE(2960),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1551),
    [sym_integer] = ACTIONS(1551),
    [sym_float] = ACTIONS(1551),
    [sym_char] = ACTIONS(1551),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1551),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [sym_keyword] = ACTIONS(455),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(340)] = {
    [sym__expression] = STATE(2984),
    [sym_block] = STATE(2984),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2984),
    [sym_nil] = STATE(2984),
    [sym__atom] = STATE(2984),
    [sym_quoted_atom] = STATE(2984),
    [sym__quoted_i_double] = STATE(2688),
    [sym__quoted_i_single] = STATE(2689),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2984),
    [sym_charlist] = STATE(2984),
    [sym_sigil] = STATE(2984),
    [sym_keywords] = STATE(2985),
    [sym_pair] = STATE(2458),
    [sym__keyword] = STATE(843),
    [sym_quoted_keyword] = STATE(843),
    [sym_list] = STATE(2984),
    [sym_tuple] = STATE(2984),
    [sym_bitstring] = STATE(2984),
    [sym_map] = STATE(2984),
    [sym__nullary_operator] = STATE(2984),
    [sym_unary_operator] = STATE(2984),
    [sym_binary_operator] = STATE(2984),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2984),
    [sym_call] = STATE(2984),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2984),
    [sym_anonymous_function] = STATE(2984),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1553),
    [sym_integer] = ACTIONS(1553),
    [sym_float] = ACTIONS(1553),
    [sym_char] = ACTIONS(1553),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1553),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [sym_keyword] = ACTIONS(455),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(341)] = {
    [sym__expression] = STATE(4067),
    [sym_block] = STATE(4067),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4067),
    [sym_nil] = STATE(4067),
    [sym__atom] = STATE(4067),
    [sym_quoted_atom] = STATE(4067),
    [sym__quoted_i_double] = STATE(1613),
    [sym__quoted_i_single] = STATE(1614),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4067),
    [sym_charlist] = STATE(4067),
    [sym_sigil] = STATE(4067),
    [sym_keywords] = STATE(2036),
    [sym_pair] = STATE(3755),
    [sym__keyword] = STATE(892),
    [sym_quoted_keyword] = STATE(892),
    [sym_list] = STATE(4067),
    [sym_tuple] = STATE(4067),
    [sym_bitstring] = STATE(4067),
    [sym_map] = STATE(4067),
    [sym__nullary_operator] = STATE(4067),
    [sym_unary_operator] = STATE(4067),
    [sym_binary_operator] = STATE(4067),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4067),
    [sym_call] = STATE(4067),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4067),
    [sym_anonymous_function] = STATE(4067),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1555),
    [sym_integer] = ACTIONS(1555),
    [sym_float] = ACTIONS(1555),
    [sym_char] = ACTIONS(1555),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1555),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [sym_keyword] = ACTIONS(1557),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(342)] = {
    [sym__expression] = STATE(4072),
    [sym_block] = STATE(4072),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4072),
    [sym_nil] = STATE(4072),
    [sym__atom] = STATE(4072),
    [sym_quoted_atom] = STATE(4072),
    [sym__quoted_i_double] = STATE(1613),
    [sym__quoted_i_single] = STATE(1614),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4072),
    [sym_charlist] = STATE(4072),
    [sym_sigil] = STATE(4072),
    [sym_keywords] = STATE(2037),
    [sym_pair] = STATE(3755),
    [sym__keyword] = STATE(892),
    [sym_quoted_keyword] = STATE(892),
    [sym_list] = STATE(4072),
    [sym_tuple] = STATE(4072),
    [sym_bitstring] = STATE(4072),
    [sym_map] = STATE(4072),
    [sym__nullary_operator] = STATE(4072),
    [sym_unary_operator] = STATE(4072),
    [sym_binary_operator] = STATE(4072),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4072),
    [sym_call] = STATE(4072),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4072),
    [sym_anonymous_function] = STATE(4072),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1559),
    [sym_integer] = ACTIONS(1559),
    [sym_float] = ACTIONS(1559),
    [sym_char] = ACTIONS(1559),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1559),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [sym_keyword] = ACTIONS(1557),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(343)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1561),
    [anon_sym_catch] = ACTIONS(1561),
    [anon_sym_else] = ACTIONS(1561),
    [anon_sym_end] = ACTIONS(1561),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1561),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(344)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_after] = ACTIONS(1563),
    [anon_sym_catch] = ACTIONS(1563),
    [anon_sym_else] = ACTIONS(1563),
    [anon_sym_end] = ACTIONS(1563),
    [anon_sym_fn] = ACTIONS(1004),
    [anon_sym_rescue] = ACTIONS(1563),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(345)] = {
    [sym__expression] = STATE(1602),
    [sym_block] = STATE(1602),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1602),
    [sym_nil] = STATE(1602),
    [sym__atom] = STATE(1602),
    [sym_quoted_atom] = STATE(1602),
    [sym__quoted_i_double] = STATE(1253),
    [sym__quoted_i_single] = STATE(1255),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1602),
    [sym_charlist] = STATE(1602),
    [sym_sigil] = STATE(1602),
    [sym_keywords] = STATE(1540),
    [sym_pair] = STATE(1573),
    [sym__keyword] = STATE(895),
    [sym_quoted_keyword] = STATE(895),
    [sym_list] = STATE(1602),
    [sym_tuple] = STATE(1602),
    [sym_bitstring] = STATE(1602),
    [sym_map] = STATE(1602),
    [sym__nullary_operator] = STATE(1602),
    [sym_unary_operator] = STATE(1602),
    [sym_binary_operator] = STATE(1602),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1602),
    [sym_call] = STATE(1602),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1602),
    [sym_anonymous_function] = STATE(1602),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1565),
    [sym_integer] = ACTIONS(1565),
    [sym_float] = ACTIONS(1565),
    [sym_char] = ACTIONS(1565),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1565),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [sym_keyword] = ACTIONS(213),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(346)] = {
    [sym__expression] = STATE(4446),
    [sym_block] = STATE(4446),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4446),
    [sym_nil] = STATE(4446),
    [sym__atom] = STATE(4446),
    [sym_quoted_atom] = STATE(4446),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4446),
    [sym_charlist] = STATE(4446),
    [sym_sigil] = STATE(4446),
    [sym_list] = STATE(4446),
    [sym_tuple] = STATE(4446),
    [sym_bitstring] = STATE(4446),
    [sym_map] = STATE(4446),
    [sym__nullary_operator] = STATE(4446),
    [sym_unary_operator] = STATE(4446),
    [sym_binary_operator] = STATE(4446),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4446),
    [sym_call] = STATE(4446),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4446),
    [sym_anonymous_function] = STATE(4446),
    [aux_sym__terminator_token1] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1361),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1567),
    [sym_integer] = ACTIONS(1567),
    [sym_float] = ACTIONS(1567),
    [sym_char] = ACTIONS(1567),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1567),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_end] = ACTIONS(1361),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(347)] = {
    [sym__expression] = STATE(4485),
    [sym_block] = STATE(4485),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4485),
    [sym_nil] = STATE(4485),
    [sym__atom] = STATE(4485),
    [sym_quoted_atom] = STATE(4485),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4485),
    [sym_charlist] = STATE(4485),
    [sym_sigil] = STATE(4485),
    [sym_list] = STATE(4485),
    [sym_tuple] = STATE(4485),
    [sym_bitstring] = STATE(4485),
    [sym_map] = STATE(4485),
    [sym__nullary_operator] = STATE(4485),
    [sym_unary_operator] = STATE(4485),
    [sym_binary_operator] = STATE(4485),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4485),
    [sym_call] = STATE(4485),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4485),
    [sym_anonymous_function] = STATE(4485),
    [aux_sym__terminator_token1] = ACTIONS(1343),
    [anon_sym_SEMI] = ACTIONS(1345),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1345),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1569),
    [sym_integer] = ACTIONS(1569),
    [sym_float] = ACTIONS(1569),
    [sym_char] = ACTIONS(1569),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1569),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(348)] = {
    [sym__expression] = STATE(3269),
    [sym_block] = STATE(3269),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(3269),
    [sym_nil] = STATE(3269),
    [sym__atom] = STATE(3269),
    [sym_quoted_atom] = STATE(3269),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(3269),
    [sym_charlist] = STATE(3269),
    [sym_sigil] = STATE(3269),
    [sym_list] = STATE(3269),
    [sym_tuple] = STATE(3269),
    [sym_bitstring] = STATE(3269),
    [sym_map] = STATE(3269),
    [sym__nullary_operator] = STATE(3269),
    [sym_unary_operator] = STATE(3269),
    [sym_binary_operator] = STATE(3269),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(3269),
    [sym_call] = STATE(3269),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(3269),
    [sym_anonymous_function] = STATE(3269),
    [aux_sym__terminator_token1] = ACTIONS(1353),
    [anon_sym_SEMI] = ACTIONS(1355),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1571),
    [sym_integer] = ACTIONS(1571),
    [sym_float] = ACTIONS(1571),
    [sym_char] = ACTIONS(1571),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1571),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_end] = ACTIONS(1355),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(349)] = {
    [sym__expression] = STATE(4446),
    [sym_block] = STATE(4446),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4446),
    [sym_nil] = STATE(4446),
    [sym__atom] = STATE(4446),
    [sym_quoted_atom] = STATE(4446),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4446),
    [sym_charlist] = STATE(4446),
    [sym_sigil] = STATE(4446),
    [sym_list] = STATE(4446),
    [sym_tuple] = STATE(4446),
    [sym_bitstring] = STATE(4446),
    [sym_map] = STATE(4446),
    [sym__nullary_operator] = STATE(4446),
    [sym_unary_operator] = STATE(4446),
    [sym_binary_operator] = STATE(4446),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4446),
    [sym_call] = STATE(4446),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4446),
    [sym_anonymous_function] = STATE(4446),
    [aux_sym__terminator_token1] = ACTIONS(1349),
    [anon_sym_SEMI] = ACTIONS(1351),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1567),
    [sym_integer] = ACTIONS(1567),
    [sym_float] = ACTIONS(1567),
    [sym_char] = ACTIONS(1567),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1567),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_end] = ACTIONS(1351),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(350)] = {
    [sym__expression] = STATE(4446),
    [sym_block] = STATE(4446),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4446),
    [sym_nil] = STATE(4446),
    [sym__atom] = STATE(4446),
    [sym_quoted_atom] = STATE(4446),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4446),
    [sym_charlist] = STATE(4446),
    [sym_sigil] = STATE(4446),
    [sym_list] = STATE(4446),
    [sym_tuple] = STATE(4446),
    [sym_bitstring] = STATE(4446),
    [sym_map] = STATE(4446),
    [sym__nullary_operator] = STATE(4446),
    [sym_unary_operator] = STATE(4446),
    [sym_binary_operator] = STATE(4446),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4446),
    [sym_call] = STATE(4446),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4446),
    [sym_anonymous_function] = STATE(4446),
    [aux_sym__terminator_token1] = ACTIONS(1343),
    [anon_sym_SEMI] = ACTIONS(1345),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1567),
    [sym_integer] = ACTIONS(1567),
    [sym_float] = ACTIONS(1567),
    [sym_char] = ACTIONS(1567),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1567),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_end] = ACTIONS(1345),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(351)] = {
    [sym__expression] = STATE(4485),
    [sym_block] = STATE(4485),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4485),
    [sym_nil] = STATE(4485),
    [sym__atom] = STATE(4485),
    [sym_quoted_atom] = STATE(4485),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4485),
    [sym_charlist] = STATE(4485),
    [sym_sigil] = STATE(4485),
    [sym_list] = STATE(4485),
    [sym_tuple] = STATE(4485),
    [sym_bitstring] = STATE(4485),
    [sym_map] = STATE(4485),
    [sym__nullary_operator] = STATE(4485),
    [sym_unary_operator] = STATE(4485),
    [sym_binary_operator] = STATE(4485),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4485),
    [sym_call] = STATE(4485),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4485),
    [sym_anonymous_function] = STATE(4485),
    [aux_sym__terminator_token1] = ACTIONS(1349),
    [anon_sym_SEMI] = ACTIONS(1351),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1351),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1569),
    [sym_integer] = ACTIONS(1569),
    [sym_float] = ACTIONS(1569),
    [sym_char] = ACTIONS(1569),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1569),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(352)] = {
    [sym__expression] = STATE(3108),
    [sym_block] = STATE(3108),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(3108),
    [sym_nil] = STATE(3108),
    [sym__atom] = STATE(3108),
    [sym_quoted_atom] = STATE(3108),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(3108),
    [sym_charlist] = STATE(3108),
    [sym_sigil] = STATE(3108),
    [sym_list] = STATE(3108),
    [sym_tuple] = STATE(3108),
    [sym_bitstring] = STATE(3108),
    [sym_map] = STATE(3108),
    [sym__nullary_operator] = STATE(3108),
    [sym_unary_operator] = STATE(3108),
    [sym_binary_operator] = STATE(3108),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(3108),
    [sym_call] = STATE(3108),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(3108),
    [sym_anonymous_function] = STATE(3108),
    [aux_sym__terminator_token1] = ACTIONS(1353),
    [anon_sym_SEMI] = ACTIONS(1355),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1355),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1573),
    [sym_integer] = ACTIONS(1573),
    [sym_float] = ACTIONS(1573),
    [sym_char] = ACTIONS(1573),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1573),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(353)] = {
    [sym__expression] = STATE(4485),
    [sym_block] = STATE(4485),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4485),
    [sym_nil] = STATE(4485),
    [sym__atom] = STATE(4485),
    [sym_quoted_atom] = STATE(4485),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4485),
    [sym_charlist] = STATE(4485),
    [sym_sigil] = STATE(4485),
    [sym_list] = STATE(4485),
    [sym_tuple] = STATE(4485),
    [sym_bitstring] = STATE(4485),
    [sym_map] = STATE(4485),
    [sym__nullary_operator] = STATE(4485),
    [sym_unary_operator] = STATE(4485),
    [sym_binary_operator] = STATE(4485),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4485),
    [sym_call] = STATE(4485),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4485),
    [sym_anonymous_function] = STATE(4485),
    [aux_sym__terminator_token1] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1361),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1361),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1569),
    [sym_integer] = ACTIONS(1569),
    [sym_float] = ACTIONS(1569),
    [sym_char] = ACTIONS(1569),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1569),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(354)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1575),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(355)] = {
    [sym__expression] = STATE(2519),
    [sym_block] = STATE(2519),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2519),
    [sym_nil] = STATE(2519),
    [sym__atom] = STATE(2519),
    [sym_quoted_atom] = STATE(2519),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2519),
    [sym_charlist] = STATE(2519),
    [sym_sigil] = STATE(2519),
    [sym_list] = STATE(2519),
    [sym_tuple] = STATE(2519),
    [sym_bitstring] = STATE(2519),
    [sym_map] = STATE(2519),
    [sym__nullary_operator] = STATE(2519),
    [sym_unary_operator] = STATE(2519),
    [sym__capture_expression] = STATE(2256),
    [sym_binary_operator] = STATE(2519),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2519),
    [sym_call] = STATE(2519),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2519),
    [sym_anonymous_function] = STATE(2519),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1579),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1581),
    [sym_integer] = ACTIONS(1583),
    [sym_float] = ACTIONS(1581),
    [sym_char] = ACTIONS(1581),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1581),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(356)] = {
    [sym__expression] = STATE(4119),
    [sym_block] = STATE(4119),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4119),
    [sym_nil] = STATE(4119),
    [sym__atom] = STATE(4119),
    [sym_quoted_atom] = STATE(4119),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4119),
    [sym_charlist] = STATE(4119),
    [sym_sigil] = STATE(4119),
    [sym_list] = STATE(4119),
    [sym_tuple] = STATE(4119),
    [sym_bitstring] = STATE(4119),
    [sym_map] = STATE(4119),
    [sym__nullary_operator] = STATE(4119),
    [sym_unary_operator] = STATE(4119),
    [sym__capture_expression] = STATE(4211),
    [sym_binary_operator] = STATE(4119),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4119),
    [sym_call] = STATE(4119),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4119),
    [sym_anonymous_function] = STATE(4119),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1585),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1587),
    [sym_integer] = ACTIONS(1589),
    [sym_float] = ACTIONS(1587),
    [sym_char] = ACTIONS(1587),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1587),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(357)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1591),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(358)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1593),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(359)] = {
    [sym__expression] = STATE(4583),
    [sym_block] = STATE(4583),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4583),
    [sym_nil] = STATE(4583),
    [sym__atom] = STATE(4583),
    [sym_quoted_atom] = STATE(4583),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4583),
    [sym_charlist] = STATE(4583),
    [sym_sigil] = STATE(4583),
    [sym_list] = STATE(4583),
    [sym_tuple] = STATE(4583),
    [sym_bitstring] = STATE(4583),
    [sym_map] = STATE(4583),
    [sym__nullary_operator] = STATE(4583),
    [sym_unary_operator] = STATE(4583),
    [sym_binary_operator] = STATE(4583),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4583),
    [sym_call] = STATE(4583),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4583),
    [sym_anonymous_function] = STATE(4583),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1595),
    [sym_integer] = ACTIONS(1595),
    [sym_float] = ACTIONS(1595),
    [sym_char] = ACTIONS(1595),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1595),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1597),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(360)] = {
    [sym__expression] = STATE(3677),
    [sym_block] = STATE(3677),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3677),
    [sym_nil] = STATE(3677),
    [sym__atom] = STATE(3677),
    [sym_quoted_atom] = STATE(3677),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3677),
    [sym_charlist] = STATE(3677),
    [sym_sigil] = STATE(3677),
    [sym_list] = STATE(3677),
    [sym_tuple] = STATE(3677),
    [sym_bitstring] = STATE(3677),
    [sym_map] = STATE(3677),
    [sym__nullary_operator] = STATE(3677),
    [sym_unary_operator] = STATE(3677),
    [sym__capture_expression] = STATE(3703),
    [sym_binary_operator] = STATE(3677),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3677),
    [sym_call] = STATE(3677),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3677),
    [sym_anonymous_function] = STATE(3677),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1599),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(1601),
    [sym_integer] = ACTIONS(1603),
    [sym_float] = ACTIONS(1601),
    [sym_char] = ACTIONS(1601),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(1601),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(361)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1605),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(362)] = {
    [sym__expression] = STATE(3210),
    [sym_block] = STATE(3210),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(3210),
    [sym_nil] = STATE(3210),
    [sym__atom] = STATE(3210),
    [sym_quoted_atom] = STATE(3210),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(3210),
    [sym_charlist] = STATE(3210),
    [sym_sigil] = STATE(3210),
    [sym_list] = STATE(3210),
    [sym_tuple] = STATE(3210),
    [sym_bitstring] = STATE(3210),
    [sym_map] = STATE(3210),
    [sym__nullary_operator] = STATE(3210),
    [sym_unary_operator] = STATE(3210),
    [sym_binary_operator] = STATE(3210),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(3210),
    [sym_call] = STATE(3210),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(3210),
    [sym_anonymous_function] = STATE(3210),
    [ts_builtin_sym_end] = ACTIONS(1607),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1609),
    [sym_integer] = ACTIONS(1609),
    [sym_float] = ACTIONS(1609),
    [sym_char] = ACTIONS(1609),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1609),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(363)] = {
    [sym__expression] = STATE(3493),
    [sym_block] = STATE(3493),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3493),
    [sym_nil] = STATE(3493),
    [sym__atom] = STATE(3493),
    [sym_quoted_atom] = STATE(3493),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3493),
    [sym_charlist] = STATE(3493),
    [sym_sigil] = STATE(3493),
    [sym_list] = STATE(3493),
    [sym_tuple] = STATE(3493),
    [sym_bitstring] = STATE(3493),
    [sym_map] = STATE(3493),
    [sym__nullary_operator] = STATE(3493),
    [sym_unary_operator] = STATE(3493),
    [sym__capture_expression] = STATE(1465),
    [sym_binary_operator] = STATE(3493),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3493),
    [sym_call] = STATE(3493),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3493),
    [sym_anonymous_function] = STATE(3493),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1611),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1613),
    [sym_integer] = ACTIONS(1615),
    [sym_float] = ACTIONS(1613),
    [sym_char] = ACTIONS(1613),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1613),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(364)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1617),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(365)] = {
    [sym__expression] = STATE(3969),
    [sym_block] = STATE(3969),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3969),
    [sym_nil] = STATE(3969),
    [sym__atom] = STATE(3969),
    [sym_quoted_atom] = STATE(3969),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3969),
    [sym_charlist] = STATE(3969),
    [sym_sigil] = STATE(3969),
    [sym_list] = STATE(3969),
    [sym_tuple] = STATE(3969),
    [sym_bitstring] = STATE(3969),
    [sym_map] = STATE(3969),
    [sym__nullary_operator] = STATE(3969),
    [sym_unary_operator] = STATE(3969),
    [sym__capture_expression] = STATE(3007),
    [sym_binary_operator] = STATE(3969),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3969),
    [sym_call] = STATE(3969),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3969),
    [sym_anonymous_function] = STATE(3969),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1619),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1621),
    [sym_integer] = ACTIONS(1623),
    [sym_float] = ACTIONS(1621),
    [sym_char] = ACTIONS(1621),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1621),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(366)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1625),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(367)] = {
    [sym__expression] = STATE(4587),
    [sym_block] = STATE(4587),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4587),
    [sym_nil] = STATE(4587),
    [sym__atom] = STATE(4587),
    [sym_quoted_atom] = STATE(4587),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4587),
    [sym_charlist] = STATE(4587),
    [sym_sigil] = STATE(4587),
    [sym_list] = STATE(4587),
    [sym_tuple] = STATE(4587),
    [sym_bitstring] = STATE(4587),
    [sym_map] = STATE(4587),
    [sym__nullary_operator] = STATE(4587),
    [sym_unary_operator] = STATE(4587),
    [sym_binary_operator] = STATE(4587),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4587),
    [sym_call] = STATE(4587),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4587),
    [sym_anonymous_function] = STATE(4587),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1627),
    [sym_integer] = ACTIONS(1627),
    [sym_float] = ACTIONS(1627),
    [sym_char] = ACTIONS(1627),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1627),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1629),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(368)] = {
    [sym__expression] = STATE(3493),
    [sym_block] = STATE(3493),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3493),
    [sym_nil] = STATE(3493),
    [sym__atom] = STATE(3493),
    [sym_quoted_atom] = STATE(3493),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3493),
    [sym_charlist] = STATE(3493),
    [sym_sigil] = STATE(3493),
    [sym_list] = STATE(3493),
    [sym_tuple] = STATE(3493),
    [sym_bitstring] = STATE(3493),
    [sym_map] = STATE(3493),
    [sym__nullary_operator] = STATE(3493),
    [sym_unary_operator] = STATE(3493),
    [sym__capture_expression] = STATE(1521),
    [sym_binary_operator] = STATE(3493),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3493),
    [sym_call] = STATE(3493),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3493),
    [sym_anonymous_function] = STATE(3493),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1611),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1613),
    [sym_integer] = ACTIONS(1631),
    [sym_float] = ACTIONS(1613),
    [sym_char] = ACTIONS(1613),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1613),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(369)] = {
    [sym__expression] = STATE(3969),
    [sym_block] = STATE(3969),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3969),
    [sym_nil] = STATE(3969),
    [sym__atom] = STATE(3969),
    [sym_quoted_atom] = STATE(3969),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3969),
    [sym_charlist] = STATE(3969),
    [sym_sigil] = STATE(3969),
    [sym_list] = STATE(3969),
    [sym_tuple] = STATE(3969),
    [sym_bitstring] = STATE(3969),
    [sym_map] = STATE(3969),
    [sym__nullary_operator] = STATE(3969),
    [sym_unary_operator] = STATE(3969),
    [sym__capture_expression] = STATE(3057),
    [sym_binary_operator] = STATE(3969),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3969),
    [sym_call] = STATE(3969),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3969),
    [sym_anonymous_function] = STATE(3969),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1619),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1621),
    [sym_integer] = ACTIONS(1633),
    [sym_float] = ACTIONS(1621),
    [sym_char] = ACTIONS(1621),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1621),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(370)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1635),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(371)] = {
    [sym__expression] = STATE(4585),
    [sym_block] = STATE(4585),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4585),
    [sym_nil] = STATE(4585),
    [sym__atom] = STATE(4585),
    [sym_quoted_atom] = STATE(4585),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4585),
    [sym_charlist] = STATE(4585),
    [sym_sigil] = STATE(4585),
    [sym_list] = STATE(4585),
    [sym_tuple] = STATE(4585),
    [sym_bitstring] = STATE(4585),
    [sym_map] = STATE(4585),
    [sym__nullary_operator] = STATE(4585),
    [sym_unary_operator] = STATE(4585),
    [sym_binary_operator] = STATE(4585),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4585),
    [sym_call] = STATE(4585),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4585),
    [sym_anonymous_function] = STATE(4585),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1637),
    [sym_integer] = ACTIONS(1637),
    [sym_float] = ACTIONS(1637),
    [sym_char] = ACTIONS(1637),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1637),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1639),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(372)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1641),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(373)] = {
    [sym__expression] = STATE(4575),
    [sym_block] = STATE(4575),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4575),
    [sym_nil] = STATE(4575),
    [sym__atom] = STATE(4575),
    [sym_quoted_atom] = STATE(4575),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4575),
    [sym_charlist] = STATE(4575),
    [sym_sigil] = STATE(4575),
    [sym_list] = STATE(4575),
    [sym_tuple] = STATE(4575),
    [sym_bitstring] = STATE(4575),
    [sym_map] = STATE(4575),
    [sym__nullary_operator] = STATE(4575),
    [sym_unary_operator] = STATE(4575),
    [sym_binary_operator] = STATE(4575),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4575),
    [sym_call] = STATE(4575),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4575),
    [sym_anonymous_function] = STATE(4575),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1643),
    [sym_integer] = ACTIONS(1643),
    [sym_float] = ACTIONS(1643),
    [sym_char] = ACTIONS(1643),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1643),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(374)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1647),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(375)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1649),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(376)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1651),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(377)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1653),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(378)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1655),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(379)] = {
    [sym__expression] = STATE(3300),
    [sym_block] = STATE(3300),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3300),
    [sym_nil] = STATE(3300),
    [sym__atom] = STATE(3300),
    [sym_quoted_atom] = STATE(3300),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3300),
    [sym_charlist] = STATE(3300),
    [sym_sigil] = STATE(3300),
    [sym_list] = STATE(3300),
    [sym_tuple] = STATE(3300),
    [sym_bitstring] = STATE(3300),
    [sym_map] = STATE(3300),
    [sym__nullary_operator] = STATE(3300),
    [sym_unary_operator] = STATE(3300),
    [sym__capture_expression] = STATE(1465),
    [sym_binary_operator] = STATE(3300),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3300),
    [sym_call] = STATE(3300),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3300),
    [sym_anonymous_function] = STATE(3300),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1611),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(1657),
    [sym_integer] = ACTIONS(1615),
    [sym_float] = ACTIONS(1657),
    [sym_char] = ACTIONS(1657),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1657),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(380)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1659),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(381)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1661),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(382)] = {
    [sym__expression] = STATE(3372),
    [sym_block] = STATE(3372),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3372),
    [sym_nil] = STATE(3372),
    [sym__atom] = STATE(3372),
    [sym_quoted_atom] = STATE(3372),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3372),
    [sym_charlist] = STATE(3372),
    [sym_sigil] = STATE(3372),
    [sym_list] = STATE(3372),
    [sym_tuple] = STATE(3372),
    [sym_bitstring] = STATE(3372),
    [sym_map] = STATE(3372),
    [sym__nullary_operator] = STATE(3372),
    [sym_unary_operator] = STATE(3372),
    [sym__capture_expression] = STATE(3373),
    [sym_binary_operator] = STATE(3372),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3372),
    [sym_call] = STATE(3372),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3372),
    [sym_anonymous_function] = STATE(3372),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1663),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1665),
    [sym_integer] = ACTIONS(1667),
    [sym_float] = ACTIONS(1665),
    [sym_char] = ACTIONS(1665),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(1665),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(383)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1669),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(384)] = {
    [sym__expression] = STATE(4105),
    [sym_block] = STATE(4105),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4105),
    [sym_nil] = STATE(4105),
    [sym__atom] = STATE(4105),
    [sym_quoted_atom] = STATE(4105),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4105),
    [sym_charlist] = STATE(4105),
    [sym_sigil] = STATE(4105),
    [sym_list] = STATE(4105),
    [sym_tuple] = STATE(4105),
    [sym_bitstring] = STATE(4105),
    [sym_map] = STATE(4105),
    [sym__nullary_operator] = STATE(4105),
    [sym_unary_operator] = STATE(4105),
    [sym__capture_expression] = STATE(2021),
    [sym_binary_operator] = STATE(4105),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4105),
    [sym_call] = STATE(4105),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4105),
    [sym_anonymous_function] = STATE(4105),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1671),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1673),
    [sym_integer] = ACTIONS(1675),
    [sym_float] = ACTIONS(1673),
    [sym_char] = ACTIONS(1673),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1673),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(385)] = {
    [sym__expression] = STATE(3987),
    [sym_block] = STATE(3987),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(3987),
    [sym_nil] = STATE(3987),
    [sym__atom] = STATE(3987),
    [sym_quoted_atom] = STATE(3987),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(3987),
    [sym_charlist] = STATE(3987),
    [sym_sigil] = STATE(3987),
    [sym_list] = STATE(3987),
    [sym_tuple] = STATE(3987),
    [sym_bitstring] = STATE(3987),
    [sym_map] = STATE(3987),
    [sym__nullary_operator] = STATE(3987),
    [sym_unary_operator] = STATE(3987),
    [sym__capture_expression] = STATE(1587),
    [sym_binary_operator] = STATE(3987),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(3987),
    [sym_call] = STATE(3987),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(3987),
    [sym_anonymous_function] = STATE(3987),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1677),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1679),
    [sym_integer] = ACTIONS(1681),
    [sym_float] = ACTIONS(1679),
    [sym_char] = ACTIONS(1679),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(1679),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(386)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1683),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(387)] = {
    [sym__expression] = STATE(4447),
    [sym_block] = STATE(4447),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4447),
    [sym_nil] = STATE(4447),
    [sym__atom] = STATE(4447),
    [sym_quoted_atom] = STATE(4447),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4447),
    [sym_charlist] = STATE(4447),
    [sym_sigil] = STATE(4447),
    [sym_list] = STATE(4447),
    [sym_tuple] = STATE(4447),
    [sym_bitstring] = STATE(4447),
    [sym_map] = STATE(4447),
    [sym__nullary_operator] = STATE(4447),
    [sym_unary_operator] = STATE(4447),
    [sym_binary_operator] = STATE(4447),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4447),
    [sym_call] = STATE(4447),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4447),
    [sym_anonymous_function] = STATE(4447),
    [ts_builtin_sym_end] = ACTIONS(1685),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1687),
    [sym_integer] = ACTIONS(1687),
    [sym_float] = ACTIONS(1687),
    [sym_char] = ACTIONS(1687),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1687),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(388)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6975),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1693),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(389)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1695),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(390)] = {
    [sym__expression] = STATE(4496),
    [sym_block] = STATE(4496),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4496),
    [sym_nil] = STATE(4496),
    [sym__atom] = STATE(4496),
    [sym_quoted_atom] = STATE(4496),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4496),
    [sym_charlist] = STATE(4496),
    [sym_sigil] = STATE(4496),
    [sym_list] = STATE(4496),
    [sym_tuple] = STATE(4496),
    [sym_bitstring] = STATE(4496),
    [sym_map] = STATE(4496),
    [sym__nullary_operator] = STATE(4496),
    [sym_unary_operator] = STATE(4496),
    [sym__capture_expression] = STATE(3007),
    [sym_binary_operator] = STATE(4496),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4496),
    [sym_call] = STATE(4496),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4496),
    [sym_anonymous_function] = STATE(4496),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1619),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1697),
    [sym_integer] = ACTIONS(1623),
    [sym_float] = ACTIONS(1697),
    [sym_char] = ACTIONS(1697),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1697),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(391)] = {
    [sym__expression] = STATE(4496),
    [sym_block] = STATE(4496),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4496),
    [sym_nil] = STATE(4496),
    [sym__atom] = STATE(4496),
    [sym_quoted_atom] = STATE(4496),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4496),
    [sym_charlist] = STATE(4496),
    [sym_sigil] = STATE(4496),
    [sym_list] = STATE(4496),
    [sym_tuple] = STATE(4496),
    [sym_bitstring] = STATE(4496),
    [sym_map] = STATE(4496),
    [sym__nullary_operator] = STATE(4496),
    [sym_unary_operator] = STATE(4496),
    [sym__capture_expression] = STATE(3057),
    [sym_binary_operator] = STATE(4496),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4496),
    [sym_call] = STATE(4496),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4496),
    [sym_anonymous_function] = STATE(4496),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1619),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1697),
    [sym_integer] = ACTIONS(1633),
    [sym_float] = ACTIONS(1697),
    [sym_char] = ACTIONS(1697),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1697),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(392)] = {
    [sym__expression] = STATE(3300),
    [sym_block] = STATE(3300),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3300),
    [sym_nil] = STATE(3300),
    [sym__atom] = STATE(3300),
    [sym_quoted_atom] = STATE(3300),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3300),
    [sym_charlist] = STATE(3300),
    [sym_sigil] = STATE(3300),
    [sym_list] = STATE(3300),
    [sym_tuple] = STATE(3300),
    [sym_bitstring] = STATE(3300),
    [sym_map] = STATE(3300),
    [sym__nullary_operator] = STATE(3300),
    [sym_unary_operator] = STATE(3300),
    [sym__capture_expression] = STATE(1521),
    [sym_binary_operator] = STATE(3300),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3300),
    [sym_call] = STATE(3300),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3300),
    [sym_anonymous_function] = STATE(3300),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1611),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(1657),
    [sym_integer] = ACTIONS(1631),
    [sym_float] = ACTIONS(1657),
    [sym_char] = ACTIONS(1657),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1657),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(393)] = {
    [sym__expression] = STATE(4588),
    [sym_block] = STATE(4588),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4588),
    [sym_nil] = STATE(4588),
    [sym__atom] = STATE(4588),
    [sym_quoted_atom] = STATE(4588),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4588),
    [sym_charlist] = STATE(4588),
    [sym_sigil] = STATE(4588),
    [sym_list] = STATE(4588),
    [sym_tuple] = STATE(4588),
    [sym_bitstring] = STATE(4588),
    [sym_map] = STATE(4588),
    [sym__nullary_operator] = STATE(4588),
    [sym_unary_operator] = STATE(4588),
    [sym_binary_operator] = STATE(4588),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4588),
    [sym_call] = STATE(4588),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4588),
    [sym_anonymous_function] = STATE(4588),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1699),
    [sym_integer] = ACTIONS(1699),
    [sym_float] = ACTIONS(1699),
    [sym_char] = ACTIONS(1699),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1699),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1701),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(394)] = {
    [sym__expression] = STATE(2519),
    [sym_block] = STATE(2519),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2519),
    [sym_nil] = STATE(2519),
    [sym__atom] = STATE(2519),
    [sym_quoted_atom] = STATE(2519),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2519),
    [sym_charlist] = STATE(2519),
    [sym_sigil] = STATE(2519),
    [sym_list] = STATE(2519),
    [sym_tuple] = STATE(2519),
    [sym_bitstring] = STATE(2519),
    [sym_map] = STATE(2519),
    [sym__nullary_operator] = STATE(2519),
    [sym_unary_operator] = STATE(2519),
    [sym__capture_expression] = STATE(2183),
    [sym_binary_operator] = STATE(2519),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2519),
    [sym_call] = STATE(2519),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2519),
    [sym_anonymous_function] = STATE(2519),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1579),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1581),
    [sym_integer] = ACTIONS(1703),
    [sym_float] = ACTIONS(1581),
    [sym_char] = ACTIONS(1581),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1581),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(395)] = {
    [sym__expression] = STATE(4032),
    [sym_block] = STATE(4032),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(4032),
    [sym_nil] = STATE(4032),
    [sym__atom] = STATE(4032),
    [sym_quoted_atom] = STATE(4032),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(4032),
    [sym_charlist] = STATE(4032),
    [sym_sigil] = STATE(4032),
    [sym_list] = STATE(4032),
    [sym_tuple] = STATE(4032),
    [sym_bitstring] = STATE(4032),
    [sym_map] = STATE(4032),
    [sym__nullary_operator] = STATE(4032),
    [sym_unary_operator] = STATE(4032),
    [sym__capture_expression] = STATE(2183),
    [sym_binary_operator] = STATE(4032),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(4032),
    [sym_call] = STATE(4032),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(4032),
    [sym_anonymous_function] = STATE(4032),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1579),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1705),
    [sym_integer] = ACTIONS(1703),
    [sym_float] = ACTIONS(1705),
    [sym_char] = ACTIONS(1705),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1705),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(396)] = {
    [sym__expression] = STATE(4032),
    [sym_block] = STATE(4032),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(4032),
    [sym_nil] = STATE(4032),
    [sym__atom] = STATE(4032),
    [sym_quoted_atom] = STATE(4032),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(4032),
    [sym_charlist] = STATE(4032),
    [sym_sigil] = STATE(4032),
    [sym_list] = STATE(4032),
    [sym_tuple] = STATE(4032),
    [sym_bitstring] = STATE(4032),
    [sym_map] = STATE(4032),
    [sym__nullary_operator] = STATE(4032),
    [sym_unary_operator] = STATE(4032),
    [sym__capture_expression] = STATE(2256),
    [sym_binary_operator] = STATE(4032),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(4032),
    [sym_call] = STATE(4032),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(4032),
    [sym_anonymous_function] = STATE(4032),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1579),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1705),
    [sym_integer] = ACTIONS(1583),
    [sym_float] = ACTIONS(1705),
    [sym_char] = ACTIONS(1705),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1705),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(397)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6909),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1707),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(398)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1709),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(399)] = {
    [sym__expression] = STATE(2291),
    [sym_block] = STATE(2291),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2291),
    [sym_nil] = STATE(2291),
    [sym__atom] = STATE(2291),
    [sym_quoted_atom] = STATE(2291),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2291),
    [sym_charlist] = STATE(2291),
    [sym_sigil] = STATE(2291),
    [sym_list] = STATE(2291),
    [sym_tuple] = STATE(2291),
    [sym_bitstring] = STATE(2291),
    [sym_map] = STATE(2291),
    [sym__nullary_operator] = STATE(2291),
    [sym_unary_operator] = STATE(2291),
    [sym__capture_expression] = STATE(1907),
    [sym_binary_operator] = STATE(2291),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2291),
    [sym_call] = STATE(2291),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2291),
    [sym_anonymous_function] = STATE(2291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1671),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1711),
    [sym_integer] = ACTIONS(1713),
    [sym_float] = ACTIONS(1711),
    [sym_char] = ACTIONS(1711),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1711),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(400)] = {
    [sym__expression] = STATE(4553),
    [sym_block] = STATE(4553),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4553),
    [sym_nil] = STATE(4553),
    [sym__atom] = STATE(4553),
    [sym_quoted_atom] = STATE(4553),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4553),
    [sym_charlist] = STATE(4553),
    [sym_sigil] = STATE(4553),
    [sym_list] = STATE(4553),
    [sym_tuple] = STATE(4553),
    [sym_bitstring] = STATE(4553),
    [sym_map] = STATE(4553),
    [sym__nullary_operator] = STATE(4553),
    [sym_unary_operator] = STATE(4553),
    [sym__capture_expression] = STATE(4554),
    [sym_binary_operator] = STATE(4553),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4553),
    [sym_call] = STATE(4553),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4553),
    [sym_anonymous_function] = STATE(4553),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1715),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1717),
    [sym_integer] = ACTIONS(1719),
    [sym_float] = ACTIONS(1717),
    [sym_char] = ACTIONS(1717),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1717),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(401)] = {
    [sym__expression] = STATE(4581),
    [sym_block] = STATE(4581),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4581),
    [sym_nil] = STATE(4581),
    [sym__atom] = STATE(4581),
    [sym_quoted_atom] = STATE(4581),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4581),
    [sym_charlist] = STATE(4581),
    [sym_sigil] = STATE(4581),
    [sym_list] = STATE(4581),
    [sym_tuple] = STATE(4581),
    [sym_bitstring] = STATE(4581),
    [sym_map] = STATE(4581),
    [sym__nullary_operator] = STATE(4581),
    [sym_unary_operator] = STATE(4581),
    [sym_binary_operator] = STATE(4581),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4581),
    [sym_call] = STATE(4581),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4581),
    [sym_anonymous_function] = STATE(4581),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1721),
    [sym_integer] = ACTIONS(1721),
    [sym_float] = ACTIONS(1721),
    [sym_char] = ACTIONS(1721),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1721),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1723),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(402)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1725),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(403)] = {
    [sym__expression] = STATE(4447),
    [sym_block] = STATE(4447),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4447),
    [sym_nil] = STATE(4447),
    [sym__atom] = STATE(4447),
    [sym_quoted_atom] = STATE(4447),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4447),
    [sym_charlist] = STATE(4447),
    [sym_sigil] = STATE(4447),
    [sym_list] = STATE(4447),
    [sym_tuple] = STATE(4447),
    [sym_bitstring] = STATE(4447),
    [sym_map] = STATE(4447),
    [sym__nullary_operator] = STATE(4447),
    [sym_unary_operator] = STATE(4447),
    [sym_binary_operator] = STATE(4447),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4447),
    [sym_call] = STATE(4447),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4447),
    [sym_anonymous_function] = STATE(4447),
    [ts_builtin_sym_end] = ACTIONS(1727),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1687),
    [sym_integer] = ACTIONS(1687),
    [sym_float] = ACTIONS(1687),
    [sym_char] = ACTIONS(1687),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1687),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(404)] = {
    [sym__expression] = STATE(3372),
    [sym_block] = STATE(3372),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3372),
    [sym_nil] = STATE(3372),
    [sym__atom] = STATE(3372),
    [sym_quoted_atom] = STATE(3372),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3372),
    [sym_charlist] = STATE(3372),
    [sym_sigil] = STATE(3372),
    [sym_list] = STATE(3372),
    [sym_tuple] = STATE(3372),
    [sym_bitstring] = STATE(3372),
    [sym_map] = STATE(3372),
    [sym__nullary_operator] = STATE(3372),
    [sym_unary_operator] = STATE(3372),
    [sym__capture_expression] = STATE(3385),
    [sym_binary_operator] = STATE(3372),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3372),
    [sym_call] = STATE(3372),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3372),
    [sym_anonymous_function] = STATE(3372),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1663),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1665),
    [sym_integer] = ACTIONS(1729),
    [sym_float] = ACTIONS(1665),
    [sym_char] = ACTIONS(1665),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(1665),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(405)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1731),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(406)] = {
    [sym__expression] = STATE(1690),
    [sym_block] = STATE(1690),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1690),
    [sym_nil] = STATE(1690),
    [sym__atom] = STATE(1690),
    [sym_quoted_atom] = STATE(1690),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1690),
    [sym_charlist] = STATE(1690),
    [sym_sigil] = STATE(1690),
    [sym_list] = STATE(1690),
    [sym_tuple] = STATE(1690),
    [sym_bitstring] = STATE(1690),
    [sym_map] = STATE(1690),
    [sym__nullary_operator] = STATE(1690),
    [sym_unary_operator] = STATE(1690),
    [sym__capture_expression] = STATE(1521),
    [sym_binary_operator] = STATE(1690),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1690),
    [sym_call] = STATE(1690),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1690),
    [sym_anonymous_function] = STATE(1690),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1611),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1733),
    [sym_integer] = ACTIONS(1631),
    [sym_float] = ACTIONS(1733),
    [sym_char] = ACTIONS(1733),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1733),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(407)] = {
    [sym__expression] = STATE(4553),
    [sym_block] = STATE(4553),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4553),
    [sym_nil] = STATE(4553),
    [sym__atom] = STATE(4553),
    [sym_quoted_atom] = STATE(4553),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4553),
    [sym_charlist] = STATE(4553),
    [sym_sigil] = STATE(4553),
    [sym_list] = STATE(4553),
    [sym_tuple] = STATE(4553),
    [sym_bitstring] = STATE(4553),
    [sym_map] = STATE(4553),
    [sym__nullary_operator] = STATE(4553),
    [sym_unary_operator] = STATE(4553),
    [sym__capture_expression] = STATE(4551),
    [sym_binary_operator] = STATE(4553),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4553),
    [sym_call] = STATE(4553),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4553),
    [sym_anonymous_function] = STATE(4553),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1715),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1717),
    [sym_integer] = ACTIONS(1735),
    [sym_float] = ACTIONS(1717),
    [sym_char] = ACTIONS(1717),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1717),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(408)] = {
    [sym__expression] = STATE(4105),
    [sym_block] = STATE(4105),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4105),
    [sym_nil] = STATE(4105),
    [sym__atom] = STATE(4105),
    [sym_quoted_atom] = STATE(4105),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4105),
    [sym_charlist] = STATE(4105),
    [sym_sigil] = STATE(4105),
    [sym_list] = STATE(4105),
    [sym_tuple] = STATE(4105),
    [sym_bitstring] = STATE(4105),
    [sym_map] = STATE(4105),
    [sym__nullary_operator] = STATE(4105),
    [sym_unary_operator] = STATE(4105),
    [sym__capture_expression] = STATE(1907),
    [sym_binary_operator] = STATE(4105),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4105),
    [sym_call] = STATE(4105),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4105),
    [sym_anonymous_function] = STATE(4105),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1671),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1673),
    [sym_integer] = ACTIONS(1713),
    [sym_float] = ACTIONS(1673),
    [sym_char] = ACTIONS(1673),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1673),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(409)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1737),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(410)] = {
    [sym__expression] = STATE(4570),
    [sym_block] = STATE(4570),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4570),
    [sym_nil] = STATE(4570),
    [sym__atom] = STATE(4570),
    [sym_quoted_atom] = STATE(4570),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4570),
    [sym_charlist] = STATE(4570),
    [sym_sigil] = STATE(4570),
    [sym_list] = STATE(4570),
    [sym_tuple] = STATE(4570),
    [sym_bitstring] = STATE(4570),
    [sym_map] = STATE(4570),
    [sym__nullary_operator] = STATE(4570),
    [sym_unary_operator] = STATE(4570),
    [sym_binary_operator] = STATE(4570),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4570),
    [sym_call] = STATE(4570),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4570),
    [sym_anonymous_function] = STATE(4570),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1739),
    [sym_integer] = ACTIONS(1739),
    [sym_float] = ACTIONS(1739),
    [sym_char] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1739),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(411)] = {
    [sym__expression] = STATE(4145),
    [sym_block] = STATE(4145),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4145),
    [sym_nil] = STATE(4145),
    [sym__atom] = STATE(4145),
    [sym_quoted_atom] = STATE(4145),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4145),
    [sym_charlist] = STATE(4145),
    [sym_sigil] = STATE(4145),
    [sym_list] = STATE(4145),
    [sym_tuple] = STATE(4145),
    [sym_bitstring] = STATE(4145),
    [sym_map] = STATE(4145),
    [sym__nullary_operator] = STATE(4145),
    [sym_unary_operator] = STATE(4145),
    [sym__capture_expression] = STATE(4146),
    [sym_binary_operator] = STATE(4145),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4145),
    [sym_call] = STATE(4145),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4145),
    [sym_anonymous_function] = STATE(4145),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1743),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1745),
    [sym_integer] = ACTIONS(1747),
    [sym_float] = ACTIONS(1745),
    [sym_char] = ACTIONS(1745),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1745),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(412)] = {
    [sym__expression] = STATE(4119),
    [sym_block] = STATE(4119),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4119),
    [sym_nil] = STATE(4119),
    [sym__atom] = STATE(4119),
    [sym_quoted_atom] = STATE(4119),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4119),
    [sym_charlist] = STATE(4119),
    [sym_sigil] = STATE(4119),
    [sym_list] = STATE(4119),
    [sym_tuple] = STATE(4119),
    [sym_bitstring] = STATE(4119),
    [sym_map] = STATE(4119),
    [sym__nullary_operator] = STATE(4119),
    [sym_unary_operator] = STATE(4119),
    [sym__capture_expression] = STATE(4326),
    [sym_binary_operator] = STATE(4119),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4119),
    [sym_call] = STATE(4119),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4119),
    [sym_anonymous_function] = STATE(4119),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1585),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1587),
    [sym_integer] = ACTIONS(1749),
    [sym_float] = ACTIONS(1587),
    [sym_char] = ACTIONS(1587),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1587),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(413)] = {
    [sym__expression] = STATE(3677),
    [sym_block] = STATE(3677),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3677),
    [sym_nil] = STATE(3677),
    [sym__atom] = STATE(3677),
    [sym_quoted_atom] = STATE(3677),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3677),
    [sym_charlist] = STATE(3677),
    [sym_sigil] = STATE(3677),
    [sym_list] = STATE(3677),
    [sym_tuple] = STATE(3677),
    [sym_bitstring] = STATE(3677),
    [sym_map] = STATE(3677),
    [sym__nullary_operator] = STATE(3677),
    [sym_unary_operator] = STATE(3677),
    [sym__capture_expression] = STATE(3678),
    [sym_binary_operator] = STATE(3677),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3677),
    [sym_call] = STATE(3677),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3677),
    [sym_anonymous_function] = STATE(3677),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1599),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(1601),
    [sym_integer] = ACTIONS(1751),
    [sym_float] = ACTIONS(1601),
    [sym_char] = ACTIONS(1601),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(1601),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(414)] = {
    [sym__expression] = STATE(4145),
    [sym_block] = STATE(4145),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4145),
    [sym_nil] = STATE(4145),
    [sym__atom] = STATE(4145),
    [sym_quoted_atom] = STATE(4145),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4145),
    [sym_charlist] = STATE(4145),
    [sym_sigil] = STATE(4145),
    [sym_list] = STATE(4145),
    [sym_tuple] = STATE(4145),
    [sym_bitstring] = STATE(4145),
    [sym_map] = STATE(4145),
    [sym__nullary_operator] = STATE(4145),
    [sym_unary_operator] = STATE(4145),
    [sym__capture_expression] = STATE(4180),
    [sym_binary_operator] = STATE(4145),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4145),
    [sym_call] = STATE(4145),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4145),
    [sym_anonymous_function] = STATE(4145),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1743),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1745),
    [sym_integer] = ACTIONS(1753),
    [sym_float] = ACTIONS(1745),
    [sym_char] = ACTIONS(1745),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1745),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(415)] = {
    [sym__expression] = STATE(1383),
    [sym_block] = STATE(1383),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1383),
    [sym_nil] = STATE(1383),
    [sym__atom] = STATE(1383),
    [sym_quoted_atom] = STATE(1383),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1383),
    [sym_charlist] = STATE(1383),
    [sym_sigil] = STATE(1383),
    [sym_list] = STATE(1383),
    [sym_tuple] = STATE(1383),
    [sym_bitstring] = STATE(1383),
    [sym_map] = STATE(1383),
    [sym__nullary_operator] = STATE(1383),
    [sym_unary_operator] = STATE(1383),
    [sym__capture_expression] = STATE(1309),
    [sym_binary_operator] = STATE(1383),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1383),
    [sym_call] = STATE(1383),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1383),
    [sym_anonymous_function] = STATE(1383),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1755),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1757),
    [sym_integer] = ACTIONS(1759),
    [sym_float] = ACTIONS(1757),
    [sym_char] = ACTIONS(1757),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1757),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(416)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1761),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(417)] = {
    [sym__expression] = STATE(1917),
    [sym_block] = STATE(1917),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1917),
    [sym_nil] = STATE(1917),
    [sym__atom] = STATE(1917),
    [sym_quoted_atom] = STATE(1917),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1917),
    [sym_charlist] = STATE(1917),
    [sym_sigil] = STATE(1917),
    [sym_list] = STATE(1917),
    [sym_tuple] = STATE(1917),
    [sym_bitstring] = STATE(1917),
    [sym_map] = STATE(1917),
    [sym__nullary_operator] = STATE(1917),
    [sym_unary_operator] = STATE(1917),
    [sym__capture_expression] = STATE(1587),
    [sym_binary_operator] = STATE(1917),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1917),
    [sym_call] = STATE(1917),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1917),
    [sym_anonymous_function] = STATE(1917),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1677),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1763),
    [sym_integer] = ACTIONS(1681),
    [sym_float] = ACTIONS(1763),
    [sym_char] = ACTIONS(1763),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(418)] = {
    [sym__expression] = STATE(3481),
    [sym_block] = STATE(3481),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(3481),
    [sym_nil] = STATE(3481),
    [sym__atom] = STATE(3481),
    [sym_quoted_atom] = STATE(3481),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(3481),
    [sym_charlist] = STATE(3481),
    [sym_sigil] = STATE(3481),
    [sym_list] = STATE(3481),
    [sym_tuple] = STATE(3481),
    [sym_bitstring] = STATE(3481),
    [sym_map] = STATE(3481),
    [sym__nullary_operator] = STATE(3481),
    [sym_unary_operator] = STATE(3481),
    [sym__capture_expression] = STATE(3482),
    [sym_binary_operator] = STATE(3481),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(3481),
    [sym_call] = STATE(3481),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(3481),
    [sym_anonymous_function] = STATE(3481),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1765),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1767),
    [sym_integer] = ACTIONS(1769),
    [sym_float] = ACTIONS(1767),
    [sym_char] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(419)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1771),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(420)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1773),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(421)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1775),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(422)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1777),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(423)] = {
    [sym__expression] = STATE(3987),
    [sym_block] = STATE(3987),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(3987),
    [sym_nil] = STATE(3987),
    [sym__atom] = STATE(3987),
    [sym_quoted_atom] = STATE(3987),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(3987),
    [sym_charlist] = STATE(3987),
    [sym_sigil] = STATE(3987),
    [sym_list] = STATE(3987),
    [sym_tuple] = STATE(3987),
    [sym_bitstring] = STATE(3987),
    [sym_map] = STATE(3987),
    [sym__nullary_operator] = STATE(3987),
    [sym_unary_operator] = STATE(3987),
    [sym__capture_expression] = STATE(1652),
    [sym_binary_operator] = STATE(3987),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(3987),
    [sym_call] = STATE(3987),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(3987),
    [sym_anonymous_function] = STATE(3987),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1677),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1679),
    [sym_integer] = ACTIONS(1779),
    [sym_float] = ACTIONS(1679),
    [sym_char] = ACTIONS(1679),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(1679),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(424)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1781),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(425)] = {
    [sym__expression] = STATE(4593),
    [sym_block] = STATE(4593),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4593),
    [sym_nil] = STATE(4593),
    [sym__atom] = STATE(4593),
    [sym_quoted_atom] = STATE(4593),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4593),
    [sym_charlist] = STATE(4593),
    [sym_sigil] = STATE(4593),
    [sym_list] = STATE(4593),
    [sym_tuple] = STATE(4593),
    [sym_bitstring] = STATE(4593),
    [sym_map] = STATE(4593),
    [sym__nullary_operator] = STATE(4593),
    [sym_unary_operator] = STATE(4593),
    [sym_binary_operator] = STATE(4593),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4593),
    [sym_call] = STATE(4593),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4593),
    [sym_anonymous_function] = STATE(4593),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1783),
    [sym_integer] = ACTIONS(1783),
    [sym_float] = ACTIONS(1783),
    [sym_char] = ACTIONS(1783),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1783),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1785),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(426)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1787),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(427)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1789),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(428)] = {
    [sym__expression] = STATE(3481),
    [sym_block] = STATE(3481),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(3481),
    [sym_nil] = STATE(3481),
    [sym__atom] = STATE(3481),
    [sym_quoted_atom] = STATE(3481),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(3481),
    [sym_charlist] = STATE(3481),
    [sym_sigil] = STATE(3481),
    [sym_list] = STATE(3481),
    [sym_tuple] = STATE(3481),
    [sym_bitstring] = STATE(3481),
    [sym_map] = STATE(3481),
    [sym__nullary_operator] = STATE(3481),
    [sym_unary_operator] = STATE(3481),
    [sym__capture_expression] = STATE(2933),
    [sym_binary_operator] = STATE(3481),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(3481),
    [sym_call] = STATE(3481),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(3481),
    [sym_anonymous_function] = STATE(3481),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1765),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1767),
    [sym_integer] = ACTIONS(1791),
    [sym_float] = ACTIONS(1767),
    [sym_char] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(429)] = {
    [sym__expression] = STATE(4577),
    [sym_block] = STATE(4577),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4577),
    [sym_nil] = STATE(4577),
    [sym__atom] = STATE(4577),
    [sym_quoted_atom] = STATE(4577),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4577),
    [sym_charlist] = STATE(4577),
    [sym_sigil] = STATE(4577),
    [sym_list] = STATE(4577),
    [sym_tuple] = STATE(4577),
    [sym_bitstring] = STATE(4577),
    [sym_map] = STATE(4577),
    [sym__nullary_operator] = STATE(4577),
    [sym_unary_operator] = STATE(4577),
    [sym_binary_operator] = STATE(4577),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4577),
    [sym_call] = STATE(4577),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4577),
    [sym_anonymous_function] = STATE(4577),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1793),
    [sym_integer] = ACTIONS(1793),
    [sym_float] = ACTIONS(1793),
    [sym_char] = ACTIONS(1793),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1793),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1795),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(430)] = {
    [sym__expression] = STATE(1383),
    [sym_block] = STATE(1383),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1383),
    [sym_nil] = STATE(1383),
    [sym__atom] = STATE(1383),
    [sym_quoted_atom] = STATE(1383),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1383),
    [sym_charlist] = STATE(1383),
    [sym_sigil] = STATE(1383),
    [sym_list] = STATE(1383),
    [sym_tuple] = STATE(1383),
    [sym_bitstring] = STATE(1383),
    [sym_map] = STATE(1383),
    [sym__nullary_operator] = STATE(1383),
    [sym_unary_operator] = STATE(1383),
    [sym__capture_expression] = STATE(1188),
    [sym_binary_operator] = STATE(1383),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1383),
    [sym_call] = STATE(1383),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1383),
    [sym_anonymous_function] = STATE(1383),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1755),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1757),
    [sym_integer] = ACTIONS(1797),
    [sym_float] = ACTIONS(1757),
    [sym_char] = ACTIONS(1757),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1757),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(431)] = {
    [sym__expression] = STATE(4377),
    [sym_block] = STATE(4377),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4377),
    [sym_nil] = STATE(4377),
    [sym__atom] = STATE(4377),
    [sym_quoted_atom] = STATE(4377),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4377),
    [sym_charlist] = STATE(4377),
    [sym_sigil] = STATE(4377),
    [sym_list] = STATE(4377),
    [sym_tuple] = STATE(4377),
    [sym_bitstring] = STATE(4377),
    [sym_map] = STATE(4377),
    [sym__nullary_operator] = STATE(4377),
    [sym_unary_operator] = STATE(4377),
    [sym__capture_expression] = STATE(1907),
    [sym_binary_operator] = STATE(4377),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4377),
    [sym_call] = STATE(4377),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4377),
    [sym_anonymous_function] = STATE(4377),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1671),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1799),
    [sym_integer] = ACTIONS(1713),
    [sym_float] = ACTIONS(1799),
    [sym_char] = ACTIONS(1799),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1799),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(432)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1801),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(433)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1803),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(434)] = {
    [sym__expression] = STATE(4377),
    [sym_block] = STATE(4377),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4377),
    [sym_nil] = STATE(4377),
    [sym__atom] = STATE(4377),
    [sym_quoted_atom] = STATE(4377),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4377),
    [sym_charlist] = STATE(4377),
    [sym_sigil] = STATE(4377),
    [sym_list] = STATE(4377),
    [sym_tuple] = STATE(4377),
    [sym_bitstring] = STATE(4377),
    [sym_map] = STATE(4377),
    [sym__nullary_operator] = STATE(4377),
    [sym_unary_operator] = STATE(4377),
    [sym__capture_expression] = STATE(2021),
    [sym_binary_operator] = STATE(4377),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4377),
    [sym_call] = STATE(4377),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4377),
    [sym_anonymous_function] = STATE(4377),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1671),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1799),
    [sym_integer] = ACTIONS(1675),
    [sym_float] = ACTIONS(1799),
    [sym_char] = ACTIONS(1799),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1799),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(435)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1805),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(436)] = {
    [sym__expression] = STATE(4572),
    [sym_block] = STATE(4572),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4572),
    [sym_nil] = STATE(4572),
    [sym__atom] = STATE(4572),
    [sym_quoted_atom] = STATE(4572),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4572),
    [sym_charlist] = STATE(4572),
    [sym_sigil] = STATE(4572),
    [sym_list] = STATE(4572),
    [sym_tuple] = STATE(4572),
    [sym_bitstring] = STATE(4572),
    [sym_map] = STATE(4572),
    [sym__nullary_operator] = STATE(4572),
    [sym_unary_operator] = STATE(4572),
    [sym_binary_operator] = STATE(4572),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4572),
    [sym_call] = STATE(4572),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4572),
    [sym_anonymous_function] = STATE(4572),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1807),
    [sym_integer] = ACTIONS(1807),
    [sym_float] = ACTIONS(1807),
    [sym_char] = ACTIONS(1807),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1807),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_RBRACE] = ACTIONS(1809),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(437)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1811),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(438)] = {
    [sym__expression] = STATE(1917),
    [sym_block] = STATE(1917),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1917),
    [sym_nil] = STATE(1917),
    [sym__atom] = STATE(1917),
    [sym_quoted_atom] = STATE(1917),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1917),
    [sym_charlist] = STATE(1917),
    [sym_sigil] = STATE(1917),
    [sym_list] = STATE(1917),
    [sym_tuple] = STATE(1917),
    [sym_bitstring] = STATE(1917),
    [sym_map] = STATE(1917),
    [sym__nullary_operator] = STATE(1917),
    [sym_unary_operator] = STATE(1917),
    [sym__capture_expression] = STATE(1652),
    [sym_binary_operator] = STATE(1917),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1917),
    [sym_call] = STATE(1917),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1917),
    [sym_anonymous_function] = STATE(1917),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1677),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1763),
    [sym_integer] = ACTIONS(1779),
    [sym_float] = ACTIONS(1763),
    [sym_char] = ACTIONS(1763),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(439)] = {
    [sym__expression] = STATE(2708),
    [sym_block] = STATE(2708),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2708),
    [sym_nil] = STATE(2708),
    [sym__atom] = STATE(2708),
    [sym_quoted_atom] = STATE(2708),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2708),
    [sym_charlist] = STATE(2708),
    [sym_sigil] = STATE(2708),
    [sym_list] = STATE(2708),
    [sym_tuple] = STATE(2708),
    [sym_bitstring] = STATE(2708),
    [sym_map] = STATE(2708),
    [sym__nullary_operator] = STATE(2708),
    [sym_unary_operator] = STATE(2708),
    [sym__capture_expression] = STATE(1309),
    [sym_binary_operator] = STATE(2708),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2708),
    [sym_call] = STATE(2708),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2708),
    [sym_anonymous_function] = STATE(2708),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1755),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1813),
    [sym_integer] = ACTIONS(1759),
    [sym_float] = ACTIONS(1813),
    [sym_char] = ACTIONS(1813),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1813),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(440)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1815),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(441)] = {
    [sym__expression] = STATE(1690),
    [sym_block] = STATE(1690),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1690),
    [sym_nil] = STATE(1690),
    [sym__atom] = STATE(1690),
    [sym_quoted_atom] = STATE(1690),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1690),
    [sym_charlist] = STATE(1690),
    [sym_sigil] = STATE(1690),
    [sym_list] = STATE(1690),
    [sym_tuple] = STATE(1690),
    [sym_bitstring] = STATE(1690),
    [sym_map] = STATE(1690),
    [sym__nullary_operator] = STATE(1690),
    [sym_unary_operator] = STATE(1690),
    [sym__capture_expression] = STATE(1465),
    [sym_binary_operator] = STATE(1690),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1690),
    [sym_call] = STATE(1690),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1690),
    [sym_anonymous_function] = STATE(1690),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1611),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1733),
    [sym_integer] = ACTIONS(1615),
    [sym_float] = ACTIONS(1733),
    [sym_char] = ACTIONS(1733),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1733),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(442)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [anon_sym_RPAREN] = ACTIONS(1817),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(443)] = {
    [sym__expression] = STATE(4447),
    [sym_block] = STATE(4447),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4447),
    [sym_nil] = STATE(4447),
    [sym__atom] = STATE(4447),
    [sym_quoted_atom] = STATE(4447),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4447),
    [sym_charlist] = STATE(4447),
    [sym_sigil] = STATE(4447),
    [sym_list] = STATE(4447),
    [sym_tuple] = STATE(4447),
    [sym_bitstring] = STATE(4447),
    [sym_map] = STATE(4447),
    [sym__nullary_operator] = STATE(4447),
    [sym_unary_operator] = STATE(4447),
    [sym_binary_operator] = STATE(4447),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4447),
    [sym_call] = STATE(4447),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4447),
    [sym_anonymous_function] = STATE(4447),
    [ts_builtin_sym_end] = ACTIONS(1819),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1687),
    [sym_integer] = ACTIONS(1687),
    [sym_float] = ACTIONS(1687),
    [sym_char] = ACTIONS(1687),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1687),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(444)] = {
    [sym__expression] = STATE(2708),
    [sym_block] = STATE(2708),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2708),
    [sym_nil] = STATE(2708),
    [sym__atom] = STATE(2708),
    [sym_quoted_atom] = STATE(2708),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2708),
    [sym_charlist] = STATE(2708),
    [sym_sigil] = STATE(2708),
    [sym_list] = STATE(2708),
    [sym_tuple] = STATE(2708),
    [sym_bitstring] = STATE(2708),
    [sym_map] = STATE(2708),
    [sym__nullary_operator] = STATE(2708),
    [sym_unary_operator] = STATE(2708),
    [sym__capture_expression] = STATE(1188),
    [sym_binary_operator] = STATE(2708),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2708),
    [sym_call] = STATE(2708),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2708),
    [sym_anonymous_function] = STATE(2708),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1755),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1813),
    [sym_integer] = ACTIONS(1797),
    [sym_float] = ACTIONS(1813),
    [sym_char] = ACTIONS(1813),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1813),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(445)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6874),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1821),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(446)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6918),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1823),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(447)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6935),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1825),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(448)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6940),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1827),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(449)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6945),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1829),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(450)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6950),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1831),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(451)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6955),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1833),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(452)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6960),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1835),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(453)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6965),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1837),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(454)] = {
    [sym__expression] = STATE(4594),
    [sym_block] = STATE(4594),
    [sym_identifier] = STATE(112),
    [sym_boolean] = STATE(4594),
    [sym_nil] = STATE(4594),
    [sym__atom] = STATE(4574),
    [sym_quoted_atom] = STATE(4574),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4594),
    [sym_charlist] = STATE(4594),
    [sym_sigil] = STATE(4594),
    [sym_list] = STATE(4594),
    [sym_tuple] = STATE(4594),
    [sym_bitstring] = STATE(4594),
    [sym_map] = STATE(4594),
    [sym_struct] = STATE(6970),
    [sym__nullary_operator] = STATE(4594),
    [sym_unary_operator] = STATE(4574),
    [sym_binary_operator] = STATE(4594),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4574),
    [sym_call] = STATE(4594),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(4568),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4594),
    [sym_anonymous_function] = STATE(4594),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1689),
    [sym_integer] = ACTIONS(1691),
    [sym_float] = ACTIONS(1691),
    [sym_char] = ACTIONS(1691),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(455)] = {
    [sym__expression] = STATE(2291),
    [sym_block] = STATE(2291),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2291),
    [sym_nil] = STATE(2291),
    [sym__atom] = STATE(2291),
    [sym_quoted_atom] = STATE(2291),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2291),
    [sym_charlist] = STATE(2291),
    [sym_sigil] = STATE(2291),
    [sym_list] = STATE(2291),
    [sym_tuple] = STATE(2291),
    [sym_bitstring] = STATE(2291),
    [sym_map] = STATE(2291),
    [sym__nullary_operator] = STATE(2291),
    [sym_unary_operator] = STATE(2291),
    [sym__capture_expression] = STATE(2021),
    [sym_binary_operator] = STATE(2291),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2291),
    [sym_call] = STATE(2291),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2291),
    [sym_anonymous_function] = STATE(2291),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1671),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1711),
    [sym_integer] = ACTIONS(1675),
    [sym_float] = ACTIONS(1711),
    [sym_char] = ACTIONS(1711),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1711),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(456)] = {
    [sym__expression] = STATE(4158),
    [sym_block] = STATE(4158),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4158),
    [sym_nil] = STATE(4158),
    [sym__atom] = STATE(4158),
    [sym_quoted_atom] = STATE(4158),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4158),
    [sym_charlist] = STATE(4158),
    [sym_sigil] = STATE(4158),
    [sym_list] = STATE(4158),
    [sym_tuple] = STATE(4158),
    [sym_bitstring] = STATE(4158),
    [sym_map] = STATE(4158),
    [sym__nullary_operator] = STATE(4158),
    [sym_unary_operator] = STATE(4158),
    [sym_binary_operator] = STATE(4158),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4158),
    [sym_call] = STATE(4158),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4158),
    [sym_anonymous_function] = STATE(4158),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1841),
    [sym_integer] = ACTIONS(1841),
    [sym_float] = ACTIONS(1841),
    [sym_char] = ACTIONS(1841),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1841),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(457)] = {
    [sym__expression] = STATE(4195),
    [sym_block] = STATE(4195),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4195),
    [sym_nil] = STATE(4195),
    [sym__atom] = STATE(4195),
    [sym_quoted_atom] = STATE(4195),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4195),
    [sym_charlist] = STATE(4195),
    [sym_sigil] = STATE(4195),
    [sym_list] = STATE(4195),
    [sym_tuple] = STATE(4195),
    [sym_bitstring] = STATE(4195),
    [sym_map] = STATE(4195),
    [sym__nullary_operator] = STATE(4195),
    [sym_unary_operator] = STATE(4195),
    [sym_binary_operator] = STATE(4195),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4195),
    [sym_call] = STATE(4195),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4195),
    [sym_anonymous_function] = STATE(4195),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1843),
    [sym_integer] = ACTIONS(1843),
    [sym_float] = ACTIONS(1843),
    [sym_char] = ACTIONS(1843),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1843),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(458)] = {
    [sym__expression] = STATE(4555),
    [sym_block] = STATE(4555),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4555),
    [sym_nil] = STATE(4555),
    [sym__atom] = STATE(4555),
    [sym_quoted_atom] = STATE(4555),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4555),
    [sym_charlist] = STATE(4555),
    [sym_sigil] = STATE(4555),
    [sym_list] = STATE(4555),
    [sym_tuple] = STATE(4555),
    [sym_bitstring] = STATE(4555),
    [sym_map] = STATE(4555),
    [sym__nullary_operator] = STATE(4555),
    [sym_unary_operator] = STATE(4555),
    [sym_binary_operator] = STATE(4555),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4555),
    [sym_call] = STATE(4555),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4555),
    [sym_anonymous_function] = STATE(4555),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1845),
    [sym_integer] = ACTIONS(1845),
    [sym_float] = ACTIONS(1845),
    [sym_char] = ACTIONS(1845),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1845),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(459)] = {
    [sym__expression] = STATE(4556),
    [sym_block] = STATE(4556),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4556),
    [sym_nil] = STATE(4556),
    [sym__atom] = STATE(4556),
    [sym_quoted_atom] = STATE(4556),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4556),
    [sym_charlist] = STATE(4556),
    [sym_sigil] = STATE(4556),
    [sym_list] = STATE(4556),
    [sym_tuple] = STATE(4556),
    [sym_bitstring] = STATE(4556),
    [sym_map] = STATE(4556),
    [sym__nullary_operator] = STATE(4556),
    [sym_unary_operator] = STATE(4556),
    [sym_binary_operator] = STATE(4556),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4556),
    [sym_call] = STATE(4556),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4556),
    [sym_anonymous_function] = STATE(4556),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1847),
    [sym_integer] = ACTIONS(1847),
    [sym_float] = ACTIONS(1847),
    [sym_char] = ACTIONS(1847),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1847),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(460)] = {
    [sym__expression] = STATE(4199),
    [sym_block] = STATE(4199),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4199),
    [sym_nil] = STATE(4199),
    [sym__atom] = STATE(4199),
    [sym_quoted_atom] = STATE(4199),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4199),
    [sym_charlist] = STATE(4199),
    [sym_sigil] = STATE(4199),
    [sym_list] = STATE(4199),
    [sym_tuple] = STATE(4199),
    [sym_bitstring] = STATE(4199),
    [sym_map] = STATE(4199),
    [sym__nullary_operator] = STATE(4199),
    [sym_unary_operator] = STATE(4199),
    [sym_binary_operator] = STATE(4199),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4199),
    [sym_call] = STATE(4199),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4199),
    [sym_anonymous_function] = STATE(4199),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1849),
    [sym_integer] = ACTIONS(1849),
    [sym_float] = ACTIONS(1849),
    [sym_char] = ACTIONS(1849),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1849),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(461)] = {
    [sym__expression] = STATE(4447),
    [sym_block] = STATE(4447),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4447),
    [sym_nil] = STATE(4447),
    [sym__atom] = STATE(4447),
    [sym_quoted_atom] = STATE(4447),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4447),
    [sym_charlist] = STATE(4447),
    [sym_sigil] = STATE(4447),
    [sym_list] = STATE(4447),
    [sym_tuple] = STATE(4447),
    [sym_bitstring] = STATE(4447),
    [sym_map] = STATE(4447),
    [sym__nullary_operator] = STATE(4447),
    [sym_unary_operator] = STATE(4447),
    [sym_binary_operator] = STATE(4447),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4447),
    [sym_call] = STATE(4447),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4447),
    [sym_anonymous_function] = STATE(4447),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1687),
    [sym_integer] = ACTIONS(1687),
    [sym_float] = ACTIONS(1687),
    [sym_char] = ACTIONS(1687),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1687),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(462)] = {
    [sym__expression] = STATE(4504),
    [sym_block] = STATE(4504),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4504),
    [sym_nil] = STATE(4504),
    [sym__atom] = STATE(4504),
    [sym_quoted_atom] = STATE(4504),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4504),
    [sym_charlist] = STATE(4504),
    [sym_sigil] = STATE(4504),
    [sym_list] = STATE(4504),
    [sym_tuple] = STATE(4504),
    [sym_bitstring] = STATE(4504),
    [sym_map] = STATE(4504),
    [sym__nullary_operator] = STATE(4504),
    [sym_unary_operator] = STATE(4504),
    [sym_binary_operator] = STATE(4504),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4504),
    [sym_call] = STATE(4504),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4504),
    [sym_anonymous_function] = STATE(4504),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1851),
    [sym_integer] = ACTIONS(1851),
    [sym_float] = ACTIONS(1851),
    [sym_char] = ACTIONS(1851),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1851),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(463)] = {
    [sym__expression] = STATE(4435),
    [sym_block] = STATE(4435),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4435),
    [sym_nil] = STATE(4435),
    [sym__atom] = STATE(4435),
    [sym_quoted_atom] = STATE(4435),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4435),
    [sym_charlist] = STATE(4435),
    [sym_sigil] = STATE(4435),
    [sym_list] = STATE(4435),
    [sym_tuple] = STATE(4435),
    [sym_bitstring] = STATE(4435),
    [sym_map] = STATE(4435),
    [sym__nullary_operator] = STATE(4435),
    [sym_unary_operator] = STATE(4435),
    [sym_binary_operator] = STATE(4435),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4435),
    [sym_call] = STATE(4435),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4435),
    [sym_anonymous_function] = STATE(4435),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1853),
    [sym_integer] = ACTIONS(1853),
    [sym_float] = ACTIONS(1853),
    [sym_char] = ACTIONS(1853),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1853),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(464)] = {
    [sym__expression] = STATE(4529),
    [sym_block] = STATE(4529),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4529),
    [sym_nil] = STATE(4529),
    [sym__atom] = STATE(4529),
    [sym_quoted_atom] = STATE(4529),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4529),
    [sym_charlist] = STATE(4529),
    [sym_sigil] = STATE(4529),
    [sym_list] = STATE(4529),
    [sym_tuple] = STATE(4529),
    [sym_bitstring] = STATE(4529),
    [sym_map] = STATE(4529),
    [sym__nullary_operator] = STATE(4529),
    [sym_unary_operator] = STATE(4529),
    [sym_binary_operator] = STATE(4529),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4529),
    [sym_call] = STATE(4529),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4529),
    [sym_anonymous_function] = STATE(4529),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1855),
    [sym_integer] = ACTIONS(1855),
    [sym_float] = ACTIONS(1855),
    [sym_char] = ACTIONS(1855),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1855),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(465)] = {
    [sym__expression] = STATE(4540),
    [sym_block] = STATE(4540),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4540),
    [sym_nil] = STATE(4540),
    [sym__atom] = STATE(4540),
    [sym_quoted_atom] = STATE(4540),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4540),
    [sym_charlist] = STATE(4540),
    [sym_sigil] = STATE(4540),
    [sym_list] = STATE(4540),
    [sym_tuple] = STATE(4540),
    [sym_bitstring] = STATE(4540),
    [sym_map] = STATE(4540),
    [sym__nullary_operator] = STATE(4540),
    [sym_unary_operator] = STATE(4540),
    [sym_binary_operator] = STATE(4540),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4540),
    [sym_call] = STATE(4540),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4540),
    [sym_anonymous_function] = STATE(4540),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1857),
    [sym_integer] = ACTIONS(1857),
    [sym_float] = ACTIONS(1857),
    [sym_char] = ACTIONS(1857),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1857),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(466)] = {
    [sym__expression] = STATE(4569),
    [sym_block] = STATE(4569),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4569),
    [sym_nil] = STATE(4569),
    [sym__atom] = STATE(4569),
    [sym_quoted_atom] = STATE(4569),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4569),
    [sym_charlist] = STATE(4569),
    [sym_sigil] = STATE(4569),
    [sym_list] = STATE(4569),
    [sym_tuple] = STATE(4569),
    [sym_bitstring] = STATE(4569),
    [sym_map] = STATE(4569),
    [sym__nullary_operator] = STATE(4569),
    [sym_unary_operator] = STATE(4569),
    [sym_binary_operator] = STATE(4569),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4569),
    [sym_call] = STATE(4569),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4569),
    [sym_anonymous_function] = STATE(4569),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1859),
    [sym_integer] = ACTIONS(1859),
    [sym_float] = ACTIONS(1859),
    [sym_char] = ACTIONS(1859),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1859),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(467)] = {
    [sym__expression] = STATE(4480),
    [sym_block] = STATE(4480),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4480),
    [sym_nil] = STATE(4480),
    [sym__atom] = STATE(4480),
    [sym_quoted_atom] = STATE(4480),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4480),
    [sym_charlist] = STATE(4480),
    [sym_sigil] = STATE(4480),
    [sym_list] = STATE(4480),
    [sym_tuple] = STATE(4480),
    [sym_bitstring] = STATE(4480),
    [sym_map] = STATE(4480),
    [sym__nullary_operator] = STATE(4480),
    [sym_unary_operator] = STATE(4480),
    [sym_binary_operator] = STATE(4480),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4480),
    [sym_call] = STATE(4480),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4480),
    [sym_anonymous_function] = STATE(4480),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1861),
    [sym_integer] = ACTIONS(1861),
    [sym_float] = ACTIONS(1861),
    [sym_char] = ACTIONS(1861),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1861),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(468)] = {
    [sym__expression] = STATE(4481),
    [sym_block] = STATE(4481),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4481),
    [sym_nil] = STATE(4481),
    [sym__atom] = STATE(4481),
    [sym_quoted_atom] = STATE(4481),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4481),
    [sym_charlist] = STATE(4481),
    [sym_sigil] = STATE(4481),
    [sym_list] = STATE(4481),
    [sym_tuple] = STATE(4481),
    [sym_bitstring] = STATE(4481),
    [sym_map] = STATE(4481),
    [sym__nullary_operator] = STATE(4481),
    [sym_unary_operator] = STATE(4481),
    [sym_binary_operator] = STATE(4481),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4481),
    [sym_call] = STATE(4481),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4481),
    [sym_anonymous_function] = STATE(4481),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1863),
    [sym_integer] = ACTIONS(1863),
    [sym_float] = ACTIONS(1863),
    [sym_char] = ACTIONS(1863),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(469)] = {
    [sym__expression] = STATE(4484),
    [sym_block] = STATE(4484),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4484),
    [sym_nil] = STATE(4484),
    [sym__atom] = STATE(4484),
    [sym_quoted_atom] = STATE(4484),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4484),
    [sym_charlist] = STATE(4484),
    [sym_sigil] = STATE(4484),
    [sym_list] = STATE(4484),
    [sym_tuple] = STATE(4484),
    [sym_bitstring] = STATE(4484),
    [sym_map] = STATE(4484),
    [sym__nullary_operator] = STATE(4484),
    [sym_unary_operator] = STATE(4484),
    [sym_binary_operator] = STATE(4484),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4484),
    [sym_call] = STATE(4484),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4484),
    [sym_anonymous_function] = STATE(4484),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1865),
    [sym_integer] = ACTIONS(1865),
    [sym_float] = ACTIONS(1865),
    [sym_char] = ACTIONS(1865),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1865),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(470)] = {
    [sym__expression] = STATE(4495),
    [sym_block] = STATE(4495),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4495),
    [sym_nil] = STATE(4495),
    [sym__atom] = STATE(4495),
    [sym_quoted_atom] = STATE(4495),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4495),
    [sym_charlist] = STATE(4495),
    [sym_sigil] = STATE(4495),
    [sym_list] = STATE(4495),
    [sym_tuple] = STATE(4495),
    [sym_bitstring] = STATE(4495),
    [sym_map] = STATE(4495),
    [sym__nullary_operator] = STATE(4495),
    [sym_unary_operator] = STATE(4495),
    [sym_binary_operator] = STATE(4495),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4495),
    [sym_call] = STATE(4495),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4495),
    [sym_anonymous_function] = STATE(4495),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1867),
    [sym_integer] = ACTIONS(1867),
    [sym_float] = ACTIONS(1867),
    [sym_char] = ACTIONS(1867),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1867),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(471)] = {
    [sym__expression] = STATE(4432),
    [sym_block] = STATE(4432),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4432),
    [sym_nil] = STATE(4432),
    [sym__atom] = STATE(4432),
    [sym_quoted_atom] = STATE(4432),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4432),
    [sym_charlist] = STATE(4432),
    [sym_sigil] = STATE(4432),
    [sym_list] = STATE(4432),
    [sym_tuple] = STATE(4432),
    [sym_bitstring] = STATE(4432),
    [sym_map] = STATE(4432),
    [sym__nullary_operator] = STATE(4432),
    [sym_unary_operator] = STATE(4432),
    [sym_binary_operator] = STATE(4432),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4432),
    [sym_call] = STATE(4432),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4432),
    [sym_anonymous_function] = STATE(4432),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1869),
    [sym_integer] = ACTIONS(1869),
    [sym_float] = ACTIONS(1869),
    [sym_char] = ACTIONS(1869),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1869),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(472)] = {
    [sym__expression] = STATE(4508),
    [sym_block] = STATE(4508),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4508),
    [sym_nil] = STATE(4508),
    [sym__atom] = STATE(4508),
    [sym_quoted_atom] = STATE(4508),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4508),
    [sym_charlist] = STATE(4508),
    [sym_sigil] = STATE(4508),
    [sym_list] = STATE(4508),
    [sym_tuple] = STATE(4508),
    [sym_bitstring] = STATE(4508),
    [sym_map] = STATE(4508),
    [sym__nullary_operator] = STATE(4508),
    [sym_unary_operator] = STATE(4508),
    [sym_binary_operator] = STATE(4508),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4508),
    [sym_call] = STATE(4508),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4508),
    [sym_anonymous_function] = STATE(4508),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1871),
    [sym_integer] = ACTIONS(1871),
    [sym_float] = ACTIONS(1871),
    [sym_char] = ACTIONS(1871),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1871),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(473)] = {
    [sym__expression] = STATE(4511),
    [sym_block] = STATE(4511),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4511),
    [sym_nil] = STATE(4511),
    [sym__atom] = STATE(4511),
    [sym_quoted_atom] = STATE(4511),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4511),
    [sym_charlist] = STATE(4511),
    [sym_sigil] = STATE(4511),
    [sym_list] = STATE(4511),
    [sym_tuple] = STATE(4511),
    [sym_bitstring] = STATE(4511),
    [sym_map] = STATE(4511),
    [sym__nullary_operator] = STATE(4511),
    [sym_unary_operator] = STATE(4511),
    [sym_binary_operator] = STATE(4511),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4511),
    [sym_call] = STATE(4511),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4511),
    [sym_anonymous_function] = STATE(4511),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1873),
    [sym_integer] = ACTIONS(1873),
    [sym_float] = ACTIONS(1873),
    [sym_char] = ACTIONS(1873),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1873),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(474)] = {
    [sym__expression] = STATE(4512),
    [sym_block] = STATE(4512),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4512),
    [sym_nil] = STATE(4512),
    [sym__atom] = STATE(4512),
    [sym_quoted_atom] = STATE(4512),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4512),
    [sym_charlist] = STATE(4512),
    [sym_sigil] = STATE(4512),
    [sym_list] = STATE(4512),
    [sym_tuple] = STATE(4512),
    [sym_bitstring] = STATE(4512),
    [sym_map] = STATE(4512),
    [sym__nullary_operator] = STATE(4512),
    [sym_unary_operator] = STATE(4512),
    [sym_binary_operator] = STATE(4512),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4512),
    [sym_call] = STATE(4512),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4512),
    [sym_anonymous_function] = STATE(4512),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1875),
    [sym_integer] = ACTIONS(1875),
    [sym_float] = ACTIONS(1875),
    [sym_char] = ACTIONS(1875),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1875),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(475)] = {
    [sym__expression] = STATE(4513),
    [sym_block] = STATE(4513),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4513),
    [sym_nil] = STATE(4513),
    [sym__atom] = STATE(4513),
    [sym_quoted_atom] = STATE(4513),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4513),
    [sym_charlist] = STATE(4513),
    [sym_sigil] = STATE(4513),
    [sym_list] = STATE(4513),
    [sym_tuple] = STATE(4513),
    [sym_bitstring] = STATE(4513),
    [sym_map] = STATE(4513),
    [sym__nullary_operator] = STATE(4513),
    [sym_unary_operator] = STATE(4513),
    [sym_binary_operator] = STATE(4513),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4513),
    [sym_call] = STATE(4513),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4513),
    [sym_anonymous_function] = STATE(4513),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1877),
    [sym_integer] = ACTIONS(1877),
    [sym_float] = ACTIONS(1877),
    [sym_char] = ACTIONS(1877),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1877),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(476)] = {
    [sym__expression] = STATE(4520),
    [sym_block] = STATE(4520),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4520),
    [sym_nil] = STATE(4520),
    [sym__atom] = STATE(4520),
    [sym_quoted_atom] = STATE(4520),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4520),
    [sym_charlist] = STATE(4520),
    [sym_sigil] = STATE(4520),
    [sym_list] = STATE(4520),
    [sym_tuple] = STATE(4520),
    [sym_bitstring] = STATE(4520),
    [sym_map] = STATE(4520),
    [sym__nullary_operator] = STATE(4520),
    [sym_unary_operator] = STATE(4520),
    [sym_binary_operator] = STATE(4520),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4520),
    [sym_call] = STATE(4520),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4520),
    [sym_anonymous_function] = STATE(4520),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1879),
    [sym_integer] = ACTIONS(1879),
    [sym_float] = ACTIONS(1879),
    [sym_char] = ACTIONS(1879),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1879),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(477)] = {
    [sym__expression] = STATE(4522),
    [sym_block] = STATE(4522),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4522),
    [sym_nil] = STATE(4522),
    [sym__atom] = STATE(4522),
    [sym_quoted_atom] = STATE(4522),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4522),
    [sym_charlist] = STATE(4522),
    [sym_sigil] = STATE(4522),
    [sym_list] = STATE(4522),
    [sym_tuple] = STATE(4522),
    [sym_bitstring] = STATE(4522),
    [sym_map] = STATE(4522),
    [sym__nullary_operator] = STATE(4522),
    [sym_unary_operator] = STATE(4522),
    [sym_binary_operator] = STATE(4522),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4522),
    [sym_call] = STATE(4522),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4522),
    [sym_anonymous_function] = STATE(4522),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1881),
    [sym_integer] = ACTIONS(1881),
    [sym_float] = ACTIONS(1881),
    [sym_char] = ACTIONS(1881),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1881),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(478)] = {
    [sym__expression] = STATE(4523),
    [sym_block] = STATE(4523),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4523),
    [sym_nil] = STATE(4523),
    [sym__atom] = STATE(4523),
    [sym_quoted_atom] = STATE(4523),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4523),
    [sym_charlist] = STATE(4523),
    [sym_sigil] = STATE(4523),
    [sym_list] = STATE(4523),
    [sym_tuple] = STATE(4523),
    [sym_bitstring] = STATE(4523),
    [sym_map] = STATE(4523),
    [sym__nullary_operator] = STATE(4523),
    [sym_unary_operator] = STATE(4523),
    [sym_binary_operator] = STATE(4523),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4523),
    [sym_call] = STATE(4523),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4523),
    [sym_anonymous_function] = STATE(4523),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1883),
    [sym_integer] = ACTIONS(1883),
    [sym_float] = ACTIONS(1883),
    [sym_char] = ACTIONS(1883),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1883),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(479)] = {
    [sym__expression] = STATE(4524),
    [sym_block] = STATE(4524),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4524),
    [sym_nil] = STATE(4524),
    [sym__atom] = STATE(4524),
    [sym_quoted_atom] = STATE(4524),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4524),
    [sym_charlist] = STATE(4524),
    [sym_sigil] = STATE(4524),
    [sym_list] = STATE(4524),
    [sym_tuple] = STATE(4524),
    [sym_bitstring] = STATE(4524),
    [sym_map] = STATE(4524),
    [sym__nullary_operator] = STATE(4524),
    [sym_unary_operator] = STATE(4524),
    [sym_binary_operator] = STATE(4524),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4524),
    [sym_call] = STATE(4524),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4524),
    [sym_anonymous_function] = STATE(4524),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1885),
    [sym_integer] = ACTIONS(1885),
    [sym_float] = ACTIONS(1885),
    [sym_char] = ACTIONS(1885),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1885),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(480)] = {
    [sym__expression] = STATE(4530),
    [sym_block] = STATE(4530),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4530),
    [sym_nil] = STATE(4530),
    [sym__atom] = STATE(4530),
    [sym_quoted_atom] = STATE(4530),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4530),
    [sym_charlist] = STATE(4530),
    [sym_sigil] = STATE(4530),
    [sym_list] = STATE(4530),
    [sym_tuple] = STATE(4530),
    [sym_bitstring] = STATE(4530),
    [sym_map] = STATE(4530),
    [sym__nullary_operator] = STATE(4530),
    [sym_unary_operator] = STATE(4530),
    [sym_binary_operator] = STATE(4530),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4530),
    [sym_call] = STATE(4530),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4530),
    [sym_anonymous_function] = STATE(4530),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1887),
    [sym_integer] = ACTIONS(1887),
    [sym_float] = ACTIONS(1887),
    [sym_char] = ACTIONS(1887),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1887),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(481)] = {
    [sym__expression] = STATE(4286),
    [sym_block] = STATE(4286),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4286),
    [sym_nil] = STATE(4286),
    [sym__atom] = STATE(4286),
    [sym_quoted_atom] = STATE(4286),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4286),
    [sym_charlist] = STATE(4286),
    [sym_sigil] = STATE(4286),
    [sym_list] = STATE(4286),
    [sym_tuple] = STATE(4286),
    [sym_bitstring] = STATE(4286),
    [sym_map] = STATE(4286),
    [sym__nullary_operator] = STATE(4286),
    [sym_unary_operator] = STATE(4286),
    [sym_binary_operator] = STATE(4286),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4286),
    [sym_call] = STATE(4286),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4286),
    [sym_anonymous_function] = STATE(4286),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1889),
    [sym_integer] = ACTIONS(1889),
    [sym_float] = ACTIONS(1889),
    [sym_char] = ACTIONS(1889),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1889),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(482)] = {
    [sym__expression] = STATE(4147),
    [sym_block] = STATE(4147),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4147),
    [sym_nil] = STATE(4147),
    [sym__atom] = STATE(4147),
    [sym_quoted_atom] = STATE(4147),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4147),
    [sym_charlist] = STATE(4147),
    [sym_sigil] = STATE(4147),
    [sym_list] = STATE(4147),
    [sym_tuple] = STATE(4147),
    [sym_bitstring] = STATE(4147),
    [sym_map] = STATE(4147),
    [sym__nullary_operator] = STATE(4147),
    [sym_unary_operator] = STATE(4147),
    [sym_binary_operator] = STATE(4147),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4147),
    [sym_call] = STATE(4147),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4147),
    [sym_anonymous_function] = STATE(4147),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1891),
    [sym_integer] = ACTIONS(1891),
    [sym_float] = ACTIONS(1891),
    [sym_char] = ACTIONS(1891),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1891),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(483)] = {
    [sym__expression] = STATE(4148),
    [sym_block] = STATE(4148),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4148),
    [sym_nil] = STATE(4148),
    [sym__atom] = STATE(4148),
    [sym_quoted_atom] = STATE(4148),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4148),
    [sym_charlist] = STATE(4148),
    [sym_sigil] = STATE(4148),
    [sym_list] = STATE(4148),
    [sym_tuple] = STATE(4148),
    [sym_bitstring] = STATE(4148),
    [sym_map] = STATE(4148),
    [sym__nullary_operator] = STATE(4148),
    [sym_unary_operator] = STATE(4148),
    [sym_binary_operator] = STATE(4148),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4148),
    [sym_call] = STATE(4148),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4148),
    [sym_anonymous_function] = STATE(4148),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1893),
    [sym_integer] = ACTIONS(1893),
    [sym_float] = ACTIONS(1893),
    [sym_char] = ACTIONS(1893),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1893),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(484)] = {
    [sym__expression] = STATE(4289),
    [sym_block] = STATE(4289),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4289),
    [sym_nil] = STATE(4289),
    [sym__atom] = STATE(4289),
    [sym_quoted_atom] = STATE(4289),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4289),
    [sym_charlist] = STATE(4289),
    [sym_sigil] = STATE(4289),
    [sym_list] = STATE(4289),
    [sym_tuple] = STATE(4289),
    [sym_bitstring] = STATE(4289),
    [sym_map] = STATE(4289),
    [sym__nullary_operator] = STATE(4289),
    [sym_unary_operator] = STATE(4289),
    [sym_binary_operator] = STATE(4289),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4289),
    [sym_call] = STATE(4289),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4289),
    [sym_anonymous_function] = STATE(4289),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1895),
    [sym_integer] = ACTIONS(1895),
    [sym_float] = ACTIONS(1895),
    [sym_char] = ACTIONS(1895),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1895),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(485)] = {
    [sym__expression] = STATE(4181),
    [sym_block] = STATE(4181),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4181),
    [sym_nil] = STATE(4181),
    [sym__atom] = STATE(4181),
    [sym_quoted_atom] = STATE(4181),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4181),
    [sym_charlist] = STATE(4181),
    [sym_sigil] = STATE(4181),
    [sym_list] = STATE(4181),
    [sym_tuple] = STATE(4181),
    [sym_bitstring] = STATE(4181),
    [sym_map] = STATE(4181),
    [sym__nullary_operator] = STATE(4181),
    [sym_unary_operator] = STATE(4181),
    [sym_binary_operator] = STATE(4181),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4181),
    [sym_call] = STATE(4181),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4181),
    [sym_anonymous_function] = STATE(4181),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1897),
    [sym_integer] = ACTIONS(1897),
    [sym_float] = ACTIONS(1897),
    [sym_char] = ACTIONS(1897),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1897),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(486)] = {
    [sym__expression] = STATE(4187),
    [sym_block] = STATE(4187),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4187),
    [sym_nil] = STATE(4187),
    [sym__atom] = STATE(4187),
    [sym_quoted_atom] = STATE(4187),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4187),
    [sym_charlist] = STATE(4187),
    [sym_sigil] = STATE(4187),
    [sym_list] = STATE(4187),
    [sym_tuple] = STATE(4187),
    [sym_bitstring] = STATE(4187),
    [sym_map] = STATE(4187),
    [sym__nullary_operator] = STATE(4187),
    [sym_unary_operator] = STATE(4187),
    [sym_binary_operator] = STATE(4187),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4187),
    [sym_call] = STATE(4187),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4187),
    [sym_anonymous_function] = STATE(4187),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1899),
    [sym_integer] = ACTIONS(1899),
    [sym_float] = ACTIONS(1899),
    [sym_char] = ACTIONS(1899),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1899),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(487)] = {
    [sym__expression] = STATE(4231),
    [sym_block] = STATE(4231),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4231),
    [sym_nil] = STATE(4231),
    [sym__atom] = STATE(4231),
    [sym_quoted_atom] = STATE(4231),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4231),
    [sym_charlist] = STATE(4231),
    [sym_sigil] = STATE(4231),
    [sym_list] = STATE(4231),
    [sym_tuple] = STATE(4231),
    [sym_bitstring] = STATE(4231),
    [sym_map] = STATE(4231),
    [sym__nullary_operator] = STATE(4231),
    [sym_unary_operator] = STATE(4231),
    [sym_binary_operator] = STATE(4231),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4231),
    [sym_call] = STATE(4231),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4231),
    [sym_anonymous_function] = STATE(4231),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1901),
    [sym_integer] = ACTIONS(1901),
    [sym_float] = ACTIONS(1901),
    [sym_char] = ACTIONS(1901),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1901),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(488)] = {
    [sym__expression] = STATE(4234),
    [sym_block] = STATE(4234),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4234),
    [sym_nil] = STATE(4234),
    [sym__atom] = STATE(4234),
    [sym_quoted_atom] = STATE(4234),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4234),
    [sym_charlist] = STATE(4234),
    [sym_sigil] = STATE(4234),
    [sym_list] = STATE(4234),
    [sym_tuple] = STATE(4234),
    [sym_bitstring] = STATE(4234),
    [sym_map] = STATE(4234),
    [sym__nullary_operator] = STATE(4234),
    [sym_unary_operator] = STATE(4234),
    [sym_binary_operator] = STATE(4234),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4234),
    [sym_call] = STATE(4234),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4234),
    [sym_anonymous_function] = STATE(4234),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1903),
    [sym_integer] = ACTIONS(1903),
    [sym_float] = ACTIONS(1903),
    [sym_char] = ACTIONS(1903),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1903),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(489)] = {
    [sym__expression] = STATE(4215),
    [sym_block] = STATE(4215),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4215),
    [sym_nil] = STATE(4215),
    [sym__atom] = STATE(4215),
    [sym_quoted_atom] = STATE(4215),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4215),
    [sym_charlist] = STATE(4215),
    [sym_sigil] = STATE(4215),
    [sym_list] = STATE(4215),
    [sym_tuple] = STATE(4215),
    [sym_bitstring] = STATE(4215),
    [sym_map] = STATE(4215),
    [sym__nullary_operator] = STATE(4215),
    [sym_unary_operator] = STATE(4215),
    [sym_binary_operator] = STATE(4215),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4215),
    [sym_call] = STATE(4215),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4215),
    [sym_anonymous_function] = STATE(4215),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(1905),
    [sym_integer] = ACTIONS(1905),
    [sym_float] = ACTIONS(1905),
    [sym_char] = ACTIONS(1905),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(1905),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(490)] = {
    [sym__expression] = STATE(4238),
    [sym_block] = STATE(4238),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4238),
    [sym_nil] = STATE(4238),
    [sym__atom] = STATE(4238),
    [sym_quoted_atom] = STATE(4238),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4238),
    [sym_charlist] = STATE(4238),
    [sym_sigil] = STATE(4238),
    [sym_list] = STATE(4238),
    [sym_tuple] = STATE(4238),
    [sym_bitstring] = STATE(4238),
    [sym_map] = STATE(4238),
    [sym__nullary_operator] = STATE(4238),
    [sym_unary_operator] = STATE(4238),
    [sym_binary_operator] = STATE(4238),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4238),
    [sym_call] = STATE(4238),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4238),
    [sym_anonymous_function] = STATE(4238),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1907),
    [sym_integer] = ACTIONS(1907),
    [sym_float] = ACTIONS(1907),
    [sym_char] = ACTIONS(1907),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1907),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(491)] = {
    [sym__expression] = STATE(4239),
    [sym_block] = STATE(4239),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4239),
    [sym_nil] = STATE(4239),
    [sym__atom] = STATE(4239),
    [sym_quoted_atom] = STATE(4239),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4239),
    [sym_charlist] = STATE(4239),
    [sym_sigil] = STATE(4239),
    [sym_list] = STATE(4239),
    [sym_tuple] = STATE(4239),
    [sym_bitstring] = STATE(4239),
    [sym_map] = STATE(4239),
    [sym__nullary_operator] = STATE(4239),
    [sym_unary_operator] = STATE(4239),
    [sym_binary_operator] = STATE(4239),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4239),
    [sym_call] = STATE(4239),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4239),
    [sym_anonymous_function] = STATE(4239),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1909),
    [sym_integer] = ACTIONS(1909),
    [sym_float] = ACTIONS(1909),
    [sym_char] = ACTIONS(1909),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1909),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(492)] = {
    [sym__expression] = STATE(4242),
    [sym_block] = STATE(4242),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4242),
    [sym_nil] = STATE(4242),
    [sym__atom] = STATE(4242),
    [sym_quoted_atom] = STATE(4242),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4242),
    [sym_charlist] = STATE(4242),
    [sym_sigil] = STATE(4242),
    [sym_list] = STATE(4242),
    [sym_tuple] = STATE(4242),
    [sym_bitstring] = STATE(4242),
    [sym_map] = STATE(4242),
    [sym__nullary_operator] = STATE(4242),
    [sym_unary_operator] = STATE(4242),
    [sym_binary_operator] = STATE(4242),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4242),
    [sym_call] = STATE(4242),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4242),
    [sym_anonymous_function] = STATE(4242),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1911),
    [sym_integer] = ACTIONS(1911),
    [sym_float] = ACTIONS(1911),
    [sym_char] = ACTIONS(1911),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1911),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(493)] = {
    [sym__expression] = STATE(4246),
    [sym_block] = STATE(4246),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4246),
    [sym_nil] = STATE(4246),
    [sym__atom] = STATE(4246),
    [sym_quoted_atom] = STATE(4246),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4246),
    [sym_charlist] = STATE(4246),
    [sym_sigil] = STATE(4246),
    [sym_list] = STATE(4246),
    [sym_tuple] = STATE(4246),
    [sym_bitstring] = STATE(4246),
    [sym_map] = STATE(4246),
    [sym__nullary_operator] = STATE(4246),
    [sym_unary_operator] = STATE(4246),
    [sym_binary_operator] = STATE(4246),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4246),
    [sym_call] = STATE(4246),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4246),
    [sym_anonymous_function] = STATE(4246),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1913),
    [sym_integer] = ACTIONS(1913),
    [sym_float] = ACTIONS(1913),
    [sym_char] = ACTIONS(1913),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1913),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(494)] = {
    [sym__expression] = STATE(4259),
    [sym_block] = STATE(4259),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4259),
    [sym_nil] = STATE(4259),
    [sym__atom] = STATE(4259),
    [sym_quoted_atom] = STATE(4259),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4259),
    [sym_charlist] = STATE(4259),
    [sym_sigil] = STATE(4259),
    [sym_list] = STATE(4259),
    [sym_tuple] = STATE(4259),
    [sym_bitstring] = STATE(4259),
    [sym_map] = STATE(4259),
    [sym__nullary_operator] = STATE(4259),
    [sym_unary_operator] = STATE(4259),
    [sym_binary_operator] = STATE(4259),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4259),
    [sym_call] = STATE(4259),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4259),
    [sym_anonymous_function] = STATE(4259),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1915),
    [sym_integer] = ACTIONS(1915),
    [sym_float] = ACTIONS(1915),
    [sym_char] = ACTIONS(1915),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1915),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(495)] = {
    [sym__expression] = STATE(4262),
    [sym_block] = STATE(4262),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4262),
    [sym_nil] = STATE(4262),
    [sym__atom] = STATE(4262),
    [sym_quoted_atom] = STATE(4262),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4262),
    [sym_charlist] = STATE(4262),
    [sym_sigil] = STATE(4262),
    [sym_list] = STATE(4262),
    [sym_tuple] = STATE(4262),
    [sym_bitstring] = STATE(4262),
    [sym_map] = STATE(4262),
    [sym__nullary_operator] = STATE(4262),
    [sym_unary_operator] = STATE(4262),
    [sym_binary_operator] = STATE(4262),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4262),
    [sym_call] = STATE(4262),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4262),
    [sym_anonymous_function] = STATE(4262),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1917),
    [sym_integer] = ACTIONS(1917),
    [sym_float] = ACTIONS(1917),
    [sym_char] = ACTIONS(1917),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1917),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(496)] = {
    [sym__expression] = STATE(4263),
    [sym_block] = STATE(4263),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4263),
    [sym_nil] = STATE(4263),
    [sym__atom] = STATE(4263),
    [sym_quoted_atom] = STATE(4263),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4263),
    [sym_charlist] = STATE(4263),
    [sym_sigil] = STATE(4263),
    [sym_list] = STATE(4263),
    [sym_tuple] = STATE(4263),
    [sym_bitstring] = STATE(4263),
    [sym_map] = STATE(4263),
    [sym__nullary_operator] = STATE(4263),
    [sym_unary_operator] = STATE(4263),
    [sym_binary_operator] = STATE(4263),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4263),
    [sym_call] = STATE(4263),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4263),
    [sym_anonymous_function] = STATE(4263),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1919),
    [sym_integer] = ACTIONS(1919),
    [sym_float] = ACTIONS(1919),
    [sym_char] = ACTIONS(1919),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1919),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(497)] = {
    [sym__expression] = STATE(4266),
    [sym_block] = STATE(4266),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4266),
    [sym_nil] = STATE(4266),
    [sym__atom] = STATE(4266),
    [sym_quoted_atom] = STATE(4266),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4266),
    [sym_charlist] = STATE(4266),
    [sym_sigil] = STATE(4266),
    [sym_list] = STATE(4266),
    [sym_tuple] = STATE(4266),
    [sym_bitstring] = STATE(4266),
    [sym_map] = STATE(4266),
    [sym__nullary_operator] = STATE(4266),
    [sym_unary_operator] = STATE(4266),
    [sym_binary_operator] = STATE(4266),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4266),
    [sym_call] = STATE(4266),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4266),
    [sym_anonymous_function] = STATE(4266),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1921),
    [sym_integer] = ACTIONS(1921),
    [sym_float] = ACTIONS(1921),
    [sym_char] = ACTIONS(1921),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1921),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(498)] = {
    [sym__expression] = STATE(4267),
    [sym_block] = STATE(4267),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4267),
    [sym_nil] = STATE(4267),
    [sym__atom] = STATE(4267),
    [sym_quoted_atom] = STATE(4267),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4267),
    [sym_charlist] = STATE(4267),
    [sym_sigil] = STATE(4267),
    [sym_list] = STATE(4267),
    [sym_tuple] = STATE(4267),
    [sym_bitstring] = STATE(4267),
    [sym_map] = STATE(4267),
    [sym__nullary_operator] = STATE(4267),
    [sym_unary_operator] = STATE(4267),
    [sym_binary_operator] = STATE(4267),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4267),
    [sym_call] = STATE(4267),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4267),
    [sym_anonymous_function] = STATE(4267),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1923),
    [sym_integer] = ACTIONS(1923),
    [sym_float] = ACTIONS(1923),
    [sym_char] = ACTIONS(1923),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1923),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(499)] = {
    [sym__expression] = STATE(4270),
    [sym_block] = STATE(4270),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4270),
    [sym_nil] = STATE(4270),
    [sym__atom] = STATE(4270),
    [sym_quoted_atom] = STATE(4270),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4270),
    [sym_charlist] = STATE(4270),
    [sym_sigil] = STATE(4270),
    [sym_list] = STATE(4270),
    [sym_tuple] = STATE(4270),
    [sym_bitstring] = STATE(4270),
    [sym_map] = STATE(4270),
    [sym__nullary_operator] = STATE(4270),
    [sym_unary_operator] = STATE(4270),
    [sym_binary_operator] = STATE(4270),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4270),
    [sym_call] = STATE(4270),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4270),
    [sym_anonymous_function] = STATE(4270),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1925),
    [sym_integer] = ACTIONS(1925),
    [sym_float] = ACTIONS(1925),
    [sym_char] = ACTIONS(1925),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1925),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(500)] = {
    [sym__expression] = STATE(4271),
    [sym_block] = STATE(4271),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4271),
    [sym_nil] = STATE(4271),
    [sym__atom] = STATE(4271),
    [sym_quoted_atom] = STATE(4271),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4271),
    [sym_charlist] = STATE(4271),
    [sym_sigil] = STATE(4271),
    [sym_list] = STATE(4271),
    [sym_tuple] = STATE(4271),
    [sym_bitstring] = STATE(4271),
    [sym_map] = STATE(4271),
    [sym__nullary_operator] = STATE(4271),
    [sym_unary_operator] = STATE(4271),
    [sym_binary_operator] = STATE(4271),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4271),
    [sym_call] = STATE(4271),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4271),
    [sym_anonymous_function] = STATE(4271),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1927),
    [sym_integer] = ACTIONS(1927),
    [sym_float] = ACTIONS(1927),
    [sym_char] = ACTIONS(1927),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1927),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(501)] = {
    [sym__expression] = STATE(4272),
    [sym_block] = STATE(4272),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4272),
    [sym_nil] = STATE(4272),
    [sym__atom] = STATE(4272),
    [sym_quoted_atom] = STATE(4272),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4272),
    [sym_charlist] = STATE(4272),
    [sym_sigil] = STATE(4272),
    [sym_list] = STATE(4272),
    [sym_tuple] = STATE(4272),
    [sym_bitstring] = STATE(4272),
    [sym_map] = STATE(4272),
    [sym__nullary_operator] = STATE(4272),
    [sym_unary_operator] = STATE(4272),
    [sym_binary_operator] = STATE(4272),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4272),
    [sym_call] = STATE(4272),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4272),
    [sym_anonymous_function] = STATE(4272),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1929),
    [sym_integer] = ACTIONS(1929),
    [sym_float] = ACTIONS(1929),
    [sym_char] = ACTIONS(1929),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1929),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(502)] = {
    [sym__expression] = STATE(4273),
    [sym_block] = STATE(4273),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4273),
    [sym_nil] = STATE(4273),
    [sym__atom] = STATE(4273),
    [sym_quoted_atom] = STATE(4273),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4273),
    [sym_charlist] = STATE(4273),
    [sym_sigil] = STATE(4273),
    [sym_list] = STATE(4273),
    [sym_tuple] = STATE(4273),
    [sym_bitstring] = STATE(4273),
    [sym_map] = STATE(4273),
    [sym__nullary_operator] = STATE(4273),
    [sym_unary_operator] = STATE(4273),
    [sym_binary_operator] = STATE(4273),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4273),
    [sym_call] = STATE(4273),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4273),
    [sym_anonymous_function] = STATE(4273),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1931),
    [sym_integer] = ACTIONS(1931),
    [sym_float] = ACTIONS(1931),
    [sym_char] = ACTIONS(1931),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1931),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(503)] = {
    [sym__expression] = STATE(4274),
    [sym_block] = STATE(4274),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4274),
    [sym_nil] = STATE(4274),
    [sym__atom] = STATE(4274),
    [sym_quoted_atom] = STATE(4274),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4274),
    [sym_charlist] = STATE(4274),
    [sym_sigil] = STATE(4274),
    [sym_list] = STATE(4274),
    [sym_tuple] = STATE(4274),
    [sym_bitstring] = STATE(4274),
    [sym_map] = STATE(4274),
    [sym__nullary_operator] = STATE(4274),
    [sym_unary_operator] = STATE(4274),
    [sym_binary_operator] = STATE(4274),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4274),
    [sym_call] = STATE(4274),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4274),
    [sym_anonymous_function] = STATE(4274),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1933),
    [sym_integer] = ACTIONS(1933),
    [sym_float] = ACTIONS(1933),
    [sym_char] = ACTIONS(1933),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(504)] = {
    [sym__expression] = STATE(3483),
    [sym_block] = STATE(3483),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(3483),
    [sym_nil] = STATE(3483),
    [sym__atom] = STATE(3483),
    [sym_quoted_atom] = STATE(3483),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(3483),
    [sym_charlist] = STATE(3483),
    [sym_sigil] = STATE(3483),
    [sym_list] = STATE(3483),
    [sym_tuple] = STATE(3483),
    [sym_bitstring] = STATE(3483),
    [sym_map] = STATE(3483),
    [sym__nullary_operator] = STATE(3483),
    [sym_unary_operator] = STATE(3483),
    [sym_binary_operator] = STATE(3483),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(3483),
    [sym_call] = STATE(3483),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(3483),
    [sym_anonymous_function] = STATE(3483),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1935),
    [sym_integer] = ACTIONS(1935),
    [sym_float] = ACTIONS(1935),
    [sym_char] = ACTIONS(1935),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1935),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(505)] = {
    [sym__expression] = STATE(3484),
    [sym_block] = STATE(3484),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(3484),
    [sym_nil] = STATE(3484),
    [sym__atom] = STATE(3484),
    [sym_quoted_atom] = STATE(3484),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(3484),
    [sym_charlist] = STATE(3484),
    [sym_sigil] = STATE(3484),
    [sym_list] = STATE(3484),
    [sym_tuple] = STATE(3484),
    [sym_bitstring] = STATE(3484),
    [sym_map] = STATE(3484),
    [sym__nullary_operator] = STATE(3484),
    [sym_unary_operator] = STATE(3484),
    [sym_binary_operator] = STATE(3484),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(3484),
    [sym_call] = STATE(3484),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(3484),
    [sym_anonymous_function] = STATE(3484),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1937),
    [sym_integer] = ACTIONS(1937),
    [sym_float] = ACTIONS(1937),
    [sym_char] = ACTIONS(1937),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1937),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(506)] = {
    [sym__expression] = STATE(2934),
    [sym_block] = STATE(2934),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2934),
    [sym_nil] = STATE(2934),
    [sym__atom] = STATE(2934),
    [sym_quoted_atom] = STATE(2934),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2934),
    [sym_charlist] = STATE(2934),
    [sym_sigil] = STATE(2934),
    [sym_list] = STATE(2934),
    [sym_tuple] = STATE(2934),
    [sym_bitstring] = STATE(2934),
    [sym_map] = STATE(2934),
    [sym__nullary_operator] = STATE(2934),
    [sym_unary_operator] = STATE(2934),
    [sym_binary_operator] = STATE(2934),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2934),
    [sym_call] = STATE(2934),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2934),
    [sym_anonymous_function] = STATE(2934),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1939),
    [sym_integer] = ACTIONS(1939),
    [sym_float] = ACTIONS(1939),
    [sym_char] = ACTIONS(1939),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1939),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(507)] = {
    [sym__expression] = STATE(2935),
    [sym_block] = STATE(2935),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2935),
    [sym_nil] = STATE(2935),
    [sym__atom] = STATE(2935),
    [sym_quoted_atom] = STATE(2935),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2935),
    [sym_charlist] = STATE(2935),
    [sym_sigil] = STATE(2935),
    [sym_list] = STATE(2935),
    [sym_tuple] = STATE(2935),
    [sym_bitstring] = STATE(2935),
    [sym_map] = STATE(2935),
    [sym__nullary_operator] = STATE(2935),
    [sym_unary_operator] = STATE(2935),
    [sym_binary_operator] = STATE(2935),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2935),
    [sym_call] = STATE(2935),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2935),
    [sym_anonymous_function] = STATE(2935),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1941),
    [sym_integer] = ACTIONS(1941),
    [sym_float] = ACTIONS(1941),
    [sym_char] = ACTIONS(1941),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1941),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(508)] = {
    [sym__expression] = STATE(2958),
    [sym_block] = STATE(2958),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2958),
    [sym_nil] = STATE(2958),
    [sym__atom] = STATE(2958),
    [sym_quoted_atom] = STATE(2958),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2958),
    [sym_charlist] = STATE(2958),
    [sym_sigil] = STATE(2958),
    [sym_list] = STATE(2958),
    [sym_tuple] = STATE(2958),
    [sym_bitstring] = STATE(2958),
    [sym_map] = STATE(2958),
    [sym__nullary_operator] = STATE(2958),
    [sym_unary_operator] = STATE(2958),
    [sym_binary_operator] = STATE(2958),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2958),
    [sym_call] = STATE(2958),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2958),
    [sym_anonymous_function] = STATE(2958),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1943),
    [sym_integer] = ACTIONS(1943),
    [sym_float] = ACTIONS(1943),
    [sym_char] = ACTIONS(1943),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1943),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(509)] = {
    [sym__expression] = STATE(2959),
    [sym_block] = STATE(2959),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2959),
    [sym_nil] = STATE(2959),
    [sym__atom] = STATE(2959),
    [sym_quoted_atom] = STATE(2959),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2959),
    [sym_charlist] = STATE(2959),
    [sym_sigil] = STATE(2959),
    [sym_list] = STATE(2959),
    [sym_tuple] = STATE(2959),
    [sym_bitstring] = STATE(2959),
    [sym_map] = STATE(2959),
    [sym__nullary_operator] = STATE(2959),
    [sym_unary_operator] = STATE(2959),
    [sym_binary_operator] = STATE(2959),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2959),
    [sym_call] = STATE(2959),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2959),
    [sym_anonymous_function] = STATE(2959),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1945),
    [sym_integer] = ACTIONS(1945),
    [sym_float] = ACTIONS(1945),
    [sym_char] = ACTIONS(1945),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1945),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(510)] = {
    [sym__expression] = STATE(2970),
    [sym_block] = STATE(2970),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2970),
    [sym_nil] = STATE(2970),
    [sym__atom] = STATE(2970),
    [sym_quoted_atom] = STATE(2970),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2970),
    [sym_charlist] = STATE(2970),
    [sym_sigil] = STATE(2970),
    [sym_list] = STATE(2970),
    [sym_tuple] = STATE(2970),
    [sym_bitstring] = STATE(2970),
    [sym_map] = STATE(2970),
    [sym__nullary_operator] = STATE(2970),
    [sym_unary_operator] = STATE(2970),
    [sym_binary_operator] = STATE(2970),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2970),
    [sym_call] = STATE(2970),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2970),
    [sym_anonymous_function] = STATE(2970),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1947),
    [sym_integer] = ACTIONS(1947),
    [sym_float] = ACTIONS(1947),
    [sym_char] = ACTIONS(1947),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1947),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(511)] = {
    [sym__expression] = STATE(2981),
    [sym_block] = STATE(2981),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2981),
    [sym_nil] = STATE(2981),
    [sym__atom] = STATE(2981),
    [sym_quoted_atom] = STATE(2981),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2981),
    [sym_charlist] = STATE(2981),
    [sym_sigil] = STATE(2981),
    [sym_list] = STATE(2981),
    [sym_tuple] = STATE(2981),
    [sym_bitstring] = STATE(2981),
    [sym_map] = STATE(2981),
    [sym__nullary_operator] = STATE(2981),
    [sym_unary_operator] = STATE(2981),
    [sym_binary_operator] = STATE(2981),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2981),
    [sym_call] = STATE(2981),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2981),
    [sym_anonymous_function] = STATE(2981),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1949),
    [sym_integer] = ACTIONS(1949),
    [sym_float] = ACTIONS(1949),
    [sym_char] = ACTIONS(1949),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(512)] = {
    [sym__expression] = STATE(2982),
    [sym_block] = STATE(2982),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2982),
    [sym_nil] = STATE(2982),
    [sym__atom] = STATE(2982),
    [sym_quoted_atom] = STATE(2982),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2982),
    [sym_charlist] = STATE(2982),
    [sym_sigil] = STATE(2982),
    [sym_list] = STATE(2982),
    [sym_tuple] = STATE(2982),
    [sym_bitstring] = STATE(2982),
    [sym_map] = STATE(2982),
    [sym__nullary_operator] = STATE(2982),
    [sym_unary_operator] = STATE(2982),
    [sym_binary_operator] = STATE(2982),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2982),
    [sym_call] = STATE(2982),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2982),
    [sym_anonymous_function] = STATE(2982),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1951),
    [sym_integer] = ACTIONS(1951),
    [sym_float] = ACTIONS(1951),
    [sym_char] = ACTIONS(1951),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1951),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(513)] = {
    [sym__expression] = STATE(2983),
    [sym_block] = STATE(2983),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2983),
    [sym_nil] = STATE(2983),
    [sym__atom] = STATE(2983),
    [sym_quoted_atom] = STATE(2983),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2983),
    [sym_charlist] = STATE(2983),
    [sym_sigil] = STATE(2983),
    [sym_list] = STATE(2983),
    [sym_tuple] = STATE(2983),
    [sym_bitstring] = STATE(2983),
    [sym_map] = STATE(2983),
    [sym__nullary_operator] = STATE(2983),
    [sym_unary_operator] = STATE(2983),
    [sym_binary_operator] = STATE(2983),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2983),
    [sym_call] = STATE(2983),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2983),
    [sym_anonymous_function] = STATE(2983),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1953),
    [sym_integer] = ACTIONS(1953),
    [sym_float] = ACTIONS(1953),
    [sym_char] = ACTIONS(1953),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1953),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(514)] = {
    [sym__expression] = STATE(2986),
    [sym_block] = STATE(2986),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2986),
    [sym_nil] = STATE(2986),
    [sym__atom] = STATE(2986),
    [sym_quoted_atom] = STATE(2986),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2986),
    [sym_charlist] = STATE(2986),
    [sym_sigil] = STATE(2986),
    [sym_list] = STATE(2986),
    [sym_tuple] = STATE(2986),
    [sym_bitstring] = STATE(2986),
    [sym_map] = STATE(2986),
    [sym__nullary_operator] = STATE(2986),
    [sym_unary_operator] = STATE(2986),
    [sym_binary_operator] = STATE(2986),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2986),
    [sym_call] = STATE(2986),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2986),
    [sym_anonymous_function] = STATE(2986),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1955),
    [sym_integer] = ACTIONS(1955),
    [sym_float] = ACTIONS(1955),
    [sym_char] = ACTIONS(1955),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1955),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(515)] = {
    [sym__expression] = STATE(2987),
    [sym_block] = STATE(2987),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2987),
    [sym_nil] = STATE(2987),
    [sym__atom] = STATE(2987),
    [sym_quoted_atom] = STATE(2987),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2987),
    [sym_charlist] = STATE(2987),
    [sym_sigil] = STATE(2987),
    [sym_list] = STATE(2987),
    [sym_tuple] = STATE(2987),
    [sym_bitstring] = STATE(2987),
    [sym_map] = STATE(2987),
    [sym__nullary_operator] = STATE(2987),
    [sym_unary_operator] = STATE(2987),
    [sym_binary_operator] = STATE(2987),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2987),
    [sym_call] = STATE(2987),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2987),
    [sym_anonymous_function] = STATE(2987),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1957),
    [sym_integer] = ACTIONS(1957),
    [sym_float] = ACTIONS(1957),
    [sym_char] = ACTIONS(1957),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1957),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(516)] = {
    [sym__expression] = STATE(2988),
    [sym_block] = STATE(2988),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2988),
    [sym_nil] = STATE(2988),
    [sym__atom] = STATE(2988),
    [sym_quoted_atom] = STATE(2988),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2988),
    [sym_charlist] = STATE(2988),
    [sym_sigil] = STATE(2988),
    [sym_list] = STATE(2988),
    [sym_tuple] = STATE(2988),
    [sym_bitstring] = STATE(2988),
    [sym_map] = STATE(2988),
    [sym__nullary_operator] = STATE(2988),
    [sym_unary_operator] = STATE(2988),
    [sym_binary_operator] = STATE(2988),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2988),
    [sym_call] = STATE(2988),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2988),
    [sym_anonymous_function] = STATE(2988),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1959),
    [sym_integer] = ACTIONS(1959),
    [sym_float] = ACTIONS(1959),
    [sym_char] = ACTIONS(1959),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1959),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(517)] = {
    [sym__expression] = STATE(2989),
    [sym_block] = STATE(2989),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2989),
    [sym_nil] = STATE(2989),
    [sym__atom] = STATE(2989),
    [sym_quoted_atom] = STATE(2989),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2989),
    [sym_charlist] = STATE(2989),
    [sym_sigil] = STATE(2989),
    [sym_list] = STATE(2989),
    [sym_tuple] = STATE(2989),
    [sym_bitstring] = STATE(2989),
    [sym_map] = STATE(2989),
    [sym__nullary_operator] = STATE(2989),
    [sym_unary_operator] = STATE(2989),
    [sym_binary_operator] = STATE(2989),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2989),
    [sym_call] = STATE(2989),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2989),
    [sym_anonymous_function] = STATE(2989),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1961),
    [sym_integer] = ACTIONS(1961),
    [sym_float] = ACTIONS(1961),
    [sym_char] = ACTIONS(1961),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1961),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(518)] = {
    [sym__expression] = STATE(2990),
    [sym_block] = STATE(2990),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2990),
    [sym_nil] = STATE(2990),
    [sym__atom] = STATE(2990),
    [sym_quoted_atom] = STATE(2990),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2990),
    [sym_charlist] = STATE(2990),
    [sym_sigil] = STATE(2990),
    [sym_list] = STATE(2990),
    [sym_tuple] = STATE(2990),
    [sym_bitstring] = STATE(2990),
    [sym_map] = STATE(2990),
    [sym__nullary_operator] = STATE(2990),
    [sym_unary_operator] = STATE(2990),
    [sym_binary_operator] = STATE(2990),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2990),
    [sym_call] = STATE(2990),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2990),
    [sym_anonymous_function] = STATE(2990),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1963),
    [sym_integer] = ACTIONS(1963),
    [sym_float] = ACTIONS(1963),
    [sym_char] = ACTIONS(1963),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1963),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(519)] = {
    [sym__expression] = STATE(2991),
    [sym_block] = STATE(2991),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2991),
    [sym_nil] = STATE(2991),
    [sym__atom] = STATE(2991),
    [sym_quoted_atom] = STATE(2991),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2991),
    [sym_charlist] = STATE(2991),
    [sym_sigil] = STATE(2991),
    [sym_list] = STATE(2991),
    [sym_tuple] = STATE(2991),
    [sym_bitstring] = STATE(2991),
    [sym_map] = STATE(2991),
    [sym__nullary_operator] = STATE(2991),
    [sym_unary_operator] = STATE(2991),
    [sym_binary_operator] = STATE(2991),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2991),
    [sym_call] = STATE(2991),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2991),
    [sym_anonymous_function] = STATE(2991),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1965),
    [sym_integer] = ACTIONS(1965),
    [sym_float] = ACTIONS(1965),
    [sym_char] = ACTIONS(1965),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1965),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(520)] = {
    [sym__expression] = STATE(2992),
    [sym_block] = STATE(2992),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2992),
    [sym_nil] = STATE(2992),
    [sym__atom] = STATE(2992),
    [sym_quoted_atom] = STATE(2992),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2992),
    [sym_charlist] = STATE(2992),
    [sym_sigil] = STATE(2992),
    [sym_list] = STATE(2992),
    [sym_tuple] = STATE(2992),
    [sym_bitstring] = STATE(2992),
    [sym_map] = STATE(2992),
    [sym__nullary_operator] = STATE(2992),
    [sym_unary_operator] = STATE(2992),
    [sym_binary_operator] = STATE(2992),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2992),
    [sym_call] = STATE(2992),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2992),
    [sym_anonymous_function] = STATE(2992),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1967),
    [sym_integer] = ACTIONS(1967),
    [sym_float] = ACTIONS(1967),
    [sym_char] = ACTIONS(1967),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1967),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(521)] = {
    [sym__expression] = STATE(2993),
    [sym_block] = STATE(2993),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2993),
    [sym_nil] = STATE(2993),
    [sym__atom] = STATE(2993),
    [sym_quoted_atom] = STATE(2993),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2993),
    [sym_charlist] = STATE(2993),
    [sym_sigil] = STATE(2993),
    [sym_list] = STATE(2993),
    [sym_tuple] = STATE(2993),
    [sym_bitstring] = STATE(2993),
    [sym_map] = STATE(2993),
    [sym__nullary_operator] = STATE(2993),
    [sym_unary_operator] = STATE(2993),
    [sym_binary_operator] = STATE(2993),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2993),
    [sym_call] = STATE(2993),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2993),
    [sym_anonymous_function] = STATE(2993),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1969),
    [sym_integer] = ACTIONS(1969),
    [sym_float] = ACTIONS(1969),
    [sym_char] = ACTIONS(1969),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1969),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(522)] = {
    [sym__expression] = STATE(2994),
    [sym_block] = STATE(2994),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2994),
    [sym_nil] = STATE(2994),
    [sym__atom] = STATE(2994),
    [sym_quoted_atom] = STATE(2994),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2994),
    [sym_charlist] = STATE(2994),
    [sym_sigil] = STATE(2994),
    [sym_list] = STATE(2994),
    [sym_tuple] = STATE(2994),
    [sym_bitstring] = STATE(2994),
    [sym_map] = STATE(2994),
    [sym__nullary_operator] = STATE(2994),
    [sym_unary_operator] = STATE(2994),
    [sym_binary_operator] = STATE(2994),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2994),
    [sym_call] = STATE(2994),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2994),
    [sym_anonymous_function] = STATE(2994),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1971),
    [sym_integer] = ACTIONS(1971),
    [sym_float] = ACTIONS(1971),
    [sym_char] = ACTIONS(1971),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1971),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(523)] = {
    [sym__expression] = STATE(2995),
    [sym_block] = STATE(2995),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(2995),
    [sym_nil] = STATE(2995),
    [sym__atom] = STATE(2995),
    [sym_quoted_atom] = STATE(2995),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(2995),
    [sym_charlist] = STATE(2995),
    [sym_sigil] = STATE(2995),
    [sym_list] = STATE(2995),
    [sym_tuple] = STATE(2995),
    [sym_bitstring] = STATE(2995),
    [sym_map] = STATE(2995),
    [sym__nullary_operator] = STATE(2995),
    [sym_unary_operator] = STATE(2995),
    [sym_binary_operator] = STATE(2995),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(2995),
    [sym_call] = STATE(2995),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(2995),
    [sym_anonymous_function] = STATE(2995),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1973),
    [sym_integer] = ACTIONS(1973),
    [sym_float] = ACTIONS(1973),
    [sym_char] = ACTIONS(1973),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1973),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(524)] = {
    [sym__expression] = STATE(4396),
    [sym_block] = STATE(4396),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4396),
    [sym_nil] = STATE(4396),
    [sym__atom] = STATE(4396),
    [sym_quoted_atom] = STATE(4396),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4396),
    [sym_charlist] = STATE(4396),
    [sym_sigil] = STATE(4396),
    [sym_list] = STATE(4396),
    [sym_tuple] = STATE(4396),
    [sym_bitstring] = STATE(4396),
    [sym_map] = STATE(4396),
    [sym__nullary_operator] = STATE(4396),
    [sym_unary_operator] = STATE(4396),
    [sym_binary_operator] = STATE(4396),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4396),
    [sym_call] = STATE(4396),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4396),
    [sym_anonymous_function] = STATE(4396),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1975),
    [sym_integer] = ACTIONS(1975),
    [sym_float] = ACTIONS(1975),
    [sym_char] = ACTIONS(1975),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1975),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(525)] = {
    [sym__expression] = STATE(1908),
    [sym_block] = STATE(1908),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(1908),
    [sym_nil] = STATE(1908),
    [sym__atom] = STATE(1908),
    [sym_quoted_atom] = STATE(1908),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(1908),
    [sym_charlist] = STATE(1908),
    [sym_sigil] = STATE(1908),
    [sym_list] = STATE(1908),
    [sym_tuple] = STATE(1908),
    [sym_bitstring] = STATE(1908),
    [sym_map] = STATE(1908),
    [sym__nullary_operator] = STATE(1908),
    [sym_unary_operator] = STATE(1908),
    [sym_binary_operator] = STATE(1908),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(1908),
    [sym_call] = STATE(1908),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(1908),
    [sym_anonymous_function] = STATE(1908),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1977),
    [sym_integer] = ACTIONS(1977),
    [sym_float] = ACTIONS(1977),
    [sym_char] = ACTIONS(1977),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1977),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(526)] = {
    [sym__expression] = STATE(4444),
    [sym_block] = STATE(4444),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4444),
    [sym_nil] = STATE(4444),
    [sym__atom] = STATE(4444),
    [sym_quoted_atom] = STATE(4444),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4444),
    [sym_charlist] = STATE(4444),
    [sym_sigil] = STATE(4444),
    [sym_list] = STATE(4444),
    [sym_tuple] = STATE(4444),
    [sym_bitstring] = STATE(4444),
    [sym_map] = STATE(4444),
    [sym__nullary_operator] = STATE(4444),
    [sym_unary_operator] = STATE(4444),
    [sym_binary_operator] = STATE(4444),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4444),
    [sym_call] = STATE(4444),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4444),
    [sym_anonymous_function] = STATE(4444),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1577),
    [sym_integer] = ACTIONS(1577),
    [sym_float] = ACTIONS(1577),
    [sym_char] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(527)] = {
    [sym__expression] = STATE(4161),
    [sym_block] = STATE(4161),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4161),
    [sym_nil] = STATE(4161),
    [sym__atom] = STATE(4161),
    [sym_quoted_atom] = STATE(4161),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4161),
    [sym_charlist] = STATE(4161),
    [sym_sigil] = STATE(4161),
    [sym_list] = STATE(4161),
    [sym_tuple] = STATE(4161),
    [sym_bitstring] = STATE(4161),
    [sym_map] = STATE(4161),
    [sym__nullary_operator] = STATE(4161),
    [sym_unary_operator] = STATE(4161),
    [sym_binary_operator] = STATE(4161),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4161),
    [sym_call] = STATE(4161),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4161),
    [sym_anonymous_function] = STATE(4161),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1979),
    [sym_integer] = ACTIONS(1979),
    [sym_float] = ACTIONS(1979),
    [sym_char] = ACTIONS(1979),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1979),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(528)] = {
    [sym__expression] = STATE(2022),
    [sym_block] = STATE(2022),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(2022),
    [sym_nil] = STATE(2022),
    [sym__atom] = STATE(2022),
    [sym_quoted_atom] = STATE(2022),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2022),
    [sym_charlist] = STATE(2022),
    [sym_sigil] = STATE(2022),
    [sym_list] = STATE(2022),
    [sym_tuple] = STATE(2022),
    [sym_bitstring] = STATE(2022),
    [sym_map] = STATE(2022),
    [sym__nullary_operator] = STATE(2022),
    [sym_unary_operator] = STATE(2022),
    [sym_binary_operator] = STATE(2022),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2022),
    [sym_call] = STATE(2022),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2022),
    [sym_anonymous_function] = STATE(2022),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1981),
    [sym_integer] = ACTIONS(1981),
    [sym_float] = ACTIONS(1981),
    [sym_char] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1981),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(529)] = {
    [sym__expression] = STATE(4065),
    [sym_block] = STATE(4065),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4065),
    [sym_nil] = STATE(4065),
    [sym__atom] = STATE(4065),
    [sym_quoted_atom] = STATE(4065),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4065),
    [sym_charlist] = STATE(4065),
    [sym_sigil] = STATE(4065),
    [sym_list] = STATE(4065),
    [sym_tuple] = STATE(4065),
    [sym_bitstring] = STATE(4065),
    [sym_map] = STATE(4065),
    [sym__nullary_operator] = STATE(4065),
    [sym_unary_operator] = STATE(4065),
    [sym_binary_operator] = STATE(4065),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4065),
    [sym_call] = STATE(4065),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4065),
    [sym_anonymous_function] = STATE(4065),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1983),
    [sym_integer] = ACTIONS(1983),
    [sym_float] = ACTIONS(1983),
    [sym_char] = ACTIONS(1983),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1983),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(530)] = {
    [sym__expression] = STATE(4066),
    [sym_block] = STATE(4066),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4066),
    [sym_nil] = STATE(4066),
    [sym__atom] = STATE(4066),
    [sym_quoted_atom] = STATE(4066),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4066),
    [sym_charlist] = STATE(4066),
    [sym_sigil] = STATE(4066),
    [sym_list] = STATE(4066),
    [sym_tuple] = STATE(4066),
    [sym_bitstring] = STATE(4066),
    [sym_map] = STATE(4066),
    [sym__nullary_operator] = STATE(4066),
    [sym_unary_operator] = STATE(4066),
    [sym_binary_operator] = STATE(4066),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4066),
    [sym_call] = STATE(4066),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4066),
    [sym_anonymous_function] = STATE(4066),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1985),
    [sym_integer] = ACTIONS(1985),
    [sym_float] = ACTIONS(1985),
    [sym_char] = ACTIONS(1985),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(531)] = {
    [sym__expression] = STATE(4445),
    [sym_block] = STATE(4445),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4445),
    [sym_nil] = STATE(4445),
    [sym__atom] = STATE(4445),
    [sym_quoted_atom] = STATE(4445),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4445),
    [sym_charlist] = STATE(4445),
    [sym_sigil] = STATE(4445),
    [sym_list] = STATE(4445),
    [sym_tuple] = STATE(4445),
    [sym_bitstring] = STATE(4445),
    [sym_map] = STATE(4445),
    [sym__nullary_operator] = STATE(4445),
    [sym_unary_operator] = STATE(4445),
    [sym_binary_operator] = STATE(4445),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4445),
    [sym_call] = STATE(4445),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4445),
    [sym_anonymous_function] = STATE(4445),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1367),
    [sym_integer] = ACTIONS(1367),
    [sym_float] = ACTIONS(1367),
    [sym_char] = ACTIONS(1367),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(1367),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(532)] = {
    [sym__expression] = STATE(4068),
    [sym_block] = STATE(4068),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4068),
    [sym_nil] = STATE(4068),
    [sym__atom] = STATE(4068),
    [sym_quoted_atom] = STATE(4068),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4068),
    [sym_charlist] = STATE(4068),
    [sym_sigil] = STATE(4068),
    [sym_list] = STATE(4068),
    [sym_tuple] = STATE(4068),
    [sym_bitstring] = STATE(4068),
    [sym_map] = STATE(4068),
    [sym__nullary_operator] = STATE(4068),
    [sym_unary_operator] = STATE(4068),
    [sym_binary_operator] = STATE(4068),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4068),
    [sym_call] = STATE(4068),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4068),
    [sym_anonymous_function] = STATE(4068),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1987),
    [sym_integer] = ACTIONS(1987),
    [sym_float] = ACTIONS(1987),
    [sym_char] = ACTIONS(1987),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1987),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(533)] = {
    [sym__expression] = STATE(4069),
    [sym_block] = STATE(4069),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4069),
    [sym_nil] = STATE(4069),
    [sym__atom] = STATE(4069),
    [sym_quoted_atom] = STATE(4069),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4069),
    [sym_charlist] = STATE(4069),
    [sym_sigil] = STATE(4069),
    [sym_list] = STATE(4069),
    [sym_tuple] = STATE(4069),
    [sym_bitstring] = STATE(4069),
    [sym_map] = STATE(4069),
    [sym__nullary_operator] = STATE(4069),
    [sym_unary_operator] = STATE(4069),
    [sym_binary_operator] = STATE(4069),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4069),
    [sym_call] = STATE(4069),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4069),
    [sym_anonymous_function] = STATE(4069),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1989),
    [sym_integer] = ACTIONS(1989),
    [sym_float] = ACTIONS(1989),
    [sym_char] = ACTIONS(1989),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(534)] = {
    [sym__expression] = STATE(4070),
    [sym_block] = STATE(4070),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4070),
    [sym_nil] = STATE(4070),
    [sym__atom] = STATE(4070),
    [sym_quoted_atom] = STATE(4070),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4070),
    [sym_charlist] = STATE(4070),
    [sym_sigil] = STATE(4070),
    [sym_list] = STATE(4070),
    [sym_tuple] = STATE(4070),
    [sym_bitstring] = STATE(4070),
    [sym_map] = STATE(4070),
    [sym__nullary_operator] = STATE(4070),
    [sym_unary_operator] = STATE(4070),
    [sym_binary_operator] = STATE(4070),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4070),
    [sym_call] = STATE(4070),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4070),
    [sym_anonymous_function] = STATE(4070),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1991),
    [sym_integer] = ACTIONS(1991),
    [sym_float] = ACTIONS(1991),
    [sym_char] = ACTIONS(1991),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1991),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(535)] = {
    [sym__expression] = STATE(4071),
    [sym_block] = STATE(4071),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4071),
    [sym_nil] = STATE(4071),
    [sym__atom] = STATE(4071),
    [sym_quoted_atom] = STATE(4071),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4071),
    [sym_charlist] = STATE(4071),
    [sym_sigil] = STATE(4071),
    [sym_list] = STATE(4071),
    [sym_tuple] = STATE(4071),
    [sym_bitstring] = STATE(4071),
    [sym_map] = STATE(4071),
    [sym__nullary_operator] = STATE(4071),
    [sym_unary_operator] = STATE(4071),
    [sym_binary_operator] = STATE(4071),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4071),
    [sym_call] = STATE(4071),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4071),
    [sym_anonymous_function] = STATE(4071),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1993),
    [sym_integer] = ACTIONS(1993),
    [sym_float] = ACTIONS(1993),
    [sym_char] = ACTIONS(1993),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(536)] = {
    [sym__expression] = STATE(3830),
    [sym_block] = STATE(3830),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3830),
    [sym_nil] = STATE(3830),
    [sym__atom] = STATE(3830),
    [sym_quoted_atom] = STATE(3830),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3830),
    [sym_charlist] = STATE(3830),
    [sym_sigil] = STATE(3830),
    [sym_list] = STATE(3830),
    [sym_tuple] = STATE(3830),
    [sym_bitstring] = STATE(3830),
    [sym_map] = STATE(3830),
    [sym__nullary_operator] = STATE(3830),
    [sym_unary_operator] = STATE(3830),
    [sym_binary_operator] = STATE(3830),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3830),
    [sym_call] = STATE(3830),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3830),
    [sym_anonymous_function] = STATE(3830),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1223),
    [sym_integer] = ACTIONS(1223),
    [sym_float] = ACTIONS(1223),
    [sym_char] = ACTIONS(1223),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1223),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(537)] = {
    [sym__expression] = STATE(4073),
    [sym_block] = STATE(4073),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4073),
    [sym_nil] = STATE(4073),
    [sym__atom] = STATE(4073),
    [sym_quoted_atom] = STATE(4073),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4073),
    [sym_charlist] = STATE(4073),
    [sym_sigil] = STATE(4073),
    [sym_list] = STATE(4073),
    [sym_tuple] = STATE(4073),
    [sym_bitstring] = STATE(4073),
    [sym_map] = STATE(4073),
    [sym__nullary_operator] = STATE(4073),
    [sym_unary_operator] = STATE(4073),
    [sym_binary_operator] = STATE(4073),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4073),
    [sym_call] = STATE(4073),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4073),
    [sym_anonymous_function] = STATE(4073),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1995),
    [sym_integer] = ACTIONS(1995),
    [sym_float] = ACTIONS(1995),
    [sym_char] = ACTIONS(1995),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1995),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(538)] = {
    [sym__expression] = STATE(4074),
    [sym_block] = STATE(4074),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4074),
    [sym_nil] = STATE(4074),
    [sym__atom] = STATE(4074),
    [sym_quoted_atom] = STATE(4074),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4074),
    [sym_charlist] = STATE(4074),
    [sym_sigil] = STATE(4074),
    [sym_list] = STATE(4074),
    [sym_tuple] = STATE(4074),
    [sym_bitstring] = STATE(4074),
    [sym_map] = STATE(4074),
    [sym__nullary_operator] = STATE(4074),
    [sym_unary_operator] = STATE(4074),
    [sym_binary_operator] = STATE(4074),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4074),
    [sym_call] = STATE(4074),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4074),
    [sym_anonymous_function] = STATE(4074),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1997),
    [sym_integer] = ACTIONS(1997),
    [sym_float] = ACTIONS(1997),
    [sym_char] = ACTIONS(1997),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1997),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(539)] = {
    [sym__expression] = STATE(4075),
    [sym_block] = STATE(4075),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4075),
    [sym_nil] = STATE(4075),
    [sym__atom] = STATE(4075),
    [sym_quoted_atom] = STATE(4075),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4075),
    [sym_charlist] = STATE(4075),
    [sym_sigil] = STATE(4075),
    [sym_list] = STATE(4075),
    [sym_tuple] = STATE(4075),
    [sym_bitstring] = STATE(4075),
    [sym_map] = STATE(4075),
    [sym__nullary_operator] = STATE(4075),
    [sym_unary_operator] = STATE(4075),
    [sym_binary_operator] = STATE(4075),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4075),
    [sym_call] = STATE(4075),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4075),
    [sym_anonymous_function] = STATE(4075),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1999),
    [sym_integer] = ACTIONS(1999),
    [sym_float] = ACTIONS(1999),
    [sym_char] = ACTIONS(1999),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1999),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(540)] = {
    [sym__expression] = STATE(4076),
    [sym_block] = STATE(4076),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4076),
    [sym_nil] = STATE(4076),
    [sym__atom] = STATE(4076),
    [sym_quoted_atom] = STATE(4076),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4076),
    [sym_charlist] = STATE(4076),
    [sym_sigil] = STATE(4076),
    [sym_list] = STATE(4076),
    [sym_tuple] = STATE(4076),
    [sym_bitstring] = STATE(4076),
    [sym_map] = STATE(4076),
    [sym__nullary_operator] = STATE(4076),
    [sym_unary_operator] = STATE(4076),
    [sym_binary_operator] = STATE(4076),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4076),
    [sym_call] = STATE(4076),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4076),
    [sym_anonymous_function] = STATE(4076),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2001),
    [sym_integer] = ACTIONS(2001),
    [sym_float] = ACTIONS(2001),
    [sym_char] = ACTIONS(2001),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2001),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(541)] = {
    [sym__expression] = STATE(4077),
    [sym_block] = STATE(4077),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4077),
    [sym_nil] = STATE(4077),
    [sym__atom] = STATE(4077),
    [sym_quoted_atom] = STATE(4077),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4077),
    [sym_charlist] = STATE(4077),
    [sym_sigil] = STATE(4077),
    [sym_list] = STATE(4077),
    [sym_tuple] = STATE(4077),
    [sym_bitstring] = STATE(4077),
    [sym_map] = STATE(4077),
    [sym__nullary_operator] = STATE(4077),
    [sym_unary_operator] = STATE(4077),
    [sym_binary_operator] = STATE(4077),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4077),
    [sym_call] = STATE(4077),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4077),
    [sym_anonymous_function] = STATE(4077),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2003),
    [sym_integer] = ACTIONS(2003),
    [sym_float] = ACTIONS(2003),
    [sym_char] = ACTIONS(2003),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2003),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(542)] = {
    [sym__expression] = STATE(4078),
    [sym_block] = STATE(4078),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4078),
    [sym_nil] = STATE(4078),
    [sym__atom] = STATE(4078),
    [sym_quoted_atom] = STATE(4078),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4078),
    [sym_charlist] = STATE(4078),
    [sym_sigil] = STATE(4078),
    [sym_list] = STATE(4078),
    [sym_tuple] = STATE(4078),
    [sym_bitstring] = STATE(4078),
    [sym_map] = STATE(4078),
    [sym__nullary_operator] = STATE(4078),
    [sym_unary_operator] = STATE(4078),
    [sym_binary_operator] = STATE(4078),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4078),
    [sym_call] = STATE(4078),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4078),
    [sym_anonymous_function] = STATE(4078),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2005),
    [sym_integer] = ACTIONS(2005),
    [sym_float] = ACTIONS(2005),
    [sym_char] = ACTIONS(2005),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2005),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(543)] = {
    [sym__expression] = STATE(4079),
    [sym_block] = STATE(4079),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4079),
    [sym_nil] = STATE(4079),
    [sym__atom] = STATE(4079),
    [sym_quoted_atom] = STATE(4079),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4079),
    [sym_charlist] = STATE(4079),
    [sym_sigil] = STATE(4079),
    [sym_list] = STATE(4079),
    [sym_tuple] = STATE(4079),
    [sym_bitstring] = STATE(4079),
    [sym_map] = STATE(4079),
    [sym__nullary_operator] = STATE(4079),
    [sym_unary_operator] = STATE(4079),
    [sym_binary_operator] = STATE(4079),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4079),
    [sym_call] = STATE(4079),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4079),
    [sym_anonymous_function] = STATE(4079),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2007),
    [sym_integer] = ACTIONS(2007),
    [sym_float] = ACTIONS(2007),
    [sym_char] = ACTIONS(2007),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2007),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(544)] = {
    [sym__expression] = STATE(4103),
    [sym_block] = STATE(4103),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4103),
    [sym_nil] = STATE(4103),
    [sym__atom] = STATE(4103),
    [sym_quoted_atom] = STATE(4103),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4103),
    [sym_charlist] = STATE(4103),
    [sym_sigil] = STATE(4103),
    [sym_list] = STATE(4103),
    [sym_tuple] = STATE(4103),
    [sym_bitstring] = STATE(4103),
    [sym_map] = STATE(4103),
    [sym__nullary_operator] = STATE(4103),
    [sym_unary_operator] = STATE(4103),
    [sym_binary_operator] = STATE(4103),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4103),
    [sym_call] = STATE(4103),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4103),
    [sym_anonymous_function] = STATE(4103),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2009),
    [sym_integer] = ACTIONS(2009),
    [sym_float] = ACTIONS(2009),
    [sym_char] = ACTIONS(2009),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(545)] = {
    [sym__expression] = STATE(4104),
    [sym_block] = STATE(4104),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4104),
    [sym_nil] = STATE(4104),
    [sym__atom] = STATE(4104),
    [sym_quoted_atom] = STATE(4104),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4104),
    [sym_charlist] = STATE(4104),
    [sym_sigil] = STATE(4104),
    [sym_list] = STATE(4104),
    [sym_tuple] = STATE(4104),
    [sym_bitstring] = STATE(4104),
    [sym_map] = STATE(4104),
    [sym__nullary_operator] = STATE(4104),
    [sym_unary_operator] = STATE(4104),
    [sym_binary_operator] = STATE(4104),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4104),
    [sym_call] = STATE(4104),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4104),
    [sym_anonymous_function] = STATE(4104),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2011),
    [sym_integer] = ACTIONS(2011),
    [sym_float] = ACTIONS(2011),
    [sym_char] = ACTIONS(2011),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2011),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(546)] = {
    [sym__expression] = STATE(4107),
    [sym_block] = STATE(4107),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4107),
    [sym_nil] = STATE(4107),
    [sym__atom] = STATE(4107),
    [sym_quoted_atom] = STATE(4107),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4107),
    [sym_charlist] = STATE(4107),
    [sym_sigil] = STATE(4107),
    [sym_list] = STATE(4107),
    [sym_tuple] = STATE(4107),
    [sym_bitstring] = STATE(4107),
    [sym_map] = STATE(4107),
    [sym__nullary_operator] = STATE(4107),
    [sym_unary_operator] = STATE(4107),
    [sym_binary_operator] = STATE(4107),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4107),
    [sym_call] = STATE(4107),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4107),
    [sym_anonymous_function] = STATE(4107),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2013),
    [sym_integer] = ACTIONS(2013),
    [sym_float] = ACTIONS(2013),
    [sym_char] = ACTIONS(2013),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(547)] = {
    [sym__expression] = STATE(2710),
    [sym_block] = STATE(2710),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2710),
    [sym_nil] = STATE(2710),
    [sym__atom] = STATE(2710),
    [sym_quoted_atom] = STATE(2710),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2710),
    [sym_charlist] = STATE(2710),
    [sym_sigil] = STATE(2710),
    [sym_list] = STATE(2710),
    [sym_tuple] = STATE(2710),
    [sym_bitstring] = STATE(2710),
    [sym_map] = STATE(2710),
    [sym__nullary_operator] = STATE(2710),
    [sym_unary_operator] = STATE(2710),
    [sym_binary_operator] = STATE(2710),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2710),
    [sym_call] = STATE(2710),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2710),
    [sym_anonymous_function] = STATE(2710),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2015),
    [sym_integer] = ACTIONS(2015),
    [sym_float] = ACTIONS(2015),
    [sym_char] = ACTIONS(2015),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2015),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(548)] = {
    [sym__expression] = STATE(1310),
    [sym_block] = STATE(1310),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(1310),
    [sym_nil] = STATE(1310),
    [sym__atom] = STATE(1310),
    [sym_quoted_atom] = STATE(1310),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1310),
    [sym_charlist] = STATE(1310),
    [sym_sigil] = STATE(1310),
    [sym_list] = STATE(1310),
    [sym_tuple] = STATE(1310),
    [sym_bitstring] = STATE(1310),
    [sym_map] = STATE(1310),
    [sym__nullary_operator] = STATE(1310),
    [sym_unary_operator] = STATE(1310),
    [sym_binary_operator] = STATE(1310),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1310),
    [sym_call] = STATE(1310),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1310),
    [sym_anonymous_function] = STATE(1310),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2017),
    [sym_integer] = ACTIONS(2017),
    [sym_float] = ACTIONS(2017),
    [sym_char] = ACTIONS(2017),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2017),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(549)] = {
    [sym__expression] = STATE(2479),
    [sym_block] = STATE(2479),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2479),
    [sym_nil] = STATE(2479),
    [sym__atom] = STATE(2479),
    [sym_quoted_atom] = STATE(2479),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2479),
    [sym_charlist] = STATE(2479),
    [sym_sigil] = STATE(2479),
    [sym_list] = STATE(2479),
    [sym_tuple] = STATE(2479),
    [sym_bitstring] = STATE(2479),
    [sym_map] = STATE(2479),
    [sym__nullary_operator] = STATE(2479),
    [sym_unary_operator] = STATE(2479),
    [sym_binary_operator] = STATE(2479),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2479),
    [sym_call] = STATE(2479),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2479),
    [sym_anonymous_function] = STATE(2479),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2019),
    [sym_integer] = ACTIONS(2019),
    [sym_float] = ACTIONS(2019),
    [sym_char] = ACTIONS(2019),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2019),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(550)] = {
    [sym__expression] = STATE(1189),
    [sym_block] = STATE(1189),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(1189),
    [sym_nil] = STATE(1189),
    [sym__atom] = STATE(1189),
    [sym_quoted_atom] = STATE(1189),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1189),
    [sym_charlist] = STATE(1189),
    [sym_sigil] = STATE(1189),
    [sym_list] = STATE(1189),
    [sym_tuple] = STATE(1189),
    [sym_bitstring] = STATE(1189),
    [sym_map] = STATE(1189),
    [sym__nullary_operator] = STATE(1189),
    [sym_unary_operator] = STATE(1189),
    [sym_binary_operator] = STATE(1189),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1189),
    [sym_call] = STATE(1189),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1189),
    [sym_anonymous_function] = STATE(1189),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2021),
    [sym_integer] = ACTIONS(2021),
    [sym_float] = ACTIONS(2021),
    [sym_char] = ACTIONS(2021),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(551)] = {
    [sym__expression] = STATE(2560),
    [sym_block] = STATE(2560),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2560),
    [sym_nil] = STATE(2560),
    [sym__atom] = STATE(2560),
    [sym_quoted_atom] = STATE(2560),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2560),
    [sym_charlist] = STATE(2560),
    [sym_sigil] = STATE(2560),
    [sym_list] = STATE(2560),
    [sym_tuple] = STATE(2560),
    [sym_bitstring] = STATE(2560),
    [sym_map] = STATE(2560),
    [sym__nullary_operator] = STATE(2560),
    [sym_unary_operator] = STATE(2560),
    [sym_binary_operator] = STATE(2560),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2560),
    [sym_call] = STATE(2560),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2560),
    [sym_anonymous_function] = STATE(2560),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2023),
    [sym_integer] = ACTIONS(2023),
    [sym_float] = ACTIONS(2023),
    [sym_char] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2023),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(552)] = {
    [sym__expression] = STATE(2561),
    [sym_block] = STATE(2561),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2561),
    [sym_nil] = STATE(2561),
    [sym__atom] = STATE(2561),
    [sym_quoted_atom] = STATE(2561),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2561),
    [sym_charlist] = STATE(2561),
    [sym_sigil] = STATE(2561),
    [sym_list] = STATE(2561),
    [sym_tuple] = STATE(2561),
    [sym_bitstring] = STATE(2561),
    [sym_map] = STATE(2561),
    [sym__nullary_operator] = STATE(2561),
    [sym_unary_operator] = STATE(2561),
    [sym_binary_operator] = STATE(2561),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2561),
    [sym_call] = STATE(2561),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2561),
    [sym_anonymous_function] = STATE(2561),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2025),
    [sym_integer] = ACTIONS(2025),
    [sym_float] = ACTIONS(2025),
    [sym_char] = ACTIONS(2025),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(553)] = {
    [sym__expression] = STATE(2567),
    [sym_block] = STATE(2567),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2567),
    [sym_nil] = STATE(2567),
    [sym__atom] = STATE(2567),
    [sym_quoted_atom] = STATE(2567),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2567),
    [sym_charlist] = STATE(2567),
    [sym_sigil] = STATE(2567),
    [sym_list] = STATE(2567),
    [sym_tuple] = STATE(2567),
    [sym_bitstring] = STATE(2567),
    [sym_map] = STATE(2567),
    [sym__nullary_operator] = STATE(2567),
    [sym_unary_operator] = STATE(2567),
    [sym_binary_operator] = STATE(2567),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2567),
    [sym_call] = STATE(2567),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2567),
    [sym_anonymous_function] = STATE(2567),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2027),
    [sym_integer] = ACTIONS(2027),
    [sym_float] = ACTIONS(2027),
    [sym_char] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2027),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(554)] = {
    [sym__expression] = STATE(2568),
    [sym_block] = STATE(2568),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2568),
    [sym_nil] = STATE(2568),
    [sym__atom] = STATE(2568),
    [sym_quoted_atom] = STATE(2568),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2568),
    [sym_charlist] = STATE(2568),
    [sym_sigil] = STATE(2568),
    [sym_list] = STATE(2568),
    [sym_tuple] = STATE(2568),
    [sym_bitstring] = STATE(2568),
    [sym_map] = STATE(2568),
    [sym__nullary_operator] = STATE(2568),
    [sym_unary_operator] = STATE(2568),
    [sym_binary_operator] = STATE(2568),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2568),
    [sym_call] = STATE(2568),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2568),
    [sym_anonymous_function] = STATE(2568),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2029),
    [sym_integer] = ACTIONS(2029),
    [sym_float] = ACTIONS(2029),
    [sym_char] = ACTIONS(2029),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(555)] = {
    [sym__expression] = STATE(2569),
    [sym_block] = STATE(2569),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2569),
    [sym_nil] = STATE(2569),
    [sym__atom] = STATE(2569),
    [sym_quoted_atom] = STATE(2569),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2569),
    [sym_charlist] = STATE(2569),
    [sym_sigil] = STATE(2569),
    [sym_list] = STATE(2569),
    [sym_tuple] = STATE(2569),
    [sym_bitstring] = STATE(2569),
    [sym_map] = STATE(2569),
    [sym__nullary_operator] = STATE(2569),
    [sym_unary_operator] = STATE(2569),
    [sym_binary_operator] = STATE(2569),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2569),
    [sym_call] = STATE(2569),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2569),
    [sym_anonymous_function] = STATE(2569),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2031),
    [sym_integer] = ACTIONS(2031),
    [sym_float] = ACTIONS(2031),
    [sym_char] = ACTIONS(2031),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2031),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(556)] = {
    [sym__expression] = STATE(2570),
    [sym_block] = STATE(2570),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2570),
    [sym_nil] = STATE(2570),
    [sym__atom] = STATE(2570),
    [sym_quoted_atom] = STATE(2570),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2570),
    [sym_charlist] = STATE(2570),
    [sym_sigil] = STATE(2570),
    [sym_list] = STATE(2570),
    [sym_tuple] = STATE(2570),
    [sym_bitstring] = STATE(2570),
    [sym_map] = STATE(2570),
    [sym__nullary_operator] = STATE(2570),
    [sym_unary_operator] = STATE(2570),
    [sym_binary_operator] = STATE(2570),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2570),
    [sym_call] = STATE(2570),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2570),
    [sym_anonymous_function] = STATE(2570),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2033),
    [sym_integer] = ACTIONS(2033),
    [sym_float] = ACTIONS(2033),
    [sym_char] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2033),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(557)] = {
    [sym__expression] = STATE(2573),
    [sym_block] = STATE(2573),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2573),
    [sym_nil] = STATE(2573),
    [sym__atom] = STATE(2573),
    [sym_quoted_atom] = STATE(2573),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2573),
    [sym_charlist] = STATE(2573),
    [sym_sigil] = STATE(2573),
    [sym_list] = STATE(2573),
    [sym_tuple] = STATE(2573),
    [sym_bitstring] = STATE(2573),
    [sym_map] = STATE(2573),
    [sym__nullary_operator] = STATE(2573),
    [sym_unary_operator] = STATE(2573),
    [sym_binary_operator] = STATE(2573),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2573),
    [sym_call] = STATE(2573),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2573),
    [sym_anonymous_function] = STATE(2573),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2035),
    [sym_integer] = ACTIONS(2035),
    [sym_float] = ACTIONS(2035),
    [sym_char] = ACTIONS(2035),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2035),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(558)] = {
    [sym__expression] = STATE(2574),
    [sym_block] = STATE(2574),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2574),
    [sym_nil] = STATE(2574),
    [sym__atom] = STATE(2574),
    [sym_quoted_atom] = STATE(2574),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2574),
    [sym_charlist] = STATE(2574),
    [sym_sigil] = STATE(2574),
    [sym_list] = STATE(2574),
    [sym_tuple] = STATE(2574),
    [sym_bitstring] = STATE(2574),
    [sym_map] = STATE(2574),
    [sym__nullary_operator] = STATE(2574),
    [sym_unary_operator] = STATE(2574),
    [sym_binary_operator] = STATE(2574),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2574),
    [sym_call] = STATE(2574),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2574),
    [sym_anonymous_function] = STATE(2574),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2037),
    [sym_integer] = ACTIONS(2037),
    [sym_float] = ACTIONS(2037),
    [sym_char] = ACTIONS(2037),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2037),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(559)] = {
    [sym__expression] = STATE(2575),
    [sym_block] = STATE(2575),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2575),
    [sym_nil] = STATE(2575),
    [sym__atom] = STATE(2575),
    [sym_quoted_atom] = STATE(2575),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2575),
    [sym_charlist] = STATE(2575),
    [sym_sigil] = STATE(2575),
    [sym_list] = STATE(2575),
    [sym_tuple] = STATE(2575),
    [sym_bitstring] = STATE(2575),
    [sym_map] = STATE(2575),
    [sym__nullary_operator] = STATE(2575),
    [sym_unary_operator] = STATE(2575),
    [sym_binary_operator] = STATE(2575),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2575),
    [sym_call] = STATE(2575),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2575),
    [sym_anonymous_function] = STATE(2575),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2039),
    [sym_integer] = ACTIONS(2039),
    [sym_float] = ACTIONS(2039),
    [sym_char] = ACTIONS(2039),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2039),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(560)] = {
    [sym__expression] = STATE(2576),
    [sym_block] = STATE(2576),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2576),
    [sym_nil] = STATE(2576),
    [sym__atom] = STATE(2576),
    [sym_quoted_atom] = STATE(2576),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2576),
    [sym_charlist] = STATE(2576),
    [sym_sigil] = STATE(2576),
    [sym_list] = STATE(2576),
    [sym_tuple] = STATE(2576),
    [sym_bitstring] = STATE(2576),
    [sym_map] = STATE(2576),
    [sym__nullary_operator] = STATE(2576),
    [sym_unary_operator] = STATE(2576),
    [sym_binary_operator] = STATE(2576),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2576),
    [sym_call] = STATE(2576),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2576),
    [sym_anonymous_function] = STATE(2576),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2041),
    [sym_integer] = ACTIONS(2041),
    [sym_float] = ACTIONS(2041),
    [sym_char] = ACTIONS(2041),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(561)] = {
    [sym__expression] = STATE(2577),
    [sym_block] = STATE(2577),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2577),
    [sym_nil] = STATE(2577),
    [sym__atom] = STATE(2577),
    [sym_quoted_atom] = STATE(2577),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2577),
    [sym_charlist] = STATE(2577),
    [sym_sigil] = STATE(2577),
    [sym_list] = STATE(2577),
    [sym_tuple] = STATE(2577),
    [sym_bitstring] = STATE(2577),
    [sym_map] = STATE(2577),
    [sym__nullary_operator] = STATE(2577),
    [sym_unary_operator] = STATE(2577),
    [sym_binary_operator] = STATE(2577),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2577),
    [sym_call] = STATE(2577),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2577),
    [sym_anonymous_function] = STATE(2577),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2043),
    [sym_integer] = ACTIONS(2043),
    [sym_float] = ACTIONS(2043),
    [sym_char] = ACTIONS(2043),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2043),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(562)] = {
    [sym__expression] = STATE(2578),
    [sym_block] = STATE(2578),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2578),
    [sym_nil] = STATE(2578),
    [sym__atom] = STATE(2578),
    [sym_quoted_atom] = STATE(2578),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2578),
    [sym_charlist] = STATE(2578),
    [sym_sigil] = STATE(2578),
    [sym_list] = STATE(2578),
    [sym_tuple] = STATE(2578),
    [sym_bitstring] = STATE(2578),
    [sym_map] = STATE(2578),
    [sym__nullary_operator] = STATE(2578),
    [sym_unary_operator] = STATE(2578),
    [sym_binary_operator] = STATE(2578),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2578),
    [sym_call] = STATE(2578),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2578),
    [sym_anonymous_function] = STATE(2578),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2045),
    [sym_integer] = ACTIONS(2045),
    [sym_float] = ACTIONS(2045),
    [sym_char] = ACTIONS(2045),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2045),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(563)] = {
    [sym__expression] = STATE(2579),
    [sym_block] = STATE(2579),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2579),
    [sym_nil] = STATE(2579),
    [sym__atom] = STATE(2579),
    [sym_quoted_atom] = STATE(2579),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2579),
    [sym_charlist] = STATE(2579),
    [sym_sigil] = STATE(2579),
    [sym_list] = STATE(2579),
    [sym_tuple] = STATE(2579),
    [sym_bitstring] = STATE(2579),
    [sym_map] = STATE(2579),
    [sym__nullary_operator] = STATE(2579),
    [sym_unary_operator] = STATE(2579),
    [sym_binary_operator] = STATE(2579),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2579),
    [sym_call] = STATE(2579),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2579),
    [sym_anonymous_function] = STATE(2579),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2047),
    [sym_integer] = ACTIONS(2047),
    [sym_float] = ACTIONS(2047),
    [sym_char] = ACTIONS(2047),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2047),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(564)] = {
    [sym__expression] = STATE(2580),
    [sym_block] = STATE(2580),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2580),
    [sym_nil] = STATE(2580),
    [sym__atom] = STATE(2580),
    [sym_quoted_atom] = STATE(2580),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2580),
    [sym_charlist] = STATE(2580),
    [sym_sigil] = STATE(2580),
    [sym_list] = STATE(2580),
    [sym_tuple] = STATE(2580),
    [sym_bitstring] = STATE(2580),
    [sym_map] = STATE(2580),
    [sym__nullary_operator] = STATE(2580),
    [sym_unary_operator] = STATE(2580),
    [sym_binary_operator] = STATE(2580),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2580),
    [sym_call] = STATE(2580),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2580),
    [sym_anonymous_function] = STATE(2580),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2049),
    [sym_integer] = ACTIONS(2049),
    [sym_float] = ACTIONS(2049),
    [sym_char] = ACTIONS(2049),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(565)] = {
    [sym__expression] = STATE(2581),
    [sym_block] = STATE(2581),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2581),
    [sym_nil] = STATE(2581),
    [sym__atom] = STATE(2581),
    [sym_quoted_atom] = STATE(2581),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2581),
    [sym_charlist] = STATE(2581),
    [sym_sigil] = STATE(2581),
    [sym_list] = STATE(2581),
    [sym_tuple] = STATE(2581),
    [sym_bitstring] = STATE(2581),
    [sym_map] = STATE(2581),
    [sym__nullary_operator] = STATE(2581),
    [sym_unary_operator] = STATE(2581),
    [sym_binary_operator] = STATE(2581),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2581),
    [sym_call] = STATE(2581),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2581),
    [sym_anonymous_function] = STATE(2581),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2051),
    [sym_integer] = ACTIONS(2051),
    [sym_float] = ACTIONS(2051),
    [sym_char] = ACTIONS(2051),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2051),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(566)] = {
    [sym__expression] = STATE(2582),
    [sym_block] = STATE(2582),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2582),
    [sym_nil] = STATE(2582),
    [sym__atom] = STATE(2582),
    [sym_quoted_atom] = STATE(2582),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2582),
    [sym_charlist] = STATE(2582),
    [sym_sigil] = STATE(2582),
    [sym_list] = STATE(2582),
    [sym_tuple] = STATE(2582),
    [sym_bitstring] = STATE(2582),
    [sym_map] = STATE(2582),
    [sym__nullary_operator] = STATE(2582),
    [sym_unary_operator] = STATE(2582),
    [sym_binary_operator] = STATE(2582),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2582),
    [sym_call] = STATE(2582),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2582),
    [sym_anonymous_function] = STATE(2582),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2053),
    [sym_integer] = ACTIONS(2053),
    [sym_float] = ACTIONS(2053),
    [sym_char] = ACTIONS(2053),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(567)] = {
    [sym__expression] = STATE(2184),
    [sym_block] = STATE(2184),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2184),
    [sym_nil] = STATE(2184),
    [sym__atom] = STATE(2184),
    [sym_quoted_atom] = STATE(2184),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2184),
    [sym_charlist] = STATE(2184),
    [sym_sigil] = STATE(2184),
    [sym_list] = STATE(2184),
    [sym_tuple] = STATE(2184),
    [sym_bitstring] = STATE(2184),
    [sym_map] = STATE(2184),
    [sym__nullary_operator] = STATE(2184),
    [sym_unary_operator] = STATE(2184),
    [sym_binary_operator] = STATE(2184),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2184),
    [sym_call] = STATE(2184),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2184),
    [sym_anonymous_function] = STATE(2184),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2055),
    [sym_integer] = ACTIONS(2055),
    [sym_float] = ACTIONS(2055),
    [sym_char] = ACTIONS(2055),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2055),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(568)] = {
    [sym__expression] = STATE(2185),
    [sym_block] = STATE(2185),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2185),
    [sym_nil] = STATE(2185),
    [sym__atom] = STATE(2185),
    [sym_quoted_atom] = STATE(2185),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2185),
    [sym_charlist] = STATE(2185),
    [sym_sigil] = STATE(2185),
    [sym_list] = STATE(2185),
    [sym_tuple] = STATE(2185),
    [sym_bitstring] = STATE(2185),
    [sym_map] = STATE(2185),
    [sym__nullary_operator] = STATE(2185),
    [sym_unary_operator] = STATE(2185),
    [sym_binary_operator] = STATE(2185),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2185),
    [sym_call] = STATE(2185),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2185),
    [sym_anonymous_function] = STATE(2185),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2057),
    [sym_integer] = ACTIONS(2057),
    [sym_float] = ACTIONS(2057),
    [sym_char] = ACTIONS(2057),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(569)] = {
    [sym__expression] = STATE(2257),
    [sym_block] = STATE(2257),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2257),
    [sym_nil] = STATE(2257),
    [sym__atom] = STATE(2257),
    [sym_quoted_atom] = STATE(2257),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2257),
    [sym_charlist] = STATE(2257),
    [sym_sigil] = STATE(2257),
    [sym_list] = STATE(2257),
    [sym_tuple] = STATE(2257),
    [sym_bitstring] = STATE(2257),
    [sym_map] = STATE(2257),
    [sym__nullary_operator] = STATE(2257),
    [sym_unary_operator] = STATE(2257),
    [sym_binary_operator] = STATE(2257),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2257),
    [sym_call] = STATE(2257),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2257),
    [sym_anonymous_function] = STATE(2257),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2059),
    [sym_integer] = ACTIONS(2059),
    [sym_float] = ACTIONS(2059),
    [sym_char] = ACTIONS(2059),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2059),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(570)] = {
    [sym__expression] = STATE(2259),
    [sym_block] = STATE(2259),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2259),
    [sym_nil] = STATE(2259),
    [sym__atom] = STATE(2259),
    [sym_quoted_atom] = STATE(2259),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2259),
    [sym_charlist] = STATE(2259),
    [sym_sigil] = STATE(2259),
    [sym_list] = STATE(2259),
    [sym_tuple] = STATE(2259),
    [sym_bitstring] = STATE(2259),
    [sym_map] = STATE(2259),
    [sym__nullary_operator] = STATE(2259),
    [sym_unary_operator] = STATE(2259),
    [sym_binary_operator] = STATE(2259),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2259),
    [sym_call] = STATE(2259),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2259),
    [sym_anonymous_function] = STATE(2259),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2061),
    [sym_integer] = ACTIONS(2061),
    [sym_float] = ACTIONS(2061),
    [sym_char] = ACTIONS(2061),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2061),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(571)] = {
    [sym__expression] = STATE(2640),
    [sym_block] = STATE(2640),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2640),
    [sym_nil] = STATE(2640),
    [sym__atom] = STATE(2640),
    [sym_quoted_atom] = STATE(2640),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2640),
    [sym_charlist] = STATE(2640),
    [sym_sigil] = STATE(2640),
    [sym_list] = STATE(2640),
    [sym_tuple] = STATE(2640),
    [sym_bitstring] = STATE(2640),
    [sym_map] = STATE(2640),
    [sym__nullary_operator] = STATE(2640),
    [sym_unary_operator] = STATE(2640),
    [sym_binary_operator] = STATE(2640),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2640),
    [sym_call] = STATE(2640),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2640),
    [sym_anonymous_function] = STATE(2640),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2063),
    [sym_integer] = ACTIONS(2063),
    [sym_float] = ACTIONS(2063),
    [sym_char] = ACTIONS(2063),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2063),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(572)] = {
    [sym__expression] = STATE(2646),
    [sym_block] = STATE(2646),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2646),
    [sym_nil] = STATE(2646),
    [sym__atom] = STATE(2646),
    [sym_quoted_atom] = STATE(2646),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2646),
    [sym_charlist] = STATE(2646),
    [sym_sigil] = STATE(2646),
    [sym_list] = STATE(2646),
    [sym_tuple] = STATE(2646),
    [sym_bitstring] = STATE(2646),
    [sym_map] = STATE(2646),
    [sym__nullary_operator] = STATE(2646),
    [sym_unary_operator] = STATE(2646),
    [sym_binary_operator] = STATE(2646),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2646),
    [sym_call] = STATE(2646),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2646),
    [sym_anonymous_function] = STATE(2646),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2065),
    [sym_integer] = ACTIONS(2065),
    [sym_float] = ACTIONS(2065),
    [sym_char] = ACTIONS(2065),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(573)] = {
    [sym__expression] = STATE(4452),
    [sym_block] = STATE(4452),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4452),
    [sym_nil] = STATE(4452),
    [sym__atom] = STATE(4452),
    [sym_quoted_atom] = STATE(4452),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4452),
    [sym_charlist] = STATE(4452),
    [sym_sigil] = STATE(4452),
    [sym_list] = STATE(4452),
    [sym_tuple] = STATE(4452),
    [sym_bitstring] = STATE(4452),
    [sym_map] = STATE(4452),
    [sym__nullary_operator] = STATE(4452),
    [sym_unary_operator] = STATE(4452),
    [sym_binary_operator] = STATE(4452),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4452),
    [sym_call] = STATE(4452),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4452),
    [sym_anonymous_function] = STATE(4452),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(1549),
    [sym_integer] = ACTIONS(1549),
    [sym_float] = ACTIONS(1549),
    [sym_char] = ACTIONS(1549),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(1549),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(574)] = {
    [sym__expression] = STATE(2649),
    [sym_block] = STATE(2649),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2649),
    [sym_nil] = STATE(2649),
    [sym__atom] = STATE(2649),
    [sym_quoted_atom] = STATE(2649),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2649),
    [sym_charlist] = STATE(2649),
    [sym_sigil] = STATE(2649),
    [sym_list] = STATE(2649),
    [sym_tuple] = STATE(2649),
    [sym_bitstring] = STATE(2649),
    [sym_map] = STATE(2649),
    [sym__nullary_operator] = STATE(2649),
    [sym_unary_operator] = STATE(2649),
    [sym_binary_operator] = STATE(2649),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2649),
    [sym_call] = STATE(2649),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2649),
    [sym_anonymous_function] = STATE(2649),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2067),
    [sym_integer] = ACTIONS(2067),
    [sym_float] = ACTIONS(2067),
    [sym_char] = ACTIONS(2067),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2067),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(575)] = {
    [sym__expression] = STATE(2650),
    [sym_block] = STATE(2650),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2650),
    [sym_nil] = STATE(2650),
    [sym__atom] = STATE(2650),
    [sym_quoted_atom] = STATE(2650),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2650),
    [sym_charlist] = STATE(2650),
    [sym_sigil] = STATE(2650),
    [sym_list] = STATE(2650),
    [sym_tuple] = STATE(2650),
    [sym_bitstring] = STATE(2650),
    [sym_map] = STATE(2650),
    [sym__nullary_operator] = STATE(2650),
    [sym_unary_operator] = STATE(2650),
    [sym_binary_operator] = STATE(2650),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2650),
    [sym_call] = STATE(2650),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2650),
    [sym_anonymous_function] = STATE(2650),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2069),
    [sym_integer] = ACTIONS(2069),
    [sym_float] = ACTIONS(2069),
    [sym_char] = ACTIONS(2069),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2069),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(576)] = {
    [sym__expression] = STATE(2651),
    [sym_block] = STATE(2651),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2651),
    [sym_nil] = STATE(2651),
    [sym__atom] = STATE(2651),
    [sym_quoted_atom] = STATE(2651),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2651),
    [sym_charlist] = STATE(2651),
    [sym_sigil] = STATE(2651),
    [sym_list] = STATE(2651),
    [sym_tuple] = STATE(2651),
    [sym_bitstring] = STATE(2651),
    [sym_map] = STATE(2651),
    [sym__nullary_operator] = STATE(2651),
    [sym_unary_operator] = STATE(2651),
    [sym_binary_operator] = STATE(2651),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2651),
    [sym_call] = STATE(2651),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2651),
    [sym_anonymous_function] = STATE(2651),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2071),
    [sym_integer] = ACTIONS(2071),
    [sym_float] = ACTIONS(2071),
    [sym_char] = ACTIONS(2071),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2071),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(577)] = {
    [sym__expression] = STATE(2652),
    [sym_block] = STATE(2652),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2652),
    [sym_nil] = STATE(2652),
    [sym__atom] = STATE(2652),
    [sym_quoted_atom] = STATE(2652),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2652),
    [sym_charlist] = STATE(2652),
    [sym_sigil] = STATE(2652),
    [sym_list] = STATE(2652),
    [sym_tuple] = STATE(2652),
    [sym_bitstring] = STATE(2652),
    [sym_map] = STATE(2652),
    [sym__nullary_operator] = STATE(2652),
    [sym_unary_operator] = STATE(2652),
    [sym_binary_operator] = STATE(2652),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2652),
    [sym_call] = STATE(2652),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2652),
    [sym_anonymous_function] = STATE(2652),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2073),
    [sym_integer] = ACTIONS(2073),
    [sym_float] = ACTIONS(2073),
    [sym_char] = ACTIONS(2073),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2073),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(578)] = {
    [sym__expression] = STATE(2655),
    [sym_block] = STATE(2655),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2655),
    [sym_nil] = STATE(2655),
    [sym__atom] = STATE(2655),
    [sym_quoted_atom] = STATE(2655),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2655),
    [sym_charlist] = STATE(2655),
    [sym_sigil] = STATE(2655),
    [sym_list] = STATE(2655),
    [sym_tuple] = STATE(2655),
    [sym_bitstring] = STATE(2655),
    [sym_map] = STATE(2655),
    [sym__nullary_operator] = STATE(2655),
    [sym_unary_operator] = STATE(2655),
    [sym_binary_operator] = STATE(2655),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2655),
    [sym_call] = STATE(2655),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2655),
    [sym_anonymous_function] = STATE(2655),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2075),
    [sym_integer] = ACTIONS(2075),
    [sym_float] = ACTIONS(2075),
    [sym_char] = ACTIONS(2075),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2075),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(579)] = {
    [sym__expression] = STATE(2656),
    [sym_block] = STATE(2656),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2656),
    [sym_nil] = STATE(2656),
    [sym__atom] = STATE(2656),
    [sym_quoted_atom] = STATE(2656),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2656),
    [sym_charlist] = STATE(2656),
    [sym_sigil] = STATE(2656),
    [sym_list] = STATE(2656),
    [sym_tuple] = STATE(2656),
    [sym_bitstring] = STATE(2656),
    [sym_map] = STATE(2656),
    [sym__nullary_operator] = STATE(2656),
    [sym_unary_operator] = STATE(2656),
    [sym_binary_operator] = STATE(2656),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2656),
    [sym_call] = STATE(2656),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2656),
    [sym_anonymous_function] = STATE(2656),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2077),
    [sym_integer] = ACTIONS(2077),
    [sym_float] = ACTIONS(2077),
    [sym_char] = ACTIONS(2077),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2077),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(580)] = {
    [sym__expression] = STATE(2657),
    [sym_block] = STATE(2657),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2657),
    [sym_nil] = STATE(2657),
    [sym__atom] = STATE(2657),
    [sym_quoted_atom] = STATE(2657),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2657),
    [sym_charlist] = STATE(2657),
    [sym_sigil] = STATE(2657),
    [sym_list] = STATE(2657),
    [sym_tuple] = STATE(2657),
    [sym_bitstring] = STATE(2657),
    [sym_map] = STATE(2657),
    [sym__nullary_operator] = STATE(2657),
    [sym_unary_operator] = STATE(2657),
    [sym_binary_operator] = STATE(2657),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2657),
    [sym_call] = STATE(2657),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2657),
    [sym_anonymous_function] = STATE(2657),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2079),
    [sym_integer] = ACTIONS(2079),
    [sym_float] = ACTIONS(2079),
    [sym_char] = ACTIONS(2079),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2079),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(581)] = {
    [sym__expression] = STATE(2658),
    [sym_block] = STATE(2658),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2658),
    [sym_nil] = STATE(2658),
    [sym__atom] = STATE(2658),
    [sym_quoted_atom] = STATE(2658),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2658),
    [sym_charlist] = STATE(2658),
    [sym_sigil] = STATE(2658),
    [sym_list] = STATE(2658),
    [sym_tuple] = STATE(2658),
    [sym_bitstring] = STATE(2658),
    [sym_map] = STATE(2658),
    [sym__nullary_operator] = STATE(2658),
    [sym_unary_operator] = STATE(2658),
    [sym_binary_operator] = STATE(2658),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2658),
    [sym_call] = STATE(2658),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2658),
    [sym_anonymous_function] = STATE(2658),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2081),
    [sym_integer] = ACTIONS(2081),
    [sym_float] = ACTIONS(2081),
    [sym_char] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(582)] = {
    [sym__expression] = STATE(2659),
    [sym_block] = STATE(2659),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2659),
    [sym_nil] = STATE(2659),
    [sym__atom] = STATE(2659),
    [sym_quoted_atom] = STATE(2659),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2659),
    [sym_charlist] = STATE(2659),
    [sym_sigil] = STATE(2659),
    [sym_list] = STATE(2659),
    [sym_tuple] = STATE(2659),
    [sym_bitstring] = STATE(2659),
    [sym_map] = STATE(2659),
    [sym__nullary_operator] = STATE(2659),
    [sym_unary_operator] = STATE(2659),
    [sym_binary_operator] = STATE(2659),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2659),
    [sym_call] = STATE(2659),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2659),
    [sym_anonymous_function] = STATE(2659),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2083),
    [sym_integer] = ACTIONS(2083),
    [sym_float] = ACTIONS(2083),
    [sym_char] = ACTIONS(2083),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2083),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(583)] = {
    [sym__expression] = STATE(2660),
    [sym_block] = STATE(2660),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2660),
    [sym_nil] = STATE(2660),
    [sym__atom] = STATE(2660),
    [sym_quoted_atom] = STATE(2660),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2660),
    [sym_charlist] = STATE(2660),
    [sym_sigil] = STATE(2660),
    [sym_list] = STATE(2660),
    [sym_tuple] = STATE(2660),
    [sym_bitstring] = STATE(2660),
    [sym_map] = STATE(2660),
    [sym__nullary_operator] = STATE(2660),
    [sym_unary_operator] = STATE(2660),
    [sym_binary_operator] = STATE(2660),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2660),
    [sym_call] = STATE(2660),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2660),
    [sym_anonymous_function] = STATE(2660),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2085),
    [sym_integer] = ACTIONS(2085),
    [sym_float] = ACTIONS(2085),
    [sym_char] = ACTIONS(2085),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2085),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(584)] = {
    [sym__expression] = STATE(2661),
    [sym_block] = STATE(2661),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2661),
    [sym_nil] = STATE(2661),
    [sym__atom] = STATE(2661),
    [sym_quoted_atom] = STATE(2661),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2661),
    [sym_charlist] = STATE(2661),
    [sym_sigil] = STATE(2661),
    [sym_list] = STATE(2661),
    [sym_tuple] = STATE(2661),
    [sym_bitstring] = STATE(2661),
    [sym_map] = STATE(2661),
    [sym__nullary_operator] = STATE(2661),
    [sym_unary_operator] = STATE(2661),
    [sym_binary_operator] = STATE(2661),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2661),
    [sym_call] = STATE(2661),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2661),
    [sym_anonymous_function] = STATE(2661),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2087),
    [sym_integer] = ACTIONS(2087),
    [sym_float] = ACTIONS(2087),
    [sym_char] = ACTIONS(2087),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2087),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(585)] = {
    [sym__expression] = STATE(2662),
    [sym_block] = STATE(2662),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2662),
    [sym_nil] = STATE(2662),
    [sym__atom] = STATE(2662),
    [sym_quoted_atom] = STATE(2662),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2662),
    [sym_charlist] = STATE(2662),
    [sym_sigil] = STATE(2662),
    [sym_list] = STATE(2662),
    [sym_tuple] = STATE(2662),
    [sym_bitstring] = STATE(2662),
    [sym_map] = STATE(2662),
    [sym__nullary_operator] = STATE(2662),
    [sym_unary_operator] = STATE(2662),
    [sym_binary_operator] = STATE(2662),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2662),
    [sym_call] = STATE(2662),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2662),
    [sym_anonymous_function] = STATE(2662),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2089),
    [sym_integer] = ACTIONS(2089),
    [sym_float] = ACTIONS(2089),
    [sym_char] = ACTIONS(2089),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2089),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(586)] = {
    [sym__expression] = STATE(2663),
    [sym_block] = STATE(2663),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2663),
    [sym_nil] = STATE(2663),
    [sym__atom] = STATE(2663),
    [sym_quoted_atom] = STATE(2663),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2663),
    [sym_charlist] = STATE(2663),
    [sym_sigil] = STATE(2663),
    [sym_list] = STATE(2663),
    [sym_tuple] = STATE(2663),
    [sym_bitstring] = STATE(2663),
    [sym_map] = STATE(2663),
    [sym__nullary_operator] = STATE(2663),
    [sym_unary_operator] = STATE(2663),
    [sym_binary_operator] = STATE(2663),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2663),
    [sym_call] = STATE(2663),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2663),
    [sym_anonymous_function] = STATE(2663),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2091),
    [sym_integer] = ACTIONS(2091),
    [sym_float] = ACTIONS(2091),
    [sym_char] = ACTIONS(2091),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2091),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(587)] = {
    [sym__expression] = STATE(2314),
    [sym_block] = STATE(2314),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2314),
    [sym_nil] = STATE(2314),
    [sym__atom] = STATE(2314),
    [sym_quoted_atom] = STATE(2314),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2314),
    [sym_charlist] = STATE(2314),
    [sym_sigil] = STATE(2314),
    [sym_list] = STATE(2314),
    [sym_tuple] = STATE(2314),
    [sym_bitstring] = STATE(2314),
    [sym_map] = STATE(2314),
    [sym__nullary_operator] = STATE(2314),
    [sym_unary_operator] = STATE(2314),
    [sym_binary_operator] = STATE(2314),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2314),
    [sym_call] = STATE(2314),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2314),
    [sym_anonymous_function] = STATE(2314),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2093),
    [sym_integer] = ACTIONS(2093),
    [sym_float] = ACTIONS(2093),
    [sym_char] = ACTIONS(2093),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2093),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(588)] = {
    [sym__expression] = STATE(4381),
    [sym_block] = STATE(4381),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4381),
    [sym_nil] = STATE(4381),
    [sym__atom] = STATE(4381),
    [sym_quoted_atom] = STATE(4381),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4381),
    [sym_charlist] = STATE(4381),
    [sym_sigil] = STATE(4381),
    [sym_list] = STATE(4381),
    [sym_tuple] = STATE(4381),
    [sym_bitstring] = STATE(4381),
    [sym_map] = STATE(4381),
    [sym__nullary_operator] = STATE(4381),
    [sym_unary_operator] = STATE(4381),
    [sym_binary_operator] = STATE(4381),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4381),
    [sym_call] = STATE(4381),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4381),
    [sym_anonymous_function] = STATE(4381),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2095),
    [sym_integer] = ACTIONS(2095),
    [sym_float] = ACTIONS(2095),
    [sym_char] = ACTIONS(2095),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2095),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(589)] = {
    [sym__expression] = STATE(3679),
    [sym_block] = STATE(3679),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3679),
    [sym_nil] = STATE(3679),
    [sym__atom] = STATE(3679),
    [sym_quoted_atom] = STATE(3679),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3679),
    [sym_charlist] = STATE(3679),
    [sym_sigil] = STATE(3679),
    [sym_list] = STATE(3679),
    [sym_tuple] = STATE(3679),
    [sym_bitstring] = STATE(3679),
    [sym_map] = STATE(3679),
    [sym__nullary_operator] = STATE(3679),
    [sym_unary_operator] = STATE(3679),
    [sym_binary_operator] = STATE(3679),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3679),
    [sym_call] = STATE(3679),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3679),
    [sym_anonymous_function] = STATE(3679),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2097),
    [sym_integer] = ACTIONS(2097),
    [sym_float] = ACTIONS(2097),
    [sym_char] = ACTIONS(2097),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2097),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(590)] = {
    [sym__expression] = STATE(3680),
    [sym_block] = STATE(3680),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3680),
    [sym_nil] = STATE(3680),
    [sym__atom] = STATE(3680),
    [sym_quoted_atom] = STATE(3680),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3680),
    [sym_charlist] = STATE(3680),
    [sym_sigil] = STATE(3680),
    [sym_list] = STATE(3680),
    [sym_tuple] = STATE(3680),
    [sym_bitstring] = STATE(3680),
    [sym_map] = STATE(3680),
    [sym__nullary_operator] = STATE(3680),
    [sym_unary_operator] = STATE(3680),
    [sym_binary_operator] = STATE(3680),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3680),
    [sym_call] = STATE(3680),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3680),
    [sym_anonymous_function] = STATE(3680),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2099),
    [sym_integer] = ACTIONS(2099),
    [sym_float] = ACTIONS(2099),
    [sym_char] = ACTIONS(2099),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2099),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(591)] = {
    [sym__expression] = STATE(4164),
    [sym_block] = STATE(4164),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4164),
    [sym_nil] = STATE(4164),
    [sym__atom] = STATE(4164),
    [sym_quoted_atom] = STATE(4164),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4164),
    [sym_charlist] = STATE(4164),
    [sym_sigil] = STATE(4164),
    [sym_list] = STATE(4164),
    [sym_tuple] = STATE(4164),
    [sym_bitstring] = STATE(4164),
    [sym_map] = STATE(4164),
    [sym__nullary_operator] = STATE(4164),
    [sym_unary_operator] = STATE(4164),
    [sym_binary_operator] = STATE(4164),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4164),
    [sym_call] = STATE(4164),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4164),
    [sym_anonymous_function] = STATE(4164),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2101),
    [sym_integer] = ACTIONS(2101),
    [sym_float] = ACTIONS(2101),
    [sym_char] = ACTIONS(2101),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2101),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(592)] = {
    [sym__expression] = STATE(3704),
    [sym_block] = STATE(3704),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3704),
    [sym_nil] = STATE(3704),
    [sym__atom] = STATE(3704),
    [sym_quoted_atom] = STATE(3704),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3704),
    [sym_charlist] = STATE(3704),
    [sym_sigil] = STATE(3704),
    [sym_list] = STATE(3704),
    [sym_tuple] = STATE(3704),
    [sym_bitstring] = STATE(3704),
    [sym_map] = STATE(3704),
    [sym__nullary_operator] = STATE(3704),
    [sym_unary_operator] = STATE(3704),
    [sym_binary_operator] = STATE(3704),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3704),
    [sym_call] = STATE(3704),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3704),
    [sym_anonymous_function] = STATE(3704),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2103),
    [sym_integer] = ACTIONS(2103),
    [sym_float] = ACTIONS(2103),
    [sym_char] = ACTIONS(2103),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2103),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(593)] = {
    [sym__expression] = STATE(3705),
    [sym_block] = STATE(3705),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3705),
    [sym_nil] = STATE(3705),
    [sym__atom] = STATE(3705),
    [sym_quoted_atom] = STATE(3705),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3705),
    [sym_charlist] = STATE(3705),
    [sym_sigil] = STATE(3705),
    [sym_list] = STATE(3705),
    [sym_tuple] = STATE(3705),
    [sym_bitstring] = STATE(3705),
    [sym_map] = STATE(3705),
    [sym__nullary_operator] = STATE(3705),
    [sym_unary_operator] = STATE(3705),
    [sym_binary_operator] = STATE(3705),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3705),
    [sym_call] = STATE(3705),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3705),
    [sym_anonymous_function] = STATE(3705),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2105),
    [sym_integer] = ACTIONS(2105),
    [sym_float] = ACTIONS(2105),
    [sym_char] = ACTIONS(2105),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2105),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(594)] = {
    [sym__expression] = STATE(3719),
    [sym_block] = STATE(3719),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3719),
    [sym_nil] = STATE(3719),
    [sym__atom] = STATE(3719),
    [sym_quoted_atom] = STATE(3719),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3719),
    [sym_charlist] = STATE(3719),
    [sym_sigil] = STATE(3719),
    [sym_list] = STATE(3719),
    [sym_tuple] = STATE(3719),
    [sym_bitstring] = STATE(3719),
    [sym_map] = STATE(3719),
    [sym__nullary_operator] = STATE(3719),
    [sym_unary_operator] = STATE(3719),
    [sym_binary_operator] = STATE(3719),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3719),
    [sym_call] = STATE(3719),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3719),
    [sym_anonymous_function] = STATE(3719),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2107),
    [sym_integer] = ACTIONS(2107),
    [sym_float] = ACTIONS(2107),
    [sym_char] = ACTIONS(2107),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2107),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(595)] = {
    [sym__expression] = STATE(3720),
    [sym_block] = STATE(3720),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3720),
    [sym_nil] = STATE(3720),
    [sym__atom] = STATE(3720),
    [sym_quoted_atom] = STATE(3720),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3720),
    [sym_charlist] = STATE(3720),
    [sym_sigil] = STATE(3720),
    [sym_list] = STATE(3720),
    [sym_tuple] = STATE(3720),
    [sym_bitstring] = STATE(3720),
    [sym_map] = STATE(3720),
    [sym__nullary_operator] = STATE(3720),
    [sym_unary_operator] = STATE(3720),
    [sym_binary_operator] = STATE(3720),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3720),
    [sym_call] = STATE(3720),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3720),
    [sym_anonymous_function] = STATE(3720),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2109),
    [sym_integer] = ACTIONS(2109),
    [sym_float] = ACTIONS(2109),
    [sym_char] = ACTIONS(2109),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(596)] = {
    [sym__expression] = STATE(3723),
    [sym_block] = STATE(3723),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3723),
    [sym_nil] = STATE(3723),
    [sym__atom] = STATE(3723),
    [sym_quoted_atom] = STATE(3723),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3723),
    [sym_charlist] = STATE(3723),
    [sym_sigil] = STATE(3723),
    [sym_list] = STATE(3723),
    [sym_tuple] = STATE(3723),
    [sym_bitstring] = STATE(3723),
    [sym_map] = STATE(3723),
    [sym__nullary_operator] = STATE(3723),
    [sym_unary_operator] = STATE(3723),
    [sym_binary_operator] = STATE(3723),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3723),
    [sym_call] = STATE(3723),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3723),
    [sym_anonymous_function] = STATE(3723),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2111),
    [sym_integer] = ACTIONS(2111),
    [sym_float] = ACTIONS(2111),
    [sym_char] = ACTIONS(2111),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2111),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(597)] = {
    [sym__expression] = STATE(3724),
    [sym_block] = STATE(3724),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3724),
    [sym_nil] = STATE(3724),
    [sym__atom] = STATE(3724),
    [sym_quoted_atom] = STATE(3724),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3724),
    [sym_charlist] = STATE(3724),
    [sym_sigil] = STATE(3724),
    [sym_list] = STATE(3724),
    [sym_tuple] = STATE(3724),
    [sym_bitstring] = STATE(3724),
    [sym_map] = STATE(3724),
    [sym__nullary_operator] = STATE(3724),
    [sym_unary_operator] = STATE(3724),
    [sym_binary_operator] = STATE(3724),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3724),
    [sym_call] = STATE(3724),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3724),
    [sym_anonymous_function] = STATE(3724),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2113),
    [sym_integer] = ACTIONS(2113),
    [sym_float] = ACTIONS(2113),
    [sym_char] = ACTIONS(2113),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(598)] = {
    [sym__expression] = STATE(3725),
    [sym_block] = STATE(3725),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3725),
    [sym_nil] = STATE(3725),
    [sym__atom] = STATE(3725),
    [sym_quoted_atom] = STATE(3725),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3725),
    [sym_charlist] = STATE(3725),
    [sym_sigil] = STATE(3725),
    [sym_list] = STATE(3725),
    [sym_tuple] = STATE(3725),
    [sym_bitstring] = STATE(3725),
    [sym_map] = STATE(3725),
    [sym__nullary_operator] = STATE(3725),
    [sym_unary_operator] = STATE(3725),
    [sym_binary_operator] = STATE(3725),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3725),
    [sym_call] = STATE(3725),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3725),
    [sym_anonymous_function] = STATE(3725),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2115),
    [sym_integer] = ACTIONS(2115),
    [sym_float] = ACTIONS(2115),
    [sym_char] = ACTIONS(2115),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2115),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(599)] = {
    [sym__expression] = STATE(3726),
    [sym_block] = STATE(3726),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3726),
    [sym_nil] = STATE(3726),
    [sym__atom] = STATE(3726),
    [sym_quoted_atom] = STATE(3726),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3726),
    [sym_charlist] = STATE(3726),
    [sym_sigil] = STATE(3726),
    [sym_list] = STATE(3726),
    [sym_tuple] = STATE(3726),
    [sym_bitstring] = STATE(3726),
    [sym_map] = STATE(3726),
    [sym__nullary_operator] = STATE(3726),
    [sym_unary_operator] = STATE(3726),
    [sym_binary_operator] = STATE(3726),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3726),
    [sym_call] = STATE(3726),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3726),
    [sym_anonymous_function] = STATE(3726),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2117),
    [sym_integer] = ACTIONS(2117),
    [sym_float] = ACTIONS(2117),
    [sym_char] = ACTIONS(2117),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2117),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(600)] = {
    [sym__expression] = STATE(3729),
    [sym_block] = STATE(3729),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3729),
    [sym_nil] = STATE(3729),
    [sym__atom] = STATE(3729),
    [sym_quoted_atom] = STATE(3729),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3729),
    [sym_charlist] = STATE(3729),
    [sym_sigil] = STATE(3729),
    [sym_list] = STATE(3729),
    [sym_tuple] = STATE(3729),
    [sym_bitstring] = STATE(3729),
    [sym_map] = STATE(3729),
    [sym__nullary_operator] = STATE(3729),
    [sym_unary_operator] = STATE(3729),
    [sym_binary_operator] = STATE(3729),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3729),
    [sym_call] = STATE(3729),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3729),
    [sym_anonymous_function] = STATE(3729),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2119),
    [sym_integer] = ACTIONS(2119),
    [sym_float] = ACTIONS(2119),
    [sym_char] = ACTIONS(2119),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2119),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(601)] = {
    [sym__expression] = STATE(3730),
    [sym_block] = STATE(3730),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3730),
    [sym_nil] = STATE(3730),
    [sym__atom] = STATE(3730),
    [sym_quoted_atom] = STATE(3730),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3730),
    [sym_charlist] = STATE(3730),
    [sym_sigil] = STATE(3730),
    [sym_list] = STATE(3730),
    [sym_tuple] = STATE(3730),
    [sym_bitstring] = STATE(3730),
    [sym_map] = STATE(3730),
    [sym__nullary_operator] = STATE(3730),
    [sym_unary_operator] = STATE(3730),
    [sym_binary_operator] = STATE(3730),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3730),
    [sym_call] = STATE(3730),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3730),
    [sym_anonymous_function] = STATE(3730),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2121),
    [sym_integer] = ACTIONS(2121),
    [sym_float] = ACTIONS(2121),
    [sym_char] = ACTIONS(2121),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2121),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(602)] = {
    [sym__expression] = STATE(3731),
    [sym_block] = STATE(3731),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3731),
    [sym_nil] = STATE(3731),
    [sym__atom] = STATE(3731),
    [sym_quoted_atom] = STATE(3731),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3731),
    [sym_charlist] = STATE(3731),
    [sym_sigil] = STATE(3731),
    [sym_list] = STATE(3731),
    [sym_tuple] = STATE(3731),
    [sym_bitstring] = STATE(3731),
    [sym_map] = STATE(3731),
    [sym__nullary_operator] = STATE(3731),
    [sym_unary_operator] = STATE(3731),
    [sym_binary_operator] = STATE(3731),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3731),
    [sym_call] = STATE(3731),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3731),
    [sym_anonymous_function] = STATE(3731),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2123),
    [sym_integer] = ACTIONS(2123),
    [sym_float] = ACTIONS(2123),
    [sym_char] = ACTIONS(2123),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2123),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(603)] = {
    [sym__expression] = STATE(3732),
    [sym_block] = STATE(3732),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3732),
    [sym_nil] = STATE(3732),
    [sym__atom] = STATE(3732),
    [sym_quoted_atom] = STATE(3732),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3732),
    [sym_charlist] = STATE(3732),
    [sym_sigil] = STATE(3732),
    [sym_list] = STATE(3732),
    [sym_tuple] = STATE(3732),
    [sym_bitstring] = STATE(3732),
    [sym_map] = STATE(3732),
    [sym__nullary_operator] = STATE(3732),
    [sym_unary_operator] = STATE(3732),
    [sym_binary_operator] = STATE(3732),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3732),
    [sym_call] = STATE(3732),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3732),
    [sym_anonymous_function] = STATE(3732),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2125),
    [sym_integer] = ACTIONS(2125),
    [sym_float] = ACTIONS(2125),
    [sym_char] = ACTIONS(2125),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2125),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(604)] = {
    [sym__expression] = STATE(3733),
    [sym_block] = STATE(3733),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3733),
    [sym_nil] = STATE(3733),
    [sym__atom] = STATE(3733),
    [sym_quoted_atom] = STATE(3733),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3733),
    [sym_charlist] = STATE(3733),
    [sym_sigil] = STATE(3733),
    [sym_list] = STATE(3733),
    [sym_tuple] = STATE(3733),
    [sym_bitstring] = STATE(3733),
    [sym_map] = STATE(3733),
    [sym__nullary_operator] = STATE(3733),
    [sym_unary_operator] = STATE(3733),
    [sym_binary_operator] = STATE(3733),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3733),
    [sym_call] = STATE(3733),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3733),
    [sym_anonymous_function] = STATE(3733),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2127),
    [sym_integer] = ACTIONS(2127),
    [sym_float] = ACTIONS(2127),
    [sym_char] = ACTIONS(2127),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2127),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(605)] = {
    [sym__expression] = STATE(3734),
    [sym_block] = STATE(3734),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3734),
    [sym_nil] = STATE(3734),
    [sym__atom] = STATE(3734),
    [sym_quoted_atom] = STATE(3734),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3734),
    [sym_charlist] = STATE(3734),
    [sym_sigil] = STATE(3734),
    [sym_list] = STATE(3734),
    [sym_tuple] = STATE(3734),
    [sym_bitstring] = STATE(3734),
    [sym_map] = STATE(3734),
    [sym__nullary_operator] = STATE(3734),
    [sym_unary_operator] = STATE(3734),
    [sym_binary_operator] = STATE(3734),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3734),
    [sym_call] = STATE(3734),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3734),
    [sym_anonymous_function] = STATE(3734),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2129),
    [sym_integer] = ACTIONS(2129),
    [sym_float] = ACTIONS(2129),
    [sym_char] = ACTIONS(2129),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2129),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(606)] = {
    [sym__expression] = STATE(3735),
    [sym_block] = STATE(3735),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3735),
    [sym_nil] = STATE(3735),
    [sym__atom] = STATE(3735),
    [sym_quoted_atom] = STATE(3735),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3735),
    [sym_charlist] = STATE(3735),
    [sym_sigil] = STATE(3735),
    [sym_list] = STATE(3735),
    [sym_tuple] = STATE(3735),
    [sym_bitstring] = STATE(3735),
    [sym_map] = STATE(3735),
    [sym__nullary_operator] = STATE(3735),
    [sym_unary_operator] = STATE(3735),
    [sym_binary_operator] = STATE(3735),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3735),
    [sym_call] = STATE(3735),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3735),
    [sym_anonymous_function] = STATE(3735),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2131),
    [sym_integer] = ACTIONS(2131),
    [sym_float] = ACTIONS(2131),
    [sym_char] = ACTIONS(2131),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2131),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(607)] = {
    [sym__expression] = STATE(3736),
    [sym_block] = STATE(3736),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3736),
    [sym_nil] = STATE(3736),
    [sym__atom] = STATE(3736),
    [sym_quoted_atom] = STATE(3736),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3736),
    [sym_charlist] = STATE(3736),
    [sym_sigil] = STATE(3736),
    [sym_list] = STATE(3736),
    [sym_tuple] = STATE(3736),
    [sym_bitstring] = STATE(3736),
    [sym_map] = STATE(3736),
    [sym__nullary_operator] = STATE(3736),
    [sym_unary_operator] = STATE(3736),
    [sym_binary_operator] = STATE(3736),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3736),
    [sym_call] = STATE(3736),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3736),
    [sym_anonymous_function] = STATE(3736),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2133),
    [sym_integer] = ACTIONS(2133),
    [sym_float] = ACTIONS(2133),
    [sym_char] = ACTIONS(2133),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2133),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(608)] = {
    [sym__expression] = STATE(3737),
    [sym_block] = STATE(3737),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3737),
    [sym_nil] = STATE(3737),
    [sym__atom] = STATE(3737),
    [sym_quoted_atom] = STATE(3737),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3737),
    [sym_charlist] = STATE(3737),
    [sym_sigil] = STATE(3737),
    [sym_list] = STATE(3737),
    [sym_tuple] = STATE(3737),
    [sym_bitstring] = STATE(3737),
    [sym_map] = STATE(3737),
    [sym__nullary_operator] = STATE(3737),
    [sym_unary_operator] = STATE(3737),
    [sym_binary_operator] = STATE(3737),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3737),
    [sym_call] = STATE(3737),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3737),
    [sym_anonymous_function] = STATE(3737),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2135),
    [sym_integer] = ACTIONS(2135),
    [sym_float] = ACTIONS(2135),
    [sym_char] = ACTIONS(2135),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2135),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(609)] = {
    [sym__expression] = STATE(3738),
    [sym_block] = STATE(3738),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3738),
    [sym_nil] = STATE(3738),
    [sym__atom] = STATE(3738),
    [sym_quoted_atom] = STATE(3738),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3738),
    [sym_charlist] = STATE(3738),
    [sym_sigil] = STATE(3738),
    [sym_list] = STATE(3738),
    [sym_tuple] = STATE(3738),
    [sym_bitstring] = STATE(3738),
    [sym_map] = STATE(3738),
    [sym__nullary_operator] = STATE(3738),
    [sym_unary_operator] = STATE(3738),
    [sym_binary_operator] = STATE(3738),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3738),
    [sym_call] = STATE(3738),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3738),
    [sym_anonymous_function] = STATE(3738),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2137),
    [sym_integer] = ACTIONS(2137),
    [sym_float] = ACTIONS(2137),
    [sym_char] = ACTIONS(2137),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2137),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(610)] = {
    [sym__expression] = STATE(4279),
    [sym_block] = STATE(4279),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4279),
    [sym_nil] = STATE(4279),
    [sym__atom] = STATE(4279),
    [sym_quoted_atom] = STATE(4279),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4279),
    [sym_charlist] = STATE(4279),
    [sym_sigil] = STATE(4279),
    [sym_list] = STATE(4279),
    [sym_tuple] = STATE(4279),
    [sym_bitstring] = STATE(4279),
    [sym_map] = STATE(4279),
    [sym__nullary_operator] = STATE(4279),
    [sym_unary_operator] = STATE(4279),
    [sym_binary_operator] = STATE(4279),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4279),
    [sym_call] = STATE(4279),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4279),
    [sym_anonymous_function] = STATE(4279),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2139),
    [sym_integer] = ACTIONS(2139),
    [sym_float] = ACTIONS(2139),
    [sym_char] = ACTIONS(2139),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2139),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(611)] = {
    [sym__expression] = STATE(4392),
    [sym_block] = STATE(4392),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4392),
    [sym_nil] = STATE(4392),
    [sym__atom] = STATE(4392),
    [sym_quoted_atom] = STATE(4392),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4392),
    [sym_charlist] = STATE(4392),
    [sym_sigil] = STATE(4392),
    [sym_list] = STATE(4392),
    [sym_tuple] = STATE(4392),
    [sym_bitstring] = STATE(4392),
    [sym_map] = STATE(4392),
    [sym__nullary_operator] = STATE(4392),
    [sym_unary_operator] = STATE(4392),
    [sym_binary_operator] = STATE(4392),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4392),
    [sym_call] = STATE(4392),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4392),
    [sym_anonymous_function] = STATE(4392),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2141),
    [sym_integer] = ACTIONS(2141),
    [sym_float] = ACTIONS(2141),
    [sym_char] = ACTIONS(2141),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2141),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(612)] = {
    [sym__expression] = STATE(1908),
    [sym_block] = STATE(1908),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(1908),
    [sym_nil] = STATE(1908),
    [sym__atom] = STATE(1908),
    [sym_quoted_atom] = STATE(1908),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(1908),
    [sym_charlist] = STATE(1908),
    [sym_sigil] = STATE(1908),
    [sym_list] = STATE(1908),
    [sym_tuple] = STATE(1908),
    [sym_bitstring] = STATE(1908),
    [sym_map] = STATE(1908),
    [sym__nullary_operator] = STATE(1908),
    [sym_unary_operator] = STATE(1908),
    [sym_binary_operator] = STATE(1908),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(1908),
    [sym_call] = STATE(1908),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(1908),
    [sym_anonymous_function] = STATE(1908),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1977),
    [sym_integer] = ACTIONS(1977),
    [sym_float] = ACTIONS(1977),
    [sym_char] = ACTIONS(1977),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1977),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(613)] = {
    [sym__expression] = STATE(4579),
    [sym_block] = STATE(4579),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4579),
    [sym_nil] = STATE(4579),
    [sym__atom] = STATE(4579),
    [sym_quoted_atom] = STATE(4579),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4579),
    [sym_charlist] = STATE(4579),
    [sym_sigil] = STATE(4579),
    [sym_list] = STATE(4579),
    [sym_tuple] = STATE(4579),
    [sym_bitstring] = STATE(4579),
    [sym_map] = STATE(4579),
    [sym__nullary_operator] = STATE(4579),
    [sym_unary_operator] = STATE(4579),
    [sym_binary_operator] = STATE(4579),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4579),
    [sym_call] = STATE(4579),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4579),
    [sym_anonymous_function] = STATE(4579),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2143),
    [sym_integer] = ACTIONS(2143),
    [sym_float] = ACTIONS(2143),
    [sym_char] = ACTIONS(2143),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(2143),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(614)] = {
    [sym__expression] = STATE(4393),
    [sym_block] = STATE(4393),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4393),
    [sym_nil] = STATE(4393),
    [sym__atom] = STATE(4393),
    [sym_quoted_atom] = STATE(4393),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4393),
    [sym_charlist] = STATE(4393),
    [sym_sigil] = STATE(4393),
    [sym_list] = STATE(4393),
    [sym_tuple] = STATE(4393),
    [sym_bitstring] = STATE(4393),
    [sym_map] = STATE(4393),
    [sym__nullary_operator] = STATE(4393),
    [sym_unary_operator] = STATE(4393),
    [sym_binary_operator] = STATE(4393),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4393),
    [sym_call] = STATE(4393),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4393),
    [sym_anonymous_function] = STATE(4393),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2145),
    [sym_integer] = ACTIONS(2145),
    [sym_float] = ACTIONS(2145),
    [sym_char] = ACTIONS(2145),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2145),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(615)] = {
    [sym__expression] = STATE(2022),
    [sym_block] = STATE(2022),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(2022),
    [sym_nil] = STATE(2022),
    [sym__atom] = STATE(2022),
    [sym_quoted_atom] = STATE(2022),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2022),
    [sym_charlist] = STATE(2022),
    [sym_sigil] = STATE(2022),
    [sym_list] = STATE(2022),
    [sym_tuple] = STATE(2022),
    [sym_bitstring] = STATE(2022),
    [sym_map] = STATE(2022),
    [sym__nullary_operator] = STATE(2022),
    [sym_unary_operator] = STATE(2022),
    [sym_binary_operator] = STATE(2022),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2022),
    [sym_call] = STATE(2022),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2022),
    [sym_anonymous_function] = STATE(2022),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1981),
    [sym_integer] = ACTIONS(1981),
    [sym_float] = ACTIONS(1981),
    [sym_char] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1981),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(616)] = {
    [sym__expression] = STATE(4299),
    [sym_block] = STATE(4299),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4299),
    [sym_nil] = STATE(4299),
    [sym__atom] = STATE(4299),
    [sym_quoted_atom] = STATE(4299),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4299),
    [sym_charlist] = STATE(4299),
    [sym_sigil] = STATE(4299),
    [sym_list] = STATE(4299),
    [sym_tuple] = STATE(4299),
    [sym_bitstring] = STATE(4299),
    [sym_map] = STATE(4299),
    [sym__nullary_operator] = STATE(4299),
    [sym_unary_operator] = STATE(4299),
    [sym_binary_operator] = STATE(4299),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4299),
    [sym_call] = STATE(4299),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4299),
    [sym_anonymous_function] = STATE(4299),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2147),
    [sym_integer] = ACTIONS(2147),
    [sym_float] = ACTIONS(2147),
    [sym_char] = ACTIONS(2147),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2147),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(617)] = {
    [sym__expression] = STATE(4300),
    [sym_block] = STATE(4300),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4300),
    [sym_nil] = STATE(4300),
    [sym__atom] = STATE(4300),
    [sym_quoted_atom] = STATE(4300),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4300),
    [sym_charlist] = STATE(4300),
    [sym_sigil] = STATE(4300),
    [sym_list] = STATE(4300),
    [sym_tuple] = STATE(4300),
    [sym_bitstring] = STATE(4300),
    [sym_map] = STATE(4300),
    [sym__nullary_operator] = STATE(4300),
    [sym_unary_operator] = STATE(4300),
    [sym_binary_operator] = STATE(4300),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4300),
    [sym_call] = STATE(4300),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4300),
    [sym_anonymous_function] = STATE(4300),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2149),
    [sym_integer] = ACTIONS(2149),
    [sym_float] = ACTIONS(2149),
    [sym_char] = ACTIONS(2149),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2149),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(618)] = {
    [sym__expression] = STATE(4303),
    [sym_block] = STATE(4303),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4303),
    [sym_nil] = STATE(4303),
    [sym__atom] = STATE(4303),
    [sym_quoted_atom] = STATE(4303),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4303),
    [sym_charlist] = STATE(4303),
    [sym_sigil] = STATE(4303),
    [sym_list] = STATE(4303),
    [sym_tuple] = STATE(4303),
    [sym_bitstring] = STATE(4303),
    [sym_map] = STATE(4303),
    [sym__nullary_operator] = STATE(4303),
    [sym_unary_operator] = STATE(4303),
    [sym_binary_operator] = STATE(4303),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4303),
    [sym_call] = STATE(4303),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4303),
    [sym_anonymous_function] = STATE(4303),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2151),
    [sym_integer] = ACTIONS(2151),
    [sym_float] = ACTIONS(2151),
    [sym_char] = ACTIONS(2151),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2151),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(619)] = {
    [sym__expression] = STATE(4304),
    [sym_block] = STATE(4304),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4304),
    [sym_nil] = STATE(4304),
    [sym__atom] = STATE(4304),
    [sym_quoted_atom] = STATE(4304),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4304),
    [sym_charlist] = STATE(4304),
    [sym_sigil] = STATE(4304),
    [sym_list] = STATE(4304),
    [sym_tuple] = STATE(4304),
    [sym_bitstring] = STATE(4304),
    [sym_map] = STATE(4304),
    [sym__nullary_operator] = STATE(4304),
    [sym_unary_operator] = STATE(4304),
    [sym_binary_operator] = STATE(4304),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4304),
    [sym_call] = STATE(4304),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4304),
    [sym_anonymous_function] = STATE(4304),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2153),
    [sym_integer] = ACTIONS(2153),
    [sym_float] = ACTIONS(2153),
    [sym_char] = ACTIONS(2153),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2153),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(620)] = {
    [sym__expression] = STATE(4305),
    [sym_block] = STATE(4305),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4305),
    [sym_nil] = STATE(4305),
    [sym__atom] = STATE(4305),
    [sym_quoted_atom] = STATE(4305),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4305),
    [sym_charlist] = STATE(4305),
    [sym_sigil] = STATE(4305),
    [sym_list] = STATE(4305),
    [sym_tuple] = STATE(4305),
    [sym_bitstring] = STATE(4305),
    [sym_map] = STATE(4305),
    [sym__nullary_operator] = STATE(4305),
    [sym_unary_operator] = STATE(4305),
    [sym_binary_operator] = STATE(4305),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4305),
    [sym_call] = STATE(4305),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4305),
    [sym_anonymous_function] = STATE(4305),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2155),
    [sym_integer] = ACTIONS(2155),
    [sym_float] = ACTIONS(2155),
    [sym_char] = ACTIONS(2155),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2155),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(621)] = {
    [sym__expression] = STATE(4306),
    [sym_block] = STATE(4306),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4306),
    [sym_nil] = STATE(4306),
    [sym__atom] = STATE(4306),
    [sym_quoted_atom] = STATE(4306),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4306),
    [sym_charlist] = STATE(4306),
    [sym_sigil] = STATE(4306),
    [sym_list] = STATE(4306),
    [sym_tuple] = STATE(4306),
    [sym_bitstring] = STATE(4306),
    [sym_map] = STATE(4306),
    [sym__nullary_operator] = STATE(4306),
    [sym_unary_operator] = STATE(4306),
    [sym_binary_operator] = STATE(4306),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4306),
    [sym_call] = STATE(4306),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4306),
    [sym_anonymous_function] = STATE(4306),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2157),
    [sym_integer] = ACTIONS(2157),
    [sym_float] = ACTIONS(2157),
    [sym_char] = ACTIONS(2157),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2157),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(622)] = {
    [sym__expression] = STATE(4309),
    [sym_block] = STATE(4309),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4309),
    [sym_nil] = STATE(4309),
    [sym__atom] = STATE(4309),
    [sym_quoted_atom] = STATE(4309),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4309),
    [sym_charlist] = STATE(4309),
    [sym_sigil] = STATE(4309),
    [sym_list] = STATE(4309),
    [sym_tuple] = STATE(4309),
    [sym_bitstring] = STATE(4309),
    [sym_map] = STATE(4309),
    [sym__nullary_operator] = STATE(4309),
    [sym_unary_operator] = STATE(4309),
    [sym_binary_operator] = STATE(4309),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4309),
    [sym_call] = STATE(4309),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4309),
    [sym_anonymous_function] = STATE(4309),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2159),
    [sym_integer] = ACTIONS(2159),
    [sym_float] = ACTIONS(2159),
    [sym_char] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(623)] = {
    [sym__expression] = STATE(4310),
    [sym_block] = STATE(4310),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4310),
    [sym_nil] = STATE(4310),
    [sym__atom] = STATE(4310),
    [sym_quoted_atom] = STATE(4310),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4310),
    [sym_charlist] = STATE(4310),
    [sym_sigil] = STATE(4310),
    [sym_list] = STATE(4310),
    [sym_tuple] = STATE(4310),
    [sym_bitstring] = STATE(4310),
    [sym_map] = STATE(4310),
    [sym__nullary_operator] = STATE(4310),
    [sym_unary_operator] = STATE(4310),
    [sym_binary_operator] = STATE(4310),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4310),
    [sym_call] = STATE(4310),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4310),
    [sym_anonymous_function] = STATE(4310),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2161),
    [sym_integer] = ACTIONS(2161),
    [sym_float] = ACTIONS(2161),
    [sym_char] = ACTIONS(2161),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2161),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(624)] = {
    [sym__expression] = STATE(4311),
    [sym_block] = STATE(4311),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4311),
    [sym_nil] = STATE(4311),
    [sym__atom] = STATE(4311),
    [sym_quoted_atom] = STATE(4311),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4311),
    [sym_charlist] = STATE(4311),
    [sym_sigil] = STATE(4311),
    [sym_list] = STATE(4311),
    [sym_tuple] = STATE(4311),
    [sym_bitstring] = STATE(4311),
    [sym_map] = STATE(4311),
    [sym__nullary_operator] = STATE(4311),
    [sym_unary_operator] = STATE(4311),
    [sym_binary_operator] = STATE(4311),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4311),
    [sym_call] = STATE(4311),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4311),
    [sym_anonymous_function] = STATE(4311),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2163),
    [sym_integer] = ACTIONS(2163),
    [sym_float] = ACTIONS(2163),
    [sym_char] = ACTIONS(2163),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2163),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(625)] = {
    [sym__expression] = STATE(4312),
    [sym_block] = STATE(4312),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4312),
    [sym_nil] = STATE(4312),
    [sym__atom] = STATE(4312),
    [sym_quoted_atom] = STATE(4312),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4312),
    [sym_charlist] = STATE(4312),
    [sym_sigil] = STATE(4312),
    [sym_list] = STATE(4312),
    [sym_tuple] = STATE(4312),
    [sym_bitstring] = STATE(4312),
    [sym_map] = STATE(4312),
    [sym__nullary_operator] = STATE(4312),
    [sym_unary_operator] = STATE(4312),
    [sym_binary_operator] = STATE(4312),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4312),
    [sym_call] = STATE(4312),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4312),
    [sym_anonymous_function] = STATE(4312),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2165),
    [sym_integer] = ACTIONS(2165),
    [sym_float] = ACTIONS(2165),
    [sym_char] = ACTIONS(2165),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2165),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(626)] = {
    [sym__expression] = STATE(4313),
    [sym_block] = STATE(4313),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4313),
    [sym_nil] = STATE(4313),
    [sym__atom] = STATE(4313),
    [sym_quoted_atom] = STATE(4313),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4313),
    [sym_charlist] = STATE(4313),
    [sym_sigil] = STATE(4313),
    [sym_list] = STATE(4313),
    [sym_tuple] = STATE(4313),
    [sym_bitstring] = STATE(4313),
    [sym_map] = STATE(4313),
    [sym__nullary_operator] = STATE(4313),
    [sym_unary_operator] = STATE(4313),
    [sym_binary_operator] = STATE(4313),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4313),
    [sym_call] = STATE(4313),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4313),
    [sym_anonymous_function] = STATE(4313),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2167),
    [sym_integer] = ACTIONS(2167),
    [sym_float] = ACTIONS(2167),
    [sym_char] = ACTIONS(2167),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2167),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(627)] = {
    [sym__expression] = STATE(4314),
    [sym_block] = STATE(4314),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4314),
    [sym_nil] = STATE(4314),
    [sym__atom] = STATE(4314),
    [sym_quoted_atom] = STATE(4314),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4314),
    [sym_charlist] = STATE(4314),
    [sym_sigil] = STATE(4314),
    [sym_list] = STATE(4314),
    [sym_tuple] = STATE(4314),
    [sym_bitstring] = STATE(4314),
    [sym_map] = STATE(4314),
    [sym__nullary_operator] = STATE(4314),
    [sym_unary_operator] = STATE(4314),
    [sym_binary_operator] = STATE(4314),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4314),
    [sym_call] = STATE(4314),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4314),
    [sym_anonymous_function] = STATE(4314),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2169),
    [sym_integer] = ACTIONS(2169),
    [sym_float] = ACTIONS(2169),
    [sym_char] = ACTIONS(2169),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2169),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(628)] = {
    [sym__expression] = STATE(4315),
    [sym_block] = STATE(4315),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4315),
    [sym_nil] = STATE(4315),
    [sym__atom] = STATE(4315),
    [sym_quoted_atom] = STATE(4315),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4315),
    [sym_charlist] = STATE(4315),
    [sym_sigil] = STATE(4315),
    [sym_list] = STATE(4315),
    [sym_tuple] = STATE(4315),
    [sym_bitstring] = STATE(4315),
    [sym_map] = STATE(4315),
    [sym__nullary_operator] = STATE(4315),
    [sym_unary_operator] = STATE(4315),
    [sym_binary_operator] = STATE(4315),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4315),
    [sym_call] = STATE(4315),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4315),
    [sym_anonymous_function] = STATE(4315),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2171),
    [sym_integer] = ACTIONS(2171),
    [sym_float] = ACTIONS(2171),
    [sym_char] = ACTIONS(2171),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2171),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(629)] = {
    [sym__expression] = STATE(4316),
    [sym_block] = STATE(4316),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4316),
    [sym_nil] = STATE(4316),
    [sym__atom] = STATE(4316),
    [sym_quoted_atom] = STATE(4316),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4316),
    [sym_charlist] = STATE(4316),
    [sym_sigil] = STATE(4316),
    [sym_list] = STATE(4316),
    [sym_tuple] = STATE(4316),
    [sym_bitstring] = STATE(4316),
    [sym_map] = STATE(4316),
    [sym__nullary_operator] = STATE(4316),
    [sym_unary_operator] = STATE(4316),
    [sym_binary_operator] = STATE(4316),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4316),
    [sym_call] = STATE(4316),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4316),
    [sym_anonymous_function] = STATE(4316),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2173),
    [sym_integer] = ACTIONS(2173),
    [sym_float] = ACTIONS(2173),
    [sym_char] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2173),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(630)] = {
    [sym__expression] = STATE(4317),
    [sym_block] = STATE(4317),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4317),
    [sym_nil] = STATE(4317),
    [sym__atom] = STATE(4317),
    [sym_quoted_atom] = STATE(4317),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4317),
    [sym_charlist] = STATE(4317),
    [sym_sigil] = STATE(4317),
    [sym_list] = STATE(4317),
    [sym_tuple] = STATE(4317),
    [sym_bitstring] = STATE(4317),
    [sym_map] = STATE(4317),
    [sym__nullary_operator] = STATE(4317),
    [sym_unary_operator] = STATE(4317),
    [sym_binary_operator] = STATE(4317),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4317),
    [sym_call] = STATE(4317),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4317),
    [sym_anonymous_function] = STATE(4317),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2175),
    [sym_integer] = ACTIONS(2175),
    [sym_float] = ACTIONS(2175),
    [sym_char] = ACTIONS(2175),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2175),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(631)] = {
    [sym__expression] = STATE(4394),
    [sym_block] = STATE(4394),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4394),
    [sym_nil] = STATE(4394),
    [sym__atom] = STATE(4394),
    [sym_quoted_atom] = STATE(4394),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4394),
    [sym_charlist] = STATE(4394),
    [sym_sigil] = STATE(4394),
    [sym_list] = STATE(4394),
    [sym_tuple] = STATE(4394),
    [sym_bitstring] = STATE(4394),
    [sym_map] = STATE(4394),
    [sym__nullary_operator] = STATE(4394),
    [sym_unary_operator] = STATE(4394),
    [sym_binary_operator] = STATE(4394),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4394),
    [sym_call] = STATE(4394),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4394),
    [sym_anonymous_function] = STATE(4394),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2177),
    [sym_integer] = ACTIONS(2177),
    [sym_float] = ACTIONS(2177),
    [sym_char] = ACTIONS(2177),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2177),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(632)] = {
    [sym__expression] = STATE(3374),
    [sym_block] = STATE(3374),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3374),
    [sym_nil] = STATE(3374),
    [sym__atom] = STATE(3374),
    [sym_quoted_atom] = STATE(3374),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3374),
    [sym_charlist] = STATE(3374),
    [sym_sigil] = STATE(3374),
    [sym_list] = STATE(3374),
    [sym_tuple] = STATE(3374),
    [sym_bitstring] = STATE(3374),
    [sym_map] = STATE(3374),
    [sym__nullary_operator] = STATE(3374),
    [sym_unary_operator] = STATE(3374),
    [sym_binary_operator] = STATE(3374),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3374),
    [sym_call] = STATE(3374),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3374),
    [sym_anonymous_function] = STATE(3374),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2179),
    [sym_integer] = ACTIONS(2179),
    [sym_float] = ACTIONS(2179),
    [sym_char] = ACTIONS(2179),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2179),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(633)] = {
    [sym__expression] = STATE(3375),
    [sym_block] = STATE(3375),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3375),
    [sym_nil] = STATE(3375),
    [sym__atom] = STATE(3375),
    [sym_quoted_atom] = STATE(3375),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3375),
    [sym_charlist] = STATE(3375),
    [sym_sigil] = STATE(3375),
    [sym_list] = STATE(3375),
    [sym_tuple] = STATE(3375),
    [sym_bitstring] = STATE(3375),
    [sym_map] = STATE(3375),
    [sym__nullary_operator] = STATE(3375),
    [sym_unary_operator] = STATE(3375),
    [sym_binary_operator] = STATE(3375),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3375),
    [sym_call] = STATE(3375),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3375),
    [sym_anonymous_function] = STATE(3375),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2181),
    [sym_integer] = ACTIONS(2181),
    [sym_float] = ACTIONS(2181),
    [sym_char] = ACTIONS(2181),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2181),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(634)] = {
    [sym__expression] = STATE(4580),
    [sym_block] = STATE(4580),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4580),
    [sym_nil] = STATE(4580),
    [sym__atom] = STATE(4580),
    [sym_quoted_atom] = STATE(4580),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4580),
    [sym_charlist] = STATE(4580),
    [sym_sigil] = STATE(4580),
    [sym_list] = STATE(4580),
    [sym_tuple] = STATE(4580),
    [sym_bitstring] = STATE(4580),
    [sym_map] = STATE(4580),
    [sym__nullary_operator] = STATE(4580),
    [sym_unary_operator] = STATE(4580),
    [sym_binary_operator] = STATE(4580),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4580),
    [sym_call] = STATE(4580),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4580),
    [sym_anonymous_function] = STATE(4580),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2183),
    [sym_integer] = ACTIONS(2183),
    [sym_float] = ACTIONS(2183),
    [sym_char] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(2183),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(635)] = {
    [sym__expression] = STATE(4031),
    [sym_block] = STATE(4031),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4031),
    [sym_nil] = STATE(4031),
    [sym__atom] = STATE(4031),
    [sym_quoted_atom] = STATE(4031),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4031),
    [sym_charlist] = STATE(4031),
    [sym_sigil] = STATE(4031),
    [sym_list] = STATE(4031),
    [sym_tuple] = STATE(4031),
    [sym_bitstring] = STATE(4031),
    [sym_map] = STATE(4031),
    [sym__nullary_operator] = STATE(4031),
    [sym_unary_operator] = STATE(4031),
    [sym_binary_operator] = STATE(4031),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4031),
    [sym_call] = STATE(4031),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4031),
    [sym_anonymous_function] = STATE(4031),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2185),
    [sym_integer] = ACTIONS(2185),
    [sym_float] = ACTIONS(2185),
    [sym_char] = ACTIONS(2185),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2185),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(636)] = {
    [sym__expression] = STATE(1588),
    [sym_block] = STATE(1588),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1588),
    [sym_nil] = STATE(1588),
    [sym__atom] = STATE(1588),
    [sym_quoted_atom] = STATE(1588),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1588),
    [sym_charlist] = STATE(1588),
    [sym_sigil] = STATE(1588),
    [sym_list] = STATE(1588),
    [sym_tuple] = STATE(1588),
    [sym_bitstring] = STATE(1588),
    [sym_map] = STATE(1588),
    [sym__nullary_operator] = STATE(1588),
    [sym_unary_operator] = STATE(1588),
    [sym_binary_operator] = STATE(1588),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1588),
    [sym_call] = STATE(1588),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1588),
    [sym_anonymous_function] = STATE(1588),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2187),
    [sym_integer] = ACTIONS(2187),
    [sym_float] = ACTIONS(2187),
    [sym_char] = ACTIONS(2187),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2187),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(637)] = {
    [sym__expression] = STATE(3386),
    [sym_block] = STATE(3386),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3386),
    [sym_nil] = STATE(3386),
    [sym__atom] = STATE(3386),
    [sym_quoted_atom] = STATE(3386),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3386),
    [sym_charlist] = STATE(3386),
    [sym_sigil] = STATE(3386),
    [sym_list] = STATE(3386),
    [sym_tuple] = STATE(3386),
    [sym_bitstring] = STATE(3386),
    [sym_map] = STATE(3386),
    [sym__nullary_operator] = STATE(3386),
    [sym_unary_operator] = STATE(3386),
    [sym_binary_operator] = STATE(3386),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3386),
    [sym_call] = STATE(3386),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3386),
    [sym_anonymous_function] = STATE(3386),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2189),
    [sym_integer] = ACTIONS(2189),
    [sym_float] = ACTIONS(2189),
    [sym_char] = ACTIONS(2189),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2189),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(638)] = {
    [sym__expression] = STATE(3387),
    [sym_block] = STATE(3387),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3387),
    [sym_nil] = STATE(3387),
    [sym__atom] = STATE(3387),
    [sym_quoted_atom] = STATE(3387),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3387),
    [sym_charlist] = STATE(3387),
    [sym_sigil] = STATE(3387),
    [sym_list] = STATE(3387),
    [sym_tuple] = STATE(3387),
    [sym_bitstring] = STATE(3387),
    [sym_map] = STATE(3387),
    [sym__nullary_operator] = STATE(3387),
    [sym_unary_operator] = STATE(3387),
    [sym_binary_operator] = STATE(3387),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3387),
    [sym_call] = STATE(3387),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3387),
    [sym_anonymous_function] = STATE(3387),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2191),
    [sym_integer] = ACTIONS(2191),
    [sym_float] = ACTIONS(2191),
    [sym_char] = ACTIONS(2191),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2191),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(639)] = {
    [sym__expression] = STATE(2961),
    [sym_block] = STATE(2961),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2961),
    [sym_nil] = STATE(2961),
    [sym__atom] = STATE(2961),
    [sym_quoted_atom] = STATE(2961),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2961),
    [sym_charlist] = STATE(2961),
    [sym_sigil] = STATE(2961),
    [sym_list] = STATE(2961),
    [sym_tuple] = STATE(2961),
    [sym_bitstring] = STATE(2961),
    [sym_map] = STATE(2961),
    [sym__nullary_operator] = STATE(2961),
    [sym_unary_operator] = STATE(2961),
    [sym_binary_operator] = STATE(2961),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2961),
    [sym_call] = STATE(2961),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2961),
    [sym_anonymous_function] = STATE(2961),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2193),
    [sym_integer] = ACTIONS(2193),
    [sym_float] = ACTIONS(2193),
    [sym_char] = ACTIONS(2193),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2193),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(640)] = {
    [sym__expression] = STATE(2962),
    [sym_block] = STATE(2962),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2962),
    [sym_nil] = STATE(2962),
    [sym__atom] = STATE(2962),
    [sym_quoted_atom] = STATE(2962),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2962),
    [sym_charlist] = STATE(2962),
    [sym_sigil] = STATE(2962),
    [sym_list] = STATE(2962),
    [sym_tuple] = STATE(2962),
    [sym_bitstring] = STATE(2962),
    [sym_map] = STATE(2962),
    [sym__nullary_operator] = STATE(2962),
    [sym_unary_operator] = STATE(2962),
    [sym_binary_operator] = STATE(2962),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2962),
    [sym_call] = STATE(2962),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2962),
    [sym_anonymous_function] = STATE(2962),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2195),
    [sym_integer] = ACTIONS(2195),
    [sym_float] = ACTIONS(2195),
    [sym_char] = ACTIONS(2195),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2195),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(641)] = {
    [sym__expression] = STATE(4124),
    [sym_block] = STATE(4124),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4124),
    [sym_nil] = STATE(4124),
    [sym__atom] = STATE(4124),
    [sym_quoted_atom] = STATE(4124),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4124),
    [sym_charlist] = STATE(4124),
    [sym_sigil] = STATE(4124),
    [sym_list] = STATE(4124),
    [sym_tuple] = STATE(4124),
    [sym_bitstring] = STATE(4124),
    [sym_map] = STATE(4124),
    [sym__nullary_operator] = STATE(4124),
    [sym_unary_operator] = STATE(4124),
    [sym_binary_operator] = STATE(4124),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4124),
    [sym_call] = STATE(4124),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4124),
    [sym_anonymous_function] = STATE(4124),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2197),
    [sym_integer] = ACTIONS(2197),
    [sym_float] = ACTIONS(2197),
    [sym_char] = ACTIONS(2197),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2197),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(642)] = {
    [sym__expression] = STATE(2965),
    [sym_block] = STATE(2965),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2965),
    [sym_nil] = STATE(2965),
    [sym__atom] = STATE(2965),
    [sym_quoted_atom] = STATE(2965),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2965),
    [sym_charlist] = STATE(2965),
    [sym_sigil] = STATE(2965),
    [sym_list] = STATE(2965),
    [sym_tuple] = STATE(2965),
    [sym_bitstring] = STATE(2965),
    [sym_map] = STATE(2965),
    [sym__nullary_operator] = STATE(2965),
    [sym_unary_operator] = STATE(2965),
    [sym_binary_operator] = STATE(2965),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2965),
    [sym_call] = STATE(2965),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2965),
    [sym_anonymous_function] = STATE(2965),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2199),
    [sym_integer] = ACTIONS(2199),
    [sym_float] = ACTIONS(2199),
    [sym_char] = ACTIONS(2199),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2199),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(643)] = {
    [sym__expression] = STATE(2966),
    [sym_block] = STATE(2966),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2966),
    [sym_nil] = STATE(2966),
    [sym__atom] = STATE(2966),
    [sym_quoted_atom] = STATE(2966),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2966),
    [sym_charlist] = STATE(2966),
    [sym_sigil] = STATE(2966),
    [sym_list] = STATE(2966),
    [sym_tuple] = STATE(2966),
    [sym_bitstring] = STATE(2966),
    [sym_map] = STATE(2966),
    [sym__nullary_operator] = STATE(2966),
    [sym_unary_operator] = STATE(2966),
    [sym_binary_operator] = STATE(2966),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2966),
    [sym_call] = STATE(2966),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2966),
    [sym_anonymous_function] = STATE(2966),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2201),
    [sym_integer] = ACTIONS(2201),
    [sym_float] = ACTIONS(2201),
    [sym_char] = ACTIONS(2201),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2201),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(644)] = {
    [sym__expression] = STATE(2967),
    [sym_block] = STATE(2967),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2967),
    [sym_nil] = STATE(2967),
    [sym__atom] = STATE(2967),
    [sym_quoted_atom] = STATE(2967),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2967),
    [sym_charlist] = STATE(2967),
    [sym_sigil] = STATE(2967),
    [sym_list] = STATE(2967),
    [sym_tuple] = STATE(2967),
    [sym_bitstring] = STATE(2967),
    [sym_map] = STATE(2967),
    [sym__nullary_operator] = STATE(2967),
    [sym_unary_operator] = STATE(2967),
    [sym_binary_operator] = STATE(2967),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2967),
    [sym_call] = STATE(2967),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2967),
    [sym_anonymous_function] = STATE(2967),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2203),
    [sym_integer] = ACTIONS(2203),
    [sym_float] = ACTIONS(2203),
    [sym_char] = ACTIONS(2203),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2203),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(645)] = {
    [sym__expression] = STATE(2968),
    [sym_block] = STATE(2968),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2968),
    [sym_nil] = STATE(2968),
    [sym__atom] = STATE(2968),
    [sym_quoted_atom] = STATE(2968),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2968),
    [sym_charlist] = STATE(2968),
    [sym_sigil] = STATE(2968),
    [sym_list] = STATE(2968),
    [sym_tuple] = STATE(2968),
    [sym_bitstring] = STATE(2968),
    [sym_map] = STATE(2968),
    [sym__nullary_operator] = STATE(2968),
    [sym_unary_operator] = STATE(2968),
    [sym_binary_operator] = STATE(2968),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2968),
    [sym_call] = STATE(2968),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2968),
    [sym_anonymous_function] = STATE(2968),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2205),
    [sym_integer] = ACTIONS(2205),
    [sym_float] = ACTIONS(2205),
    [sym_char] = ACTIONS(2205),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2205),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(646)] = {
    [sym__expression] = STATE(2971),
    [sym_block] = STATE(2971),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2971),
    [sym_nil] = STATE(2971),
    [sym__atom] = STATE(2971),
    [sym_quoted_atom] = STATE(2971),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2971),
    [sym_charlist] = STATE(2971),
    [sym_sigil] = STATE(2971),
    [sym_list] = STATE(2971),
    [sym_tuple] = STATE(2971),
    [sym_bitstring] = STATE(2971),
    [sym_map] = STATE(2971),
    [sym__nullary_operator] = STATE(2971),
    [sym_unary_operator] = STATE(2971),
    [sym_binary_operator] = STATE(2971),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2971),
    [sym_call] = STATE(2971),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2971),
    [sym_anonymous_function] = STATE(2971),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2207),
    [sym_integer] = ACTIONS(2207),
    [sym_float] = ACTIONS(2207),
    [sym_char] = ACTIONS(2207),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2207),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(647)] = {
    [sym__expression] = STATE(2972),
    [sym_block] = STATE(2972),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2972),
    [sym_nil] = STATE(2972),
    [sym__atom] = STATE(2972),
    [sym_quoted_atom] = STATE(2972),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2972),
    [sym_charlist] = STATE(2972),
    [sym_sigil] = STATE(2972),
    [sym_list] = STATE(2972),
    [sym_tuple] = STATE(2972),
    [sym_bitstring] = STATE(2972),
    [sym_map] = STATE(2972),
    [sym__nullary_operator] = STATE(2972),
    [sym_unary_operator] = STATE(2972),
    [sym_binary_operator] = STATE(2972),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2972),
    [sym_call] = STATE(2972),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2972),
    [sym_anonymous_function] = STATE(2972),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2209),
    [sym_integer] = ACTIONS(2209),
    [sym_float] = ACTIONS(2209),
    [sym_char] = ACTIONS(2209),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2209),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(648)] = {
    [sym__expression] = STATE(2973),
    [sym_block] = STATE(2973),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2973),
    [sym_nil] = STATE(2973),
    [sym__atom] = STATE(2973),
    [sym_quoted_atom] = STATE(2973),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2973),
    [sym_charlist] = STATE(2973),
    [sym_sigil] = STATE(2973),
    [sym_list] = STATE(2973),
    [sym_tuple] = STATE(2973),
    [sym_bitstring] = STATE(2973),
    [sym_map] = STATE(2973),
    [sym__nullary_operator] = STATE(2973),
    [sym_unary_operator] = STATE(2973),
    [sym_binary_operator] = STATE(2973),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2973),
    [sym_call] = STATE(2973),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2973),
    [sym_anonymous_function] = STATE(2973),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2211),
    [sym_integer] = ACTIONS(2211),
    [sym_float] = ACTIONS(2211),
    [sym_char] = ACTIONS(2211),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2211),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(649)] = {
    [sym__expression] = STATE(2974),
    [sym_block] = STATE(2974),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2974),
    [sym_nil] = STATE(2974),
    [sym__atom] = STATE(2974),
    [sym_quoted_atom] = STATE(2974),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2974),
    [sym_charlist] = STATE(2974),
    [sym_sigil] = STATE(2974),
    [sym_list] = STATE(2974),
    [sym_tuple] = STATE(2974),
    [sym_bitstring] = STATE(2974),
    [sym_map] = STATE(2974),
    [sym__nullary_operator] = STATE(2974),
    [sym_unary_operator] = STATE(2974),
    [sym_binary_operator] = STATE(2974),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2974),
    [sym_call] = STATE(2974),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2974),
    [sym_anonymous_function] = STATE(2974),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2213),
    [sym_integer] = ACTIONS(2213),
    [sym_float] = ACTIONS(2213),
    [sym_char] = ACTIONS(2213),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2213),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(650)] = {
    [sym__expression] = STATE(2975),
    [sym_block] = STATE(2975),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2975),
    [sym_nil] = STATE(2975),
    [sym__atom] = STATE(2975),
    [sym_quoted_atom] = STATE(2975),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2975),
    [sym_charlist] = STATE(2975),
    [sym_sigil] = STATE(2975),
    [sym_list] = STATE(2975),
    [sym_tuple] = STATE(2975),
    [sym_bitstring] = STATE(2975),
    [sym_map] = STATE(2975),
    [sym__nullary_operator] = STATE(2975),
    [sym_unary_operator] = STATE(2975),
    [sym_binary_operator] = STATE(2975),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2975),
    [sym_call] = STATE(2975),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2975),
    [sym_anonymous_function] = STATE(2975),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2215),
    [sym_integer] = ACTIONS(2215),
    [sym_float] = ACTIONS(2215),
    [sym_char] = ACTIONS(2215),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2215),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(651)] = {
    [sym__expression] = STATE(2976),
    [sym_block] = STATE(2976),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2976),
    [sym_nil] = STATE(2976),
    [sym__atom] = STATE(2976),
    [sym_quoted_atom] = STATE(2976),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2976),
    [sym_charlist] = STATE(2976),
    [sym_sigil] = STATE(2976),
    [sym_list] = STATE(2976),
    [sym_tuple] = STATE(2976),
    [sym_bitstring] = STATE(2976),
    [sym_map] = STATE(2976),
    [sym__nullary_operator] = STATE(2976),
    [sym_unary_operator] = STATE(2976),
    [sym_binary_operator] = STATE(2976),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2976),
    [sym_call] = STATE(2976),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2976),
    [sym_anonymous_function] = STATE(2976),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2217),
    [sym_integer] = ACTIONS(2217),
    [sym_float] = ACTIONS(2217),
    [sym_char] = ACTIONS(2217),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2217),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(652)] = {
    [sym__expression] = STATE(2977),
    [sym_block] = STATE(2977),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2977),
    [sym_nil] = STATE(2977),
    [sym__atom] = STATE(2977),
    [sym_quoted_atom] = STATE(2977),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2977),
    [sym_charlist] = STATE(2977),
    [sym_sigil] = STATE(2977),
    [sym_list] = STATE(2977),
    [sym_tuple] = STATE(2977),
    [sym_bitstring] = STATE(2977),
    [sym_map] = STATE(2977),
    [sym__nullary_operator] = STATE(2977),
    [sym_unary_operator] = STATE(2977),
    [sym_binary_operator] = STATE(2977),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2977),
    [sym_call] = STATE(2977),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2977),
    [sym_anonymous_function] = STATE(2977),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2219),
    [sym_integer] = ACTIONS(2219),
    [sym_float] = ACTIONS(2219),
    [sym_char] = ACTIONS(2219),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2219),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(653)] = {
    [sym__expression] = STATE(2978),
    [sym_block] = STATE(2978),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2978),
    [sym_nil] = STATE(2978),
    [sym__atom] = STATE(2978),
    [sym_quoted_atom] = STATE(2978),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2978),
    [sym_charlist] = STATE(2978),
    [sym_sigil] = STATE(2978),
    [sym_list] = STATE(2978),
    [sym_tuple] = STATE(2978),
    [sym_bitstring] = STATE(2978),
    [sym_map] = STATE(2978),
    [sym__nullary_operator] = STATE(2978),
    [sym_unary_operator] = STATE(2978),
    [sym_binary_operator] = STATE(2978),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2978),
    [sym_call] = STATE(2978),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2978),
    [sym_anonymous_function] = STATE(2978),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2221),
    [sym_integer] = ACTIONS(2221),
    [sym_float] = ACTIONS(2221),
    [sym_char] = ACTIONS(2221),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2221),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(654)] = {
    [sym__expression] = STATE(2979),
    [sym_block] = STATE(2979),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(2979),
    [sym_nil] = STATE(2979),
    [sym__atom] = STATE(2979),
    [sym_quoted_atom] = STATE(2979),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(2979),
    [sym_charlist] = STATE(2979),
    [sym_sigil] = STATE(2979),
    [sym_list] = STATE(2979),
    [sym_tuple] = STATE(2979),
    [sym_bitstring] = STATE(2979),
    [sym_map] = STATE(2979),
    [sym__nullary_operator] = STATE(2979),
    [sym_unary_operator] = STATE(2979),
    [sym_binary_operator] = STATE(2979),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(2979),
    [sym_call] = STATE(2979),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(2979),
    [sym_anonymous_function] = STATE(2979),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2223),
    [sym_integer] = ACTIONS(2223),
    [sym_float] = ACTIONS(2223),
    [sym_char] = ACTIONS(2223),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2223),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(655)] = {
    [sym__expression] = STATE(3398),
    [sym_block] = STATE(3398),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3398),
    [sym_nil] = STATE(3398),
    [sym__atom] = STATE(3398),
    [sym_quoted_atom] = STATE(3398),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3398),
    [sym_charlist] = STATE(3398),
    [sym_sigil] = STATE(3398),
    [sym_list] = STATE(3398),
    [sym_tuple] = STATE(3398),
    [sym_bitstring] = STATE(3398),
    [sym_map] = STATE(3398),
    [sym__nullary_operator] = STATE(3398),
    [sym_unary_operator] = STATE(3398),
    [sym_binary_operator] = STATE(3398),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3398),
    [sym_call] = STATE(3398),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3398),
    [sym_anonymous_function] = STATE(3398),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2225),
    [sym_integer] = ACTIONS(2225),
    [sym_float] = ACTIONS(2225),
    [sym_char] = ACTIONS(2225),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2225),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(656)] = {
    [sym__expression] = STATE(2117),
    [sym_block] = STATE(2117),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(2117),
    [sym_nil] = STATE(2117),
    [sym__atom] = STATE(2117),
    [sym_quoted_atom] = STATE(2117),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2117),
    [sym_charlist] = STATE(2117),
    [sym_sigil] = STATE(2117),
    [sym_list] = STATE(2117),
    [sym_tuple] = STATE(2117),
    [sym_bitstring] = STATE(2117),
    [sym_map] = STATE(2117),
    [sym__nullary_operator] = STATE(2117),
    [sym_unary_operator] = STATE(2117),
    [sym_binary_operator] = STATE(2117),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2117),
    [sym_call] = STATE(2117),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2117),
    [sym_anonymous_function] = STATE(2117),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2227),
    [sym_integer] = ACTIONS(2227),
    [sym_float] = ACTIONS(2227),
    [sym_char] = ACTIONS(2227),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2227),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(657)] = {
    [sym__expression] = STATE(1588),
    [sym_block] = STATE(1588),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1588),
    [sym_nil] = STATE(1588),
    [sym__atom] = STATE(1588),
    [sym_quoted_atom] = STATE(1588),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1588),
    [sym_charlist] = STATE(1588),
    [sym_sigil] = STATE(1588),
    [sym_list] = STATE(1588),
    [sym_tuple] = STATE(1588),
    [sym_bitstring] = STATE(1588),
    [sym_map] = STATE(1588),
    [sym__nullary_operator] = STATE(1588),
    [sym_unary_operator] = STATE(1588),
    [sym_binary_operator] = STATE(1588),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1588),
    [sym_call] = STATE(1588),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1588),
    [sym_anonymous_function] = STATE(1588),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2187),
    [sym_integer] = ACTIONS(2187),
    [sym_float] = ACTIONS(2187),
    [sym_char] = ACTIONS(2187),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2187),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(658)] = {
    [sym__expression] = STATE(2118),
    [sym_block] = STATE(2118),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(2118),
    [sym_nil] = STATE(2118),
    [sym__atom] = STATE(2118),
    [sym_quoted_atom] = STATE(2118),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2118),
    [sym_charlist] = STATE(2118),
    [sym_sigil] = STATE(2118),
    [sym_list] = STATE(2118),
    [sym_tuple] = STATE(2118),
    [sym_bitstring] = STATE(2118),
    [sym_map] = STATE(2118),
    [sym__nullary_operator] = STATE(2118),
    [sym_unary_operator] = STATE(2118),
    [sym_binary_operator] = STATE(2118),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2118),
    [sym_call] = STATE(2118),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2118),
    [sym_anonymous_function] = STATE(2118),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2229),
    [sym_integer] = ACTIONS(2229),
    [sym_float] = ACTIONS(2229),
    [sym_char] = ACTIONS(2229),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2229),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(659)] = {
    [sym__expression] = STATE(1654),
    [sym_block] = STATE(1654),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1654),
    [sym_nil] = STATE(1654),
    [sym__atom] = STATE(1654),
    [sym_quoted_atom] = STATE(1654),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1654),
    [sym_charlist] = STATE(1654),
    [sym_sigil] = STATE(1654),
    [sym_list] = STATE(1654),
    [sym_tuple] = STATE(1654),
    [sym_bitstring] = STATE(1654),
    [sym_map] = STATE(1654),
    [sym__nullary_operator] = STATE(1654),
    [sym_unary_operator] = STATE(1654),
    [sym_binary_operator] = STATE(1654),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1654),
    [sym_call] = STATE(1654),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1654),
    [sym_anonymous_function] = STATE(1654),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2231),
    [sym_integer] = ACTIONS(2231),
    [sym_float] = ACTIONS(2231),
    [sym_char] = ACTIONS(2231),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2231),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(660)] = {
    [sym__expression] = STATE(1946),
    [sym_block] = STATE(1946),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1946),
    [sym_nil] = STATE(1946),
    [sym__atom] = STATE(1946),
    [sym_quoted_atom] = STATE(1946),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1946),
    [sym_charlist] = STATE(1946),
    [sym_sigil] = STATE(1946),
    [sym_list] = STATE(1946),
    [sym_tuple] = STATE(1946),
    [sym_bitstring] = STATE(1946),
    [sym_map] = STATE(1946),
    [sym__nullary_operator] = STATE(1946),
    [sym_unary_operator] = STATE(1946),
    [sym_binary_operator] = STATE(1946),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1946),
    [sym_call] = STATE(1946),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1946),
    [sym_anonymous_function] = STATE(1946),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2233),
    [sym_integer] = ACTIONS(2233),
    [sym_float] = ACTIONS(2233),
    [sym_char] = ACTIONS(2233),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2233),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(661)] = {
    [sym__expression] = STATE(1947),
    [sym_block] = STATE(1947),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1947),
    [sym_nil] = STATE(1947),
    [sym__atom] = STATE(1947),
    [sym_quoted_atom] = STATE(1947),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1947),
    [sym_charlist] = STATE(1947),
    [sym_sigil] = STATE(1947),
    [sym_list] = STATE(1947),
    [sym_tuple] = STATE(1947),
    [sym_bitstring] = STATE(1947),
    [sym_map] = STATE(1947),
    [sym__nullary_operator] = STATE(1947),
    [sym_unary_operator] = STATE(1947),
    [sym_binary_operator] = STATE(1947),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1947),
    [sym_call] = STATE(1947),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1947),
    [sym_anonymous_function] = STATE(1947),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2235),
    [sym_integer] = ACTIONS(2235),
    [sym_float] = ACTIONS(2235),
    [sym_char] = ACTIONS(2235),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2235),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(662)] = {
    [sym__expression] = STATE(1949),
    [sym_block] = STATE(1949),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1949),
    [sym_nil] = STATE(1949),
    [sym__atom] = STATE(1949),
    [sym_quoted_atom] = STATE(1949),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1949),
    [sym_charlist] = STATE(1949),
    [sym_sigil] = STATE(1949),
    [sym_list] = STATE(1949),
    [sym_tuple] = STATE(1949),
    [sym_bitstring] = STATE(1949),
    [sym_map] = STATE(1949),
    [sym__nullary_operator] = STATE(1949),
    [sym_unary_operator] = STATE(1949),
    [sym_binary_operator] = STATE(1949),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1949),
    [sym_call] = STATE(1949),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1949),
    [sym_anonymous_function] = STATE(1949),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2237),
    [sym_integer] = ACTIONS(2237),
    [sym_float] = ACTIONS(2237),
    [sym_char] = ACTIONS(2237),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2237),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(663)] = {
    [sym__expression] = STATE(1951),
    [sym_block] = STATE(1951),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1951),
    [sym_nil] = STATE(1951),
    [sym__atom] = STATE(1951),
    [sym_quoted_atom] = STATE(1951),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1951),
    [sym_charlist] = STATE(1951),
    [sym_sigil] = STATE(1951),
    [sym_list] = STATE(1951),
    [sym_tuple] = STATE(1951),
    [sym_bitstring] = STATE(1951),
    [sym_map] = STATE(1951),
    [sym__nullary_operator] = STATE(1951),
    [sym_unary_operator] = STATE(1951),
    [sym_binary_operator] = STATE(1951),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1951),
    [sym_call] = STATE(1951),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1951),
    [sym_anonymous_function] = STATE(1951),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2239),
    [sym_integer] = ACTIONS(2239),
    [sym_float] = ACTIONS(2239),
    [sym_char] = ACTIONS(2239),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2239),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(664)] = {
    [sym__expression] = STATE(1952),
    [sym_block] = STATE(1952),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1952),
    [sym_nil] = STATE(1952),
    [sym__atom] = STATE(1952),
    [sym_quoted_atom] = STATE(1952),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1952),
    [sym_charlist] = STATE(1952),
    [sym_sigil] = STATE(1952),
    [sym_list] = STATE(1952),
    [sym_tuple] = STATE(1952),
    [sym_bitstring] = STATE(1952),
    [sym_map] = STATE(1952),
    [sym__nullary_operator] = STATE(1952),
    [sym_unary_operator] = STATE(1952),
    [sym_binary_operator] = STATE(1952),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1952),
    [sym_call] = STATE(1952),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1952),
    [sym_anonymous_function] = STATE(1952),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2241),
    [sym_integer] = ACTIONS(2241),
    [sym_float] = ACTIONS(2241),
    [sym_char] = ACTIONS(2241),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2241),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(665)] = {
    [sym__expression] = STATE(1953),
    [sym_block] = STATE(1953),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1953),
    [sym_nil] = STATE(1953),
    [sym__atom] = STATE(1953),
    [sym_quoted_atom] = STATE(1953),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1953),
    [sym_charlist] = STATE(1953),
    [sym_sigil] = STATE(1953),
    [sym_list] = STATE(1953),
    [sym_tuple] = STATE(1953),
    [sym_bitstring] = STATE(1953),
    [sym_map] = STATE(1953),
    [sym__nullary_operator] = STATE(1953),
    [sym_unary_operator] = STATE(1953),
    [sym_binary_operator] = STATE(1953),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1953),
    [sym_call] = STATE(1953),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1953),
    [sym_anonymous_function] = STATE(1953),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2243),
    [sym_integer] = ACTIONS(2243),
    [sym_float] = ACTIONS(2243),
    [sym_char] = ACTIONS(2243),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2243),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(666)] = {
    [sym__expression] = STATE(1957),
    [sym_block] = STATE(1957),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1957),
    [sym_nil] = STATE(1957),
    [sym__atom] = STATE(1957),
    [sym_quoted_atom] = STATE(1957),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1957),
    [sym_charlist] = STATE(1957),
    [sym_sigil] = STATE(1957),
    [sym_list] = STATE(1957),
    [sym_tuple] = STATE(1957),
    [sym_bitstring] = STATE(1957),
    [sym_map] = STATE(1957),
    [sym__nullary_operator] = STATE(1957),
    [sym_unary_operator] = STATE(1957),
    [sym_binary_operator] = STATE(1957),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1957),
    [sym_call] = STATE(1957),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1957),
    [sym_anonymous_function] = STATE(1957),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2245),
    [sym_integer] = ACTIONS(2245),
    [sym_float] = ACTIONS(2245),
    [sym_char] = ACTIONS(2245),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2245),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(667)] = {
    [sym__expression] = STATE(1958),
    [sym_block] = STATE(1958),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1958),
    [sym_nil] = STATE(1958),
    [sym__atom] = STATE(1958),
    [sym_quoted_atom] = STATE(1958),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1958),
    [sym_charlist] = STATE(1958),
    [sym_sigil] = STATE(1958),
    [sym_list] = STATE(1958),
    [sym_tuple] = STATE(1958),
    [sym_bitstring] = STATE(1958),
    [sym_map] = STATE(1958),
    [sym__nullary_operator] = STATE(1958),
    [sym_unary_operator] = STATE(1958),
    [sym_binary_operator] = STATE(1958),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1958),
    [sym_call] = STATE(1958),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1958),
    [sym_anonymous_function] = STATE(1958),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2247),
    [sym_integer] = ACTIONS(2247),
    [sym_float] = ACTIONS(2247),
    [sym_char] = ACTIONS(2247),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2247),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(668)] = {
    [sym__expression] = STATE(1959),
    [sym_block] = STATE(1959),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1959),
    [sym_nil] = STATE(1959),
    [sym__atom] = STATE(1959),
    [sym_quoted_atom] = STATE(1959),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1959),
    [sym_charlist] = STATE(1959),
    [sym_sigil] = STATE(1959),
    [sym_list] = STATE(1959),
    [sym_tuple] = STATE(1959),
    [sym_bitstring] = STATE(1959),
    [sym_map] = STATE(1959),
    [sym__nullary_operator] = STATE(1959),
    [sym_unary_operator] = STATE(1959),
    [sym_binary_operator] = STATE(1959),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1959),
    [sym_call] = STATE(1959),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1959),
    [sym_anonymous_function] = STATE(1959),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2249),
    [sym_integer] = ACTIONS(2249),
    [sym_float] = ACTIONS(2249),
    [sym_char] = ACTIONS(2249),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2249),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(669)] = {
    [sym__expression] = STATE(1960),
    [sym_block] = STATE(1960),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1960),
    [sym_nil] = STATE(1960),
    [sym__atom] = STATE(1960),
    [sym_quoted_atom] = STATE(1960),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1960),
    [sym_charlist] = STATE(1960),
    [sym_sigil] = STATE(1960),
    [sym_list] = STATE(1960),
    [sym_tuple] = STATE(1960),
    [sym_bitstring] = STATE(1960),
    [sym_map] = STATE(1960),
    [sym__nullary_operator] = STATE(1960),
    [sym_unary_operator] = STATE(1960),
    [sym_binary_operator] = STATE(1960),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1960),
    [sym_call] = STATE(1960),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1960),
    [sym_anonymous_function] = STATE(1960),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2251),
    [sym_integer] = ACTIONS(2251),
    [sym_float] = ACTIONS(2251),
    [sym_char] = ACTIONS(2251),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2251),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(670)] = {
    [sym__expression] = STATE(1961),
    [sym_block] = STATE(1961),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1961),
    [sym_nil] = STATE(1961),
    [sym__atom] = STATE(1961),
    [sym_quoted_atom] = STATE(1961),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1961),
    [sym_charlist] = STATE(1961),
    [sym_sigil] = STATE(1961),
    [sym_list] = STATE(1961),
    [sym_tuple] = STATE(1961),
    [sym_bitstring] = STATE(1961),
    [sym_map] = STATE(1961),
    [sym__nullary_operator] = STATE(1961),
    [sym_unary_operator] = STATE(1961),
    [sym_binary_operator] = STATE(1961),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1961),
    [sym_call] = STATE(1961),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1961),
    [sym_anonymous_function] = STATE(1961),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2253),
    [sym_integer] = ACTIONS(2253),
    [sym_float] = ACTIONS(2253),
    [sym_char] = ACTIONS(2253),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2253),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(671)] = {
    [sym__expression] = STATE(1962),
    [sym_block] = STATE(1962),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1962),
    [sym_nil] = STATE(1962),
    [sym__atom] = STATE(1962),
    [sym_quoted_atom] = STATE(1962),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1962),
    [sym_charlist] = STATE(1962),
    [sym_sigil] = STATE(1962),
    [sym_list] = STATE(1962),
    [sym_tuple] = STATE(1962),
    [sym_bitstring] = STATE(1962),
    [sym_map] = STATE(1962),
    [sym__nullary_operator] = STATE(1962),
    [sym_unary_operator] = STATE(1962),
    [sym_binary_operator] = STATE(1962),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1962),
    [sym_call] = STATE(1962),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1962),
    [sym_anonymous_function] = STATE(1962),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2255),
    [sym_integer] = ACTIONS(2255),
    [sym_float] = ACTIONS(2255),
    [sym_char] = ACTIONS(2255),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2255),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(672)] = {
    [sym__expression] = STATE(1963),
    [sym_block] = STATE(1963),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1963),
    [sym_nil] = STATE(1963),
    [sym__atom] = STATE(1963),
    [sym_quoted_atom] = STATE(1963),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1963),
    [sym_charlist] = STATE(1963),
    [sym_sigil] = STATE(1963),
    [sym_list] = STATE(1963),
    [sym_tuple] = STATE(1963),
    [sym_bitstring] = STATE(1963),
    [sym_map] = STATE(1963),
    [sym__nullary_operator] = STATE(1963),
    [sym_unary_operator] = STATE(1963),
    [sym_binary_operator] = STATE(1963),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1963),
    [sym_call] = STATE(1963),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1963),
    [sym_anonymous_function] = STATE(1963),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2257),
    [sym_integer] = ACTIONS(2257),
    [sym_float] = ACTIONS(2257),
    [sym_char] = ACTIONS(2257),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2257),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(673)] = {
    [sym__expression] = STATE(1964),
    [sym_block] = STATE(1964),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1964),
    [sym_nil] = STATE(1964),
    [sym__atom] = STATE(1964),
    [sym_quoted_atom] = STATE(1964),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1964),
    [sym_charlist] = STATE(1964),
    [sym_sigil] = STATE(1964),
    [sym_list] = STATE(1964),
    [sym_tuple] = STATE(1964),
    [sym_bitstring] = STATE(1964),
    [sym_map] = STATE(1964),
    [sym__nullary_operator] = STATE(1964),
    [sym_unary_operator] = STATE(1964),
    [sym_binary_operator] = STATE(1964),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1964),
    [sym_call] = STATE(1964),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1964),
    [sym_anonymous_function] = STATE(1964),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2259),
    [sym_integer] = ACTIONS(2259),
    [sym_float] = ACTIONS(2259),
    [sym_char] = ACTIONS(2259),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2259),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(674)] = {
    [sym__expression] = STATE(1965),
    [sym_block] = STATE(1965),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(1965),
    [sym_nil] = STATE(1965),
    [sym__atom] = STATE(1965),
    [sym_quoted_atom] = STATE(1965),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1965),
    [sym_charlist] = STATE(1965),
    [sym_sigil] = STATE(1965),
    [sym_list] = STATE(1965),
    [sym_tuple] = STATE(1965),
    [sym_bitstring] = STATE(1965),
    [sym_map] = STATE(1965),
    [sym__nullary_operator] = STATE(1965),
    [sym_unary_operator] = STATE(1965),
    [sym_binary_operator] = STATE(1965),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1965),
    [sym_call] = STATE(1965),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1965),
    [sym_anonymous_function] = STATE(1965),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2261),
    [sym_integer] = ACTIONS(2261),
    [sym_float] = ACTIONS(2261),
    [sym_char] = ACTIONS(2261),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2261),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(675)] = {
    [sym__expression] = STATE(2119),
    [sym_block] = STATE(2119),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(2119),
    [sym_nil] = STATE(2119),
    [sym__atom] = STATE(2119),
    [sym_quoted_atom] = STATE(2119),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2119),
    [sym_charlist] = STATE(2119),
    [sym_sigil] = STATE(2119),
    [sym_list] = STATE(2119),
    [sym_tuple] = STATE(2119),
    [sym_bitstring] = STATE(2119),
    [sym_map] = STATE(2119),
    [sym__nullary_operator] = STATE(2119),
    [sym_unary_operator] = STATE(2119),
    [sym_binary_operator] = STATE(2119),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2119),
    [sym_call] = STATE(2119),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2119),
    [sym_anonymous_function] = STATE(2119),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2263),
    [sym_integer] = ACTIONS(2263),
    [sym_float] = ACTIONS(2263),
    [sym_char] = ACTIONS(2263),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2263),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(676)] = {
    [sym__expression] = STATE(3495),
    [sym_block] = STATE(3495),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3495),
    [sym_nil] = STATE(3495),
    [sym__atom] = STATE(3495),
    [sym_quoted_atom] = STATE(3495),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3495),
    [sym_charlist] = STATE(3495),
    [sym_sigil] = STATE(3495),
    [sym_list] = STATE(3495),
    [sym_tuple] = STATE(3495),
    [sym_bitstring] = STATE(3495),
    [sym_map] = STATE(3495),
    [sym__nullary_operator] = STATE(3495),
    [sym_unary_operator] = STATE(3495),
    [sym_binary_operator] = STATE(3495),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3495),
    [sym_call] = STATE(3495),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3495),
    [sym_anonymous_function] = STATE(3495),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2265),
    [sym_integer] = ACTIONS(2265),
    [sym_float] = ACTIONS(2265),
    [sym_char] = ACTIONS(2265),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2265),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(677)] = {
    [sym__expression] = STATE(1466),
    [sym_block] = STATE(1466),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(1466),
    [sym_nil] = STATE(1466),
    [sym__atom] = STATE(1466),
    [sym_quoted_atom] = STATE(1466),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1466),
    [sym_charlist] = STATE(1466),
    [sym_sigil] = STATE(1466),
    [sym_list] = STATE(1466),
    [sym_tuple] = STATE(1466),
    [sym_bitstring] = STATE(1466),
    [sym_map] = STATE(1466),
    [sym__nullary_operator] = STATE(1466),
    [sym_unary_operator] = STATE(1466),
    [sym_binary_operator] = STATE(1466),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1466),
    [sym_call] = STATE(1466),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1466),
    [sym_anonymous_function] = STATE(1466),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2267),
    [sym_integer] = ACTIONS(2267),
    [sym_float] = ACTIONS(2267),
    [sym_char] = ACTIONS(2267),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2267),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(678)] = {
    [sym__expression] = STATE(4127),
    [sym_block] = STATE(4127),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4127),
    [sym_nil] = STATE(4127),
    [sym__atom] = STATE(4127),
    [sym_quoted_atom] = STATE(4127),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4127),
    [sym_charlist] = STATE(4127),
    [sym_sigil] = STATE(4127),
    [sym_list] = STATE(4127),
    [sym_tuple] = STATE(4127),
    [sym_bitstring] = STATE(4127),
    [sym_map] = STATE(4127),
    [sym__nullary_operator] = STATE(4127),
    [sym_unary_operator] = STATE(4127),
    [sym_binary_operator] = STATE(4127),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4127),
    [sym_call] = STATE(4127),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4127),
    [sym_anonymous_function] = STATE(4127),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2269),
    [sym_integer] = ACTIONS(2269),
    [sym_float] = ACTIONS(2269),
    [sym_char] = ACTIONS(2269),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2269),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(679)] = {
    [sym__expression] = STATE(2939),
    [sym_block] = STATE(2939),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(2939),
    [sym_nil] = STATE(2939),
    [sym__atom] = STATE(2939),
    [sym_quoted_atom] = STATE(2939),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2939),
    [sym_charlist] = STATE(2939),
    [sym_sigil] = STATE(2939),
    [sym_list] = STATE(2939),
    [sym_tuple] = STATE(2939),
    [sym_bitstring] = STATE(2939),
    [sym_map] = STATE(2939),
    [sym__nullary_operator] = STATE(2939),
    [sym_unary_operator] = STATE(2939),
    [sym_binary_operator] = STATE(2939),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2939),
    [sym_call] = STATE(2939),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(2939),
    [sym_anonymous_function] = STATE(2939),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2271),
    [sym_integer] = ACTIONS(2271),
    [sym_float] = ACTIONS(2271),
    [sym_char] = ACTIONS(2271),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2271),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(680)] = {
    [sym__expression] = STATE(1523),
    [sym_block] = STATE(1523),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(1523),
    [sym_nil] = STATE(1523),
    [sym__atom] = STATE(1523),
    [sym_quoted_atom] = STATE(1523),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1523),
    [sym_charlist] = STATE(1523),
    [sym_sigil] = STATE(1523),
    [sym_list] = STATE(1523),
    [sym_tuple] = STATE(1523),
    [sym_bitstring] = STATE(1523),
    [sym_map] = STATE(1523),
    [sym__nullary_operator] = STATE(1523),
    [sym_unary_operator] = STATE(1523),
    [sym_binary_operator] = STATE(1523),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1523),
    [sym_call] = STATE(1523),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1523),
    [sym_anonymous_function] = STATE(1523),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2273),
    [sym_integer] = ACTIONS(2273),
    [sym_float] = ACTIONS(2273),
    [sym_char] = ACTIONS(2273),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2273),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(681)] = {
    [sym__expression] = STATE(3113),
    [sym_block] = STATE(3113),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3113),
    [sym_nil] = STATE(3113),
    [sym__atom] = STATE(3113),
    [sym_quoted_atom] = STATE(3113),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3113),
    [sym_charlist] = STATE(3113),
    [sym_sigil] = STATE(3113),
    [sym_list] = STATE(3113),
    [sym_tuple] = STATE(3113),
    [sym_bitstring] = STATE(3113),
    [sym_map] = STATE(3113),
    [sym__nullary_operator] = STATE(3113),
    [sym_unary_operator] = STATE(3113),
    [sym_binary_operator] = STATE(3113),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3113),
    [sym_call] = STATE(3113),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3113),
    [sym_anonymous_function] = STATE(3113),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2275),
    [sym_integer] = ACTIONS(2275),
    [sym_float] = ACTIONS(2275),
    [sym_char] = ACTIONS(2275),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2275),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(682)] = {
    [sym__expression] = STATE(3114),
    [sym_block] = STATE(3114),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3114),
    [sym_nil] = STATE(3114),
    [sym__atom] = STATE(3114),
    [sym_quoted_atom] = STATE(3114),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3114),
    [sym_charlist] = STATE(3114),
    [sym_sigil] = STATE(3114),
    [sym_list] = STATE(3114),
    [sym_tuple] = STATE(3114),
    [sym_bitstring] = STATE(3114),
    [sym_map] = STATE(3114),
    [sym__nullary_operator] = STATE(3114),
    [sym_unary_operator] = STATE(3114),
    [sym_binary_operator] = STATE(3114),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3114),
    [sym_call] = STATE(3114),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3114),
    [sym_anonymous_function] = STATE(3114),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2277),
    [sym_integer] = ACTIONS(2277),
    [sym_float] = ACTIONS(2277),
    [sym_char] = ACTIONS(2277),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2277),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(683)] = {
    [sym__expression] = STATE(3116),
    [sym_block] = STATE(3116),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3116),
    [sym_nil] = STATE(3116),
    [sym__atom] = STATE(3116),
    [sym_quoted_atom] = STATE(3116),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3116),
    [sym_charlist] = STATE(3116),
    [sym_sigil] = STATE(3116),
    [sym_list] = STATE(3116),
    [sym_tuple] = STATE(3116),
    [sym_bitstring] = STATE(3116),
    [sym_map] = STATE(3116),
    [sym__nullary_operator] = STATE(3116),
    [sym_unary_operator] = STATE(3116),
    [sym_binary_operator] = STATE(3116),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3116),
    [sym_call] = STATE(3116),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3116),
    [sym_anonymous_function] = STATE(3116),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2279),
    [sym_integer] = ACTIONS(2279),
    [sym_float] = ACTIONS(2279),
    [sym_char] = ACTIONS(2279),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2279),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(684)] = {
    [sym__expression] = STATE(3117),
    [sym_block] = STATE(3117),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3117),
    [sym_nil] = STATE(3117),
    [sym__atom] = STATE(3117),
    [sym_quoted_atom] = STATE(3117),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3117),
    [sym_charlist] = STATE(3117),
    [sym_sigil] = STATE(3117),
    [sym_list] = STATE(3117),
    [sym_tuple] = STATE(3117),
    [sym_bitstring] = STATE(3117),
    [sym_map] = STATE(3117),
    [sym__nullary_operator] = STATE(3117),
    [sym_unary_operator] = STATE(3117),
    [sym_binary_operator] = STATE(3117),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3117),
    [sym_call] = STATE(3117),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3117),
    [sym_anonymous_function] = STATE(3117),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2281),
    [sym_integer] = ACTIONS(2281),
    [sym_float] = ACTIONS(2281),
    [sym_char] = ACTIONS(2281),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2281),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(685)] = {
    [sym__expression] = STATE(3118),
    [sym_block] = STATE(3118),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3118),
    [sym_nil] = STATE(3118),
    [sym__atom] = STATE(3118),
    [sym_quoted_atom] = STATE(3118),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3118),
    [sym_charlist] = STATE(3118),
    [sym_sigil] = STATE(3118),
    [sym_list] = STATE(3118),
    [sym_tuple] = STATE(3118),
    [sym_bitstring] = STATE(3118),
    [sym_map] = STATE(3118),
    [sym__nullary_operator] = STATE(3118),
    [sym_unary_operator] = STATE(3118),
    [sym_binary_operator] = STATE(3118),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3118),
    [sym_call] = STATE(3118),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3118),
    [sym_anonymous_function] = STATE(3118),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2283),
    [sym_integer] = ACTIONS(2283),
    [sym_float] = ACTIONS(2283),
    [sym_char] = ACTIONS(2283),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2283),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(686)] = {
    [sym__expression] = STATE(3119),
    [sym_block] = STATE(3119),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3119),
    [sym_nil] = STATE(3119),
    [sym__atom] = STATE(3119),
    [sym_quoted_atom] = STATE(3119),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3119),
    [sym_charlist] = STATE(3119),
    [sym_sigil] = STATE(3119),
    [sym_list] = STATE(3119),
    [sym_tuple] = STATE(3119),
    [sym_bitstring] = STATE(3119),
    [sym_map] = STATE(3119),
    [sym__nullary_operator] = STATE(3119),
    [sym_unary_operator] = STATE(3119),
    [sym_binary_operator] = STATE(3119),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3119),
    [sym_call] = STATE(3119),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3119),
    [sym_anonymous_function] = STATE(3119),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2285),
    [sym_integer] = ACTIONS(2285),
    [sym_float] = ACTIONS(2285),
    [sym_char] = ACTIONS(2285),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2285),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(687)] = {
    [sym__expression] = STATE(3121),
    [sym_block] = STATE(3121),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3121),
    [sym_nil] = STATE(3121),
    [sym__atom] = STATE(3121),
    [sym_quoted_atom] = STATE(3121),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3121),
    [sym_charlist] = STATE(3121),
    [sym_sigil] = STATE(3121),
    [sym_list] = STATE(3121),
    [sym_tuple] = STATE(3121),
    [sym_bitstring] = STATE(3121),
    [sym_map] = STATE(3121),
    [sym__nullary_operator] = STATE(3121),
    [sym_unary_operator] = STATE(3121),
    [sym_binary_operator] = STATE(3121),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3121),
    [sym_call] = STATE(3121),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3121),
    [sym_anonymous_function] = STATE(3121),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2287),
    [sym_integer] = ACTIONS(2287),
    [sym_float] = ACTIONS(2287),
    [sym_char] = ACTIONS(2287),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2287),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(688)] = {
    [sym__expression] = STATE(3122),
    [sym_block] = STATE(3122),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3122),
    [sym_nil] = STATE(3122),
    [sym__atom] = STATE(3122),
    [sym_quoted_atom] = STATE(3122),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3122),
    [sym_charlist] = STATE(3122),
    [sym_sigil] = STATE(3122),
    [sym_list] = STATE(3122),
    [sym_tuple] = STATE(3122),
    [sym_bitstring] = STATE(3122),
    [sym_map] = STATE(3122),
    [sym__nullary_operator] = STATE(3122),
    [sym_unary_operator] = STATE(3122),
    [sym_binary_operator] = STATE(3122),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3122),
    [sym_call] = STATE(3122),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3122),
    [sym_anonymous_function] = STATE(3122),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2289),
    [sym_integer] = ACTIONS(2289),
    [sym_float] = ACTIONS(2289),
    [sym_char] = ACTIONS(2289),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2289),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(689)] = {
    [sym__expression] = STATE(3123),
    [sym_block] = STATE(3123),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3123),
    [sym_nil] = STATE(3123),
    [sym__atom] = STATE(3123),
    [sym_quoted_atom] = STATE(3123),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3123),
    [sym_charlist] = STATE(3123),
    [sym_sigil] = STATE(3123),
    [sym_list] = STATE(3123),
    [sym_tuple] = STATE(3123),
    [sym_bitstring] = STATE(3123),
    [sym_map] = STATE(3123),
    [sym__nullary_operator] = STATE(3123),
    [sym_unary_operator] = STATE(3123),
    [sym_binary_operator] = STATE(3123),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3123),
    [sym_call] = STATE(3123),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3123),
    [sym_anonymous_function] = STATE(3123),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2291),
    [sym_integer] = ACTIONS(2291),
    [sym_float] = ACTIONS(2291),
    [sym_char] = ACTIONS(2291),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2291),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(690)] = {
    [sym__expression] = STATE(3124),
    [sym_block] = STATE(3124),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3124),
    [sym_nil] = STATE(3124),
    [sym__atom] = STATE(3124),
    [sym_quoted_atom] = STATE(3124),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3124),
    [sym_charlist] = STATE(3124),
    [sym_sigil] = STATE(3124),
    [sym_list] = STATE(3124),
    [sym_tuple] = STATE(3124),
    [sym_bitstring] = STATE(3124),
    [sym_map] = STATE(3124),
    [sym__nullary_operator] = STATE(3124),
    [sym_unary_operator] = STATE(3124),
    [sym_binary_operator] = STATE(3124),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3124),
    [sym_call] = STATE(3124),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3124),
    [sym_anonymous_function] = STATE(3124),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2293),
    [sym_integer] = ACTIONS(2293),
    [sym_float] = ACTIONS(2293),
    [sym_char] = ACTIONS(2293),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2293),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(691)] = {
    [sym__expression] = STATE(3125),
    [sym_block] = STATE(3125),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3125),
    [sym_nil] = STATE(3125),
    [sym__atom] = STATE(3125),
    [sym_quoted_atom] = STATE(3125),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3125),
    [sym_charlist] = STATE(3125),
    [sym_sigil] = STATE(3125),
    [sym_list] = STATE(3125),
    [sym_tuple] = STATE(3125),
    [sym_bitstring] = STATE(3125),
    [sym_map] = STATE(3125),
    [sym__nullary_operator] = STATE(3125),
    [sym_unary_operator] = STATE(3125),
    [sym_binary_operator] = STATE(3125),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3125),
    [sym_call] = STATE(3125),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3125),
    [sym_anonymous_function] = STATE(3125),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2295),
    [sym_integer] = ACTIONS(2295),
    [sym_float] = ACTIONS(2295),
    [sym_char] = ACTIONS(2295),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2295),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(692)] = {
    [sym__expression] = STATE(3126),
    [sym_block] = STATE(3126),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3126),
    [sym_nil] = STATE(3126),
    [sym__atom] = STATE(3126),
    [sym_quoted_atom] = STATE(3126),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3126),
    [sym_charlist] = STATE(3126),
    [sym_sigil] = STATE(3126),
    [sym_list] = STATE(3126),
    [sym_tuple] = STATE(3126),
    [sym_bitstring] = STATE(3126),
    [sym_map] = STATE(3126),
    [sym__nullary_operator] = STATE(3126),
    [sym_unary_operator] = STATE(3126),
    [sym_binary_operator] = STATE(3126),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3126),
    [sym_call] = STATE(3126),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3126),
    [sym_anonymous_function] = STATE(3126),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2297),
    [sym_integer] = ACTIONS(2297),
    [sym_float] = ACTIONS(2297),
    [sym_char] = ACTIONS(2297),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2297),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(693)] = {
    [sym__expression] = STATE(3127),
    [sym_block] = STATE(3127),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3127),
    [sym_nil] = STATE(3127),
    [sym__atom] = STATE(3127),
    [sym_quoted_atom] = STATE(3127),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3127),
    [sym_charlist] = STATE(3127),
    [sym_sigil] = STATE(3127),
    [sym_list] = STATE(3127),
    [sym_tuple] = STATE(3127),
    [sym_bitstring] = STATE(3127),
    [sym_map] = STATE(3127),
    [sym__nullary_operator] = STATE(3127),
    [sym_unary_operator] = STATE(3127),
    [sym_binary_operator] = STATE(3127),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3127),
    [sym_call] = STATE(3127),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3127),
    [sym_anonymous_function] = STATE(3127),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2299),
    [sym_integer] = ACTIONS(2299),
    [sym_float] = ACTIONS(2299),
    [sym_char] = ACTIONS(2299),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2299),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(694)] = {
    [sym__expression] = STATE(3128),
    [sym_block] = STATE(3128),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3128),
    [sym_nil] = STATE(3128),
    [sym__atom] = STATE(3128),
    [sym_quoted_atom] = STATE(3128),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3128),
    [sym_charlist] = STATE(3128),
    [sym_sigil] = STATE(3128),
    [sym_list] = STATE(3128),
    [sym_tuple] = STATE(3128),
    [sym_bitstring] = STATE(3128),
    [sym_map] = STATE(3128),
    [sym__nullary_operator] = STATE(3128),
    [sym_unary_operator] = STATE(3128),
    [sym_binary_operator] = STATE(3128),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3128),
    [sym_call] = STATE(3128),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3128),
    [sym_anonymous_function] = STATE(3128),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2301),
    [sym_integer] = ACTIONS(2301),
    [sym_float] = ACTIONS(2301),
    [sym_char] = ACTIONS(2301),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2301),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(695)] = {
    [sym__expression] = STATE(3129),
    [sym_block] = STATE(3129),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3129),
    [sym_nil] = STATE(3129),
    [sym__atom] = STATE(3129),
    [sym_quoted_atom] = STATE(3129),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3129),
    [sym_charlist] = STATE(3129),
    [sym_sigil] = STATE(3129),
    [sym_list] = STATE(3129),
    [sym_tuple] = STATE(3129),
    [sym_bitstring] = STATE(3129),
    [sym_map] = STATE(3129),
    [sym__nullary_operator] = STATE(3129),
    [sym_unary_operator] = STATE(3129),
    [sym_binary_operator] = STATE(3129),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3129),
    [sym_call] = STATE(3129),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3129),
    [sym_anonymous_function] = STATE(3129),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2303),
    [sym_integer] = ACTIONS(2303),
    [sym_float] = ACTIONS(2303),
    [sym_char] = ACTIONS(2303),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2303),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(696)] = {
    [sym__expression] = STATE(2980),
    [sym_block] = STATE(2980),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(2980),
    [sym_nil] = STATE(2980),
    [sym__atom] = STATE(2980),
    [sym_quoted_atom] = STATE(2980),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(2980),
    [sym_charlist] = STATE(2980),
    [sym_sigil] = STATE(2980),
    [sym_list] = STATE(2980),
    [sym_tuple] = STATE(2980),
    [sym_bitstring] = STATE(2980),
    [sym_map] = STATE(2980),
    [sym__nullary_operator] = STATE(2980),
    [sym_unary_operator] = STATE(2980),
    [sym_binary_operator] = STATE(2980),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(2980),
    [sym_call] = STATE(2980),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(2980),
    [sym_anonymous_function] = STATE(2980),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2305),
    [sym_integer] = ACTIONS(2305),
    [sym_float] = ACTIONS(2305),
    [sym_char] = ACTIONS(2305),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2305),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(697)] = {
    [sym__expression] = STATE(3469),
    [sym_block] = STATE(3469),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3469),
    [sym_nil] = STATE(3469),
    [sym__atom] = STATE(3469),
    [sym_quoted_atom] = STATE(3469),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3469),
    [sym_charlist] = STATE(3469),
    [sym_sigil] = STATE(3469),
    [sym_list] = STATE(3469),
    [sym_tuple] = STATE(3469),
    [sym_bitstring] = STATE(3469),
    [sym_map] = STATE(3469),
    [sym__nullary_operator] = STATE(3469),
    [sym_unary_operator] = STATE(3469),
    [sym_binary_operator] = STATE(3469),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3469),
    [sym_call] = STATE(3469),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3469),
    [sym_anonymous_function] = STATE(3469),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2307),
    [sym_integer] = ACTIONS(2307),
    [sym_float] = ACTIONS(2307),
    [sym_char] = ACTIONS(2307),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2307),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(698)] = {
    [sym__expression] = STATE(1466),
    [sym_block] = STATE(1466),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(1466),
    [sym_nil] = STATE(1466),
    [sym__atom] = STATE(1466),
    [sym_quoted_atom] = STATE(1466),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1466),
    [sym_charlist] = STATE(1466),
    [sym_sigil] = STATE(1466),
    [sym_list] = STATE(1466),
    [sym_tuple] = STATE(1466),
    [sym_bitstring] = STATE(1466),
    [sym_map] = STATE(1466),
    [sym__nullary_operator] = STATE(1466),
    [sym_unary_operator] = STATE(1466),
    [sym_binary_operator] = STATE(1466),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1466),
    [sym_call] = STATE(1466),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1466),
    [sym_anonymous_function] = STATE(1466),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2267),
    [sym_integer] = ACTIONS(2267),
    [sym_float] = ACTIONS(2267),
    [sym_char] = ACTIONS(2267),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2267),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(699)] = {
    [sym__expression] = STATE(3851),
    [sym_block] = STATE(3851),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(3851),
    [sym_nil] = STATE(3851),
    [sym__atom] = STATE(3851),
    [sym_quoted_atom] = STATE(3851),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(3851),
    [sym_charlist] = STATE(3851),
    [sym_sigil] = STATE(3851),
    [sym_list] = STATE(3851),
    [sym_tuple] = STATE(3851),
    [sym_bitstring] = STATE(3851),
    [sym_map] = STATE(3851),
    [sym__nullary_operator] = STATE(3851),
    [sym_unary_operator] = STATE(3851),
    [sym_binary_operator] = STATE(3851),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(3851),
    [sym_call] = STATE(3851),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(3851),
    [sym_anonymous_function] = STATE(3851),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2309),
    [sym_integer] = ACTIONS(2309),
    [sym_float] = ACTIONS(2309),
    [sym_char] = ACTIONS(2309),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2309),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(700)] = {
    [sym__expression] = STATE(1654),
    [sym_block] = STATE(1654),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(1654),
    [sym_nil] = STATE(1654),
    [sym__atom] = STATE(1654),
    [sym_quoted_atom] = STATE(1654),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(1654),
    [sym_charlist] = STATE(1654),
    [sym_sigil] = STATE(1654),
    [sym_list] = STATE(1654),
    [sym_tuple] = STATE(1654),
    [sym_bitstring] = STATE(1654),
    [sym_map] = STATE(1654),
    [sym__nullary_operator] = STATE(1654),
    [sym_unary_operator] = STATE(1654),
    [sym_binary_operator] = STATE(1654),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(1654),
    [sym_call] = STATE(1654),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(1654),
    [sym_anonymous_function] = STATE(1654),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2231),
    [sym_integer] = ACTIONS(2231),
    [sym_float] = ACTIONS(2231),
    [sym_char] = ACTIONS(2231),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2231),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(701)] = {
    [sym__expression] = STATE(3470),
    [sym_block] = STATE(3470),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3470),
    [sym_nil] = STATE(3470),
    [sym__atom] = STATE(3470),
    [sym_quoted_atom] = STATE(3470),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3470),
    [sym_charlist] = STATE(3470),
    [sym_sigil] = STATE(3470),
    [sym_list] = STATE(3470),
    [sym_tuple] = STATE(3470),
    [sym_bitstring] = STATE(3470),
    [sym_map] = STATE(3470),
    [sym__nullary_operator] = STATE(3470),
    [sym_unary_operator] = STATE(3470),
    [sym_binary_operator] = STATE(3470),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3470),
    [sym_call] = STATE(3470),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3470),
    [sym_anonymous_function] = STATE(3470),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2311),
    [sym_integer] = ACTIONS(2311),
    [sym_float] = ACTIONS(2311),
    [sym_char] = ACTIONS(2311),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2311),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(702)] = {
    [sym__expression] = STATE(1523),
    [sym_block] = STATE(1523),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(1523),
    [sym_nil] = STATE(1523),
    [sym__atom] = STATE(1523),
    [sym_quoted_atom] = STATE(1523),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1523),
    [sym_charlist] = STATE(1523),
    [sym_sigil] = STATE(1523),
    [sym_list] = STATE(1523),
    [sym_tuple] = STATE(1523),
    [sym_bitstring] = STATE(1523),
    [sym_map] = STATE(1523),
    [sym__nullary_operator] = STATE(1523),
    [sym_unary_operator] = STATE(1523),
    [sym_binary_operator] = STATE(1523),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1523),
    [sym_call] = STATE(1523),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1523),
    [sym_anonymous_function] = STATE(1523),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2273),
    [sym_integer] = ACTIONS(2273),
    [sym_float] = ACTIONS(2273),
    [sym_char] = ACTIONS(2273),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2273),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(703)] = {
    [sym__expression] = STATE(3164),
    [sym_block] = STATE(3164),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3164),
    [sym_nil] = STATE(3164),
    [sym__atom] = STATE(3164),
    [sym_quoted_atom] = STATE(3164),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3164),
    [sym_charlist] = STATE(3164),
    [sym_sigil] = STATE(3164),
    [sym_list] = STATE(3164),
    [sym_tuple] = STATE(3164),
    [sym_bitstring] = STATE(3164),
    [sym_map] = STATE(3164),
    [sym__nullary_operator] = STATE(3164),
    [sym_unary_operator] = STATE(3164),
    [sym_binary_operator] = STATE(3164),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3164),
    [sym_call] = STATE(3164),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3164),
    [sym_anonymous_function] = STATE(3164),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2313),
    [sym_integer] = ACTIONS(2313),
    [sym_float] = ACTIONS(2313),
    [sym_char] = ACTIONS(2313),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2313),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(704)] = {
    [sym__expression] = STATE(4192),
    [sym_block] = STATE(4192),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4192),
    [sym_nil] = STATE(4192),
    [sym__atom] = STATE(4192),
    [sym_quoted_atom] = STATE(4192),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4192),
    [sym_charlist] = STATE(4192),
    [sym_sigil] = STATE(4192),
    [sym_list] = STATE(4192),
    [sym_tuple] = STATE(4192),
    [sym_bitstring] = STATE(4192),
    [sym_map] = STATE(4192),
    [sym__nullary_operator] = STATE(4192),
    [sym_unary_operator] = STATE(4192),
    [sym_binary_operator] = STATE(4192),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4192),
    [sym_call] = STATE(4192),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4192),
    [sym_anonymous_function] = STATE(4192),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2315),
    [sym_integer] = ACTIONS(2315),
    [sym_float] = ACTIONS(2315),
    [sym_char] = ACTIONS(2315),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2315),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(705)] = {
    [sym__expression] = STATE(3167),
    [sym_block] = STATE(3167),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3167),
    [sym_nil] = STATE(3167),
    [sym__atom] = STATE(3167),
    [sym_quoted_atom] = STATE(3167),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3167),
    [sym_charlist] = STATE(3167),
    [sym_sigil] = STATE(3167),
    [sym_list] = STATE(3167),
    [sym_tuple] = STATE(3167),
    [sym_bitstring] = STATE(3167),
    [sym_map] = STATE(3167),
    [sym__nullary_operator] = STATE(3167),
    [sym_unary_operator] = STATE(3167),
    [sym_binary_operator] = STATE(3167),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3167),
    [sym_call] = STATE(3167),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3167),
    [sym_anonymous_function] = STATE(3167),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2317),
    [sym_integer] = ACTIONS(2317),
    [sym_float] = ACTIONS(2317),
    [sym_char] = ACTIONS(2317),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2317),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(706)] = {
    [sym__expression] = STATE(3168),
    [sym_block] = STATE(3168),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3168),
    [sym_nil] = STATE(3168),
    [sym__atom] = STATE(3168),
    [sym_quoted_atom] = STATE(3168),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3168),
    [sym_charlist] = STATE(3168),
    [sym_sigil] = STATE(3168),
    [sym_list] = STATE(3168),
    [sym_tuple] = STATE(3168),
    [sym_bitstring] = STATE(3168),
    [sym_map] = STATE(3168),
    [sym__nullary_operator] = STATE(3168),
    [sym_unary_operator] = STATE(3168),
    [sym_binary_operator] = STATE(3168),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3168),
    [sym_call] = STATE(3168),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3168),
    [sym_anonymous_function] = STATE(3168),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2319),
    [sym_integer] = ACTIONS(2319),
    [sym_float] = ACTIONS(2319),
    [sym_char] = ACTIONS(2319),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2319),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(707)] = {
    [sym__expression] = STATE(3169),
    [sym_block] = STATE(3169),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3169),
    [sym_nil] = STATE(3169),
    [sym__atom] = STATE(3169),
    [sym_quoted_atom] = STATE(3169),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3169),
    [sym_charlist] = STATE(3169),
    [sym_sigil] = STATE(3169),
    [sym_list] = STATE(3169),
    [sym_tuple] = STATE(3169),
    [sym_bitstring] = STATE(3169),
    [sym_map] = STATE(3169),
    [sym__nullary_operator] = STATE(3169),
    [sym_unary_operator] = STATE(3169),
    [sym_binary_operator] = STATE(3169),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3169),
    [sym_call] = STATE(3169),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3169),
    [sym_anonymous_function] = STATE(3169),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2321),
    [sym_integer] = ACTIONS(2321),
    [sym_float] = ACTIONS(2321),
    [sym_char] = ACTIONS(2321),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2321),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(708)] = {
    [sym__expression] = STATE(3170),
    [sym_block] = STATE(3170),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3170),
    [sym_nil] = STATE(3170),
    [sym__atom] = STATE(3170),
    [sym_quoted_atom] = STATE(3170),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3170),
    [sym_charlist] = STATE(3170),
    [sym_sigil] = STATE(3170),
    [sym_list] = STATE(3170),
    [sym_tuple] = STATE(3170),
    [sym_bitstring] = STATE(3170),
    [sym_map] = STATE(3170),
    [sym__nullary_operator] = STATE(3170),
    [sym_unary_operator] = STATE(3170),
    [sym_binary_operator] = STATE(3170),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3170),
    [sym_call] = STATE(3170),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3170),
    [sym_anonymous_function] = STATE(3170),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2323),
    [sym_integer] = ACTIONS(2323),
    [sym_float] = ACTIONS(2323),
    [sym_char] = ACTIONS(2323),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2323),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(709)] = {
    [sym__expression] = STATE(3172),
    [sym_block] = STATE(3172),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3172),
    [sym_nil] = STATE(3172),
    [sym__atom] = STATE(3172),
    [sym_quoted_atom] = STATE(3172),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3172),
    [sym_charlist] = STATE(3172),
    [sym_sigil] = STATE(3172),
    [sym_list] = STATE(3172),
    [sym_tuple] = STATE(3172),
    [sym_bitstring] = STATE(3172),
    [sym_map] = STATE(3172),
    [sym__nullary_operator] = STATE(3172),
    [sym_unary_operator] = STATE(3172),
    [sym_binary_operator] = STATE(3172),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3172),
    [sym_call] = STATE(3172),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3172),
    [sym_anonymous_function] = STATE(3172),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2325),
    [sym_integer] = ACTIONS(2325),
    [sym_float] = ACTIONS(2325),
    [sym_char] = ACTIONS(2325),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2325),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(710)] = {
    [sym__expression] = STATE(3173),
    [sym_block] = STATE(3173),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3173),
    [sym_nil] = STATE(3173),
    [sym__atom] = STATE(3173),
    [sym_quoted_atom] = STATE(3173),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3173),
    [sym_charlist] = STATE(3173),
    [sym_sigil] = STATE(3173),
    [sym_list] = STATE(3173),
    [sym_tuple] = STATE(3173),
    [sym_bitstring] = STATE(3173),
    [sym_map] = STATE(3173),
    [sym__nullary_operator] = STATE(3173),
    [sym_unary_operator] = STATE(3173),
    [sym_binary_operator] = STATE(3173),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3173),
    [sym_call] = STATE(3173),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3173),
    [sym_anonymous_function] = STATE(3173),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2327),
    [sym_integer] = ACTIONS(2327),
    [sym_float] = ACTIONS(2327),
    [sym_char] = ACTIONS(2327),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2327),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(711)] = {
    [sym__expression] = STATE(3174),
    [sym_block] = STATE(3174),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3174),
    [sym_nil] = STATE(3174),
    [sym__atom] = STATE(3174),
    [sym_quoted_atom] = STATE(3174),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3174),
    [sym_charlist] = STATE(3174),
    [sym_sigil] = STATE(3174),
    [sym_list] = STATE(3174),
    [sym_tuple] = STATE(3174),
    [sym_bitstring] = STATE(3174),
    [sym_map] = STATE(3174),
    [sym__nullary_operator] = STATE(3174),
    [sym_unary_operator] = STATE(3174),
    [sym_binary_operator] = STATE(3174),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3174),
    [sym_call] = STATE(3174),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3174),
    [sym_anonymous_function] = STATE(3174),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2329),
    [sym_integer] = ACTIONS(2329),
    [sym_float] = ACTIONS(2329),
    [sym_char] = ACTIONS(2329),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2329),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(712)] = {
    [sym__expression] = STATE(3175),
    [sym_block] = STATE(3175),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3175),
    [sym_nil] = STATE(3175),
    [sym__atom] = STATE(3175),
    [sym_quoted_atom] = STATE(3175),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3175),
    [sym_charlist] = STATE(3175),
    [sym_sigil] = STATE(3175),
    [sym_list] = STATE(3175),
    [sym_tuple] = STATE(3175),
    [sym_bitstring] = STATE(3175),
    [sym_map] = STATE(3175),
    [sym__nullary_operator] = STATE(3175),
    [sym_unary_operator] = STATE(3175),
    [sym_binary_operator] = STATE(3175),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3175),
    [sym_call] = STATE(3175),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3175),
    [sym_anonymous_function] = STATE(3175),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2331),
    [sym_integer] = ACTIONS(2331),
    [sym_float] = ACTIONS(2331),
    [sym_char] = ACTIONS(2331),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2331),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(713)] = {
    [sym__expression] = STATE(3176),
    [sym_block] = STATE(3176),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3176),
    [sym_nil] = STATE(3176),
    [sym__atom] = STATE(3176),
    [sym_quoted_atom] = STATE(3176),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3176),
    [sym_charlist] = STATE(3176),
    [sym_sigil] = STATE(3176),
    [sym_list] = STATE(3176),
    [sym_tuple] = STATE(3176),
    [sym_bitstring] = STATE(3176),
    [sym_map] = STATE(3176),
    [sym__nullary_operator] = STATE(3176),
    [sym_unary_operator] = STATE(3176),
    [sym_binary_operator] = STATE(3176),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3176),
    [sym_call] = STATE(3176),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3176),
    [sym_anonymous_function] = STATE(3176),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2333),
    [sym_integer] = ACTIONS(2333),
    [sym_float] = ACTIONS(2333),
    [sym_char] = ACTIONS(2333),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2333),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(714)] = {
    [sym__expression] = STATE(3177),
    [sym_block] = STATE(3177),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3177),
    [sym_nil] = STATE(3177),
    [sym__atom] = STATE(3177),
    [sym_quoted_atom] = STATE(3177),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3177),
    [sym_charlist] = STATE(3177),
    [sym_sigil] = STATE(3177),
    [sym_list] = STATE(3177),
    [sym_tuple] = STATE(3177),
    [sym_bitstring] = STATE(3177),
    [sym_map] = STATE(3177),
    [sym__nullary_operator] = STATE(3177),
    [sym_unary_operator] = STATE(3177),
    [sym_binary_operator] = STATE(3177),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3177),
    [sym_call] = STATE(3177),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3177),
    [sym_anonymous_function] = STATE(3177),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2335),
    [sym_integer] = ACTIONS(2335),
    [sym_float] = ACTIONS(2335),
    [sym_char] = ACTIONS(2335),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2335),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(715)] = {
    [sym__expression] = STATE(3178),
    [sym_block] = STATE(3178),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3178),
    [sym_nil] = STATE(3178),
    [sym__atom] = STATE(3178),
    [sym_quoted_atom] = STATE(3178),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3178),
    [sym_charlist] = STATE(3178),
    [sym_sigil] = STATE(3178),
    [sym_list] = STATE(3178),
    [sym_tuple] = STATE(3178),
    [sym_bitstring] = STATE(3178),
    [sym_map] = STATE(3178),
    [sym__nullary_operator] = STATE(3178),
    [sym_unary_operator] = STATE(3178),
    [sym_binary_operator] = STATE(3178),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3178),
    [sym_call] = STATE(3178),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3178),
    [sym_anonymous_function] = STATE(3178),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2337),
    [sym_integer] = ACTIONS(2337),
    [sym_float] = ACTIONS(2337),
    [sym_char] = ACTIONS(2337),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2337),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(716)] = {
    [sym__expression] = STATE(3179),
    [sym_block] = STATE(3179),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3179),
    [sym_nil] = STATE(3179),
    [sym__atom] = STATE(3179),
    [sym_quoted_atom] = STATE(3179),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3179),
    [sym_charlist] = STATE(3179),
    [sym_sigil] = STATE(3179),
    [sym_list] = STATE(3179),
    [sym_tuple] = STATE(3179),
    [sym_bitstring] = STATE(3179),
    [sym_map] = STATE(3179),
    [sym__nullary_operator] = STATE(3179),
    [sym_unary_operator] = STATE(3179),
    [sym_binary_operator] = STATE(3179),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3179),
    [sym_call] = STATE(3179),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3179),
    [sym_anonymous_function] = STATE(3179),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2339),
    [sym_integer] = ACTIONS(2339),
    [sym_float] = ACTIONS(2339),
    [sym_char] = ACTIONS(2339),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2339),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(717)] = {
    [sym__expression] = STATE(3180),
    [sym_block] = STATE(3180),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3180),
    [sym_nil] = STATE(3180),
    [sym__atom] = STATE(3180),
    [sym_quoted_atom] = STATE(3180),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3180),
    [sym_charlist] = STATE(3180),
    [sym_sigil] = STATE(3180),
    [sym_list] = STATE(3180),
    [sym_tuple] = STATE(3180),
    [sym_bitstring] = STATE(3180),
    [sym_map] = STATE(3180),
    [sym__nullary_operator] = STATE(3180),
    [sym_unary_operator] = STATE(3180),
    [sym_binary_operator] = STATE(3180),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3180),
    [sym_call] = STATE(3180),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3180),
    [sym_anonymous_function] = STATE(3180),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2341),
    [sym_integer] = ACTIONS(2341),
    [sym_float] = ACTIONS(2341),
    [sym_char] = ACTIONS(2341),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2341),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(718)] = {
    [sym__expression] = STATE(3471),
    [sym_block] = STATE(3471),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3471),
    [sym_nil] = STATE(3471),
    [sym__atom] = STATE(3471),
    [sym_quoted_atom] = STATE(3471),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3471),
    [sym_charlist] = STATE(3471),
    [sym_sigil] = STATE(3471),
    [sym_list] = STATE(3471),
    [sym_tuple] = STATE(3471),
    [sym_bitstring] = STATE(3471),
    [sym_map] = STATE(3471),
    [sym__nullary_operator] = STATE(3471),
    [sym_unary_operator] = STATE(3471),
    [sym_binary_operator] = STATE(3471),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3471),
    [sym_call] = STATE(3471),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3471),
    [sym_anonymous_function] = STATE(3471),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2343),
    [sym_integer] = ACTIONS(2343),
    [sym_float] = ACTIONS(2343),
    [sym_char] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2343),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(719)] = {
    [sym__expression] = STATE(2335),
    [sym_block] = STATE(2335),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2335),
    [sym_nil] = STATE(2335),
    [sym__atom] = STATE(2335),
    [sym_quoted_atom] = STATE(2335),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2335),
    [sym_charlist] = STATE(2335),
    [sym_sigil] = STATE(2335),
    [sym_list] = STATE(2335),
    [sym_tuple] = STATE(2335),
    [sym_bitstring] = STATE(2335),
    [sym_map] = STATE(2335),
    [sym__nullary_operator] = STATE(2335),
    [sym_unary_operator] = STATE(2335),
    [sym_binary_operator] = STATE(2335),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2335),
    [sym_call] = STATE(2335),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2335),
    [sym_anonymous_function] = STATE(2335),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2345),
    [sym_integer] = ACTIONS(2345),
    [sym_float] = ACTIONS(2345),
    [sym_char] = ACTIONS(2345),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2345),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(720)] = {
    [sym__expression] = STATE(1908),
    [sym_block] = STATE(1908),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(1908),
    [sym_nil] = STATE(1908),
    [sym__atom] = STATE(1908),
    [sym_quoted_atom] = STATE(1908),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(1908),
    [sym_charlist] = STATE(1908),
    [sym_sigil] = STATE(1908),
    [sym_list] = STATE(1908),
    [sym_tuple] = STATE(1908),
    [sym_bitstring] = STATE(1908),
    [sym_map] = STATE(1908),
    [sym__nullary_operator] = STATE(1908),
    [sym_unary_operator] = STATE(1908),
    [sym_binary_operator] = STATE(1908),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(1908),
    [sym_call] = STATE(1908),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(1908),
    [sym_anonymous_function] = STATE(1908),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1977),
    [sym_integer] = ACTIONS(1977),
    [sym_float] = ACTIONS(1977),
    [sym_char] = ACTIONS(1977),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1977),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(721)] = {
    [sym__expression] = STATE(4139),
    [sym_block] = STATE(4139),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4139),
    [sym_nil] = STATE(4139),
    [sym__atom] = STATE(4139),
    [sym_quoted_atom] = STATE(4139),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4139),
    [sym_charlist] = STATE(4139),
    [sym_sigil] = STATE(4139),
    [sym_list] = STATE(4139),
    [sym_tuple] = STATE(4139),
    [sym_bitstring] = STATE(4139),
    [sym_map] = STATE(4139),
    [sym__nullary_operator] = STATE(4139),
    [sym_unary_operator] = STATE(4139),
    [sym_binary_operator] = STATE(4139),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4139),
    [sym_call] = STATE(4139),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4139),
    [sym_anonymous_function] = STATE(4139),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2347),
    [sym_integer] = ACTIONS(2347),
    [sym_float] = ACTIONS(2347),
    [sym_char] = ACTIONS(2347),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2347),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(722)] = {
    [sym__expression] = STATE(2336),
    [sym_block] = STATE(2336),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2336),
    [sym_nil] = STATE(2336),
    [sym__atom] = STATE(2336),
    [sym_quoted_atom] = STATE(2336),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2336),
    [sym_charlist] = STATE(2336),
    [sym_sigil] = STATE(2336),
    [sym_list] = STATE(2336),
    [sym_tuple] = STATE(2336),
    [sym_bitstring] = STATE(2336),
    [sym_map] = STATE(2336),
    [sym__nullary_operator] = STATE(2336),
    [sym_unary_operator] = STATE(2336),
    [sym_binary_operator] = STATE(2336),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2336),
    [sym_call] = STATE(2336),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2336),
    [sym_anonymous_function] = STATE(2336),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2349),
    [sym_integer] = ACTIONS(2349),
    [sym_float] = ACTIONS(2349),
    [sym_char] = ACTIONS(2349),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2349),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(723)] = {
    [sym__expression] = STATE(2022),
    [sym_block] = STATE(2022),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2022),
    [sym_nil] = STATE(2022),
    [sym__atom] = STATE(2022),
    [sym_quoted_atom] = STATE(2022),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2022),
    [sym_charlist] = STATE(2022),
    [sym_sigil] = STATE(2022),
    [sym_list] = STATE(2022),
    [sym_tuple] = STATE(2022),
    [sym_bitstring] = STATE(2022),
    [sym_map] = STATE(2022),
    [sym__nullary_operator] = STATE(2022),
    [sym_unary_operator] = STATE(2022),
    [sym_binary_operator] = STATE(2022),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2022),
    [sym_call] = STATE(2022),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2022),
    [sym_anonymous_function] = STATE(2022),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1981),
    [sym_integer] = ACTIONS(1981),
    [sym_float] = ACTIONS(1981),
    [sym_char] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1981),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(724)] = {
    [sym__expression] = STATE(2187),
    [sym_block] = STATE(2187),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2187),
    [sym_nil] = STATE(2187),
    [sym__atom] = STATE(2187),
    [sym_quoted_atom] = STATE(2187),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2187),
    [sym_charlist] = STATE(2187),
    [sym_sigil] = STATE(2187),
    [sym_list] = STATE(2187),
    [sym_tuple] = STATE(2187),
    [sym_bitstring] = STATE(2187),
    [sym_map] = STATE(2187),
    [sym__nullary_operator] = STATE(2187),
    [sym_unary_operator] = STATE(2187),
    [sym_binary_operator] = STATE(2187),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2187),
    [sym_call] = STATE(2187),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2187),
    [sym_anonymous_function] = STATE(2187),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2351),
    [sym_integer] = ACTIONS(2351),
    [sym_float] = ACTIONS(2351),
    [sym_char] = ACTIONS(2351),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2351),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(725)] = {
    [sym__expression] = STATE(2190),
    [sym_block] = STATE(2190),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2190),
    [sym_nil] = STATE(2190),
    [sym__atom] = STATE(2190),
    [sym_quoted_atom] = STATE(2190),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2190),
    [sym_charlist] = STATE(2190),
    [sym_sigil] = STATE(2190),
    [sym_list] = STATE(2190),
    [sym_tuple] = STATE(2190),
    [sym_bitstring] = STATE(2190),
    [sym_map] = STATE(2190),
    [sym__nullary_operator] = STATE(2190),
    [sym_unary_operator] = STATE(2190),
    [sym_binary_operator] = STATE(2190),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2190),
    [sym_call] = STATE(2190),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2190),
    [sym_anonymous_function] = STATE(2190),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2353),
    [sym_integer] = ACTIONS(2353),
    [sym_float] = ACTIONS(2353),
    [sym_char] = ACTIONS(2353),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2353),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(726)] = {
    [sym__expression] = STATE(2192),
    [sym_block] = STATE(2192),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2192),
    [sym_nil] = STATE(2192),
    [sym__atom] = STATE(2192),
    [sym_quoted_atom] = STATE(2192),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2192),
    [sym_charlist] = STATE(2192),
    [sym_sigil] = STATE(2192),
    [sym_list] = STATE(2192),
    [sym_tuple] = STATE(2192),
    [sym_bitstring] = STATE(2192),
    [sym_map] = STATE(2192),
    [sym__nullary_operator] = STATE(2192),
    [sym_unary_operator] = STATE(2192),
    [sym_binary_operator] = STATE(2192),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2192),
    [sym_call] = STATE(2192),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2192),
    [sym_anonymous_function] = STATE(2192),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2355),
    [sym_integer] = ACTIONS(2355),
    [sym_float] = ACTIONS(2355),
    [sym_char] = ACTIONS(2355),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2355),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(727)] = {
    [sym__expression] = STATE(2196),
    [sym_block] = STATE(2196),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2196),
    [sym_nil] = STATE(2196),
    [sym__atom] = STATE(2196),
    [sym_quoted_atom] = STATE(2196),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2196),
    [sym_charlist] = STATE(2196),
    [sym_sigil] = STATE(2196),
    [sym_list] = STATE(2196),
    [sym_tuple] = STATE(2196),
    [sym_bitstring] = STATE(2196),
    [sym_map] = STATE(2196),
    [sym__nullary_operator] = STATE(2196),
    [sym_unary_operator] = STATE(2196),
    [sym_binary_operator] = STATE(2196),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2196),
    [sym_call] = STATE(2196),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2196),
    [sym_anonymous_function] = STATE(2196),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2357),
    [sym_integer] = ACTIONS(2357),
    [sym_float] = ACTIONS(2357),
    [sym_char] = ACTIONS(2357),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2357),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(728)] = {
    [sym__expression] = STATE(2226),
    [sym_block] = STATE(2226),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2226),
    [sym_nil] = STATE(2226),
    [sym__atom] = STATE(2226),
    [sym_quoted_atom] = STATE(2226),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2226),
    [sym_charlist] = STATE(2226),
    [sym_sigil] = STATE(2226),
    [sym_list] = STATE(2226),
    [sym_tuple] = STATE(2226),
    [sym_bitstring] = STATE(2226),
    [sym_map] = STATE(2226),
    [sym__nullary_operator] = STATE(2226),
    [sym_unary_operator] = STATE(2226),
    [sym_binary_operator] = STATE(2226),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2226),
    [sym_call] = STATE(2226),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2226),
    [sym_anonymous_function] = STATE(2226),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2359),
    [sym_integer] = ACTIONS(2359),
    [sym_float] = ACTIONS(2359),
    [sym_char] = ACTIONS(2359),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2359),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(729)] = {
    [sym__expression] = STATE(2141),
    [sym_block] = STATE(2141),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2141),
    [sym_nil] = STATE(2141),
    [sym__atom] = STATE(2141),
    [sym_quoted_atom] = STATE(2141),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2141),
    [sym_charlist] = STATE(2141),
    [sym_sigil] = STATE(2141),
    [sym_list] = STATE(2141),
    [sym_tuple] = STATE(2141),
    [sym_bitstring] = STATE(2141),
    [sym_map] = STATE(2141),
    [sym__nullary_operator] = STATE(2141),
    [sym_unary_operator] = STATE(2141),
    [sym_binary_operator] = STATE(2141),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2141),
    [sym_call] = STATE(2141),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2141),
    [sym_anonymous_function] = STATE(2141),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2361),
    [sym_integer] = ACTIONS(2361),
    [sym_float] = ACTIONS(2361),
    [sym_char] = ACTIONS(2361),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2361),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(730)] = {
    [sym__expression] = STATE(2275),
    [sym_block] = STATE(2275),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2275),
    [sym_nil] = STATE(2275),
    [sym__atom] = STATE(2275),
    [sym_quoted_atom] = STATE(2275),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2275),
    [sym_charlist] = STATE(2275),
    [sym_sigil] = STATE(2275),
    [sym_list] = STATE(2275),
    [sym_tuple] = STATE(2275),
    [sym_bitstring] = STATE(2275),
    [sym_map] = STATE(2275),
    [sym__nullary_operator] = STATE(2275),
    [sym_unary_operator] = STATE(2275),
    [sym_binary_operator] = STATE(2275),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2275),
    [sym_call] = STATE(2275),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2275),
    [sym_anonymous_function] = STATE(2275),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2363),
    [sym_integer] = ACTIONS(2363),
    [sym_float] = ACTIONS(2363),
    [sym_char] = ACTIONS(2363),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2363),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(731)] = {
    [sym__expression] = STATE(2276),
    [sym_block] = STATE(2276),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2276),
    [sym_nil] = STATE(2276),
    [sym__atom] = STATE(2276),
    [sym_quoted_atom] = STATE(2276),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2276),
    [sym_charlist] = STATE(2276),
    [sym_sigil] = STATE(2276),
    [sym_list] = STATE(2276),
    [sym_tuple] = STATE(2276),
    [sym_bitstring] = STATE(2276),
    [sym_map] = STATE(2276),
    [sym__nullary_operator] = STATE(2276),
    [sym_unary_operator] = STATE(2276),
    [sym_binary_operator] = STATE(2276),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2276),
    [sym_call] = STATE(2276),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2276),
    [sym_anonymous_function] = STATE(2276),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2365),
    [sym_integer] = ACTIONS(2365),
    [sym_float] = ACTIONS(2365),
    [sym_char] = ACTIONS(2365),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(732)] = {
    [sym__expression] = STATE(2280),
    [sym_block] = STATE(2280),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2280),
    [sym_nil] = STATE(2280),
    [sym__atom] = STATE(2280),
    [sym_quoted_atom] = STATE(2280),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2280),
    [sym_charlist] = STATE(2280),
    [sym_sigil] = STATE(2280),
    [sym_list] = STATE(2280),
    [sym_tuple] = STATE(2280),
    [sym_bitstring] = STATE(2280),
    [sym_map] = STATE(2280),
    [sym__nullary_operator] = STATE(2280),
    [sym_unary_operator] = STATE(2280),
    [sym_binary_operator] = STATE(2280),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2280),
    [sym_call] = STATE(2280),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2280),
    [sym_anonymous_function] = STATE(2280),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2367),
    [sym_integer] = ACTIONS(2367),
    [sym_float] = ACTIONS(2367),
    [sym_char] = ACTIONS(2367),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2367),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(733)] = {
    [sym__expression] = STATE(2281),
    [sym_block] = STATE(2281),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2281),
    [sym_nil] = STATE(2281),
    [sym__atom] = STATE(2281),
    [sym_quoted_atom] = STATE(2281),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2281),
    [sym_charlist] = STATE(2281),
    [sym_sigil] = STATE(2281),
    [sym_list] = STATE(2281),
    [sym_tuple] = STATE(2281),
    [sym_bitstring] = STATE(2281),
    [sym_map] = STATE(2281),
    [sym__nullary_operator] = STATE(2281),
    [sym_unary_operator] = STATE(2281),
    [sym_binary_operator] = STATE(2281),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2281),
    [sym_call] = STATE(2281),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2281),
    [sym_anonymous_function] = STATE(2281),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2369),
    [sym_integer] = ACTIONS(2369),
    [sym_float] = ACTIONS(2369),
    [sym_char] = ACTIONS(2369),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2369),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(734)] = {
    [sym__expression] = STATE(2357),
    [sym_block] = STATE(2357),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2357),
    [sym_nil] = STATE(2357),
    [sym__atom] = STATE(2357),
    [sym_quoted_atom] = STATE(2357),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2357),
    [sym_charlist] = STATE(2357),
    [sym_sigil] = STATE(2357),
    [sym_list] = STATE(2357),
    [sym_tuple] = STATE(2357),
    [sym_bitstring] = STATE(2357),
    [sym_map] = STATE(2357),
    [sym__nullary_operator] = STATE(2357),
    [sym_unary_operator] = STATE(2357),
    [sym_binary_operator] = STATE(2357),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2357),
    [sym_call] = STATE(2357),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2357),
    [sym_anonymous_function] = STATE(2357),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2371),
    [sym_integer] = ACTIONS(2371),
    [sym_float] = ACTIONS(2371),
    [sym_char] = ACTIONS(2371),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2371),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(735)] = {
    [sym__expression] = STATE(2155),
    [sym_block] = STATE(2155),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2155),
    [sym_nil] = STATE(2155),
    [sym__atom] = STATE(2155),
    [sym_quoted_atom] = STATE(2155),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2155),
    [sym_charlist] = STATE(2155),
    [sym_sigil] = STATE(2155),
    [sym_list] = STATE(2155),
    [sym_tuple] = STATE(2155),
    [sym_bitstring] = STATE(2155),
    [sym_map] = STATE(2155),
    [sym__nullary_operator] = STATE(2155),
    [sym_unary_operator] = STATE(2155),
    [sym_binary_operator] = STATE(2155),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2155),
    [sym_call] = STATE(2155),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2155),
    [sym_anonymous_function] = STATE(2155),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2373),
    [sym_integer] = ACTIONS(2373),
    [sym_float] = ACTIONS(2373),
    [sym_char] = ACTIONS(2373),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2373),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(736)] = {
    [sym__expression] = STATE(2167),
    [sym_block] = STATE(2167),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2167),
    [sym_nil] = STATE(2167),
    [sym__atom] = STATE(2167),
    [sym_quoted_atom] = STATE(2167),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2167),
    [sym_charlist] = STATE(2167),
    [sym_sigil] = STATE(2167),
    [sym_list] = STATE(2167),
    [sym_tuple] = STATE(2167),
    [sym_bitstring] = STATE(2167),
    [sym_map] = STATE(2167),
    [sym__nullary_operator] = STATE(2167),
    [sym_unary_operator] = STATE(2167),
    [sym_binary_operator] = STATE(2167),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2167),
    [sym_call] = STATE(2167),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2167),
    [sym_anonymous_function] = STATE(2167),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2375),
    [sym_integer] = ACTIONS(2375),
    [sym_float] = ACTIONS(2375),
    [sym_char] = ACTIONS(2375),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2375),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(737)] = {
    [sym__expression] = STATE(2169),
    [sym_block] = STATE(2169),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2169),
    [sym_nil] = STATE(2169),
    [sym__atom] = STATE(2169),
    [sym_quoted_atom] = STATE(2169),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2169),
    [sym_charlist] = STATE(2169),
    [sym_sigil] = STATE(2169),
    [sym_list] = STATE(2169),
    [sym_tuple] = STATE(2169),
    [sym_bitstring] = STATE(2169),
    [sym_map] = STATE(2169),
    [sym__nullary_operator] = STATE(2169),
    [sym_unary_operator] = STATE(2169),
    [sym_binary_operator] = STATE(2169),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2169),
    [sym_call] = STATE(2169),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2169),
    [sym_anonymous_function] = STATE(2169),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2377),
    [sym_integer] = ACTIONS(2377),
    [sym_float] = ACTIONS(2377),
    [sym_char] = ACTIONS(2377),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2377),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(738)] = {
    [sym__expression] = STATE(2171),
    [sym_block] = STATE(2171),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2171),
    [sym_nil] = STATE(2171),
    [sym__atom] = STATE(2171),
    [sym_quoted_atom] = STATE(2171),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2171),
    [sym_charlist] = STATE(2171),
    [sym_sigil] = STATE(2171),
    [sym_list] = STATE(2171),
    [sym_tuple] = STATE(2171),
    [sym_bitstring] = STATE(2171),
    [sym_map] = STATE(2171),
    [sym__nullary_operator] = STATE(2171),
    [sym_unary_operator] = STATE(2171),
    [sym_binary_operator] = STATE(2171),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2171),
    [sym_call] = STATE(2171),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2171),
    [sym_anonymous_function] = STATE(2171),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2379),
    [sym_integer] = ACTIONS(2379),
    [sym_float] = ACTIONS(2379),
    [sym_char] = ACTIONS(2379),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2379),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(739)] = {
    [sym__expression] = STATE(2337),
    [sym_block] = STATE(2337),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2337),
    [sym_nil] = STATE(2337),
    [sym__atom] = STATE(2337),
    [sym_quoted_atom] = STATE(2337),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2337),
    [sym_charlist] = STATE(2337),
    [sym_sigil] = STATE(2337),
    [sym_list] = STATE(2337),
    [sym_tuple] = STATE(2337),
    [sym_bitstring] = STATE(2337),
    [sym_map] = STATE(2337),
    [sym__nullary_operator] = STATE(2337),
    [sym_unary_operator] = STATE(2337),
    [sym_binary_operator] = STATE(2337),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2337),
    [sym_call] = STATE(2337),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2337),
    [sym_anonymous_function] = STATE(2337),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2381),
    [sym_integer] = ACTIONS(2381),
    [sym_float] = ACTIONS(2381),
    [sym_char] = ACTIONS(2381),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2381),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(740)] = {
    [sym__expression] = STATE(1554),
    [sym_block] = STATE(1554),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1554),
    [sym_nil] = STATE(1554),
    [sym__atom] = STATE(1554),
    [sym_quoted_atom] = STATE(1554),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1554),
    [sym_charlist] = STATE(1554),
    [sym_sigil] = STATE(1554),
    [sym_list] = STATE(1554),
    [sym_tuple] = STATE(1554),
    [sym_bitstring] = STATE(1554),
    [sym_map] = STATE(1554),
    [sym__nullary_operator] = STATE(1554),
    [sym_unary_operator] = STATE(1554),
    [sym_binary_operator] = STATE(1554),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1554),
    [sym_call] = STATE(1554),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1554),
    [sym_anonymous_function] = STATE(1554),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2383),
    [sym_integer] = ACTIONS(2383),
    [sym_float] = ACTIONS(2383),
    [sym_char] = ACTIONS(2383),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2383),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(741)] = {
    [sym__expression] = STATE(1310),
    [sym_block] = STATE(1310),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1310),
    [sym_nil] = STATE(1310),
    [sym__atom] = STATE(1310),
    [sym_quoted_atom] = STATE(1310),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1310),
    [sym_charlist] = STATE(1310),
    [sym_sigil] = STATE(1310),
    [sym_list] = STATE(1310),
    [sym_tuple] = STATE(1310),
    [sym_bitstring] = STATE(1310),
    [sym_map] = STATE(1310),
    [sym__nullary_operator] = STATE(1310),
    [sym_unary_operator] = STATE(1310),
    [sym_binary_operator] = STATE(1310),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1310),
    [sym_call] = STATE(1310),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1310),
    [sym_anonymous_function] = STATE(1310),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2017),
    [sym_integer] = ACTIONS(2017),
    [sym_float] = ACTIONS(2017),
    [sym_char] = ACTIONS(2017),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2017),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(742)] = {
    [sym__expression] = STATE(1555),
    [sym_block] = STATE(1555),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1555),
    [sym_nil] = STATE(1555),
    [sym__atom] = STATE(1555),
    [sym_quoted_atom] = STATE(1555),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1555),
    [sym_charlist] = STATE(1555),
    [sym_sigil] = STATE(1555),
    [sym_list] = STATE(1555),
    [sym_tuple] = STATE(1555),
    [sym_bitstring] = STATE(1555),
    [sym_map] = STATE(1555),
    [sym__nullary_operator] = STATE(1555),
    [sym_unary_operator] = STATE(1555),
    [sym_binary_operator] = STATE(1555),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1555),
    [sym_call] = STATE(1555),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1555),
    [sym_anonymous_function] = STATE(1555),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2385),
    [sym_integer] = ACTIONS(2385),
    [sym_float] = ACTIONS(2385),
    [sym_char] = ACTIONS(2385),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2385),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(743)] = {
    [sym__expression] = STATE(1189),
    [sym_block] = STATE(1189),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1189),
    [sym_nil] = STATE(1189),
    [sym__atom] = STATE(1189),
    [sym_quoted_atom] = STATE(1189),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1189),
    [sym_charlist] = STATE(1189),
    [sym_sigil] = STATE(1189),
    [sym_list] = STATE(1189),
    [sym_tuple] = STATE(1189),
    [sym_bitstring] = STATE(1189),
    [sym_map] = STATE(1189),
    [sym__nullary_operator] = STATE(1189),
    [sym_unary_operator] = STATE(1189),
    [sym_binary_operator] = STATE(1189),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1189),
    [sym_call] = STATE(1189),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1189),
    [sym_anonymous_function] = STATE(1189),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2021),
    [sym_integer] = ACTIONS(2021),
    [sym_float] = ACTIONS(2021),
    [sym_char] = ACTIONS(2021),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(744)] = {
    [sym__expression] = STATE(1514),
    [sym_block] = STATE(1514),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1514),
    [sym_nil] = STATE(1514),
    [sym__atom] = STATE(1514),
    [sym_quoted_atom] = STATE(1514),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1514),
    [sym_charlist] = STATE(1514),
    [sym_sigil] = STATE(1514),
    [sym_list] = STATE(1514),
    [sym_tuple] = STATE(1514),
    [sym_bitstring] = STATE(1514),
    [sym_map] = STATE(1514),
    [sym__nullary_operator] = STATE(1514),
    [sym_unary_operator] = STATE(1514),
    [sym_binary_operator] = STATE(1514),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1514),
    [sym_call] = STATE(1514),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1514),
    [sym_anonymous_function] = STATE(1514),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2387),
    [sym_integer] = ACTIONS(2387),
    [sym_float] = ACTIONS(2387),
    [sym_char] = ACTIONS(2387),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2387),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(745)] = {
    [sym__expression] = STATE(1515),
    [sym_block] = STATE(1515),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1515),
    [sym_nil] = STATE(1515),
    [sym__atom] = STATE(1515),
    [sym_quoted_atom] = STATE(1515),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1515),
    [sym_charlist] = STATE(1515),
    [sym_sigil] = STATE(1515),
    [sym_list] = STATE(1515),
    [sym_tuple] = STATE(1515),
    [sym_bitstring] = STATE(1515),
    [sym_map] = STATE(1515),
    [sym__nullary_operator] = STATE(1515),
    [sym_unary_operator] = STATE(1515),
    [sym_binary_operator] = STATE(1515),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1515),
    [sym_call] = STATE(1515),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1515),
    [sym_anonymous_function] = STATE(1515),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2389),
    [sym_integer] = ACTIONS(2389),
    [sym_float] = ACTIONS(2389),
    [sym_char] = ACTIONS(2389),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2389),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(746)] = {
    [sym__expression] = STATE(1517),
    [sym_block] = STATE(1517),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1517),
    [sym_nil] = STATE(1517),
    [sym__atom] = STATE(1517),
    [sym_quoted_atom] = STATE(1517),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1517),
    [sym_charlist] = STATE(1517),
    [sym_sigil] = STATE(1517),
    [sym_list] = STATE(1517),
    [sym_tuple] = STATE(1517),
    [sym_bitstring] = STATE(1517),
    [sym_map] = STATE(1517),
    [sym__nullary_operator] = STATE(1517),
    [sym_unary_operator] = STATE(1517),
    [sym_binary_operator] = STATE(1517),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1517),
    [sym_call] = STATE(1517),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1517),
    [sym_anonymous_function] = STATE(1517),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2391),
    [sym_integer] = ACTIONS(2391),
    [sym_float] = ACTIONS(2391),
    [sym_char] = ACTIONS(2391),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2391),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(747)] = {
    [sym__expression] = STATE(1518),
    [sym_block] = STATE(1518),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1518),
    [sym_nil] = STATE(1518),
    [sym__atom] = STATE(1518),
    [sym_quoted_atom] = STATE(1518),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1518),
    [sym_charlist] = STATE(1518),
    [sym_sigil] = STATE(1518),
    [sym_list] = STATE(1518),
    [sym_tuple] = STATE(1518),
    [sym_bitstring] = STATE(1518),
    [sym_map] = STATE(1518),
    [sym__nullary_operator] = STATE(1518),
    [sym_unary_operator] = STATE(1518),
    [sym_binary_operator] = STATE(1518),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1518),
    [sym_call] = STATE(1518),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1518),
    [sym_anonymous_function] = STATE(1518),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2393),
    [sym_integer] = ACTIONS(2393),
    [sym_float] = ACTIONS(2393),
    [sym_char] = ACTIONS(2393),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2393),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(748)] = {
    [sym__expression] = STATE(1519),
    [sym_block] = STATE(1519),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1519),
    [sym_nil] = STATE(1519),
    [sym__atom] = STATE(1519),
    [sym_quoted_atom] = STATE(1519),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1519),
    [sym_charlist] = STATE(1519),
    [sym_sigil] = STATE(1519),
    [sym_list] = STATE(1519),
    [sym_tuple] = STATE(1519),
    [sym_bitstring] = STATE(1519),
    [sym_map] = STATE(1519),
    [sym__nullary_operator] = STATE(1519),
    [sym_unary_operator] = STATE(1519),
    [sym_binary_operator] = STATE(1519),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1519),
    [sym_call] = STATE(1519),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1519),
    [sym_anonymous_function] = STATE(1519),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2395),
    [sym_integer] = ACTIONS(2395),
    [sym_float] = ACTIONS(2395),
    [sym_char] = ACTIONS(2395),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2395),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(749)] = {
    [sym__expression] = STATE(1520),
    [sym_block] = STATE(1520),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1520),
    [sym_nil] = STATE(1520),
    [sym__atom] = STATE(1520),
    [sym_quoted_atom] = STATE(1520),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1520),
    [sym_charlist] = STATE(1520),
    [sym_sigil] = STATE(1520),
    [sym_list] = STATE(1520),
    [sym_tuple] = STATE(1520),
    [sym_bitstring] = STATE(1520),
    [sym_map] = STATE(1520),
    [sym__nullary_operator] = STATE(1520),
    [sym_unary_operator] = STATE(1520),
    [sym_binary_operator] = STATE(1520),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1520),
    [sym_call] = STATE(1520),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1520),
    [sym_anonymous_function] = STATE(1520),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2397),
    [sym_integer] = ACTIONS(2397),
    [sym_float] = ACTIONS(2397),
    [sym_char] = ACTIONS(2397),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2397),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(750)] = {
    [sym__expression] = STATE(4159),
    [sym_block] = STATE(4159),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4159),
    [sym_nil] = STATE(4159),
    [sym__atom] = STATE(4159),
    [sym_quoted_atom] = STATE(4159),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4159),
    [sym_charlist] = STATE(4159),
    [sym_sigil] = STATE(4159),
    [sym_list] = STATE(4159),
    [sym_tuple] = STATE(4159),
    [sym_bitstring] = STATE(4159),
    [sym_map] = STATE(4159),
    [sym__nullary_operator] = STATE(4159),
    [sym_unary_operator] = STATE(4159),
    [sym_binary_operator] = STATE(4159),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4159),
    [sym_call] = STATE(4159),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4159),
    [sym_anonymous_function] = STATE(4159),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2399),
    [sym_integer] = ACTIONS(2399),
    [sym_float] = ACTIONS(2399),
    [sym_char] = ACTIONS(2399),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2399),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(751)] = {
    [sym__expression] = STATE(1526),
    [sym_block] = STATE(1526),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1526),
    [sym_nil] = STATE(1526),
    [sym__atom] = STATE(1526),
    [sym_quoted_atom] = STATE(1526),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1526),
    [sym_charlist] = STATE(1526),
    [sym_sigil] = STATE(1526),
    [sym_list] = STATE(1526),
    [sym_tuple] = STATE(1526),
    [sym_bitstring] = STATE(1526),
    [sym_map] = STATE(1526),
    [sym__nullary_operator] = STATE(1526),
    [sym_unary_operator] = STATE(1526),
    [sym_binary_operator] = STATE(1526),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1526),
    [sym_call] = STATE(1526),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1526),
    [sym_anonymous_function] = STATE(1526),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2401),
    [sym_integer] = ACTIONS(2401),
    [sym_float] = ACTIONS(2401),
    [sym_char] = ACTIONS(2401),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2401),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(752)] = {
    [sym__expression] = STATE(1527),
    [sym_block] = STATE(1527),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1527),
    [sym_nil] = STATE(1527),
    [sym__atom] = STATE(1527),
    [sym_quoted_atom] = STATE(1527),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1527),
    [sym_charlist] = STATE(1527),
    [sym_sigil] = STATE(1527),
    [sym_list] = STATE(1527),
    [sym_tuple] = STATE(1527),
    [sym_bitstring] = STATE(1527),
    [sym_map] = STATE(1527),
    [sym__nullary_operator] = STATE(1527),
    [sym_unary_operator] = STATE(1527),
    [sym_binary_operator] = STATE(1527),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1527),
    [sym_call] = STATE(1527),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1527),
    [sym_anonymous_function] = STATE(1527),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2403),
    [sym_integer] = ACTIONS(2403),
    [sym_float] = ACTIONS(2403),
    [sym_char] = ACTIONS(2403),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2403),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(753)] = {
    [sym__expression] = STATE(1528),
    [sym_block] = STATE(1528),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1528),
    [sym_nil] = STATE(1528),
    [sym__atom] = STATE(1528),
    [sym_quoted_atom] = STATE(1528),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1528),
    [sym_charlist] = STATE(1528),
    [sym_sigil] = STATE(1528),
    [sym_list] = STATE(1528),
    [sym_tuple] = STATE(1528),
    [sym_bitstring] = STATE(1528),
    [sym_map] = STATE(1528),
    [sym__nullary_operator] = STATE(1528),
    [sym_unary_operator] = STATE(1528),
    [sym_binary_operator] = STATE(1528),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1528),
    [sym_call] = STATE(1528),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1528),
    [sym_anonymous_function] = STATE(1528),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2405),
    [sym_integer] = ACTIONS(2405),
    [sym_float] = ACTIONS(2405),
    [sym_char] = ACTIONS(2405),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2405),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(754)] = {
    [sym__expression] = STATE(1532),
    [sym_block] = STATE(1532),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1532),
    [sym_nil] = STATE(1532),
    [sym__atom] = STATE(1532),
    [sym_quoted_atom] = STATE(1532),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1532),
    [sym_charlist] = STATE(1532),
    [sym_sigil] = STATE(1532),
    [sym_list] = STATE(1532),
    [sym_tuple] = STATE(1532),
    [sym_bitstring] = STATE(1532),
    [sym_map] = STATE(1532),
    [sym__nullary_operator] = STATE(1532),
    [sym_unary_operator] = STATE(1532),
    [sym_binary_operator] = STATE(1532),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1532),
    [sym_call] = STATE(1532),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1532),
    [sym_anonymous_function] = STATE(1532),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2407),
    [sym_integer] = ACTIONS(2407),
    [sym_float] = ACTIONS(2407),
    [sym_char] = ACTIONS(2407),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2407),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(755)] = {
    [sym__expression] = STATE(1533),
    [sym_block] = STATE(1533),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1533),
    [sym_nil] = STATE(1533),
    [sym__atom] = STATE(1533),
    [sym_quoted_atom] = STATE(1533),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1533),
    [sym_charlist] = STATE(1533),
    [sym_sigil] = STATE(1533),
    [sym_list] = STATE(1533),
    [sym_tuple] = STATE(1533),
    [sym_bitstring] = STATE(1533),
    [sym_map] = STATE(1533),
    [sym__nullary_operator] = STATE(1533),
    [sym_unary_operator] = STATE(1533),
    [sym_binary_operator] = STATE(1533),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1533),
    [sym_call] = STATE(1533),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1533),
    [sym_anonymous_function] = STATE(1533),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2409),
    [sym_integer] = ACTIONS(2409),
    [sym_float] = ACTIONS(2409),
    [sym_char] = ACTIONS(2409),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2409),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(756)] = {
    [sym__expression] = STATE(1534),
    [sym_block] = STATE(1534),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1534),
    [sym_nil] = STATE(1534),
    [sym__atom] = STATE(1534),
    [sym_quoted_atom] = STATE(1534),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1534),
    [sym_charlist] = STATE(1534),
    [sym_sigil] = STATE(1534),
    [sym_list] = STATE(1534),
    [sym_tuple] = STATE(1534),
    [sym_bitstring] = STATE(1534),
    [sym_map] = STATE(1534),
    [sym__nullary_operator] = STATE(1534),
    [sym_unary_operator] = STATE(1534),
    [sym_binary_operator] = STATE(1534),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1534),
    [sym_call] = STATE(1534),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1534),
    [sym_anonymous_function] = STATE(1534),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2411),
    [sym_integer] = ACTIONS(2411),
    [sym_float] = ACTIONS(2411),
    [sym_char] = ACTIONS(2411),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2411),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(757)] = {
    [sym__expression] = STATE(1535),
    [sym_block] = STATE(1535),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1535),
    [sym_nil] = STATE(1535),
    [sym__atom] = STATE(1535),
    [sym_quoted_atom] = STATE(1535),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1535),
    [sym_charlist] = STATE(1535),
    [sym_sigil] = STATE(1535),
    [sym_list] = STATE(1535),
    [sym_tuple] = STATE(1535),
    [sym_bitstring] = STATE(1535),
    [sym_map] = STATE(1535),
    [sym__nullary_operator] = STATE(1535),
    [sym_unary_operator] = STATE(1535),
    [sym_binary_operator] = STATE(1535),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1535),
    [sym_call] = STATE(1535),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1535),
    [sym_anonymous_function] = STATE(1535),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2413),
    [sym_integer] = ACTIONS(2413),
    [sym_float] = ACTIONS(2413),
    [sym_char] = ACTIONS(2413),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2413),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(758)] = {
    [sym__expression] = STATE(1539),
    [sym_block] = STATE(1539),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1539),
    [sym_nil] = STATE(1539),
    [sym__atom] = STATE(1539),
    [sym_quoted_atom] = STATE(1539),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1539),
    [sym_charlist] = STATE(1539),
    [sym_sigil] = STATE(1539),
    [sym_list] = STATE(1539),
    [sym_tuple] = STATE(1539),
    [sym_bitstring] = STATE(1539),
    [sym_map] = STATE(1539),
    [sym__nullary_operator] = STATE(1539),
    [sym_unary_operator] = STATE(1539),
    [sym_binary_operator] = STATE(1539),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1539),
    [sym_call] = STATE(1539),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1539),
    [sym_anonymous_function] = STATE(1539),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2415),
    [sym_integer] = ACTIONS(2415),
    [sym_float] = ACTIONS(2415),
    [sym_char] = ACTIONS(2415),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2415),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(759)] = {
    [sym__expression] = STATE(1541),
    [sym_block] = STATE(1541),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1541),
    [sym_nil] = STATE(1541),
    [sym__atom] = STATE(1541),
    [sym_quoted_atom] = STATE(1541),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1541),
    [sym_charlist] = STATE(1541),
    [sym_sigil] = STATE(1541),
    [sym_list] = STATE(1541),
    [sym_tuple] = STATE(1541),
    [sym_bitstring] = STATE(1541),
    [sym_map] = STATE(1541),
    [sym__nullary_operator] = STATE(1541),
    [sym_unary_operator] = STATE(1541),
    [sym_binary_operator] = STATE(1541),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1541),
    [sym_call] = STATE(1541),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1541),
    [sym_anonymous_function] = STATE(1541),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2417),
    [sym_integer] = ACTIONS(2417),
    [sym_float] = ACTIONS(2417),
    [sym_char] = ACTIONS(2417),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2417),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(760)] = {
    [sym__expression] = STATE(1556),
    [sym_block] = STATE(1556),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1556),
    [sym_nil] = STATE(1556),
    [sym__atom] = STATE(1556),
    [sym_quoted_atom] = STATE(1556),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1556),
    [sym_charlist] = STATE(1556),
    [sym_sigil] = STATE(1556),
    [sym_list] = STATE(1556),
    [sym_tuple] = STATE(1556),
    [sym_bitstring] = STATE(1556),
    [sym_map] = STATE(1556),
    [sym__nullary_operator] = STATE(1556),
    [sym_unary_operator] = STATE(1556),
    [sym_binary_operator] = STATE(1556),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1556),
    [sym_call] = STATE(1556),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1556),
    [sym_anonymous_function] = STATE(1556),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2419),
    [sym_integer] = ACTIONS(2419),
    [sym_float] = ACTIONS(2419),
    [sym_char] = ACTIONS(2419),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2419),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(761)] = {
    [sym__expression] = STATE(1795),
    [sym_block] = STATE(1795),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1795),
    [sym_nil] = STATE(1795),
    [sym__atom] = STATE(1795),
    [sym_quoted_atom] = STATE(1795),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1795),
    [sym_charlist] = STATE(1795),
    [sym_sigil] = STATE(1795),
    [sym_list] = STATE(1795),
    [sym_tuple] = STATE(1795),
    [sym_bitstring] = STATE(1795),
    [sym_map] = STATE(1795),
    [sym__nullary_operator] = STATE(1795),
    [sym_unary_operator] = STATE(1795),
    [sym_binary_operator] = STATE(1795),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1795),
    [sym_call] = STATE(1795),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1795),
    [sym_anonymous_function] = STATE(1795),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2421),
    [sym_integer] = ACTIONS(2421),
    [sym_float] = ACTIONS(2421),
    [sym_char] = ACTIONS(2421),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2421),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(762)] = {
    [sym__expression] = STATE(1466),
    [sym_block] = STATE(1466),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1466),
    [sym_nil] = STATE(1466),
    [sym__atom] = STATE(1466),
    [sym_quoted_atom] = STATE(1466),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1466),
    [sym_charlist] = STATE(1466),
    [sym_sigil] = STATE(1466),
    [sym_list] = STATE(1466),
    [sym_tuple] = STATE(1466),
    [sym_bitstring] = STATE(1466),
    [sym_map] = STATE(1466),
    [sym__nullary_operator] = STATE(1466),
    [sym_unary_operator] = STATE(1466),
    [sym_binary_operator] = STATE(1466),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1466),
    [sym_call] = STATE(1466),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1466),
    [sym_anonymous_function] = STATE(1466),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2267),
    [sym_integer] = ACTIONS(2267),
    [sym_float] = ACTIONS(2267),
    [sym_char] = ACTIONS(2267),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2267),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(763)] = {
    [sym__expression] = STATE(1796),
    [sym_block] = STATE(1796),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1796),
    [sym_nil] = STATE(1796),
    [sym__atom] = STATE(1796),
    [sym_quoted_atom] = STATE(1796),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1796),
    [sym_charlist] = STATE(1796),
    [sym_sigil] = STATE(1796),
    [sym_list] = STATE(1796),
    [sym_tuple] = STATE(1796),
    [sym_bitstring] = STATE(1796),
    [sym_map] = STATE(1796),
    [sym__nullary_operator] = STATE(1796),
    [sym_unary_operator] = STATE(1796),
    [sym_binary_operator] = STATE(1796),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1796),
    [sym_call] = STATE(1796),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1796),
    [sym_anonymous_function] = STATE(1796),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2423),
    [sym_integer] = ACTIONS(2423),
    [sym_float] = ACTIONS(2423),
    [sym_char] = ACTIONS(2423),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2423),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(764)] = {
    [sym__expression] = STATE(1523),
    [sym_block] = STATE(1523),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1523),
    [sym_nil] = STATE(1523),
    [sym__atom] = STATE(1523),
    [sym_quoted_atom] = STATE(1523),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1523),
    [sym_charlist] = STATE(1523),
    [sym_sigil] = STATE(1523),
    [sym_list] = STATE(1523),
    [sym_tuple] = STATE(1523),
    [sym_bitstring] = STATE(1523),
    [sym_map] = STATE(1523),
    [sym__nullary_operator] = STATE(1523),
    [sym_unary_operator] = STATE(1523),
    [sym_binary_operator] = STATE(1523),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1523),
    [sym_call] = STATE(1523),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1523),
    [sym_anonymous_function] = STATE(1523),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2273),
    [sym_integer] = ACTIONS(2273),
    [sym_float] = ACTIONS(2273),
    [sym_char] = ACTIONS(2273),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2273),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(765)] = {
    [sym__expression] = STATE(1595),
    [sym_block] = STATE(1595),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1595),
    [sym_nil] = STATE(1595),
    [sym__atom] = STATE(1595),
    [sym_quoted_atom] = STATE(1595),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1595),
    [sym_charlist] = STATE(1595),
    [sym_sigil] = STATE(1595),
    [sym_list] = STATE(1595),
    [sym_tuple] = STATE(1595),
    [sym_bitstring] = STATE(1595),
    [sym_map] = STATE(1595),
    [sym__nullary_operator] = STATE(1595),
    [sym_unary_operator] = STATE(1595),
    [sym_binary_operator] = STATE(1595),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1595),
    [sym_call] = STATE(1595),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1595),
    [sym_anonymous_function] = STATE(1595),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2425),
    [sym_integer] = ACTIONS(2425),
    [sym_float] = ACTIONS(2425),
    [sym_char] = ACTIONS(2425),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2425),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(766)] = {
    [sym__expression] = STATE(1596),
    [sym_block] = STATE(1596),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1596),
    [sym_nil] = STATE(1596),
    [sym__atom] = STATE(1596),
    [sym_quoted_atom] = STATE(1596),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1596),
    [sym_charlist] = STATE(1596),
    [sym_sigil] = STATE(1596),
    [sym_list] = STATE(1596),
    [sym_tuple] = STATE(1596),
    [sym_bitstring] = STATE(1596),
    [sym_map] = STATE(1596),
    [sym__nullary_operator] = STATE(1596),
    [sym_unary_operator] = STATE(1596),
    [sym_binary_operator] = STATE(1596),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1596),
    [sym_call] = STATE(1596),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1596),
    [sym_anonymous_function] = STATE(1596),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2427),
    [sym_integer] = ACTIONS(2427),
    [sym_float] = ACTIONS(2427),
    [sym_char] = ACTIONS(2427),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2427),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(767)] = {
    [sym__expression] = STATE(4022),
    [sym_block] = STATE(4022),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4022),
    [sym_nil] = STATE(4022),
    [sym__atom] = STATE(4022),
    [sym_quoted_atom] = STATE(4022),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4022),
    [sym_charlist] = STATE(4022),
    [sym_sigil] = STATE(4022),
    [sym_list] = STATE(4022),
    [sym_tuple] = STATE(4022),
    [sym_bitstring] = STATE(4022),
    [sym_map] = STATE(4022),
    [sym__nullary_operator] = STATE(4022),
    [sym_unary_operator] = STATE(4022),
    [sym_binary_operator] = STATE(4022),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4022),
    [sym_call] = STATE(4022),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4022),
    [sym_anonymous_function] = STATE(4022),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2429),
    [sym_integer] = ACTIONS(2429),
    [sym_float] = ACTIONS(2429),
    [sym_char] = ACTIONS(2429),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2429),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(768)] = {
    [sym__expression] = STATE(1598),
    [sym_block] = STATE(1598),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1598),
    [sym_nil] = STATE(1598),
    [sym__atom] = STATE(1598),
    [sym_quoted_atom] = STATE(1598),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1598),
    [sym_charlist] = STATE(1598),
    [sym_sigil] = STATE(1598),
    [sym_list] = STATE(1598),
    [sym_tuple] = STATE(1598),
    [sym_bitstring] = STATE(1598),
    [sym_map] = STATE(1598),
    [sym__nullary_operator] = STATE(1598),
    [sym_unary_operator] = STATE(1598),
    [sym_binary_operator] = STATE(1598),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1598),
    [sym_call] = STATE(1598),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1598),
    [sym_anonymous_function] = STATE(1598),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2431),
    [sym_integer] = ACTIONS(2431),
    [sym_float] = ACTIONS(2431),
    [sym_char] = ACTIONS(2431),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2431),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(769)] = {
    [sym__expression] = STATE(1599),
    [sym_block] = STATE(1599),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1599),
    [sym_nil] = STATE(1599),
    [sym__atom] = STATE(1599),
    [sym_quoted_atom] = STATE(1599),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1599),
    [sym_charlist] = STATE(1599),
    [sym_sigil] = STATE(1599),
    [sym_list] = STATE(1599),
    [sym_tuple] = STATE(1599),
    [sym_bitstring] = STATE(1599),
    [sym_map] = STATE(1599),
    [sym__nullary_operator] = STATE(1599),
    [sym_unary_operator] = STATE(1599),
    [sym_binary_operator] = STATE(1599),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1599),
    [sym_call] = STATE(1599),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1599),
    [sym_anonymous_function] = STATE(1599),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2433),
    [sym_integer] = ACTIONS(2433),
    [sym_float] = ACTIONS(2433),
    [sym_char] = ACTIONS(2433),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2433),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(770)] = {
    [sym__expression] = STATE(1600),
    [sym_block] = STATE(1600),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1600),
    [sym_nil] = STATE(1600),
    [sym__atom] = STATE(1600),
    [sym_quoted_atom] = STATE(1600),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1600),
    [sym_charlist] = STATE(1600),
    [sym_sigil] = STATE(1600),
    [sym_list] = STATE(1600),
    [sym_tuple] = STATE(1600),
    [sym_bitstring] = STATE(1600),
    [sym_map] = STATE(1600),
    [sym__nullary_operator] = STATE(1600),
    [sym_unary_operator] = STATE(1600),
    [sym_binary_operator] = STATE(1600),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1600),
    [sym_call] = STATE(1600),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1600),
    [sym_anonymous_function] = STATE(1600),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2435),
    [sym_integer] = ACTIONS(2435),
    [sym_float] = ACTIONS(2435),
    [sym_char] = ACTIONS(2435),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2435),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(771)] = {
    [sym__expression] = STATE(1601),
    [sym_block] = STATE(1601),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1601),
    [sym_nil] = STATE(1601),
    [sym__atom] = STATE(1601),
    [sym_quoted_atom] = STATE(1601),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1601),
    [sym_charlist] = STATE(1601),
    [sym_sigil] = STATE(1601),
    [sym_list] = STATE(1601),
    [sym_tuple] = STATE(1601),
    [sym_bitstring] = STATE(1601),
    [sym_map] = STATE(1601),
    [sym__nullary_operator] = STATE(1601),
    [sym_unary_operator] = STATE(1601),
    [sym_binary_operator] = STATE(1601),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1601),
    [sym_call] = STATE(1601),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1601),
    [sym_anonymous_function] = STATE(1601),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2437),
    [sym_integer] = ACTIONS(2437),
    [sym_float] = ACTIONS(2437),
    [sym_char] = ACTIONS(2437),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2437),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(772)] = {
    [sym__expression] = STATE(4023),
    [sym_block] = STATE(4023),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4023),
    [sym_nil] = STATE(4023),
    [sym__atom] = STATE(4023),
    [sym_quoted_atom] = STATE(4023),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4023),
    [sym_charlist] = STATE(4023),
    [sym_sigil] = STATE(4023),
    [sym_list] = STATE(4023),
    [sym_tuple] = STATE(4023),
    [sym_bitstring] = STATE(4023),
    [sym_map] = STATE(4023),
    [sym__nullary_operator] = STATE(4023),
    [sym_unary_operator] = STATE(4023),
    [sym_binary_operator] = STATE(4023),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4023),
    [sym_call] = STATE(4023),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4023),
    [sym_anonymous_function] = STATE(4023),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2439),
    [sym_integer] = ACTIONS(2439),
    [sym_float] = ACTIONS(2439),
    [sym_char] = ACTIONS(2439),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2439),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(773)] = {
    [sym__expression] = STATE(1603),
    [sym_block] = STATE(1603),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1603),
    [sym_nil] = STATE(1603),
    [sym__atom] = STATE(1603),
    [sym_quoted_atom] = STATE(1603),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1603),
    [sym_charlist] = STATE(1603),
    [sym_sigil] = STATE(1603),
    [sym_list] = STATE(1603),
    [sym_tuple] = STATE(1603),
    [sym_bitstring] = STATE(1603),
    [sym_map] = STATE(1603),
    [sym__nullary_operator] = STATE(1603),
    [sym_unary_operator] = STATE(1603),
    [sym_binary_operator] = STATE(1603),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1603),
    [sym_call] = STATE(1603),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1603),
    [sym_anonymous_function] = STATE(1603),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2441),
    [sym_integer] = ACTIONS(2441),
    [sym_float] = ACTIONS(2441),
    [sym_char] = ACTIONS(2441),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2441),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(774)] = {
    [sym__expression] = STATE(1604),
    [sym_block] = STATE(1604),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1604),
    [sym_nil] = STATE(1604),
    [sym__atom] = STATE(1604),
    [sym_quoted_atom] = STATE(1604),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1604),
    [sym_charlist] = STATE(1604),
    [sym_sigil] = STATE(1604),
    [sym_list] = STATE(1604),
    [sym_tuple] = STATE(1604),
    [sym_bitstring] = STATE(1604),
    [sym_map] = STATE(1604),
    [sym__nullary_operator] = STATE(1604),
    [sym_unary_operator] = STATE(1604),
    [sym_binary_operator] = STATE(1604),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1604),
    [sym_call] = STATE(1604),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1604),
    [sym_anonymous_function] = STATE(1604),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2443),
    [sym_integer] = ACTIONS(2443),
    [sym_float] = ACTIONS(2443),
    [sym_char] = ACTIONS(2443),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2443),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(775)] = {
    [sym__expression] = STATE(1605),
    [sym_block] = STATE(1605),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1605),
    [sym_nil] = STATE(1605),
    [sym__atom] = STATE(1605),
    [sym_quoted_atom] = STATE(1605),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1605),
    [sym_charlist] = STATE(1605),
    [sym_sigil] = STATE(1605),
    [sym_list] = STATE(1605),
    [sym_tuple] = STATE(1605),
    [sym_bitstring] = STATE(1605),
    [sym_map] = STATE(1605),
    [sym__nullary_operator] = STATE(1605),
    [sym_unary_operator] = STATE(1605),
    [sym_binary_operator] = STATE(1605),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1605),
    [sym_call] = STATE(1605),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1605),
    [sym_anonymous_function] = STATE(1605),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2445),
    [sym_integer] = ACTIONS(2445),
    [sym_float] = ACTIONS(2445),
    [sym_char] = ACTIONS(2445),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2445),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(776)] = {
    [sym__expression] = STATE(1606),
    [sym_block] = STATE(1606),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1606),
    [sym_nil] = STATE(1606),
    [sym__atom] = STATE(1606),
    [sym_quoted_atom] = STATE(1606),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1606),
    [sym_charlist] = STATE(1606),
    [sym_sigil] = STATE(1606),
    [sym_list] = STATE(1606),
    [sym_tuple] = STATE(1606),
    [sym_bitstring] = STATE(1606),
    [sym_map] = STATE(1606),
    [sym__nullary_operator] = STATE(1606),
    [sym_unary_operator] = STATE(1606),
    [sym_binary_operator] = STATE(1606),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1606),
    [sym_call] = STATE(1606),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1606),
    [sym_anonymous_function] = STATE(1606),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2447),
    [sym_integer] = ACTIONS(2447),
    [sym_float] = ACTIONS(2447),
    [sym_char] = ACTIONS(2447),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2447),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(777)] = {
    [sym__expression] = STATE(1607),
    [sym_block] = STATE(1607),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1607),
    [sym_nil] = STATE(1607),
    [sym__atom] = STATE(1607),
    [sym_quoted_atom] = STATE(1607),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1607),
    [sym_charlist] = STATE(1607),
    [sym_sigil] = STATE(1607),
    [sym_list] = STATE(1607),
    [sym_tuple] = STATE(1607),
    [sym_bitstring] = STATE(1607),
    [sym_map] = STATE(1607),
    [sym__nullary_operator] = STATE(1607),
    [sym_unary_operator] = STATE(1607),
    [sym_binary_operator] = STATE(1607),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1607),
    [sym_call] = STATE(1607),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1607),
    [sym_anonymous_function] = STATE(1607),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2449),
    [sym_integer] = ACTIONS(2449),
    [sym_float] = ACTIONS(2449),
    [sym_char] = ACTIONS(2449),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2449),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(778)] = {
    [sym__expression] = STATE(1608),
    [sym_block] = STATE(1608),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1608),
    [sym_nil] = STATE(1608),
    [sym__atom] = STATE(1608),
    [sym_quoted_atom] = STATE(1608),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1608),
    [sym_charlist] = STATE(1608),
    [sym_sigil] = STATE(1608),
    [sym_list] = STATE(1608),
    [sym_tuple] = STATE(1608),
    [sym_bitstring] = STATE(1608),
    [sym_map] = STATE(1608),
    [sym__nullary_operator] = STATE(1608),
    [sym_unary_operator] = STATE(1608),
    [sym_binary_operator] = STATE(1608),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1608),
    [sym_call] = STATE(1608),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1608),
    [sym_anonymous_function] = STATE(1608),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2451),
    [sym_integer] = ACTIONS(2451),
    [sym_float] = ACTIONS(2451),
    [sym_char] = ACTIONS(2451),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2451),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(779)] = {
    [sym__expression] = STATE(1609),
    [sym_block] = STATE(1609),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1609),
    [sym_nil] = STATE(1609),
    [sym__atom] = STATE(1609),
    [sym_quoted_atom] = STATE(1609),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1609),
    [sym_charlist] = STATE(1609),
    [sym_sigil] = STATE(1609),
    [sym_list] = STATE(1609),
    [sym_tuple] = STATE(1609),
    [sym_bitstring] = STATE(1609),
    [sym_map] = STATE(1609),
    [sym__nullary_operator] = STATE(1609),
    [sym_unary_operator] = STATE(1609),
    [sym_binary_operator] = STATE(1609),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1609),
    [sym_call] = STATE(1609),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1609),
    [sym_anonymous_function] = STATE(1609),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2453),
    [sym_integer] = ACTIONS(2453),
    [sym_float] = ACTIONS(2453),
    [sym_char] = ACTIONS(2453),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2453),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(780)] = {
    [sym__expression] = STATE(1610),
    [sym_block] = STATE(1610),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1610),
    [sym_nil] = STATE(1610),
    [sym__atom] = STATE(1610),
    [sym_quoted_atom] = STATE(1610),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1610),
    [sym_charlist] = STATE(1610),
    [sym_sigil] = STATE(1610),
    [sym_list] = STATE(1610),
    [sym_tuple] = STATE(1610),
    [sym_bitstring] = STATE(1610),
    [sym_map] = STATE(1610),
    [sym__nullary_operator] = STATE(1610),
    [sym_unary_operator] = STATE(1610),
    [sym_binary_operator] = STATE(1610),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1610),
    [sym_call] = STATE(1610),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1610),
    [sym_anonymous_function] = STATE(1610),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2455),
    [sym_integer] = ACTIONS(2455),
    [sym_float] = ACTIONS(2455),
    [sym_char] = ACTIONS(2455),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2455),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(781)] = {
    [sym__expression] = STATE(1611),
    [sym_block] = STATE(1611),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1611),
    [sym_nil] = STATE(1611),
    [sym__atom] = STATE(1611),
    [sym_quoted_atom] = STATE(1611),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1611),
    [sym_charlist] = STATE(1611),
    [sym_sigil] = STATE(1611),
    [sym_list] = STATE(1611),
    [sym_tuple] = STATE(1611),
    [sym_bitstring] = STATE(1611),
    [sym_map] = STATE(1611),
    [sym__nullary_operator] = STATE(1611),
    [sym_unary_operator] = STATE(1611),
    [sym_binary_operator] = STATE(1611),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1611),
    [sym_call] = STATE(1611),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1611),
    [sym_anonymous_function] = STATE(1611),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2457),
    [sym_integer] = ACTIONS(2457),
    [sym_float] = ACTIONS(2457),
    [sym_char] = ACTIONS(2457),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2457),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(782)] = {
    [sym__expression] = STATE(1797),
    [sym_block] = STATE(1797),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1797),
    [sym_nil] = STATE(1797),
    [sym__atom] = STATE(1797),
    [sym_quoted_atom] = STATE(1797),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1797),
    [sym_charlist] = STATE(1797),
    [sym_sigil] = STATE(1797),
    [sym_list] = STATE(1797),
    [sym_tuple] = STATE(1797),
    [sym_bitstring] = STATE(1797),
    [sym_map] = STATE(1797),
    [sym__nullary_operator] = STATE(1797),
    [sym_unary_operator] = STATE(1797),
    [sym_binary_operator] = STATE(1797),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1797),
    [sym_call] = STATE(1797),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1797),
    [sym_anonymous_function] = STATE(1797),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2459),
    [sym_integer] = ACTIONS(2459),
    [sym_float] = ACTIONS(2459),
    [sym_char] = ACTIONS(2459),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2459),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(783)] = {
    [sym__expression] = STATE(4026),
    [sym_block] = STATE(4026),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4026),
    [sym_nil] = STATE(4026),
    [sym__atom] = STATE(4026),
    [sym_quoted_atom] = STATE(4026),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4026),
    [sym_charlist] = STATE(4026),
    [sym_sigil] = STATE(4026),
    [sym_list] = STATE(4026),
    [sym_tuple] = STATE(4026),
    [sym_bitstring] = STATE(4026),
    [sym_map] = STATE(4026),
    [sym__nullary_operator] = STATE(4026),
    [sym_unary_operator] = STATE(4026),
    [sym_binary_operator] = STATE(4026),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4026),
    [sym_call] = STATE(4026),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4026),
    [sym_anonymous_function] = STATE(4026),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2461),
    [sym_integer] = ACTIONS(2461),
    [sym_float] = ACTIONS(2461),
    [sym_char] = ACTIONS(2461),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2461),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(784)] = {
    [sym__expression] = STATE(4027),
    [sym_block] = STATE(4027),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4027),
    [sym_nil] = STATE(4027),
    [sym__atom] = STATE(4027),
    [sym_quoted_atom] = STATE(4027),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4027),
    [sym_charlist] = STATE(4027),
    [sym_sigil] = STATE(4027),
    [sym_list] = STATE(4027),
    [sym_tuple] = STATE(4027),
    [sym_bitstring] = STATE(4027),
    [sym_map] = STATE(4027),
    [sym__nullary_operator] = STATE(4027),
    [sym_unary_operator] = STATE(4027),
    [sym_binary_operator] = STATE(4027),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4027),
    [sym_call] = STATE(4027),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4027),
    [sym_anonymous_function] = STATE(4027),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2463),
    [sym_integer] = ACTIONS(2463),
    [sym_float] = ACTIONS(2463),
    [sym_char] = ACTIONS(2463),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2463),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(785)] = {
    [sym__expression] = STATE(4028),
    [sym_block] = STATE(4028),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4028),
    [sym_nil] = STATE(4028),
    [sym__atom] = STATE(4028),
    [sym_quoted_atom] = STATE(4028),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4028),
    [sym_charlist] = STATE(4028),
    [sym_sigil] = STATE(4028),
    [sym_list] = STATE(4028),
    [sym_tuple] = STATE(4028),
    [sym_bitstring] = STATE(4028),
    [sym_map] = STATE(4028),
    [sym__nullary_operator] = STATE(4028),
    [sym_unary_operator] = STATE(4028),
    [sym_binary_operator] = STATE(4028),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4028),
    [sym_call] = STATE(4028),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4028),
    [sym_anonymous_function] = STATE(4028),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2465),
    [sym_integer] = ACTIONS(2465),
    [sym_float] = ACTIONS(2465),
    [sym_char] = ACTIONS(2465),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2465),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(786)] = {
    [sym__expression] = STATE(4038),
    [sym_block] = STATE(4038),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4038),
    [sym_nil] = STATE(4038),
    [sym__atom] = STATE(4038),
    [sym_quoted_atom] = STATE(4038),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4038),
    [sym_charlist] = STATE(4038),
    [sym_sigil] = STATE(4038),
    [sym_list] = STATE(4038),
    [sym_tuple] = STATE(4038),
    [sym_bitstring] = STATE(4038),
    [sym_map] = STATE(4038),
    [sym__nullary_operator] = STATE(4038),
    [sym_unary_operator] = STATE(4038),
    [sym_binary_operator] = STATE(4038),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4038),
    [sym_call] = STATE(4038),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4038),
    [sym_anonymous_function] = STATE(4038),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2467),
    [sym_integer] = ACTIONS(2467),
    [sym_float] = ACTIONS(2467),
    [sym_char] = ACTIONS(2467),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2467),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(787)] = {
    [sym__expression] = STATE(4042),
    [sym_block] = STATE(4042),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4042),
    [sym_nil] = STATE(4042),
    [sym__atom] = STATE(4042),
    [sym_quoted_atom] = STATE(4042),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4042),
    [sym_charlist] = STATE(4042),
    [sym_sigil] = STATE(4042),
    [sym_list] = STATE(4042),
    [sym_tuple] = STATE(4042),
    [sym_bitstring] = STATE(4042),
    [sym_map] = STATE(4042),
    [sym__nullary_operator] = STATE(4042),
    [sym_unary_operator] = STATE(4042),
    [sym_binary_operator] = STATE(4042),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4042),
    [sym_call] = STATE(4042),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4042),
    [sym_anonymous_function] = STATE(4042),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2469),
    [sym_integer] = ACTIONS(2469),
    [sym_float] = ACTIONS(2469),
    [sym_char] = ACTIONS(2469),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2469),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(788)] = {
    [sym__expression] = STATE(4043),
    [sym_block] = STATE(4043),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4043),
    [sym_nil] = STATE(4043),
    [sym__atom] = STATE(4043),
    [sym_quoted_atom] = STATE(4043),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4043),
    [sym_charlist] = STATE(4043),
    [sym_sigil] = STATE(4043),
    [sym_list] = STATE(4043),
    [sym_tuple] = STATE(4043),
    [sym_bitstring] = STATE(4043),
    [sym_map] = STATE(4043),
    [sym__nullary_operator] = STATE(4043),
    [sym_unary_operator] = STATE(4043),
    [sym_binary_operator] = STATE(4043),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4043),
    [sym_call] = STATE(4043),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4043),
    [sym_anonymous_function] = STATE(4043),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2471),
    [sym_integer] = ACTIONS(2471),
    [sym_float] = ACTIONS(2471),
    [sym_char] = ACTIONS(2471),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2471),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(789)] = {
    [sym__expression] = STATE(4050),
    [sym_block] = STATE(4050),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4050),
    [sym_nil] = STATE(4050),
    [sym__atom] = STATE(4050),
    [sym_quoted_atom] = STATE(4050),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4050),
    [sym_charlist] = STATE(4050),
    [sym_sigil] = STATE(4050),
    [sym_list] = STATE(4050),
    [sym_tuple] = STATE(4050),
    [sym_bitstring] = STATE(4050),
    [sym_map] = STATE(4050),
    [sym__nullary_operator] = STATE(4050),
    [sym_unary_operator] = STATE(4050),
    [sym_binary_operator] = STATE(4050),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4050),
    [sym_call] = STATE(4050),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4050),
    [sym_anonymous_function] = STATE(4050),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2473),
    [sym_integer] = ACTIONS(2473),
    [sym_float] = ACTIONS(2473),
    [sym_char] = ACTIONS(2473),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2473),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(790)] = {
    [sym__expression] = STATE(4051),
    [sym_block] = STATE(4051),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4051),
    [sym_nil] = STATE(4051),
    [sym__atom] = STATE(4051),
    [sym_quoted_atom] = STATE(4051),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4051),
    [sym_charlist] = STATE(4051),
    [sym_sigil] = STATE(4051),
    [sym_list] = STATE(4051),
    [sym_tuple] = STATE(4051),
    [sym_bitstring] = STATE(4051),
    [sym_map] = STATE(4051),
    [sym__nullary_operator] = STATE(4051),
    [sym_unary_operator] = STATE(4051),
    [sym_binary_operator] = STATE(4051),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4051),
    [sym_call] = STATE(4051),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4051),
    [sym_anonymous_function] = STATE(4051),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2475),
    [sym_integer] = ACTIONS(2475),
    [sym_float] = ACTIONS(2475),
    [sym_char] = ACTIONS(2475),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2475),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(791)] = {
    [sym__expression] = STATE(4052),
    [sym_block] = STATE(4052),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4052),
    [sym_nil] = STATE(4052),
    [sym__atom] = STATE(4052),
    [sym_quoted_atom] = STATE(4052),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4052),
    [sym_charlist] = STATE(4052),
    [sym_sigil] = STATE(4052),
    [sym_list] = STATE(4052),
    [sym_tuple] = STATE(4052),
    [sym_bitstring] = STATE(4052),
    [sym_map] = STATE(4052),
    [sym__nullary_operator] = STATE(4052),
    [sym_unary_operator] = STATE(4052),
    [sym_binary_operator] = STATE(4052),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4052),
    [sym_call] = STATE(4052),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4052),
    [sym_anonymous_function] = STATE(4052),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2477),
    [sym_integer] = ACTIONS(2477),
    [sym_float] = ACTIONS(2477),
    [sym_char] = ACTIONS(2477),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2477),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(792)] = {
    [sym__expression] = STATE(3859),
    [sym_block] = STATE(3859),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3859),
    [sym_nil] = STATE(3859),
    [sym__atom] = STATE(3859),
    [sym_quoted_atom] = STATE(3859),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3859),
    [sym_charlist] = STATE(3859),
    [sym_sigil] = STATE(3859),
    [sym_list] = STATE(3859),
    [sym_tuple] = STATE(3859),
    [sym_bitstring] = STATE(3859),
    [sym_map] = STATE(3859),
    [sym__nullary_operator] = STATE(3859),
    [sym_unary_operator] = STATE(3859),
    [sym_binary_operator] = STATE(3859),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3859),
    [sym_call] = STATE(3859),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3859),
    [sym_anonymous_function] = STATE(3859),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2479),
    [sym_integer] = ACTIONS(2479),
    [sym_float] = ACTIONS(2479),
    [sym_char] = ACTIONS(2479),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2479),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(793)] = {
    [sym__expression] = STATE(4053),
    [sym_block] = STATE(4053),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4053),
    [sym_nil] = STATE(4053),
    [sym__atom] = STATE(4053),
    [sym_quoted_atom] = STATE(4053),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4053),
    [sym_charlist] = STATE(4053),
    [sym_sigil] = STATE(4053),
    [sym_list] = STATE(4053),
    [sym_tuple] = STATE(4053),
    [sym_bitstring] = STATE(4053),
    [sym_map] = STATE(4053),
    [sym__nullary_operator] = STATE(4053),
    [sym_unary_operator] = STATE(4053),
    [sym_binary_operator] = STATE(4053),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4053),
    [sym_call] = STATE(4053),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4053),
    [sym_anonymous_function] = STATE(4053),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2481),
    [sym_integer] = ACTIONS(2481),
    [sym_float] = ACTIONS(2481),
    [sym_char] = ACTIONS(2481),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2481),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(794)] = {
    [sym__expression] = STATE(4054),
    [sym_block] = STATE(4054),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4054),
    [sym_nil] = STATE(4054),
    [sym__atom] = STATE(4054),
    [sym_quoted_atom] = STATE(4054),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4054),
    [sym_charlist] = STATE(4054),
    [sym_sigil] = STATE(4054),
    [sym_list] = STATE(4054),
    [sym_tuple] = STATE(4054),
    [sym_bitstring] = STATE(4054),
    [sym_map] = STATE(4054),
    [sym__nullary_operator] = STATE(4054),
    [sym_unary_operator] = STATE(4054),
    [sym_binary_operator] = STATE(4054),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4054),
    [sym_call] = STATE(4054),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4054),
    [sym_anonymous_function] = STATE(4054),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2483),
    [sym_integer] = ACTIONS(2483),
    [sym_float] = ACTIONS(2483),
    [sym_char] = ACTIONS(2483),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2483),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(795)] = {
    [sym__expression] = STATE(4055),
    [sym_block] = STATE(4055),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4055),
    [sym_nil] = STATE(4055),
    [sym__atom] = STATE(4055),
    [sym_quoted_atom] = STATE(4055),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4055),
    [sym_charlist] = STATE(4055),
    [sym_sigil] = STATE(4055),
    [sym_list] = STATE(4055),
    [sym_tuple] = STATE(4055),
    [sym_bitstring] = STATE(4055),
    [sym_map] = STATE(4055),
    [sym__nullary_operator] = STATE(4055),
    [sym_unary_operator] = STATE(4055),
    [sym_binary_operator] = STATE(4055),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4055),
    [sym_call] = STATE(4055),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4055),
    [sym_anonymous_function] = STATE(4055),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2485),
    [sym_integer] = ACTIONS(2485),
    [sym_float] = ACTIONS(2485),
    [sym_char] = ACTIONS(2485),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2485),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(796)] = {
    [sym__expression] = STATE(4056),
    [sym_block] = STATE(4056),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4056),
    [sym_nil] = STATE(4056),
    [sym__atom] = STATE(4056),
    [sym_quoted_atom] = STATE(4056),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4056),
    [sym_charlist] = STATE(4056),
    [sym_sigil] = STATE(4056),
    [sym_list] = STATE(4056),
    [sym_tuple] = STATE(4056),
    [sym_bitstring] = STATE(4056),
    [sym_map] = STATE(4056),
    [sym__nullary_operator] = STATE(4056),
    [sym_unary_operator] = STATE(4056),
    [sym_binary_operator] = STATE(4056),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4056),
    [sym_call] = STATE(4056),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4056),
    [sym_anonymous_function] = STATE(4056),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2487),
    [sym_integer] = ACTIONS(2487),
    [sym_float] = ACTIONS(2487),
    [sym_char] = ACTIONS(2487),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2487),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(797)] = {
    [sym__expression] = STATE(4057),
    [sym_block] = STATE(4057),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(4057),
    [sym_nil] = STATE(4057),
    [sym__atom] = STATE(4057),
    [sym_quoted_atom] = STATE(4057),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(4057),
    [sym_charlist] = STATE(4057),
    [sym_sigil] = STATE(4057),
    [sym_list] = STATE(4057),
    [sym_tuple] = STATE(4057),
    [sym_bitstring] = STATE(4057),
    [sym_map] = STATE(4057),
    [sym__nullary_operator] = STATE(4057),
    [sym_unary_operator] = STATE(4057),
    [sym_binary_operator] = STATE(4057),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(4057),
    [sym_call] = STATE(4057),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(4057),
    [sym_anonymous_function] = STATE(4057),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2489),
    [sym_integer] = ACTIONS(2489),
    [sym_float] = ACTIONS(2489),
    [sym_char] = ACTIONS(2489),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2489),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(798)] = {
    [sym__expression] = STATE(2487),
    [sym_block] = STATE(2487),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2487),
    [sym_nil] = STATE(2487),
    [sym__atom] = STATE(2487),
    [sym_quoted_atom] = STATE(2487),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2487),
    [sym_charlist] = STATE(2487),
    [sym_sigil] = STATE(2487),
    [sym_list] = STATE(2487),
    [sym_tuple] = STATE(2487),
    [sym_bitstring] = STATE(2487),
    [sym_map] = STATE(2487),
    [sym__nullary_operator] = STATE(2487),
    [sym_unary_operator] = STATE(2487),
    [sym_binary_operator] = STATE(2487),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2487),
    [sym_call] = STATE(2487),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2487),
    [sym_anonymous_function] = STATE(2487),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(970),
    [sym_integer] = ACTIONS(970),
    [sym_float] = ACTIONS(970),
    [sym_char] = ACTIONS(970),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(970),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(799)] = {
    [sym__expression] = STATE(4442),
    [sym_block] = STATE(4442),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4442),
    [sym_nil] = STATE(4442),
    [sym__atom] = STATE(4442),
    [sym_quoted_atom] = STATE(4442),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4442),
    [sym_charlist] = STATE(4442),
    [sym_sigil] = STATE(4442),
    [sym_list] = STATE(4442),
    [sym_tuple] = STATE(4442),
    [sym_bitstring] = STATE(4442),
    [sym_map] = STATE(4442),
    [sym__nullary_operator] = STATE(4442),
    [sym_unary_operator] = STATE(4442),
    [sym_binary_operator] = STATE(4442),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4442),
    [sym_call] = STATE(4442),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4442),
    [sym_anonymous_function] = STATE(4442),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(1363),
    [sym_integer] = ACTIONS(1363),
    [sym_float] = ACTIONS(1363),
    [sym_char] = ACTIONS(1363),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(1363),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(800)] = {
    [sym__expression] = STATE(3011),
    [sym_block] = STATE(3011),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(3011),
    [sym_nil] = STATE(3011),
    [sym__atom] = STATE(3011),
    [sym_quoted_atom] = STATE(3011),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3011),
    [sym_charlist] = STATE(3011),
    [sym_sigil] = STATE(3011),
    [sym_list] = STATE(3011),
    [sym_tuple] = STATE(3011),
    [sym_bitstring] = STATE(3011),
    [sym_map] = STATE(3011),
    [sym__nullary_operator] = STATE(3011),
    [sym_unary_operator] = STATE(3011),
    [sym_binary_operator] = STATE(3011),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3011),
    [sym_call] = STATE(3011),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3011),
    [sym_anonymous_function] = STATE(3011),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2491),
    [sym_integer] = ACTIONS(2491),
    [sym_float] = ACTIONS(2491),
    [sym_char] = ACTIONS(2491),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2491),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(801)] = {
    [sym__expression] = STATE(3016),
    [sym_block] = STATE(3016),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(3016),
    [sym_nil] = STATE(3016),
    [sym__atom] = STATE(3016),
    [sym_quoted_atom] = STATE(3016),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3016),
    [sym_charlist] = STATE(3016),
    [sym_sigil] = STATE(3016),
    [sym_list] = STATE(3016),
    [sym_tuple] = STATE(3016),
    [sym_bitstring] = STATE(3016),
    [sym_map] = STATE(3016),
    [sym__nullary_operator] = STATE(3016),
    [sym_unary_operator] = STATE(3016),
    [sym_binary_operator] = STATE(3016),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3016),
    [sym_call] = STATE(3016),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3016),
    [sym_anonymous_function] = STATE(3016),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2493),
    [sym_integer] = ACTIONS(2493),
    [sym_float] = ACTIONS(2493),
    [sym_char] = ACTIONS(2493),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2493),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(802)] = {
    [sym__expression] = STATE(3058),
    [sym_block] = STATE(3058),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(3058),
    [sym_nil] = STATE(3058),
    [sym__atom] = STATE(3058),
    [sym_quoted_atom] = STATE(3058),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3058),
    [sym_charlist] = STATE(3058),
    [sym_sigil] = STATE(3058),
    [sym_list] = STATE(3058),
    [sym_tuple] = STATE(3058),
    [sym_bitstring] = STATE(3058),
    [sym_map] = STATE(3058),
    [sym__nullary_operator] = STATE(3058),
    [sym_unary_operator] = STATE(3058),
    [sym_binary_operator] = STATE(3058),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3058),
    [sym_call] = STATE(3058),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3058),
    [sym_anonymous_function] = STATE(3058),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2495),
    [sym_integer] = ACTIONS(2495),
    [sym_float] = ACTIONS(2495),
    [sym_char] = ACTIONS(2495),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2495),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(803)] = {
    [sym__expression] = STATE(3059),
    [sym_block] = STATE(3059),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(3059),
    [sym_nil] = STATE(3059),
    [sym__atom] = STATE(3059),
    [sym_quoted_atom] = STATE(3059),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3059),
    [sym_charlist] = STATE(3059),
    [sym_sigil] = STATE(3059),
    [sym_list] = STATE(3059),
    [sym_tuple] = STATE(3059),
    [sym_bitstring] = STATE(3059),
    [sym_map] = STATE(3059),
    [sym__nullary_operator] = STATE(3059),
    [sym_unary_operator] = STATE(3059),
    [sym_binary_operator] = STATE(3059),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3059),
    [sym_call] = STATE(3059),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3059),
    [sym_anonymous_function] = STATE(3059),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2497),
    [sym_integer] = ACTIONS(2497),
    [sym_float] = ACTIONS(2497),
    [sym_char] = ACTIONS(2497),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2497),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(804)] = {
    [sym__expression] = STATE(4448),
    [sym_block] = STATE(4448),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4448),
    [sym_nil] = STATE(4448),
    [sym__atom] = STATE(4448),
    [sym_quoted_atom] = STATE(4448),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4448),
    [sym_charlist] = STATE(4448),
    [sym_sigil] = STATE(4448),
    [sym_list] = STATE(4448),
    [sym_tuple] = STATE(4448),
    [sym_bitstring] = STATE(4448),
    [sym_map] = STATE(4448),
    [sym__nullary_operator] = STATE(4448),
    [sym_unary_operator] = STATE(4448),
    [sym_binary_operator] = STATE(4448),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4448),
    [sym_call] = STATE(4448),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4448),
    [sym_anonymous_function] = STATE(4448),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2499),
    [sym_integer] = ACTIONS(2499),
    [sym_float] = ACTIONS(2499),
    [sym_char] = ACTIONS(2499),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2499),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(805)] = {
    [sym__expression] = STATE(4450),
    [sym_block] = STATE(4450),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4450),
    [sym_nil] = STATE(4450),
    [sym__atom] = STATE(4450),
    [sym_quoted_atom] = STATE(4450),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4450),
    [sym_charlist] = STATE(4450),
    [sym_sigil] = STATE(4450),
    [sym_list] = STATE(4450),
    [sym_tuple] = STATE(4450),
    [sym_bitstring] = STATE(4450),
    [sym_map] = STATE(4450),
    [sym__nullary_operator] = STATE(4450),
    [sym_unary_operator] = STATE(4450),
    [sym_binary_operator] = STATE(4450),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4450),
    [sym_call] = STATE(4450),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4450),
    [sym_anonymous_function] = STATE(4450),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2501),
    [sym_integer] = ACTIONS(2501),
    [sym_float] = ACTIONS(2501),
    [sym_char] = ACTIONS(2501),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2501),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(806)] = {
    [sym__expression] = STATE(4453),
    [sym_block] = STATE(4453),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4453),
    [sym_nil] = STATE(4453),
    [sym__atom] = STATE(4453),
    [sym_quoted_atom] = STATE(4453),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4453),
    [sym_charlist] = STATE(4453),
    [sym_sigil] = STATE(4453),
    [sym_list] = STATE(4453),
    [sym_tuple] = STATE(4453),
    [sym_bitstring] = STATE(4453),
    [sym_map] = STATE(4453),
    [sym__nullary_operator] = STATE(4453),
    [sym_unary_operator] = STATE(4453),
    [sym_binary_operator] = STATE(4453),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4453),
    [sym_call] = STATE(4453),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4453),
    [sym_anonymous_function] = STATE(4453),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2503),
    [sym_integer] = ACTIONS(2503),
    [sym_float] = ACTIONS(2503),
    [sym_char] = ACTIONS(2503),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2503),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(807)] = {
    [sym__expression] = STATE(4456),
    [sym_block] = STATE(4456),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4456),
    [sym_nil] = STATE(4456),
    [sym__atom] = STATE(4456),
    [sym_quoted_atom] = STATE(4456),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4456),
    [sym_charlist] = STATE(4456),
    [sym_sigil] = STATE(4456),
    [sym_list] = STATE(4456),
    [sym_tuple] = STATE(4456),
    [sym_bitstring] = STATE(4456),
    [sym_map] = STATE(4456),
    [sym__nullary_operator] = STATE(4456),
    [sym_unary_operator] = STATE(4456),
    [sym_binary_operator] = STATE(4456),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4456),
    [sym_call] = STATE(4456),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4456),
    [sym_anonymous_function] = STATE(4456),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2505),
    [sym_integer] = ACTIONS(2505),
    [sym_float] = ACTIONS(2505),
    [sym_char] = ACTIONS(2505),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2505),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(808)] = {
    [sym__expression] = STATE(4457),
    [sym_block] = STATE(4457),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4457),
    [sym_nil] = STATE(4457),
    [sym__atom] = STATE(4457),
    [sym_quoted_atom] = STATE(4457),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4457),
    [sym_charlist] = STATE(4457),
    [sym_sigil] = STATE(4457),
    [sym_list] = STATE(4457),
    [sym_tuple] = STATE(4457),
    [sym_bitstring] = STATE(4457),
    [sym_map] = STATE(4457),
    [sym__nullary_operator] = STATE(4457),
    [sym_unary_operator] = STATE(4457),
    [sym_binary_operator] = STATE(4457),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4457),
    [sym_call] = STATE(4457),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4457),
    [sym_anonymous_function] = STATE(4457),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2507),
    [sym_integer] = ACTIONS(2507),
    [sym_float] = ACTIONS(2507),
    [sym_char] = ACTIONS(2507),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2507),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(809)] = {
    [sym__expression] = STATE(4458),
    [sym_block] = STATE(4458),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4458),
    [sym_nil] = STATE(4458),
    [sym__atom] = STATE(4458),
    [sym_quoted_atom] = STATE(4458),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4458),
    [sym_charlist] = STATE(4458),
    [sym_sigil] = STATE(4458),
    [sym_list] = STATE(4458),
    [sym_tuple] = STATE(4458),
    [sym_bitstring] = STATE(4458),
    [sym_map] = STATE(4458),
    [sym__nullary_operator] = STATE(4458),
    [sym_unary_operator] = STATE(4458),
    [sym_binary_operator] = STATE(4458),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4458),
    [sym_call] = STATE(4458),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4458),
    [sym_anonymous_function] = STATE(4458),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2509),
    [sym_integer] = ACTIONS(2509),
    [sym_float] = ACTIONS(2509),
    [sym_char] = ACTIONS(2509),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2509),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(810)] = {
    [sym__expression] = STATE(4514),
    [sym_block] = STATE(4514),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4514),
    [sym_nil] = STATE(4514),
    [sym__atom] = STATE(4514),
    [sym_quoted_atom] = STATE(4514),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4514),
    [sym_charlist] = STATE(4514),
    [sym_sigil] = STATE(4514),
    [sym_list] = STATE(4514),
    [sym_tuple] = STATE(4514),
    [sym_bitstring] = STATE(4514),
    [sym_map] = STATE(4514),
    [sym__nullary_operator] = STATE(4514),
    [sym_unary_operator] = STATE(4514),
    [sym_binary_operator] = STATE(4514),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4514),
    [sym_call] = STATE(4514),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4514),
    [sym_anonymous_function] = STATE(4514),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2511),
    [sym_integer] = ACTIONS(2511),
    [sym_float] = ACTIONS(2511),
    [sym_char] = ACTIONS(2511),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2511),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(811)] = {
    [sym__expression] = STATE(4534),
    [sym_block] = STATE(4534),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4534),
    [sym_nil] = STATE(4534),
    [sym__atom] = STATE(4534),
    [sym_quoted_atom] = STATE(4534),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4534),
    [sym_charlist] = STATE(4534),
    [sym_sigil] = STATE(4534),
    [sym_list] = STATE(4534),
    [sym_tuple] = STATE(4534),
    [sym_bitstring] = STATE(4534),
    [sym_map] = STATE(4534),
    [sym__nullary_operator] = STATE(4534),
    [sym_unary_operator] = STATE(4534),
    [sym_binary_operator] = STATE(4534),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4534),
    [sym_call] = STATE(4534),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4534),
    [sym_anonymous_function] = STATE(4534),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2513),
    [sym_integer] = ACTIONS(2513),
    [sym_float] = ACTIONS(2513),
    [sym_char] = ACTIONS(2513),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2513),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(812)] = {
    [sym__expression] = STATE(4536),
    [sym_block] = STATE(4536),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4536),
    [sym_nil] = STATE(4536),
    [sym__atom] = STATE(4536),
    [sym_quoted_atom] = STATE(4536),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4536),
    [sym_charlist] = STATE(4536),
    [sym_sigil] = STATE(4536),
    [sym_list] = STATE(4536),
    [sym_tuple] = STATE(4536),
    [sym_bitstring] = STATE(4536),
    [sym_map] = STATE(4536),
    [sym__nullary_operator] = STATE(4536),
    [sym_unary_operator] = STATE(4536),
    [sym_binary_operator] = STATE(4536),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4536),
    [sym_call] = STATE(4536),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4536),
    [sym_anonymous_function] = STATE(4536),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2515),
    [sym_integer] = ACTIONS(2515),
    [sym_float] = ACTIONS(2515),
    [sym_char] = ACTIONS(2515),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2515),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(813)] = {
    [sym__expression] = STATE(4537),
    [sym_block] = STATE(4537),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4537),
    [sym_nil] = STATE(4537),
    [sym__atom] = STATE(4537),
    [sym_quoted_atom] = STATE(4537),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4537),
    [sym_charlist] = STATE(4537),
    [sym_sigil] = STATE(4537),
    [sym_list] = STATE(4537),
    [sym_tuple] = STATE(4537),
    [sym_bitstring] = STATE(4537),
    [sym_map] = STATE(4537),
    [sym__nullary_operator] = STATE(4537),
    [sym_unary_operator] = STATE(4537),
    [sym_binary_operator] = STATE(4537),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4537),
    [sym_call] = STATE(4537),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4537),
    [sym_anonymous_function] = STATE(4537),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2517),
    [sym_integer] = ACTIONS(2517),
    [sym_float] = ACTIONS(2517),
    [sym_char] = ACTIONS(2517),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2517),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(814)] = {
    [sym__expression] = STATE(4538),
    [sym_block] = STATE(4538),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4538),
    [sym_nil] = STATE(4538),
    [sym__atom] = STATE(4538),
    [sym_quoted_atom] = STATE(4538),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4538),
    [sym_charlist] = STATE(4538),
    [sym_sigil] = STATE(4538),
    [sym_list] = STATE(4538),
    [sym_tuple] = STATE(4538),
    [sym_bitstring] = STATE(4538),
    [sym_map] = STATE(4538),
    [sym__nullary_operator] = STATE(4538),
    [sym_unary_operator] = STATE(4538),
    [sym_binary_operator] = STATE(4538),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4538),
    [sym_call] = STATE(4538),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4538),
    [sym_anonymous_function] = STATE(4538),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2519),
    [sym_integer] = ACTIONS(2519),
    [sym_float] = ACTIONS(2519),
    [sym_char] = ACTIONS(2519),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2519),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(815)] = {
    [sym__expression] = STATE(4539),
    [sym_block] = STATE(4539),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4539),
    [sym_nil] = STATE(4539),
    [sym__atom] = STATE(4539),
    [sym_quoted_atom] = STATE(4539),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4539),
    [sym_charlist] = STATE(4539),
    [sym_sigil] = STATE(4539),
    [sym_list] = STATE(4539),
    [sym_tuple] = STATE(4539),
    [sym_bitstring] = STATE(4539),
    [sym_map] = STATE(4539),
    [sym__nullary_operator] = STATE(4539),
    [sym_unary_operator] = STATE(4539),
    [sym_binary_operator] = STATE(4539),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4539),
    [sym_call] = STATE(4539),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4539),
    [sym_anonymous_function] = STATE(4539),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2521),
    [sym_integer] = ACTIONS(2521),
    [sym_float] = ACTIONS(2521),
    [sym_char] = ACTIONS(2521),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2521),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(816)] = {
    [sym__expression] = STATE(4542),
    [sym_block] = STATE(4542),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4542),
    [sym_nil] = STATE(4542),
    [sym__atom] = STATE(4542),
    [sym_quoted_atom] = STATE(4542),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4542),
    [sym_charlist] = STATE(4542),
    [sym_sigil] = STATE(4542),
    [sym_list] = STATE(4542),
    [sym_tuple] = STATE(4542),
    [sym_bitstring] = STATE(4542),
    [sym_map] = STATE(4542),
    [sym__nullary_operator] = STATE(4542),
    [sym_unary_operator] = STATE(4542),
    [sym_binary_operator] = STATE(4542),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4542),
    [sym_call] = STATE(4542),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4542),
    [sym_anonymous_function] = STATE(4542),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2523),
    [sym_integer] = ACTIONS(2523),
    [sym_float] = ACTIONS(2523),
    [sym_char] = ACTIONS(2523),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2523),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(817)] = {
    [sym__expression] = STATE(4433),
    [sym_block] = STATE(4433),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4433),
    [sym_nil] = STATE(4433),
    [sym__atom] = STATE(4433),
    [sym_quoted_atom] = STATE(4433),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4433),
    [sym_charlist] = STATE(4433),
    [sym_sigil] = STATE(4433),
    [sym_list] = STATE(4433),
    [sym_tuple] = STATE(4433),
    [sym_bitstring] = STATE(4433),
    [sym_map] = STATE(4433),
    [sym__nullary_operator] = STATE(4433),
    [sym_unary_operator] = STATE(4433),
    [sym_binary_operator] = STATE(4433),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4433),
    [sym_call] = STATE(4433),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4433),
    [sym_anonymous_function] = STATE(4433),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2525),
    [sym_integer] = ACTIONS(2525),
    [sym_float] = ACTIONS(2525),
    [sym_char] = ACTIONS(2525),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2525),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(818)] = {
    [sym__expression] = STATE(4438),
    [sym_block] = STATE(4438),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4438),
    [sym_nil] = STATE(4438),
    [sym__atom] = STATE(4438),
    [sym_quoted_atom] = STATE(4438),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4438),
    [sym_charlist] = STATE(4438),
    [sym_sigil] = STATE(4438),
    [sym_list] = STATE(4438),
    [sym_tuple] = STATE(4438),
    [sym_bitstring] = STATE(4438),
    [sym_map] = STATE(4438),
    [sym__nullary_operator] = STATE(4438),
    [sym_unary_operator] = STATE(4438),
    [sym_binary_operator] = STATE(4438),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4438),
    [sym_call] = STATE(4438),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4438),
    [sym_anonymous_function] = STATE(4438),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2527),
    [sym_integer] = ACTIONS(2527),
    [sym_float] = ACTIONS(2527),
    [sym_char] = ACTIONS(2527),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2527),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(819)] = {
    [sym__expression] = STATE(3080),
    [sym_block] = STATE(3080),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(3080),
    [sym_nil] = STATE(3080),
    [sym__atom] = STATE(3080),
    [sym_quoted_atom] = STATE(3080),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3080),
    [sym_charlist] = STATE(3080),
    [sym_sigil] = STATE(3080),
    [sym_list] = STATE(3080),
    [sym_tuple] = STATE(3080),
    [sym_bitstring] = STATE(3080),
    [sym_map] = STATE(3080),
    [sym__nullary_operator] = STATE(3080),
    [sym_unary_operator] = STATE(3080),
    [sym_binary_operator] = STATE(3080),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3080),
    [sym_call] = STATE(3080),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3080),
    [sym_anonymous_function] = STATE(3080),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2529),
    [sym_integer] = ACTIONS(2529),
    [sym_float] = ACTIONS(2529),
    [sym_char] = ACTIONS(2529),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2529),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(820)] = {
    [sym__expression] = STATE(2587),
    [sym_block] = STATE(2587),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2587),
    [sym_nil] = STATE(2587),
    [sym__atom] = STATE(2587),
    [sym_quoted_atom] = STATE(2587),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2587),
    [sym_charlist] = STATE(2587),
    [sym_sigil] = STATE(2587),
    [sym_list] = STATE(2587),
    [sym_tuple] = STATE(2587),
    [sym_bitstring] = STATE(2587),
    [sym_map] = STATE(2587),
    [sym__nullary_operator] = STATE(2587),
    [sym_unary_operator] = STATE(2587),
    [sym_binary_operator] = STATE(2587),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2587),
    [sym_call] = STATE(2587),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2587),
    [sym_anonymous_function] = STATE(2587),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(1347),
    [sym_integer] = ACTIONS(1347),
    [sym_float] = ACTIONS(1347),
    [sym_char] = ACTIONS(1347),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1347),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(821)] = {
    [sym__expression] = STATE(2184),
    [sym_block] = STATE(2184),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(2184),
    [sym_nil] = STATE(2184),
    [sym__atom] = STATE(2184),
    [sym_quoted_atom] = STATE(2184),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2184),
    [sym_charlist] = STATE(2184),
    [sym_sigil] = STATE(2184),
    [sym_list] = STATE(2184),
    [sym_tuple] = STATE(2184),
    [sym_bitstring] = STATE(2184),
    [sym_map] = STATE(2184),
    [sym__nullary_operator] = STATE(2184),
    [sym_unary_operator] = STATE(2184),
    [sym_binary_operator] = STATE(2184),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2184),
    [sym_call] = STATE(2184),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2184),
    [sym_anonymous_function] = STATE(2184),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2055),
    [sym_integer] = ACTIONS(2055),
    [sym_float] = ACTIONS(2055),
    [sym_char] = ACTIONS(2055),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2055),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(822)] = {
    [sym__expression] = STATE(2185),
    [sym_block] = STATE(2185),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(2185),
    [sym_nil] = STATE(2185),
    [sym__atom] = STATE(2185),
    [sym_quoted_atom] = STATE(2185),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2185),
    [sym_charlist] = STATE(2185),
    [sym_sigil] = STATE(2185),
    [sym_list] = STATE(2185),
    [sym_tuple] = STATE(2185),
    [sym_bitstring] = STATE(2185),
    [sym_map] = STATE(2185),
    [sym__nullary_operator] = STATE(2185),
    [sym_unary_operator] = STATE(2185),
    [sym_binary_operator] = STATE(2185),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2185),
    [sym_call] = STATE(2185),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2185),
    [sym_anonymous_function] = STATE(2185),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2057),
    [sym_integer] = ACTIONS(2057),
    [sym_float] = ACTIONS(2057),
    [sym_char] = ACTIONS(2057),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(823)] = {
    [sym__expression] = STATE(2257),
    [sym_block] = STATE(2257),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(2257),
    [sym_nil] = STATE(2257),
    [sym__atom] = STATE(2257),
    [sym_quoted_atom] = STATE(2257),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2257),
    [sym_charlist] = STATE(2257),
    [sym_sigil] = STATE(2257),
    [sym_list] = STATE(2257),
    [sym_tuple] = STATE(2257),
    [sym_bitstring] = STATE(2257),
    [sym_map] = STATE(2257),
    [sym__nullary_operator] = STATE(2257),
    [sym_unary_operator] = STATE(2257),
    [sym_binary_operator] = STATE(2257),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2257),
    [sym_call] = STATE(2257),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2257),
    [sym_anonymous_function] = STATE(2257),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2059),
    [sym_integer] = ACTIONS(2059),
    [sym_float] = ACTIONS(2059),
    [sym_char] = ACTIONS(2059),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2059),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(824)] = {
    [sym__expression] = STATE(2259),
    [sym_block] = STATE(2259),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(2259),
    [sym_nil] = STATE(2259),
    [sym__atom] = STATE(2259),
    [sym_quoted_atom] = STATE(2259),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2259),
    [sym_charlist] = STATE(2259),
    [sym_sigil] = STATE(2259),
    [sym_list] = STATE(2259),
    [sym_tuple] = STATE(2259),
    [sym_bitstring] = STATE(2259),
    [sym_map] = STATE(2259),
    [sym__nullary_operator] = STATE(2259),
    [sym_unary_operator] = STATE(2259),
    [sym_binary_operator] = STATE(2259),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2259),
    [sym_call] = STATE(2259),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2259),
    [sym_anonymous_function] = STATE(2259),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2061),
    [sym_integer] = ACTIONS(2061),
    [sym_float] = ACTIONS(2061),
    [sym_char] = ACTIONS(2061),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2061),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(825)] = {
    [sym__expression] = STATE(3760),
    [sym_block] = STATE(3760),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3760),
    [sym_nil] = STATE(3760),
    [sym__atom] = STATE(3760),
    [sym_quoted_atom] = STATE(3760),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3760),
    [sym_charlist] = STATE(3760),
    [sym_sigil] = STATE(3760),
    [sym_list] = STATE(3760),
    [sym_tuple] = STATE(3760),
    [sym_bitstring] = STATE(3760),
    [sym_map] = STATE(3760),
    [sym__nullary_operator] = STATE(3760),
    [sym_unary_operator] = STATE(3760),
    [sym_binary_operator] = STATE(3760),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3760),
    [sym_call] = STATE(3760),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3760),
    [sym_anonymous_function] = STATE(3760),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2531),
    [sym_integer] = ACTIONS(2531),
    [sym_float] = ACTIONS(2531),
    [sym_char] = ACTIONS(2531),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2531),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(826)] = {
    [sym__expression] = STATE(3761),
    [sym_block] = STATE(3761),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3761),
    [sym_nil] = STATE(3761),
    [sym__atom] = STATE(3761),
    [sym_quoted_atom] = STATE(3761),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3761),
    [sym_charlist] = STATE(3761),
    [sym_sigil] = STATE(3761),
    [sym_list] = STATE(3761),
    [sym_tuple] = STATE(3761),
    [sym_bitstring] = STATE(3761),
    [sym_map] = STATE(3761),
    [sym__nullary_operator] = STATE(3761),
    [sym_unary_operator] = STATE(3761),
    [sym_binary_operator] = STATE(3761),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3761),
    [sym_call] = STATE(3761),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3761),
    [sym_anonymous_function] = STATE(3761),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2533),
    [sym_integer] = ACTIONS(2533),
    [sym_float] = ACTIONS(2533),
    [sym_char] = ACTIONS(2533),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2533),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(827)] = {
    [sym__expression] = STATE(3763),
    [sym_block] = STATE(3763),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3763),
    [sym_nil] = STATE(3763),
    [sym__atom] = STATE(3763),
    [sym_quoted_atom] = STATE(3763),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3763),
    [sym_charlist] = STATE(3763),
    [sym_sigil] = STATE(3763),
    [sym_list] = STATE(3763),
    [sym_tuple] = STATE(3763),
    [sym_bitstring] = STATE(3763),
    [sym_map] = STATE(3763),
    [sym__nullary_operator] = STATE(3763),
    [sym_unary_operator] = STATE(3763),
    [sym_binary_operator] = STATE(3763),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3763),
    [sym_call] = STATE(3763),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3763),
    [sym_anonymous_function] = STATE(3763),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2535),
    [sym_integer] = ACTIONS(2535),
    [sym_float] = ACTIONS(2535),
    [sym_char] = ACTIONS(2535),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2535),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(828)] = {
    [sym__expression] = STATE(3764),
    [sym_block] = STATE(3764),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3764),
    [sym_nil] = STATE(3764),
    [sym__atom] = STATE(3764),
    [sym_quoted_atom] = STATE(3764),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3764),
    [sym_charlist] = STATE(3764),
    [sym_sigil] = STATE(3764),
    [sym_list] = STATE(3764),
    [sym_tuple] = STATE(3764),
    [sym_bitstring] = STATE(3764),
    [sym_map] = STATE(3764),
    [sym__nullary_operator] = STATE(3764),
    [sym_unary_operator] = STATE(3764),
    [sym_binary_operator] = STATE(3764),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3764),
    [sym_call] = STATE(3764),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3764),
    [sym_anonymous_function] = STATE(3764),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2537),
    [sym_integer] = ACTIONS(2537),
    [sym_float] = ACTIONS(2537),
    [sym_char] = ACTIONS(2537),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2537),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(829)] = {
    [sym__expression] = STATE(3765),
    [sym_block] = STATE(3765),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3765),
    [sym_nil] = STATE(3765),
    [sym__atom] = STATE(3765),
    [sym_quoted_atom] = STATE(3765),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3765),
    [sym_charlist] = STATE(3765),
    [sym_sigil] = STATE(3765),
    [sym_list] = STATE(3765),
    [sym_tuple] = STATE(3765),
    [sym_bitstring] = STATE(3765),
    [sym_map] = STATE(3765),
    [sym__nullary_operator] = STATE(3765),
    [sym_unary_operator] = STATE(3765),
    [sym_binary_operator] = STATE(3765),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3765),
    [sym_call] = STATE(3765),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3765),
    [sym_anonymous_function] = STATE(3765),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2539),
    [sym_integer] = ACTIONS(2539),
    [sym_float] = ACTIONS(2539),
    [sym_char] = ACTIONS(2539),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2539),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(830)] = {
    [sym__expression] = STATE(3766),
    [sym_block] = STATE(3766),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3766),
    [sym_nil] = STATE(3766),
    [sym__atom] = STATE(3766),
    [sym_quoted_atom] = STATE(3766),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3766),
    [sym_charlist] = STATE(3766),
    [sym_sigil] = STATE(3766),
    [sym_list] = STATE(3766),
    [sym_tuple] = STATE(3766),
    [sym_bitstring] = STATE(3766),
    [sym_map] = STATE(3766),
    [sym__nullary_operator] = STATE(3766),
    [sym_unary_operator] = STATE(3766),
    [sym_binary_operator] = STATE(3766),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3766),
    [sym_call] = STATE(3766),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3766),
    [sym_anonymous_function] = STATE(3766),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2541),
    [sym_integer] = ACTIONS(2541),
    [sym_float] = ACTIONS(2541),
    [sym_char] = ACTIONS(2541),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2541),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(831)] = {
    [sym__expression] = STATE(3768),
    [sym_block] = STATE(3768),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3768),
    [sym_nil] = STATE(3768),
    [sym__atom] = STATE(3768),
    [sym_quoted_atom] = STATE(3768),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3768),
    [sym_charlist] = STATE(3768),
    [sym_sigil] = STATE(3768),
    [sym_list] = STATE(3768),
    [sym_tuple] = STATE(3768),
    [sym_bitstring] = STATE(3768),
    [sym_map] = STATE(3768),
    [sym__nullary_operator] = STATE(3768),
    [sym_unary_operator] = STATE(3768),
    [sym_binary_operator] = STATE(3768),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3768),
    [sym_call] = STATE(3768),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3768),
    [sym_anonymous_function] = STATE(3768),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2543),
    [sym_integer] = ACTIONS(2543),
    [sym_float] = ACTIONS(2543),
    [sym_char] = ACTIONS(2543),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2543),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(832)] = {
    [sym__expression] = STATE(3769),
    [sym_block] = STATE(3769),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3769),
    [sym_nil] = STATE(3769),
    [sym__atom] = STATE(3769),
    [sym_quoted_atom] = STATE(3769),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3769),
    [sym_charlist] = STATE(3769),
    [sym_sigil] = STATE(3769),
    [sym_list] = STATE(3769),
    [sym_tuple] = STATE(3769),
    [sym_bitstring] = STATE(3769),
    [sym_map] = STATE(3769),
    [sym__nullary_operator] = STATE(3769),
    [sym_unary_operator] = STATE(3769),
    [sym_binary_operator] = STATE(3769),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3769),
    [sym_call] = STATE(3769),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3769),
    [sym_anonymous_function] = STATE(3769),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2545),
    [sym_integer] = ACTIONS(2545),
    [sym_float] = ACTIONS(2545),
    [sym_char] = ACTIONS(2545),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2545),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(833)] = {
    [sym__expression] = STATE(3770),
    [sym_block] = STATE(3770),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3770),
    [sym_nil] = STATE(3770),
    [sym__atom] = STATE(3770),
    [sym_quoted_atom] = STATE(3770),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3770),
    [sym_charlist] = STATE(3770),
    [sym_sigil] = STATE(3770),
    [sym_list] = STATE(3770),
    [sym_tuple] = STATE(3770),
    [sym_bitstring] = STATE(3770),
    [sym_map] = STATE(3770),
    [sym__nullary_operator] = STATE(3770),
    [sym_unary_operator] = STATE(3770),
    [sym_binary_operator] = STATE(3770),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3770),
    [sym_call] = STATE(3770),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3770),
    [sym_anonymous_function] = STATE(3770),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2547),
    [sym_integer] = ACTIONS(2547),
    [sym_float] = ACTIONS(2547),
    [sym_char] = ACTIONS(2547),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2547),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(834)] = {
    [sym__expression] = STATE(3771),
    [sym_block] = STATE(3771),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3771),
    [sym_nil] = STATE(3771),
    [sym__atom] = STATE(3771),
    [sym_quoted_atom] = STATE(3771),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3771),
    [sym_charlist] = STATE(3771),
    [sym_sigil] = STATE(3771),
    [sym_list] = STATE(3771),
    [sym_tuple] = STATE(3771),
    [sym_bitstring] = STATE(3771),
    [sym_map] = STATE(3771),
    [sym__nullary_operator] = STATE(3771),
    [sym_unary_operator] = STATE(3771),
    [sym_binary_operator] = STATE(3771),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3771),
    [sym_call] = STATE(3771),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3771),
    [sym_anonymous_function] = STATE(3771),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2549),
    [sym_integer] = ACTIONS(2549),
    [sym_float] = ACTIONS(2549),
    [sym_char] = ACTIONS(2549),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2549),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(835)] = {
    [sym__expression] = STATE(3772),
    [sym_block] = STATE(3772),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3772),
    [sym_nil] = STATE(3772),
    [sym__atom] = STATE(3772),
    [sym_quoted_atom] = STATE(3772),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3772),
    [sym_charlist] = STATE(3772),
    [sym_sigil] = STATE(3772),
    [sym_list] = STATE(3772),
    [sym_tuple] = STATE(3772),
    [sym_bitstring] = STATE(3772),
    [sym_map] = STATE(3772),
    [sym__nullary_operator] = STATE(3772),
    [sym_unary_operator] = STATE(3772),
    [sym_binary_operator] = STATE(3772),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3772),
    [sym_call] = STATE(3772),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3772),
    [sym_anonymous_function] = STATE(3772),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2551),
    [sym_integer] = ACTIONS(2551),
    [sym_float] = ACTIONS(2551),
    [sym_char] = ACTIONS(2551),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2551),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(836)] = {
    [sym__expression] = STATE(3773),
    [sym_block] = STATE(3773),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3773),
    [sym_nil] = STATE(3773),
    [sym__atom] = STATE(3773),
    [sym_quoted_atom] = STATE(3773),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3773),
    [sym_charlist] = STATE(3773),
    [sym_sigil] = STATE(3773),
    [sym_list] = STATE(3773),
    [sym_tuple] = STATE(3773),
    [sym_bitstring] = STATE(3773),
    [sym_map] = STATE(3773),
    [sym__nullary_operator] = STATE(3773),
    [sym_unary_operator] = STATE(3773),
    [sym_binary_operator] = STATE(3773),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3773),
    [sym_call] = STATE(3773),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3773),
    [sym_anonymous_function] = STATE(3773),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2553),
    [sym_integer] = ACTIONS(2553),
    [sym_float] = ACTIONS(2553),
    [sym_char] = ACTIONS(2553),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2553),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(837)] = {
    [sym__expression] = STATE(3774),
    [sym_block] = STATE(3774),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3774),
    [sym_nil] = STATE(3774),
    [sym__atom] = STATE(3774),
    [sym_quoted_atom] = STATE(3774),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3774),
    [sym_charlist] = STATE(3774),
    [sym_sigil] = STATE(3774),
    [sym_list] = STATE(3774),
    [sym_tuple] = STATE(3774),
    [sym_bitstring] = STATE(3774),
    [sym_map] = STATE(3774),
    [sym__nullary_operator] = STATE(3774),
    [sym_unary_operator] = STATE(3774),
    [sym_binary_operator] = STATE(3774),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3774),
    [sym_call] = STATE(3774),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3774),
    [sym_anonymous_function] = STATE(3774),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2555),
    [sym_integer] = ACTIONS(2555),
    [sym_float] = ACTIONS(2555),
    [sym_char] = ACTIONS(2555),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2555),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(838)] = {
    [sym__expression] = STATE(3775),
    [sym_block] = STATE(3775),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3775),
    [sym_nil] = STATE(3775),
    [sym__atom] = STATE(3775),
    [sym_quoted_atom] = STATE(3775),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3775),
    [sym_charlist] = STATE(3775),
    [sym_sigil] = STATE(3775),
    [sym_list] = STATE(3775),
    [sym_tuple] = STATE(3775),
    [sym_bitstring] = STATE(3775),
    [sym_map] = STATE(3775),
    [sym__nullary_operator] = STATE(3775),
    [sym_unary_operator] = STATE(3775),
    [sym_binary_operator] = STATE(3775),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3775),
    [sym_call] = STATE(3775),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3775),
    [sym_anonymous_function] = STATE(3775),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2557),
    [sym_integer] = ACTIONS(2557),
    [sym_float] = ACTIONS(2557),
    [sym_char] = ACTIONS(2557),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2557),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(839)] = {
    [sym__expression] = STATE(3776),
    [sym_block] = STATE(3776),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(3776),
    [sym_nil] = STATE(3776),
    [sym__atom] = STATE(3776),
    [sym_quoted_atom] = STATE(3776),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(3776),
    [sym_charlist] = STATE(3776),
    [sym_sigil] = STATE(3776),
    [sym_list] = STATE(3776),
    [sym_tuple] = STATE(3776),
    [sym_bitstring] = STATE(3776),
    [sym_map] = STATE(3776),
    [sym__nullary_operator] = STATE(3776),
    [sym_unary_operator] = STATE(3776),
    [sym_binary_operator] = STATE(3776),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(3776),
    [sym_call] = STATE(3776),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(3776),
    [sym_anonymous_function] = STATE(3776),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2559),
    [sym_integer] = ACTIONS(2559),
    [sym_float] = ACTIONS(2559),
    [sym_char] = ACTIONS(2559),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2559),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(840)] = {
    [sym__expression] = STATE(2314),
    [sym_block] = STATE(2314),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(2314),
    [sym_nil] = STATE(2314),
    [sym__atom] = STATE(2314),
    [sym_quoted_atom] = STATE(2314),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2314),
    [sym_charlist] = STATE(2314),
    [sym_sigil] = STATE(2314),
    [sym_list] = STATE(2314),
    [sym_tuple] = STATE(2314),
    [sym_bitstring] = STATE(2314),
    [sym_map] = STATE(2314),
    [sym__nullary_operator] = STATE(2314),
    [sym_unary_operator] = STATE(2314),
    [sym_binary_operator] = STATE(2314),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2314),
    [sym_call] = STATE(2314),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2314),
    [sym_anonymous_function] = STATE(2314),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2093),
    [sym_integer] = ACTIONS(2093),
    [sym_float] = ACTIONS(2093),
    [sym_char] = ACTIONS(2093),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2093),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(841)] = {
    [sym__expression] = STATE(4535),
    [sym_block] = STATE(4535),
    [sym_identifier] = STATE(71),
    [sym_boolean] = STATE(4535),
    [sym_nil] = STATE(4535),
    [sym__atom] = STATE(4535),
    [sym_quoted_atom] = STATE(4535),
    [sym__quoted_i_double] = STATE(4494),
    [sym__quoted_i_single] = STATE(4502),
    [sym__quoted_i_heredoc_single] = STATE(4499),
    [sym__quoted_i_heredoc_double] = STATE(4500),
    [sym_string] = STATE(4535),
    [sym_charlist] = STATE(4535),
    [sym_sigil] = STATE(4535),
    [sym_list] = STATE(4535),
    [sym_tuple] = STATE(4535),
    [sym_bitstring] = STATE(4535),
    [sym_map] = STATE(4535),
    [sym__nullary_operator] = STATE(4535),
    [sym_unary_operator] = STATE(4535),
    [sym_binary_operator] = STATE(4535),
    [sym_operator_identifier] = STATE(6934),
    [sym_dot] = STATE(4535),
    [sym_call] = STATE(4535),
    [sym__call_without_parentheses] = STATE(4509),
    [sym__call_with_parentheses] = STATE(4510),
    [sym__local_call_without_parentheses] = STATE(4515),
    [sym__local_call_with_parentheses] = STATE(4033),
    [sym__local_call_just_do_block] = STATE(4517),
    [sym__remote_call_without_parentheses] = STATE(4521),
    [sym__remote_call_with_parentheses] = STATE(4037),
    [sym__remote_dot] = STATE(62),
    [sym__anonymous_call] = STATE(4039),
    [sym__anonymous_dot] = STATE(6787),
    [sym__double_call] = STATE(4527),
    [sym_access_call] = STATE(4535),
    [sym_anonymous_function] = STATE(4535),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1157),
    [aux_sym_identifier_token1] = ACTIONS(1159),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1159),
    [sym_alias] = ACTIONS(2561),
    [sym_integer] = ACTIONS(2561),
    [sym_float] = ACTIONS(2561),
    [sym_char] = ACTIONS(2561),
    [anon_sym_true] = ACTIONS(1163),
    [anon_sym_false] = ACTIONS(1163),
    [anon_sym_nil] = ACTIONS(1165),
    [sym_atom] = ACTIONS(2561),
    [anon_sym_DQUOTE] = ACTIONS(1167),
    [anon_sym_SQUOTE] = ACTIONS(1169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1171),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1173),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LBRACK] = ACTIONS(1177),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1179),
    [anon_sym_LT_LT] = ACTIONS(1183),
    [anon_sym_PERCENT] = ACTIONS(1187),
    [anon_sym_DOT_DOT] = ACTIONS(1189),
    [anon_sym_AMP] = ACTIONS(1191),
    [anon_sym_PLUS] = ACTIONS(1193),
    [anon_sym_DASH] = ACTIONS(1193),
    [anon_sym_BANG] = ACTIONS(1193),
    [anon_sym_CARET] = ACTIONS(1193),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1193),
    [anon_sym_not] = ACTIONS(1193),
    [anon_sym_AT] = ACTIONS(1195),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1197),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1199),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1201),
  },
  [STATE(842)] = {
    [sym__expression] = STATE(3097),
    [sym_block] = STATE(3097),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(3097),
    [sym_nil] = STATE(3097),
    [sym__atom] = STATE(3097),
    [sym_quoted_atom] = STATE(3097),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(3097),
    [sym_charlist] = STATE(3097),
    [sym_sigil] = STATE(3097),
    [sym_list] = STATE(3097),
    [sym_tuple] = STATE(3097),
    [sym_bitstring] = STATE(3097),
    [sym_map] = STATE(3097),
    [sym__nullary_operator] = STATE(3097),
    [sym_unary_operator] = STATE(3097),
    [sym_binary_operator] = STATE(3097),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(3097),
    [sym_call] = STATE(3097),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(3097),
    [sym_anonymous_function] = STATE(3097),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(1489),
    [sym_integer] = ACTIONS(1489),
    [sym_float] = ACTIONS(1489),
    [sym_char] = ACTIONS(1489),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(1489),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(843)] = {
    [sym__expression] = STATE(3206),
    [sym_block] = STATE(3206),
    [sym_identifier] = STATE(53),
    [sym_boolean] = STATE(3206),
    [sym_nil] = STATE(3206),
    [sym__atom] = STATE(3206),
    [sym_quoted_atom] = STATE(3206),
    [sym__quoted_i_double] = STATE(3076),
    [sym__quoted_i_single] = STATE(3077),
    [sym__quoted_i_heredoc_single] = STATE(3437),
    [sym__quoted_i_heredoc_double] = STATE(3438),
    [sym_string] = STATE(3206),
    [sym_charlist] = STATE(3206),
    [sym_sigil] = STATE(3206),
    [sym_list] = STATE(3206),
    [sym_tuple] = STATE(3206),
    [sym_bitstring] = STATE(3206),
    [sym_map] = STATE(3206),
    [sym__nullary_operator] = STATE(3206),
    [sym_unary_operator] = STATE(3206),
    [sym_binary_operator] = STATE(3206),
    [sym_operator_identifier] = STATE(6944),
    [sym_dot] = STATE(3206),
    [sym_call] = STATE(3206),
    [sym__call_without_parentheses] = STATE(3440),
    [sym__call_with_parentheses] = STATE(3443),
    [sym__local_call_without_parentheses] = STATE(3449),
    [sym__local_call_with_parentheses] = STATE(2278),
    [sym__local_call_just_do_block] = STATE(3453),
    [sym__remote_call_without_parentheses] = STATE(3454),
    [sym__remote_call_with_parentheses] = STATE(2285),
    [sym__remote_dot] = STATE(41),
    [sym__anonymous_call] = STATE(2261),
    [sym__anonymous_dot] = STATE(6817),
    [sym__double_call] = STATE(3458),
    [sym_access_call] = STATE(3206),
    [sym_anonymous_function] = STATE(3206),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(431),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(2563),
    [sym_integer] = ACTIONS(2563),
    [sym_float] = ACTIONS(2563),
    [sym_char] = ACTIONS(2563),
    [anon_sym_true] = ACTIONS(437),
    [anon_sym_false] = ACTIONS(437),
    [anon_sym_nil] = ACTIONS(439),
    [sym_atom] = ACTIONS(2563),
    [anon_sym_DQUOTE] = ACTIONS(441),
    [anon_sym_SQUOTE] = ACTIONS(443),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(445),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(447),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(451),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(453),
    [anon_sym_LT_LT] = ACTIONS(457),
    [anon_sym_PERCENT] = ACTIONS(459),
    [anon_sym_DOT_DOT] = ACTIONS(1491),
    [anon_sym_AMP] = ACTIONS(461),
    [anon_sym_PLUS] = ACTIONS(463),
    [anon_sym_DASH] = ACTIONS(463),
    [anon_sym_BANG] = ACTIONS(463),
    [anon_sym_CARET] = ACTIONS(463),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(463),
    [anon_sym_not] = ACTIONS(463),
    [anon_sym_AT] = ACTIONS(465),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(467),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(471),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(473),
  },
  [STATE(844)] = {
    [sym__expression] = STATE(2478),
    [sym_block] = STATE(2478),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2478),
    [sym_nil] = STATE(2478),
    [sym__atom] = STATE(2478),
    [sym_quoted_atom] = STATE(2478),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2478),
    [sym_charlist] = STATE(2478),
    [sym_sigil] = STATE(2478),
    [sym_list] = STATE(2478),
    [sym_tuple] = STATE(2478),
    [sym_bitstring] = STATE(2478),
    [sym_map] = STATE(2478),
    [sym__nullary_operator] = STATE(2478),
    [sym_unary_operator] = STATE(2478),
    [sym_binary_operator] = STATE(2478),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2478),
    [sym_call] = STATE(2478),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2478),
    [sym_anonymous_function] = STATE(2478),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1499),
    [sym_integer] = ACTIONS(1499),
    [sym_float] = ACTIONS(1499),
    [sym_char] = ACTIONS(1499),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1499),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(845)] = {
    [sym__expression] = STATE(2684),
    [sym_block] = STATE(2684),
    [sym_identifier] = STATE(44),
    [sym_boolean] = STATE(2684),
    [sym_nil] = STATE(2684),
    [sym__atom] = STATE(2684),
    [sym_quoted_atom] = STATE(2684),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(2684),
    [sym_charlist] = STATE(2684),
    [sym_sigil] = STATE(2684),
    [sym_list] = STATE(2684),
    [sym_tuple] = STATE(2684),
    [sym_bitstring] = STATE(2684),
    [sym_map] = STATE(2684),
    [sym__nullary_operator] = STATE(2684),
    [sym_unary_operator] = STATE(2684),
    [sym_binary_operator] = STATE(2684),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(2684),
    [sym_call] = STATE(2684),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(42),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(2684),
    [sym_anonymous_function] = STATE(2684),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2565),
    [sym_integer] = ACTIONS(2565),
    [sym_float] = ACTIONS(2565),
    [sym_char] = ACTIONS(2565),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2565),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(479),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(483),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_BANG] = ACTIONS(485),
    [anon_sym_CARET] = ACTIONS(485),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(485),
    [anon_sym_not] = ACTIONS(485),
    [anon_sym_AT] = ACTIONS(487),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(489),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(846)] = {
    [sym__expression] = STATE(2716),
    [sym_block] = STATE(2716),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2716),
    [sym_nil] = STATE(2716),
    [sym__atom] = STATE(2716),
    [sym_quoted_atom] = STATE(2716),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2716),
    [sym_charlist] = STATE(2716),
    [sym_sigil] = STATE(2716),
    [sym_list] = STATE(2716),
    [sym_tuple] = STATE(2716),
    [sym_bitstring] = STATE(2716),
    [sym_map] = STATE(2716),
    [sym__nullary_operator] = STATE(2716),
    [sym_unary_operator] = STATE(2716),
    [sym_binary_operator] = STATE(2716),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2716),
    [sym_call] = STATE(2716),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2716),
    [sym_anonymous_function] = STATE(2716),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1547),
    [sym_integer] = ACTIONS(1547),
    [sym_float] = ACTIONS(1547),
    [sym_char] = ACTIONS(1547),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1547),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(847)] = {
    [sym__expression] = STATE(4485),
    [sym_block] = STATE(4485),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4485),
    [sym_nil] = STATE(4485),
    [sym__atom] = STATE(4485),
    [sym_quoted_atom] = STATE(4485),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4485),
    [sym_charlist] = STATE(4485),
    [sym_sigil] = STATE(4485),
    [sym_list] = STATE(4485),
    [sym_tuple] = STATE(4485),
    [sym_bitstring] = STATE(4485),
    [sym_map] = STATE(4485),
    [sym__nullary_operator] = STATE(4485),
    [sym_unary_operator] = STATE(4485),
    [sym_binary_operator] = STATE(4485),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4485),
    [sym_call] = STATE(4485),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4485),
    [sym_anonymous_function] = STATE(4485),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(1569),
    [sym_integer] = ACTIONS(1569),
    [sym_float] = ACTIONS(1569),
    [sym_char] = ACTIONS(1569),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1569),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(848)] = {
    [sym__expression] = STATE(2719),
    [sym_block] = STATE(2719),
    [sym_identifier] = STATE(36),
    [sym_boolean] = STATE(2719),
    [sym_nil] = STATE(2719),
    [sym__atom] = STATE(2719),
    [sym_quoted_atom] = STATE(2719),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(2719),
    [sym_charlist] = STATE(2719),
    [sym_sigil] = STATE(2719),
    [sym_list] = STATE(2719),
    [sym_tuple] = STATE(2719),
    [sym_bitstring] = STATE(2719),
    [sym_map] = STATE(2719),
    [sym__nullary_operator] = STATE(2719),
    [sym_unary_operator] = STATE(2719),
    [sym_binary_operator] = STATE(2719),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(2719),
    [sym_call] = STATE(2719),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(35),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(2719),
    [sym_anonymous_function] = STATE(2719),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2567),
    [sym_integer] = ACTIONS(2567),
    [sym_float] = ACTIONS(2567),
    [sym_char] = ACTIONS(2567),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2567),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(389),
    [anon_sym_PLUS] = ACTIONS(391),
    [anon_sym_DASH] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(391),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(391),
    [anon_sym_not] = ACTIONS(391),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(399),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(849)] = {
    [sym__expression] = STATE(4133),
    [sym_block] = STATE(4133),
    [sym_identifier] = STATE(72),
    [sym_boolean] = STATE(4133),
    [sym_nil] = STATE(4133),
    [sym__atom] = STATE(4133),
    [sym_quoted_atom] = STATE(4133),
    [sym__quoted_i_double] = STATE(4106),
    [sym__quoted_i_single] = STATE(4108),
    [sym__quoted_i_heredoc_single] = STATE(4109),
    [sym__quoted_i_heredoc_double] = STATE(4110),
    [sym_string] = STATE(4133),
    [sym_charlist] = STATE(4133),
    [sym_sigil] = STATE(4133),
    [sym_list] = STATE(4133),
    [sym_tuple] = STATE(4133),
    [sym_bitstring] = STATE(4133),
    [sym_map] = STATE(4133),
    [sym__nullary_operator] = STATE(4133),
    [sym_unary_operator] = STATE(4133),
    [sym_binary_operator] = STATE(4133),
    [sym_operator_identifier] = STATE(6939),
    [sym_dot] = STATE(4133),
    [sym_call] = STATE(4133),
    [sym__call_without_parentheses] = STATE(4111),
    [sym__call_with_parentheses] = STATE(4112),
    [sym__local_call_without_parentheses] = STATE(4114),
    [sym__local_call_with_parentheses] = STATE(3217),
    [sym__local_call_just_do_block] = STATE(4116),
    [sym__remote_call_without_parentheses] = STATE(4117),
    [sym__remote_call_with_parentheses] = STATE(3218),
    [sym__remote_dot] = STATE(55),
    [sym__anonymous_call] = STATE(3220),
    [sym__anonymous_dot] = STATE(6779),
    [sym__double_call] = STATE(4121),
    [sym_access_call] = STATE(4133),
    [sym_anonymous_function] = STATE(4133),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1365),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2569),
    [sym_integer] = ACTIONS(2569),
    [sym_float] = ACTIONS(2569),
    [sym_char] = ACTIONS(2569),
    [anon_sym_true] = ACTIONS(724),
    [anon_sym_false] = ACTIONS(724),
    [anon_sym_nil] = ACTIONS(726),
    [sym_atom] = ACTIONS(2569),
    [anon_sym_DQUOTE] = ACTIONS(728),
    [anon_sym_SQUOTE] = ACTIONS(730),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(732),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LBRACK] = ACTIONS(738),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(740),
    [anon_sym_LT_LT] = ACTIONS(742),
    [anon_sym_PERCENT] = ACTIONS(744),
    [anon_sym_DOT_DOT] = ACTIONS(746),
    [anon_sym_AMP] = ACTIONS(748),
    [anon_sym_PLUS] = ACTIONS(750),
    [anon_sym_DASH] = ACTIONS(750),
    [anon_sym_BANG] = ACTIONS(750),
    [anon_sym_CARET] = ACTIONS(750),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(750),
    [anon_sym_not] = ACTIONS(750),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(758),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(760),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(762),
  },
  [STATE(850)] = {
    [sym__expression] = STATE(3538),
    [sym_block] = STATE(3538),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3538),
    [sym_nil] = STATE(3538),
    [sym__atom] = STATE(3538),
    [sym_quoted_atom] = STATE(3538),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3538),
    [sym_charlist] = STATE(3538),
    [sym_sigil] = STATE(3538),
    [sym_list] = STATE(3538),
    [sym_tuple] = STATE(3538),
    [sym_bitstring] = STATE(3538),
    [sym_map] = STATE(3538),
    [sym__nullary_operator] = STATE(3538),
    [sym_unary_operator] = STATE(3538),
    [sym_binary_operator] = STATE(3538),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3538),
    [sym_call] = STATE(3538),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3538),
    [sym_anonymous_function] = STATE(3538),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(1373),
    [sym_integer] = ACTIONS(1373),
    [sym_float] = ACTIONS(1373),
    [sym_char] = ACTIONS(1373),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(1373),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(851)] = {
    [sym__expression] = STATE(4178),
    [sym_block] = STATE(4178),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4178),
    [sym_nil] = STATE(4178),
    [sym__atom] = STATE(4178),
    [sym_quoted_atom] = STATE(4178),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4178),
    [sym_charlist] = STATE(4178),
    [sym_sigil] = STATE(4178),
    [sym_list] = STATE(4178),
    [sym_tuple] = STATE(4178),
    [sym_bitstring] = STATE(4178),
    [sym_map] = STATE(4178),
    [sym__nullary_operator] = STATE(4178),
    [sym_unary_operator] = STATE(4178),
    [sym_binary_operator] = STATE(4178),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4178),
    [sym_call] = STATE(4178),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4178),
    [sym_anonymous_function] = STATE(4178),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2571),
    [sym_integer] = ACTIONS(2571),
    [sym_float] = ACTIONS(2571),
    [sym_char] = ACTIONS(2571),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2571),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(852)] = {
    [sym__expression] = STATE(3011),
    [sym_block] = STATE(3011),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3011),
    [sym_nil] = STATE(3011),
    [sym__atom] = STATE(3011),
    [sym_quoted_atom] = STATE(3011),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3011),
    [sym_charlist] = STATE(3011),
    [sym_sigil] = STATE(3011),
    [sym_list] = STATE(3011),
    [sym_tuple] = STATE(3011),
    [sym_bitstring] = STATE(3011),
    [sym_map] = STATE(3011),
    [sym__nullary_operator] = STATE(3011),
    [sym_unary_operator] = STATE(3011),
    [sym_binary_operator] = STATE(3011),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3011),
    [sym_call] = STATE(3011),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3011),
    [sym_anonymous_function] = STATE(3011),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2491),
    [sym_integer] = ACTIONS(2491),
    [sym_float] = ACTIONS(2491),
    [sym_char] = ACTIONS(2491),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2491),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(853)] = {
    [sym__expression] = STATE(3016),
    [sym_block] = STATE(3016),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3016),
    [sym_nil] = STATE(3016),
    [sym__atom] = STATE(3016),
    [sym_quoted_atom] = STATE(3016),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3016),
    [sym_charlist] = STATE(3016),
    [sym_sigil] = STATE(3016),
    [sym_list] = STATE(3016),
    [sym_tuple] = STATE(3016),
    [sym_bitstring] = STATE(3016),
    [sym_map] = STATE(3016),
    [sym__nullary_operator] = STATE(3016),
    [sym_unary_operator] = STATE(3016),
    [sym_binary_operator] = STATE(3016),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3016),
    [sym_call] = STATE(3016),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3016),
    [sym_anonymous_function] = STATE(3016),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2493),
    [sym_integer] = ACTIONS(2493),
    [sym_float] = ACTIONS(2493),
    [sym_char] = ACTIONS(2493),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2493),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(1086),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(854)] = {
    [sym__expression] = STATE(4182),
    [sym_block] = STATE(4182),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4182),
    [sym_nil] = STATE(4182),
    [sym__atom] = STATE(4182),
    [sym_quoted_atom] = STATE(4182),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4182),
    [sym_charlist] = STATE(4182),
    [sym_sigil] = STATE(4182),
    [sym_list] = STATE(4182),
    [sym_tuple] = STATE(4182),
    [sym_bitstring] = STATE(4182),
    [sym_map] = STATE(4182),
    [sym__nullary_operator] = STATE(4182),
    [sym_unary_operator] = STATE(4182),
    [sym_binary_operator] = STATE(4182),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4182),
    [sym_call] = STATE(4182),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4182),
    [sym_anonymous_function] = STATE(4182),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2573),
    [sym_integer] = ACTIONS(2573),
    [sym_float] = ACTIONS(2573),
    [sym_char] = ACTIONS(2573),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2573),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(855)] = {
    [sym__expression] = STATE(3544),
    [sym_block] = STATE(3544),
    [sym_identifier] = STATE(54),
    [sym_boolean] = STATE(3544),
    [sym_nil] = STATE(3544),
    [sym__atom] = STATE(3544),
    [sym_quoted_atom] = STATE(3544),
    [sym__quoted_i_double] = STATE(3653),
    [sym__quoted_i_single] = STATE(3654),
    [sym__quoted_i_heredoc_single] = STATE(3655),
    [sym__quoted_i_heredoc_double] = STATE(3656),
    [sym_string] = STATE(3544),
    [sym_charlist] = STATE(3544),
    [sym_sigil] = STATE(3544),
    [sym_list] = STATE(3544),
    [sym_tuple] = STATE(3544),
    [sym_bitstring] = STATE(3544),
    [sym_map] = STATE(3544),
    [sym__nullary_operator] = STATE(3544),
    [sym_unary_operator] = STATE(3544),
    [sym_binary_operator] = STATE(3544),
    [sym_operator_identifier] = STATE(6964),
    [sym_dot] = STATE(3544),
    [sym_call] = STATE(3544),
    [sym__call_without_parentheses] = STATE(3657),
    [sym__call_with_parentheses] = STATE(3658),
    [sym__local_call_without_parentheses] = STATE(3659),
    [sym__local_call_with_parentheses] = STATE(2563),
    [sym__local_call_just_do_block] = STATE(3660),
    [sym__remote_call_without_parentheses] = STATE(3661),
    [sym__remote_call_with_parentheses] = STATE(2564),
    [sym__remote_dot] = STATE(61),
    [sym__anonymous_call] = STATE(2566),
    [sym__anonymous_dot] = STATE(6866),
    [sym__double_call] = STATE(3662),
    [sym_access_call] = STATE(3544),
    [sym_anonymous_function] = STATE(3544),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(540),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2575),
    [sym_integer] = ACTIONS(2575),
    [sym_float] = ACTIONS(2575),
    [sym_char] = ACTIONS(2575),
    [anon_sym_true] = ACTIONS(546),
    [anon_sym_false] = ACTIONS(546),
    [anon_sym_nil] = ACTIONS(548),
    [sym_atom] = ACTIONS(2575),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_SQUOTE] = ACTIONS(552),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(554),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(556),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LBRACK] = ACTIONS(560),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(562),
    [anon_sym_LT_LT] = ACTIONS(566),
    [anon_sym_PERCENT] = ACTIONS(568),
    [anon_sym_DOT_DOT] = ACTIONS(1375),
    [anon_sym_AMP] = ACTIONS(573),
    [anon_sym_PLUS] = ACTIONS(578),
    [anon_sym_DASH] = ACTIONS(578),
    [anon_sym_BANG] = ACTIONS(578),
    [anon_sym_CARET] = ACTIONS(578),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(578),
    [anon_sym_not] = ACTIONS(578),
    [anon_sym_AT] = ACTIONS(580),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(582),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(586),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(588),
  },
  [STATE(856)] = {
    [sym__expression] = STATE(4183),
    [sym_block] = STATE(4183),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4183),
    [sym_nil] = STATE(4183),
    [sym__atom] = STATE(4183),
    [sym_quoted_atom] = STATE(4183),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4183),
    [sym_charlist] = STATE(4183),
    [sym_sigil] = STATE(4183),
    [sym_list] = STATE(4183),
    [sym_tuple] = STATE(4183),
    [sym_bitstring] = STATE(4183),
    [sym_map] = STATE(4183),
    [sym__nullary_operator] = STATE(4183),
    [sym_unary_operator] = STATE(4183),
    [sym_binary_operator] = STATE(4183),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4183),
    [sym_call] = STATE(4183),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4183),
    [sym_anonymous_function] = STATE(4183),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2577),
    [sym_integer] = ACTIONS(2577),
    [sym_float] = ACTIONS(2577),
    [sym_char] = ACTIONS(2577),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2577),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(857)] = {
    [sym__expression] = STATE(4184),
    [sym_block] = STATE(4184),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4184),
    [sym_nil] = STATE(4184),
    [sym__atom] = STATE(4184),
    [sym_quoted_atom] = STATE(4184),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4184),
    [sym_charlist] = STATE(4184),
    [sym_sigil] = STATE(4184),
    [sym_list] = STATE(4184),
    [sym_tuple] = STATE(4184),
    [sym_bitstring] = STATE(4184),
    [sym_map] = STATE(4184),
    [sym__nullary_operator] = STATE(4184),
    [sym_unary_operator] = STATE(4184),
    [sym_binary_operator] = STATE(4184),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4184),
    [sym_call] = STATE(4184),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4184),
    [sym_anonymous_function] = STATE(4184),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2579),
    [sym_integer] = ACTIONS(2579),
    [sym_float] = ACTIONS(2579),
    [sym_char] = ACTIONS(2579),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2579),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(858)] = {
    [sym__expression] = STATE(3058),
    [sym_block] = STATE(3058),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3058),
    [sym_nil] = STATE(3058),
    [sym__atom] = STATE(3058),
    [sym_quoted_atom] = STATE(3058),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3058),
    [sym_charlist] = STATE(3058),
    [sym_sigil] = STATE(3058),
    [sym_list] = STATE(3058),
    [sym_tuple] = STATE(3058),
    [sym_bitstring] = STATE(3058),
    [sym_map] = STATE(3058),
    [sym__nullary_operator] = STATE(3058),
    [sym_unary_operator] = STATE(3058),
    [sym_binary_operator] = STATE(3058),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3058),
    [sym_call] = STATE(3058),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3058),
    [sym_anonymous_function] = STATE(3058),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2495),
    [sym_integer] = ACTIONS(2495),
    [sym_float] = ACTIONS(2495),
    [sym_char] = ACTIONS(2495),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2495),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(859)] = {
    [sym__expression] = STATE(3059),
    [sym_block] = STATE(3059),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3059),
    [sym_nil] = STATE(3059),
    [sym__atom] = STATE(3059),
    [sym_quoted_atom] = STATE(3059),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3059),
    [sym_charlist] = STATE(3059),
    [sym_sigil] = STATE(3059),
    [sym_list] = STATE(3059),
    [sym_tuple] = STATE(3059),
    [sym_bitstring] = STATE(3059),
    [sym_map] = STATE(3059),
    [sym__nullary_operator] = STATE(3059),
    [sym_unary_operator] = STATE(3059),
    [sym_binary_operator] = STATE(3059),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3059),
    [sym_call] = STATE(3059),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3059),
    [sym_anonymous_function] = STATE(3059),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2497),
    [sym_integer] = ACTIONS(2497),
    [sym_float] = ACTIONS(2497),
    [sym_char] = ACTIONS(2497),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2497),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(860)] = {
    [sym__expression] = STATE(3147),
    [sym_block] = STATE(3147),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3147),
    [sym_nil] = STATE(3147),
    [sym__atom] = STATE(3147),
    [sym_quoted_atom] = STATE(3147),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3147),
    [sym_charlist] = STATE(3147),
    [sym_sigil] = STATE(3147),
    [sym_list] = STATE(3147),
    [sym_tuple] = STATE(3147),
    [sym_bitstring] = STATE(3147),
    [sym_map] = STATE(3147),
    [sym__nullary_operator] = STATE(3147),
    [sym_unary_operator] = STATE(3147),
    [sym_binary_operator] = STATE(3147),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3147),
    [sym_call] = STATE(3147),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3147),
    [sym_anonymous_function] = STATE(3147),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1399),
    [sym_integer] = ACTIONS(1399),
    [sym_float] = ACTIONS(1399),
    [sym_char] = ACTIONS(1399),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(861)] = {
    [sym__expression] = STATE(3927),
    [sym_block] = STATE(3927),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3927),
    [sym_nil] = STATE(3927),
    [sym__atom] = STATE(3927),
    [sym_quoted_atom] = STATE(3927),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3927),
    [sym_charlist] = STATE(3927),
    [sym_sigil] = STATE(3927),
    [sym_list] = STATE(3927),
    [sym_tuple] = STATE(3927),
    [sym_bitstring] = STATE(3927),
    [sym_map] = STATE(3927),
    [sym__nullary_operator] = STATE(3927),
    [sym_unary_operator] = STATE(3927),
    [sym_binary_operator] = STATE(3927),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3927),
    [sym_call] = STATE(3927),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3927),
    [sym_anonymous_function] = STATE(3927),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2581),
    [sym_integer] = ACTIONS(2581),
    [sym_float] = ACTIONS(2581),
    [sym_char] = ACTIONS(2581),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2581),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(862)] = {
    [sym__expression] = STATE(3928),
    [sym_block] = STATE(3928),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3928),
    [sym_nil] = STATE(3928),
    [sym__atom] = STATE(3928),
    [sym_quoted_atom] = STATE(3928),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3928),
    [sym_charlist] = STATE(3928),
    [sym_sigil] = STATE(3928),
    [sym_list] = STATE(3928),
    [sym_tuple] = STATE(3928),
    [sym_bitstring] = STATE(3928),
    [sym_map] = STATE(3928),
    [sym__nullary_operator] = STATE(3928),
    [sym_unary_operator] = STATE(3928),
    [sym_binary_operator] = STATE(3928),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3928),
    [sym_call] = STATE(3928),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3928),
    [sym_anonymous_function] = STATE(3928),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2583),
    [sym_integer] = ACTIONS(2583),
    [sym_float] = ACTIONS(2583),
    [sym_char] = ACTIONS(2583),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2583),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(863)] = {
    [sym__expression] = STATE(4185),
    [sym_block] = STATE(4185),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4185),
    [sym_nil] = STATE(4185),
    [sym__atom] = STATE(4185),
    [sym_quoted_atom] = STATE(4185),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4185),
    [sym_charlist] = STATE(4185),
    [sym_sigil] = STATE(4185),
    [sym_list] = STATE(4185),
    [sym_tuple] = STATE(4185),
    [sym_bitstring] = STATE(4185),
    [sym_map] = STATE(4185),
    [sym__nullary_operator] = STATE(4185),
    [sym_unary_operator] = STATE(4185),
    [sym_binary_operator] = STATE(4185),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4185),
    [sym_call] = STATE(4185),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4185),
    [sym_anonymous_function] = STATE(4185),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2585),
    [sym_integer] = ACTIONS(2585),
    [sym_float] = ACTIONS(2585),
    [sym_char] = ACTIONS(2585),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2585),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(864)] = {
    [sym__expression] = STATE(3954),
    [sym_block] = STATE(3954),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3954),
    [sym_nil] = STATE(3954),
    [sym__atom] = STATE(3954),
    [sym_quoted_atom] = STATE(3954),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3954),
    [sym_charlist] = STATE(3954),
    [sym_sigil] = STATE(3954),
    [sym_list] = STATE(3954),
    [sym_tuple] = STATE(3954),
    [sym_bitstring] = STATE(3954),
    [sym_map] = STATE(3954),
    [sym__nullary_operator] = STATE(3954),
    [sym_unary_operator] = STATE(3954),
    [sym_binary_operator] = STATE(3954),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3954),
    [sym_call] = STATE(3954),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3954),
    [sym_anonymous_function] = STATE(3954),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2587),
    [sym_integer] = ACTIONS(2587),
    [sym_float] = ACTIONS(2587),
    [sym_char] = ACTIONS(2587),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2587),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(865)] = {
    [sym__expression] = STATE(3955),
    [sym_block] = STATE(3955),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3955),
    [sym_nil] = STATE(3955),
    [sym__atom] = STATE(3955),
    [sym_quoted_atom] = STATE(3955),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3955),
    [sym_charlist] = STATE(3955),
    [sym_sigil] = STATE(3955),
    [sym_list] = STATE(3955),
    [sym_tuple] = STATE(3955),
    [sym_bitstring] = STATE(3955),
    [sym_map] = STATE(3955),
    [sym__nullary_operator] = STATE(3955),
    [sym_unary_operator] = STATE(3955),
    [sym_binary_operator] = STATE(3955),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3955),
    [sym_call] = STATE(3955),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3955),
    [sym_anonymous_function] = STATE(3955),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2589),
    [sym_integer] = ACTIONS(2589),
    [sym_float] = ACTIONS(2589),
    [sym_char] = ACTIONS(2589),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2589),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(866)] = {
    [sym__expression] = STATE(3956),
    [sym_block] = STATE(3956),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3956),
    [sym_nil] = STATE(3956),
    [sym__atom] = STATE(3956),
    [sym_quoted_atom] = STATE(3956),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3956),
    [sym_charlist] = STATE(3956),
    [sym_sigil] = STATE(3956),
    [sym_list] = STATE(3956),
    [sym_tuple] = STATE(3956),
    [sym_bitstring] = STATE(3956),
    [sym_map] = STATE(3956),
    [sym__nullary_operator] = STATE(3956),
    [sym_unary_operator] = STATE(3956),
    [sym_binary_operator] = STATE(3956),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3956),
    [sym_call] = STATE(3956),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3956),
    [sym_anonymous_function] = STATE(3956),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2591),
    [sym_integer] = ACTIONS(2591),
    [sym_float] = ACTIONS(2591),
    [sym_char] = ACTIONS(2591),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2591),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(867)] = {
    [sym__expression] = STATE(3957),
    [sym_block] = STATE(3957),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3957),
    [sym_nil] = STATE(3957),
    [sym__atom] = STATE(3957),
    [sym_quoted_atom] = STATE(3957),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3957),
    [sym_charlist] = STATE(3957),
    [sym_sigil] = STATE(3957),
    [sym_list] = STATE(3957),
    [sym_tuple] = STATE(3957),
    [sym_bitstring] = STATE(3957),
    [sym_map] = STATE(3957),
    [sym__nullary_operator] = STATE(3957),
    [sym_unary_operator] = STATE(3957),
    [sym_binary_operator] = STATE(3957),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3957),
    [sym_call] = STATE(3957),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3957),
    [sym_anonymous_function] = STATE(3957),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2593),
    [sym_integer] = ACTIONS(2593),
    [sym_float] = ACTIONS(2593),
    [sym_char] = ACTIONS(2593),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2593),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(868)] = {
    [sym__expression] = STATE(3199),
    [sym_block] = STATE(3199),
    [sym_identifier] = STATE(57),
    [sym_boolean] = STATE(3199),
    [sym_nil] = STATE(3199),
    [sym__atom] = STATE(3199),
    [sym_quoted_atom] = STATE(3199),
    [sym__quoted_i_double] = STATE(3352),
    [sym__quoted_i_single] = STATE(3353),
    [sym__quoted_i_heredoc_single] = STATE(3354),
    [sym__quoted_i_heredoc_double] = STATE(3355),
    [sym_string] = STATE(3199),
    [sym_charlist] = STATE(3199),
    [sym_sigil] = STATE(3199),
    [sym_list] = STATE(3199),
    [sym_tuple] = STATE(3199),
    [sym_bitstring] = STATE(3199),
    [sym_map] = STATE(3199),
    [sym__nullary_operator] = STATE(3199),
    [sym_unary_operator] = STATE(3199),
    [sym_binary_operator] = STATE(3199),
    [sym_operator_identifier] = STATE(6969),
    [sym_dot] = STATE(3199),
    [sym_call] = STATE(3199),
    [sym__call_without_parentheses] = STATE(3356),
    [sym__call_with_parentheses] = STATE(3357),
    [sym__local_call_without_parentheses] = STATE(3358),
    [sym__local_call_with_parentheses] = STATE(2198),
    [sym__local_call_just_do_block] = STATE(3359),
    [sym__remote_call_without_parentheses] = STATE(3360),
    [sym__remote_call_with_parentheses] = STATE(2199),
    [sym__remote_dot] = STATE(58),
    [sym__anonymous_call] = STATE(2202),
    [sym__anonymous_dot] = STATE(6777),
    [sym__double_call] = STATE(3361),
    [sym_access_call] = STATE(3199),
    [sym_anonymous_function] = STATE(3199),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(590),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2595),
    [sym_integer] = ACTIONS(2595),
    [sym_float] = ACTIONS(2595),
    [sym_char] = ACTIONS(2595),
    [anon_sym_true] = ACTIONS(594),
    [anon_sym_false] = ACTIONS(594),
    [anon_sym_nil] = ACTIONS(596),
    [sym_atom] = ACTIONS(2595),
    [anon_sym_DQUOTE] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(600),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(602),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(610),
    [anon_sym_LT_LT] = ACTIONS(614),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_DOT_DOT] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(618),
    [anon_sym_PLUS] = ACTIONS(620),
    [anon_sym_DASH] = ACTIONS(620),
    [anon_sym_BANG] = ACTIONS(620),
    [anon_sym_CARET] = ACTIONS(620),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(620),
    [anon_sym_not] = ACTIONS(620),
    [anon_sym_AT] = ACTIONS(622),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(626),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(632),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(634),
  },
  [STATE(869)] = {
    [sym__expression] = STATE(3970),
    [sym_block] = STATE(3970),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3970),
    [sym_nil] = STATE(3970),
    [sym__atom] = STATE(3970),
    [sym_quoted_atom] = STATE(3970),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3970),
    [sym_charlist] = STATE(3970),
    [sym_sigil] = STATE(3970),
    [sym_list] = STATE(3970),
    [sym_tuple] = STATE(3970),
    [sym_bitstring] = STATE(3970),
    [sym_map] = STATE(3970),
    [sym__nullary_operator] = STATE(3970),
    [sym_unary_operator] = STATE(3970),
    [sym_binary_operator] = STATE(3970),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3970),
    [sym_call] = STATE(3970),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3970),
    [sym_anonymous_function] = STATE(3970),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2597),
    [sym_integer] = ACTIONS(2597),
    [sym_float] = ACTIONS(2597),
    [sym_char] = ACTIONS(2597),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2597),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(870)] = {
    [sym__expression] = STATE(3971),
    [sym_block] = STATE(3971),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3971),
    [sym_nil] = STATE(3971),
    [sym__atom] = STATE(3971),
    [sym_quoted_atom] = STATE(3971),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3971),
    [sym_charlist] = STATE(3971),
    [sym_sigil] = STATE(3971),
    [sym_list] = STATE(3971),
    [sym_tuple] = STATE(3971),
    [sym_bitstring] = STATE(3971),
    [sym_map] = STATE(3971),
    [sym__nullary_operator] = STATE(3971),
    [sym_unary_operator] = STATE(3971),
    [sym_binary_operator] = STATE(3971),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3971),
    [sym_call] = STATE(3971),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3971),
    [sym_anonymous_function] = STATE(3971),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2599),
    [sym_integer] = ACTIONS(2599),
    [sym_float] = ACTIONS(2599),
    [sym_char] = ACTIONS(2599),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2599),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(871)] = {
    [sym__expression] = STATE(3972),
    [sym_block] = STATE(3972),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3972),
    [sym_nil] = STATE(3972),
    [sym__atom] = STATE(3972),
    [sym_quoted_atom] = STATE(3972),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3972),
    [sym_charlist] = STATE(3972),
    [sym_sigil] = STATE(3972),
    [sym_list] = STATE(3972),
    [sym_tuple] = STATE(3972),
    [sym_bitstring] = STATE(3972),
    [sym_map] = STATE(3972),
    [sym__nullary_operator] = STATE(3972),
    [sym_unary_operator] = STATE(3972),
    [sym_binary_operator] = STATE(3972),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3972),
    [sym_call] = STATE(3972),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3972),
    [sym_anonymous_function] = STATE(3972),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2601),
    [sym_integer] = ACTIONS(2601),
    [sym_float] = ACTIONS(2601),
    [sym_char] = ACTIONS(2601),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2601),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(872)] = {
    [sym__expression] = STATE(3973),
    [sym_block] = STATE(3973),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3973),
    [sym_nil] = STATE(3973),
    [sym__atom] = STATE(3973),
    [sym_quoted_atom] = STATE(3973),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3973),
    [sym_charlist] = STATE(3973),
    [sym_sigil] = STATE(3973),
    [sym_list] = STATE(3973),
    [sym_tuple] = STATE(3973),
    [sym_bitstring] = STATE(3973),
    [sym_map] = STATE(3973),
    [sym__nullary_operator] = STATE(3973),
    [sym_unary_operator] = STATE(3973),
    [sym_binary_operator] = STATE(3973),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3973),
    [sym_call] = STATE(3973),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3973),
    [sym_anonymous_function] = STATE(3973),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2603),
    [sym_integer] = ACTIONS(2603),
    [sym_float] = ACTIONS(2603),
    [sym_char] = ACTIONS(2603),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2603),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(873)] = {
    [sym__expression] = STATE(3974),
    [sym_block] = STATE(3974),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3974),
    [sym_nil] = STATE(3974),
    [sym__atom] = STATE(3974),
    [sym_quoted_atom] = STATE(3974),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3974),
    [sym_charlist] = STATE(3974),
    [sym_sigil] = STATE(3974),
    [sym_list] = STATE(3974),
    [sym_tuple] = STATE(3974),
    [sym_bitstring] = STATE(3974),
    [sym_map] = STATE(3974),
    [sym__nullary_operator] = STATE(3974),
    [sym_unary_operator] = STATE(3974),
    [sym_binary_operator] = STATE(3974),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3974),
    [sym_call] = STATE(3974),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3974),
    [sym_anonymous_function] = STATE(3974),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2605),
    [sym_integer] = ACTIONS(2605),
    [sym_float] = ACTIONS(2605),
    [sym_char] = ACTIONS(2605),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2605),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(874)] = {
    [sym__expression] = STATE(3975),
    [sym_block] = STATE(3975),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3975),
    [sym_nil] = STATE(3975),
    [sym__atom] = STATE(3975),
    [sym_quoted_atom] = STATE(3975),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3975),
    [sym_charlist] = STATE(3975),
    [sym_sigil] = STATE(3975),
    [sym_list] = STATE(3975),
    [sym_tuple] = STATE(3975),
    [sym_bitstring] = STATE(3975),
    [sym_map] = STATE(3975),
    [sym__nullary_operator] = STATE(3975),
    [sym_unary_operator] = STATE(3975),
    [sym_binary_operator] = STATE(3975),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3975),
    [sym_call] = STATE(3975),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3975),
    [sym_anonymous_function] = STATE(3975),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2607),
    [sym_integer] = ACTIONS(2607),
    [sym_float] = ACTIONS(2607),
    [sym_char] = ACTIONS(2607),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2607),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(875)] = {
    [sym__expression] = STATE(3976),
    [sym_block] = STATE(3976),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3976),
    [sym_nil] = STATE(3976),
    [sym__atom] = STATE(3976),
    [sym_quoted_atom] = STATE(3976),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3976),
    [sym_charlist] = STATE(3976),
    [sym_sigil] = STATE(3976),
    [sym_list] = STATE(3976),
    [sym_tuple] = STATE(3976),
    [sym_bitstring] = STATE(3976),
    [sym_map] = STATE(3976),
    [sym__nullary_operator] = STATE(3976),
    [sym_unary_operator] = STATE(3976),
    [sym_binary_operator] = STATE(3976),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3976),
    [sym_call] = STATE(3976),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3976),
    [sym_anonymous_function] = STATE(3976),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2609),
    [sym_integer] = ACTIONS(2609),
    [sym_float] = ACTIONS(2609),
    [sym_char] = ACTIONS(2609),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2609),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(876)] = {
    [sym__expression] = STATE(3977),
    [sym_block] = STATE(3977),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3977),
    [sym_nil] = STATE(3977),
    [sym__atom] = STATE(3977),
    [sym_quoted_atom] = STATE(3977),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3977),
    [sym_charlist] = STATE(3977),
    [sym_sigil] = STATE(3977),
    [sym_list] = STATE(3977),
    [sym_tuple] = STATE(3977),
    [sym_bitstring] = STATE(3977),
    [sym_map] = STATE(3977),
    [sym__nullary_operator] = STATE(3977),
    [sym_unary_operator] = STATE(3977),
    [sym_binary_operator] = STATE(3977),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3977),
    [sym_call] = STATE(3977),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3977),
    [sym_anonymous_function] = STATE(3977),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2611),
    [sym_integer] = ACTIONS(2611),
    [sym_float] = ACTIONS(2611),
    [sym_char] = ACTIONS(2611),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2611),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(877)] = {
    [sym__expression] = STATE(3978),
    [sym_block] = STATE(3978),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3978),
    [sym_nil] = STATE(3978),
    [sym__atom] = STATE(3978),
    [sym_quoted_atom] = STATE(3978),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3978),
    [sym_charlist] = STATE(3978),
    [sym_sigil] = STATE(3978),
    [sym_list] = STATE(3978),
    [sym_tuple] = STATE(3978),
    [sym_bitstring] = STATE(3978),
    [sym_map] = STATE(3978),
    [sym__nullary_operator] = STATE(3978),
    [sym_unary_operator] = STATE(3978),
    [sym_binary_operator] = STATE(3978),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3978),
    [sym_call] = STATE(3978),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3978),
    [sym_anonymous_function] = STATE(3978),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2613),
    [sym_integer] = ACTIONS(2613),
    [sym_float] = ACTIONS(2613),
    [sym_char] = ACTIONS(2613),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2613),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(878)] = {
    [sym__expression] = STATE(3080),
    [sym_block] = STATE(3080),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(3080),
    [sym_nil] = STATE(3080),
    [sym__atom] = STATE(3080),
    [sym_quoted_atom] = STATE(3080),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(3080),
    [sym_charlist] = STATE(3080),
    [sym_sigil] = STATE(3080),
    [sym_list] = STATE(3080),
    [sym_tuple] = STATE(3080),
    [sym_bitstring] = STATE(3080),
    [sym_map] = STATE(3080),
    [sym__nullary_operator] = STATE(3080),
    [sym_unary_operator] = STATE(3080),
    [sym_binary_operator] = STATE(3080),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(3080),
    [sym_call] = STATE(3080),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(3080),
    [sym_anonymous_function] = STATE(3080),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2529),
    [sym_integer] = ACTIONS(2529),
    [sym_float] = ACTIONS(2529),
    [sym_char] = ACTIONS(2529),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2529),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(879)] = {
    [sym__expression] = STATE(3207),
    [sym_block] = STATE(3207),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3207),
    [sym_nil] = STATE(3207),
    [sym__atom] = STATE(3207),
    [sym_quoted_atom] = STATE(3207),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3207),
    [sym_charlist] = STATE(3207),
    [sym_sigil] = STATE(3207),
    [sym_list] = STATE(3207),
    [sym_tuple] = STATE(3207),
    [sym_bitstring] = STATE(3207),
    [sym_map] = STATE(3207),
    [sym__nullary_operator] = STATE(3207),
    [sym_unary_operator] = STATE(3207),
    [sym_binary_operator] = STATE(3207),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3207),
    [sym_call] = STATE(3207),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3207),
    [sym_anonymous_function] = STATE(3207),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(1443),
    [sym_integer] = ACTIONS(1443),
    [sym_float] = ACTIONS(1443),
    [sym_char] = ACTIONS(1443),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1443),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(880)] = {
    [sym__expression] = STATE(3264),
    [sym_block] = STATE(3264),
    [sym_identifier] = STATE(46),
    [sym_boolean] = STATE(3264),
    [sym_nil] = STATE(3264),
    [sym__atom] = STATE(3264),
    [sym_quoted_atom] = STATE(3264),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3264),
    [sym_charlist] = STATE(3264),
    [sym_sigil] = STATE(3264),
    [sym_list] = STATE(3264),
    [sym_tuple] = STATE(3264),
    [sym_bitstring] = STATE(3264),
    [sym_map] = STATE(3264),
    [sym__nullary_operator] = STATE(3264),
    [sym_unary_operator] = STATE(3264),
    [sym_binary_operator] = STATE(3264),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3264),
    [sym_call] = STATE(3264),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(47),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3264),
    [sym_anonymous_function] = STATE(3264),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2615),
    [sym_integer] = ACTIONS(2615),
    [sym_float] = ACTIONS(2615),
    [sym_char] = ACTIONS(2615),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2615),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(493),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(497),
    [anon_sym_PLUS] = ACTIONS(499),
    [anon_sym_DASH] = ACTIONS(499),
    [anon_sym_BANG] = ACTIONS(499),
    [anon_sym_CARET] = ACTIONS(499),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(499),
    [anon_sym_not] = ACTIONS(499),
    [anon_sym_AT] = ACTIONS(501),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(503),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(881)] = {
    [sym__expression] = STATE(4029),
    [sym_block] = STATE(4029),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(4029),
    [sym_nil] = STATE(4029),
    [sym__atom] = STATE(4029),
    [sym_quoted_atom] = STATE(4029),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(4029),
    [sym_charlist] = STATE(4029),
    [sym_sigil] = STATE(4029),
    [sym_list] = STATE(4029),
    [sym_tuple] = STATE(4029),
    [sym_bitstring] = STATE(4029),
    [sym_map] = STATE(4029),
    [sym__nullary_operator] = STATE(4029),
    [sym_unary_operator] = STATE(4029),
    [sym_binary_operator] = STATE(4029),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(4029),
    [sym_call] = STATE(4029),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(4029),
    [sym_anonymous_function] = STATE(4029),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(1465),
    [sym_integer] = ACTIONS(1465),
    [sym_float] = ACTIONS(1465),
    [sym_char] = ACTIONS(1465),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(1465),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(882)] = {
    [sym__expression] = STATE(3895),
    [sym_block] = STATE(3895),
    [sym_identifier] = STATE(59),
    [sym_boolean] = STATE(3895),
    [sym_nil] = STATE(3895),
    [sym__atom] = STATE(3895),
    [sym_quoted_atom] = STATE(3895),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(3895),
    [sym_charlist] = STATE(3895),
    [sym_sigil] = STATE(3895),
    [sym_list] = STATE(3895),
    [sym_tuple] = STATE(3895),
    [sym_bitstring] = STATE(3895),
    [sym_map] = STATE(3895),
    [sym__nullary_operator] = STATE(3895),
    [sym_unary_operator] = STATE(3895),
    [sym_binary_operator] = STATE(3895),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(3895),
    [sym_call] = STATE(3895),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(50),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(3895),
    [sym_anonymous_function] = STATE(3895),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2617),
    [sym_integer] = ACTIONS(2617),
    [sym_float] = ACTIONS(2617),
    [sym_char] = ACTIONS(2617),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2617),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(694),
    [anon_sym_DASH] = ACTIONS(694),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(694),
    [anon_sym_not] = ACTIONS(694),
    [anon_sym_AT] = ACTIONS(696),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(700),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(883)] = {
    [sym__expression] = STATE(3289),
    [sym_block] = STATE(3289),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3289),
    [sym_nil] = STATE(3289),
    [sym__atom] = STATE(3289),
    [sym_quoted_atom] = STATE(3289),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3289),
    [sym_charlist] = STATE(3289),
    [sym_sigil] = STATE(3289),
    [sym_list] = STATE(3289),
    [sym_tuple] = STATE(3289),
    [sym_bitstring] = STATE(3289),
    [sym_map] = STATE(3289),
    [sym__nullary_operator] = STATE(3289),
    [sym_unary_operator] = STATE(3289),
    [sym_binary_operator] = STATE(3289),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3289),
    [sym_call] = STATE(3289),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3289),
    [sym_anonymous_function] = STATE(3289),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(1487),
    [sym_integer] = ACTIONS(1487),
    [sym_float] = ACTIONS(1487),
    [sym_char] = ACTIONS(1487),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1487),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(884)] = {
    [sym__expression] = STATE(4186),
    [sym_block] = STATE(4186),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4186),
    [sym_nil] = STATE(4186),
    [sym__atom] = STATE(4186),
    [sym_quoted_atom] = STATE(4186),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4186),
    [sym_charlist] = STATE(4186),
    [sym_sigil] = STATE(4186),
    [sym_list] = STATE(4186),
    [sym_tuple] = STATE(4186),
    [sym_bitstring] = STATE(4186),
    [sym_map] = STATE(4186),
    [sym__nullary_operator] = STATE(4186),
    [sym_unary_operator] = STATE(4186),
    [sym_binary_operator] = STATE(4186),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4186),
    [sym_call] = STATE(4186),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4186),
    [sym_anonymous_function] = STATE(4186),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2619),
    [sym_integer] = ACTIONS(2619),
    [sym_float] = ACTIONS(2619),
    [sym_char] = ACTIONS(2619),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2619),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(885)] = {
    [sym__expression] = STATE(4030),
    [sym_block] = STATE(4030),
    [sym_identifier] = STATE(65),
    [sym_boolean] = STATE(4030),
    [sym_nil] = STATE(4030),
    [sym__atom] = STATE(4030),
    [sym_quoted_atom] = STATE(4030),
    [sym__quoted_i_double] = STATE(2144),
    [sym__quoted_i_single] = STATE(2145),
    [sym__quoted_i_heredoc_single] = STATE(2146),
    [sym__quoted_i_heredoc_double] = STATE(2147),
    [sym_string] = STATE(4030),
    [sym_charlist] = STATE(4030),
    [sym_sigil] = STATE(4030),
    [sym_list] = STATE(4030),
    [sym_tuple] = STATE(4030),
    [sym_bitstring] = STATE(4030),
    [sym_map] = STATE(4030),
    [sym__nullary_operator] = STATE(4030),
    [sym_unary_operator] = STATE(4030),
    [sym_binary_operator] = STATE(4030),
    [sym_operator_identifier] = STATE(6959),
    [sym_dot] = STATE(4030),
    [sym_call] = STATE(4030),
    [sym__call_without_parentheses] = STATE(2148),
    [sym__call_with_parentheses] = STATE(2149),
    [sym__local_call_without_parentheses] = STATE(2150),
    [sym__local_call_with_parentheses] = STATE(1643),
    [sym__local_call_just_do_block] = STATE(2153),
    [sym__remote_call_without_parentheses] = STATE(2156),
    [sym__remote_call_with_parentheses] = STATE(1645),
    [sym__remote_dot] = STATE(64),
    [sym__anonymous_call] = STATE(1649),
    [sym__anonymous_dot] = STATE(6814),
    [sym__double_call] = STATE(2164),
    [sym_access_call] = STATE(4030),
    [sym_anonymous_function] = STATE(4030),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(359),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2621),
    [sym_integer] = ACTIONS(2621),
    [sym_float] = ACTIONS(2621),
    [sym_char] = ACTIONS(2621),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_nil] = ACTIONS(367),
    [sym_atom] = ACTIONS(2621),
    [anon_sym_DQUOTE] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LBRACK] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(381),
    [anon_sym_LT_LT] = ACTIONS(385),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(1383),
    [anon_sym_AMP] = ACTIONS(650),
    [anon_sym_PLUS] = ACTIONS(652),
    [anon_sym_DASH] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(652),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(652),
    [anon_sym_not] = ACTIONS(652),
    [anon_sym_AT] = ACTIONS(654),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(395),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(656),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(401),
  },
  [STATE(886)] = {
    [sym__expression] = STATE(1380),
    [sym_block] = STATE(1380),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1380),
    [sym_nil] = STATE(1380),
    [sym__atom] = STATE(1380),
    [sym_quoted_atom] = STATE(1380),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1380),
    [sym_charlist] = STATE(1380),
    [sym_sigil] = STATE(1380),
    [sym_list] = STATE(1380),
    [sym_tuple] = STATE(1380),
    [sym_bitstring] = STATE(1380),
    [sym_map] = STATE(1380),
    [sym__nullary_operator] = STATE(1380),
    [sym_unary_operator] = STATE(1380),
    [sym_binary_operator] = STATE(1380),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1380),
    [sym_call] = STATE(1380),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1380),
    [sym_anonymous_function] = STATE(1380),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1501),
    [sym_integer] = ACTIONS(1501),
    [sym_float] = ACTIONS(1501),
    [sym_char] = ACTIONS(1501),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(1501),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(887)] = {
    [sym__expression] = STATE(3299),
    [sym_block] = STATE(3299),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3299),
    [sym_nil] = STATE(3299),
    [sym__atom] = STATE(3299),
    [sym_quoted_atom] = STATE(3299),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3299),
    [sym_charlist] = STATE(3299),
    [sym_sigil] = STATE(3299),
    [sym_list] = STATE(3299),
    [sym_tuple] = STATE(3299),
    [sym_bitstring] = STATE(3299),
    [sym_map] = STATE(3299),
    [sym__nullary_operator] = STATE(3299),
    [sym_unary_operator] = STATE(3299),
    [sym_binary_operator] = STATE(3299),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3299),
    [sym_call] = STATE(3299),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3299),
    [sym_anonymous_function] = STATE(3299),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2623),
    [sym_integer] = ACTIONS(2623),
    [sym_float] = ACTIONS(2623),
    [sym_char] = ACTIONS(2623),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2623),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(888)] = {
    [sym__expression] = STATE(4446),
    [sym_block] = STATE(4446),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4446),
    [sym_nil] = STATE(4446),
    [sym__atom] = STATE(4446),
    [sym_quoted_atom] = STATE(4446),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4446),
    [sym_charlist] = STATE(4446),
    [sym_sigil] = STATE(4446),
    [sym_list] = STATE(4446),
    [sym_tuple] = STATE(4446),
    [sym_bitstring] = STATE(4446),
    [sym_map] = STATE(4446),
    [sym__nullary_operator] = STATE(4446),
    [sym_unary_operator] = STATE(4446),
    [sym_binary_operator] = STATE(4446),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4446),
    [sym_call] = STATE(4446),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4446),
    [sym_anonymous_function] = STATE(4446),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(1567),
    [sym_integer] = ACTIONS(1567),
    [sym_float] = ACTIONS(1567),
    [sym_char] = ACTIONS(1567),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(1567),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(889)] = {
    [sym__expression] = STATE(4190),
    [sym_block] = STATE(4190),
    [sym_identifier] = STATE(60),
    [sym_boolean] = STATE(4190),
    [sym_nil] = STATE(4190),
    [sym__atom] = STATE(4190),
    [sym_quoted_atom] = STATE(4190),
    [sym__quoted_i_double] = STATE(4100),
    [sym__quoted_i_single] = STATE(4150),
    [sym__quoted_i_heredoc_single] = STATE(4155),
    [sym__quoted_i_heredoc_double] = STATE(4172),
    [sym_string] = STATE(4190),
    [sym_charlist] = STATE(4190),
    [sym_sigil] = STATE(4190),
    [sym_list] = STATE(4190),
    [sym_tuple] = STATE(4190),
    [sym_bitstring] = STATE(4190),
    [sym_map] = STATE(4190),
    [sym__nullary_operator] = STATE(4190),
    [sym_unary_operator] = STATE(4190),
    [sym_binary_operator] = STATE(4190),
    [sym_operator_identifier] = STATE(6927),
    [sym_dot] = STATE(4190),
    [sym_call] = STATE(4190),
    [sym__call_without_parentheses] = STATE(4193),
    [sym__call_with_parentheses] = STATE(4325),
    [sym__local_call_without_parentheses] = STATE(4339),
    [sym__local_call_with_parentheses] = STATE(3049),
    [sym__local_call_just_do_block] = STATE(4214),
    [sym__remote_call_without_parentheses] = STATE(4363),
    [sym__remote_call_with_parentheses] = STATE(3085),
    [sym__remote_dot] = STATE(49),
    [sym__anonymous_call] = STATE(3086),
    [sym__anonymous_dot] = STATE(6853),
    [sym__double_call] = STATE(4062),
    [sym_access_call] = STATE(4190),
    [sym_anonymous_function] = STATE(4190),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(13),
    [aux_sym_identifier_token1] = ACTIONS(15),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(15),
    [sym_alias] = ACTIONS(2625),
    [sym_integer] = ACTIONS(2625),
    [sym_float] = ACTIONS(2625),
    [sym_char] = ACTIONS(2625),
    [anon_sym_true] = ACTIONS(19),
    [anon_sym_false] = ACTIONS(19),
    [anon_sym_nil] = ACTIONS(21),
    [sym_atom] = ACTIONS(2625),
    [anon_sym_DQUOTE] = ACTIONS(23),
    [anon_sym_SQUOTE] = ACTIONS(25),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(27),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(29),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LBRACK] = ACTIONS(33),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LT_LT] = ACTIONS(39),
    [anon_sym_PERCENT] = ACTIONS(41),
    [anon_sym_DOT_DOT] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(45),
    [anon_sym_PLUS] = ACTIONS(47),
    [anon_sym_DASH] = ACTIONS(47),
    [anon_sym_BANG] = ACTIONS(47),
    [anon_sym_CARET] = ACTIONS(47),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(47),
    [anon_sym_not] = ACTIONS(47),
    [anon_sym_AT] = ACTIONS(49),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(51),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(53),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(57),
  },
  [STATE(890)] = {
    [sym__expression] = STATE(1688),
    [sym_block] = STATE(1688),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1688),
    [sym_nil] = STATE(1688),
    [sym__atom] = STATE(1688),
    [sym_quoted_atom] = STATE(1688),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1688),
    [sym_charlist] = STATE(1688),
    [sym_sigil] = STATE(1688),
    [sym_list] = STATE(1688),
    [sym_tuple] = STATE(1688),
    [sym_bitstring] = STATE(1688),
    [sym_map] = STATE(1688),
    [sym__nullary_operator] = STATE(1688),
    [sym_unary_operator] = STATE(1688),
    [sym_binary_operator] = STATE(1688),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1688),
    [sym_call] = STATE(1688),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1688),
    [sym_anonymous_function] = STATE(1688),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(1521),
    [sym_integer] = ACTIONS(1521),
    [sym_float] = ACTIONS(1521),
    [sym_char] = ACTIONS(1521),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(1521),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(891)] = {
    [sym__expression] = STATE(1381),
    [sym_block] = STATE(1381),
    [sym_identifier] = STATE(15),
    [sym_boolean] = STATE(1381),
    [sym_nil] = STATE(1381),
    [sym__atom] = STATE(1381),
    [sym_quoted_atom] = STATE(1381),
    [sym__quoted_i_double] = STATE(1257),
    [sym__quoted_i_single] = STATE(1258),
    [sym__quoted_i_heredoc_single] = STATE(1259),
    [sym__quoted_i_heredoc_double] = STATE(1260),
    [sym_string] = STATE(1381),
    [sym_charlist] = STATE(1381),
    [sym_sigil] = STATE(1381),
    [sym_list] = STATE(1381),
    [sym_tuple] = STATE(1381),
    [sym_bitstring] = STATE(1381),
    [sym_map] = STATE(1381),
    [sym__nullary_operator] = STATE(1381),
    [sym_unary_operator] = STATE(1381),
    [sym_binary_operator] = STATE(1381),
    [sym_operator_identifier] = STATE(6954),
    [sym_dot] = STATE(1381),
    [sym_call] = STATE(1381),
    [sym__call_without_parentheses] = STATE(1261),
    [sym__call_with_parentheses] = STATE(1262),
    [sym__local_call_without_parentheses] = STATE(1263),
    [sym__local_call_with_parentheses] = STATE(1091),
    [sym__local_call_just_do_block] = STATE(1265),
    [sym__remote_call_without_parentheses] = STATE(1266),
    [sym__remote_call_with_parentheses] = STATE(1093),
    [sym__remote_dot] = STATE(18),
    [sym__anonymous_call] = STATE(1088),
    [sym__anonymous_dot] = STATE(6791),
    [sym__double_call] = STATE(1270),
    [sym_access_call] = STATE(1381),
    [sym_anonymous_function] = STATE(1381),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(237),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2627),
    [sym_integer] = ACTIONS(2627),
    [sym_float] = ACTIONS(2627),
    [sym_char] = ACTIONS(2627),
    [anon_sym_true] = ACTIONS(241),
    [anon_sym_false] = ACTIONS(241),
    [anon_sym_nil] = ACTIONS(243),
    [sym_atom] = ACTIONS(2627),
    [anon_sym_DQUOTE] = ACTIONS(245),
    [anon_sym_SQUOTE] = ACTIONS(247),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(249),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(251),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(271),
    [anon_sym_PLUS] = ACTIONS(276),
    [anon_sym_DASH] = ACTIONS(276),
    [anon_sym_BANG] = ACTIONS(276),
    [anon_sym_CARET] = ACTIONS(276),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(276),
    [anon_sym_not] = ACTIONS(276),
    [anon_sym_AT] = ACTIONS(278),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(280),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(284),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(289),
  },
  [STATE(892)] = {
    [sym__expression] = STATE(4157),
    [sym_block] = STATE(4157),
    [sym_identifier] = STATE(63),
    [sym_boolean] = STATE(4157),
    [sym_nil] = STATE(4157),
    [sym__atom] = STATE(4157),
    [sym_quoted_atom] = STATE(4157),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4157),
    [sym_charlist] = STATE(4157),
    [sym_sigil] = STATE(4157),
    [sym_list] = STATE(4157),
    [sym_tuple] = STATE(4157),
    [sym_bitstring] = STATE(4157),
    [sym_map] = STATE(4157),
    [sym__nullary_operator] = STATE(4157),
    [sym_unary_operator] = STATE(4157),
    [sym_binary_operator] = STATE(4157),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4157),
    [sym_call] = STATE(4157),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(43),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4157),
    [sym_anonymous_function] = STATE(4157),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(686),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(686),
    [sym_alias] = ACTIONS(2629),
    [sym_integer] = ACTIONS(2629),
    [sym_float] = ACTIONS(2629),
    [sym_char] = ACTIONS(2629),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2629),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1529),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_PLUS] = ACTIONS(1533),
    [anon_sym_DASH] = ACTIONS(1533),
    [anon_sym_BANG] = ACTIONS(1533),
    [anon_sym_CARET] = ACTIONS(1533),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1533),
    [anon_sym_not] = ACTIONS(1533),
    [anon_sym_AT] = ACTIONS(1535),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1537),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(893)] = {
    [sym__expression] = STATE(4493),
    [sym_block] = STATE(4493),
    [sym_identifier] = STATE(88),
    [sym_boolean] = STATE(4493),
    [sym_nil] = STATE(4493),
    [sym__atom] = STATE(4493),
    [sym_quoted_atom] = STATE(4493),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4493),
    [sym_charlist] = STATE(4493),
    [sym_sigil] = STATE(4493),
    [sym_list] = STATE(4493),
    [sym_tuple] = STATE(4493),
    [sym_bitstring] = STATE(4493),
    [sym_map] = STATE(4493),
    [sym__nullary_operator] = STATE(4493),
    [sym_unary_operator] = STATE(4493),
    [sym_binary_operator] = STATE(4493),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4493),
    [sym_call] = STATE(4493),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(66),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4493),
    [sym_anonymous_function] = STATE(4493),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2631),
    [sym_integer] = ACTIONS(2631),
    [sym_float] = ACTIONS(2631),
    [sym_char] = ACTIONS(2631),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2631),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1507),
    [anon_sym_PLUS] = ACTIONS(1509),
    [anon_sym_DASH] = ACTIONS(1509),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_CARET] = ACTIONS(1509),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1509),
    [anon_sym_not] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1511),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1513),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(894)] = {
    [sym__expression] = STATE(4189),
    [sym_block] = STATE(4189),
    [sym_identifier] = STATE(56),
    [sym_boolean] = STATE(4189),
    [sym_nil] = STATE(4189),
    [sym__atom] = STATE(4189),
    [sym_quoted_atom] = STATE(4189),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(4189),
    [sym_charlist] = STATE(4189),
    [sym_sigil] = STATE(4189),
    [sym_list] = STATE(4189),
    [sym_tuple] = STATE(4189),
    [sym_bitstring] = STATE(4189),
    [sym_map] = STATE(4189),
    [sym__nullary_operator] = STATE(4189),
    [sym_unary_operator] = STATE(4189),
    [sym_binary_operator] = STATE(4189),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(4189),
    [sym_call] = STATE(4189),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(45),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(4189),
    [sym_anonymous_function] = STATE(4189),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(1403),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1403),
    [sym_alias] = ACTIONS(2633),
    [sym_integer] = ACTIONS(2633),
    [sym_float] = ACTIONS(2633),
    [sym_char] = ACTIONS(2633),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2633),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1413),
    [anon_sym_DASH] = ACTIONS(1413),
    [anon_sym_BANG] = ACTIONS(1413),
    [anon_sym_CARET] = ACTIONS(1413),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1413),
    [anon_sym_not] = ACTIONS(1413),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1417),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(895)] = {
    [sym__expression] = STATE(1689),
    [sym_block] = STATE(1689),
    [sym_identifier] = STATE(19),
    [sym_boolean] = STATE(1689),
    [sym_nil] = STATE(1689),
    [sym__atom] = STATE(1689),
    [sym_quoted_atom] = STATE(1689),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(1689),
    [sym_charlist] = STATE(1689),
    [sym_sigil] = STATE(1689),
    [sym_list] = STATE(1689),
    [sym_tuple] = STATE(1689),
    [sym_bitstring] = STATE(1689),
    [sym_map] = STATE(1689),
    [sym__nullary_operator] = STATE(1689),
    [sym_unary_operator] = STATE(1689),
    [sym_binary_operator] = STATE(1689),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(1689),
    [sym_call] = STATE(1689),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(14),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(1689),
    [sym_anonymous_function] = STATE(1689),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2635),
    [sym_integer] = ACTIONS(2635),
    [sym_float] = ACTIONS(2635),
    [sym_char] = ACTIONS(2635),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2635),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(211),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(219),
    [anon_sym_PLUS] = ACTIONS(221),
    [anon_sym_DASH] = ACTIONS(221),
    [anon_sym_BANG] = ACTIONS(221),
    [anon_sym_CARET] = ACTIONS(221),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(221),
    [anon_sym_not] = ACTIONS(221),
    [anon_sym_AT] = ACTIONS(223),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(229),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(896)] = {
    [sym__expression] = STATE(2076),
    [sym_block] = STATE(2076),
    [sym_identifier] = STATE(21),
    [sym_boolean] = STATE(2076),
    [sym_nil] = STATE(2076),
    [sym__atom] = STATE(2076),
    [sym_quoted_atom] = STATE(2076),
    [sym__quoted_i_double] = STATE(1778),
    [sym__quoted_i_single] = STATE(1779),
    [sym__quoted_i_heredoc_single] = STATE(1617),
    [sym__quoted_i_heredoc_double] = STATE(1644),
    [sym_string] = STATE(2076),
    [sym_charlist] = STATE(2076),
    [sym_sigil] = STATE(2076),
    [sym_list] = STATE(2076),
    [sym_tuple] = STATE(2076),
    [sym_bitstring] = STATE(2076),
    [sym_map] = STATE(2076),
    [sym__nullary_operator] = STATE(2076),
    [sym_unary_operator] = STATE(2076),
    [sym_binary_operator] = STATE(2076),
    [sym_operator_identifier] = STATE(6872),
    [sym_dot] = STATE(2076),
    [sym_call] = STATE(2076),
    [sym__call_without_parentheses] = STATE(1646),
    [sym__call_with_parentheses] = STATE(1647),
    [sym__local_call_without_parentheses] = STATE(1648),
    [sym__local_call_with_parentheses] = STATE(1300),
    [sym__local_call_just_do_block] = STATE(1653),
    [sym__remote_call_without_parentheses] = STATE(1784),
    [sym__remote_call_with_parentheses] = STATE(1193),
    [sym__remote_dot] = STATE(17),
    [sym__anonymous_call] = STATE(1194),
    [sym__anonymous_dot] = STATE(6792),
    [sym__double_call] = STATE(1786),
    [sym_access_call] = STATE(2076),
    [sym_anonymous_function] = STATE(2076),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1431),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2637),
    [sym_integer] = ACTIONS(2637),
    [sym_float] = ACTIONS(2637),
    [sym_char] = ACTIONS(2637),
    [anon_sym_true] = ACTIONS(69),
    [anon_sym_false] = ACTIONS(69),
    [anon_sym_nil] = ACTIONS(71),
    [sym_atom] = ACTIONS(2637),
    [anon_sym_DQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE] = ACTIONS(75),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(85),
    [anon_sym_LT_LT] = ACTIONS(89),
    [anon_sym_PERCENT] = ACTIONS(91),
    [anon_sym_DOT_DOT] = ACTIONS(93),
    [anon_sym_AMP] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(97),
    [anon_sym_not] = ACTIONS(97),
    [anon_sym_AT] = ACTIONS(99),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(111),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(115),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(117),
  },
  [STATE(897)] = {
    [sym__expression] = STATE(2307),
    [sym_block] = STATE(2307),
    [sym_identifier] = STATE(23),
    [sym_boolean] = STATE(2307),
    [sym_nil] = STATE(2307),
    [sym__atom] = STATE(2307),
    [sym_quoted_atom] = STATE(2307),
    [sym__quoted_i_double] = STATE(1931),
    [sym__quoted_i_single] = STATE(2019),
    [sym__quoted_i_heredoc_single] = STATE(2053),
    [sym__quoted_i_heredoc_double] = STATE(2054),
    [sym_string] = STATE(2307),
    [sym_charlist] = STATE(2307),
    [sym_sigil] = STATE(2307),
    [sym_list] = STATE(2307),
    [sym_tuple] = STATE(2307),
    [sym_bitstring] = STATE(2307),
    [sym_map] = STATE(2307),
    [sym__nullary_operator] = STATE(2307),
    [sym_unary_operator] = STATE(2307),
    [sym_binary_operator] = STATE(2307),
    [sym_operator_identifier] = STATE(6949),
    [sym_dot] = STATE(2307),
    [sym_call] = STATE(2307),
    [sym__call_without_parentheses] = STATE(2055),
    [sym__call_with_parentheses] = STATE(2056),
    [sym__local_call_without_parentheses] = STATE(2057),
    [sym__local_call_with_parentheses] = STATE(1479),
    [sym__local_call_just_do_block] = STATE(2061),
    [sym__remote_call_without_parentheses] = STATE(2062),
    [sym__remote_call_with_parentheses] = STATE(1481),
    [sym__remote_dot] = STATE(16),
    [sym__anonymous_call] = STATE(1390),
    [sym__anonymous_dot] = STATE(6865),
    [sym__double_call] = STATE(2015),
    [sym_access_call] = STATE(2307),
    [sym_anonymous_function] = STATE(2307),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(968),
    [aux_sym_identifier_token1] = ACTIONS(65),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(65),
    [sym_alias] = ACTIONS(2639),
    [sym_integer] = ACTIONS(2639),
    [sym_float] = ACTIONS(2639),
    [sym_char] = ACTIONS(2639),
    [anon_sym_true] = ACTIONS(972),
    [anon_sym_false] = ACTIONS(972),
    [anon_sym_nil] = ACTIONS(974),
    [sym_atom] = ACTIONS(2639),
    [anon_sym_DQUOTE] = ACTIONS(976),
    [anon_sym_SQUOTE] = ACTIONS(978),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(980),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(982),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LBRACK] = ACTIONS(986),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(988),
    [anon_sym_LT_LT] = ACTIONS(990),
    [anon_sym_PERCENT] = ACTIONS(992),
    [anon_sym_DOT_DOT] = ACTIONS(994),
    [anon_sym_AMP] = ACTIONS(996),
    [anon_sym_PLUS] = ACTIONS(998),
    [anon_sym_DASH] = ACTIONS(998),
    [anon_sym_BANG] = ACTIONS(998),
    [anon_sym_CARET] = ACTIONS(998),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(998),
    [anon_sym_not] = ACTIONS(998),
    [anon_sym_AT] = ACTIONS(1000),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1004),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1006),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1008),
  },
  [STATE(898)] = {
    [sym__expression] = STATE(4589),
    [sym_block] = STATE(4589),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4589),
    [sym_nil] = STATE(4589),
    [sym__atom] = STATE(4589),
    [sym_quoted_atom] = STATE(4589),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4589),
    [sym_charlist] = STATE(4589),
    [sym_sigil] = STATE(4589),
    [sym_list] = STATE(4589),
    [sym_tuple] = STATE(4589),
    [sym_bitstring] = STATE(4589),
    [sym_map] = STATE(4589),
    [sym__nullary_operator] = STATE(4589),
    [sym_unary_operator] = STATE(4589),
    [sym_binary_operator] = STATE(4589),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4589),
    [sym_call] = STATE(4589),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4589),
    [sym_anonymous_function] = STATE(4589),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2641),
    [sym_integer] = ACTIONS(2641),
    [sym_float] = ACTIONS(2641),
    [sym_char] = ACTIONS(2641),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2641),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(899)] = {
    [sym__expression] = STATE(4571),
    [sym_block] = STATE(4571),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4571),
    [sym_nil] = STATE(4571),
    [sym__atom] = STATE(4571),
    [sym_quoted_atom] = STATE(4571),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4571),
    [sym_charlist] = STATE(4571),
    [sym_sigil] = STATE(4571),
    [sym_list] = STATE(4571),
    [sym_tuple] = STATE(4571),
    [sym_bitstring] = STATE(4571),
    [sym_map] = STATE(4571),
    [sym__nullary_operator] = STATE(4571),
    [sym_unary_operator] = STATE(4571),
    [sym_binary_operator] = STATE(4571),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4571),
    [sym_call] = STATE(4571),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4571),
    [sym_anonymous_function] = STATE(4571),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2643),
    [sym_integer] = ACTIONS(2643),
    [sym_float] = ACTIONS(2643),
    [sym_char] = ACTIONS(2643),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2643),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(900)] = {
    [sym__expression] = STATE(4573),
    [sym_block] = STATE(4573),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4573),
    [sym_nil] = STATE(4573),
    [sym__atom] = STATE(4573),
    [sym_quoted_atom] = STATE(4573),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4573),
    [sym_charlist] = STATE(4573),
    [sym_sigil] = STATE(4573),
    [sym_list] = STATE(4573),
    [sym_tuple] = STATE(4573),
    [sym_bitstring] = STATE(4573),
    [sym_map] = STATE(4573),
    [sym__nullary_operator] = STATE(4573),
    [sym_unary_operator] = STATE(4573),
    [sym_binary_operator] = STATE(4573),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4573),
    [sym_call] = STATE(4573),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4573),
    [sym_anonymous_function] = STATE(4573),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2645),
    [sym_integer] = ACTIONS(2645),
    [sym_float] = ACTIONS(2645),
    [sym_char] = ACTIONS(2645),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2645),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(901)] = {
    [sym__expression] = STATE(4576),
    [sym_block] = STATE(4576),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4576),
    [sym_nil] = STATE(4576),
    [sym__atom] = STATE(4576),
    [sym_quoted_atom] = STATE(4576),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4576),
    [sym_charlist] = STATE(4576),
    [sym_sigil] = STATE(4576),
    [sym_list] = STATE(4576),
    [sym_tuple] = STATE(4576),
    [sym_bitstring] = STATE(4576),
    [sym_map] = STATE(4576),
    [sym__nullary_operator] = STATE(4576),
    [sym_unary_operator] = STATE(4576),
    [sym_binary_operator] = STATE(4576),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4576),
    [sym_call] = STATE(4576),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4576),
    [sym_anonymous_function] = STATE(4576),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2647),
    [sym_integer] = ACTIONS(2647),
    [sym_float] = ACTIONS(2647),
    [sym_char] = ACTIONS(2647),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2647),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(902)] = {
    [sym__expression] = STATE(4578),
    [sym_block] = STATE(4578),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4578),
    [sym_nil] = STATE(4578),
    [sym__atom] = STATE(4578),
    [sym_quoted_atom] = STATE(4578),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4578),
    [sym_charlist] = STATE(4578),
    [sym_sigil] = STATE(4578),
    [sym_list] = STATE(4578),
    [sym_tuple] = STATE(4578),
    [sym_bitstring] = STATE(4578),
    [sym_map] = STATE(4578),
    [sym__nullary_operator] = STATE(4578),
    [sym_unary_operator] = STATE(4578),
    [sym_binary_operator] = STATE(4578),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4578),
    [sym_call] = STATE(4578),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4578),
    [sym_anonymous_function] = STATE(4578),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2649),
    [sym_integer] = ACTIONS(2649),
    [sym_float] = ACTIONS(2649),
    [sym_char] = ACTIONS(2649),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2649),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(903)] = {
    [sym__expression] = STATE(4582),
    [sym_block] = STATE(4582),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4582),
    [sym_nil] = STATE(4582),
    [sym__atom] = STATE(4582),
    [sym_quoted_atom] = STATE(4582),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4582),
    [sym_charlist] = STATE(4582),
    [sym_sigil] = STATE(4582),
    [sym_list] = STATE(4582),
    [sym_tuple] = STATE(4582),
    [sym_bitstring] = STATE(4582),
    [sym_map] = STATE(4582),
    [sym__nullary_operator] = STATE(4582),
    [sym_unary_operator] = STATE(4582),
    [sym_binary_operator] = STATE(4582),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4582),
    [sym_call] = STATE(4582),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4582),
    [sym_anonymous_function] = STATE(4582),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2651),
    [sym_integer] = ACTIONS(2651),
    [sym_float] = ACTIONS(2651),
    [sym_char] = ACTIONS(2651),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2651),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(904)] = {
    [sym__expression] = STATE(4584),
    [sym_block] = STATE(4584),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4584),
    [sym_nil] = STATE(4584),
    [sym__atom] = STATE(4584),
    [sym_quoted_atom] = STATE(4584),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4584),
    [sym_charlist] = STATE(4584),
    [sym_sigil] = STATE(4584),
    [sym_list] = STATE(4584),
    [sym_tuple] = STATE(4584),
    [sym_bitstring] = STATE(4584),
    [sym_map] = STATE(4584),
    [sym__nullary_operator] = STATE(4584),
    [sym_unary_operator] = STATE(4584),
    [sym_binary_operator] = STATE(4584),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4584),
    [sym_call] = STATE(4584),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4584),
    [sym_anonymous_function] = STATE(4584),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2653),
    [sym_integer] = ACTIONS(2653),
    [sym_float] = ACTIONS(2653),
    [sym_char] = ACTIONS(2653),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2653),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(905)] = {
    [sym__expression] = STATE(4586),
    [sym_block] = STATE(4586),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4586),
    [sym_nil] = STATE(4586),
    [sym__atom] = STATE(4586),
    [sym_quoted_atom] = STATE(4586),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4586),
    [sym_charlist] = STATE(4586),
    [sym_sigil] = STATE(4586),
    [sym_list] = STATE(4586),
    [sym_tuple] = STATE(4586),
    [sym_bitstring] = STATE(4586),
    [sym_map] = STATE(4586),
    [sym__nullary_operator] = STATE(4586),
    [sym_unary_operator] = STATE(4586),
    [sym_binary_operator] = STATE(4586),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4586),
    [sym_call] = STATE(4586),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4586),
    [sym_anonymous_function] = STATE(4586),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2655),
    [sym_integer] = ACTIONS(2655),
    [sym_float] = ACTIONS(2655),
    [sym_char] = ACTIONS(2655),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2655),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(906)] = {
    [sym__expression] = STATE(4590),
    [sym_block] = STATE(4590),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4590),
    [sym_nil] = STATE(4590),
    [sym__atom] = STATE(4590),
    [sym_quoted_atom] = STATE(4590),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4590),
    [sym_charlist] = STATE(4590),
    [sym_sigil] = STATE(4590),
    [sym_list] = STATE(4590),
    [sym_tuple] = STATE(4590),
    [sym_bitstring] = STATE(4590),
    [sym_map] = STATE(4590),
    [sym__nullary_operator] = STATE(4590),
    [sym_unary_operator] = STATE(4590),
    [sym_binary_operator] = STATE(4590),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4590),
    [sym_call] = STATE(4590),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4590),
    [sym_anonymous_function] = STATE(4590),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2657),
    [sym_integer] = ACTIONS(2657),
    [sym_float] = ACTIONS(2657),
    [sym_char] = ACTIONS(2657),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2657),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(907)] = {
    [sym__expression] = STATE(4591),
    [sym_block] = STATE(4591),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4591),
    [sym_nil] = STATE(4591),
    [sym__atom] = STATE(4591),
    [sym_quoted_atom] = STATE(4591),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4591),
    [sym_charlist] = STATE(4591),
    [sym_sigil] = STATE(4591),
    [sym_list] = STATE(4591),
    [sym_tuple] = STATE(4591),
    [sym_bitstring] = STATE(4591),
    [sym_map] = STATE(4591),
    [sym__nullary_operator] = STATE(4591),
    [sym_unary_operator] = STATE(4591),
    [sym_binary_operator] = STATE(4591),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4591),
    [sym_call] = STATE(4591),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4591),
    [sym_anonymous_function] = STATE(4591),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2659),
    [sym_integer] = ACTIONS(2659),
    [sym_float] = ACTIONS(2659),
    [sym_char] = ACTIONS(2659),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2659),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(908)] = {
    [sym__expression] = STATE(4592),
    [sym_block] = STATE(4592),
    [sym_identifier] = STATE(51),
    [sym_boolean] = STATE(4592),
    [sym_nil] = STATE(4592),
    [sym__atom] = STATE(4592),
    [sym_quoted_atom] = STATE(4592),
    [sym__quoted_i_double] = STATE(3053),
    [sym__quoted_i_single] = STATE(3055),
    [sym__quoted_i_heredoc_single] = STATE(2943),
    [sym__quoted_i_heredoc_double] = STATE(2946),
    [sym_string] = STATE(4592),
    [sym_charlist] = STATE(4592),
    [sym_sigil] = STATE(4592),
    [sym_list] = STATE(4592),
    [sym_tuple] = STATE(4592),
    [sym_bitstring] = STATE(4592),
    [sym_map] = STATE(4592),
    [sym__nullary_operator] = STATE(4592),
    [sym_unary_operator] = STATE(4592),
    [sym_binary_operator] = STATE(4592),
    [sym_operator_identifier] = STATE(6917),
    [sym_dot] = STATE(4592),
    [sym_call] = STATE(4592),
    [sym__call_without_parentheses] = STATE(2947),
    [sym__call_with_parentheses] = STATE(2948),
    [sym__local_call_without_parentheses] = STATE(2949),
    [sym__local_call_with_parentheses] = STATE(2204),
    [sym__local_call_just_do_block] = STATE(2950),
    [sym__remote_call_without_parentheses] = STATE(2951),
    [sym__remote_call_with_parentheses] = STATE(2206),
    [sym__remote_dot] = STATE(38),
    [sym__anonymous_call] = STATE(2209),
    [sym__anonymous_dot] = STATE(6786),
    [sym__double_call] = STATE(2952),
    [sym_access_call] = STATE(4592),
    [sym_anonymous_function] = STATE(4592),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(1093),
    [aux_sym_identifier_token1] = ACTIONS(720),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(720),
    [sym_alias] = ACTIONS(2661),
    [sym_integer] = ACTIONS(2661),
    [sym_float] = ACTIONS(2661),
    [sym_char] = ACTIONS(2661),
    [anon_sym_true] = ACTIONS(1097),
    [anon_sym_false] = ACTIONS(1097),
    [anon_sym_nil] = ACTIONS(1099),
    [sym_atom] = ACTIONS(2661),
    [anon_sym_DQUOTE] = ACTIONS(1101),
    [anon_sym_SQUOTE] = ACTIONS(1103),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1105),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1107),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LBRACK] = ACTIONS(1113),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1115),
    [anon_sym_LT_LT] = ACTIONS(1119),
    [anon_sym_PERCENT] = ACTIONS(1121),
    [anon_sym_DOT_DOT] = ACTIONS(1123),
    [anon_sym_AMP] = ACTIONS(1125),
    [anon_sym_PLUS] = ACTIONS(1127),
    [anon_sym_DASH] = ACTIONS(1127),
    [anon_sym_BANG] = ACTIONS(1127),
    [anon_sym_CARET] = ACTIONS(1127),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(1127),
    [anon_sym_not] = ACTIONS(1127),
    [anon_sym_AT] = ACTIONS(1129),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(1131),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(1133),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(1135),
  },
  [STATE(909)] = {
    [sym__expression] = STATE(3165),
    [sym_block] = STATE(3165),
    [sym_identifier] = STATE(52),
    [sym_boolean] = STATE(3165),
    [sym_nil] = STATE(3165),
    [sym__atom] = STATE(3165),
    [sym_quoted_atom] = STATE(3165),
    [sym__quoted_i_double] = STATE(1392),
    [sym__quoted_i_single] = STATE(1393),
    [sym__quoted_i_heredoc_single] = STATE(1394),
    [sym__quoted_i_heredoc_double] = STATE(1395),
    [sym_string] = STATE(3165),
    [sym_charlist] = STATE(3165),
    [sym_sigil] = STATE(3165),
    [sym_list] = STATE(3165),
    [sym_tuple] = STATE(3165),
    [sym_bitstring] = STATE(3165),
    [sym_map] = STATE(3165),
    [sym__nullary_operator] = STATE(3165),
    [sym_unary_operator] = STATE(3165),
    [sym_binary_operator] = STATE(3165),
    [sym_operator_identifier] = STATE(6974),
    [sym_dot] = STATE(3165),
    [sym_call] = STATE(3165),
    [sym__call_without_parentheses] = STATE(1396),
    [sym__call_with_parentheses] = STATE(1397),
    [sym__local_call_without_parentheses] = STATE(1398),
    [sym__local_call_with_parentheses] = STATE(1134),
    [sym__local_call_just_do_block] = STATE(1400),
    [sym__remote_call_without_parentheses] = STATE(1401),
    [sym__remote_call_with_parentheses] = STATE(1136),
    [sym__remote_dot] = STATE(48),
    [sym__anonymous_call] = STATE(1139),
    [sym__anonymous_dot] = STATE(6816),
    [sym__double_call] = STATE(1409),
    [sym_access_call] = STATE(3165),
    [sym_anonymous_function] = STATE(3165),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(189),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2663),
    [sym_integer] = ACTIONS(2663),
    [sym_float] = ACTIONS(2663),
    [sym_char] = ACTIONS(2663),
    [anon_sym_true] = ACTIONS(195),
    [anon_sym_false] = ACTIONS(195),
    [anon_sym_nil] = ACTIONS(197),
    [sym_atom] = ACTIONS(2663),
    [anon_sym_DQUOTE] = ACTIONS(199),
    [anon_sym_SQUOTE] = ACTIONS(201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(203),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(205),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LBRACK] = ACTIONS(209),
    [anon_sym_LT] = ACTIONS(35),
    [anon_sym_GT] = ACTIONS(35),
    [anon_sym_PIPE] = ACTIONS(35),
    [anon_sym_SLASH] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(512),
    [anon_sym_LT_LT] = ACTIONS(215),
    [anon_sym_PERCENT] = ACTIONS(217),
    [anon_sym_DOT_DOT] = ACTIONS(1445),
    [anon_sym_AMP] = ACTIONS(516),
    [anon_sym_PLUS] = ACTIONS(518),
    [anon_sym_DASH] = ACTIONS(518),
    [anon_sym_BANG] = ACTIONS(518),
    [anon_sym_CARET] = ACTIONS(518),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(518),
    [anon_sym_not] = ACTIONS(518),
    [anon_sym_AT] = ACTIONS(520),
    [anon_sym_LT_DASH] = ACTIONS(35),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(35),
    [anon_sym_when] = ACTIONS(35),
    [anon_sym_COLON_COLON] = ACTIONS(35),
    [anon_sym_EQ] = ACTIONS(35),
    [anon_sym_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(35),
    [anon_sym_or] = ACTIONS(35),
    [anon_sym_AMP_AMP] = ACTIONS(35),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(35),
    [anon_sym_and] = ACTIONS(35),
    [anon_sym_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ] = ACTIONS(35),
    [anon_sym_EQ_TILDE] = ACTIONS(35),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(35),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(35),
    [anon_sym_LT_EQ] = ACTIONS(35),
    [anon_sym_GT_EQ] = ACTIONS(35),
    [anon_sym_PIPE_GT] = ACTIONS(35),
    [anon_sym_LT_LT_LT] = ACTIONS(35),
    [anon_sym_GT_GT_GT] = ACTIONS(35),
    [anon_sym_LT_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE] = ACTIONS(35),
    [anon_sym_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_TILDE_GT] = ACTIONS(35),
    [anon_sym_LT_PIPE_GT] = ACTIONS(35),
    [anon_sym_in] = ACTIONS(35),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(35),
    [anon_sym_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH] = ACTIONS(35),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(35),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(35),
    [anon_sym_LT_GT] = ACTIONS(35),
    [anon_sym_STAR] = ACTIONS(35),
    [anon_sym_STAR_STAR] = ACTIONS(35),
    [anon_sym_DASH_GT] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(225),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(522),
    [sym__not_in] = ACTIONS(55),
    [sym__quoted_atom_start] = ACTIONS(231),
  },
  [STATE(910)] = {
    [aux_sym__terminator_token1] = ACTIONS(2665),
    [anon_sym_SEMI] = ACTIONS(2667),
    [anon_sym_LPAREN] = ACTIONS(2667),
    [aux_sym_identifier_token1] = ACTIONS(2667),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2667),
    [sym_alias] = ACTIONS(2667),
    [sym_integer] = ACTIONS(2667),
    [sym_float] = ACTIONS(2667),
    [sym_char] = ACTIONS(2667),
    [anon_sym_true] = ACTIONS(2667),
    [anon_sym_false] = ACTIONS(2667),
    [anon_sym_nil] = ACTIONS(2667),
    [sym_atom] = ACTIONS(2667),
    [anon_sym_DQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2667),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2667),
    [anon_sym_LBRACE] = ACTIONS(2667),
    [anon_sym_LBRACK] = ACTIONS(2667),
    [anon_sym_LT] = ACTIONS(2667),
    [anon_sym_GT] = ACTIONS(2667),
    [anon_sym_PIPE] = ACTIONS(2667),
    [anon_sym_SLASH] = ACTIONS(2667),
    [anon_sym_TILDE] = ACTIONS(2667),
    [anon_sym_COMMA] = ACTIONS(2667),
    [sym_keyword] = ACTIONS(2667),
    [anon_sym_LT_LT] = ACTIONS(2667),
    [anon_sym_PERCENT] = ACTIONS(2667),
    [anon_sym_DOT_DOT] = ACTIONS(2667),
    [anon_sym_AMP] = ACTIONS(2667),
    [anon_sym_PLUS] = ACTIONS(2667),
    [anon_sym_DASH] = ACTIONS(2667),
    [anon_sym_BANG] = ACTIONS(2667),
    [anon_sym_CARET] = ACTIONS(2667),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2667),
    [anon_sym_not] = ACTIONS(2667),
    [anon_sym_AT] = ACTIONS(2667),
    [anon_sym_LT_DASH] = ACTIONS(2667),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2667),
    [anon_sym_when] = ACTIONS(2667),
    [anon_sym_COLON_COLON] = ACTIONS(2667),
    [anon_sym_EQ_GT] = ACTIONS(2667),
    [anon_sym_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_or] = ACTIONS(2667),
    [anon_sym_AMP_AMP] = ACTIONS(2667),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2667),
    [anon_sym_and] = ACTIONS(2667),
    [anon_sym_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ] = ACTIONS(2667),
    [anon_sym_EQ_TILDE] = ACTIONS(2667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2667),
    [anon_sym_LT_EQ] = ACTIONS(2667),
    [anon_sym_GT_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_LT] = ACTIONS(2667),
    [anon_sym_GT_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2667),
    [anon_sym_in] = ACTIONS(2667),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2667),
    [anon_sym_SLASH_SLASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2667),
    [anon_sym_LT_GT] = ACTIONS(2667),
    [anon_sym_STAR] = ACTIONS(2667),
    [anon_sym_STAR_STAR] = ACTIONS(2667),
    [anon_sym_DASH_GT] = ACTIONS(2667),
    [anon_sym_DOT] = ACTIONS(2667),
    [anon_sym_after] = ACTIONS(2667),
    [anon_sym_catch] = ACTIONS(2667),
    [anon_sym_do] = ACTIONS(2667),
    [anon_sym_else] = ACTIONS(2667),
    [anon_sym_end] = ACTIONS(2667),
    [anon_sym_fn] = ACTIONS(2667),
    [anon_sym_rescue] = ACTIONS(2667),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_LBRACK2] = ACTIONS(2665),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2665),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2665),
    [sym__not_in] = ACTIONS(2665),
    [sym__quoted_atom_start] = ACTIONS(2665),
  },
  [STATE(911)] = {
    [aux_sym__terminator_token1] = ACTIONS(2669),
    [anon_sym_SEMI] = ACTIONS(2671),
    [anon_sym_LPAREN] = ACTIONS(2671),
    [aux_sym_identifier_token1] = ACTIONS(2671),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2671),
    [sym_alias] = ACTIONS(2671),
    [sym_integer] = ACTIONS(2671),
    [sym_float] = ACTIONS(2671),
    [sym_char] = ACTIONS(2671),
    [anon_sym_true] = ACTIONS(2671),
    [anon_sym_false] = ACTIONS(2671),
    [anon_sym_nil] = ACTIONS(2671),
    [sym_atom] = ACTIONS(2671),
    [anon_sym_DQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2671),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2671),
    [anon_sym_LBRACE] = ACTIONS(2671),
    [anon_sym_LBRACK] = ACTIONS(2671),
    [anon_sym_LT] = ACTIONS(2671),
    [anon_sym_GT] = ACTIONS(2671),
    [anon_sym_PIPE] = ACTIONS(2671),
    [anon_sym_SLASH] = ACTIONS(2671),
    [anon_sym_TILDE] = ACTIONS(2671),
    [anon_sym_COMMA] = ACTIONS(2671),
    [sym_keyword] = ACTIONS(2671),
    [anon_sym_LT_LT] = ACTIONS(2671),
    [anon_sym_PERCENT] = ACTIONS(2671),
    [anon_sym_DOT_DOT] = ACTIONS(2671),
    [anon_sym_AMP] = ACTIONS(2671),
    [anon_sym_PLUS] = ACTIONS(2671),
    [anon_sym_DASH] = ACTIONS(2671),
    [anon_sym_BANG] = ACTIONS(2671),
    [anon_sym_CARET] = ACTIONS(2671),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2671),
    [anon_sym_not] = ACTIONS(2671),
    [anon_sym_AT] = ACTIONS(2671),
    [anon_sym_LT_DASH] = ACTIONS(2671),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2671),
    [anon_sym_when] = ACTIONS(2671),
    [anon_sym_COLON_COLON] = ACTIONS(2671),
    [anon_sym_EQ_GT] = ACTIONS(2671),
    [anon_sym_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_or] = ACTIONS(2671),
    [anon_sym_AMP_AMP] = ACTIONS(2671),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2671),
    [anon_sym_and] = ACTIONS(2671),
    [anon_sym_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ] = ACTIONS(2671),
    [anon_sym_EQ_TILDE] = ACTIONS(2671),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2671),
    [anon_sym_LT_EQ] = ACTIONS(2671),
    [anon_sym_GT_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_LT] = ACTIONS(2671),
    [anon_sym_GT_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2671),
    [anon_sym_in] = ACTIONS(2671),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2671),
    [anon_sym_SLASH_SLASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2671),
    [anon_sym_LT_GT] = ACTIONS(2671),
    [anon_sym_STAR] = ACTIONS(2671),
    [anon_sym_STAR_STAR] = ACTIONS(2671),
    [anon_sym_DASH_GT] = ACTIONS(2671),
    [anon_sym_DOT] = ACTIONS(2671),
    [anon_sym_after] = ACTIONS(2671),
    [anon_sym_catch] = ACTIONS(2671),
    [anon_sym_do] = ACTIONS(2671),
    [anon_sym_else] = ACTIONS(2671),
    [anon_sym_end] = ACTIONS(2671),
    [anon_sym_fn] = ACTIONS(2671),
    [anon_sym_rescue] = ACTIONS(2671),
    [anon_sym_LPAREN2] = ACTIONS(2669),
    [anon_sym_LBRACK2] = ACTIONS(2669),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2669),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2669),
    [sym__not_in] = ACTIONS(2669),
    [sym__quoted_atom_start] = ACTIONS(2669),
  },
  [STATE(912)] = {
    [aux_sym__terminator_token1] = ACTIONS(2673),
    [anon_sym_SEMI] = ACTIONS(2675),
    [anon_sym_LPAREN] = ACTIONS(2675),
    [aux_sym_identifier_token1] = ACTIONS(2675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2675),
    [sym_alias] = ACTIONS(2675),
    [sym_integer] = ACTIONS(2675),
    [sym_float] = ACTIONS(2675),
    [sym_char] = ACTIONS(2675),
    [anon_sym_true] = ACTIONS(2675),
    [anon_sym_false] = ACTIONS(2675),
    [anon_sym_nil] = ACTIONS(2675),
    [sym_atom] = ACTIONS(2675),
    [anon_sym_DQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2675),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2675),
    [anon_sym_LBRACE] = ACTIONS(2675),
    [anon_sym_LBRACK] = ACTIONS(2675),
    [anon_sym_LT] = ACTIONS(2675),
    [anon_sym_GT] = ACTIONS(2675),
    [anon_sym_PIPE] = ACTIONS(2675),
    [anon_sym_SLASH] = ACTIONS(2675),
    [anon_sym_TILDE] = ACTIONS(2675),
    [anon_sym_COMMA] = ACTIONS(2675),
    [sym_keyword] = ACTIONS(2675),
    [anon_sym_LT_LT] = ACTIONS(2675),
    [anon_sym_PERCENT] = ACTIONS(2675),
    [anon_sym_DOT_DOT] = ACTIONS(2675),
    [anon_sym_AMP] = ACTIONS(2675),
    [anon_sym_PLUS] = ACTIONS(2675),
    [anon_sym_DASH] = ACTIONS(2675),
    [anon_sym_BANG] = ACTIONS(2675),
    [anon_sym_CARET] = ACTIONS(2675),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2675),
    [anon_sym_not] = ACTIONS(2675),
    [anon_sym_AT] = ACTIONS(2675),
    [anon_sym_LT_DASH] = ACTIONS(2675),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2675),
    [anon_sym_when] = ACTIONS(2675),
    [anon_sym_COLON_COLON] = ACTIONS(2675),
    [anon_sym_EQ_GT] = ACTIONS(2675),
    [anon_sym_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_or] = ACTIONS(2675),
    [anon_sym_AMP_AMP] = ACTIONS(2675),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2675),
    [anon_sym_and] = ACTIONS(2675),
    [anon_sym_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ] = ACTIONS(2675),
    [anon_sym_EQ_TILDE] = ACTIONS(2675),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2675),
    [anon_sym_LT_EQ] = ACTIONS(2675),
    [anon_sym_GT_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_LT] = ACTIONS(2675),
    [anon_sym_GT_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2675),
    [anon_sym_in] = ACTIONS(2675),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2675),
    [anon_sym_SLASH_SLASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2675),
    [anon_sym_LT_GT] = ACTIONS(2675),
    [anon_sym_STAR] = ACTIONS(2675),
    [anon_sym_STAR_STAR] = ACTIONS(2675),
    [anon_sym_DASH_GT] = ACTIONS(2675),
    [anon_sym_DOT] = ACTIONS(2675),
    [anon_sym_after] = ACTIONS(2675),
    [anon_sym_catch] = ACTIONS(2675),
    [anon_sym_do] = ACTIONS(2675),
    [anon_sym_else] = ACTIONS(2675),
    [anon_sym_end] = ACTIONS(2675),
    [anon_sym_fn] = ACTIONS(2675),
    [anon_sym_rescue] = ACTIONS(2675),
    [anon_sym_LPAREN2] = ACTIONS(2673),
    [anon_sym_LBRACK2] = ACTIONS(2673),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2673),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2673),
    [sym__not_in] = ACTIONS(2673),
    [sym__quoted_atom_start] = ACTIONS(2673),
  },
  [STATE(913)] = {
    [aux_sym__terminator_token1] = ACTIONS(2677),
    [anon_sym_SEMI] = ACTIONS(2679),
    [anon_sym_LPAREN] = ACTIONS(2679),
    [aux_sym_identifier_token1] = ACTIONS(2679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2679),
    [sym_alias] = ACTIONS(2679),
    [sym_integer] = ACTIONS(2679),
    [sym_float] = ACTIONS(2679),
    [sym_char] = ACTIONS(2679),
    [anon_sym_true] = ACTIONS(2679),
    [anon_sym_false] = ACTIONS(2679),
    [anon_sym_nil] = ACTIONS(2679),
    [sym_atom] = ACTIONS(2679),
    [anon_sym_DQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2679),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2679),
    [anon_sym_LBRACE] = ACTIONS(2679),
    [anon_sym_LBRACK] = ACTIONS(2679),
    [anon_sym_LT] = ACTIONS(2679),
    [anon_sym_GT] = ACTIONS(2679),
    [anon_sym_PIPE] = ACTIONS(2679),
    [anon_sym_SLASH] = ACTIONS(2679),
    [anon_sym_TILDE] = ACTIONS(2679),
    [anon_sym_COMMA] = ACTIONS(2679),
    [sym_keyword] = ACTIONS(2679),
    [anon_sym_LT_LT] = ACTIONS(2679),
    [anon_sym_PERCENT] = ACTIONS(2679),
    [anon_sym_DOT_DOT] = ACTIONS(2679),
    [anon_sym_AMP] = ACTIONS(2679),
    [anon_sym_PLUS] = ACTIONS(2679),
    [anon_sym_DASH] = ACTIONS(2679),
    [anon_sym_BANG] = ACTIONS(2679),
    [anon_sym_CARET] = ACTIONS(2679),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2679),
    [anon_sym_not] = ACTIONS(2679),
    [anon_sym_AT] = ACTIONS(2679),
    [anon_sym_LT_DASH] = ACTIONS(2679),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2679),
    [anon_sym_when] = ACTIONS(2679),
    [anon_sym_COLON_COLON] = ACTIONS(2679),
    [anon_sym_EQ_GT] = ACTIONS(2679),
    [anon_sym_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_or] = ACTIONS(2679),
    [anon_sym_AMP_AMP] = ACTIONS(2679),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2679),
    [anon_sym_and] = ACTIONS(2679),
    [anon_sym_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ] = ACTIONS(2679),
    [anon_sym_EQ_TILDE] = ACTIONS(2679),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2679),
    [anon_sym_LT_EQ] = ACTIONS(2679),
    [anon_sym_GT_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_LT] = ACTIONS(2679),
    [anon_sym_GT_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2679),
    [anon_sym_in] = ACTIONS(2679),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2679),
    [anon_sym_SLASH_SLASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2679),
    [anon_sym_LT_GT] = ACTIONS(2679),
    [anon_sym_STAR] = ACTIONS(2679),
    [anon_sym_STAR_STAR] = ACTIONS(2679),
    [anon_sym_DASH_GT] = ACTIONS(2679),
    [anon_sym_DOT] = ACTIONS(2679),
    [anon_sym_after] = ACTIONS(2679),
    [anon_sym_catch] = ACTIONS(2679),
    [anon_sym_do] = ACTIONS(2679),
    [anon_sym_else] = ACTIONS(2679),
    [anon_sym_end] = ACTIONS(2679),
    [anon_sym_fn] = ACTIONS(2679),
    [anon_sym_rescue] = ACTIONS(2679),
    [anon_sym_LPAREN2] = ACTIONS(2677),
    [anon_sym_LBRACK2] = ACTIONS(2677),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2677),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2677),
    [sym__not_in] = ACTIONS(2677),
    [sym__quoted_atom_start] = ACTIONS(2677),
  },
  [STATE(914)] = {
    [aux_sym__terminator_token1] = ACTIONS(2681),
    [anon_sym_SEMI] = ACTIONS(2683),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_after] = ACTIONS(2683),
    [anon_sym_catch] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_else] = ACTIONS(2683),
    [anon_sym_end] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_rescue] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2681),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(915)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_after] = ACTIONS(2687),
    [anon_sym_catch] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_else] = ACTIONS(2687),
    [anon_sym_end] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_rescue] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(916)] = {
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_after] = ACTIONS(2691),
    [anon_sym_catch] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_else] = ACTIONS(2691),
    [anon_sym_end] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_rescue] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(917)] = {
    [aux_sym__terminator_token1] = ACTIONS(2693),
    [anon_sym_SEMI] = ACTIONS(2695),
    [anon_sym_LPAREN] = ACTIONS(2695),
    [aux_sym_identifier_token1] = ACTIONS(2695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2695),
    [sym_alias] = ACTIONS(2695),
    [sym_integer] = ACTIONS(2695),
    [sym_float] = ACTIONS(2695),
    [sym_char] = ACTIONS(2695),
    [anon_sym_true] = ACTIONS(2695),
    [anon_sym_false] = ACTIONS(2695),
    [anon_sym_nil] = ACTIONS(2695),
    [sym_atom] = ACTIONS(2695),
    [anon_sym_DQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2695),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2695),
    [anon_sym_LBRACE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(2695),
    [anon_sym_LT] = ACTIONS(2695),
    [anon_sym_GT] = ACTIONS(2695),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_SLASH] = ACTIONS(2695),
    [anon_sym_TILDE] = ACTIONS(2695),
    [anon_sym_COMMA] = ACTIONS(2695),
    [sym_keyword] = ACTIONS(2695),
    [anon_sym_LT_LT] = ACTIONS(2695),
    [anon_sym_PERCENT] = ACTIONS(2695),
    [anon_sym_DOT_DOT] = ACTIONS(2695),
    [anon_sym_AMP] = ACTIONS(2695),
    [anon_sym_PLUS] = ACTIONS(2695),
    [anon_sym_DASH] = ACTIONS(2695),
    [anon_sym_BANG] = ACTIONS(2695),
    [anon_sym_CARET] = ACTIONS(2695),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2695),
    [anon_sym_not] = ACTIONS(2695),
    [anon_sym_AT] = ACTIONS(2695),
    [anon_sym_LT_DASH] = ACTIONS(2695),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2695),
    [anon_sym_when] = ACTIONS(2695),
    [anon_sym_COLON_COLON] = ACTIONS(2695),
    [anon_sym_EQ_GT] = ACTIONS(2695),
    [anon_sym_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_or] = ACTIONS(2695),
    [anon_sym_AMP_AMP] = ACTIONS(2695),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2695),
    [anon_sym_and] = ACTIONS(2695),
    [anon_sym_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ] = ACTIONS(2695),
    [anon_sym_EQ_TILDE] = ACTIONS(2695),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2695),
    [anon_sym_LT_EQ] = ACTIONS(2695),
    [anon_sym_GT_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_LT] = ACTIONS(2695),
    [anon_sym_GT_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2695),
    [anon_sym_in] = ACTIONS(2695),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2695),
    [anon_sym_SLASH_SLASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2695),
    [anon_sym_LT_GT] = ACTIONS(2695),
    [anon_sym_STAR] = ACTIONS(2695),
    [anon_sym_STAR_STAR] = ACTIONS(2695),
    [anon_sym_DASH_GT] = ACTIONS(2695),
    [anon_sym_DOT] = ACTIONS(2695),
    [anon_sym_after] = ACTIONS(2695),
    [anon_sym_catch] = ACTIONS(2695),
    [anon_sym_do] = ACTIONS(2695),
    [anon_sym_else] = ACTIONS(2695),
    [anon_sym_end] = ACTIONS(2695),
    [anon_sym_fn] = ACTIONS(2695),
    [anon_sym_rescue] = ACTIONS(2695),
    [anon_sym_LPAREN2] = ACTIONS(2693),
    [anon_sym_LBRACK2] = ACTIONS(2693),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2693),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2693),
    [sym__not_in] = ACTIONS(2693),
    [sym__quoted_atom_start] = ACTIONS(2693),
  },
  [STATE(918)] = {
    [aux_sym__terminator_token1] = ACTIONS(2697),
    [anon_sym_SEMI] = ACTIONS(2699),
    [anon_sym_LPAREN] = ACTIONS(2699),
    [aux_sym_identifier_token1] = ACTIONS(2699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2699),
    [sym_alias] = ACTIONS(2699),
    [sym_integer] = ACTIONS(2699),
    [sym_float] = ACTIONS(2699),
    [sym_char] = ACTIONS(2699),
    [anon_sym_true] = ACTIONS(2699),
    [anon_sym_false] = ACTIONS(2699),
    [anon_sym_nil] = ACTIONS(2699),
    [sym_atom] = ACTIONS(2699),
    [anon_sym_DQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2699),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2699),
    [anon_sym_LBRACE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(2699),
    [anon_sym_LT] = ACTIONS(2699),
    [anon_sym_GT] = ACTIONS(2699),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_SLASH] = ACTIONS(2699),
    [anon_sym_TILDE] = ACTIONS(2699),
    [anon_sym_COMMA] = ACTIONS(2699),
    [sym_keyword] = ACTIONS(2699),
    [anon_sym_LT_LT] = ACTIONS(2699),
    [anon_sym_PERCENT] = ACTIONS(2699),
    [anon_sym_DOT_DOT] = ACTIONS(2699),
    [anon_sym_AMP] = ACTIONS(2699),
    [anon_sym_PLUS] = ACTIONS(2699),
    [anon_sym_DASH] = ACTIONS(2699),
    [anon_sym_BANG] = ACTIONS(2699),
    [anon_sym_CARET] = ACTIONS(2699),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2699),
    [anon_sym_not] = ACTIONS(2699),
    [anon_sym_AT] = ACTIONS(2699),
    [anon_sym_LT_DASH] = ACTIONS(2699),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2699),
    [anon_sym_when] = ACTIONS(2699),
    [anon_sym_COLON_COLON] = ACTIONS(2699),
    [anon_sym_EQ_GT] = ACTIONS(2699),
    [anon_sym_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_or] = ACTIONS(2699),
    [anon_sym_AMP_AMP] = ACTIONS(2699),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2699),
    [anon_sym_and] = ACTIONS(2699),
    [anon_sym_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ] = ACTIONS(2699),
    [anon_sym_EQ_TILDE] = ACTIONS(2699),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2699),
    [anon_sym_LT_EQ] = ACTIONS(2699),
    [anon_sym_GT_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_LT] = ACTIONS(2699),
    [anon_sym_GT_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2699),
    [anon_sym_in] = ACTIONS(2699),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2699),
    [anon_sym_SLASH_SLASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2699),
    [anon_sym_LT_GT] = ACTIONS(2699),
    [anon_sym_STAR] = ACTIONS(2699),
    [anon_sym_STAR_STAR] = ACTIONS(2699),
    [anon_sym_DASH_GT] = ACTIONS(2699),
    [anon_sym_DOT] = ACTIONS(2699),
    [anon_sym_after] = ACTIONS(2699),
    [anon_sym_catch] = ACTIONS(2699),
    [anon_sym_do] = ACTIONS(2699),
    [anon_sym_else] = ACTIONS(2699),
    [anon_sym_end] = ACTIONS(2699),
    [anon_sym_fn] = ACTIONS(2699),
    [anon_sym_rescue] = ACTIONS(2699),
    [anon_sym_LPAREN2] = ACTIONS(2697),
    [anon_sym_LBRACK2] = ACTIONS(2697),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2697),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2697),
    [sym__not_in] = ACTIONS(2697),
    [sym__quoted_atom_start] = ACTIONS(2697),
  },
  [STATE(919)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_after] = ACTIONS(2687),
    [anon_sym_catch] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_else] = ACTIONS(2687),
    [anon_sym_end] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_rescue] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(920)] = {
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_after] = ACTIONS(2691),
    [anon_sym_catch] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_else] = ACTIONS(2691),
    [anon_sym_end] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_rescue] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(921)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_after] = ACTIONS(2687),
    [anon_sym_catch] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_else] = ACTIONS(2687),
    [anon_sym_end] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_rescue] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(922)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_after] = ACTIONS(2687),
    [anon_sym_catch] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_else] = ACTIONS(2687),
    [anon_sym_end] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_rescue] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(923)] = {
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_after] = ACTIONS(2691),
    [anon_sym_catch] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_else] = ACTIONS(2691),
    [anon_sym_end] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_rescue] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(924)] = {
    [aux_sym__terminator_token1] = ACTIONS(2701),
    [anon_sym_SEMI] = ACTIONS(2703),
    [anon_sym_LPAREN] = ACTIONS(2703),
    [aux_sym_identifier_token1] = ACTIONS(2703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2703),
    [sym_alias] = ACTIONS(2703),
    [sym_integer] = ACTIONS(2703),
    [sym_float] = ACTIONS(2703),
    [sym_char] = ACTIONS(2703),
    [anon_sym_true] = ACTIONS(2703),
    [anon_sym_false] = ACTIONS(2703),
    [anon_sym_nil] = ACTIONS(2703),
    [sym_atom] = ACTIONS(2703),
    [anon_sym_DQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2703),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2703),
    [anon_sym_LBRACE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(2703),
    [anon_sym_LT] = ACTIONS(2703),
    [anon_sym_GT] = ACTIONS(2703),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_SLASH] = ACTIONS(2703),
    [anon_sym_TILDE] = ACTIONS(2703),
    [anon_sym_COMMA] = ACTIONS(2703),
    [sym_keyword] = ACTIONS(2703),
    [anon_sym_LT_LT] = ACTIONS(2703),
    [anon_sym_PERCENT] = ACTIONS(2703),
    [anon_sym_DOT_DOT] = ACTIONS(2703),
    [anon_sym_AMP] = ACTIONS(2703),
    [anon_sym_PLUS] = ACTIONS(2703),
    [anon_sym_DASH] = ACTIONS(2703),
    [anon_sym_BANG] = ACTIONS(2703),
    [anon_sym_CARET] = ACTIONS(2703),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2703),
    [anon_sym_not] = ACTIONS(2703),
    [anon_sym_AT] = ACTIONS(2703),
    [anon_sym_LT_DASH] = ACTIONS(2703),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2703),
    [anon_sym_when] = ACTIONS(2703),
    [anon_sym_COLON_COLON] = ACTIONS(2703),
    [anon_sym_EQ_GT] = ACTIONS(2703),
    [anon_sym_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_or] = ACTIONS(2703),
    [anon_sym_AMP_AMP] = ACTIONS(2703),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2703),
    [anon_sym_and] = ACTIONS(2703),
    [anon_sym_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ] = ACTIONS(2703),
    [anon_sym_EQ_TILDE] = ACTIONS(2703),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2703),
    [anon_sym_LT_EQ] = ACTIONS(2703),
    [anon_sym_GT_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_LT] = ACTIONS(2703),
    [anon_sym_GT_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2703),
    [anon_sym_in] = ACTIONS(2703),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2703),
    [anon_sym_SLASH_SLASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2703),
    [anon_sym_LT_GT] = ACTIONS(2703),
    [anon_sym_STAR] = ACTIONS(2703),
    [anon_sym_STAR_STAR] = ACTIONS(2703),
    [anon_sym_DASH_GT] = ACTIONS(2703),
    [anon_sym_DOT] = ACTIONS(2703),
    [anon_sym_after] = ACTIONS(2703),
    [anon_sym_catch] = ACTIONS(2703),
    [anon_sym_do] = ACTIONS(2703),
    [anon_sym_else] = ACTIONS(2703),
    [anon_sym_end] = ACTIONS(2703),
    [anon_sym_fn] = ACTIONS(2703),
    [anon_sym_rescue] = ACTIONS(2703),
    [anon_sym_LPAREN2] = ACTIONS(2701),
    [anon_sym_LBRACK2] = ACTIONS(2701),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2701),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2701),
    [sym__not_in] = ACTIONS(2701),
    [sym__quoted_atom_start] = ACTIONS(2701),
  },
  [STATE(925)] = {
    [aux_sym__terminator_token1] = ACTIONS(2705),
    [anon_sym_SEMI] = ACTIONS(2707),
    [anon_sym_LPAREN] = ACTIONS(2707),
    [aux_sym_identifier_token1] = ACTIONS(2707),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2707),
    [sym_alias] = ACTIONS(2707),
    [sym_integer] = ACTIONS(2707),
    [sym_float] = ACTIONS(2707),
    [sym_char] = ACTIONS(2707),
    [anon_sym_true] = ACTIONS(2707),
    [anon_sym_false] = ACTIONS(2707),
    [anon_sym_nil] = ACTIONS(2707),
    [sym_atom] = ACTIONS(2707),
    [anon_sym_DQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2707),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2707),
    [anon_sym_LBRACE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(2707),
    [anon_sym_LT] = ACTIONS(2707),
    [anon_sym_GT] = ACTIONS(2707),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_SLASH] = ACTIONS(2707),
    [anon_sym_TILDE] = ACTIONS(2707),
    [anon_sym_COMMA] = ACTIONS(2707),
    [sym_keyword] = ACTIONS(2707),
    [anon_sym_LT_LT] = ACTIONS(2707),
    [anon_sym_PERCENT] = ACTIONS(2707),
    [anon_sym_DOT_DOT] = ACTIONS(2707),
    [anon_sym_AMP] = ACTIONS(2707),
    [anon_sym_PLUS] = ACTIONS(2707),
    [anon_sym_DASH] = ACTIONS(2707),
    [anon_sym_BANG] = ACTIONS(2707),
    [anon_sym_CARET] = ACTIONS(2707),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2707),
    [anon_sym_not] = ACTIONS(2707),
    [anon_sym_AT] = ACTIONS(2707),
    [anon_sym_LT_DASH] = ACTIONS(2707),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2707),
    [anon_sym_when] = ACTIONS(2707),
    [anon_sym_COLON_COLON] = ACTIONS(2707),
    [anon_sym_EQ_GT] = ACTIONS(2707),
    [anon_sym_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_or] = ACTIONS(2707),
    [anon_sym_AMP_AMP] = ACTIONS(2707),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2707),
    [anon_sym_and] = ACTIONS(2707),
    [anon_sym_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ] = ACTIONS(2707),
    [anon_sym_EQ_TILDE] = ACTIONS(2707),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2707),
    [anon_sym_LT_EQ] = ACTIONS(2707),
    [anon_sym_GT_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_LT] = ACTIONS(2707),
    [anon_sym_GT_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2707),
    [anon_sym_in] = ACTIONS(2707),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2707),
    [anon_sym_SLASH_SLASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2707),
    [anon_sym_LT_GT] = ACTIONS(2707),
    [anon_sym_STAR] = ACTIONS(2707),
    [anon_sym_STAR_STAR] = ACTIONS(2707),
    [anon_sym_DASH_GT] = ACTIONS(2707),
    [anon_sym_DOT] = ACTIONS(2707),
    [anon_sym_after] = ACTIONS(2707),
    [anon_sym_catch] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2707),
    [anon_sym_else] = ACTIONS(2707),
    [anon_sym_end] = ACTIONS(2707),
    [anon_sym_fn] = ACTIONS(2707),
    [anon_sym_rescue] = ACTIONS(2707),
    [anon_sym_LPAREN2] = ACTIONS(2705),
    [anon_sym_LBRACK2] = ACTIONS(2705),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2705),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2705),
    [sym__not_in] = ACTIONS(2705),
    [sym__quoted_atom_start] = ACTIONS(2705),
  },
  [STATE(926)] = {
    [aux_sym__terminator_token1] = ACTIONS(2709),
    [anon_sym_SEMI] = ACTIONS(2711),
    [anon_sym_LPAREN] = ACTIONS(2711),
    [aux_sym_identifier_token1] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [sym_alias] = ACTIONS(2711),
    [sym_integer] = ACTIONS(2711),
    [sym_float] = ACTIONS(2711),
    [sym_char] = ACTIONS(2711),
    [anon_sym_true] = ACTIONS(2711),
    [anon_sym_false] = ACTIONS(2711),
    [anon_sym_nil] = ACTIONS(2711),
    [sym_atom] = ACTIONS(2711),
    [anon_sym_DQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2711),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2711),
    [anon_sym_LBRACE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2711),
    [anon_sym_GT] = ACTIONS(2711),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_SLASH] = ACTIONS(2711),
    [anon_sym_TILDE] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [sym_keyword] = ACTIONS(2711),
    [anon_sym_LT_LT] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2711),
    [anon_sym_AMP] = ACTIONS(2711),
    [anon_sym_PLUS] = ACTIONS(2711),
    [anon_sym_DASH] = ACTIONS(2711),
    [anon_sym_BANG] = ACTIONS(2711),
    [anon_sym_CARET] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_not] = ACTIONS(2711),
    [anon_sym_AT] = ACTIONS(2711),
    [anon_sym_LT_DASH] = ACTIONS(2711),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2711),
    [anon_sym_when] = ACTIONS(2711),
    [anon_sym_COLON_COLON] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2711),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2711),
    [anon_sym_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_LT] = ACTIONS(2711),
    [anon_sym_GT_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2711),
    [anon_sym_in] = ACTIONS(2711),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2711),
    [anon_sym_LT_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2711),
    [anon_sym_STAR_STAR] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_DOT] = ACTIONS(2711),
    [anon_sym_after] = ACTIONS(2711),
    [anon_sym_catch] = ACTIONS(2711),
    [anon_sym_do] = ACTIONS(2711),
    [anon_sym_else] = ACTIONS(2711),
    [anon_sym_end] = ACTIONS(2711),
    [anon_sym_fn] = ACTIONS(2711),
    [anon_sym_rescue] = ACTIONS(2711),
    [anon_sym_LPAREN2] = ACTIONS(2709),
    [anon_sym_LBRACK2] = ACTIONS(2709),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2709),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2709),
    [sym__not_in] = ACTIONS(2709),
    [sym__quoted_atom_start] = ACTIONS(2709),
  },
  [STATE(927)] = {
    [aux_sym__terminator_token1] = ACTIONS(2681),
    [anon_sym_SEMI] = ACTIONS(2683),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_after] = ACTIONS(2683),
    [anon_sym_catch] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_else] = ACTIONS(2683),
    [anon_sym_end] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_rescue] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(928)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [anon_sym_RPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_RBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_RBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(929)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2699),
    [anon_sym_RPAREN] = ACTIONS(2699),
    [aux_sym_identifier_token1] = ACTIONS(2699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2699),
    [sym_alias] = ACTIONS(2699),
    [sym_integer] = ACTIONS(2699),
    [sym_float] = ACTIONS(2699),
    [sym_char] = ACTIONS(2699),
    [anon_sym_true] = ACTIONS(2699),
    [anon_sym_false] = ACTIONS(2699),
    [anon_sym_nil] = ACTIONS(2699),
    [sym_atom] = ACTIONS(2699),
    [anon_sym_DQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2699),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2699),
    [anon_sym_LBRACE] = ACTIONS(2699),
    [anon_sym_RBRACE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(2699),
    [anon_sym_RBRACK] = ACTIONS(2699),
    [anon_sym_LT] = ACTIONS(2699),
    [anon_sym_GT] = ACTIONS(2699),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_SLASH] = ACTIONS(2699),
    [anon_sym_TILDE] = ACTIONS(2699),
    [anon_sym_COMMA] = ACTIONS(2699),
    [sym_keyword] = ACTIONS(2699),
    [anon_sym_LT_LT] = ACTIONS(2699),
    [anon_sym_PERCENT] = ACTIONS(2699),
    [anon_sym_DOT_DOT] = ACTIONS(2699),
    [anon_sym_AMP] = ACTIONS(2699),
    [anon_sym_PLUS] = ACTIONS(2699),
    [anon_sym_DASH] = ACTIONS(2699),
    [anon_sym_BANG] = ACTIONS(2699),
    [anon_sym_CARET] = ACTIONS(2699),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2699),
    [anon_sym_not] = ACTIONS(2699),
    [anon_sym_AT] = ACTIONS(2699),
    [anon_sym_LT_DASH] = ACTIONS(2699),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2699),
    [anon_sym_when] = ACTIONS(2699),
    [anon_sym_COLON_COLON] = ACTIONS(2699),
    [anon_sym_EQ_GT] = ACTIONS(2699),
    [anon_sym_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_or] = ACTIONS(2699),
    [anon_sym_AMP_AMP] = ACTIONS(2699),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2699),
    [anon_sym_and] = ACTIONS(2699),
    [anon_sym_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ] = ACTIONS(2699),
    [anon_sym_EQ_TILDE] = ACTIONS(2699),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2699),
    [anon_sym_LT_EQ] = ACTIONS(2699),
    [anon_sym_GT_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_LT] = ACTIONS(2699),
    [anon_sym_GT_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2699),
    [anon_sym_in] = ACTIONS(2699),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2699),
    [anon_sym_SLASH_SLASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2699),
    [anon_sym_LT_GT] = ACTIONS(2699),
    [anon_sym_STAR] = ACTIONS(2699),
    [anon_sym_STAR_STAR] = ACTIONS(2699),
    [anon_sym_DASH_GT] = ACTIONS(2699),
    [anon_sym_DOT] = ACTIONS(2699),
    [anon_sym_do] = ACTIONS(2699),
    [anon_sym_fn] = ACTIONS(2699),
    [anon_sym_LPAREN2] = ACTIONS(2697),
    [anon_sym_LBRACK2] = ACTIONS(2697),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2697),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2697),
    [sym__not_in] = ACTIONS(2697),
    [sym__quoted_atom_start] = ACTIONS(2697),
  },
  [STATE(930)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [anon_sym_RPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_RBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_RBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(931)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [anon_sym_RPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_RBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_RBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2681),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(932)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2703),
    [anon_sym_RPAREN] = ACTIONS(2703),
    [aux_sym_identifier_token1] = ACTIONS(2703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2703),
    [sym_alias] = ACTIONS(2703),
    [sym_integer] = ACTIONS(2703),
    [sym_float] = ACTIONS(2703),
    [sym_char] = ACTIONS(2703),
    [anon_sym_true] = ACTIONS(2703),
    [anon_sym_false] = ACTIONS(2703),
    [anon_sym_nil] = ACTIONS(2703),
    [sym_atom] = ACTIONS(2703),
    [anon_sym_DQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2703),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2703),
    [anon_sym_LBRACE] = ACTIONS(2703),
    [anon_sym_RBRACE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(2703),
    [anon_sym_RBRACK] = ACTIONS(2703),
    [anon_sym_LT] = ACTIONS(2703),
    [anon_sym_GT] = ACTIONS(2703),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_SLASH] = ACTIONS(2703),
    [anon_sym_TILDE] = ACTIONS(2703),
    [anon_sym_COMMA] = ACTIONS(2703),
    [sym_keyword] = ACTIONS(2703),
    [anon_sym_LT_LT] = ACTIONS(2703),
    [anon_sym_PERCENT] = ACTIONS(2703),
    [anon_sym_DOT_DOT] = ACTIONS(2703),
    [anon_sym_AMP] = ACTIONS(2703),
    [anon_sym_PLUS] = ACTIONS(2703),
    [anon_sym_DASH] = ACTIONS(2703),
    [anon_sym_BANG] = ACTIONS(2703),
    [anon_sym_CARET] = ACTIONS(2703),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2703),
    [anon_sym_not] = ACTIONS(2703),
    [anon_sym_AT] = ACTIONS(2703),
    [anon_sym_LT_DASH] = ACTIONS(2703),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2703),
    [anon_sym_when] = ACTIONS(2703),
    [anon_sym_COLON_COLON] = ACTIONS(2703),
    [anon_sym_EQ_GT] = ACTIONS(2703),
    [anon_sym_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_or] = ACTIONS(2703),
    [anon_sym_AMP_AMP] = ACTIONS(2703),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2703),
    [anon_sym_and] = ACTIONS(2703),
    [anon_sym_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ] = ACTIONS(2703),
    [anon_sym_EQ_TILDE] = ACTIONS(2703),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2703),
    [anon_sym_LT_EQ] = ACTIONS(2703),
    [anon_sym_GT_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_LT] = ACTIONS(2703),
    [anon_sym_GT_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2703),
    [anon_sym_in] = ACTIONS(2703),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2703),
    [anon_sym_SLASH_SLASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2703),
    [anon_sym_LT_GT] = ACTIONS(2703),
    [anon_sym_STAR] = ACTIONS(2703),
    [anon_sym_STAR_STAR] = ACTIONS(2703),
    [anon_sym_DASH_GT] = ACTIONS(2703),
    [anon_sym_DOT] = ACTIONS(2703),
    [anon_sym_do] = ACTIONS(2703),
    [anon_sym_fn] = ACTIONS(2703),
    [anon_sym_LPAREN2] = ACTIONS(2701),
    [anon_sym_LBRACK2] = ACTIONS(2701),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2701),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2701),
    [sym__not_in] = ACTIONS(2701),
    [sym__quoted_atom_start] = ACTIONS(2701),
  },
  [STATE(933)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2707),
    [anon_sym_RPAREN] = ACTIONS(2707),
    [aux_sym_identifier_token1] = ACTIONS(2707),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2707),
    [sym_alias] = ACTIONS(2707),
    [sym_integer] = ACTIONS(2707),
    [sym_float] = ACTIONS(2707),
    [sym_char] = ACTIONS(2707),
    [anon_sym_true] = ACTIONS(2707),
    [anon_sym_false] = ACTIONS(2707),
    [anon_sym_nil] = ACTIONS(2707),
    [sym_atom] = ACTIONS(2707),
    [anon_sym_DQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2707),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2707),
    [anon_sym_LBRACE] = ACTIONS(2707),
    [anon_sym_RBRACE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(2707),
    [anon_sym_RBRACK] = ACTIONS(2707),
    [anon_sym_LT] = ACTIONS(2707),
    [anon_sym_GT] = ACTIONS(2707),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_SLASH] = ACTIONS(2707),
    [anon_sym_TILDE] = ACTIONS(2707),
    [anon_sym_COMMA] = ACTIONS(2707),
    [sym_keyword] = ACTIONS(2707),
    [anon_sym_LT_LT] = ACTIONS(2707),
    [anon_sym_PERCENT] = ACTIONS(2707),
    [anon_sym_DOT_DOT] = ACTIONS(2707),
    [anon_sym_AMP] = ACTIONS(2707),
    [anon_sym_PLUS] = ACTIONS(2707),
    [anon_sym_DASH] = ACTIONS(2707),
    [anon_sym_BANG] = ACTIONS(2707),
    [anon_sym_CARET] = ACTIONS(2707),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2707),
    [anon_sym_not] = ACTIONS(2707),
    [anon_sym_AT] = ACTIONS(2707),
    [anon_sym_LT_DASH] = ACTIONS(2707),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2707),
    [anon_sym_when] = ACTIONS(2707),
    [anon_sym_COLON_COLON] = ACTIONS(2707),
    [anon_sym_EQ_GT] = ACTIONS(2707),
    [anon_sym_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_or] = ACTIONS(2707),
    [anon_sym_AMP_AMP] = ACTIONS(2707),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2707),
    [anon_sym_and] = ACTIONS(2707),
    [anon_sym_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ] = ACTIONS(2707),
    [anon_sym_EQ_TILDE] = ACTIONS(2707),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2707),
    [anon_sym_LT_EQ] = ACTIONS(2707),
    [anon_sym_GT_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_LT] = ACTIONS(2707),
    [anon_sym_GT_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2707),
    [anon_sym_in] = ACTIONS(2707),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2707),
    [anon_sym_SLASH_SLASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2707),
    [anon_sym_LT_GT] = ACTIONS(2707),
    [anon_sym_STAR] = ACTIONS(2707),
    [anon_sym_STAR_STAR] = ACTIONS(2707),
    [anon_sym_DASH_GT] = ACTIONS(2707),
    [anon_sym_DOT] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2707),
    [anon_sym_fn] = ACTIONS(2707),
    [anon_sym_LPAREN2] = ACTIONS(2705),
    [anon_sym_LBRACK2] = ACTIONS(2705),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2705),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2705),
    [sym__not_in] = ACTIONS(2705),
    [sym__quoted_atom_start] = ACTIONS(2705),
  },
  [STATE(934)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [aux_sym_identifier_token1] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [sym_alias] = ACTIONS(2711),
    [sym_integer] = ACTIONS(2711),
    [sym_float] = ACTIONS(2711),
    [sym_char] = ACTIONS(2711),
    [anon_sym_true] = ACTIONS(2711),
    [anon_sym_false] = ACTIONS(2711),
    [anon_sym_nil] = ACTIONS(2711),
    [sym_atom] = ACTIONS(2711),
    [anon_sym_DQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2711),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2711),
    [anon_sym_LBRACE] = ACTIONS(2711),
    [anon_sym_RBRACE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(2711),
    [anon_sym_RBRACK] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2711),
    [anon_sym_GT] = ACTIONS(2711),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_SLASH] = ACTIONS(2711),
    [anon_sym_TILDE] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [sym_keyword] = ACTIONS(2711),
    [anon_sym_LT_LT] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2711),
    [anon_sym_AMP] = ACTIONS(2711),
    [anon_sym_PLUS] = ACTIONS(2711),
    [anon_sym_DASH] = ACTIONS(2711),
    [anon_sym_BANG] = ACTIONS(2711),
    [anon_sym_CARET] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_not] = ACTIONS(2711),
    [anon_sym_AT] = ACTIONS(2711),
    [anon_sym_LT_DASH] = ACTIONS(2711),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2711),
    [anon_sym_when] = ACTIONS(2711),
    [anon_sym_COLON_COLON] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2711),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2711),
    [anon_sym_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_LT] = ACTIONS(2711),
    [anon_sym_GT_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2711),
    [anon_sym_in] = ACTIONS(2711),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2711),
    [anon_sym_LT_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2711),
    [anon_sym_STAR_STAR] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_DOT] = ACTIONS(2711),
    [anon_sym_do] = ACTIONS(2711),
    [anon_sym_fn] = ACTIONS(2711),
    [anon_sym_LPAREN2] = ACTIONS(2709),
    [anon_sym_LBRACK2] = ACTIONS(2709),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2709),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2709),
    [sym__not_in] = ACTIONS(2709),
    [sym__quoted_atom_start] = ACTIONS(2709),
  },
  [STATE(935)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2667),
    [anon_sym_RPAREN] = ACTIONS(2667),
    [aux_sym_identifier_token1] = ACTIONS(2667),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2667),
    [sym_alias] = ACTIONS(2667),
    [sym_integer] = ACTIONS(2667),
    [sym_float] = ACTIONS(2667),
    [sym_char] = ACTIONS(2667),
    [anon_sym_true] = ACTIONS(2667),
    [anon_sym_false] = ACTIONS(2667),
    [anon_sym_nil] = ACTIONS(2667),
    [sym_atom] = ACTIONS(2667),
    [anon_sym_DQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2667),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2667),
    [anon_sym_LBRACE] = ACTIONS(2667),
    [anon_sym_RBRACE] = ACTIONS(2667),
    [anon_sym_LBRACK] = ACTIONS(2667),
    [anon_sym_RBRACK] = ACTIONS(2667),
    [anon_sym_LT] = ACTIONS(2667),
    [anon_sym_GT] = ACTIONS(2667),
    [anon_sym_PIPE] = ACTIONS(2667),
    [anon_sym_SLASH] = ACTIONS(2667),
    [anon_sym_TILDE] = ACTIONS(2667),
    [anon_sym_COMMA] = ACTIONS(2667),
    [sym_keyword] = ACTIONS(2667),
    [anon_sym_LT_LT] = ACTIONS(2667),
    [anon_sym_PERCENT] = ACTIONS(2667),
    [anon_sym_DOT_DOT] = ACTIONS(2667),
    [anon_sym_AMP] = ACTIONS(2667),
    [anon_sym_PLUS] = ACTIONS(2667),
    [anon_sym_DASH] = ACTIONS(2667),
    [anon_sym_BANG] = ACTIONS(2667),
    [anon_sym_CARET] = ACTIONS(2667),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2667),
    [anon_sym_not] = ACTIONS(2667),
    [anon_sym_AT] = ACTIONS(2667),
    [anon_sym_LT_DASH] = ACTIONS(2667),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2667),
    [anon_sym_when] = ACTIONS(2667),
    [anon_sym_COLON_COLON] = ACTIONS(2667),
    [anon_sym_EQ_GT] = ACTIONS(2667),
    [anon_sym_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_or] = ACTIONS(2667),
    [anon_sym_AMP_AMP] = ACTIONS(2667),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2667),
    [anon_sym_and] = ACTIONS(2667),
    [anon_sym_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ] = ACTIONS(2667),
    [anon_sym_EQ_TILDE] = ACTIONS(2667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2667),
    [anon_sym_LT_EQ] = ACTIONS(2667),
    [anon_sym_GT_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_LT] = ACTIONS(2667),
    [anon_sym_GT_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2667),
    [anon_sym_in] = ACTIONS(2667),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2667),
    [anon_sym_SLASH_SLASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2667),
    [anon_sym_LT_GT] = ACTIONS(2667),
    [anon_sym_STAR] = ACTIONS(2667),
    [anon_sym_STAR_STAR] = ACTIONS(2667),
    [anon_sym_DASH_GT] = ACTIONS(2667),
    [anon_sym_DOT] = ACTIONS(2667),
    [anon_sym_do] = ACTIONS(2667),
    [anon_sym_fn] = ACTIONS(2667),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_LBRACK2] = ACTIONS(2665),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2665),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2665),
    [sym__not_in] = ACTIONS(2665),
    [sym__quoted_atom_start] = ACTIONS(2665),
  },
  [STATE(936)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2671),
    [anon_sym_RPAREN] = ACTIONS(2671),
    [aux_sym_identifier_token1] = ACTIONS(2671),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2671),
    [sym_alias] = ACTIONS(2671),
    [sym_integer] = ACTIONS(2671),
    [sym_float] = ACTIONS(2671),
    [sym_char] = ACTIONS(2671),
    [anon_sym_true] = ACTIONS(2671),
    [anon_sym_false] = ACTIONS(2671),
    [anon_sym_nil] = ACTIONS(2671),
    [sym_atom] = ACTIONS(2671),
    [anon_sym_DQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2671),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2671),
    [anon_sym_LBRACE] = ACTIONS(2671),
    [anon_sym_RBRACE] = ACTIONS(2671),
    [anon_sym_LBRACK] = ACTIONS(2671),
    [anon_sym_RBRACK] = ACTIONS(2671),
    [anon_sym_LT] = ACTIONS(2671),
    [anon_sym_GT] = ACTIONS(2671),
    [anon_sym_PIPE] = ACTIONS(2671),
    [anon_sym_SLASH] = ACTIONS(2671),
    [anon_sym_TILDE] = ACTIONS(2671),
    [anon_sym_COMMA] = ACTIONS(2671),
    [sym_keyword] = ACTIONS(2671),
    [anon_sym_LT_LT] = ACTIONS(2671),
    [anon_sym_PERCENT] = ACTIONS(2671),
    [anon_sym_DOT_DOT] = ACTIONS(2671),
    [anon_sym_AMP] = ACTIONS(2671),
    [anon_sym_PLUS] = ACTIONS(2671),
    [anon_sym_DASH] = ACTIONS(2671),
    [anon_sym_BANG] = ACTIONS(2671),
    [anon_sym_CARET] = ACTIONS(2671),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2671),
    [anon_sym_not] = ACTIONS(2671),
    [anon_sym_AT] = ACTIONS(2671),
    [anon_sym_LT_DASH] = ACTIONS(2671),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2671),
    [anon_sym_when] = ACTIONS(2671),
    [anon_sym_COLON_COLON] = ACTIONS(2671),
    [anon_sym_EQ_GT] = ACTIONS(2671),
    [anon_sym_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_or] = ACTIONS(2671),
    [anon_sym_AMP_AMP] = ACTIONS(2671),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2671),
    [anon_sym_and] = ACTIONS(2671),
    [anon_sym_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ] = ACTIONS(2671),
    [anon_sym_EQ_TILDE] = ACTIONS(2671),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2671),
    [anon_sym_LT_EQ] = ACTIONS(2671),
    [anon_sym_GT_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_LT] = ACTIONS(2671),
    [anon_sym_GT_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2671),
    [anon_sym_in] = ACTIONS(2671),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2671),
    [anon_sym_SLASH_SLASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2671),
    [anon_sym_LT_GT] = ACTIONS(2671),
    [anon_sym_STAR] = ACTIONS(2671),
    [anon_sym_STAR_STAR] = ACTIONS(2671),
    [anon_sym_DASH_GT] = ACTIONS(2671),
    [anon_sym_DOT] = ACTIONS(2671),
    [anon_sym_do] = ACTIONS(2671),
    [anon_sym_fn] = ACTIONS(2671),
    [anon_sym_LPAREN2] = ACTIONS(2669),
    [anon_sym_LBRACK2] = ACTIONS(2669),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2669),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2669),
    [sym__not_in] = ACTIONS(2669),
    [sym__quoted_atom_start] = ACTIONS(2669),
  },
  [STATE(937)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2675),
    [anon_sym_RPAREN] = ACTIONS(2675),
    [aux_sym_identifier_token1] = ACTIONS(2675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2675),
    [sym_alias] = ACTIONS(2675),
    [sym_integer] = ACTIONS(2675),
    [sym_float] = ACTIONS(2675),
    [sym_char] = ACTIONS(2675),
    [anon_sym_true] = ACTIONS(2675),
    [anon_sym_false] = ACTIONS(2675),
    [anon_sym_nil] = ACTIONS(2675),
    [sym_atom] = ACTIONS(2675),
    [anon_sym_DQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2675),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2675),
    [anon_sym_LBRACE] = ACTIONS(2675),
    [anon_sym_RBRACE] = ACTIONS(2675),
    [anon_sym_LBRACK] = ACTIONS(2675),
    [anon_sym_RBRACK] = ACTIONS(2675),
    [anon_sym_LT] = ACTIONS(2675),
    [anon_sym_GT] = ACTIONS(2675),
    [anon_sym_PIPE] = ACTIONS(2675),
    [anon_sym_SLASH] = ACTIONS(2675),
    [anon_sym_TILDE] = ACTIONS(2675),
    [anon_sym_COMMA] = ACTIONS(2675),
    [sym_keyword] = ACTIONS(2675),
    [anon_sym_LT_LT] = ACTIONS(2675),
    [anon_sym_PERCENT] = ACTIONS(2675),
    [anon_sym_DOT_DOT] = ACTIONS(2675),
    [anon_sym_AMP] = ACTIONS(2675),
    [anon_sym_PLUS] = ACTIONS(2675),
    [anon_sym_DASH] = ACTIONS(2675),
    [anon_sym_BANG] = ACTIONS(2675),
    [anon_sym_CARET] = ACTIONS(2675),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2675),
    [anon_sym_not] = ACTIONS(2675),
    [anon_sym_AT] = ACTIONS(2675),
    [anon_sym_LT_DASH] = ACTIONS(2675),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2675),
    [anon_sym_when] = ACTIONS(2675),
    [anon_sym_COLON_COLON] = ACTIONS(2675),
    [anon_sym_EQ_GT] = ACTIONS(2675),
    [anon_sym_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_or] = ACTIONS(2675),
    [anon_sym_AMP_AMP] = ACTIONS(2675),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2675),
    [anon_sym_and] = ACTIONS(2675),
    [anon_sym_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ] = ACTIONS(2675),
    [anon_sym_EQ_TILDE] = ACTIONS(2675),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2675),
    [anon_sym_LT_EQ] = ACTIONS(2675),
    [anon_sym_GT_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_LT] = ACTIONS(2675),
    [anon_sym_GT_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2675),
    [anon_sym_in] = ACTIONS(2675),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2675),
    [anon_sym_SLASH_SLASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2675),
    [anon_sym_LT_GT] = ACTIONS(2675),
    [anon_sym_STAR] = ACTIONS(2675),
    [anon_sym_STAR_STAR] = ACTIONS(2675),
    [anon_sym_DASH_GT] = ACTIONS(2675),
    [anon_sym_DOT] = ACTIONS(2675),
    [anon_sym_do] = ACTIONS(2675),
    [anon_sym_fn] = ACTIONS(2675),
    [anon_sym_LPAREN2] = ACTIONS(2673),
    [anon_sym_LBRACK2] = ACTIONS(2673),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2673),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2673),
    [sym__not_in] = ACTIONS(2673),
    [sym__quoted_atom_start] = ACTIONS(2673),
  },
  [STATE(938)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [anon_sym_RPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_RBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_RBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(939)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [anon_sym_RPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_RBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_RBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(940)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [anon_sym_RPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_RBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_RBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(941)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [anon_sym_RPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_RBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_RBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(942)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [anon_sym_RPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_RBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_RBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(943)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2695),
    [anon_sym_RPAREN] = ACTIONS(2695),
    [aux_sym_identifier_token1] = ACTIONS(2695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2695),
    [sym_alias] = ACTIONS(2695),
    [sym_integer] = ACTIONS(2695),
    [sym_float] = ACTIONS(2695),
    [sym_char] = ACTIONS(2695),
    [anon_sym_true] = ACTIONS(2695),
    [anon_sym_false] = ACTIONS(2695),
    [anon_sym_nil] = ACTIONS(2695),
    [sym_atom] = ACTIONS(2695),
    [anon_sym_DQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2695),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2695),
    [anon_sym_LBRACE] = ACTIONS(2695),
    [anon_sym_RBRACE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(2695),
    [anon_sym_RBRACK] = ACTIONS(2695),
    [anon_sym_LT] = ACTIONS(2695),
    [anon_sym_GT] = ACTIONS(2695),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_SLASH] = ACTIONS(2695),
    [anon_sym_TILDE] = ACTIONS(2695),
    [anon_sym_COMMA] = ACTIONS(2695),
    [sym_keyword] = ACTIONS(2695),
    [anon_sym_LT_LT] = ACTIONS(2695),
    [anon_sym_PERCENT] = ACTIONS(2695),
    [anon_sym_DOT_DOT] = ACTIONS(2695),
    [anon_sym_AMP] = ACTIONS(2695),
    [anon_sym_PLUS] = ACTIONS(2695),
    [anon_sym_DASH] = ACTIONS(2695),
    [anon_sym_BANG] = ACTIONS(2695),
    [anon_sym_CARET] = ACTIONS(2695),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2695),
    [anon_sym_not] = ACTIONS(2695),
    [anon_sym_AT] = ACTIONS(2695),
    [anon_sym_LT_DASH] = ACTIONS(2695),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2695),
    [anon_sym_when] = ACTIONS(2695),
    [anon_sym_COLON_COLON] = ACTIONS(2695),
    [anon_sym_EQ_GT] = ACTIONS(2695),
    [anon_sym_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_or] = ACTIONS(2695),
    [anon_sym_AMP_AMP] = ACTIONS(2695),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2695),
    [anon_sym_and] = ACTIONS(2695),
    [anon_sym_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ] = ACTIONS(2695),
    [anon_sym_EQ_TILDE] = ACTIONS(2695),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2695),
    [anon_sym_LT_EQ] = ACTIONS(2695),
    [anon_sym_GT_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_LT] = ACTIONS(2695),
    [anon_sym_GT_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2695),
    [anon_sym_in] = ACTIONS(2695),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2695),
    [anon_sym_SLASH_SLASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2695),
    [anon_sym_LT_GT] = ACTIONS(2695),
    [anon_sym_STAR] = ACTIONS(2695),
    [anon_sym_STAR_STAR] = ACTIONS(2695),
    [anon_sym_DASH_GT] = ACTIONS(2695),
    [anon_sym_DOT] = ACTIONS(2695),
    [anon_sym_do] = ACTIONS(2695),
    [anon_sym_fn] = ACTIONS(2695),
    [anon_sym_LPAREN2] = ACTIONS(2693),
    [anon_sym_LBRACK2] = ACTIONS(2693),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2693),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2693),
    [sym__not_in] = ACTIONS(2693),
    [sym__quoted_atom_start] = ACTIONS(2693),
  },
  [STATE(944)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2679),
    [anon_sym_RPAREN] = ACTIONS(2679),
    [aux_sym_identifier_token1] = ACTIONS(2679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2679),
    [sym_alias] = ACTIONS(2679),
    [sym_integer] = ACTIONS(2679),
    [sym_float] = ACTIONS(2679),
    [sym_char] = ACTIONS(2679),
    [anon_sym_true] = ACTIONS(2679),
    [anon_sym_false] = ACTIONS(2679),
    [anon_sym_nil] = ACTIONS(2679),
    [sym_atom] = ACTIONS(2679),
    [anon_sym_DQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2679),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2679),
    [anon_sym_LBRACE] = ACTIONS(2679),
    [anon_sym_RBRACE] = ACTIONS(2679),
    [anon_sym_LBRACK] = ACTIONS(2679),
    [anon_sym_RBRACK] = ACTIONS(2679),
    [anon_sym_LT] = ACTIONS(2679),
    [anon_sym_GT] = ACTIONS(2679),
    [anon_sym_PIPE] = ACTIONS(2679),
    [anon_sym_SLASH] = ACTIONS(2679),
    [anon_sym_TILDE] = ACTIONS(2679),
    [anon_sym_COMMA] = ACTIONS(2679),
    [sym_keyword] = ACTIONS(2679),
    [anon_sym_LT_LT] = ACTIONS(2679),
    [anon_sym_PERCENT] = ACTIONS(2679),
    [anon_sym_DOT_DOT] = ACTIONS(2679),
    [anon_sym_AMP] = ACTIONS(2679),
    [anon_sym_PLUS] = ACTIONS(2679),
    [anon_sym_DASH] = ACTIONS(2679),
    [anon_sym_BANG] = ACTIONS(2679),
    [anon_sym_CARET] = ACTIONS(2679),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2679),
    [anon_sym_not] = ACTIONS(2679),
    [anon_sym_AT] = ACTIONS(2679),
    [anon_sym_LT_DASH] = ACTIONS(2679),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2679),
    [anon_sym_when] = ACTIONS(2679),
    [anon_sym_COLON_COLON] = ACTIONS(2679),
    [anon_sym_EQ_GT] = ACTIONS(2679),
    [anon_sym_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_or] = ACTIONS(2679),
    [anon_sym_AMP_AMP] = ACTIONS(2679),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2679),
    [anon_sym_and] = ACTIONS(2679),
    [anon_sym_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ] = ACTIONS(2679),
    [anon_sym_EQ_TILDE] = ACTIONS(2679),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2679),
    [anon_sym_LT_EQ] = ACTIONS(2679),
    [anon_sym_GT_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_LT] = ACTIONS(2679),
    [anon_sym_GT_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2679),
    [anon_sym_in] = ACTIONS(2679),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2679),
    [anon_sym_SLASH_SLASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2679),
    [anon_sym_LT_GT] = ACTIONS(2679),
    [anon_sym_STAR] = ACTIONS(2679),
    [anon_sym_STAR_STAR] = ACTIONS(2679),
    [anon_sym_DASH_GT] = ACTIONS(2679),
    [anon_sym_DOT] = ACTIONS(2679),
    [anon_sym_do] = ACTIONS(2679),
    [anon_sym_fn] = ACTIONS(2679),
    [anon_sym_LPAREN2] = ACTIONS(2677),
    [anon_sym_LBRACK2] = ACTIONS(2677),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2677),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2677),
    [sym__not_in] = ACTIONS(2677),
    [sym__quoted_atom_start] = ACTIONS(2677),
  },
  [STATE(945)] = {
    [ts_builtin_sym_end] = ACTIONS(2709),
    [aux_sym__terminator_token1] = ACTIONS(2709),
    [anon_sym_SEMI] = ACTIONS(2711),
    [anon_sym_LPAREN] = ACTIONS(2711),
    [aux_sym_identifier_token1] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [sym_alias] = ACTIONS(2711),
    [sym_integer] = ACTIONS(2711),
    [sym_float] = ACTIONS(2711),
    [sym_char] = ACTIONS(2711),
    [anon_sym_true] = ACTIONS(2711),
    [anon_sym_false] = ACTIONS(2711),
    [anon_sym_nil] = ACTIONS(2711),
    [sym_atom] = ACTIONS(2711),
    [anon_sym_DQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2711),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2711),
    [anon_sym_LBRACE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2711),
    [anon_sym_GT] = ACTIONS(2711),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_SLASH] = ACTIONS(2711),
    [anon_sym_TILDE] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [sym_keyword] = ACTIONS(2711),
    [anon_sym_LT_LT] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2711),
    [anon_sym_AMP] = ACTIONS(2711),
    [anon_sym_PLUS] = ACTIONS(2711),
    [anon_sym_DASH] = ACTIONS(2711),
    [anon_sym_BANG] = ACTIONS(2711),
    [anon_sym_CARET] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_not] = ACTIONS(2711),
    [anon_sym_AT] = ACTIONS(2711),
    [anon_sym_LT_DASH] = ACTIONS(2711),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2711),
    [anon_sym_when] = ACTIONS(2711),
    [anon_sym_COLON_COLON] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2711),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2711),
    [anon_sym_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_LT] = ACTIONS(2711),
    [anon_sym_GT_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2711),
    [anon_sym_in] = ACTIONS(2711),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2711),
    [anon_sym_LT_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2711),
    [anon_sym_STAR_STAR] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_DOT] = ACTIONS(2711),
    [anon_sym_do] = ACTIONS(2711),
    [anon_sym_fn] = ACTIONS(2711),
    [anon_sym_LPAREN2] = ACTIONS(2709),
    [anon_sym_LBRACK2] = ACTIONS(2709),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2709),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2709),
    [sym__not_in] = ACTIONS(2709),
    [sym__quoted_atom_start] = ACTIONS(2709),
  },
  [STATE(946)] = {
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_end] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(947)] = {
    [ts_builtin_sym_end] = ACTIONS(2705),
    [aux_sym__terminator_token1] = ACTIONS(2705),
    [anon_sym_SEMI] = ACTIONS(2707),
    [anon_sym_LPAREN] = ACTIONS(2707),
    [aux_sym_identifier_token1] = ACTIONS(2707),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2707),
    [sym_alias] = ACTIONS(2707),
    [sym_integer] = ACTIONS(2707),
    [sym_float] = ACTIONS(2707),
    [sym_char] = ACTIONS(2707),
    [anon_sym_true] = ACTIONS(2707),
    [anon_sym_false] = ACTIONS(2707),
    [anon_sym_nil] = ACTIONS(2707),
    [sym_atom] = ACTIONS(2707),
    [anon_sym_DQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2707),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2707),
    [anon_sym_LBRACE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(2707),
    [anon_sym_LT] = ACTIONS(2707),
    [anon_sym_GT] = ACTIONS(2707),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_SLASH] = ACTIONS(2707),
    [anon_sym_TILDE] = ACTIONS(2707),
    [anon_sym_COMMA] = ACTIONS(2707),
    [sym_keyword] = ACTIONS(2707),
    [anon_sym_LT_LT] = ACTIONS(2707),
    [anon_sym_PERCENT] = ACTIONS(2707),
    [anon_sym_DOT_DOT] = ACTIONS(2707),
    [anon_sym_AMP] = ACTIONS(2707),
    [anon_sym_PLUS] = ACTIONS(2707),
    [anon_sym_DASH] = ACTIONS(2707),
    [anon_sym_BANG] = ACTIONS(2707),
    [anon_sym_CARET] = ACTIONS(2707),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2707),
    [anon_sym_not] = ACTIONS(2707),
    [anon_sym_AT] = ACTIONS(2707),
    [anon_sym_LT_DASH] = ACTIONS(2707),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2707),
    [anon_sym_when] = ACTIONS(2707),
    [anon_sym_COLON_COLON] = ACTIONS(2707),
    [anon_sym_EQ_GT] = ACTIONS(2707),
    [anon_sym_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_or] = ACTIONS(2707),
    [anon_sym_AMP_AMP] = ACTIONS(2707),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2707),
    [anon_sym_and] = ACTIONS(2707),
    [anon_sym_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ] = ACTIONS(2707),
    [anon_sym_EQ_TILDE] = ACTIONS(2707),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2707),
    [anon_sym_LT_EQ] = ACTIONS(2707),
    [anon_sym_GT_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_LT] = ACTIONS(2707),
    [anon_sym_GT_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2707),
    [anon_sym_in] = ACTIONS(2707),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2707),
    [anon_sym_SLASH_SLASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2707),
    [anon_sym_LT_GT] = ACTIONS(2707),
    [anon_sym_STAR] = ACTIONS(2707),
    [anon_sym_STAR_STAR] = ACTIONS(2707),
    [anon_sym_DASH_GT] = ACTIONS(2707),
    [anon_sym_DOT] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2707),
    [anon_sym_fn] = ACTIONS(2707),
    [anon_sym_LPAREN2] = ACTIONS(2705),
    [anon_sym_LBRACK2] = ACTIONS(2705),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2705),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2705),
    [sym__not_in] = ACTIONS(2705),
    [sym__quoted_atom_start] = ACTIONS(2705),
  },
  [STATE(948)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_end] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(949)] = {
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_end] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(950)] = {
    [aux_sym__terminator_token1] = ACTIONS(2693),
    [anon_sym_SEMI] = ACTIONS(2695),
    [anon_sym_LPAREN] = ACTIONS(2695),
    [aux_sym_identifier_token1] = ACTIONS(2695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2695),
    [sym_alias] = ACTIONS(2695),
    [sym_integer] = ACTIONS(2695),
    [sym_float] = ACTIONS(2695),
    [sym_char] = ACTIONS(2695),
    [anon_sym_true] = ACTIONS(2695),
    [anon_sym_false] = ACTIONS(2695),
    [anon_sym_nil] = ACTIONS(2695),
    [sym_atom] = ACTIONS(2695),
    [anon_sym_DQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2695),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2695),
    [anon_sym_LBRACE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(2695),
    [anon_sym_LT] = ACTIONS(2695),
    [anon_sym_GT] = ACTIONS(2695),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_SLASH] = ACTIONS(2695),
    [anon_sym_TILDE] = ACTIONS(2695),
    [anon_sym_COMMA] = ACTIONS(2695),
    [sym_keyword] = ACTIONS(2695),
    [anon_sym_LT_LT] = ACTIONS(2695),
    [anon_sym_PERCENT] = ACTIONS(2695),
    [anon_sym_DOT_DOT] = ACTIONS(2695),
    [anon_sym_AMP] = ACTIONS(2695),
    [anon_sym_PLUS] = ACTIONS(2695),
    [anon_sym_DASH] = ACTIONS(2695),
    [anon_sym_BANG] = ACTIONS(2695),
    [anon_sym_CARET] = ACTIONS(2695),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2695),
    [anon_sym_not] = ACTIONS(2695),
    [anon_sym_AT] = ACTIONS(2695),
    [anon_sym_LT_DASH] = ACTIONS(2695),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2695),
    [anon_sym_when] = ACTIONS(2695),
    [anon_sym_COLON_COLON] = ACTIONS(2695),
    [anon_sym_EQ_GT] = ACTIONS(2695),
    [anon_sym_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_or] = ACTIONS(2695),
    [anon_sym_AMP_AMP] = ACTIONS(2695),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2695),
    [anon_sym_and] = ACTIONS(2695),
    [anon_sym_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ] = ACTIONS(2695),
    [anon_sym_EQ_TILDE] = ACTIONS(2695),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2695),
    [anon_sym_LT_EQ] = ACTIONS(2695),
    [anon_sym_GT_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_LT] = ACTIONS(2695),
    [anon_sym_GT_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2695),
    [anon_sym_in] = ACTIONS(2695),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2695),
    [anon_sym_SLASH_SLASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2695),
    [anon_sym_LT_GT] = ACTIONS(2695),
    [anon_sym_STAR] = ACTIONS(2695),
    [anon_sym_STAR_STAR] = ACTIONS(2695),
    [anon_sym_DASH_GT] = ACTIONS(2695),
    [anon_sym_DOT] = ACTIONS(2695),
    [anon_sym_do] = ACTIONS(2695),
    [anon_sym_end] = ACTIONS(2695),
    [anon_sym_fn] = ACTIONS(2695),
    [anon_sym_LPAREN2] = ACTIONS(2693),
    [anon_sym_LBRACK2] = ACTIONS(2693),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2693),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2693),
    [sym__not_in] = ACTIONS(2693),
    [sym__quoted_atom_start] = ACTIONS(2693),
  },
  [STATE(951)] = {
    [aux_sym__terminator_token1] = ACTIONS(2677),
    [anon_sym_SEMI] = ACTIONS(2679),
    [anon_sym_LPAREN] = ACTIONS(2679),
    [aux_sym_identifier_token1] = ACTIONS(2679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2679),
    [sym_alias] = ACTIONS(2679),
    [sym_integer] = ACTIONS(2679),
    [sym_float] = ACTIONS(2679),
    [sym_char] = ACTIONS(2679),
    [anon_sym_true] = ACTIONS(2679),
    [anon_sym_false] = ACTIONS(2679),
    [anon_sym_nil] = ACTIONS(2679),
    [sym_atom] = ACTIONS(2679),
    [anon_sym_DQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2679),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2679),
    [anon_sym_LBRACE] = ACTIONS(2679),
    [anon_sym_LBRACK] = ACTIONS(2679),
    [anon_sym_LT] = ACTIONS(2679),
    [anon_sym_GT] = ACTIONS(2679),
    [anon_sym_PIPE] = ACTIONS(2679),
    [anon_sym_SLASH] = ACTIONS(2679),
    [anon_sym_TILDE] = ACTIONS(2679),
    [anon_sym_COMMA] = ACTIONS(2679),
    [sym_keyword] = ACTIONS(2679),
    [anon_sym_LT_LT] = ACTIONS(2679),
    [anon_sym_PERCENT] = ACTIONS(2679),
    [anon_sym_DOT_DOT] = ACTIONS(2679),
    [anon_sym_AMP] = ACTIONS(2679),
    [anon_sym_PLUS] = ACTIONS(2679),
    [anon_sym_DASH] = ACTIONS(2679),
    [anon_sym_BANG] = ACTIONS(2679),
    [anon_sym_CARET] = ACTIONS(2679),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2679),
    [anon_sym_not] = ACTIONS(2679),
    [anon_sym_AT] = ACTIONS(2679),
    [anon_sym_LT_DASH] = ACTIONS(2679),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2679),
    [anon_sym_when] = ACTIONS(2679),
    [anon_sym_COLON_COLON] = ACTIONS(2679),
    [anon_sym_EQ_GT] = ACTIONS(2679),
    [anon_sym_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_or] = ACTIONS(2679),
    [anon_sym_AMP_AMP] = ACTIONS(2679),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2679),
    [anon_sym_and] = ACTIONS(2679),
    [anon_sym_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ] = ACTIONS(2679),
    [anon_sym_EQ_TILDE] = ACTIONS(2679),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2679),
    [anon_sym_LT_EQ] = ACTIONS(2679),
    [anon_sym_GT_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_LT] = ACTIONS(2679),
    [anon_sym_GT_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2679),
    [anon_sym_in] = ACTIONS(2679),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2679),
    [anon_sym_SLASH_SLASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2679),
    [anon_sym_LT_GT] = ACTIONS(2679),
    [anon_sym_STAR] = ACTIONS(2679),
    [anon_sym_STAR_STAR] = ACTIONS(2679),
    [anon_sym_DASH_GT] = ACTIONS(2679),
    [anon_sym_DOT] = ACTIONS(2679),
    [anon_sym_do] = ACTIONS(2679),
    [anon_sym_end] = ACTIONS(2679),
    [anon_sym_fn] = ACTIONS(2679),
    [anon_sym_LPAREN2] = ACTIONS(2677),
    [anon_sym_LBRACK2] = ACTIONS(2677),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2677),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2677),
    [sym__not_in] = ACTIONS(2677),
    [sym__quoted_atom_start] = ACTIONS(2677),
  },
  [STATE(952)] = {
    [aux_sym__terminator_token1] = ACTIONS(2697),
    [anon_sym_SEMI] = ACTIONS(2699),
    [anon_sym_LPAREN] = ACTIONS(2699),
    [aux_sym_identifier_token1] = ACTIONS(2699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2699),
    [sym_alias] = ACTIONS(2699),
    [sym_integer] = ACTIONS(2699),
    [sym_float] = ACTIONS(2699),
    [sym_char] = ACTIONS(2699),
    [anon_sym_true] = ACTIONS(2699),
    [anon_sym_false] = ACTIONS(2699),
    [anon_sym_nil] = ACTIONS(2699),
    [sym_atom] = ACTIONS(2699),
    [anon_sym_DQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2699),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2699),
    [anon_sym_LBRACE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(2699),
    [anon_sym_LT] = ACTIONS(2699),
    [anon_sym_GT] = ACTIONS(2699),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_SLASH] = ACTIONS(2699),
    [anon_sym_TILDE] = ACTIONS(2699),
    [anon_sym_COMMA] = ACTIONS(2699),
    [sym_keyword] = ACTIONS(2699),
    [anon_sym_LT_LT] = ACTIONS(2699),
    [anon_sym_PERCENT] = ACTIONS(2699),
    [anon_sym_DOT_DOT] = ACTIONS(2699),
    [anon_sym_AMP] = ACTIONS(2699),
    [anon_sym_PLUS] = ACTIONS(2699),
    [anon_sym_DASH] = ACTIONS(2699),
    [anon_sym_BANG] = ACTIONS(2699),
    [anon_sym_CARET] = ACTIONS(2699),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2699),
    [anon_sym_not] = ACTIONS(2699),
    [anon_sym_AT] = ACTIONS(2699),
    [anon_sym_LT_DASH] = ACTIONS(2699),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2699),
    [anon_sym_when] = ACTIONS(2699),
    [anon_sym_COLON_COLON] = ACTIONS(2699),
    [anon_sym_EQ_GT] = ACTIONS(2699),
    [anon_sym_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_or] = ACTIONS(2699),
    [anon_sym_AMP_AMP] = ACTIONS(2699),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2699),
    [anon_sym_and] = ACTIONS(2699),
    [anon_sym_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ] = ACTIONS(2699),
    [anon_sym_EQ_TILDE] = ACTIONS(2699),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2699),
    [anon_sym_LT_EQ] = ACTIONS(2699),
    [anon_sym_GT_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_LT] = ACTIONS(2699),
    [anon_sym_GT_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2699),
    [anon_sym_in] = ACTIONS(2699),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2699),
    [anon_sym_SLASH_SLASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2699),
    [anon_sym_LT_GT] = ACTIONS(2699),
    [anon_sym_STAR] = ACTIONS(2699),
    [anon_sym_STAR_STAR] = ACTIONS(2699),
    [anon_sym_DASH_GT] = ACTIONS(2699),
    [anon_sym_DOT] = ACTIONS(2699),
    [anon_sym_do] = ACTIONS(2699),
    [anon_sym_end] = ACTIONS(2699),
    [anon_sym_fn] = ACTIONS(2699),
    [anon_sym_LPAREN2] = ACTIONS(2697),
    [anon_sym_LBRACK2] = ACTIONS(2697),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2697),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2697),
    [sym__not_in] = ACTIONS(2697),
    [sym__quoted_atom_start] = ACTIONS(2697),
  },
  [STATE(953)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_end] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(954)] = {
    [aux_sym__terminator_token1] = ACTIONS(2681),
    [anon_sym_SEMI] = ACTIONS(2683),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [anon_sym_RPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2681),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(955)] = {
    [ts_builtin_sym_end] = ACTIONS(2689),
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(956)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_end] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(957)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_end] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(958)] = {
    [ts_builtin_sym_end] = ACTIONS(2701),
    [aux_sym__terminator_token1] = ACTIONS(2701),
    [anon_sym_SEMI] = ACTIONS(2703),
    [anon_sym_LPAREN] = ACTIONS(2703),
    [aux_sym_identifier_token1] = ACTIONS(2703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2703),
    [sym_alias] = ACTIONS(2703),
    [sym_integer] = ACTIONS(2703),
    [sym_float] = ACTIONS(2703),
    [sym_char] = ACTIONS(2703),
    [anon_sym_true] = ACTIONS(2703),
    [anon_sym_false] = ACTIONS(2703),
    [anon_sym_nil] = ACTIONS(2703),
    [sym_atom] = ACTIONS(2703),
    [anon_sym_DQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2703),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2703),
    [anon_sym_LBRACE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(2703),
    [anon_sym_LT] = ACTIONS(2703),
    [anon_sym_GT] = ACTIONS(2703),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_SLASH] = ACTIONS(2703),
    [anon_sym_TILDE] = ACTIONS(2703),
    [anon_sym_COMMA] = ACTIONS(2703),
    [sym_keyword] = ACTIONS(2703),
    [anon_sym_LT_LT] = ACTIONS(2703),
    [anon_sym_PERCENT] = ACTIONS(2703),
    [anon_sym_DOT_DOT] = ACTIONS(2703),
    [anon_sym_AMP] = ACTIONS(2703),
    [anon_sym_PLUS] = ACTIONS(2703),
    [anon_sym_DASH] = ACTIONS(2703),
    [anon_sym_BANG] = ACTIONS(2703),
    [anon_sym_CARET] = ACTIONS(2703),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2703),
    [anon_sym_not] = ACTIONS(2703),
    [anon_sym_AT] = ACTIONS(2703),
    [anon_sym_LT_DASH] = ACTIONS(2703),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2703),
    [anon_sym_when] = ACTIONS(2703),
    [anon_sym_COLON_COLON] = ACTIONS(2703),
    [anon_sym_EQ_GT] = ACTIONS(2703),
    [anon_sym_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_or] = ACTIONS(2703),
    [anon_sym_AMP_AMP] = ACTIONS(2703),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2703),
    [anon_sym_and] = ACTIONS(2703),
    [anon_sym_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ] = ACTIONS(2703),
    [anon_sym_EQ_TILDE] = ACTIONS(2703),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2703),
    [anon_sym_LT_EQ] = ACTIONS(2703),
    [anon_sym_GT_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_LT] = ACTIONS(2703),
    [anon_sym_GT_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2703),
    [anon_sym_in] = ACTIONS(2703),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2703),
    [anon_sym_SLASH_SLASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2703),
    [anon_sym_LT_GT] = ACTIONS(2703),
    [anon_sym_STAR] = ACTIONS(2703),
    [anon_sym_STAR_STAR] = ACTIONS(2703),
    [anon_sym_DASH_GT] = ACTIONS(2703),
    [anon_sym_DOT] = ACTIONS(2703),
    [anon_sym_do] = ACTIONS(2703),
    [anon_sym_fn] = ACTIONS(2703),
    [anon_sym_LPAREN2] = ACTIONS(2701),
    [anon_sym_LBRACK2] = ACTIONS(2701),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2701),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2701),
    [sym__not_in] = ACTIONS(2701),
    [sym__quoted_atom_start] = ACTIONS(2701),
  },
  [STATE(959)] = {
    [ts_builtin_sym_end] = ACTIONS(2665),
    [aux_sym__terminator_token1] = ACTIONS(2665),
    [anon_sym_SEMI] = ACTIONS(2667),
    [anon_sym_LPAREN] = ACTIONS(2667),
    [aux_sym_identifier_token1] = ACTIONS(2667),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2667),
    [sym_alias] = ACTIONS(2667),
    [sym_integer] = ACTIONS(2667),
    [sym_float] = ACTIONS(2667),
    [sym_char] = ACTIONS(2667),
    [anon_sym_true] = ACTIONS(2667),
    [anon_sym_false] = ACTIONS(2667),
    [anon_sym_nil] = ACTIONS(2667),
    [sym_atom] = ACTIONS(2667),
    [anon_sym_DQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2667),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2667),
    [anon_sym_LBRACE] = ACTIONS(2667),
    [anon_sym_LBRACK] = ACTIONS(2667),
    [anon_sym_LT] = ACTIONS(2667),
    [anon_sym_GT] = ACTIONS(2667),
    [anon_sym_PIPE] = ACTIONS(2667),
    [anon_sym_SLASH] = ACTIONS(2667),
    [anon_sym_TILDE] = ACTIONS(2667),
    [anon_sym_COMMA] = ACTIONS(2667),
    [sym_keyword] = ACTIONS(2667),
    [anon_sym_LT_LT] = ACTIONS(2667),
    [anon_sym_PERCENT] = ACTIONS(2667),
    [anon_sym_DOT_DOT] = ACTIONS(2667),
    [anon_sym_AMP] = ACTIONS(2667),
    [anon_sym_PLUS] = ACTIONS(2667),
    [anon_sym_DASH] = ACTIONS(2667),
    [anon_sym_BANG] = ACTIONS(2667),
    [anon_sym_CARET] = ACTIONS(2667),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2667),
    [anon_sym_not] = ACTIONS(2667),
    [anon_sym_AT] = ACTIONS(2667),
    [anon_sym_LT_DASH] = ACTIONS(2667),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2667),
    [anon_sym_when] = ACTIONS(2667),
    [anon_sym_COLON_COLON] = ACTIONS(2667),
    [anon_sym_EQ_GT] = ACTIONS(2667),
    [anon_sym_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_or] = ACTIONS(2667),
    [anon_sym_AMP_AMP] = ACTIONS(2667),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2667),
    [anon_sym_and] = ACTIONS(2667),
    [anon_sym_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ] = ACTIONS(2667),
    [anon_sym_EQ_TILDE] = ACTIONS(2667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2667),
    [anon_sym_LT_EQ] = ACTIONS(2667),
    [anon_sym_GT_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_LT] = ACTIONS(2667),
    [anon_sym_GT_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2667),
    [anon_sym_in] = ACTIONS(2667),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2667),
    [anon_sym_SLASH_SLASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2667),
    [anon_sym_LT_GT] = ACTIONS(2667),
    [anon_sym_STAR] = ACTIONS(2667),
    [anon_sym_STAR_STAR] = ACTIONS(2667),
    [anon_sym_DASH_GT] = ACTIONS(2667),
    [anon_sym_DOT] = ACTIONS(2667),
    [anon_sym_do] = ACTIONS(2667),
    [anon_sym_fn] = ACTIONS(2667),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_LBRACK2] = ACTIONS(2665),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2665),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2665),
    [sym__not_in] = ACTIONS(2665),
    [sym__quoted_atom_start] = ACTIONS(2665),
  },
  [STATE(960)] = {
    [ts_builtin_sym_end] = ACTIONS(2669),
    [aux_sym__terminator_token1] = ACTIONS(2669),
    [anon_sym_SEMI] = ACTIONS(2671),
    [anon_sym_LPAREN] = ACTIONS(2671),
    [aux_sym_identifier_token1] = ACTIONS(2671),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2671),
    [sym_alias] = ACTIONS(2671),
    [sym_integer] = ACTIONS(2671),
    [sym_float] = ACTIONS(2671),
    [sym_char] = ACTIONS(2671),
    [anon_sym_true] = ACTIONS(2671),
    [anon_sym_false] = ACTIONS(2671),
    [anon_sym_nil] = ACTIONS(2671),
    [sym_atom] = ACTIONS(2671),
    [anon_sym_DQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2671),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2671),
    [anon_sym_LBRACE] = ACTIONS(2671),
    [anon_sym_LBRACK] = ACTIONS(2671),
    [anon_sym_LT] = ACTIONS(2671),
    [anon_sym_GT] = ACTIONS(2671),
    [anon_sym_PIPE] = ACTIONS(2671),
    [anon_sym_SLASH] = ACTIONS(2671),
    [anon_sym_TILDE] = ACTIONS(2671),
    [anon_sym_COMMA] = ACTIONS(2671),
    [sym_keyword] = ACTIONS(2671),
    [anon_sym_LT_LT] = ACTIONS(2671),
    [anon_sym_PERCENT] = ACTIONS(2671),
    [anon_sym_DOT_DOT] = ACTIONS(2671),
    [anon_sym_AMP] = ACTIONS(2671),
    [anon_sym_PLUS] = ACTIONS(2671),
    [anon_sym_DASH] = ACTIONS(2671),
    [anon_sym_BANG] = ACTIONS(2671),
    [anon_sym_CARET] = ACTIONS(2671),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2671),
    [anon_sym_not] = ACTIONS(2671),
    [anon_sym_AT] = ACTIONS(2671),
    [anon_sym_LT_DASH] = ACTIONS(2671),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2671),
    [anon_sym_when] = ACTIONS(2671),
    [anon_sym_COLON_COLON] = ACTIONS(2671),
    [anon_sym_EQ_GT] = ACTIONS(2671),
    [anon_sym_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_or] = ACTIONS(2671),
    [anon_sym_AMP_AMP] = ACTIONS(2671),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2671),
    [anon_sym_and] = ACTIONS(2671),
    [anon_sym_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ] = ACTIONS(2671),
    [anon_sym_EQ_TILDE] = ACTIONS(2671),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2671),
    [anon_sym_LT_EQ] = ACTIONS(2671),
    [anon_sym_GT_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_LT] = ACTIONS(2671),
    [anon_sym_GT_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2671),
    [anon_sym_in] = ACTIONS(2671),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2671),
    [anon_sym_SLASH_SLASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2671),
    [anon_sym_LT_GT] = ACTIONS(2671),
    [anon_sym_STAR] = ACTIONS(2671),
    [anon_sym_STAR_STAR] = ACTIONS(2671),
    [anon_sym_DASH_GT] = ACTIONS(2671),
    [anon_sym_DOT] = ACTIONS(2671),
    [anon_sym_do] = ACTIONS(2671),
    [anon_sym_fn] = ACTIONS(2671),
    [anon_sym_LPAREN2] = ACTIONS(2669),
    [anon_sym_LBRACK2] = ACTIONS(2669),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2669),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2669),
    [sym__not_in] = ACTIONS(2669),
    [sym__quoted_atom_start] = ACTIONS(2669),
  },
  [STATE(961)] = {
    [aux_sym__terminator_token1] = ACTIONS(2701),
    [anon_sym_SEMI] = ACTIONS(2703),
    [anon_sym_LPAREN] = ACTIONS(2703),
    [aux_sym_identifier_token1] = ACTIONS(2703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2703),
    [sym_alias] = ACTIONS(2703),
    [sym_integer] = ACTIONS(2703),
    [sym_float] = ACTIONS(2703),
    [sym_char] = ACTIONS(2703),
    [anon_sym_true] = ACTIONS(2703),
    [anon_sym_false] = ACTIONS(2703),
    [anon_sym_nil] = ACTIONS(2703),
    [sym_atom] = ACTIONS(2703),
    [anon_sym_DQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2703),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2703),
    [anon_sym_LBRACE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(2703),
    [anon_sym_LT] = ACTIONS(2703),
    [anon_sym_GT] = ACTIONS(2703),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_SLASH] = ACTIONS(2703),
    [anon_sym_TILDE] = ACTIONS(2703),
    [anon_sym_COMMA] = ACTIONS(2703),
    [sym_keyword] = ACTIONS(2703),
    [anon_sym_LT_LT] = ACTIONS(2703),
    [anon_sym_PERCENT] = ACTIONS(2703),
    [anon_sym_DOT_DOT] = ACTIONS(2703),
    [anon_sym_AMP] = ACTIONS(2703),
    [anon_sym_PLUS] = ACTIONS(2703),
    [anon_sym_DASH] = ACTIONS(2703),
    [anon_sym_BANG] = ACTIONS(2703),
    [anon_sym_CARET] = ACTIONS(2703),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2703),
    [anon_sym_not] = ACTIONS(2703),
    [anon_sym_AT] = ACTIONS(2703),
    [anon_sym_LT_DASH] = ACTIONS(2703),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2703),
    [anon_sym_when] = ACTIONS(2703),
    [anon_sym_COLON_COLON] = ACTIONS(2703),
    [anon_sym_EQ_GT] = ACTIONS(2703),
    [anon_sym_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_or] = ACTIONS(2703),
    [anon_sym_AMP_AMP] = ACTIONS(2703),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2703),
    [anon_sym_and] = ACTIONS(2703),
    [anon_sym_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ] = ACTIONS(2703),
    [anon_sym_EQ_TILDE] = ACTIONS(2703),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2703),
    [anon_sym_LT_EQ] = ACTIONS(2703),
    [anon_sym_GT_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_LT] = ACTIONS(2703),
    [anon_sym_GT_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2703),
    [anon_sym_in] = ACTIONS(2703),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2703),
    [anon_sym_SLASH_SLASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2703),
    [anon_sym_LT_GT] = ACTIONS(2703),
    [anon_sym_STAR] = ACTIONS(2703),
    [anon_sym_STAR_STAR] = ACTIONS(2703),
    [anon_sym_DASH_GT] = ACTIONS(2703),
    [anon_sym_DOT] = ACTIONS(2703),
    [anon_sym_do] = ACTIONS(2703),
    [anon_sym_end] = ACTIONS(2703),
    [anon_sym_fn] = ACTIONS(2703),
    [anon_sym_LPAREN2] = ACTIONS(2701),
    [anon_sym_LBRACK2] = ACTIONS(2701),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2701),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2701),
    [sym__not_in] = ACTIONS(2701),
    [sym__quoted_atom_start] = ACTIONS(2701),
  },
  [STATE(962)] = {
    [aux_sym__terminator_token1] = ACTIONS(2705),
    [anon_sym_SEMI] = ACTIONS(2707),
    [anon_sym_LPAREN] = ACTIONS(2707),
    [aux_sym_identifier_token1] = ACTIONS(2707),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2707),
    [sym_alias] = ACTIONS(2707),
    [sym_integer] = ACTIONS(2707),
    [sym_float] = ACTIONS(2707),
    [sym_char] = ACTIONS(2707),
    [anon_sym_true] = ACTIONS(2707),
    [anon_sym_false] = ACTIONS(2707),
    [anon_sym_nil] = ACTIONS(2707),
    [sym_atom] = ACTIONS(2707),
    [anon_sym_DQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2707),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2707),
    [anon_sym_LBRACE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(2707),
    [anon_sym_LT] = ACTIONS(2707),
    [anon_sym_GT] = ACTIONS(2707),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_SLASH] = ACTIONS(2707),
    [anon_sym_TILDE] = ACTIONS(2707),
    [anon_sym_COMMA] = ACTIONS(2707),
    [sym_keyword] = ACTIONS(2707),
    [anon_sym_LT_LT] = ACTIONS(2707),
    [anon_sym_PERCENT] = ACTIONS(2707),
    [anon_sym_DOT_DOT] = ACTIONS(2707),
    [anon_sym_AMP] = ACTIONS(2707),
    [anon_sym_PLUS] = ACTIONS(2707),
    [anon_sym_DASH] = ACTIONS(2707),
    [anon_sym_BANG] = ACTIONS(2707),
    [anon_sym_CARET] = ACTIONS(2707),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2707),
    [anon_sym_not] = ACTIONS(2707),
    [anon_sym_AT] = ACTIONS(2707),
    [anon_sym_LT_DASH] = ACTIONS(2707),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2707),
    [anon_sym_when] = ACTIONS(2707),
    [anon_sym_COLON_COLON] = ACTIONS(2707),
    [anon_sym_EQ_GT] = ACTIONS(2707),
    [anon_sym_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_or] = ACTIONS(2707),
    [anon_sym_AMP_AMP] = ACTIONS(2707),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2707),
    [anon_sym_and] = ACTIONS(2707),
    [anon_sym_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ] = ACTIONS(2707),
    [anon_sym_EQ_TILDE] = ACTIONS(2707),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2707),
    [anon_sym_LT_EQ] = ACTIONS(2707),
    [anon_sym_GT_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_LT] = ACTIONS(2707),
    [anon_sym_GT_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2707),
    [anon_sym_in] = ACTIONS(2707),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2707),
    [anon_sym_SLASH_SLASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2707),
    [anon_sym_LT_GT] = ACTIONS(2707),
    [anon_sym_STAR] = ACTIONS(2707),
    [anon_sym_STAR_STAR] = ACTIONS(2707),
    [anon_sym_DASH_GT] = ACTIONS(2707),
    [anon_sym_DOT] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2707),
    [anon_sym_end] = ACTIONS(2707),
    [anon_sym_fn] = ACTIONS(2707),
    [anon_sym_LPAREN2] = ACTIONS(2705),
    [anon_sym_LBRACK2] = ACTIONS(2705),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2705),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2705),
    [sym__not_in] = ACTIONS(2705),
    [sym__quoted_atom_start] = ACTIONS(2705),
  },
  [STATE(963)] = {
    [aux_sym__terminator_token1] = ACTIONS(2709),
    [anon_sym_SEMI] = ACTIONS(2711),
    [anon_sym_LPAREN] = ACTIONS(2711),
    [aux_sym_identifier_token1] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [sym_alias] = ACTIONS(2711),
    [sym_integer] = ACTIONS(2711),
    [sym_float] = ACTIONS(2711),
    [sym_char] = ACTIONS(2711),
    [anon_sym_true] = ACTIONS(2711),
    [anon_sym_false] = ACTIONS(2711),
    [anon_sym_nil] = ACTIONS(2711),
    [sym_atom] = ACTIONS(2711),
    [anon_sym_DQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2711),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2711),
    [anon_sym_LBRACE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2711),
    [anon_sym_GT] = ACTIONS(2711),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_SLASH] = ACTIONS(2711),
    [anon_sym_TILDE] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [sym_keyword] = ACTIONS(2711),
    [anon_sym_LT_LT] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2711),
    [anon_sym_AMP] = ACTIONS(2711),
    [anon_sym_PLUS] = ACTIONS(2711),
    [anon_sym_DASH] = ACTIONS(2711),
    [anon_sym_BANG] = ACTIONS(2711),
    [anon_sym_CARET] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_not] = ACTIONS(2711),
    [anon_sym_AT] = ACTIONS(2711),
    [anon_sym_LT_DASH] = ACTIONS(2711),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2711),
    [anon_sym_when] = ACTIONS(2711),
    [anon_sym_COLON_COLON] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2711),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2711),
    [anon_sym_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_LT] = ACTIONS(2711),
    [anon_sym_GT_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2711),
    [anon_sym_in] = ACTIONS(2711),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2711),
    [anon_sym_LT_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2711),
    [anon_sym_STAR_STAR] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_DOT] = ACTIONS(2711),
    [anon_sym_do] = ACTIONS(2711),
    [anon_sym_end] = ACTIONS(2711),
    [anon_sym_fn] = ACTIONS(2711),
    [anon_sym_LPAREN2] = ACTIONS(2709),
    [anon_sym_LBRACK2] = ACTIONS(2709),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2709),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2709),
    [sym__not_in] = ACTIONS(2709),
    [sym__quoted_atom_start] = ACTIONS(2709),
  },
  [STATE(964)] = {
    [aux_sym__terminator_token1] = ACTIONS(2665),
    [anon_sym_SEMI] = ACTIONS(2667),
    [anon_sym_LPAREN] = ACTIONS(2667),
    [aux_sym_identifier_token1] = ACTIONS(2667),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2667),
    [sym_alias] = ACTIONS(2667),
    [sym_integer] = ACTIONS(2667),
    [sym_float] = ACTIONS(2667),
    [sym_char] = ACTIONS(2667),
    [anon_sym_true] = ACTIONS(2667),
    [anon_sym_false] = ACTIONS(2667),
    [anon_sym_nil] = ACTIONS(2667),
    [sym_atom] = ACTIONS(2667),
    [anon_sym_DQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2667),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2667),
    [anon_sym_LBRACE] = ACTIONS(2667),
    [anon_sym_LBRACK] = ACTIONS(2667),
    [anon_sym_LT] = ACTIONS(2667),
    [anon_sym_GT] = ACTIONS(2667),
    [anon_sym_PIPE] = ACTIONS(2667),
    [anon_sym_SLASH] = ACTIONS(2667),
    [anon_sym_TILDE] = ACTIONS(2667),
    [anon_sym_COMMA] = ACTIONS(2667),
    [sym_keyword] = ACTIONS(2667),
    [anon_sym_LT_LT] = ACTIONS(2667),
    [anon_sym_PERCENT] = ACTIONS(2667),
    [anon_sym_DOT_DOT] = ACTIONS(2667),
    [anon_sym_AMP] = ACTIONS(2667),
    [anon_sym_PLUS] = ACTIONS(2667),
    [anon_sym_DASH] = ACTIONS(2667),
    [anon_sym_BANG] = ACTIONS(2667),
    [anon_sym_CARET] = ACTIONS(2667),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2667),
    [anon_sym_not] = ACTIONS(2667),
    [anon_sym_AT] = ACTIONS(2667),
    [anon_sym_LT_DASH] = ACTIONS(2667),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2667),
    [anon_sym_when] = ACTIONS(2667),
    [anon_sym_COLON_COLON] = ACTIONS(2667),
    [anon_sym_EQ_GT] = ACTIONS(2667),
    [anon_sym_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_or] = ACTIONS(2667),
    [anon_sym_AMP_AMP] = ACTIONS(2667),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2667),
    [anon_sym_and] = ACTIONS(2667),
    [anon_sym_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ] = ACTIONS(2667),
    [anon_sym_EQ_TILDE] = ACTIONS(2667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2667),
    [anon_sym_LT_EQ] = ACTIONS(2667),
    [anon_sym_GT_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_LT] = ACTIONS(2667),
    [anon_sym_GT_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2667),
    [anon_sym_in] = ACTIONS(2667),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2667),
    [anon_sym_SLASH_SLASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2667),
    [anon_sym_LT_GT] = ACTIONS(2667),
    [anon_sym_STAR] = ACTIONS(2667),
    [anon_sym_STAR_STAR] = ACTIONS(2667),
    [anon_sym_DASH_GT] = ACTIONS(2667),
    [anon_sym_DOT] = ACTIONS(2667),
    [anon_sym_do] = ACTIONS(2667),
    [anon_sym_end] = ACTIONS(2667),
    [anon_sym_fn] = ACTIONS(2667),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_LBRACK2] = ACTIONS(2665),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2665),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2665),
    [sym__not_in] = ACTIONS(2665),
    [sym__quoted_atom_start] = ACTIONS(2665),
  },
  [STATE(965)] = {
    [aux_sym__terminator_token1] = ACTIONS(2669),
    [anon_sym_SEMI] = ACTIONS(2671),
    [anon_sym_LPAREN] = ACTIONS(2671),
    [aux_sym_identifier_token1] = ACTIONS(2671),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2671),
    [sym_alias] = ACTIONS(2671),
    [sym_integer] = ACTIONS(2671),
    [sym_float] = ACTIONS(2671),
    [sym_char] = ACTIONS(2671),
    [anon_sym_true] = ACTIONS(2671),
    [anon_sym_false] = ACTIONS(2671),
    [anon_sym_nil] = ACTIONS(2671),
    [sym_atom] = ACTIONS(2671),
    [anon_sym_DQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2671),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2671),
    [anon_sym_LBRACE] = ACTIONS(2671),
    [anon_sym_LBRACK] = ACTIONS(2671),
    [anon_sym_LT] = ACTIONS(2671),
    [anon_sym_GT] = ACTIONS(2671),
    [anon_sym_PIPE] = ACTIONS(2671),
    [anon_sym_SLASH] = ACTIONS(2671),
    [anon_sym_TILDE] = ACTIONS(2671),
    [anon_sym_COMMA] = ACTIONS(2671),
    [sym_keyword] = ACTIONS(2671),
    [anon_sym_LT_LT] = ACTIONS(2671),
    [anon_sym_PERCENT] = ACTIONS(2671),
    [anon_sym_DOT_DOT] = ACTIONS(2671),
    [anon_sym_AMP] = ACTIONS(2671),
    [anon_sym_PLUS] = ACTIONS(2671),
    [anon_sym_DASH] = ACTIONS(2671),
    [anon_sym_BANG] = ACTIONS(2671),
    [anon_sym_CARET] = ACTIONS(2671),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2671),
    [anon_sym_not] = ACTIONS(2671),
    [anon_sym_AT] = ACTIONS(2671),
    [anon_sym_LT_DASH] = ACTIONS(2671),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2671),
    [anon_sym_when] = ACTIONS(2671),
    [anon_sym_COLON_COLON] = ACTIONS(2671),
    [anon_sym_EQ_GT] = ACTIONS(2671),
    [anon_sym_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_or] = ACTIONS(2671),
    [anon_sym_AMP_AMP] = ACTIONS(2671),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2671),
    [anon_sym_and] = ACTIONS(2671),
    [anon_sym_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ] = ACTIONS(2671),
    [anon_sym_EQ_TILDE] = ACTIONS(2671),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2671),
    [anon_sym_LT_EQ] = ACTIONS(2671),
    [anon_sym_GT_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_LT] = ACTIONS(2671),
    [anon_sym_GT_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2671),
    [anon_sym_in] = ACTIONS(2671),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2671),
    [anon_sym_SLASH_SLASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2671),
    [anon_sym_LT_GT] = ACTIONS(2671),
    [anon_sym_STAR] = ACTIONS(2671),
    [anon_sym_STAR_STAR] = ACTIONS(2671),
    [anon_sym_DASH_GT] = ACTIONS(2671),
    [anon_sym_DOT] = ACTIONS(2671),
    [anon_sym_do] = ACTIONS(2671),
    [anon_sym_end] = ACTIONS(2671),
    [anon_sym_fn] = ACTIONS(2671),
    [anon_sym_LPAREN2] = ACTIONS(2669),
    [anon_sym_LBRACK2] = ACTIONS(2669),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2669),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2669),
    [sym__not_in] = ACTIONS(2669),
    [sym__quoted_atom_start] = ACTIONS(2669),
  },
  [STATE(966)] = {
    [aux_sym__terminator_token1] = ACTIONS(2673),
    [anon_sym_SEMI] = ACTIONS(2675),
    [anon_sym_LPAREN] = ACTIONS(2675),
    [aux_sym_identifier_token1] = ACTIONS(2675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2675),
    [sym_alias] = ACTIONS(2675),
    [sym_integer] = ACTIONS(2675),
    [sym_float] = ACTIONS(2675),
    [sym_char] = ACTIONS(2675),
    [anon_sym_true] = ACTIONS(2675),
    [anon_sym_false] = ACTIONS(2675),
    [anon_sym_nil] = ACTIONS(2675),
    [sym_atom] = ACTIONS(2675),
    [anon_sym_DQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2675),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2675),
    [anon_sym_LBRACE] = ACTIONS(2675),
    [anon_sym_LBRACK] = ACTIONS(2675),
    [anon_sym_LT] = ACTIONS(2675),
    [anon_sym_GT] = ACTIONS(2675),
    [anon_sym_PIPE] = ACTIONS(2675),
    [anon_sym_SLASH] = ACTIONS(2675),
    [anon_sym_TILDE] = ACTIONS(2675),
    [anon_sym_COMMA] = ACTIONS(2675),
    [sym_keyword] = ACTIONS(2675),
    [anon_sym_LT_LT] = ACTIONS(2675),
    [anon_sym_PERCENT] = ACTIONS(2675),
    [anon_sym_DOT_DOT] = ACTIONS(2675),
    [anon_sym_AMP] = ACTIONS(2675),
    [anon_sym_PLUS] = ACTIONS(2675),
    [anon_sym_DASH] = ACTIONS(2675),
    [anon_sym_BANG] = ACTIONS(2675),
    [anon_sym_CARET] = ACTIONS(2675),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2675),
    [anon_sym_not] = ACTIONS(2675),
    [anon_sym_AT] = ACTIONS(2675),
    [anon_sym_LT_DASH] = ACTIONS(2675),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2675),
    [anon_sym_when] = ACTIONS(2675),
    [anon_sym_COLON_COLON] = ACTIONS(2675),
    [anon_sym_EQ_GT] = ACTIONS(2675),
    [anon_sym_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_or] = ACTIONS(2675),
    [anon_sym_AMP_AMP] = ACTIONS(2675),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2675),
    [anon_sym_and] = ACTIONS(2675),
    [anon_sym_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ] = ACTIONS(2675),
    [anon_sym_EQ_TILDE] = ACTIONS(2675),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2675),
    [anon_sym_LT_EQ] = ACTIONS(2675),
    [anon_sym_GT_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_LT] = ACTIONS(2675),
    [anon_sym_GT_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2675),
    [anon_sym_in] = ACTIONS(2675),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2675),
    [anon_sym_SLASH_SLASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2675),
    [anon_sym_LT_GT] = ACTIONS(2675),
    [anon_sym_STAR] = ACTIONS(2675),
    [anon_sym_STAR_STAR] = ACTIONS(2675),
    [anon_sym_DASH_GT] = ACTIONS(2675),
    [anon_sym_DOT] = ACTIONS(2675),
    [anon_sym_do] = ACTIONS(2675),
    [anon_sym_end] = ACTIONS(2675),
    [anon_sym_fn] = ACTIONS(2675),
    [anon_sym_LPAREN2] = ACTIONS(2673),
    [anon_sym_LBRACK2] = ACTIONS(2673),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2673),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2673),
    [sym__not_in] = ACTIONS(2673),
    [sym__quoted_atom_start] = ACTIONS(2673),
  },
  [STATE(967)] = {
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_end] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(968)] = {
    [aux_sym__terminator_token1] = ACTIONS(2681),
    [anon_sym_SEMI] = ACTIONS(2683),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_end] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2681),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(969)] = {
    [ts_builtin_sym_end] = ACTIONS(2685),
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(970)] = {
    [ts_builtin_sym_end] = ACTIONS(2685),
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(971)] = {
    [ts_builtin_sym_end] = ACTIONS(2689),
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(972)] = {
    [ts_builtin_sym_end] = ACTIONS(2685),
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(973)] = {
    [ts_builtin_sym_end] = ACTIONS(2689),
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(974)] = {
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [anon_sym_RPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(975)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [anon_sym_RPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(976)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [anon_sym_RPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(977)] = {
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [anon_sym_RPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(978)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [anon_sym_RPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(979)] = {
    [aux_sym__terminator_token1] = ACTIONS(2689),
    [anon_sym_SEMI] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [anon_sym_RPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(980)] = {
    [aux_sym__terminator_token1] = ACTIONS(2693),
    [anon_sym_SEMI] = ACTIONS(2695),
    [anon_sym_LPAREN] = ACTIONS(2695),
    [anon_sym_RPAREN] = ACTIONS(2695),
    [aux_sym_identifier_token1] = ACTIONS(2695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2695),
    [sym_alias] = ACTIONS(2695),
    [sym_integer] = ACTIONS(2695),
    [sym_float] = ACTIONS(2695),
    [sym_char] = ACTIONS(2695),
    [anon_sym_true] = ACTIONS(2695),
    [anon_sym_false] = ACTIONS(2695),
    [anon_sym_nil] = ACTIONS(2695),
    [sym_atom] = ACTIONS(2695),
    [anon_sym_DQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2695),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2695),
    [anon_sym_LBRACE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(2695),
    [anon_sym_LT] = ACTIONS(2695),
    [anon_sym_GT] = ACTIONS(2695),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_SLASH] = ACTIONS(2695),
    [anon_sym_TILDE] = ACTIONS(2695),
    [anon_sym_COMMA] = ACTIONS(2695),
    [sym_keyword] = ACTIONS(2695),
    [anon_sym_LT_LT] = ACTIONS(2695),
    [anon_sym_PERCENT] = ACTIONS(2695),
    [anon_sym_DOT_DOT] = ACTIONS(2695),
    [anon_sym_AMP] = ACTIONS(2695),
    [anon_sym_PLUS] = ACTIONS(2695),
    [anon_sym_DASH] = ACTIONS(2695),
    [anon_sym_BANG] = ACTIONS(2695),
    [anon_sym_CARET] = ACTIONS(2695),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2695),
    [anon_sym_not] = ACTIONS(2695),
    [anon_sym_AT] = ACTIONS(2695),
    [anon_sym_LT_DASH] = ACTIONS(2695),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2695),
    [anon_sym_when] = ACTIONS(2695),
    [anon_sym_COLON_COLON] = ACTIONS(2695),
    [anon_sym_EQ_GT] = ACTIONS(2695),
    [anon_sym_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_or] = ACTIONS(2695),
    [anon_sym_AMP_AMP] = ACTIONS(2695),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2695),
    [anon_sym_and] = ACTIONS(2695),
    [anon_sym_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ] = ACTIONS(2695),
    [anon_sym_EQ_TILDE] = ACTIONS(2695),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2695),
    [anon_sym_LT_EQ] = ACTIONS(2695),
    [anon_sym_GT_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_LT] = ACTIONS(2695),
    [anon_sym_GT_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2695),
    [anon_sym_in] = ACTIONS(2695),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2695),
    [anon_sym_SLASH_SLASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2695),
    [anon_sym_LT_GT] = ACTIONS(2695),
    [anon_sym_STAR] = ACTIONS(2695),
    [anon_sym_STAR_STAR] = ACTIONS(2695),
    [anon_sym_DASH_GT] = ACTIONS(2695),
    [anon_sym_DOT] = ACTIONS(2695),
    [anon_sym_do] = ACTIONS(2695),
    [anon_sym_fn] = ACTIONS(2695),
    [anon_sym_LPAREN2] = ACTIONS(2693),
    [anon_sym_LBRACK2] = ACTIONS(2693),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2693),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2693),
    [sym__not_in] = ACTIONS(2693),
    [sym__quoted_atom_start] = ACTIONS(2693),
  },
  [STATE(981)] = {
    [aux_sym__terminator_token1] = ACTIONS(2677),
    [anon_sym_SEMI] = ACTIONS(2679),
    [anon_sym_LPAREN] = ACTIONS(2679),
    [anon_sym_RPAREN] = ACTIONS(2679),
    [aux_sym_identifier_token1] = ACTIONS(2679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2679),
    [sym_alias] = ACTIONS(2679),
    [sym_integer] = ACTIONS(2679),
    [sym_float] = ACTIONS(2679),
    [sym_char] = ACTIONS(2679),
    [anon_sym_true] = ACTIONS(2679),
    [anon_sym_false] = ACTIONS(2679),
    [anon_sym_nil] = ACTIONS(2679),
    [sym_atom] = ACTIONS(2679),
    [anon_sym_DQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2679),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2679),
    [anon_sym_LBRACE] = ACTIONS(2679),
    [anon_sym_LBRACK] = ACTIONS(2679),
    [anon_sym_LT] = ACTIONS(2679),
    [anon_sym_GT] = ACTIONS(2679),
    [anon_sym_PIPE] = ACTIONS(2679),
    [anon_sym_SLASH] = ACTIONS(2679),
    [anon_sym_TILDE] = ACTIONS(2679),
    [anon_sym_COMMA] = ACTIONS(2679),
    [sym_keyword] = ACTIONS(2679),
    [anon_sym_LT_LT] = ACTIONS(2679),
    [anon_sym_PERCENT] = ACTIONS(2679),
    [anon_sym_DOT_DOT] = ACTIONS(2679),
    [anon_sym_AMP] = ACTIONS(2679),
    [anon_sym_PLUS] = ACTIONS(2679),
    [anon_sym_DASH] = ACTIONS(2679),
    [anon_sym_BANG] = ACTIONS(2679),
    [anon_sym_CARET] = ACTIONS(2679),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2679),
    [anon_sym_not] = ACTIONS(2679),
    [anon_sym_AT] = ACTIONS(2679),
    [anon_sym_LT_DASH] = ACTIONS(2679),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2679),
    [anon_sym_when] = ACTIONS(2679),
    [anon_sym_COLON_COLON] = ACTIONS(2679),
    [anon_sym_EQ_GT] = ACTIONS(2679),
    [anon_sym_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_or] = ACTIONS(2679),
    [anon_sym_AMP_AMP] = ACTIONS(2679),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2679),
    [anon_sym_and] = ACTIONS(2679),
    [anon_sym_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ] = ACTIONS(2679),
    [anon_sym_EQ_TILDE] = ACTIONS(2679),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2679),
    [anon_sym_LT_EQ] = ACTIONS(2679),
    [anon_sym_GT_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_LT] = ACTIONS(2679),
    [anon_sym_GT_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2679),
    [anon_sym_in] = ACTIONS(2679),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2679),
    [anon_sym_SLASH_SLASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2679),
    [anon_sym_LT_GT] = ACTIONS(2679),
    [anon_sym_STAR] = ACTIONS(2679),
    [anon_sym_STAR_STAR] = ACTIONS(2679),
    [anon_sym_DASH_GT] = ACTIONS(2679),
    [anon_sym_DOT] = ACTIONS(2679),
    [anon_sym_do] = ACTIONS(2679),
    [anon_sym_fn] = ACTIONS(2679),
    [anon_sym_LPAREN2] = ACTIONS(2677),
    [anon_sym_LBRACK2] = ACTIONS(2677),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2677),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2677),
    [sym__not_in] = ACTIONS(2677),
    [sym__quoted_atom_start] = ACTIONS(2677),
  },
  [STATE(982)] = {
    [aux_sym__terminator_token1] = ACTIONS(2697),
    [anon_sym_SEMI] = ACTIONS(2699),
    [anon_sym_LPAREN] = ACTIONS(2699),
    [anon_sym_RPAREN] = ACTIONS(2699),
    [aux_sym_identifier_token1] = ACTIONS(2699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2699),
    [sym_alias] = ACTIONS(2699),
    [sym_integer] = ACTIONS(2699),
    [sym_float] = ACTIONS(2699),
    [sym_char] = ACTIONS(2699),
    [anon_sym_true] = ACTIONS(2699),
    [anon_sym_false] = ACTIONS(2699),
    [anon_sym_nil] = ACTIONS(2699),
    [sym_atom] = ACTIONS(2699),
    [anon_sym_DQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2699),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2699),
    [anon_sym_LBRACE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(2699),
    [anon_sym_LT] = ACTIONS(2699),
    [anon_sym_GT] = ACTIONS(2699),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_SLASH] = ACTIONS(2699),
    [anon_sym_TILDE] = ACTIONS(2699),
    [anon_sym_COMMA] = ACTIONS(2699),
    [sym_keyword] = ACTIONS(2699),
    [anon_sym_LT_LT] = ACTIONS(2699),
    [anon_sym_PERCENT] = ACTIONS(2699),
    [anon_sym_DOT_DOT] = ACTIONS(2699),
    [anon_sym_AMP] = ACTIONS(2699),
    [anon_sym_PLUS] = ACTIONS(2699),
    [anon_sym_DASH] = ACTIONS(2699),
    [anon_sym_BANG] = ACTIONS(2699),
    [anon_sym_CARET] = ACTIONS(2699),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2699),
    [anon_sym_not] = ACTIONS(2699),
    [anon_sym_AT] = ACTIONS(2699),
    [anon_sym_LT_DASH] = ACTIONS(2699),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2699),
    [anon_sym_when] = ACTIONS(2699),
    [anon_sym_COLON_COLON] = ACTIONS(2699),
    [anon_sym_EQ_GT] = ACTIONS(2699),
    [anon_sym_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_or] = ACTIONS(2699),
    [anon_sym_AMP_AMP] = ACTIONS(2699),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2699),
    [anon_sym_and] = ACTIONS(2699),
    [anon_sym_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ] = ACTIONS(2699),
    [anon_sym_EQ_TILDE] = ACTIONS(2699),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2699),
    [anon_sym_LT_EQ] = ACTIONS(2699),
    [anon_sym_GT_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_LT] = ACTIONS(2699),
    [anon_sym_GT_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2699),
    [anon_sym_in] = ACTIONS(2699),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2699),
    [anon_sym_SLASH_SLASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2699),
    [anon_sym_LT_GT] = ACTIONS(2699),
    [anon_sym_STAR] = ACTIONS(2699),
    [anon_sym_STAR_STAR] = ACTIONS(2699),
    [anon_sym_DASH_GT] = ACTIONS(2699),
    [anon_sym_DOT] = ACTIONS(2699),
    [anon_sym_do] = ACTIONS(2699),
    [anon_sym_fn] = ACTIONS(2699),
    [anon_sym_LPAREN2] = ACTIONS(2697),
    [anon_sym_LBRACK2] = ACTIONS(2697),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2697),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2697),
    [sym__not_in] = ACTIONS(2697),
    [sym__quoted_atom_start] = ACTIONS(2697),
  },
  [STATE(983)] = {
    [ts_builtin_sym_end] = ACTIONS(2693),
    [aux_sym__terminator_token1] = ACTIONS(2693),
    [anon_sym_SEMI] = ACTIONS(2695),
    [anon_sym_LPAREN] = ACTIONS(2695),
    [aux_sym_identifier_token1] = ACTIONS(2695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2695),
    [sym_alias] = ACTIONS(2695),
    [sym_integer] = ACTIONS(2695),
    [sym_float] = ACTIONS(2695),
    [sym_char] = ACTIONS(2695),
    [anon_sym_true] = ACTIONS(2695),
    [anon_sym_false] = ACTIONS(2695),
    [anon_sym_nil] = ACTIONS(2695),
    [sym_atom] = ACTIONS(2695),
    [anon_sym_DQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2695),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2695),
    [anon_sym_LBRACE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(2695),
    [anon_sym_LT] = ACTIONS(2695),
    [anon_sym_GT] = ACTIONS(2695),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_SLASH] = ACTIONS(2695),
    [anon_sym_TILDE] = ACTIONS(2695),
    [anon_sym_COMMA] = ACTIONS(2695),
    [sym_keyword] = ACTIONS(2695),
    [anon_sym_LT_LT] = ACTIONS(2695),
    [anon_sym_PERCENT] = ACTIONS(2695),
    [anon_sym_DOT_DOT] = ACTIONS(2695),
    [anon_sym_AMP] = ACTIONS(2695),
    [anon_sym_PLUS] = ACTIONS(2695),
    [anon_sym_DASH] = ACTIONS(2695),
    [anon_sym_BANG] = ACTIONS(2695),
    [anon_sym_CARET] = ACTIONS(2695),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2695),
    [anon_sym_not] = ACTIONS(2695),
    [anon_sym_AT] = ACTIONS(2695),
    [anon_sym_LT_DASH] = ACTIONS(2695),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2695),
    [anon_sym_when] = ACTIONS(2695),
    [anon_sym_COLON_COLON] = ACTIONS(2695),
    [anon_sym_EQ_GT] = ACTIONS(2695),
    [anon_sym_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_or] = ACTIONS(2695),
    [anon_sym_AMP_AMP] = ACTIONS(2695),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2695),
    [anon_sym_and] = ACTIONS(2695),
    [anon_sym_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ] = ACTIONS(2695),
    [anon_sym_EQ_TILDE] = ACTIONS(2695),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2695),
    [anon_sym_LT_EQ] = ACTIONS(2695),
    [anon_sym_GT_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_LT] = ACTIONS(2695),
    [anon_sym_GT_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2695),
    [anon_sym_in] = ACTIONS(2695),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2695),
    [anon_sym_SLASH_SLASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2695),
    [anon_sym_LT_GT] = ACTIONS(2695),
    [anon_sym_STAR] = ACTIONS(2695),
    [anon_sym_STAR_STAR] = ACTIONS(2695),
    [anon_sym_DASH_GT] = ACTIONS(2695),
    [anon_sym_DOT] = ACTIONS(2695),
    [anon_sym_do] = ACTIONS(2695),
    [anon_sym_fn] = ACTIONS(2695),
    [anon_sym_LPAREN2] = ACTIONS(2693),
    [anon_sym_LBRACK2] = ACTIONS(2693),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2693),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2693),
    [sym__not_in] = ACTIONS(2693),
    [sym__quoted_atom_start] = ACTIONS(2693),
  },
  [STATE(984)] = {
    [ts_builtin_sym_end] = ACTIONS(2677),
    [aux_sym__terminator_token1] = ACTIONS(2677),
    [anon_sym_SEMI] = ACTIONS(2679),
    [anon_sym_LPAREN] = ACTIONS(2679),
    [aux_sym_identifier_token1] = ACTIONS(2679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2679),
    [sym_alias] = ACTIONS(2679),
    [sym_integer] = ACTIONS(2679),
    [sym_float] = ACTIONS(2679),
    [sym_char] = ACTIONS(2679),
    [anon_sym_true] = ACTIONS(2679),
    [anon_sym_false] = ACTIONS(2679),
    [anon_sym_nil] = ACTIONS(2679),
    [sym_atom] = ACTIONS(2679),
    [anon_sym_DQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2679),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2679),
    [anon_sym_LBRACE] = ACTIONS(2679),
    [anon_sym_LBRACK] = ACTIONS(2679),
    [anon_sym_LT] = ACTIONS(2679),
    [anon_sym_GT] = ACTIONS(2679),
    [anon_sym_PIPE] = ACTIONS(2679),
    [anon_sym_SLASH] = ACTIONS(2679),
    [anon_sym_TILDE] = ACTIONS(2679),
    [anon_sym_COMMA] = ACTIONS(2679),
    [sym_keyword] = ACTIONS(2679),
    [anon_sym_LT_LT] = ACTIONS(2679),
    [anon_sym_PERCENT] = ACTIONS(2679),
    [anon_sym_DOT_DOT] = ACTIONS(2679),
    [anon_sym_AMP] = ACTIONS(2679),
    [anon_sym_PLUS] = ACTIONS(2679),
    [anon_sym_DASH] = ACTIONS(2679),
    [anon_sym_BANG] = ACTIONS(2679),
    [anon_sym_CARET] = ACTIONS(2679),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2679),
    [anon_sym_not] = ACTIONS(2679),
    [anon_sym_AT] = ACTIONS(2679),
    [anon_sym_LT_DASH] = ACTIONS(2679),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2679),
    [anon_sym_when] = ACTIONS(2679),
    [anon_sym_COLON_COLON] = ACTIONS(2679),
    [anon_sym_EQ_GT] = ACTIONS(2679),
    [anon_sym_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_or] = ACTIONS(2679),
    [anon_sym_AMP_AMP] = ACTIONS(2679),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2679),
    [anon_sym_and] = ACTIONS(2679),
    [anon_sym_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ] = ACTIONS(2679),
    [anon_sym_EQ_TILDE] = ACTIONS(2679),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2679),
    [anon_sym_LT_EQ] = ACTIONS(2679),
    [anon_sym_GT_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_LT] = ACTIONS(2679),
    [anon_sym_GT_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2679),
    [anon_sym_in] = ACTIONS(2679),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2679),
    [anon_sym_SLASH_SLASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2679),
    [anon_sym_LT_GT] = ACTIONS(2679),
    [anon_sym_STAR] = ACTIONS(2679),
    [anon_sym_STAR_STAR] = ACTIONS(2679),
    [anon_sym_DASH_GT] = ACTIONS(2679),
    [anon_sym_DOT] = ACTIONS(2679),
    [anon_sym_do] = ACTIONS(2679),
    [anon_sym_fn] = ACTIONS(2679),
    [anon_sym_LPAREN2] = ACTIONS(2677),
    [anon_sym_LBRACK2] = ACTIONS(2677),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2677),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2677),
    [sym__not_in] = ACTIONS(2677),
    [sym__quoted_atom_start] = ACTIONS(2677),
  },
  [STATE(985)] = {
    [ts_builtin_sym_end] = ACTIONS(2697),
    [aux_sym__terminator_token1] = ACTIONS(2697),
    [anon_sym_SEMI] = ACTIONS(2699),
    [anon_sym_LPAREN] = ACTIONS(2699),
    [aux_sym_identifier_token1] = ACTIONS(2699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2699),
    [sym_alias] = ACTIONS(2699),
    [sym_integer] = ACTIONS(2699),
    [sym_float] = ACTIONS(2699),
    [sym_char] = ACTIONS(2699),
    [anon_sym_true] = ACTIONS(2699),
    [anon_sym_false] = ACTIONS(2699),
    [anon_sym_nil] = ACTIONS(2699),
    [sym_atom] = ACTIONS(2699),
    [anon_sym_DQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2699),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2699),
    [anon_sym_LBRACE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(2699),
    [anon_sym_LT] = ACTIONS(2699),
    [anon_sym_GT] = ACTIONS(2699),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_SLASH] = ACTIONS(2699),
    [anon_sym_TILDE] = ACTIONS(2699),
    [anon_sym_COMMA] = ACTIONS(2699),
    [sym_keyword] = ACTIONS(2699),
    [anon_sym_LT_LT] = ACTIONS(2699),
    [anon_sym_PERCENT] = ACTIONS(2699),
    [anon_sym_DOT_DOT] = ACTIONS(2699),
    [anon_sym_AMP] = ACTIONS(2699),
    [anon_sym_PLUS] = ACTIONS(2699),
    [anon_sym_DASH] = ACTIONS(2699),
    [anon_sym_BANG] = ACTIONS(2699),
    [anon_sym_CARET] = ACTIONS(2699),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2699),
    [anon_sym_not] = ACTIONS(2699),
    [anon_sym_AT] = ACTIONS(2699),
    [anon_sym_LT_DASH] = ACTIONS(2699),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2699),
    [anon_sym_when] = ACTIONS(2699),
    [anon_sym_COLON_COLON] = ACTIONS(2699),
    [anon_sym_EQ_GT] = ACTIONS(2699),
    [anon_sym_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_or] = ACTIONS(2699),
    [anon_sym_AMP_AMP] = ACTIONS(2699),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2699),
    [anon_sym_and] = ACTIONS(2699),
    [anon_sym_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ] = ACTIONS(2699),
    [anon_sym_EQ_TILDE] = ACTIONS(2699),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2699),
    [anon_sym_LT_EQ] = ACTIONS(2699),
    [anon_sym_GT_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_LT] = ACTIONS(2699),
    [anon_sym_GT_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2699),
    [anon_sym_in] = ACTIONS(2699),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2699),
    [anon_sym_SLASH_SLASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2699),
    [anon_sym_LT_GT] = ACTIONS(2699),
    [anon_sym_STAR] = ACTIONS(2699),
    [anon_sym_STAR_STAR] = ACTIONS(2699),
    [anon_sym_DASH_GT] = ACTIONS(2699),
    [anon_sym_DOT] = ACTIONS(2699),
    [anon_sym_do] = ACTIONS(2699),
    [anon_sym_fn] = ACTIONS(2699),
    [anon_sym_LPAREN2] = ACTIONS(2697),
    [anon_sym_LBRACK2] = ACTIONS(2697),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2697),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2697),
    [sym__not_in] = ACTIONS(2697),
    [sym__quoted_atom_start] = ACTIONS(2697),
  },
  [STATE(986)] = {
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [anon_sym_RPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(987)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [anon_sym_RPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_RBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_RBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(988)] = {
    [ts_builtin_sym_end] = ACTIONS(2681),
    [aux_sym__terminator_token1] = ACTIONS(2681),
    [anon_sym_SEMI] = ACTIONS(2683),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2681),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(989)] = {
    [ts_builtin_sym_end] = ACTIONS(2673),
    [aux_sym__terminator_token1] = ACTIONS(2673),
    [anon_sym_SEMI] = ACTIONS(2675),
    [anon_sym_LPAREN] = ACTIONS(2675),
    [aux_sym_identifier_token1] = ACTIONS(2675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2675),
    [sym_alias] = ACTIONS(2675),
    [sym_integer] = ACTIONS(2675),
    [sym_float] = ACTIONS(2675),
    [sym_char] = ACTIONS(2675),
    [anon_sym_true] = ACTIONS(2675),
    [anon_sym_false] = ACTIONS(2675),
    [anon_sym_nil] = ACTIONS(2675),
    [sym_atom] = ACTIONS(2675),
    [anon_sym_DQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2675),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2675),
    [anon_sym_LBRACE] = ACTIONS(2675),
    [anon_sym_LBRACK] = ACTIONS(2675),
    [anon_sym_LT] = ACTIONS(2675),
    [anon_sym_GT] = ACTIONS(2675),
    [anon_sym_PIPE] = ACTIONS(2675),
    [anon_sym_SLASH] = ACTIONS(2675),
    [anon_sym_TILDE] = ACTIONS(2675),
    [anon_sym_COMMA] = ACTIONS(2675),
    [sym_keyword] = ACTIONS(2675),
    [anon_sym_LT_LT] = ACTIONS(2675),
    [anon_sym_PERCENT] = ACTIONS(2675),
    [anon_sym_DOT_DOT] = ACTIONS(2675),
    [anon_sym_AMP] = ACTIONS(2675),
    [anon_sym_PLUS] = ACTIONS(2675),
    [anon_sym_DASH] = ACTIONS(2675),
    [anon_sym_BANG] = ACTIONS(2675),
    [anon_sym_CARET] = ACTIONS(2675),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2675),
    [anon_sym_not] = ACTIONS(2675),
    [anon_sym_AT] = ACTIONS(2675),
    [anon_sym_LT_DASH] = ACTIONS(2675),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2675),
    [anon_sym_when] = ACTIONS(2675),
    [anon_sym_COLON_COLON] = ACTIONS(2675),
    [anon_sym_EQ_GT] = ACTIONS(2675),
    [anon_sym_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_or] = ACTIONS(2675),
    [anon_sym_AMP_AMP] = ACTIONS(2675),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2675),
    [anon_sym_and] = ACTIONS(2675),
    [anon_sym_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ] = ACTIONS(2675),
    [anon_sym_EQ_TILDE] = ACTIONS(2675),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2675),
    [anon_sym_LT_EQ] = ACTIONS(2675),
    [anon_sym_GT_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_LT] = ACTIONS(2675),
    [anon_sym_GT_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2675),
    [anon_sym_in] = ACTIONS(2675),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2675),
    [anon_sym_SLASH_SLASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2675),
    [anon_sym_LT_GT] = ACTIONS(2675),
    [anon_sym_STAR] = ACTIONS(2675),
    [anon_sym_STAR_STAR] = ACTIONS(2675),
    [anon_sym_DASH_GT] = ACTIONS(2675),
    [anon_sym_DOT] = ACTIONS(2675),
    [anon_sym_do] = ACTIONS(2675),
    [anon_sym_fn] = ACTIONS(2675),
    [anon_sym_LPAREN2] = ACTIONS(2673),
    [anon_sym_LBRACK2] = ACTIONS(2673),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2673),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2673),
    [sym__not_in] = ACTIONS(2673),
    [sym__quoted_atom_start] = ACTIONS(2673),
  },
  [STATE(990)] = {
    [aux_sym__terminator_token1] = ACTIONS(2701),
    [anon_sym_SEMI] = ACTIONS(2703),
    [anon_sym_LPAREN] = ACTIONS(2703),
    [anon_sym_RPAREN] = ACTIONS(2703),
    [aux_sym_identifier_token1] = ACTIONS(2703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2703),
    [sym_alias] = ACTIONS(2703),
    [sym_integer] = ACTIONS(2703),
    [sym_float] = ACTIONS(2703),
    [sym_char] = ACTIONS(2703),
    [anon_sym_true] = ACTIONS(2703),
    [anon_sym_false] = ACTIONS(2703),
    [anon_sym_nil] = ACTIONS(2703),
    [sym_atom] = ACTIONS(2703),
    [anon_sym_DQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2703),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2703),
    [anon_sym_LBRACE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(2703),
    [anon_sym_LT] = ACTIONS(2703),
    [anon_sym_GT] = ACTIONS(2703),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_SLASH] = ACTIONS(2703),
    [anon_sym_TILDE] = ACTIONS(2703),
    [anon_sym_COMMA] = ACTIONS(2703),
    [sym_keyword] = ACTIONS(2703),
    [anon_sym_LT_LT] = ACTIONS(2703),
    [anon_sym_PERCENT] = ACTIONS(2703),
    [anon_sym_DOT_DOT] = ACTIONS(2703),
    [anon_sym_AMP] = ACTIONS(2703),
    [anon_sym_PLUS] = ACTIONS(2703),
    [anon_sym_DASH] = ACTIONS(2703),
    [anon_sym_BANG] = ACTIONS(2703),
    [anon_sym_CARET] = ACTIONS(2703),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2703),
    [anon_sym_not] = ACTIONS(2703),
    [anon_sym_AT] = ACTIONS(2703),
    [anon_sym_LT_DASH] = ACTIONS(2703),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2703),
    [anon_sym_when] = ACTIONS(2703),
    [anon_sym_COLON_COLON] = ACTIONS(2703),
    [anon_sym_EQ_GT] = ACTIONS(2703),
    [anon_sym_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_or] = ACTIONS(2703),
    [anon_sym_AMP_AMP] = ACTIONS(2703),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2703),
    [anon_sym_and] = ACTIONS(2703),
    [anon_sym_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ] = ACTIONS(2703),
    [anon_sym_EQ_TILDE] = ACTIONS(2703),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2703),
    [anon_sym_LT_EQ] = ACTIONS(2703),
    [anon_sym_GT_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_LT] = ACTIONS(2703),
    [anon_sym_GT_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2703),
    [anon_sym_in] = ACTIONS(2703),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2703),
    [anon_sym_SLASH_SLASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2703),
    [anon_sym_LT_GT] = ACTIONS(2703),
    [anon_sym_STAR] = ACTIONS(2703),
    [anon_sym_STAR_STAR] = ACTIONS(2703),
    [anon_sym_DASH_GT] = ACTIONS(2703),
    [anon_sym_DOT] = ACTIONS(2703),
    [anon_sym_do] = ACTIONS(2703),
    [anon_sym_fn] = ACTIONS(2703),
    [anon_sym_LPAREN2] = ACTIONS(2701),
    [anon_sym_LBRACK2] = ACTIONS(2701),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2701),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2701),
    [sym__not_in] = ACTIONS(2701),
    [sym__quoted_atom_start] = ACTIONS(2701),
  },
  [STATE(991)] = {
    [aux_sym__terminator_token1] = ACTIONS(2705),
    [anon_sym_SEMI] = ACTIONS(2707),
    [anon_sym_LPAREN] = ACTIONS(2707),
    [anon_sym_RPAREN] = ACTIONS(2707),
    [aux_sym_identifier_token1] = ACTIONS(2707),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2707),
    [sym_alias] = ACTIONS(2707),
    [sym_integer] = ACTIONS(2707),
    [sym_float] = ACTIONS(2707),
    [sym_char] = ACTIONS(2707),
    [anon_sym_true] = ACTIONS(2707),
    [anon_sym_false] = ACTIONS(2707),
    [anon_sym_nil] = ACTIONS(2707),
    [sym_atom] = ACTIONS(2707),
    [anon_sym_DQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2707),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2707),
    [anon_sym_LBRACE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(2707),
    [anon_sym_LT] = ACTIONS(2707),
    [anon_sym_GT] = ACTIONS(2707),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_SLASH] = ACTIONS(2707),
    [anon_sym_TILDE] = ACTIONS(2707),
    [anon_sym_COMMA] = ACTIONS(2707),
    [sym_keyword] = ACTIONS(2707),
    [anon_sym_LT_LT] = ACTIONS(2707),
    [anon_sym_PERCENT] = ACTIONS(2707),
    [anon_sym_DOT_DOT] = ACTIONS(2707),
    [anon_sym_AMP] = ACTIONS(2707),
    [anon_sym_PLUS] = ACTIONS(2707),
    [anon_sym_DASH] = ACTIONS(2707),
    [anon_sym_BANG] = ACTIONS(2707),
    [anon_sym_CARET] = ACTIONS(2707),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2707),
    [anon_sym_not] = ACTIONS(2707),
    [anon_sym_AT] = ACTIONS(2707),
    [anon_sym_LT_DASH] = ACTIONS(2707),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2707),
    [anon_sym_when] = ACTIONS(2707),
    [anon_sym_COLON_COLON] = ACTIONS(2707),
    [anon_sym_EQ_GT] = ACTIONS(2707),
    [anon_sym_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_or] = ACTIONS(2707),
    [anon_sym_AMP_AMP] = ACTIONS(2707),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2707),
    [anon_sym_and] = ACTIONS(2707),
    [anon_sym_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ] = ACTIONS(2707),
    [anon_sym_EQ_TILDE] = ACTIONS(2707),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2707),
    [anon_sym_LT_EQ] = ACTIONS(2707),
    [anon_sym_GT_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_LT] = ACTIONS(2707),
    [anon_sym_GT_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2707),
    [anon_sym_in] = ACTIONS(2707),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2707),
    [anon_sym_SLASH_SLASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2707),
    [anon_sym_LT_GT] = ACTIONS(2707),
    [anon_sym_STAR] = ACTIONS(2707),
    [anon_sym_STAR_STAR] = ACTIONS(2707),
    [anon_sym_DASH_GT] = ACTIONS(2707),
    [anon_sym_DOT] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2707),
    [anon_sym_fn] = ACTIONS(2707),
    [anon_sym_LPAREN2] = ACTIONS(2705),
    [anon_sym_LBRACK2] = ACTIONS(2705),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2705),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2705),
    [sym__not_in] = ACTIONS(2705),
    [sym__quoted_atom_start] = ACTIONS(2705),
  },
  [STATE(992)] = {
    [aux_sym__terminator_token1] = ACTIONS(2709),
    [anon_sym_SEMI] = ACTIONS(2711),
    [anon_sym_LPAREN] = ACTIONS(2711),
    [anon_sym_RPAREN] = ACTIONS(2711),
    [aux_sym_identifier_token1] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [sym_alias] = ACTIONS(2711),
    [sym_integer] = ACTIONS(2711),
    [sym_float] = ACTIONS(2711),
    [sym_char] = ACTIONS(2711),
    [anon_sym_true] = ACTIONS(2711),
    [anon_sym_false] = ACTIONS(2711),
    [anon_sym_nil] = ACTIONS(2711),
    [sym_atom] = ACTIONS(2711),
    [anon_sym_DQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2711),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2711),
    [anon_sym_LBRACE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2711),
    [anon_sym_GT] = ACTIONS(2711),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_SLASH] = ACTIONS(2711),
    [anon_sym_TILDE] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [sym_keyword] = ACTIONS(2711),
    [anon_sym_LT_LT] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2711),
    [anon_sym_AMP] = ACTIONS(2711),
    [anon_sym_PLUS] = ACTIONS(2711),
    [anon_sym_DASH] = ACTIONS(2711),
    [anon_sym_BANG] = ACTIONS(2711),
    [anon_sym_CARET] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_not] = ACTIONS(2711),
    [anon_sym_AT] = ACTIONS(2711),
    [anon_sym_LT_DASH] = ACTIONS(2711),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2711),
    [anon_sym_when] = ACTIONS(2711),
    [anon_sym_COLON_COLON] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2711),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2711),
    [anon_sym_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_LT] = ACTIONS(2711),
    [anon_sym_GT_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2711),
    [anon_sym_in] = ACTIONS(2711),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2711),
    [anon_sym_LT_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2711),
    [anon_sym_STAR_STAR] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_DOT] = ACTIONS(2711),
    [anon_sym_do] = ACTIONS(2711),
    [anon_sym_fn] = ACTIONS(2711),
    [anon_sym_LPAREN2] = ACTIONS(2709),
    [anon_sym_LBRACK2] = ACTIONS(2709),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2709),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2709),
    [sym__not_in] = ACTIONS(2709),
    [sym__quoted_atom_start] = ACTIONS(2709),
  },
  [STATE(993)] = {
    [aux_sym__terminator_token1] = ACTIONS(2665),
    [anon_sym_SEMI] = ACTIONS(2667),
    [anon_sym_LPAREN] = ACTIONS(2667),
    [anon_sym_RPAREN] = ACTIONS(2667),
    [aux_sym_identifier_token1] = ACTIONS(2667),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2667),
    [sym_alias] = ACTIONS(2667),
    [sym_integer] = ACTIONS(2667),
    [sym_float] = ACTIONS(2667),
    [sym_char] = ACTIONS(2667),
    [anon_sym_true] = ACTIONS(2667),
    [anon_sym_false] = ACTIONS(2667),
    [anon_sym_nil] = ACTIONS(2667),
    [sym_atom] = ACTIONS(2667),
    [anon_sym_DQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2667),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2667),
    [anon_sym_LBRACE] = ACTIONS(2667),
    [anon_sym_LBRACK] = ACTIONS(2667),
    [anon_sym_LT] = ACTIONS(2667),
    [anon_sym_GT] = ACTIONS(2667),
    [anon_sym_PIPE] = ACTIONS(2667),
    [anon_sym_SLASH] = ACTIONS(2667),
    [anon_sym_TILDE] = ACTIONS(2667),
    [anon_sym_COMMA] = ACTIONS(2667),
    [sym_keyword] = ACTIONS(2667),
    [anon_sym_LT_LT] = ACTIONS(2667),
    [anon_sym_PERCENT] = ACTIONS(2667),
    [anon_sym_DOT_DOT] = ACTIONS(2667),
    [anon_sym_AMP] = ACTIONS(2667),
    [anon_sym_PLUS] = ACTIONS(2667),
    [anon_sym_DASH] = ACTIONS(2667),
    [anon_sym_BANG] = ACTIONS(2667),
    [anon_sym_CARET] = ACTIONS(2667),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2667),
    [anon_sym_not] = ACTIONS(2667),
    [anon_sym_AT] = ACTIONS(2667),
    [anon_sym_LT_DASH] = ACTIONS(2667),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2667),
    [anon_sym_when] = ACTIONS(2667),
    [anon_sym_COLON_COLON] = ACTIONS(2667),
    [anon_sym_EQ_GT] = ACTIONS(2667),
    [anon_sym_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_or] = ACTIONS(2667),
    [anon_sym_AMP_AMP] = ACTIONS(2667),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2667),
    [anon_sym_and] = ACTIONS(2667),
    [anon_sym_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ] = ACTIONS(2667),
    [anon_sym_EQ_TILDE] = ACTIONS(2667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2667),
    [anon_sym_LT_EQ] = ACTIONS(2667),
    [anon_sym_GT_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_LT] = ACTIONS(2667),
    [anon_sym_GT_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2667),
    [anon_sym_in] = ACTIONS(2667),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2667),
    [anon_sym_SLASH_SLASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2667),
    [anon_sym_LT_GT] = ACTIONS(2667),
    [anon_sym_STAR] = ACTIONS(2667),
    [anon_sym_STAR_STAR] = ACTIONS(2667),
    [anon_sym_DASH_GT] = ACTIONS(2667),
    [anon_sym_DOT] = ACTIONS(2667),
    [anon_sym_do] = ACTIONS(2667),
    [anon_sym_fn] = ACTIONS(2667),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_LBRACK2] = ACTIONS(2665),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2665),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2665),
    [sym__not_in] = ACTIONS(2665),
    [sym__quoted_atom_start] = ACTIONS(2665),
  },
  [STATE(994)] = {
    [aux_sym__terminator_token1] = ACTIONS(2669),
    [anon_sym_SEMI] = ACTIONS(2671),
    [anon_sym_LPAREN] = ACTIONS(2671),
    [anon_sym_RPAREN] = ACTIONS(2671),
    [aux_sym_identifier_token1] = ACTIONS(2671),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2671),
    [sym_alias] = ACTIONS(2671),
    [sym_integer] = ACTIONS(2671),
    [sym_float] = ACTIONS(2671),
    [sym_char] = ACTIONS(2671),
    [anon_sym_true] = ACTIONS(2671),
    [anon_sym_false] = ACTIONS(2671),
    [anon_sym_nil] = ACTIONS(2671),
    [sym_atom] = ACTIONS(2671),
    [anon_sym_DQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2671),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2671),
    [anon_sym_LBRACE] = ACTIONS(2671),
    [anon_sym_LBRACK] = ACTIONS(2671),
    [anon_sym_LT] = ACTIONS(2671),
    [anon_sym_GT] = ACTIONS(2671),
    [anon_sym_PIPE] = ACTIONS(2671),
    [anon_sym_SLASH] = ACTIONS(2671),
    [anon_sym_TILDE] = ACTIONS(2671),
    [anon_sym_COMMA] = ACTIONS(2671),
    [sym_keyword] = ACTIONS(2671),
    [anon_sym_LT_LT] = ACTIONS(2671),
    [anon_sym_PERCENT] = ACTIONS(2671),
    [anon_sym_DOT_DOT] = ACTIONS(2671),
    [anon_sym_AMP] = ACTIONS(2671),
    [anon_sym_PLUS] = ACTIONS(2671),
    [anon_sym_DASH] = ACTIONS(2671),
    [anon_sym_BANG] = ACTIONS(2671),
    [anon_sym_CARET] = ACTIONS(2671),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2671),
    [anon_sym_not] = ACTIONS(2671),
    [anon_sym_AT] = ACTIONS(2671),
    [anon_sym_LT_DASH] = ACTIONS(2671),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2671),
    [anon_sym_when] = ACTIONS(2671),
    [anon_sym_COLON_COLON] = ACTIONS(2671),
    [anon_sym_EQ_GT] = ACTIONS(2671),
    [anon_sym_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_or] = ACTIONS(2671),
    [anon_sym_AMP_AMP] = ACTIONS(2671),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2671),
    [anon_sym_and] = ACTIONS(2671),
    [anon_sym_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ] = ACTIONS(2671),
    [anon_sym_EQ_TILDE] = ACTIONS(2671),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2671),
    [anon_sym_LT_EQ] = ACTIONS(2671),
    [anon_sym_GT_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_LT] = ACTIONS(2671),
    [anon_sym_GT_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2671),
    [anon_sym_in] = ACTIONS(2671),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2671),
    [anon_sym_SLASH_SLASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2671),
    [anon_sym_LT_GT] = ACTIONS(2671),
    [anon_sym_STAR] = ACTIONS(2671),
    [anon_sym_STAR_STAR] = ACTIONS(2671),
    [anon_sym_DASH_GT] = ACTIONS(2671),
    [anon_sym_DOT] = ACTIONS(2671),
    [anon_sym_do] = ACTIONS(2671),
    [anon_sym_fn] = ACTIONS(2671),
    [anon_sym_LPAREN2] = ACTIONS(2669),
    [anon_sym_LBRACK2] = ACTIONS(2669),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2669),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2669),
    [sym__not_in] = ACTIONS(2669),
    [sym__quoted_atom_start] = ACTIONS(2669),
  },
  [STATE(995)] = {
    [aux_sym__terminator_token1] = ACTIONS(2673),
    [anon_sym_SEMI] = ACTIONS(2675),
    [anon_sym_LPAREN] = ACTIONS(2675),
    [anon_sym_RPAREN] = ACTIONS(2675),
    [aux_sym_identifier_token1] = ACTIONS(2675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2675),
    [sym_alias] = ACTIONS(2675),
    [sym_integer] = ACTIONS(2675),
    [sym_float] = ACTIONS(2675),
    [sym_char] = ACTIONS(2675),
    [anon_sym_true] = ACTIONS(2675),
    [anon_sym_false] = ACTIONS(2675),
    [anon_sym_nil] = ACTIONS(2675),
    [sym_atom] = ACTIONS(2675),
    [anon_sym_DQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2675),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2675),
    [anon_sym_LBRACE] = ACTIONS(2675),
    [anon_sym_LBRACK] = ACTIONS(2675),
    [anon_sym_LT] = ACTIONS(2675),
    [anon_sym_GT] = ACTIONS(2675),
    [anon_sym_PIPE] = ACTIONS(2675),
    [anon_sym_SLASH] = ACTIONS(2675),
    [anon_sym_TILDE] = ACTIONS(2675),
    [anon_sym_COMMA] = ACTIONS(2675),
    [sym_keyword] = ACTIONS(2675),
    [anon_sym_LT_LT] = ACTIONS(2675),
    [anon_sym_PERCENT] = ACTIONS(2675),
    [anon_sym_DOT_DOT] = ACTIONS(2675),
    [anon_sym_AMP] = ACTIONS(2675),
    [anon_sym_PLUS] = ACTIONS(2675),
    [anon_sym_DASH] = ACTIONS(2675),
    [anon_sym_BANG] = ACTIONS(2675),
    [anon_sym_CARET] = ACTIONS(2675),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2675),
    [anon_sym_not] = ACTIONS(2675),
    [anon_sym_AT] = ACTIONS(2675),
    [anon_sym_LT_DASH] = ACTIONS(2675),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2675),
    [anon_sym_when] = ACTIONS(2675),
    [anon_sym_COLON_COLON] = ACTIONS(2675),
    [anon_sym_EQ_GT] = ACTIONS(2675),
    [anon_sym_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_or] = ACTIONS(2675),
    [anon_sym_AMP_AMP] = ACTIONS(2675),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2675),
    [anon_sym_and] = ACTIONS(2675),
    [anon_sym_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ] = ACTIONS(2675),
    [anon_sym_EQ_TILDE] = ACTIONS(2675),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2675),
    [anon_sym_LT_EQ] = ACTIONS(2675),
    [anon_sym_GT_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_LT] = ACTIONS(2675),
    [anon_sym_GT_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2675),
    [anon_sym_in] = ACTIONS(2675),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2675),
    [anon_sym_SLASH_SLASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2675),
    [anon_sym_LT_GT] = ACTIONS(2675),
    [anon_sym_STAR] = ACTIONS(2675),
    [anon_sym_STAR_STAR] = ACTIONS(2675),
    [anon_sym_DASH_GT] = ACTIONS(2675),
    [anon_sym_DOT] = ACTIONS(2675),
    [anon_sym_do] = ACTIONS(2675),
    [anon_sym_fn] = ACTIONS(2675),
    [anon_sym_LPAREN2] = ACTIONS(2673),
    [anon_sym_LBRACK2] = ACTIONS(2673),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2673),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2673),
    [sym__not_in] = ACTIONS(2673),
    [sym__quoted_atom_start] = ACTIONS(2673),
  },
  [STATE(996)] = {
    [ts_builtin_sym_end] = ACTIONS(2685),
    [aux_sym__terminator_token1] = ACTIONS(2685),
    [anon_sym_SEMI] = ACTIONS(2687),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(997)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2675),
    [aux_sym_identifier_token1] = ACTIONS(2675),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2675),
    [sym_alias] = ACTIONS(2675),
    [sym_integer] = ACTIONS(2675),
    [sym_float] = ACTIONS(2675),
    [sym_char] = ACTIONS(2675),
    [anon_sym_true] = ACTIONS(2675),
    [anon_sym_false] = ACTIONS(2675),
    [anon_sym_nil] = ACTIONS(2675),
    [sym_atom] = ACTIONS(2675),
    [anon_sym_DQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE] = ACTIONS(2675),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2675),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2675),
    [anon_sym_LBRACE] = ACTIONS(2675),
    [anon_sym_LBRACK] = ACTIONS(2675),
    [anon_sym_LT] = ACTIONS(2675),
    [anon_sym_GT] = ACTIONS(2675),
    [anon_sym_PIPE] = ACTIONS(2675),
    [anon_sym_SLASH] = ACTIONS(2675),
    [anon_sym_TILDE] = ACTIONS(2675),
    [anon_sym_COMMA] = ACTIONS(2675),
    [sym_keyword] = ACTIONS(2675),
    [anon_sym_LT_LT] = ACTIONS(2675),
    [anon_sym_GT_GT] = ACTIONS(2675),
    [anon_sym_PERCENT] = ACTIONS(2675),
    [anon_sym_DOT_DOT] = ACTIONS(2675),
    [anon_sym_AMP] = ACTIONS(2675),
    [anon_sym_PLUS] = ACTIONS(2675),
    [anon_sym_DASH] = ACTIONS(2675),
    [anon_sym_BANG] = ACTIONS(2675),
    [anon_sym_CARET] = ACTIONS(2675),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2675),
    [anon_sym_not] = ACTIONS(2675),
    [anon_sym_AT] = ACTIONS(2675),
    [anon_sym_LT_DASH] = ACTIONS(2675),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2675),
    [anon_sym_when] = ACTIONS(2675),
    [anon_sym_COLON_COLON] = ACTIONS(2675),
    [anon_sym_EQ_GT] = ACTIONS(2675),
    [anon_sym_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2675),
    [anon_sym_or] = ACTIONS(2675),
    [anon_sym_AMP_AMP] = ACTIONS(2675),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2675),
    [anon_sym_and] = ACTIONS(2675),
    [anon_sym_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ] = ACTIONS(2675),
    [anon_sym_EQ_TILDE] = ACTIONS(2675),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2675),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2675),
    [anon_sym_LT_EQ] = ACTIONS(2675),
    [anon_sym_GT_EQ] = ACTIONS(2675),
    [anon_sym_PIPE_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_LT] = ACTIONS(2675),
    [anon_sym_GT_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE] = ACTIONS(2675),
    [anon_sym_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2675),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2675),
    [anon_sym_in] = ACTIONS(2675),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2675),
    [anon_sym_SLASH_SLASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH] = ACTIONS(2675),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2675),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2675),
    [anon_sym_LT_GT] = ACTIONS(2675),
    [anon_sym_STAR] = ACTIONS(2675),
    [anon_sym_STAR_STAR] = ACTIONS(2675),
    [anon_sym_DASH_GT] = ACTIONS(2675),
    [anon_sym_DOT] = ACTIONS(2675),
    [anon_sym_do] = ACTIONS(2675),
    [anon_sym_fn] = ACTIONS(2675),
    [anon_sym_LPAREN2] = ACTIONS(2673),
    [anon_sym_LBRACK2] = ACTIONS(2673),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2673),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2673),
    [sym__not_in] = ACTIONS(2673),
    [sym__quoted_atom_start] = ACTIONS(2673),
  },
  [STATE(998)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(999)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(1000)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(1001)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2681),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(1002)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(1003)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2703),
    [aux_sym_identifier_token1] = ACTIONS(2703),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2703),
    [sym_alias] = ACTIONS(2703),
    [sym_integer] = ACTIONS(2703),
    [sym_float] = ACTIONS(2703),
    [sym_char] = ACTIONS(2703),
    [anon_sym_true] = ACTIONS(2703),
    [anon_sym_false] = ACTIONS(2703),
    [anon_sym_nil] = ACTIONS(2703),
    [sym_atom] = ACTIONS(2703),
    [anon_sym_DQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE] = ACTIONS(2703),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2703),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2703),
    [anon_sym_LBRACE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(2703),
    [anon_sym_LT] = ACTIONS(2703),
    [anon_sym_GT] = ACTIONS(2703),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_SLASH] = ACTIONS(2703),
    [anon_sym_TILDE] = ACTIONS(2703),
    [anon_sym_COMMA] = ACTIONS(2703),
    [sym_keyword] = ACTIONS(2703),
    [anon_sym_LT_LT] = ACTIONS(2703),
    [anon_sym_GT_GT] = ACTIONS(2703),
    [anon_sym_PERCENT] = ACTIONS(2703),
    [anon_sym_DOT_DOT] = ACTIONS(2703),
    [anon_sym_AMP] = ACTIONS(2703),
    [anon_sym_PLUS] = ACTIONS(2703),
    [anon_sym_DASH] = ACTIONS(2703),
    [anon_sym_BANG] = ACTIONS(2703),
    [anon_sym_CARET] = ACTIONS(2703),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2703),
    [anon_sym_not] = ACTIONS(2703),
    [anon_sym_AT] = ACTIONS(2703),
    [anon_sym_LT_DASH] = ACTIONS(2703),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2703),
    [anon_sym_when] = ACTIONS(2703),
    [anon_sym_COLON_COLON] = ACTIONS(2703),
    [anon_sym_EQ_GT] = ACTIONS(2703),
    [anon_sym_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2703),
    [anon_sym_or] = ACTIONS(2703),
    [anon_sym_AMP_AMP] = ACTIONS(2703),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2703),
    [anon_sym_and] = ACTIONS(2703),
    [anon_sym_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ] = ACTIONS(2703),
    [anon_sym_EQ_TILDE] = ACTIONS(2703),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2703),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2703),
    [anon_sym_LT_EQ] = ACTIONS(2703),
    [anon_sym_GT_EQ] = ACTIONS(2703),
    [anon_sym_PIPE_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_LT] = ACTIONS(2703),
    [anon_sym_GT_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE] = ACTIONS(2703),
    [anon_sym_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2703),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2703),
    [anon_sym_in] = ACTIONS(2703),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2703),
    [anon_sym_SLASH_SLASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH] = ACTIONS(2703),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2703),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2703),
    [anon_sym_LT_GT] = ACTIONS(2703),
    [anon_sym_STAR] = ACTIONS(2703),
    [anon_sym_STAR_STAR] = ACTIONS(2703),
    [anon_sym_DASH_GT] = ACTIONS(2703),
    [anon_sym_DOT] = ACTIONS(2703),
    [anon_sym_do] = ACTIONS(2703),
    [anon_sym_fn] = ACTIONS(2703),
    [anon_sym_LPAREN2] = ACTIONS(2701),
    [anon_sym_LBRACK2] = ACTIONS(2701),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2701),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2701),
    [sym__not_in] = ACTIONS(2701),
    [sym__quoted_atom_start] = ACTIONS(2701),
  },
  [STATE(1004)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(1005)] = {
    [aux_sym__terminator_token1] = ACTIONS(2681),
    [anon_sym_SEMI] = ACTIONS(2683),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_end] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(1006)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2707),
    [aux_sym_identifier_token1] = ACTIONS(2707),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2707),
    [sym_alias] = ACTIONS(2707),
    [sym_integer] = ACTIONS(2707),
    [sym_float] = ACTIONS(2707),
    [sym_char] = ACTIONS(2707),
    [anon_sym_true] = ACTIONS(2707),
    [anon_sym_false] = ACTIONS(2707),
    [anon_sym_nil] = ACTIONS(2707),
    [sym_atom] = ACTIONS(2707),
    [anon_sym_DQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE] = ACTIONS(2707),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2707),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2707),
    [anon_sym_LBRACE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(2707),
    [anon_sym_LT] = ACTIONS(2707),
    [anon_sym_GT] = ACTIONS(2707),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_SLASH] = ACTIONS(2707),
    [anon_sym_TILDE] = ACTIONS(2707),
    [anon_sym_COMMA] = ACTIONS(2707),
    [sym_keyword] = ACTIONS(2707),
    [anon_sym_LT_LT] = ACTIONS(2707),
    [anon_sym_GT_GT] = ACTIONS(2707),
    [anon_sym_PERCENT] = ACTIONS(2707),
    [anon_sym_DOT_DOT] = ACTIONS(2707),
    [anon_sym_AMP] = ACTIONS(2707),
    [anon_sym_PLUS] = ACTIONS(2707),
    [anon_sym_DASH] = ACTIONS(2707),
    [anon_sym_BANG] = ACTIONS(2707),
    [anon_sym_CARET] = ACTIONS(2707),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2707),
    [anon_sym_not] = ACTIONS(2707),
    [anon_sym_AT] = ACTIONS(2707),
    [anon_sym_LT_DASH] = ACTIONS(2707),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2707),
    [anon_sym_when] = ACTIONS(2707),
    [anon_sym_COLON_COLON] = ACTIONS(2707),
    [anon_sym_EQ_GT] = ACTIONS(2707),
    [anon_sym_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2707),
    [anon_sym_or] = ACTIONS(2707),
    [anon_sym_AMP_AMP] = ACTIONS(2707),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2707),
    [anon_sym_and] = ACTIONS(2707),
    [anon_sym_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ] = ACTIONS(2707),
    [anon_sym_EQ_TILDE] = ACTIONS(2707),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2707),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2707),
    [anon_sym_LT_EQ] = ACTIONS(2707),
    [anon_sym_GT_EQ] = ACTIONS(2707),
    [anon_sym_PIPE_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_LT] = ACTIONS(2707),
    [anon_sym_GT_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE] = ACTIONS(2707),
    [anon_sym_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2707),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2707),
    [anon_sym_in] = ACTIONS(2707),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2707),
    [anon_sym_SLASH_SLASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH] = ACTIONS(2707),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2707),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2707),
    [anon_sym_LT_GT] = ACTIONS(2707),
    [anon_sym_STAR] = ACTIONS(2707),
    [anon_sym_STAR_STAR] = ACTIONS(2707),
    [anon_sym_DASH_GT] = ACTIONS(2707),
    [anon_sym_DOT] = ACTIONS(2707),
    [anon_sym_do] = ACTIONS(2707),
    [anon_sym_fn] = ACTIONS(2707),
    [anon_sym_LPAREN2] = ACTIONS(2705),
    [anon_sym_LBRACK2] = ACTIONS(2705),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2705),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2705),
    [sym__not_in] = ACTIONS(2705),
    [sym__quoted_atom_start] = ACTIONS(2705),
  },
  [STATE(1007)] = {
    [aux_sym__terminator_token1] = ACTIONS(2681),
    [anon_sym_SEMI] = ACTIONS(2683),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [anon_sym_RPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(1008)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2699),
    [aux_sym_identifier_token1] = ACTIONS(2699),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2699),
    [sym_alias] = ACTIONS(2699),
    [sym_integer] = ACTIONS(2699),
    [sym_float] = ACTIONS(2699),
    [sym_char] = ACTIONS(2699),
    [anon_sym_true] = ACTIONS(2699),
    [anon_sym_false] = ACTIONS(2699),
    [anon_sym_nil] = ACTIONS(2699),
    [sym_atom] = ACTIONS(2699),
    [anon_sym_DQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE] = ACTIONS(2699),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2699),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2699),
    [anon_sym_LBRACE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(2699),
    [anon_sym_LT] = ACTIONS(2699),
    [anon_sym_GT] = ACTIONS(2699),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_SLASH] = ACTIONS(2699),
    [anon_sym_TILDE] = ACTIONS(2699),
    [anon_sym_COMMA] = ACTIONS(2699),
    [sym_keyword] = ACTIONS(2699),
    [anon_sym_LT_LT] = ACTIONS(2699),
    [anon_sym_GT_GT] = ACTIONS(2699),
    [anon_sym_PERCENT] = ACTIONS(2699),
    [anon_sym_DOT_DOT] = ACTIONS(2699),
    [anon_sym_AMP] = ACTIONS(2699),
    [anon_sym_PLUS] = ACTIONS(2699),
    [anon_sym_DASH] = ACTIONS(2699),
    [anon_sym_BANG] = ACTIONS(2699),
    [anon_sym_CARET] = ACTIONS(2699),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2699),
    [anon_sym_not] = ACTIONS(2699),
    [anon_sym_AT] = ACTIONS(2699),
    [anon_sym_LT_DASH] = ACTIONS(2699),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2699),
    [anon_sym_when] = ACTIONS(2699),
    [anon_sym_COLON_COLON] = ACTIONS(2699),
    [anon_sym_EQ_GT] = ACTIONS(2699),
    [anon_sym_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2699),
    [anon_sym_or] = ACTIONS(2699),
    [anon_sym_AMP_AMP] = ACTIONS(2699),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2699),
    [anon_sym_and] = ACTIONS(2699),
    [anon_sym_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ] = ACTIONS(2699),
    [anon_sym_EQ_TILDE] = ACTIONS(2699),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2699),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2699),
    [anon_sym_LT_EQ] = ACTIONS(2699),
    [anon_sym_GT_EQ] = ACTIONS(2699),
    [anon_sym_PIPE_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_LT] = ACTIONS(2699),
    [anon_sym_GT_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE] = ACTIONS(2699),
    [anon_sym_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2699),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2699),
    [anon_sym_in] = ACTIONS(2699),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2699),
    [anon_sym_SLASH_SLASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH] = ACTIONS(2699),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2699),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2699),
    [anon_sym_LT_GT] = ACTIONS(2699),
    [anon_sym_STAR] = ACTIONS(2699),
    [anon_sym_STAR_STAR] = ACTIONS(2699),
    [anon_sym_DASH_GT] = ACTIONS(2699),
    [anon_sym_DOT] = ACTIONS(2699),
    [anon_sym_do] = ACTIONS(2699),
    [anon_sym_fn] = ACTIONS(2699),
    [anon_sym_LPAREN2] = ACTIONS(2697),
    [anon_sym_LBRACK2] = ACTIONS(2697),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2697),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2697),
    [sym__not_in] = ACTIONS(2697),
    [sym__quoted_atom_start] = ACTIONS(2697),
  },
  [STATE(1009)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2687),
    [aux_sym_identifier_token1] = ACTIONS(2687),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2687),
    [sym_alias] = ACTIONS(2687),
    [sym_integer] = ACTIONS(2687),
    [sym_float] = ACTIONS(2687),
    [sym_char] = ACTIONS(2687),
    [anon_sym_true] = ACTIONS(2687),
    [anon_sym_false] = ACTIONS(2687),
    [anon_sym_nil] = ACTIONS(2687),
    [sym_atom] = ACTIONS(2687),
    [anon_sym_DQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE] = ACTIONS(2687),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2687),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2687),
    [anon_sym_LBRACE] = ACTIONS(2687),
    [anon_sym_LBRACK] = ACTIONS(2687),
    [anon_sym_LT] = ACTIONS(2687),
    [anon_sym_GT] = ACTIONS(2687),
    [anon_sym_PIPE] = ACTIONS(2687),
    [anon_sym_SLASH] = ACTIONS(2687),
    [anon_sym_TILDE] = ACTIONS(2687),
    [anon_sym_COMMA] = ACTIONS(2687),
    [sym_keyword] = ACTIONS(2687),
    [anon_sym_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT] = ACTIONS(2687),
    [anon_sym_PERCENT] = ACTIONS(2687),
    [anon_sym_DOT_DOT] = ACTIONS(2687),
    [anon_sym_AMP] = ACTIONS(2687),
    [anon_sym_PLUS] = ACTIONS(2687),
    [anon_sym_DASH] = ACTIONS(2687),
    [anon_sym_BANG] = ACTIONS(2687),
    [anon_sym_CARET] = ACTIONS(2687),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2687),
    [anon_sym_not] = ACTIONS(2687),
    [anon_sym_AT] = ACTIONS(2687),
    [anon_sym_LT_DASH] = ACTIONS(2687),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2687),
    [anon_sym_when] = ACTIONS(2687),
    [anon_sym_COLON_COLON] = ACTIONS(2687),
    [anon_sym_EQ_GT] = ACTIONS(2687),
    [anon_sym_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2687),
    [anon_sym_or] = ACTIONS(2687),
    [anon_sym_AMP_AMP] = ACTIONS(2687),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2687),
    [anon_sym_and] = ACTIONS(2687),
    [anon_sym_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ] = ACTIONS(2687),
    [anon_sym_EQ_TILDE] = ACTIONS(2687),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2687),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2687),
    [anon_sym_LT_EQ] = ACTIONS(2687),
    [anon_sym_GT_EQ] = ACTIONS(2687),
    [anon_sym_PIPE_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_LT] = ACTIONS(2687),
    [anon_sym_GT_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE] = ACTIONS(2687),
    [anon_sym_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2687),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2687),
    [anon_sym_in] = ACTIONS(2687),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2687),
    [anon_sym_SLASH_SLASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH] = ACTIONS(2687),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2687),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2687),
    [anon_sym_LT_GT] = ACTIONS(2687),
    [anon_sym_STAR] = ACTIONS(2687),
    [anon_sym_STAR_STAR] = ACTIONS(2687),
    [anon_sym_DASH_GT] = ACTIONS(2687),
    [anon_sym_DOT] = ACTIONS(2687),
    [anon_sym_do] = ACTIONS(2687),
    [anon_sym_fn] = ACTIONS(2687),
    [anon_sym_LPAREN2] = ACTIONS(2685),
    [anon_sym_LBRACK2] = ACTIONS(2685),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2685),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2685),
    [sym__not_in] = ACTIONS(2685),
    [sym__quoted_atom_start] = ACTIONS(2685),
  },
  [STATE(1010)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2667),
    [aux_sym_identifier_token1] = ACTIONS(2667),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2667),
    [sym_alias] = ACTIONS(2667),
    [sym_integer] = ACTIONS(2667),
    [sym_float] = ACTIONS(2667),
    [sym_char] = ACTIONS(2667),
    [anon_sym_true] = ACTIONS(2667),
    [anon_sym_false] = ACTIONS(2667),
    [anon_sym_nil] = ACTIONS(2667),
    [sym_atom] = ACTIONS(2667),
    [anon_sym_DQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE] = ACTIONS(2667),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2667),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2667),
    [anon_sym_LBRACE] = ACTIONS(2667),
    [anon_sym_LBRACK] = ACTIONS(2667),
    [anon_sym_LT] = ACTIONS(2667),
    [anon_sym_GT] = ACTIONS(2667),
    [anon_sym_PIPE] = ACTIONS(2667),
    [anon_sym_SLASH] = ACTIONS(2667),
    [anon_sym_TILDE] = ACTIONS(2667),
    [anon_sym_COMMA] = ACTIONS(2667),
    [sym_keyword] = ACTIONS(2667),
    [anon_sym_LT_LT] = ACTIONS(2667),
    [anon_sym_GT_GT] = ACTIONS(2667),
    [anon_sym_PERCENT] = ACTIONS(2667),
    [anon_sym_DOT_DOT] = ACTIONS(2667),
    [anon_sym_AMP] = ACTIONS(2667),
    [anon_sym_PLUS] = ACTIONS(2667),
    [anon_sym_DASH] = ACTIONS(2667),
    [anon_sym_BANG] = ACTIONS(2667),
    [anon_sym_CARET] = ACTIONS(2667),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2667),
    [anon_sym_not] = ACTIONS(2667),
    [anon_sym_AT] = ACTIONS(2667),
    [anon_sym_LT_DASH] = ACTIONS(2667),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2667),
    [anon_sym_when] = ACTIONS(2667),
    [anon_sym_COLON_COLON] = ACTIONS(2667),
    [anon_sym_EQ_GT] = ACTIONS(2667),
    [anon_sym_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2667),
    [anon_sym_or] = ACTIONS(2667),
    [anon_sym_AMP_AMP] = ACTIONS(2667),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2667),
    [anon_sym_and] = ACTIONS(2667),
    [anon_sym_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ] = ACTIONS(2667),
    [anon_sym_EQ_TILDE] = ACTIONS(2667),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2667),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2667),
    [anon_sym_LT_EQ] = ACTIONS(2667),
    [anon_sym_GT_EQ] = ACTIONS(2667),
    [anon_sym_PIPE_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_LT] = ACTIONS(2667),
    [anon_sym_GT_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE] = ACTIONS(2667),
    [anon_sym_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2667),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2667),
    [anon_sym_in] = ACTIONS(2667),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2667),
    [anon_sym_SLASH_SLASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH] = ACTIONS(2667),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2667),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2667),
    [anon_sym_LT_GT] = ACTIONS(2667),
    [anon_sym_STAR] = ACTIONS(2667),
    [anon_sym_STAR_STAR] = ACTIONS(2667),
    [anon_sym_DASH_GT] = ACTIONS(2667),
    [anon_sym_DOT] = ACTIONS(2667),
    [anon_sym_do] = ACTIONS(2667),
    [anon_sym_fn] = ACTIONS(2667),
    [anon_sym_LPAREN2] = ACTIONS(2665),
    [anon_sym_LBRACK2] = ACTIONS(2665),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2665),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2665),
    [sym__not_in] = ACTIONS(2665),
    [sym__quoted_atom_start] = ACTIONS(2665),
  },
  [STATE(1011)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2691),
    [aux_sym_identifier_token1] = ACTIONS(2691),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2691),
    [sym_alias] = ACTIONS(2691),
    [sym_integer] = ACTIONS(2691),
    [sym_float] = ACTIONS(2691),
    [sym_char] = ACTIONS(2691),
    [anon_sym_true] = ACTIONS(2691),
    [anon_sym_false] = ACTIONS(2691),
    [anon_sym_nil] = ACTIONS(2691),
    [sym_atom] = ACTIONS(2691),
    [anon_sym_DQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE] = ACTIONS(2691),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2691),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2691),
    [anon_sym_LBRACE] = ACTIONS(2691),
    [anon_sym_LBRACK] = ACTIONS(2691),
    [anon_sym_LT] = ACTIONS(2691),
    [anon_sym_GT] = ACTIONS(2691),
    [anon_sym_PIPE] = ACTIONS(2691),
    [anon_sym_SLASH] = ACTIONS(2691),
    [anon_sym_TILDE] = ACTIONS(2691),
    [anon_sym_COMMA] = ACTIONS(2691),
    [sym_keyword] = ACTIONS(2691),
    [anon_sym_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT] = ACTIONS(2691),
    [anon_sym_PERCENT] = ACTIONS(2691),
    [anon_sym_DOT_DOT] = ACTIONS(2691),
    [anon_sym_AMP] = ACTIONS(2691),
    [anon_sym_PLUS] = ACTIONS(2691),
    [anon_sym_DASH] = ACTIONS(2691),
    [anon_sym_BANG] = ACTIONS(2691),
    [anon_sym_CARET] = ACTIONS(2691),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2691),
    [anon_sym_not] = ACTIONS(2691),
    [anon_sym_AT] = ACTIONS(2691),
    [anon_sym_LT_DASH] = ACTIONS(2691),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2691),
    [anon_sym_when] = ACTIONS(2691),
    [anon_sym_COLON_COLON] = ACTIONS(2691),
    [anon_sym_EQ_GT] = ACTIONS(2691),
    [anon_sym_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2691),
    [anon_sym_or] = ACTIONS(2691),
    [anon_sym_AMP_AMP] = ACTIONS(2691),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2691),
    [anon_sym_and] = ACTIONS(2691),
    [anon_sym_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ] = ACTIONS(2691),
    [anon_sym_EQ_TILDE] = ACTIONS(2691),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2691),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2691),
    [anon_sym_LT_EQ] = ACTIONS(2691),
    [anon_sym_GT_EQ] = ACTIONS(2691),
    [anon_sym_PIPE_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_LT] = ACTIONS(2691),
    [anon_sym_GT_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE] = ACTIONS(2691),
    [anon_sym_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2691),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2691),
    [anon_sym_in] = ACTIONS(2691),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2691),
    [anon_sym_SLASH_SLASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH] = ACTIONS(2691),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2691),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2691),
    [anon_sym_LT_GT] = ACTIONS(2691),
    [anon_sym_STAR] = ACTIONS(2691),
    [anon_sym_STAR_STAR] = ACTIONS(2691),
    [anon_sym_DASH_GT] = ACTIONS(2691),
    [anon_sym_DOT] = ACTIONS(2691),
    [anon_sym_do] = ACTIONS(2691),
    [anon_sym_fn] = ACTIONS(2691),
    [anon_sym_LPAREN2] = ACTIONS(2689),
    [anon_sym_LBRACK2] = ACTIONS(2689),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2689),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2689),
    [sym__not_in] = ACTIONS(2689),
    [sym__quoted_atom_start] = ACTIONS(2689),
  },
  [STATE(1012)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2695),
    [aux_sym_identifier_token1] = ACTIONS(2695),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2695),
    [sym_alias] = ACTIONS(2695),
    [sym_integer] = ACTIONS(2695),
    [sym_float] = ACTIONS(2695),
    [sym_char] = ACTIONS(2695),
    [anon_sym_true] = ACTIONS(2695),
    [anon_sym_false] = ACTIONS(2695),
    [anon_sym_nil] = ACTIONS(2695),
    [sym_atom] = ACTIONS(2695),
    [anon_sym_DQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE] = ACTIONS(2695),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2695),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2695),
    [anon_sym_LBRACE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(2695),
    [anon_sym_LT] = ACTIONS(2695),
    [anon_sym_GT] = ACTIONS(2695),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_SLASH] = ACTIONS(2695),
    [anon_sym_TILDE] = ACTIONS(2695),
    [anon_sym_COMMA] = ACTIONS(2695),
    [sym_keyword] = ACTIONS(2695),
    [anon_sym_LT_LT] = ACTIONS(2695),
    [anon_sym_GT_GT] = ACTIONS(2695),
    [anon_sym_PERCENT] = ACTIONS(2695),
    [anon_sym_DOT_DOT] = ACTIONS(2695),
    [anon_sym_AMP] = ACTIONS(2695),
    [anon_sym_PLUS] = ACTIONS(2695),
    [anon_sym_DASH] = ACTIONS(2695),
    [anon_sym_BANG] = ACTIONS(2695),
    [anon_sym_CARET] = ACTIONS(2695),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2695),
    [anon_sym_not] = ACTIONS(2695),
    [anon_sym_AT] = ACTIONS(2695),
    [anon_sym_LT_DASH] = ACTIONS(2695),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2695),
    [anon_sym_when] = ACTIONS(2695),
    [anon_sym_COLON_COLON] = ACTIONS(2695),
    [anon_sym_EQ_GT] = ACTIONS(2695),
    [anon_sym_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2695),
    [anon_sym_or] = ACTIONS(2695),
    [anon_sym_AMP_AMP] = ACTIONS(2695),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2695),
    [anon_sym_and] = ACTIONS(2695),
    [anon_sym_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ] = ACTIONS(2695),
    [anon_sym_EQ_TILDE] = ACTIONS(2695),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2695),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2695),
    [anon_sym_LT_EQ] = ACTIONS(2695),
    [anon_sym_GT_EQ] = ACTIONS(2695),
    [anon_sym_PIPE_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_LT] = ACTIONS(2695),
    [anon_sym_GT_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE] = ACTIONS(2695),
    [anon_sym_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2695),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2695),
    [anon_sym_in] = ACTIONS(2695),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2695),
    [anon_sym_SLASH_SLASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH] = ACTIONS(2695),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2695),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2695),
    [anon_sym_LT_GT] = ACTIONS(2695),
    [anon_sym_STAR] = ACTIONS(2695),
    [anon_sym_STAR_STAR] = ACTIONS(2695),
    [anon_sym_DASH_GT] = ACTIONS(2695),
    [anon_sym_DOT] = ACTIONS(2695),
    [anon_sym_do] = ACTIONS(2695),
    [anon_sym_fn] = ACTIONS(2695),
    [anon_sym_LPAREN2] = ACTIONS(2693),
    [anon_sym_LBRACK2] = ACTIONS(2693),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2693),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2693),
    [sym__not_in] = ACTIONS(2693),
    [sym__quoted_atom_start] = ACTIONS(2693),
  },
  [STATE(1013)] = {
    [ts_builtin_sym_end] = ACTIONS(2681),
    [aux_sym__terminator_token1] = ACTIONS(2681),
    [anon_sym_SEMI] = ACTIONS(2683),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(1014)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2711),
    [aux_sym_identifier_token1] = ACTIONS(2711),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2711),
    [sym_alias] = ACTIONS(2711),
    [sym_integer] = ACTIONS(2711),
    [sym_float] = ACTIONS(2711),
    [sym_char] = ACTIONS(2711),
    [anon_sym_true] = ACTIONS(2711),
    [anon_sym_false] = ACTIONS(2711),
    [anon_sym_nil] = ACTIONS(2711),
    [sym_atom] = ACTIONS(2711),
    [anon_sym_DQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE] = ACTIONS(2711),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2711),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2711),
    [anon_sym_LBRACE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(2711),
    [anon_sym_LT] = ACTIONS(2711),
    [anon_sym_GT] = ACTIONS(2711),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_SLASH] = ACTIONS(2711),
    [anon_sym_TILDE] = ACTIONS(2711),
    [anon_sym_COMMA] = ACTIONS(2711),
    [sym_keyword] = ACTIONS(2711),
    [anon_sym_LT_LT] = ACTIONS(2711),
    [anon_sym_GT_GT] = ACTIONS(2711),
    [anon_sym_PERCENT] = ACTIONS(2711),
    [anon_sym_DOT_DOT] = ACTIONS(2711),
    [anon_sym_AMP] = ACTIONS(2711),
    [anon_sym_PLUS] = ACTIONS(2711),
    [anon_sym_DASH] = ACTIONS(2711),
    [anon_sym_BANG] = ACTIONS(2711),
    [anon_sym_CARET] = ACTIONS(2711),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2711),
    [anon_sym_not] = ACTIONS(2711),
    [anon_sym_AT] = ACTIONS(2711),
    [anon_sym_LT_DASH] = ACTIONS(2711),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2711),
    [anon_sym_when] = ACTIONS(2711),
    [anon_sym_COLON_COLON] = ACTIONS(2711),
    [anon_sym_EQ_GT] = ACTIONS(2711),
    [anon_sym_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2711),
    [anon_sym_or] = ACTIONS(2711),
    [anon_sym_AMP_AMP] = ACTIONS(2711),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2711),
    [anon_sym_and] = ACTIONS(2711),
    [anon_sym_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ] = ACTIONS(2711),
    [anon_sym_EQ_TILDE] = ACTIONS(2711),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2711),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2711),
    [anon_sym_LT_EQ] = ACTIONS(2711),
    [anon_sym_GT_EQ] = ACTIONS(2711),
    [anon_sym_PIPE_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_LT] = ACTIONS(2711),
    [anon_sym_GT_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE] = ACTIONS(2711),
    [anon_sym_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2711),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2711),
    [anon_sym_in] = ACTIONS(2711),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2711),
    [anon_sym_SLASH_SLASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH] = ACTIONS(2711),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2711),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2711),
    [anon_sym_LT_GT] = ACTIONS(2711),
    [anon_sym_STAR] = ACTIONS(2711),
    [anon_sym_STAR_STAR] = ACTIONS(2711),
    [anon_sym_DASH_GT] = ACTIONS(2711),
    [anon_sym_DOT] = ACTIONS(2711),
    [anon_sym_do] = ACTIONS(2711),
    [anon_sym_fn] = ACTIONS(2711),
    [anon_sym_LPAREN2] = ACTIONS(2709),
    [anon_sym_LBRACK2] = ACTIONS(2709),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2709),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2709),
    [sym__not_in] = ACTIONS(2709),
    [sym__quoted_atom_start] = ACTIONS(2709),
  },
  [STATE(1015)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2679),
    [aux_sym_identifier_token1] = ACTIONS(2679),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2679),
    [sym_alias] = ACTIONS(2679),
    [sym_integer] = ACTIONS(2679),
    [sym_float] = ACTIONS(2679),
    [sym_char] = ACTIONS(2679),
    [anon_sym_true] = ACTIONS(2679),
    [anon_sym_false] = ACTIONS(2679),
    [anon_sym_nil] = ACTIONS(2679),
    [sym_atom] = ACTIONS(2679),
    [anon_sym_DQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE] = ACTIONS(2679),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2679),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2679),
    [anon_sym_LBRACE] = ACTIONS(2679),
    [anon_sym_LBRACK] = ACTIONS(2679),
    [anon_sym_LT] = ACTIONS(2679),
    [anon_sym_GT] = ACTIONS(2679),
    [anon_sym_PIPE] = ACTIONS(2679),
    [anon_sym_SLASH] = ACTIONS(2679),
    [anon_sym_TILDE] = ACTIONS(2679),
    [anon_sym_COMMA] = ACTIONS(2679),
    [sym_keyword] = ACTIONS(2679),
    [anon_sym_LT_LT] = ACTIONS(2679),
    [anon_sym_GT_GT] = ACTIONS(2679),
    [anon_sym_PERCENT] = ACTIONS(2679),
    [anon_sym_DOT_DOT] = ACTIONS(2679),
    [anon_sym_AMP] = ACTIONS(2679),
    [anon_sym_PLUS] = ACTIONS(2679),
    [anon_sym_DASH] = ACTIONS(2679),
    [anon_sym_BANG] = ACTIONS(2679),
    [anon_sym_CARET] = ACTIONS(2679),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2679),
    [anon_sym_not] = ACTIONS(2679),
    [anon_sym_AT] = ACTIONS(2679),
    [anon_sym_LT_DASH] = ACTIONS(2679),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2679),
    [anon_sym_when] = ACTIONS(2679),
    [anon_sym_COLON_COLON] = ACTIONS(2679),
    [anon_sym_EQ_GT] = ACTIONS(2679),
    [anon_sym_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2679),
    [anon_sym_or] = ACTIONS(2679),
    [anon_sym_AMP_AMP] = ACTIONS(2679),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2679),
    [anon_sym_and] = ACTIONS(2679),
    [anon_sym_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ] = ACTIONS(2679),
    [anon_sym_EQ_TILDE] = ACTIONS(2679),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2679),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2679),
    [anon_sym_LT_EQ] = ACTIONS(2679),
    [anon_sym_GT_EQ] = ACTIONS(2679),
    [anon_sym_PIPE_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_LT] = ACTIONS(2679),
    [anon_sym_GT_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE] = ACTIONS(2679),
    [anon_sym_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2679),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2679),
    [anon_sym_in] = ACTIONS(2679),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2679),
    [anon_sym_SLASH_SLASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH] = ACTIONS(2679),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2679),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2679),
    [anon_sym_LT_GT] = ACTIONS(2679),
    [anon_sym_STAR] = ACTIONS(2679),
    [anon_sym_STAR_STAR] = ACTIONS(2679),
    [anon_sym_DASH_GT] = ACTIONS(2679),
    [anon_sym_DOT] = ACTIONS(2679),
    [anon_sym_do] = ACTIONS(2679),
    [anon_sym_fn] = ACTIONS(2679),
    [anon_sym_LPAREN2] = ACTIONS(2677),
    [anon_sym_LBRACK2] = ACTIONS(2677),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2677),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2677),
    [sym__not_in] = ACTIONS(2677),
    [sym__quoted_atom_start] = ACTIONS(2677),
  },
  [STATE(1016)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2671),
    [aux_sym_identifier_token1] = ACTIONS(2671),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2671),
    [sym_alias] = ACTIONS(2671),
    [sym_integer] = ACTIONS(2671),
    [sym_float] = ACTIONS(2671),
    [sym_char] = ACTIONS(2671),
    [anon_sym_true] = ACTIONS(2671),
    [anon_sym_false] = ACTIONS(2671),
    [anon_sym_nil] = ACTIONS(2671),
    [sym_atom] = ACTIONS(2671),
    [anon_sym_DQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE] = ACTIONS(2671),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2671),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2671),
    [anon_sym_LBRACE] = ACTIONS(2671),
    [anon_sym_LBRACK] = ACTIONS(2671),
    [anon_sym_LT] = ACTIONS(2671),
    [anon_sym_GT] = ACTIONS(2671),
    [anon_sym_PIPE] = ACTIONS(2671),
    [anon_sym_SLASH] = ACTIONS(2671),
    [anon_sym_TILDE] = ACTIONS(2671),
    [anon_sym_COMMA] = ACTIONS(2671),
    [sym_keyword] = ACTIONS(2671),
    [anon_sym_LT_LT] = ACTIONS(2671),
    [anon_sym_GT_GT] = ACTIONS(2671),
    [anon_sym_PERCENT] = ACTIONS(2671),
    [anon_sym_DOT_DOT] = ACTIONS(2671),
    [anon_sym_AMP] = ACTIONS(2671),
    [anon_sym_PLUS] = ACTIONS(2671),
    [anon_sym_DASH] = ACTIONS(2671),
    [anon_sym_BANG] = ACTIONS(2671),
    [anon_sym_CARET] = ACTIONS(2671),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2671),
    [anon_sym_not] = ACTIONS(2671),
    [anon_sym_AT] = ACTIONS(2671),
    [anon_sym_LT_DASH] = ACTIONS(2671),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2671),
    [anon_sym_when] = ACTIONS(2671),
    [anon_sym_COLON_COLON] = ACTIONS(2671),
    [anon_sym_EQ_GT] = ACTIONS(2671),
    [anon_sym_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2671),
    [anon_sym_or] = ACTIONS(2671),
    [anon_sym_AMP_AMP] = ACTIONS(2671),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2671),
    [anon_sym_and] = ACTIONS(2671),
    [anon_sym_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ] = ACTIONS(2671),
    [anon_sym_EQ_TILDE] = ACTIONS(2671),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2671),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2671),
    [anon_sym_LT_EQ] = ACTIONS(2671),
    [anon_sym_GT_EQ] = ACTIONS(2671),
    [anon_sym_PIPE_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_LT] = ACTIONS(2671),
    [anon_sym_GT_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE] = ACTIONS(2671),
    [anon_sym_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2671),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2671),
    [anon_sym_in] = ACTIONS(2671),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2671),
    [anon_sym_SLASH_SLASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH] = ACTIONS(2671),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2671),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2671),
    [anon_sym_LT_GT] = ACTIONS(2671),
    [anon_sym_STAR] = ACTIONS(2671),
    [anon_sym_STAR_STAR] = ACTIONS(2671),
    [anon_sym_DASH_GT] = ACTIONS(2671),
    [anon_sym_DOT] = ACTIONS(2671),
    [anon_sym_do] = ACTIONS(2671),
    [anon_sym_fn] = ACTIONS(2671),
    [anon_sym_LPAREN2] = ACTIONS(2669),
    [anon_sym_LBRACK2] = ACTIONS(2669),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(2669),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2669),
    [sym__not_in] = ACTIONS(2669),
    [sym__quoted_atom_start] = ACTIONS(2669),
  },
  [STATE(1017)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2683),
    [aux_sym_identifier_token1] = ACTIONS(2683),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2683),
    [sym_alias] = ACTIONS(2683),
    [sym_integer] = ACTIONS(2683),
    [sym_float] = ACTIONS(2683),
    [sym_char] = ACTIONS(2683),
    [anon_sym_true] = ACTIONS(2683),
    [anon_sym_false] = ACTIONS(2683),
    [anon_sym_nil] = ACTIONS(2683),
    [sym_atom] = ACTIONS(2683),
    [anon_sym_DQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE] = ACTIONS(2683),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2683),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2683),
    [anon_sym_LBRACE] = ACTIONS(2683),
    [anon_sym_LBRACK] = ACTIONS(2683),
    [anon_sym_LT] = ACTIONS(2683),
    [anon_sym_GT] = ACTIONS(2683),
    [anon_sym_PIPE] = ACTIONS(2683),
    [anon_sym_SLASH] = ACTIONS(2683),
    [anon_sym_TILDE] = ACTIONS(2683),
    [anon_sym_COMMA] = ACTIONS(2683),
    [sym_keyword] = ACTIONS(2683),
    [anon_sym_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT] = ACTIONS(2683),
    [anon_sym_PERCENT] = ACTIONS(2683),
    [anon_sym_DOT_DOT] = ACTIONS(2683),
    [anon_sym_AMP] = ACTIONS(2683),
    [anon_sym_PLUS] = ACTIONS(2683),
    [anon_sym_DASH] = ACTIONS(2683),
    [anon_sym_BANG] = ACTIONS(2683),
    [anon_sym_CARET] = ACTIONS(2683),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2683),
    [anon_sym_not] = ACTIONS(2683),
    [anon_sym_AT] = ACTIONS(2683),
    [anon_sym_LT_DASH] = ACTIONS(2683),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2683),
    [anon_sym_when] = ACTIONS(2683),
    [anon_sym_COLON_COLON] = ACTIONS(2683),
    [anon_sym_EQ_GT] = ACTIONS(2683),
    [anon_sym_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2683),
    [anon_sym_or] = ACTIONS(2683),
    [anon_sym_AMP_AMP] = ACTIONS(2683),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2683),
    [anon_sym_and] = ACTIONS(2683),
    [anon_sym_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ] = ACTIONS(2683),
    [anon_sym_EQ_TILDE] = ACTIONS(2683),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2683),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2683),
    [anon_sym_LT_EQ] = ACTIONS(2683),
    [anon_sym_GT_EQ] = ACTIONS(2683),
    [anon_sym_PIPE_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_LT] = ACTIONS(2683),
    [anon_sym_GT_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE] = ACTIONS(2683),
    [anon_sym_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2683),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2683),
    [anon_sym_in] = ACTIONS(2683),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2683),
    [anon_sym_SLASH_SLASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH] = ACTIONS(2683),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2683),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2683),
    [anon_sym_LT_GT] = ACTIONS(2683),
    [anon_sym_STAR] = ACTIONS(2683),
    [anon_sym_STAR_STAR] = ACTIONS(2683),
    [anon_sym_DASH_GT] = ACTIONS(2683),
    [anon_sym_DOT] = ACTIONS(2683),
    [anon_sym_do] = ACTIONS(2683),
    [anon_sym_fn] = ACTIONS(2683),
    [anon_sym_LPAREN2] = ACTIONS(2681),
    [anon_sym_LBRACK2] = ACTIONS(2681),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2681),
    [sym__not_in] = ACTIONS(2681),
    [sym__quoted_atom_start] = ACTIONS(2681),
  },
  [STATE(1018)] = {
    [aux_sym__terminator_repeat1] = STATE(1019),
    [aux_sym__terminator_token1] = ACTIONS(2713),
    [anon_sym_SEMI] = ACTIONS(2715),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [aux_sym_identifier_token1] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [sym_alias] = ACTIONS(2717),
    [sym_integer] = ACTIONS(2717),
    [sym_float] = ACTIONS(2717),
    [sym_char] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [anon_sym_nil] = ACTIONS(2717),
    [sym_atom] = ACTIONS(2717),
    [anon_sym_DQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2717),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [sym_keyword] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_not] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_LT_DASH] = ACTIONS(2717),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2717),
    [anon_sym_when] = ACTIONS(2717),
    [anon_sym_COLON_COLON] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_LT] = ACTIONS(2717),
    [anon_sym_GT_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2717),
    [anon_sym_in] = ACTIONS(2717),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2717),
    [anon_sym_LT_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_after] = ACTIONS(2717),
    [anon_sym_catch] = ACTIONS(2717),
    [anon_sym_else] = ACTIONS(2717),
    [anon_sym_end] = ACTIONS(2717),
    [anon_sym_fn] = ACTIONS(2717),
    [anon_sym_rescue] = ACTIONS(2717),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2719),
    [sym__not_in] = ACTIONS(2719),
    [sym__quoted_atom_start] = ACTIONS(2719),
  },
  [STATE(1019)] = {
    [aux_sym__terminator_repeat1] = STATE(1019),
    [aux_sym__terminator_token1] = ACTIONS(2721),
    [anon_sym_SEMI] = ACTIONS(2724),
    [anon_sym_LPAREN] = ACTIONS(2724),
    [aux_sym_identifier_token1] = ACTIONS(2724),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2724),
    [sym_alias] = ACTIONS(2724),
    [sym_integer] = ACTIONS(2724),
    [sym_float] = ACTIONS(2724),
    [sym_char] = ACTIONS(2724),
    [anon_sym_true] = ACTIONS(2724),
    [anon_sym_false] = ACTIONS(2724),
    [anon_sym_nil] = ACTIONS(2724),
    [sym_atom] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2724),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2724),
    [anon_sym_LT] = ACTIONS(2724),
    [anon_sym_GT] = ACTIONS(2724),
    [anon_sym_PIPE] = ACTIONS(2724),
    [anon_sym_SLASH] = ACTIONS(2724),
    [anon_sym_TILDE] = ACTIONS(2724),
    [sym_keyword] = ACTIONS(2724),
    [anon_sym_LT_LT] = ACTIONS(2724),
    [anon_sym_PERCENT] = ACTIONS(2724),
    [anon_sym_DOT_DOT] = ACTIONS(2724),
    [anon_sym_AMP] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_BANG] = ACTIONS(2724),
    [anon_sym_CARET] = ACTIONS(2724),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2724),
    [anon_sym_not] = ACTIONS(2724),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_LT_DASH] = ACTIONS(2724),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2724),
    [anon_sym_when] = ACTIONS(2724),
    [anon_sym_COLON_COLON] = ACTIONS(2724),
    [anon_sym_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_or] = ACTIONS(2724),
    [anon_sym_AMP_AMP] = ACTIONS(2724),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2724),
    [anon_sym_and] = ACTIONS(2724),
    [anon_sym_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ] = ACTIONS(2724),
    [anon_sym_EQ_TILDE] = ACTIONS(2724),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2724),
    [anon_sym_LT_EQ] = ACTIONS(2724),
    [anon_sym_GT_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_LT] = ACTIONS(2724),
    [anon_sym_GT_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2724),
    [anon_sym_LT_GT] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2724),
    [anon_sym_STAR_STAR] = ACTIONS(2724),
    [anon_sym_DASH_GT] = ACTIONS(2724),
    [anon_sym_after] = ACTIONS(2724),
    [anon_sym_catch] = ACTIONS(2724),
    [anon_sym_else] = ACTIONS(2724),
    [anon_sym_end] = ACTIONS(2724),
    [anon_sym_fn] = ACTIONS(2724),
    [anon_sym_rescue] = ACTIONS(2724),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2726),
    [sym__not_in] = ACTIONS(2726),
    [sym__quoted_atom_start] = ACTIONS(2726),
  },
  [STATE(1020)] = {
    [aux_sym__terminator_repeat1] = STATE(1020),
    [aux_sym__terminator_token1] = ACTIONS(2728),
    [anon_sym_SEMI] = ACTIONS(2724),
    [anon_sym_LPAREN] = ACTIONS(2724),
    [aux_sym_identifier_token1] = ACTIONS(2724),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2724),
    [sym_alias] = ACTIONS(2724),
    [sym_integer] = ACTIONS(2724),
    [sym_float] = ACTIONS(2724),
    [sym_char] = ACTIONS(2724),
    [anon_sym_true] = ACTIONS(2724),
    [anon_sym_false] = ACTIONS(2724),
    [anon_sym_nil] = ACTIONS(2724),
    [sym_atom] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2724),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2724),
    [anon_sym_LT] = ACTIONS(2724),
    [anon_sym_GT] = ACTIONS(2724),
    [anon_sym_PIPE] = ACTIONS(2724),
    [anon_sym_SLASH] = ACTIONS(2724),
    [anon_sym_TILDE] = ACTIONS(2724),
    [anon_sym_LT_LT] = ACTIONS(2724),
    [anon_sym_PERCENT] = ACTIONS(2724),
    [anon_sym_DOT_DOT] = ACTIONS(2724),
    [anon_sym_AMP] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_BANG] = ACTIONS(2724),
    [anon_sym_CARET] = ACTIONS(2724),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2724),
    [anon_sym_not] = ACTIONS(2724),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_LT_DASH] = ACTIONS(2724),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2724),
    [anon_sym_when] = ACTIONS(2724),
    [anon_sym_COLON_COLON] = ACTIONS(2724),
    [anon_sym_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_or] = ACTIONS(2724),
    [anon_sym_AMP_AMP] = ACTIONS(2724),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2724),
    [anon_sym_and] = ACTIONS(2724),
    [anon_sym_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ] = ACTIONS(2724),
    [anon_sym_EQ_TILDE] = ACTIONS(2724),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2724),
    [anon_sym_LT_EQ] = ACTIONS(2724),
    [anon_sym_GT_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_LT] = ACTIONS(2724),
    [anon_sym_GT_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2724),
    [anon_sym_LT_GT] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2724),
    [anon_sym_STAR_STAR] = ACTIONS(2724),
    [anon_sym_DASH_GT] = ACTIONS(2724),
    [anon_sym_after] = ACTIONS(2724),
    [anon_sym_catch] = ACTIONS(2724),
    [anon_sym_else] = ACTIONS(2724),
    [anon_sym_end] = ACTIONS(2724),
    [anon_sym_fn] = ACTIONS(2724),
    [anon_sym_rescue] = ACTIONS(2724),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2726),
    [sym__not_in] = ACTIONS(2726),
    [sym__quoted_atom_start] = ACTIONS(2726),
  },
  [STATE(1021)] = {
    [aux_sym__terminator_repeat1] = STATE(1020),
    [aux_sym__terminator_token1] = ACTIONS(2731),
    [anon_sym_SEMI] = ACTIONS(2733),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [aux_sym_identifier_token1] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [sym_alias] = ACTIONS(2717),
    [sym_integer] = ACTIONS(2717),
    [sym_float] = ACTIONS(2717),
    [sym_char] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [anon_sym_nil] = ACTIONS(2717),
    [sym_atom] = ACTIONS(2717),
    [anon_sym_DQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2717),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_not] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_LT_DASH] = ACTIONS(2717),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2717),
    [anon_sym_when] = ACTIONS(2717),
    [anon_sym_COLON_COLON] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_LT] = ACTIONS(2717),
    [anon_sym_GT_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2717),
    [anon_sym_in] = ACTIONS(2717),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2717),
    [anon_sym_LT_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_after] = ACTIONS(2717),
    [anon_sym_catch] = ACTIONS(2717),
    [anon_sym_else] = ACTIONS(2717),
    [anon_sym_end] = ACTIONS(2717),
    [anon_sym_fn] = ACTIONS(2717),
    [anon_sym_rescue] = ACTIONS(2717),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2719),
    [sym__not_in] = ACTIONS(2719),
    [sym__quoted_atom_start] = ACTIONS(2719),
  },
  [STATE(1022)] = {
    [aux_sym__terminator_repeat1] = STATE(1020),
    [aux_sym__terminator_token1] = ACTIONS(2731),
    [anon_sym_SEMI] = ACTIONS(2735),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [aux_sym_identifier_token1] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [sym_alias] = ACTIONS(2717),
    [sym_integer] = ACTIONS(2717),
    [sym_float] = ACTIONS(2717),
    [sym_char] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [anon_sym_nil] = ACTIONS(2717),
    [sym_atom] = ACTIONS(2717),
    [anon_sym_DQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2717),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_not] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_LT_DASH] = ACTIONS(2717),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2717),
    [anon_sym_when] = ACTIONS(2717),
    [anon_sym_COLON_COLON] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_LT] = ACTIONS(2717),
    [anon_sym_GT_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2717),
    [anon_sym_in] = ACTIONS(2717),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2717),
    [anon_sym_LT_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_after] = ACTIONS(2717),
    [anon_sym_catch] = ACTIONS(2717),
    [anon_sym_else] = ACTIONS(2717),
    [anon_sym_end] = ACTIONS(2717),
    [anon_sym_fn] = ACTIONS(2717),
    [anon_sym_rescue] = ACTIONS(2717),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2719),
    [sym__not_in] = ACTIONS(2719),
    [sym__quoted_atom_start] = ACTIONS(2719),
  },
  [STATE(1023)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2737),
    [aux_sym_identifier_token1] = ACTIONS(2737),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2737),
    [sym_alias] = ACTIONS(2737),
    [sym_integer] = ACTIONS(2737),
    [sym_float] = ACTIONS(2737),
    [sym_char] = ACTIONS(2737),
    [anon_sym_true] = ACTIONS(2737),
    [anon_sym_false] = ACTIONS(2737),
    [anon_sym_nil] = ACTIONS(2737),
    [sym_atom] = ACTIONS(2737),
    [anon_sym_DQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2737),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2737),
    [anon_sym_LBRACE] = ACTIONS(2737),
    [anon_sym_LBRACK] = ACTIONS(2737),
    [anon_sym_LT] = ACTIONS(2737),
    [anon_sym_GT] = ACTIONS(2737),
    [anon_sym_PIPE] = ACTIONS(2737),
    [anon_sym_SLASH] = ACTIONS(2737),
    [anon_sym_TILDE] = ACTIONS(2737),
    [sym_keyword] = ACTIONS(2737),
    [anon_sym_LT_LT] = ACTIONS(2737),
    [anon_sym_PERCENT] = ACTIONS(2737),
    [anon_sym_DOT_DOT] = ACTIONS(2737),
    [anon_sym_AMP] = ACTIONS(2737),
    [anon_sym_PLUS] = ACTIONS(2737),
    [anon_sym_DASH] = ACTIONS(2737),
    [anon_sym_BANG] = ACTIONS(2737),
    [anon_sym_CARET] = ACTIONS(2737),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2737),
    [anon_sym_not] = ACTIONS(2737),
    [anon_sym_AT] = ACTIONS(2737),
    [anon_sym_LT_DASH] = ACTIONS(2737),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2737),
    [anon_sym_when] = ACTIONS(2737),
    [anon_sym_COLON_COLON] = ACTIONS(2737),
    [anon_sym_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_or] = ACTIONS(2737),
    [anon_sym_AMP_AMP] = ACTIONS(2737),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2737),
    [anon_sym_and] = ACTIONS(2737),
    [anon_sym_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ] = ACTIONS(2737),
    [anon_sym_EQ_TILDE] = ACTIONS(2737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2737),
    [anon_sym_LT_EQ] = ACTIONS(2737),
    [anon_sym_GT_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_LT] = ACTIONS(2737),
    [anon_sym_GT_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2737),
    [anon_sym_in] = ACTIONS(2737),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2737),
    [anon_sym_LT_GT] = ACTIONS(2737),
    [anon_sym_STAR] = ACTIONS(2737),
    [anon_sym_STAR_STAR] = ACTIONS(2737),
    [anon_sym_DASH_GT] = ACTIONS(2737),
    [anon_sym_after] = ACTIONS(2737),
    [anon_sym_catch] = ACTIONS(2737),
    [anon_sym_else] = ACTIONS(2737),
    [anon_sym_end] = ACTIONS(2737),
    [anon_sym_fn] = ACTIONS(2737),
    [anon_sym_rescue] = ACTIONS(2737),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2739),
    [sym__not_in] = ACTIONS(2739),
    [sym__quoted_atom_start] = ACTIONS(2739),
  },
  [STATE(1024)] = {
    [aux_sym__terminator_token1] = ACTIONS(2739),
    [anon_sym_SEMI] = ACTIONS(2737),
    [anon_sym_LPAREN] = ACTIONS(2737),
    [aux_sym_identifier_token1] = ACTIONS(2737),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2737),
    [sym_alias] = ACTIONS(2737),
    [sym_integer] = ACTIONS(2737),
    [sym_float] = ACTIONS(2737),
    [sym_char] = ACTIONS(2737),
    [anon_sym_true] = ACTIONS(2737),
    [anon_sym_false] = ACTIONS(2737),
    [anon_sym_nil] = ACTIONS(2737),
    [sym_atom] = ACTIONS(2737),
    [anon_sym_DQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2737),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2737),
    [anon_sym_LBRACE] = ACTIONS(2737),
    [anon_sym_LBRACK] = ACTIONS(2737),
    [anon_sym_LT] = ACTIONS(2737),
    [anon_sym_GT] = ACTIONS(2737),
    [anon_sym_PIPE] = ACTIONS(2737),
    [anon_sym_SLASH] = ACTIONS(2737),
    [anon_sym_TILDE] = ACTIONS(2737),
    [anon_sym_LT_LT] = ACTIONS(2737),
    [anon_sym_PERCENT] = ACTIONS(2737),
    [anon_sym_DOT_DOT] = ACTIONS(2737),
    [anon_sym_AMP] = ACTIONS(2737),
    [anon_sym_PLUS] = ACTIONS(2737),
    [anon_sym_DASH] = ACTIONS(2737),
    [anon_sym_BANG] = ACTIONS(2737),
    [anon_sym_CARET] = ACTIONS(2737),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2737),
    [anon_sym_not] = ACTIONS(2737),
    [anon_sym_AT] = ACTIONS(2737),
    [anon_sym_LT_DASH] = ACTIONS(2737),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2737),
    [anon_sym_when] = ACTIONS(2737),
    [anon_sym_COLON_COLON] = ACTIONS(2737),
    [anon_sym_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_or] = ACTIONS(2737),
    [anon_sym_AMP_AMP] = ACTIONS(2737),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2737),
    [anon_sym_and] = ACTIONS(2737),
    [anon_sym_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ] = ACTIONS(2737),
    [anon_sym_EQ_TILDE] = ACTIONS(2737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2737),
    [anon_sym_LT_EQ] = ACTIONS(2737),
    [anon_sym_GT_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_LT] = ACTIONS(2737),
    [anon_sym_GT_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2737),
    [anon_sym_in] = ACTIONS(2737),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2737),
    [anon_sym_LT_GT] = ACTIONS(2737),
    [anon_sym_STAR] = ACTIONS(2737),
    [anon_sym_STAR_STAR] = ACTIONS(2737),
    [anon_sym_DASH_GT] = ACTIONS(2737),
    [anon_sym_after] = ACTIONS(2737),
    [anon_sym_catch] = ACTIONS(2737),
    [anon_sym_else] = ACTIONS(2737),
    [anon_sym_end] = ACTIONS(2737),
    [anon_sym_fn] = ACTIONS(2737),
    [anon_sym_rescue] = ACTIONS(2737),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2739),
    [sym__not_in] = ACTIONS(2739),
    [sym__quoted_atom_start] = ACTIONS(2739),
  },
  [STATE(1025)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2737),
    [aux_sym_identifier_token1] = ACTIONS(2737),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2737),
    [sym_alias] = ACTIONS(2737),
    [sym_integer] = ACTIONS(2737),
    [sym_float] = ACTIONS(2737),
    [sym_char] = ACTIONS(2737),
    [anon_sym_true] = ACTIONS(2737),
    [anon_sym_false] = ACTIONS(2737),
    [anon_sym_nil] = ACTIONS(2737),
    [sym_atom] = ACTIONS(2737),
    [anon_sym_DQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2737),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2737),
    [anon_sym_LBRACE] = ACTIONS(2737),
    [anon_sym_LBRACK] = ACTIONS(2737),
    [anon_sym_LT] = ACTIONS(2737),
    [anon_sym_GT] = ACTIONS(2737),
    [anon_sym_PIPE] = ACTIONS(2737),
    [anon_sym_SLASH] = ACTIONS(2737),
    [anon_sym_TILDE] = ACTIONS(2737),
    [anon_sym_LT_LT] = ACTIONS(2737),
    [anon_sym_PERCENT] = ACTIONS(2737),
    [anon_sym_DOT_DOT] = ACTIONS(2737),
    [anon_sym_AMP] = ACTIONS(2737),
    [anon_sym_PLUS] = ACTIONS(2737),
    [anon_sym_DASH] = ACTIONS(2737),
    [anon_sym_BANG] = ACTIONS(2737),
    [anon_sym_CARET] = ACTIONS(2737),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2737),
    [anon_sym_not] = ACTIONS(2737),
    [anon_sym_AT] = ACTIONS(2737),
    [anon_sym_LT_DASH] = ACTIONS(2737),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2737),
    [anon_sym_when] = ACTIONS(2737),
    [anon_sym_COLON_COLON] = ACTIONS(2737),
    [anon_sym_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_or] = ACTIONS(2737),
    [anon_sym_AMP_AMP] = ACTIONS(2737),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2737),
    [anon_sym_and] = ACTIONS(2737),
    [anon_sym_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ] = ACTIONS(2737),
    [anon_sym_EQ_TILDE] = ACTIONS(2737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2737),
    [anon_sym_LT_EQ] = ACTIONS(2737),
    [anon_sym_GT_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_LT] = ACTIONS(2737),
    [anon_sym_GT_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2737),
    [anon_sym_in] = ACTIONS(2737),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2737),
    [anon_sym_LT_GT] = ACTIONS(2737),
    [anon_sym_STAR] = ACTIONS(2737),
    [anon_sym_STAR_STAR] = ACTIONS(2737),
    [anon_sym_DASH_GT] = ACTIONS(2737),
    [anon_sym_after] = ACTIONS(2737),
    [anon_sym_catch] = ACTIONS(2737),
    [anon_sym_else] = ACTIONS(2737),
    [anon_sym_end] = ACTIONS(2737),
    [anon_sym_fn] = ACTIONS(2737),
    [anon_sym_rescue] = ACTIONS(2737),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2739),
    [sym__not_in] = ACTIONS(2739),
    [sym__quoted_atom_start] = ACTIONS(2739),
  },
  [STATE(1026)] = {
    [aux_sym__terminator_repeat1] = STATE(1026),
    [aux_sym__terminator_token1] = ACTIONS(2741),
    [anon_sym_SEMI] = ACTIONS(2724),
    [anon_sym_LPAREN] = ACTIONS(2724),
    [anon_sym_RPAREN] = ACTIONS(2724),
    [aux_sym_identifier_token1] = ACTIONS(2724),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2724),
    [sym_alias] = ACTIONS(2724),
    [sym_integer] = ACTIONS(2724),
    [sym_float] = ACTIONS(2724),
    [sym_char] = ACTIONS(2724),
    [anon_sym_true] = ACTIONS(2724),
    [anon_sym_false] = ACTIONS(2724),
    [anon_sym_nil] = ACTIONS(2724),
    [sym_atom] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2724),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2724),
    [anon_sym_LT] = ACTIONS(2724),
    [anon_sym_GT] = ACTIONS(2724),
    [anon_sym_PIPE] = ACTIONS(2724),
    [anon_sym_SLASH] = ACTIONS(2724),
    [anon_sym_TILDE] = ACTIONS(2724),
    [sym_keyword] = ACTIONS(2724),
    [anon_sym_LT_LT] = ACTIONS(2724),
    [anon_sym_PERCENT] = ACTIONS(2724),
    [anon_sym_DOT_DOT] = ACTIONS(2724),
    [anon_sym_AMP] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_BANG] = ACTIONS(2724),
    [anon_sym_CARET] = ACTIONS(2724),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2724),
    [anon_sym_not] = ACTIONS(2724),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_LT_DASH] = ACTIONS(2724),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2724),
    [anon_sym_when] = ACTIONS(2724),
    [anon_sym_COLON_COLON] = ACTIONS(2724),
    [anon_sym_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_or] = ACTIONS(2724),
    [anon_sym_AMP_AMP] = ACTIONS(2724),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2724),
    [anon_sym_and] = ACTIONS(2724),
    [anon_sym_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ] = ACTIONS(2724),
    [anon_sym_EQ_TILDE] = ACTIONS(2724),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2724),
    [anon_sym_LT_EQ] = ACTIONS(2724),
    [anon_sym_GT_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_LT] = ACTIONS(2724),
    [anon_sym_GT_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2724),
    [anon_sym_LT_GT] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2724),
    [anon_sym_STAR_STAR] = ACTIONS(2724),
    [anon_sym_DASH_GT] = ACTIONS(2724),
    [anon_sym_fn] = ACTIONS(2724),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2726),
    [sym__not_in] = ACTIONS(2726),
    [sym__quoted_atom_start] = ACTIONS(2726),
  },
  [STATE(1027)] = {
    [aux_sym__terminator_repeat1] = STATE(1028),
    [aux_sym__terminator_token1] = ACTIONS(2744),
    [anon_sym_SEMI] = ACTIONS(2746),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [aux_sym_identifier_token1] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [sym_alias] = ACTIONS(2717),
    [sym_integer] = ACTIONS(2717),
    [sym_float] = ACTIONS(2717),
    [sym_char] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [anon_sym_nil] = ACTIONS(2717),
    [sym_atom] = ACTIONS(2717),
    [anon_sym_DQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2717),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [sym_keyword] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_not] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_LT_DASH] = ACTIONS(2717),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2717),
    [anon_sym_when] = ACTIONS(2717),
    [anon_sym_COLON_COLON] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_LT] = ACTIONS(2717),
    [anon_sym_GT_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2717),
    [anon_sym_in] = ACTIONS(2717),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2717),
    [anon_sym_LT_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_end] = ACTIONS(2717),
    [anon_sym_fn] = ACTIONS(2717),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2719),
    [sym__not_in] = ACTIONS(2719),
    [sym__quoted_atom_start] = ACTIONS(2719),
  },
  [STATE(1028)] = {
    [aux_sym__terminator_repeat1] = STATE(1028),
    [aux_sym__terminator_token1] = ACTIONS(2748),
    [anon_sym_SEMI] = ACTIONS(2724),
    [anon_sym_LPAREN] = ACTIONS(2724),
    [aux_sym_identifier_token1] = ACTIONS(2724),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2724),
    [sym_alias] = ACTIONS(2724),
    [sym_integer] = ACTIONS(2724),
    [sym_float] = ACTIONS(2724),
    [sym_char] = ACTIONS(2724),
    [anon_sym_true] = ACTIONS(2724),
    [anon_sym_false] = ACTIONS(2724),
    [anon_sym_nil] = ACTIONS(2724),
    [sym_atom] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2724),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2724),
    [anon_sym_LT] = ACTIONS(2724),
    [anon_sym_GT] = ACTIONS(2724),
    [anon_sym_PIPE] = ACTIONS(2724),
    [anon_sym_SLASH] = ACTIONS(2724),
    [anon_sym_TILDE] = ACTIONS(2724),
    [sym_keyword] = ACTIONS(2724),
    [anon_sym_LT_LT] = ACTIONS(2724),
    [anon_sym_PERCENT] = ACTIONS(2724),
    [anon_sym_DOT_DOT] = ACTIONS(2724),
    [anon_sym_AMP] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_BANG] = ACTIONS(2724),
    [anon_sym_CARET] = ACTIONS(2724),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2724),
    [anon_sym_not] = ACTIONS(2724),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_LT_DASH] = ACTIONS(2724),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2724),
    [anon_sym_when] = ACTIONS(2724),
    [anon_sym_COLON_COLON] = ACTIONS(2724),
    [anon_sym_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_or] = ACTIONS(2724),
    [anon_sym_AMP_AMP] = ACTIONS(2724),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2724),
    [anon_sym_and] = ACTIONS(2724),
    [anon_sym_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ] = ACTIONS(2724),
    [anon_sym_EQ_TILDE] = ACTIONS(2724),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2724),
    [anon_sym_LT_EQ] = ACTIONS(2724),
    [anon_sym_GT_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_LT] = ACTIONS(2724),
    [anon_sym_GT_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2724),
    [anon_sym_LT_GT] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2724),
    [anon_sym_STAR_STAR] = ACTIONS(2724),
    [anon_sym_DASH_GT] = ACTIONS(2724),
    [anon_sym_end] = ACTIONS(2724),
    [anon_sym_fn] = ACTIONS(2724),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2726),
    [sym__not_in] = ACTIONS(2726),
    [sym__quoted_atom_start] = ACTIONS(2726),
  },
  [STATE(1029)] = {
    [aux_sym__terminator_repeat1] = STATE(1026),
    [aux_sym__terminator_token1] = ACTIONS(2751),
    [anon_sym_SEMI] = ACTIONS(2753),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [aux_sym_identifier_token1] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [sym_alias] = ACTIONS(2717),
    [sym_integer] = ACTIONS(2717),
    [sym_float] = ACTIONS(2717),
    [sym_char] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [anon_sym_nil] = ACTIONS(2717),
    [sym_atom] = ACTIONS(2717),
    [anon_sym_DQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2717),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [sym_keyword] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_not] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_LT_DASH] = ACTIONS(2717),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2717),
    [anon_sym_when] = ACTIONS(2717),
    [anon_sym_COLON_COLON] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_LT] = ACTIONS(2717),
    [anon_sym_GT_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2717),
    [anon_sym_in] = ACTIONS(2717),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2717),
    [anon_sym_LT_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_fn] = ACTIONS(2717),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2719),
    [sym__not_in] = ACTIONS(2719),
    [sym__quoted_atom_start] = ACTIONS(2719),
  },
  [STATE(1030)] = {
    [aux_sym__terminator_repeat1] = STATE(1031),
    [aux_sym__terminator_token1] = ACTIONS(2755),
    [anon_sym_SEMI] = ACTIONS(2757),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [aux_sym_identifier_token1] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [sym_alias] = ACTIONS(2717),
    [sym_integer] = ACTIONS(2717),
    [sym_float] = ACTIONS(2717),
    [sym_char] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [anon_sym_nil] = ACTIONS(2717),
    [sym_atom] = ACTIONS(2717),
    [anon_sym_DQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2717),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_not] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_LT_DASH] = ACTIONS(2717),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2717),
    [anon_sym_when] = ACTIONS(2717),
    [anon_sym_COLON_COLON] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_LT] = ACTIONS(2717),
    [anon_sym_GT_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2717),
    [anon_sym_in] = ACTIONS(2717),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2717),
    [anon_sym_LT_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_fn] = ACTIONS(2717),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2719),
    [sym__not_in] = ACTIONS(2719),
    [sym__quoted_atom_start] = ACTIONS(2719),
  },
  [STATE(1031)] = {
    [aux_sym__terminator_repeat1] = STATE(1031),
    [aux_sym__terminator_token1] = ACTIONS(2759),
    [anon_sym_SEMI] = ACTIONS(2724),
    [anon_sym_LPAREN] = ACTIONS(2724),
    [anon_sym_RPAREN] = ACTIONS(2724),
    [aux_sym_identifier_token1] = ACTIONS(2724),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2724),
    [sym_alias] = ACTIONS(2724),
    [sym_integer] = ACTIONS(2724),
    [sym_float] = ACTIONS(2724),
    [sym_char] = ACTIONS(2724),
    [anon_sym_true] = ACTIONS(2724),
    [anon_sym_false] = ACTIONS(2724),
    [anon_sym_nil] = ACTIONS(2724),
    [sym_atom] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2724),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2724),
    [anon_sym_LT] = ACTIONS(2724),
    [anon_sym_GT] = ACTIONS(2724),
    [anon_sym_PIPE] = ACTIONS(2724),
    [anon_sym_SLASH] = ACTIONS(2724),
    [anon_sym_TILDE] = ACTIONS(2724),
    [anon_sym_LT_LT] = ACTIONS(2724),
    [anon_sym_PERCENT] = ACTIONS(2724),
    [anon_sym_DOT_DOT] = ACTIONS(2724),
    [anon_sym_AMP] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_BANG] = ACTIONS(2724),
    [anon_sym_CARET] = ACTIONS(2724),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2724),
    [anon_sym_not] = ACTIONS(2724),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_LT_DASH] = ACTIONS(2724),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2724),
    [anon_sym_when] = ACTIONS(2724),
    [anon_sym_COLON_COLON] = ACTIONS(2724),
    [anon_sym_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_or] = ACTIONS(2724),
    [anon_sym_AMP_AMP] = ACTIONS(2724),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2724),
    [anon_sym_and] = ACTIONS(2724),
    [anon_sym_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ] = ACTIONS(2724),
    [anon_sym_EQ_TILDE] = ACTIONS(2724),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2724),
    [anon_sym_LT_EQ] = ACTIONS(2724),
    [anon_sym_GT_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_LT] = ACTIONS(2724),
    [anon_sym_GT_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2724),
    [anon_sym_LT_GT] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2724),
    [anon_sym_STAR_STAR] = ACTIONS(2724),
    [anon_sym_DASH_GT] = ACTIONS(2724),
    [anon_sym_fn] = ACTIONS(2724),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2726),
    [sym__not_in] = ACTIONS(2726),
    [sym__quoted_atom_start] = ACTIONS(2726),
  },
  [STATE(1032)] = {
    [aux_sym__terminator_repeat1] = STATE(1033),
    [aux_sym__terminator_token1] = ACTIONS(2762),
    [anon_sym_SEMI] = ACTIONS(2764),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [aux_sym_identifier_token1] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [sym_alias] = ACTIONS(2717),
    [sym_integer] = ACTIONS(2717),
    [sym_float] = ACTIONS(2717),
    [sym_char] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [anon_sym_nil] = ACTIONS(2717),
    [sym_atom] = ACTIONS(2717),
    [anon_sym_DQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2717),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_not] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_LT_DASH] = ACTIONS(2717),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2717),
    [anon_sym_when] = ACTIONS(2717),
    [anon_sym_COLON_COLON] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_LT] = ACTIONS(2717),
    [anon_sym_GT_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2717),
    [anon_sym_in] = ACTIONS(2717),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2717),
    [anon_sym_LT_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_end] = ACTIONS(2717),
    [anon_sym_fn] = ACTIONS(2717),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2719),
    [sym__not_in] = ACTIONS(2719),
    [sym__quoted_atom_start] = ACTIONS(2719),
  },
  [STATE(1033)] = {
    [aux_sym__terminator_repeat1] = STATE(1033),
    [aux_sym__terminator_token1] = ACTIONS(2766),
    [anon_sym_SEMI] = ACTIONS(2724),
    [anon_sym_LPAREN] = ACTIONS(2724),
    [aux_sym_identifier_token1] = ACTIONS(2724),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2724),
    [sym_alias] = ACTIONS(2724),
    [sym_integer] = ACTIONS(2724),
    [sym_float] = ACTIONS(2724),
    [sym_char] = ACTIONS(2724),
    [anon_sym_true] = ACTIONS(2724),
    [anon_sym_false] = ACTIONS(2724),
    [anon_sym_nil] = ACTIONS(2724),
    [sym_atom] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2724),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2724),
    [anon_sym_LT] = ACTIONS(2724),
    [anon_sym_GT] = ACTIONS(2724),
    [anon_sym_PIPE] = ACTIONS(2724),
    [anon_sym_SLASH] = ACTIONS(2724),
    [anon_sym_TILDE] = ACTIONS(2724),
    [anon_sym_LT_LT] = ACTIONS(2724),
    [anon_sym_PERCENT] = ACTIONS(2724),
    [anon_sym_DOT_DOT] = ACTIONS(2724),
    [anon_sym_AMP] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_BANG] = ACTIONS(2724),
    [anon_sym_CARET] = ACTIONS(2724),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2724),
    [anon_sym_not] = ACTIONS(2724),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_LT_DASH] = ACTIONS(2724),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2724),
    [anon_sym_when] = ACTIONS(2724),
    [anon_sym_COLON_COLON] = ACTIONS(2724),
    [anon_sym_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_or] = ACTIONS(2724),
    [anon_sym_AMP_AMP] = ACTIONS(2724),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2724),
    [anon_sym_and] = ACTIONS(2724),
    [anon_sym_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ] = ACTIONS(2724),
    [anon_sym_EQ_TILDE] = ACTIONS(2724),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2724),
    [anon_sym_LT_EQ] = ACTIONS(2724),
    [anon_sym_GT_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_LT] = ACTIONS(2724),
    [anon_sym_GT_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2724),
    [anon_sym_LT_GT] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2724),
    [anon_sym_STAR_STAR] = ACTIONS(2724),
    [anon_sym_DASH_GT] = ACTIONS(2724),
    [anon_sym_end] = ACTIONS(2724),
    [anon_sym_fn] = ACTIONS(2724),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2726),
    [sym__not_in] = ACTIONS(2726),
    [sym__quoted_atom_start] = ACTIONS(2726),
  },
  [STATE(1034)] = {
    [aux_sym__terminator_repeat1] = STATE(1034),
    [ts_builtin_sym_end] = ACTIONS(2726),
    [aux_sym__terminator_token1] = ACTIONS(2769),
    [anon_sym_SEMI] = ACTIONS(2724),
    [anon_sym_LPAREN] = ACTIONS(2724),
    [aux_sym_identifier_token1] = ACTIONS(2724),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2724),
    [sym_alias] = ACTIONS(2724),
    [sym_integer] = ACTIONS(2724),
    [sym_float] = ACTIONS(2724),
    [sym_char] = ACTIONS(2724),
    [anon_sym_true] = ACTIONS(2724),
    [anon_sym_false] = ACTIONS(2724),
    [anon_sym_nil] = ACTIONS(2724),
    [sym_atom] = ACTIONS(2724),
    [anon_sym_DQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE] = ACTIONS(2724),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2724),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2724),
    [anon_sym_LBRACE] = ACTIONS(2724),
    [anon_sym_LBRACK] = ACTIONS(2724),
    [anon_sym_LT] = ACTIONS(2724),
    [anon_sym_GT] = ACTIONS(2724),
    [anon_sym_PIPE] = ACTIONS(2724),
    [anon_sym_SLASH] = ACTIONS(2724),
    [anon_sym_TILDE] = ACTIONS(2724),
    [anon_sym_LT_LT] = ACTIONS(2724),
    [anon_sym_PERCENT] = ACTIONS(2724),
    [anon_sym_DOT_DOT] = ACTIONS(2724),
    [anon_sym_AMP] = ACTIONS(2724),
    [anon_sym_PLUS] = ACTIONS(2724),
    [anon_sym_DASH] = ACTIONS(2724),
    [anon_sym_BANG] = ACTIONS(2724),
    [anon_sym_CARET] = ACTIONS(2724),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2724),
    [anon_sym_not] = ACTIONS(2724),
    [anon_sym_AT] = ACTIONS(2724),
    [anon_sym_LT_DASH] = ACTIONS(2724),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2724),
    [anon_sym_when] = ACTIONS(2724),
    [anon_sym_COLON_COLON] = ACTIONS(2724),
    [anon_sym_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2724),
    [anon_sym_or] = ACTIONS(2724),
    [anon_sym_AMP_AMP] = ACTIONS(2724),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2724),
    [anon_sym_and] = ACTIONS(2724),
    [anon_sym_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ] = ACTIONS(2724),
    [anon_sym_EQ_TILDE] = ACTIONS(2724),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2724),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2724),
    [anon_sym_LT_EQ] = ACTIONS(2724),
    [anon_sym_GT_EQ] = ACTIONS(2724),
    [anon_sym_PIPE_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_LT] = ACTIONS(2724),
    [anon_sym_GT_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE] = ACTIONS(2724),
    [anon_sym_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2724),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2724),
    [anon_sym_in] = ACTIONS(2724),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH] = ACTIONS(2724),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2724),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2724),
    [anon_sym_LT_GT] = ACTIONS(2724),
    [anon_sym_STAR] = ACTIONS(2724),
    [anon_sym_STAR_STAR] = ACTIONS(2724),
    [anon_sym_DASH_GT] = ACTIONS(2724),
    [anon_sym_fn] = ACTIONS(2724),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2726),
    [sym__not_in] = ACTIONS(2726),
    [sym__quoted_atom_start] = ACTIONS(2726),
  },
  [STATE(1035)] = {
    [aux_sym__terminator_repeat1] = STATE(1034),
    [ts_builtin_sym_end] = ACTIONS(2719),
    [aux_sym__terminator_token1] = ACTIONS(2772),
    [anon_sym_SEMI] = ACTIONS(2774),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [aux_sym_identifier_token1] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [sym_alias] = ACTIONS(2717),
    [sym_integer] = ACTIONS(2717),
    [sym_float] = ACTIONS(2717),
    [sym_char] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [anon_sym_nil] = ACTIONS(2717),
    [sym_atom] = ACTIONS(2717),
    [anon_sym_DQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2717),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_not] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_LT_DASH] = ACTIONS(2717),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2717),
    [anon_sym_when] = ACTIONS(2717),
    [anon_sym_COLON_COLON] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_LT] = ACTIONS(2717),
    [anon_sym_GT_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2717),
    [anon_sym_in] = ACTIONS(2717),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2717),
    [anon_sym_LT_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_fn] = ACTIONS(2717),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2719),
    [sym__not_in] = ACTIONS(2719),
    [sym__quoted_atom_start] = ACTIONS(2719),
  },
  [STATE(1036)] = {
    [aux_sym__terminator_repeat1] = STATE(1031),
    [aux_sym__terminator_token1] = ACTIONS(2755),
    [anon_sym_SEMI] = ACTIONS(2776),
    [anon_sym_LPAREN] = ACTIONS(2717),
    [anon_sym_RPAREN] = ACTIONS(2717),
    [aux_sym_identifier_token1] = ACTIONS(2717),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2717),
    [sym_alias] = ACTIONS(2717),
    [sym_integer] = ACTIONS(2717),
    [sym_float] = ACTIONS(2717),
    [sym_char] = ACTIONS(2717),
    [anon_sym_true] = ACTIONS(2717),
    [anon_sym_false] = ACTIONS(2717),
    [anon_sym_nil] = ACTIONS(2717),
    [sym_atom] = ACTIONS(2717),
    [anon_sym_DQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE] = ACTIONS(2717),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2717),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2717),
    [anon_sym_LBRACE] = ACTIONS(2717),
    [anon_sym_LBRACK] = ACTIONS(2717),
    [anon_sym_LT] = ACTIONS(2717),
    [anon_sym_GT] = ACTIONS(2717),
    [anon_sym_PIPE] = ACTIONS(2717),
    [anon_sym_SLASH] = ACTIONS(2717),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_LT_LT] = ACTIONS(2717),
    [anon_sym_PERCENT] = ACTIONS(2717),
    [anon_sym_DOT_DOT] = ACTIONS(2717),
    [anon_sym_AMP] = ACTIONS(2717),
    [anon_sym_PLUS] = ACTIONS(2717),
    [anon_sym_DASH] = ACTIONS(2717),
    [anon_sym_BANG] = ACTIONS(2717),
    [anon_sym_CARET] = ACTIONS(2717),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2717),
    [anon_sym_not] = ACTIONS(2717),
    [anon_sym_AT] = ACTIONS(2717),
    [anon_sym_LT_DASH] = ACTIONS(2717),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2717),
    [anon_sym_when] = ACTIONS(2717),
    [anon_sym_COLON_COLON] = ACTIONS(2717),
    [anon_sym_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2717),
    [anon_sym_or] = ACTIONS(2717),
    [anon_sym_AMP_AMP] = ACTIONS(2717),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2717),
    [anon_sym_and] = ACTIONS(2717),
    [anon_sym_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ] = ACTIONS(2717),
    [anon_sym_EQ_TILDE] = ACTIONS(2717),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2717),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2717),
    [anon_sym_LT_EQ] = ACTIONS(2717),
    [anon_sym_GT_EQ] = ACTIONS(2717),
    [anon_sym_PIPE_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_LT] = ACTIONS(2717),
    [anon_sym_GT_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE] = ACTIONS(2717),
    [anon_sym_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2717),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2717),
    [anon_sym_in] = ACTIONS(2717),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH] = ACTIONS(2717),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2717),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2717),
    [anon_sym_LT_GT] = ACTIONS(2717),
    [anon_sym_STAR] = ACTIONS(2717),
    [anon_sym_STAR_STAR] = ACTIONS(2717),
    [anon_sym_DASH_GT] = ACTIONS(2717),
    [anon_sym_fn] = ACTIONS(2717),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2719),
    [sym__not_in] = ACTIONS(2719),
    [sym__quoted_atom_start] = ACTIONS(2719),
  },
  [STATE(1037)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2737),
    [anon_sym_RPAREN] = ACTIONS(2737),
    [aux_sym_identifier_token1] = ACTIONS(2737),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2737),
    [sym_alias] = ACTIONS(2737),
    [sym_integer] = ACTIONS(2737),
    [sym_float] = ACTIONS(2737),
    [sym_char] = ACTIONS(2737),
    [anon_sym_true] = ACTIONS(2737),
    [anon_sym_false] = ACTIONS(2737),
    [anon_sym_nil] = ACTIONS(2737),
    [sym_atom] = ACTIONS(2737),
    [anon_sym_DQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2737),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2737),
    [anon_sym_LBRACE] = ACTIONS(2737),
    [anon_sym_LBRACK] = ACTIONS(2737),
    [anon_sym_LT] = ACTIONS(2737),
    [anon_sym_GT] = ACTIONS(2737),
    [anon_sym_PIPE] = ACTIONS(2737),
    [anon_sym_SLASH] = ACTIONS(2737),
    [anon_sym_TILDE] = ACTIONS(2737),
    [sym_keyword] = ACTIONS(2737),
    [anon_sym_LT_LT] = ACTIONS(2737),
    [anon_sym_PERCENT] = ACTIONS(2737),
    [anon_sym_DOT_DOT] = ACTIONS(2737),
    [anon_sym_AMP] = ACTIONS(2737),
    [anon_sym_PLUS] = ACTIONS(2737),
    [anon_sym_DASH] = ACTIONS(2737),
    [anon_sym_BANG] = ACTIONS(2737),
    [anon_sym_CARET] = ACTIONS(2737),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2737),
    [anon_sym_not] = ACTIONS(2737),
    [anon_sym_AT] = ACTIONS(2737),
    [anon_sym_LT_DASH] = ACTIONS(2737),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2737),
    [anon_sym_when] = ACTIONS(2737),
    [anon_sym_COLON_COLON] = ACTIONS(2737),
    [anon_sym_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_or] = ACTIONS(2737),
    [anon_sym_AMP_AMP] = ACTIONS(2737),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2737),
    [anon_sym_and] = ACTIONS(2737),
    [anon_sym_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ] = ACTIONS(2737),
    [anon_sym_EQ_TILDE] = ACTIONS(2737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2737),
    [anon_sym_LT_EQ] = ACTIONS(2737),
    [anon_sym_GT_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_LT] = ACTIONS(2737),
    [anon_sym_GT_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2737),
    [anon_sym_in] = ACTIONS(2737),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2737),
    [anon_sym_LT_GT] = ACTIONS(2737),
    [anon_sym_STAR] = ACTIONS(2737),
    [anon_sym_STAR_STAR] = ACTIONS(2737),
    [anon_sym_DASH_GT] = ACTIONS(2737),
    [anon_sym_fn] = ACTIONS(2737),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2739),
    [sym__not_in] = ACTIONS(2739),
    [sym__quoted_atom_start] = ACTIONS(2739),
  },
  [STATE(1038)] = {
    [aux_sym__terminator_token1] = ACTIONS(2739),
    [anon_sym_SEMI] = ACTIONS(2737),
    [anon_sym_LPAREN] = ACTIONS(2737),
    [anon_sym_RPAREN] = ACTIONS(2737),
    [aux_sym_identifier_token1] = ACTIONS(2737),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2737),
    [sym_alias] = ACTIONS(2737),
    [sym_integer] = ACTIONS(2737),
    [sym_float] = ACTIONS(2737),
    [sym_char] = ACTIONS(2737),
    [anon_sym_true] = ACTIONS(2737),
    [anon_sym_false] = ACTIONS(2737),
    [anon_sym_nil] = ACTIONS(2737),
    [sym_atom] = ACTIONS(2737),
    [anon_sym_DQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2737),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2737),
    [anon_sym_LBRACE] = ACTIONS(2737),
    [anon_sym_LBRACK] = ACTIONS(2737),
    [anon_sym_LT] = ACTIONS(2737),
    [anon_sym_GT] = ACTIONS(2737),
    [anon_sym_PIPE] = ACTIONS(2737),
    [anon_sym_SLASH] = ACTIONS(2737),
    [anon_sym_TILDE] = ACTIONS(2737),
    [anon_sym_LT_LT] = ACTIONS(2737),
    [anon_sym_PERCENT] = ACTIONS(2737),
    [anon_sym_DOT_DOT] = ACTIONS(2737),
    [anon_sym_AMP] = ACTIONS(2737),
    [anon_sym_PLUS] = ACTIONS(2737),
    [anon_sym_DASH] = ACTIONS(2737),
    [anon_sym_BANG] = ACTIONS(2737),
    [anon_sym_CARET] = ACTIONS(2737),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2737),
    [anon_sym_not] = ACTIONS(2737),
    [anon_sym_AT] = ACTIONS(2737),
    [anon_sym_LT_DASH] = ACTIONS(2737),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2737),
    [anon_sym_when] = ACTIONS(2737),
    [anon_sym_COLON_COLON] = ACTIONS(2737),
    [anon_sym_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_or] = ACTIONS(2737),
    [anon_sym_AMP_AMP] = ACTIONS(2737),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2737),
    [anon_sym_and] = ACTIONS(2737),
    [anon_sym_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ] = ACTIONS(2737),
    [anon_sym_EQ_TILDE] = ACTIONS(2737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2737),
    [anon_sym_LT_EQ] = ACTIONS(2737),
    [anon_sym_GT_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_LT] = ACTIONS(2737),
    [anon_sym_GT_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2737),
    [anon_sym_in] = ACTIONS(2737),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2737),
    [anon_sym_LT_GT] = ACTIONS(2737),
    [anon_sym_STAR] = ACTIONS(2737),
    [anon_sym_STAR_STAR] = ACTIONS(2737),
    [anon_sym_DASH_GT] = ACTIONS(2737),
    [anon_sym_fn] = ACTIONS(2737),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2739),
    [sym__not_in] = ACTIONS(2739),
    [sym__quoted_atom_start] = ACTIONS(2739),
  },
  [STATE(1039)] = {
    [aux_sym__terminator_token1] = ACTIONS(2739),
    [anon_sym_SEMI] = ACTIONS(2737),
    [anon_sym_LPAREN] = ACTIONS(2737),
    [aux_sym_identifier_token1] = ACTIONS(2737),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2737),
    [sym_alias] = ACTIONS(2737),
    [sym_integer] = ACTIONS(2737),
    [sym_float] = ACTIONS(2737),
    [sym_char] = ACTIONS(2737),
    [anon_sym_true] = ACTIONS(2737),
    [anon_sym_false] = ACTIONS(2737),
    [anon_sym_nil] = ACTIONS(2737),
    [sym_atom] = ACTIONS(2737),
    [anon_sym_DQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2737),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2737),
    [anon_sym_LBRACE] = ACTIONS(2737),
    [anon_sym_LBRACK] = ACTIONS(2737),
    [anon_sym_LT] = ACTIONS(2737),
    [anon_sym_GT] = ACTIONS(2737),
    [anon_sym_PIPE] = ACTIONS(2737),
    [anon_sym_SLASH] = ACTIONS(2737),
    [anon_sym_TILDE] = ACTIONS(2737),
    [anon_sym_LT_LT] = ACTIONS(2737),
    [anon_sym_PERCENT] = ACTIONS(2737),
    [anon_sym_DOT_DOT] = ACTIONS(2737),
    [anon_sym_AMP] = ACTIONS(2737),
    [anon_sym_PLUS] = ACTIONS(2737),
    [anon_sym_DASH] = ACTIONS(2737),
    [anon_sym_BANG] = ACTIONS(2737),
    [anon_sym_CARET] = ACTIONS(2737),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2737),
    [anon_sym_not] = ACTIONS(2737),
    [anon_sym_AT] = ACTIONS(2737),
    [anon_sym_LT_DASH] = ACTIONS(2737),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2737),
    [anon_sym_when] = ACTIONS(2737),
    [anon_sym_COLON_COLON] = ACTIONS(2737),
    [anon_sym_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_or] = ACTIONS(2737),
    [anon_sym_AMP_AMP] = ACTIONS(2737),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2737),
    [anon_sym_and] = ACTIONS(2737),
    [anon_sym_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ] = ACTIONS(2737),
    [anon_sym_EQ_TILDE] = ACTIONS(2737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2737),
    [anon_sym_LT_EQ] = ACTIONS(2737),
    [anon_sym_GT_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_LT] = ACTIONS(2737),
    [anon_sym_GT_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2737),
    [anon_sym_in] = ACTIONS(2737),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2737),
    [anon_sym_LT_GT] = ACTIONS(2737),
    [anon_sym_STAR] = ACTIONS(2737),
    [anon_sym_STAR_STAR] = ACTIONS(2737),
    [anon_sym_DASH_GT] = ACTIONS(2737),
    [anon_sym_end] = ACTIONS(2737),
    [anon_sym_fn] = ACTIONS(2737),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2739),
    [sym__not_in] = ACTIONS(2739),
    [sym__quoted_atom_start] = ACTIONS(2739),
  },
  [STATE(1040)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2737),
    [aux_sym_identifier_token1] = ACTIONS(2737),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2737),
    [sym_alias] = ACTIONS(2737),
    [sym_integer] = ACTIONS(2737),
    [sym_float] = ACTIONS(2737),
    [sym_char] = ACTIONS(2737),
    [anon_sym_true] = ACTIONS(2737),
    [anon_sym_false] = ACTIONS(2737),
    [anon_sym_nil] = ACTIONS(2737),
    [sym_atom] = ACTIONS(2737),
    [anon_sym_DQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2737),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2737),
    [anon_sym_LBRACE] = ACTIONS(2737),
    [anon_sym_LBRACK] = ACTIONS(2737),
    [anon_sym_LT] = ACTIONS(2737),
    [anon_sym_GT] = ACTIONS(2737),
    [anon_sym_PIPE] = ACTIONS(2737),
    [anon_sym_SLASH] = ACTIONS(2737),
    [anon_sym_TILDE] = ACTIONS(2737),
    [sym_keyword] = ACTIONS(2737),
    [anon_sym_LT_LT] = ACTIONS(2737),
    [anon_sym_PERCENT] = ACTIONS(2737),
    [anon_sym_DOT_DOT] = ACTIONS(2737),
    [anon_sym_AMP] = ACTIONS(2737),
    [anon_sym_PLUS] = ACTIONS(2737),
    [anon_sym_DASH] = ACTIONS(2737),
    [anon_sym_BANG] = ACTIONS(2737),
    [anon_sym_CARET] = ACTIONS(2737),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2737),
    [anon_sym_not] = ACTIONS(2737),
    [anon_sym_AT] = ACTIONS(2737),
    [anon_sym_LT_DASH] = ACTIONS(2737),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2737),
    [anon_sym_when] = ACTIONS(2737),
    [anon_sym_COLON_COLON] = ACTIONS(2737),
    [anon_sym_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_or] = ACTIONS(2737),
    [anon_sym_AMP_AMP] = ACTIONS(2737),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2737),
    [anon_sym_and] = ACTIONS(2737),
    [anon_sym_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ] = ACTIONS(2737),
    [anon_sym_EQ_TILDE] = ACTIONS(2737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2737),
    [anon_sym_LT_EQ] = ACTIONS(2737),
    [anon_sym_GT_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_LT] = ACTIONS(2737),
    [anon_sym_GT_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2737),
    [anon_sym_in] = ACTIONS(2737),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2737),
    [anon_sym_LT_GT] = ACTIONS(2737),
    [anon_sym_STAR] = ACTIONS(2737),
    [anon_sym_STAR_STAR] = ACTIONS(2737),
    [anon_sym_DASH_GT] = ACTIONS(2737),
    [anon_sym_end] = ACTIONS(2737),
    [anon_sym_fn] = ACTIONS(2737),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2739),
    [sym__not_in] = ACTIONS(2739),
    [sym__quoted_atom_start] = ACTIONS(2739),
  },
  [STATE(1041)] = {
    [ts_builtin_sym_end] = ACTIONS(2739),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2737),
    [aux_sym_identifier_token1] = ACTIONS(2737),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2737),
    [sym_alias] = ACTIONS(2737),
    [sym_integer] = ACTIONS(2737),
    [sym_float] = ACTIONS(2737),
    [sym_char] = ACTIONS(2737),
    [anon_sym_true] = ACTIONS(2737),
    [anon_sym_false] = ACTIONS(2737),
    [anon_sym_nil] = ACTIONS(2737),
    [sym_atom] = ACTIONS(2737),
    [anon_sym_DQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2737),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2737),
    [anon_sym_LBRACE] = ACTIONS(2737),
    [anon_sym_LBRACK] = ACTIONS(2737),
    [anon_sym_LT] = ACTIONS(2737),
    [anon_sym_GT] = ACTIONS(2737),
    [anon_sym_PIPE] = ACTIONS(2737),
    [anon_sym_SLASH] = ACTIONS(2737),
    [anon_sym_TILDE] = ACTIONS(2737),
    [anon_sym_LT_LT] = ACTIONS(2737),
    [anon_sym_PERCENT] = ACTIONS(2737),
    [anon_sym_DOT_DOT] = ACTIONS(2737),
    [anon_sym_AMP] = ACTIONS(2737),
    [anon_sym_PLUS] = ACTIONS(2737),
    [anon_sym_DASH] = ACTIONS(2737),
    [anon_sym_BANG] = ACTIONS(2737),
    [anon_sym_CARET] = ACTIONS(2737),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2737),
    [anon_sym_not] = ACTIONS(2737),
    [anon_sym_AT] = ACTIONS(2737),
    [anon_sym_LT_DASH] = ACTIONS(2737),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2737),
    [anon_sym_when] = ACTIONS(2737),
    [anon_sym_COLON_COLON] = ACTIONS(2737),
    [anon_sym_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_or] = ACTIONS(2737),
    [anon_sym_AMP_AMP] = ACTIONS(2737),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2737),
    [anon_sym_and] = ACTIONS(2737),
    [anon_sym_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ] = ACTIONS(2737),
    [anon_sym_EQ_TILDE] = ACTIONS(2737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2737),
    [anon_sym_LT_EQ] = ACTIONS(2737),
    [anon_sym_GT_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_LT] = ACTIONS(2737),
    [anon_sym_GT_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2737),
    [anon_sym_in] = ACTIONS(2737),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2737),
    [anon_sym_LT_GT] = ACTIONS(2737),
    [anon_sym_STAR] = ACTIONS(2737),
    [anon_sym_STAR_STAR] = ACTIONS(2737),
    [anon_sym_DASH_GT] = ACTIONS(2737),
    [anon_sym_fn] = ACTIONS(2737),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2739),
    [sym__not_in] = ACTIONS(2739),
    [sym__quoted_atom_start] = ACTIONS(2739),
  },
  [STATE(1042)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2737),
    [anon_sym_RPAREN] = ACTIONS(2737),
    [aux_sym_identifier_token1] = ACTIONS(2737),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2737),
    [sym_alias] = ACTIONS(2737),
    [sym_integer] = ACTIONS(2737),
    [sym_float] = ACTIONS(2737),
    [sym_char] = ACTIONS(2737),
    [anon_sym_true] = ACTIONS(2737),
    [anon_sym_false] = ACTIONS(2737),
    [anon_sym_nil] = ACTIONS(2737),
    [sym_atom] = ACTIONS(2737),
    [anon_sym_DQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE] = ACTIONS(2737),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2737),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2737),
    [anon_sym_LBRACE] = ACTIONS(2737),
    [anon_sym_LBRACK] = ACTIONS(2737),
    [anon_sym_LT] = ACTIONS(2737),
    [anon_sym_GT] = ACTIONS(2737),
    [anon_sym_PIPE] = ACTIONS(2737),
    [anon_sym_SLASH] = ACTIONS(2737),
    [anon_sym_TILDE] = ACTIONS(2737),
    [anon_sym_LT_LT] = ACTIONS(2737),
    [anon_sym_PERCENT] = ACTIONS(2737),
    [anon_sym_DOT_DOT] = ACTIONS(2737),
    [anon_sym_AMP] = ACTIONS(2737),
    [anon_sym_PLUS] = ACTIONS(2737),
    [anon_sym_DASH] = ACTIONS(2737),
    [anon_sym_BANG] = ACTIONS(2737),
    [anon_sym_CARET] = ACTIONS(2737),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2737),
    [anon_sym_not] = ACTIONS(2737),
    [anon_sym_AT] = ACTIONS(2737),
    [anon_sym_LT_DASH] = ACTIONS(2737),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2737),
    [anon_sym_when] = ACTIONS(2737),
    [anon_sym_COLON_COLON] = ACTIONS(2737),
    [anon_sym_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2737),
    [anon_sym_or] = ACTIONS(2737),
    [anon_sym_AMP_AMP] = ACTIONS(2737),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2737),
    [anon_sym_and] = ACTIONS(2737),
    [anon_sym_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ] = ACTIONS(2737),
    [anon_sym_EQ_TILDE] = ACTIONS(2737),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2737),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2737),
    [anon_sym_LT_EQ] = ACTIONS(2737),
    [anon_sym_GT_EQ] = ACTIONS(2737),
    [anon_sym_PIPE_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_LT] = ACTIONS(2737),
    [anon_sym_GT_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE] = ACTIONS(2737),
    [anon_sym_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2737),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2737),
    [anon_sym_in] = ACTIONS(2737),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH] = ACTIONS(2737),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2737),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2737),
    [anon_sym_LT_GT] = ACTIONS(2737),
    [anon_sym_STAR] = ACTIONS(2737),
    [anon_sym_STAR_STAR] = ACTIONS(2737),
    [anon_sym_DASH_GT] = ACTIONS(2737),
    [anon_sym_fn] = ACTIONS(2737),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2739),
    [sym__not_in] = ACTIONS(2739),
    [sym__quoted_atom_start] = ACTIONS(2739),
  },
  [STATE(1043)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2778),
    [aux_sym_identifier_token1] = ACTIONS(2778),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2778),
    [sym_alias] = ACTIONS(2778),
    [sym_integer] = ACTIONS(2778),
    [sym_float] = ACTIONS(2778),
    [sym_char] = ACTIONS(2778),
    [anon_sym_true] = ACTIONS(2778),
    [anon_sym_false] = ACTIONS(2778),
    [anon_sym_nil] = ACTIONS(2778),
    [sym_atom] = ACTIONS(2778),
    [anon_sym_DQUOTE] = ACTIONS(2778),
    [anon_sym_SQUOTE] = ACTIONS(2778),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2778),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2778),
    [anon_sym_LBRACE] = ACTIONS(2778),
    [anon_sym_LBRACK] = ACTIONS(2778),
    [anon_sym_LT] = ACTIONS(2778),
    [anon_sym_GT] = ACTIONS(2778),
    [anon_sym_PIPE] = ACTIONS(2778),
    [anon_sym_SLASH] = ACTIONS(2778),
    [anon_sym_TILDE] = ACTIONS(2778),
    [anon_sym_LT_LT] = ACTIONS(2778),
    [anon_sym_PERCENT] = ACTIONS(2778),
    [anon_sym_DOT_DOT] = ACTIONS(2778),
    [anon_sym_AMP] = ACTIONS(2778),
    [anon_sym_PLUS] = ACTIONS(2778),
    [anon_sym_DASH] = ACTIONS(2778),
    [anon_sym_BANG] = ACTIONS(2778),
    [anon_sym_CARET] = ACTIONS(2778),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2778),
    [anon_sym_not] = ACTIONS(2778),
    [anon_sym_AT] = ACTIONS(2778),
    [anon_sym_LT_DASH] = ACTIONS(2778),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2778),
    [anon_sym_when] = ACTIONS(2778),
    [anon_sym_COLON_COLON] = ACTIONS(2778),
    [anon_sym_EQ] = ACTIONS(2778),
    [anon_sym_PIPE_PIPE] = ACTIONS(2778),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2778),
    [anon_sym_or] = ACTIONS(2778),
    [anon_sym_AMP_AMP] = ACTIONS(2778),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2778),
    [anon_sym_and] = ACTIONS(2778),
    [anon_sym_EQ_EQ] = ACTIONS(2778),
    [anon_sym_BANG_EQ] = ACTIONS(2778),
    [anon_sym_EQ_TILDE] = ACTIONS(2778),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2778),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2778),
    [anon_sym_LT_EQ] = ACTIONS(2778),
    [anon_sym_GT_EQ] = ACTIONS(2778),
    [anon_sym_PIPE_GT] = ACTIONS(2778),
    [anon_sym_LT_LT_LT] = ACTIONS(2778),
    [anon_sym_GT_GT_GT] = ACTIONS(2778),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2778),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2778),
    [anon_sym_LT_TILDE] = ACTIONS(2778),
    [anon_sym_TILDE_GT] = ACTIONS(2778),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2778),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2778),
    [anon_sym_in] = ACTIONS(2778),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2778),
    [anon_sym_PLUS_PLUS] = ACTIONS(2778),
    [anon_sym_DASH_DASH] = ACTIONS(2778),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2778),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2778),
    [anon_sym_LT_GT] = ACTIONS(2778),
    [anon_sym_STAR] = ACTIONS(2778),
    [anon_sym_STAR_STAR] = ACTIONS(2778),
    [anon_sym_DASH_GT] = ACTIONS(2778),
    [anon_sym_fn] = ACTIONS(2778),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2780),
    [sym__not_in] = ACTIONS(2780),
    [sym__quoted_atom_start] = ACTIONS(2780),
  },
  [STATE(1044)] = {
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2778),
    [aux_sym_identifier_token1] = ACTIONS(2778),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(2778),
    [sym_alias] = ACTIONS(2778),
    [sym_integer] = ACTIONS(2778),
    [sym_float] = ACTIONS(2778),
    [sym_char] = ACTIONS(2778),
    [anon_sym_true] = ACTIONS(2778),
    [anon_sym_false] = ACTIONS(2778),
    [anon_sym_nil] = ACTIONS(2778),
    [sym_atom] = ACTIONS(2778),
    [anon_sym_DQUOTE] = ACTIONS(2778),
    [anon_sym_SQUOTE] = ACTIONS(2778),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2778),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2778),
    [anon_sym_LBRACE] = ACTIONS(2778),
    [anon_sym_LBRACK] = ACTIONS(2778),
    [anon_sym_LT] = ACTIONS(2778),
    [anon_sym_GT] = ACTIONS(2778),
    [anon_sym_PIPE] = ACTIONS(2778),
    [anon_sym_SLASH] = ACTIONS(2778),
    [anon_sym_TILDE] = ACTIONS(2778),
    [anon_sym_LT_LT] = ACTIONS(2778),
    [anon_sym_PERCENT] = ACTIONS(2778),
    [anon_sym_DOT_DOT] = ACTIONS(2778),
    [anon_sym_AMP] = ACTIONS(2778),
    [anon_sym_PLUS] = ACTIONS(2778),
    [anon_sym_DASH] = ACTIONS(2778),
    [anon_sym_BANG] = ACTIONS(2778),
    [anon_sym_CARET] = ACTIONS(2778),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2778),
    [anon_sym_not] = ACTIONS(2778),
    [anon_sym_AT] = ACTIONS(2778),
    [anon_sym_LT_DASH] = ACTIONS(2778),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2778),
    [anon_sym_when] = ACTIONS(2778),
    [anon_sym_COLON_COLON] = ACTIONS(2778),
    [anon_sym_EQ] = ACTIONS(2778),
    [anon_sym_PIPE_PIPE] = ACTIONS(2778),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2778),
    [anon_sym_or] = ACTIONS(2778),
    [anon_sym_AMP_AMP] = ACTIONS(2778),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2778),
    [anon_sym_and] = ACTIONS(2778),
    [anon_sym_EQ_EQ] = ACTIONS(2778),
    [anon_sym_BANG_EQ] = ACTIONS(2778),
    [anon_sym_EQ_TILDE] = ACTIONS(2778),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2778),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2778),
    [anon_sym_LT_EQ] = ACTIONS(2778),
    [anon_sym_GT_EQ] = ACTIONS(2778),
    [anon_sym_PIPE_GT] = ACTIONS(2778),
    [anon_sym_LT_LT_LT] = ACTIONS(2778),
    [anon_sym_GT_GT_GT] = ACTIONS(2778),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2778),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2778),
    [anon_sym_LT_TILDE] = ACTIONS(2778),
    [anon_sym_TILDE_GT] = ACTIONS(2778),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2778),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2778),
    [anon_sym_in] = ACTIONS(2778),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2778),
    [anon_sym_PLUS_PLUS] = ACTIONS(2778),
    [anon_sym_DASH_DASH] = ACTIONS(2778),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2778),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2778),
    [anon_sym_LT_GT] = ACTIONS(2778),
    [anon_sym_STAR] = ACTIONS(2778),
    [anon_sym_STAR_STAR] = ACTIONS(2778),
    [anon_sym_DASH_GT] = ACTIONS(2778),
    [anon_sym_fn] = ACTIONS(2778),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__before_unary_op] = ACTIONS(2780),
    [sym__not_in] = ACTIONS(2780),
    [sym__quoted_atom_start] = ACTIONS(2780),
  },
  [STATE(1045)] = {
    [sym_identifier] = STATE(1012),
    [sym__quoted_i_double] = STATE(1015),
    [sym__quoted_i_single] = STATE(1008),
    [sym_tuple] = STATE(4531),
    [sym_operator_identifier] = STATE(1012),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2784),
    [anon_sym_true] = ACTIONS(2786),
    [anon_sym_false] = ACTIONS(2786),
    [anon_sym_nil] = ACTIONS(2786),
    [anon_sym_DQUOTE] = ACTIONS(2788),
    [anon_sym_SQUOTE] = ACTIONS(2790),
    [anon_sym_LBRACE] = ACTIONS(1175),
    [anon_sym_LT] = ACTIONS(2792),
    [anon_sym_GT] = ACTIONS(2792),
    [anon_sym_PIPE] = ACTIONS(2792),
    [anon_sym_SLASH] = ACTIONS(2792),
    [anon_sym_AMP] = ACTIONS(2794),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_BANG] = ACTIONS(2796),
    [anon_sym_CARET] = ACTIONS(2796),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2796),
    [anon_sym_not] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2800),
    [anon_sym_LT_DASH] = ACTIONS(2792),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2792),
    [anon_sym_when] = ACTIONS(2802),
    [anon_sym_COLON_COLON] = ACTIONS(2792),
    [anon_sym_EQ] = ACTIONS(2792),
    [anon_sym_PIPE_PIPE] = ACTIONS(2792),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2792),
    [anon_sym_or] = ACTIONS(2802),
    [anon_sym_AMP_AMP] = ACTIONS(2792),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2792),
    [anon_sym_and] = ACTIONS(2802),
    [anon_sym_EQ_EQ] = ACTIONS(2792),
    [anon_sym_BANG_EQ] = ACTIONS(2792),
    [anon_sym_EQ_TILDE] = ACTIONS(2792),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2792),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2792),
    [anon_sym_LT_EQ] = ACTIONS(2792),
    [anon_sym_GT_EQ] = ACTIONS(2792),
    [anon_sym_PIPE_GT] = ACTIONS(2792),
    [anon_sym_LT_LT_LT] = ACTIONS(2792),
    [anon_sym_GT_GT_GT] = ACTIONS(2792),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2792),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2792),
    [anon_sym_LT_TILDE] = ACTIONS(2792),
    [anon_sym_TILDE_GT] = ACTIONS(2792),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2792),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2792),
    [anon_sym_in] = ACTIONS(2802),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2792),
    [anon_sym_PLUS_PLUS] = ACTIONS(2792),
    [anon_sym_DASH_DASH] = ACTIONS(2792),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2792),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2792),
    [anon_sym_LT_GT] = ACTIONS(2792),
    [anon_sym_STAR] = ACTIONS(2792),
    [anon_sym_STAR_STAR] = ACTIONS(2792),
    [anon_sym_DASH_GT] = ACTIONS(2792),
    [anon_sym_after] = ACTIONS(2786),
    [anon_sym_catch] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2786),
    [anon_sym_else] = ACTIONS(2786),
    [anon_sym_end] = ACTIONS(2786),
    [anon_sym_fn] = ACTIONS(2786),
    [anon_sym_rescue] = ACTIONS(2786),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2804),
  },
  [STATE(1046)] = {
    [sym_identifier] = STATE(943),
    [sym__quoted_i_double] = STATE(944),
    [sym__quoted_i_single] = STATE(929),
    [sym_tuple] = STATE(3081),
    [sym_operator_identifier] = STATE(943),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2806),
    [anon_sym_true] = ACTIONS(2808),
    [anon_sym_false] = ACTIONS(2808),
    [anon_sym_nil] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2812),
    [anon_sym_LBRACE] = ACTIONS(1109),
    [anon_sym_LT] = ACTIONS(2814),
    [anon_sym_GT] = ACTIONS(2814),
    [anon_sym_PIPE] = ACTIONS(2814),
    [anon_sym_SLASH] = ACTIONS(2814),
    [anon_sym_AMP] = ACTIONS(2816),
    [anon_sym_PLUS] = ACTIONS(2818),
    [anon_sym_DASH] = ACTIONS(2818),
    [anon_sym_BANG] = ACTIONS(2818),
    [anon_sym_CARET] = ACTIONS(2818),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2818),
    [anon_sym_not] = ACTIONS(2820),
    [anon_sym_AT] = ACTIONS(2822),
    [anon_sym_LT_DASH] = ACTIONS(2814),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2814),
    [anon_sym_when] = ACTIONS(2824),
    [anon_sym_COLON_COLON] = ACTIONS(2814),
    [anon_sym_EQ] = ACTIONS(2814),
    [anon_sym_PIPE_PIPE] = ACTIONS(2814),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2814),
    [anon_sym_or] = ACTIONS(2824),
    [anon_sym_AMP_AMP] = ACTIONS(2814),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2814),
    [anon_sym_and] = ACTIONS(2824),
    [anon_sym_EQ_EQ] = ACTIONS(2814),
    [anon_sym_BANG_EQ] = ACTIONS(2814),
    [anon_sym_EQ_TILDE] = ACTIONS(2814),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2814),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2814),
    [anon_sym_LT_EQ] = ACTIONS(2814),
    [anon_sym_GT_EQ] = ACTIONS(2814),
    [anon_sym_PIPE_GT] = ACTIONS(2814),
    [anon_sym_LT_LT_LT] = ACTIONS(2814),
    [anon_sym_GT_GT_GT] = ACTIONS(2814),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2814),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2814),
    [anon_sym_LT_TILDE] = ACTIONS(2814),
    [anon_sym_TILDE_GT] = ACTIONS(2814),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2814),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2814),
    [anon_sym_in] = ACTIONS(2824),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_DASH_DASH] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2814),
    [anon_sym_LT_GT] = ACTIONS(2814),
    [anon_sym_STAR] = ACTIONS(2814),
    [anon_sym_STAR_STAR] = ACTIONS(2814),
    [anon_sym_DASH_GT] = ACTIONS(2814),
    [anon_sym_after] = ACTIONS(2808),
    [anon_sym_catch] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_else] = ACTIONS(2808),
    [anon_sym_end] = ACTIONS(2808),
    [anon_sym_fn] = ACTIONS(2808),
    [anon_sym_rescue] = ACTIONS(2808),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2826),
  },
  [STATE(1047)] = {
    [sym_identifier] = STATE(917),
    [sym__quoted_i_double] = STATE(913),
    [sym__quoted_i_single] = STATE(918),
    [sym_tuple] = STATE(1542),
    [sym_operator_identifier] = STATE(917),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2828),
    [anon_sym_true] = ACTIONS(2830),
    [anon_sym_false] = ACTIONS(2830),
    [anon_sym_nil] = ACTIONS(2830),
    [anon_sym_DQUOTE] = ACTIONS(2832),
    [anon_sym_SQUOTE] = ACTIONS(2834),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LT] = ACTIONS(2836),
    [anon_sym_GT] = ACTIONS(2836),
    [anon_sym_PIPE] = ACTIONS(2836),
    [anon_sym_SLASH] = ACTIONS(2836),
    [anon_sym_AMP] = ACTIONS(2838),
    [anon_sym_PLUS] = ACTIONS(2840),
    [anon_sym_DASH] = ACTIONS(2840),
    [anon_sym_BANG] = ACTIONS(2840),
    [anon_sym_CARET] = ACTIONS(2840),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2840),
    [anon_sym_not] = ACTIONS(2842),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_LT_DASH] = ACTIONS(2836),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2836),
    [anon_sym_when] = ACTIONS(2846),
    [anon_sym_COLON_COLON] = ACTIONS(2836),
    [anon_sym_EQ] = ACTIONS(2836),
    [anon_sym_PIPE_PIPE] = ACTIONS(2836),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2836),
    [anon_sym_or] = ACTIONS(2846),
    [anon_sym_AMP_AMP] = ACTIONS(2836),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2836),
    [anon_sym_and] = ACTIONS(2846),
    [anon_sym_EQ_EQ] = ACTIONS(2836),
    [anon_sym_BANG_EQ] = ACTIONS(2836),
    [anon_sym_EQ_TILDE] = ACTIONS(2836),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2836),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2836),
    [anon_sym_LT_EQ] = ACTIONS(2836),
    [anon_sym_GT_EQ] = ACTIONS(2836),
    [anon_sym_PIPE_GT] = ACTIONS(2836),
    [anon_sym_LT_LT_LT] = ACTIONS(2836),
    [anon_sym_GT_GT_GT] = ACTIONS(2836),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2836),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2836),
    [anon_sym_LT_TILDE] = ACTIONS(2836),
    [anon_sym_TILDE_GT] = ACTIONS(2836),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2836),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2836),
    [anon_sym_in] = ACTIONS(2846),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2836),
    [anon_sym_PLUS_PLUS] = ACTIONS(2836),
    [anon_sym_DASH_DASH] = ACTIONS(2836),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2836),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2836),
    [anon_sym_LT_GT] = ACTIONS(2836),
    [anon_sym_STAR] = ACTIONS(2836),
    [anon_sym_STAR_STAR] = ACTIONS(2836),
    [anon_sym_DASH_GT] = ACTIONS(2836),
    [anon_sym_after] = ACTIONS(2830),
    [anon_sym_catch] = ACTIONS(2830),
    [anon_sym_do] = ACTIONS(2830),
    [anon_sym_else] = ACTIONS(2830),
    [anon_sym_end] = ACTIONS(2830),
    [anon_sym_fn] = ACTIONS(2830),
    [anon_sym_rescue] = ACTIONS(2830),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2848),
  },
  [STATE(1048)] = {
    [sym_identifier] = STATE(943),
    [sym__quoted_i_double] = STATE(944),
    [sym__quoted_i_single] = STATE(929),
    [sym_tuple] = STATE(4275),
    [sym_operator_identifier] = STATE(943),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2850),
    [anon_sym_true] = ACTIONS(2808),
    [anon_sym_false] = ACTIONS(2808),
    [anon_sym_nil] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2812),
    [anon_sym_LBRACE] = ACTIONS(736),
    [anon_sym_LT] = ACTIONS(2814),
    [anon_sym_GT] = ACTIONS(2814),
    [anon_sym_PIPE] = ACTIONS(2814),
    [anon_sym_SLASH] = ACTIONS(2814),
    [anon_sym_AMP] = ACTIONS(2816),
    [anon_sym_PLUS] = ACTIONS(2818),
    [anon_sym_DASH] = ACTIONS(2818),
    [anon_sym_BANG] = ACTIONS(2818),
    [anon_sym_CARET] = ACTIONS(2818),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2818),
    [anon_sym_not] = ACTIONS(2820),
    [anon_sym_AT] = ACTIONS(2822),
    [anon_sym_LT_DASH] = ACTIONS(2814),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2814),
    [anon_sym_when] = ACTIONS(2824),
    [anon_sym_COLON_COLON] = ACTIONS(2814),
    [anon_sym_EQ] = ACTIONS(2814),
    [anon_sym_PIPE_PIPE] = ACTIONS(2814),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2814),
    [anon_sym_or] = ACTIONS(2824),
    [anon_sym_AMP_AMP] = ACTIONS(2814),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2814),
    [anon_sym_and] = ACTIONS(2824),
    [anon_sym_EQ_EQ] = ACTIONS(2814),
    [anon_sym_BANG_EQ] = ACTIONS(2814),
    [anon_sym_EQ_TILDE] = ACTIONS(2814),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2814),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2814),
    [anon_sym_LT_EQ] = ACTIONS(2814),
    [anon_sym_GT_EQ] = ACTIONS(2814),
    [anon_sym_PIPE_GT] = ACTIONS(2814),
    [anon_sym_LT_LT_LT] = ACTIONS(2814),
    [anon_sym_GT_GT_GT] = ACTIONS(2814),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2814),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2814),
    [anon_sym_LT_TILDE] = ACTIONS(2814),
    [anon_sym_TILDE_GT] = ACTIONS(2814),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2814),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2814),
    [anon_sym_in] = ACTIONS(2824),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_DASH_DASH] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2814),
    [anon_sym_LT_GT] = ACTIONS(2814),
    [anon_sym_STAR] = ACTIONS(2814),
    [anon_sym_STAR_STAR] = ACTIONS(2814),
    [anon_sym_DASH_GT] = ACTIONS(2814),
    [anon_sym_after] = ACTIONS(2808),
    [anon_sym_catch] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_else] = ACTIONS(2808),
    [anon_sym_end] = ACTIONS(2808),
    [anon_sym_fn] = ACTIONS(2808),
    [anon_sym_rescue] = ACTIONS(2808),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2826),
  },
  [STATE(1049)] = {
    [sym_identifier] = STATE(983),
    [sym__quoted_i_double] = STATE(984),
    [sym__quoted_i_single] = STATE(985),
    [sym_tuple] = STATE(2998),
    [sym_operator_identifier] = STATE(983),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(2852),
    [anon_sym_true] = ACTIONS(2854),
    [anon_sym_false] = ACTIONS(2854),
    [anon_sym_nil] = ACTIONS(2854),
    [anon_sym_DQUOTE] = ACTIONS(2856),
    [anon_sym_SQUOTE] = ACTIONS(2858),
    [anon_sym_LBRACE] = ACTIONS(449),
    [anon_sym_LT] = ACTIONS(2860),
    [anon_sym_GT] = ACTIONS(2860),
    [anon_sym_PIPE] = ACTIONS(2860),
    [anon_sym_SLASH] = ACTIONS(2860),
    [anon_sym_AMP] = ACTIONS(2862),
    [anon_sym_PLUS] = ACTIONS(2864),
    [anon_sym_DASH] = ACTIONS(2864),
    [anon_sym_BANG] = ACTIONS(2864),
    [anon_sym_CARET] = ACTIONS(2864),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2864),
    [anon_sym_not] = ACTIONS(2866),
    [anon_sym_AT] = ACTIONS(2868),
    [anon_sym_LT_DASH] = ACTIONS(2860),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2860),
    [anon_sym_when] = ACTIONS(2870),
    [anon_sym_COLON_COLON] = ACTIONS(2860),
    [anon_sym_EQ] = ACTIONS(2860),
    [anon_sym_PIPE_PIPE] = ACTIONS(2860),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2860),
    [anon_sym_or] = ACTIONS(2870),
    [anon_sym_AMP_AMP] = ACTIONS(2860),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2860),
    [anon_sym_and] = ACTIONS(2870),
    [anon_sym_EQ_EQ] = ACTIONS(2860),
    [anon_sym_BANG_EQ] = ACTIONS(2860),
    [anon_sym_EQ_TILDE] = ACTIONS(2860),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2860),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2860),
    [anon_sym_LT_EQ] = ACTIONS(2860),
    [anon_sym_GT_EQ] = ACTIONS(2860),
    [anon_sym_PIPE_GT] = ACTIONS(2860),
    [anon_sym_LT_LT_LT] = ACTIONS(2860),
    [anon_sym_GT_GT_GT] = ACTIONS(2860),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2860),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2860),
    [anon_sym_LT_TILDE] = ACTIONS(2860),
    [anon_sym_TILDE_GT] = ACTIONS(2860),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2860),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2860),
    [anon_sym_in] = ACTIONS(2870),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2860),
    [anon_sym_PLUS_PLUS] = ACTIONS(2860),
    [anon_sym_DASH_DASH] = ACTIONS(2860),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2860),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2860),
    [anon_sym_LT_GT] = ACTIONS(2860),
    [anon_sym_STAR] = ACTIONS(2860),
    [anon_sym_STAR_STAR] = ACTIONS(2860),
    [anon_sym_DASH_GT] = ACTIONS(2860),
    [anon_sym_after] = ACTIONS(2854),
    [anon_sym_catch] = ACTIONS(2854),
    [anon_sym_do] = ACTIONS(2854),
    [anon_sym_else] = ACTIONS(2854),
    [anon_sym_end] = ACTIONS(2854),
    [anon_sym_fn] = ACTIONS(2854),
    [anon_sym_rescue] = ACTIONS(2854),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2872),
  },
  [STATE(1050)] = {
    [sym_identifier] = STATE(980),
    [sym__quoted_i_double] = STATE(981),
    [sym__quoted_i_single] = STATE(982),
    [sym_tuple] = STATE(2038),
    [sym_operator_identifier] = STATE(980),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2874),
    [anon_sym_true] = ACTIONS(2876),
    [anon_sym_false] = ACTIONS(2876),
    [anon_sym_nil] = ACTIONS(2876),
    [anon_sym_DQUOTE] = ACTIONS(2878),
    [anon_sym_SQUOTE] = ACTIONS(2880),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LT] = ACTIONS(2882),
    [anon_sym_GT] = ACTIONS(2882),
    [anon_sym_PIPE] = ACTIONS(2882),
    [anon_sym_SLASH] = ACTIONS(2882),
    [anon_sym_AMP] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2886),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_CARET] = ACTIONS(2886),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2886),
    [anon_sym_not] = ACTIONS(2888),
    [anon_sym_AT] = ACTIONS(2890),
    [anon_sym_LT_DASH] = ACTIONS(2882),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2882),
    [anon_sym_when] = ACTIONS(2892),
    [anon_sym_COLON_COLON] = ACTIONS(2882),
    [anon_sym_EQ] = ACTIONS(2882),
    [anon_sym_PIPE_PIPE] = ACTIONS(2882),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2882),
    [anon_sym_or] = ACTIONS(2892),
    [anon_sym_AMP_AMP] = ACTIONS(2882),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2882),
    [anon_sym_and] = ACTIONS(2892),
    [anon_sym_EQ_EQ] = ACTIONS(2882),
    [anon_sym_BANG_EQ] = ACTIONS(2882),
    [anon_sym_EQ_TILDE] = ACTIONS(2882),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2882),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2882),
    [anon_sym_LT_EQ] = ACTIONS(2882),
    [anon_sym_GT_EQ] = ACTIONS(2882),
    [anon_sym_PIPE_GT] = ACTIONS(2882),
    [anon_sym_LT_LT_LT] = ACTIONS(2882),
    [anon_sym_GT_GT_GT] = ACTIONS(2882),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2882),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2882),
    [anon_sym_LT_TILDE] = ACTIONS(2882),
    [anon_sym_TILDE_GT] = ACTIONS(2882),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2882),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2882),
    [anon_sym_in] = ACTIONS(2892),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2882),
    [anon_sym_PLUS_PLUS] = ACTIONS(2882),
    [anon_sym_DASH_DASH] = ACTIONS(2882),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2882),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2882),
    [anon_sym_LT_GT] = ACTIONS(2882),
    [anon_sym_STAR] = ACTIONS(2882),
    [anon_sym_STAR_STAR] = ACTIONS(2882),
    [anon_sym_DASH_GT] = ACTIONS(2882),
    [anon_sym_after] = ACTIONS(2876),
    [anon_sym_catch] = ACTIONS(2876),
    [anon_sym_do] = ACTIONS(2876),
    [anon_sym_else] = ACTIONS(2876),
    [anon_sym_end] = ACTIONS(2876),
    [anon_sym_fn] = ACTIONS(2876),
    [anon_sym_rescue] = ACTIONS(2876),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2894),
  },
  [STATE(1051)] = {
    [sym_identifier] = STATE(980),
    [sym__quoted_i_double] = STATE(981),
    [sym__quoted_i_single] = STATE(982),
    [sym_tuple] = STATE(1203),
    [sym_operator_identifier] = STATE(980),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2896),
    [anon_sym_true] = ACTIONS(2876),
    [anon_sym_false] = ACTIONS(2876),
    [anon_sym_nil] = ACTIONS(2876),
    [anon_sym_DQUOTE] = ACTIONS(2878),
    [anon_sym_SQUOTE] = ACTIONS(2880),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LT] = ACTIONS(2882),
    [anon_sym_GT] = ACTIONS(2882),
    [anon_sym_PIPE] = ACTIONS(2882),
    [anon_sym_SLASH] = ACTIONS(2882),
    [anon_sym_AMP] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2886),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_CARET] = ACTIONS(2886),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2886),
    [anon_sym_not] = ACTIONS(2888),
    [anon_sym_AT] = ACTIONS(2890),
    [anon_sym_LT_DASH] = ACTIONS(2882),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2882),
    [anon_sym_when] = ACTIONS(2892),
    [anon_sym_COLON_COLON] = ACTIONS(2882),
    [anon_sym_EQ] = ACTIONS(2882),
    [anon_sym_PIPE_PIPE] = ACTIONS(2882),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2882),
    [anon_sym_or] = ACTIONS(2892),
    [anon_sym_AMP_AMP] = ACTIONS(2882),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2882),
    [anon_sym_and] = ACTIONS(2892),
    [anon_sym_EQ_EQ] = ACTIONS(2882),
    [anon_sym_BANG_EQ] = ACTIONS(2882),
    [anon_sym_EQ_TILDE] = ACTIONS(2882),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2882),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2882),
    [anon_sym_LT_EQ] = ACTIONS(2882),
    [anon_sym_GT_EQ] = ACTIONS(2882),
    [anon_sym_PIPE_GT] = ACTIONS(2882),
    [anon_sym_LT_LT_LT] = ACTIONS(2882),
    [anon_sym_GT_GT_GT] = ACTIONS(2882),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2882),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2882),
    [anon_sym_LT_TILDE] = ACTIONS(2882),
    [anon_sym_TILDE_GT] = ACTIONS(2882),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2882),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2882),
    [anon_sym_in] = ACTIONS(2892),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2882),
    [anon_sym_PLUS_PLUS] = ACTIONS(2882),
    [anon_sym_DASH_DASH] = ACTIONS(2882),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2882),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2882),
    [anon_sym_LT_GT] = ACTIONS(2882),
    [anon_sym_STAR] = ACTIONS(2882),
    [anon_sym_STAR_STAR] = ACTIONS(2882),
    [anon_sym_DASH_GT] = ACTIONS(2882),
    [anon_sym_after] = ACTIONS(2876),
    [anon_sym_catch] = ACTIONS(2876),
    [anon_sym_do] = ACTIONS(2876),
    [anon_sym_else] = ACTIONS(2876),
    [anon_sym_end] = ACTIONS(2876),
    [anon_sym_fn] = ACTIONS(2876),
    [anon_sym_rescue] = ACTIONS(2876),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2894),
  },
  [STATE(1052)] = {
    [sym_identifier] = STATE(983),
    [sym__quoted_i_double] = STATE(984),
    [sym__quoted_i_single] = STATE(985),
    [sym_tuple] = STATE(4212),
    [sym_operator_identifier] = STATE(983),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(433),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(433),
    [sym_alias] = ACTIONS(2898),
    [anon_sym_true] = ACTIONS(2854),
    [anon_sym_false] = ACTIONS(2854),
    [anon_sym_nil] = ACTIONS(2854),
    [anon_sym_DQUOTE] = ACTIONS(2856),
    [anon_sym_SQUOTE] = ACTIONS(2858),
    [anon_sym_LBRACE] = ACTIONS(31),
    [anon_sym_LT] = ACTIONS(2860),
    [anon_sym_GT] = ACTIONS(2860),
    [anon_sym_PIPE] = ACTIONS(2860),
    [anon_sym_SLASH] = ACTIONS(2860),
    [anon_sym_AMP] = ACTIONS(2862),
    [anon_sym_PLUS] = ACTIONS(2864),
    [anon_sym_DASH] = ACTIONS(2864),
    [anon_sym_BANG] = ACTIONS(2864),
    [anon_sym_CARET] = ACTIONS(2864),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2864),
    [anon_sym_not] = ACTIONS(2866),
    [anon_sym_AT] = ACTIONS(2868),
    [anon_sym_LT_DASH] = ACTIONS(2860),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2860),
    [anon_sym_when] = ACTIONS(2870),
    [anon_sym_COLON_COLON] = ACTIONS(2860),
    [anon_sym_EQ] = ACTIONS(2860),
    [anon_sym_PIPE_PIPE] = ACTIONS(2860),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2860),
    [anon_sym_or] = ACTIONS(2870),
    [anon_sym_AMP_AMP] = ACTIONS(2860),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2860),
    [anon_sym_and] = ACTIONS(2870),
    [anon_sym_EQ_EQ] = ACTIONS(2860),
    [anon_sym_BANG_EQ] = ACTIONS(2860),
    [anon_sym_EQ_TILDE] = ACTIONS(2860),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2860),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2860),
    [anon_sym_LT_EQ] = ACTIONS(2860),
    [anon_sym_GT_EQ] = ACTIONS(2860),
    [anon_sym_PIPE_GT] = ACTIONS(2860),
    [anon_sym_LT_LT_LT] = ACTIONS(2860),
    [anon_sym_GT_GT_GT] = ACTIONS(2860),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2860),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2860),
    [anon_sym_LT_TILDE] = ACTIONS(2860),
    [anon_sym_TILDE_GT] = ACTIONS(2860),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2860),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2860),
    [anon_sym_in] = ACTIONS(2870),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2860),
    [anon_sym_PLUS_PLUS] = ACTIONS(2860),
    [anon_sym_DASH_DASH] = ACTIONS(2860),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2860),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2860),
    [anon_sym_LT_GT] = ACTIONS(2860),
    [anon_sym_STAR] = ACTIONS(2860),
    [anon_sym_STAR_STAR] = ACTIONS(2860),
    [anon_sym_DASH_GT] = ACTIONS(2860),
    [anon_sym_after] = ACTIONS(2854),
    [anon_sym_catch] = ACTIONS(2854),
    [anon_sym_do] = ACTIONS(2854),
    [anon_sym_else] = ACTIONS(2854),
    [anon_sym_end] = ACTIONS(2854),
    [anon_sym_fn] = ACTIONS(2854),
    [anon_sym_rescue] = ACTIONS(2854),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2872),
  },
  [STATE(1053)] = {
    [sym_identifier] = STATE(1012),
    [sym__quoted_i_double] = STATE(1015),
    [sym__quoted_i_single] = STATE(1008),
    [sym_tuple] = STATE(3739),
    [sym_operator_identifier] = STATE(1012),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(542),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(542),
    [sym_alias] = ACTIONS(2900),
    [anon_sym_true] = ACTIONS(2786),
    [anon_sym_false] = ACTIONS(2786),
    [anon_sym_nil] = ACTIONS(2786),
    [anon_sym_DQUOTE] = ACTIONS(2788),
    [anon_sym_SQUOTE] = ACTIONS(2790),
    [anon_sym_LBRACE] = ACTIONS(558),
    [anon_sym_LT] = ACTIONS(2792),
    [anon_sym_GT] = ACTIONS(2792),
    [anon_sym_PIPE] = ACTIONS(2792),
    [anon_sym_SLASH] = ACTIONS(2792),
    [anon_sym_AMP] = ACTIONS(2794),
    [anon_sym_PLUS] = ACTIONS(2796),
    [anon_sym_DASH] = ACTIONS(2796),
    [anon_sym_BANG] = ACTIONS(2796),
    [anon_sym_CARET] = ACTIONS(2796),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2796),
    [anon_sym_not] = ACTIONS(2798),
    [anon_sym_AT] = ACTIONS(2800),
    [anon_sym_LT_DASH] = ACTIONS(2792),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2792),
    [anon_sym_when] = ACTIONS(2802),
    [anon_sym_COLON_COLON] = ACTIONS(2792),
    [anon_sym_EQ] = ACTIONS(2792),
    [anon_sym_PIPE_PIPE] = ACTIONS(2792),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2792),
    [anon_sym_or] = ACTIONS(2802),
    [anon_sym_AMP_AMP] = ACTIONS(2792),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2792),
    [anon_sym_and] = ACTIONS(2802),
    [anon_sym_EQ_EQ] = ACTIONS(2792),
    [anon_sym_BANG_EQ] = ACTIONS(2792),
    [anon_sym_EQ_TILDE] = ACTIONS(2792),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2792),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2792),
    [anon_sym_LT_EQ] = ACTIONS(2792),
    [anon_sym_GT_EQ] = ACTIONS(2792),
    [anon_sym_PIPE_GT] = ACTIONS(2792),
    [anon_sym_LT_LT_LT] = ACTIONS(2792),
    [anon_sym_GT_GT_GT] = ACTIONS(2792),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2792),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2792),
    [anon_sym_LT_TILDE] = ACTIONS(2792),
    [anon_sym_TILDE_GT] = ACTIONS(2792),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2792),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2792),
    [anon_sym_in] = ACTIONS(2802),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2792),
    [anon_sym_PLUS_PLUS] = ACTIONS(2792),
    [anon_sym_DASH_DASH] = ACTIONS(2792),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2792),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2792),
    [anon_sym_LT_GT] = ACTIONS(2792),
    [anon_sym_STAR] = ACTIONS(2792),
    [anon_sym_STAR_STAR] = ACTIONS(2792),
    [anon_sym_DASH_GT] = ACTIONS(2792),
    [anon_sym_after] = ACTIONS(2786),
    [anon_sym_catch] = ACTIONS(2786),
    [anon_sym_do] = ACTIONS(2786),
    [anon_sym_else] = ACTIONS(2786),
    [anon_sym_end] = ACTIONS(2786),
    [anon_sym_fn] = ACTIONS(2786),
    [anon_sym_rescue] = ACTIONS(2786),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2804),
  },
  [STATE(1054)] = {
    [sym_identifier] = STATE(943),
    [sym__quoted_i_double] = STATE(944),
    [sym__quoted_i_single] = STATE(929),
    [sym_tuple] = STATE(2315),
    [sym_operator_identifier] = STATE(943),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2902),
    [anon_sym_true] = ACTIONS(2808),
    [anon_sym_false] = ACTIONS(2808),
    [anon_sym_nil] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2812),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_LT] = ACTIONS(2814),
    [anon_sym_GT] = ACTIONS(2814),
    [anon_sym_PIPE] = ACTIONS(2814),
    [anon_sym_SLASH] = ACTIONS(2814),
    [anon_sym_AMP] = ACTIONS(2816),
    [anon_sym_PLUS] = ACTIONS(2818),
    [anon_sym_DASH] = ACTIONS(2818),
    [anon_sym_BANG] = ACTIONS(2818),
    [anon_sym_CARET] = ACTIONS(2818),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2818),
    [anon_sym_not] = ACTIONS(2820),
    [anon_sym_AT] = ACTIONS(2822),
    [anon_sym_LT_DASH] = ACTIONS(2814),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2814),
    [anon_sym_when] = ACTIONS(2824),
    [anon_sym_COLON_COLON] = ACTIONS(2814),
    [anon_sym_EQ] = ACTIONS(2814),
    [anon_sym_PIPE_PIPE] = ACTIONS(2814),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2814),
    [anon_sym_or] = ACTIONS(2824),
    [anon_sym_AMP_AMP] = ACTIONS(2814),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2814),
    [anon_sym_and] = ACTIONS(2824),
    [anon_sym_EQ_EQ] = ACTIONS(2814),
    [anon_sym_BANG_EQ] = ACTIONS(2814),
    [anon_sym_EQ_TILDE] = ACTIONS(2814),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2814),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2814),
    [anon_sym_LT_EQ] = ACTIONS(2814),
    [anon_sym_GT_EQ] = ACTIONS(2814),
    [anon_sym_PIPE_GT] = ACTIONS(2814),
    [anon_sym_LT_LT_LT] = ACTIONS(2814),
    [anon_sym_GT_GT_GT] = ACTIONS(2814),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2814),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2814),
    [anon_sym_LT_TILDE] = ACTIONS(2814),
    [anon_sym_TILDE_GT] = ACTIONS(2814),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2814),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2814),
    [anon_sym_in] = ACTIONS(2824),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_DASH_DASH] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2814),
    [anon_sym_LT_GT] = ACTIONS(2814),
    [anon_sym_STAR] = ACTIONS(2814),
    [anon_sym_STAR_STAR] = ACTIONS(2814),
    [anon_sym_DASH_GT] = ACTIONS(2814),
    [anon_sym_after] = ACTIONS(2808),
    [anon_sym_catch] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_else] = ACTIONS(2808),
    [anon_sym_end] = ACTIONS(2808),
    [anon_sym_fn] = ACTIONS(2808),
    [anon_sym_rescue] = ACTIONS(2808),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2826),
  },
  [STATE(1055)] = {
    [sym_identifier] = STATE(980),
    [sym__quoted_i_double] = STATE(981),
    [sym__quoted_i_single] = STATE(982),
    [sym_tuple] = STATE(1686),
    [sym_operator_identifier] = STATE(980),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2904),
    [anon_sym_true] = ACTIONS(2876),
    [anon_sym_false] = ACTIONS(2876),
    [anon_sym_nil] = ACTIONS(2876),
    [anon_sym_DQUOTE] = ACTIONS(2878),
    [anon_sym_SQUOTE] = ACTIONS(2880),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LT] = ACTIONS(2882),
    [anon_sym_GT] = ACTIONS(2882),
    [anon_sym_PIPE] = ACTIONS(2882),
    [anon_sym_SLASH] = ACTIONS(2882),
    [anon_sym_AMP] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2886),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_CARET] = ACTIONS(2886),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2886),
    [anon_sym_not] = ACTIONS(2888),
    [anon_sym_AT] = ACTIONS(2890),
    [anon_sym_LT_DASH] = ACTIONS(2882),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2882),
    [anon_sym_when] = ACTIONS(2892),
    [anon_sym_COLON_COLON] = ACTIONS(2882),
    [anon_sym_EQ] = ACTIONS(2882),
    [anon_sym_PIPE_PIPE] = ACTIONS(2882),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2882),
    [anon_sym_or] = ACTIONS(2892),
    [anon_sym_AMP_AMP] = ACTIONS(2882),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2882),
    [anon_sym_and] = ACTIONS(2892),
    [anon_sym_EQ_EQ] = ACTIONS(2882),
    [anon_sym_BANG_EQ] = ACTIONS(2882),
    [anon_sym_EQ_TILDE] = ACTIONS(2882),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2882),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2882),
    [anon_sym_LT_EQ] = ACTIONS(2882),
    [anon_sym_GT_EQ] = ACTIONS(2882),
    [anon_sym_PIPE_GT] = ACTIONS(2882),
    [anon_sym_LT_LT_LT] = ACTIONS(2882),
    [anon_sym_GT_GT_GT] = ACTIONS(2882),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2882),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2882),
    [anon_sym_LT_TILDE] = ACTIONS(2882),
    [anon_sym_TILDE_GT] = ACTIONS(2882),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2882),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2882),
    [anon_sym_in] = ACTIONS(2892),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2882),
    [anon_sym_PLUS_PLUS] = ACTIONS(2882),
    [anon_sym_DASH_DASH] = ACTIONS(2882),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2882),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2882),
    [anon_sym_LT_GT] = ACTIONS(2882),
    [anon_sym_STAR] = ACTIONS(2882),
    [anon_sym_STAR_STAR] = ACTIONS(2882),
    [anon_sym_DASH_GT] = ACTIONS(2882),
    [anon_sym_after] = ACTIONS(2876),
    [anon_sym_catch] = ACTIONS(2876),
    [anon_sym_do] = ACTIONS(2876),
    [anon_sym_else] = ACTIONS(2876),
    [anon_sym_end] = ACTIONS(2876),
    [anon_sym_fn] = ACTIONS(2876),
    [anon_sym_rescue] = ACTIONS(2876),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2894),
  },
  [STATE(1056)] = {
    [sym_identifier] = STATE(950),
    [sym__quoted_i_double] = STATE(951),
    [sym__quoted_i_single] = STATE(952),
    [sym_tuple] = STATE(2038),
    [sym_operator_identifier] = STATE(950),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2874),
    [anon_sym_true] = ACTIONS(2906),
    [anon_sym_false] = ACTIONS(2906),
    [anon_sym_nil] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(2908),
    [anon_sym_SQUOTE] = ACTIONS(2910),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LT] = ACTIONS(2912),
    [anon_sym_GT] = ACTIONS(2912),
    [anon_sym_PIPE] = ACTIONS(2912),
    [anon_sym_SLASH] = ACTIONS(2912),
    [anon_sym_AMP] = ACTIONS(2914),
    [anon_sym_PLUS] = ACTIONS(2916),
    [anon_sym_DASH] = ACTIONS(2916),
    [anon_sym_BANG] = ACTIONS(2916),
    [anon_sym_CARET] = ACTIONS(2916),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2916),
    [anon_sym_not] = ACTIONS(2918),
    [anon_sym_AT] = ACTIONS(2920),
    [anon_sym_LT_DASH] = ACTIONS(2912),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2912),
    [anon_sym_when] = ACTIONS(2922),
    [anon_sym_COLON_COLON] = ACTIONS(2912),
    [anon_sym_EQ] = ACTIONS(2912),
    [anon_sym_PIPE_PIPE] = ACTIONS(2912),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2912),
    [anon_sym_or] = ACTIONS(2922),
    [anon_sym_AMP_AMP] = ACTIONS(2912),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2912),
    [anon_sym_and] = ACTIONS(2922),
    [anon_sym_EQ_EQ] = ACTIONS(2912),
    [anon_sym_BANG_EQ] = ACTIONS(2912),
    [anon_sym_EQ_TILDE] = ACTIONS(2912),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2912),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2912),
    [anon_sym_LT_EQ] = ACTIONS(2912),
    [anon_sym_GT_EQ] = ACTIONS(2912),
    [anon_sym_PIPE_GT] = ACTIONS(2912),
    [anon_sym_LT_LT_LT] = ACTIONS(2912),
    [anon_sym_GT_GT_GT] = ACTIONS(2912),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2912),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2912),
    [anon_sym_LT_TILDE] = ACTIONS(2912),
    [anon_sym_TILDE_GT] = ACTIONS(2912),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2912),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2912),
    [anon_sym_in] = ACTIONS(2922),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2912),
    [anon_sym_PLUS_PLUS] = ACTIONS(2912),
    [anon_sym_DASH_DASH] = ACTIONS(2912),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2912),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2912),
    [anon_sym_LT_GT] = ACTIONS(2912),
    [anon_sym_STAR] = ACTIONS(2912),
    [anon_sym_STAR_STAR] = ACTIONS(2912),
    [anon_sym_DASH_GT] = ACTIONS(2912),
    [anon_sym_after] = ACTIONS(2906),
    [anon_sym_catch] = ACTIONS(2906),
    [anon_sym_do] = ACTIONS(2906),
    [anon_sym_else] = ACTIONS(2906),
    [anon_sym_end] = ACTIONS(2906),
    [anon_sym_fn] = ACTIONS(2906),
    [anon_sym_rescue] = ACTIONS(2906),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2924),
  },
  [STATE(1057)] = {
    [sym_identifier] = STATE(943),
    [sym__quoted_i_double] = STATE(944),
    [sym__quoted_i_single] = STATE(929),
    [sym_tuple] = STATE(3399),
    [sym_operator_identifier] = STATE(943),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(361),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(361),
    [sym_alias] = ACTIONS(2926),
    [anon_sym_true] = ACTIONS(2808),
    [anon_sym_false] = ACTIONS(2808),
    [anon_sym_nil] = ACTIONS(2808),
    [anon_sym_DQUOTE] = ACTIONS(2810),
    [anon_sym_SQUOTE] = ACTIONS(2812),
    [anon_sym_LBRACE] = ACTIONS(606),
    [anon_sym_LT] = ACTIONS(2814),
    [anon_sym_GT] = ACTIONS(2814),
    [anon_sym_PIPE] = ACTIONS(2814),
    [anon_sym_SLASH] = ACTIONS(2814),
    [anon_sym_AMP] = ACTIONS(2816),
    [anon_sym_PLUS] = ACTIONS(2818),
    [anon_sym_DASH] = ACTIONS(2818),
    [anon_sym_BANG] = ACTIONS(2818),
    [anon_sym_CARET] = ACTIONS(2818),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2818),
    [anon_sym_not] = ACTIONS(2820),
    [anon_sym_AT] = ACTIONS(2822),
    [anon_sym_LT_DASH] = ACTIONS(2814),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2814),
    [anon_sym_when] = ACTIONS(2824),
    [anon_sym_COLON_COLON] = ACTIONS(2814),
    [anon_sym_EQ] = ACTIONS(2814),
    [anon_sym_PIPE_PIPE] = ACTIONS(2814),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2814),
    [anon_sym_or] = ACTIONS(2824),
    [anon_sym_AMP_AMP] = ACTIONS(2814),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2814),
    [anon_sym_and] = ACTIONS(2824),
    [anon_sym_EQ_EQ] = ACTIONS(2814),
    [anon_sym_BANG_EQ] = ACTIONS(2814),
    [anon_sym_EQ_TILDE] = ACTIONS(2814),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2814),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2814),
    [anon_sym_LT_EQ] = ACTIONS(2814),
    [anon_sym_GT_EQ] = ACTIONS(2814),
    [anon_sym_PIPE_GT] = ACTIONS(2814),
    [anon_sym_LT_LT_LT] = ACTIONS(2814),
    [anon_sym_GT_GT_GT] = ACTIONS(2814),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2814),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2814),
    [anon_sym_LT_TILDE] = ACTIONS(2814),
    [anon_sym_TILDE_GT] = ACTIONS(2814),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2814),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2814),
    [anon_sym_in] = ACTIONS(2824),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_DASH_DASH] = ACTIONS(2814),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2814),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2814),
    [anon_sym_LT_GT] = ACTIONS(2814),
    [anon_sym_STAR] = ACTIONS(2814),
    [anon_sym_STAR_STAR] = ACTIONS(2814),
    [anon_sym_DASH_GT] = ACTIONS(2814),
    [anon_sym_after] = ACTIONS(2808),
    [anon_sym_catch] = ACTIONS(2808),
    [anon_sym_do] = ACTIONS(2808),
    [anon_sym_else] = ACTIONS(2808),
    [anon_sym_end] = ACTIONS(2808),
    [anon_sym_fn] = ACTIONS(2808),
    [anon_sym_rescue] = ACTIONS(2808),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2826),
  },
  [STATE(1058)] = {
    [sym_identifier] = STATE(917),
    [sym__quoted_i_double] = STATE(913),
    [sym__quoted_i_single] = STATE(918),
    [sym_tuple] = STATE(1686),
    [sym_operator_identifier] = STATE(917),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2904),
    [anon_sym_true] = ACTIONS(2830),
    [anon_sym_false] = ACTIONS(2830),
    [anon_sym_nil] = ACTIONS(2830),
    [anon_sym_DQUOTE] = ACTIONS(2832),
    [anon_sym_SQUOTE] = ACTIONS(2834),
    [anon_sym_LBRACE] = ACTIONS(81),
    [anon_sym_LT] = ACTIONS(2836),
    [anon_sym_GT] = ACTIONS(2836),
    [anon_sym_PIPE] = ACTIONS(2836),
    [anon_sym_SLASH] = ACTIONS(2836),
    [anon_sym_AMP] = ACTIONS(2838),
    [anon_sym_PLUS] = ACTIONS(2840),
    [anon_sym_DASH] = ACTIONS(2840),
    [anon_sym_BANG] = ACTIONS(2840),
    [anon_sym_CARET] = ACTIONS(2840),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2840),
    [anon_sym_not] = ACTIONS(2842),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_LT_DASH] = ACTIONS(2836),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2836),
    [anon_sym_when] = ACTIONS(2846),
    [anon_sym_COLON_COLON] = ACTIONS(2836),
    [anon_sym_EQ] = ACTIONS(2836),
    [anon_sym_PIPE_PIPE] = ACTIONS(2836),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2836),
    [anon_sym_or] = ACTIONS(2846),
    [anon_sym_AMP_AMP] = ACTIONS(2836),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2836),
    [anon_sym_and] = ACTIONS(2846),
    [anon_sym_EQ_EQ] = ACTIONS(2836),
    [anon_sym_BANG_EQ] = ACTIONS(2836),
    [anon_sym_EQ_TILDE] = ACTIONS(2836),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2836),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2836),
    [anon_sym_LT_EQ] = ACTIONS(2836),
    [anon_sym_GT_EQ] = ACTIONS(2836),
    [anon_sym_PIPE_GT] = ACTIONS(2836),
    [anon_sym_LT_LT_LT] = ACTIONS(2836),
    [anon_sym_GT_GT_GT] = ACTIONS(2836),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2836),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2836),
    [anon_sym_LT_TILDE] = ACTIONS(2836),
    [anon_sym_TILDE_GT] = ACTIONS(2836),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2836),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2836),
    [anon_sym_in] = ACTIONS(2846),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2836),
    [anon_sym_PLUS_PLUS] = ACTIONS(2836),
    [anon_sym_DASH_DASH] = ACTIONS(2836),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2836),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2836),
    [anon_sym_LT_GT] = ACTIONS(2836),
    [anon_sym_STAR] = ACTIONS(2836),
    [anon_sym_STAR_STAR] = ACTIONS(2836),
    [anon_sym_DASH_GT] = ACTIONS(2836),
    [anon_sym_after] = ACTIONS(2830),
    [anon_sym_catch] = ACTIONS(2830),
    [anon_sym_do] = ACTIONS(2830),
    [anon_sym_else] = ACTIONS(2830),
    [anon_sym_end] = ACTIONS(2830),
    [anon_sym_fn] = ACTIONS(2830),
    [anon_sym_rescue] = ACTIONS(2830),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2848),
  },
  [STATE(1059)] = {
    [sym_identifier] = STATE(980),
    [sym__quoted_i_double] = STATE(981),
    [sym__quoted_i_single] = STATE(982),
    [sym_tuple] = STATE(1542),
    [sym_operator_identifier] = STATE(980),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(475),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(475),
    [sym_alias] = ACTIONS(2828),
    [anon_sym_true] = ACTIONS(2876),
    [anon_sym_false] = ACTIONS(2876),
    [anon_sym_nil] = ACTIONS(2876),
    [anon_sym_DQUOTE] = ACTIONS(2878),
    [anon_sym_SQUOTE] = ACTIONS(2880),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LT] = ACTIONS(2882),
    [anon_sym_GT] = ACTIONS(2882),
    [anon_sym_PIPE] = ACTIONS(2882),
    [anon_sym_SLASH] = ACTIONS(2882),
    [anon_sym_AMP] = ACTIONS(2884),
    [anon_sym_PLUS] = ACTIONS(2886),
    [anon_sym_DASH] = ACTIONS(2886),
    [anon_sym_BANG] = ACTIONS(2886),
    [anon_sym_CARET] = ACTIONS(2886),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2886),
    [anon_sym_not] = ACTIONS(2888),
    [anon_sym_AT] = ACTIONS(2890),
    [anon_sym_LT_DASH] = ACTIONS(2882),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2882),
    [anon_sym_when] = ACTIONS(2892),
    [anon_sym_COLON_COLON] = ACTIONS(2882),
    [anon_sym_EQ] = ACTIONS(2882),
    [anon_sym_PIPE_PIPE] = ACTIONS(2882),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2882),
    [anon_sym_or] = ACTIONS(2892),
    [anon_sym_AMP_AMP] = ACTIONS(2882),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2882),
    [anon_sym_and] = ACTIONS(2892),
    [anon_sym_EQ_EQ] = ACTIONS(2882),
    [anon_sym_BANG_EQ] = ACTIONS(2882),
    [anon_sym_EQ_TILDE] = ACTIONS(2882),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2882),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2882),
    [anon_sym_LT_EQ] = ACTIONS(2882),
    [anon_sym_GT_EQ] = ACTIONS(2882),
    [anon_sym_PIPE_GT] = ACTIONS(2882),
    [anon_sym_LT_LT_LT] = ACTIONS(2882),
    [anon_sym_GT_GT_GT] = ACTIONS(2882),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2882),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2882),
    [anon_sym_LT_TILDE] = ACTIONS(2882),
    [anon_sym_TILDE_GT] = ACTIONS(2882),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2882),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2882),
    [anon_sym_in] = ACTIONS(2892),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2882),
    [anon_sym_PLUS_PLUS] = ACTIONS(2882),
    [anon_sym_DASH_DASH] = ACTIONS(2882),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2882),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2882),
    [anon_sym_LT_GT] = ACTIONS(2882),
    [anon_sym_STAR] = ACTIONS(2882),
    [anon_sym_STAR_STAR] = ACTIONS(2882),
    [anon_sym_DASH_GT] = ACTIONS(2882),
    [anon_sym_after] = ACTIONS(2876),
    [anon_sym_catch] = ACTIONS(2876),
    [anon_sym_do] = ACTIONS(2876),
    [anon_sym_else] = ACTIONS(2876),
    [anon_sym_end] = ACTIONS(2876),
    [anon_sym_fn] = ACTIONS(2876),
    [anon_sym_rescue] = ACTIONS(2876),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2894),
  },
  [STATE(1060)] = {
    [sym_identifier] = STATE(950),
    [sym__quoted_i_double] = STATE(951),
    [sym__quoted_i_single] = STATE(952),
    [sym_tuple] = STATE(1542),
    [sym_operator_identifier] = STATE(950),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(508),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(508),
    [sym_alias] = ACTIONS(2828),
    [anon_sym_true] = ACTIONS(2906),
    [anon_sym_false] = ACTIONS(2906),
    [anon_sym_nil] = ACTIONS(2906),
    [anon_sym_DQUOTE] = ACTIONS(2908),
    [anon_sym_SQUOTE] = ACTIONS(2910),
    [anon_sym_LBRACE] = ACTIONS(207),
    [anon_sym_LT] = ACTIONS(2912),
    [anon_sym_GT] = ACTIONS(2912),
    [anon_sym_PIPE] = ACTIONS(2912),
    [anon_sym_SLASH] = ACTIONS(2912),
    [anon_sym_AMP] = ACTIONS(2914),
    [anon_sym_PLUS] = ACTIONS(2916),
    [anon_sym_DASH] = ACTIONS(2916),
    [anon_sym_BANG] = ACTIONS(2916),
    [anon_sym_CARET] = ACTIONS(2916),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2916),
    [anon_sym_not] = ACTIONS(2918),
    [anon_sym_AT] = ACTIONS(2920),
    [anon_sym_LT_DASH] = ACTIONS(2912),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2912),
    [anon_sym_when] = ACTIONS(2922),
    [anon_sym_COLON_COLON] = ACTIONS(2912),
    [anon_sym_EQ] = ACTIONS(2912),
    [anon_sym_PIPE_PIPE] = ACTIONS(2912),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2912),
    [anon_sym_or] = ACTIONS(2922),
    [anon_sym_AMP_AMP] = ACTIONS(2912),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2912),
    [anon_sym_and] = ACTIONS(2922),
    [anon_sym_EQ_EQ] = ACTIONS(2912),
    [anon_sym_BANG_EQ] = ACTIONS(2912),
    [anon_sym_EQ_TILDE] = ACTIONS(2912),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2912),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2912),
    [anon_sym_LT_EQ] = ACTIONS(2912),
    [anon_sym_GT_EQ] = ACTIONS(2912),
    [anon_sym_PIPE_GT] = ACTIONS(2912),
    [anon_sym_LT_LT_LT] = ACTIONS(2912),
    [anon_sym_GT_GT_GT] = ACTIONS(2912),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2912),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2912),
    [anon_sym_LT_TILDE] = ACTIONS(2912),
    [anon_sym_TILDE_GT] = ACTIONS(2912),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2912),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2912),
    [anon_sym_in] = ACTIONS(2922),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2912),
    [anon_sym_PLUS_PLUS] = ACTIONS(2912),
    [anon_sym_DASH_DASH] = ACTIONS(2912),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2912),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2912),
    [anon_sym_LT_GT] = ACTIONS(2912),
    [anon_sym_STAR] = ACTIONS(2912),
    [anon_sym_STAR_STAR] = ACTIONS(2912),
    [anon_sym_DASH_GT] = ACTIONS(2912),
    [anon_sym_after] = ACTIONS(2906),
    [anon_sym_catch] = ACTIONS(2906),
    [anon_sym_do] = ACTIONS(2906),
    [anon_sym_else] = ACTIONS(2906),
    [anon_sym_end] = ACTIONS(2906),
    [anon_sym_fn] = ACTIONS(2906),
    [anon_sym_rescue] = ACTIONS(2906),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2924),
  },
  [STATE(1061)] = {
    [sym_identifier] = STATE(917),
    [sym__quoted_i_double] = STATE(913),
    [sym__quoted_i_single] = STATE(918),
    [sym_tuple] = STATE(2038),
    [sym_operator_identifier] = STATE(917),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2874),
    [anon_sym_true] = ACTIONS(2830),
    [anon_sym_false] = ACTIONS(2830),
    [anon_sym_nil] = ACTIONS(2830),
    [anon_sym_DQUOTE] = ACTIONS(2832),
    [anon_sym_SQUOTE] = ACTIONS(2834),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_LT] = ACTIONS(2836),
    [anon_sym_GT] = ACTIONS(2836),
    [anon_sym_PIPE] = ACTIONS(2836),
    [anon_sym_SLASH] = ACTIONS(2836),
    [anon_sym_AMP] = ACTIONS(2838),
    [anon_sym_PLUS] = ACTIONS(2840),
    [anon_sym_DASH] = ACTIONS(2840),
    [anon_sym_BANG] = ACTIONS(2840),
    [anon_sym_CARET] = ACTIONS(2840),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2840),
    [anon_sym_not] = ACTIONS(2842),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_LT_DASH] = ACTIONS(2836),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2836),
    [anon_sym_when] = ACTIONS(2846),
    [anon_sym_COLON_COLON] = ACTIONS(2836),
    [anon_sym_EQ] = ACTIONS(2836),
    [anon_sym_PIPE_PIPE] = ACTIONS(2836),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2836),
    [anon_sym_or] = ACTIONS(2846),
    [anon_sym_AMP_AMP] = ACTIONS(2836),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2836),
    [anon_sym_and] = ACTIONS(2846),
    [anon_sym_EQ_EQ] = ACTIONS(2836),
    [anon_sym_BANG_EQ] = ACTIONS(2836),
    [anon_sym_EQ_TILDE] = ACTIONS(2836),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2836),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2836),
    [anon_sym_LT_EQ] = ACTIONS(2836),
    [anon_sym_GT_EQ] = ACTIONS(2836),
    [anon_sym_PIPE_GT] = ACTIONS(2836),
    [anon_sym_LT_LT_LT] = ACTIONS(2836),
    [anon_sym_GT_GT_GT] = ACTIONS(2836),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2836),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2836),
    [anon_sym_LT_TILDE] = ACTIONS(2836),
    [anon_sym_TILDE_GT] = ACTIONS(2836),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2836),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2836),
    [anon_sym_in] = ACTIONS(2846),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2836),
    [anon_sym_PLUS_PLUS] = ACTIONS(2836),
    [anon_sym_DASH_DASH] = ACTIONS(2836),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2836),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2836),
    [anon_sym_LT_GT] = ACTIONS(2836),
    [anon_sym_STAR] = ACTIONS(2836),
    [anon_sym_STAR_STAR] = ACTIONS(2836),
    [anon_sym_DASH_GT] = ACTIONS(2836),
    [anon_sym_after] = ACTIONS(2830),
    [anon_sym_catch] = ACTIONS(2830),
    [anon_sym_do] = ACTIONS(2830),
    [anon_sym_else] = ACTIONS(2830),
    [anon_sym_end] = ACTIONS(2830),
    [anon_sym_fn] = ACTIONS(2830),
    [anon_sym_rescue] = ACTIONS(2830),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2848),
  },
  [STATE(1062)] = {
    [sym_identifier] = STATE(917),
    [sym__quoted_i_double] = STATE(913),
    [sym__quoted_i_single] = STATE(918),
    [sym_tuple] = STATE(1203),
    [sym_operator_identifier] = STATE(917),
    [aux_sym__terminator_token1] = ACTIONS(3),
    [anon_sym_LPAREN] = ACTIONS(2782),
    [aux_sym_identifier_token1] = ACTIONS(191),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(191),
    [sym_alias] = ACTIONS(2896),
    [anon_sym_true] = ACTIONS(2830),
    [anon_sym_false] = ACTIONS(2830),
    [anon_sym_nil] = ACTIONS(2830),
    [anon_sym_DQUOTE] = ACTIONS(2832),
    [anon_sym_SQUOTE] = ACTIONS(2834),
    [anon_sym_LBRACE] = ACTIONS(253),
    [anon_sym_LT] = ACTIONS(2836),
    [anon_sym_GT] = ACTIONS(2836),
    [anon_sym_PIPE] = ACTIONS(2836),
    [anon_sym_SLASH] = ACTIONS(2836),
    [anon_sym_AMP] = ACTIONS(2838),
    [anon_sym_PLUS] = ACTIONS(2840),
    [anon_sym_DASH] = ACTIONS(2840),
    [anon_sym_BANG] = ACTIONS(2840),
    [anon_sym_CARET] = ACTIONS(2840),
    [anon_sym_TILDE_TILDE_TILDE] = ACTIONS(2840),
    [anon_sym_not] = ACTIONS(2842),
    [anon_sym_AT] = ACTIONS(2844),
    [anon_sym_LT_DASH] = ACTIONS(2836),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2836),
    [anon_sym_when] = ACTIONS(2846),
    [anon_sym_COLON_COLON] = ACTIONS(2836),
    [anon_sym_EQ] = ACTIONS(2836),
    [anon_sym_PIPE_PIPE] = ACTIONS(2836),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2836),
    [anon_sym_or] = ACTIONS(2846),
    [anon_sym_AMP_AMP] = ACTIONS(2836),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2836),
    [anon_sym_and] = ACTIONS(2846),
    [anon_sym_EQ_EQ] = ACTIONS(2836),
    [anon_sym_BANG_EQ] = ACTIONS(2836),
    [anon_sym_EQ_TILDE] = ACTIONS(2836),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2836),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2836),
    [anon_sym_LT_EQ] = ACTIONS(2836),
    [anon_sym_GT_EQ] = ACTIONS(2836),
    [anon_sym_PIPE_GT] = ACTIONS(2836),
    [anon_sym_LT_LT_LT] = ACTIONS(2836),
    [anon_sym_GT_GT_GT] = ACTIONS(2836),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2836),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2836),
    [anon_sym_LT_TILDE] = ACTIONS(2836),
    [anon_sym_TILDE_GT] = ACTIONS(2836),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2836),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2836),
    [anon_sym_in] = ACTIONS(2846),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2836),
    [anon_sym_PLUS_PLUS] = ACTIONS(2836),
    [anon_sym_DASH_DASH] = ACTIONS(2836),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2836),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2836),
    [anon_sym_LT_GT] = ACTIONS(2836),
    [anon_sym_STAR] = ACTIONS(2836),
    [anon_sym_STAR_STAR] = ACTIONS(2836),
    [anon_sym_DASH_GT] = ACTIONS(2836),
    [anon_sym_after] = ACTIONS(2830),
    [anon_sym_catch] = ACTIONS(2830),
    [anon_sym_do] = ACTIONS(2830),
    [anon_sym_else] = ACTIONS(2830),
    [anon_sym_end] = ACTIONS(2830),
    [anon_sym_fn] = ACTIONS(2830),
    [anon_sym_rescue] = ACTIONS(2830),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2848),
  },
  [STATE(1063)] = {
    [sym__terminator] = STATE(171),
    [sym_after_block] = STATE(4746),
    [sym_rescue_block] = STATE(4746),
    [sym_catch_block] = STATE(4746),
    [sym_else_block] = STATE(4746),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4637),
    [aux_sym_do_block_repeat1] = STATE(4746),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2930),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1068),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1064)] = {
    [sym__terminator] = STATE(145),
    [sym_after_block] = STATE(4762),
    [sym_rescue_block] = STATE(4762),
    [sym_catch_block] = STATE(4762),
    [sym_else_block] = STATE(4762),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4700),
    [aux_sym_do_block_repeat1] = STATE(4762),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2979),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1016),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1065)] = {
    [sym__terminator] = STATE(170),
    [sym_after_block] = STATE(4713),
    [sym_rescue_block] = STATE(4713),
    [sym_catch_block] = STATE(4713),
    [sym_else_block] = STATE(4713),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4662),
    [aux_sym_do_block_repeat1] = STATE(4713),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2981),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(311),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1066)] = {
    [sym__terminator] = STATE(165),
    [sym_after_block] = STATE(4735),
    [sym_rescue_block] = STATE(4735),
    [sym_catch_block] = STATE(4735),
    [sym_else_block] = STATE(4735),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4674),
    [aux_sym_do_block_repeat1] = STATE(4735),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2983),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1056),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1067)] = {
    [sym__terminator] = STATE(154),
    [sym_after_block] = STATE(4709),
    [sym_rescue_block] = STATE(4709),
    [sym_catch_block] = STATE(4709),
    [sym_else_block] = STATE(4709),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4660),
    [aux_sym_do_block_repeat1] = STATE(4709),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2985),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1034),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1068)] = {
    [sym__terminator] = STATE(160),
    [sym_after_block] = STATE(4721),
    [sym_rescue_block] = STATE(4721),
    [sym_catch_block] = STATE(4721),
    [sym_else_block] = STATE(4721),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4669),
    [aux_sym_do_block_repeat1] = STATE(4721),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2987),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1046),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1069)] = {
    [sym__terminator] = STATE(144),
    [sym_after_block] = STATE(4749),
    [sym_rescue_block] = STATE(4749),
    [sym_catch_block] = STATE(4749),
    [sym_else_block] = STATE(4749),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4694),
    [aux_sym_do_block_repeat1] = STATE(4749),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2989),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(315),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1070)] = {
    [sym__terminator] = STATE(153),
    [sym_after_block] = STATE(4707),
    [sym_rescue_block] = STATE(4707),
    [sym_catch_block] = STATE(4707),
    [sym_else_block] = STATE(4707),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4658),
    [aux_sym_do_block_repeat1] = STATE(4707),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2991),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(331),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1071)] = {
    [sym__terminator] = STATE(164),
    [sym_after_block] = STATE(4725),
    [sym_rescue_block] = STATE(4725),
    [sym_catch_block] = STATE(4725),
    [sym_else_block] = STATE(4725),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4672),
    [aux_sym_do_block_repeat1] = STATE(4725),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2993),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(335),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1072)] = {
    [sym__terminator] = STATE(159),
    [sym_after_block] = STATE(4719),
    [sym_rescue_block] = STATE(4719),
    [sym_catch_block] = STATE(4719),
    [sym_else_block] = STATE(4719),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4667),
    [aux_sym_do_block_repeat1] = STATE(4719),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2995),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(343),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1073)] = {
    [sym__terminator] = STATE(162),
    [sym_after_block] = STATE(4729),
    [sym_rescue_block] = STATE(4729),
    [sym_catch_block] = STATE(4729),
    [sym_else_block] = STATE(4729),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4678),
    [aux_sym_do_block_repeat1] = STATE(4729),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2997),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(347),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1074)] = {
    [sym__terminator] = STATE(140),
    [sym_after_block] = STATE(4732),
    [sym_rescue_block] = STATE(4732),
    [sym_catch_block] = STATE(4732),
    [sym_else_block] = STATE(4732),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4680),
    [aux_sym_do_block_repeat1] = STATE(4732),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(2999),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1052),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1075)] = {
    [sym__terminator] = STATE(141),
    [sym_after_block] = STATE(4731),
    [sym_rescue_block] = STATE(4731),
    [sym_catch_block] = STATE(4731),
    [sym_else_block] = STATE(4731),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4634),
    [aux_sym_do_block_repeat1] = STATE(4731),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3001),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(327),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1076)] = {
    [sym__terminator] = STATE(147),
    [sym_after_block] = STATE(4715),
    [sym_rescue_block] = STATE(4715),
    [sym_catch_block] = STATE(4715),
    [sym_else_block] = STATE(4715),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4645),
    [aux_sym_do_block_repeat1] = STATE(4715),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3003),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(323),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1077)] = {
    [sym__terminator] = STATE(142),
    [sym_after_block] = STATE(4708),
    [sym_rescue_block] = STATE(4708),
    [sym_catch_block] = STATE(4708),
    [sym_else_block] = STATE(4708),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4636),
    [aux_sym_do_block_repeat1] = STATE(4708),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3005),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1010),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1078)] = {
    [sym__terminator] = STATE(167),
    [sym_after_block] = STATE(4738),
    [sym_rescue_block] = STATE(4738),
    [sym_catch_block] = STATE(4738),
    [sym_else_block] = STATE(4738),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4685),
    [aux_sym_do_block_repeat1] = STATE(4738),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3007),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(351),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1079)] = {
    [sym__terminator] = STATE(168),
    [sym_after_block] = STATE(4741),
    [sym_rescue_block] = STATE(4741),
    [sym_catch_block] = STATE(4741),
    [sym_else_block] = STATE(4741),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4687),
    [aux_sym_do_block_repeat1] = STATE(4741),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3009),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1062),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1080)] = {
    [sym__terminator] = STATE(156),
    [sym_after_block] = STATE(4752),
    [sym_rescue_block] = STATE(4752),
    [sym_catch_block] = STATE(4752),
    [sym_else_block] = STATE(4752),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4640),
    [aux_sym_do_block_repeat1] = STATE(4752),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3011),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(319),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1081)] = {
    [sym__terminator] = STATE(173),
    [sym_after_block] = STATE(4750),
    [sym_rescue_block] = STATE(4750),
    [sym_catch_block] = STATE(4750),
    [sym_else_block] = STATE(4750),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4692),
    [aux_sym_do_block_repeat1] = STATE(4750),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3013),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(307),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1082)] = {
    [sym__terminator] = STATE(174),
    [sym_after_block] = STATE(4754),
    [sym_rescue_block] = STATE(4754),
    [sym_catch_block] = STATE(4754),
    [sym_else_block] = STATE(4754),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4696),
    [aux_sym_do_block_repeat1] = STATE(4754),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3015),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1074),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1083)] = {
    [sym__terminator] = STATE(148),
    [sym_after_block] = STATE(4720),
    [sym_rescue_block] = STATE(4720),
    [sym_catch_block] = STATE(4720),
    [sym_else_block] = STATE(4720),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4647),
    [aux_sym_do_block_repeat1] = STATE(4720),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3017),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1022),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1084)] = {
    [sym__terminator] = STATE(157),
    [sym_after_block] = STATE(4730),
    [sym_rescue_block] = STATE(4730),
    [sym_catch_block] = STATE(4730),
    [sym_else_block] = STATE(4730),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4643),
    [aux_sym_do_block_repeat1] = STATE(4730),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3019),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1040),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1085)] = {
    [sym__terminator] = STATE(150),
    [sym_after_block] = STATE(4751),
    [sym_rescue_block] = STATE(4751),
    [sym_catch_block] = STATE(4751),
    [sym_else_block] = STATE(4751),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4652),
    [aux_sym_do_block_repeat1] = STATE(4751),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3021),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(339),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1086)] = {
    [sym__terminator] = STATE(151),
    [sym_after_block] = STATE(4755),
    [sym_rescue_block] = STATE(4755),
    [sym_catch_block] = STATE(4755),
    [sym_else_block] = STATE(4755),
    [aux_sym__terminator_repeat1] = STATE(1022),
    [aux_sym_block_repeat2] = STATE(4654),
    [aux_sym_do_block_repeat1] = STATE(4755),
    [aux_sym__stab_clause_arguments_without_parentheses_repeat1] = STATE(6177),
    [aux_sym__terminator_token1] = ACTIONS(2928),
    [anon_sym_SEMI] = ACTIONS(3023),
    [anon_sym_LT] = ACTIONS(2932),
    [anon_sym_GT] = ACTIONS(2932),
    [anon_sym_PIPE] = ACTIONS(2934),
    [anon_sym_SLASH] = ACTIONS(2936),
    [anon_sym_COMMA] = ACTIONS(2938),
    [anon_sym_DOT_DOT] = ACTIONS(2940),
    [anon_sym_PLUS] = ACTIONS(2942),
    [anon_sym_DASH] = ACTIONS(2942),
    [anon_sym_LT_DASH] = ACTIONS(2944),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(2944),
    [anon_sym_when] = ACTIONS(2946),
    [anon_sym_COLON_COLON] = ACTIONS(2949),
    [anon_sym_EQ_GT] = ACTIONS(2951),
    [anon_sym_EQ] = ACTIONS(2953),
    [anon_sym_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(2955),
    [anon_sym_or] = ACTIONS(2955),
    [anon_sym_AMP_AMP] = ACTIONS(2957),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(2957),
    [anon_sym_and] = ACTIONS(2957),
    [anon_sym_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ] = ACTIONS(2959),
    [anon_sym_EQ_TILDE] = ACTIONS(2959),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(2959),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(2959),
    [anon_sym_LT_EQ] = ACTIONS(2932),
    [anon_sym_GT_EQ] = ACTIONS(2932),
    [anon_sym_PIPE_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_LT] = ACTIONS(2961),
    [anon_sym_GT_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE] = ACTIONS(2961),
    [anon_sym_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_TILDE_GT] = ACTIONS(2961),
    [anon_sym_LT_PIPE_GT] = ACTIONS(2961),
    [anon_sym_in] = ACTIONS(2963),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(2965),
    [anon_sym_SLASH_SLASH] = ACTIONS(2967),
    [anon_sym_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH] = ACTIONS(2940),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(2940),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(2940),
    [anon_sym_LT_GT] = ACTIONS(2940),
    [anon_sym_STAR] = ACTIONS(2936),
    [anon_sym_STAR_STAR] = ACTIONS(2969),
    [anon_sym_DASH_GT] = ACTIONS(2971),
    [anon_sym_DOT] = ACTIONS(2973),
    [anon_sym_after] = ACTIONS(103),
    [anon_sym_catch] = ACTIONS(105),
    [anon_sym_else] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(1028),
    [anon_sym_rescue] = ACTIONS(113),
    [anon_sym_LBRACK2] = ACTIONS(2975),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(2977),
  },
  [STATE(1087)] = {
    [sym_do_block] = STATE(1103),
    [aux_sym__terminator_token1] = ACTIONS(3025),
    [anon_sym_SEMI] = ACTIONS(3027),
    [anon_sym_LPAREN] = ACTIONS(3027),
    [anon_sym_RPAREN] = ACTIONS(3027),
    [anon_sym_LT] = ACTIONS(3027),
    [anon_sym_GT] = ACTIONS(3027),
    [anon_sym_PIPE] = ACTIONS(3027),
    [anon_sym_SLASH] = ACTIONS(3027),
    [anon_sym_COMMA] = ACTIONS(3027),
    [anon_sym_DOT_DOT] = ACTIONS(3027),
    [anon_sym_PLUS] = ACTIONS(3027),
    [anon_sym_DASH] = ACTIONS(3027),
    [anon_sym_LT_DASH] = ACTIONS(3027),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(3027),
    [anon_sym_when] = ACTIONS(3027),
    [anon_sym_COLON_COLON] = ACTIONS(3027),
    [anon_sym_EQ_GT] = ACTIONS(3027),
    [anon_sym_EQ] = ACTIONS(3027),
    [anon_sym_PIPE_PIPE] = ACTIONS(3027),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(3027),
    [anon_sym_or] = ACTIONS(3027),
    [anon_sym_AMP_AMP] = ACTIONS(3027),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(3027),
    [anon_sym_and] = ACTIONS(3027),
    [anon_sym_EQ_EQ] = ACTIONS(3027),
    [anon_sym_BANG_EQ] = ACTIONS(3027),
    [anon_sym_EQ_TILDE] = ACTIONS(3027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3027),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(3027),
    [anon_sym_LT_EQ] = ACTIONS(3027),
    [anon_sym_GT_EQ] = ACTIONS(3027),
    [anon_sym_PIPE_GT] = ACTIONS(3027),
    [anon_sym_LT_LT_LT] = ACTIONS(3027),
    [anon_sym_GT_GT_GT] = ACTIONS(3027),
    [anon_sym_LT_LT_TILDE] = ACTIONS(3027),
    [anon_sym_TILDE_GT_GT] = ACTIONS(3027),
    [anon_sym_LT_TILDE] = ACTIONS(3027),
    [anon_sym_TILDE_GT] = ACTIONS(3027),
    [anon_sym_LT_TILDE_GT] = ACTIONS(3027),
    [anon_sym_LT_PIPE_GT] = ACTIONS(3027),
    [anon_sym_in] = ACTIONS(3027),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(3027),
    [anon_sym_SLASH_SLASH] = ACTIONS(3027),
    [anon_sym_PLUS_PLUS] = ACTIONS(3027),
    [anon_sym_DASH_DASH] = ACTIONS(3027),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(3027),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(3027),
    [anon_sym_LT_GT] = ACTIONS(3027),
    [anon_sym_STAR] = ACTIONS(3027),
    [anon_sym_STAR_STAR] = ACTIONS(3027),
    [anon_sym_DASH_GT] = ACTIONS(3027),
    [anon_sym_DOT] = ACTIONS(3027),
    [anon_sym_after] = ACTIONS(3027),
    [anon_sym_catch] = ACTIONS(3027),
    [anon_sym_do] = ACTIONS(3027),
    [anon_sym_else] = ACTIONS(3027),
    [anon_sym_end] = ACTIONS(3027),
    [anon_sym_rescue] = ACTIONS(3027),
    [anon_sym_LBRACK2] = ACTIONS(3025),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(3025),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(3025),
  },
  [STATE(1088)] = {
    [sym__call_arguments_with_parentheses] = STATE(1096),
    [aux_sym__terminator_token1] = ACTIONS(3029),
    [anon_sym_SEMI] = ACTIONS(3031),
    [anon_sym_LPAREN] = ACTIONS(3033),
    [anon_sym_RPAREN] = ACTIONS(3031),
    [anon_sym_LT] = ACTIONS(3031),
    [anon_sym_GT] = ACTIONS(3031),
    [anon_sym_PIPE] = ACTIONS(3031),
    [anon_sym_SLASH] = ACTIONS(3031),
    [anon_sym_COMMA] = ACTIONS(3031),
    [anon_sym_DOT_DOT] = ACTIONS(3031),
    [anon_sym_PLUS] = ACTIONS(3031),
    [anon_sym_DASH] = ACTIONS(3031),
    [anon_sym_LT_DASH] = ACTIONS(3031),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(3031),
    [anon_sym_when] = ACTIONS(3031),
    [anon_sym_COLON_COLON] = ACTIONS(3031),
    [anon_sym_EQ_GT] = ACTIONS(3031),
    [anon_sym_EQ] = ACTIONS(3031),
    [anon_sym_PIPE_PIPE] = ACTIONS(3031),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(3031),
    [anon_sym_or] = ACTIONS(3031),
    [anon_sym_AMP_AMP] = ACTIONS(3031),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(3031),
    [anon_sym_and] = ACTIONS(3031),
    [anon_sym_EQ_EQ] = ACTIONS(3031),
    [anon_sym_BANG_EQ] = ACTIONS(3031),
    [anon_sym_EQ_TILDE] = ACTIONS(3031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3031),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(3031),
    [anon_sym_LT_EQ] = ACTIONS(3031),
    [anon_sym_GT_EQ] = ACTIONS(3031),
    [anon_sym_PIPE_GT] = ACTIONS(3031),
    [anon_sym_LT_LT_LT] = ACTIONS(3031),
    [anon_sym_GT_GT_GT] = ACTIONS(3031),
    [anon_sym_LT_LT_TILDE] = ACTIONS(3031),
    [anon_sym_TILDE_GT_GT] = ACTIONS(3031),
    [anon_sym_LT_TILDE] = ACTIONS(3031),
    [anon_sym_TILDE_GT] = ACTIONS(3031),
    [anon_sym_LT_TILDE_GT] = ACTIONS(3031),
    [anon_sym_LT_PIPE_GT] = ACTIONS(3031),
    [anon_sym_in] = ACTIONS(3031),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(3031),
    [anon_sym_SLASH_SLASH] = ACTIONS(3031),
    [anon_sym_PLUS_PLUS] = ACTIONS(3031),
    [anon_sym_DASH_DASH] = ACTIONS(3031),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(3031),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(3031),
    [anon_sym_LT_GT] = ACTIONS(3031),
    [anon_sym_STAR] = ACTIONS(3031),
    [anon_sym_STAR_STAR] = ACTIONS(3031),
    [anon_sym_DASH_GT] = ACTIONS(3031),
    [anon_sym_DOT] = ACTIONS(3031),
    [anon_sym_after] = ACTIONS(3031),
    [anon_sym_catch] = ACTIONS(3031),
    [anon_sym_do] = ACTIONS(3031),
    [anon_sym_else] = ACTIONS(3031),
    [anon_sym_end] = ACTIONS(3031),
    [anon_sym_rescue] = ACTIONS(3031),
    [anon_sym_LBRACK2] = ACTIONS(3029),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(3029),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(3029),
  },
  [STATE(1089)] = {
    [sym_do_block] = STATE(1430),
    [aux_sym__terminator_token1] = ACTIONS(3035),
    [anon_sym_SEMI] = ACTIONS(3037),
    [anon_sym_LPAREN] = ACTIONS(3037),
    [anon_sym_RPAREN] = ACTIONS(3037),
    [anon_sym_LT] = ACTIONS(3037),
    [anon_sym_GT] = ACTIONS(3037),
    [anon_sym_PIPE] = ACTIONS(3037),
    [anon_sym_SLASH] = ACTIONS(3037),
    [anon_sym_COMMA] = ACTIONS(3037),
    [anon_sym_DOT_DOT] = ACTIONS(3037),
    [anon_sym_PLUS] = ACTIONS(3037),
    [anon_sym_DASH] = ACTIONS(3037),
    [anon_sym_LT_DASH] = ACTIONS(3037),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(3037),
    [anon_sym_when] = ACTIONS(3037),
    [anon_sym_COLON_COLON] = ACTIONS(3037),
    [anon_sym_EQ_GT] = ACTIONS(3037),
    [anon_sym_EQ] = ACTIONS(3037),
    [anon_sym_PIPE_PIPE] = ACTIONS(3037),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(3037),
    [anon_sym_or] = ACTIONS(3037),
    [anon_sym_AMP_AMP] = ACTIONS(3037),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(3037),
    [anon_sym_and] = ACTIONS(3037),
    [anon_sym_EQ_EQ] = ACTIONS(3037),
    [anon_sym_BANG_EQ] = ACTIONS(3037),
    [anon_sym_EQ_TILDE] = ACTIONS(3037),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3037),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(3037),
    [anon_sym_LT_EQ] = ACTIONS(3037),
    [anon_sym_GT_EQ] = ACTIONS(3037),
    [anon_sym_PIPE_GT] = ACTIONS(3037),
    [anon_sym_LT_LT_LT] = ACTIONS(3037),
    [anon_sym_GT_GT_GT] = ACTIONS(3037),
    [anon_sym_LT_LT_TILDE] = ACTIONS(3037),
    [anon_sym_TILDE_GT_GT] = ACTIONS(3037),
    [anon_sym_LT_TILDE] = ACTIONS(3037),
    [anon_sym_TILDE_GT] = ACTIONS(3037),
    [anon_sym_LT_TILDE_GT] = ACTIONS(3037),
    [anon_sym_LT_PIPE_GT] = ACTIONS(3037),
    [anon_sym_in] = ACTIONS(3037),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(3037),
    [anon_sym_SLASH_SLASH] = ACTIONS(3037),
    [anon_sym_PLUS_PLUS] = ACTIONS(3037),
    [anon_sym_DASH_DASH] = ACTIONS(3037),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(3037),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(3037),
    [anon_sym_LT_GT] = ACTIONS(3037),
    [anon_sym_STAR] = ACTIONS(3037),
    [anon_sym_STAR_STAR] = ACTIONS(3037),
    [anon_sym_DASH_GT] = ACTIONS(3037),
    [anon_sym_DOT] = ACTIONS(3037),
    [anon_sym_after] = ACTIONS(3037),
    [anon_sym_catch] = ACTIONS(3037),
    [anon_sym_do] = ACTIONS(295),
    [anon_sym_else] = ACTIONS(3037),
    [anon_sym_end] = ACTIONS(3037),
    [anon_sym_rescue] = ACTIONS(3037),
    [anon_sym_LBRACK2] = ACTIONS(3035),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(3039),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(3035),
  },
  [STATE(1090)] = {
    [sym_do_block] = STATE(1427),
    [aux_sym__terminator_token1] = ACTIONS(3025),
    [anon_sym_SEMI] = ACTIONS(3027),
    [anon_sym_LPAREN] = ACTIONS(3027),
    [anon_sym_RPAREN] = ACTIONS(3027),
    [anon_sym_LT] = ACTIONS(3027),
    [anon_sym_GT] = ACTIONS(3027),
    [anon_sym_PIPE] = ACTIONS(3027),
    [anon_sym_SLASH] = ACTIONS(3027),
    [anon_sym_COMMA] = ACTIONS(3027),
    [anon_sym_DOT_DOT] = ACTIONS(3027),
    [anon_sym_PLUS] = ACTIONS(3027),
    [anon_sym_DASH] = ACTIONS(3027),
    [anon_sym_LT_DASH] = ACTIONS(3027),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(3027),
    [anon_sym_when] = ACTIONS(3027),
    [anon_sym_COLON_COLON] = ACTIONS(3027),
    [anon_sym_EQ_GT] = ACTIONS(3027),
    [anon_sym_EQ] = ACTIONS(3027),
    [anon_sym_PIPE_PIPE] = ACTIONS(3027),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(3027),
    [anon_sym_or] = ACTIONS(3027),
    [anon_sym_AMP_AMP] = ACTIONS(3027),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(3027),
    [anon_sym_and] = ACTIONS(3027),
    [anon_sym_EQ_EQ] = ACTIONS(3027),
    [anon_sym_BANG_EQ] = ACTIONS(3027),
    [anon_sym_EQ_TILDE] = ACTIONS(3027),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3027),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(3027),
    [anon_sym_LT_EQ] = ACTIONS(3027),
    [anon_sym_GT_EQ] = ACTIONS(3027),
    [anon_sym_PIPE_GT] = ACTIONS(3027),
    [anon_sym_LT_LT_LT] = ACTIONS(3027),
    [anon_sym_GT_GT_GT] = ACTIONS(3027),
    [anon_sym_LT_LT_TILDE] = ACTIONS(3027),
    [anon_sym_TILDE_GT_GT] = ACTIONS(3027),
    [anon_sym_LT_TILDE] = ACTIONS(3027),
    [anon_sym_TILDE_GT] = ACTIONS(3027),
    [anon_sym_LT_TILDE_GT] = ACTIONS(3027),
    [anon_sym_LT_PIPE_GT] = ACTIONS(3027),
    [anon_sym_in] = ACTIONS(3027),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(3027),
    [anon_sym_SLASH_SLASH] = ACTIONS(3027),
    [anon_sym_PLUS_PLUS] = ACTIONS(3027),
    [anon_sym_DASH_DASH] = ACTIONS(3027),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(3027),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(3027),
    [anon_sym_LT_GT] = ACTIONS(3027),
    [anon_sym_STAR] = ACTIONS(3027),
    [anon_sym_STAR_STAR] = ACTIONS(3027),
    [anon_sym_DASH_GT] = ACTIONS(3027),
    [anon_sym_DOT] = ACTIONS(3027),
    [anon_sym_after] = ACTIONS(3027),
    [anon_sym_catch] = ACTIONS(3027),
    [anon_sym_do] = ACTIONS(295),
    [anon_sym_else] = ACTIONS(3027),
    [anon_sym_end] = ACTIONS(3027),
    [anon_sym_rescue] = ACTIONS(3027),
    [anon_sym_LBRACK2] = ACTIONS(3025),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(3041),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(3025),
  },
  [STATE(1091)] = {
    [sym__call_arguments_with_parentheses] = STATE(1138),
    [aux_sym__terminator_token1] = ACTIONS(3029),
    [anon_sym_SEMI] = ACTIONS(3031),
    [anon_sym_LPAREN] = ACTIONS(3033),
    [anon_sym_RPAREN] = ACTIONS(3031),
    [anon_sym_LT] = ACTIONS(3031),
    [anon_sym_GT] = ACTIONS(3031),
    [anon_sym_PIPE] = ACTIONS(3031),
    [anon_sym_SLASH] = ACTIONS(3031),
    [anon_sym_COMMA] = ACTIONS(3031),
    [anon_sym_DOT_DOT] = ACTIONS(3031),
    [anon_sym_PLUS] = ACTIONS(3031),
    [anon_sym_DASH] = ACTIONS(3031),
    [anon_sym_LT_DASH] = ACTIONS(3031),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(3031),
    [anon_sym_when] = ACTIONS(3031),
    [anon_sym_COLON_COLON] = ACTIONS(3031),
    [anon_sym_EQ_GT] = ACTIONS(3031),
    [anon_sym_EQ] = ACTIONS(3031),
    [anon_sym_PIPE_PIPE] = ACTIONS(3031),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(3031),
    [anon_sym_or] = ACTIONS(3031),
    [anon_sym_AMP_AMP] = ACTIONS(3031),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(3031),
    [anon_sym_and] = ACTIONS(3031),
    [anon_sym_EQ_EQ] = ACTIONS(3031),
    [anon_sym_BANG_EQ] = ACTIONS(3031),
    [anon_sym_EQ_TILDE] = ACTIONS(3031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3031),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(3031),
    [anon_sym_LT_EQ] = ACTIONS(3031),
    [anon_sym_GT_EQ] = ACTIONS(3031),
    [anon_sym_PIPE_GT] = ACTIONS(3031),
    [anon_sym_LT_LT_LT] = ACTIONS(3031),
    [anon_sym_GT_GT_GT] = ACTIONS(3031),
    [anon_sym_LT_LT_TILDE] = ACTIONS(3031),
    [anon_sym_TILDE_GT_GT] = ACTIONS(3031),
    [anon_sym_LT_TILDE] = ACTIONS(3031),
    [anon_sym_TILDE_GT] = ACTIONS(3031),
    [anon_sym_LT_TILDE_GT] = ACTIONS(3031),
    [anon_sym_LT_PIPE_GT] = ACTIONS(3031),
    [anon_sym_in] = ACTIONS(3031),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(3031),
    [anon_sym_SLASH_SLASH] = ACTIONS(3031),
    [anon_sym_PLUS_PLUS] = ACTIONS(3031),
    [anon_sym_DASH_DASH] = ACTIONS(3031),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(3031),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(3031),
    [anon_sym_LT_GT] = ACTIONS(3031),
    [anon_sym_STAR] = ACTIONS(3031),
    [anon_sym_STAR_STAR] = ACTIONS(3031),
    [anon_sym_DASH_GT] = ACTIONS(3031),
    [anon_sym_DOT] = ACTIONS(3031),
    [anon_sym_after] = ACTIONS(3031),
    [anon_sym_catch] = ACTIONS(3031),
    [anon_sym_do] = ACTIONS(3031),
    [anon_sym_else] = ACTIONS(3031),
    [anon_sym_end] = ACTIONS(3031),
    [anon_sym_rescue] = ACTIONS(3031),
    [anon_sym_LBRACK2] = ACTIONS(3029),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(3029),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(3029),
  },
  [STATE(1092)] = {
    [sym_do_block] = STATE(1112),
    [aux_sym__terminator_token1] = ACTIONS(3035),
    [anon_sym_SEMI] = ACTIONS(3037),
    [anon_sym_LPAREN] = ACTIONS(3037),
    [anon_sym_RPAREN] = ACTIONS(3037),
    [anon_sym_LT] = ACTIONS(3037),
    [anon_sym_GT] = ACTIONS(3037),
    [anon_sym_PIPE] = ACTIONS(3037),
    [anon_sym_SLASH] = ACTIONS(3037),
    [anon_sym_COMMA] = ACTIONS(3037),
    [anon_sym_DOT_DOT] = ACTIONS(3037),
    [anon_sym_PLUS] = ACTIONS(3037),
    [anon_sym_DASH] = ACTIONS(3037),
    [anon_sym_LT_DASH] = ACTIONS(3037),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(3037),
    [anon_sym_when] = ACTIONS(3037),
    [anon_sym_COLON_COLON] = ACTIONS(3037),
    [anon_sym_EQ_GT] = ACTIONS(3037),
    [anon_sym_EQ] = ACTIONS(3037),
    [anon_sym_PIPE_PIPE] = ACTIONS(3037),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(3037),
    [anon_sym_or] = ACTIONS(3037),
    [anon_sym_AMP_AMP] = ACTIONS(3037),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(3037),
    [anon_sym_and] = ACTIONS(3037),
    [anon_sym_EQ_EQ] = ACTIONS(3037),
    [anon_sym_BANG_EQ] = ACTIONS(3037),
    [anon_sym_EQ_TILDE] = ACTIONS(3037),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3037),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(3037),
    [anon_sym_LT_EQ] = ACTIONS(3037),
    [anon_sym_GT_EQ] = ACTIONS(3037),
    [anon_sym_PIPE_GT] = ACTIONS(3037),
    [anon_sym_LT_LT_LT] = ACTIONS(3037),
    [anon_sym_GT_GT_GT] = ACTIONS(3037),
    [anon_sym_LT_LT_TILDE] = ACTIONS(3037),
    [anon_sym_TILDE_GT_GT] = ACTIONS(3037),
    [anon_sym_LT_TILDE] = ACTIONS(3037),
    [anon_sym_TILDE_GT] = ACTIONS(3037),
    [anon_sym_LT_TILDE_GT] = ACTIONS(3037),
    [anon_sym_LT_PIPE_GT] = ACTIONS(3037),
    [anon_sym_in] = ACTIONS(3037),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(3037),
    [anon_sym_SLASH_SLASH] = ACTIONS(3037),
    [anon_sym_PLUS_PLUS] = ACTIONS(3037),
    [anon_sym_DASH_DASH] = ACTIONS(3037),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(3037),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(3037),
    [anon_sym_LT_GT] = ACTIONS(3037),
    [anon_sym_STAR] = ACTIONS(3037),
    [anon_sym_STAR_STAR] = ACTIONS(3037),
    [anon_sym_DASH_GT] = ACTIONS(3037),
    [anon_sym_DOT] = ACTIONS(3037),
    [anon_sym_after] = ACTIONS(3037),
    [anon_sym_catch] = ACTIONS(3037),
    [anon_sym_do] = ACTIONS(3037),
    [anon_sym_else] = ACTIONS(3037),
    [anon_sym_end] = ACTIONS(3037),
    [anon_sym_rescue] = ACTIONS(3037),
    [anon_sym_LBRACK2] = ACTIONS(3035),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(3035),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(3035),
  },
  [STATE(1093)] = {
    [sym__call_arguments_with_parentheses] = STATE(1100),
    [aux_sym__terminator_token1] = ACTIONS(3029),
    [anon_sym_SEMI] = ACTIONS(3031),
    [anon_sym_LPAREN] = ACTIONS(3033),
    [anon_sym_RPAREN] = ACTIONS(3031),
    [anon_sym_LT] = ACTIONS(3031),
    [anon_sym_GT] = ACTIONS(3031),
    [anon_sym_PIPE] = ACTIONS(3031),
    [anon_sym_SLASH] = ACTIONS(3031),
    [anon_sym_COMMA] = ACTIONS(3031),
    [anon_sym_DOT_DOT] = ACTIONS(3031),
    [anon_sym_PLUS] = ACTIONS(3031),
    [anon_sym_DASH] = ACTIONS(3031),
    [anon_sym_LT_DASH] = ACTIONS(3031),
    [anon_sym_BSLASH_BSLASH] = ACTIONS(3031),
    [anon_sym_when] = ACTIONS(3031),
    [anon_sym_COLON_COLON] = ACTIONS(3031),
    [anon_sym_EQ_GT] = ACTIONS(3031),
    [anon_sym_EQ] = ACTIONS(3031),
    [anon_sym_PIPE_PIPE] = ACTIONS(3031),
    [anon_sym_PIPE_PIPE_PIPE] = ACTIONS(3031),
    [anon_sym_or] = ACTIONS(3031),
    [anon_sym_AMP_AMP] = ACTIONS(3031),
    [anon_sym_AMP_AMP_AMP] = ACTIONS(3031),
    [anon_sym_and] = ACTIONS(3031),
    [anon_sym_EQ_EQ] = ACTIONS(3031),
    [anon_sym_BANG_EQ] = ACTIONS(3031),
    [anon_sym_EQ_TILDE] = ACTIONS(3031),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(3031),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(3031),
    [anon_sym_LT_EQ] = ACTIONS(3031),
    [anon_sym_GT_EQ] = ACTIONS(3031),
    [anon_sym_PIPE_GT] = ACTIONS(3031),
    [anon_sym_LT_LT_LT] = ACTIONS(3031),
    [anon_sym_GT_GT_GT] = ACTIONS(3031),
    [anon_sym_LT_LT_TILDE] = ACTIONS(3031),
    [anon_sym_TILDE_GT_GT] = ACTIONS(3031),
    [anon_sym_LT_TILDE] = ACTIONS(3031),
    [anon_sym_TILDE_GT] = ACTIONS(3031),
    [anon_sym_LT_TILDE_GT] = ACTIONS(3031),
    [anon_sym_LT_PIPE_GT] = ACTIONS(3031),
    [anon_sym_in] = ACTIONS(3031),
    [anon_sym_CARET_CARET_CARET] = ACTIONS(3031),
    [anon_sym_SLASH_SLASH] = ACTIONS(3031),
    [anon_sym_PLUS_PLUS] = ACTIONS(3031),
    [anon_sym_DASH_DASH] = ACTIONS(3031),
    [anon_sym_PLUS_PLUS_PLUS] = ACTIONS(3031),
    [anon_sym_DASH_DASH_DASH] = ACTIONS(3031),
    [anon_sym_LT_GT] = ACTIONS(3031),
    [anon_sym_STAR] = ACTIONS(3031),
    [anon_sym_STAR_STAR] = ACTIONS(3031),
    [anon_sym_DASH_GT] = ACTIONS(3031),
    [anon_sym_DOT] = ACTIONS(3031),
    [anon_sym_after] = ACTIONS(3031),
    [anon_sym_catch] = ACTIONS(3031),
    [anon_sym_do] = ACTIONS(3031),
    [anon_sym_else] = ACTIONS(3031),
    [anon_sym_end] = ACTIONS(3031),
    [anon_sym_rescue] = ACTIONS(3031),
    [anon_sym_LBRACK2] = ACTIONS(3029),
    [sym_comment] = ACTIONS(5),
    [sym__newline_before_do] = ACTIONS(3029),
    [sym__newline_before_binary_operator] = ACTIONS(3),
    [sym__newline_before_comment] = ACTIONS(3),
    [sym__not_in] = ACTIONS(3029),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2946), 1,
      anon_sym_when,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3043), 1,
      anon_sym_SEMI,
    STATE(344), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4786), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(670), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [129] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3045), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3047), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [202] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1217), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [277] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    ACTIONS(3057), 1,
      sym__newline_before_do,
    STATE(1694), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3053), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3055), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [356] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [429] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2946), 1,
      anon_sym_when,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3059), 1,
      anon_sym_SEMI,
    STATE(282), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4780), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1377), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [558] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1208), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [633] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2946), 1,
      anon_sym_when,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3061), 1,
      anon_sym_SEMI,
    STATE(268), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4766), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(674), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [762] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3063), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3065), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [835] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3067), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3069), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [908] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3071), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3073), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [981] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1056] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2946), 1,
      anon_sym_when,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3081), 1,
      anon_sym_SEMI,
    STATE(269), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4782), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(678), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [1185] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1342), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3025), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3027), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1260] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    ACTIONS(3083), 1,
      sym__newline_before_do,
    STATE(1698), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1339] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1347), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3035), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3037), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1414] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3085), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3087), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1487] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3089), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3091), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1560] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3093), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3095), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1633] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3097), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3099), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1706] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1779] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3101), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3103), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [1852] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2946), 1,
      anon_sym_when,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3105), 1,
      anon_sym_SEMI,
    STATE(283), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4767), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1379), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [1981] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3107), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3109), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2054] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2127] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    ACTIONS(3111), 1,
      sym__newline_before_do,
    STATE(1585), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3025), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3027), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2206] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2946), 1,
      anon_sym_when,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3113), 1,
      anon_sym_SEMI,
    STATE(277), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4775), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1563), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [2335] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    ACTIONS(3115), 1,
      sym__newline_before_do,
    STATE(1590), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3035), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3037), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2414] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2487] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    ACTIONS(3121), 1,
      sym__newline_before_do,
    STATE(1702), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2566] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2641] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1214), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2716] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2946), 1,
      anon_sym_when,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3129), 1,
      anon_sym_SEMI,
    STATE(343), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4783), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(666), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [2845] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3131), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3133), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2918] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    ACTIONS(3135), 1,
      sym__newline_before_do,
    STATE(1697), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [2997] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2946), 1,
      anon_sym_when,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3137), 1,
      anon_sym_SEMI,
    STATE(276), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4771), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1561), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [3126] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    ACTIONS(3139), 1,
      sym__newline_before_do,
    STATE(1703), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3205] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3278] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3141), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3143), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3351] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3145), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3147), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3424] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3149), 1,
      anon_sym_LPAREN,
    STATE(1321), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3501] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1205), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3053), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3055), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3576] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3149), 1,
      anon_sym_LPAREN,
    STATE(1323), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3653] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3726] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1207), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3801] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3149), 1,
      anon_sym_LPAREN,
    STATE(1328), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3878] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3151), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3153), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [3951] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3155), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3157), 57,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4024] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3159), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3161), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4096] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4168] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4240] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4312] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4384] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4456] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4528] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4600] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4672] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4744] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4816] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4888] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [4960] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5032] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5104] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5176] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5248] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5320] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5392] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5464] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5536] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5608] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5680] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5752] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5824] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5896] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [5968] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6040] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6112] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6184] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6256] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6328] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6400] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6472] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6544] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6616] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6688] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6760] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6832] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6904] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [6976] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7048] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7120] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7192] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3331), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3333), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7264] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7336] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7408] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7480] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3101), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3103), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7552] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3151), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3153), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7624] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3339), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3341), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7696] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3033), 1,
      anon_sym_LPAREN,
    STATE(1108), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7772] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3033), 1,
      anon_sym_LPAREN,
    STATE(1130), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7848] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7920] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [7992] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3351), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3353), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8064] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3355), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3357), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8136] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3359), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3361), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8208] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3363), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3365), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8280] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8352] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8424] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3371), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3373), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8496] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3045), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3047), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8568] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3375), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3377), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8640] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8712] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8784] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8856] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [8928] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9000] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9072] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9144] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9216] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9288] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9360] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9432] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9504] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3403), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3405), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9576] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    ACTIONS(3407), 1,
      sym__newline_before_do,
    STATE(2041), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3053), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3055), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9654] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    ACTIONS(3409), 1,
      sym__newline_before_do,
    STATE(2043), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9732] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    ACTIONS(3411), 1,
      sym__newline_before_do,
    STATE(2044), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9810] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9882] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [9954] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10026] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10098] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10170] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10242] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    ACTIONS(3413), 1,
      sym__newline_before_do,
    STATE(2046), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10320] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    ACTIONS(3415), 1,
      sym__newline_before_do,
    STATE(2048), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10398] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10470] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10542] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10614] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10686] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3417), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3419), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10758] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3421), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3423), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10830] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10902] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3429), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3431), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [10974] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11046] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3433), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3435), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11118] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11190] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11262] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11334] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11406] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11478] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11550] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11622] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11694] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11766] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [11838] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_PIPE,
    ACTIONS(3443), 1,
      anon_sym_COMMA,
    ACTIONS(3451), 1,
      anon_sym_when,
    ACTIONS(3453), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3455), 1,
      anon_sym_EQ_GT,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    STATE(1345), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3433), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3449), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3435), 8,
      anon_sym_SEMI,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [11956] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12028] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12100] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12174] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12246] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12320] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12392] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12464] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12536] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12608] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12680] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12752] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12824] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12896] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [12968] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13040] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13112] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 5,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13184] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13256] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13328] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13400] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13472] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3131), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3133), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13544] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3141), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3143), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13616] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13688] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13760] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3085), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3087), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13832] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13904] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [13976] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14048] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3493), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3495), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14120] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3497), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3499), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14192] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3501), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3503), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14264] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3505), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3507), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14336] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3509), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3511), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14408] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3089), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3091), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14480] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3513), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3515), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14552] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14624] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14696] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3097), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3099), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14768] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3521), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3523), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14840] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14912] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [14984] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3525), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3527), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15056] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3529), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3531), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15128] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3155), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3157), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15200] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15272] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3533), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3535), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15344] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3537), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3539), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15416] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3541), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3543), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15488] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3033), 1,
      anon_sym_LPAREN,
    STATE(1128), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15564] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3545), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3547), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15636] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3107), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3109), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15708] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3145), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3147), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15780] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3549), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3551), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15852] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3553), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3555), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15924] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3557), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3559), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [15996] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3561), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3563), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16068] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16140] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16212] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16284] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3569), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3571), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16356] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3577), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16430] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3579), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16504] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3581), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16578] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3583), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16652] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3585), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16726] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3587), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16800] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3589), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16874] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3591), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [16948] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1548), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3053), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3055), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17022] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1551), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17096] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3593), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17170] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1557), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17244] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3595), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17318] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3597), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17392] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1562), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17466] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3599), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17540] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(1563), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17614] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3601), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17688] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3603), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17762] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3605), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17836] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3607), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17910] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3071), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3073), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [17982] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18056] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3611), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18130] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3613), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18204] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3615), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18278] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 56,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18350] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18422] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3063), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3065), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18494] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18566] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3067), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3069), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18638] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18710] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18782] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3641), 1,
      anon_sym_COMMA,
    STATE(1346), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3637), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3639), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18858] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3647), 1,
      anon_sym_COMMA,
    STATE(1346), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3645), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [18934] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3093), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3095), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19006] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19078] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3658), 1,
      anon_sym_COMMA,
    STATE(1350), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19154] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3658), 1,
      anon_sym_COMMA,
    STATE(1351), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19230] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3664), 1,
      anon_sym_COMMA,
    STATE(1351), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19306] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 56,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19378] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19449] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3071), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3073), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19520] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19591] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19662] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19733] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19804] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19875] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [19946] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20017] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20088] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20159] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20230] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20301] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20372] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20443] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20514] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20585] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20656] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20727] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20798] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20869] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [20940] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21011] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21082] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21153] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21224] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21295] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_PIPE,
    ACTIONS(3451), 1,
      anon_sym_when,
    ACTIONS(3453), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3455), 1,
      anon_sym_EQ_GT,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3449), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3643), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3645), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21408] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_PIPE,
    ACTIONS(3451), 1,
      anon_sym_when,
    ACTIONS(3453), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3455), 1,
      anon_sym_EQ_GT,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3449), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3671), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3673), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21521] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21592] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_PIPE,
    ACTIONS(3451), 1,
      anon_sym_when,
    ACTIONS(3453), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3455), 1,
      anon_sym_EQ_GT,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3449), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3675), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3677), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21705] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3533), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3535), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21776] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3417), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3419), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21847] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3501), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3503), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21918] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3421), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3423), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [21989] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3429), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3431), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22060] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22131] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3149), 1,
      anon_sym_LPAREN,
    STATE(1229), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22206] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22277] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22348] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22419] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22490] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22561] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22632] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22703] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22774] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3513), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3515), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22845] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22916] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [22987] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23058] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3493), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3495), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23129] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23200] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23271] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3497), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3499), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23342] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23413] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23484] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23555] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23626] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3537), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3539), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23697] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3505), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3507), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23768] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3521), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3523), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23839] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3509), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3511), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23910] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [23981] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24052] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3679), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24125] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3681), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24198] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24269] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24340] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3683), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24413] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3685), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24486] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3689), 1,
      anon_sym_PIPE,
    ACTIONS(3693), 1,
      anon_sym_COMMA,
    ACTIONS(3701), 1,
      anon_sym_when,
    ACTIONS(3703), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3705), 1,
      anon_sym_EQ_GT,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    STATE(1558), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3433), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3699), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3435), 7,
      anon_sym_SEMI,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [24603] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24674] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3063), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3065), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24745] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24816] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3067), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3069), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24887] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [24958] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25029] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3093), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3095), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25100] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25171] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25242] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25313] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3731), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25386] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25457] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3733), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25530] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25601] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3735), 1,
      anon_sym_COMMA,
    STATE(1438), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25676] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25747] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25818] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25889] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [25960] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26031] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3331), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3333), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26102] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3101), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3103), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26173] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26244] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3738), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26317] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3151), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3153), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26388] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3541), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3543), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26459] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3740), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26532] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3045), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3047), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26603] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3742), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26676] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26747] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3085), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3087), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26818] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26889] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [26960] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3553), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3555), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27031] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3561), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3563), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27102] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3089), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3091), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27173] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3097), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3099), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27244] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3155), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3157), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27315] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3545), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3547), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27386] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3549), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3551), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27457] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27528] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27599] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27670] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3569), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3571), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27741] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27812] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27883] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [27954] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28025] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28096] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28167] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28238] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3744), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28311] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28382] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28453] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3746), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28526] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3149), 1,
      anon_sym_LPAREN,
    STATE(1220), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28601] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3131), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3133), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28672] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3149), 1,
      anon_sym_LPAREN,
    STATE(1221), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28747] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3141), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3143), 56,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28818] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28889] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3433), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3435), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [28960] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3748), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29033] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3750), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29106] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3752), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29179] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3754), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29252] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3756), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29325] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3758), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29398] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3760), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29471] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3762), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29544] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3557), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3559), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29615] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3764), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29688] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29759] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29830] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29901] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [29972] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30043] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30114] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30185] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3525), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3527), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30256] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3529), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3531), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30327] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30398] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30469] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30540] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30611] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30682] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30753] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30824] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3766), 1,
      anon_sym_COMMA,
    STATE(1438), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30899] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [30972] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31043] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 41,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31130] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 31,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31223] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_PIPE,
    ACTIONS(3455), 1,
      anon_sym_EQ_GT,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31330] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31407] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 43,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31490] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31569] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_PIPE,
    ACTIONS(3451), 1,
      anon_sym_when,
    ACTIONS(3453), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3455), 1,
      anon_sym_EQ_GT,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 11,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31680] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31751] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_PIPE,
    ACTIONS(3451), 1,
      anon_sym_when,
    ACTIONS(3453), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3455), 1,
      anon_sym_EQ_GT,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 11,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31862] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [31933] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32004] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3339), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3341), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32075] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_PIPE,
    ACTIONS(3453), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3455), 1,
      anon_sym_EQ_GT,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 12,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32184] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3455), 1,
      anon_sym_EQ_GT,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 14,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32289] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3457), 1,
      anon_sym_EQ,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3459), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 15,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32392] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3351), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3353), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32463] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3159), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3161), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32534] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3355), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3357), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32605] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3461), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 19,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32704] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3463), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 22,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32801] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3437), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3465), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 27,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32896] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3467), 1,
      anon_sym_in,
    ACTIONS(3469), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3479), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 40,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [32987] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3359), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3361), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33058] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3363), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3365), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33129] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33200] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 42,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33285] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33356] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3471), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3473), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3441), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3447), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3445), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 42,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33441] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3371), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3373), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33512] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33583] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33654] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33725] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33796] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33867] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3375), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3377), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [33938] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34009] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34080] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34151] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34224] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34295] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34370] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34445] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_DOT,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34520] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34591] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3768), 1,
      anon_sym_COMMA,
    STATE(1559), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3637), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3639), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34666] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3770), 1,
      anon_sym_COMMA,
    STATE(1559), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3645), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34741] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34812] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34883] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [34954] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35025] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3403), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3405), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35096] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35167] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35238] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35309] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35380] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35451] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35522] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35593] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35664] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3766), 1,
      anon_sym_COMMA,
    STATE(1511), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35739] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35810] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35880] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [35950] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36020] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3541), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3543), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36090] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3553), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3555), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36160] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3561), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3563), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36230] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3071), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3073), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36300] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36370] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36440] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3063), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3065), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36510] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3067), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3069), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36580] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36650] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36720] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36790] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3569), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3571), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36860] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3093), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3095), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [36930] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3417), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3419), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37000] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3421), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3423), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37070] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37140] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37210] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 40,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37296] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 30,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37388] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3689), 1,
      anon_sym_PIPE,
    ACTIONS(3705), 1,
      anon_sym_EQ_GT,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 12,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37494] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 52,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37570] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 42,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37652] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37730] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3689), 1,
      anon_sym_PIPE,
    ACTIONS(3701), 1,
      anon_sym_when,
    ACTIONS(3703), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3705), 1,
      anon_sym_EQ_GT,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37840] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3689), 1,
      anon_sym_PIPE,
    ACTIONS(3701), 1,
      anon_sym_when,
    ACTIONS(3703), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3705), 1,
      anon_sym_EQ_GT,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [37950] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3689), 1,
      anon_sym_PIPE,
    ACTIONS(3703), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3705), 1,
      anon_sym_EQ_GT,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 11,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38058] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3705), 1,
      anon_sym_EQ_GT,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38162] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 14,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38264] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 18,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38362] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 21,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38458] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 26,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38552] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 39,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38642] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 41,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38726] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 41,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38810] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3429), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3431), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38880] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [38952] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39024] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3101), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3103), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39094] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3151), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3153), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39164] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39234] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3045), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3047), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39304] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3085), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3087), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39374] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3089), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3091), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39444] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3097), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3099), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39514] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3155), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3157), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39584] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3331), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3333), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39654] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39724] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39794] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39864] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [39934] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40004] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40074] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40144] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 4,
      sym__not_in,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40214] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3131), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3133), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40284] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40354] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40424] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40494] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40564] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40634] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40704] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3141), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3143), 55,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40774] = 33,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3775), 1,
      anon_sym_SEMI,
    ACTIONS(3777), 1,
      anon_sym_RPAREN,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3785), 1,
      anon_sym_COMMA,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    STATE(435), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5753), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    STATE(6658), 1,
      aux_sym__stab_clause_arguments_with_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [40902] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3557), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3559), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [40972] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41042] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    STATE(2065), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [41116] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41186] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    STATE(2066), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [41260] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41330] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41400] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41470] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    STATE(2069), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [41544] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41614] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3824), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3331), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3333), 53,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41686] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41756] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41826] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41896] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3826), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [41968] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3828), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42040] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3830), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42112] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3832), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42184] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3834), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42256] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3836), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42328] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3339), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3341), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42398] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3838), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42470] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3840), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42542] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3842), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42614] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3844), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42686] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3846), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42758] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3848), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42830] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3850), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42902] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3852), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [42974] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3854), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43046] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3856), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43118] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3858), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43190] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3860), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43262] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3862), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43334] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3864), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43406] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43476] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43546] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3351), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3353), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43616] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3159), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3161), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43686] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3355), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3357), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43756] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3359), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3361), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43826] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3363), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3365), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [43896] = 29,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3866), 1,
      aux_sym__terminator_token1,
    ACTIONS(3869), 1,
      anon_sym_SEMI,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3884), 1,
      anon_sym_when,
    ACTIONS(3886), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    STATE(259), 1,
      sym__terminator,
    STATE(1021), 1,
      aux_sym__terminator_repeat1,
    STATE(4787), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3882), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1355), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [44016] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [44086] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [44156] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3371), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3373), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [44226] = 29,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3884), 1,
      anon_sym_when,
    ACTIONS(3886), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3914), 1,
      aux_sym__terminator_token1,
    ACTIONS(3917), 1,
      anon_sym_SEMI,
    STATE(258), 1,
      sym__terminator,
    STATE(1021), 1,
      aux_sym__terminator_repeat1,
    STATE(4770), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3882), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1351), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [44346] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3689), 1,
      anon_sym_PIPE,
    ACTIONS(3701), 1,
      anon_sym_when,
    ACTIONS(3703), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3705), 1,
      anon_sym_EQ_GT,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3699), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3645), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [44458] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3689), 1,
      anon_sym_PIPE,
    ACTIONS(3701), 1,
      anon_sym_when,
    ACTIONS(3703), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3705), 1,
      anon_sym_EQ_GT,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3671), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3699), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3673), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [44570] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3689), 1,
      anon_sym_PIPE,
    ACTIONS(3701), 1,
      anon_sym_when,
    ACTIONS(3703), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3705), 1,
      anon_sym_EQ_GT,
    ACTIONS(3707), 1,
      anon_sym_EQ,
    ACTIONS(3717), 1,
      anon_sym_in,
    ACTIONS(3719), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3721), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3723), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3729), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3675), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(3691), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3697), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3699), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3709), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3711), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3687), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3713), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3695), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3677), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3715), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [44682] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [44752] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [44822] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [44892] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3375), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3377), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [44962] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45032] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45102] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45172] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45242] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3920), 1,
      anon_sym_COMMA,
    STATE(1699), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45316] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45386] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45456] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45526] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45596] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45666] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3403), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3405), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45736] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45806] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45876] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [45946] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46016] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46086] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46156] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46226] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46296] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46366] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46436] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46506] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46576] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46646] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46716] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46786] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46856] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46926] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [46996] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47066] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47136] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47206] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47276] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47346] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47416] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47486] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47556] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47626] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47696] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47766] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47836] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47906] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [47976] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3493), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3495), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48046] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48116] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3497), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3499), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48186] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3501), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3503), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48256] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3505), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3507), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48326] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3509), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3511), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48396] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48466] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48536] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48606] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48676] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48746] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48816] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3513), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3515), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48886] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [48956] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49026] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49096] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49166] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3521), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3523), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49236] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49306] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49376] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3525), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3527), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49446] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3529), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3531), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49516] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49586] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3533), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3535), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49656] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49726] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49796] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49866] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [49936] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3537), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3539), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50006] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50076] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50146] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50216] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50286] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50356] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50426] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50496] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50566] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50636] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50706] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50776] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50846] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50916] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [50986] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [51056] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [51126] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [51196] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [51266] = 33,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3785), 1,
      anon_sym_COMMA,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3923), 1,
      anon_sym_SEMI,
    ACTIONS(3925), 1,
      anon_sym_RPAREN,
    STATE(374), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5122), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    STATE(6658), 1,
      aux_sym__stab_clause_arguments_with_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [51394] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [51464] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2115), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3025), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3027), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [51536] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2116), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3035), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3037), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [51608] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    ACTIONS(3927), 1,
      sym__newline_before_do,
    STATE(2806), 1,
      sym_do_block,
    ACTIONS(3025), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3027), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [51684] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    ACTIONS(3929), 1,
      sym__newline_before_do,
    STATE(2808), 1,
      sym_do_block,
    ACTIONS(3035), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3037), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [51760] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3931), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3339), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3341), 53,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [51832] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [51902] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [51972] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 55,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52042] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52116] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52190] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3725), 1,
      anon_sym_DOT,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_do,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52264] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3933), 1,
      anon_sym_COMMA,
    STATE(1799), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52338] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3933), 1,
      anon_sym_COMMA,
    STATE(1699), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52412] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 55,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52482] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52551] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3403), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3405), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52620] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52689] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52758] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52827] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52896] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [52965] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53034] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53103] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53172] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53241] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53310] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53379] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53448] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53517] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53586] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53655] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53724] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53793] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53862] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [53931] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54000] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54069] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54138] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54207] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54276] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54345] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54414] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54483] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54552] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54621] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54690] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54759] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54828] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3493), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3495), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54897] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3497), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3499), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [54966] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3501), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3503), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55035] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3505), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3507), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55104] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3509), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3511), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55173] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55242] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55311] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55380] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3513), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3515), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55449] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55518] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55587] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55656] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3521), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3523), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55725] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55794] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3525), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3527), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55863] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3529), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3531), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [55932] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3533), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3535), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [56001] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [56070] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3537), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3539), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [56139] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [56208] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [56277] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [56346] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [56415] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3387), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3389), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [56484] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3391), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3393), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [56553] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [56622] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3343), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3345), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [56691] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3347), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3349), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [56760] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [56829] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3395), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3397), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [56898] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3399), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3401), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [56967] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [57036] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [57105] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [57174] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [57243] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [57312] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2701), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2703), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [57381] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2705), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2707), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [57450] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2709), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2711), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [57519] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2665), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2667), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [57588] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2669), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2671), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [57657] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2673), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2675), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [57726] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [57795] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [57864] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [57933] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58002] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58071] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58140] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3075), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3077), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [58211] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3123), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3125), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [58282] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58351] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58420] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58489] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58558] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58627] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58696] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58765] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58834] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58903] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [58972] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59041] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59110] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59179] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59248] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59317] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59386] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59455] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3541), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3543), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59524] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3553), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3555), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59593] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3561), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3563), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59662] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(914), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3935), 1,
      anon_sym_SEMI,
    STATE(416), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5129), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [59787] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59856] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59925] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [59994] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [60063] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [60132] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1647), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3937), 1,
      anon_sym_SEMI,
    STATE(381), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5250), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [60257] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3569), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3571), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [60326] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3923), 1,
      anon_sym_SEMI,
    ACTIONS(3939), 1,
      anon_sym_RPAREN,
    STATE(374), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5122), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [60451] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [60520] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [60589] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [60658] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3675), 1,
      aux_sym__terminator_token1,
    ACTIONS(3941), 1,
      anon_sym_when,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3677), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [60769] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [60838] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [60907] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [60976] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61045] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61114] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61183] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61252] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61321] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3331), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3333), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61390] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 54,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61459] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2701), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [61528] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2705), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [61597] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3943), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61668] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61737] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3945), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61808] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3947), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61879] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3949), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [61950] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3951), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [62021] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3167), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3169), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62090] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3171), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3173), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62159] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3175), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3177), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62228] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3179), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3181), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62297] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3183), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3185), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62366] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3187), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3189), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62435] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3191), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3193), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62504] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3195), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3197), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62573] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3199), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3201), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62642] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3203), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3205), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [62711] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 40,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [62796] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 30,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [62887] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 12,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [62992] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 52,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [63067] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(964), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3953), 1,
      anon_sym_SEMI,
    STATE(364), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5229), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [63192] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 42,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [63273] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [63350] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3941), 1,
      anon_sym_when,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [63459] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1761), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3955), 1,
      anon_sym_SEMI,
    STATE(421), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5493), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [63584] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3935), 1,
      anon_sym_SEMI,
    ACTIONS(3957), 1,
      anon_sym_RPAREN,
    STATE(416), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5129), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [63709] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3941), 1,
      anon_sym_when,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [63818] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 11,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [63925] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64028] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 14,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64129] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 18,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64226] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 21,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64321] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 26,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64414] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 39,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64503] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 41,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64586] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 41,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64669] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64738] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3417), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3419), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64807] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3101), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3103), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [64876] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3421), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3423), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [64945] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3151), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3153), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [65014] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(900), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3959), 1,
      anon_sym_SEMI,
    STATE(370), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5230), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [65139] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3045), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3047), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [65208] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3207), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3209), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [65277] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3211), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3213), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [65346] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3215), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3217), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [65415] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3961), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [65486] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1617), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3963), 1,
      anon_sym_SEMI,
    STATE(402), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5363), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [65611] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3223), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3225), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [65680] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3227), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3229), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [65749] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3953), 1,
      anon_sym_SEMI,
    ACTIONS(3965), 1,
      anon_sym_RPAREN,
    STATE(364), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5229), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [65874] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3231), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3233), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [65943] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3235), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3237), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66012] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3239), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3241), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66081] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3243), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3245), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66150] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3247), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3249), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66219] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3251), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3253), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66288] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3667), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3669), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66357] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3255), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3257), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66426] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3259), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3261), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66495] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3263), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3265), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66564] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3085), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3087), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66633] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3267), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3269), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66702] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3271), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3273), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66771] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(940), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3923), 1,
      anon_sym_SEMI,
    STATE(374), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5122), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [66896] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3275), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3277), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [66965] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3279), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3281), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67034] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3283), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3285), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67103] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3287), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3289), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67172] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3291), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3293), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67241] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3295), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3297), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67310] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3299), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3301), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67379] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3303), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3305), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67448] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1805), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3967), 1,
      anon_sym_SEMI,
    STATE(357), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5955), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [67573] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3307), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3309), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67642] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3311), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3313), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67711] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3775), 1,
      anon_sym_SEMI,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3969), 1,
      anon_sym_RPAREN,
    STATE(435), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5753), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [67836] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3315), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3317), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67905] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3319), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3321), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [67974] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3323), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3325), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [68043] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3327), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3329), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [68112] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3089), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3091), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [68181] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3097), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3099), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [68250] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3155), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3157), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [68319] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [68388] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [68457] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3971), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [68528] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1695), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3973), 1,
      anon_sym_SEMI,
    STATE(372), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5123), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [68653] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(924), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3975), 1,
      anon_sym_SEMI,
    STATE(389), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5943), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [68778] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [68847] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [68916] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [68985] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69054] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(928), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3977), 1,
      anon_sym_SEMI,
    STATE(409), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5465), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [69179] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3339), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3341), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69248] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3979), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69319] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3981), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69390] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69459] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69528] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3351), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3353), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69597] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3159), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3161), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69666] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3355), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3357), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69735] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1635), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3983), 1,
      anon_sym_SEMI,
    STATE(366), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5311), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [69860] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3359), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3361), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69929] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3363), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3365), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [69998] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3959), 1,
      anon_sym_SEMI,
    ACTIONS(3985), 1,
      anon_sym_RPAREN,
    STATE(370), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5230), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [70123] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70192] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70261] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3371), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3373), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70330] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3987), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70401] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3989), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70472] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3375), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3377), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70541] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70610] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70679] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70748] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70817] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [70886] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(954), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3991), 1,
      anon_sym_SEMI,
    STATE(432), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5752), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [71011] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [71080] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1737), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3993), 1,
      anon_sym_SEMI,
    STATE(420), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5536), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [71205] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3977), 1,
      anon_sym_SEMI,
    ACTIONS(3995), 1,
      anon_sym_RPAREN,
    STATE(409), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5465), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [71330] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3429), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3431), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [71399] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3131), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3133), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [71468] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [71537] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [71606] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [71675] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [71744] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [71813] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3141), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3143), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [71882] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(936), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3775), 1,
      anon_sym_SEMI,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    STATE(435), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5753), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [72007] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3557), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3559), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [72076] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [72145] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 54,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [72214] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2318), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3053), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3055), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [72285] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3997), 1,
      anon_sym_SEMI,
    STATE(358), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5848), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [72410] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2320), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3049), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3051), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [72481] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2321), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3049), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3051), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [72552] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3991), 1,
      anon_sym_SEMI,
    ACTIONS(3999), 1,
      anon_sym_RPAREN,
    STATE(432), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5752), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [72677] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2324), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3117), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3119), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [72748] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2328), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3049), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3051), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [72819] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3071), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3073), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [72888] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    ACTIONS(4001), 1,
      sym__newline_before_do,
    STATE(3087), 1,
      sym_do_block,
    ACTIONS(3053), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3055), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [72963] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    ACTIONS(4003), 1,
      sym__newline_before_do,
    STATE(3089), 1,
      sym_do_block,
    ACTIONS(3049), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3051), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [73038] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    ACTIONS(4005), 1,
      sym__newline_before_do,
    STATE(3090), 1,
      sym_do_block,
    ACTIONS(3049), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3051), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [73113] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    ACTIONS(4007), 1,
      sym__newline_before_do,
    STATE(3092), 1,
      sym_do_block,
    ACTIONS(3117), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3119), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [73188] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    ACTIONS(4009), 1,
      sym__newline_before_do,
    STATE(3093), 1,
      sym_do_block,
    ACTIONS(3049), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3051), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [73263] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2934), 1,
      anon_sym_PIPE,
    ACTIONS(2949), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2951), 1,
      anon_sym_EQ_GT,
    ACTIONS(2953), 1,
      anon_sym_EQ,
    ACTIONS(2963), 1,
      anon_sym_in,
    ACTIONS(2965), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(2967), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(2969), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(2977), 1,
      sym__not_in,
    ACTIONS(3671), 1,
      aux_sym__terminator_token1,
    ACTIONS(3941), 1,
      anon_sym_when,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2936), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2942), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2944), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(2955), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(2957), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(2932), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2959), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2940), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3673), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(2961), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [73374] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(906), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4011), 1,
      anon_sym_SEMI,
    STATE(376), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5108), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [73499] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1651), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4013), 1,
      anon_sym_SEMI,
    STATE(380), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5132), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [73624] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4011), 1,
      anon_sym_SEMI,
    ACTIONS(4015), 1,
      anon_sym_RPAREN,
    STATE(376), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5108), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [73749] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3975), 1,
      anon_sym_SEMI,
    ACTIONS(4017), 1,
      anon_sym_RPAREN,
    STATE(389), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5943), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [73874] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2709), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [73943] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(920), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4019), 1,
      anon_sym_SEMI,
    STATE(398), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5190), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [74068] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2665), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [74137] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1709), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4021), 1,
      anon_sym_SEMI,
    STATE(354), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5213), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [74262] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4019), 1,
      anon_sym_SEMI,
    ACTIONS(4023), 1,
      anon_sym_RPAREN,
    STATE(398), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5190), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [74387] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4025), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [74458] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4027), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [74529] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4029), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [74600] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4031), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [74671] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4033), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [74742] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4035), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [74813] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4037), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [74884] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3107), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3109), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [74953] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4039), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [75024] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4041), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [75095] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4043), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [75166] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(946), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4045), 1,
      anon_sym_SEMI,
    STATE(419), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5272), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [75291] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4047), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [75362] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4049), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [75433] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1771), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4051), 1,
      anon_sym_SEMI,
    STATE(422), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5295), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [75558] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4045), 1,
      anon_sym_SEMI,
    ACTIONS(4053), 1,
      anon_sym_RPAREN,
    STATE(419), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5272), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [75683] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4055), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [75754] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4057), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [75825] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4059), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [75896] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(960), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4061), 1,
      anon_sym_SEMI,
    STATE(437), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5345), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [76021] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4063), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [76092] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4065), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [76163] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4067), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [76234] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3145), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3147), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [76303] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_RPAREN,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4069), 1,
      anon_sym_SEMI,
    STATE(440), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5365), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [76428] = 32,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3793), 1,
      anon_sym_when,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(4061), 1,
      anon_sym_SEMI,
    ACTIONS(4071), 1,
      anon_sym_RPAREN,
    STATE(437), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5345), 1,
      aux_sym_block_repeat2,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [76553] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4073), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [76624] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4075), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [76695] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3063), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3065), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [76764] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3067), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3069), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [76833] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3093), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3095), 53,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [76902] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [76975] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [77048] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2973), 1,
      anon_sym_DOT,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [77121] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2669), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [77190] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2673), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [77259] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3621), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3623), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [77328] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4077), 1,
      anon_sym_COMMA,
    STATE(2124), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 52,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [77401] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4077), 1,
      anon_sym_COMMA,
    STATE(2125), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 52,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [77474] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4079), 1,
      anon_sym_COMMA,
    STATE(2125), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 52,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [77547] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3625), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3627), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [77616] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3629), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3631), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [77685] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3633), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3635), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [77754] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3650), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3652), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [77823] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3163), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3165), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [77892] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4082), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [77963] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4084), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [78034] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [78105] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4088), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [78176] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4090), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [78247] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4092), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [78318] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4094), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [78389] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4096), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [78460] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4098), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 53,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [78531] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3219), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3221), 53,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [78600] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3884), 1,
      anon_sym_when,
    ACTIONS(3886), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [78708] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    STATE(2142), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3617), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3619), 50,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [78780] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [78848] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3075), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3077), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [78916] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3123), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3125), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [78984] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3123), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3125), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [79052] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3075), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3077), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [79120] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3481), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3483), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [79188] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3481), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3483), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [79256] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3485), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3487), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [79324] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [79392] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [79460] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3485), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3487), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [79528] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [79596] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 25,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [79688] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3485), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3487), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [79756] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [79824] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4103), 1,
      anon_sym_COMMA,
    STATE(2243), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3654), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3656), 50,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [79896] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [79964] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [80032] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [80100] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4105), 1,
      anon_sym_COMMA,
    STATE(2162), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3643), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3645), 50,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [80172] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [80240] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [80308] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3617), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3619), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [80376] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [80444] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [80532] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    ACTIONS(4108), 1,
      sym__newline_before_do,
    STATE(3605), 1,
      sym_do_block,
    ACTIONS(3025), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3027), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [80606] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 40,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [80688] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3489), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3491), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [80756] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 40,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [80838] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3489), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3491), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [80906] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3331), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3333), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [80974] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81042] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81110] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    ACTIONS(4110), 1,
      sym__newline_before_do,
    STATE(3534), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3035), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3037), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [81184] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3387), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3389), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81252] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3391), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3393), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81320] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3541), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3543), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81388] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3553), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3555), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81456] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3561), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3563), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81524] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3567), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81592] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3567), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81660] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(3565), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [81732] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3567), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81800] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3569), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3571), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [81868] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 39,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [81952] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    ACTIONS(4116), 1,
      sym__newline_before_do,
    STATE(3615), 1,
      sym_do_block,
    ACTIONS(3035), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3037), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [82026] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [82094] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 29,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [82184] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 11,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [82288] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [82362] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2635), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3025), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3027), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [82432] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [82500] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [82568] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 41,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [82648] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [82716] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4118), 1,
      anon_sym_LPAREN,
    STATE(2628), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [82788] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4118), 1,
      anon_sym_LPAREN,
    STATE(2629), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [82860] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [82928] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [82996] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4118), 1,
      anon_sym_LPAREN,
    STATE(2632), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [83068] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3577), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [83138] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    STATE(2072), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [83210] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3579), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [83280] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    STATE(2073), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [83352] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3581), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [83422] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [83490] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    STATE(2075), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [83562] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3583), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [83632] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [83700] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3585), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [83770] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3587), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [83840] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [83908] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2500), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3025), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3027), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [83978] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2502), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3035), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3037), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [84048] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3589), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84118] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4120), 1,
      anon_sym_COMMA,
    STATE(2258), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84190] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84258] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [84326] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84394] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84462] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84530] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84598] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84666] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3369), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [84742] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84810] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84878] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [84946] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [85014] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3591), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [85084] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [85152] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3557), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3559), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [85220] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [85288] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [85356] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [85424] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [85492] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3593), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [85562] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [85630] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3117), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3119), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [85698] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [85766] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [85834] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4103), 1,
      anon_sym_COMMA,
    STATE(2142), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3660), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3662), 50,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [85906] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3433), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4124), 1,
      anon_sym_PIPE,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    ACTIONS(4136), 1,
      anon_sym_when,
    ACTIONS(4138), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4140), 1,
      anon_sym_EQ_GT,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    STATE(2284), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4134), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3435), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_do,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [86020] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3433), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3435), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [86088] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3337), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [86156] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3595), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [86226] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(2637), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3035), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3037), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [86296] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3884), 1,
      anon_sym_when,
    ACTIONS(3886), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [86404] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [86472] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3597), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [86542] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [86610] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [86678] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [86746] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [86814] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3337), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [86882] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(3335), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [86954] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4120), 1,
      anon_sym_COMMA,
    STATE(2288), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [87026] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3337), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87094] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [87162] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_LPAREN,
    STATE(2712), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87234] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87302] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3339), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3341), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87370] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87438] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [87506] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87574] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3493), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3495), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87642] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3343), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3345), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87710] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3347), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3349), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87778] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [87846] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3599), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [87916] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3351), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3353), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [87984] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3497), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3499), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [88052] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3159), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3161), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [88120] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3886), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [88226] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 12,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [88328] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3501), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3503), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [88396] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_LPAREN,
    STATE(2705), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [88468] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3601), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [88538] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [88638] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 17,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [88734] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [88802] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [88870] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4164), 1,
      anon_sym_COMMA,
    STATE(2162), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3637), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3639), 50,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [88942] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_LPAREN,
    STATE(2706), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [89014] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [89082] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [89150] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4166), 1,
      anon_sym_COMMA,
    STATE(2288), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [89222] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [89290] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [89358] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3675), 1,
      aux_sym__terminator_token1,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3884), 1,
      anon_sym_when,
    ACTIONS(3886), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3882), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3677), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [89468] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4169), 1,
      anon_sym_COMMA,
    STATE(2292), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3645), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [89540] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3505), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3507), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [89608] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3509), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3511), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [89676] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [89744] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [89812] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [89880] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3513), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3515), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [89948] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4172), 1,
      anon_sym_COMMA,
    STATE(2292), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3637), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3639), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [90020] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3517), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3519), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90088] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3517), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3519), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90156] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [90224] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [90292] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3545), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3547), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90360] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3521), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3523), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90428] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3517), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3519), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90496] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3671), 1,
      aux_sym__terminator_token1,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3884), 1,
      anon_sym_when,
    ACTIONS(3886), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3882), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3673), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [90606] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3525), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3527), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90674] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3355), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3357), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90742] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3359), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3361), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90810] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3363), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3365), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90878] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3369), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [90946] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3369), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [91014] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [91086] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3371), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3373), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [91154] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [91222] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [91290] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3375), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3377), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [91358] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3369), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [91426] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3379), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3381), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [91494] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3379), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3381), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [91562] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3383), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3385), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [91630] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [91698] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3383), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3385), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [91766] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [91834] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [91902] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [91970] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3379), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3381), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [92038] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3403), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3405), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [92106] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3603), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [92176] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [92244] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [92312] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3395), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3397), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [92380] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3399), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3401), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [92448] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 52,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [92520] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 52,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [92592] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 52,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [92664] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [92732] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3605), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [92802] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3607), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [92872] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [92942] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [93010] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [93078] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [93146] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [93214] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [93282] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3417), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3419), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [93350] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3421), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3423), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [93418] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3429), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3431), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [93486] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [93554] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [93622] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3529), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3531), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [93690] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3533), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3535), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [93758] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4176), 1,
      anon_sym_PIPE,
    ACTIONS(4180), 1,
      anon_sym_COMMA,
    ACTIONS(4188), 1,
      anon_sym_when,
    ACTIONS(4190), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4192), 1,
      anon_sym_EQ_GT,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    STATE(2299), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3433), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4186), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3435), 4,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [93872] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3537), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3539), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [93940] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3549), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3551), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [94008] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 20,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [94102] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3611), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [94172] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3613), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [94242] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3615), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [94312] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 52,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [94380] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    ACTIONS(4216), 1,
      sym__newline_before_do,
    STATE(3530), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3025), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3027), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [94454] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [94522] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3247), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3249), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [94589] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3752), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [94658] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3754), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [94727] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3756), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [94796] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3758), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [94865] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3760), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [94934] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3762), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [95003] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3764), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [95072] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4218), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [95141] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4220), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [95210] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [95277] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [95344] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3395), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3397), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [95411] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3399), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3401), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [95478] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [95545] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [95612] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [95679] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [95746] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [95813] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [95880] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [95947] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96014] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96081] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96148] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96215] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96282] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96349] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96416] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96483] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96550] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96617] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96684] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96751] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96818] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96885] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [96952] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97019] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97086] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97153] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97220] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97287] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97354] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97421] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97488] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2701), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2703), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [97555] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2705), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2707), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [97622] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97689] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97756] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2709), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2711), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [97823] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2665), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2667), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [97890] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [97957] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98024] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2669), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2671), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [98091] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2673), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2675), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [98158] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98225] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98292] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98359] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98426] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98493] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98560] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98627] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98694] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98761] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98828] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98895] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [98962] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [99029] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [99096] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [99163] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [99230] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [99297] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [99364] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [99431] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [99498] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [99565] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [99632] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [99699] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [99766] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [99833] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [99900] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [99967] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [100034] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [100101] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [100168] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [100235] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [100302] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [100369] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [100436] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [100503] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 51,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [100570] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3075), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3077), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [100639] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4222), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [100708] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4224), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [100777] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3123), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3125), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [100846] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4226), 1,
      anon_sym_COMMA,
    STATE(2543), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [100917] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3141), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3143), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [100984] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4228), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [101053] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4230), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [101122] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4232), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [101191] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4234), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [101260] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4236), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [101329] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4238), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [101398] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4240), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [101467] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2669), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [101534] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2673), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [101601] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3621), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3623), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [101668] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3625), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3627), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [101735] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3629), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3631), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [101802] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3633), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3635), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [101869] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3650), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3652), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [101936] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3163), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3165), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [102003] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3267), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3269), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [102070] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3271), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3273), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [102137] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3275), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3277), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [102204] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4176), 1,
      anon_sym_PIPE,
    ACTIONS(4188), 1,
      anon_sym_when,
    ACTIONS(4190), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4192), 1,
      anon_sym_EQ_GT,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4186), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3645), 5,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [102313] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_do,
  [102384] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3279), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3281), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [102451] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3283), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3285), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [102518] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4242), 1,
      anon_sym_COMMA,
    STATE(2599), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [102589] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4246), 1,
      anon_sym_PIPE,
    ACTIONS(4250), 1,
      anon_sym_COMMA,
    ACTIONS(4258), 1,
      anon_sym_when,
    ACTIONS(4260), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4262), 1,
      anon_sym_EQ_GT,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    STATE(2724), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3435), 2,
      anon_sym_SEMI,
      anon_sym_do,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4256), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3433), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [102702] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3287), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3289), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [102769] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3167), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3169), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [102836] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3171), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3173), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [102903] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3884), 1,
      anon_sym_when,
    ACTIONS(3886), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(4288), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3882), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4290), 6,
      anon_sym_SEMI,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [103012] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3175), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3177), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [103079] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3179), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3181), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [103146] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3183), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3185), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [103213] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3187), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3189), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [103280] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3191), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3193), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [103347] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3195), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3197), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [103414] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3199), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3201), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [103481] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3203), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3205), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [103548] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3291), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3293), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [103615] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3063), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3065), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [103682] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3295), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3297), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [103749] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4292), 1,
      anon_sym_COMMA,
    STATE(2584), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3637), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3639), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [103820] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3067), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3069), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [103887] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3131), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3133), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [103954] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3093), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3095), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104021] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4294), 1,
      anon_sym_COMMA,
    STATE(2503), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3617), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3619), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [104092] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 6,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104159] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 6,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104226] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4297), 1,
      anon_sym_COMMA,
    STATE(2506), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3643), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3645), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [104297] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104364] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104431] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3075), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3077), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [104500] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3123), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3125), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [104569] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104636] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104703] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104770] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104837] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104904] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [104971] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [105038] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [105105] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3675), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4124), 1,
      anon_sym_PIPE,
    ACTIONS(4136), 1,
      anon_sym_when,
    ACTIONS(4138), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4140), 1,
      anon_sym_EQ_GT,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4134), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3677), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [105214] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [105281] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3207), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3209), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [105348] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3211), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3213), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [105415] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3215), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3217), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [105482] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3219), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3221), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [105549] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3223), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3225), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [105616] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3227), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3229), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [105683] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4302), 1,
      anon_sym_PIPE,
    ACTIONS(4306), 1,
      anon_sym_COMMA,
    ACTIONS(4314), 1,
      anon_sym_when,
    ACTIONS(4316), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4318), 1,
      anon_sym_EQ_GT,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    STATE(2878), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3433), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4312), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3435), 3,
      anon_sym_SEMI,
      anon_sym_do,
      anon_sym_end,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [105796] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3231), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3233), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [105863] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3235), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3237), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [105930] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3239), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3241), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [105997] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3243), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3245), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [106064] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3247), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3249), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [106131] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3251), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3253), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [106198] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3667), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3669), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [106265] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3255), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3257), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [106332] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3259), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3261), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [106399] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3263), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3265), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [106466] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [106533] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [106600] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [106667] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [106734] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3267), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3269), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [106801] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4226), 1,
      anon_sym_COMMA,
    STATE(2614), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [106872] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3271), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3273), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [106939] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3275), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3277), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107006] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3279), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3281), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107073] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3283), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3285), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107140] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3287), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3289), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107207] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3291), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3293), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107274] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3295), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3297), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107341] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3299), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3301), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107408] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3303), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3305), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107475] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3307), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3309), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107542] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3311), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3313), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107609] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3315), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3317), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107676] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3319), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3321), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107743] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3323), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3325), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107810] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3327), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3329), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [107877] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [107944] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_DASH_GT,
      anon_sym_do,
  [108027] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 27,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_DASH_GT,
      anon_sym_do,
  [108116] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4176), 1,
      anon_sym_PIPE,
    ACTIONS(4192), 1,
      anon_sym_EQ_GT,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [108219] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4342), 1,
      anon_sym_LPAREN,
    STATE(3258), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [108290] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4342), 1,
      anon_sym_LPAREN,
    STATE(3259), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [108361] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3101), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3103), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [108428] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4342), 1,
      anon_sym_LPAREN,
    STATE(3261), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [108499] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_DASH_GT,
      anon_sym_do,
  [108572] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 39,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_DASH_GT,
      anon_sym_do,
  [108651] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_DASH_GT,
      anon_sym_do,
  [108726] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4176), 1,
      anon_sym_PIPE,
    ACTIONS(4188), 1,
      anon_sym_when,
    ACTIONS(4190), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4192), 1,
      anon_sym_EQ_GT,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [108833] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4176), 1,
      anon_sym_PIPE,
    ACTIONS(4188), 1,
      anon_sym_when,
    ACTIONS(4190), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4192), 1,
      anon_sym_EQ_GT,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [108940] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3299), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3301), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [109007] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4176), 1,
      anon_sym_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4192), 1,
      anon_sym_EQ_GT,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [109112] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4192), 1,
      anon_sym_EQ_GT,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_do,
  [109213] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_DASH_GT,
      anon_sym_do,
  [109312] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_DASH_GT,
      anon_sym_do,
  [109407] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 18,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_DASH_GT,
      anon_sym_do,
  [109500] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 23,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_DASH_GT,
      anon_sym_do,
  [109591] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_DASH_GT,
      anon_sym_do,
  [109678] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
      anon_sym_do,
  [109759] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
      anon_sym_do,
  [109840] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_do,
  [109911] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    ACTIONS(4344), 1,
      sym__newline_before_do,
    STATE(4294), 1,
      sym_do_block,
    ACTIONS(3025), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3027), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [109984] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4346), 1,
      anon_sym_COMMA,
    STATE(2584), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3645), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110055] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3151), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3153), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110122] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3141), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3143), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [110189] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3874), 1,
      anon_sym_PIPE,
    ACTIONS(3884), 1,
      anon_sym_when,
    ACTIONS(3886), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3888), 1,
      anon_sym_EQ_GT,
    ACTIONS(3890), 1,
      anon_sym_EQ,
    ACTIONS(3900), 1,
      anon_sym_in,
    ACTIONS(3902), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3904), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3906), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3908), 1,
      anon_sym_DOT,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(3912), 1,
      sym__not_in,
    ACTIONS(4349), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3876), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3880), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3882), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3892), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3894), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3872), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3896), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3878), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4351), 6,
      anon_sym_SEMI,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
    ACTIONS(3898), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [110298] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3045), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3047), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110365] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3288), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3025), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3027), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110434] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3290), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3035), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3037), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110503] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110570] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110637] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110704] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110771] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110838] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110905] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [110972] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111039] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4353), 1,
      anon_sym_COMMA,
    STATE(2599), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111110] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111177] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111244] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111311] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111378] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111445] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111512] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111579] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111646] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3085), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3087), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111713] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3145), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3147), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111780] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111847] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111914] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [111981] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112048] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4356), 1,
      anon_sym_COMMA,
    STATE(2614), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112119] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112186] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112253] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112320] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112387] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112454] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112521] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112588] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4359), 1,
      anon_sym_COMMA,
    STATE(2622), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3645), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112659] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112726] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112793] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112860] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [112927] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3402), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3053), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3055), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [112996] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3404), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3049), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3051), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113065] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3405), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3049), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3051), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113134] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [113201] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3407), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3117), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3119), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113270] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3408), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3049), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3051), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113339] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3071), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3073), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113406] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3063), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3065), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113473] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3067), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3069), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113540] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3089), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3091), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [113607] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3093), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3095), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113674] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3101), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3103), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113741] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3151), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3153), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113808] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_do,
  [113891] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3045), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3047), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [113958] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3085), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3087), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [114025] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3089), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3091), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [114092] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3097), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3099), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [114159] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3155), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3157), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [114226] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 27,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_do,
  [114315] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4124), 1,
      anon_sym_PIPE,
    ACTIONS(4140), 1,
      anon_sym_EQ_GT,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [114418] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3097), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3099), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [114485] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 49,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_do,
  [114558] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 39,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_do,
  [114637] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_do,
  [114712] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4124), 1,
      anon_sym_PIPE,
    ACTIONS(4136), 1,
      anon_sym_when,
    ACTIONS(4138), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4140), 1,
      anon_sym_EQ_GT,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [114819] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4124), 1,
      anon_sym_PIPE,
    ACTIONS(4136), 1,
      anon_sym_when,
    ACTIONS(4138), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4140), 1,
      anon_sym_EQ_GT,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [114926] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3155), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3157), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [114993] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4124), 1,
      anon_sym_PIPE,
    ACTIONS(4138), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4140), 1,
      anon_sym_EQ_GT,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_do,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [115098] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4140), 1,
      anon_sym_EQ_GT,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
  [115199] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 11,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_do,
  [115298] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 15,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_do,
  [115393] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 18,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_do,
  [115486] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 23,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_do,
  [115577] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_do,
  [115664] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [115745] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [115826] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    ACTIONS(4362), 1,
      sym__newline_before_do,
    STATE(4323), 1,
      sym_do_block,
    ACTIONS(3035), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3037), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [115899] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3433), 1,
      sym__newline_before_do,
    ACTIONS(4366), 1,
      anon_sym_PIPE,
    ACTIONS(4370), 1,
      anon_sym_COMMA,
    ACTIONS(4378), 1,
      anon_sym_when,
    ACTIONS(4380), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4382), 1,
      anon_sym_EQ_GT,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    STATE(2703), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4376), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3435), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [116012] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3303), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3305), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [116079] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    ACTIONS(4408), 1,
      sym__newline_before_do,
    STATE(4288), 1,
      sym_do_block,
    ACTIONS(3053), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3055), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [116152] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    ACTIONS(4410), 1,
      sym__newline_before_do,
    STATE(4295), 1,
      sym_do_block,
    ACTIONS(3049), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3051), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [116225] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4412), 1,
      anon_sym_COMMA,
    STATE(2718), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3654), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3656), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [116296] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3131), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3133), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [116363] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 6,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [116430] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 6,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [116497] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    ACTIONS(4414), 1,
      sym__newline_before_do,
    STATE(4123), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [116570] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    ACTIONS(4416), 1,
      sym__newline_before_do,
    STATE(4296), 1,
      sym_do_block,
    ACTIONS(3049), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3051), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [116643] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3307), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3309), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [116710] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2701), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [116777] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2705), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [116844] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    ACTIONS(4418), 1,
      sym__newline_before_do,
    STATE(4095), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3053), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3055), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [116917] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    ACTIONS(4420), 1,
      sym__newline_before_do,
    STATE(4302), 1,
      sym_do_block,
    ACTIONS(3117), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3119), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [116990] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3259), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3261), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117057] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3311), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3313), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117124] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3315), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3317), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117191] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3141), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3143), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [117258] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4176), 1,
      anon_sym_PIPE,
    ACTIONS(4188), 1,
      anon_sym_when,
    ACTIONS(4190), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4192), 1,
      anon_sym_EQ_GT,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3671), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4186), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3673), 5,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [117367] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    ACTIONS(4422), 1,
      sym__newline_before_do,
    STATE(4125), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117440] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3319), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3321), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117507] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3323), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3325), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117574] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [117643] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [117712] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    ACTIONS(4424), 1,
      sym__newline_before_do,
    STATE(4134), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117785] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3327), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3329), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117852] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3089), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3091), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117919] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3097), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3099), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [117986] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3155), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3157), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [118053] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4242), 1,
      anon_sym_COMMA,
    STATE(2482), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [118124] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2701), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2703), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [118191] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2705), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2707), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [118258] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [118325] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [118392] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3263), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3265), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [118459] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4428), 1,
      anon_sym_PIPE,
    ACTIONS(4432), 1,
      anon_sym_COMMA,
    ACTIONS(4440), 1,
      anon_sym_when,
    ACTIONS(4442), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4444), 1,
      anon_sym_EQ_GT,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    STATE(2499), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3433), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4438), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3435), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_do,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [118572] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3107), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3109), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [118639] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4468), 1,
      anon_sym_COMMA,
    STATE(2506), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3637), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3639), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [118710] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3005), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3053), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3055), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [118779] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3008), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [118848] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3009), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [118917] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3085), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3087), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [118984] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4176), 1,
      anon_sym_PIPE,
    ACTIONS(4188), 1,
      anon_sym_when,
    ACTIONS(4190), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4192), 1,
      anon_sym_EQ_GT,
    ACTIONS(4194), 1,
      anon_sym_EQ,
    ACTIONS(4204), 1,
      anon_sym_in,
    ACTIONS(4206), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4208), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4210), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(4214), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3675), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4178), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4184), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4186), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4196), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4198), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4174), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3677), 5,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4200), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4182), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4202), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [119093] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3012), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [119162] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3477), 1,
      anon_sym_LBRACK2,
    ACTIONS(4212), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_do,
  [119233] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2701), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [119300] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3013), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [119369] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2705), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [119436] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3107), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3109), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [119503] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3071), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3073), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [119570] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3643), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4124), 1,
      anon_sym_PIPE,
    ACTIONS(4136), 1,
      anon_sym_when,
    ACTIONS(4138), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4140), 1,
      anon_sym_EQ_GT,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4134), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3645), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [119679] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3145), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3147), 51,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [119746] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4412), 1,
      anon_sym_COMMA,
    STATE(2503), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3660), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3662), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [119817] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3671), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4124), 1,
      anon_sym_PIPE,
    ACTIONS(4136), 1,
      anon_sym_when,
    ACTIONS(4138), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4140), 1,
      anon_sym_EQ_GT,
    ACTIONS(4142), 1,
      anon_sym_EQ,
    ACTIONS(4152), 1,
      anon_sym_in,
    ACTIONS(4154), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4156), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4158), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4160), 1,
      sym__not_in,
    ACTIONS(4126), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4132), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4134), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4144), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4146), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4122), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3673), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4148), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4130), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4150), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [119926] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 6,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [119993] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2709), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2711), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [120060] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2665), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2667), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [120127] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120194] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4470), 1,
      anon_sym_COMMA,
    STATE(2622), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3637), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3639), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120265] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120332] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3679), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120401] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3681), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120470] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3683), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120539] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3685), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120608] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3731), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120677] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3733), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120746] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3738), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120815] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3740), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120884] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3742), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [120953] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3744), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121022] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3746), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121091] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3748), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121160] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3750), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121229] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3752), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121298] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3754), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121367] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3756), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121436] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3758), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121505] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3760), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121574] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3762), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121643] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3764), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121712] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 6,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121779] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4472), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121848] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121915] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [121982] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3071), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3073), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [122049] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122116] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122183] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122250] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122317] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3131), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3133), 50,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122384] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122451] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122518] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122585] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2709), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [122652] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122719] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2665), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [122786] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4474), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [122855] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [122922] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4476), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [122991] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [123058] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4478), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123127] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4480), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123196] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [123263] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4482), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123332] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4484), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123401] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4486), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123470] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [123537] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4488), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123606] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4490), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123675] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4492), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123744] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4494), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123813] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4496), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123882] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4498), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [123951] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4500), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [124020] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4502), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [124089] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4504), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [124158] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [124225] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2669), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2671), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [124292] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4506), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [124361] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4508), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [124430] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4510), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [124499] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2673), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2675), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [124566] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4512), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [124635] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [124702] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [124769] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [124836] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [124903] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [124970] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125037] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125104] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125171] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125238] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125305] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125372] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125439] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125506] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125573] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3063), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3065), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [125640] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125707] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125774] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3067), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3069), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [125841] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [125908] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3093), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3095), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [125975] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [126042] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [126109] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [126176] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [126243] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [126310] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2669), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [126377] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2673), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [126444] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [126511] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [126578] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3621), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3623), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [126645] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [126712] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3625), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3627), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [126779] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3629), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3631), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [126846] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [126913] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3633), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3635), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [126980] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [127047] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3650), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3652), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127114] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [127181] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3163), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3165), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127248] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3167), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3169), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127315] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [127382] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3171), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3173), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127449] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3175), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3177), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127516] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3179), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3181), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127583] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3183), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3185), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127650] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3187), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3189), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127717] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3191), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3193), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127784] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3195), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3197), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127851] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3199), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3201), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [127918] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [127985] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [128052] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3203), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3205), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [128119] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [128186] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [128253] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3101), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3103), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [128320] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [128387] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [128454] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    ACTIONS(4514), 1,
      sym__newline_before_do,
    STATE(4138), 1,
      sym_do_block,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3049), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3051), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [128527] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3151), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3153), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [128594] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2709), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [128661] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2665), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [128728] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4516), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [128797] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3045), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3047), 52,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [128864] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4518), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [128933] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4520), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129002] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3207), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3209), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [129069] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4522), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129138] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3211), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3213), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [129205] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3215), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3217), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [129272] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4524), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129341] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4526), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129410] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4528), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129479] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3387), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3389), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [129546] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3391), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3393), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [129613] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4530), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129682] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4532), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129751] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4534), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129820] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4536), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129889] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4538), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [129958] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4540), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130027] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4542), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130096] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4544), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130165] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3219), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3221), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [130232] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4546), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130301] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4548), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130370] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4550), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130439] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4552), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130508] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4554), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130577] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4556), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130646] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4558), 1,
      anon_sym_COMMA,
    STATE(2879), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3637), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3639), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [130717] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4560), 1,
      anon_sym_COMMA,
    STATE(2879), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3645), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [130788] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4563), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [130857] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4565), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [130926] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4567), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [130995] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3223), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3225), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [131062] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4569), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [131131] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4571), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [131200] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4573), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [131269] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4575), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [131338] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3343), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3345), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [131405] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3347), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3349), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [131472] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3679), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [131541] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3681), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [131610] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3683), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [131679] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3685), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [131748] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3731), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [131817] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3733), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [131886] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3227), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3229), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [131953] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3738), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [132022] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3231), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3233), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [132089] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3740), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [132158] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3235), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3237), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [132225] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [132292] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3239), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3241), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [132359] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [132426] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3243), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3245), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [132493] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [132560] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [132627] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [132694] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3251), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3253), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [132761] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [132828] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3667), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3669), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [132895] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3387), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3389), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [132962] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3255), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3257), 52,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [133029] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3391), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3393), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [133096] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3343), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3345), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [133163] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3347), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3349), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [133230] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3395), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3397), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [133297] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4577), 1,
      anon_sym_COMMA,
    STATE(2918), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [133368] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4577), 1,
      anon_sym_COMMA,
    STATE(2919), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [133439] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4579), 1,
      anon_sym_COMMA,
    STATE(2919), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [133510] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3399), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3401), 51,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [133577] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3742), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [133646] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3744), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [133715] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3746), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [133784] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3748), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [133853] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3750), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
      anon_sym_end,
  [133922] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    ACTIONS(4582), 1,
      sym__newline_before_do,
    STATE(4318), 1,
      sym_do_block,
    ACTIONS(3049), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3051), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [133995] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3231), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3233), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134061] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3417), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3419), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [134127] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134193] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3421), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3423), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [134259] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3429), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3431), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [134325] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134391] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134457] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3337), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [134527] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134593] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3339), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3341), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134659] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3549), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3551), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134725] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134791] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [134861] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134927] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3351), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3353), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [134993] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3159), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3161), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [135059] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3123), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3125), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135125] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3343), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3345), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [135191] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3347), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3349), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [135257] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3075), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3077), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135323] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3481), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3483), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135389] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3481), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3483), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135455] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3485), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3487), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135521] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3485), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3487), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135587] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3485), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3487), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135653] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135719] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3489), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3491), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135785] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3489), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3491), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [135851] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3355), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3357), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [135917] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3359), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3361), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [135983] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3363), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3365), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [136049] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_do,
  [136131] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 25,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_do,
  [136219] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4246), 1,
      anon_sym_PIPE,
    ACTIONS(4262), 1,
      anon_sym_EQ_GT,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [136321] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_DASH_GT,
      anon_sym_do,
  [136403] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 26,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_DASH_GT,
      anon_sym_do,
  [136491] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4366), 1,
      anon_sym_PIPE,
    ACTIONS(4382), 1,
      anon_sym_EQ_GT,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [136593] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [136659] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_DASH_GT,
      anon_sym_do,
  [136731] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_DASH_GT,
      anon_sym_do,
  [136809] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 46,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_DASH_GT,
      anon_sym_do,
  [136883] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4366), 1,
      anon_sym_PIPE,
    ACTIONS(4378), 1,
      anon_sym_when,
    ACTIONS(4380), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4382), 1,
      anon_sym_EQ_GT,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [136989] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4366), 1,
      anon_sym_PIPE,
    ACTIONS(4378), 1,
      anon_sym_when,
    ACTIONS(4380), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4382), 1,
      anon_sym_EQ_GT,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [137095] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_do,
  [137167] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4366), 1,
      anon_sym_PIPE,
    ACTIONS(4380), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4382), 1,
      anon_sym_EQ_GT,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [137271] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4382), 1,
      anon_sym_EQ_GT,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [137371] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_DASH_GT,
      anon_sym_do,
  [137469] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 14,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_DASH_GT,
      anon_sym_do,
  [137563] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 17,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_DASH_GT,
      anon_sym_do,
  [137655] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_DASH_GT,
      anon_sym_do,
  [137745] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_DASH_GT,
      anon_sym_do,
  [137831] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
      anon_sym_do,
  [137911] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
      anon_sym_do,
  [137991] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [138061] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_do,
  [138139] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3369), 45,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_do,
  [138213] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4246), 1,
      anon_sym_PIPE,
    ACTIONS(4258), 1,
      anon_sym_when,
    ACTIONS(4260), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4262), 1,
      anon_sym_EQ_GT,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [138319] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4246), 1,
      anon_sym_PIPE,
    ACTIONS(4258), 1,
      anon_sym_when,
    ACTIONS(4260), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4262), 1,
      anon_sym_EQ_GT,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [138425] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [138491] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4246), 1,
      anon_sym_PIPE,
    ACTIONS(4260), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4262), 1,
      anon_sym_EQ_GT,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_do,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [138595] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4262), 1,
      anon_sym_EQ_GT,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [138695] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_do,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [138793] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_do,
  [138887] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 16,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_do,
  [138979] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 21,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_do,
  [139069] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 34,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_do,
  [139155] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [139235] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [139315] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [139385] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3395), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3397), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [139451] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3399), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3401), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [139517] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3371), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3373), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [139583] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3387), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3389), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [139649] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3391), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3393), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [139715] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3541), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3543), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [139781] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3553), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3555), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [139847] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3561), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3563), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [139913] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [139979] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3375), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3377), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140045] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140111] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [140177] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140243] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140309] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140375] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 50,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [140445] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140511] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140577] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3403), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3405), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140643] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3617), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3619), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [140709] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [140775] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3569), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3571), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [140841] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140907] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [140973] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141039] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141105] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141171] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141237] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141303] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141369] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141435] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141501] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141567] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141633] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141699] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141765] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141831] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141897] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [141963] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142029] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142095] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142161] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142227] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142293] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3493), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3495), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142359] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3497), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3499), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142425] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3501), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3503), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142491] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3505), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3507), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142557] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3509), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3511), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142623] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3513), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3515), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142689] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3557), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3559), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [142755] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142821] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [142887] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_LPAREN,
    STATE(2673), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [142957] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3521), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3523), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [143023] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [143089] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3525), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3527), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [143155] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3075), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3077), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [143221] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3117), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3119), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [143287] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3123), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3125), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [143353] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [143419] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [143485] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 50,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [143555] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [143621] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3339), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3341), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [143687] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [143753] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [143819] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3529), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3531), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [143885] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3533), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3535), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [143951] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3537), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3539), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [144017] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3343), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3345), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144083] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3347), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3349), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144149] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3351), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3353), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144215] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3159), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3161), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144281] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [144347] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [144413] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3331), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3333), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144479] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3355), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3357), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144545] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3359), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3361), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144611] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3363), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3365), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144677] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [144743] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [144809] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144875] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [144941] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 50,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [145011] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3371), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3373), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145077] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [145143] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [145209] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3433), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3435), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [145275] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_LPAREN,
    STATE(2685), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145345] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_LPAREN,
    STATE(2846), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145415] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3375), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3377), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145481] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145547] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3379), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3381), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145613] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3379), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3381), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145679] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3383), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3385), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145745] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3383), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3385), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145811] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3379), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3381), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145877] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3403), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3405), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [145943] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3395), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3397), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146009] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3399), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3401), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146075] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4246), 1,
      anon_sym_PIPE,
    ACTIONS(4258), 1,
      anon_sym_when,
    ACTIONS(4260), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4262), 1,
      anon_sym_EQ_GT,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4256), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3643), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3645), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [146183] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146249] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146315] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146381] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146447] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146513] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146579] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146645] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146711] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146777] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [146843] = 29,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1351), 1,
      anon_sym_RPAREN,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4588), 1,
      aux_sym__terminator_token1,
    ACTIONS(4591), 1,
      anon_sym_SEMI,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4606), 1,
      anon_sym_when,
    ACTIONS(4608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    STATE(347), 1,
      sym__terminator,
    STATE(1036), 1,
      aux_sym__terminator_repeat1,
    STATE(5245), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4604), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [146959] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [147025] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [147091] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [147157] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [147223] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_do,
  [147305] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 26,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_do,
  [147393] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4428), 1,
      anon_sym_PIPE,
    ACTIONS(4444), 1,
      anon_sym_EQ_GT,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [147495] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_do,
  [147567] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_do,
  [147645] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 46,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_do,
  [147719] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4428), 1,
      anon_sym_PIPE,
    ACTIONS(4440), 1,
      anon_sym_when,
    ACTIONS(4442), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4444), 1,
      anon_sym_EQ_GT,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [147825] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4428), 1,
      anon_sym_PIPE,
    ACTIONS(4440), 1,
      anon_sym_when,
    ACTIONS(4442), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4444), 1,
      anon_sym_EQ_GT,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [147931] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4428), 1,
      anon_sym_PIPE,
    ACTIONS(4442), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4444), 1,
      anon_sym_EQ_GT,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_do,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [148035] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4444), 1,
      anon_sym_EQ_GT,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [148135] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_do,
  [148233] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_do,
  [148327] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 17,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_do,
  [148419] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 22,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_do,
  [148509] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_do,
  [148595] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [148675] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [148755] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [148821] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [148887] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [148953] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149019] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149085] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3131), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3133), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [149151] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149217] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149283] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149349] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149415] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149481] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149547] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149613] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3493), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3495), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149679] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3497), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3499), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149745] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3501), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3503), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149811] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3505), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3507), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [149877] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3643), 1,
      sym__newline_before_do,
    ACTIONS(4366), 1,
      anon_sym_PIPE,
    ACTIONS(4378), 1,
      anon_sym_when,
    ACTIONS(4380), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4382), 1,
      anon_sym_EQ_GT,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4376), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3645), 4,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [149985] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3509), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3511), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150051] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3513), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3515), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150117] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3517), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3519), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150183] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3517), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3519), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150249] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3521), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3523), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150315] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3517), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3519), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150381] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3525), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3527), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150447] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3141), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3143), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [150513] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3529), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3531), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150579] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3533), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3535), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150645] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3826), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [150713] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3828), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [150781] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3830), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [150849] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3537), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3539), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [150915] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3832), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [150983] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3834), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [151051] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_do,
      anon_sym_end,
  [151133] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 26,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_do,
      anon_sym_end,
  [151221] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4302), 1,
      anon_sym_PIPE,
    ACTIONS(4318), 1,
      anon_sym_EQ_GT,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_end,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [151323] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_do,
      anon_sym_end,
  [151395] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_do,
      anon_sym_end,
  [151473] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 46,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_do,
      anon_sym_end,
  [151547] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4302), 1,
      anon_sym_PIPE,
    ACTIONS(4314), 1,
      anon_sym_when,
    ACTIONS(4316), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4318), 1,
      anon_sym_EQ_GT,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
      anon_sym_end,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [151653] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4302), 1,
      anon_sym_PIPE,
    ACTIONS(4314), 1,
      anon_sym_when,
    ACTIONS(4316), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4318), 1,
      anon_sym_EQ_GT,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
      anon_sym_end,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [151759] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4302), 1,
      anon_sym_PIPE,
    ACTIONS(4316), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4318), 1,
      anon_sym_EQ_GT,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_do,
      anon_sym_end,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [151863] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4318), 1,
      anon_sym_EQ_GT,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_end,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [151963] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_do,
      anon_sym_end,
  [152061] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 14,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_do,
      anon_sym_end,
  [152155] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 17,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_do,
      anon_sym_end,
  [152247] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 22,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_do,
      anon_sym_end,
  [152337] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_do,
      anon_sym_end,
  [152423] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
      anon_sym_end,
  [152503] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
      anon_sym_end,
  [152583] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3836), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [152651] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3838), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [152719] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3840), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [152787] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3842), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [152855] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3844), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [152923] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3846), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [152991] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3848), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [153059] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3850), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [153127] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3852), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [153195] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3854), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [153263] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3856), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [153331] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3858), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [153399] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3860), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [153467] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3862), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [153535] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3864), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [153603] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3075), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3077), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [153671] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3123), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3125), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [153739] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3331), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3333), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [153805] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3671), 1,
      sym__newline_before_do,
    ACTIONS(4366), 1,
      anon_sym_PIPE,
    ACTIONS(4378), 1,
      anon_sym_when,
    ACTIONS(4380), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4382), 1,
      anon_sym_EQ_GT,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4376), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3673), 4,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [153913] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3433), 1,
      sym__newline_before_do,
    ACTIONS(4636), 1,
      anon_sym_PIPE,
    ACTIONS(4640), 1,
      anon_sym_COMMA,
    ACTIONS(4648), 1,
      anon_sym_when,
    ACTIONS(4650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4652), 1,
      anon_sym_EQ_GT,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    STATE(3209), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3435), 2,
      anon_sym_GT_GT,
      anon_sym_do,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4646), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [154025] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    ACTIONS(4678), 1,
      sym__newline_before_do,
    STATE(4437), 1,
      sym_do_block,
    ACTIONS(3053), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3055), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [154097] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    ACTIONS(4680), 1,
      sym__newline_before_do,
    STATE(4443), 1,
      sym_do_block,
    ACTIONS(3049), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3051), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [154169] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    ACTIONS(4682), 1,
      sym__newline_before_do,
    STATE(4483), 1,
      sym_do_block,
    ACTIONS(3049), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3051), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [154241] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    ACTIONS(4684), 1,
      sym__newline_before_do,
    STATE(4503), 1,
      sym_do_block,
    ACTIONS(3117), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3119), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [154313] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    ACTIONS(4686), 1,
      sym__newline_before_do,
    STATE(4567), 1,
      sym_do_block,
    ACTIONS(3049), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3051), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [154385] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4246), 1,
      anon_sym_PIPE,
    ACTIONS(4258), 1,
      anon_sym_when,
    ACTIONS(4260), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4262), 1,
      anon_sym_EQ_GT,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4256), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3671), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3673), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [154493] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4428), 1,
      anon_sym_PIPE,
    ACTIONS(4440), 1,
      anon_sym_when,
    ACTIONS(4442), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4444), 1,
      anon_sym_EQ_GT,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4438), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3645), 4,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [154601] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3107), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3109), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [154667] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4688), 1,
      anon_sym_COMMA,
    STATE(3248), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3637), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3639), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [154737] = 29,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9), 1,
      aux_sym__terminator_token1,
    ACTIONS(1819), 1,
      ts_builtin_sym_end,
    ACTIONS(4690), 1,
      anon_sym_SEMI,
    ACTIONS(4694), 1,
      anon_sym_PIPE,
    ACTIONS(4704), 1,
      anon_sym_when,
    ACTIONS(4706), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    STATE(403), 1,
      sym__terminator,
    STATE(1035), 1,
      aux_sym__terminator_repeat1,
    STATE(5931), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4702), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [154853] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [154919] = 29,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1355), 1,
      anon_sym_RPAREN,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4606), 1,
      anon_sym_when,
    ACTIONS(4608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(4734), 1,
      aux_sym__terminator_token1,
    ACTIONS(4737), 1,
      anon_sym_SEMI,
    STATE(351), 1,
      sym__terminator,
    STATE(1036), 1,
      aux_sym__terminator_repeat1,
    STATE(5674), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4604), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [155035] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3145), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3147), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [155101] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4740), 1,
      anon_sym_COMMA,
    STATE(3226), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155171] = 29,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9), 1,
      aux_sym__terminator_token1,
    ACTIONS(1607), 1,
      ts_builtin_sym_end,
    ACTIONS(4694), 1,
      anon_sym_PIPE,
    ACTIONS(4704), 1,
      anon_sym_when,
    ACTIONS(4706), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(4742), 1,
      anon_sym_SEMI,
    STATE(443), 1,
      sym__terminator,
    STATE(1035), 1,
      aux_sym__terminator_repeat1,
    STATE(5352), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4702), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [155287] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155353] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4118), 1,
      anon_sym_LPAREN,
    STATE(2668), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155423] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4118), 1,
      anon_sym_LPAREN,
    STATE(2674), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155493] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155559] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4118), 1,
      anon_sym_LPAREN,
    STATE(2926), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155629] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155695] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155761] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155827] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [155893] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4744), 1,
      anon_sym_COMMA,
    STATE(3239), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3654), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3656), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [155963] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4740), 1,
      anon_sym_COMMA,
    STATE(3240), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156033] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3545), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3547), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [156099] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156165] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156231] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156297] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156363] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156429] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156495] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156561] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156627] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156693] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156759] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3433), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3435), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [156825] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4744), 1,
      anon_sym_COMMA,
    STATE(3247), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3660), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3662), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [156895] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4746), 1,
      anon_sym_COMMA,
    STATE(3240), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [156965] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3387), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3389), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157031] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3391), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3393), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157097] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3343), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3345), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157163] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3347), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3349), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157229] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3395), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3397), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157295] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3399), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3401), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157361] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4749), 1,
      anon_sym_COMMA,
    STATE(3247), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3617), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3619), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157431] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4752), 1,
      anon_sym_COMMA,
    STATE(3248), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3643), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3645), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157501] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2701), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2703), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157567] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2705), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2707), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157633] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2709), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2711), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157699] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2665), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2667), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157765] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2669), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2671), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157831] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2673), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(2675), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [157897] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [157963] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3545), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3547), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [158029] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3740), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3053), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3055), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [158097] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3743), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3049), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3051), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [158165] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3744), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3049), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3051), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [158233] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3746), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3117), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3119), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [158301] = 5,
    ACTIONS(5), 1,
      sym_comment,
    STATE(3747), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3049), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3051), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [158369] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3071), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3073), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [158435] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4755), 1,
      anon_sym_COMMA,
    STATE(3292), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3660), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3662), 49,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [158505] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4428), 1,
      anon_sym_PIPE,
    ACTIONS(4440), 1,
      anon_sym_when,
    ACTIONS(4442), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4444), 1,
      anon_sym_EQ_GT,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3671), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4438), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3673), 4,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [158613] = 29,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1355), 1,
      anon_sym_end,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4757), 1,
      aux_sym__terminator_token1,
    ACTIONS(4760), 1,
      anon_sym_SEMI,
    ACTIONS(4765), 1,
      anon_sym_PIPE,
    ACTIONS(4775), 1,
      anon_sym_when,
    ACTIONS(4777), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    STATE(349), 1,
      sym__terminator,
    STATE(1032), 1,
      aux_sym__terminator_repeat1,
    STATE(5249), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4773), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [158729] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4803), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [158797] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4805), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [158865] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4807), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [158933] = 29,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1351), 1,
      anon_sym_end,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4765), 1,
      anon_sym_PIPE,
    ACTIONS(4775), 1,
      anon_sym_when,
    ACTIONS(4777), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(4809), 1,
      aux_sym__terminator_token1,
    ACTIONS(4812), 1,
      anon_sym_SEMI,
    STATE(350), 1,
      sym__terminator,
    STATE(1032), 1,
      aux_sym__terminator_repeat1,
    STATE(5398), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4773), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [159049] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4815), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159117] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4817), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159185] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4819), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159253] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4821), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159321] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4823), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159389] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4825), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159457] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4827), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159525] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4829), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159593] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4831), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159661] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4833), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159729] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4835), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159797] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4837), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159865] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4839), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [159933] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4841), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160001] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4843), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160069] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4845), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160137] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4847), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3573), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160205] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3063), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3065), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160271] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3067), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3069), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160337] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4302), 1,
      anon_sym_PIPE,
    ACTIONS(4314), 1,
      anon_sym_when,
    ACTIONS(4316), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4318), 1,
      anon_sym_EQ_GT,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3643), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4312), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3645), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_do,
      anon_sym_end,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [160445] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3093), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3095), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160511] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3617), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 51,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [160577] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4849), 1,
      anon_sym_COMMA,
    STATE(3292), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3617), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 49,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [160647] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3621), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3623), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160713] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3625), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3627), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160779] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3629), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3631), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160845] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3633), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3635), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160911] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3650), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3652), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [160977] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3163), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3165), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161043] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4302), 1,
      anon_sym_PIPE,
    ACTIONS(4314), 1,
      anon_sym_when,
    ACTIONS(4316), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4318), 1,
      anon_sym_EQ_GT,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3671), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4312), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3673), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_do,
      anon_sym_end,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [161151] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4302), 1,
      anon_sym_PIPE,
    ACTIONS(4314), 1,
      anon_sym_when,
    ACTIONS(4316), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4318), 1,
      anon_sym_EQ_GT,
    ACTIONS(4320), 1,
      anon_sym_EQ,
    ACTIONS(4330), 1,
      anon_sym_in,
    ACTIONS(4332), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4334), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4336), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(4340), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3675), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4304), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4310), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4312), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4322), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4324), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3677), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_do,
      anon_sym_end,
    ACTIONS(4300), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4326), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4308), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4328), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [161259] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3167), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3169), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161325] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3171), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3173), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161391] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3175), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3177), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161457] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3179), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3181), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161523] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3183), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3185), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161589] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3187), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3189), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161655] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3191), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3193), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161721] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3195), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3197), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161787] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3199), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3201), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161853] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3203), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3205), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161919] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3101), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3103), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [161985] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3151), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3153), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162051] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3045), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3047), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162117] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3207), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3209), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162183] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3211), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3213), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162249] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3215), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3217), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162315] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3219), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3221), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162381] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3223), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3225), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162447] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3227), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3229), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162513] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3235), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3237), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162579] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3239), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3241), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162645] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3243), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3245), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162711] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3247), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3249), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162777] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3251), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3253), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162843] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3667), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3669), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162909] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3255), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3257), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [162975] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3259), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3261), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163041] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3263), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3265), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163107] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3085), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3087), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163173] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3267), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3269), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163239] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3271), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3273), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163305] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3275), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3277), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163371] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3279), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3281), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163437] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3283), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3285), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163503] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3287), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3289), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163569] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3291), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3293), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163635] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3295), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3297), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163701] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3299), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3301), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163767] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3303), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3305), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163833] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3307), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3309), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163899] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3311), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3313), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [163965] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3315), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3317), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [164031] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3319), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3321), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [164097] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3323), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3325), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [164163] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3327), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3329), 50,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [164229] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3089), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3091), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [164295] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3097), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3099), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [164361] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3155), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3157), 50,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [164427] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3417), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3419), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [164493] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3421), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3423), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [164559] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3429), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3431), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [164625] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3075), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3077), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [164691] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3123), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3125), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [164757] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3123), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3125), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [164823] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3075), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3077), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [164889] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3481), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3483), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [164955] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3481), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3483), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165021] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3485), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3487), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165087] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3485), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3487), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165153] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3485), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3487), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165219] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165285] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3489), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3491), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165351] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3489), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3491), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165417] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3331), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3333), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165483] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3549), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3551), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165549] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2701), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [165615] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2705), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [165681] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3541), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3543), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165747] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3553), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3555), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165813] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3561), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3563), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165879] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3567), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [165945] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3675), 1,
      sym__newline_before_do,
    ACTIONS(4366), 1,
      anon_sym_PIPE,
    ACTIONS(4378), 1,
      anon_sym_when,
    ACTIONS(4380), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4382), 1,
      anon_sym_EQ_GT,
    ACTIONS(4384), 1,
      anon_sym_EQ,
    ACTIONS(4394), 1,
      anon_sym_in,
    ACTIONS(4396), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4398), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4400), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(4406), 1,
      sym__not_in,
    ACTIONS(4368), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4374), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4376), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4386), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4388), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3677), 4,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_do,
    ACTIONS(4364), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4390), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4372), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4392), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [166053] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3567), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [166119] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(3565), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_do,
  [166189] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3567), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [166255] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3569), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3571), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [166321] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3557), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3559), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [166387] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3417), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3419), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [166453] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3421), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3423), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [166519] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3429), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3431), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [166585] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3117), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3119), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [166651] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [166717] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [166783] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3337), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [166849] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3337), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [166915] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(3335), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_do,
  [166985] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3337), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167051] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3339), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3341), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167117] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2709), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [167183] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2665), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [167249] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3351), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3353), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167315] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3159), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3161), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167381] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3355), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3357), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167447] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3359), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3361), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167513] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3363), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3365), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167579] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3369), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167645] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3369), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167711] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4402), 1,
      anon_sym_DOT,
    ACTIONS(4404), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_do,
  [167781] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3371), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3373), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [167847] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [167913] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [167979] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3375), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3377), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168045] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3369), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168111] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3379), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3381), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168177] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3379), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3381), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168243] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3383), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3385), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168309] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3383), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3385), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168375] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3379), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3381), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168441] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3403), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3405), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168507] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2669), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [168573] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2673), 4,
      sym__newline_before_do,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [168639] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168705] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168771] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168837] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168903] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [168969] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169035] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169101] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169167] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169233] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169299] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169365] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169431] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169497] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169563] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169629] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169695] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169761] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169827] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169893] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [169959] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3493), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3495), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170025] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3497), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3499), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170091] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3501), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3503), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170157] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3505), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3507), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170223] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3509), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3511), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170289] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [170355] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [170421] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3513), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3515), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170487] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [170553] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3517), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3519), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170619] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3517), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3519), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170685] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [170751] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3521), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3523), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170817] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3517), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3519), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170883] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3525), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3527), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [170949] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3529), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3531), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [171015] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3533), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3535), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [171081] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [171147] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3537), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3539), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [171213] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3557), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3559), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [171279] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [171345] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [171411] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [171477] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [171543] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [171609] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [171675] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [171741] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [171807] = 27,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3433), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4854), 1,
      anon_sym_PIPE,
    ACTIONS(4858), 1,
      anon_sym_COMMA,
    ACTIONS(4866), 1,
      anon_sym_when,
    ACTIONS(4868), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4870), 1,
      anon_sym_EQ_GT,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    STATE(3461), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3435), 2,
      anon_sym_LBRACE,
      anon_sym_do,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4864), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [171919] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4892), 1,
      anon_sym_COMMA,
    STATE(3463), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3637), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3639), 48,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [171989] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [172055] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4894), 1,
      anon_sym_COMMA,
    STATE(3463), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3643), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3645), 48,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [172125] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [172191] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [172257] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [172323] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [172389] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [172455] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
      anon_sym_end,
  [172525] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
      anon_sym_end,
  [172595] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4338), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
      anon_sym_end,
  [172665] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [172731] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [172797] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [172863] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [172929] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [172995] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3541), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3543), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [173061] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3553), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3555), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [173127] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3561), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3563), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [173193] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [173259] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4246), 1,
      anon_sym_PIPE,
    ACTIONS(4258), 1,
      anon_sym_when,
    ACTIONS(4260), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4262), 1,
      anon_sym_EQ_GT,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4274), 1,
      anon_sym_in,
    ACTIONS(4276), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4278), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4280), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(4286), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4248), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4254), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4256), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3675), 3,
      sym__newline_before_do,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3677), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4266), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4268), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4244), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4270), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4252), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4272), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [173367] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [173433] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4282), 1,
      anon_sym_DOT,
    ACTIONS(4284), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3567), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [173503] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [173569] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3569), 5,
      sym__newline_before_do,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3571), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [173635] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [173701] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3387), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3389), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [173767] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3391), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3393), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
      anon_sym_do,
  [173833] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [173899] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [173965] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [174031] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [174097] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4428), 1,
      anon_sym_PIPE,
    ACTIONS(4440), 1,
      anon_sym_when,
    ACTIONS(4442), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4444), 1,
      anon_sym_EQ_GT,
    ACTIONS(4446), 1,
      anon_sym_EQ,
    ACTIONS(4456), 1,
      anon_sym_in,
    ACTIONS(4458), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4460), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4462), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(4466), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3675), 2,
      sym__newline_before_do,
      aux_sym__terminator_token1,
    ACTIONS(4430), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4436), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4438), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4448), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4450), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3677), 4,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4426), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4452), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4434), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4454), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [174205] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [174271] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3727), 1,
      anon_sym_LBRACK2,
    ACTIONS(4464), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [174341] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [174407] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [174473] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [174539] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4897), 1,
      anon_sym_COMMA,
    STATE(3500), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3654), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3656), 48,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [174609] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4897), 1,
      anon_sym_COMMA,
    STATE(3501), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3660), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3662), 48,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [174679] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4899), 1,
      anon_sym_COMMA,
    STATE(3501), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3617), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3619), 48,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [174749] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [174815] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [174881] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [174947] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [175013] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [175079] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 51,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [175145] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4755), 1,
      anon_sym_COMMA,
    STATE(3263), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3654), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3656), 49,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [175215] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [175280] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [175345] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [175410] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [175475] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [175540] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [175605] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [175670] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3433), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3435), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [175735] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4902), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [175802] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4904), 1,
      anon_sym_COMMA,
    STATE(3518), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3617), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 48,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [175871] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4907), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [175938] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3085), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3087), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [176003] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3131), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3133), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [176068] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3387), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3389), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [176133] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3391), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3393), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [176198] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3343), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3345), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [176263] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3347), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3349), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [176328] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3395), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3397), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [176393] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3399), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3401), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [176458] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3545), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3547), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [176523] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3549), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3551), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [176588] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3067), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3069), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [176653] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3101), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3103), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [176718] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4909), 1,
      anon_sym_COMMA,
    STATE(3518), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3660), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3662), 48,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [176787] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [176852] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3093), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3095), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [176917] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2701), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [176982] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2705), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [177047] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3151), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3153), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177112] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3643), 1,
      sym__newline_before_do,
    ACTIONS(4636), 1,
      anon_sym_PIPE,
    ACTIONS(4648), 1,
      anon_sym_when,
    ACTIONS(4650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4652), 1,
      anon_sym_EQ_GT,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4646), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3645), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_do,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [177219] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177284] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177349] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177414] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 5,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177479] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4911), 1,
      anon_sym_COMMA,
    STATE(3816), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 47,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177548] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3671), 1,
      sym__newline_before_do,
    ACTIONS(4636), 1,
      anon_sym_PIPE,
    ACTIONS(4648), 1,
      anon_sym_when,
    ACTIONS(4650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4652), 1,
      anon_sym_EQ_GT,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4646), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3673), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_do,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [177655] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3045), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3047), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177720] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4913), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177787] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 5,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177852] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 5,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [177917] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3071), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3073), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [177982] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4915), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [178049] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2709), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178114] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2665), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178179] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4917), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178246] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4919), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178313] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4921), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178380] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4923), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178447] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4925), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178514] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4927), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178581] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4929), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178648] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4931), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178715] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4933), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178782] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4935), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178849] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4937), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178916] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4939), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [178983] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4941), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [179050] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4943), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [179117] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4945), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [179184] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4947), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [179251] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4949), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [179318] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4951), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [179385] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4953), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [179452] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4955), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [179519] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [179584] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [179649] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [179714] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [179779] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [179844] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [179909] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [179974] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180039] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180104] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180169] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180234] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180299] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180364] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180429] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2701), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2703), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [180494] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2705), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2707), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [180559] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3071), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3073), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180624] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4957), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180691] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2709), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2711), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [180756] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2665), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2667), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [180821] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4082), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180888] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4084), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [180955] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2669), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2671), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [181020] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2673), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2675), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [181085] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [181150] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3063), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3065), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [181215] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [181280] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [181345] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3141), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3143), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [181410] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [181475] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [181540] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [181605] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3067), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3069), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [181670] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [181735] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [181800] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [181865] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [181930] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [181995] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [182060] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [182125] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [182190] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [182255] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3093), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3095), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [182320] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3617), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3619), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [182385] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [182452] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2669), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [182517] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2673), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [182582] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [182647] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4088), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [182714] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3621), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3623), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [182779] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3625), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3627), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [182844] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3629), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3631), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [182909] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3633), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3635), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [182974] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3650), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3652), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [183039] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3163), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3165), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [183104] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4090), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [183171] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4092), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [183238] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4094), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [183305] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4096), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [183372] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3167), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3169), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [183437] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4098), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [183504] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3961), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [183571] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3171), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3173), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [183636] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3175), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3177), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [183701] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [183766] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [183831] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3179), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3181), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [183896] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3183), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3185), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [183961] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [184026] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [184091] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3187), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3189), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [184156] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3191), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3193), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [184221] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [184286] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [184351] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3195), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3197), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [184416] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3199), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3201), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [184481] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3203), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3205), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [184546] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3417), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3419), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [184611] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3421), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3423), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [184676] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3429), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3431), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [184741] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3075), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3077), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [184806] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3123), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3125), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [184871] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3123), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3125), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [184936] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3075), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3077), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185001] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3481), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3483), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185066] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3481), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3483), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185131] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3485), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3487), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185196] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3485), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3487), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185261] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3485), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3487), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185326] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3029), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3031), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185391] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3489), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3491), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185456] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3489), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3491), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185521] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3331), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3333), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [185586] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [185653] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [185720] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3943), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [185787] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2701), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [185852] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [185917] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [185982] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [186047] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3541), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3543), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [186112] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3553), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3555), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [186177] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3561), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3563), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [186242] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3567), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [186307] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3675), 1,
      sym__newline_before_do,
    ACTIONS(4636), 1,
      anon_sym_PIPE,
    ACTIONS(4648), 1,
      anon_sym_when,
    ACTIONS(4650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4652), 1,
      anon_sym_EQ_GT,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4646), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3677), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_do,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [186414] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3567), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [186479] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(3565), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [186548] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3565), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3567), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [186613] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3569), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3571), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [186678] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3101), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3103), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [186743] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [186808] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3945), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [186875] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3151), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3153), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [186940] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [187005] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3045), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3047), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [187070] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3207), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3209), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [187135] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3211), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3213), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [187200] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3215), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3217), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [187265] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3219), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3221), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [187330] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3223), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3225), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [187395] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3947), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [187462] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3557), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3559), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [187527] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3949), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [187594] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3951), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [187661] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3227), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3229), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [187726] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3971), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [187793] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3117), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3119), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [187858] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4959), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [187925] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3981), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [187992] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3337), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [188057] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3337), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [188122] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(3335), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [188191] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3335), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3337), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [188256] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3339), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3341), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [188321] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3987), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [188388] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3989), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [188455] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2709), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [188520] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2665), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [188585] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [188650] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [188715] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3351), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3353), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [188780] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3159), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3161), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [188845] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4961), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [188912] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3355), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3357), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [188977] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3359), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3361), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [189042] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3363), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3365), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [189107] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_do,
  [189188] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 25,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_do,
  [189275] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4636), 1,
      anon_sym_PIPE,
    ACTIONS(4652), 1,
      anon_sym_EQ_GT,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [189376] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3369), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [189441] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_do,
  [189512] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_do,
  [189589] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 45,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_do,
  [189662] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4636), 1,
      anon_sym_PIPE,
    ACTIONS(4648), 1,
      anon_sym_when,
    ACTIONS(4650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4652), 1,
      anon_sym_EQ_GT,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [189767] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4636), 1,
      anon_sym_PIPE,
    ACTIONS(4648), 1,
      anon_sym_when,
    ACTIONS(4650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4652), 1,
      anon_sym_EQ_GT,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [189872] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3369), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [189937] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4636), 1,
      anon_sym_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4652), 1,
      anon_sym_EQ_GT,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_do,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [190040] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4652), 1,
      anon_sym_EQ_GT,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [190139] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4654), 1,
      anon_sym_EQ,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4656), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_do,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [190236] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4658), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_do,
  [190329] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4660), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 16,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_do,
  [190420] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4634), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4662), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 21,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_do,
  [190509] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4664), 1,
      anon_sym_in,
    ACTIONS(4666), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(4676), 1,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 34,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_do,
  [190594] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [190673] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4668), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4670), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4638), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4644), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4642), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [190752] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4672), 1,
      anon_sym_DOT,
    ACTIONS(4674), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_do,
  [190821] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3371), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3373), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [190886] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3375), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3377), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [190951] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3367), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3369), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [191016] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3131), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3133), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [191081] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3379), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3381), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [191146] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3379), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3381), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [191211] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3383), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3385), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [191276] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3383), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3385), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [191341] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3379), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3381), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [191406] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3403), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3405), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [191471] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 5,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [191536] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2669), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [191601] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [191666] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [191731] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [191796] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [191861] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4963), 1,
      anon_sym_COMMA,
    STATE(3814), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 48,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [191930] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [191995] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [192060] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [192125] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [192190] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_do,
  [192271] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 25,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_do,
  [192358] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4854), 1,
      anon_sym_PIPE,
    ACTIONS(4870), 1,
      anon_sym_EQ_GT,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [192459] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_do,
  [192530] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_do,
  [192607] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 45,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_do,
  [192680] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4854), 1,
      anon_sym_PIPE,
    ACTIONS(4866), 1,
      anon_sym_when,
    ACTIONS(4868), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4870), 1,
      anon_sym_EQ_GT,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [192785] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4854), 1,
      anon_sym_PIPE,
    ACTIONS(4866), 1,
      anon_sym_when,
    ACTIONS(4868), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4870), 1,
      anon_sym_EQ_GT,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_do,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [192890] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4854), 1,
      anon_sym_PIPE,
    ACTIONS(4868), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4870), 1,
      anon_sym_EQ_GT,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_do,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [192993] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4870), 1,
      anon_sym_EQ_GT,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_do,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [193092] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_do,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [193189] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_do,
  [193282] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 16,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_do,
  [193373] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 21,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_do,
  [193462] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 34,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_do,
  [193547] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [193626] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3367), 2,
      sym__newline_before_do,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_do,
  [193705] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [193770] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [193835] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [193900] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [193965] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [194030] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [194095] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [194160] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [194225] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [194290] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [194355] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [194420] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4965), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [194487] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3089), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3091), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [194552] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4967), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [194619] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [194684] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [194749] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [194814] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [194879] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [194944] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195009] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195074] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195139] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195204] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3425), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3427), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195269] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3493), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3495), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195334] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3497), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3499), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195399] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3501), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3503), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195464] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3505), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3507), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195529] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3509), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3511), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195594] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4969), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [195661] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3513), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3515), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195726] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3141), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3143), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [195791] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3517), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3519), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195856] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3517), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3519), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195921] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3521), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3523), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [195986] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3517), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3519), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [196051] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3525), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3527), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [196116] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4963), 1,
      anon_sym_COMMA,
    STATE(3983), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 48,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [196185] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3231), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3233), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [196250] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4971), 1,
      anon_sym_COMMA,
    STATE(3816), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 47,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [196319] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4974), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [196386] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4976), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [196453] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4978), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [196520] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3529), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3531), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [196585] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3533), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3535), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [196650] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3537), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3539), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [196715] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [196780] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [196845] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [196910] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [196975] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197040] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197105] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2673), 5,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197170] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3645), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [197275] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197340] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(2705), 5,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197405] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197470] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197535] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197600] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197665] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3235), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3237), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [197730] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197795] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197860] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [197925] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3239), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3241), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [197990] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3243), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3245), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [198055] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3247), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3249), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [198120] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3251), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3253), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [198185] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [198250] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [198315] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [198380] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [198445] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [198510] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3667), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3669), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [198575] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
  [198644] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3255), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3257), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [198709] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3259), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3261), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [198774] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3263), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3265), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [198839] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [198904] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3085), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3087), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [198969] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [199034] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [199099] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3673), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [199204] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3387), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3389), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [199269] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3391), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3393), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [199334] = 26,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5020), 1,
      anon_sym_COMMA,
    STATE(6013), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(5018), 2,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [199443] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3267), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3269), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [199508] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3271), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3273), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [199573] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3275), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3277), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [199638] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3279), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3281), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [199703] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3283), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3285), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [199768] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3287), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3289), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [199833] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3291), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3293), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [199898] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3295), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3297), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [199963] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3299), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3301), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200028] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3303), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3305), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200093] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3307), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3309), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200158] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3311), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3313), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200223] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3315), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3317), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200288] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3319), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3321), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200353] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3323), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3325), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200418] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3327), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3329), 50,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200483] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3089), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3091), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200548] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3097), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3099), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200613] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3155), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3157), 50,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [200678] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [200743] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [200808] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3097), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3099), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [200873] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3155), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3157), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [200938] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5022), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201005] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201070] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201135] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201200] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201265] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201330] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201395] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201460] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201525] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3671), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(5024), 1,
      anon_sym_when,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3673), 4,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [201632] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201697] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201762] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3343), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3345), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [201827] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3347), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3349), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [201892] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [201957] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [202022] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4082), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202089] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4084), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202156] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4086), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202223] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4088), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202290] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4090), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202357] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4092), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202424] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4094), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202491] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4096), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202558] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4098), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202625] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3961), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202692] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3943), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202759] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3945), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202826] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3947), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202893] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3949), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [202960] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3951), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [203027] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3971), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [203094] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3979), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [203161] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3981), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [203228] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3987), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [203295] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3989), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [203362] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [203427] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [203492] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [203557] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [203622] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [203687] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
  [203768] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 26,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [203853] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [203918] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [203983] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [204048] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [204113] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [204178] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [204243] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [204308] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [204373] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [204438] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [204503] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [204568] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3395), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3397), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [204633] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3399), 3,
      sym__newline_before_do,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3401), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_do,
  [204698] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3621), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3623), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [204763] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3625), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3627), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [204828] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3629), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3631), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [204893] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3633), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3635), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [204958] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3650), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3652), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [205023] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3163), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3165), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [205088] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [205153] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [205218] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [205283] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [205348] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [205447] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3279), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3281), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [205512] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
  [205583] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
  [205660] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 46,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
  [205733] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [205836] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [205901] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [205966] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [206031] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [206096] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [206161] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3187), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3189), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [206226] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3191), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3193), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [206291] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3195), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3197), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [206356] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3199), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3201), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [206421] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3203), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3205), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [206486] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [206589] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3677), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [206694] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [206795] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [206892] = 19,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
  [206987] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 14,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
  [207078] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 17,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
  [207167] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 22,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
  [207254] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [207337] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [207416] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [207495] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 50,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [207560] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [207625] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [207690] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4911), 1,
      anon_sym_COMMA,
    STATE(3543), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 47,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [207759] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5026), 1,
      anon_sym_COMMA,
    STATE(3983), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 48,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [207828] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [207893] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5029), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [207960] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5031), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [208027] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3675), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(5024), 1,
      anon_sym_when,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3791), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3677), 4,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [208134] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3207), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3209), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208199] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3211), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3213), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208264] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3215), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3217), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208329] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3219), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3221), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208394] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3223), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3225), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208459] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3227), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3229), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208524] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3231), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3233), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208589] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3235), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3237), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208654] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3239), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3241), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208719] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3243), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3245), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208784] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3247), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3249), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208849] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3251), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3253), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208914] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3667), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3669), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [208979] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3255), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3257), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209044] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3259), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3261), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209109] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3263), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3265), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209174] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3167), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3169), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [209239] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3171), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3173), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [209304] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3267), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3269), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209369] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3271), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3273), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209434] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3275), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3277), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209499] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3175), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3177), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [209564] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3283), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3285), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209629] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3287), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3289), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209694] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3291), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3293), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209759] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3295), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3297), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209824] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3299), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3301), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209889] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3303), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3305), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [209954] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3307), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3309), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [210019] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3311), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3313), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [210084] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3315), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3317), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [210149] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3319), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3321), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [210214] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3323), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3325), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [210279] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3327), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3329), 50,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [210344] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_DASH_GT,
  [210425] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 26,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_DASH_GT,
  [210512] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [210613] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3179), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3181), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [210678] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_DASH_GT,
  [210749] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 38,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_DASH_GT,
  [210826] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3369), 46,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_DASH_GT,
  [210899] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3643), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4854), 1,
      anon_sym_PIPE,
    ACTIONS(4866), 1,
      anon_sym_when,
    ACTIONS(4868), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4870), 1,
      anon_sym_EQ_GT,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4864), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3645), 3,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [211006] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3671), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4854), 1,
      anon_sym_PIPE,
    ACTIONS(4866), 1,
      anon_sym_when,
    ACTIONS(4868), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4870), 1,
      anon_sym_EQ_GT,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4864), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3673), 3,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [211113] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
  [211182] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3675), 1,
      sym__newline_before_do,
    ACTIONS(4112), 1,
      anon_sym_DOT,
    ACTIONS(4114), 1,
      anon_sym_LBRACK2,
    ACTIONS(4854), 1,
      anon_sym_PIPE,
    ACTIONS(4866), 1,
      anon_sym_when,
    ACTIONS(4868), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4870), 1,
      anon_sym_EQ_GT,
    ACTIONS(4872), 1,
      anon_sym_EQ,
    ACTIONS(4882), 1,
      anon_sym_in,
    ACTIONS(4884), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4886), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4888), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4890), 1,
      sym__not_in,
    ACTIONS(4856), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4862), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4864), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3677), 3,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_do,
    ACTIONS(4874), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4876), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4852), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4878), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4860), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4880), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [211289] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4342), 1,
      anon_sym_LPAREN,
    STATE(3202), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [211358] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5033), 1,
      anon_sym_COMMA,
    STATE(4035), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3654), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3656), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [211427] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5033), 1,
      anon_sym_COMMA,
    STATE(4036), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3660), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3662), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [211496] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5035), 1,
      anon_sym_COMMA,
    STATE(4036), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
      anon_sym_end,
  [211565] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4342), 1,
      anon_sym_LPAREN,
    STATE(3203), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [211634] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(5024), 1,
      anon_sym_when,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [211739] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4342), 1,
      anon_sym_LPAREN,
    STATE(3205), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [211808] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(5024), 1,
      anon_sym_when,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [211913] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3183), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3185), 49,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [211978] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3781), 1,
      anon_sym_PIPE,
    ACTIONS(3796), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [212081] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3798), 1,
      anon_sym_EQ_GT,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [212180] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3063), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3065), 49,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [212245] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3075), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3077), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [212312] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3123), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3125), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [212379] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5038), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [212446] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4909), 1,
      anon_sym_COMMA,
    STATE(3532), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3654), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3656), 48,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [212515] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5040), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [212582] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3800), 1,
      anon_sym_EQ,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3802), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_DASH_GT,
  [212679] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3804), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_DASH_GT,
  [212772] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3806), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 17,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_DASH_GT,
  [212863] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3779), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3808), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 22,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_DASH_GT,
  [212952] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3810), 1,
      anon_sym_in,
    ACTIONS(3812), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3820), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_DASH_GT,
  [213037] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
  [213116] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3814), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(3816), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3783), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3787), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
  [213195] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_LBRACK2,
    ACTIONS(3818), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
  [213264] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5042), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [213331] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5044), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [213398] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3979), 1,
      aux_sym_sigil_token3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3573), 3,
      sym__not_in,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3575), 49,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [213465] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3279), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3281), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [213529] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3029), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3031), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [213593] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [213657] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3283), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3285), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [213721] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
  [213801] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 25,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [213887] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [213987] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
  [214057] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
  [214133] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3369), 45,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
  [214205] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4606), 1,
      anon_sym_when,
    ACTIONS(4608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [214309] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4606), 1,
      anon_sym_when,
    ACTIONS(4608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [214413] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [214515] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [214613] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [214709] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
  [214801] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 16,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
  [214891] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 21,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
  [214979] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 34,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [215063] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3287), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3289), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215127] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3291), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3293), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215191] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3295), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3297), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215255] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3617), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [215319] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3299), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3301), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215383] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3303), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3305), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215447] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3307), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3309), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215511] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3311), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3313), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215575] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3315), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3317), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215639] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3319), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3321), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215703] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3323), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3325), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215767] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3327), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3329), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215831] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3089), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3091), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215895] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3097), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3099), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [215959] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3155), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3157), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [216023] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3375), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3377), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [216087] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3429), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3431), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [216151] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3417), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3419), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [216215] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3421), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3423), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [216279] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3429), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3431), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [216343] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [216407] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [216471] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [216535] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [216613] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [216691] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3675), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4765), 1,
      anon_sym_PIPE,
    ACTIONS(4775), 1,
      anon_sym_when,
    ACTIONS(4777), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4773), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3677), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_end,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [216797] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3075), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3077), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [216861] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [216929] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3123), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3125), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [216993] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3123), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3125), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [217057] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3075), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3077), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [217121] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3481), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3483), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [217185] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3481), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3483), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [217249] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [217313] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3485), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3487), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [217377] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3363), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3365), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [217441] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3485), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3487), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [217505] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3485), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3487), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [217569] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3075), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3077), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [217635] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4694), 1,
      anon_sym_PIPE,
    ACTIONS(4704), 1,
      anon_sym_when,
    ACTIONS(4706), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3675), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3677), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4702), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [217741] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3117), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3119), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [217805] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [217869] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3123), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3125), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [217935] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [217999] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3369), 46,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
  [218069] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [218133] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [218197] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
  [218273] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3343), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3345), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [218337] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3489), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3491), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [218401] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3489), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3491), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [218465] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [218529] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3489), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3491), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [218593] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5058), 1,
      anon_sym_when,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5056), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3673), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [218697] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3383), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3385), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [218761] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3071), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3073), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [218825] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [218889] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3387), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3389), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [218953] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3379), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3381), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [219017] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4694), 1,
      anon_sym_PIPE,
    ACTIONS(4704), 1,
      anon_sym_when,
    ACTIONS(4706), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3671), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3673), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4702), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [219123] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3391), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3393), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [219187] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3541), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3543), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [219251] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3553), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3555), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [219315] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3561), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3563), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [219379] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [219443] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5058), 1,
      anon_sym_when,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5056), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3677), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [219547] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [219611] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 1,
      sym__not_in,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 48,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
  [219679] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [219743] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3569), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3571), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [219807] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [219871] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3537), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3539), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [219935] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3331), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3333), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [219999] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [220063] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3513), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3515), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [220127] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3123), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3125), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [220191] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [220255] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3671), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4606), 1,
      anon_sym_when,
    ACTIONS(4608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4604), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3673), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [220361] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3367), 3,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3369), 44,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
  [220433] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4694), 1,
      anon_sym_PIPE,
    ACTIONS(4704), 1,
      anon_sym_when,
    ACTIONS(4706), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3369), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [220537] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4694), 1,
      anon_sym_PIPE,
    ACTIONS(4704), 1,
      anon_sym_when,
    ACTIONS(4706), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3369), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [220641] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 48,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [220709] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [220773] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3347), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3349), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [220837] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 34,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
  [220917] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [220981] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2709), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2711), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [221045] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2665), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2667), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [221109] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3521), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3523), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [221173] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3517), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3519), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [221237] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3557), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3559), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [221301] = 26,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5088), 1,
      anon_sym_RPAREN,
    ACTIONS(5090), 1,
      anon_sym_COMMA,
    ACTIONS(5093), 1,
      anon_sym_when,
    ACTIONS(5096), 1,
      anon_sym_DASH_GT,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5056), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [221409] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3075), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3077), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [221473] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3525), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3527), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [221537] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3117), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3119), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [221601] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3369), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [221665] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3343), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3345), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [221729] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3347), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3349), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [221793] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4694), 1,
      anon_sym_PIPE,
    ACTIONS(4706), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [221895] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [221959] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [222023] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 1,
      sym__not_in,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 48,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
  [222091] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [222189] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [222285] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 12,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
  [222377] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 15,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
  [222467] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 20,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
  [222555] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [222619] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3339), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3341), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [222683] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3671), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4765), 1,
      anon_sym_PIPE,
    ACTIONS(4775), 1,
      anon_sym_when,
    ACTIONS(4777), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4773), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3673), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_end,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [222789] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 33,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [222873] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3617), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3619), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [222937] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [223015] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [223079] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5098), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [223145] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3367), 3,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [223223] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5100), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [223289] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3343), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3345), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [223353] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3347), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3349), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [223417] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 3,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [223485] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3351), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3353), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [223549] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3159), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3161), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [223613] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5102), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [223679] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5104), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [223745] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5106), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [223811] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5108), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [223877] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5110), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [223943] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5112), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224009] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3395), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3397), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224073] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3399), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3401), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224137] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3395), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3397), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224201] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224265] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3371), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3373), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224329] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3399), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3401), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224393] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224457] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224521] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5114), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224587] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5116), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224653] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5118), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224719] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5120), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224785] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5122), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224851] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5124), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224917] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5126), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [224983] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5128), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [225049] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5130), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [225115] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5132), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [225181] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5134), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [225247] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5136), 1,
      aux_sym_sigil_token3,
    ACTIONS(3573), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3575), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [225313] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3355), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3357), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [225377] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3359), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3361), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [225441] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3363), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3365), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [225505] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_DASH_GT,
  [225585] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3387), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3389), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [225649] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3391), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3393), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [225713] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 25,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_DASH_GT,
  [225797] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [225895] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3541), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3543), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [225959] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [226023] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_DASH_GT,
  [226093] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_DASH_GT,
  [226169] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3529), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3531), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [226233] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3533), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3535), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [226297] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 45,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_DASH_GT,
  [226369] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [226433] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3537), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3539), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [226497] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5138), 1,
      anon_sym_COMMA,
    STATE(4278), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3654), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3656), 47,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [226565] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5058), 1,
      anon_sym_when,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [226667] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [226731] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5058), 1,
      anon_sym_when,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_DASH_GT,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [226833] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [226897] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [226961] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227025] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227089] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227153] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227217] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3131), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3133), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227281] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227345] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227409] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [227473] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [227573] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2701), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2703), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227637] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2705), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2707), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227701] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [227797] = 19,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_DASH_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [227891] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2709), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2711), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [227955] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2665), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2667), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [228019] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_DASH_GT,
  [228109] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 16,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_DASH_GT,
  [228197] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2669), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2671), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [228261] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2673), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2675), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [228325] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 21,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_DASH_GT,
  [228411] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 34,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_DASH_GT,
  [228493] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
  [228571] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_DASH_GT,
  [228649] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
  [228717] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3371), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3373), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [228781] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3553), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3555), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [228845] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [228909] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5138), 1,
      anon_sym_COMMA,
    STATE(4324), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3660), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3662), 47,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [228977] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 24,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [229063] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4694), 1,
      anon_sym_PIPE,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [229163] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3493), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3495), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [229227] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3557), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3559), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [229291] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3403), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3405), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [229355] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3501), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3503), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [229419] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3505), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3507), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [229483] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 3,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3337), 47,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [229551] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3509), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3511), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [229615] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3375), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3377), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [229679] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3337), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [229743] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3063), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3065), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [229807] = 26,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5140), 1,
      anon_sym_RPAREN,
    ACTIONS(5142), 1,
      anon_sym_COMMA,
    STATE(6663), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [229915] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [229979] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3359), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3361), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [230043] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3067), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3069), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [230107] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3379), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3381), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [230171] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3379), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3381), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [230235] = 26,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(2971), 1,
      anon_sym_DASH_GT,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5144), 1,
      anon_sym_when,
    STATE(6177), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5056), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [230343] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3383), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3385), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [230407] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_end,
  [230487] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 25,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_end,
  [230573] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4765), 1,
      anon_sym_PIPE,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_end,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [230673] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3383), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3385), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [230737] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_end,
  [230807] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 37,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_end,
  [230883] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3369), 45,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_end,
  [230955] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4765), 1,
      anon_sym_PIPE,
    ACTIONS(4775), 1,
      anon_sym_when,
    ACTIONS(4777), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_end,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [231059] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4765), 1,
      anon_sym_PIPE,
    ACTIONS(4775), 1,
      anon_sym_when,
    ACTIONS(4777), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_end,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [231163] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3355), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3357), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [231227] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4765), 1,
      anon_sym_PIPE,
    ACTIONS(4777), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_end,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [231329] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_end,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [231427] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 9,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_end,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [231523] = 18,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 13,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_end,
  [231615] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 16,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_end,
  [231705] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 21,
      anon_sym_SEMI,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_end,
  [231793] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 34,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_end,
  [231877] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_end,
  [231955] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_end,
  [232033] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3379), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3381), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [232097] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3403), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3405), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [232161] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [232225] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3395), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3397), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [232289] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3399), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3401), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [232353] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3093), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3095), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [232417] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5147), 1,
      anon_sym_COMMA,
    STATE(4324), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3617), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 47,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [232485] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3481), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3483), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [232549] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [232613] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2669), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2671), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [232677] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2673), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2675), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [232741] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [232805] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [232869] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [232933] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [232997] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [233061] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [233125] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [233189] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [233253] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [233317] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3421), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3423), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [233381] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [233445] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [233509] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3621), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3623), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [233573] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3625), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3627), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [233637] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3629), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3631), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [233701] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3633), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3635), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [233765] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3650), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3652), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [233829] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [233893] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [233957] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3163), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3165), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [234021] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3141), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3143), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [234085] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [234149] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [234213] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3167), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3169), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [234277] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3171), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3173), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [234341] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [234405] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [234469] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [234533] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [234597] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3175), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3177), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [234661] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3561), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3563), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [234725] = 26,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5018), 1,
      anon_sym_GT_GT,
    ACTIONS(5152), 1,
      anon_sym_PIPE,
    ACTIONS(5156), 1,
      anon_sym_COMMA,
    ACTIONS(5164), 1,
      anon_sym_when,
    ACTIONS(5166), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5168), 1,
      anon_sym_EQ_GT,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    STATE(6479), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5162), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [234833] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3179), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3181), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [234897] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3183), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3185), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [234961] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3485), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3487), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [235025] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3187), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3189), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [235089] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3191), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3193), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [235153] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3195), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3197), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [235217] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3199), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3201), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [235281] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [235345] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3203), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3205), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [235409] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [235473] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [235537] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [235601] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3493), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3495), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [235665] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3497), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3499), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [235729] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3501), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3503), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [235793] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3417), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3419), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [235857] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3675), 1,
      aux_sym__terminator_token1,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4606), 1,
      anon_sym_when,
    ACTIONS(4608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4604), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3677), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [235963] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [236027] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3567), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [236091] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3101), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3103), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [236155] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 3,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(3567), 47,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [236223] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3151), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3153), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [236287] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3505), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3507), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [236351] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3509), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3511), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [236415] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3339), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3341), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [236479] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3513), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3515), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [236543] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3045), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3047), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [236607] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3517), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3519), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [236671] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3517), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3519), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [236735] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3521), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3523), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [236799] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3331), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3333), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [236863] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_end,
  [236931] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3337), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_end,
  [236999] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_end,
  [237067] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3517), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3519), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [237131] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      aux_sym__terminator_token1,
    ACTIONS(3567), 48,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [237199] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3525), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3527), 49,
      anon_sym_RPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DASH_GT,
      anon_sym_DOT,
  [237263] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5194), 1,
      anon_sym_COMMA,
    STATE(4399), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3654), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3656), 47,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237331] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5194), 1,
      anon_sym_COMMA,
    STATE(4400), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3660), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3662), 47,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237399] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5196), 1,
      anon_sym_COMMA,
    STATE(4400), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3617), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 47,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237467] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3207), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3209), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237531] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3211), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3213), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237595] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3215), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3217), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237659] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3219), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3221), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237723] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3529), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3531), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237787] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3533), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3535), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237851] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3223), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3225), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237915] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3227), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3229), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [237979] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3231), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3233), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238043] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3235), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3237), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238107] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3239), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3241), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238171] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3243), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3245), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238235] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3247), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3249), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238299] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3251), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3253), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238363] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3351), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3353), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238427] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3159), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3161), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238491] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3667), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3669), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238555] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3255), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3257), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238619] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3259), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3261), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238683] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3263), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3265), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238747] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2701), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2703), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238811] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(2705), 3,
      sym__not_in,
      aux_sym_quoted_keyword_token1,
      anon_sym_LBRACK2,
    ACTIONS(2707), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238875] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3387), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3389), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [238939] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3391), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3393), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239003] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3085), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3087), 49,
      anon_sym_LPAREN,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239067] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3425), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3427), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239131] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3569), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3571), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239195] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3267), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3269), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239259] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3271), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3273), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239323] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3275), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3277), 49,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      aux_sym_sigil_token3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239387] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(3497), 4,
      sym__not_in,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
      anon_sym_LBRACK2,
    ACTIONS(3499), 48,
      anon_sym_SEMI,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239451] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_PIPE,
    ACTIONS(5166), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5168), 1,
      anon_sym_EQ_GT,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [239550] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [239627] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239690] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239753] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3339), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3341), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239816] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3375), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3377), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [239879] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [239956] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3159), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3161), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [240019] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [240082] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3117), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3119), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [240145] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_PIPE,
    ACTIONS(5164), 1,
      anon_sym_when,
    ACTIONS(5166), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5168), 1,
      anon_sym_EQ_GT,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(3645), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5162), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [240248] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3379), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3381), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [240311] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4288), 1,
      aux_sym__terminator_token1,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4606), 1,
      anon_sym_when,
    ACTIONS(4608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4290), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4604), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [240416] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5093), 1,
      anon_sym_when,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5056), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(5096), 2,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [240519] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4349), 1,
      aux_sym__terminator_token1,
    ACTIONS(4765), 1,
      anon_sym_PIPE,
    ACTIONS(4775), 1,
      anon_sym_when,
    ACTIONS(4777), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4779), 1,
      anon_sym_EQ_GT,
    ACTIONS(4781), 1,
      anon_sym_EQ,
    ACTIONS(4791), 1,
      anon_sym_in,
    ACTIONS(4793), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4795), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4797), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4799), 1,
      anon_sym_DOT,
    ACTIONS(4801), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4351), 2,
      anon_sym_SEMI,
      anon_sym_end,
    ACTIONS(4767), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4773), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4783), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4785), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4763), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4787), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4769), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4789), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [240624] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4351), 1,
      anon_sym_SEMI,
    ACTIONS(4694), 1,
      anon_sym_PIPE,
    ACTIONS(4704), 1,
      anon_sym_when,
    ACTIONS(4706), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4708), 1,
      anon_sym_EQ_GT,
    ACTIONS(4710), 1,
      anon_sym_EQ,
    ACTIONS(4720), 1,
      anon_sym_in,
    ACTIONS(4722), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4724), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4726), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4728), 1,
      anon_sym_DOT,
    ACTIONS(4730), 1,
      anon_sym_LBRACK2,
    ACTIONS(4732), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4349), 2,
      ts_builtin_sym_end,
      aux_sym__terminator_token1,
    ACTIONS(4696), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4700), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4702), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4712), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4714), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4692), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4716), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4698), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4718), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [240729] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 34,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
  [240808] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3331), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3333), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [240871] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 24,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [240954] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5219), 1,
      anon_sym_PIPE,
    ACTIONS(5221), 1,
      anon_sym_EQ_GT,
    ACTIONS(5223), 1,
      anon_sym_EQ,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5225), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5227), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [241051] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(5088), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [241154] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 46,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
  [241223] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3343), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3345), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [241286] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [241349] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
  [241424] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 44,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
  [241495] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5219), 1,
      anon_sym_PIPE,
    ACTIONS(5221), 1,
      anon_sym_EQ_GT,
    ACTIONS(5223), 1,
      anon_sym_EQ,
    ACTIONS(5231), 1,
      anon_sym_when,
    ACTIONS(5233), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5225), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5227), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3369), 4,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [241596] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [241659] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3355), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3357), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [241722] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3359), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3361), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [241785] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3347), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3349), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [241848] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [241911] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [241974] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242037] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242100] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3351), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3353), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242163] = 21,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_PIPE,
    ACTIONS(5168), 1,
      anon_sym_EQ_GT,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(3369), 6,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [242260] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242323] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242386] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242449] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242512] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3493), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3495), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242575] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3497), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3499), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242638] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3501), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3503), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242701] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3505), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3507), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242764] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3509), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3511), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242827] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [242890] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5219), 1,
      anon_sym_PIPE,
    ACTIONS(5221), 1,
      anon_sym_EQ_GT,
    ACTIONS(5223), 1,
      anon_sym_EQ,
    ACTIONS(5231), 1,
      anon_sym_when,
    ACTIONS(5233), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5225), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5227), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3369), 4,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [242991] = 7,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 46,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
  [243060] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 36,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
  [243135] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3557), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3559), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [243198] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3379), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3381), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [243261] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 44,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
  [243332] = 25,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3910), 1,
      anon_sym_LBRACK2,
    ACTIONS(4349), 1,
      aux_sym__terminator_token1,
    ACTIONS(4596), 1,
      anon_sym_PIPE,
    ACTIONS(4606), 1,
      anon_sym_when,
    ACTIONS(4608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4610), 1,
      anon_sym_EQ_GT,
    ACTIONS(4612), 1,
      anon_sym_EQ,
    ACTIONS(4622), 1,
      anon_sym_in,
    ACTIONS(4624), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(4626), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(4628), 1,
      anon_sym_STAR_STAR,
    ACTIONS(4630), 1,
      anon_sym_DOT,
    ACTIONS(4632), 1,
      sym__not_in,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4351), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    ACTIONS(4598), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4602), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4604), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(4614), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(4616), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4594), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(4618), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4600), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(4620), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [243437] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3513), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3515), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [243500] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3383), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3385), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [243563] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3417), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3419), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [243626] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3421), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3423), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [243689] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3429), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3431), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [243752] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3517), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3519), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [243815] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3517), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3519), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [243878] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5219), 1,
      anon_sym_PIPE,
    ACTIONS(5221), 1,
      anon_sym_EQ_GT,
    ACTIONS(5223), 1,
      anon_sym_EQ,
    ACTIONS(5231), 1,
      anon_sym_when,
    ACTIONS(5233), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3673), 2,
      anon_sym_LBRACE,
      anon_sym_COMMA,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5235), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5225), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5227), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [243981] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3075), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3077), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244044] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_PIPE,
    ACTIONS(5164), 1,
      anon_sym_when,
    ACTIONS(5166), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5168), 1,
      anon_sym_EQ_GT,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3369), 4,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [244145] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5219), 1,
      anon_sym_PIPE,
    ACTIONS(5221), 1,
      anon_sym_EQ_GT,
    ACTIONS(5223), 1,
      anon_sym_EQ,
    ACTIONS(5231), 1,
      anon_sym_when,
    ACTIONS(5233), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3677), 2,
      anon_sym_LBRACE,
      anon_sym_COMMA,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5235), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5225), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5227), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [244248] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_PIPE,
    ACTIONS(5164), 1,
      anon_sym_when,
    ACTIONS(5166), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5168), 1,
      anon_sym_EQ_GT,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(3369), 4,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [244349] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244412] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3123), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3125), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244475] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3075), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3077), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244538] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3521), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3523), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244601] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3123), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3125), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244664] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3383), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3385), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244727] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 1,
      sym__not_in,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 47,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [244794] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3403), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3405), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244857] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3617), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244920] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [244983] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5168), 1,
      anon_sym_EQ_GT,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [245078] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3481), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3483), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [245141] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3481), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3483), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [245204] = 19,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [245297] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 12,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
  [245386] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 15,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
  [245473] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5219), 1,
      anon_sym_PIPE,
    ACTIONS(5221), 1,
      anon_sym_EQ_GT,
    ACTIONS(5223), 1,
      anon_sym_EQ,
    ACTIONS(5233), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5225), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5227), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3369), 5,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [245572] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3485), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3487), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [245635] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3517), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3519), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [245698] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3485), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3487), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [245761] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3525), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3527), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [245824] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3331), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3824), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3333), 46,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [245889] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 20,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
  [245974] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3485), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3487), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [246037] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 33,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [246118] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [246195] = 11,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 35,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
  [246272] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3363), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3365), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [246335] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3395), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3397), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [246398] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3029), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3031), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [246461] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3399), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3401), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [246524] = 12,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 34,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
  [246603] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3367), 1,
      sym__not_in,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3369), 47,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [246670] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3371), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3373), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [246733] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3489), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3491), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [246796] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3489), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3491), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [246859] = 20,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5221), 1,
      anon_sym_EQ_GT,
    ACTIONS(5223), 1,
      anon_sym_EQ,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5225), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5227), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 7,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [246954] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_PIPE,
    ACTIONS(5164), 1,
      anon_sym_when,
    ACTIONS(5166), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5168), 1,
      anon_sym_EQ_GT,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(3673), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5162), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [247057] = 19,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5223), 1,
      anon_sym_EQ,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5225), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5227), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 8,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [247150] = 17,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5227), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 12,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
  [247239] = 16,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5229), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 15,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
  [247326] = 15,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5217), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5211), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 20,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
  [247411] = 14,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
    ACTIONS(3369), 24,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
  [247494] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [247557] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(5205), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5207), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5209), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5213), 1,
      anon_sym_in,
    ACTIONS(5215), 1,
      sym__not_in,
    ACTIONS(5199), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5203), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5201), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(3369), 33,
      anon_sym_LBRACE,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [247638] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3339), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3931), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3341), 46,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [247703] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [247766] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3387), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3389), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [247829] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3391), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3393), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [247892] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3541), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3543), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [247955] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3553), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3555), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248018] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3561), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3563), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248081] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248144] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3335), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3337), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248207] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248270] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_PIPE,
    ACTIONS(5164), 1,
      anon_sym_when,
    ACTIONS(5166), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5168), 1,
      anon_sym_EQ_GT,
    ACTIONS(5170), 1,
      anon_sym_EQ,
    ACTIONS(5180), 1,
      anon_sym_in,
    ACTIONS(5182), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5184), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5186), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(5192), 1,
      sym__not_in,
    ACTIONS(3677), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
    ACTIONS(5154), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5162), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5172), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5174), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5150), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5176), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5158), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5178), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [248373] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248436] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 1,
      sym__not_in,
    ACTIONS(5188), 1,
      anon_sym_DOT,
    ACTIONS(5190), 1,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 47,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
  [248503] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3565), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3567), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248566] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3569), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3571), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248629] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3529), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3531), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248692] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3533), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3535), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248755] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248818] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248881] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3537), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3539), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [248944] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [249007] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [249070] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [249133] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3425), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3427), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [249196] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3379), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3381), 48,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [249259] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5237), 1,
      anon_sym_LBRACE,
    ACTIONS(3481), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3483), 46,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [249323] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5239), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [249425] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5241), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [249527] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5243), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [249629] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5245), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [249731] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5247), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [249833] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(894), 1,
      anon_sym_LBRACE,
    ACTIONS(233), 2,
      sym__not_in,
      anon_sym_LBRACK2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(235), 46,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_SLASH,
      anon_sym_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
      anon_sym_when,
      anon_sym_COLON_COLON,
      anon_sym_EQ_GT,
      anon_sym_EQ,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
      anon_sym_in,
      anon_sym_CARET_CARET_CARET,
      anon_sym_SLASH_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_DOT,
  [249897] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5249), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [249999] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5251), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [250101] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5253), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [250203] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5255), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [250305] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5058), 1,
      anon_sym_when,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5257), 1,
      anon_sym_DASH_GT,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5056), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [250407] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_PIPE,
    ACTIONS(5058), 1,
      anon_sym_when,
    ACTIONS(5060), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5062), 1,
      anon_sym_EQ_GT,
    ACTIONS(5064), 1,
      anon_sym_EQ,
    ACTIONS(5074), 1,
      anon_sym_in,
    ACTIONS(5076), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5078), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5080), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5082), 1,
      anon_sym_DOT,
    ACTIONS(5084), 1,
      anon_sym_LBRACK2,
    ACTIONS(5086), 1,
      sym__not_in,
    ACTIONS(5259), 1,
      anon_sym_DASH_GT,
    ACTIONS(5050), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(5054), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(5056), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5066), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5068), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(5046), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5070), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(5052), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5072), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [250509] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5261), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [250611] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5263), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [250713] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5265), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [250815] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5267), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [250917] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5269), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251019] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5271), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251121] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5273), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251223] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5275), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251325] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5277), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251427] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5279), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251529] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5281), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251631] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5283), 1,
      anon_sym_RBRACK,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251733] = 24,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(5285), 1,
      anon_sym_RBRACE,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251835] = 23,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4584), 1,
      anon_sym_DOT,
    ACTIONS(4586), 1,
      anon_sym_LBRACK2,
    ACTIONS(4982), 1,
      anon_sym_PIPE,
    ACTIONS(4992), 1,
      anon_sym_when,
    ACTIONS(4994), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4996), 1,
      anon_sym_EQ_GT,
    ACTIONS(4998), 1,
      anon_sym_EQ,
    ACTIONS(5008), 1,
      anon_sym_in,
    ACTIONS(5010), 1,
      anon_sym_CARET_CARET_CARET,
    ACTIONS(5012), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(5014), 1,
      anon_sym_STAR_STAR,
    ACTIONS(5016), 1,
      sym__not_in,
    ACTIONS(4984), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(4988), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4990), 2,
      anon_sym_LT_DASH,
      anon_sym_BSLASH_BSLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5000), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_PIPE_PIPE_PIPE,
      anon_sym_or,
    ACTIONS(5002), 3,
      anon_sym_AMP_AMP,
      anon_sym_AMP_AMP_AMP,
      anon_sym_and,
    ACTIONS(4980), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(5004), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(4986), 6,
      anon_sym_DOT_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS_PLUS,
      anon_sym_DASH_DASH_DASH,
      anon_sym_LT_GT,
    ACTIONS(5006), 9,
      anon_sym_PIPE_GT,
      anon_sym_LT_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT_TILDE,
      anon_sym_TILDE_GT_GT,
      anon_sym_LT_TILDE,
      anon_sym_TILDE_GT,
      anon_sym_LT_TILDE_GT,
      anon_sym_LT_PIPE_GT,
  [251934] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5287), 1,
      anon_sym_LPAREN,
    ACTIONS(5289), 1,
      anon_sym_DQUOTE,
    ACTIONS(5291), 1,
      anon_sym_SQUOTE,
    ACTIONS(5293), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5295), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5297), 1,
      anon_sym_LBRACE,
    ACTIONS(5299), 1,
      anon_sym_LBRACK,
    ACTIONS(5301), 1,
      anon_sym_LT,
    ACTIONS(5303), 1,
      anon_sym_PIPE,
    ACTIONS(5305), 1,
      anon_sym_SLASH,
    STATE(2762), 1,
      sym__quoted_i_double,
    STATE(2764), 1,
      sym__quoted_i_single,
    STATE(2766), 1,
      sym__quoted_i_heredoc_single,
    STATE(2767), 1,
      sym__quoted_i_heredoc_double,
    STATE(2769), 1,
      sym__quoted_i_parenthesis,
    STATE(2770), 1,
      sym__quoted_i_curly,
    STATE(2771), 1,
      sym__quoted_i_square,
    STATE(2773), 1,
      sym__quoted_i_angle,
    STATE(2774), 1,
      sym__quoted_i_bar,
    STATE(2775), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252003] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5307), 1,
      anon_sym_LPAREN,
    ACTIONS(5309), 1,
      anon_sym_DQUOTE,
    ACTIONS(5311), 1,
      anon_sym_SQUOTE,
    ACTIONS(5313), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5315), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5317), 1,
      anon_sym_LBRACE,
    ACTIONS(5319), 1,
      anon_sym_LBRACK,
    ACTIONS(5321), 1,
      anon_sym_LT,
    ACTIONS(5323), 1,
      anon_sym_PIPE,
    ACTIONS(5325), 1,
      anon_sym_SLASH,
    STATE(2203), 1,
      sym__quoted_i_double,
    STATE(2205), 1,
      sym__quoted_i_single,
    STATE(2207), 1,
      sym__quoted_i_heredoc_single,
    STATE(2210), 1,
      sym__quoted_i_heredoc_double,
    STATE(2212), 1,
      sym__quoted_i_parenthesis,
    STATE(2213), 1,
      sym__quoted_i_curly,
    STATE(2217), 1,
      sym__quoted_i_square,
    STATE(2231), 1,
      sym__quoted_i_angle,
    STATE(2238), 1,
      sym__quoted_i_bar,
    STATE(2247), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252072] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5327), 1,
      anon_sym_LPAREN,
    ACTIONS(5329), 1,
      anon_sym_DQUOTE,
    ACTIONS(5331), 1,
      anon_sym_SQUOTE,
    ACTIONS(5333), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5335), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5337), 1,
      anon_sym_LBRACE,
    ACTIONS(5339), 1,
      anon_sym_LBRACK,
    ACTIONS(5341), 1,
      anon_sym_LT,
    ACTIONS(5343), 1,
      anon_sym_PIPE,
    ACTIONS(5345), 1,
      anon_sym_SLASH,
    STATE(2251), 1,
      sym__quoted_double,
    STATE(2271), 1,
      sym__quoted_single,
    STATE(2279), 1,
      sym__quoted_heredoc_single,
    STATE(2330), 1,
      sym__quoted_heredoc_double,
    STATE(2339), 1,
      sym__quoted_parenthesis,
    STATE(2340), 1,
      sym__quoted_curly,
    STATE(2341), 1,
      sym__quoted_square,
    STATE(2358), 1,
      sym__quoted_angle,
    STATE(2359), 1,
      sym__quoted_bar,
    STATE(2360), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252141] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5347), 1,
      anon_sym_LPAREN,
    ACTIONS(5349), 1,
      anon_sym_DQUOTE,
    ACTIONS(5351), 1,
      anon_sym_SQUOTE,
    ACTIONS(5353), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5355), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5357), 1,
      anon_sym_LBRACE,
    ACTIONS(5359), 1,
      anon_sym_LBRACK,
    ACTIONS(5361), 1,
      anon_sym_LT,
    ACTIONS(5363), 1,
      anon_sym_PIPE,
    ACTIONS(5365), 1,
      anon_sym_SLASH,
    STATE(3563), 1,
      sym__quoted_double,
    STATE(3564), 1,
      sym__quoted_single,
    STATE(3565), 1,
      sym__quoted_heredoc_single,
    STATE(3566), 1,
      sym__quoted_heredoc_double,
    STATE(3567), 1,
      sym__quoted_parenthesis,
    STATE(3568), 1,
      sym__quoted_curly,
    STATE(3569), 1,
      sym__quoted_square,
    STATE(3570), 1,
      sym__quoted_angle,
    STATE(3571), 1,
      sym__quoted_bar,
    STATE(3572), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252210] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5367), 1,
      anon_sym_LPAREN,
    ACTIONS(5369), 1,
      anon_sym_DQUOTE,
    ACTIONS(5371), 1,
      anon_sym_SQUOTE,
    ACTIONS(5373), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5375), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5377), 1,
      anon_sym_LBRACE,
    ACTIONS(5379), 1,
      anon_sym_LBRACK,
    ACTIONS(5381), 1,
      anon_sym_LT,
    ACTIONS(5383), 1,
      anon_sym_PIPE,
    ACTIONS(5385), 1,
      anon_sym_SLASH,
    STATE(4194), 1,
      sym__quoted_i_double,
    STATE(4196), 1,
      sym__quoted_i_single,
    STATE(4202), 1,
      sym__quoted_i_heredoc_single,
    STATE(4203), 1,
      sym__quoted_i_heredoc_double,
    STATE(4204), 1,
      sym__quoted_i_parenthesis,
    STATE(4205), 1,
      sym__quoted_i_curly,
    STATE(4206), 1,
      sym__quoted_i_square,
    STATE(4207), 1,
      sym__quoted_i_angle,
    STATE(4216), 1,
      sym__quoted_i_bar,
    STATE(4217), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252279] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5387), 1,
      anon_sym_LPAREN,
    ACTIONS(5389), 1,
      anon_sym_DQUOTE,
    ACTIONS(5391), 1,
      anon_sym_SQUOTE,
    ACTIONS(5393), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5395), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5397), 1,
      anon_sym_LBRACE,
    ACTIONS(5399), 1,
      anon_sym_LBRACK,
    ACTIONS(5401), 1,
      anon_sym_LT,
    ACTIONS(5403), 1,
      anon_sym_PIPE,
    ACTIONS(5405), 1,
      anon_sym_SLASH,
    STATE(3158), 1,
      sym__quoted_i_double,
    STATE(3159), 1,
      sym__quoted_i_single,
    STATE(3160), 1,
      sym__quoted_i_heredoc_single,
    STATE(3162), 1,
      sym__quoted_i_heredoc_double,
    STATE(3163), 1,
      sym__quoted_i_parenthesis,
    STATE(3181), 1,
      sym__quoted_i_curly,
    STATE(3182), 1,
      sym__quoted_i_square,
    STATE(3183), 1,
      sym__quoted_i_angle,
    STATE(3184), 1,
      sym__quoted_i_bar,
    STATE(3185), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252348] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5407), 1,
      anon_sym_LPAREN,
    ACTIONS(5409), 1,
      anon_sym_DQUOTE,
    ACTIONS(5411), 1,
      anon_sym_SQUOTE,
    ACTIONS(5413), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5415), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5417), 1,
      anon_sym_LBRACE,
    ACTIONS(5419), 1,
      anon_sym_LBRACK,
    ACTIONS(5421), 1,
      anon_sym_LT,
    ACTIONS(5423), 1,
      anon_sym_PIPE,
    ACTIONS(5425), 1,
      anon_sym_SLASH,
    STATE(3186), 1,
      sym__quoted_double,
    STATE(3187), 1,
      sym__quoted_single,
    STATE(3188), 1,
      sym__quoted_heredoc_single,
    STATE(3189), 1,
      sym__quoted_heredoc_double,
    STATE(3190), 1,
      sym__quoted_parenthesis,
    STATE(3191), 1,
      sym__quoted_curly,
    STATE(3192), 1,
      sym__quoted_square,
    STATE(3193), 1,
      sym__quoted_angle,
    STATE(3194), 1,
      sym__quoted_bar,
    STATE(3195), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252417] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5427), 1,
      anon_sym_LPAREN,
    ACTIONS(5429), 1,
      anon_sym_DQUOTE,
    ACTIONS(5431), 1,
      anon_sym_SQUOTE,
    ACTIONS(5433), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5435), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5437), 1,
      anon_sym_LBRACE,
    ACTIONS(5439), 1,
      anon_sym_LBRACK,
    ACTIONS(5441), 1,
      anon_sym_LT,
    ACTIONS(5443), 1,
      anon_sym_PIPE,
    ACTIONS(5445), 1,
      anon_sym_SLASH,
    STATE(2086), 1,
      sym__quoted_i_double,
    STATE(2087), 1,
      sym__quoted_i_single,
    STATE(2088), 1,
      sym__quoted_i_heredoc_single,
    STATE(2089), 1,
      sym__quoted_i_heredoc_double,
    STATE(2090), 1,
      sym__quoted_i_parenthesis,
    STATE(2091), 1,
      sym__quoted_i_curly,
    STATE(2092), 1,
      sym__quoted_i_square,
    STATE(2094), 1,
      sym__quoted_i_angle,
    STATE(2095), 1,
      sym__quoted_i_bar,
    STATE(2096), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252486] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5447), 1,
      anon_sym_LPAREN,
    ACTIONS(5449), 1,
      anon_sym_DQUOTE,
    ACTIONS(5451), 1,
      anon_sym_SQUOTE,
    ACTIONS(5453), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5455), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5457), 1,
      anon_sym_LBRACE,
    ACTIONS(5459), 1,
      anon_sym_LBRACK,
    ACTIONS(5461), 1,
      anon_sym_LT,
    ACTIONS(5463), 1,
      anon_sym_PIPE,
    ACTIONS(5465), 1,
      anon_sym_SLASH,
    STATE(2098), 1,
      sym__quoted_double,
    STATE(2099), 1,
      sym__quoted_single,
    STATE(2102), 1,
      sym__quoted_heredoc_single,
    STATE(2103), 1,
      sym__quoted_heredoc_double,
    STATE(2104), 1,
      sym__quoted_parenthesis,
    STATE(2106), 1,
      sym__quoted_curly,
    STATE(2107), 1,
      sym__quoted_square,
    STATE(2108), 1,
      sym__quoted_angle,
    STATE(2112), 1,
      sym__quoted_bar,
    STATE(2113), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252555] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5467), 1,
      anon_sym_LPAREN,
    ACTIONS(5469), 1,
      anon_sym_DQUOTE,
    ACTIONS(5471), 1,
      anon_sym_SQUOTE,
    ACTIONS(5473), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5475), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5477), 1,
      anon_sym_LBRACE,
    ACTIONS(5479), 1,
      anon_sym_LBRACK,
    ACTIONS(5481), 1,
      anon_sym_LT,
    ACTIONS(5483), 1,
      anon_sym_PIPE,
    ACTIONS(5485), 1,
      anon_sym_SLASH,
    STATE(4218), 1,
      sym__quoted_double,
    STATE(4219), 1,
      sym__quoted_single,
    STATE(4220), 1,
      sym__quoted_heredoc_single,
    STATE(4221), 1,
      sym__quoted_heredoc_double,
    STATE(4222), 1,
      sym__quoted_parenthesis,
    STATE(4223), 1,
      sym__quoted_curly,
    STATE(4224), 1,
      sym__quoted_square,
    STATE(4225), 1,
      sym__quoted_angle,
    STATE(4226), 1,
      sym__quoted_bar,
    STATE(4227), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252624] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5487), 1,
      anon_sym_LPAREN,
    ACTIONS(5489), 1,
      anon_sym_DQUOTE,
    ACTIONS(5491), 1,
      anon_sym_SQUOTE,
    ACTIONS(5493), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5495), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5497), 1,
      anon_sym_LBRACE,
    ACTIONS(5499), 1,
      anon_sym_LBRACK,
    ACTIONS(5501), 1,
      anon_sym_LT,
    ACTIONS(5503), 1,
      anon_sym_PIPE,
    ACTIONS(5505), 1,
      anon_sym_SLASH,
    STATE(2372), 1,
      sym__quoted_i_slash,
    STATE(2747), 1,
      sym__quoted_i_double,
    STATE(2850), 1,
      sym__quoted_i_single,
    STATE(2852), 1,
      sym__quoted_i_heredoc_single,
    STATE(2881), 1,
      sym__quoted_i_heredoc_double,
    STATE(2882), 1,
      sym__quoted_i_parenthesis,
    STATE(2884), 1,
      sym__quoted_i_curly,
    STATE(2885), 1,
      sym__quoted_i_square,
    STATE(2886), 1,
      sym__quoted_i_angle,
    STATE(2887), 1,
      sym__quoted_i_bar,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252693] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5507), 1,
      anon_sym_LPAREN,
    ACTIONS(5509), 1,
      anon_sym_DQUOTE,
    ACTIONS(5511), 1,
      anon_sym_SQUOTE,
    ACTIONS(5513), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5515), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5517), 1,
      anon_sym_LBRACE,
    ACTIONS(5519), 1,
      anon_sym_LBRACK,
    ACTIONS(5521), 1,
      anon_sym_LT,
    ACTIONS(5523), 1,
      anon_sym_PIPE,
    ACTIONS(5525), 1,
      anon_sym_SLASH,
    STATE(2373), 1,
      sym__quoted_double,
    STATE(2455), 1,
      sym__quoted_single,
    STATE(2456), 1,
      sym__quoted_heredoc_single,
    STATE(2460), 1,
      sym__quoted_heredoc_double,
    STATE(2461), 1,
      sym__quoted_parenthesis,
    STATE(2462), 1,
      sym__quoted_curly,
    STATE(2463), 1,
      sym__quoted_square,
    STATE(2464), 1,
      sym__quoted_angle,
    STATE(2465), 1,
      sym__quoted_bar,
    STATE(2466), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252762] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5527), 1,
      anon_sym_LPAREN,
    ACTIONS(5529), 1,
      anon_sym_DQUOTE,
    ACTIONS(5531), 1,
      anon_sym_SQUOTE,
    ACTIONS(5533), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5535), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5537), 1,
      anon_sym_LBRACE,
    ACTIONS(5539), 1,
      anon_sym_LBRACK,
    ACTIONS(5541), 1,
      anon_sym_LT,
    ACTIONS(5543), 1,
      anon_sym_PIPE,
    ACTIONS(5545), 1,
      anon_sym_SLASH,
    STATE(3266), 1,
      sym__quoted_i_double,
    STATE(3267), 1,
      sym__quoted_i_single,
    STATE(3268), 1,
      sym__quoted_i_heredoc_single,
    STATE(3270), 1,
      sym__quoted_i_heredoc_double,
    STATE(3271), 1,
      sym__quoted_i_parenthesis,
    STATE(3272), 1,
      sym__quoted_i_curly,
    STATE(3273), 1,
      sym__quoted_i_square,
    STATE(3274), 1,
      sym__quoted_i_angle,
    STATE(3275), 1,
      sym__quoted_i_bar,
    STATE(3276), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252831] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5547), 1,
      anon_sym_LPAREN,
    ACTIONS(5549), 1,
      anon_sym_DQUOTE,
    ACTIONS(5551), 1,
      anon_sym_SQUOTE,
    ACTIONS(5553), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5555), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5557), 1,
      anon_sym_LBRACE,
    ACTIONS(5559), 1,
      anon_sym_LBRACK,
    ACTIONS(5561), 1,
      anon_sym_LT,
    ACTIONS(5563), 1,
      anon_sym_PIPE,
    ACTIONS(5565), 1,
      anon_sym_SLASH,
    STATE(3277), 1,
      sym__quoted_double,
    STATE(3278), 1,
      sym__quoted_single,
    STATE(3279), 1,
      sym__quoted_heredoc_single,
    STATE(3280), 1,
      sym__quoted_heredoc_double,
    STATE(3281), 1,
      sym__quoted_parenthesis,
    STATE(3282), 1,
      sym__quoted_curly,
    STATE(3283), 1,
      sym__quoted_square,
    STATE(3284), 1,
      sym__quoted_angle,
    STATE(3285), 1,
      sym__quoted_bar,
    STATE(3286), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252900] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5567), 1,
      anon_sym_LPAREN,
    ACTIONS(5569), 1,
      anon_sym_DQUOTE,
    ACTIONS(5571), 1,
      anon_sym_SQUOTE,
    ACTIONS(5573), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5575), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5577), 1,
      anon_sym_LBRACE,
    ACTIONS(5579), 1,
      anon_sym_LBRACK,
    ACTIONS(5581), 1,
      anon_sym_LT,
    ACTIONS(5583), 1,
      anon_sym_PIPE,
    ACTIONS(5585), 1,
      anon_sym_SLASH,
    STATE(3668), 1,
      sym__quoted_double,
    STATE(3684), 1,
      sym__quoted_single,
    STATE(3693), 1,
      sym__quoted_heredoc_single,
    STATE(3695), 1,
      sym__quoted_heredoc_double,
    STATE(3696), 1,
      sym__quoted_parenthesis,
    STATE(3698), 1,
      sym__quoted_curly,
    STATE(3701), 1,
      sym__quoted_angle,
    STATE(3707), 1,
      sym__quoted_bar,
    STATE(3708), 1,
      sym__quoted_slash,
    STATE(4060), 1,
      sym__quoted_square,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [252969] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5587), 1,
      anon_sym_LPAREN,
    ACTIONS(5589), 1,
      anon_sym_DQUOTE,
    ACTIONS(5591), 1,
      anon_sym_SQUOTE,
    ACTIONS(5593), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5595), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5597), 1,
      anon_sym_LBRACE,
    ACTIONS(5599), 1,
      anon_sym_LBRACK,
    ACTIONS(5601), 1,
      anon_sym_LT,
    ACTIONS(5603), 1,
      anon_sym_PIPE,
    ACTIONS(5605), 1,
      anon_sym_SLASH,
    STATE(2776), 1,
      sym__quoted_double,
    STATE(2777), 1,
      sym__quoted_single,
    STATE(2778), 1,
      sym__quoted_heredoc_single,
    STATE(2779), 1,
      sym__quoted_heredoc_double,
    STATE(2780), 1,
      sym__quoted_parenthesis,
    STATE(2781), 1,
      sym__quoted_curly,
    STATE(2784), 1,
      sym__quoted_square,
    STATE(2785), 1,
      sym__quoted_angle,
    STATE(2786), 1,
      sym__quoted_bar,
    STATE(2788), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253038] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5607), 1,
      anon_sym_LPAREN,
    ACTIONS(5609), 1,
      anon_sym_DQUOTE,
    ACTIONS(5611), 1,
      anon_sym_SQUOTE,
    ACTIONS(5613), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5615), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5617), 1,
      anon_sym_LBRACE,
    ACTIONS(5619), 1,
      anon_sym_LBRACK,
    ACTIONS(5621), 1,
      anon_sym_LT,
    ACTIONS(5623), 1,
      anon_sym_PIPE,
    ACTIONS(5625), 1,
      anon_sym_SLASH,
    STATE(3553), 1,
      sym__quoted_i_double,
    STATE(3554), 1,
      sym__quoted_i_single,
    STATE(3555), 1,
      sym__quoted_i_heredoc_single,
    STATE(3556), 1,
      sym__quoted_i_heredoc_double,
    STATE(3557), 1,
      sym__quoted_i_parenthesis,
    STATE(3558), 1,
      sym__quoted_i_curly,
    STATE(3559), 1,
      sym__quoted_i_square,
    STATE(3560), 1,
      sym__quoted_i_angle,
    STATE(3561), 1,
      sym__quoted_i_bar,
    STATE(3562), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253107] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5627), 1,
      anon_sym_LPAREN,
    ACTIONS(5629), 1,
      anon_sym_DQUOTE,
    ACTIONS(5631), 1,
      anon_sym_SQUOTE,
    ACTIONS(5633), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5635), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5637), 1,
      anon_sym_LBRACE,
    ACTIONS(5639), 1,
      anon_sym_LBRACK,
    ACTIONS(5641), 1,
      anon_sym_LT,
    ACTIONS(5643), 1,
      anon_sym_PIPE,
    ACTIONS(5645), 1,
      anon_sym_SLASH,
    STATE(3902), 1,
      sym__quoted_i_double,
    STATE(3903), 1,
      sym__quoted_i_single,
    STATE(3904), 1,
      sym__quoted_i_heredoc_single,
    STATE(3905), 1,
      sym__quoted_i_heredoc_double,
    STATE(3906), 1,
      sym__quoted_i_parenthesis,
    STATE(3907), 1,
      sym__quoted_i_curly,
    STATE(3908), 1,
      sym__quoted_i_square,
    STATE(3909), 1,
      sym__quoted_i_angle,
    STATE(3910), 1,
      sym__quoted_i_bar,
    STATE(3911), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253176] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5647), 1,
      anon_sym_LPAREN,
    ACTIONS(5649), 1,
      anon_sym_DQUOTE,
    ACTIONS(5651), 1,
      anon_sym_SQUOTE,
    ACTIONS(5653), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5655), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5657), 1,
      anon_sym_LBRACE,
    ACTIONS(5659), 1,
      anon_sym_LBRACK,
    ACTIONS(5661), 1,
      anon_sym_LT,
    ACTIONS(5663), 1,
      anon_sym_PIPE,
    ACTIONS(5665), 1,
      anon_sym_SLASH,
    STATE(3912), 1,
      sym__quoted_double,
    STATE(3913), 1,
      sym__quoted_single,
    STATE(3914), 1,
      sym__quoted_heredoc_single,
    STATE(3915), 1,
      sym__quoted_heredoc_double,
    STATE(3916), 1,
      sym__quoted_parenthesis,
    STATE(3917), 1,
      sym__quoted_curly,
    STATE(3918), 1,
      sym__quoted_square,
    STATE(3919), 1,
      sym__quoted_angle,
    STATE(3920), 1,
      sym__quoted_bar,
    STATE(3921), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253245] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5667), 1,
      anon_sym_LPAREN,
    ACTIONS(5669), 1,
      anon_sym_DQUOTE,
    ACTIONS(5671), 1,
      anon_sym_SQUOTE,
    ACTIONS(5673), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5675), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5677), 1,
      anon_sym_LBRACE,
    ACTIONS(5679), 1,
      anon_sym_LBRACK,
    ACTIONS(5681), 1,
      anon_sym_LT,
    ACTIONS(5683), 1,
      anon_sym_PIPE,
    ACTIONS(5685), 1,
      anon_sym_SLASH,
    STATE(3517), 1,
      sym__quoted_i_angle,
    STATE(3519), 1,
      sym__quoted_i_curly,
    STATE(3700), 1,
      sym__quoted_i_parenthesis,
    STATE(3788), 1,
      sym__quoted_i_bar,
    STATE(3985), 1,
      sym__quoted_i_slash,
    STATE(3986), 1,
      sym__quoted_i_square,
    STATE(4047), 1,
      sym__quoted_i_double,
    STATE(4049), 1,
      sym__quoted_i_single,
    STATE(4058), 1,
      sym__quoted_i_heredoc_single,
    STATE(4059), 1,
      sym__quoted_i_heredoc_double,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253314] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5687), 1,
      anon_sym_LPAREN,
    ACTIONS(5689), 1,
      anon_sym_DQUOTE,
    ACTIONS(5691), 1,
      anon_sym_SQUOTE,
    ACTIONS(5693), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5695), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5697), 1,
      anon_sym_LBRACE,
    ACTIONS(5699), 1,
      anon_sym_LBRACK,
    ACTIONS(5701), 1,
      anon_sym_LT,
    ACTIONS(5703), 1,
      anon_sym_PIPE,
    ACTIONS(5705), 1,
      anon_sym_SLASH,
    STATE(2853), 1,
      sym__quoted_i_double,
    STATE(2855), 1,
      sym__quoted_i_single,
    STATE(2858), 1,
      sym__quoted_i_heredoc_single,
    STATE(2859), 1,
      sym__quoted_i_heredoc_double,
    STATE(2860), 1,
      sym__quoted_i_parenthesis,
    STATE(2863), 1,
      sym__quoted_i_curly,
    STATE(2864), 1,
      sym__quoted_i_square,
    STATE(2865), 1,
      sym__quoted_i_angle,
    STATE(2866), 1,
      sym__quoted_i_bar,
    STATE(2867), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253383] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5707), 1,
      anon_sym_LPAREN,
    ACTIONS(5709), 1,
      anon_sym_DQUOTE,
    ACTIONS(5711), 1,
      anon_sym_SQUOTE,
    ACTIONS(5713), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5715), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5717), 1,
      anon_sym_LBRACE,
    ACTIONS(5719), 1,
      anon_sym_LBRACK,
    ACTIONS(5721), 1,
      anon_sym_LT,
    ACTIONS(5723), 1,
      anon_sym_PIPE,
    ACTIONS(5725), 1,
      anon_sym_SLASH,
    STATE(2868), 1,
      sym__quoted_double,
    STATE(2869), 1,
      sym__quoted_single,
    STATE(2870), 1,
      sym__quoted_heredoc_single,
    STATE(2872), 1,
      sym__quoted_heredoc_double,
    STATE(2873), 1,
      sym__quoted_parenthesis,
    STATE(2874), 1,
      sym__quoted_curly,
    STATE(2875), 1,
      sym__quoted_square,
    STATE(2876), 1,
      sym__quoted_angle,
    STATE(2877), 1,
      sym__quoted_bar,
    STATE(2880), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253452] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5727), 1,
      anon_sym_LPAREN,
    ACTIONS(5729), 1,
      anon_sym_DQUOTE,
    ACTIONS(5731), 1,
      anon_sym_SQUOTE,
    ACTIONS(5733), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5735), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5737), 1,
      anon_sym_LBRACE,
    ACTIONS(5739), 1,
      anon_sym_LBRACK,
    ACTIONS(5741), 1,
      anon_sym_LT,
    ACTIONS(5743), 1,
      anon_sym_PIPE,
    ACTIONS(5745), 1,
      anon_sym_SLASH,
    STATE(3593), 1,
      sym__quoted_i_double,
    STATE(3594), 1,
      sym__quoted_i_single,
    STATE(3617), 1,
      sym__quoted_i_heredoc_single,
    STATE(3621), 1,
      sym__quoted_i_heredoc_double,
    STATE(3628), 1,
      sym__quoted_i_parenthesis,
    STATE(3629), 1,
      sym__quoted_i_curly,
    STATE(3630), 1,
      sym__quoted_i_square,
    STATE(3631), 1,
      sym__quoted_i_angle,
    STATE(3633), 1,
      sym__quoted_i_bar,
    STATE(3634), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253521] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5747), 1,
      anon_sym_LPAREN,
    ACTIONS(5749), 1,
      anon_sym_DQUOTE,
    ACTIONS(5751), 1,
      anon_sym_SQUOTE,
    ACTIONS(5753), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5755), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5757), 1,
      anon_sym_LBRACE,
    ACTIONS(5759), 1,
      anon_sym_LBRACK,
    ACTIONS(5761), 1,
      anon_sym_LT,
    ACTIONS(5763), 1,
      anon_sym_PIPE,
    ACTIONS(5765), 1,
      anon_sym_SLASH,
    STATE(1655), 1,
      sym__quoted_i_double,
    STATE(1656), 1,
      sym__quoted_i_single,
    STATE(1657), 1,
      sym__quoted_i_heredoc_single,
    STATE(1658), 1,
      sym__quoted_i_heredoc_double,
    STATE(1659), 1,
      sym__quoted_i_parenthesis,
    STATE(1660), 1,
      sym__quoted_i_curly,
    STATE(1662), 1,
      sym__quoted_i_square,
    STATE(1663), 1,
      sym__quoted_i_angle,
    STATE(1664), 1,
      sym__quoted_i_bar,
    STATE(1665), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253590] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5767), 1,
      anon_sym_LPAREN,
    ACTIONS(5769), 1,
      anon_sym_DQUOTE,
    ACTIONS(5771), 1,
      anon_sym_SQUOTE,
    ACTIONS(5773), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5775), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5777), 1,
      anon_sym_LBRACE,
    ACTIONS(5779), 1,
      anon_sym_LBRACK,
    ACTIONS(5781), 1,
      anon_sym_LT,
    ACTIONS(5783), 1,
      anon_sym_PIPE,
    ACTIONS(5785), 1,
      anon_sym_SLASH,
    STATE(1666), 1,
      sym__quoted_double,
    STATE(1667), 1,
      sym__quoted_single,
    STATE(1668), 1,
      sym__quoted_heredoc_single,
    STATE(1669), 1,
      sym__quoted_heredoc_double,
    STATE(1670), 1,
      sym__quoted_parenthesis,
    STATE(1671), 1,
      sym__quoted_curly,
    STATE(1672), 1,
      sym__quoted_square,
    STATE(1673), 1,
      sym__quoted_angle,
    STATE(1674), 1,
      sym__quoted_bar,
    STATE(1675), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253659] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5787), 1,
      anon_sym_LPAREN,
    ACTIONS(5789), 1,
      anon_sym_DQUOTE,
    ACTIONS(5791), 1,
      anon_sym_SQUOTE,
    ACTIONS(5793), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5795), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5797), 1,
      anon_sym_LBRACE,
    ACTIONS(5799), 1,
      anon_sym_LBRACK,
    ACTIONS(5801), 1,
      anon_sym_LT,
    ACTIONS(5803), 1,
      anon_sym_PIPE,
    ACTIONS(5805), 1,
      anon_sym_SLASH,
    STATE(2726), 1,
      sym__quoted_i_double,
    STATE(2727), 1,
      sym__quoted_i_single,
    STATE(2728), 1,
      sym__quoted_i_heredoc_single,
    STATE(2729), 1,
      sym__quoted_i_heredoc_double,
    STATE(2730), 1,
      sym__quoted_i_parenthesis,
    STATE(2731), 1,
      sym__quoted_i_curly,
    STATE(2732), 1,
      sym__quoted_i_square,
    STATE(2733), 1,
      sym__quoted_i_angle,
    STATE(2734), 1,
      sym__quoted_i_bar,
    STATE(2735), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253728] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5807), 1,
      anon_sym_LPAREN,
    ACTIONS(5809), 1,
      anon_sym_DQUOTE,
    ACTIONS(5811), 1,
      anon_sym_SQUOTE,
    ACTIONS(5813), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5815), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5817), 1,
      anon_sym_LBRACE,
    ACTIONS(5819), 1,
      anon_sym_LBRACK,
    ACTIONS(5821), 1,
      anon_sym_LT,
    ACTIONS(5823), 1,
      anon_sym_PIPE,
    ACTIONS(5825), 1,
      anon_sym_SLASH,
    STATE(2736), 1,
      sym__quoted_double,
    STATE(2737), 1,
      sym__quoted_single,
    STATE(2738), 1,
      sym__quoted_heredoc_single,
    STATE(2739), 1,
      sym__quoted_heredoc_double,
    STATE(2740), 1,
      sym__quoted_parenthesis,
    STATE(2741), 1,
      sym__quoted_curly,
    STATE(2742), 1,
      sym__quoted_square,
    STATE(2743), 1,
      sym__quoted_angle,
    STATE(2744), 1,
      sym__quoted_bar,
    STATE(2745), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253797] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5827), 1,
      anon_sym_LPAREN,
    ACTIONS(5829), 1,
      anon_sym_DQUOTE,
    ACTIONS(5831), 1,
      anon_sym_SQUOTE,
    ACTIONS(5833), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5835), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5837), 1,
      anon_sym_LBRACE,
    ACTIONS(5839), 1,
      anon_sym_LBRACK,
    ACTIONS(5841), 1,
      anon_sym_LT,
    ACTIONS(5843), 1,
      anon_sym_PIPE,
    ACTIONS(5845), 1,
      anon_sym_SLASH,
    STATE(2890), 1,
      sym__quoted_i_double,
    STATE(2891), 1,
      sym__quoted_i_single,
    STATE(2892), 1,
      sym__quoted_i_heredoc_single,
    STATE(2893), 1,
      sym__quoted_i_heredoc_double,
    STATE(2894), 1,
      sym__quoted_i_parenthesis,
    STATE(2895), 1,
      sym__quoted_i_curly,
    STATE(2897), 1,
      sym__quoted_i_square,
    STATE(2899), 1,
      sym__quoted_i_angle,
    STATE(2921), 1,
      sym__quoted_i_bar,
    STATE(2922), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253866] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5847), 1,
      anon_sym_LPAREN,
    ACTIONS(5849), 1,
      anon_sym_DQUOTE,
    ACTIONS(5851), 1,
      anon_sym_SQUOTE,
    ACTIONS(5853), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5855), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5857), 1,
      anon_sym_LBRACE,
    ACTIONS(5859), 1,
      anon_sym_LBRACK,
    ACTIONS(5861), 1,
      anon_sym_LT,
    ACTIONS(5863), 1,
      anon_sym_PIPE,
    ACTIONS(5865), 1,
      anon_sym_SLASH,
    STATE(2365), 1,
      sym__quoted_heredoc_double,
    STATE(2366), 1,
      sym__quoted_parenthesis,
    STATE(2367), 1,
      sym__quoted_curly,
    STATE(2368), 1,
      sym__quoted_square,
    STATE(2369), 1,
      sym__quoted_angle,
    STATE(2370), 1,
      sym__quoted_bar,
    STATE(2371), 1,
      sym__quoted_slash,
    STATE(2923), 1,
      sym__quoted_double,
    STATE(2924), 1,
      sym__quoted_single,
    STATE(2925), 1,
      sym__quoted_heredoc_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [253935] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5867), 1,
      anon_sym_LPAREN,
    ACTIONS(5869), 1,
      anon_sym_DQUOTE,
    ACTIONS(5871), 1,
      anon_sym_SQUOTE,
    ACTIONS(5873), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5875), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5877), 1,
      anon_sym_LBRACE,
    ACTIONS(5879), 1,
      anon_sym_LBRACK,
    ACTIONS(5881), 1,
      anon_sym_LT,
    ACTIONS(5883), 1,
      anon_sym_PIPE,
    ACTIONS(5885), 1,
      anon_sym_SLASH,
    STATE(1976), 1,
      sym__quoted_i_slash,
    STATE(2131), 1,
      sym__quoted_i_double,
    STATE(2132), 1,
      sym__quoted_i_single,
    STATE(2133), 1,
      sym__quoted_i_heredoc_single,
    STATE(2134), 1,
      sym__quoted_i_heredoc_double,
    STATE(2135), 1,
      sym__quoted_i_parenthesis,
    STATE(2136), 1,
      sym__quoted_i_curly,
    STATE(2137), 1,
      sym__quoted_i_square,
    STATE(2138), 1,
      sym__quoted_i_angle,
    STATE(2139), 1,
      sym__quoted_i_bar,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [254004] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5887), 1,
      anon_sym_LPAREN,
    ACTIONS(5889), 1,
      anon_sym_DQUOTE,
    ACTIONS(5891), 1,
      anon_sym_SQUOTE,
    ACTIONS(5893), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5895), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5897), 1,
      anon_sym_LBRACE,
    ACTIONS(5899), 1,
      anon_sym_LBRACK,
    ACTIONS(5901), 1,
      anon_sym_LT,
    ACTIONS(5903), 1,
      anon_sym_PIPE,
    ACTIONS(5905), 1,
      anon_sym_SLASH,
    STATE(1930), 1,
      sym__quoted_double,
    STATE(1932), 1,
      sym__quoted_single,
    STATE(1933), 1,
      sym__quoted_heredoc_single,
    STATE(1934), 1,
      sym__quoted_heredoc_double,
    STATE(1935), 1,
      sym__quoted_parenthesis,
    STATE(2016), 1,
      sym__quoted_curly,
    STATE(2025), 1,
      sym__quoted_square,
    STATE(2026), 1,
      sym__quoted_angle,
    STATE(2039), 1,
      sym__quoted_bar,
    STATE(2040), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [254073] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5907), 1,
      anon_sym_LPAREN,
    ACTIONS(5909), 1,
      anon_sym_DQUOTE,
    ACTIONS(5911), 1,
      anon_sym_SQUOTE,
    ACTIONS(5913), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5915), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5917), 1,
      anon_sym_LBRACE,
    ACTIONS(5919), 1,
      anon_sym_LBRACK,
    ACTIONS(5921), 1,
      anon_sym_LT,
    ACTIONS(5923), 1,
      anon_sym_PIPE,
    ACTIONS(5925), 1,
      anon_sym_SLASH,
    STATE(1312), 1,
      sym__quoted_i_double,
    STATE(1313), 1,
      sym__quoted_i_single,
    STATE(1314), 1,
      sym__quoted_i_heredoc_single,
    STATE(1315), 1,
      sym__quoted_i_heredoc_double,
    STATE(1316), 1,
      sym__quoted_i_parenthesis,
    STATE(1317), 1,
      sym__quoted_i_curly,
    STATE(1318), 1,
      sym__quoted_i_square,
    STATE(1319), 1,
      sym__quoted_i_angle,
    STATE(1322), 1,
      sym__quoted_i_bar,
    STATE(1324), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [254142] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5927), 1,
      anon_sym_LPAREN,
    ACTIONS(5929), 1,
      anon_sym_DQUOTE,
    ACTIONS(5931), 1,
      anon_sym_SQUOTE,
    ACTIONS(5933), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5935), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5937), 1,
      anon_sym_LBRACE,
    ACTIONS(5939), 1,
      anon_sym_LBRACK,
    ACTIONS(5941), 1,
      anon_sym_LT,
    ACTIONS(5943), 1,
      anon_sym_PIPE,
    ACTIONS(5945), 1,
      anon_sym_SLASH,
    STATE(1325), 1,
      sym__quoted_double,
    STATE(1327), 1,
      sym__quoted_single,
    STATE(1329), 1,
      sym__quoted_heredoc_single,
    STATE(1330), 1,
      sym__quoted_heredoc_double,
    STATE(1331), 1,
      sym__quoted_parenthesis,
    STATE(1332), 1,
      sym__quoted_curly,
    STATE(1334), 1,
      sym__quoted_square,
    STATE(1335), 1,
      sym__quoted_angle,
    STATE(1336), 1,
      sym__quoted_bar,
    STATE(1337), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [254211] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5947), 1,
      anon_sym_LPAREN,
    ACTIONS(5949), 1,
      anon_sym_DQUOTE,
    ACTIONS(5951), 1,
      anon_sym_SQUOTE,
    ACTIONS(5953), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5955), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5957), 1,
      anon_sym_LBRACE,
    ACTIONS(5959), 1,
      anon_sym_LBRACK,
    ACTIONS(5961), 1,
      anon_sym_LT,
    ACTIONS(5963), 1,
      anon_sym_PIPE,
    ACTIONS(5965), 1,
      anon_sym_SLASH,
    STATE(1417), 1,
      sym__quoted_i_double,
    STATE(1418), 1,
      sym__quoted_i_single,
    STATE(1421), 1,
      sym__quoted_i_heredoc_single,
    STATE(1422), 1,
      sym__quoted_i_heredoc_double,
    STATE(1434), 1,
      sym__quoted_i_parenthesis,
    STATE(1436), 1,
      sym__quoted_i_curly,
    STATE(1447), 1,
      sym__quoted_i_square,
    STATE(1450), 1,
      sym__quoted_i_angle,
    STATE(1452), 1,
      sym__quoted_i_bar,
    STATE(1475), 1,
      sym__quoted_i_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [254280] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5967), 1,
      anon_sym_LPAREN,
    ACTIONS(5969), 1,
      anon_sym_DQUOTE,
    ACTIONS(5971), 1,
      anon_sym_SQUOTE,
    ACTIONS(5973), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5975), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5977), 1,
      anon_sym_LBRACE,
    ACTIONS(5979), 1,
      anon_sym_LBRACK,
    ACTIONS(5981), 1,
      anon_sym_LT,
    ACTIONS(5983), 1,
      anon_sym_PIPE,
    ACTIONS(5985), 1,
      anon_sym_SLASH,
    STATE(1478), 1,
      sym__quoted_double,
    STATE(1485), 1,
      sym__quoted_single,
    STATE(1486), 1,
      sym__quoted_heredoc_single,
    STATE(1487), 1,
      sym__quoted_heredoc_double,
    STATE(1488), 1,
      sym__quoted_parenthesis,
    STATE(1489), 1,
      sym__quoted_curly,
    STATE(1490), 1,
      sym__quoted_square,
    STATE(1491), 1,
      sym__quoted_angle,
    STATE(1492), 1,
      sym__quoted_bar,
    STATE(1494), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [254349] = 22,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5987), 1,
      anon_sym_LPAREN,
    ACTIONS(5989), 1,
      anon_sym_DQUOTE,
    ACTIONS(5991), 1,
      anon_sym_SQUOTE,
    ACTIONS(5993), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(5995), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(5997), 1,
      anon_sym_LBRACE,
    ACTIONS(5999), 1,
      anon_sym_LBRACK,
    ACTIONS(6001), 1,
      anon_sym_LT,
    ACTIONS(6003), 1,
      anon_sym_PIPE,
    ACTIONS(6005), 1,
      anon_sym_SLASH,
    STATE(3546), 1,
      sym__quoted_double,
    STATE(3550), 1,
      sym__quoted_single,
    STATE(3590), 1,
      sym__quoted_heredoc_single,
    STATE(3715), 1,
      sym__quoted_heredoc_double,
    STATE(3790), 1,
      sym__quoted_parenthesis,
    STATE(3806), 1,
      sym__quoted_curly,
    STATE(3817), 1,
      sym__quoted_square,
    STATE(3818), 1,
      sym__quoted_angle,
    STATE(3819), 1,
      sym__quoted_bar,
    STATE(3886), 1,
      sym__quoted_slash,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [254418] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1068), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4638), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4746), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254463] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(327), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4635), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4731), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254508] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1010), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4702), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4708), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254553] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1010), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3005), 1,
      anon_sym_SEMI,
    STATE(142), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4708), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254598] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1010), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4708), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254643] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1012), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6009), 1,
      anon_sym_SEMI,
    STATE(143), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4714), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254688] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1070), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6011), 1,
      anon_sym_SEMI,
    STATE(172), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4740), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254733] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1070), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4740), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254778] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1040), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4649), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4730), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254823] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1040), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3019), 1,
      anon_sym_SEMI,
    STATE(157), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4730), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254868] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1040), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4730), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254913] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(323), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4646), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4715), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [254958] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1042), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6013), 1,
      anon_sym_SEMI,
    STATE(158), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4739), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255003] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1022), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4648), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4720), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255048] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1022), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3017), 1,
      anon_sym_SEMI,
    STATE(148), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4720), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255093] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1022), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4720), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255138] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1024), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6015), 1,
      anon_sym_SEMI,
    STATE(149), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4724), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255183] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1024), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4724), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255228] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1042), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4739), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255273] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(339), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4653), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4751), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255318] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1028), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4655), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4755), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255363] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1028), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3023), 1,
      anon_sym_SEMI,
    STATE(151), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4755), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255408] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1028), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4755), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255453] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1030), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6017), 1,
      anon_sym_SEMI,
    STATE(152), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4759), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255498] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1030), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4759), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255543] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(331), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4659), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4707), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255588] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1034), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4661), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4709), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255633] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1034), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2985), 1,
      anon_sym_SEMI,
    STATE(154), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4709), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255678] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1034), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4709), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255723] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1036), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6019), 1,
      anon_sym_SEMI,
    STATE(175), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4711), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255768] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1036), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4711), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255813] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1068), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2930), 1,
      anon_sym_SEMI,
    STATE(171), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4746), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255858] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1068), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4746), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255903] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(335), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4673), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4725), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255948] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(343), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4668), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4719), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [255993] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1046), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4670), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4721), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256038] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1046), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2987), 1,
      anon_sym_SEMI,
    STATE(160), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4721), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256083] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1046), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4721), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256128] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1048), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6021), 1,
      anon_sym_SEMI,
    STATE(161), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4722), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256173] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1048), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4722), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256218] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1056), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4682), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4735), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256263] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1056), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2983), 1,
      anon_sym_SEMI,
    STATE(165), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4735), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256308] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1056), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4735), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256353] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1058), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6023), 1,
      anon_sym_SEMI,
    STATE(166), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4736), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256398] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(347), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4679), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4729), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256443] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(315), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4695), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4749), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256488] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1052), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4681), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4732), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256533] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1052), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2999), 1,
      anon_sym_SEMI,
    STATE(140), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4732), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256578] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1052), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4732), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256623] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1002), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6025), 1,
      anon_sym_SEMI,
    STATE(163), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4765), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256668] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1002), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4765), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256713] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1058), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4736), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256758] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(351), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4686), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4738), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256803] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1062), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4688), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4741), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256848] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1062), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3009), 1,
      anon_sym_SEMI,
    STATE(168), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4741), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256893] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1062), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4741), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256938] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1064), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6027), 1,
      anon_sym_SEMI,
    STATE(169), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4743), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [256983] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1064), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4743), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257028] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(307), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4693), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4750), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257073] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1016), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4701), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4762), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257118] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1074), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4697), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4754), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257163] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1074), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3015), 1,
      anon_sym_SEMI,
    STATE(174), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4754), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257208] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1074), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4754), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257253] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1016), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(2979), 1,
      anon_sym_SEMI,
    STATE(145), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4762), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257298] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1016), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4762), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257343] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1076), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6029), 1,
      anon_sym_SEMI,
    STATE(155), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4756), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257388] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1076), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4756), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257433] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(311), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4663), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4713), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257478] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(319), 1,
      anon_sym_end,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4641), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4752), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257523] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1018), 1,
      anon_sym_end,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6031), 1,
      anon_sym_SEMI,
    STATE(146), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4763), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257568] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1018), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4763), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257613] = 13,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1012), 1,
      anon_sym_end,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    STATE(4714), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257658] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1117), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(4840), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(792), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6033), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
  [257694] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1117), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(4840), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(792), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6039), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
  [257730] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(327), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257761] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(331), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257792] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1034), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257823] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1012), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257854] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1036), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257885] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(323), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257916] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1078), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257947] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6041), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [257978] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1068), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258009] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1014), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258040] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1022), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258071] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(343), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258102] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(335), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258133] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6043), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258164] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1046), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258195] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1024), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258226] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1048), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258257] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1050), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258288] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6045), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258319] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1026), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258350] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1056), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258381] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6047), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258412] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(347), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258443] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(315), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258474] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1052), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258505] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1042), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258536] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1010), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258567] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1002), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258598] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(311), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258629] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6049), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258660] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1058), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258691] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1060), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258722] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(351), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258753] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1062), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258784] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1044), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258815] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1072), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258846] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1064), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258877] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6051), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258908] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1066), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258939] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6053), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [258970] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6055), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259001] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1070), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259032] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(307), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259063] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(339), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259094] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1016), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259125] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1074), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259156] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1028), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259187] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1040), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259218] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6057), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259249] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1076), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259280] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1030), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259311] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1038), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259342] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6059), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259373] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6061), 1,
      anon_sym_after,
    ACTIONS(6064), 1,
      anon_sym_catch,
    ACTIONS(6067), 1,
      anon_sym_else,
    ACTIONS(6070), 1,
      anon_sym_end,
    ACTIONS(6072), 1,
      anon_sym_rescue,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259404] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1032), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259435] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(319), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259466] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6075), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259497] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1018), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259528] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1020), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259559] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(6077), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259590] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_after,
    ACTIONS(105), 1,
      anon_sym_catch,
    ACTIONS(107), 1,
      anon_sym_else,
    ACTIONS(113), 1,
      anon_sym_rescue,
    ACTIONS(1054), 1,
      anon_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    STATE(4758), 5,
      sym_after_block,
      sym_rescue_block,
      sym_catch_block,
      sym_else_block,
      aux_sym_do_block_repeat1,
  [259621] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3059), 1,
      anon_sym_SEMI,
    STATE(282), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1377), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259651] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6079), 1,
      anon_sym_SEMI,
    STATE(295), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1423), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259681] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1379), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259711] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1423), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259741] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6081), 1,
      aux_sym__terminator_token1,
    ACTIONS(6084), 1,
      anon_sym_SEMI,
    STATE(261), 1,
      sym__terminator,
    STATE(1021), 1,
      aux_sym__terminator_repeat1,
    STATE(4773), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1345), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259771] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6087), 1,
      anon_sym_SEMI,
    STATE(286), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1395), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259801] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1395), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259831] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6089), 1,
      aux_sym__terminator_token1,
    ACTIONS(6092), 1,
      anon_sym_SEMI,
    STATE(820), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(4773), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4351), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259861] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4769), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1379), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259891] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6095), 1,
      anon_sym_SEMI,
    STATE(289), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1397), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259921] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1377), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259951] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4793), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1377), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [259981] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1397), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260011] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4768), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(678), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260041] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(6097), 1,
      anon_sym_SEMI,
    STATE(290), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1421), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260071] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4772), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1561), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260101] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3105), 1,
      anon_sym_SEMI,
    STATE(283), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1379), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260131] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3137), 1,
      anon_sym_SEMI,
    STATE(276), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1561), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260161] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1561), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260191] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4778), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1563), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260221] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2928), 1,
      aux_sym__terminator_token1,
    ACTIONS(3113), 1,
      anon_sym_SEMI,
    STATE(277), 1,
      sym__terminator,
    STATE(1022), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1563), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260251] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3914), 1,
      aux_sym__terminator_token1,
    ACTIONS(3917), 1,
      anon_sym_SEMI,
    STATE(258), 1,
      sym__terminator,
    STATE(1021), 1,
      aux_sym__terminator_repeat1,
    STATE(4773), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1351), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260281] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1563), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260311] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4776), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(674), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260341] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1181), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    ACTIONS(6039), 1,
      anon_sym_GT_GT,
    STATE(4840), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(841), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260375] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6099), 1,
      aux_sym__terminator_token1,
    ACTIONS(6102), 1,
      anon_sym_SEMI,
    STATE(798), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(4791), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(4290), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260405] = 10,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1181), 1,
      sym_keyword,
    ACTIONS(6033), 1,
      anon_sym_GT_GT,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(4840), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(841), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260439] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(1421), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260469] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6105), 1,
      aux_sym__terminator_token1,
    ACTIONS(6108), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4794), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(6111), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260499] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4784), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(666), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260529] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(6007), 1,
      anon_sym_SEMI,
    STATE(138), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(4788), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(670), 5,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [260559] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6113), 1,
      anon_sym_AMP,
    ACTIONS(6117), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6115), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [260582] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6119), 1,
      anon_sym_AMP,
    ACTIONS(6123), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6121), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [260605] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6125), 1,
      anon_sym_AMP,
    ACTIONS(6129), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6127), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [260628] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(612), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(3015), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(868), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260659] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1181), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(4840), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(841), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260690] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6131), 1,
      anon_sym_AMP,
    ACTIONS(6135), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6133), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [260713] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1181), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(4506), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(841), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260744] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(481), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1338), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(845), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260775] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(87), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(4083), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(849), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260806] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6137), 1,
      anon_sym_AMP,
    ACTIONS(6141), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6139), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [260829] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(495), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1483), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(880), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260860] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1389), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(4191), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(721), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260891] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1557), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1966), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(892), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260922] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6143), 1,
      anon_sym_AMP,
    ACTIONS(6147), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6145), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [260945] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(564), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(3616), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(855), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [260976] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1505), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(3291), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(893), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261007] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6149), 1,
      anon_sym_AMP,
    ACTIONS(6153), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6151), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261030] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6155), 1,
      anon_sym_AMP,
    ACTIONS(6159), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6157), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261053] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6161), 1,
      anon_sym_AMP,
    ACTIONS(6165), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6163), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261076] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6167), 1,
      anon_sym_AMP,
    ACTIONS(6171), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6169), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261099] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6173), 1,
      anon_sym_AMP,
    ACTIONS(6177), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6175), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261122] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6179), 1,
      anon_sym_AMP,
    ACTIONS(6183), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6181), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261145] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1479), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1794), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(882), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261176] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1461), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1966), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(897), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261207] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6185), 1,
      anon_sym_AMP,
    ACTIONS(6189), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6187), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261230] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6191), 1,
      anon_sym_AMP,
    ACTIONS(6195), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6193), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261253] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6197), 1,
      anon_sym_AMP,
    ACTIONS(6201), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6199), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261276] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1117), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(3291), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(792), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261307] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(455), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(3211), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(843), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261338] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6203), 1,
      anon_sym_AMP,
    ACTIONS(6207), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6205), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261361] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(383), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(2165), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(848), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261392] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6209), 1,
      anon_sym_AMP,
    ACTIONS(6213), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6211), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261415] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1117), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(4840), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(792), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261446] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(262), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1338), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(891), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261477] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(648), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(2165), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(885), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261508] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(514), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1483), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(887), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261539] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1435), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1794), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(896), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261570] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(213), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1483), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(895), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261601] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(87), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(4840), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(849), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261632] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6215), 1,
      anon_sym_AMP,
    ACTIONS(6219), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6217), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261655] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6221), 1,
      anon_sym_AMP,
    ACTIONS(6225), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6223), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261678] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6227), 1,
      anon_sym_AMP,
    ACTIONS(6231), 1,
      anon_sym_AT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6229), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_TILDE_TILDE_TILDE,
      anon_sym_not,
  [261701] = 9,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1409), 1,
      sym_keyword,
    ACTIONS(6035), 1,
      anon_sym_DQUOTE,
    ACTIONS(6037), 1,
      anon_sym_SQUOTE,
    STATE(1966), 1,
      sym_pair,
    STATE(6869), 1,
      sym__quoted_i_single,
    STATE(6891), 1,
      sym__quoted_i_double,
    STATE(894), 2,
      sym__keyword,
      sym_quoted_keyword,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [261732] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_when,
      anon_sym_DASH_GT,
  [261750] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6233), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(6111), 7,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [261770] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6235), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(6237), 7,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [261790] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6239), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(6241), 7,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_after,
      anon_sym_catch,
      anon_sym_else,
      anon_sym_end,
      anon_sym_rescue,
  [261810] = 8,
    ACTIONS(6243), 1,
      anon_sym_DQUOTE,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6249), 1,
      sym__quoted_content_i_double,
    STATE(5875), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [261837] = 8,
    ACTIONS(6251), 1,
      anon_sym_SQUOTE,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6257), 1,
      sym__quoted_content_i_single,
    STATE(5444), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [261864] = 8,
    ACTIONS(6259), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6265), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5673), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [261891] = 8,
    ACTIONS(6267), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6273), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5786), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [261918] = 8,
    ACTIONS(6275), 1,
      anon_sym_RPAREN,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6281), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5381), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [261945] = 8,
    ACTIONS(6283), 1,
      anon_sym_RBRACE,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6289), 1,
      sym__quoted_content_i_curly,
    STATE(5383), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [261972] = 8,
    ACTIONS(6291), 1,
      anon_sym_RBRACK,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6297), 1,
      sym__quoted_content_i_square,
    STATE(5386), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [261999] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6299), 1,
      anon_sym_RPAREN,
    ACTIONS(6301), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5315), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262026] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6303), 1,
      anon_sym_RBRACE,
    ACTIONS(6305), 1,
      sym__quoted_content_i_curly,
    STATE(5323), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262053] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6307), 1,
      anon_sym_RBRACK,
    ACTIONS(6309), 1,
      sym__quoted_content_i_square,
    STATE(5325), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262080] = 8,
    ACTIONS(6311), 1,
      anon_sym_GT,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6317), 1,
      sym__quoted_content_i_angle,
    STATE(5327), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262107] = 8,
    ACTIONS(6319), 1,
      anon_sym_PIPE,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6325), 1,
      sym__quoted_content_i_bar,
    STATE(5329), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262134] = 8,
    ACTIONS(6327), 1,
      anon_sym_SLASH,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6333), 1,
      sym__quoted_content_i_slash,
    STATE(5331), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262161] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6335), 1,
      anon_sym_DQUOTE,
    ACTIONS(6337), 1,
      sym__quoted_content_i_double,
    STATE(5690), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262188] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6339), 1,
      anon_sym_SQUOTE,
    ACTIONS(6341), 1,
      sym__quoted_content_i_single,
    STATE(5704), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262215] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6343), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6345), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5706), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262242] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6347), 1,
      anon_sym_DQUOTE,
    ACTIONS(6349), 1,
      sym__quoted_content_i_double,
    STATE(5254), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262269] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6351), 1,
      anon_sym_SQUOTE,
    ACTIONS(6353), 1,
      sym__quoted_content_i_single,
    STATE(5256), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262296] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6355), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6357), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5258), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262323] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6359), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6361), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5260), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262350] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6363), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6365), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5720), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262377] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6367), 1,
      anon_sym_DQUOTE,
    ACTIONS(6369), 1,
      sym__quoted_content_i_double,
    STATE(5834), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262404] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6371), 1,
      anon_sym_SQUOTE,
    ACTIONS(6373), 1,
      sym__quoted_content_i_single,
    STATE(5905), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262431] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6375), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6377), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5923), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262458] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6379), 1,
      anon_sym_RPAREN,
    ACTIONS(6381), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5298), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262485] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6383), 1,
      anon_sym_RBRACE,
    ACTIONS(6385), 1,
      sym__quoted_content_i_curly,
    STATE(5300), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262512] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6387), 1,
      anon_sym_RBRACK,
    ACTIONS(6389), 1,
      sym__quoted_content_i_square,
    STATE(5302), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262539] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6391), 1,
      anon_sym_GT,
    ACTIONS(6393), 1,
      sym__quoted_content_i_angle,
    STATE(5304), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262566] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6395), 1,
      anon_sym_PIPE,
    ACTIONS(6397), 1,
      sym__quoted_content_i_bar,
    STATE(5306), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262593] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6399), 1,
      anon_sym_SLASH,
    ACTIONS(6401), 1,
      sym__quoted_content_i_slash,
    STATE(5308), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262620] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6403), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6405), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5925), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262647] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6407), 1,
      anon_sym_GT,
    ACTIONS(6409), 1,
      sym__quoted_content_i_angle,
    STATE(5395), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262674] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6411), 1,
      anon_sym_RPAREN,
    ACTIONS(6413), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5087), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262701] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6415), 1,
      anon_sym_RBRACE,
    ACTIONS(6417), 1,
      sym__quoted_content_i_curly,
    STATE(5099), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262728] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6419), 1,
      anon_sym_RBRACK,
    ACTIONS(6421), 1,
      sym__quoted_content_i_square,
    STATE(5103), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262755] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6423), 1,
      anon_sym_RPAREN,
    ACTIONS(6425), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5851), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262782] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6427), 1,
      anon_sym_RBRACE,
    ACTIONS(6429), 1,
      sym__quoted_content_i_curly,
    STATE(5865), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262809] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6431), 1,
      anon_sym_RBRACK,
    ACTIONS(6433), 1,
      sym__quoted_content_i_square,
    STATE(5873), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262836] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6435), 1,
      anon_sym_GT,
    ACTIONS(6437), 1,
      sym__quoted_content_i_angle,
    STATE(5881), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262863] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6439), 1,
      anon_sym_PIPE,
    ACTIONS(6441), 1,
      sym__quoted_content_i_bar,
    STATE(5889), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262890] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6443), 1,
      anon_sym_SLASH,
    ACTIONS(6445), 1,
      sym__quoted_content_i_slash,
    STATE(5897), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262917] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6447), 1,
      anon_sym_GT,
    ACTIONS(6449), 1,
      sym__quoted_content_i_angle,
    STATE(5109), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262944] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6451), 1,
      anon_sym_PIPE,
    ACTIONS(6453), 1,
      sym__quoted_content_i_bar,
    STATE(5116), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262971] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6455), 1,
      anon_sym_DQUOTE,
    ACTIONS(6457), 1,
      sym__quoted_content_i_double,
    STATE(5334), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [262998] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6459), 1,
      anon_sym_SQUOTE,
    ACTIONS(6461), 1,
      sym__quoted_content_i_single,
    STATE(5336), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263025] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6463), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6465), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5338), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263052] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6467), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6469), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5340), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263079] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6471), 1,
      anon_sym_SLASH,
    ACTIONS(6473), 1,
      sym__quoted_content_i_slash,
    STATE(5118), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263106] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6475), 1,
      anon_sym_PIPE,
    ACTIONS(6477), 1,
      sym__quoted_content_i_bar,
    STATE(5399), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263133] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6479), 1,
      anon_sym_SLASH,
    ACTIONS(6481), 1,
      sym__quoted_content_i_slash,
    STATE(5401), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263160] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6483), 1,
      anon_sym_RPAREN,
    ACTIONS(6485), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5368), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263187] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6487), 1,
      anon_sym_RBRACE,
    ACTIONS(6489), 1,
      sym__quoted_content_i_curly,
    STATE(5370), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263214] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6491), 1,
      anon_sym_RBRACK,
    ACTIONS(6493), 1,
      sym__quoted_content_i_square,
    STATE(5372), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263241] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6495), 1,
      anon_sym_GT,
    ACTIONS(6497), 1,
      sym__quoted_content_i_angle,
    STATE(5374), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263268] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6499), 1,
      anon_sym_PIPE,
    ACTIONS(6501), 1,
      sym__quoted_content_i_bar,
    STATE(5376), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263295] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6503), 1,
      anon_sym_SLASH,
    ACTIONS(6505), 1,
      sym__quoted_content_i_slash,
    STATE(5378), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263322] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6507), 1,
      anon_sym_RPAREN,
    ACTIONS(6509), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5506), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263349] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6511), 1,
      anon_sym_RBRACE,
    ACTIONS(6513), 1,
      sym__quoted_content_i_curly,
    STATE(5514), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263376] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6515), 1,
      anon_sym_RBRACK,
    ACTIONS(6517), 1,
      sym__quoted_content_i_square,
    STATE(5516), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263403] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6519), 1,
      anon_sym_GT,
    ACTIONS(6521), 1,
      sym__quoted_content_i_angle,
    STATE(5585), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263430] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6523), 1,
      anon_sym_PIPE,
    ACTIONS(6525), 1,
      sym__quoted_content_i_bar,
    STATE(5648), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263457] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6527), 1,
      anon_sym_SLASH,
    ACTIONS(6529), 1,
      sym__quoted_content_i_slash,
    STATE(5651), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263484] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6531), 1,
      anon_sym_RPAREN,
    ACTIONS(6533), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5447), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263511] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6535), 1,
      anon_sym_PIPE,
    ACTIONS(6537), 1,
      sym__quoted_content_i_bar,
    STATE(5736), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263538] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6539), 1,
      anon_sym_DQUOTE,
    ACTIONS(6541), 1,
      sym__quoted_content_i_double,
    STATE(5413), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263565] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6543), 1,
      anon_sym_SQUOTE,
    ACTIONS(6545), 1,
      sym__quoted_content_i_single,
    STATE(5415), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263592] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6547), 1,
      anon_sym_DQUOTE,
    ACTIONS(6549), 1,
      sym__quoted_content_i_double,
    STATE(5092), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263619] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6551), 1,
      anon_sym_SQUOTE,
    ACTIONS(6553), 1,
      sym__quoted_content_i_single,
    STATE(5094), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263646] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6555), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6557), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5096), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263673] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6559), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6561), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5098), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263700] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6563), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6565), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5421), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263727] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6567), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6569), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5423), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263754] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6571), 1,
      anon_sym_GT,
    ACTIONS(6573), 1,
      sym__quoted_content_i_angle,
    STATE(5670), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263781] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6575), 1,
      anon_sym_DQUOTE,
    ACTIONS(6577), 1,
      sym__quoted_content_i_double,
    STATE(5406), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263808] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6579), 1,
      anon_sym_SQUOTE,
    ACTIONS(6581), 1,
      sym__quoted_content_i_single,
    STATE(5408), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263835] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6583), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6585), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5410), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263862] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6587), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6589), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5412), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263889] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6591), 1,
      anon_sym_DQUOTE,
    ACTIONS(6593), 1,
      sym__quoted_content_i_double,
    STATE(5248), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263916] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6595), 1,
      anon_sym_SLASH,
    ACTIONS(6597), 1,
      sym__quoted_content_i_slash,
    STATE(5819), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263943] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6599), 1,
      anon_sym_RBRACE,
    ACTIONS(6601), 1,
      sym__quoted_content_i_curly,
    STATE(5535), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263970] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6603), 1,
      anon_sym_RPAREN,
    ACTIONS(6605), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5427), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [263997] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6607), 1,
      anon_sym_RBRACE,
    ACTIONS(6609), 1,
      sym__quoted_content_i_curly,
    STATE(5429), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264024] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6611), 1,
      anon_sym_RBRACK,
    ACTIONS(6613), 1,
      sym__quoted_content_i_square,
    STATE(5431), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264051] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6615), 1,
      anon_sym_GT,
    ACTIONS(6617), 1,
      sym__quoted_content_i_angle,
    STATE(5433), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264078] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6619), 1,
      anon_sym_PIPE,
    ACTIONS(6621), 1,
      sym__quoted_content_i_bar,
    STATE(5435), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264105] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6623), 1,
      anon_sym_SLASH,
    ACTIONS(6625), 1,
      sym__quoted_content_i_slash,
    STATE(5437), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264132] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6627), 1,
      anon_sym_RPAREN,
    ACTIONS(6629), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5135), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264159] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6631), 1,
      anon_sym_DQUOTE,
    ACTIONS(6633), 1,
      sym__quoted_content_i_double,
    STATE(5450), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264186] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6635), 1,
      anon_sym_SQUOTE,
    ACTIONS(6637), 1,
      sym__quoted_content_i_single,
    STATE(5452), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264213] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6639), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6641), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5454), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264240] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6643), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6645), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5456), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264267] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6647), 1,
      anon_sym_RBRACE,
    ACTIONS(6649), 1,
      sym__quoted_content_i_curly,
    STATE(5137), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264294] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6651), 1,
      anon_sym_RBRACK,
    ACTIONS(6653), 1,
      sym__quoted_content_i_square,
    STATE(5139), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264321] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6655), 1,
      anon_sym_GT,
    ACTIONS(6657), 1,
      sym__quoted_content_i_angle,
    STATE(5141), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264348] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6659), 1,
      anon_sym_RPAREN,
    ACTIONS(6661), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5472), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264375] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6663), 1,
      anon_sym_RBRACE,
    ACTIONS(6665), 1,
      sym__quoted_content_i_curly,
    STATE(5474), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264402] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6667), 1,
      anon_sym_RBRACK,
    ACTIONS(6669), 1,
      sym__quoted_content_i_square,
    STATE(5476), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264429] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6671), 1,
      anon_sym_GT,
    ACTIONS(6673), 1,
      sym__quoted_content_i_angle,
    STATE(5478), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264456] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6675), 1,
      anon_sym_PIPE,
    ACTIONS(6677), 1,
      sym__quoted_content_i_bar,
    STATE(5480), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264483] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6679), 1,
      anon_sym_SLASH,
    ACTIONS(6681), 1,
      sym__quoted_content_i_slash,
    STATE(5482), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264510] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6683), 1,
      anon_sym_PIPE,
    ACTIONS(6685), 1,
      sym__quoted_content_i_bar,
    STATE(5143), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264537] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6687), 1,
      anon_sym_DQUOTE,
    ACTIONS(6689), 1,
      sym__quoted_content_i_double,
    STATE(5496), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264564] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6691), 1,
      anon_sym_SQUOTE,
    ACTIONS(6693), 1,
      sym__quoted_content_i_single,
    STATE(5498), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264591] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6695), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6697), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5500), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264618] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6699), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6701), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5502), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264645] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6703), 1,
      anon_sym_SLASH,
    ACTIONS(6705), 1,
      sym__quoted_content_i_slash,
    STATE(5145), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264672] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6707), 1,
      anon_sym_DQUOTE,
    ACTIONS(6709), 1,
      sym__quoted_content_i_double,
    STATE(5170), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264699] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6711), 1,
      anon_sym_SQUOTE,
    ACTIONS(6713), 1,
      sym__quoted_content_i_single,
    STATE(5172), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264726] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6715), 1,
      anon_sym_RPAREN,
    ACTIONS(6717), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5518), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264753] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6719), 1,
      anon_sym_RBRACE,
    ACTIONS(6721), 1,
      sym__quoted_content_i_curly,
    STATE(5520), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264780] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6723), 1,
      anon_sym_RBRACK,
    ACTIONS(6725), 1,
      sym__quoted_content_i_square,
    STATE(5522), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264807] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6727), 1,
      anon_sym_GT,
    ACTIONS(6729), 1,
      sym__quoted_content_i_angle,
    STATE(5524), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264834] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6731), 1,
      anon_sym_PIPE,
    ACTIONS(6733), 1,
      sym__quoted_content_i_bar,
    STATE(5526), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264861] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6735), 1,
      anon_sym_SLASH,
    ACTIONS(6737), 1,
      sym__quoted_content_i_slash,
    STATE(5528), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264888] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6739), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6741), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5174), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264915] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6743), 1,
      anon_sym_DQUOTE,
    ACTIONS(6745), 1,
      sym__quoted_content_i_double,
    STATE(5542), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264942] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6747), 1,
      anon_sym_SQUOTE,
    ACTIONS(6749), 1,
      sym__quoted_content_i_single,
    STATE(5544), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264969] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6751), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6753), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5546), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [264996] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6755), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6757), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5548), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265023] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6759), 1,
      anon_sym_DQUOTE,
    ACTIONS(6761), 1,
      sym__quoted_content_i_double,
    STATE(5629), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265050] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6763), 1,
      anon_sym_SQUOTE,
    ACTIONS(6765), 1,
      sym__quoted_content_i_single,
    STATE(5639), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265077] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6767), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6769), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5641), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265104] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6771), 1,
      anon_sym_RPAREN,
    ACTIONS(6773), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5564), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265131] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6775), 1,
      anon_sym_RBRACE,
    ACTIONS(6777), 1,
      sym__quoted_content_i_curly,
    STATE(5566), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265158] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6779), 1,
      anon_sym_RBRACK,
    ACTIONS(6781), 1,
      sym__quoted_content_i_square,
    STATE(5568), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265185] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6783), 1,
      anon_sym_GT,
    ACTIONS(6785), 1,
      sym__quoted_content_i_angle,
    STATE(5570), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265212] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6787), 1,
      anon_sym_PIPE,
    ACTIONS(6789), 1,
      sym__quoted_content_i_bar,
    STATE(5572), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265239] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6791), 1,
      anon_sym_SLASH,
    ACTIONS(6793), 1,
      sym__quoted_content_i_slash,
    STATE(5574), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265266] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6795), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6797), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5647), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265293] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6799), 1,
      anon_sym_DQUOTE,
    ACTIONS(6801), 1,
      sym__quoted_content_i_double,
    STATE(5588), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265320] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6803), 1,
      anon_sym_SQUOTE,
    ACTIONS(6805), 1,
      sym__quoted_content_i_single,
    STATE(5590), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265347] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6807), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6809), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5592), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265374] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6811), 1,
      anon_sym_DQUOTE,
    ACTIONS(6813), 1,
      sym__quoted_content_i_double,
    STATE(5353), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265401] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6815), 1,
      anon_sym_DQUOTE,
    ACTIONS(6817), 1,
      sym__quoted_content_i_double,
    STATE(5200), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265428] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6819), 1,
      anon_sym_SQUOTE,
    ACTIONS(6821), 1,
      sym__quoted_content_i_single,
    STATE(5202), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265455] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6823), 1,
      anon_sym_RPAREN,
    ACTIONS(6825), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5267), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265482] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6827), 1,
      anon_sym_RPAREN,
    ACTIONS(6829), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5609), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265509] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6831), 1,
      anon_sym_RBRACE,
    ACTIONS(6833), 1,
      sym__quoted_content_i_curly,
    STATE(5611), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265536] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6835), 1,
      anon_sym_RBRACK,
    ACTIONS(6837), 1,
      sym__quoted_content_i_square,
    STATE(5613), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265563] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6839), 1,
      anon_sym_GT,
    ACTIONS(6841), 1,
      sym__quoted_content_i_angle,
    STATE(5615), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265590] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6843), 1,
      anon_sym_PIPE,
    ACTIONS(6845), 1,
      sym__quoted_content_i_bar,
    STATE(5617), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265617] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6847), 1,
      anon_sym_SLASH,
    ACTIONS(6849), 1,
      sym__quoted_content_i_slash,
    STATE(5619), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265644] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6851), 1,
      anon_sym_RBRACE,
    ACTIONS(6853), 1,
      sym__quoted_content_i_curly,
    STATE(5269), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265671] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6855), 1,
      anon_sym_DQUOTE,
    ACTIONS(6857), 1,
      sym__quoted_content_i_double,
    STATE(5632), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265698] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6859), 1,
      anon_sym_SQUOTE,
    ACTIONS(6861), 1,
      sym__quoted_content_i_single,
    STATE(5634), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265725] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6863), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6865), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5636), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265752] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6867), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6869), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5638), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265779] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6871), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6873), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5205), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265806] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6875), 1,
      anon_sym_RPAREN,
    ACTIONS(6877), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5539), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265833] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6879), 1,
      anon_sym_RBRACE,
    ACTIONS(6881), 1,
      sym__quoted_content_i_curly,
    STATE(5550), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265860] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(6883), 1,
      anon_sym_RPAREN,
    ACTIONS(6885), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5653), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265887] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(6887), 1,
      anon_sym_RBRACE,
    ACTIONS(6889), 1,
      sym__quoted_content_i_curly,
    STATE(5655), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265914] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6891), 1,
      anon_sym_RBRACK,
    ACTIONS(6893), 1,
      sym__quoted_content_i_square,
    STATE(5657), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265941] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(6895), 1,
      anon_sym_GT,
    ACTIONS(6897), 1,
      sym__quoted_content_i_angle,
    STATE(5659), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265968] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(6899), 1,
      anon_sym_PIPE,
    ACTIONS(6901), 1,
      sym__quoted_content_i_bar,
    STATE(5661), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [265995] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(6903), 1,
      anon_sym_SLASH,
    ACTIONS(6905), 1,
      sym__quoted_content_i_slash,
    STATE(5663), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266022] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6907), 1,
      anon_sym_RBRACK,
    ACTIONS(6909), 1,
      sym__quoted_content_i_square,
    STATE(5557), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266049] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6911), 1,
      anon_sym_DQUOTE,
    ACTIONS(6913), 1,
      sym__quoted_content_i_double,
    STATE(5676), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266076] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6915), 1,
      anon_sym_SQUOTE,
    ACTIONS(6917), 1,
      sym__quoted_content_i_single,
    STATE(5678), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266103] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6919), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6921), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5680), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266130] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6923), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6925), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5682), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266157] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(6927), 1,
      anon_sym_RBRACK,
    ACTIONS(6929), 1,
      sym__quoted_content_i_square,
    STATE(5285), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266184] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6931), 1,
      anon_sym_DQUOTE,
    ACTIONS(6933), 1,
      sym__quoted_content_i_double,
    STATE(5693), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266211] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6935), 1,
      anon_sym_SQUOTE,
    ACTIONS(6937), 1,
      sym__quoted_content_i_single,
    STATE(5695), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266238] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6939), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6941), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5697), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266265] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6943), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6945), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5699), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266292] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6947), 1,
      anon_sym_DQUOTE,
    ACTIONS(6949), 1,
      sym__quoted_content_i_double,
    STATE(5709), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266319] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6951), 1,
      anon_sym_SQUOTE,
    ACTIONS(6953), 1,
      sym__quoted_content_i_single,
    STATE(5711), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266346] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6955), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6957), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5713), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266373] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6959), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6961), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5715), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266400] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6963), 1,
      anon_sym_DQUOTE,
    ACTIONS(6965), 1,
      sym__quoted_content_i_double,
    STATE(5725), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266427] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6967), 1,
      anon_sym_SQUOTE,
    ACTIONS(6969), 1,
      sym__quoted_content_i_single,
    STATE(5727), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266454] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6971), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6973), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5729), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266481] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6975), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6977), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5731), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266508] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6979), 1,
      anon_sym_DQUOTE,
    ACTIONS(6981), 1,
      sym__quoted_content_i_double,
    STATE(5741), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266535] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6983), 1,
      anon_sym_SQUOTE,
    ACTIONS(6985), 1,
      sym__quoted_content_i_single,
    STATE(5743), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266562] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(6987), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(6989), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5745), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266589] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(6991), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(6993), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5747), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266616] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(6995), 1,
      anon_sym_DQUOTE,
    ACTIONS(6997), 1,
      sym__quoted_content_i_double,
    STATE(5757), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266643] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(6999), 1,
      anon_sym_SQUOTE,
    ACTIONS(7001), 1,
      sym__quoted_content_i_single,
    STATE(5759), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266670] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7003), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7005), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5761), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266697] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7007), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7009), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5763), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266724] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7011), 1,
      anon_sym_DQUOTE,
    ACTIONS(7013), 1,
      sym__quoted_content_i_double,
    STATE(5773), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266751] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7015), 1,
      anon_sym_SQUOTE,
    ACTIONS(7017), 1,
      sym__quoted_content_i_single,
    STATE(5775), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266778] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7019), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7021), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5777), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266805] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7023), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7025), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5779), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266832] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7027), 1,
      anon_sym_DQUOTE,
    ACTIONS(7029), 1,
      sym__quoted_content_i_double,
    STATE(5789), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266859] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7031), 1,
      anon_sym_SQUOTE,
    ACTIONS(7033), 1,
      sym__quoted_content_i_single,
    STATE(5791), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266886] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7035), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7037), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5793), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266913] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7039), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7041), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5795), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266940] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7043), 1,
      anon_sym_DQUOTE,
    ACTIONS(7045), 1,
      sym__quoted_content_i_double,
    STATE(5805), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266967] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7047), 1,
      anon_sym_SQUOTE,
    ACTIONS(7049), 1,
      sym__quoted_content_i_single,
    STATE(5807), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [266994] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7053), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5809), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267021] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7057), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5811), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267048] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7059), 1,
      anon_sym_DQUOTE,
    ACTIONS(7061), 1,
      sym__quoted_content_i_double,
    STATE(5821), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267075] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7063), 1,
      anon_sym_SQUOTE,
    ACTIONS(7065), 1,
      sym__quoted_content_i_single,
    STATE(5823), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267102] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7067), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7069), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5825), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267129] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7071), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7073), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5827), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267156] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7075), 1,
      anon_sym_DQUOTE,
    ACTIONS(7077), 1,
      sym__quoted_content_i_double,
    STATE(5837), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267183] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7079), 1,
      anon_sym_SQUOTE,
    ACTIONS(7081), 1,
      sym__quoted_content_i_single,
    STATE(5839), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267210] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7083), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7085), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5841), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267237] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7087), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7089), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5843), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267264] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7091), 1,
      anon_sym_DQUOTE,
    ACTIONS(7093), 1,
      sym__quoted_content_i_double,
    STATE(5853), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267291] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7095), 1,
      anon_sym_SQUOTE,
    ACTIONS(7097), 1,
      sym__quoted_content_i_single,
    STATE(5855), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267318] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7099), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7101), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5857), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267345] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7103), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7105), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5859), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267372] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7107), 1,
      anon_sym_DQUOTE,
    ACTIONS(7109), 1,
      sym__quoted_content_i_double,
    STATE(5867), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267399] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7111), 1,
      anon_sym_SQUOTE,
    ACTIONS(7113), 1,
      sym__quoted_content_i_single,
    STATE(5869), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267426] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7115), 1,
      anon_sym_GT,
    ACTIONS(7117), 1,
      sym__quoted_content_i_angle,
    STATE(5559), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267453] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7119), 1,
      anon_sym_SQUOTE,
    ACTIONS(7121), 1,
      sym__quoted_content_i_single,
    STATE(5877), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267480] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7123), 1,
      anon_sym_DQUOTE,
    ACTIONS(7125), 1,
      sym__quoted_content_i_double,
    STATE(5883), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267507] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7127), 1,
      anon_sym_SQUOTE,
    ACTIONS(7129), 1,
      sym__quoted_content_i_single,
    STATE(5885), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267534] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7131), 1,
      anon_sym_DQUOTE,
    ACTIONS(7133), 1,
      sym__quoted_content_i_double,
    STATE(5891), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267561] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7135), 1,
      anon_sym_SQUOTE,
    ACTIONS(7137), 1,
      sym__quoted_content_i_single,
    STATE(5893), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267588] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7139), 1,
      anon_sym_DQUOTE,
    ACTIONS(7141), 1,
      sym__quoted_content_i_double,
    STATE(5899), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267615] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7143), 1,
      anon_sym_SQUOTE,
    ACTIONS(7145), 1,
      sym__quoted_content_i_single,
    STATE(5901), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267642] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7147), 1,
      anon_sym_DQUOTE,
    ACTIONS(7149), 1,
      sym__quoted_content_i_double,
    STATE(5907), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267669] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7151), 1,
      anon_sym_SQUOTE,
    ACTIONS(7153), 1,
      sym__quoted_content_i_single,
    STATE(5909), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267696] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7155), 1,
      anon_sym_DQUOTE,
    ACTIONS(7157), 1,
      sym__quoted_content_i_double,
    STATE(5915), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267723] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7159), 1,
      anon_sym_SQUOTE,
    ACTIONS(7161), 1,
      sym__quoted_content_i_single,
    STATE(5917), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267750] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7163), 1,
      anon_sym_GT,
    ACTIONS(7165), 1,
      sym__quoted_content_i_angle,
    STATE(5290), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267777] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7167), 1,
      anon_sym_PIPE,
    ACTIONS(7169), 1,
      sym__quoted_content_i_bar,
    STATE(5561), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267804] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7171), 1,
      anon_sym_PIPE,
    ACTIONS(7173), 1,
      sym__quoted_content_i_bar,
    STATE(5294), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267831] = 8,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7175), 1,
      anon_sym_DQUOTE,
    ACTIONS(7177), 1,
      sym__quoted_content_i_double,
    STATE(5178), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267858] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7179), 1,
      anon_sym_SQUOTE,
    ACTIONS(7181), 1,
      sym__quoted_content_i_single,
    STATE(5180), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267885] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7183), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7185), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5182), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267912] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7187), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7189), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5184), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267939] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7191), 1,
      anon_sym_SLASH,
    ACTIONS(7193), 1,
      sym__quoted_content_i_slash,
    STATE(5313), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267966] = 8,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7195), 1,
      anon_sym_SQUOTE,
    ACTIONS(7197), 1,
      sym__quoted_content_i_single,
    STATE(5938), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [267993] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7199), 1,
      anon_sym_SLASH,
    ACTIONS(7201), 1,
      sym__quoted_content_i_slash,
    STATE(5581), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268020] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7203), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7205), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5209), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268047] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7207), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7209), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5185), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268074] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7211), 1,
      anon_sym_RBRACK,
    ACTIONS(7213), 1,
      sym__quoted_content_i_square,
    STATE(5603), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268101] = 8,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7215), 1,
      anon_sym_RPAREN,
    ACTIONS(7217), 1,
      sym__quoted_content_i_parenthesis,
    STATE(5216), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268128] = 8,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7219), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7221), 1,
      sym__quoted_content_i_heredoc_single,
    STATE(5944), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268155] = 8,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7223), 1,
      anon_sym_RBRACE,
    ACTIONS(7225), 1,
      sym__quoted_content_i_curly,
    STATE(5218), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268182] = 8,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7227), 1,
      anon_sym_RBRACK,
    ACTIONS(7229), 1,
      sym__quoted_content_i_square,
    STATE(5220), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268209] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7231), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7233), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5952), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268236] = 8,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7235), 1,
      anon_sym_GT,
    ACTIONS(7237), 1,
      sym__quoted_content_i_angle,
    STATE(5222), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268263] = 8,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7239), 1,
      anon_sym_PIPE,
    ACTIONS(7241), 1,
      sym__quoted_content_i_bar,
    STATE(5224), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268290] = 8,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7243), 1,
      anon_sym_SLASH,
    ACTIONS(7245), 1,
      sym__quoted_content_i_slash,
    STATE(5226), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268317] = 8,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7247), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(7249), 1,
      sym__quoted_content_i_heredoc_double,
    STATE(5594), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268344] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7251), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268368] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7253), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268392] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7255), 1,
      anon_sym_RBRACE,
    STATE(5151), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268416] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(912), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5204), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [268442] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(906), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5110), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [268468] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7261), 1,
      anon_sym_DQUOTE,
    STATE(5111), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268492] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7261), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268516] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7263), 1,
      anon_sym_SQUOTE,
    STATE(5113), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268540] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7263), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268564] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7265), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5114), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268588] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7265), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268612] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7267), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5115), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268636] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7267), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268660] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7255), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268684] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7269), 1,
      anon_sym_RBRACK,
    STATE(5152), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268708] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7271), 1,
      anon_sym_DQUOTE,
    STATE(5953), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268732] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(910), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5125), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [268758] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7269), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268782] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(952), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5928), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [268808] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1647), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5251), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [268834] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1651), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5133), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [268860] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7273), 1,
      anon_sym_GT,
    STATE(5154), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268884] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1651), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(4013), 1,
      anon_sym_SEMI,
    STATE(380), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [268910] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7273), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268934] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1651), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [268960] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7275), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [268984] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7277), 1,
      anon_sym_PIPE,
    STATE(5161), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269008] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7279), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269032] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7281), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269056] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7283), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269080] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7277), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269104] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7285), 1,
      anon_sym_SLASH,
    STATE(5162), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269128] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7285), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269152] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7287), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269178] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4351), 1,
      anon_sym_RPAREN,
    ACTIONS(6089), 1,
      aux_sym__terminator_token1,
    ACTIONS(7289), 1,
      anon_sym_SEMI,
    STATE(847), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5120), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269204] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7292), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269228] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1647), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3937), 1,
      anon_sym_SEMI,
    STATE(381), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269254] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1641), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7294), 1,
      anon_sym_SEMI,
    STATE(433), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269280] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7296), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5146), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269306] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7296), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269332] = 7,
    ACTIONS(7298), 1,
      anon_sym_RBRACK,
    ACTIONS(7300), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(7303), 1,
      sym_escape_sequence,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269356] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1647), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269382] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1761), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5503), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269408] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1761), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3955), 1,
      anon_sym_SEMI,
    STATE(421), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269434] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7306), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269458] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4349), 1,
      ts_builtin_sym_end,
    ACTIONS(6089), 1,
      aux_sym__terminator_token1,
    ACTIONS(7308), 1,
      anon_sym_SEMI,
    STATE(461), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5131), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269484] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1659), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7311), 1,
      anon_sym_SEMI,
    STATE(383), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269510] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1659), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269536] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7313), 1,
      anon_sym_RPAREN,
    STATE(5155), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269560] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7313), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269584] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7315), 1,
      anon_sym_RBRACE,
    STATE(5156), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269608] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7315), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269632] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7317), 1,
      anon_sym_RBRACK,
    STATE(5157), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269656] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7317), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269680] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7319), 1,
      anon_sym_GT,
    STATE(5158), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269704] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7319), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269728] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7321), 1,
      anon_sym_PIPE,
    STATE(5159), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269752] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7321), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269776] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7323), 1,
      anon_sym_SLASH,
    STATE(5160), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269800] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7323), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269824] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7325), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269850] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7327), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269874] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1641), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [269900] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7329), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269924] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7331), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269948] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7333), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269972] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7335), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [269996] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7337), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270020] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7339), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270044] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7341), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270068] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7343), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270092] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7345), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270116] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7347), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270140] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7349), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270164] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7351), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270188] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7353), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270212] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7355), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270236] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1761), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [270262] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7357), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270286] = 7,
    ACTIONS(7359), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(7361), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(7364), 1,
      sym_escape_sequence,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270310] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7367), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270334] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7369), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270358] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(964), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5240), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [270384] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7371), 1,
      anon_sym_DQUOTE,
    STATE(5244), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270408] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7371), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270432] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7373), 1,
      anon_sym_SQUOTE,
    STATE(5246), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270456] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7373), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270480] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7375), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5247), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270504] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7375), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270528] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7377), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5261), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270552] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(920), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5192), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [270578] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7379), 1,
      anon_sym_DQUOTE,
    STATE(5193), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270602] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7379), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270626] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7381), 1,
      anon_sym_SQUOTE,
    STATE(5195), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270650] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7381), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270674] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7383), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5196), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270698] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7383), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270722] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7385), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5197), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270746] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7385), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270770] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7377), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270794] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7387), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270818] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(932), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5207), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [270844] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1709), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5214), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [270870] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7389), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5361), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [270896] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1709), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(4021), 1,
      anon_sym_SEMI,
    STATE(354), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [270922] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(7391), 1,
      anon_sym_COMMA,
    STATE(5263), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(6039), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
  [270942] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1709), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [270968] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7393), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [270992] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(900), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5238), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271018] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7395), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271042] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7397), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271066] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7399), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271090] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7401), 1,
      anon_sym_DQUOTE,
    STATE(5239), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271114] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4290), 1,
      anon_sym_RPAREN,
    ACTIONS(6099), 1,
      aux_sym__terminator_token1,
    ACTIONS(7403), 1,
      anon_sym_SEMI,
    STATE(526), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271140] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7401), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271164] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7406), 1,
      anon_sym_SQUOTE,
    STATE(5241), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271188] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7406), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271212] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7408), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5242), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271236] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7389), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271262] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7408), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271286] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7410), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5227), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271312] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7410), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271338] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7412), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5243), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271362] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7412), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271386] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(898), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5293), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271412] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1617), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5364), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271438] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(904), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5280), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271464] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1575), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7414), 1,
      anon_sym_SEMI,
    STATE(405), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271490] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1575), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271516] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7416), 1,
      anon_sym_RPAREN,
    STATE(5232), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271540] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7416), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271564] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7418), 1,
      anon_sym_RBRACE,
    STATE(5233), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271588] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7418), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271612] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7420), 1,
      anon_sym_RBRACK,
    STATE(5234), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271636] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7420), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271660] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7422), 1,
      anon_sym_GT,
    STATE(5235), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271684] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7422), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271708] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7424), 1,
      anon_sym_PIPE,
    STATE(5236), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271732] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7424), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271756] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7426), 1,
      anon_sym_SLASH,
    STATE(5237), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271780] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7426), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271804] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7428), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271830] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1635), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5312), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271856] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1617), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3963), 1,
      anon_sym_SEMI,
    STATE(402), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271882] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1635), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3983), 1,
      anon_sym_SEMI,
    STATE(366), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [271908] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7430), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271932] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7432), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271956] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7434), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [271980] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7436), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272004] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7438), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272028] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7440), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272052] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7442), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272076] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1635), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272102] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7444), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272126] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1617), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272152] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7446), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272176] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7448), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272200] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7450), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272224] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7452), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272248] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1345), 1,
      anon_sym_RPAREN,
    ACTIONS(7454), 1,
      aux_sym__terminator_token1,
    ACTIONS(7457), 1,
      anon_sym_SEMI,
    STATE(353), 1,
      sym__terminator,
    STATE(1036), 1,
      aux_sym__terminator_repeat1,
    STATE(5120), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272274] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7460), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272298] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7462), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272322] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7271), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272346] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1351), 1,
      anon_sym_end,
    ACTIONS(4809), 1,
      aux_sym__terminator_token1,
    ACTIONS(4812), 1,
      anon_sym_SEMI,
    STATE(350), 1,
      sym__terminator,
    STATE(1032), 1,
      aux_sym__terminator_repeat1,
    STATE(5929), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272372] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1661), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7464), 1,
      anon_sym_SEMI,
    STATE(386), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272398] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1661), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272424] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(946), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5274), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272450] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7466), 1,
      anon_sym_DQUOTE,
    STATE(5275), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272474] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7466), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272498] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7468), 1,
      anon_sym_SQUOTE,
    STATE(5276), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272522] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7468), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272546] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7470), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5277), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272570] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7470), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272594] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7472), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5278), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272618] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7472), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272642] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7474), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272666] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7476), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272690] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(7478), 1,
      anon_sym_COMMA,
    STATE(5263), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
  [272710] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(7481), 1,
      anon_sym_COMMA,
    STATE(5264), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3619), 3,
      anon_sym_RPAREN,
      anon_sym_when,
      anon_sym_DASH_GT,
  [272730] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(896), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5288), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272756] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7484), 1,
      anon_sym_RPAREN,
    STATE(5459), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272780] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7484), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272804] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7486), 1,
      anon_sym_RBRACE,
    STATE(5494), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272828] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7486), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272852] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1771), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5296), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272878] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7488), 1,
      anon_sym_RBRACK,
    STATE(5505), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272902] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1771), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(4051), 1,
      anon_sym_SEMI,
    STATE(422), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272928] = 7,
    ACTIONS(7490), 1,
      anon_sym_RPAREN,
    ACTIONS(7492), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(7495), 1,
      sym_escape_sequence,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [272952] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1771), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [272978] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7498), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273002] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7500), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273026] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7502), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273050] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7504), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273074] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7506), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5341), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273100] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7506), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273126] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(914), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5163), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273152] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6105), 1,
      aux_sym__terminator_token1,
    ACTIONS(6111), 1,
      anon_sym_RPAREN,
    ACTIONS(7508), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273178] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7511), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273202] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7513), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5671), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273228] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7488), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273252] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7515), 1,
      anon_sym_GT,
    STATE(5540), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273276] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7517), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5309), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273302] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7517), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273328] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7513), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273354] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7515), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273378] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7519), 1,
      anon_sym_PIPE,
    STATE(5551), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273402] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7521), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5402), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273428] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7521), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273454] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7519), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273478] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1777), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7523), 1,
      anon_sym_SEMI,
    STATE(424), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273504] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1777), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273530] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7525), 1,
      anon_sym_RPAREN,
    STATE(5316), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273554] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7525), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273578] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7527), 1,
      anon_sym_RBRACE,
    STATE(5317), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273602] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7527), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273626] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7529), 1,
      anon_sym_RBRACK,
    STATE(5318), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273650] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7529), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273674] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7531), 1,
      anon_sym_GT,
    STATE(5319), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273698] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7531), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273722] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7533), 1,
      anon_sym_PIPE,
    STATE(5320), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273746] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7533), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273770] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7535), 1,
      anon_sym_SLASH,
    STATE(5321), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273794] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7535), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273818] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7537), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273844] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7539), 1,
      anon_sym_SLASH,
    STATE(5586), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273868] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1625), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7541), 1,
      anon_sym_SEMI,
    STATE(375), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273894] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1625), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [273920] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7539), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273944] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7543), 1,
      anon_sym_RPAREN,
    STATE(5354), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273968] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7543), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [273992] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7545), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274016] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7547), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274040] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7549), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274064] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7551), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274088] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7553), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274112] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7555), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274136] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7557), 1,
      anon_sym_RBRACE,
    STATE(5355), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274160] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7557), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274184] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7559), 1,
      anon_sym_RBRACK,
    STATE(5356), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274208] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7559), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274232] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7561), 1,
      anon_sym_GT,
    STATE(5357), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274256] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7561), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274280] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7563), 1,
      anon_sym_PIPE,
    STATE(5360), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274304] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7563), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274328] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7565), 1,
      anon_sym_SLASH,
    STATE(5362), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274352] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7565), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274376] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(960), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5347), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [274402] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7567), 1,
      anon_sym_DQUOTE,
    STATE(5348), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274426] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7567), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274450] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7569), 1,
      anon_sym_SQUOTE,
    STATE(5349), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274474] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7569), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274498] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7571), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5350), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274522] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7571), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274546] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7573), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5351), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274570] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7573), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274594] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7575), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [274620] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7577), 1,
      anon_sym_DQUOTE,
    STATE(5164), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274644] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(950), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5359), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [274670] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1811), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5366), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [274696] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(4069), 1,
      anon_sym_SEMI,
    STATE(440), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [274722] = 7,
    ACTIONS(7579), 1,
      anon_sym_GT,
    ACTIONS(7581), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(7584), 1,
      sym_escape_sequence,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274746] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1811), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [274772] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7587), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274796] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7589), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274820] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7591), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274844] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7593), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274868] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9), 1,
      aux_sym__terminator_token1,
    ACTIONS(1819), 1,
      ts_builtin_sym_end,
    ACTIONS(4690), 1,
      anon_sym_SEMI,
    STATE(403), 1,
      sym__terminator,
    STATE(1035), 1,
      aux_sym__terminator_repeat1,
    STATE(5131), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [274894] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7577), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274918] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7595), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274942] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7597), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274966] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7599), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [274990] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7601), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275014] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7603), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5379), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [275040] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7603), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [275066] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7605), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275090] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7607), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [275116] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7609), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275140] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1725), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7611), 1,
      anon_sym_SEMI,
    STATE(378), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [275166] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1725), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [275192] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7613), 1,
      anon_sym_SEMI,
    STATE(442), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [275218] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1815), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [275244] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7615), 1,
      anon_sym_RPAREN,
    STATE(5387), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275268] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7615), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275292] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7617), 1,
      anon_sym_RBRACE,
    STATE(5388), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275316] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7617), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275340] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7619), 1,
      anon_sym_RBRACK,
    STATE(5389), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275364] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7619), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275388] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7621), 1,
      anon_sym_GT,
    STATE(5390), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275412] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7621), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275436] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7623), 1,
      anon_sym_PIPE,
    STATE(5391), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275460] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7623), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275484] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7625), 1,
      anon_sym_SLASH,
    STATE(5392), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275508] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7625), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275532] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7627), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [275558] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7629), 1,
      anon_sym_RPAREN,
    STATE(5448), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275582] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7629), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275606] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7631), 1,
      anon_sym_RBRACE,
    STATE(5457), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275630] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7631), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275654] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(7633), 1,
      anon_sym_COMMA,
    STATE(5384), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3645), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
  [275674] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7636), 1,
      anon_sym_RBRACK,
    STATE(5466), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275698] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7636), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275722] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7638), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275746] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7640), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275770] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7642), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275794] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7644), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275818] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7646), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275842] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7648), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275866] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7650), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5950), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [275892] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7652), 1,
      anon_sym_GT,
    STATE(5469), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275916] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7652), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275940] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7654), 1,
      anon_sym_PIPE,
    STATE(5491), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275964] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7656), 1,
      anon_sym_RPAREN,
    STATE(5262), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [275988] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1345), 1,
      anon_sym_end,
    ACTIONS(7658), 1,
      aux_sym__terminator_token1,
    ACTIONS(7661), 1,
      anon_sym_SEMI,
    STATE(346), 1,
      sym__terminator,
    STATE(1032), 1,
      aux_sym__terminator_repeat1,
    STATE(5929), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [276014] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7654), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276038] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7664), 1,
      anon_sym_SLASH,
    STATE(5492), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276062] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7664), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276086] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7666), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [276112] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(928), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5467), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [276138] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7668), 1,
      anon_sym_DQUOTE,
    STATE(5468), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276162] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7670), 1,
      anon_sym_DQUOTE,
    STATE(5416), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276186] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7670), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276210] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7672), 1,
      anon_sym_SQUOTE,
    STATE(5417), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276234] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7672), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276258] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7674), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5418), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276282] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7674), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276306] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7676), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5086), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276330] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7676), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276354] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7668), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276378] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7678), 1,
      anon_sym_SQUOTE,
    STATE(5470), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276402] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7678), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276426] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7680), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276450] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7682), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276474] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7684), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276498] = 7,
    ACTIONS(7686), 1,
      anon_sym_SQUOTE,
    ACTIONS(7688), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(7691), 1,
      sym_escape_sequence,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276522] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7694), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5489), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276546] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7694), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276570] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7696), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5490), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276594] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7696), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276618] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7698), 1,
      anon_sym_SQUOTE,
    STATE(5166), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276642] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7700), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276666] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7702), 1,
      anon_sym_RPAREN,
    STATE(5438), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276690] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7702), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276714] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7704), 1,
      anon_sym_RBRACE,
    STATE(5439), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276738] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7704), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276762] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7706), 1,
      anon_sym_RBRACK,
    STATE(5440), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276786] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7706), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276810] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7708), 1,
      anon_sym_GT,
    STATE(5441), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276834] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7708), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276858] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7710), 1,
      anon_sym_PIPE,
    STATE(5442), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276882] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7710), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276906] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7712), 1,
      anon_sym_SLASH,
    STATE(5443), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276930] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7712), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276954] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7714), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [276978] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7716), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277002] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7718), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277026] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7720), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277050] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7722), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277074] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7724), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277098] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7698), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277122] = 7,
    ACTIONS(7726), 1,
      anon_sym_PIPE,
    ACTIONS(7728), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(7731), 1,
      sym_escape_sequence,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277146] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(934), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5513), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [277172] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7656), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277196] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7734), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277220] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7736), 1,
      anon_sym_DQUOTE,
    STATE(5461), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277244] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7736), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277268] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7738), 1,
      anon_sym_SQUOTE,
    STATE(5462), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277292] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7738), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277316] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7740), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5463), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277340] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7740), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277364] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7742), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5464), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277388] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7742), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277412] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7744), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277436] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1737), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5537), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [277462] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7746), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277486] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7748), 1,
      anon_sym_RBRACE,
    STATE(5945), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277510] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7750), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277534] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7752), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277558] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7754), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277582] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7756), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277606] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1737), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3993), 1,
      anon_sym_SEMI,
    STATE(420), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [277632] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7758), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277656] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1737), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [277682] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7760), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277706] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7762), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277730] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7764), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277754] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7766), 1,
      anon_sym_RPAREN,
    STATE(5483), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277778] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7766), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277802] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7768), 1,
      anon_sym_RBRACE,
    STATE(5484), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277826] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7768), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277850] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7770), 1,
      anon_sym_RBRACK,
    STATE(5485), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277874] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7770), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277898] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7772), 1,
      anon_sym_GT,
    STATE(5486), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277922] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7772), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277946] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7774), 1,
      anon_sym_PIPE,
    STATE(5487), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277970] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7774), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [277994] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7776), 1,
      anon_sym_SLASH,
    STATE(5488), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278018] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7776), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278042] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7778), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278066] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7780), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278090] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7782), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278114] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7784), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278138] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7786), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278162] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7788), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278186] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7790), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278210] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7792), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278234] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7794), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278258] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7796), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278282] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1775), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7798), 1,
      anon_sym_SEMI,
    STATE(426), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [278308] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7800), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278332] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7802), 1,
      anon_sym_DQUOTE,
    STATE(5507), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278356] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7802), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278380] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7804), 1,
      anon_sym_SQUOTE,
    STATE(5508), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278404] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7804), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278428] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7806), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5509), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278452] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7806), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278476] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7808), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5510), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278500] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7808), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278524] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1775), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [278550] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7810), 1,
      anon_sym_RPAREN,
    STATE(5833), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278574] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7812), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278598] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7810), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278622] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7814), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278646] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7816), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278670] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7818), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278694] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7820), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278718] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7822), 1,
      anon_sym_RBRACE,
    STATE(5912), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278742] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7824), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5582), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [278768] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7824), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [278794] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7822), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278818] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7826), 1,
      anon_sym_RBRACK,
    STATE(5913), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278842] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7826), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278866] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7828), 1,
      anon_sym_RPAREN,
    STATE(5529), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278890] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7828), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278914] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7830), 1,
      anon_sym_RBRACE,
    STATE(5530), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278938] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7830), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278962] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7832), 1,
      anon_sym_RBRACK,
    STATE(5531), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [278986] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7832), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279010] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7834), 1,
      anon_sym_GT,
    STATE(5532), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279034] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7834), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279058] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7836), 1,
      anon_sym_PIPE,
    STATE(5533), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279082] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7836), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279106] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7838), 1,
      anon_sym_SLASH,
    STATE(5534), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279130] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7838), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279154] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7840), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279178] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7842), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279202] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7844), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279226] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7846), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279250] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7848), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279274] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7850), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279298] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7748), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279322] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1773), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(7852), 1,
      anon_sym_SEMI,
    STATE(427), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [279348] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1773), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [279374] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7854), 1,
      anon_sym_RPAREN,
    STATE(5604), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279398] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7854), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279422] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7856), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279446] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7858), 1,
      anon_sym_DQUOTE,
    STATE(5553), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279470] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7858), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279494] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7860), 1,
      anon_sym_SQUOTE,
    STATE(5554), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279518] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7860), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279542] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7862), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5555), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279566] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7862), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279590] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7864), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5556), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279614] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7864), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279638] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7866), 1,
      anon_sym_RBRACE,
    STATE(5605), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279662] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7866), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279686] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7868), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279710] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7870), 1,
      anon_sym_RBRACK,
    STATE(5606), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279734] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7872), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279758] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7874), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279782] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7876), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279806] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7878), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279830] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7870), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279854] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7880), 1,
      anon_sym_GT,
    STATE(5607), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279878] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7880), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279902] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7882), 1,
      anon_sym_PIPE,
    STATE(5626), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279926] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7882), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279950] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7884), 1,
      anon_sym_SLASH,
    STATE(5627), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279974] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7886), 1,
      anon_sym_RPAREN,
    STATE(5575), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [279998] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7886), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280022] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7888), 1,
      anon_sym_RBRACE,
    STATE(5576), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280046] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7888), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280070] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7890), 1,
      anon_sym_RBRACK,
    STATE(5577), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280094] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7890), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280118] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7892), 1,
      anon_sym_GT,
    STATE(5960), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280142] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7892), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280166] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7894), 1,
      anon_sym_PIPE,
    STATE(5579), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280190] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7894), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280214] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7896), 1,
      anon_sym_SLASH,
    STATE(5580), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280238] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7896), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280262] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7898), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280286] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7900), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280310] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7902), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280334] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(7906), 1,
      anon_sym_COMMA,
    STATE(5191), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(7904), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
  [280354] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7908), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280378] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7910), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280402] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7884), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280426] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7912), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [280452] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7914), 1,
      anon_sym_RBRACK,
    STATE(5121), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280476] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7916), 1,
      anon_sym_GT,
    STATE(5921), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280500] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7916), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280524] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7918), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280548] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7920), 1,
      anon_sym_DQUOTE,
    STATE(5598), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280572] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7920), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280596] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7922), 1,
      anon_sym_SQUOTE,
    STATE(5599), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280620] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7922), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280644] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7924), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5600), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280668] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7924), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280692] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7926), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5601), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280716] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7926), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280740] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(7928), 1,
      anon_sym_COMMA,
    STATE(5946), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3656), 3,
      anon_sym_RPAREN,
      anon_sym_when,
      anon_sym_DASH_GT,
  [280760] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7930), 1,
      anon_sym_PIPE,
    STATE(5922), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280784] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(936), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5784), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [280810] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7932), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280834] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7934), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280858] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7936), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280882] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7938), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280906] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7940), 1,
      anon_sym_DQUOTE,
    STATE(5785), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280930] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7914), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280954] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7942), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [280978] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7944), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281002] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7946), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281026] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7948), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281050] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7950), 1,
      anon_sym_RPAREN,
    STATE(5620), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281074] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7950), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281098] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7952), 1,
      anon_sym_RBRACE,
    STATE(5621), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281122] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7952), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281146] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7954), 1,
      anon_sym_RBRACK,
    STATE(5622), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281170] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7954), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281194] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7956), 1,
      anon_sym_GT,
    STATE(5623), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281218] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7956), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281242] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7958), 1,
      anon_sym_PIPE,
    STATE(5624), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281266] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7958), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281290] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7960), 1,
      anon_sym_SLASH,
    STATE(5625), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281314] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7960), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281338] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7962), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281362] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(7964), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281386] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(7966), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281410] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7968), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281434] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7970), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281458] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7972), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281482] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7974), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281506] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(7976), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281530] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7978), 1,
      anon_sym_GT,
    STATE(5283), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281554] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7940), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281578] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7980), 1,
      anon_sym_SQUOTE,
    STATE(5787), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281602] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7982), 1,
      anon_sym_DQUOTE,
    STATE(5642), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281626] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7982), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281650] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7984), 1,
      anon_sym_SQUOTE,
    STATE(5643), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281674] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7984), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281698] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7986), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5644), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281722] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7986), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281746] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7988), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5645), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281770] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7988), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281794] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7980), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281818] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7990), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5800), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281842] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7990), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281866] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(7992), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281890] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(7994), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281914] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(7996), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281938] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(7998), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281962] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8000), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5801), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [281986] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8000), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282010] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(7930), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282034] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8002), 1,
      anon_sym_SLASH,
    STATE(5941), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282058] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(944), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5937), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [282084] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8002), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282108] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(8004), 1,
      anon_sym_RPAREN,
    STATE(5664), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282132] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(8004), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282156] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(8006), 1,
      anon_sym_RBRACE,
    STATE(5665), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282180] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(8006), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282204] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(8008), 1,
      anon_sym_RBRACK,
    STATE(5666), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282228] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(8008), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282252] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(8010), 1,
      anon_sym_GT,
    STATE(5667), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282276] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(8010), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282300] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(8012), 1,
      anon_sym_PIPE,
    STATE(5668), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282324] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(8012), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282348] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8014), 1,
      anon_sym_SLASH,
    STATE(5669), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282372] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8014), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282396] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(8016), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282420] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(8018), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282444] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(8020), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282468] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(8022), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282492] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(8024), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282516] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8026), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282540] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(7978), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282564] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(8028), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [282590] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8030), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5167), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282614] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8030), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282638] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1351), 1,
      anon_sym_RPAREN,
    ACTIONS(4588), 1,
      aux_sym__terminator_token1,
    ACTIONS(4591), 1,
      anon_sym_SEMI,
    STATE(347), 1,
      sym__terminator,
    STATE(1036), 1,
      aux_sym__terminator_repeat1,
    STATE(5120), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [282664] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8032), 1,
      anon_sym_DQUOTE,
    STATE(5683), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282688] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8032), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282712] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8034), 1,
      anon_sym_SQUOTE,
    STATE(5684), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282736] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8034), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282760] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8036), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5685), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282784] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8036), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282808] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8038), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5686), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282832] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8038), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282856] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8040), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282880] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8042), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282904] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8044), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282928] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8046), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282952] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(8048), 1,
      anon_sym_PIPE,
    STATE(5425), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [282976] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(954), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5754), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [283002] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8050), 1,
      anon_sym_DQUOTE,
    STATE(5755), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283026] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8050), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283050] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8052), 1,
      anon_sym_SQUOTE,
    STATE(5769), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283074] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8054), 1,
      anon_sym_DQUOTE,
    STATE(5700), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283098] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8054), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283122] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8056), 1,
      anon_sym_SQUOTE,
    STATE(5701), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283146] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8056), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283170] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8058), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5702), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283194] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8058), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283218] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8060), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5703), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283242] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8060), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283266] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8062), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283290] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8064), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283314] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8066), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283338] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8068), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283362] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8052), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283386] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8070), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5770), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283410] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8070), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283434] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8072), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5771), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283458] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8074), 1,
      anon_sym_DQUOTE,
    STATE(5716), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283482] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8074), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283506] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8076), 1,
      anon_sym_SQUOTE,
    STATE(5717), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283530] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8076), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283554] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8078), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5718), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283578] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8078), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283602] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8080), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5719), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283626] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8080), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283650] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8082), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283674] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8084), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283698] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8086), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283722] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8088), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283746] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8072), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283770] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1805), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5956), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [283796] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8090), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283820] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(958), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5816), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [283846] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8092), 1,
      anon_sym_DQUOTE,
    STATE(5732), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283870] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8092), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283894] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8094), 1,
      anon_sym_SQUOTE,
    STATE(5733), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283918] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8094), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283942] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8096), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5734), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283966] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8096), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [283990] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8098), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5735), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284014] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8098), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284038] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8100), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284062] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8102), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284086] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8104), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284110] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8106), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284134] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(8048), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284158] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8108), 1,
      anon_sym_SLASH,
    STATE(5722), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284182] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(924), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [284208] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1801), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5849), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [284234] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8110), 1,
      anon_sym_DQUOTE,
    STATE(5748), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284258] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8110), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284282] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8112), 1,
      anon_sym_SQUOTE,
    STATE(5749), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284306] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8112), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284330] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8114), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5750), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284354] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8114), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284378] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8116), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5751), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284402] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8116), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284426] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8118), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284450] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8120), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284474] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8122), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284498] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8124), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284522] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1801), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3997), 1,
      anon_sym_SEMI,
    STATE(358), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [284548] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1805), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3967), 1,
      anon_sym_SEMI,
    STATE(357), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [284574] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1801), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [284600] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8126), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284624] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8128), 1,
      anon_sym_DQUOTE,
    STATE(5764), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284648] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8128), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284672] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8130), 1,
      anon_sym_SQUOTE,
    STATE(5765), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284696] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8130), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284720] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8132), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5766), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284744] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8132), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284768] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8134), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5767), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284792] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8134), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284816] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8136), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284840] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8138), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284864] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8140), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284888] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8142), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284912] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8144), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5186), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284936] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8146), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284960] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8148), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [284984] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8150), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285008] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8152), 1,
      anon_sym_DQUOTE,
    STATE(5780), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285032] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8152), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285056] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8154), 1,
      anon_sym_SQUOTE,
    STATE(5781), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285080] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8154), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285104] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8156), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5782), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285128] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8156), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285152] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8158), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5783), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285176] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8158), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285200] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8160), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285224] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8162), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285248] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8164), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285272] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8166), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285296] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1805), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [285322] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8168), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285346] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8144), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285370] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8170), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285394] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8172), 1,
      anon_sym_DQUOTE,
    STATE(5796), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285418] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8172), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285442] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8174), 1,
      anon_sym_SQUOTE,
    STATE(5797), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285466] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8174), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285490] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8176), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5798), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285514] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8176), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285538] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8178), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5799), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285562] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8178), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285586] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8180), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285610] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8182), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285634] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8184), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285658] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8186), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285682] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8188), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285706] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8190), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285730] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(940), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5127), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [285756] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(8192), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5904), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [285782] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8194), 1,
      anon_sym_DQUOTE,
    STATE(5812), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285806] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8194), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285830] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8196), 1,
      anon_sym_SQUOTE,
    STATE(5813), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285854] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8196), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285878] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8198), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5814), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285902] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8198), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285926] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8200), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5815), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285950] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8200), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285974] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8202), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [285998] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8204), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286022] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8206), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286046] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8208), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286070] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(8192), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [286096] = 7,
    ACTIONS(8210), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8212), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(8215), 1,
      sym_escape_sequence,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286120] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8218), 1,
      anon_sym_DQUOTE,
    STATE(5130), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286144] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8108), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286168] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8220), 1,
      anon_sym_DQUOTE,
    STATE(5828), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286192] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8220), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286216] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8222), 1,
      anon_sym_SQUOTE,
    STATE(5829), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286240] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8222), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286264] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8224), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5830), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286288] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8224), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286312] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8226), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5831), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286336] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8226), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286360] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8228), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286384] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8230), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286408] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8232), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286432] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8234), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286456] = 7,
    ACTIONS(8236), 1,
      anon_sym_SLASH,
    ACTIONS(8238), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(8241), 1,
      sym_escape_sequence,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286480] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(8244), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286504] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8218), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286528] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8246), 1,
      anon_sym_SQUOTE,
    STATE(5147), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286552] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8248), 1,
      anon_sym_DQUOTE,
    STATE(5844), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286576] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8248), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286600] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8250), 1,
      anon_sym_SQUOTE,
    STATE(5845), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286624] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8250), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286648] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8252), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5846), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286672] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8252), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286696] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8254), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5847), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286720] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8254), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286744] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8256), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286768] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8258), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286792] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8260), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286816] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8262), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286840] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1593), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(8264), 1,
      anon_sym_SEMI,
    STATE(361), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [286866] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1593), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [286892] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(8266), 1,
      anon_sym_RPAREN,
    STATE(5926), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286916] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(8266), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286940] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8268), 1,
      anon_sym_DQUOTE,
    STATE(5860), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286964] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8268), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [286988] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8270), 1,
      anon_sym_SQUOTE,
    STATE(5861), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287012] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8270), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287036] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8272), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5862), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287060] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8272), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287084] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5863), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287108] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8274), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287132] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8276), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287156] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8278), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287180] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8280), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287204] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8282), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287228] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(8284), 1,
      anon_sym_RBRACE,
    STATE(5927), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287252] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(8284), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287276] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8286), 1,
      anon_sym_DQUOTE,
    STATE(5870), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287300] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8286), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287324] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8288), 1,
      anon_sym_SQUOTE,
    STATE(5871), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287348] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8288), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287372] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8290), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287396] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8292), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287420] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(8294), 1,
      anon_sym_RBRACK,
    STATE(5932), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287444] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(8294), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287468] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8296), 1,
      anon_sym_DQUOTE,
    STATE(5878), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287492] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8296), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287516] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8298), 1,
      anon_sym_SQUOTE,
    STATE(5879), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287540] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8298), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287564] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8300), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287588] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8302), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287612] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(8304), 1,
      anon_sym_GT,
    STATE(5933), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287636] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(8304), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287660] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8306), 1,
      anon_sym_DQUOTE,
    STATE(5886), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287684] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8306), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287708] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8308), 1,
      anon_sym_SQUOTE,
    STATE(5887), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287732] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8308), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287756] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8310), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287780] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8312), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287804] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(8314), 1,
      anon_sym_PIPE,
    STATE(5934), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287828] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(8314), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287852] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8316), 1,
      anon_sym_DQUOTE,
    STATE(5894), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287876] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8316), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287900] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8318), 1,
      anon_sym_SQUOTE,
    STATE(5895), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287924] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8318), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287948] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8320), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287972] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8322), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [287996] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8324), 1,
      anon_sym_SLASH,
    STATE(5935), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288020] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8324), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288044] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8326), 1,
      anon_sym_DQUOTE,
    STATE(5902), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288068] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8326), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288092] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8328), 1,
      anon_sym_SQUOTE,
    STATE(5903), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288116] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8328), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288140] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8330), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288164] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8332), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288188] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(8334), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [288214] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8246), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288238] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8336), 1,
      anon_sym_DQUOTE,
    STATE(5910), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288262] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8336), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288286] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8338), 1,
      anon_sym_SQUOTE,
    STATE(5911), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288310] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8338), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288334] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8340), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288358] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8342), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288382] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(8344), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288406] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(8346), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288430] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8348), 1,
      anon_sym_DQUOTE,
    STATE(5918), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288454] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8348), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288478] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8350), 1,
      anon_sym_SQUOTE,
    STATE(5919), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288502] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8350), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288526] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8352), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288550] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8354), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288574] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8356), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5150), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288598] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(8358), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288622] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(8360), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288646] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8356), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288670] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8362), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5153), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288694] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8362), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288718] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(8364), 1,
      anon_sym_RPAREN,
    STATE(5273), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288742] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(8366), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288766] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(7650), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [288792] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4351), 1,
      anon_sym_end,
    ACTIONS(6089), 1,
      aux_sym__terminator_token1,
    ACTIONS(8368), 1,
      anon_sym_SEMI,
    STATE(888), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5929), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [288818] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8371), 1,
      anon_sym_SQUOTE,
    STATE(5959), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288842] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9), 1,
      aux_sym__terminator_token1,
    ACTIONS(1727), 1,
      ts_builtin_sym_end,
    ACTIONS(8373), 1,
      anon_sym_SEMI,
    STATE(387), 1,
      sym__terminator,
    STATE(1035), 1,
      aux_sym__terminator_repeat1,
    STATE(5131), 1,
      aux_sym_source_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [288868] = 7,
    ACTIONS(6293), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6295), 1,
      sym_escape_sequence,
    ACTIONS(8375), 1,
      anon_sym_RBRACK,
    STATE(5126), 1,
      aux_sym__quoted_i_square_repeat1,
    STATE(5964), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288892] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(8377), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288916] = 7,
    ACTIONS(6321), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6323), 1,
      sym_escape_sequence,
    ACTIONS(8379), 1,
      anon_sym_PIPE,
    STATE(5445), 1,
      aux_sym__quoted_i_bar_repeat1,
    STATE(5990), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288940] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8381), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [288964] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(8383), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5119), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [288990] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(8383), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [289016] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8371), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289040] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1695), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5148), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [289066] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8385), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5942), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289090] = 7,
    ACTIONS(6329), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6331), 1,
      sym_escape_sequence,
    ACTIONS(8387), 1,
      anon_sym_SLASH,
    STATE(5832), 1,
      aux_sym__quoted_i_slash_repeat1,
    STATE(5992), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289114] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8389), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289138] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1695), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(3973), 1,
      anon_sym_SEMI,
    STATE(372), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [289164] = 7,
    ACTIONS(6261), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6263), 1,
      sym_escape_sequence,
    ACTIONS(8385), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(5165), 1,
      aux_sym__quoted_i_heredoc_single_repeat1,
    STATE(6178), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289188] = 7,
    ACTIONS(6285), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6287), 1,
      sym_escape_sequence,
    ACTIONS(8391), 1,
      anon_sym_RBRACE,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289212] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(7928), 1,
      anon_sym_COMMA,
    STATE(5264), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(3662), 3,
      anon_sym_RPAREN,
      anon_sym_when,
      anon_sym_DASH_GT,
  [289232] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8393), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5231), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289256] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(918), 1,
      anon_sym_end,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5289), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [289282] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1695), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [289308] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(7257), 1,
      anon_sym_SEMI,
    ACTIONS(8395), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [289334] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(8397), 1,
      aux_sym__terminator_token1,
    ACTIONS(8400), 1,
      anon_sym_SEMI,
    ACTIONS(8403), 1,
      anon_sym_end,
    STATE(139), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5951), 1,
      aux_sym_anonymous_function_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [289360] = 7,
    ACTIONS(6269), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6271), 1,
      sym_escape_sequence,
    ACTIONS(8393), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(5817), 1,
      aux_sym__quoted_i_heredoc_double_repeat1,
    STATE(6180), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289384] = 7,
    ACTIONS(6245), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6247), 1,
      sym_escape_sequence,
    ACTIONS(8405), 1,
      anon_sym_DQUOTE,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289408] = 7,
    ACTIONS(8407), 1,
      anon_sym_RBRACE,
    ACTIONS(8409), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(8412), 1,
      sym_escape_sequence,
    STATE(5954), 1,
      aux_sym__quoted_i_curly_repeat1,
    STATE(6172), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289432] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1591), 1,
      anon_sym_RPAREN,
    ACTIONS(3773), 1,
      aux_sym__terminator_token1,
    ACTIONS(8415), 1,
      anon_sym_SEMI,
    STATE(377), 1,
      sym__terminator,
    STATE(1030), 1,
      aux_sym__terminator_repeat1,
    STATE(5199), 1,
      aux_sym_block_repeat2,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [289458] = 8,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(680), 1,
      aux_sym__terminator_token1,
    ACTIONS(1591), 1,
      anon_sym_RPAREN,
    ACTIONS(7259), 1,
      anon_sym_SEMI,
    STATE(137), 1,
      sym__terminator,
    STATE(1029), 1,
      aux_sym__terminator_repeat1,
    STATE(5282), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
  [289484] = 7,
    ACTIONS(8417), 1,
      anon_sym_DQUOTE,
    ACTIONS(8419), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(8422), 1,
      sym_escape_sequence,
    STATE(5957), 1,
      aux_sym__quoted_i_double_repeat1,
    STATE(6114), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289508] = 7,
    ACTIONS(6277), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6279), 1,
      sym_escape_sequence,
    ACTIONS(7253), 1,
      anon_sym_RPAREN,
    STATE(5149), 1,
      aux_sym__quoted_i_parenthesis_repeat1,
    STATE(6146), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289532] = 7,
    ACTIONS(6253), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6255), 1,
      sym_escape_sequence,
    ACTIONS(8425), 1,
      anon_sym_SQUOTE,
    STATE(5419), 1,
      aux_sym__quoted_i_single_repeat1,
    STATE(6175), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289556] = 7,
    ACTIONS(6313), 1,
      anon_sym_POUND_LBRACE,
    ACTIONS(6315), 1,
      sym_escape_sequence,
    ACTIONS(8427), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym__quoted_i_angle_repeat1,
    STATE(5985), 1,
      sym_interpolation,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289580] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_curly,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [289597] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(199), 1,
      anon_sym_DQUOTE,
    ACTIONS(201), 1,
      anon_sym_SQUOTE,
    STATE(1439), 1,
      sym__quoted_i_double,
    STATE(1443), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [289618] = 6,
    ACTIONS(8433), 1,
      anon_sym_RPAREN,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8437), 1,
      sym__quoted_content_parenthesis,
    STATE(6567), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289639] = 4,
    ACTIONS(8441), 1,
      sym__quoted_content_i_square,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8439), 3,
      anon_sym_RBRACK,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [289656] = 6,
    ACTIONS(8443), 1,
      anon_sym_DQUOTE,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8447), 1,
      sym__quoted_content_double,
    STATE(6614), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289677] = 6,
    ACTIONS(8449), 1,
      anon_sym_SQUOTE,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8453), 1,
      sym__quoted_content_single,
    STATE(6635), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289698] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8455), 1,
      anon_sym_RPAREN,
    ACTIONS(8457), 1,
      sym__quoted_content_parenthesis,
    STATE(6347), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289719] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8459), 1,
      anon_sym_DQUOTE,
    ACTIONS(8461), 1,
      sym__quoted_content_double,
    STATE(6349), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289740] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8463), 1,
      anon_sym_SQUOTE,
    ACTIONS(8465), 1,
      sym__quoted_content_single,
    STATE(6351), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289761] = 6,
    ACTIONS(8467), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8471), 1,
      sym__quoted_content_heredoc_single,
    STATE(6353), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289782] = 6,
    ACTIONS(8473), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8477), 1,
      sym__quoted_content_heredoc_double,
    STATE(6355), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289803] = 6,
    ACTIONS(8479), 1,
      anon_sym_RBRACE,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8483), 1,
      sym__quoted_content_curly,
    STATE(6357), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289824] = 6,
    ACTIONS(8485), 1,
      anon_sym_RBRACK,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8489), 1,
      sym__quoted_content_square,
    STATE(6379), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289845] = 6,
    ACTIONS(8491), 1,
      anon_sym_GT,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8495), 1,
      sym__quoted_content_angle,
    STATE(6391), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289866] = 6,
    ACTIONS(8497), 1,
      anon_sym_PIPE,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8501), 1,
      sym__quoted_content_bar,
    STATE(6393), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289887] = 6,
    ACTIONS(8503), 1,
      anon_sym_SLASH,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8507), 1,
      sym__quoted_content_slash,
    STATE(6395), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289908] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8509), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8511), 1,
      sym__quoted_content_heredoc_single,
    STATE(6639), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289929] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8513), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8515), 1,
      sym__quoted_content_heredoc_double,
    STATE(6644), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289950] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8517), 1,
      anon_sym_RBRACE,
    ACTIONS(8519), 1,
      sym__quoted_content_curly,
    STATE(6655), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289971] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8521), 1,
      anon_sym_RBRACK,
    ACTIONS(8523), 1,
      sym__quoted_content_square,
    STATE(6193), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [289992] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8525), 1,
      anon_sym_GT,
    ACTIONS(8527), 1,
      sym__quoted_content_angle,
    STATE(6195), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290013] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8529), 1,
      anon_sym_PIPE,
    ACTIONS(8531), 1,
      sym__quoted_content_bar,
    STATE(6197), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290034] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8533), 1,
      anon_sym_SLASH,
    ACTIONS(8535), 1,
      sym__quoted_content_slash,
    STATE(6199), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290055] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_slash,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_SLASH,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [290072] = 4,
    ACTIONS(8543), 1,
      sym__quoted_content_i_angle,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8541), 3,
      anon_sym_GT,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [290089] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(245), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      anon_sym_SQUOTE,
    STATE(1277), 1,
      sym__quoted_i_double,
    STATE(1278), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [290110] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(73), 1,
      anon_sym_DQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE,
    STATE(1792), 1,
      sym__quoted_i_double,
    STATE(1793), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [290131] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_parenthesis,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_RPAREN,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [290148] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_heredoc_single,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [290165] = 4,
    ACTIONS(8547), 1,
      sym__quoted_content_i_bar,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8545), 3,
      anon_sym_PIPE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [290182] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_bar,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_PIPE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [290199] = 4,
    ACTIONS(8551), 1,
      sym__quoted_content_i_slash,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8549), 3,
      anon_sym_SLASH,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [290216] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8553), 1,
      anon_sym_RPAREN,
    ACTIONS(8555), 1,
      sym__quoted_content_parenthesis,
    STATE(6667), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290237] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8557), 1,
      anon_sym_DQUOTE,
    ACTIONS(8559), 1,
      sym__quoted_content_double,
    STATE(6669), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290258] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8561), 1,
      anon_sym_SQUOTE,
    ACTIONS(8563), 1,
      sym__quoted_content_single,
    STATE(6671), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290279] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8565), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8567), 1,
      sym__quoted_content_heredoc_single,
    STATE(6673), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290300] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8569), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8571), 1,
      sym__quoted_content_heredoc_double,
    STATE(6675), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290321] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8573), 1,
      anon_sym_RBRACE,
    ACTIONS(8575), 1,
      sym__quoted_content_curly,
    STATE(6677), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290342] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8577), 1,
      anon_sym_RBRACK,
    ACTIONS(8579), 1,
      sym__quoted_content_square,
    STATE(6679), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290363] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8581), 1,
      anon_sym_GT,
    ACTIONS(8583), 1,
      sym__quoted_content_angle,
    STATE(6681), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290384] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8585), 1,
      anon_sym_PIPE,
    ACTIONS(8587), 1,
      sym__quoted_content_bar,
    STATE(6683), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290405] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8589), 1,
      anon_sym_SLASH,
    ACTIONS(8591), 1,
      sym__quoted_content_slash,
    STATE(6685), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290426] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8593), 1,
      anon_sym_RPAREN,
    ACTIONS(8595), 1,
      sym__quoted_content_parenthesis,
    STATE(6217), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290447] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8597), 1,
      anon_sym_DQUOTE,
    ACTIONS(8599), 1,
      sym__quoted_content_double,
    STATE(6219), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290468] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8601), 1,
      anon_sym_SQUOTE,
    ACTIONS(8603), 1,
      sym__quoted_content_single,
    STATE(6221), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290489] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8605), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8607), 1,
      sym__quoted_content_heredoc_single,
    STATE(6223), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290510] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8609), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8611), 1,
      sym__quoted_content_heredoc_double,
    STATE(6225), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290531] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8613), 1,
      anon_sym_RBRACE,
    ACTIONS(8615), 1,
      sym__quoted_content_curly,
    STATE(6247), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290552] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8617), 1,
      anon_sym_RBRACK,
    ACTIONS(8619), 1,
      sym__quoted_content_square,
    STATE(6249), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290573] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8621), 1,
      anon_sym_GT,
    ACTIONS(8623), 1,
      sym__quoted_content_angle,
    STATE(6251), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290594] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8625), 1,
      anon_sym_PIPE,
    ACTIONS(8627), 1,
      sym__quoted_content_bar,
    STATE(6263), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290615] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8629), 1,
      anon_sym_SLASH,
    ACTIONS(8631), 1,
      sym__quoted_content_slash,
    STATE(6265), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290636] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(8633), 1,
      anon_sym_COMMA,
    STATE(5384), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(1281), 2,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [290655] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1101), 1,
      anon_sym_DQUOTE,
    ACTIONS(1103), 1,
      anon_sym_SQUOTE,
    STATE(2953), 1,
      sym__quoted_i_double,
    STATE(2954), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [290676] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8635), 1,
      anon_sym_RPAREN,
    ACTIONS(8637), 1,
      sym__quoted_content_parenthesis,
    STATE(6227), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290697] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8639), 1,
      anon_sym_DQUOTE,
    ACTIONS(8641), 1,
      sym__quoted_content_double,
    STATE(6229), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290718] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8643), 1,
      anon_sym_SQUOTE,
    ACTIONS(8645), 1,
      sym__quoted_content_single,
    STATE(6231), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290739] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8647), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8649), 1,
      sym__quoted_content_heredoc_single,
    STATE(6233), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290760] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8651), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8653), 1,
      sym__quoted_content_heredoc_double,
    STATE(6235), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290781] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8655), 1,
      anon_sym_RBRACE,
    ACTIONS(8657), 1,
      sym__quoted_content_curly,
    STATE(6237), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290802] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8659), 1,
      anon_sym_RBRACK,
    ACTIONS(8661), 1,
      sym__quoted_content_square,
    STATE(6239), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290823] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8663), 1,
      anon_sym_GT,
    ACTIONS(8665), 1,
      sym__quoted_content_angle,
    STATE(6241), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290844] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8667), 1,
      anon_sym_PIPE,
    ACTIONS(8669), 1,
      sym__quoted_content_bar,
    STATE(6243), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290865] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8671), 1,
      anon_sym_SLASH,
    ACTIONS(8673), 1,
      sym__quoted_content_slash,
    STATE(6245), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290886] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8675), 1,
      anon_sym_RPAREN,
    ACTIONS(8677), 1,
      sym__quoted_content_parenthesis,
    STATE(6474), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290907] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8679), 1,
      anon_sym_DQUOTE,
    ACTIONS(8681), 1,
      sym__quoted_content_double,
    STATE(6476), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290928] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8683), 1,
      anon_sym_SQUOTE,
    ACTIONS(8685), 1,
      sym__quoted_content_single,
    STATE(6478), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290949] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8687), 1,
      anon_sym_RPAREN,
    ACTIONS(8689), 1,
      sym__quoted_content_parenthesis,
    STATE(6307), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290970] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8691), 1,
      anon_sym_DQUOTE,
    ACTIONS(8693), 1,
      sym__quoted_content_double,
    STATE(6309), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [290991] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8695), 1,
      anon_sym_SQUOTE,
    ACTIONS(8697), 1,
      sym__quoted_content_single,
    STATE(6311), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291012] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8699), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8701), 1,
      sym__quoted_content_heredoc_single,
    STATE(6313), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291033] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8703), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8705), 1,
      sym__quoted_content_heredoc_double,
    STATE(6315), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291054] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8707), 1,
      anon_sym_RBRACE,
    ACTIONS(8709), 1,
      sym__quoted_content_curly,
    STATE(6317), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291075] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8711), 1,
      anon_sym_RBRACK,
    ACTIONS(8713), 1,
      sym__quoted_content_square,
    STATE(6319), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291096] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8715), 1,
      anon_sym_GT,
    ACTIONS(8717), 1,
      sym__quoted_content_angle,
    STATE(6321), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291117] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8719), 1,
      anon_sym_PIPE,
    ACTIONS(8721), 1,
      sym__quoted_content_bar,
    STATE(6323), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291138] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8723), 1,
      anon_sym_SLASH,
    ACTIONS(8725), 1,
      sym__quoted_content_slash,
    STATE(6325), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291159] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8727), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8729), 1,
      sym__quoted_content_heredoc_single,
    STATE(6481), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291180] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_bar,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_PIPE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [291197] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8731), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8733), 1,
      sym__quoted_content_heredoc_double,
    STATE(6485), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291218] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(441), 1,
      anon_sym_DQUOTE,
    ACTIONS(443), 1,
      anon_sym_SQUOTE,
    STATE(3466), 1,
      sym__quoted_i_double,
    STATE(3467), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [291239] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8735), 1,
      anon_sym_RBRACE,
    ACTIONS(8737), 1,
      sym__quoted_content_curly,
    STATE(6516), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291260] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8739), 1,
      anon_sym_RBRACK,
    ACTIONS(8741), 1,
      sym__quoted_content_square,
    STATE(6518), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291281] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(369), 1,
      anon_sym_DQUOTE,
    ACTIONS(371), 1,
      anon_sym_SQUOTE,
    STATE(2170), 1,
      sym__quoted_i_double,
    STATE(2172), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [291302] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8743), 1,
      anon_sym_GT,
    ACTIONS(8745), 1,
      sym__quoted_content_angle,
    STATE(6530), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291323] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8747), 1,
      anon_sym_PIPE,
    ACTIONS(8749), 1,
      sym__quoted_content_bar,
    STATE(6532), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291344] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8751), 1,
      anon_sym_SLASH,
    ACTIONS(8753), 1,
      sym__quoted_content_slash,
    STATE(6534), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291365] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8755), 1,
      anon_sym_RPAREN,
    ACTIONS(8757), 1,
      sym__quoted_content_parenthesis,
    STATE(6574), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291386] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_single,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_SQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [291403] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8759), 1,
      anon_sym_RPAREN,
    ACTIONS(8761), 1,
      sym__quoted_content_parenthesis,
    STATE(6360), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291424] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8763), 1,
      anon_sym_DQUOTE,
    ACTIONS(8765), 1,
      sym__quoted_content_double,
    STATE(6362), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291445] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8767), 1,
      anon_sym_SQUOTE,
    ACTIONS(8769), 1,
      sym__quoted_content_single,
    STATE(6364), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291466] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8771), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8773), 1,
      sym__quoted_content_heredoc_single,
    STATE(6366), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291487] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8775), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8777), 1,
      sym__quoted_content_heredoc_double,
    STATE(6368), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291508] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8779), 1,
      anon_sym_RBRACE,
    ACTIONS(8781), 1,
      sym__quoted_content_curly,
    STATE(6370), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291529] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8783), 1,
      anon_sym_RBRACK,
    ACTIONS(8785), 1,
      sym__quoted_content_square,
    STATE(6372), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291550] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8787), 1,
      anon_sym_GT,
    ACTIONS(8789), 1,
      sym__quoted_content_angle,
    STATE(6374), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291571] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8791), 1,
      anon_sym_PIPE,
    ACTIONS(8793), 1,
      sym__quoted_content_bar,
    STATE(6376), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291592] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8795), 1,
      anon_sym_SLASH,
    ACTIONS(8797), 1,
      sym__quoted_content_slash,
    STATE(6378), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291613] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8799), 1,
      anon_sym_DQUOTE,
    ACTIONS(8801), 1,
      sym__quoted_content_double,
    STATE(6290), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291634] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(8803), 1,
      anon_sym_COMMA,
    STATE(6061), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(5096), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [291653] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_slash,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_SLASH,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [291670] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8806), 1,
      anon_sym_RPAREN,
    ACTIONS(8808), 1,
      sym__quoted_content_parenthesis,
    STATE(6442), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291691] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8810), 1,
      anon_sym_DQUOTE,
    ACTIONS(8812), 1,
      sym__quoted_content_double,
    STATE(6444), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291712] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8814), 1,
      anon_sym_SQUOTE,
    ACTIONS(8816), 1,
      sym__quoted_content_single,
    STATE(6446), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291733] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8818), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8820), 1,
      sym__quoted_content_heredoc_single,
    STATE(6458), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291754] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8822), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8824), 1,
      sym__quoted_content_heredoc_double,
    STATE(6460), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291775] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8826), 1,
      anon_sym_RBRACE,
    ACTIONS(8828), 1,
      sym__quoted_content_curly,
    STATE(6462), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291796] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8830), 1,
      anon_sym_RBRACK,
    ACTIONS(8832), 1,
      sym__quoted_content_square,
    STATE(6465), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291817] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8834), 1,
      anon_sym_GT,
    ACTIONS(8836), 1,
      sym__quoted_content_angle,
    STATE(6467), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291838] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8838), 1,
      anon_sym_PIPE,
    ACTIONS(8840), 1,
      sym__quoted_content_bar,
    STATE(6469), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291859] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8842), 1,
      anon_sym_SLASH,
    ACTIONS(8844), 1,
      sym__quoted_content_slash,
    STATE(6471), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291880] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8846), 1,
      anon_sym_RPAREN,
    ACTIONS(8848), 1,
      sym__quoted_content_parenthesis,
    STATE(6423), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291901] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8850), 1,
      anon_sym_DQUOTE,
    ACTIONS(8852), 1,
      sym__quoted_content_double,
    STATE(6425), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291922] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8854), 1,
      anon_sym_SQUOTE,
    ACTIONS(8856), 1,
      sym__quoted_content_single,
    STATE(6427), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291943] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8858), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8860), 1,
      sym__quoted_content_heredoc_single,
    STATE(6429), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291964] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8862), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8864), 1,
      sym__quoted_content_heredoc_double,
    STATE(6431), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [291985] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8866), 1,
      anon_sym_RBRACE,
    ACTIONS(8868), 1,
      sym__quoted_content_curly,
    STATE(6433), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292006] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8870), 1,
      anon_sym_RBRACK,
    ACTIONS(8872), 1,
      sym__quoted_content_square,
    STATE(6435), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292027] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8874), 1,
      anon_sym_GT,
    ACTIONS(8876), 1,
      sym__quoted_content_angle,
    STATE(6437), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292048] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8878), 1,
      anon_sym_PIPE,
    ACTIONS(8880), 1,
      sym__quoted_content_bar,
    STATE(6439), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292069] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8882), 1,
      anon_sym_SLASH,
    ACTIONS(8884), 1,
      sym__quoted_content_slash,
    STATE(6441), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292090] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8886), 1,
      anon_sym_RPAREN,
    ACTIONS(8888), 1,
      sym__quoted_content_parenthesis,
    STATE(6537), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292111] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8890), 1,
      anon_sym_DQUOTE,
    ACTIONS(8892), 1,
      sym__quoted_content_double,
    STATE(6541), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292132] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_heredoc_single,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [292149] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8894), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8896), 1,
      sym__quoted_content_heredoc_single,
    STATE(6545), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292170] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8898), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8900), 1,
      sym__quoted_content_heredoc_double,
    STATE(6547), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292191] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8902), 1,
      anon_sym_RBRACE,
    ACTIONS(8904), 1,
      sym__quoted_content_curly,
    STATE(6549), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292212] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8906), 1,
      anon_sym_RBRACK,
    ACTIONS(8908), 1,
      sym__quoted_content_square,
    STATE(6551), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292233] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8910), 1,
      anon_sym_GT,
    ACTIONS(8912), 1,
      sym__quoted_content_angle,
    STATE(6554), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292254] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8914), 1,
      anon_sym_PIPE,
    ACTIONS(8916), 1,
      sym__quoted_content_bar,
    STATE(6556), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292275] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8918), 1,
      anon_sym_SLASH,
    ACTIONS(8920), 1,
      sym__quoted_content_slash,
    STATE(6558), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292296] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8922), 1,
      anon_sym_RPAREN,
    ACTIONS(8924), 1,
      sym__quoted_content_parenthesis,
    STATE(6496), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292317] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8926), 1,
      anon_sym_DQUOTE,
    ACTIONS(8928), 1,
      sym__quoted_content_double,
    STATE(6498), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292338] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8930), 1,
      anon_sym_SQUOTE,
    ACTIONS(8932), 1,
      sym__quoted_content_single,
    STATE(6500), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292359] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8934), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8936), 1,
      sym__quoted_content_heredoc_single,
    STATE(6770), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292380] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8938), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8940), 1,
      sym__quoted_content_heredoc_double,
    STATE(6504), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292401] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8942), 1,
      anon_sym_RBRACE,
    ACTIONS(8944), 1,
      sym__quoted_content_curly,
    STATE(6506), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292422] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8946), 1,
      anon_sym_RBRACK,
    ACTIONS(8948), 1,
      sym__quoted_content_square,
    STATE(6508), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292443] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8950), 1,
      anon_sym_GT,
    ACTIONS(8952), 1,
      sym__quoted_content_angle,
    STATE(6509), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292464] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8954), 1,
      anon_sym_PIPE,
    ACTIONS(8956), 1,
      sym__quoted_content_bar,
    STATE(6511), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292485] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8958), 1,
      anon_sym_SLASH,
    ACTIONS(8960), 1,
      sym__quoted_content_slash,
    STATE(6513), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292506] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(550), 1,
      anon_sym_DQUOTE,
    ACTIONS(552), 1,
      anon_sym_SQUOTE,
    STATE(3663), 1,
      sym__quoted_i_double,
    STATE(3664), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [292527] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(8962), 1,
      anon_sym_RPAREN,
    ACTIONS(8964), 1,
      sym__quoted_content_parenthesis,
    STATE(6735), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292548] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(8966), 1,
      anon_sym_DQUOTE,
    ACTIONS(8968), 1,
      sym__quoted_content_double,
    STATE(6766), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292569] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(8970), 1,
      anon_sym_SQUOTE,
    ACTIONS(8972), 1,
      sym__quoted_content_single,
    STATE(6203), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292590] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(8974), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(8976), 1,
      sym__quoted_content_heredoc_single,
    STATE(6207), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292611] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(8978), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(8980), 1,
      sym__quoted_content_heredoc_double,
    STATE(6209), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292632] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(8982), 1,
      anon_sym_RBRACE,
    ACTIONS(8984), 1,
      sym__quoted_content_curly,
    STATE(6211), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292653] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(8986), 1,
      anon_sym_RBRACK,
    ACTIONS(8988), 1,
      sym__quoted_content_square,
    STATE(6560), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292674] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(8990), 1,
      anon_sym_GT,
    ACTIONS(8992), 1,
      sym__quoted_content_angle,
    STATE(6569), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292695] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(8994), 1,
      anon_sym_PIPE,
    ACTIONS(8996), 1,
      sym__quoted_content_bar,
    STATE(6572), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292716] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(8998), 1,
      anon_sym_SLASH,
    ACTIONS(9000), 1,
      sym__quoted_content_slash,
    STATE(6213), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292737] = 4,
    ACTIONS(9004), 1,
      sym__quoted_content_i_double,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9002), 3,
      anon_sym_DQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [292754] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9006), 1,
      anon_sym_RPAREN,
    ACTIONS(9008), 1,
      sym__quoted_content_parenthesis,
    STATE(6215), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292775] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9010), 1,
      anon_sym_SQUOTE,
    ACTIONS(9012), 1,
      sym__quoted_content_single,
    STATE(6562), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292796] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9014), 1,
      anon_sym_DQUOTE,
    ACTIONS(9016), 1,
      sym__quoted_content_double,
    STATE(6277), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292817] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_double,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_DQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [292834] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9018), 1,
      anon_sym_SQUOTE,
    ACTIONS(9020), 1,
      sym__quoted_content_single,
    STATE(6340), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292855] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9022), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(9024), 1,
      sym__quoted_content_heredoc_single,
    STATE(6717), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292876] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9026), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(9028), 1,
      sym__quoted_content_heredoc_double,
    STATE(6720), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292897] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9030), 1,
      anon_sym_RBRACE,
    ACTIONS(9032), 1,
      sym__quoted_content_curly,
    STATE(6723), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292918] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9034), 1,
      anon_sym_RPAREN,
    ACTIONS(9036), 1,
      sym__quoted_content_parenthesis,
    STATE(6593), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292939] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9038), 1,
      anon_sym_DQUOTE,
    ACTIONS(9040), 1,
      sym__quoted_content_double,
    STATE(6595), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292960] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9042), 1,
      anon_sym_SQUOTE,
    ACTIONS(9044), 1,
      sym__quoted_content_single,
    STATE(6597), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [292981] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9046), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(9048), 1,
      sym__quoted_content_heredoc_single,
    STATE(6599), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293002] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9050), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(9052), 1,
      sym__quoted_content_heredoc_double,
    STATE(6601), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293023] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9054), 1,
      anon_sym_RBRACE,
    ACTIONS(9056), 1,
      sym__quoted_content_curly,
    STATE(6603), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293044] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9058), 1,
      anon_sym_RBRACK,
    ACTIONS(9060), 1,
      sym__quoted_content_square,
    STATE(6605), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293065] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9062), 1,
      anon_sym_GT,
    ACTIONS(9064), 1,
      sym__quoted_content_angle,
    STATE(6607), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293086] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9066), 1,
      anon_sym_PIPE,
    ACTIONS(9068), 1,
      sym__quoted_content_bar,
    STATE(6609), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293107] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9070), 1,
      anon_sym_SLASH,
    ACTIONS(9072), 1,
      sym__quoted_content_slash,
    STATE(6611), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293128] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_DQUOTE,
    ACTIONS(25), 1,
      anon_sym_SQUOTE,
    STATE(4131), 1,
      sym__quoted_i_double,
    STATE(4132), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [293149] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9074), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(9076), 1,
      sym__quoted_content_heredoc_single,
    STATE(6342), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293170] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9078), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(9080), 1,
      sym__quoted_content_heredoc_double,
    STATE(6345), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293191] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9082), 1,
      anon_sym_RBRACK,
    ACTIONS(9084), 1,
      sym__quoted_content_square,
    STATE(6726), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293212] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9086), 1,
      anon_sym_RBRACE,
    ACTIONS(9088), 1,
      sym__quoted_content_curly,
    STATE(6397), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293233] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9090), 1,
      anon_sym_RBRACK,
    ACTIONS(9092), 1,
      sym__quoted_content_square,
    STATE(6399), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293254] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9094), 1,
      anon_sym_GT,
    ACTIONS(9096), 1,
      sym__quoted_content_angle,
    STATE(6403), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293275] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9098), 1,
      anon_sym_GT,
    ACTIONS(9100), 1,
      sym__quoted_content_angle,
    STATE(6733), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293296] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(976), 1,
      anon_sym_DQUOTE,
    ACTIONS(978), 1,
      anon_sym_SQUOTE,
    STATE(1860), 1,
      sym__quoted_i_double,
    STATE(1863), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [293317] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9102), 1,
      anon_sym_PIPE,
    ACTIONS(9104), 1,
      sym__quoted_content_bar,
    STATE(6405), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293338] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9106), 1,
      anon_sym_PIPE,
    ACTIONS(9108), 1,
      sym__quoted_content_bar,
    STATE(6737), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293359] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9110), 1,
      anon_sym_SLASH,
    ACTIONS(9112), 1,
      sym__quoted_content_slash,
    STATE(6407), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293380] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(598), 1,
      anon_sym_DQUOTE,
    ACTIONS(600), 1,
      anon_sym_SQUOTE,
    STATE(3362), 1,
      sym__quoted_i_double,
    STATE(3363), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [293401] = 4,
    ACTIONS(9116), 1,
      sym__quoted_content_i_parenthesis,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9114), 3,
      anon_sym_RPAREN,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [293418] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(728), 1,
      anon_sym_DQUOTE,
    ACTIONS(730), 1,
      anon_sym_SQUOTE,
    STATE(4129), 1,
      sym__quoted_i_double,
    STATE(4130), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [293439] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_square,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_RBRACK,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [293456] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_angle,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_GT,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [293473] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9118), 1,
      anon_sym_RPAREN,
    ACTIONS(9120), 1,
      sym__quoted_content_parenthesis,
    STATE(6742), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293494] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9122), 1,
      anon_sym_DQUOTE,
    ACTIONS(9124), 1,
      sym__quoted_content_double,
    STATE(6190), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293515] = 6,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9126), 1,
      anon_sym_RPAREN,
    ACTIONS(9128), 1,
      sym__quoted_content_parenthesis,
    STATE(6688), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293536] = 6,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9130), 1,
      anon_sym_DQUOTE,
    ACTIONS(9132), 1,
      sym__quoted_content_double,
    STATE(6690), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293557] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9134), 1,
      anon_sym_SQUOTE,
    ACTIONS(9136), 1,
      sym__quoted_content_single,
    STATE(6692), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293578] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9138), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(9140), 1,
      sym__quoted_content_heredoc_single,
    STATE(6694), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293599] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9142), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(9144), 1,
      sym__quoted_content_heredoc_double,
    STATE(6696), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293620] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9146), 1,
      anon_sym_RBRACE,
    ACTIONS(9148), 1,
      sym__quoted_content_curly,
    STATE(6698), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293641] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9150), 1,
      anon_sym_RBRACK,
    ACTIONS(9152), 1,
      sym__quoted_content_square,
    STATE(6703), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293662] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9154), 1,
      anon_sym_GT,
    ACTIONS(9156), 1,
      sym__quoted_content_angle,
    STATE(6711), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293683] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9158), 1,
      anon_sym_PIPE,
    ACTIONS(9160), 1,
      sym__quoted_content_bar,
    STATE(6713), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293704] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9162), 1,
      anon_sym_SLASH,
    ACTIONS(9164), 1,
      sym__quoted_content_slash,
    STATE(6716), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293725] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9166), 1,
      anon_sym_SQUOTE,
    ACTIONS(9168), 1,
      sym__quoted_content_single,
    STATE(6212), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293746] = 6,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9170), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(9172), 1,
      sym__quoted_content_heredoc_single,
    STATE(6280), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293767] = 6,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9174), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(9176), 1,
      sym__quoted_content_heredoc_double,
    STATE(6289), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293788] = 6,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9178), 1,
      anon_sym_RBRACE,
    ACTIONS(9180), 1,
      sym__quoted_content_curly,
    STATE(6292), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293809] = 6,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9182), 1,
      anon_sym_RBRACK,
    ACTIONS(9184), 1,
      sym__quoted_content_square,
    STATE(6327), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293830] = 6,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9186), 1,
      anon_sym_GT,
    ACTIONS(9188), 1,
      sym__quoted_content_angle,
    STATE(6412), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293851] = 6,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9190), 1,
      anon_sym_PIPE,
    ACTIONS(9192), 1,
      sym__quoted_content_bar,
    STATE(6575), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293872] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9194), 1,
      anon_sym_SLASH,
    ACTIONS(9196), 1,
      sym__quoted_content_slash,
    STATE(6588), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293893] = 6,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9198), 1,
      anon_sym_SLASH,
    ACTIONS(9200), 1,
      sym__quoted_content_slash,
    STATE(6740), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [293914] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_heredoc_double,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [293931] = 4,
    ACTIONS(9204), 1,
      sym__quoted_content_i_curly,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9202), 3,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [293948] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_angle,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_GT,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [293965] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_curly,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [293982] = 4,
    ACTIONS(9208), 1,
      sym__quoted_content_i_single,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9206), 3,
      anon_sym_SQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [293999] = 6,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1167), 1,
      anon_sym_DQUOTE,
    ACTIONS(1169), 1,
      anon_sym_SQUOTE,
    STATE(4532), 1,
      sym__quoted_i_double,
    STATE(4533), 1,
      sym__quoted_i_single,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [294020] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9210), 1,
      anon_sym_COMMA,
    STATE(6061), 1,
      aux_sym__stab_clause_arguments_without_parentheses_repeat1,
    ACTIONS(9212), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [294039] = 4,
    ACTIONS(9216), 1,
      sym__quoted_content_i_heredoc_single,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9214), 3,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [294056] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_parenthesis,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_RPAREN,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [294073] = 4,
    ACTIONS(9220), 1,
      sym__quoted_content_i_heredoc_double,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9218), 3,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [294090] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_double,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_DQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [294107] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_heredoc_double,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [294124] = 4,
    ACTIONS(8431), 1,
      sym__quoted_content_i_square,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8429), 3,
      anon_sym_RBRACK,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [294141] = 4,
    ACTIONS(8539), 1,
      sym__quoted_content_i_single,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8537), 3,
      anon_sym_SQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [294158] = 6,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9222), 1,
      anon_sym_SQUOTE,
    ACTIONS(9224), 1,
      sym__quoted_content_single,
    STATE(6543), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294179] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9226), 1,
      anon_sym_GT,
    STATE(6527), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294197] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9228), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294215] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9230), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294233] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9232), 1,
      anon_sym_DQUOTE,
    STATE(6294), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294251] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9232), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294269] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9234), 1,
      anon_sym_RBRACK,
    STATE(6273), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294287] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9236), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6728), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294305] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9234), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294323] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9238), 1,
      anon_sym_GT,
    STATE(6274), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294341] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9238), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294359] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9240), 1,
      anon_sym_PIPE,
    STATE(6275), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294377] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9240), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294395] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9242), 1,
      anon_sym_SLASH,
    STATE(6276), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294413] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9242), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294431] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9244), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294449] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(7359), 3,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [294463] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9246), 1,
      anon_sym_SQUOTE,
    STATE(6295), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294481] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9248), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294499] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8210), 3,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [294513] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9250), 1,
      anon_sym_RPAREN,
    ACTIONS(9252), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [294529] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9254), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6535), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294547] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9254), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294565] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9256), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6699), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294583] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9256), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294601] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9258), 1,
      anon_sym_RBRACE,
    STATE(6701), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294619] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9258), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294637] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9246), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294655] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9260), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294673] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9262), 1,
      anon_sym_RPAREN,
    STATE(6637), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294691] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9262), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294709] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9264), 1,
      anon_sym_RPAREN,
    STATE(6281), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294727] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9264), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294745] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9266), 1,
      anon_sym_DQUOTE,
    STATE(6282), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294763] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9266), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294781] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9268), 1,
      anon_sym_SQUOTE,
    STATE(6283), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294799] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9268), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294817] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9270), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6284), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294835] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9270), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294853] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9272), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6285), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294871] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9272), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294889] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9274), 1,
      anon_sym_RPAREN,
    STATE(6252), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294907] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9274), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294925] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9276), 1,
      anon_sym_DQUOTE,
    STATE(6253), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294943] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9276), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294961] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9278), 1,
      anon_sym_SQUOTE,
    STATE(6254), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294979] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9278), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [294997] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9280), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6255), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295015] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9280), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295033] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9282), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6256), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295051] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9282), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295069] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9284), 1,
      anon_sym_RBRACE,
    STATE(6257), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295087] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9284), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295105] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9286), 1,
      anon_sym_RBRACK,
    STATE(6258), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295123] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9286), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295141] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9288), 1,
      anon_sym_GT,
    STATE(6259), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295159] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9288), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295177] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9290), 1,
      anon_sym_PIPE,
    STATE(6260), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295195] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9290), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295213] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9292), 1,
      anon_sym_SLASH,
    STATE(6261), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295231] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9292), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295249] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9294), 1,
      anon_sym_RBRACE,
    STATE(6286), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295267] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9294), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295285] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9296), 1,
      anon_sym_RBRACK,
    STATE(6287), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295303] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9296), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295321] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9298), 1,
      anon_sym_GT,
    STATE(6297), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295339] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9298), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295357] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9300), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295375] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9302), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295393] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9304), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295411] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9306), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295429] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9308), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295447] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9310), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295465] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9312), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295483] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9314), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295501] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9316), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295519] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9318), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295537] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9320), 1,
      anon_sym_PIPE,
    STATE(6298), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295555] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9320), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295573] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9322), 1,
      anon_sym_SLASH,
    STATE(6299), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295591] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9322), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295609] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9324), 1,
      anon_sym_DQUOTE,
    STATE(6638), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295627] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9326), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295645] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9328), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295663] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9330), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295681] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9332), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295699] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9334), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295717] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9336), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295735] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9338), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295753] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9340), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295771] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9342), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295789] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9344), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295807] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9324), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295825] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9346), 1,
      anon_sym_SQUOTE,
    STATE(6641), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295843] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9348), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6296), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295861] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9348), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295879] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9350), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295897] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9352), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295915] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9354), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295933] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9356), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295951] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9358), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295969] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9360), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [295987] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9362), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296005] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9364), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6300), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296023] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9364), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296041] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9366), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296059] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9368), 1,
      anon_sym_RBRACE,
    STATE(6301), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296077] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9368), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296095] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9370), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296113] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9372), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296131] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9374), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296149] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9376), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296167] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9378), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296185] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9380), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296203] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9382), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296221] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9384), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296239] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9386), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296257] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9388), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296275] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9390), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296293] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9392), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296311] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9394), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296329] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9396), 1,
      anon_sym_RPAREN,
    STATE(6328), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296347] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9396), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296365] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9398), 1,
      anon_sym_DQUOTE,
    STATE(6329), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296383] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9398), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296401] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9400), 1,
      anon_sym_SQUOTE,
    STATE(6330), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296419] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9400), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296437] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9402), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6331), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296455] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9402), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296473] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9404), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6332), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296491] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9404), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296509] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9406), 1,
      anon_sym_RBRACE,
    STATE(6333), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296527] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9406), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296545] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9408), 1,
      anon_sym_RBRACK,
    STATE(6334), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296563] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9408), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296581] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9410), 1,
      anon_sym_GT,
    STATE(6335), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296599] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9410), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296617] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9412), 1,
      anon_sym_PIPE,
    STATE(6336), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296635] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9412), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296653] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9414), 1,
      anon_sym_SLASH,
    STATE(6337), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296671] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9414), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296689] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9416), 1,
      anon_sym_RBRACK,
    STATE(6302), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296707] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9416), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296725] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9418), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296743] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9420), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296761] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9422), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296779] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9424), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296797] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9426), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296815] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9428), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296833] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9430), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296851] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9432), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296869] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9434), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296887] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9436), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296905] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9438), 1,
      anon_sym_RPAREN,
    STATE(6413), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296923] = 5,
    ACTIONS(9440), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(9442), 1,
      sym_escape_sequence,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296941] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9346), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296959] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9445), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6643), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296977] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9445), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [296995] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9447), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6646), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297013] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9449), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297031] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9447), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297049] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9451), 1,
      anon_sym_RBRACE,
    STATE(6648), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297067] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9438), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297085] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9453), 1,
      anon_sym_DQUOTE,
    STATE(6414), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297103] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9453), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297121] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9455), 1,
      anon_sym_SQUOTE,
    STATE(6415), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297139] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9455), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297157] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9457), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6416), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297175] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9457), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297193] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9459), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6417), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297211] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9459), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297229] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9461), 1,
      anon_sym_RBRACE,
    STATE(6418), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297247] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9461), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297265] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9463), 1,
      anon_sym_RBRACK,
    STATE(6419), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297283] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9465), 1,
      anon_sym_RPAREN,
    STATE(6380), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297301] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9465), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297319] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9467), 1,
      anon_sym_DQUOTE,
    STATE(6381), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297337] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9467), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297355] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9469), 1,
      anon_sym_SQUOTE,
    STATE(6382), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297373] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9469), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297391] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9471), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6383), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297409] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9471), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297427] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9473), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6384), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297445] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9473), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297463] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9475), 1,
      anon_sym_RBRACE,
    STATE(6385), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297481] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9475), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297499] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9477), 1,
      anon_sym_RBRACK,
    STATE(6386), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297517] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9477), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297535] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9479), 1,
      anon_sym_GT,
    STATE(6387), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297553] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9479), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297571] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9481), 1,
      anon_sym_PIPE,
    STATE(6388), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297589] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9481), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297607] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9483), 1,
      anon_sym_SLASH,
    STATE(6389), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297625] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9483), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297643] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9463), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297661] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9485), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297679] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9487), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297697] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9489), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297715] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9491), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297733] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9493), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297751] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9495), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297769] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9497), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297787] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9499), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297805] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9501), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297823] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9503), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297841] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9505), 1,
      anon_sym_GT,
    STATE(6420), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297859] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9505), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297877] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9507), 1,
      anon_sym_PIPE,
    STATE(6482), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297895] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9507), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297913] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9509), 1,
      anon_sym_SLASH,
    STATE(6483), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297931] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9509), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297949] = 5,
    ACTIONS(9511), 1,
      anon_sym_RBRACE,
    ACTIONS(9513), 1,
      sym_escape_sequence,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297967] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9451), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [297985] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9516), 1,
      anon_sym_RBRACK,
    STATE(6649), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298003] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9516), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298021] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9518), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298039] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9520), 1,
      anon_sym_GT,
    STATE(6650), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298057] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3662), 1,
      anon_sym_RPAREN,
    ACTIONS(9522), 1,
      anon_sym_COMMA,
    STATE(5263), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [298075] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9520), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298093] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9524), 1,
      anon_sym_PIPE,
    STATE(6653), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298111] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9524), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298129] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9526), 1,
      anon_sym_SLASH,
    STATE(6657), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298147] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9526), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298165] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9528), 1,
      anon_sym_SQUOTE,
    STATE(6765), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298183] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(7726), 3,
      anon_sym_PIPE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [298197] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(7298), 3,
      anon_sym_RBRACK,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [298211] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9530), 1,
      anon_sym_GT,
    STATE(6303), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298229] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9530), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298247] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9532), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298265] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9534), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298283] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9536), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298301] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9538), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298319] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9540), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298337] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9542), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298355] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9544), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298373] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9546), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298391] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9548), 1,
      anon_sym_RPAREN,
    STATE(6487), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298409] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9550), 1,
      anon_sym_RPAREN,
    STATE(6448), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298427] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9550), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298445] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9552), 1,
      anon_sym_DQUOTE,
    STATE(6449), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298463] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9552), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298481] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9554), 1,
      anon_sym_SQUOTE,
    STATE(6450), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298499] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9554), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298517] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9556), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6451), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298535] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9556), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298553] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9558), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6452), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298571] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9558), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298589] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9560), 1,
      anon_sym_RBRACE,
    STATE(6453), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298607] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9560), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298625] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9562), 1,
      anon_sym_RBRACK,
    STATE(6454), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298643] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9562), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298661] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9564), 1,
      anon_sym_GT,
    STATE(6455), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298679] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9564), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298697] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9566), 1,
      anon_sym_PIPE,
    STATE(6456), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298715] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9566), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298733] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9568), 1,
      anon_sym_SLASH,
    STATE(6457), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298751] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9568), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298769] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9548), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298787] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9570), 1,
      anon_sym_DQUOTE,
    STATE(6488), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298805] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9570), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298823] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9572), 1,
      anon_sym_SQUOTE,
    STATE(6489), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298841] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9572), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298859] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9574), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6490), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298877] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9576), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298895] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9578), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298913] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9580), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298931] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9582), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298949] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9584), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298967] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9586), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [298985] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9588), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299003] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9590), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299021] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9592), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299039] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9594), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299057] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9574), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299075] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9596), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6491), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299093] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9596), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299111] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9598), 1,
      anon_sym_RBRACE,
    STATE(6492), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299129] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9598), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299147] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9600), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299165] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9602), 1,
      anon_sym_RBRACK,
    STATE(6493), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299183] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9602), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299201] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9604), 1,
      anon_sym_GT,
    STATE(6494), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299219] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9604), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299237] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9606), 1,
      anon_sym_PIPE,
    STATE(6514), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299255] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9606), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299273] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9608), 1,
      anon_sym_SLASH,
    STATE(6515), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299291] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9608), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299309] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9610), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299327] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9612), 1,
      anon_sym_RPAREN,
    STATE(6589), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299345] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9612), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299363] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9614), 1,
      anon_sym_DQUOTE,
    STATE(6591), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299381] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9614), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299399] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9616), 1,
      anon_sym_SQUOTE,
    STATE(6613), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299417] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9616), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299435] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(1281), 1,
      anon_sym_GT_GT,
    ACTIONS(9618), 1,
      anon_sym_COMMA,
    STATE(6704), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [299453] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9620), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6616), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299471] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9620), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299489] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9622), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299507] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9624), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299525] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9626), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6617), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299543] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9626), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299561] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9628), 1,
      anon_sym_RBRACE,
    STATE(6618), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299579] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9630), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299597] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9632), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299615] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9634), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299633] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9636), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299651] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9638), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299669] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9640), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299687] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9642), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299705] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9644), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299723] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9646), 1,
      anon_sym_RPAREN,
    STATE(6520), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299741] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9646), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299759] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9648), 1,
      anon_sym_DQUOTE,
    STATE(6521), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299777] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9648), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299795] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9650), 1,
      anon_sym_SQUOTE,
    STATE(6522), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299813] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9650), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299831] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9652), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6523), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299849] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9654), 1,
      anon_sym_RPAREN,
    STATE(6293), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299867] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9656), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6524), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299885] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9656), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299903] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9658), 1,
      anon_sym_RBRACE,
    STATE(6525), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299921] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9658), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299939] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9660), 1,
      anon_sym_RBRACK,
    STATE(6526), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299957] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9660), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299975] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9226), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [299993] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9662), 1,
      anon_sym_PIPE,
    STATE(6528), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300011] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9662), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300029] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9664), 1,
      anon_sym_SLASH,
    STATE(6529), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300047] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9664), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300065] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9666), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300083] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9668), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300101] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9628), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300119] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9670), 1,
      anon_sym_RBRACK,
    STATE(6619), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300137] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9670), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300155] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9672), 1,
      anon_sym_GT,
    STATE(6620), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300173] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9674), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300191] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9676), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300209] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9678), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300227] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9680), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300245] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9682), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300263] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9684), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300281] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9686), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300299] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9688), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300317] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9690), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300335] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9692), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300353] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9672), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300371] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9694), 1,
      anon_sym_PIPE,
    STATE(6621), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300389] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9694), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300407] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9696), 1,
      anon_sym_SLASH,
    STATE(6622), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300425] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9696), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300443] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9698), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300461] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9700), 1,
      anon_sym_RPAREN,
    STATE(6576), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300479] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9700), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300497] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9702), 1,
      anon_sym_DQUOTE,
    STATE(6577), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300515] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(7490), 3,
      anon_sym_RPAREN,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [300529] = 4,
    ACTIONS(9706), 1,
      sym__quoted_content_single,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9704), 2,
      anon_sym_SQUOTE,
      sym_escape_sequence,
  [300545] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9702), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300563] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9708), 1,
      anon_sym_SQUOTE,
    STATE(6578), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300581] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9708), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300599] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9710), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6579), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300617] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9710), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300635] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9712), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6580), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300653] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9712), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300671] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9714), 1,
      anon_sym_RBRACE,
    STATE(6581), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300689] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9714), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300707] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9716), 1,
      anon_sym_RBRACK,
    STATE(6582), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300725] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9716), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300743] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9718), 1,
      anon_sym_GT,
    STATE(6583), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300761] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9720), 1,
      anon_sym_RPAREN,
    STATE(6664), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300779] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9718), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300797] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9722), 1,
      anon_sym_PIPE,
    STATE(6584), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300815] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9722), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300833] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9724), 1,
      anon_sym_SLASH,
    STATE(6585), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300851] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9724), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300869] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9726), 1,
      anon_sym_RBRACK,
    STATE(6706), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300887] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9726), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300905] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9728), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300923] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9528), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300941] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8236), 3,
      anon_sym_SLASH,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [300955] = 5,
    ACTIONS(9730), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(9732), 1,
      sym_escape_sequence,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300973] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9735), 1,
      anon_sym_PIPE,
    STATE(6304), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [300991] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9737), 1,
      anon_sym_RPAREN,
    STATE(6267), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301009] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9737), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301027] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9739), 1,
      anon_sym_GT,
    STATE(6707), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301045] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9739), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301063] = 5,
    ACTIONS(9741), 1,
      anon_sym_RBRACK,
    ACTIONS(9743), 1,
      sym_escape_sequence,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301081] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9746), 1,
      anon_sym_PIPE,
    STATE(6708), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301099] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9746), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301117] = 4,
    ACTIONS(9750), 1,
      sym__quoted_content_parenthesis,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9748), 2,
      anon_sym_RPAREN,
      sym_escape_sequence,
  [301133] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9720), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301151] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9735), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301169] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9752), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301187] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9754), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301205] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9756), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301223] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9758), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301241] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9760), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301259] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9762), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301277] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9764), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301295] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9766), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301313] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9768), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301331] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9770), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301349] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9260), 1,
      anon_sym_SLASH,
    STATE(6709), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301367] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9772), 1,
      anon_sym_SLASH,
    STATE(6305), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301385] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9772), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301403] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9774), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301421] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(1225), 3,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
  [301435] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9776), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301453] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9778), 1,
      anon_sym_RPAREN,
    STATE(6624), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301471] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9778), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301489] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9780), 1,
      anon_sym_DQUOTE,
    STATE(6625), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301507] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9780), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301525] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9782), 1,
      anon_sym_SQUOTE,
    STATE(6626), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301543] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9782), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301561] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9784), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6627), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301579] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9784), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301597] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9786), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6628), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301615] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9786), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301633] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9788), 1,
      anon_sym_RBRACE,
    STATE(6629), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301651] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9788), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301669] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9790), 1,
      anon_sym_RBRACK,
    STATE(6630), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301687] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9790), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301705] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9792), 1,
      anon_sym_GT,
    STATE(6631), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301723] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9792), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301741] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9794), 1,
      anon_sym_PIPE,
    STATE(6632), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301759] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9794), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301777] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9796), 1,
      anon_sym_SLASH,
    STATE(6633), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301795] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9796), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301813] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9798), 1,
      anon_sym_DQUOTE,
    STATE(6268), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301831] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9800), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301849] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9798), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301867] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9366), 1,
      anon_sym_DQUOTE,
    STATE(6731), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301885] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9802), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301903] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9804), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301921] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9806), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301939] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9808), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301957] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9810), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301975] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9812), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [301993] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9814), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302011] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9816), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302029] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9818), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302047] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9820), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302065] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9822), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302083] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9824), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302101] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9826), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302119] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9828), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302137] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9830), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302155] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9832), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302173] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9834), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302191] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9836), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302209] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9838), 1,
      anon_sym_SQUOTE,
    STATE(6269), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302227] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9838), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302245] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9840), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6270), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302263] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9842), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302281] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9844), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302299] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9840), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302317] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9846), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6271), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302335] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9848), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302353] = 5,
    ACTIONS(9850), 1,
      anon_sym_GT,
    ACTIONS(9852), 1,
      sym_escape_sequence,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302371] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9855), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302389] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9846), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302407] = 4,
    ACTIONS(9859), 1,
      sym__quoted_content_double,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9857), 2,
      anon_sym_DQUOTE,
      sym_escape_sequence,
  [302423] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9861), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302441] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9863), 1,
      anon_sym_RBRACE,
    STATE(6272), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302459] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9865), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302477] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9867), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302495] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9869), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302513] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9871), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302531] = 5,
    ACTIONS(9873), 1,
      anon_sym_PIPE,
    ACTIONS(9875), 1,
      sym_escape_sequence,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302549] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9878), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302567] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9880), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302585] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9863), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302603] = 5,
    ACTIONS(9882), 1,
      anon_sym_SLASH,
    ACTIONS(9884), 1,
      sym_escape_sequence,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302621] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9887), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302639] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9889), 1,
      anon_sym_RPAREN,
    ACTIONS(9891), 1,
      anon_sym_COMMA,
    STATE(6661), 1,
      aux_sym__stab_clause_arguments_with_parentheses_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [302657] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(9893), 3,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
  [302671] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(7579), 3,
      anon_sym_GT,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [302685] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(5088), 1,
      anon_sym_RPAREN,
    ACTIONS(9895), 1,
      anon_sym_COMMA,
    STATE(6661), 1,
      aux_sym__stab_clause_arguments_with_parentheses_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [302703] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8407), 3,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [302717] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9898), 1,
      anon_sym_RPAREN,
    ACTIONS(9900), 1,
      anon_sym_COMMA,
    STATE(5384), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [302735] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9902), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302753] = 5,
    ACTIONS(9904), 1,
      anon_sym_SQUOTE,
    ACTIONS(9906), 1,
      sym_escape_sequence,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302771] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9909), 1,
      anon_sym_RPAREN,
    STATE(6743), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302789] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9909), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302807] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9911), 1,
      anon_sym_DQUOTE,
    STATE(6744), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302825] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9911), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302843] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9913), 1,
      anon_sym_SQUOTE,
    STATE(6751), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302861] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9913), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302879] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6758), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302897] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302915] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9917), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6759), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302933] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9917), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302951] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9919), 1,
      anon_sym_RBRACE,
    STATE(6763), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302969] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9919), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [302987] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9921), 1,
      anon_sym_RBRACK,
    STATE(6768), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303005] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9921), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303023] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9923), 1,
      anon_sym_GT,
    STATE(6769), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303041] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9923), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303059] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9925), 1,
      anon_sym_PIPE,
    STATE(6187), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303077] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9925), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303095] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9927), 1,
      anon_sym_SLASH,
    STATE(6188), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303113] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9927), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303131] = 5,
    ACTIONS(9929), 1,
      anon_sym_RPAREN,
    ACTIONS(9931), 1,
      sym_escape_sequence,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303149] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9934), 1,
      anon_sym_RPAREN,
    STATE(6745), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303167] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9934), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303185] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9936), 1,
      anon_sym_DQUOTE,
    STATE(6746), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303203] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(9936), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303221] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9938), 1,
      anon_sym_SQUOTE,
    STATE(6747), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303239] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9938), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303257] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9940), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6748), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303275] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9940), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303293] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9942), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6749), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303311] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9942), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303329] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9944), 1,
      anon_sym_RBRACE,
    STATE(6750), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303347] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9944), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303365] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9946), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303383] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3656), 1,
      anon_sym_RPAREN,
    ACTIONS(9522), 1,
      anon_sym_COMMA,
    STATE(6402), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [303401] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9948), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303419] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9950), 1,
      anon_sym_RBRACK,
    STATE(6752), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303437] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9950), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303455] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3645), 1,
      anon_sym_GT_GT,
    ACTIONS(9952), 1,
      anon_sym_COMMA,
    STATE(6704), 1,
      aux_sym__items_with_trailing_separator_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [303473] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9955), 1,
      anon_sym_GT,
    STATE(6753), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303491] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9957), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303509] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9959), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303527] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9961), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303545] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9963), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303563] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9965), 1,
      anon_sym_RPAREN,
    ACTIONS(2971), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [303579] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9955), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303597] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9967), 1,
      anon_sym_PIPE,
    STATE(6754), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303615] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(9967), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303633] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9969), 1,
      anon_sym_SLASH,
    STATE(6757), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303651] = 4,
    ACTIONS(9973), 1,
      sym__quoted_content_heredoc_single,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9971), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      sym_escape_sequence,
  [303667] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(9969), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303685] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9236), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303703] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9975), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6200), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303721] = 4,
    ACTIONS(9979), 1,
      sym__quoted_content_heredoc_double,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9977), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      sym_escape_sequence,
  [303737] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(9975), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303755] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9981), 1,
      anon_sym_RBRACE,
    STATE(6344), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303773] = 4,
    ACTIONS(9985), 1,
      sym__quoted_content_curly,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9983), 2,
      anon_sym_RBRACE,
      sym_escape_sequence,
  [303789] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(9981), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303807] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9987), 1,
      anon_sym_RBRACK,
    STATE(6561), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303825] = 4,
    ACTIONS(9991), 1,
      sym__quoted_content_square,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9989), 2,
      anon_sym_RBRACK,
      sym_escape_sequence,
  [303841] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(9987), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303859] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9993), 1,
      anon_sym_GT,
    STATE(6623), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303877] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9995), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303895] = 4,
    ACTIONS(9999), 1,
      sym__quoted_content_angle,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9997), 2,
      anon_sym_GT,
      sym_escape_sequence,
  [303911] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_GT_GT,
    ACTIONS(10001), 1,
      anon_sym_COMMA,
    STATE(6730), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [303929] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(10004), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303947] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(10006), 1,
      anon_sym_RPAREN,
    STATE(6400), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303965] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(9993), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [303983] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(10008), 1,
      anon_sym_PIPE,
    STATE(6651), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304001] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(10006), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304019] = 4,
    ACTIONS(10012), 1,
      sym__quoted_content_bar,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(10010), 2,
      anon_sym_PIPE,
      sym_escape_sequence,
  [304035] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(10008), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304053] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(10014), 1,
      anon_sym_SLASH,
    STATE(6654), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304071] = 4,
    ACTIONS(10018), 1,
      sym__quoted_content_slash,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(10016), 2,
      anon_sym_SLASH,
      sym_escape_sequence,
  [304087] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(10014), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304105] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(9248), 1,
      anon_sym_SQUOTE,
    STATE(6472), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304123] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(9654), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304141] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(10020), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304159] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(10022), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304177] = 5,
    ACTIONS(8435), 1,
      sym_escape_sequence,
    ACTIONS(10024), 1,
      anon_sym_RPAREN,
    STATE(6686), 1,
      aux_sym__quoted_parenthesis_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304195] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(10026), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304213] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(10028), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304231] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(10030), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304249] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(10032), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304267] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(10034), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304285] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(10036), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304303] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(10038), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304321] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(10040), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304339] = 5,
    ACTIONS(8499), 1,
      sym_escape_sequence,
    ACTIONS(10042), 1,
      anon_sym_PIPE,
    STATE(6652), 1,
      aux_sym__quoted_bar_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304357] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(7904), 1,
      anon_sym_GT_GT,
    ACTIONS(10044), 1,
      anon_sym_COMMA,
    STATE(6756), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304375] = 5,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(6039), 1,
      anon_sym_GT_GT,
    ACTIONS(10046), 1,
      anon_sym_COMMA,
    STATE(6730), 1,
      aux_sym_keywords_repeat1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304393] = 5,
    ACTIONS(8505), 1,
      sym_escape_sequence,
    ACTIONS(10048), 1,
      anon_sym_SLASH,
    STATE(6656), 1,
      aux_sym__quoted_slash_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304411] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(10050), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304429] = 5,
    ACTIONS(8475), 1,
      sym_escape_sequence,
    ACTIONS(10052), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(6339), 1,
      aux_sym__quoted_heredoc_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304447] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(10054), 1,
      anon_sym_DQUOTE,
    STATE(6463), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304465] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(8417), 3,
      anon_sym_DQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [304479] = 5,
    ACTIONS(10056), 1,
      anon_sym_DQUOTE,
    ACTIONS(10058), 1,
      sym_escape_sequence,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304497] = 5,
    ACTIONS(8481), 1,
      sym_escape_sequence,
    ACTIONS(10061), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      aux_sym__quoted_curly_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304515] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
    ACTIONS(5018), 3,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
  [304529] = 5,
    ACTIONS(8451), 1,
      sym_escape_sequence,
    ACTIONS(10063), 1,
      anon_sym_SQUOTE,
    STATE(6665), 1,
      aux_sym__quoted_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304547] = 5,
    ACTIONS(8445), 1,
      sym_escape_sequence,
    ACTIONS(10054), 1,
      anon_sym_DQUOTE,
    STATE(6762), 1,
      aux_sym__quoted_double_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304565] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(7686), 3,
      anon_sym_SQUOTE,
      anon_sym_POUND_LBRACE,
      sym_escape_sequence,
  [304579] = 5,
    ACTIONS(8487), 1,
      sym_escape_sequence,
    ACTIONS(10065), 1,
      anon_sym_RBRACK,
    STATE(6570), 1,
      aux_sym__quoted_square_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304597] = 5,
    ACTIONS(8493), 1,
      sym_escape_sequence,
    ACTIONS(10067), 1,
      anon_sym_GT,
    STATE(6642), 1,
      aux_sym__quoted_angle_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304615] = 5,
    ACTIONS(8469), 1,
      sym_escape_sequence,
    ACTIONS(9652), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(6564), 1,
      aux_sym__quoted_heredoc_single_repeat1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
  [304633] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    STATE(4387), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304648] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9850), 2,
      anon_sym_GT,
      sym_escape_sequence,
  [304661] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    STATE(4491), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304676] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    STATE(1756), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304691] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9440), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      sym_escape_sequence,
  [304704] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10069), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304717] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4118), 1,
      anon_sym_LPAREN,
    STATE(2633), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304732] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9873), 2,
      anon_sym_PIPE,
      sym_escape_sequence,
  [304745] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10071), 1,
      anon_sym_LPAREN,
    STATE(3549), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304760] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    STATE(2045), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304775] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9730), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      sym_escape_sequence,
  [304788] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9882), 2,
      anon_sym_SLASH,
      sym_escape_sequence,
  [304801] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    STATE(4165), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304816] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    STATE(3153), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304831] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    STATE(4492), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304846] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10073), 1,
      anon_sym_LPAREN,
    STATE(2750), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304861] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10075), 1,
      anon_sym_LPAREN,
    STATE(4135), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304876] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2971), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304889] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    STATE(3091), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304904] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    STATE(4156), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304919] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3033), 1,
      anon_sym_LPAREN,
    STATE(1104), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304934] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10077), 1,
      anon_sym_LPAREN,
    STATE(1354), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304949] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    STATE(4388), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304964] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    STATE(4168), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304979] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    STATE(4487), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [304994] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    STATE(4389), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305009] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    STATE(3687), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305024] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3931), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305037] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9929), 2,
      anon_sym_RPAREN,
      sym_escape_sequence,
  [305050] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10079), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305063] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    STATE(4390), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305078] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10081), 1,
      aux_sym_sigil_token1,
    ACTIONS(10083), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305093] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    STATE(1616), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305108] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    STATE(1843), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305123] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    STATE(1844), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305138] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    STATE(1845), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305153] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9741), 2,
      anon_sym_RBRACK,
      sym_escape_sequence,
  [305166] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    STATE(4395), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305181] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    STATE(3685), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305196] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    STATE(4386), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305211] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10085), 1,
      anon_sym_when,
    ACTIONS(10087), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305226] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10089), 1,
      aux_sym_sigil_token1,
    ACTIONS(10091), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305241] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    STATE(2851), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305256] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    STATE(2070), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305271] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10087), 1,
      anon_sym_DASH_GT,
    ACTIONS(10093), 1,
      anon_sym_when,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305286] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3149), 1,
      anon_sym_LPAREN,
    STATE(1333), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305301] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_LPAREN,
    STATE(2715), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305316] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    STATE(3537), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305331] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    STATE(1618), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305346] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9904), 2,
      anon_sym_SQUOTE,
      sym_escape_sequence,
  [305359] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    STATE(1847), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305374] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    STATE(1451), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305389] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    STATE(1755), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305404] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(291), 1,
      anon_sym_do,
    STATE(1848), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305419] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10095), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305432] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    STATE(3150), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305447] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    STATE(1700), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305462] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(9511), 2,
      anon_sym_RBRACE,
      sym_escape_sequence,
  [305475] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    STATE(4382), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305490] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_do,
    STATE(4298), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305505] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    STATE(4486), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305520] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10097), 1,
      aux_sym_sigil_token1,
    ACTIONS(10099), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305535] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    STATE(4516), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305550] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    STATE(3151), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305565] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(9252), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305578] = 3,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(5), 2,
      aux_sym__terminator_token1,
      sym_comment,
    ACTIONS(10056), 2,
      anon_sym_DQUOTE,
      sym_escape_sequence,
  [305591] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10101), 1,
      aux_sym_sigil_token1,
    ACTIONS(10103), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305606] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    STATE(2847), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305621] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10105), 1,
      aux_sym_sigil_token1,
    ACTIONS(10107), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305636] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    STATE(1448), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305651] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    STATE(1750), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305666] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10109), 1,
      aux_sym_sigil_token1,
    ACTIONS(10111), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305681] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10113), 1,
      aux_sym_sigil_token1,
    ACTIONS(10115), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305696] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    STATE(4126), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305711] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    STATE(3152), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305726] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10117), 1,
      aux_sym_sigil_token1,
    ACTIONS(10119), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305741] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10121), 1,
      aux_sym_sigil_token1,
    ACTIONS(10123), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305756] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10125), 1,
      aux_sym_sigil_token1,
    ACTIONS(10127), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305771] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10129), 2,
      anon_sym_when,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305784] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10131), 1,
      aux_sym__terminator_token1,
    ACTIONS(3), 2,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
    ACTIONS(8403), 2,
      anon_sym_SEMI,
      anon_sym_end,
  [305799] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10133), 1,
      aux_sym_sigil_token1,
    ACTIONS(10135), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305814] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10137), 1,
      aux_sym_sigil_token1,
    ACTIONS(10139), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305829] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10141), 1,
      anon_sym_LPAREN,
    STATE(3589), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305844] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10143), 1,
      aux_sym_sigil_token1,
    ACTIONS(10145), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305859] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10147), 1,
      aux_sym_sigil_token1,
    ACTIONS(10149), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305874] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10151), 1,
      aux_sym_sigil_token1,
    ACTIONS(10153), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305889] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10155), 1,
      aux_sym_sigil_token1,
    ACTIONS(10157), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305904] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10159), 1,
      aux_sym_sigil_token1,
    ACTIONS(10161), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305919] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10163), 1,
      aux_sym_sigil_token1,
    ACTIONS(10165), 1,
      aux_sym_sigil_token2,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305934] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    STATE(4154), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305949] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    STATE(4169), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305964] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_do,
    STATE(3149), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305979] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    STATE(1751), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [305994] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(527), 1,
      anon_sym_do,
    STATE(3545), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306009] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10167), 1,
      anon_sym_LPAREN,
    STATE(1581), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306024] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(4342), 1,
      anon_sym_LPAREN,
    STATE(3262), 1,
      sym__call_arguments_with_parentheses,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306039] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(295), 1,
      anon_sym_do,
    STATE(1752), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306054] = 4,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(642), 1,
      anon_sym_do,
    STATE(4501), 1,
      sym_do_block,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306069] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3127), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306081] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10169), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306093] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10171), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306105] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10173), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306117] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10175), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306129] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10177), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306141] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10179), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306153] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10181), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306165] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10183), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306177] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10183), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306189] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10185), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306201] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10187), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306213] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10189), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306225] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10191), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306237] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10193), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306249] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10195), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306261] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10197), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306273] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10199), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306285] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10201), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306297] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10203), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306309] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10205), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306321] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10207), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306333] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(3079), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306345] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10209), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306357] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10211), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306369] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10213), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306381] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10215), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306393] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10217), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306405] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10219), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306417] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10221), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306429] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10223), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306441] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10225), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306453] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10227), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306465] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10229), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306477] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10231), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306489] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10233), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306501] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10235), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306513] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10237), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306525] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10239), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306537] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10241), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306549] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10243), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306561] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10245), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306573] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10247), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306585] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10249), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306597] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10251), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306609] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10253), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306621] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10255), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306633] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10257), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306645] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10259), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306657] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10261), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306669] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2709), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306681] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10263), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306693] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10265), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306705] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10267), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306717] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10269), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306729] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2665), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306741] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10271), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306753] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10273), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306765] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10275), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306777] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10277), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306789] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10279), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306801] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10281), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306813] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10283), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306825] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10285), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306837] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10287), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306849] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10289), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306861] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10291), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306873] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10293), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306885] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10295), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306897] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10297), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306909] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10299), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306921] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10301), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306933] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10303), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306945] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10305), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306957] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10307), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306969] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10309), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306981] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10311), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [306993] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10313), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307005] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10315), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307017] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10317), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307029] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10319), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307041] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10321), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307053] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10323), 1,
      ts_builtin_sym_end,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307065] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10325), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307077] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10327), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307089] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10329), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307101] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10331), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307113] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10333), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307125] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10335), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307137] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10337), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307149] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10339), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307161] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10341), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307173] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2673), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307185] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10343), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307197] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10345), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307209] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10347), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307221] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10349), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307233] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10351), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307245] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10353), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307257] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2669), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307269] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10355), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307281] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10357), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307293] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10359), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307305] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10361), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307317] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2687), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307329] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10363), 1,
      anon_sym_SLASH,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307341] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10365), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307353] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10367), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307365] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10369), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307377] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10371), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307389] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2705), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307401] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10373), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307413] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10375), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307425] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10377), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307437] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10379), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307449] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10381), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307461] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10383), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307473] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10385), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307485] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10387), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307497] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10389), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307509] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10391), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307521] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10393), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307533] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(2701), 1,
      aux_sym_quoted_keyword_token1,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307545] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10395), 1,
      sym_integer,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307557] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10397), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307569] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10399), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307581] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10401), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307593] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10403), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307605] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10405), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307617] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10407), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307629] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10409), 1,
      anon_sym_GT_GT,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
  [307641] = 3,
    ACTIONS(5), 1,
      sym_comment,
    ACTIONS(10411), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 3,
      sym__newline_before_binary_operator,
      sym__newline_before_comment,
      aux_sym__terminator_token1,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(1094)] = 0,
  [SMALL_STATE(1095)] = 129,
  [SMALL_STATE(1096)] = 202,
  [SMALL_STATE(1097)] = 277,
  [SMALL_STATE(1098)] = 356,
  [SMALL_STATE(1099)] = 429,
  [SMALL_STATE(1100)] = 558,
  [SMALL_STATE(1101)] = 633,
  [SMALL_STATE(1102)] = 762,
  [SMALL_STATE(1103)] = 835,
  [SMALL_STATE(1104)] = 908,
  [SMALL_STATE(1105)] = 981,
  [SMALL_STATE(1106)] = 1056,
  [SMALL_STATE(1107)] = 1185,
  [SMALL_STATE(1108)] = 1260,
  [SMALL_STATE(1109)] = 1339,
  [SMALL_STATE(1110)] = 1414,
  [SMALL_STATE(1111)] = 1487,
  [SMALL_STATE(1112)] = 1560,
  [SMALL_STATE(1113)] = 1633,
  [SMALL_STATE(1114)] = 1706,
  [SMALL_STATE(1115)] = 1779,
  [SMALL_STATE(1116)] = 1852,
  [SMALL_STATE(1117)] = 1981,
  [SMALL_STATE(1118)] = 2054,
  [SMALL_STATE(1119)] = 2127,
  [SMALL_STATE(1120)] = 2206,
  [SMALL_STATE(1121)] = 2335,
  [SMALL_STATE(1122)] = 2414,
  [SMALL_STATE(1123)] = 2487,
  [SMALL_STATE(1124)] = 2566,
  [SMALL_STATE(1125)] = 2641,
  [SMALL_STATE(1126)] = 2716,
  [SMALL_STATE(1127)] = 2845,
  [SMALL_STATE(1128)] = 2918,
  [SMALL_STATE(1129)] = 2997,
  [SMALL_STATE(1130)] = 3126,
  [SMALL_STATE(1131)] = 3205,
  [SMALL_STATE(1132)] = 3278,
  [SMALL_STATE(1133)] = 3351,
  [SMALL_STATE(1134)] = 3424,
  [SMALL_STATE(1135)] = 3501,
  [SMALL_STATE(1136)] = 3576,
  [SMALL_STATE(1137)] = 3653,
  [SMALL_STATE(1138)] = 3726,
  [SMALL_STATE(1139)] = 3801,
  [SMALL_STATE(1140)] = 3878,
  [SMALL_STATE(1141)] = 3951,
  [SMALL_STATE(1142)] = 4024,
  [SMALL_STATE(1143)] = 4096,
  [SMALL_STATE(1144)] = 4168,
  [SMALL_STATE(1145)] = 4240,
  [SMALL_STATE(1146)] = 4312,
  [SMALL_STATE(1147)] = 4384,
  [SMALL_STATE(1148)] = 4456,
  [SMALL_STATE(1149)] = 4528,
  [SMALL_STATE(1150)] = 4600,
  [SMALL_STATE(1151)] = 4672,
  [SMALL_STATE(1152)] = 4744,
  [SMALL_STATE(1153)] = 4816,
  [SMALL_STATE(1154)] = 4888,
  [SMALL_STATE(1155)] = 4960,
  [SMALL_STATE(1156)] = 5032,
  [SMALL_STATE(1157)] = 5104,
  [SMALL_STATE(1158)] = 5176,
  [SMALL_STATE(1159)] = 5248,
  [SMALL_STATE(1160)] = 5320,
  [SMALL_STATE(1161)] = 5392,
  [SMALL_STATE(1162)] = 5464,
  [SMALL_STATE(1163)] = 5536,
  [SMALL_STATE(1164)] = 5608,
  [SMALL_STATE(1165)] = 5680,
  [SMALL_STATE(1166)] = 5752,
  [SMALL_STATE(1167)] = 5824,
  [SMALL_STATE(1168)] = 5896,
  [SMALL_STATE(1169)] = 5968,
  [SMALL_STATE(1170)] = 6040,
  [SMALL_STATE(1171)] = 6112,
  [SMALL_STATE(1172)] = 6184,
  [SMALL_STATE(1173)] = 6256,
  [SMALL_STATE(1174)] = 6328,
  [SMALL_STATE(1175)] = 6400,
  [SMALL_STATE(1176)] = 6472,
  [SMALL_STATE(1177)] = 6544,
  [SMALL_STATE(1178)] = 6616,
  [SMALL_STATE(1179)] = 6688,
  [SMALL_STATE(1180)] = 6760,
  [SMALL_STATE(1181)] = 6832,
  [SMALL_STATE(1182)] = 6904,
  [SMALL_STATE(1183)] = 6976,
  [SMALL_STATE(1184)] = 7048,
  [SMALL_STATE(1185)] = 7120,
  [SMALL_STATE(1186)] = 7192,
  [SMALL_STATE(1187)] = 7264,
  [SMALL_STATE(1188)] = 7336,
  [SMALL_STATE(1189)] = 7408,
  [SMALL_STATE(1190)] = 7480,
  [SMALL_STATE(1191)] = 7552,
  [SMALL_STATE(1192)] = 7624,
  [SMALL_STATE(1193)] = 7696,
  [SMALL_STATE(1194)] = 7772,
  [SMALL_STATE(1195)] = 7848,
  [SMALL_STATE(1196)] = 7920,
  [SMALL_STATE(1197)] = 7992,
  [SMALL_STATE(1198)] = 8064,
  [SMALL_STATE(1199)] = 8136,
  [SMALL_STATE(1200)] = 8208,
  [SMALL_STATE(1201)] = 8280,
  [SMALL_STATE(1202)] = 8352,
  [SMALL_STATE(1203)] = 8424,
  [SMALL_STATE(1204)] = 8496,
  [SMALL_STATE(1205)] = 8568,
  [SMALL_STATE(1206)] = 8640,
  [SMALL_STATE(1207)] = 8712,
  [SMALL_STATE(1208)] = 8784,
  [SMALL_STATE(1209)] = 8856,
  [SMALL_STATE(1210)] = 8928,
  [SMALL_STATE(1211)] = 9000,
  [SMALL_STATE(1212)] = 9072,
  [SMALL_STATE(1213)] = 9144,
  [SMALL_STATE(1214)] = 9216,
  [SMALL_STATE(1215)] = 9288,
  [SMALL_STATE(1216)] = 9360,
  [SMALL_STATE(1217)] = 9432,
  [SMALL_STATE(1218)] = 9504,
  [SMALL_STATE(1219)] = 9576,
  [SMALL_STATE(1220)] = 9654,
  [SMALL_STATE(1221)] = 9732,
  [SMALL_STATE(1222)] = 9810,
  [SMALL_STATE(1223)] = 9882,
  [SMALL_STATE(1224)] = 9954,
  [SMALL_STATE(1225)] = 10026,
  [SMALL_STATE(1226)] = 10098,
  [SMALL_STATE(1227)] = 10170,
  [SMALL_STATE(1228)] = 10242,
  [SMALL_STATE(1229)] = 10320,
  [SMALL_STATE(1230)] = 10398,
  [SMALL_STATE(1231)] = 10470,
  [SMALL_STATE(1232)] = 10542,
  [SMALL_STATE(1233)] = 10614,
  [SMALL_STATE(1234)] = 10686,
  [SMALL_STATE(1235)] = 10758,
  [SMALL_STATE(1236)] = 10830,
  [SMALL_STATE(1237)] = 10902,
  [SMALL_STATE(1238)] = 10974,
  [SMALL_STATE(1239)] = 11046,
  [SMALL_STATE(1240)] = 11118,
  [SMALL_STATE(1241)] = 11190,
  [SMALL_STATE(1242)] = 11262,
  [SMALL_STATE(1243)] = 11334,
  [SMALL_STATE(1244)] = 11406,
  [SMALL_STATE(1245)] = 11478,
  [SMALL_STATE(1246)] = 11550,
  [SMALL_STATE(1247)] = 11622,
  [SMALL_STATE(1248)] = 11694,
  [SMALL_STATE(1249)] = 11766,
  [SMALL_STATE(1250)] = 11838,
  [SMALL_STATE(1251)] = 11956,
  [SMALL_STATE(1252)] = 12028,
  [SMALL_STATE(1253)] = 12100,
  [SMALL_STATE(1254)] = 12174,
  [SMALL_STATE(1255)] = 12246,
  [SMALL_STATE(1256)] = 12320,
  [SMALL_STATE(1257)] = 12392,
  [SMALL_STATE(1258)] = 12464,
  [SMALL_STATE(1259)] = 12536,
  [SMALL_STATE(1260)] = 12608,
  [SMALL_STATE(1261)] = 12680,
  [SMALL_STATE(1262)] = 12752,
  [SMALL_STATE(1263)] = 12824,
  [SMALL_STATE(1264)] = 12896,
  [SMALL_STATE(1265)] = 12968,
  [SMALL_STATE(1266)] = 13040,
  [SMALL_STATE(1267)] = 13112,
  [SMALL_STATE(1268)] = 13184,
  [SMALL_STATE(1269)] = 13256,
  [SMALL_STATE(1270)] = 13328,
  [SMALL_STATE(1271)] = 13400,
  [SMALL_STATE(1272)] = 13472,
  [SMALL_STATE(1273)] = 13544,
  [SMALL_STATE(1274)] = 13616,
  [SMALL_STATE(1275)] = 13688,
  [SMALL_STATE(1276)] = 13760,
  [SMALL_STATE(1277)] = 13832,
  [SMALL_STATE(1278)] = 13904,
  [SMALL_STATE(1279)] = 13976,
  [SMALL_STATE(1280)] = 14048,
  [SMALL_STATE(1281)] = 14120,
  [SMALL_STATE(1282)] = 14192,
  [SMALL_STATE(1283)] = 14264,
  [SMALL_STATE(1284)] = 14336,
  [SMALL_STATE(1285)] = 14408,
  [SMALL_STATE(1286)] = 14480,
  [SMALL_STATE(1287)] = 14552,
  [SMALL_STATE(1288)] = 14624,
  [SMALL_STATE(1289)] = 14696,
  [SMALL_STATE(1290)] = 14768,
  [SMALL_STATE(1291)] = 14840,
  [SMALL_STATE(1292)] = 14912,
  [SMALL_STATE(1293)] = 14984,
  [SMALL_STATE(1294)] = 15056,
  [SMALL_STATE(1295)] = 15128,
  [SMALL_STATE(1296)] = 15200,
  [SMALL_STATE(1297)] = 15272,
  [SMALL_STATE(1298)] = 15344,
  [SMALL_STATE(1299)] = 15416,
  [SMALL_STATE(1300)] = 15488,
  [SMALL_STATE(1301)] = 15564,
  [SMALL_STATE(1302)] = 15636,
  [SMALL_STATE(1303)] = 15708,
  [SMALL_STATE(1304)] = 15780,
  [SMALL_STATE(1305)] = 15852,
  [SMALL_STATE(1306)] = 15924,
  [SMALL_STATE(1307)] = 15996,
  [SMALL_STATE(1308)] = 16068,
  [SMALL_STATE(1309)] = 16140,
  [SMALL_STATE(1310)] = 16212,
  [SMALL_STATE(1311)] = 16284,
  [SMALL_STATE(1312)] = 16356,
  [SMALL_STATE(1313)] = 16430,
  [SMALL_STATE(1314)] = 16504,
  [SMALL_STATE(1315)] = 16578,
  [SMALL_STATE(1316)] = 16652,
  [SMALL_STATE(1317)] = 16726,
  [SMALL_STATE(1318)] = 16800,
  [SMALL_STATE(1319)] = 16874,
  [SMALL_STATE(1320)] = 16948,
  [SMALL_STATE(1321)] = 17022,
  [SMALL_STATE(1322)] = 17096,
  [SMALL_STATE(1323)] = 17170,
  [SMALL_STATE(1324)] = 17244,
  [SMALL_STATE(1325)] = 17318,
  [SMALL_STATE(1326)] = 17392,
  [SMALL_STATE(1327)] = 17466,
  [SMALL_STATE(1328)] = 17540,
  [SMALL_STATE(1329)] = 17614,
  [SMALL_STATE(1330)] = 17688,
  [SMALL_STATE(1331)] = 17762,
  [SMALL_STATE(1332)] = 17836,
  [SMALL_STATE(1333)] = 17910,
  [SMALL_STATE(1334)] = 17982,
  [SMALL_STATE(1335)] = 18056,
  [SMALL_STATE(1336)] = 18130,
  [SMALL_STATE(1337)] = 18204,
  [SMALL_STATE(1338)] = 18278,
  [SMALL_STATE(1339)] = 18350,
  [SMALL_STATE(1340)] = 18422,
  [SMALL_STATE(1341)] = 18494,
  [SMALL_STATE(1342)] = 18566,
  [SMALL_STATE(1343)] = 18638,
  [SMALL_STATE(1344)] = 18710,
  [SMALL_STATE(1345)] = 18782,
  [SMALL_STATE(1346)] = 18858,
  [SMALL_STATE(1347)] = 18934,
  [SMALL_STATE(1348)] = 19006,
  [SMALL_STATE(1349)] = 19078,
  [SMALL_STATE(1350)] = 19154,
  [SMALL_STATE(1351)] = 19230,
  [SMALL_STATE(1352)] = 19306,
  [SMALL_STATE(1353)] = 19378,
  [SMALL_STATE(1354)] = 19449,
  [SMALL_STATE(1355)] = 19520,
  [SMALL_STATE(1356)] = 19591,
  [SMALL_STATE(1357)] = 19662,
  [SMALL_STATE(1358)] = 19733,
  [SMALL_STATE(1359)] = 19804,
  [SMALL_STATE(1360)] = 19875,
  [SMALL_STATE(1361)] = 19946,
  [SMALL_STATE(1362)] = 20017,
  [SMALL_STATE(1363)] = 20088,
  [SMALL_STATE(1364)] = 20159,
  [SMALL_STATE(1365)] = 20230,
  [SMALL_STATE(1366)] = 20301,
  [SMALL_STATE(1367)] = 20372,
  [SMALL_STATE(1368)] = 20443,
  [SMALL_STATE(1369)] = 20514,
  [SMALL_STATE(1370)] = 20585,
  [SMALL_STATE(1371)] = 20656,
  [SMALL_STATE(1372)] = 20727,
  [SMALL_STATE(1373)] = 20798,
  [SMALL_STATE(1374)] = 20869,
  [SMALL_STATE(1375)] = 20940,
  [SMALL_STATE(1376)] = 21011,
  [SMALL_STATE(1377)] = 21082,
  [SMALL_STATE(1378)] = 21153,
  [SMALL_STATE(1379)] = 21224,
  [SMALL_STATE(1380)] = 21295,
  [SMALL_STATE(1381)] = 21408,
  [SMALL_STATE(1382)] = 21521,
  [SMALL_STATE(1383)] = 21592,
  [SMALL_STATE(1384)] = 21705,
  [SMALL_STATE(1385)] = 21776,
  [SMALL_STATE(1386)] = 21847,
  [SMALL_STATE(1387)] = 21918,
  [SMALL_STATE(1388)] = 21989,
  [SMALL_STATE(1389)] = 22060,
  [SMALL_STATE(1390)] = 22131,
  [SMALL_STATE(1391)] = 22206,
  [SMALL_STATE(1392)] = 22277,
  [SMALL_STATE(1393)] = 22348,
  [SMALL_STATE(1394)] = 22419,
  [SMALL_STATE(1395)] = 22490,
  [SMALL_STATE(1396)] = 22561,
  [SMALL_STATE(1397)] = 22632,
  [SMALL_STATE(1398)] = 22703,
  [SMALL_STATE(1399)] = 22774,
  [SMALL_STATE(1400)] = 22845,
  [SMALL_STATE(1401)] = 22916,
  [SMALL_STATE(1402)] = 22987,
  [SMALL_STATE(1403)] = 23058,
  [SMALL_STATE(1404)] = 23129,
  [SMALL_STATE(1405)] = 23200,
  [SMALL_STATE(1406)] = 23271,
  [SMALL_STATE(1407)] = 23342,
  [SMALL_STATE(1408)] = 23413,
  [SMALL_STATE(1409)] = 23484,
  [SMALL_STATE(1410)] = 23555,
  [SMALL_STATE(1411)] = 23626,
  [SMALL_STATE(1412)] = 23697,
  [SMALL_STATE(1413)] = 23768,
  [SMALL_STATE(1414)] = 23839,
  [SMALL_STATE(1415)] = 23910,
  [SMALL_STATE(1416)] = 23981,
  [SMALL_STATE(1417)] = 24052,
  [SMALL_STATE(1418)] = 24125,
  [SMALL_STATE(1419)] = 24198,
  [SMALL_STATE(1420)] = 24269,
  [SMALL_STATE(1421)] = 24340,
  [SMALL_STATE(1422)] = 24413,
  [SMALL_STATE(1423)] = 24486,
  [SMALL_STATE(1424)] = 24603,
  [SMALL_STATE(1425)] = 24674,
  [SMALL_STATE(1426)] = 24745,
  [SMALL_STATE(1427)] = 24816,
  [SMALL_STATE(1428)] = 24887,
  [SMALL_STATE(1429)] = 24958,
  [SMALL_STATE(1430)] = 25029,
  [SMALL_STATE(1431)] = 25100,
  [SMALL_STATE(1432)] = 25171,
  [SMALL_STATE(1433)] = 25242,
  [SMALL_STATE(1434)] = 25313,
  [SMALL_STATE(1435)] = 25386,
  [SMALL_STATE(1436)] = 25457,
  [SMALL_STATE(1437)] = 25530,
  [SMALL_STATE(1438)] = 25601,
  [SMALL_STATE(1439)] = 25676,
  [SMALL_STATE(1440)] = 25747,
  [SMALL_STATE(1441)] = 25818,
  [SMALL_STATE(1442)] = 25889,
  [SMALL_STATE(1443)] = 25960,
  [SMALL_STATE(1444)] = 26031,
  [SMALL_STATE(1445)] = 26102,
  [SMALL_STATE(1446)] = 26173,
  [SMALL_STATE(1447)] = 26244,
  [SMALL_STATE(1448)] = 26317,
  [SMALL_STATE(1449)] = 26388,
  [SMALL_STATE(1450)] = 26459,
  [SMALL_STATE(1451)] = 26532,
  [SMALL_STATE(1452)] = 26603,
  [SMALL_STATE(1453)] = 26676,
  [SMALL_STATE(1454)] = 26747,
  [SMALL_STATE(1455)] = 26818,
  [SMALL_STATE(1456)] = 26889,
  [SMALL_STATE(1457)] = 26960,
  [SMALL_STATE(1458)] = 27031,
  [SMALL_STATE(1459)] = 27102,
  [SMALL_STATE(1460)] = 27173,
  [SMALL_STATE(1461)] = 27244,
  [SMALL_STATE(1462)] = 27315,
  [SMALL_STATE(1463)] = 27386,
  [SMALL_STATE(1464)] = 27457,
  [SMALL_STATE(1465)] = 27528,
  [SMALL_STATE(1466)] = 27599,
  [SMALL_STATE(1467)] = 27670,
  [SMALL_STATE(1468)] = 27741,
  [SMALL_STATE(1469)] = 27812,
  [SMALL_STATE(1470)] = 27883,
  [SMALL_STATE(1471)] = 27954,
  [SMALL_STATE(1472)] = 28025,
  [SMALL_STATE(1473)] = 28096,
  [SMALL_STATE(1474)] = 28167,
  [SMALL_STATE(1475)] = 28238,
  [SMALL_STATE(1476)] = 28311,
  [SMALL_STATE(1477)] = 28382,
  [SMALL_STATE(1478)] = 28453,
  [SMALL_STATE(1479)] = 28526,
  [SMALL_STATE(1480)] = 28601,
  [SMALL_STATE(1481)] = 28672,
  [SMALL_STATE(1482)] = 28747,
  [SMALL_STATE(1483)] = 28818,
  [SMALL_STATE(1484)] = 28889,
  [SMALL_STATE(1485)] = 28960,
  [SMALL_STATE(1486)] = 29033,
  [SMALL_STATE(1487)] = 29106,
  [SMALL_STATE(1488)] = 29179,
  [SMALL_STATE(1489)] = 29252,
  [SMALL_STATE(1490)] = 29325,
  [SMALL_STATE(1491)] = 29398,
  [SMALL_STATE(1492)] = 29471,
  [SMALL_STATE(1493)] = 29544,
  [SMALL_STATE(1494)] = 29615,
  [SMALL_STATE(1495)] = 29688,
  [SMALL_STATE(1496)] = 29759,
  [SMALL_STATE(1497)] = 29830,
  [SMALL_STATE(1498)] = 29901,
  [SMALL_STATE(1499)] = 29972,
  [SMALL_STATE(1500)] = 30043,
  [SMALL_STATE(1501)] = 30114,
  [SMALL_STATE(1502)] = 30185,
  [SMALL_STATE(1503)] = 30256,
  [SMALL_STATE(1504)] = 30327,
  [SMALL_STATE(1505)] = 30398,
  [SMALL_STATE(1506)] = 30469,
  [SMALL_STATE(1507)] = 30540,
  [SMALL_STATE(1508)] = 30611,
  [SMALL_STATE(1509)] = 30682,
  [SMALL_STATE(1510)] = 30753,
  [SMALL_STATE(1511)] = 30824,
  [SMALL_STATE(1512)] = 30899,
  [SMALL_STATE(1513)] = 30972,
  [SMALL_STATE(1514)] = 31043,
  [SMALL_STATE(1515)] = 31130,
  [SMALL_STATE(1516)] = 31223,
  [SMALL_STATE(1517)] = 31330,
  [SMALL_STATE(1518)] = 31407,
  [SMALL_STATE(1519)] = 31490,
  [SMALL_STATE(1520)] = 31569,
  [SMALL_STATE(1521)] = 31680,
  [SMALL_STATE(1522)] = 31751,
  [SMALL_STATE(1523)] = 31862,
  [SMALL_STATE(1524)] = 31933,
  [SMALL_STATE(1525)] = 32004,
  [SMALL_STATE(1526)] = 32075,
  [SMALL_STATE(1527)] = 32184,
  [SMALL_STATE(1528)] = 32289,
  [SMALL_STATE(1529)] = 32392,
  [SMALL_STATE(1530)] = 32463,
  [SMALL_STATE(1531)] = 32534,
  [SMALL_STATE(1532)] = 32605,
  [SMALL_STATE(1533)] = 32704,
  [SMALL_STATE(1534)] = 32801,
  [SMALL_STATE(1535)] = 32896,
  [SMALL_STATE(1536)] = 32987,
  [SMALL_STATE(1537)] = 33058,
  [SMALL_STATE(1538)] = 33129,
  [SMALL_STATE(1539)] = 33200,
  [SMALL_STATE(1540)] = 33285,
  [SMALL_STATE(1541)] = 33356,
  [SMALL_STATE(1542)] = 33441,
  [SMALL_STATE(1543)] = 33512,
  [SMALL_STATE(1544)] = 33583,
  [SMALL_STATE(1545)] = 33654,
  [SMALL_STATE(1546)] = 33725,
  [SMALL_STATE(1547)] = 33796,
  [SMALL_STATE(1548)] = 33867,
  [SMALL_STATE(1549)] = 33938,
  [SMALL_STATE(1550)] = 34009,
  [SMALL_STATE(1551)] = 34080,
  [SMALL_STATE(1552)] = 34151,
  [SMALL_STATE(1553)] = 34224,
  [SMALL_STATE(1554)] = 34295,
  [SMALL_STATE(1555)] = 34370,
  [SMALL_STATE(1556)] = 34445,
  [SMALL_STATE(1557)] = 34520,
  [SMALL_STATE(1558)] = 34591,
  [SMALL_STATE(1559)] = 34666,
  [SMALL_STATE(1560)] = 34741,
  [SMALL_STATE(1561)] = 34812,
  [SMALL_STATE(1562)] = 34883,
  [SMALL_STATE(1563)] = 34954,
  [SMALL_STATE(1564)] = 35025,
  [SMALL_STATE(1565)] = 35096,
  [SMALL_STATE(1566)] = 35167,
  [SMALL_STATE(1567)] = 35238,
  [SMALL_STATE(1568)] = 35309,
  [SMALL_STATE(1569)] = 35380,
  [SMALL_STATE(1570)] = 35451,
  [SMALL_STATE(1571)] = 35522,
  [SMALL_STATE(1572)] = 35593,
  [SMALL_STATE(1573)] = 35664,
  [SMALL_STATE(1574)] = 35739,
  [SMALL_STATE(1575)] = 35810,
  [SMALL_STATE(1576)] = 35880,
  [SMALL_STATE(1577)] = 35950,
  [SMALL_STATE(1578)] = 36020,
  [SMALL_STATE(1579)] = 36090,
  [SMALL_STATE(1580)] = 36160,
  [SMALL_STATE(1581)] = 36230,
  [SMALL_STATE(1582)] = 36300,
  [SMALL_STATE(1583)] = 36370,
  [SMALL_STATE(1584)] = 36440,
  [SMALL_STATE(1585)] = 36510,
  [SMALL_STATE(1586)] = 36580,
  [SMALL_STATE(1587)] = 36650,
  [SMALL_STATE(1588)] = 36720,
  [SMALL_STATE(1589)] = 36790,
  [SMALL_STATE(1590)] = 36860,
  [SMALL_STATE(1591)] = 36930,
  [SMALL_STATE(1592)] = 37000,
  [SMALL_STATE(1593)] = 37070,
  [SMALL_STATE(1594)] = 37140,
  [SMALL_STATE(1595)] = 37210,
  [SMALL_STATE(1596)] = 37296,
  [SMALL_STATE(1597)] = 37388,
  [SMALL_STATE(1598)] = 37494,
  [SMALL_STATE(1599)] = 37570,
  [SMALL_STATE(1600)] = 37652,
  [SMALL_STATE(1601)] = 37730,
  [SMALL_STATE(1602)] = 37840,
  [SMALL_STATE(1603)] = 37950,
  [SMALL_STATE(1604)] = 38058,
  [SMALL_STATE(1605)] = 38162,
  [SMALL_STATE(1606)] = 38264,
  [SMALL_STATE(1607)] = 38362,
  [SMALL_STATE(1608)] = 38458,
  [SMALL_STATE(1609)] = 38552,
  [SMALL_STATE(1610)] = 38642,
  [SMALL_STATE(1611)] = 38726,
  [SMALL_STATE(1612)] = 38810,
  [SMALL_STATE(1613)] = 38880,
  [SMALL_STATE(1614)] = 38952,
  [SMALL_STATE(1615)] = 39024,
  [SMALL_STATE(1616)] = 39094,
  [SMALL_STATE(1617)] = 39164,
  [SMALL_STATE(1618)] = 39234,
  [SMALL_STATE(1619)] = 39304,
  [SMALL_STATE(1620)] = 39374,
  [SMALL_STATE(1621)] = 39444,
  [SMALL_STATE(1622)] = 39514,
  [SMALL_STATE(1623)] = 39584,
  [SMALL_STATE(1624)] = 39654,
  [SMALL_STATE(1625)] = 39724,
  [SMALL_STATE(1626)] = 39794,
  [SMALL_STATE(1627)] = 39864,
  [SMALL_STATE(1628)] = 39934,
  [SMALL_STATE(1629)] = 40004,
  [SMALL_STATE(1630)] = 40074,
  [SMALL_STATE(1631)] = 40144,
  [SMALL_STATE(1632)] = 40214,
  [SMALL_STATE(1633)] = 40284,
  [SMALL_STATE(1634)] = 40354,
  [SMALL_STATE(1635)] = 40424,
  [SMALL_STATE(1636)] = 40494,
  [SMALL_STATE(1637)] = 40564,
  [SMALL_STATE(1638)] = 40634,
  [SMALL_STATE(1639)] = 40704,
  [SMALL_STATE(1640)] = 40774,
  [SMALL_STATE(1641)] = 40902,
  [SMALL_STATE(1642)] = 40972,
  [SMALL_STATE(1643)] = 41042,
  [SMALL_STATE(1644)] = 41116,
  [SMALL_STATE(1645)] = 41186,
  [SMALL_STATE(1646)] = 41260,
  [SMALL_STATE(1647)] = 41330,
  [SMALL_STATE(1648)] = 41400,
  [SMALL_STATE(1649)] = 41470,
  [SMALL_STATE(1650)] = 41544,
  [SMALL_STATE(1651)] = 41614,
  [SMALL_STATE(1652)] = 41686,
  [SMALL_STATE(1653)] = 41756,
  [SMALL_STATE(1654)] = 41826,
  [SMALL_STATE(1655)] = 41896,
  [SMALL_STATE(1656)] = 41968,
  [SMALL_STATE(1657)] = 42040,
  [SMALL_STATE(1658)] = 42112,
  [SMALL_STATE(1659)] = 42184,
  [SMALL_STATE(1660)] = 42256,
  [SMALL_STATE(1661)] = 42328,
  [SMALL_STATE(1662)] = 42398,
  [SMALL_STATE(1663)] = 42470,
  [SMALL_STATE(1664)] = 42542,
  [SMALL_STATE(1665)] = 42614,
  [SMALL_STATE(1666)] = 42686,
  [SMALL_STATE(1667)] = 42758,
  [SMALL_STATE(1668)] = 42830,
  [SMALL_STATE(1669)] = 42902,
  [SMALL_STATE(1670)] = 42974,
  [SMALL_STATE(1671)] = 43046,
  [SMALL_STATE(1672)] = 43118,
  [SMALL_STATE(1673)] = 43190,
  [SMALL_STATE(1674)] = 43262,
  [SMALL_STATE(1675)] = 43334,
  [SMALL_STATE(1676)] = 43406,
  [SMALL_STATE(1677)] = 43476,
  [SMALL_STATE(1678)] = 43546,
  [SMALL_STATE(1679)] = 43616,
  [SMALL_STATE(1680)] = 43686,
  [SMALL_STATE(1681)] = 43756,
  [SMALL_STATE(1682)] = 43826,
  [SMALL_STATE(1683)] = 43896,
  [SMALL_STATE(1684)] = 44016,
  [SMALL_STATE(1685)] = 44086,
  [SMALL_STATE(1686)] = 44156,
  [SMALL_STATE(1687)] = 44226,
  [SMALL_STATE(1688)] = 44346,
  [SMALL_STATE(1689)] = 44458,
  [SMALL_STATE(1690)] = 44570,
  [SMALL_STATE(1691)] = 44682,
  [SMALL_STATE(1692)] = 44752,
  [SMALL_STATE(1693)] = 44822,
  [SMALL_STATE(1694)] = 44892,
  [SMALL_STATE(1695)] = 44962,
  [SMALL_STATE(1696)] = 45032,
  [SMALL_STATE(1697)] = 45102,
  [SMALL_STATE(1698)] = 45172,
  [SMALL_STATE(1699)] = 45242,
  [SMALL_STATE(1700)] = 45316,
  [SMALL_STATE(1701)] = 45386,
  [SMALL_STATE(1702)] = 45456,
  [SMALL_STATE(1703)] = 45526,
  [SMALL_STATE(1704)] = 45596,
  [SMALL_STATE(1705)] = 45666,
  [SMALL_STATE(1706)] = 45736,
  [SMALL_STATE(1707)] = 45806,
  [SMALL_STATE(1708)] = 45876,
  [SMALL_STATE(1709)] = 45946,
  [SMALL_STATE(1710)] = 46016,
  [SMALL_STATE(1711)] = 46086,
  [SMALL_STATE(1712)] = 46156,
  [SMALL_STATE(1713)] = 46226,
  [SMALL_STATE(1714)] = 46296,
  [SMALL_STATE(1715)] = 46366,
  [SMALL_STATE(1716)] = 46436,
  [SMALL_STATE(1717)] = 46506,
  [SMALL_STATE(1718)] = 46576,
  [SMALL_STATE(1719)] = 46646,
  [SMALL_STATE(1720)] = 46716,
  [SMALL_STATE(1721)] = 46786,
  [SMALL_STATE(1722)] = 46856,
  [SMALL_STATE(1723)] = 46926,
  [SMALL_STATE(1724)] = 46996,
  [SMALL_STATE(1725)] = 47066,
  [SMALL_STATE(1726)] = 47136,
  [SMALL_STATE(1727)] = 47206,
  [SMALL_STATE(1728)] = 47276,
  [SMALL_STATE(1729)] = 47346,
  [SMALL_STATE(1730)] = 47416,
  [SMALL_STATE(1731)] = 47486,
  [SMALL_STATE(1732)] = 47556,
  [SMALL_STATE(1733)] = 47626,
  [SMALL_STATE(1734)] = 47696,
  [SMALL_STATE(1735)] = 47766,
  [SMALL_STATE(1736)] = 47836,
  [SMALL_STATE(1737)] = 47906,
  [SMALL_STATE(1738)] = 47976,
  [SMALL_STATE(1739)] = 48046,
  [SMALL_STATE(1740)] = 48116,
  [SMALL_STATE(1741)] = 48186,
  [SMALL_STATE(1742)] = 48256,
  [SMALL_STATE(1743)] = 48326,
  [SMALL_STATE(1744)] = 48396,
  [SMALL_STATE(1745)] = 48466,
  [SMALL_STATE(1746)] = 48536,
  [SMALL_STATE(1747)] = 48606,
  [SMALL_STATE(1748)] = 48676,
  [SMALL_STATE(1749)] = 48746,
  [SMALL_STATE(1750)] = 48816,
  [SMALL_STATE(1751)] = 48886,
  [SMALL_STATE(1752)] = 48956,
  [SMALL_STATE(1753)] = 49026,
  [SMALL_STATE(1754)] = 49096,
  [SMALL_STATE(1755)] = 49166,
  [SMALL_STATE(1756)] = 49236,
  [SMALL_STATE(1757)] = 49306,
  [SMALL_STATE(1758)] = 49376,
  [SMALL_STATE(1759)] = 49446,
  [SMALL_STATE(1760)] = 49516,
  [SMALL_STATE(1761)] = 49586,
  [SMALL_STATE(1762)] = 49656,
  [SMALL_STATE(1763)] = 49726,
  [SMALL_STATE(1764)] = 49796,
  [SMALL_STATE(1765)] = 49866,
  [SMALL_STATE(1766)] = 49936,
  [SMALL_STATE(1767)] = 50006,
  [SMALL_STATE(1768)] = 50076,
  [SMALL_STATE(1769)] = 50146,
  [SMALL_STATE(1770)] = 50216,
  [SMALL_STATE(1771)] = 50286,
  [SMALL_STATE(1772)] = 50356,
  [SMALL_STATE(1773)] = 50426,
  [SMALL_STATE(1774)] = 50496,
  [SMALL_STATE(1775)] = 50566,
  [SMALL_STATE(1776)] = 50636,
  [SMALL_STATE(1777)] = 50706,
  [SMALL_STATE(1778)] = 50776,
  [SMALL_STATE(1779)] = 50846,
  [SMALL_STATE(1780)] = 50916,
  [SMALL_STATE(1781)] = 50986,
  [SMALL_STATE(1782)] = 51056,
  [SMALL_STATE(1783)] = 51126,
  [SMALL_STATE(1784)] = 51196,
  [SMALL_STATE(1785)] = 51266,
  [SMALL_STATE(1786)] = 51394,
  [SMALL_STATE(1787)] = 51464,
  [SMALL_STATE(1788)] = 51536,
  [SMALL_STATE(1789)] = 51608,
  [SMALL_STATE(1790)] = 51684,
  [SMALL_STATE(1791)] = 51760,
  [SMALL_STATE(1792)] = 51832,
  [SMALL_STATE(1793)] = 51902,
  [SMALL_STATE(1794)] = 51972,
  [SMALL_STATE(1795)] = 52042,
  [SMALL_STATE(1796)] = 52116,
  [SMALL_STATE(1797)] = 52190,
  [SMALL_STATE(1798)] = 52264,
  [SMALL_STATE(1799)] = 52338,
  [SMALL_STATE(1800)] = 52412,
  [SMALL_STATE(1801)] = 52482,
  [SMALL_STATE(1802)] = 52551,
  [SMALL_STATE(1803)] = 52620,
  [SMALL_STATE(1804)] = 52689,
  [SMALL_STATE(1805)] = 52758,
  [SMALL_STATE(1806)] = 52827,
  [SMALL_STATE(1807)] = 52896,
  [SMALL_STATE(1808)] = 52965,
  [SMALL_STATE(1809)] = 53034,
  [SMALL_STATE(1810)] = 53103,
  [SMALL_STATE(1811)] = 53172,
  [SMALL_STATE(1812)] = 53241,
  [SMALL_STATE(1813)] = 53310,
  [SMALL_STATE(1814)] = 53379,
  [SMALL_STATE(1815)] = 53448,
  [SMALL_STATE(1816)] = 53517,
  [SMALL_STATE(1817)] = 53586,
  [SMALL_STATE(1818)] = 53655,
  [SMALL_STATE(1819)] = 53724,
  [SMALL_STATE(1820)] = 53793,
  [SMALL_STATE(1821)] = 53862,
  [SMALL_STATE(1822)] = 53931,
  [SMALL_STATE(1823)] = 54000,
  [SMALL_STATE(1824)] = 54069,
  [SMALL_STATE(1825)] = 54138,
  [SMALL_STATE(1826)] = 54207,
  [SMALL_STATE(1827)] = 54276,
  [SMALL_STATE(1828)] = 54345,
  [SMALL_STATE(1829)] = 54414,
  [SMALL_STATE(1830)] = 54483,
  [SMALL_STATE(1831)] = 54552,
  [SMALL_STATE(1832)] = 54621,
  [SMALL_STATE(1833)] = 54690,
  [SMALL_STATE(1834)] = 54759,
  [SMALL_STATE(1835)] = 54828,
  [SMALL_STATE(1836)] = 54897,
  [SMALL_STATE(1837)] = 54966,
  [SMALL_STATE(1838)] = 55035,
  [SMALL_STATE(1839)] = 55104,
  [SMALL_STATE(1840)] = 55173,
  [SMALL_STATE(1841)] = 55242,
  [SMALL_STATE(1842)] = 55311,
  [SMALL_STATE(1843)] = 55380,
  [SMALL_STATE(1844)] = 55449,
  [SMALL_STATE(1845)] = 55518,
  [SMALL_STATE(1846)] = 55587,
  [SMALL_STATE(1847)] = 55656,
  [SMALL_STATE(1848)] = 55725,
  [SMALL_STATE(1849)] = 55794,
  [SMALL_STATE(1850)] = 55863,
  [SMALL_STATE(1851)] = 55932,
  [SMALL_STATE(1852)] = 56001,
  [SMALL_STATE(1853)] = 56070,
  [SMALL_STATE(1854)] = 56139,
  [SMALL_STATE(1855)] = 56208,
  [SMALL_STATE(1856)] = 56277,
  [SMALL_STATE(1857)] = 56346,
  [SMALL_STATE(1858)] = 56415,
  [SMALL_STATE(1859)] = 56484,
  [SMALL_STATE(1860)] = 56553,
  [SMALL_STATE(1861)] = 56622,
  [SMALL_STATE(1862)] = 56691,
  [SMALL_STATE(1863)] = 56760,
  [SMALL_STATE(1864)] = 56829,
  [SMALL_STATE(1865)] = 56898,
  [SMALL_STATE(1866)] = 56967,
  [SMALL_STATE(1867)] = 57036,
  [SMALL_STATE(1868)] = 57105,
  [SMALL_STATE(1869)] = 57174,
  [SMALL_STATE(1870)] = 57243,
  [SMALL_STATE(1871)] = 57312,
  [SMALL_STATE(1872)] = 57381,
  [SMALL_STATE(1873)] = 57450,
  [SMALL_STATE(1874)] = 57519,
  [SMALL_STATE(1875)] = 57588,
  [SMALL_STATE(1876)] = 57657,
  [SMALL_STATE(1877)] = 57726,
  [SMALL_STATE(1878)] = 57795,
  [SMALL_STATE(1879)] = 57864,
  [SMALL_STATE(1880)] = 57933,
  [SMALL_STATE(1881)] = 58002,
  [SMALL_STATE(1882)] = 58071,
  [SMALL_STATE(1883)] = 58140,
  [SMALL_STATE(1884)] = 58211,
  [SMALL_STATE(1885)] = 58282,
  [SMALL_STATE(1886)] = 58351,
  [SMALL_STATE(1887)] = 58420,
  [SMALL_STATE(1888)] = 58489,
  [SMALL_STATE(1889)] = 58558,
  [SMALL_STATE(1890)] = 58627,
  [SMALL_STATE(1891)] = 58696,
  [SMALL_STATE(1892)] = 58765,
  [SMALL_STATE(1893)] = 58834,
  [SMALL_STATE(1894)] = 58903,
  [SMALL_STATE(1895)] = 58972,
  [SMALL_STATE(1896)] = 59041,
  [SMALL_STATE(1897)] = 59110,
  [SMALL_STATE(1898)] = 59179,
  [SMALL_STATE(1899)] = 59248,
  [SMALL_STATE(1900)] = 59317,
  [SMALL_STATE(1901)] = 59386,
  [SMALL_STATE(1902)] = 59455,
  [SMALL_STATE(1903)] = 59524,
  [SMALL_STATE(1904)] = 59593,
  [SMALL_STATE(1905)] = 59662,
  [SMALL_STATE(1906)] = 59787,
  [SMALL_STATE(1907)] = 59856,
  [SMALL_STATE(1908)] = 59925,
  [SMALL_STATE(1909)] = 59994,
  [SMALL_STATE(1910)] = 60063,
  [SMALL_STATE(1911)] = 60132,
  [SMALL_STATE(1912)] = 60257,
  [SMALL_STATE(1913)] = 60326,
  [SMALL_STATE(1914)] = 60451,
  [SMALL_STATE(1915)] = 60520,
  [SMALL_STATE(1916)] = 60589,
  [SMALL_STATE(1917)] = 60658,
  [SMALL_STATE(1918)] = 60769,
  [SMALL_STATE(1919)] = 60838,
  [SMALL_STATE(1920)] = 60907,
  [SMALL_STATE(1921)] = 60976,
  [SMALL_STATE(1922)] = 61045,
  [SMALL_STATE(1923)] = 61114,
  [SMALL_STATE(1924)] = 61183,
  [SMALL_STATE(1925)] = 61252,
  [SMALL_STATE(1926)] = 61321,
  [SMALL_STATE(1927)] = 61390,
  [SMALL_STATE(1928)] = 61459,
  [SMALL_STATE(1929)] = 61528,
  [SMALL_STATE(1930)] = 61597,
  [SMALL_STATE(1931)] = 61668,
  [SMALL_STATE(1932)] = 61737,
  [SMALL_STATE(1933)] = 61808,
  [SMALL_STATE(1934)] = 61879,
  [SMALL_STATE(1935)] = 61950,
  [SMALL_STATE(1936)] = 62021,
  [SMALL_STATE(1937)] = 62090,
  [SMALL_STATE(1938)] = 62159,
  [SMALL_STATE(1939)] = 62228,
  [SMALL_STATE(1940)] = 62297,
  [SMALL_STATE(1941)] = 62366,
  [SMALL_STATE(1942)] = 62435,
  [SMALL_STATE(1943)] = 62504,
  [SMALL_STATE(1944)] = 62573,
  [SMALL_STATE(1945)] = 62642,
  [SMALL_STATE(1946)] = 62711,
  [SMALL_STATE(1947)] = 62796,
  [SMALL_STATE(1948)] = 62887,
  [SMALL_STATE(1949)] = 62992,
  [SMALL_STATE(1950)] = 63067,
  [SMALL_STATE(1951)] = 63192,
  [SMALL_STATE(1952)] = 63273,
  [SMALL_STATE(1953)] = 63350,
  [SMALL_STATE(1954)] = 63459,
  [SMALL_STATE(1955)] = 63584,
  [SMALL_STATE(1956)] = 63709,
  [SMALL_STATE(1957)] = 63818,
  [SMALL_STATE(1958)] = 63925,
  [SMALL_STATE(1959)] = 64028,
  [SMALL_STATE(1960)] = 64129,
  [SMALL_STATE(1961)] = 64226,
  [SMALL_STATE(1962)] = 64321,
  [SMALL_STATE(1963)] = 64414,
  [SMALL_STATE(1964)] = 64503,
  [SMALL_STATE(1965)] = 64586,
  [SMALL_STATE(1966)] = 64669,
  [SMALL_STATE(1967)] = 64738,
  [SMALL_STATE(1968)] = 64807,
  [SMALL_STATE(1969)] = 64876,
  [SMALL_STATE(1970)] = 64945,
  [SMALL_STATE(1971)] = 65014,
  [SMALL_STATE(1972)] = 65139,
  [SMALL_STATE(1973)] = 65208,
  [SMALL_STATE(1974)] = 65277,
  [SMALL_STATE(1975)] = 65346,
  [SMALL_STATE(1976)] = 65415,
  [SMALL_STATE(1977)] = 65486,
  [SMALL_STATE(1978)] = 65611,
  [SMALL_STATE(1979)] = 65680,
  [SMALL_STATE(1980)] = 65749,
  [SMALL_STATE(1981)] = 65874,
  [SMALL_STATE(1982)] = 65943,
  [SMALL_STATE(1983)] = 66012,
  [SMALL_STATE(1984)] = 66081,
  [SMALL_STATE(1985)] = 66150,
  [SMALL_STATE(1986)] = 66219,
  [SMALL_STATE(1987)] = 66288,
  [SMALL_STATE(1988)] = 66357,
  [SMALL_STATE(1989)] = 66426,
  [SMALL_STATE(1990)] = 66495,
  [SMALL_STATE(1991)] = 66564,
  [SMALL_STATE(1992)] = 66633,
  [SMALL_STATE(1993)] = 66702,
  [SMALL_STATE(1994)] = 66771,
  [SMALL_STATE(1995)] = 66896,
  [SMALL_STATE(1996)] = 66965,
  [SMALL_STATE(1997)] = 67034,
  [SMALL_STATE(1998)] = 67103,
  [SMALL_STATE(1999)] = 67172,
  [SMALL_STATE(2000)] = 67241,
  [SMALL_STATE(2001)] = 67310,
  [SMALL_STATE(2002)] = 67379,
  [SMALL_STATE(2003)] = 67448,
  [SMALL_STATE(2004)] = 67573,
  [SMALL_STATE(2005)] = 67642,
  [SMALL_STATE(2006)] = 67711,
  [SMALL_STATE(2007)] = 67836,
  [SMALL_STATE(2008)] = 67905,
  [SMALL_STATE(2009)] = 67974,
  [SMALL_STATE(2010)] = 68043,
  [SMALL_STATE(2011)] = 68112,
  [SMALL_STATE(2012)] = 68181,
  [SMALL_STATE(2013)] = 68250,
  [SMALL_STATE(2014)] = 68319,
  [SMALL_STATE(2015)] = 68388,
  [SMALL_STATE(2016)] = 68457,
  [SMALL_STATE(2017)] = 68528,
  [SMALL_STATE(2018)] = 68653,
  [SMALL_STATE(2019)] = 68778,
  [SMALL_STATE(2020)] = 68847,
  [SMALL_STATE(2021)] = 68916,
  [SMALL_STATE(2022)] = 68985,
  [SMALL_STATE(2023)] = 69054,
  [SMALL_STATE(2024)] = 69179,
  [SMALL_STATE(2025)] = 69248,
  [SMALL_STATE(2026)] = 69319,
  [SMALL_STATE(2027)] = 69390,
  [SMALL_STATE(2028)] = 69459,
  [SMALL_STATE(2029)] = 69528,
  [SMALL_STATE(2030)] = 69597,
  [SMALL_STATE(2031)] = 69666,
  [SMALL_STATE(2032)] = 69735,
  [SMALL_STATE(2033)] = 69860,
  [SMALL_STATE(2034)] = 69929,
  [SMALL_STATE(2035)] = 69998,
  [SMALL_STATE(2036)] = 70123,
  [SMALL_STATE(2037)] = 70192,
  [SMALL_STATE(2038)] = 70261,
  [SMALL_STATE(2039)] = 70330,
  [SMALL_STATE(2040)] = 70401,
  [SMALL_STATE(2041)] = 70472,
  [SMALL_STATE(2042)] = 70541,
  [SMALL_STATE(2043)] = 70610,
  [SMALL_STATE(2044)] = 70679,
  [SMALL_STATE(2045)] = 70748,
  [SMALL_STATE(2046)] = 70817,
  [SMALL_STATE(2047)] = 70886,
  [SMALL_STATE(2048)] = 71011,
  [SMALL_STATE(2049)] = 71080,
  [SMALL_STATE(2050)] = 71205,
  [SMALL_STATE(2051)] = 71330,
  [SMALL_STATE(2052)] = 71399,
  [SMALL_STATE(2053)] = 71468,
  [SMALL_STATE(2054)] = 71537,
  [SMALL_STATE(2055)] = 71606,
  [SMALL_STATE(2056)] = 71675,
  [SMALL_STATE(2057)] = 71744,
  [SMALL_STATE(2058)] = 71813,
  [SMALL_STATE(2059)] = 71882,
  [SMALL_STATE(2060)] = 72007,
  [SMALL_STATE(2061)] = 72076,
  [SMALL_STATE(2062)] = 72145,
  [SMALL_STATE(2063)] = 72214,
  [SMALL_STATE(2064)] = 72285,
  [SMALL_STATE(2065)] = 72410,
  [SMALL_STATE(2066)] = 72481,
  [SMALL_STATE(2067)] = 72552,
  [SMALL_STATE(2068)] = 72677,
  [SMALL_STATE(2069)] = 72748,
  [SMALL_STATE(2070)] = 72819,
  [SMALL_STATE(2071)] = 72888,
  [SMALL_STATE(2072)] = 72963,
  [SMALL_STATE(2073)] = 73038,
  [SMALL_STATE(2074)] = 73113,
  [SMALL_STATE(2075)] = 73188,
  [SMALL_STATE(2076)] = 73263,
  [SMALL_STATE(2077)] = 73374,
  [SMALL_STATE(2078)] = 73499,
  [SMALL_STATE(2079)] = 73624,
  [SMALL_STATE(2080)] = 73749,
  [SMALL_STATE(2081)] = 73874,
  [SMALL_STATE(2082)] = 73943,
  [SMALL_STATE(2083)] = 74068,
  [SMALL_STATE(2084)] = 74137,
  [SMALL_STATE(2085)] = 74262,
  [SMALL_STATE(2086)] = 74387,
  [SMALL_STATE(2087)] = 74458,
  [SMALL_STATE(2088)] = 74529,
  [SMALL_STATE(2089)] = 74600,
  [SMALL_STATE(2090)] = 74671,
  [SMALL_STATE(2091)] = 74742,
  [SMALL_STATE(2092)] = 74813,
  [SMALL_STATE(2093)] = 74884,
  [SMALL_STATE(2094)] = 74953,
  [SMALL_STATE(2095)] = 75024,
  [SMALL_STATE(2096)] = 75095,
  [SMALL_STATE(2097)] = 75166,
  [SMALL_STATE(2098)] = 75291,
  [SMALL_STATE(2099)] = 75362,
  [SMALL_STATE(2100)] = 75433,
  [SMALL_STATE(2101)] = 75558,
  [SMALL_STATE(2102)] = 75683,
  [SMALL_STATE(2103)] = 75754,
  [SMALL_STATE(2104)] = 75825,
  [SMALL_STATE(2105)] = 75896,
  [SMALL_STATE(2106)] = 76021,
  [SMALL_STATE(2107)] = 76092,
  [SMALL_STATE(2108)] = 76163,
  [SMALL_STATE(2109)] = 76234,
  [SMALL_STATE(2110)] = 76303,
  [SMALL_STATE(2111)] = 76428,
  [SMALL_STATE(2112)] = 76553,
  [SMALL_STATE(2113)] = 76624,
  [SMALL_STATE(2114)] = 76695,
  [SMALL_STATE(2115)] = 76764,
  [SMALL_STATE(2116)] = 76833,
  [SMALL_STATE(2117)] = 76902,
  [SMALL_STATE(2118)] = 76975,
  [SMALL_STATE(2119)] = 77048,
  [SMALL_STATE(2120)] = 77121,
  [SMALL_STATE(2121)] = 77190,
  [SMALL_STATE(2122)] = 77259,
  [SMALL_STATE(2123)] = 77328,
  [SMALL_STATE(2124)] = 77401,
  [SMALL_STATE(2125)] = 77474,
  [SMALL_STATE(2126)] = 77547,
  [SMALL_STATE(2127)] = 77616,
  [SMALL_STATE(2128)] = 77685,
  [SMALL_STATE(2129)] = 77754,
  [SMALL_STATE(2130)] = 77823,
  [SMALL_STATE(2131)] = 77892,
  [SMALL_STATE(2132)] = 77963,
  [SMALL_STATE(2133)] = 78034,
  [SMALL_STATE(2134)] = 78105,
  [SMALL_STATE(2135)] = 78176,
  [SMALL_STATE(2136)] = 78247,
  [SMALL_STATE(2137)] = 78318,
  [SMALL_STATE(2138)] = 78389,
  [SMALL_STATE(2139)] = 78460,
  [SMALL_STATE(2140)] = 78531,
  [SMALL_STATE(2141)] = 78600,
  [SMALL_STATE(2142)] = 78708,
  [SMALL_STATE(2143)] = 78780,
  [SMALL_STATE(2144)] = 78848,
  [SMALL_STATE(2145)] = 78916,
  [SMALL_STATE(2146)] = 78984,
  [SMALL_STATE(2147)] = 79052,
  [SMALL_STATE(2148)] = 79120,
  [SMALL_STATE(2149)] = 79188,
  [SMALL_STATE(2150)] = 79256,
  [SMALL_STATE(2151)] = 79324,
  [SMALL_STATE(2152)] = 79392,
  [SMALL_STATE(2153)] = 79460,
  [SMALL_STATE(2154)] = 79528,
  [SMALL_STATE(2155)] = 79596,
  [SMALL_STATE(2156)] = 79688,
  [SMALL_STATE(2157)] = 79756,
  [SMALL_STATE(2158)] = 79824,
  [SMALL_STATE(2159)] = 79896,
  [SMALL_STATE(2160)] = 79964,
  [SMALL_STATE(2161)] = 80032,
  [SMALL_STATE(2162)] = 80100,
  [SMALL_STATE(2163)] = 80172,
  [SMALL_STATE(2164)] = 80240,
  [SMALL_STATE(2165)] = 80308,
  [SMALL_STATE(2166)] = 80376,
  [SMALL_STATE(2167)] = 80444,
  [SMALL_STATE(2168)] = 80532,
  [SMALL_STATE(2169)] = 80606,
  [SMALL_STATE(2170)] = 80688,
  [SMALL_STATE(2171)] = 80756,
  [SMALL_STATE(2172)] = 80838,
  [SMALL_STATE(2173)] = 80906,
  [SMALL_STATE(2174)] = 80974,
  [SMALL_STATE(2175)] = 81042,
  [SMALL_STATE(2176)] = 81110,
  [SMALL_STATE(2177)] = 81184,
  [SMALL_STATE(2178)] = 81252,
  [SMALL_STATE(2179)] = 81320,
  [SMALL_STATE(2180)] = 81388,
  [SMALL_STATE(2181)] = 81456,
  [SMALL_STATE(2182)] = 81524,
  [SMALL_STATE(2183)] = 81592,
  [SMALL_STATE(2184)] = 81660,
  [SMALL_STATE(2185)] = 81732,
  [SMALL_STATE(2186)] = 81800,
  [SMALL_STATE(2187)] = 81868,
  [SMALL_STATE(2188)] = 81952,
  [SMALL_STATE(2189)] = 82026,
  [SMALL_STATE(2190)] = 82094,
  [SMALL_STATE(2191)] = 82184,
  [SMALL_STATE(2192)] = 82288,
  [SMALL_STATE(2193)] = 82362,
  [SMALL_STATE(2194)] = 82432,
  [SMALL_STATE(2195)] = 82500,
  [SMALL_STATE(2196)] = 82568,
  [SMALL_STATE(2197)] = 82648,
  [SMALL_STATE(2198)] = 82716,
  [SMALL_STATE(2199)] = 82788,
  [SMALL_STATE(2200)] = 82860,
  [SMALL_STATE(2201)] = 82928,
  [SMALL_STATE(2202)] = 82996,
  [SMALL_STATE(2203)] = 83068,
  [SMALL_STATE(2204)] = 83138,
  [SMALL_STATE(2205)] = 83210,
  [SMALL_STATE(2206)] = 83280,
  [SMALL_STATE(2207)] = 83352,
  [SMALL_STATE(2208)] = 83422,
  [SMALL_STATE(2209)] = 83490,
  [SMALL_STATE(2210)] = 83562,
  [SMALL_STATE(2211)] = 83632,
  [SMALL_STATE(2212)] = 83700,
  [SMALL_STATE(2213)] = 83770,
  [SMALL_STATE(2214)] = 83840,
  [SMALL_STATE(2215)] = 83908,
  [SMALL_STATE(2216)] = 83978,
  [SMALL_STATE(2217)] = 84048,
  [SMALL_STATE(2218)] = 84118,
  [SMALL_STATE(2219)] = 84190,
  [SMALL_STATE(2220)] = 84258,
  [SMALL_STATE(2221)] = 84326,
  [SMALL_STATE(2222)] = 84394,
  [SMALL_STATE(2223)] = 84462,
  [SMALL_STATE(2224)] = 84530,
  [SMALL_STATE(2225)] = 84598,
  [SMALL_STATE(2226)] = 84666,
  [SMALL_STATE(2227)] = 84742,
  [SMALL_STATE(2228)] = 84810,
  [SMALL_STATE(2229)] = 84878,
  [SMALL_STATE(2230)] = 84946,
  [SMALL_STATE(2231)] = 85014,
  [SMALL_STATE(2232)] = 85084,
  [SMALL_STATE(2233)] = 85152,
  [SMALL_STATE(2234)] = 85220,
  [SMALL_STATE(2235)] = 85288,
  [SMALL_STATE(2236)] = 85356,
  [SMALL_STATE(2237)] = 85424,
  [SMALL_STATE(2238)] = 85492,
  [SMALL_STATE(2239)] = 85562,
  [SMALL_STATE(2240)] = 85630,
  [SMALL_STATE(2241)] = 85698,
  [SMALL_STATE(2242)] = 85766,
  [SMALL_STATE(2243)] = 85834,
  [SMALL_STATE(2244)] = 85906,
  [SMALL_STATE(2245)] = 86020,
  [SMALL_STATE(2246)] = 86088,
  [SMALL_STATE(2247)] = 86156,
  [SMALL_STATE(2248)] = 86226,
  [SMALL_STATE(2249)] = 86296,
  [SMALL_STATE(2250)] = 86404,
  [SMALL_STATE(2251)] = 86472,
  [SMALL_STATE(2252)] = 86542,
  [SMALL_STATE(2253)] = 86610,
  [SMALL_STATE(2254)] = 86678,
  [SMALL_STATE(2255)] = 86746,
  [SMALL_STATE(2256)] = 86814,
  [SMALL_STATE(2257)] = 86882,
  [SMALL_STATE(2258)] = 86954,
  [SMALL_STATE(2259)] = 87026,
  [SMALL_STATE(2260)] = 87094,
  [SMALL_STATE(2261)] = 87162,
  [SMALL_STATE(2262)] = 87234,
  [SMALL_STATE(2263)] = 87302,
  [SMALL_STATE(2264)] = 87370,
  [SMALL_STATE(2265)] = 87438,
  [SMALL_STATE(2266)] = 87506,
  [SMALL_STATE(2267)] = 87574,
  [SMALL_STATE(2268)] = 87642,
  [SMALL_STATE(2269)] = 87710,
  [SMALL_STATE(2270)] = 87778,
  [SMALL_STATE(2271)] = 87846,
  [SMALL_STATE(2272)] = 87916,
  [SMALL_STATE(2273)] = 87984,
  [SMALL_STATE(2274)] = 88052,
  [SMALL_STATE(2275)] = 88120,
  [SMALL_STATE(2276)] = 88226,
  [SMALL_STATE(2277)] = 88328,
  [SMALL_STATE(2278)] = 88396,
  [SMALL_STATE(2279)] = 88468,
  [SMALL_STATE(2280)] = 88538,
  [SMALL_STATE(2281)] = 88638,
  [SMALL_STATE(2282)] = 88734,
  [SMALL_STATE(2283)] = 88802,
  [SMALL_STATE(2284)] = 88870,
  [SMALL_STATE(2285)] = 88942,
  [SMALL_STATE(2286)] = 89014,
  [SMALL_STATE(2287)] = 89082,
  [SMALL_STATE(2288)] = 89150,
  [SMALL_STATE(2289)] = 89222,
  [SMALL_STATE(2290)] = 89290,
  [SMALL_STATE(2291)] = 89358,
  [SMALL_STATE(2292)] = 89468,
  [SMALL_STATE(2293)] = 89540,
  [SMALL_STATE(2294)] = 89608,
  [SMALL_STATE(2295)] = 89676,
  [SMALL_STATE(2296)] = 89744,
  [SMALL_STATE(2297)] = 89812,
  [SMALL_STATE(2298)] = 89880,
  [SMALL_STATE(2299)] = 89948,
  [SMALL_STATE(2300)] = 90020,
  [SMALL_STATE(2301)] = 90088,
  [SMALL_STATE(2302)] = 90156,
  [SMALL_STATE(2303)] = 90224,
  [SMALL_STATE(2304)] = 90292,
  [SMALL_STATE(2305)] = 90360,
  [SMALL_STATE(2306)] = 90428,
  [SMALL_STATE(2307)] = 90496,
  [SMALL_STATE(2308)] = 90606,
  [SMALL_STATE(2309)] = 90674,
  [SMALL_STATE(2310)] = 90742,
  [SMALL_STATE(2311)] = 90810,
  [SMALL_STATE(2312)] = 90878,
  [SMALL_STATE(2313)] = 90946,
  [SMALL_STATE(2314)] = 91014,
  [SMALL_STATE(2315)] = 91086,
  [SMALL_STATE(2316)] = 91154,
  [SMALL_STATE(2317)] = 91222,
  [SMALL_STATE(2318)] = 91290,
  [SMALL_STATE(2319)] = 91358,
  [SMALL_STATE(2320)] = 91426,
  [SMALL_STATE(2321)] = 91494,
  [SMALL_STATE(2322)] = 91562,
  [SMALL_STATE(2323)] = 91630,
  [SMALL_STATE(2324)] = 91698,
  [SMALL_STATE(2325)] = 91766,
  [SMALL_STATE(2326)] = 91834,
  [SMALL_STATE(2327)] = 91902,
  [SMALL_STATE(2328)] = 91970,
  [SMALL_STATE(2329)] = 92038,
  [SMALL_STATE(2330)] = 92106,
  [SMALL_STATE(2331)] = 92176,
  [SMALL_STATE(2332)] = 92244,
  [SMALL_STATE(2333)] = 92312,
  [SMALL_STATE(2334)] = 92380,
  [SMALL_STATE(2335)] = 92448,
  [SMALL_STATE(2336)] = 92520,
  [SMALL_STATE(2337)] = 92592,
  [SMALL_STATE(2338)] = 92664,
  [SMALL_STATE(2339)] = 92732,
  [SMALL_STATE(2340)] = 92802,
  [SMALL_STATE(2341)] = 92872,
  [SMALL_STATE(2342)] = 92942,
  [SMALL_STATE(2343)] = 93010,
  [SMALL_STATE(2344)] = 93078,
  [SMALL_STATE(2345)] = 93146,
  [SMALL_STATE(2346)] = 93214,
  [SMALL_STATE(2347)] = 93282,
  [SMALL_STATE(2348)] = 93350,
  [SMALL_STATE(2349)] = 93418,
  [SMALL_STATE(2350)] = 93486,
  [SMALL_STATE(2351)] = 93554,
  [SMALL_STATE(2352)] = 93622,
  [SMALL_STATE(2353)] = 93690,
  [SMALL_STATE(2354)] = 93758,
  [SMALL_STATE(2355)] = 93872,
  [SMALL_STATE(2356)] = 93940,
  [SMALL_STATE(2357)] = 94008,
  [SMALL_STATE(2358)] = 94102,
  [SMALL_STATE(2359)] = 94172,
  [SMALL_STATE(2360)] = 94242,
  [SMALL_STATE(2361)] = 94312,
  [SMALL_STATE(2362)] = 94380,
  [SMALL_STATE(2363)] = 94454,
  [SMALL_STATE(2364)] = 94522,
  [SMALL_STATE(2365)] = 94589,
  [SMALL_STATE(2366)] = 94658,
  [SMALL_STATE(2367)] = 94727,
  [SMALL_STATE(2368)] = 94796,
  [SMALL_STATE(2369)] = 94865,
  [SMALL_STATE(2370)] = 94934,
  [SMALL_STATE(2371)] = 95003,
  [SMALL_STATE(2372)] = 95072,
  [SMALL_STATE(2373)] = 95141,
  [SMALL_STATE(2374)] = 95210,
  [SMALL_STATE(2375)] = 95277,
  [SMALL_STATE(2376)] = 95344,
  [SMALL_STATE(2377)] = 95411,
  [SMALL_STATE(2378)] = 95478,
  [SMALL_STATE(2379)] = 95545,
  [SMALL_STATE(2380)] = 95612,
  [SMALL_STATE(2381)] = 95679,
  [SMALL_STATE(2382)] = 95746,
  [SMALL_STATE(2383)] = 95813,
  [SMALL_STATE(2384)] = 95880,
  [SMALL_STATE(2385)] = 95947,
  [SMALL_STATE(2386)] = 96014,
  [SMALL_STATE(2387)] = 96081,
  [SMALL_STATE(2388)] = 96148,
  [SMALL_STATE(2389)] = 96215,
  [SMALL_STATE(2390)] = 96282,
  [SMALL_STATE(2391)] = 96349,
  [SMALL_STATE(2392)] = 96416,
  [SMALL_STATE(2393)] = 96483,
  [SMALL_STATE(2394)] = 96550,
  [SMALL_STATE(2395)] = 96617,
  [SMALL_STATE(2396)] = 96684,
  [SMALL_STATE(2397)] = 96751,
  [SMALL_STATE(2398)] = 96818,
  [SMALL_STATE(2399)] = 96885,
  [SMALL_STATE(2400)] = 96952,
  [SMALL_STATE(2401)] = 97019,
  [SMALL_STATE(2402)] = 97086,
  [SMALL_STATE(2403)] = 97153,
  [SMALL_STATE(2404)] = 97220,
  [SMALL_STATE(2405)] = 97287,
  [SMALL_STATE(2406)] = 97354,
  [SMALL_STATE(2407)] = 97421,
  [SMALL_STATE(2408)] = 97488,
  [SMALL_STATE(2409)] = 97555,
  [SMALL_STATE(2410)] = 97622,
  [SMALL_STATE(2411)] = 97689,
  [SMALL_STATE(2412)] = 97756,
  [SMALL_STATE(2413)] = 97823,
  [SMALL_STATE(2414)] = 97890,
  [SMALL_STATE(2415)] = 97957,
  [SMALL_STATE(2416)] = 98024,
  [SMALL_STATE(2417)] = 98091,
  [SMALL_STATE(2418)] = 98158,
  [SMALL_STATE(2419)] = 98225,
  [SMALL_STATE(2420)] = 98292,
  [SMALL_STATE(2421)] = 98359,
  [SMALL_STATE(2422)] = 98426,
  [SMALL_STATE(2423)] = 98493,
  [SMALL_STATE(2424)] = 98560,
  [SMALL_STATE(2425)] = 98627,
  [SMALL_STATE(2426)] = 98694,
  [SMALL_STATE(2427)] = 98761,
  [SMALL_STATE(2428)] = 98828,
  [SMALL_STATE(2429)] = 98895,
  [SMALL_STATE(2430)] = 98962,
  [SMALL_STATE(2431)] = 99029,
  [SMALL_STATE(2432)] = 99096,
  [SMALL_STATE(2433)] = 99163,
  [SMALL_STATE(2434)] = 99230,
  [SMALL_STATE(2435)] = 99297,
  [SMALL_STATE(2436)] = 99364,
  [SMALL_STATE(2437)] = 99431,
  [SMALL_STATE(2438)] = 99498,
  [SMALL_STATE(2439)] = 99565,
  [SMALL_STATE(2440)] = 99632,
  [SMALL_STATE(2441)] = 99699,
  [SMALL_STATE(2442)] = 99766,
  [SMALL_STATE(2443)] = 99833,
  [SMALL_STATE(2444)] = 99900,
  [SMALL_STATE(2445)] = 99967,
  [SMALL_STATE(2446)] = 100034,
  [SMALL_STATE(2447)] = 100101,
  [SMALL_STATE(2448)] = 100168,
  [SMALL_STATE(2449)] = 100235,
  [SMALL_STATE(2450)] = 100302,
  [SMALL_STATE(2451)] = 100369,
  [SMALL_STATE(2452)] = 100436,
  [SMALL_STATE(2453)] = 100503,
  [SMALL_STATE(2454)] = 100570,
  [SMALL_STATE(2455)] = 100639,
  [SMALL_STATE(2456)] = 100708,
  [SMALL_STATE(2457)] = 100777,
  [SMALL_STATE(2458)] = 100846,
  [SMALL_STATE(2459)] = 100917,
  [SMALL_STATE(2460)] = 100984,
  [SMALL_STATE(2461)] = 101053,
  [SMALL_STATE(2462)] = 101122,
  [SMALL_STATE(2463)] = 101191,
  [SMALL_STATE(2464)] = 101260,
  [SMALL_STATE(2465)] = 101329,
  [SMALL_STATE(2466)] = 101398,
  [SMALL_STATE(2467)] = 101467,
  [SMALL_STATE(2468)] = 101534,
  [SMALL_STATE(2469)] = 101601,
  [SMALL_STATE(2470)] = 101668,
  [SMALL_STATE(2471)] = 101735,
  [SMALL_STATE(2472)] = 101802,
  [SMALL_STATE(2473)] = 101869,
  [SMALL_STATE(2474)] = 101936,
  [SMALL_STATE(2475)] = 102003,
  [SMALL_STATE(2476)] = 102070,
  [SMALL_STATE(2477)] = 102137,
  [SMALL_STATE(2478)] = 102204,
  [SMALL_STATE(2479)] = 102313,
  [SMALL_STATE(2480)] = 102384,
  [SMALL_STATE(2481)] = 102451,
  [SMALL_STATE(2482)] = 102518,
  [SMALL_STATE(2483)] = 102589,
  [SMALL_STATE(2484)] = 102702,
  [SMALL_STATE(2485)] = 102769,
  [SMALL_STATE(2486)] = 102836,
  [SMALL_STATE(2487)] = 102903,
  [SMALL_STATE(2488)] = 103012,
  [SMALL_STATE(2489)] = 103079,
  [SMALL_STATE(2490)] = 103146,
  [SMALL_STATE(2491)] = 103213,
  [SMALL_STATE(2492)] = 103280,
  [SMALL_STATE(2493)] = 103347,
  [SMALL_STATE(2494)] = 103414,
  [SMALL_STATE(2495)] = 103481,
  [SMALL_STATE(2496)] = 103548,
  [SMALL_STATE(2497)] = 103615,
  [SMALL_STATE(2498)] = 103682,
  [SMALL_STATE(2499)] = 103749,
  [SMALL_STATE(2500)] = 103820,
  [SMALL_STATE(2501)] = 103887,
  [SMALL_STATE(2502)] = 103954,
  [SMALL_STATE(2503)] = 104021,
  [SMALL_STATE(2504)] = 104092,
  [SMALL_STATE(2505)] = 104159,
  [SMALL_STATE(2506)] = 104226,
  [SMALL_STATE(2507)] = 104297,
  [SMALL_STATE(2508)] = 104364,
  [SMALL_STATE(2509)] = 104431,
  [SMALL_STATE(2510)] = 104500,
  [SMALL_STATE(2511)] = 104569,
  [SMALL_STATE(2512)] = 104636,
  [SMALL_STATE(2513)] = 104703,
  [SMALL_STATE(2514)] = 104770,
  [SMALL_STATE(2515)] = 104837,
  [SMALL_STATE(2516)] = 104904,
  [SMALL_STATE(2517)] = 104971,
  [SMALL_STATE(2518)] = 105038,
  [SMALL_STATE(2519)] = 105105,
  [SMALL_STATE(2520)] = 105214,
  [SMALL_STATE(2521)] = 105281,
  [SMALL_STATE(2522)] = 105348,
  [SMALL_STATE(2523)] = 105415,
  [SMALL_STATE(2524)] = 105482,
  [SMALL_STATE(2525)] = 105549,
  [SMALL_STATE(2526)] = 105616,
  [SMALL_STATE(2527)] = 105683,
  [SMALL_STATE(2528)] = 105796,
  [SMALL_STATE(2529)] = 105863,
  [SMALL_STATE(2530)] = 105930,
  [SMALL_STATE(2531)] = 105997,
  [SMALL_STATE(2532)] = 106064,
  [SMALL_STATE(2533)] = 106131,
  [SMALL_STATE(2534)] = 106198,
  [SMALL_STATE(2535)] = 106265,
  [SMALL_STATE(2536)] = 106332,
  [SMALL_STATE(2537)] = 106399,
  [SMALL_STATE(2538)] = 106466,
  [SMALL_STATE(2539)] = 106533,
  [SMALL_STATE(2540)] = 106600,
  [SMALL_STATE(2541)] = 106667,
  [SMALL_STATE(2542)] = 106734,
  [SMALL_STATE(2543)] = 106801,
  [SMALL_STATE(2544)] = 106872,
  [SMALL_STATE(2545)] = 106939,
  [SMALL_STATE(2546)] = 107006,
  [SMALL_STATE(2547)] = 107073,
  [SMALL_STATE(2548)] = 107140,
  [SMALL_STATE(2549)] = 107207,
  [SMALL_STATE(2550)] = 107274,
  [SMALL_STATE(2551)] = 107341,
  [SMALL_STATE(2552)] = 107408,
  [SMALL_STATE(2553)] = 107475,
  [SMALL_STATE(2554)] = 107542,
  [SMALL_STATE(2555)] = 107609,
  [SMALL_STATE(2556)] = 107676,
  [SMALL_STATE(2557)] = 107743,
  [SMALL_STATE(2558)] = 107810,
  [SMALL_STATE(2559)] = 107877,
  [SMALL_STATE(2560)] = 107944,
  [SMALL_STATE(2561)] = 108027,
  [SMALL_STATE(2562)] = 108116,
  [SMALL_STATE(2563)] = 108219,
  [SMALL_STATE(2564)] = 108290,
  [SMALL_STATE(2565)] = 108361,
  [SMALL_STATE(2566)] = 108428,
  [SMALL_STATE(2567)] = 108499,
  [SMALL_STATE(2568)] = 108572,
  [SMALL_STATE(2569)] = 108651,
  [SMALL_STATE(2570)] = 108726,
  [SMALL_STATE(2571)] = 108833,
  [SMALL_STATE(2572)] = 108940,
  [SMALL_STATE(2573)] = 109007,
  [SMALL_STATE(2574)] = 109112,
  [SMALL_STATE(2575)] = 109213,
  [SMALL_STATE(2576)] = 109312,
  [SMALL_STATE(2577)] = 109407,
  [SMALL_STATE(2578)] = 109500,
  [SMALL_STATE(2579)] = 109591,
  [SMALL_STATE(2580)] = 109678,
  [SMALL_STATE(2581)] = 109759,
  [SMALL_STATE(2582)] = 109840,
  [SMALL_STATE(2583)] = 109911,
  [SMALL_STATE(2584)] = 109984,
  [SMALL_STATE(2585)] = 110055,
  [SMALL_STATE(2586)] = 110122,
  [SMALL_STATE(2587)] = 110189,
  [SMALL_STATE(2588)] = 110298,
  [SMALL_STATE(2589)] = 110365,
  [SMALL_STATE(2590)] = 110434,
  [SMALL_STATE(2591)] = 110503,
  [SMALL_STATE(2592)] = 110570,
  [SMALL_STATE(2593)] = 110637,
  [SMALL_STATE(2594)] = 110704,
  [SMALL_STATE(2595)] = 110771,
  [SMALL_STATE(2596)] = 110838,
  [SMALL_STATE(2597)] = 110905,
  [SMALL_STATE(2598)] = 110972,
  [SMALL_STATE(2599)] = 111039,
  [SMALL_STATE(2600)] = 111110,
  [SMALL_STATE(2601)] = 111177,
  [SMALL_STATE(2602)] = 111244,
  [SMALL_STATE(2603)] = 111311,
  [SMALL_STATE(2604)] = 111378,
  [SMALL_STATE(2605)] = 111445,
  [SMALL_STATE(2606)] = 111512,
  [SMALL_STATE(2607)] = 111579,
  [SMALL_STATE(2608)] = 111646,
  [SMALL_STATE(2609)] = 111713,
  [SMALL_STATE(2610)] = 111780,
  [SMALL_STATE(2611)] = 111847,
  [SMALL_STATE(2612)] = 111914,
  [SMALL_STATE(2613)] = 111981,
  [SMALL_STATE(2614)] = 112048,
  [SMALL_STATE(2615)] = 112119,
  [SMALL_STATE(2616)] = 112186,
  [SMALL_STATE(2617)] = 112253,
  [SMALL_STATE(2618)] = 112320,
  [SMALL_STATE(2619)] = 112387,
  [SMALL_STATE(2620)] = 112454,
  [SMALL_STATE(2621)] = 112521,
  [SMALL_STATE(2622)] = 112588,
  [SMALL_STATE(2623)] = 112659,
  [SMALL_STATE(2624)] = 112726,
  [SMALL_STATE(2625)] = 112793,
  [SMALL_STATE(2626)] = 112860,
  [SMALL_STATE(2627)] = 112927,
  [SMALL_STATE(2628)] = 112996,
  [SMALL_STATE(2629)] = 113065,
  [SMALL_STATE(2630)] = 113134,
  [SMALL_STATE(2631)] = 113201,
  [SMALL_STATE(2632)] = 113270,
  [SMALL_STATE(2633)] = 113339,
  [SMALL_STATE(2634)] = 113406,
  [SMALL_STATE(2635)] = 113473,
  [SMALL_STATE(2636)] = 113540,
  [SMALL_STATE(2637)] = 113607,
  [SMALL_STATE(2638)] = 113674,
  [SMALL_STATE(2639)] = 113741,
  [SMALL_STATE(2640)] = 113808,
  [SMALL_STATE(2641)] = 113891,
  [SMALL_STATE(2642)] = 113958,
  [SMALL_STATE(2643)] = 114025,
  [SMALL_STATE(2644)] = 114092,
  [SMALL_STATE(2645)] = 114159,
  [SMALL_STATE(2646)] = 114226,
  [SMALL_STATE(2647)] = 114315,
  [SMALL_STATE(2648)] = 114418,
  [SMALL_STATE(2649)] = 114485,
  [SMALL_STATE(2650)] = 114558,
  [SMALL_STATE(2651)] = 114637,
  [SMALL_STATE(2652)] = 114712,
  [SMALL_STATE(2653)] = 114819,
  [SMALL_STATE(2654)] = 114926,
  [SMALL_STATE(2655)] = 114993,
  [SMALL_STATE(2656)] = 115098,
  [SMALL_STATE(2657)] = 115199,
  [SMALL_STATE(2658)] = 115298,
  [SMALL_STATE(2659)] = 115393,
  [SMALL_STATE(2660)] = 115486,
  [SMALL_STATE(2661)] = 115577,
  [SMALL_STATE(2662)] = 115664,
  [SMALL_STATE(2663)] = 115745,
  [SMALL_STATE(2664)] = 115826,
  [SMALL_STATE(2665)] = 115899,
  [SMALL_STATE(2666)] = 116012,
  [SMALL_STATE(2667)] = 116079,
  [SMALL_STATE(2668)] = 116152,
  [SMALL_STATE(2669)] = 116225,
  [SMALL_STATE(2670)] = 116296,
  [SMALL_STATE(2671)] = 116363,
  [SMALL_STATE(2672)] = 116430,
  [SMALL_STATE(2673)] = 116497,
  [SMALL_STATE(2674)] = 116570,
  [SMALL_STATE(2675)] = 116643,
  [SMALL_STATE(2676)] = 116710,
  [SMALL_STATE(2677)] = 116777,
  [SMALL_STATE(2678)] = 116844,
  [SMALL_STATE(2679)] = 116917,
  [SMALL_STATE(2680)] = 116990,
  [SMALL_STATE(2681)] = 117057,
  [SMALL_STATE(2682)] = 117124,
  [SMALL_STATE(2683)] = 117191,
  [SMALL_STATE(2684)] = 117258,
  [SMALL_STATE(2685)] = 117367,
  [SMALL_STATE(2686)] = 117440,
  [SMALL_STATE(2687)] = 117507,
  [SMALL_STATE(2688)] = 117574,
  [SMALL_STATE(2689)] = 117643,
  [SMALL_STATE(2690)] = 117712,
  [SMALL_STATE(2691)] = 117785,
  [SMALL_STATE(2692)] = 117852,
  [SMALL_STATE(2693)] = 117919,
  [SMALL_STATE(2694)] = 117986,
  [SMALL_STATE(2695)] = 118053,
  [SMALL_STATE(2696)] = 118124,
  [SMALL_STATE(2697)] = 118191,
  [SMALL_STATE(2698)] = 118258,
  [SMALL_STATE(2699)] = 118325,
  [SMALL_STATE(2700)] = 118392,
  [SMALL_STATE(2701)] = 118459,
  [SMALL_STATE(2702)] = 118572,
  [SMALL_STATE(2703)] = 118639,
  [SMALL_STATE(2704)] = 118710,
  [SMALL_STATE(2705)] = 118779,
  [SMALL_STATE(2706)] = 118848,
  [SMALL_STATE(2707)] = 118917,
  [SMALL_STATE(2708)] = 118984,
  [SMALL_STATE(2709)] = 119093,
  [SMALL_STATE(2710)] = 119162,
  [SMALL_STATE(2711)] = 119233,
  [SMALL_STATE(2712)] = 119300,
  [SMALL_STATE(2713)] = 119369,
  [SMALL_STATE(2714)] = 119436,
  [SMALL_STATE(2715)] = 119503,
  [SMALL_STATE(2716)] = 119570,
  [SMALL_STATE(2717)] = 119679,
  [SMALL_STATE(2718)] = 119746,
  [SMALL_STATE(2719)] = 119817,
  [SMALL_STATE(2720)] = 119926,
  [SMALL_STATE(2721)] = 119993,
  [SMALL_STATE(2722)] = 120060,
  [SMALL_STATE(2723)] = 120127,
  [SMALL_STATE(2724)] = 120194,
  [SMALL_STATE(2725)] = 120265,
  [SMALL_STATE(2726)] = 120332,
  [SMALL_STATE(2727)] = 120401,
  [SMALL_STATE(2728)] = 120470,
  [SMALL_STATE(2729)] = 120539,
  [SMALL_STATE(2730)] = 120608,
  [SMALL_STATE(2731)] = 120677,
  [SMALL_STATE(2732)] = 120746,
  [SMALL_STATE(2733)] = 120815,
  [SMALL_STATE(2734)] = 120884,
  [SMALL_STATE(2735)] = 120953,
  [SMALL_STATE(2736)] = 121022,
  [SMALL_STATE(2737)] = 121091,
  [SMALL_STATE(2738)] = 121160,
  [SMALL_STATE(2739)] = 121229,
  [SMALL_STATE(2740)] = 121298,
  [SMALL_STATE(2741)] = 121367,
  [SMALL_STATE(2742)] = 121436,
  [SMALL_STATE(2743)] = 121505,
  [SMALL_STATE(2744)] = 121574,
  [SMALL_STATE(2745)] = 121643,
  [SMALL_STATE(2746)] = 121712,
  [SMALL_STATE(2747)] = 121779,
  [SMALL_STATE(2748)] = 121848,
  [SMALL_STATE(2749)] = 121915,
  [SMALL_STATE(2750)] = 121982,
  [SMALL_STATE(2751)] = 122049,
  [SMALL_STATE(2752)] = 122116,
  [SMALL_STATE(2753)] = 122183,
  [SMALL_STATE(2754)] = 122250,
  [SMALL_STATE(2755)] = 122317,
  [SMALL_STATE(2756)] = 122384,
  [SMALL_STATE(2757)] = 122451,
  [SMALL_STATE(2758)] = 122518,
  [SMALL_STATE(2759)] = 122585,
  [SMALL_STATE(2760)] = 122652,
  [SMALL_STATE(2761)] = 122719,
  [SMALL_STATE(2762)] = 122786,
  [SMALL_STATE(2763)] = 122855,
  [SMALL_STATE(2764)] = 122922,
  [SMALL_STATE(2765)] = 122991,
  [SMALL_STATE(2766)] = 123058,
  [SMALL_STATE(2767)] = 123127,
  [SMALL_STATE(2768)] = 123196,
  [SMALL_STATE(2769)] = 123263,
  [SMALL_STATE(2770)] = 123332,
  [SMALL_STATE(2771)] = 123401,
  [SMALL_STATE(2772)] = 123470,
  [SMALL_STATE(2773)] = 123537,
  [SMALL_STATE(2774)] = 123606,
  [SMALL_STATE(2775)] = 123675,
  [SMALL_STATE(2776)] = 123744,
  [SMALL_STATE(2777)] = 123813,
  [SMALL_STATE(2778)] = 123882,
  [SMALL_STATE(2779)] = 123951,
  [SMALL_STATE(2780)] = 124020,
  [SMALL_STATE(2781)] = 124089,
  [SMALL_STATE(2782)] = 124158,
  [SMALL_STATE(2783)] = 124225,
  [SMALL_STATE(2784)] = 124292,
  [SMALL_STATE(2785)] = 124361,
  [SMALL_STATE(2786)] = 124430,
  [SMALL_STATE(2787)] = 124499,
  [SMALL_STATE(2788)] = 124566,
  [SMALL_STATE(2789)] = 124635,
  [SMALL_STATE(2790)] = 124702,
  [SMALL_STATE(2791)] = 124769,
  [SMALL_STATE(2792)] = 124836,
  [SMALL_STATE(2793)] = 124903,
  [SMALL_STATE(2794)] = 124970,
  [SMALL_STATE(2795)] = 125037,
  [SMALL_STATE(2796)] = 125104,
  [SMALL_STATE(2797)] = 125171,
  [SMALL_STATE(2798)] = 125238,
  [SMALL_STATE(2799)] = 125305,
  [SMALL_STATE(2800)] = 125372,
  [SMALL_STATE(2801)] = 125439,
  [SMALL_STATE(2802)] = 125506,
  [SMALL_STATE(2803)] = 125573,
  [SMALL_STATE(2804)] = 125640,
  [SMALL_STATE(2805)] = 125707,
  [SMALL_STATE(2806)] = 125774,
  [SMALL_STATE(2807)] = 125841,
  [SMALL_STATE(2808)] = 125908,
  [SMALL_STATE(2809)] = 125975,
  [SMALL_STATE(2810)] = 126042,
  [SMALL_STATE(2811)] = 126109,
  [SMALL_STATE(2812)] = 126176,
  [SMALL_STATE(2813)] = 126243,
  [SMALL_STATE(2814)] = 126310,
  [SMALL_STATE(2815)] = 126377,
  [SMALL_STATE(2816)] = 126444,
  [SMALL_STATE(2817)] = 126511,
  [SMALL_STATE(2818)] = 126578,
  [SMALL_STATE(2819)] = 126645,
  [SMALL_STATE(2820)] = 126712,
  [SMALL_STATE(2821)] = 126779,
  [SMALL_STATE(2822)] = 126846,
  [SMALL_STATE(2823)] = 126913,
  [SMALL_STATE(2824)] = 126980,
  [SMALL_STATE(2825)] = 127047,
  [SMALL_STATE(2826)] = 127114,
  [SMALL_STATE(2827)] = 127181,
  [SMALL_STATE(2828)] = 127248,
  [SMALL_STATE(2829)] = 127315,
  [SMALL_STATE(2830)] = 127382,
  [SMALL_STATE(2831)] = 127449,
  [SMALL_STATE(2832)] = 127516,
  [SMALL_STATE(2833)] = 127583,
  [SMALL_STATE(2834)] = 127650,
  [SMALL_STATE(2835)] = 127717,
  [SMALL_STATE(2836)] = 127784,
  [SMALL_STATE(2837)] = 127851,
  [SMALL_STATE(2838)] = 127918,
  [SMALL_STATE(2839)] = 127985,
  [SMALL_STATE(2840)] = 128052,
  [SMALL_STATE(2841)] = 128119,
  [SMALL_STATE(2842)] = 128186,
  [SMALL_STATE(2843)] = 128253,
  [SMALL_STATE(2844)] = 128320,
  [SMALL_STATE(2845)] = 128387,
  [SMALL_STATE(2846)] = 128454,
  [SMALL_STATE(2847)] = 128527,
  [SMALL_STATE(2848)] = 128594,
  [SMALL_STATE(2849)] = 128661,
  [SMALL_STATE(2850)] = 128728,
  [SMALL_STATE(2851)] = 128797,
  [SMALL_STATE(2852)] = 128864,
  [SMALL_STATE(2853)] = 128933,
  [SMALL_STATE(2854)] = 129002,
  [SMALL_STATE(2855)] = 129069,
  [SMALL_STATE(2856)] = 129138,
  [SMALL_STATE(2857)] = 129205,
  [SMALL_STATE(2858)] = 129272,
  [SMALL_STATE(2859)] = 129341,
  [SMALL_STATE(2860)] = 129410,
  [SMALL_STATE(2861)] = 129479,
  [SMALL_STATE(2862)] = 129546,
  [SMALL_STATE(2863)] = 129613,
  [SMALL_STATE(2864)] = 129682,
  [SMALL_STATE(2865)] = 129751,
  [SMALL_STATE(2866)] = 129820,
  [SMALL_STATE(2867)] = 129889,
  [SMALL_STATE(2868)] = 129958,
  [SMALL_STATE(2869)] = 130027,
  [SMALL_STATE(2870)] = 130096,
  [SMALL_STATE(2871)] = 130165,
  [SMALL_STATE(2872)] = 130232,
  [SMALL_STATE(2873)] = 130301,
  [SMALL_STATE(2874)] = 130370,
  [SMALL_STATE(2875)] = 130439,
  [SMALL_STATE(2876)] = 130508,
  [SMALL_STATE(2877)] = 130577,
  [SMALL_STATE(2878)] = 130646,
  [SMALL_STATE(2879)] = 130717,
  [SMALL_STATE(2880)] = 130788,
  [SMALL_STATE(2881)] = 130857,
  [SMALL_STATE(2882)] = 130926,
  [SMALL_STATE(2883)] = 130995,
  [SMALL_STATE(2884)] = 131062,
  [SMALL_STATE(2885)] = 131131,
  [SMALL_STATE(2886)] = 131200,
  [SMALL_STATE(2887)] = 131269,
  [SMALL_STATE(2888)] = 131338,
  [SMALL_STATE(2889)] = 131405,
  [SMALL_STATE(2890)] = 131472,
  [SMALL_STATE(2891)] = 131541,
  [SMALL_STATE(2892)] = 131610,
  [SMALL_STATE(2893)] = 131679,
  [SMALL_STATE(2894)] = 131748,
  [SMALL_STATE(2895)] = 131817,
  [SMALL_STATE(2896)] = 131886,
  [SMALL_STATE(2897)] = 131953,
  [SMALL_STATE(2898)] = 132022,
  [SMALL_STATE(2899)] = 132089,
  [SMALL_STATE(2900)] = 132158,
  [SMALL_STATE(2901)] = 132225,
  [SMALL_STATE(2902)] = 132292,
  [SMALL_STATE(2903)] = 132359,
  [SMALL_STATE(2904)] = 132426,
  [SMALL_STATE(2905)] = 132493,
  [SMALL_STATE(2906)] = 132560,
  [SMALL_STATE(2907)] = 132627,
  [SMALL_STATE(2908)] = 132694,
  [SMALL_STATE(2909)] = 132761,
  [SMALL_STATE(2910)] = 132828,
  [SMALL_STATE(2911)] = 132895,
  [SMALL_STATE(2912)] = 132962,
  [SMALL_STATE(2913)] = 133029,
  [SMALL_STATE(2914)] = 133096,
  [SMALL_STATE(2915)] = 133163,
  [SMALL_STATE(2916)] = 133230,
  [SMALL_STATE(2917)] = 133297,
  [SMALL_STATE(2918)] = 133368,
  [SMALL_STATE(2919)] = 133439,
  [SMALL_STATE(2920)] = 133510,
  [SMALL_STATE(2921)] = 133577,
  [SMALL_STATE(2922)] = 133646,
  [SMALL_STATE(2923)] = 133715,
  [SMALL_STATE(2924)] = 133784,
  [SMALL_STATE(2925)] = 133853,
  [SMALL_STATE(2926)] = 133922,
  [SMALL_STATE(2927)] = 133995,
  [SMALL_STATE(2928)] = 134061,
  [SMALL_STATE(2929)] = 134127,
  [SMALL_STATE(2930)] = 134193,
  [SMALL_STATE(2931)] = 134259,
  [SMALL_STATE(2932)] = 134325,
  [SMALL_STATE(2933)] = 134391,
  [SMALL_STATE(2934)] = 134457,
  [SMALL_STATE(2935)] = 134527,
  [SMALL_STATE(2936)] = 134593,
  [SMALL_STATE(2937)] = 134659,
  [SMALL_STATE(2938)] = 134725,
  [SMALL_STATE(2939)] = 134791,
  [SMALL_STATE(2940)] = 134861,
  [SMALL_STATE(2941)] = 134927,
  [SMALL_STATE(2942)] = 134993,
  [SMALL_STATE(2943)] = 135059,
  [SMALL_STATE(2944)] = 135125,
  [SMALL_STATE(2945)] = 135191,
  [SMALL_STATE(2946)] = 135257,
  [SMALL_STATE(2947)] = 135323,
  [SMALL_STATE(2948)] = 135389,
  [SMALL_STATE(2949)] = 135455,
  [SMALL_STATE(2950)] = 135521,
  [SMALL_STATE(2951)] = 135587,
  [SMALL_STATE(2952)] = 135653,
  [SMALL_STATE(2953)] = 135719,
  [SMALL_STATE(2954)] = 135785,
  [SMALL_STATE(2955)] = 135851,
  [SMALL_STATE(2956)] = 135917,
  [SMALL_STATE(2957)] = 135983,
  [SMALL_STATE(2958)] = 136049,
  [SMALL_STATE(2959)] = 136131,
  [SMALL_STATE(2960)] = 136219,
  [SMALL_STATE(2961)] = 136321,
  [SMALL_STATE(2962)] = 136403,
  [SMALL_STATE(2963)] = 136491,
  [SMALL_STATE(2964)] = 136593,
  [SMALL_STATE(2965)] = 136659,
  [SMALL_STATE(2966)] = 136731,
  [SMALL_STATE(2967)] = 136809,
  [SMALL_STATE(2968)] = 136883,
  [SMALL_STATE(2969)] = 136989,
  [SMALL_STATE(2970)] = 137095,
  [SMALL_STATE(2971)] = 137167,
  [SMALL_STATE(2972)] = 137271,
  [SMALL_STATE(2973)] = 137371,
  [SMALL_STATE(2974)] = 137469,
  [SMALL_STATE(2975)] = 137563,
  [SMALL_STATE(2976)] = 137655,
  [SMALL_STATE(2977)] = 137745,
  [SMALL_STATE(2978)] = 137831,
  [SMALL_STATE(2979)] = 137911,
  [SMALL_STATE(2980)] = 137991,
  [SMALL_STATE(2981)] = 138061,
  [SMALL_STATE(2982)] = 138139,
  [SMALL_STATE(2983)] = 138213,
  [SMALL_STATE(2984)] = 138319,
  [SMALL_STATE(2985)] = 138425,
  [SMALL_STATE(2986)] = 138491,
  [SMALL_STATE(2987)] = 138595,
  [SMALL_STATE(2988)] = 138695,
  [SMALL_STATE(2989)] = 138793,
  [SMALL_STATE(2990)] = 138887,
  [SMALL_STATE(2991)] = 138979,
  [SMALL_STATE(2992)] = 139069,
  [SMALL_STATE(2993)] = 139155,
  [SMALL_STATE(2994)] = 139235,
  [SMALL_STATE(2995)] = 139315,
  [SMALL_STATE(2996)] = 139385,
  [SMALL_STATE(2997)] = 139451,
  [SMALL_STATE(2998)] = 139517,
  [SMALL_STATE(2999)] = 139583,
  [SMALL_STATE(3000)] = 139649,
  [SMALL_STATE(3001)] = 139715,
  [SMALL_STATE(3002)] = 139781,
  [SMALL_STATE(3003)] = 139847,
  [SMALL_STATE(3004)] = 139913,
  [SMALL_STATE(3005)] = 139979,
  [SMALL_STATE(3006)] = 140045,
  [SMALL_STATE(3007)] = 140111,
  [SMALL_STATE(3008)] = 140177,
  [SMALL_STATE(3009)] = 140243,
  [SMALL_STATE(3010)] = 140309,
  [SMALL_STATE(3011)] = 140375,
  [SMALL_STATE(3012)] = 140445,
  [SMALL_STATE(3013)] = 140511,
  [SMALL_STATE(3014)] = 140577,
  [SMALL_STATE(3015)] = 140643,
  [SMALL_STATE(3016)] = 140709,
  [SMALL_STATE(3017)] = 140775,
  [SMALL_STATE(3018)] = 140841,
  [SMALL_STATE(3019)] = 140907,
  [SMALL_STATE(3020)] = 140973,
  [SMALL_STATE(3021)] = 141039,
  [SMALL_STATE(3022)] = 141105,
  [SMALL_STATE(3023)] = 141171,
  [SMALL_STATE(3024)] = 141237,
  [SMALL_STATE(3025)] = 141303,
  [SMALL_STATE(3026)] = 141369,
  [SMALL_STATE(3027)] = 141435,
  [SMALL_STATE(3028)] = 141501,
  [SMALL_STATE(3029)] = 141567,
  [SMALL_STATE(3030)] = 141633,
  [SMALL_STATE(3031)] = 141699,
  [SMALL_STATE(3032)] = 141765,
  [SMALL_STATE(3033)] = 141831,
  [SMALL_STATE(3034)] = 141897,
  [SMALL_STATE(3035)] = 141963,
  [SMALL_STATE(3036)] = 142029,
  [SMALL_STATE(3037)] = 142095,
  [SMALL_STATE(3038)] = 142161,
  [SMALL_STATE(3039)] = 142227,
  [SMALL_STATE(3040)] = 142293,
  [SMALL_STATE(3041)] = 142359,
  [SMALL_STATE(3042)] = 142425,
  [SMALL_STATE(3043)] = 142491,
  [SMALL_STATE(3044)] = 142557,
  [SMALL_STATE(3045)] = 142623,
  [SMALL_STATE(3046)] = 142689,
  [SMALL_STATE(3047)] = 142755,
  [SMALL_STATE(3048)] = 142821,
  [SMALL_STATE(3049)] = 142887,
  [SMALL_STATE(3050)] = 142957,
  [SMALL_STATE(3051)] = 143023,
  [SMALL_STATE(3052)] = 143089,
  [SMALL_STATE(3053)] = 143155,
  [SMALL_STATE(3054)] = 143221,
  [SMALL_STATE(3055)] = 143287,
  [SMALL_STATE(3056)] = 143353,
  [SMALL_STATE(3057)] = 143419,
  [SMALL_STATE(3058)] = 143485,
  [SMALL_STATE(3059)] = 143555,
  [SMALL_STATE(3060)] = 143621,
  [SMALL_STATE(3061)] = 143687,
  [SMALL_STATE(3062)] = 143753,
  [SMALL_STATE(3063)] = 143819,
  [SMALL_STATE(3064)] = 143885,
  [SMALL_STATE(3065)] = 143951,
  [SMALL_STATE(3066)] = 144017,
  [SMALL_STATE(3067)] = 144083,
  [SMALL_STATE(3068)] = 144149,
  [SMALL_STATE(3069)] = 144215,
  [SMALL_STATE(3070)] = 144281,
  [SMALL_STATE(3071)] = 144347,
  [SMALL_STATE(3072)] = 144413,
  [SMALL_STATE(3073)] = 144479,
  [SMALL_STATE(3074)] = 144545,
  [SMALL_STATE(3075)] = 144611,
  [SMALL_STATE(3076)] = 144677,
  [SMALL_STATE(3077)] = 144743,
  [SMALL_STATE(3078)] = 144809,
  [SMALL_STATE(3079)] = 144875,
  [SMALL_STATE(3080)] = 144941,
  [SMALL_STATE(3081)] = 145011,
  [SMALL_STATE(3082)] = 145077,
  [SMALL_STATE(3083)] = 145143,
  [SMALL_STATE(3084)] = 145209,
  [SMALL_STATE(3085)] = 145275,
  [SMALL_STATE(3086)] = 145345,
  [SMALL_STATE(3087)] = 145415,
  [SMALL_STATE(3088)] = 145481,
  [SMALL_STATE(3089)] = 145547,
  [SMALL_STATE(3090)] = 145613,
  [SMALL_STATE(3091)] = 145679,
  [SMALL_STATE(3092)] = 145745,
  [SMALL_STATE(3093)] = 145811,
  [SMALL_STATE(3094)] = 145877,
  [SMALL_STATE(3095)] = 145943,
  [SMALL_STATE(3096)] = 146009,
  [SMALL_STATE(3097)] = 146075,
  [SMALL_STATE(3098)] = 146183,
  [SMALL_STATE(3099)] = 146249,
  [SMALL_STATE(3100)] = 146315,
  [SMALL_STATE(3101)] = 146381,
  [SMALL_STATE(3102)] = 146447,
  [SMALL_STATE(3103)] = 146513,
  [SMALL_STATE(3104)] = 146579,
  [SMALL_STATE(3105)] = 146645,
  [SMALL_STATE(3106)] = 146711,
  [SMALL_STATE(3107)] = 146777,
  [SMALL_STATE(3108)] = 146843,
  [SMALL_STATE(3109)] = 146959,
  [SMALL_STATE(3110)] = 147025,
  [SMALL_STATE(3111)] = 147091,
  [SMALL_STATE(3112)] = 147157,
  [SMALL_STATE(3113)] = 147223,
  [SMALL_STATE(3114)] = 147305,
  [SMALL_STATE(3115)] = 147393,
  [SMALL_STATE(3116)] = 147495,
  [SMALL_STATE(3117)] = 147567,
  [SMALL_STATE(3118)] = 147645,
  [SMALL_STATE(3119)] = 147719,
  [SMALL_STATE(3120)] = 147825,
  [SMALL_STATE(3121)] = 147931,
  [SMALL_STATE(3122)] = 148035,
  [SMALL_STATE(3123)] = 148135,
  [SMALL_STATE(3124)] = 148233,
  [SMALL_STATE(3125)] = 148327,
  [SMALL_STATE(3126)] = 148419,
  [SMALL_STATE(3127)] = 148509,
  [SMALL_STATE(3128)] = 148595,
  [SMALL_STATE(3129)] = 148675,
  [SMALL_STATE(3130)] = 148755,
  [SMALL_STATE(3131)] = 148821,
  [SMALL_STATE(3132)] = 148887,
  [SMALL_STATE(3133)] = 148953,
  [SMALL_STATE(3134)] = 149019,
  [SMALL_STATE(3135)] = 149085,
  [SMALL_STATE(3136)] = 149151,
  [SMALL_STATE(3137)] = 149217,
  [SMALL_STATE(3138)] = 149283,
  [SMALL_STATE(3139)] = 149349,
  [SMALL_STATE(3140)] = 149415,
  [SMALL_STATE(3141)] = 149481,
  [SMALL_STATE(3142)] = 149547,
  [SMALL_STATE(3143)] = 149613,
  [SMALL_STATE(3144)] = 149679,
  [SMALL_STATE(3145)] = 149745,
  [SMALL_STATE(3146)] = 149811,
  [SMALL_STATE(3147)] = 149877,
  [SMALL_STATE(3148)] = 149985,
  [SMALL_STATE(3149)] = 150051,
  [SMALL_STATE(3150)] = 150117,
  [SMALL_STATE(3151)] = 150183,
  [SMALL_STATE(3152)] = 150249,
  [SMALL_STATE(3153)] = 150315,
  [SMALL_STATE(3154)] = 150381,
  [SMALL_STATE(3155)] = 150447,
  [SMALL_STATE(3156)] = 150513,
  [SMALL_STATE(3157)] = 150579,
  [SMALL_STATE(3158)] = 150645,
  [SMALL_STATE(3159)] = 150713,
  [SMALL_STATE(3160)] = 150781,
  [SMALL_STATE(3161)] = 150849,
  [SMALL_STATE(3162)] = 150915,
  [SMALL_STATE(3163)] = 150983,
  [SMALL_STATE(3164)] = 151051,
  [SMALL_STATE(3165)] = 151133,
  [SMALL_STATE(3166)] = 151221,
  [SMALL_STATE(3167)] = 151323,
  [SMALL_STATE(3168)] = 151395,
  [SMALL_STATE(3169)] = 151473,
  [SMALL_STATE(3170)] = 151547,
  [SMALL_STATE(3171)] = 151653,
  [SMALL_STATE(3172)] = 151759,
  [SMALL_STATE(3173)] = 151863,
  [SMALL_STATE(3174)] = 151963,
  [SMALL_STATE(3175)] = 152061,
  [SMALL_STATE(3176)] = 152155,
  [SMALL_STATE(3177)] = 152247,
  [SMALL_STATE(3178)] = 152337,
  [SMALL_STATE(3179)] = 152423,
  [SMALL_STATE(3180)] = 152503,
  [SMALL_STATE(3181)] = 152583,
  [SMALL_STATE(3182)] = 152651,
  [SMALL_STATE(3183)] = 152719,
  [SMALL_STATE(3184)] = 152787,
  [SMALL_STATE(3185)] = 152855,
  [SMALL_STATE(3186)] = 152923,
  [SMALL_STATE(3187)] = 152991,
  [SMALL_STATE(3188)] = 153059,
  [SMALL_STATE(3189)] = 153127,
  [SMALL_STATE(3190)] = 153195,
  [SMALL_STATE(3191)] = 153263,
  [SMALL_STATE(3192)] = 153331,
  [SMALL_STATE(3193)] = 153399,
  [SMALL_STATE(3194)] = 153467,
  [SMALL_STATE(3195)] = 153535,
  [SMALL_STATE(3196)] = 153603,
  [SMALL_STATE(3197)] = 153671,
  [SMALL_STATE(3198)] = 153739,
  [SMALL_STATE(3199)] = 153805,
  [SMALL_STATE(3200)] = 153913,
  [SMALL_STATE(3201)] = 154025,
  [SMALL_STATE(3202)] = 154097,
  [SMALL_STATE(3203)] = 154169,
  [SMALL_STATE(3204)] = 154241,
  [SMALL_STATE(3205)] = 154313,
  [SMALL_STATE(3206)] = 154385,
  [SMALL_STATE(3207)] = 154493,
  [SMALL_STATE(3208)] = 154601,
  [SMALL_STATE(3209)] = 154667,
  [SMALL_STATE(3210)] = 154737,
  [SMALL_STATE(3211)] = 154853,
  [SMALL_STATE(3212)] = 154919,
  [SMALL_STATE(3213)] = 155035,
  [SMALL_STATE(3214)] = 155101,
  [SMALL_STATE(3215)] = 155171,
  [SMALL_STATE(3216)] = 155287,
  [SMALL_STATE(3217)] = 155353,
  [SMALL_STATE(3218)] = 155423,
  [SMALL_STATE(3219)] = 155493,
  [SMALL_STATE(3220)] = 155559,
  [SMALL_STATE(3221)] = 155629,
  [SMALL_STATE(3222)] = 155695,
  [SMALL_STATE(3223)] = 155761,
  [SMALL_STATE(3224)] = 155827,
  [SMALL_STATE(3225)] = 155893,
  [SMALL_STATE(3226)] = 155963,
  [SMALL_STATE(3227)] = 156033,
  [SMALL_STATE(3228)] = 156099,
  [SMALL_STATE(3229)] = 156165,
  [SMALL_STATE(3230)] = 156231,
  [SMALL_STATE(3231)] = 156297,
  [SMALL_STATE(3232)] = 156363,
  [SMALL_STATE(3233)] = 156429,
  [SMALL_STATE(3234)] = 156495,
  [SMALL_STATE(3235)] = 156561,
  [SMALL_STATE(3236)] = 156627,
  [SMALL_STATE(3237)] = 156693,
  [SMALL_STATE(3238)] = 156759,
  [SMALL_STATE(3239)] = 156825,
  [SMALL_STATE(3240)] = 156895,
  [SMALL_STATE(3241)] = 156965,
  [SMALL_STATE(3242)] = 157031,
  [SMALL_STATE(3243)] = 157097,
  [SMALL_STATE(3244)] = 157163,
  [SMALL_STATE(3245)] = 157229,
  [SMALL_STATE(3246)] = 157295,
  [SMALL_STATE(3247)] = 157361,
  [SMALL_STATE(3248)] = 157431,
  [SMALL_STATE(3249)] = 157501,
  [SMALL_STATE(3250)] = 157567,
  [SMALL_STATE(3251)] = 157633,
  [SMALL_STATE(3252)] = 157699,
  [SMALL_STATE(3253)] = 157765,
  [SMALL_STATE(3254)] = 157831,
  [SMALL_STATE(3255)] = 157897,
  [SMALL_STATE(3256)] = 157963,
  [SMALL_STATE(3257)] = 158029,
  [SMALL_STATE(3258)] = 158097,
  [SMALL_STATE(3259)] = 158165,
  [SMALL_STATE(3260)] = 158233,
  [SMALL_STATE(3261)] = 158301,
  [SMALL_STATE(3262)] = 158369,
  [SMALL_STATE(3263)] = 158435,
  [SMALL_STATE(3264)] = 158505,
  [SMALL_STATE(3265)] = 158613,
  [SMALL_STATE(3266)] = 158729,
  [SMALL_STATE(3267)] = 158797,
  [SMALL_STATE(3268)] = 158865,
  [SMALL_STATE(3269)] = 158933,
  [SMALL_STATE(3270)] = 159049,
  [SMALL_STATE(3271)] = 159117,
  [SMALL_STATE(3272)] = 159185,
  [SMALL_STATE(3273)] = 159253,
  [SMALL_STATE(3274)] = 159321,
  [SMALL_STATE(3275)] = 159389,
  [SMALL_STATE(3276)] = 159457,
  [SMALL_STATE(3277)] = 159525,
  [SMALL_STATE(3278)] = 159593,
  [SMALL_STATE(3279)] = 159661,
  [SMALL_STATE(3280)] = 159729,
  [SMALL_STATE(3281)] = 159797,
  [SMALL_STATE(3282)] = 159865,
  [SMALL_STATE(3283)] = 159933,
  [SMALL_STATE(3284)] = 160001,
  [SMALL_STATE(3285)] = 160069,
  [SMALL_STATE(3286)] = 160137,
  [SMALL_STATE(3287)] = 160205,
  [SMALL_STATE(3288)] = 160271,
  [SMALL_STATE(3289)] = 160337,
  [SMALL_STATE(3290)] = 160445,
  [SMALL_STATE(3291)] = 160511,
  [SMALL_STATE(3292)] = 160577,
  [SMALL_STATE(3293)] = 160647,
  [SMALL_STATE(3294)] = 160713,
  [SMALL_STATE(3295)] = 160779,
  [SMALL_STATE(3296)] = 160845,
  [SMALL_STATE(3297)] = 160911,
  [SMALL_STATE(3298)] = 160977,
  [SMALL_STATE(3299)] = 161043,
  [SMALL_STATE(3300)] = 161151,
  [SMALL_STATE(3301)] = 161259,
  [SMALL_STATE(3302)] = 161325,
  [SMALL_STATE(3303)] = 161391,
  [SMALL_STATE(3304)] = 161457,
  [SMALL_STATE(3305)] = 161523,
  [SMALL_STATE(3306)] = 161589,
  [SMALL_STATE(3307)] = 161655,
  [SMALL_STATE(3308)] = 161721,
  [SMALL_STATE(3309)] = 161787,
  [SMALL_STATE(3310)] = 161853,
  [SMALL_STATE(3311)] = 161919,
  [SMALL_STATE(3312)] = 161985,
  [SMALL_STATE(3313)] = 162051,
  [SMALL_STATE(3314)] = 162117,
  [SMALL_STATE(3315)] = 162183,
  [SMALL_STATE(3316)] = 162249,
  [SMALL_STATE(3317)] = 162315,
  [SMALL_STATE(3318)] = 162381,
  [SMALL_STATE(3319)] = 162447,
  [SMALL_STATE(3320)] = 162513,
  [SMALL_STATE(3321)] = 162579,
  [SMALL_STATE(3322)] = 162645,
  [SMALL_STATE(3323)] = 162711,
  [SMALL_STATE(3324)] = 162777,
  [SMALL_STATE(3325)] = 162843,
  [SMALL_STATE(3326)] = 162909,
  [SMALL_STATE(3327)] = 162975,
  [SMALL_STATE(3328)] = 163041,
  [SMALL_STATE(3329)] = 163107,
  [SMALL_STATE(3330)] = 163173,
  [SMALL_STATE(3331)] = 163239,
  [SMALL_STATE(3332)] = 163305,
  [SMALL_STATE(3333)] = 163371,
  [SMALL_STATE(3334)] = 163437,
  [SMALL_STATE(3335)] = 163503,
  [SMALL_STATE(3336)] = 163569,
  [SMALL_STATE(3337)] = 163635,
  [SMALL_STATE(3338)] = 163701,
  [SMALL_STATE(3339)] = 163767,
  [SMALL_STATE(3340)] = 163833,
  [SMALL_STATE(3341)] = 163899,
  [SMALL_STATE(3342)] = 163965,
  [SMALL_STATE(3343)] = 164031,
  [SMALL_STATE(3344)] = 164097,
  [SMALL_STATE(3345)] = 164163,
  [SMALL_STATE(3346)] = 164229,
  [SMALL_STATE(3347)] = 164295,
  [SMALL_STATE(3348)] = 164361,
  [SMALL_STATE(3349)] = 164427,
  [SMALL_STATE(3350)] = 164493,
  [SMALL_STATE(3351)] = 164559,
  [SMALL_STATE(3352)] = 164625,
  [SMALL_STATE(3353)] = 164691,
  [SMALL_STATE(3354)] = 164757,
  [SMALL_STATE(3355)] = 164823,
  [SMALL_STATE(3356)] = 164889,
  [SMALL_STATE(3357)] = 164955,
  [SMALL_STATE(3358)] = 165021,
  [SMALL_STATE(3359)] = 165087,
  [SMALL_STATE(3360)] = 165153,
  [SMALL_STATE(3361)] = 165219,
  [SMALL_STATE(3362)] = 165285,
  [SMALL_STATE(3363)] = 165351,
  [SMALL_STATE(3364)] = 165417,
  [SMALL_STATE(3365)] = 165483,
  [SMALL_STATE(3366)] = 165549,
  [SMALL_STATE(3367)] = 165615,
  [SMALL_STATE(3368)] = 165681,
  [SMALL_STATE(3369)] = 165747,
  [SMALL_STATE(3370)] = 165813,
  [SMALL_STATE(3371)] = 165879,
  [SMALL_STATE(3372)] = 165945,
  [SMALL_STATE(3373)] = 166053,
  [SMALL_STATE(3374)] = 166119,
  [SMALL_STATE(3375)] = 166189,
  [SMALL_STATE(3376)] = 166255,
  [SMALL_STATE(3377)] = 166321,
  [SMALL_STATE(3378)] = 166387,
  [SMALL_STATE(3379)] = 166453,
  [SMALL_STATE(3380)] = 166519,
  [SMALL_STATE(3381)] = 166585,
  [SMALL_STATE(3382)] = 166651,
  [SMALL_STATE(3383)] = 166717,
  [SMALL_STATE(3384)] = 166783,
  [SMALL_STATE(3385)] = 166849,
  [SMALL_STATE(3386)] = 166915,
  [SMALL_STATE(3387)] = 166985,
  [SMALL_STATE(3388)] = 167051,
  [SMALL_STATE(3389)] = 167117,
  [SMALL_STATE(3390)] = 167183,
  [SMALL_STATE(3391)] = 167249,
  [SMALL_STATE(3392)] = 167315,
  [SMALL_STATE(3393)] = 167381,
  [SMALL_STATE(3394)] = 167447,
  [SMALL_STATE(3395)] = 167513,
  [SMALL_STATE(3396)] = 167579,
  [SMALL_STATE(3397)] = 167645,
  [SMALL_STATE(3398)] = 167711,
  [SMALL_STATE(3399)] = 167781,
  [SMALL_STATE(3400)] = 167847,
  [SMALL_STATE(3401)] = 167913,
  [SMALL_STATE(3402)] = 167979,
  [SMALL_STATE(3403)] = 168045,
  [SMALL_STATE(3404)] = 168111,
  [SMALL_STATE(3405)] = 168177,
  [SMALL_STATE(3406)] = 168243,
  [SMALL_STATE(3407)] = 168309,
  [SMALL_STATE(3408)] = 168375,
  [SMALL_STATE(3409)] = 168441,
  [SMALL_STATE(3410)] = 168507,
  [SMALL_STATE(3411)] = 168573,
  [SMALL_STATE(3412)] = 168639,
  [SMALL_STATE(3413)] = 168705,
  [SMALL_STATE(3414)] = 168771,
  [SMALL_STATE(3415)] = 168837,
  [SMALL_STATE(3416)] = 168903,
  [SMALL_STATE(3417)] = 168969,
  [SMALL_STATE(3418)] = 169035,
  [SMALL_STATE(3419)] = 169101,
  [SMALL_STATE(3420)] = 169167,
  [SMALL_STATE(3421)] = 169233,
  [SMALL_STATE(3422)] = 169299,
  [SMALL_STATE(3423)] = 169365,
  [SMALL_STATE(3424)] = 169431,
  [SMALL_STATE(3425)] = 169497,
  [SMALL_STATE(3426)] = 169563,
  [SMALL_STATE(3427)] = 169629,
  [SMALL_STATE(3428)] = 169695,
  [SMALL_STATE(3429)] = 169761,
  [SMALL_STATE(3430)] = 169827,
  [SMALL_STATE(3431)] = 169893,
  [SMALL_STATE(3432)] = 169959,
  [SMALL_STATE(3433)] = 170025,
  [SMALL_STATE(3434)] = 170091,
  [SMALL_STATE(3435)] = 170157,
  [SMALL_STATE(3436)] = 170223,
  [SMALL_STATE(3437)] = 170289,
  [SMALL_STATE(3438)] = 170355,
  [SMALL_STATE(3439)] = 170421,
  [SMALL_STATE(3440)] = 170487,
  [SMALL_STATE(3441)] = 170553,
  [SMALL_STATE(3442)] = 170619,
  [SMALL_STATE(3443)] = 170685,
  [SMALL_STATE(3444)] = 170751,
  [SMALL_STATE(3445)] = 170817,
  [SMALL_STATE(3446)] = 170883,
  [SMALL_STATE(3447)] = 170949,
  [SMALL_STATE(3448)] = 171015,
  [SMALL_STATE(3449)] = 171081,
  [SMALL_STATE(3450)] = 171147,
  [SMALL_STATE(3451)] = 171213,
  [SMALL_STATE(3452)] = 171279,
  [SMALL_STATE(3453)] = 171345,
  [SMALL_STATE(3454)] = 171411,
  [SMALL_STATE(3455)] = 171477,
  [SMALL_STATE(3456)] = 171543,
  [SMALL_STATE(3457)] = 171609,
  [SMALL_STATE(3458)] = 171675,
  [SMALL_STATE(3459)] = 171741,
  [SMALL_STATE(3460)] = 171807,
  [SMALL_STATE(3461)] = 171919,
  [SMALL_STATE(3462)] = 171989,
  [SMALL_STATE(3463)] = 172055,
  [SMALL_STATE(3464)] = 172125,
  [SMALL_STATE(3465)] = 172191,
  [SMALL_STATE(3466)] = 172257,
  [SMALL_STATE(3467)] = 172323,
  [SMALL_STATE(3468)] = 172389,
  [SMALL_STATE(3469)] = 172455,
  [SMALL_STATE(3470)] = 172525,
  [SMALL_STATE(3471)] = 172595,
  [SMALL_STATE(3472)] = 172665,
  [SMALL_STATE(3473)] = 172731,
  [SMALL_STATE(3474)] = 172797,
  [SMALL_STATE(3475)] = 172863,
  [SMALL_STATE(3476)] = 172929,
  [SMALL_STATE(3477)] = 172995,
  [SMALL_STATE(3478)] = 173061,
  [SMALL_STATE(3479)] = 173127,
  [SMALL_STATE(3480)] = 173193,
  [SMALL_STATE(3481)] = 173259,
  [SMALL_STATE(3482)] = 173367,
  [SMALL_STATE(3483)] = 173433,
  [SMALL_STATE(3484)] = 173503,
  [SMALL_STATE(3485)] = 173569,
  [SMALL_STATE(3486)] = 173635,
  [SMALL_STATE(3487)] = 173701,
  [SMALL_STATE(3488)] = 173767,
  [SMALL_STATE(3489)] = 173833,
  [SMALL_STATE(3490)] = 173899,
  [SMALL_STATE(3491)] = 173965,
  [SMALL_STATE(3492)] = 174031,
  [SMALL_STATE(3493)] = 174097,
  [SMALL_STATE(3494)] = 174205,
  [SMALL_STATE(3495)] = 174271,
  [SMALL_STATE(3496)] = 174341,
  [SMALL_STATE(3497)] = 174407,
  [SMALL_STATE(3498)] = 174473,
  [SMALL_STATE(3499)] = 174539,
  [SMALL_STATE(3500)] = 174609,
  [SMALL_STATE(3501)] = 174679,
  [SMALL_STATE(3502)] = 174749,
  [SMALL_STATE(3503)] = 174815,
  [SMALL_STATE(3504)] = 174881,
  [SMALL_STATE(3505)] = 174947,
  [SMALL_STATE(3506)] = 175013,
  [SMALL_STATE(3507)] = 175079,
  [SMALL_STATE(3508)] = 175145,
  [SMALL_STATE(3509)] = 175215,
  [SMALL_STATE(3510)] = 175280,
  [SMALL_STATE(3511)] = 175345,
  [SMALL_STATE(3512)] = 175410,
  [SMALL_STATE(3513)] = 175475,
  [SMALL_STATE(3514)] = 175540,
  [SMALL_STATE(3515)] = 175605,
  [SMALL_STATE(3516)] = 175670,
  [SMALL_STATE(3517)] = 175735,
  [SMALL_STATE(3518)] = 175802,
  [SMALL_STATE(3519)] = 175871,
  [SMALL_STATE(3520)] = 175938,
  [SMALL_STATE(3521)] = 176003,
  [SMALL_STATE(3522)] = 176068,
  [SMALL_STATE(3523)] = 176133,
  [SMALL_STATE(3524)] = 176198,
  [SMALL_STATE(3525)] = 176263,
  [SMALL_STATE(3526)] = 176328,
  [SMALL_STATE(3527)] = 176393,
  [SMALL_STATE(3528)] = 176458,
  [SMALL_STATE(3529)] = 176523,
  [SMALL_STATE(3530)] = 176588,
  [SMALL_STATE(3531)] = 176653,
  [SMALL_STATE(3532)] = 176718,
  [SMALL_STATE(3533)] = 176787,
  [SMALL_STATE(3534)] = 176852,
  [SMALL_STATE(3535)] = 176917,
  [SMALL_STATE(3536)] = 176982,
  [SMALL_STATE(3537)] = 177047,
  [SMALL_STATE(3538)] = 177112,
  [SMALL_STATE(3539)] = 177219,
  [SMALL_STATE(3540)] = 177284,
  [SMALL_STATE(3541)] = 177349,
  [SMALL_STATE(3542)] = 177414,
  [SMALL_STATE(3543)] = 177479,
  [SMALL_STATE(3544)] = 177548,
  [SMALL_STATE(3545)] = 177655,
  [SMALL_STATE(3546)] = 177720,
  [SMALL_STATE(3547)] = 177787,
  [SMALL_STATE(3548)] = 177852,
  [SMALL_STATE(3549)] = 177917,
  [SMALL_STATE(3550)] = 177982,
  [SMALL_STATE(3551)] = 178049,
  [SMALL_STATE(3552)] = 178114,
  [SMALL_STATE(3553)] = 178179,
  [SMALL_STATE(3554)] = 178246,
  [SMALL_STATE(3555)] = 178313,
  [SMALL_STATE(3556)] = 178380,
  [SMALL_STATE(3557)] = 178447,
  [SMALL_STATE(3558)] = 178514,
  [SMALL_STATE(3559)] = 178581,
  [SMALL_STATE(3560)] = 178648,
  [SMALL_STATE(3561)] = 178715,
  [SMALL_STATE(3562)] = 178782,
  [SMALL_STATE(3563)] = 178849,
  [SMALL_STATE(3564)] = 178916,
  [SMALL_STATE(3565)] = 178983,
  [SMALL_STATE(3566)] = 179050,
  [SMALL_STATE(3567)] = 179117,
  [SMALL_STATE(3568)] = 179184,
  [SMALL_STATE(3569)] = 179251,
  [SMALL_STATE(3570)] = 179318,
  [SMALL_STATE(3571)] = 179385,
  [SMALL_STATE(3572)] = 179452,
  [SMALL_STATE(3573)] = 179519,
  [SMALL_STATE(3574)] = 179584,
  [SMALL_STATE(3575)] = 179649,
  [SMALL_STATE(3576)] = 179714,
  [SMALL_STATE(3577)] = 179779,
  [SMALL_STATE(3578)] = 179844,
  [SMALL_STATE(3579)] = 179909,
  [SMALL_STATE(3580)] = 179974,
  [SMALL_STATE(3581)] = 180039,
  [SMALL_STATE(3582)] = 180104,
  [SMALL_STATE(3583)] = 180169,
  [SMALL_STATE(3584)] = 180234,
  [SMALL_STATE(3585)] = 180299,
  [SMALL_STATE(3586)] = 180364,
  [SMALL_STATE(3587)] = 180429,
  [SMALL_STATE(3588)] = 180494,
  [SMALL_STATE(3589)] = 180559,
  [SMALL_STATE(3590)] = 180624,
  [SMALL_STATE(3591)] = 180691,
  [SMALL_STATE(3592)] = 180756,
  [SMALL_STATE(3593)] = 180821,
  [SMALL_STATE(3594)] = 180888,
  [SMALL_STATE(3595)] = 180955,
  [SMALL_STATE(3596)] = 181020,
  [SMALL_STATE(3597)] = 181085,
  [SMALL_STATE(3598)] = 181150,
  [SMALL_STATE(3599)] = 181215,
  [SMALL_STATE(3600)] = 181280,
  [SMALL_STATE(3601)] = 181345,
  [SMALL_STATE(3602)] = 181410,
  [SMALL_STATE(3603)] = 181475,
  [SMALL_STATE(3604)] = 181540,
  [SMALL_STATE(3605)] = 181605,
  [SMALL_STATE(3606)] = 181670,
  [SMALL_STATE(3607)] = 181735,
  [SMALL_STATE(3608)] = 181800,
  [SMALL_STATE(3609)] = 181865,
  [SMALL_STATE(3610)] = 181930,
  [SMALL_STATE(3611)] = 181995,
  [SMALL_STATE(3612)] = 182060,
  [SMALL_STATE(3613)] = 182125,
  [SMALL_STATE(3614)] = 182190,
  [SMALL_STATE(3615)] = 182255,
  [SMALL_STATE(3616)] = 182320,
  [SMALL_STATE(3617)] = 182385,
  [SMALL_STATE(3618)] = 182452,
  [SMALL_STATE(3619)] = 182517,
  [SMALL_STATE(3620)] = 182582,
  [SMALL_STATE(3621)] = 182647,
  [SMALL_STATE(3622)] = 182714,
  [SMALL_STATE(3623)] = 182779,
  [SMALL_STATE(3624)] = 182844,
  [SMALL_STATE(3625)] = 182909,
  [SMALL_STATE(3626)] = 182974,
  [SMALL_STATE(3627)] = 183039,
  [SMALL_STATE(3628)] = 183104,
  [SMALL_STATE(3629)] = 183171,
  [SMALL_STATE(3630)] = 183238,
  [SMALL_STATE(3631)] = 183305,
  [SMALL_STATE(3632)] = 183372,
  [SMALL_STATE(3633)] = 183437,
  [SMALL_STATE(3634)] = 183504,
  [SMALL_STATE(3635)] = 183571,
  [SMALL_STATE(3636)] = 183636,
  [SMALL_STATE(3637)] = 183701,
  [SMALL_STATE(3638)] = 183766,
  [SMALL_STATE(3639)] = 183831,
  [SMALL_STATE(3640)] = 183896,
  [SMALL_STATE(3641)] = 183961,
  [SMALL_STATE(3642)] = 184026,
  [SMALL_STATE(3643)] = 184091,
  [SMALL_STATE(3644)] = 184156,
  [SMALL_STATE(3645)] = 184221,
  [SMALL_STATE(3646)] = 184286,
  [SMALL_STATE(3647)] = 184351,
  [SMALL_STATE(3648)] = 184416,
  [SMALL_STATE(3649)] = 184481,
  [SMALL_STATE(3650)] = 184546,
  [SMALL_STATE(3651)] = 184611,
  [SMALL_STATE(3652)] = 184676,
  [SMALL_STATE(3653)] = 184741,
  [SMALL_STATE(3654)] = 184806,
  [SMALL_STATE(3655)] = 184871,
  [SMALL_STATE(3656)] = 184936,
  [SMALL_STATE(3657)] = 185001,
  [SMALL_STATE(3658)] = 185066,
  [SMALL_STATE(3659)] = 185131,
  [SMALL_STATE(3660)] = 185196,
  [SMALL_STATE(3661)] = 185261,
  [SMALL_STATE(3662)] = 185326,
  [SMALL_STATE(3663)] = 185391,
  [SMALL_STATE(3664)] = 185456,
  [SMALL_STATE(3665)] = 185521,
  [SMALL_STATE(3666)] = 185586,
  [SMALL_STATE(3667)] = 185653,
  [SMALL_STATE(3668)] = 185720,
  [SMALL_STATE(3669)] = 185787,
  [SMALL_STATE(3670)] = 185852,
  [SMALL_STATE(3671)] = 185917,
  [SMALL_STATE(3672)] = 185982,
  [SMALL_STATE(3673)] = 186047,
  [SMALL_STATE(3674)] = 186112,
  [SMALL_STATE(3675)] = 186177,
  [SMALL_STATE(3676)] = 186242,
  [SMALL_STATE(3677)] = 186307,
  [SMALL_STATE(3678)] = 186414,
  [SMALL_STATE(3679)] = 186479,
  [SMALL_STATE(3680)] = 186548,
  [SMALL_STATE(3681)] = 186613,
  [SMALL_STATE(3682)] = 186678,
  [SMALL_STATE(3683)] = 186743,
  [SMALL_STATE(3684)] = 186808,
  [SMALL_STATE(3685)] = 186875,
  [SMALL_STATE(3686)] = 186940,
  [SMALL_STATE(3687)] = 187005,
  [SMALL_STATE(3688)] = 187070,
  [SMALL_STATE(3689)] = 187135,
  [SMALL_STATE(3690)] = 187200,
  [SMALL_STATE(3691)] = 187265,
  [SMALL_STATE(3692)] = 187330,
  [SMALL_STATE(3693)] = 187395,
  [SMALL_STATE(3694)] = 187462,
  [SMALL_STATE(3695)] = 187527,
  [SMALL_STATE(3696)] = 187594,
  [SMALL_STATE(3697)] = 187661,
  [SMALL_STATE(3698)] = 187726,
  [SMALL_STATE(3699)] = 187793,
  [SMALL_STATE(3700)] = 187858,
  [SMALL_STATE(3701)] = 187925,
  [SMALL_STATE(3702)] = 187992,
  [SMALL_STATE(3703)] = 188057,
  [SMALL_STATE(3704)] = 188122,
  [SMALL_STATE(3705)] = 188191,
  [SMALL_STATE(3706)] = 188256,
  [SMALL_STATE(3707)] = 188321,
  [SMALL_STATE(3708)] = 188388,
  [SMALL_STATE(3709)] = 188455,
  [SMALL_STATE(3710)] = 188520,
  [SMALL_STATE(3711)] = 188585,
  [SMALL_STATE(3712)] = 188650,
  [SMALL_STATE(3713)] = 188715,
  [SMALL_STATE(3714)] = 188780,
  [SMALL_STATE(3715)] = 188845,
  [SMALL_STATE(3716)] = 188912,
  [SMALL_STATE(3717)] = 188977,
  [SMALL_STATE(3718)] = 189042,
  [SMALL_STATE(3719)] = 189107,
  [SMALL_STATE(3720)] = 189188,
  [SMALL_STATE(3721)] = 189275,
  [SMALL_STATE(3722)] = 189376,
  [SMALL_STATE(3723)] = 189441,
  [SMALL_STATE(3724)] = 189512,
  [SMALL_STATE(3725)] = 189589,
  [SMALL_STATE(3726)] = 189662,
  [SMALL_STATE(3727)] = 189767,
  [SMALL_STATE(3728)] = 189872,
  [SMALL_STATE(3729)] = 189937,
  [SMALL_STATE(3730)] = 190040,
  [SMALL_STATE(3731)] = 190139,
  [SMALL_STATE(3732)] = 190236,
  [SMALL_STATE(3733)] = 190329,
  [SMALL_STATE(3734)] = 190420,
  [SMALL_STATE(3735)] = 190509,
  [SMALL_STATE(3736)] = 190594,
  [SMALL_STATE(3737)] = 190673,
  [SMALL_STATE(3738)] = 190752,
  [SMALL_STATE(3739)] = 190821,
  [SMALL_STATE(3740)] = 190886,
  [SMALL_STATE(3741)] = 190951,
  [SMALL_STATE(3742)] = 191016,
  [SMALL_STATE(3743)] = 191081,
  [SMALL_STATE(3744)] = 191146,
  [SMALL_STATE(3745)] = 191211,
  [SMALL_STATE(3746)] = 191276,
  [SMALL_STATE(3747)] = 191341,
  [SMALL_STATE(3748)] = 191406,
  [SMALL_STATE(3749)] = 191471,
  [SMALL_STATE(3750)] = 191536,
  [SMALL_STATE(3751)] = 191601,
  [SMALL_STATE(3752)] = 191666,
  [SMALL_STATE(3753)] = 191731,
  [SMALL_STATE(3754)] = 191796,
  [SMALL_STATE(3755)] = 191861,
  [SMALL_STATE(3756)] = 191930,
  [SMALL_STATE(3757)] = 191995,
  [SMALL_STATE(3758)] = 192060,
  [SMALL_STATE(3759)] = 192125,
  [SMALL_STATE(3760)] = 192190,
  [SMALL_STATE(3761)] = 192271,
  [SMALL_STATE(3762)] = 192358,
  [SMALL_STATE(3763)] = 192459,
  [SMALL_STATE(3764)] = 192530,
  [SMALL_STATE(3765)] = 192607,
  [SMALL_STATE(3766)] = 192680,
  [SMALL_STATE(3767)] = 192785,
  [SMALL_STATE(3768)] = 192890,
  [SMALL_STATE(3769)] = 192993,
  [SMALL_STATE(3770)] = 193092,
  [SMALL_STATE(3771)] = 193189,
  [SMALL_STATE(3772)] = 193282,
  [SMALL_STATE(3773)] = 193373,
  [SMALL_STATE(3774)] = 193462,
  [SMALL_STATE(3775)] = 193547,
  [SMALL_STATE(3776)] = 193626,
  [SMALL_STATE(3777)] = 193705,
  [SMALL_STATE(3778)] = 193770,
  [SMALL_STATE(3779)] = 193835,
  [SMALL_STATE(3780)] = 193900,
  [SMALL_STATE(3781)] = 193965,
  [SMALL_STATE(3782)] = 194030,
  [SMALL_STATE(3783)] = 194095,
  [SMALL_STATE(3784)] = 194160,
  [SMALL_STATE(3785)] = 194225,
  [SMALL_STATE(3786)] = 194290,
  [SMALL_STATE(3787)] = 194355,
  [SMALL_STATE(3788)] = 194420,
  [SMALL_STATE(3789)] = 194487,
  [SMALL_STATE(3790)] = 194552,
  [SMALL_STATE(3791)] = 194619,
  [SMALL_STATE(3792)] = 194684,
  [SMALL_STATE(3793)] = 194749,
  [SMALL_STATE(3794)] = 194814,
  [SMALL_STATE(3795)] = 194879,
  [SMALL_STATE(3796)] = 194944,
  [SMALL_STATE(3797)] = 195009,
  [SMALL_STATE(3798)] = 195074,
  [SMALL_STATE(3799)] = 195139,
  [SMALL_STATE(3800)] = 195204,
  [SMALL_STATE(3801)] = 195269,
  [SMALL_STATE(3802)] = 195334,
  [SMALL_STATE(3803)] = 195399,
  [SMALL_STATE(3804)] = 195464,
  [SMALL_STATE(3805)] = 195529,
  [SMALL_STATE(3806)] = 195594,
  [SMALL_STATE(3807)] = 195661,
  [SMALL_STATE(3808)] = 195726,
  [SMALL_STATE(3809)] = 195791,
  [SMALL_STATE(3810)] = 195856,
  [SMALL_STATE(3811)] = 195921,
  [SMALL_STATE(3812)] = 195986,
  [SMALL_STATE(3813)] = 196051,
  [SMALL_STATE(3814)] = 196116,
  [SMALL_STATE(3815)] = 196185,
  [SMALL_STATE(3816)] = 196250,
  [SMALL_STATE(3817)] = 196319,
  [SMALL_STATE(3818)] = 196386,
  [SMALL_STATE(3819)] = 196453,
  [SMALL_STATE(3820)] = 196520,
  [SMALL_STATE(3821)] = 196585,
  [SMALL_STATE(3822)] = 196650,
  [SMALL_STATE(3823)] = 196715,
  [SMALL_STATE(3824)] = 196780,
  [SMALL_STATE(3825)] = 196845,
  [SMALL_STATE(3826)] = 196910,
  [SMALL_STATE(3827)] = 196975,
  [SMALL_STATE(3828)] = 197040,
  [SMALL_STATE(3829)] = 197105,
  [SMALL_STATE(3830)] = 197170,
  [SMALL_STATE(3831)] = 197275,
  [SMALL_STATE(3832)] = 197340,
  [SMALL_STATE(3833)] = 197405,
  [SMALL_STATE(3834)] = 197470,
  [SMALL_STATE(3835)] = 197535,
  [SMALL_STATE(3836)] = 197600,
  [SMALL_STATE(3837)] = 197665,
  [SMALL_STATE(3838)] = 197730,
  [SMALL_STATE(3839)] = 197795,
  [SMALL_STATE(3840)] = 197860,
  [SMALL_STATE(3841)] = 197925,
  [SMALL_STATE(3842)] = 197990,
  [SMALL_STATE(3843)] = 198055,
  [SMALL_STATE(3844)] = 198120,
  [SMALL_STATE(3845)] = 198185,
  [SMALL_STATE(3846)] = 198250,
  [SMALL_STATE(3847)] = 198315,
  [SMALL_STATE(3848)] = 198380,
  [SMALL_STATE(3849)] = 198445,
  [SMALL_STATE(3850)] = 198510,
  [SMALL_STATE(3851)] = 198575,
  [SMALL_STATE(3852)] = 198644,
  [SMALL_STATE(3853)] = 198709,
  [SMALL_STATE(3854)] = 198774,
  [SMALL_STATE(3855)] = 198839,
  [SMALL_STATE(3856)] = 198904,
  [SMALL_STATE(3857)] = 198969,
  [SMALL_STATE(3858)] = 199034,
  [SMALL_STATE(3859)] = 199099,
  [SMALL_STATE(3860)] = 199204,
  [SMALL_STATE(3861)] = 199269,
  [SMALL_STATE(3862)] = 199334,
  [SMALL_STATE(3863)] = 199443,
  [SMALL_STATE(3864)] = 199508,
  [SMALL_STATE(3865)] = 199573,
  [SMALL_STATE(3866)] = 199638,
  [SMALL_STATE(3867)] = 199703,
  [SMALL_STATE(3868)] = 199768,
  [SMALL_STATE(3869)] = 199833,
  [SMALL_STATE(3870)] = 199898,
  [SMALL_STATE(3871)] = 199963,
  [SMALL_STATE(3872)] = 200028,
  [SMALL_STATE(3873)] = 200093,
  [SMALL_STATE(3874)] = 200158,
  [SMALL_STATE(3875)] = 200223,
  [SMALL_STATE(3876)] = 200288,
  [SMALL_STATE(3877)] = 200353,
  [SMALL_STATE(3878)] = 200418,
  [SMALL_STATE(3879)] = 200483,
  [SMALL_STATE(3880)] = 200548,
  [SMALL_STATE(3881)] = 200613,
  [SMALL_STATE(3882)] = 200678,
  [SMALL_STATE(3883)] = 200743,
  [SMALL_STATE(3884)] = 200808,
  [SMALL_STATE(3885)] = 200873,
  [SMALL_STATE(3886)] = 200938,
  [SMALL_STATE(3887)] = 201005,
  [SMALL_STATE(3888)] = 201070,
  [SMALL_STATE(3889)] = 201135,
  [SMALL_STATE(3890)] = 201200,
  [SMALL_STATE(3891)] = 201265,
  [SMALL_STATE(3892)] = 201330,
  [SMALL_STATE(3893)] = 201395,
  [SMALL_STATE(3894)] = 201460,
  [SMALL_STATE(3895)] = 201525,
  [SMALL_STATE(3896)] = 201632,
  [SMALL_STATE(3897)] = 201697,
  [SMALL_STATE(3898)] = 201762,
  [SMALL_STATE(3899)] = 201827,
  [SMALL_STATE(3900)] = 201892,
  [SMALL_STATE(3901)] = 201957,
  [SMALL_STATE(3902)] = 202022,
  [SMALL_STATE(3903)] = 202089,
  [SMALL_STATE(3904)] = 202156,
  [SMALL_STATE(3905)] = 202223,
  [SMALL_STATE(3906)] = 202290,
  [SMALL_STATE(3907)] = 202357,
  [SMALL_STATE(3908)] = 202424,
  [SMALL_STATE(3909)] = 202491,
  [SMALL_STATE(3910)] = 202558,
  [SMALL_STATE(3911)] = 202625,
  [SMALL_STATE(3912)] = 202692,
  [SMALL_STATE(3913)] = 202759,
  [SMALL_STATE(3914)] = 202826,
  [SMALL_STATE(3915)] = 202893,
  [SMALL_STATE(3916)] = 202960,
  [SMALL_STATE(3917)] = 203027,
  [SMALL_STATE(3918)] = 203094,
  [SMALL_STATE(3919)] = 203161,
  [SMALL_STATE(3920)] = 203228,
  [SMALL_STATE(3921)] = 203295,
  [SMALL_STATE(3922)] = 203362,
  [SMALL_STATE(3923)] = 203427,
  [SMALL_STATE(3924)] = 203492,
  [SMALL_STATE(3925)] = 203557,
  [SMALL_STATE(3926)] = 203622,
  [SMALL_STATE(3927)] = 203687,
  [SMALL_STATE(3928)] = 203768,
  [SMALL_STATE(3929)] = 203853,
  [SMALL_STATE(3930)] = 203918,
  [SMALL_STATE(3931)] = 203983,
  [SMALL_STATE(3932)] = 204048,
  [SMALL_STATE(3933)] = 204113,
  [SMALL_STATE(3934)] = 204178,
  [SMALL_STATE(3935)] = 204243,
  [SMALL_STATE(3936)] = 204308,
  [SMALL_STATE(3937)] = 204373,
  [SMALL_STATE(3938)] = 204438,
  [SMALL_STATE(3939)] = 204503,
  [SMALL_STATE(3940)] = 204568,
  [SMALL_STATE(3941)] = 204633,
  [SMALL_STATE(3942)] = 204698,
  [SMALL_STATE(3943)] = 204763,
  [SMALL_STATE(3944)] = 204828,
  [SMALL_STATE(3945)] = 204893,
  [SMALL_STATE(3946)] = 204958,
  [SMALL_STATE(3947)] = 205023,
  [SMALL_STATE(3948)] = 205088,
  [SMALL_STATE(3949)] = 205153,
  [SMALL_STATE(3950)] = 205218,
  [SMALL_STATE(3951)] = 205283,
  [SMALL_STATE(3952)] = 205348,
  [SMALL_STATE(3953)] = 205447,
  [SMALL_STATE(3954)] = 205512,
  [SMALL_STATE(3955)] = 205583,
  [SMALL_STATE(3956)] = 205660,
  [SMALL_STATE(3957)] = 205733,
  [SMALL_STATE(3958)] = 205836,
  [SMALL_STATE(3959)] = 205901,
  [SMALL_STATE(3960)] = 205966,
  [SMALL_STATE(3961)] = 206031,
  [SMALL_STATE(3962)] = 206096,
  [SMALL_STATE(3963)] = 206161,
  [SMALL_STATE(3964)] = 206226,
  [SMALL_STATE(3965)] = 206291,
  [SMALL_STATE(3966)] = 206356,
  [SMALL_STATE(3967)] = 206421,
  [SMALL_STATE(3968)] = 206486,
  [SMALL_STATE(3969)] = 206589,
  [SMALL_STATE(3970)] = 206694,
  [SMALL_STATE(3971)] = 206795,
  [SMALL_STATE(3972)] = 206892,
  [SMALL_STATE(3973)] = 206987,
  [SMALL_STATE(3974)] = 207078,
  [SMALL_STATE(3975)] = 207167,
  [SMALL_STATE(3976)] = 207254,
  [SMALL_STATE(3977)] = 207337,
  [SMALL_STATE(3978)] = 207416,
  [SMALL_STATE(3979)] = 207495,
  [SMALL_STATE(3980)] = 207560,
  [SMALL_STATE(3981)] = 207625,
  [SMALL_STATE(3982)] = 207690,
  [SMALL_STATE(3983)] = 207759,
  [SMALL_STATE(3984)] = 207828,
  [SMALL_STATE(3985)] = 207893,
  [SMALL_STATE(3986)] = 207960,
  [SMALL_STATE(3987)] = 208027,
  [SMALL_STATE(3988)] = 208134,
  [SMALL_STATE(3989)] = 208199,
  [SMALL_STATE(3990)] = 208264,
  [SMALL_STATE(3991)] = 208329,
  [SMALL_STATE(3992)] = 208394,
  [SMALL_STATE(3993)] = 208459,
  [SMALL_STATE(3994)] = 208524,
  [SMALL_STATE(3995)] = 208589,
  [SMALL_STATE(3996)] = 208654,
  [SMALL_STATE(3997)] = 208719,
  [SMALL_STATE(3998)] = 208784,
  [SMALL_STATE(3999)] = 208849,
  [SMALL_STATE(4000)] = 208914,
  [SMALL_STATE(4001)] = 208979,
  [SMALL_STATE(4002)] = 209044,
  [SMALL_STATE(4003)] = 209109,
  [SMALL_STATE(4004)] = 209174,
  [SMALL_STATE(4005)] = 209239,
  [SMALL_STATE(4006)] = 209304,
  [SMALL_STATE(4007)] = 209369,
  [SMALL_STATE(4008)] = 209434,
  [SMALL_STATE(4009)] = 209499,
  [SMALL_STATE(4010)] = 209564,
  [SMALL_STATE(4011)] = 209629,
  [SMALL_STATE(4012)] = 209694,
  [SMALL_STATE(4013)] = 209759,
  [SMALL_STATE(4014)] = 209824,
  [SMALL_STATE(4015)] = 209889,
  [SMALL_STATE(4016)] = 209954,
  [SMALL_STATE(4017)] = 210019,
  [SMALL_STATE(4018)] = 210084,
  [SMALL_STATE(4019)] = 210149,
  [SMALL_STATE(4020)] = 210214,
  [SMALL_STATE(4021)] = 210279,
  [SMALL_STATE(4022)] = 210344,
  [SMALL_STATE(4023)] = 210425,
  [SMALL_STATE(4024)] = 210512,
  [SMALL_STATE(4025)] = 210613,
  [SMALL_STATE(4026)] = 210678,
  [SMALL_STATE(4027)] = 210749,
  [SMALL_STATE(4028)] = 210826,
  [SMALL_STATE(4029)] = 210899,
  [SMALL_STATE(4030)] = 211006,
  [SMALL_STATE(4031)] = 211113,
  [SMALL_STATE(4032)] = 211182,
  [SMALL_STATE(4033)] = 211289,
  [SMALL_STATE(4034)] = 211358,
  [SMALL_STATE(4035)] = 211427,
  [SMALL_STATE(4036)] = 211496,
  [SMALL_STATE(4037)] = 211565,
  [SMALL_STATE(4038)] = 211634,
  [SMALL_STATE(4039)] = 211739,
  [SMALL_STATE(4040)] = 211808,
  [SMALL_STATE(4041)] = 211913,
  [SMALL_STATE(4042)] = 211978,
  [SMALL_STATE(4043)] = 212081,
  [SMALL_STATE(4044)] = 212180,
  [SMALL_STATE(4045)] = 212245,
  [SMALL_STATE(4046)] = 212312,
  [SMALL_STATE(4047)] = 212379,
  [SMALL_STATE(4048)] = 212446,
  [SMALL_STATE(4049)] = 212515,
  [SMALL_STATE(4050)] = 212582,
  [SMALL_STATE(4051)] = 212679,
  [SMALL_STATE(4052)] = 212772,
  [SMALL_STATE(4053)] = 212863,
  [SMALL_STATE(4054)] = 212952,
  [SMALL_STATE(4055)] = 213037,
  [SMALL_STATE(4056)] = 213116,
  [SMALL_STATE(4057)] = 213195,
  [SMALL_STATE(4058)] = 213264,
  [SMALL_STATE(4059)] = 213331,
  [SMALL_STATE(4060)] = 213398,
  [SMALL_STATE(4061)] = 213465,
  [SMALL_STATE(4062)] = 213529,
  [SMALL_STATE(4063)] = 213593,
  [SMALL_STATE(4064)] = 213657,
  [SMALL_STATE(4065)] = 213721,
  [SMALL_STATE(4066)] = 213801,
  [SMALL_STATE(4067)] = 213887,
  [SMALL_STATE(4068)] = 213987,
  [SMALL_STATE(4069)] = 214057,
  [SMALL_STATE(4070)] = 214133,
  [SMALL_STATE(4071)] = 214205,
  [SMALL_STATE(4072)] = 214309,
  [SMALL_STATE(4073)] = 214413,
  [SMALL_STATE(4074)] = 214515,
  [SMALL_STATE(4075)] = 214613,
  [SMALL_STATE(4076)] = 214709,
  [SMALL_STATE(4077)] = 214801,
  [SMALL_STATE(4078)] = 214891,
  [SMALL_STATE(4079)] = 214979,
  [SMALL_STATE(4080)] = 215063,
  [SMALL_STATE(4081)] = 215127,
  [SMALL_STATE(4082)] = 215191,
  [SMALL_STATE(4083)] = 215255,
  [SMALL_STATE(4084)] = 215319,
  [SMALL_STATE(4085)] = 215383,
  [SMALL_STATE(4086)] = 215447,
  [SMALL_STATE(4087)] = 215511,
  [SMALL_STATE(4088)] = 215575,
  [SMALL_STATE(4089)] = 215639,
  [SMALL_STATE(4090)] = 215703,
  [SMALL_STATE(4091)] = 215767,
  [SMALL_STATE(4092)] = 215831,
  [SMALL_STATE(4093)] = 215895,
  [SMALL_STATE(4094)] = 215959,
  [SMALL_STATE(4095)] = 216023,
  [SMALL_STATE(4096)] = 216087,
  [SMALL_STATE(4097)] = 216151,
  [SMALL_STATE(4098)] = 216215,
  [SMALL_STATE(4099)] = 216279,
  [SMALL_STATE(4100)] = 216343,
  [SMALL_STATE(4101)] = 216407,
  [SMALL_STATE(4102)] = 216471,
  [SMALL_STATE(4103)] = 216535,
  [SMALL_STATE(4104)] = 216613,
  [SMALL_STATE(4105)] = 216691,
  [SMALL_STATE(4106)] = 216797,
  [SMALL_STATE(4107)] = 216861,
  [SMALL_STATE(4108)] = 216929,
  [SMALL_STATE(4109)] = 216993,
  [SMALL_STATE(4110)] = 217057,
  [SMALL_STATE(4111)] = 217121,
  [SMALL_STATE(4112)] = 217185,
  [SMALL_STATE(4113)] = 217249,
  [SMALL_STATE(4114)] = 217313,
  [SMALL_STATE(4115)] = 217377,
  [SMALL_STATE(4116)] = 217441,
  [SMALL_STATE(4117)] = 217505,
  [SMALL_STATE(4118)] = 217569,
  [SMALL_STATE(4119)] = 217635,
  [SMALL_STATE(4120)] = 217741,
  [SMALL_STATE(4121)] = 217805,
  [SMALL_STATE(4122)] = 217869,
  [SMALL_STATE(4123)] = 217935,
  [SMALL_STATE(4124)] = 217999,
  [SMALL_STATE(4125)] = 218069,
  [SMALL_STATE(4126)] = 218133,
  [SMALL_STATE(4127)] = 218197,
  [SMALL_STATE(4128)] = 218273,
  [SMALL_STATE(4129)] = 218337,
  [SMALL_STATE(4130)] = 218401,
  [SMALL_STATE(4131)] = 218465,
  [SMALL_STATE(4132)] = 218529,
  [SMALL_STATE(4133)] = 218593,
  [SMALL_STATE(4134)] = 218697,
  [SMALL_STATE(4135)] = 218761,
  [SMALL_STATE(4136)] = 218825,
  [SMALL_STATE(4137)] = 218889,
  [SMALL_STATE(4138)] = 218953,
  [SMALL_STATE(4139)] = 219017,
  [SMALL_STATE(4140)] = 219123,
  [SMALL_STATE(4141)] = 219187,
  [SMALL_STATE(4142)] = 219251,
  [SMALL_STATE(4143)] = 219315,
  [SMALL_STATE(4144)] = 219379,
  [SMALL_STATE(4145)] = 219443,
  [SMALL_STATE(4146)] = 219547,
  [SMALL_STATE(4147)] = 219611,
  [SMALL_STATE(4148)] = 219679,
  [SMALL_STATE(4149)] = 219743,
  [SMALL_STATE(4150)] = 219807,
  [SMALL_STATE(4151)] = 219871,
  [SMALL_STATE(4152)] = 219935,
  [SMALL_STATE(4153)] = 219999,
  [SMALL_STATE(4154)] = 220063,
  [SMALL_STATE(4155)] = 220127,
  [SMALL_STATE(4156)] = 220191,
  [SMALL_STATE(4157)] = 220255,
  [SMALL_STATE(4158)] = 220361,
  [SMALL_STATE(4159)] = 220433,
  [SMALL_STATE(4160)] = 220537,
  [SMALL_STATE(4161)] = 220641,
  [SMALL_STATE(4162)] = 220709,
  [SMALL_STATE(4163)] = 220773,
  [SMALL_STATE(4164)] = 220837,
  [SMALL_STATE(4165)] = 220917,
  [SMALL_STATE(4166)] = 220981,
  [SMALL_STATE(4167)] = 221045,
  [SMALL_STATE(4168)] = 221109,
  [SMALL_STATE(4169)] = 221173,
  [SMALL_STATE(4170)] = 221237,
  [SMALL_STATE(4171)] = 221301,
  [SMALL_STATE(4172)] = 221409,
  [SMALL_STATE(4173)] = 221473,
  [SMALL_STATE(4174)] = 221537,
  [SMALL_STATE(4175)] = 221601,
  [SMALL_STATE(4176)] = 221665,
  [SMALL_STATE(4177)] = 221729,
  [SMALL_STATE(4178)] = 221793,
  [SMALL_STATE(4179)] = 221895,
  [SMALL_STATE(4180)] = 221959,
  [SMALL_STATE(4181)] = 222023,
  [SMALL_STATE(4182)] = 222091,
  [SMALL_STATE(4183)] = 222189,
  [SMALL_STATE(4184)] = 222285,
  [SMALL_STATE(4185)] = 222377,
  [SMALL_STATE(4186)] = 222467,
  [SMALL_STATE(4187)] = 222555,
  [SMALL_STATE(4188)] = 222619,
  [SMALL_STATE(4189)] = 222683,
  [SMALL_STATE(4190)] = 222789,
  [SMALL_STATE(4191)] = 222873,
  [SMALL_STATE(4192)] = 222937,
  [SMALL_STATE(4193)] = 223015,
  [SMALL_STATE(4194)] = 223079,
  [SMALL_STATE(4195)] = 223145,
  [SMALL_STATE(4196)] = 223223,
  [SMALL_STATE(4197)] = 223289,
  [SMALL_STATE(4198)] = 223353,
  [SMALL_STATE(4199)] = 223417,
  [SMALL_STATE(4200)] = 223485,
  [SMALL_STATE(4201)] = 223549,
  [SMALL_STATE(4202)] = 223613,
  [SMALL_STATE(4203)] = 223679,
  [SMALL_STATE(4204)] = 223745,
  [SMALL_STATE(4205)] = 223811,
  [SMALL_STATE(4206)] = 223877,
  [SMALL_STATE(4207)] = 223943,
  [SMALL_STATE(4208)] = 224009,
  [SMALL_STATE(4209)] = 224073,
  [SMALL_STATE(4210)] = 224137,
  [SMALL_STATE(4211)] = 224201,
  [SMALL_STATE(4212)] = 224265,
  [SMALL_STATE(4213)] = 224329,
  [SMALL_STATE(4214)] = 224393,
  [SMALL_STATE(4215)] = 224457,
  [SMALL_STATE(4216)] = 224521,
  [SMALL_STATE(4217)] = 224587,
  [SMALL_STATE(4218)] = 224653,
  [SMALL_STATE(4219)] = 224719,
  [SMALL_STATE(4220)] = 224785,
  [SMALL_STATE(4221)] = 224851,
  [SMALL_STATE(4222)] = 224917,
  [SMALL_STATE(4223)] = 224983,
  [SMALL_STATE(4224)] = 225049,
  [SMALL_STATE(4225)] = 225115,
  [SMALL_STATE(4226)] = 225181,
  [SMALL_STATE(4227)] = 225247,
  [SMALL_STATE(4228)] = 225313,
  [SMALL_STATE(4229)] = 225377,
  [SMALL_STATE(4230)] = 225441,
  [SMALL_STATE(4231)] = 225505,
  [SMALL_STATE(4232)] = 225585,
  [SMALL_STATE(4233)] = 225649,
  [SMALL_STATE(4234)] = 225713,
  [SMALL_STATE(4235)] = 225797,
  [SMALL_STATE(4236)] = 225895,
  [SMALL_STATE(4237)] = 225959,
  [SMALL_STATE(4238)] = 226023,
  [SMALL_STATE(4239)] = 226093,
  [SMALL_STATE(4240)] = 226169,
  [SMALL_STATE(4241)] = 226233,
  [SMALL_STATE(4242)] = 226297,
  [SMALL_STATE(4243)] = 226369,
  [SMALL_STATE(4244)] = 226433,
  [SMALL_STATE(4245)] = 226497,
  [SMALL_STATE(4246)] = 226565,
  [SMALL_STATE(4247)] = 226667,
  [SMALL_STATE(4248)] = 226731,
  [SMALL_STATE(4249)] = 226833,
  [SMALL_STATE(4250)] = 226897,
  [SMALL_STATE(4251)] = 226961,
  [SMALL_STATE(4252)] = 227025,
  [SMALL_STATE(4253)] = 227089,
  [SMALL_STATE(4254)] = 227153,
  [SMALL_STATE(4255)] = 227217,
  [SMALL_STATE(4256)] = 227281,
  [SMALL_STATE(4257)] = 227345,
  [SMALL_STATE(4258)] = 227409,
  [SMALL_STATE(4259)] = 227473,
  [SMALL_STATE(4260)] = 227573,
  [SMALL_STATE(4261)] = 227637,
  [SMALL_STATE(4262)] = 227701,
  [SMALL_STATE(4263)] = 227797,
  [SMALL_STATE(4264)] = 227891,
  [SMALL_STATE(4265)] = 227955,
  [SMALL_STATE(4266)] = 228019,
  [SMALL_STATE(4267)] = 228109,
  [SMALL_STATE(4268)] = 228197,
  [SMALL_STATE(4269)] = 228261,
  [SMALL_STATE(4270)] = 228325,
  [SMALL_STATE(4271)] = 228411,
  [SMALL_STATE(4272)] = 228493,
  [SMALL_STATE(4273)] = 228571,
  [SMALL_STATE(4274)] = 228649,
  [SMALL_STATE(4275)] = 228717,
  [SMALL_STATE(4276)] = 228781,
  [SMALL_STATE(4277)] = 228845,
  [SMALL_STATE(4278)] = 228909,
  [SMALL_STATE(4279)] = 228977,
  [SMALL_STATE(4280)] = 229063,
  [SMALL_STATE(4281)] = 229163,
  [SMALL_STATE(4282)] = 229227,
  [SMALL_STATE(4283)] = 229291,
  [SMALL_STATE(4284)] = 229355,
  [SMALL_STATE(4285)] = 229419,
  [SMALL_STATE(4286)] = 229483,
  [SMALL_STATE(4287)] = 229551,
  [SMALL_STATE(4288)] = 229615,
  [SMALL_STATE(4289)] = 229679,
  [SMALL_STATE(4290)] = 229743,
  [SMALL_STATE(4291)] = 229807,
  [SMALL_STATE(4292)] = 229915,
  [SMALL_STATE(4293)] = 229979,
  [SMALL_STATE(4294)] = 230043,
  [SMALL_STATE(4295)] = 230107,
  [SMALL_STATE(4296)] = 230171,
  [SMALL_STATE(4297)] = 230235,
  [SMALL_STATE(4298)] = 230343,
  [SMALL_STATE(4299)] = 230407,
  [SMALL_STATE(4300)] = 230487,
  [SMALL_STATE(4301)] = 230573,
  [SMALL_STATE(4302)] = 230673,
  [SMALL_STATE(4303)] = 230737,
  [SMALL_STATE(4304)] = 230807,
  [SMALL_STATE(4305)] = 230883,
  [SMALL_STATE(4306)] = 230955,
  [SMALL_STATE(4307)] = 231059,
  [SMALL_STATE(4308)] = 231163,
  [SMALL_STATE(4309)] = 231227,
  [SMALL_STATE(4310)] = 231329,
  [SMALL_STATE(4311)] = 231427,
  [SMALL_STATE(4312)] = 231523,
  [SMALL_STATE(4313)] = 231615,
  [SMALL_STATE(4314)] = 231705,
  [SMALL_STATE(4315)] = 231793,
  [SMALL_STATE(4316)] = 231877,
  [SMALL_STATE(4317)] = 231955,
  [SMALL_STATE(4318)] = 232033,
  [SMALL_STATE(4319)] = 232097,
  [SMALL_STATE(4320)] = 232161,
  [SMALL_STATE(4321)] = 232225,
  [SMALL_STATE(4322)] = 232289,
  [SMALL_STATE(4323)] = 232353,
  [SMALL_STATE(4324)] = 232417,
  [SMALL_STATE(4325)] = 232485,
  [SMALL_STATE(4326)] = 232549,
  [SMALL_STATE(4327)] = 232613,
  [SMALL_STATE(4328)] = 232677,
  [SMALL_STATE(4329)] = 232741,
  [SMALL_STATE(4330)] = 232805,
  [SMALL_STATE(4331)] = 232869,
  [SMALL_STATE(4332)] = 232933,
  [SMALL_STATE(4333)] = 232997,
  [SMALL_STATE(4334)] = 233061,
  [SMALL_STATE(4335)] = 233125,
  [SMALL_STATE(4336)] = 233189,
  [SMALL_STATE(4337)] = 233253,
  [SMALL_STATE(4338)] = 233317,
  [SMALL_STATE(4339)] = 233381,
  [SMALL_STATE(4340)] = 233445,
  [SMALL_STATE(4341)] = 233509,
  [SMALL_STATE(4342)] = 233573,
  [SMALL_STATE(4343)] = 233637,
  [SMALL_STATE(4344)] = 233701,
  [SMALL_STATE(4345)] = 233765,
  [SMALL_STATE(4346)] = 233829,
  [SMALL_STATE(4347)] = 233893,
  [SMALL_STATE(4348)] = 233957,
  [SMALL_STATE(4349)] = 234021,
  [SMALL_STATE(4350)] = 234085,
  [SMALL_STATE(4351)] = 234149,
  [SMALL_STATE(4352)] = 234213,
  [SMALL_STATE(4353)] = 234277,
  [SMALL_STATE(4354)] = 234341,
  [SMALL_STATE(4355)] = 234405,
  [SMALL_STATE(4356)] = 234469,
  [SMALL_STATE(4357)] = 234533,
  [SMALL_STATE(4358)] = 234597,
  [SMALL_STATE(4359)] = 234661,
  [SMALL_STATE(4360)] = 234725,
  [SMALL_STATE(4361)] = 234833,
  [SMALL_STATE(4362)] = 234897,
  [SMALL_STATE(4363)] = 234961,
  [SMALL_STATE(4364)] = 235025,
  [SMALL_STATE(4365)] = 235089,
  [SMALL_STATE(4366)] = 235153,
  [SMALL_STATE(4367)] = 235217,
  [SMALL_STATE(4368)] = 235281,
  [SMALL_STATE(4369)] = 235345,
  [SMALL_STATE(4370)] = 235409,
  [SMALL_STATE(4371)] = 235473,
  [SMALL_STATE(4372)] = 235537,
  [SMALL_STATE(4373)] = 235601,
  [SMALL_STATE(4374)] = 235665,
  [SMALL_STATE(4375)] = 235729,
  [SMALL_STATE(4376)] = 235793,
  [SMALL_STATE(4377)] = 235857,
  [SMALL_STATE(4378)] = 235963,
  [SMALL_STATE(4379)] = 236027,
  [SMALL_STATE(4380)] = 236091,
  [SMALL_STATE(4381)] = 236155,
  [SMALL_STATE(4382)] = 236223,
  [SMALL_STATE(4383)] = 236287,
  [SMALL_STATE(4384)] = 236351,
  [SMALL_STATE(4385)] = 236415,
  [SMALL_STATE(4386)] = 236479,
  [SMALL_STATE(4387)] = 236543,
  [SMALL_STATE(4388)] = 236607,
  [SMALL_STATE(4389)] = 236671,
  [SMALL_STATE(4390)] = 236735,
  [SMALL_STATE(4391)] = 236799,
  [SMALL_STATE(4392)] = 236863,
  [SMALL_STATE(4393)] = 236931,
  [SMALL_STATE(4394)] = 236999,
  [SMALL_STATE(4395)] = 237067,
  [SMALL_STATE(4396)] = 237131,
  [SMALL_STATE(4397)] = 237199,
  [SMALL_STATE(4398)] = 237263,
  [SMALL_STATE(4399)] = 237331,
  [SMALL_STATE(4400)] = 237399,
  [SMALL_STATE(4401)] = 237467,
  [SMALL_STATE(4402)] = 237531,
  [SMALL_STATE(4403)] = 237595,
  [SMALL_STATE(4404)] = 237659,
  [SMALL_STATE(4405)] = 237723,
  [SMALL_STATE(4406)] = 237787,
  [SMALL_STATE(4407)] = 237851,
  [SMALL_STATE(4408)] = 237915,
  [SMALL_STATE(4409)] = 237979,
  [SMALL_STATE(4410)] = 238043,
  [SMALL_STATE(4411)] = 238107,
  [SMALL_STATE(4412)] = 238171,
  [SMALL_STATE(4413)] = 238235,
  [SMALL_STATE(4414)] = 238299,
  [SMALL_STATE(4415)] = 238363,
  [SMALL_STATE(4416)] = 238427,
  [SMALL_STATE(4417)] = 238491,
  [SMALL_STATE(4418)] = 238555,
  [SMALL_STATE(4419)] = 238619,
  [SMALL_STATE(4420)] = 238683,
  [SMALL_STATE(4421)] = 238747,
  [SMALL_STATE(4422)] = 238811,
  [SMALL_STATE(4423)] = 238875,
  [SMALL_STATE(4424)] = 238939,
  [SMALL_STATE(4425)] = 239003,
  [SMALL_STATE(4426)] = 239067,
  [SMALL_STATE(4427)] = 239131,
  [SMALL_STATE(4428)] = 239195,
  [SMALL_STATE(4429)] = 239259,
  [SMALL_STATE(4430)] = 239323,
  [SMALL_STATE(4431)] = 239387,
  [SMALL_STATE(4432)] = 239451,
  [SMALL_STATE(4433)] = 239550,
  [SMALL_STATE(4434)] = 239627,
  [SMALL_STATE(4435)] = 239690,
  [SMALL_STATE(4436)] = 239753,
  [SMALL_STATE(4437)] = 239816,
  [SMALL_STATE(4438)] = 239879,
  [SMALL_STATE(4439)] = 239956,
  [SMALL_STATE(4440)] = 240019,
  [SMALL_STATE(4441)] = 240082,
  [SMALL_STATE(4442)] = 240145,
  [SMALL_STATE(4443)] = 240248,
  [SMALL_STATE(4444)] = 240311,
  [SMALL_STATE(4445)] = 240416,
  [SMALL_STATE(4446)] = 240519,
  [SMALL_STATE(4447)] = 240624,
  [SMALL_STATE(4448)] = 240729,
  [SMALL_STATE(4449)] = 240808,
  [SMALL_STATE(4450)] = 240871,
  [SMALL_STATE(4451)] = 240954,
  [SMALL_STATE(4452)] = 241051,
  [SMALL_STATE(4453)] = 241154,
  [SMALL_STATE(4454)] = 241223,
  [SMALL_STATE(4455)] = 241286,
  [SMALL_STATE(4456)] = 241349,
  [SMALL_STATE(4457)] = 241424,
  [SMALL_STATE(4458)] = 241495,
  [SMALL_STATE(4459)] = 241596,
  [SMALL_STATE(4460)] = 241659,
  [SMALL_STATE(4461)] = 241722,
  [SMALL_STATE(4462)] = 241785,
  [SMALL_STATE(4463)] = 241848,
  [SMALL_STATE(4464)] = 241911,
  [SMALL_STATE(4465)] = 241974,
  [SMALL_STATE(4466)] = 242037,
  [SMALL_STATE(4467)] = 242100,
  [SMALL_STATE(4468)] = 242163,
  [SMALL_STATE(4469)] = 242260,
  [SMALL_STATE(4470)] = 242323,
  [SMALL_STATE(4471)] = 242386,
  [SMALL_STATE(4472)] = 242449,
  [SMALL_STATE(4473)] = 242512,
  [SMALL_STATE(4474)] = 242575,
  [SMALL_STATE(4475)] = 242638,
  [SMALL_STATE(4476)] = 242701,
  [SMALL_STATE(4477)] = 242764,
  [SMALL_STATE(4478)] = 242827,
  [SMALL_STATE(4479)] = 242890,
  [SMALL_STATE(4480)] = 242991,
  [SMALL_STATE(4481)] = 243060,
  [SMALL_STATE(4482)] = 243135,
  [SMALL_STATE(4483)] = 243198,
  [SMALL_STATE(4484)] = 243261,
  [SMALL_STATE(4485)] = 243332,
  [SMALL_STATE(4486)] = 243437,
  [SMALL_STATE(4487)] = 243500,
  [SMALL_STATE(4488)] = 243563,
  [SMALL_STATE(4489)] = 243626,
  [SMALL_STATE(4490)] = 243689,
  [SMALL_STATE(4491)] = 243752,
  [SMALL_STATE(4492)] = 243815,
  [SMALL_STATE(4493)] = 243878,
  [SMALL_STATE(4494)] = 243981,
  [SMALL_STATE(4495)] = 244044,
  [SMALL_STATE(4496)] = 244145,
  [SMALL_STATE(4497)] = 244248,
  [SMALL_STATE(4498)] = 244349,
  [SMALL_STATE(4499)] = 244412,
  [SMALL_STATE(4500)] = 244475,
  [SMALL_STATE(4501)] = 244538,
  [SMALL_STATE(4502)] = 244601,
  [SMALL_STATE(4503)] = 244664,
  [SMALL_STATE(4504)] = 244727,
  [SMALL_STATE(4505)] = 244794,
  [SMALL_STATE(4506)] = 244857,
  [SMALL_STATE(4507)] = 244920,
  [SMALL_STATE(4508)] = 244983,
  [SMALL_STATE(4509)] = 245078,
  [SMALL_STATE(4510)] = 245141,
  [SMALL_STATE(4511)] = 245204,
  [SMALL_STATE(4512)] = 245297,
  [SMALL_STATE(4513)] = 245386,
  [SMALL_STATE(4514)] = 245473,
  [SMALL_STATE(4515)] = 245572,
  [SMALL_STATE(4516)] = 245635,
  [SMALL_STATE(4517)] = 245698,
  [SMALL_STATE(4518)] = 245761,
  [SMALL_STATE(4519)] = 245824,
  [SMALL_STATE(4520)] = 245889,
  [SMALL_STATE(4521)] = 245974,
  [SMALL_STATE(4522)] = 246037,
  [SMALL_STATE(4523)] = 246118,
  [SMALL_STATE(4524)] = 246195,
  [SMALL_STATE(4525)] = 246272,
  [SMALL_STATE(4526)] = 246335,
  [SMALL_STATE(4527)] = 246398,
  [SMALL_STATE(4528)] = 246461,
  [SMALL_STATE(4529)] = 246524,
  [SMALL_STATE(4530)] = 246603,
  [SMALL_STATE(4531)] = 246670,
  [SMALL_STATE(4532)] = 246733,
  [SMALL_STATE(4533)] = 246796,
  [SMALL_STATE(4534)] = 246859,
  [SMALL_STATE(4535)] = 246954,
  [SMALL_STATE(4536)] = 247057,
  [SMALL_STATE(4537)] = 247150,
  [SMALL_STATE(4538)] = 247239,
  [SMALL_STATE(4539)] = 247326,
  [SMALL_STATE(4540)] = 247411,
  [SMALL_STATE(4541)] = 247494,
  [SMALL_STATE(4542)] = 247557,
  [SMALL_STATE(4543)] = 247638,
  [SMALL_STATE(4544)] = 247703,
  [SMALL_STATE(4545)] = 247766,
  [SMALL_STATE(4546)] = 247829,
  [SMALL_STATE(4547)] = 247892,
  [SMALL_STATE(4548)] = 247955,
  [SMALL_STATE(4549)] = 248018,
  [SMALL_STATE(4550)] = 248081,
  [SMALL_STATE(4551)] = 248144,
  [SMALL_STATE(4552)] = 248207,
  [SMALL_STATE(4553)] = 248270,
  [SMALL_STATE(4554)] = 248373,
  [SMALL_STATE(4555)] = 248436,
  [SMALL_STATE(4556)] = 248503,
  [SMALL_STATE(4557)] = 248566,
  [SMALL_STATE(4558)] = 248629,
  [SMALL_STATE(4559)] = 248692,
  [SMALL_STATE(4560)] = 248755,
  [SMALL_STATE(4561)] = 248818,
  [SMALL_STATE(4562)] = 248881,
  [SMALL_STATE(4563)] = 248944,
  [SMALL_STATE(4564)] = 249007,
  [SMALL_STATE(4565)] = 249070,
  [SMALL_STATE(4566)] = 249133,
  [SMALL_STATE(4567)] = 249196,
  [SMALL_STATE(4568)] = 249259,
  [SMALL_STATE(4569)] = 249323,
  [SMALL_STATE(4570)] = 249425,
  [SMALL_STATE(4571)] = 249527,
  [SMALL_STATE(4572)] = 249629,
  [SMALL_STATE(4573)] = 249731,
  [SMALL_STATE(4574)] = 249833,
  [SMALL_STATE(4575)] = 249897,
  [SMALL_STATE(4576)] = 249999,
  [SMALL_STATE(4577)] = 250101,
  [SMALL_STATE(4578)] = 250203,
  [SMALL_STATE(4579)] = 250305,
  [SMALL_STATE(4580)] = 250407,
  [SMALL_STATE(4581)] = 250509,
  [SMALL_STATE(4582)] = 250611,
  [SMALL_STATE(4583)] = 250713,
  [SMALL_STATE(4584)] = 250815,
  [SMALL_STATE(4585)] = 250917,
  [SMALL_STATE(4586)] = 251019,
  [SMALL_STATE(4587)] = 251121,
  [SMALL_STATE(4588)] = 251223,
  [SMALL_STATE(4589)] = 251325,
  [SMALL_STATE(4590)] = 251427,
  [SMALL_STATE(4591)] = 251529,
  [SMALL_STATE(4592)] = 251631,
  [SMALL_STATE(4593)] = 251733,
  [SMALL_STATE(4594)] = 251835,
  [SMALL_STATE(4595)] = 251934,
  [SMALL_STATE(4596)] = 252003,
  [SMALL_STATE(4597)] = 252072,
  [SMALL_STATE(4598)] = 252141,
  [SMALL_STATE(4599)] = 252210,
  [SMALL_STATE(4600)] = 252279,
  [SMALL_STATE(4601)] = 252348,
  [SMALL_STATE(4602)] = 252417,
  [SMALL_STATE(4603)] = 252486,
  [SMALL_STATE(4604)] = 252555,
  [SMALL_STATE(4605)] = 252624,
  [SMALL_STATE(4606)] = 252693,
  [SMALL_STATE(4607)] = 252762,
  [SMALL_STATE(4608)] = 252831,
  [SMALL_STATE(4609)] = 252900,
  [SMALL_STATE(4610)] = 252969,
  [SMALL_STATE(4611)] = 253038,
  [SMALL_STATE(4612)] = 253107,
  [SMALL_STATE(4613)] = 253176,
  [SMALL_STATE(4614)] = 253245,
  [SMALL_STATE(4615)] = 253314,
  [SMALL_STATE(4616)] = 253383,
  [SMALL_STATE(4617)] = 253452,
  [SMALL_STATE(4618)] = 253521,
  [SMALL_STATE(4619)] = 253590,
  [SMALL_STATE(4620)] = 253659,
  [SMALL_STATE(4621)] = 253728,
  [SMALL_STATE(4622)] = 253797,
  [SMALL_STATE(4623)] = 253866,
  [SMALL_STATE(4624)] = 253935,
  [SMALL_STATE(4625)] = 254004,
  [SMALL_STATE(4626)] = 254073,
  [SMALL_STATE(4627)] = 254142,
  [SMALL_STATE(4628)] = 254211,
  [SMALL_STATE(4629)] = 254280,
  [SMALL_STATE(4630)] = 254349,
  [SMALL_STATE(4631)] = 254418,
  [SMALL_STATE(4632)] = 254463,
  [SMALL_STATE(4633)] = 254508,
  [SMALL_STATE(4634)] = 254553,
  [SMALL_STATE(4635)] = 254598,
  [SMALL_STATE(4636)] = 254643,
  [SMALL_STATE(4637)] = 254688,
  [SMALL_STATE(4638)] = 254733,
  [SMALL_STATE(4639)] = 254778,
  [SMALL_STATE(4640)] = 254823,
  [SMALL_STATE(4641)] = 254868,
  [SMALL_STATE(4642)] = 254913,
  [SMALL_STATE(4643)] = 254958,
  [SMALL_STATE(4644)] = 255003,
  [SMALL_STATE(4645)] = 255048,
  [SMALL_STATE(4646)] = 255093,
  [SMALL_STATE(4647)] = 255138,
  [SMALL_STATE(4648)] = 255183,
  [SMALL_STATE(4649)] = 255228,
  [SMALL_STATE(4650)] = 255273,
  [SMALL_STATE(4651)] = 255318,
  [SMALL_STATE(4652)] = 255363,
  [SMALL_STATE(4653)] = 255408,
  [SMALL_STATE(4654)] = 255453,
  [SMALL_STATE(4655)] = 255498,
  [SMALL_STATE(4656)] = 255543,
  [SMALL_STATE(4657)] = 255588,
  [SMALL_STATE(4658)] = 255633,
  [SMALL_STATE(4659)] = 255678,
  [SMALL_STATE(4660)] = 255723,
  [SMALL_STATE(4661)] = 255768,
  [SMALL_STATE(4662)] = 255813,
  [SMALL_STATE(4663)] = 255858,
  [SMALL_STATE(4664)] = 255903,
  [SMALL_STATE(4665)] = 255948,
  [SMALL_STATE(4666)] = 255993,
  [SMALL_STATE(4667)] = 256038,
  [SMALL_STATE(4668)] = 256083,
  [SMALL_STATE(4669)] = 256128,
  [SMALL_STATE(4670)] = 256173,
  [SMALL_STATE(4671)] = 256218,
  [SMALL_STATE(4672)] = 256263,
  [SMALL_STATE(4673)] = 256308,
  [SMALL_STATE(4674)] = 256353,
  [SMALL_STATE(4675)] = 256398,
  [SMALL_STATE(4676)] = 256443,
  [SMALL_STATE(4677)] = 256488,
  [SMALL_STATE(4678)] = 256533,
  [SMALL_STATE(4679)] = 256578,
  [SMALL_STATE(4680)] = 256623,
  [SMALL_STATE(4681)] = 256668,
  [SMALL_STATE(4682)] = 256713,
  [SMALL_STATE(4683)] = 256758,
  [SMALL_STATE(4684)] = 256803,
  [SMALL_STATE(4685)] = 256848,
  [SMALL_STATE(4686)] = 256893,
  [SMALL_STATE(4687)] = 256938,
  [SMALL_STATE(4688)] = 256983,
  [SMALL_STATE(4689)] = 257028,
  [SMALL_STATE(4690)] = 257073,
  [SMALL_STATE(4691)] = 257118,
  [SMALL_STATE(4692)] = 257163,
  [SMALL_STATE(4693)] = 257208,
  [SMALL_STATE(4694)] = 257253,
  [SMALL_STATE(4695)] = 257298,
  [SMALL_STATE(4696)] = 257343,
  [SMALL_STATE(4697)] = 257388,
  [SMALL_STATE(4698)] = 257433,
  [SMALL_STATE(4699)] = 257478,
  [SMALL_STATE(4700)] = 257523,
  [SMALL_STATE(4701)] = 257568,
  [SMALL_STATE(4702)] = 257613,
  [SMALL_STATE(4703)] = 257658,
  [SMALL_STATE(4704)] = 257694,
  [SMALL_STATE(4705)] = 257730,
  [SMALL_STATE(4706)] = 257761,
  [SMALL_STATE(4707)] = 257792,
  [SMALL_STATE(4708)] = 257823,
  [SMALL_STATE(4709)] = 257854,
  [SMALL_STATE(4710)] = 257885,
  [SMALL_STATE(4711)] = 257916,
  [SMALL_STATE(4712)] = 257947,
  [SMALL_STATE(4713)] = 257978,
  [SMALL_STATE(4714)] = 258009,
  [SMALL_STATE(4715)] = 258040,
  [SMALL_STATE(4716)] = 258071,
  [SMALL_STATE(4717)] = 258102,
  [SMALL_STATE(4718)] = 258133,
  [SMALL_STATE(4719)] = 258164,
  [SMALL_STATE(4720)] = 258195,
  [SMALL_STATE(4721)] = 258226,
  [SMALL_STATE(4722)] = 258257,
  [SMALL_STATE(4723)] = 258288,
  [SMALL_STATE(4724)] = 258319,
  [SMALL_STATE(4725)] = 258350,
  [SMALL_STATE(4726)] = 258381,
  [SMALL_STATE(4727)] = 258412,
  [SMALL_STATE(4728)] = 258443,
  [SMALL_STATE(4729)] = 258474,
  [SMALL_STATE(4730)] = 258505,
  [SMALL_STATE(4731)] = 258536,
  [SMALL_STATE(4732)] = 258567,
  [SMALL_STATE(4733)] = 258598,
  [SMALL_STATE(4734)] = 258629,
  [SMALL_STATE(4735)] = 258660,
  [SMALL_STATE(4736)] = 258691,
  [SMALL_STATE(4737)] = 258722,
  [SMALL_STATE(4738)] = 258753,
  [SMALL_STATE(4739)] = 258784,
  [SMALL_STATE(4740)] = 258815,
  [SMALL_STATE(4741)] = 258846,
  [SMALL_STATE(4742)] = 258877,
  [SMALL_STATE(4743)] = 258908,
  [SMALL_STATE(4744)] = 258939,
  [SMALL_STATE(4745)] = 258970,
  [SMALL_STATE(4746)] = 259001,
  [SMALL_STATE(4747)] = 259032,
  [SMALL_STATE(4748)] = 259063,
  [SMALL_STATE(4749)] = 259094,
  [SMALL_STATE(4750)] = 259125,
  [SMALL_STATE(4751)] = 259156,
  [SMALL_STATE(4752)] = 259187,
  [SMALL_STATE(4753)] = 259218,
  [SMALL_STATE(4754)] = 259249,
  [SMALL_STATE(4755)] = 259280,
  [SMALL_STATE(4756)] = 259311,
  [SMALL_STATE(4757)] = 259342,
  [SMALL_STATE(4758)] = 259373,
  [SMALL_STATE(4759)] = 259404,
  [SMALL_STATE(4760)] = 259435,
  [SMALL_STATE(4761)] = 259466,
  [SMALL_STATE(4762)] = 259497,
  [SMALL_STATE(4763)] = 259528,
  [SMALL_STATE(4764)] = 259559,
  [SMALL_STATE(4765)] = 259590,
  [SMALL_STATE(4766)] = 259621,
  [SMALL_STATE(4767)] = 259651,
  [SMALL_STATE(4768)] = 259681,
  [SMALL_STATE(4769)] = 259711,
  [SMALL_STATE(4770)] = 259741,
  [SMALL_STATE(4771)] = 259771,
  [SMALL_STATE(4772)] = 259801,
  [SMALL_STATE(4773)] = 259831,
  [SMALL_STATE(4774)] = 259861,
  [SMALL_STATE(4775)] = 259891,
  [SMALL_STATE(4776)] = 259921,
  [SMALL_STATE(4777)] = 259951,
  [SMALL_STATE(4778)] = 259981,
  [SMALL_STATE(4779)] = 260011,
  [SMALL_STATE(4780)] = 260041,
  [SMALL_STATE(4781)] = 260071,
  [SMALL_STATE(4782)] = 260101,
  [SMALL_STATE(4783)] = 260131,
  [SMALL_STATE(4784)] = 260161,
  [SMALL_STATE(4785)] = 260191,
  [SMALL_STATE(4786)] = 260221,
  [SMALL_STATE(4787)] = 260251,
  [SMALL_STATE(4788)] = 260281,
  [SMALL_STATE(4789)] = 260311,
  [SMALL_STATE(4790)] = 260341,
  [SMALL_STATE(4791)] = 260375,
  [SMALL_STATE(4792)] = 260405,
  [SMALL_STATE(4793)] = 260439,
  [SMALL_STATE(4794)] = 260469,
  [SMALL_STATE(4795)] = 260499,
  [SMALL_STATE(4796)] = 260529,
  [SMALL_STATE(4797)] = 260559,
  [SMALL_STATE(4798)] = 260582,
  [SMALL_STATE(4799)] = 260605,
  [SMALL_STATE(4800)] = 260628,
  [SMALL_STATE(4801)] = 260659,
  [SMALL_STATE(4802)] = 260690,
  [SMALL_STATE(4803)] = 260713,
  [SMALL_STATE(4804)] = 260744,
  [SMALL_STATE(4805)] = 260775,
  [SMALL_STATE(4806)] = 260806,
  [SMALL_STATE(4807)] = 260829,
  [SMALL_STATE(4808)] = 260860,
  [SMALL_STATE(4809)] = 260891,
  [SMALL_STATE(4810)] = 260922,
  [SMALL_STATE(4811)] = 260945,
  [SMALL_STATE(4812)] = 260976,
  [SMALL_STATE(4813)] = 261007,
  [SMALL_STATE(4814)] = 261030,
  [SMALL_STATE(4815)] = 261053,
  [SMALL_STATE(4816)] = 261076,
  [SMALL_STATE(4817)] = 261099,
  [SMALL_STATE(4818)] = 261122,
  [SMALL_STATE(4819)] = 261145,
  [SMALL_STATE(4820)] = 261176,
  [SMALL_STATE(4821)] = 261207,
  [SMALL_STATE(4822)] = 261230,
  [SMALL_STATE(4823)] = 261253,
  [SMALL_STATE(4824)] = 261276,
  [SMALL_STATE(4825)] = 261307,
  [SMALL_STATE(4826)] = 261338,
  [SMALL_STATE(4827)] = 261361,
  [SMALL_STATE(4828)] = 261392,
  [SMALL_STATE(4829)] = 261415,
  [SMALL_STATE(4830)] = 261446,
  [SMALL_STATE(4831)] = 261477,
  [SMALL_STATE(4832)] = 261508,
  [SMALL_STATE(4833)] = 261539,
  [SMALL_STATE(4834)] = 261570,
  [SMALL_STATE(4835)] = 261601,
  [SMALL_STATE(4836)] = 261632,
  [SMALL_STATE(4837)] = 261655,
  [SMALL_STATE(4838)] = 261678,
  [SMALL_STATE(4839)] = 261701,
  [SMALL_STATE(4840)] = 261732,
  [SMALL_STATE(4841)] = 261750,
  [SMALL_STATE(4842)] = 261770,
  [SMALL_STATE(4843)] = 261790,
  [SMALL_STATE(4844)] = 261810,
  [SMALL_STATE(4845)] = 261837,
  [SMALL_STATE(4846)] = 261864,
  [SMALL_STATE(4847)] = 261891,
  [SMALL_STATE(4848)] = 261918,
  [SMALL_STATE(4849)] = 261945,
  [SMALL_STATE(4850)] = 261972,
  [SMALL_STATE(4851)] = 261999,
  [SMALL_STATE(4852)] = 262026,
  [SMALL_STATE(4853)] = 262053,
  [SMALL_STATE(4854)] = 262080,
  [SMALL_STATE(4855)] = 262107,
  [SMALL_STATE(4856)] = 262134,
  [SMALL_STATE(4857)] = 262161,
  [SMALL_STATE(4858)] = 262188,
  [SMALL_STATE(4859)] = 262215,
  [SMALL_STATE(4860)] = 262242,
  [SMALL_STATE(4861)] = 262269,
  [SMALL_STATE(4862)] = 262296,
  [SMALL_STATE(4863)] = 262323,
  [SMALL_STATE(4864)] = 262350,
  [SMALL_STATE(4865)] = 262377,
  [SMALL_STATE(4866)] = 262404,
  [SMALL_STATE(4867)] = 262431,
  [SMALL_STATE(4868)] = 262458,
  [SMALL_STATE(4869)] = 262485,
  [SMALL_STATE(4870)] = 262512,
  [SMALL_STATE(4871)] = 262539,
  [SMALL_STATE(4872)] = 262566,
  [SMALL_STATE(4873)] = 262593,
  [SMALL_STATE(4874)] = 262620,
  [SMALL_STATE(4875)] = 262647,
  [SMALL_STATE(4876)] = 262674,
  [SMALL_STATE(4877)] = 262701,
  [SMALL_STATE(4878)] = 262728,
  [SMALL_STATE(4879)] = 262755,
  [SMALL_STATE(4880)] = 262782,
  [SMALL_STATE(4881)] = 262809,
  [SMALL_STATE(4882)] = 262836,
  [SMALL_STATE(4883)] = 262863,
  [SMALL_STATE(4884)] = 262890,
  [SMALL_STATE(4885)] = 262917,
  [SMALL_STATE(4886)] = 262944,
  [SMALL_STATE(4887)] = 262971,
  [SMALL_STATE(4888)] = 262998,
  [SMALL_STATE(4889)] = 263025,
  [SMALL_STATE(4890)] = 263052,
  [SMALL_STATE(4891)] = 263079,
  [SMALL_STATE(4892)] = 263106,
  [SMALL_STATE(4893)] = 263133,
  [SMALL_STATE(4894)] = 263160,
  [SMALL_STATE(4895)] = 263187,
  [SMALL_STATE(4896)] = 263214,
  [SMALL_STATE(4897)] = 263241,
  [SMALL_STATE(4898)] = 263268,
  [SMALL_STATE(4899)] = 263295,
  [SMALL_STATE(4900)] = 263322,
  [SMALL_STATE(4901)] = 263349,
  [SMALL_STATE(4902)] = 263376,
  [SMALL_STATE(4903)] = 263403,
  [SMALL_STATE(4904)] = 263430,
  [SMALL_STATE(4905)] = 263457,
  [SMALL_STATE(4906)] = 263484,
  [SMALL_STATE(4907)] = 263511,
  [SMALL_STATE(4908)] = 263538,
  [SMALL_STATE(4909)] = 263565,
  [SMALL_STATE(4910)] = 263592,
  [SMALL_STATE(4911)] = 263619,
  [SMALL_STATE(4912)] = 263646,
  [SMALL_STATE(4913)] = 263673,
  [SMALL_STATE(4914)] = 263700,
  [SMALL_STATE(4915)] = 263727,
  [SMALL_STATE(4916)] = 263754,
  [SMALL_STATE(4917)] = 263781,
  [SMALL_STATE(4918)] = 263808,
  [SMALL_STATE(4919)] = 263835,
  [SMALL_STATE(4920)] = 263862,
  [SMALL_STATE(4921)] = 263889,
  [SMALL_STATE(4922)] = 263916,
  [SMALL_STATE(4923)] = 263943,
  [SMALL_STATE(4924)] = 263970,
  [SMALL_STATE(4925)] = 263997,
  [SMALL_STATE(4926)] = 264024,
  [SMALL_STATE(4927)] = 264051,
  [SMALL_STATE(4928)] = 264078,
  [SMALL_STATE(4929)] = 264105,
  [SMALL_STATE(4930)] = 264132,
  [SMALL_STATE(4931)] = 264159,
  [SMALL_STATE(4932)] = 264186,
  [SMALL_STATE(4933)] = 264213,
  [SMALL_STATE(4934)] = 264240,
  [SMALL_STATE(4935)] = 264267,
  [SMALL_STATE(4936)] = 264294,
  [SMALL_STATE(4937)] = 264321,
  [SMALL_STATE(4938)] = 264348,
  [SMALL_STATE(4939)] = 264375,
  [SMALL_STATE(4940)] = 264402,
  [SMALL_STATE(4941)] = 264429,
  [SMALL_STATE(4942)] = 264456,
  [SMALL_STATE(4943)] = 264483,
  [SMALL_STATE(4944)] = 264510,
  [SMALL_STATE(4945)] = 264537,
  [SMALL_STATE(4946)] = 264564,
  [SMALL_STATE(4947)] = 264591,
  [SMALL_STATE(4948)] = 264618,
  [SMALL_STATE(4949)] = 264645,
  [SMALL_STATE(4950)] = 264672,
  [SMALL_STATE(4951)] = 264699,
  [SMALL_STATE(4952)] = 264726,
  [SMALL_STATE(4953)] = 264753,
  [SMALL_STATE(4954)] = 264780,
  [SMALL_STATE(4955)] = 264807,
  [SMALL_STATE(4956)] = 264834,
  [SMALL_STATE(4957)] = 264861,
  [SMALL_STATE(4958)] = 264888,
  [SMALL_STATE(4959)] = 264915,
  [SMALL_STATE(4960)] = 264942,
  [SMALL_STATE(4961)] = 264969,
  [SMALL_STATE(4962)] = 264996,
  [SMALL_STATE(4963)] = 265023,
  [SMALL_STATE(4964)] = 265050,
  [SMALL_STATE(4965)] = 265077,
  [SMALL_STATE(4966)] = 265104,
  [SMALL_STATE(4967)] = 265131,
  [SMALL_STATE(4968)] = 265158,
  [SMALL_STATE(4969)] = 265185,
  [SMALL_STATE(4970)] = 265212,
  [SMALL_STATE(4971)] = 265239,
  [SMALL_STATE(4972)] = 265266,
  [SMALL_STATE(4973)] = 265293,
  [SMALL_STATE(4974)] = 265320,
  [SMALL_STATE(4975)] = 265347,
  [SMALL_STATE(4976)] = 265374,
  [SMALL_STATE(4977)] = 265401,
  [SMALL_STATE(4978)] = 265428,
  [SMALL_STATE(4979)] = 265455,
  [SMALL_STATE(4980)] = 265482,
  [SMALL_STATE(4981)] = 265509,
  [SMALL_STATE(4982)] = 265536,
  [SMALL_STATE(4983)] = 265563,
  [SMALL_STATE(4984)] = 265590,
  [SMALL_STATE(4985)] = 265617,
  [SMALL_STATE(4986)] = 265644,
  [SMALL_STATE(4987)] = 265671,
  [SMALL_STATE(4988)] = 265698,
  [SMALL_STATE(4989)] = 265725,
  [SMALL_STATE(4990)] = 265752,
  [SMALL_STATE(4991)] = 265779,
  [SMALL_STATE(4992)] = 265806,
  [SMALL_STATE(4993)] = 265833,
  [SMALL_STATE(4994)] = 265860,
  [SMALL_STATE(4995)] = 265887,
  [SMALL_STATE(4996)] = 265914,
  [SMALL_STATE(4997)] = 265941,
  [SMALL_STATE(4998)] = 265968,
  [SMALL_STATE(4999)] = 265995,
  [SMALL_STATE(5000)] = 266022,
  [SMALL_STATE(5001)] = 266049,
  [SMALL_STATE(5002)] = 266076,
  [SMALL_STATE(5003)] = 266103,
  [SMALL_STATE(5004)] = 266130,
  [SMALL_STATE(5005)] = 266157,
  [SMALL_STATE(5006)] = 266184,
  [SMALL_STATE(5007)] = 266211,
  [SMALL_STATE(5008)] = 266238,
  [SMALL_STATE(5009)] = 266265,
  [SMALL_STATE(5010)] = 266292,
  [SMALL_STATE(5011)] = 266319,
  [SMALL_STATE(5012)] = 266346,
  [SMALL_STATE(5013)] = 266373,
  [SMALL_STATE(5014)] = 266400,
  [SMALL_STATE(5015)] = 266427,
  [SMALL_STATE(5016)] = 266454,
  [SMALL_STATE(5017)] = 266481,
  [SMALL_STATE(5018)] = 266508,
  [SMALL_STATE(5019)] = 266535,
  [SMALL_STATE(5020)] = 266562,
  [SMALL_STATE(5021)] = 266589,
  [SMALL_STATE(5022)] = 266616,
  [SMALL_STATE(5023)] = 266643,
  [SMALL_STATE(5024)] = 266670,
  [SMALL_STATE(5025)] = 266697,
  [SMALL_STATE(5026)] = 266724,
  [SMALL_STATE(5027)] = 266751,
  [SMALL_STATE(5028)] = 266778,
  [SMALL_STATE(5029)] = 266805,
  [SMALL_STATE(5030)] = 266832,
  [SMALL_STATE(5031)] = 266859,
  [SMALL_STATE(5032)] = 266886,
  [SMALL_STATE(5033)] = 266913,
  [SMALL_STATE(5034)] = 266940,
  [SMALL_STATE(5035)] = 266967,
  [SMALL_STATE(5036)] = 266994,
  [SMALL_STATE(5037)] = 267021,
  [SMALL_STATE(5038)] = 267048,
  [SMALL_STATE(5039)] = 267075,
  [SMALL_STATE(5040)] = 267102,
  [SMALL_STATE(5041)] = 267129,
  [SMALL_STATE(5042)] = 267156,
  [SMALL_STATE(5043)] = 267183,
  [SMALL_STATE(5044)] = 267210,
  [SMALL_STATE(5045)] = 267237,
  [SMALL_STATE(5046)] = 267264,
  [SMALL_STATE(5047)] = 267291,
  [SMALL_STATE(5048)] = 267318,
  [SMALL_STATE(5049)] = 267345,
  [SMALL_STATE(5050)] = 267372,
  [SMALL_STATE(5051)] = 267399,
  [SMALL_STATE(5052)] = 267426,
  [SMALL_STATE(5053)] = 267453,
  [SMALL_STATE(5054)] = 267480,
  [SMALL_STATE(5055)] = 267507,
  [SMALL_STATE(5056)] = 267534,
  [SMALL_STATE(5057)] = 267561,
  [SMALL_STATE(5058)] = 267588,
  [SMALL_STATE(5059)] = 267615,
  [SMALL_STATE(5060)] = 267642,
  [SMALL_STATE(5061)] = 267669,
  [SMALL_STATE(5062)] = 267696,
  [SMALL_STATE(5063)] = 267723,
  [SMALL_STATE(5064)] = 267750,
  [SMALL_STATE(5065)] = 267777,
  [SMALL_STATE(5066)] = 267804,
  [SMALL_STATE(5067)] = 267831,
  [SMALL_STATE(5068)] = 267858,
  [SMALL_STATE(5069)] = 267885,
  [SMALL_STATE(5070)] = 267912,
  [SMALL_STATE(5071)] = 267939,
  [SMALL_STATE(5072)] = 267966,
  [SMALL_STATE(5073)] = 267993,
  [SMALL_STATE(5074)] = 268020,
  [SMALL_STATE(5075)] = 268047,
  [SMALL_STATE(5076)] = 268074,
  [SMALL_STATE(5077)] = 268101,
  [SMALL_STATE(5078)] = 268128,
  [SMALL_STATE(5079)] = 268155,
  [SMALL_STATE(5080)] = 268182,
  [SMALL_STATE(5081)] = 268209,
  [SMALL_STATE(5082)] = 268236,
  [SMALL_STATE(5083)] = 268263,
  [SMALL_STATE(5084)] = 268290,
  [SMALL_STATE(5085)] = 268317,
  [SMALL_STATE(5086)] = 268344,
  [SMALL_STATE(5087)] = 268368,
  [SMALL_STATE(5088)] = 268392,
  [SMALL_STATE(5089)] = 268416,
  [SMALL_STATE(5090)] = 268442,
  [SMALL_STATE(5091)] = 268468,
  [SMALL_STATE(5092)] = 268492,
  [SMALL_STATE(5093)] = 268516,
  [SMALL_STATE(5094)] = 268540,
  [SMALL_STATE(5095)] = 268564,
  [SMALL_STATE(5096)] = 268588,
  [SMALL_STATE(5097)] = 268612,
  [SMALL_STATE(5098)] = 268636,
  [SMALL_STATE(5099)] = 268660,
  [SMALL_STATE(5100)] = 268684,
  [SMALL_STATE(5101)] = 268708,
  [SMALL_STATE(5102)] = 268732,
  [SMALL_STATE(5103)] = 268758,
  [SMALL_STATE(5104)] = 268782,
  [SMALL_STATE(5105)] = 268808,
  [SMALL_STATE(5106)] = 268834,
  [SMALL_STATE(5107)] = 268860,
  [SMALL_STATE(5108)] = 268884,
  [SMALL_STATE(5109)] = 268910,
  [SMALL_STATE(5110)] = 268934,
  [SMALL_STATE(5111)] = 268960,
  [SMALL_STATE(5112)] = 268984,
  [SMALL_STATE(5113)] = 269008,
  [SMALL_STATE(5114)] = 269032,
  [SMALL_STATE(5115)] = 269056,
  [SMALL_STATE(5116)] = 269080,
  [SMALL_STATE(5117)] = 269104,
  [SMALL_STATE(5118)] = 269128,
  [SMALL_STATE(5119)] = 269152,
  [SMALL_STATE(5120)] = 269178,
  [SMALL_STATE(5121)] = 269204,
  [SMALL_STATE(5122)] = 269228,
  [SMALL_STATE(5123)] = 269254,
  [SMALL_STATE(5124)] = 269280,
  [SMALL_STATE(5125)] = 269306,
  [SMALL_STATE(5126)] = 269332,
  [SMALL_STATE(5127)] = 269356,
  [SMALL_STATE(5128)] = 269382,
  [SMALL_STATE(5129)] = 269408,
  [SMALL_STATE(5130)] = 269434,
  [SMALL_STATE(5131)] = 269458,
  [SMALL_STATE(5132)] = 269484,
  [SMALL_STATE(5133)] = 269510,
  [SMALL_STATE(5134)] = 269536,
  [SMALL_STATE(5135)] = 269560,
  [SMALL_STATE(5136)] = 269584,
  [SMALL_STATE(5137)] = 269608,
  [SMALL_STATE(5138)] = 269632,
  [SMALL_STATE(5139)] = 269656,
  [SMALL_STATE(5140)] = 269680,
  [SMALL_STATE(5141)] = 269704,
  [SMALL_STATE(5142)] = 269728,
  [SMALL_STATE(5143)] = 269752,
  [SMALL_STATE(5144)] = 269776,
  [SMALL_STATE(5145)] = 269800,
  [SMALL_STATE(5146)] = 269824,
  [SMALL_STATE(5147)] = 269850,
  [SMALL_STATE(5148)] = 269874,
  [SMALL_STATE(5149)] = 269900,
  [SMALL_STATE(5150)] = 269924,
  [SMALL_STATE(5151)] = 269948,
  [SMALL_STATE(5152)] = 269972,
  [SMALL_STATE(5153)] = 269996,
  [SMALL_STATE(5154)] = 270020,
  [SMALL_STATE(5155)] = 270044,
  [SMALL_STATE(5156)] = 270068,
  [SMALL_STATE(5157)] = 270092,
  [SMALL_STATE(5158)] = 270116,
  [SMALL_STATE(5159)] = 270140,
  [SMALL_STATE(5160)] = 270164,
  [SMALL_STATE(5161)] = 270188,
  [SMALL_STATE(5162)] = 270212,
  [SMALL_STATE(5163)] = 270236,
  [SMALL_STATE(5164)] = 270262,
  [SMALL_STATE(5165)] = 270286,
  [SMALL_STATE(5166)] = 270310,
  [SMALL_STATE(5167)] = 270334,
  [SMALL_STATE(5168)] = 270358,
  [SMALL_STATE(5169)] = 270384,
  [SMALL_STATE(5170)] = 270408,
  [SMALL_STATE(5171)] = 270432,
  [SMALL_STATE(5172)] = 270456,
  [SMALL_STATE(5173)] = 270480,
  [SMALL_STATE(5174)] = 270504,
  [SMALL_STATE(5175)] = 270528,
  [SMALL_STATE(5176)] = 270552,
  [SMALL_STATE(5177)] = 270578,
  [SMALL_STATE(5178)] = 270602,
  [SMALL_STATE(5179)] = 270626,
  [SMALL_STATE(5180)] = 270650,
  [SMALL_STATE(5181)] = 270674,
  [SMALL_STATE(5182)] = 270698,
  [SMALL_STATE(5183)] = 270722,
  [SMALL_STATE(5184)] = 270746,
  [SMALL_STATE(5185)] = 270770,
  [SMALL_STATE(5186)] = 270794,
  [SMALL_STATE(5187)] = 270818,
  [SMALL_STATE(5188)] = 270844,
  [SMALL_STATE(5189)] = 270870,
  [SMALL_STATE(5190)] = 270896,
  [SMALL_STATE(5191)] = 270922,
  [SMALL_STATE(5192)] = 270942,
  [SMALL_STATE(5193)] = 270968,
  [SMALL_STATE(5194)] = 270992,
  [SMALL_STATE(5195)] = 271018,
  [SMALL_STATE(5196)] = 271042,
  [SMALL_STATE(5197)] = 271066,
  [SMALL_STATE(5198)] = 271090,
  [SMALL_STATE(5199)] = 271114,
  [SMALL_STATE(5200)] = 271140,
  [SMALL_STATE(5201)] = 271164,
  [SMALL_STATE(5202)] = 271188,
  [SMALL_STATE(5203)] = 271212,
  [SMALL_STATE(5204)] = 271236,
  [SMALL_STATE(5205)] = 271262,
  [SMALL_STATE(5206)] = 271286,
  [SMALL_STATE(5207)] = 271312,
  [SMALL_STATE(5208)] = 271338,
  [SMALL_STATE(5209)] = 271362,
  [SMALL_STATE(5210)] = 271386,
  [SMALL_STATE(5211)] = 271412,
  [SMALL_STATE(5212)] = 271438,
  [SMALL_STATE(5213)] = 271464,
  [SMALL_STATE(5214)] = 271490,
  [SMALL_STATE(5215)] = 271516,
  [SMALL_STATE(5216)] = 271540,
  [SMALL_STATE(5217)] = 271564,
  [SMALL_STATE(5218)] = 271588,
  [SMALL_STATE(5219)] = 271612,
  [SMALL_STATE(5220)] = 271636,
  [SMALL_STATE(5221)] = 271660,
  [SMALL_STATE(5222)] = 271684,
  [SMALL_STATE(5223)] = 271708,
  [SMALL_STATE(5224)] = 271732,
  [SMALL_STATE(5225)] = 271756,
  [SMALL_STATE(5226)] = 271780,
  [SMALL_STATE(5227)] = 271804,
  [SMALL_STATE(5228)] = 271830,
  [SMALL_STATE(5229)] = 271856,
  [SMALL_STATE(5230)] = 271882,
  [SMALL_STATE(5231)] = 271908,
  [SMALL_STATE(5232)] = 271932,
  [SMALL_STATE(5233)] = 271956,
  [SMALL_STATE(5234)] = 271980,
  [SMALL_STATE(5235)] = 272004,
  [SMALL_STATE(5236)] = 272028,
  [SMALL_STATE(5237)] = 272052,
  [SMALL_STATE(5238)] = 272076,
  [SMALL_STATE(5239)] = 272102,
  [SMALL_STATE(5240)] = 272126,
  [SMALL_STATE(5241)] = 272152,
  [SMALL_STATE(5242)] = 272176,
  [SMALL_STATE(5243)] = 272200,
  [SMALL_STATE(5244)] = 272224,
  [SMALL_STATE(5245)] = 272248,
  [SMALL_STATE(5246)] = 272274,
  [SMALL_STATE(5247)] = 272298,
  [SMALL_STATE(5248)] = 272322,
  [SMALL_STATE(5249)] = 272346,
  [SMALL_STATE(5250)] = 272372,
  [SMALL_STATE(5251)] = 272398,
  [SMALL_STATE(5252)] = 272424,
  [SMALL_STATE(5253)] = 272450,
  [SMALL_STATE(5254)] = 272474,
  [SMALL_STATE(5255)] = 272498,
  [SMALL_STATE(5256)] = 272522,
  [SMALL_STATE(5257)] = 272546,
  [SMALL_STATE(5258)] = 272570,
  [SMALL_STATE(5259)] = 272594,
  [SMALL_STATE(5260)] = 272618,
  [SMALL_STATE(5261)] = 272642,
  [SMALL_STATE(5262)] = 272666,
  [SMALL_STATE(5263)] = 272690,
  [SMALL_STATE(5264)] = 272710,
  [SMALL_STATE(5265)] = 272730,
  [SMALL_STATE(5266)] = 272756,
  [SMALL_STATE(5267)] = 272780,
  [SMALL_STATE(5268)] = 272804,
  [SMALL_STATE(5269)] = 272828,
  [SMALL_STATE(5270)] = 272852,
  [SMALL_STATE(5271)] = 272878,
  [SMALL_STATE(5272)] = 272902,
  [SMALL_STATE(5273)] = 272928,
  [SMALL_STATE(5274)] = 272952,
  [SMALL_STATE(5275)] = 272978,
  [SMALL_STATE(5276)] = 273002,
  [SMALL_STATE(5277)] = 273026,
  [SMALL_STATE(5278)] = 273050,
  [SMALL_STATE(5279)] = 273074,
  [SMALL_STATE(5280)] = 273100,
  [SMALL_STATE(5281)] = 273126,
  [SMALL_STATE(5282)] = 273152,
  [SMALL_STATE(5283)] = 273178,
  [SMALL_STATE(5284)] = 273202,
  [SMALL_STATE(5285)] = 273228,
  [SMALL_STATE(5286)] = 273252,
  [SMALL_STATE(5287)] = 273276,
  [SMALL_STATE(5288)] = 273302,
  [SMALL_STATE(5289)] = 273328,
  [SMALL_STATE(5290)] = 273354,
  [SMALL_STATE(5291)] = 273378,
  [SMALL_STATE(5292)] = 273402,
  [SMALL_STATE(5293)] = 273428,
  [SMALL_STATE(5294)] = 273454,
  [SMALL_STATE(5295)] = 273478,
  [SMALL_STATE(5296)] = 273504,
  [SMALL_STATE(5297)] = 273530,
  [SMALL_STATE(5298)] = 273554,
  [SMALL_STATE(5299)] = 273578,
  [SMALL_STATE(5300)] = 273602,
  [SMALL_STATE(5301)] = 273626,
  [SMALL_STATE(5302)] = 273650,
  [SMALL_STATE(5303)] = 273674,
  [SMALL_STATE(5304)] = 273698,
  [SMALL_STATE(5305)] = 273722,
  [SMALL_STATE(5306)] = 273746,
  [SMALL_STATE(5307)] = 273770,
  [SMALL_STATE(5308)] = 273794,
  [SMALL_STATE(5309)] = 273818,
  [SMALL_STATE(5310)] = 273844,
  [SMALL_STATE(5311)] = 273868,
  [SMALL_STATE(5312)] = 273894,
  [SMALL_STATE(5313)] = 273920,
  [SMALL_STATE(5314)] = 273944,
  [SMALL_STATE(5315)] = 273968,
  [SMALL_STATE(5316)] = 273992,
  [SMALL_STATE(5317)] = 274016,
  [SMALL_STATE(5318)] = 274040,
  [SMALL_STATE(5319)] = 274064,
  [SMALL_STATE(5320)] = 274088,
  [SMALL_STATE(5321)] = 274112,
  [SMALL_STATE(5322)] = 274136,
  [SMALL_STATE(5323)] = 274160,
  [SMALL_STATE(5324)] = 274184,
  [SMALL_STATE(5325)] = 274208,
  [SMALL_STATE(5326)] = 274232,
  [SMALL_STATE(5327)] = 274256,
  [SMALL_STATE(5328)] = 274280,
  [SMALL_STATE(5329)] = 274304,
  [SMALL_STATE(5330)] = 274328,
  [SMALL_STATE(5331)] = 274352,
  [SMALL_STATE(5332)] = 274376,
  [SMALL_STATE(5333)] = 274402,
  [SMALL_STATE(5334)] = 274426,
  [SMALL_STATE(5335)] = 274450,
  [SMALL_STATE(5336)] = 274474,
  [SMALL_STATE(5337)] = 274498,
  [SMALL_STATE(5338)] = 274522,
  [SMALL_STATE(5339)] = 274546,
  [SMALL_STATE(5340)] = 274570,
  [SMALL_STATE(5341)] = 274594,
  [SMALL_STATE(5342)] = 274620,
  [SMALL_STATE(5343)] = 274644,
  [SMALL_STATE(5344)] = 274670,
  [SMALL_STATE(5345)] = 274696,
  [SMALL_STATE(5346)] = 274722,
  [SMALL_STATE(5347)] = 274746,
  [SMALL_STATE(5348)] = 274772,
  [SMALL_STATE(5349)] = 274796,
  [SMALL_STATE(5350)] = 274820,
  [SMALL_STATE(5351)] = 274844,
  [SMALL_STATE(5352)] = 274868,
  [SMALL_STATE(5353)] = 274894,
  [SMALL_STATE(5354)] = 274918,
  [SMALL_STATE(5355)] = 274942,
  [SMALL_STATE(5356)] = 274966,
  [SMALL_STATE(5357)] = 274990,
  [SMALL_STATE(5358)] = 275014,
  [SMALL_STATE(5359)] = 275040,
  [SMALL_STATE(5360)] = 275066,
  [SMALL_STATE(5361)] = 275090,
  [SMALL_STATE(5362)] = 275116,
  [SMALL_STATE(5363)] = 275140,
  [SMALL_STATE(5364)] = 275166,
  [SMALL_STATE(5365)] = 275192,
  [SMALL_STATE(5366)] = 275218,
  [SMALL_STATE(5367)] = 275244,
  [SMALL_STATE(5368)] = 275268,
  [SMALL_STATE(5369)] = 275292,
  [SMALL_STATE(5370)] = 275316,
  [SMALL_STATE(5371)] = 275340,
  [SMALL_STATE(5372)] = 275364,
  [SMALL_STATE(5373)] = 275388,
  [SMALL_STATE(5374)] = 275412,
  [SMALL_STATE(5375)] = 275436,
  [SMALL_STATE(5376)] = 275460,
  [SMALL_STATE(5377)] = 275484,
  [SMALL_STATE(5378)] = 275508,
  [SMALL_STATE(5379)] = 275532,
  [SMALL_STATE(5380)] = 275558,
  [SMALL_STATE(5381)] = 275582,
  [SMALL_STATE(5382)] = 275606,
  [SMALL_STATE(5383)] = 275630,
  [SMALL_STATE(5384)] = 275654,
  [SMALL_STATE(5385)] = 275674,
  [SMALL_STATE(5386)] = 275698,
  [SMALL_STATE(5387)] = 275722,
  [SMALL_STATE(5388)] = 275746,
  [SMALL_STATE(5389)] = 275770,
  [SMALL_STATE(5390)] = 275794,
  [SMALL_STATE(5391)] = 275818,
  [SMALL_STATE(5392)] = 275842,
  [SMALL_STATE(5393)] = 275866,
  [SMALL_STATE(5394)] = 275892,
  [SMALL_STATE(5395)] = 275916,
  [SMALL_STATE(5396)] = 275940,
  [SMALL_STATE(5397)] = 275964,
  [SMALL_STATE(5398)] = 275988,
  [SMALL_STATE(5399)] = 276014,
  [SMALL_STATE(5400)] = 276038,
  [SMALL_STATE(5401)] = 276062,
  [SMALL_STATE(5402)] = 276086,
  [SMALL_STATE(5403)] = 276112,
  [SMALL_STATE(5404)] = 276138,
  [SMALL_STATE(5405)] = 276162,
  [SMALL_STATE(5406)] = 276186,
  [SMALL_STATE(5407)] = 276210,
  [SMALL_STATE(5408)] = 276234,
  [SMALL_STATE(5409)] = 276258,
  [SMALL_STATE(5410)] = 276282,
  [SMALL_STATE(5411)] = 276306,
  [SMALL_STATE(5412)] = 276330,
  [SMALL_STATE(5413)] = 276354,
  [SMALL_STATE(5414)] = 276378,
  [SMALL_STATE(5415)] = 276402,
  [SMALL_STATE(5416)] = 276426,
  [SMALL_STATE(5417)] = 276450,
  [SMALL_STATE(5418)] = 276474,
  [SMALL_STATE(5419)] = 276498,
  [SMALL_STATE(5420)] = 276522,
  [SMALL_STATE(5421)] = 276546,
  [SMALL_STATE(5422)] = 276570,
  [SMALL_STATE(5423)] = 276594,
  [SMALL_STATE(5424)] = 276618,
  [SMALL_STATE(5425)] = 276642,
  [SMALL_STATE(5426)] = 276666,
  [SMALL_STATE(5427)] = 276690,
  [SMALL_STATE(5428)] = 276714,
  [SMALL_STATE(5429)] = 276738,
  [SMALL_STATE(5430)] = 276762,
  [SMALL_STATE(5431)] = 276786,
  [SMALL_STATE(5432)] = 276810,
  [SMALL_STATE(5433)] = 276834,
  [SMALL_STATE(5434)] = 276858,
  [SMALL_STATE(5435)] = 276882,
  [SMALL_STATE(5436)] = 276906,
  [SMALL_STATE(5437)] = 276930,
  [SMALL_STATE(5438)] = 276954,
  [SMALL_STATE(5439)] = 276978,
  [SMALL_STATE(5440)] = 277002,
  [SMALL_STATE(5441)] = 277026,
  [SMALL_STATE(5442)] = 277050,
  [SMALL_STATE(5443)] = 277074,
  [SMALL_STATE(5444)] = 277098,
  [SMALL_STATE(5445)] = 277122,
  [SMALL_STATE(5446)] = 277146,
  [SMALL_STATE(5447)] = 277172,
  [SMALL_STATE(5448)] = 277196,
  [SMALL_STATE(5449)] = 277220,
  [SMALL_STATE(5450)] = 277244,
  [SMALL_STATE(5451)] = 277268,
  [SMALL_STATE(5452)] = 277292,
  [SMALL_STATE(5453)] = 277316,
  [SMALL_STATE(5454)] = 277340,
  [SMALL_STATE(5455)] = 277364,
  [SMALL_STATE(5456)] = 277388,
  [SMALL_STATE(5457)] = 277412,
  [SMALL_STATE(5458)] = 277436,
  [SMALL_STATE(5459)] = 277462,
  [SMALL_STATE(5460)] = 277486,
  [SMALL_STATE(5461)] = 277510,
  [SMALL_STATE(5462)] = 277534,
  [SMALL_STATE(5463)] = 277558,
  [SMALL_STATE(5464)] = 277582,
  [SMALL_STATE(5465)] = 277606,
  [SMALL_STATE(5466)] = 277632,
  [SMALL_STATE(5467)] = 277656,
  [SMALL_STATE(5468)] = 277682,
  [SMALL_STATE(5469)] = 277706,
  [SMALL_STATE(5470)] = 277730,
  [SMALL_STATE(5471)] = 277754,
  [SMALL_STATE(5472)] = 277778,
  [SMALL_STATE(5473)] = 277802,
  [SMALL_STATE(5474)] = 277826,
  [SMALL_STATE(5475)] = 277850,
  [SMALL_STATE(5476)] = 277874,
  [SMALL_STATE(5477)] = 277898,
  [SMALL_STATE(5478)] = 277922,
  [SMALL_STATE(5479)] = 277946,
  [SMALL_STATE(5480)] = 277970,
  [SMALL_STATE(5481)] = 277994,
  [SMALL_STATE(5482)] = 278018,
  [SMALL_STATE(5483)] = 278042,
  [SMALL_STATE(5484)] = 278066,
  [SMALL_STATE(5485)] = 278090,
  [SMALL_STATE(5486)] = 278114,
  [SMALL_STATE(5487)] = 278138,
  [SMALL_STATE(5488)] = 278162,
  [SMALL_STATE(5489)] = 278186,
  [SMALL_STATE(5490)] = 278210,
  [SMALL_STATE(5491)] = 278234,
  [SMALL_STATE(5492)] = 278258,
  [SMALL_STATE(5493)] = 278282,
  [SMALL_STATE(5494)] = 278308,
  [SMALL_STATE(5495)] = 278332,
  [SMALL_STATE(5496)] = 278356,
  [SMALL_STATE(5497)] = 278380,
  [SMALL_STATE(5498)] = 278404,
  [SMALL_STATE(5499)] = 278428,
  [SMALL_STATE(5500)] = 278452,
  [SMALL_STATE(5501)] = 278476,
  [SMALL_STATE(5502)] = 278500,
  [SMALL_STATE(5503)] = 278524,
  [SMALL_STATE(5504)] = 278550,
  [SMALL_STATE(5505)] = 278574,
  [SMALL_STATE(5506)] = 278598,
  [SMALL_STATE(5507)] = 278622,
  [SMALL_STATE(5508)] = 278646,
  [SMALL_STATE(5509)] = 278670,
  [SMALL_STATE(5510)] = 278694,
  [SMALL_STATE(5511)] = 278718,
  [SMALL_STATE(5512)] = 278742,
  [SMALL_STATE(5513)] = 278768,
  [SMALL_STATE(5514)] = 278794,
  [SMALL_STATE(5515)] = 278818,
  [SMALL_STATE(5516)] = 278842,
  [SMALL_STATE(5517)] = 278866,
  [SMALL_STATE(5518)] = 278890,
  [SMALL_STATE(5519)] = 278914,
  [SMALL_STATE(5520)] = 278938,
  [SMALL_STATE(5521)] = 278962,
  [SMALL_STATE(5522)] = 278986,
  [SMALL_STATE(5523)] = 279010,
  [SMALL_STATE(5524)] = 279034,
  [SMALL_STATE(5525)] = 279058,
  [SMALL_STATE(5526)] = 279082,
  [SMALL_STATE(5527)] = 279106,
  [SMALL_STATE(5528)] = 279130,
  [SMALL_STATE(5529)] = 279154,
  [SMALL_STATE(5530)] = 279178,
  [SMALL_STATE(5531)] = 279202,
  [SMALL_STATE(5532)] = 279226,
  [SMALL_STATE(5533)] = 279250,
  [SMALL_STATE(5534)] = 279274,
  [SMALL_STATE(5535)] = 279298,
  [SMALL_STATE(5536)] = 279322,
  [SMALL_STATE(5537)] = 279348,
  [SMALL_STATE(5538)] = 279374,
  [SMALL_STATE(5539)] = 279398,
  [SMALL_STATE(5540)] = 279422,
  [SMALL_STATE(5541)] = 279446,
  [SMALL_STATE(5542)] = 279470,
  [SMALL_STATE(5543)] = 279494,
  [SMALL_STATE(5544)] = 279518,
  [SMALL_STATE(5545)] = 279542,
  [SMALL_STATE(5546)] = 279566,
  [SMALL_STATE(5547)] = 279590,
  [SMALL_STATE(5548)] = 279614,
  [SMALL_STATE(5549)] = 279638,
  [SMALL_STATE(5550)] = 279662,
  [SMALL_STATE(5551)] = 279686,
  [SMALL_STATE(5552)] = 279710,
  [SMALL_STATE(5553)] = 279734,
  [SMALL_STATE(5554)] = 279758,
  [SMALL_STATE(5555)] = 279782,
  [SMALL_STATE(5556)] = 279806,
  [SMALL_STATE(5557)] = 279830,
  [SMALL_STATE(5558)] = 279854,
  [SMALL_STATE(5559)] = 279878,
  [SMALL_STATE(5560)] = 279902,
  [SMALL_STATE(5561)] = 279926,
  [SMALL_STATE(5562)] = 279950,
  [SMALL_STATE(5563)] = 279974,
  [SMALL_STATE(5564)] = 279998,
  [SMALL_STATE(5565)] = 280022,
  [SMALL_STATE(5566)] = 280046,
  [SMALL_STATE(5567)] = 280070,
  [SMALL_STATE(5568)] = 280094,
  [SMALL_STATE(5569)] = 280118,
  [SMALL_STATE(5570)] = 280142,
  [SMALL_STATE(5571)] = 280166,
  [SMALL_STATE(5572)] = 280190,
  [SMALL_STATE(5573)] = 280214,
  [SMALL_STATE(5574)] = 280238,
  [SMALL_STATE(5575)] = 280262,
  [SMALL_STATE(5576)] = 280286,
  [SMALL_STATE(5577)] = 280310,
  [SMALL_STATE(5578)] = 280334,
  [SMALL_STATE(5579)] = 280354,
  [SMALL_STATE(5580)] = 280378,
  [SMALL_STATE(5581)] = 280402,
  [SMALL_STATE(5582)] = 280426,
  [SMALL_STATE(5583)] = 280452,
  [SMALL_STATE(5584)] = 280476,
  [SMALL_STATE(5585)] = 280500,
  [SMALL_STATE(5586)] = 280524,
  [SMALL_STATE(5587)] = 280548,
  [SMALL_STATE(5588)] = 280572,
  [SMALL_STATE(5589)] = 280596,
  [SMALL_STATE(5590)] = 280620,
  [SMALL_STATE(5591)] = 280644,
  [SMALL_STATE(5592)] = 280668,
  [SMALL_STATE(5593)] = 280692,
  [SMALL_STATE(5594)] = 280716,
  [SMALL_STATE(5595)] = 280740,
  [SMALL_STATE(5596)] = 280760,
  [SMALL_STATE(5597)] = 280784,
  [SMALL_STATE(5598)] = 280810,
  [SMALL_STATE(5599)] = 280834,
  [SMALL_STATE(5600)] = 280858,
  [SMALL_STATE(5601)] = 280882,
  [SMALL_STATE(5602)] = 280906,
  [SMALL_STATE(5603)] = 280930,
  [SMALL_STATE(5604)] = 280954,
  [SMALL_STATE(5605)] = 280978,
  [SMALL_STATE(5606)] = 281002,
  [SMALL_STATE(5607)] = 281026,
  [SMALL_STATE(5608)] = 281050,
  [SMALL_STATE(5609)] = 281074,
  [SMALL_STATE(5610)] = 281098,
  [SMALL_STATE(5611)] = 281122,
  [SMALL_STATE(5612)] = 281146,
  [SMALL_STATE(5613)] = 281170,
  [SMALL_STATE(5614)] = 281194,
  [SMALL_STATE(5615)] = 281218,
  [SMALL_STATE(5616)] = 281242,
  [SMALL_STATE(5617)] = 281266,
  [SMALL_STATE(5618)] = 281290,
  [SMALL_STATE(5619)] = 281314,
  [SMALL_STATE(5620)] = 281338,
  [SMALL_STATE(5621)] = 281362,
  [SMALL_STATE(5622)] = 281386,
  [SMALL_STATE(5623)] = 281410,
  [SMALL_STATE(5624)] = 281434,
  [SMALL_STATE(5625)] = 281458,
  [SMALL_STATE(5626)] = 281482,
  [SMALL_STATE(5627)] = 281506,
  [SMALL_STATE(5628)] = 281530,
  [SMALL_STATE(5629)] = 281554,
  [SMALL_STATE(5630)] = 281578,
  [SMALL_STATE(5631)] = 281602,
  [SMALL_STATE(5632)] = 281626,
  [SMALL_STATE(5633)] = 281650,
  [SMALL_STATE(5634)] = 281674,
  [SMALL_STATE(5635)] = 281698,
  [SMALL_STATE(5636)] = 281722,
  [SMALL_STATE(5637)] = 281746,
  [SMALL_STATE(5638)] = 281770,
  [SMALL_STATE(5639)] = 281794,
  [SMALL_STATE(5640)] = 281818,
  [SMALL_STATE(5641)] = 281842,
  [SMALL_STATE(5642)] = 281866,
  [SMALL_STATE(5643)] = 281890,
  [SMALL_STATE(5644)] = 281914,
  [SMALL_STATE(5645)] = 281938,
  [SMALL_STATE(5646)] = 281962,
  [SMALL_STATE(5647)] = 281986,
  [SMALL_STATE(5648)] = 282010,
  [SMALL_STATE(5649)] = 282034,
  [SMALL_STATE(5650)] = 282058,
  [SMALL_STATE(5651)] = 282084,
  [SMALL_STATE(5652)] = 282108,
  [SMALL_STATE(5653)] = 282132,
  [SMALL_STATE(5654)] = 282156,
  [SMALL_STATE(5655)] = 282180,
  [SMALL_STATE(5656)] = 282204,
  [SMALL_STATE(5657)] = 282228,
  [SMALL_STATE(5658)] = 282252,
  [SMALL_STATE(5659)] = 282276,
  [SMALL_STATE(5660)] = 282300,
  [SMALL_STATE(5661)] = 282324,
  [SMALL_STATE(5662)] = 282348,
  [SMALL_STATE(5663)] = 282372,
  [SMALL_STATE(5664)] = 282396,
  [SMALL_STATE(5665)] = 282420,
  [SMALL_STATE(5666)] = 282444,
  [SMALL_STATE(5667)] = 282468,
  [SMALL_STATE(5668)] = 282492,
  [SMALL_STATE(5669)] = 282516,
  [SMALL_STATE(5670)] = 282540,
  [SMALL_STATE(5671)] = 282564,
  [SMALL_STATE(5672)] = 282590,
  [SMALL_STATE(5673)] = 282614,
  [SMALL_STATE(5674)] = 282638,
  [SMALL_STATE(5675)] = 282664,
  [SMALL_STATE(5676)] = 282688,
  [SMALL_STATE(5677)] = 282712,
  [SMALL_STATE(5678)] = 282736,
  [SMALL_STATE(5679)] = 282760,
  [SMALL_STATE(5680)] = 282784,
  [SMALL_STATE(5681)] = 282808,
  [SMALL_STATE(5682)] = 282832,
  [SMALL_STATE(5683)] = 282856,
  [SMALL_STATE(5684)] = 282880,
  [SMALL_STATE(5685)] = 282904,
  [SMALL_STATE(5686)] = 282928,
  [SMALL_STATE(5687)] = 282952,
  [SMALL_STATE(5688)] = 282976,
  [SMALL_STATE(5689)] = 283002,
  [SMALL_STATE(5690)] = 283026,
  [SMALL_STATE(5691)] = 283050,
  [SMALL_STATE(5692)] = 283074,
  [SMALL_STATE(5693)] = 283098,
  [SMALL_STATE(5694)] = 283122,
  [SMALL_STATE(5695)] = 283146,
  [SMALL_STATE(5696)] = 283170,
  [SMALL_STATE(5697)] = 283194,
  [SMALL_STATE(5698)] = 283218,
  [SMALL_STATE(5699)] = 283242,
  [SMALL_STATE(5700)] = 283266,
  [SMALL_STATE(5701)] = 283290,
  [SMALL_STATE(5702)] = 283314,
  [SMALL_STATE(5703)] = 283338,
  [SMALL_STATE(5704)] = 283362,
  [SMALL_STATE(5705)] = 283386,
  [SMALL_STATE(5706)] = 283410,
  [SMALL_STATE(5707)] = 283434,
  [SMALL_STATE(5708)] = 283458,
  [SMALL_STATE(5709)] = 283482,
  [SMALL_STATE(5710)] = 283506,
  [SMALL_STATE(5711)] = 283530,
  [SMALL_STATE(5712)] = 283554,
  [SMALL_STATE(5713)] = 283578,
  [SMALL_STATE(5714)] = 283602,
  [SMALL_STATE(5715)] = 283626,
  [SMALL_STATE(5716)] = 283650,
  [SMALL_STATE(5717)] = 283674,
  [SMALL_STATE(5718)] = 283698,
  [SMALL_STATE(5719)] = 283722,
  [SMALL_STATE(5720)] = 283746,
  [SMALL_STATE(5721)] = 283770,
  [SMALL_STATE(5722)] = 283796,
  [SMALL_STATE(5723)] = 283820,
  [SMALL_STATE(5724)] = 283846,
  [SMALL_STATE(5725)] = 283870,
  [SMALL_STATE(5726)] = 283894,
  [SMALL_STATE(5727)] = 283918,
  [SMALL_STATE(5728)] = 283942,
  [SMALL_STATE(5729)] = 283966,
  [SMALL_STATE(5730)] = 283990,
  [SMALL_STATE(5731)] = 284014,
  [SMALL_STATE(5732)] = 284038,
  [SMALL_STATE(5733)] = 284062,
  [SMALL_STATE(5734)] = 284086,
  [SMALL_STATE(5735)] = 284110,
  [SMALL_STATE(5736)] = 284134,
  [SMALL_STATE(5737)] = 284158,
  [SMALL_STATE(5738)] = 284182,
  [SMALL_STATE(5739)] = 284208,
  [SMALL_STATE(5740)] = 284234,
  [SMALL_STATE(5741)] = 284258,
  [SMALL_STATE(5742)] = 284282,
  [SMALL_STATE(5743)] = 284306,
  [SMALL_STATE(5744)] = 284330,
  [SMALL_STATE(5745)] = 284354,
  [SMALL_STATE(5746)] = 284378,
  [SMALL_STATE(5747)] = 284402,
  [SMALL_STATE(5748)] = 284426,
  [SMALL_STATE(5749)] = 284450,
  [SMALL_STATE(5750)] = 284474,
  [SMALL_STATE(5751)] = 284498,
  [SMALL_STATE(5752)] = 284522,
  [SMALL_STATE(5753)] = 284548,
  [SMALL_STATE(5754)] = 284574,
  [SMALL_STATE(5755)] = 284600,
  [SMALL_STATE(5756)] = 284624,
  [SMALL_STATE(5757)] = 284648,
  [SMALL_STATE(5758)] = 284672,
  [SMALL_STATE(5759)] = 284696,
  [SMALL_STATE(5760)] = 284720,
  [SMALL_STATE(5761)] = 284744,
  [SMALL_STATE(5762)] = 284768,
  [SMALL_STATE(5763)] = 284792,
  [SMALL_STATE(5764)] = 284816,
  [SMALL_STATE(5765)] = 284840,
  [SMALL_STATE(5766)] = 284864,
  [SMALL_STATE(5767)] = 284888,
  [SMALL_STATE(5768)] = 284912,
  [SMALL_STATE(5769)] = 284936,
  [SMALL_STATE(5770)] = 284960,
  [SMALL_STATE(5771)] = 284984,
  [SMALL_STATE(5772)] = 285008,
  [SMALL_STATE(5773)] = 285032,
  [SMALL_STATE(5774)] = 285056,
  [SMALL_STATE(5775)] = 285080,
  [SMALL_STATE(5776)] = 285104,
  [SMALL_STATE(5777)] = 285128,
  [SMALL_STATE(5778)] = 285152,
  [SMALL_STATE(5779)] = 285176,
  [SMALL_STATE(5780)] = 285200,
  [SMALL_STATE(5781)] = 285224,
  [SMALL_STATE(5782)] = 285248,
  [SMALL_STATE(5783)] = 285272,
  [SMALL_STATE(5784)] = 285296,
  [SMALL_STATE(5785)] = 285322,
  [SMALL_STATE(5786)] = 285346,
  [SMALL_STATE(5787)] = 285370,
  [SMALL_STATE(5788)] = 285394,
  [SMALL_STATE(5789)] = 285418,
  [SMALL_STATE(5790)] = 285442,
  [SMALL_STATE(5791)] = 285466,
  [SMALL_STATE(5792)] = 285490,
  [SMALL_STATE(5793)] = 285514,
  [SMALL_STATE(5794)] = 285538,
  [SMALL_STATE(5795)] = 285562,
  [SMALL_STATE(5796)] = 285586,
  [SMALL_STATE(5797)] = 285610,
  [SMALL_STATE(5798)] = 285634,
  [SMALL_STATE(5799)] = 285658,
  [SMALL_STATE(5800)] = 285682,
  [SMALL_STATE(5801)] = 285706,
  [SMALL_STATE(5802)] = 285730,
  [SMALL_STATE(5803)] = 285756,
  [SMALL_STATE(5804)] = 285782,
  [SMALL_STATE(5805)] = 285806,
  [SMALL_STATE(5806)] = 285830,
  [SMALL_STATE(5807)] = 285854,
  [SMALL_STATE(5808)] = 285878,
  [SMALL_STATE(5809)] = 285902,
  [SMALL_STATE(5810)] = 285926,
  [SMALL_STATE(5811)] = 285950,
  [SMALL_STATE(5812)] = 285974,
  [SMALL_STATE(5813)] = 285998,
  [SMALL_STATE(5814)] = 286022,
  [SMALL_STATE(5815)] = 286046,
  [SMALL_STATE(5816)] = 286070,
  [SMALL_STATE(5817)] = 286096,
  [SMALL_STATE(5818)] = 286120,
  [SMALL_STATE(5819)] = 286144,
  [SMALL_STATE(5820)] = 286168,
  [SMALL_STATE(5821)] = 286192,
  [SMALL_STATE(5822)] = 286216,
  [SMALL_STATE(5823)] = 286240,
  [SMALL_STATE(5824)] = 286264,
  [SMALL_STATE(5825)] = 286288,
  [SMALL_STATE(5826)] = 286312,
  [SMALL_STATE(5827)] = 286336,
  [SMALL_STATE(5828)] = 286360,
  [SMALL_STATE(5829)] = 286384,
  [SMALL_STATE(5830)] = 286408,
  [SMALL_STATE(5831)] = 286432,
  [SMALL_STATE(5832)] = 286456,
  [SMALL_STATE(5833)] = 286480,
  [SMALL_STATE(5834)] = 286504,
  [SMALL_STATE(5835)] = 286528,
  [SMALL_STATE(5836)] = 286552,
  [SMALL_STATE(5837)] = 286576,
  [SMALL_STATE(5838)] = 286600,
  [SMALL_STATE(5839)] = 286624,
  [SMALL_STATE(5840)] = 286648,
  [SMALL_STATE(5841)] = 286672,
  [SMALL_STATE(5842)] = 286696,
  [SMALL_STATE(5843)] = 286720,
  [SMALL_STATE(5844)] = 286744,
  [SMALL_STATE(5845)] = 286768,
  [SMALL_STATE(5846)] = 286792,
  [SMALL_STATE(5847)] = 286816,
  [SMALL_STATE(5848)] = 286840,
  [SMALL_STATE(5849)] = 286866,
  [SMALL_STATE(5850)] = 286892,
  [SMALL_STATE(5851)] = 286916,
  [SMALL_STATE(5852)] = 286940,
  [SMALL_STATE(5853)] = 286964,
  [SMALL_STATE(5854)] = 286988,
  [SMALL_STATE(5855)] = 287012,
  [SMALL_STATE(5856)] = 287036,
  [SMALL_STATE(5857)] = 287060,
  [SMALL_STATE(5858)] = 287084,
  [SMALL_STATE(5859)] = 287108,
  [SMALL_STATE(5860)] = 287132,
  [SMALL_STATE(5861)] = 287156,
  [SMALL_STATE(5862)] = 287180,
  [SMALL_STATE(5863)] = 287204,
  [SMALL_STATE(5864)] = 287228,
  [SMALL_STATE(5865)] = 287252,
  [SMALL_STATE(5866)] = 287276,
  [SMALL_STATE(5867)] = 287300,
  [SMALL_STATE(5868)] = 287324,
  [SMALL_STATE(5869)] = 287348,
  [SMALL_STATE(5870)] = 287372,
  [SMALL_STATE(5871)] = 287396,
  [SMALL_STATE(5872)] = 287420,
  [SMALL_STATE(5873)] = 287444,
  [SMALL_STATE(5874)] = 287468,
  [SMALL_STATE(5875)] = 287492,
  [SMALL_STATE(5876)] = 287516,
  [SMALL_STATE(5877)] = 287540,
  [SMALL_STATE(5878)] = 287564,
  [SMALL_STATE(5879)] = 287588,
  [SMALL_STATE(5880)] = 287612,
  [SMALL_STATE(5881)] = 287636,
  [SMALL_STATE(5882)] = 287660,
  [SMALL_STATE(5883)] = 287684,
  [SMALL_STATE(5884)] = 287708,
  [SMALL_STATE(5885)] = 287732,
  [SMALL_STATE(5886)] = 287756,
  [SMALL_STATE(5887)] = 287780,
  [SMALL_STATE(5888)] = 287804,
  [SMALL_STATE(5889)] = 287828,
  [SMALL_STATE(5890)] = 287852,
  [SMALL_STATE(5891)] = 287876,
  [SMALL_STATE(5892)] = 287900,
  [SMALL_STATE(5893)] = 287924,
  [SMALL_STATE(5894)] = 287948,
  [SMALL_STATE(5895)] = 287972,
  [SMALL_STATE(5896)] = 287996,
  [SMALL_STATE(5897)] = 288020,
  [SMALL_STATE(5898)] = 288044,
  [SMALL_STATE(5899)] = 288068,
  [SMALL_STATE(5900)] = 288092,
  [SMALL_STATE(5901)] = 288116,
  [SMALL_STATE(5902)] = 288140,
  [SMALL_STATE(5903)] = 288164,
  [SMALL_STATE(5904)] = 288188,
  [SMALL_STATE(5905)] = 288214,
  [SMALL_STATE(5906)] = 288238,
  [SMALL_STATE(5907)] = 288262,
  [SMALL_STATE(5908)] = 288286,
  [SMALL_STATE(5909)] = 288310,
  [SMALL_STATE(5910)] = 288334,
  [SMALL_STATE(5911)] = 288358,
  [SMALL_STATE(5912)] = 288382,
  [SMALL_STATE(5913)] = 288406,
  [SMALL_STATE(5914)] = 288430,
  [SMALL_STATE(5915)] = 288454,
  [SMALL_STATE(5916)] = 288478,
  [SMALL_STATE(5917)] = 288502,
  [SMALL_STATE(5918)] = 288526,
  [SMALL_STATE(5919)] = 288550,
  [SMALL_STATE(5920)] = 288574,
  [SMALL_STATE(5921)] = 288598,
  [SMALL_STATE(5922)] = 288622,
  [SMALL_STATE(5923)] = 288646,
  [SMALL_STATE(5924)] = 288670,
  [SMALL_STATE(5925)] = 288694,
  [SMALL_STATE(5926)] = 288718,
  [SMALL_STATE(5927)] = 288742,
  [SMALL_STATE(5928)] = 288766,
  [SMALL_STATE(5929)] = 288792,
  [SMALL_STATE(5930)] = 288818,
  [SMALL_STATE(5931)] = 288842,
  [SMALL_STATE(5932)] = 288868,
  [SMALL_STATE(5933)] = 288892,
  [SMALL_STATE(5934)] = 288916,
  [SMALL_STATE(5935)] = 288940,
  [SMALL_STATE(5936)] = 288964,
  [SMALL_STATE(5937)] = 288990,
  [SMALL_STATE(5938)] = 289016,
  [SMALL_STATE(5939)] = 289040,
  [SMALL_STATE(5940)] = 289066,
  [SMALL_STATE(5941)] = 289090,
  [SMALL_STATE(5942)] = 289114,
  [SMALL_STATE(5943)] = 289138,
  [SMALL_STATE(5944)] = 289164,
  [SMALL_STATE(5945)] = 289188,
  [SMALL_STATE(5946)] = 289212,
  [SMALL_STATE(5947)] = 289232,
  [SMALL_STATE(5948)] = 289256,
  [SMALL_STATE(5949)] = 289282,
  [SMALL_STATE(5950)] = 289308,
  [SMALL_STATE(5951)] = 289334,
  [SMALL_STATE(5952)] = 289360,
  [SMALL_STATE(5953)] = 289384,
  [SMALL_STATE(5954)] = 289408,
  [SMALL_STATE(5955)] = 289432,
  [SMALL_STATE(5956)] = 289458,
  [SMALL_STATE(5957)] = 289484,
  [SMALL_STATE(5958)] = 289508,
  [SMALL_STATE(5959)] = 289532,
  [SMALL_STATE(5960)] = 289556,
  [SMALL_STATE(5961)] = 289580,
  [SMALL_STATE(5962)] = 289597,
  [SMALL_STATE(5963)] = 289618,
  [SMALL_STATE(5964)] = 289639,
  [SMALL_STATE(5965)] = 289656,
  [SMALL_STATE(5966)] = 289677,
  [SMALL_STATE(5967)] = 289698,
  [SMALL_STATE(5968)] = 289719,
  [SMALL_STATE(5969)] = 289740,
  [SMALL_STATE(5970)] = 289761,
  [SMALL_STATE(5971)] = 289782,
  [SMALL_STATE(5972)] = 289803,
  [SMALL_STATE(5973)] = 289824,
  [SMALL_STATE(5974)] = 289845,
  [SMALL_STATE(5975)] = 289866,
  [SMALL_STATE(5976)] = 289887,
  [SMALL_STATE(5977)] = 289908,
  [SMALL_STATE(5978)] = 289929,
  [SMALL_STATE(5979)] = 289950,
  [SMALL_STATE(5980)] = 289971,
  [SMALL_STATE(5981)] = 289992,
  [SMALL_STATE(5982)] = 290013,
  [SMALL_STATE(5983)] = 290034,
  [SMALL_STATE(5984)] = 290055,
  [SMALL_STATE(5985)] = 290072,
  [SMALL_STATE(5986)] = 290089,
  [SMALL_STATE(5987)] = 290110,
  [SMALL_STATE(5988)] = 290131,
  [SMALL_STATE(5989)] = 290148,
  [SMALL_STATE(5990)] = 290165,
  [SMALL_STATE(5991)] = 290182,
  [SMALL_STATE(5992)] = 290199,
  [SMALL_STATE(5993)] = 290216,
  [SMALL_STATE(5994)] = 290237,
  [SMALL_STATE(5995)] = 290258,
  [SMALL_STATE(5996)] = 290279,
  [SMALL_STATE(5997)] = 290300,
  [SMALL_STATE(5998)] = 290321,
  [SMALL_STATE(5999)] = 290342,
  [SMALL_STATE(6000)] = 290363,
  [SMALL_STATE(6001)] = 290384,
  [SMALL_STATE(6002)] = 290405,
  [SMALL_STATE(6003)] = 290426,
  [SMALL_STATE(6004)] = 290447,
  [SMALL_STATE(6005)] = 290468,
  [SMALL_STATE(6006)] = 290489,
  [SMALL_STATE(6007)] = 290510,
  [SMALL_STATE(6008)] = 290531,
  [SMALL_STATE(6009)] = 290552,
  [SMALL_STATE(6010)] = 290573,
  [SMALL_STATE(6011)] = 290594,
  [SMALL_STATE(6012)] = 290615,
  [SMALL_STATE(6013)] = 290636,
  [SMALL_STATE(6014)] = 290655,
  [SMALL_STATE(6015)] = 290676,
  [SMALL_STATE(6016)] = 290697,
  [SMALL_STATE(6017)] = 290718,
  [SMALL_STATE(6018)] = 290739,
  [SMALL_STATE(6019)] = 290760,
  [SMALL_STATE(6020)] = 290781,
  [SMALL_STATE(6021)] = 290802,
  [SMALL_STATE(6022)] = 290823,
  [SMALL_STATE(6023)] = 290844,
  [SMALL_STATE(6024)] = 290865,
  [SMALL_STATE(6025)] = 290886,
  [SMALL_STATE(6026)] = 290907,
  [SMALL_STATE(6027)] = 290928,
  [SMALL_STATE(6028)] = 290949,
  [SMALL_STATE(6029)] = 290970,
  [SMALL_STATE(6030)] = 290991,
  [SMALL_STATE(6031)] = 291012,
  [SMALL_STATE(6032)] = 291033,
  [SMALL_STATE(6033)] = 291054,
  [SMALL_STATE(6034)] = 291075,
  [SMALL_STATE(6035)] = 291096,
  [SMALL_STATE(6036)] = 291117,
  [SMALL_STATE(6037)] = 291138,
  [SMALL_STATE(6038)] = 291159,
  [SMALL_STATE(6039)] = 291180,
  [SMALL_STATE(6040)] = 291197,
  [SMALL_STATE(6041)] = 291218,
  [SMALL_STATE(6042)] = 291239,
  [SMALL_STATE(6043)] = 291260,
  [SMALL_STATE(6044)] = 291281,
  [SMALL_STATE(6045)] = 291302,
  [SMALL_STATE(6046)] = 291323,
  [SMALL_STATE(6047)] = 291344,
  [SMALL_STATE(6048)] = 291365,
  [SMALL_STATE(6049)] = 291386,
  [SMALL_STATE(6050)] = 291403,
  [SMALL_STATE(6051)] = 291424,
  [SMALL_STATE(6052)] = 291445,
  [SMALL_STATE(6053)] = 291466,
  [SMALL_STATE(6054)] = 291487,
  [SMALL_STATE(6055)] = 291508,
  [SMALL_STATE(6056)] = 291529,
  [SMALL_STATE(6057)] = 291550,
  [SMALL_STATE(6058)] = 291571,
  [SMALL_STATE(6059)] = 291592,
  [SMALL_STATE(6060)] = 291613,
  [SMALL_STATE(6061)] = 291634,
  [SMALL_STATE(6062)] = 291653,
  [SMALL_STATE(6063)] = 291670,
  [SMALL_STATE(6064)] = 291691,
  [SMALL_STATE(6065)] = 291712,
  [SMALL_STATE(6066)] = 291733,
  [SMALL_STATE(6067)] = 291754,
  [SMALL_STATE(6068)] = 291775,
  [SMALL_STATE(6069)] = 291796,
  [SMALL_STATE(6070)] = 291817,
  [SMALL_STATE(6071)] = 291838,
  [SMALL_STATE(6072)] = 291859,
  [SMALL_STATE(6073)] = 291880,
  [SMALL_STATE(6074)] = 291901,
  [SMALL_STATE(6075)] = 291922,
  [SMALL_STATE(6076)] = 291943,
  [SMALL_STATE(6077)] = 291964,
  [SMALL_STATE(6078)] = 291985,
  [SMALL_STATE(6079)] = 292006,
  [SMALL_STATE(6080)] = 292027,
  [SMALL_STATE(6081)] = 292048,
  [SMALL_STATE(6082)] = 292069,
  [SMALL_STATE(6083)] = 292090,
  [SMALL_STATE(6084)] = 292111,
  [SMALL_STATE(6085)] = 292132,
  [SMALL_STATE(6086)] = 292149,
  [SMALL_STATE(6087)] = 292170,
  [SMALL_STATE(6088)] = 292191,
  [SMALL_STATE(6089)] = 292212,
  [SMALL_STATE(6090)] = 292233,
  [SMALL_STATE(6091)] = 292254,
  [SMALL_STATE(6092)] = 292275,
  [SMALL_STATE(6093)] = 292296,
  [SMALL_STATE(6094)] = 292317,
  [SMALL_STATE(6095)] = 292338,
  [SMALL_STATE(6096)] = 292359,
  [SMALL_STATE(6097)] = 292380,
  [SMALL_STATE(6098)] = 292401,
  [SMALL_STATE(6099)] = 292422,
  [SMALL_STATE(6100)] = 292443,
  [SMALL_STATE(6101)] = 292464,
  [SMALL_STATE(6102)] = 292485,
  [SMALL_STATE(6103)] = 292506,
  [SMALL_STATE(6104)] = 292527,
  [SMALL_STATE(6105)] = 292548,
  [SMALL_STATE(6106)] = 292569,
  [SMALL_STATE(6107)] = 292590,
  [SMALL_STATE(6108)] = 292611,
  [SMALL_STATE(6109)] = 292632,
  [SMALL_STATE(6110)] = 292653,
  [SMALL_STATE(6111)] = 292674,
  [SMALL_STATE(6112)] = 292695,
  [SMALL_STATE(6113)] = 292716,
  [SMALL_STATE(6114)] = 292737,
  [SMALL_STATE(6115)] = 292754,
  [SMALL_STATE(6116)] = 292775,
  [SMALL_STATE(6117)] = 292796,
  [SMALL_STATE(6118)] = 292817,
  [SMALL_STATE(6119)] = 292834,
  [SMALL_STATE(6120)] = 292855,
  [SMALL_STATE(6121)] = 292876,
  [SMALL_STATE(6122)] = 292897,
  [SMALL_STATE(6123)] = 292918,
  [SMALL_STATE(6124)] = 292939,
  [SMALL_STATE(6125)] = 292960,
  [SMALL_STATE(6126)] = 292981,
  [SMALL_STATE(6127)] = 293002,
  [SMALL_STATE(6128)] = 293023,
  [SMALL_STATE(6129)] = 293044,
  [SMALL_STATE(6130)] = 293065,
  [SMALL_STATE(6131)] = 293086,
  [SMALL_STATE(6132)] = 293107,
  [SMALL_STATE(6133)] = 293128,
  [SMALL_STATE(6134)] = 293149,
  [SMALL_STATE(6135)] = 293170,
  [SMALL_STATE(6136)] = 293191,
  [SMALL_STATE(6137)] = 293212,
  [SMALL_STATE(6138)] = 293233,
  [SMALL_STATE(6139)] = 293254,
  [SMALL_STATE(6140)] = 293275,
  [SMALL_STATE(6141)] = 293296,
  [SMALL_STATE(6142)] = 293317,
  [SMALL_STATE(6143)] = 293338,
  [SMALL_STATE(6144)] = 293359,
  [SMALL_STATE(6145)] = 293380,
  [SMALL_STATE(6146)] = 293401,
  [SMALL_STATE(6147)] = 293418,
  [SMALL_STATE(6148)] = 293439,
  [SMALL_STATE(6149)] = 293456,
  [SMALL_STATE(6150)] = 293473,
  [SMALL_STATE(6151)] = 293494,
  [SMALL_STATE(6152)] = 293515,
  [SMALL_STATE(6153)] = 293536,
  [SMALL_STATE(6154)] = 293557,
  [SMALL_STATE(6155)] = 293578,
  [SMALL_STATE(6156)] = 293599,
  [SMALL_STATE(6157)] = 293620,
  [SMALL_STATE(6158)] = 293641,
  [SMALL_STATE(6159)] = 293662,
  [SMALL_STATE(6160)] = 293683,
  [SMALL_STATE(6161)] = 293704,
  [SMALL_STATE(6162)] = 293725,
  [SMALL_STATE(6163)] = 293746,
  [SMALL_STATE(6164)] = 293767,
  [SMALL_STATE(6165)] = 293788,
  [SMALL_STATE(6166)] = 293809,
  [SMALL_STATE(6167)] = 293830,
  [SMALL_STATE(6168)] = 293851,
  [SMALL_STATE(6169)] = 293872,
  [SMALL_STATE(6170)] = 293893,
  [SMALL_STATE(6171)] = 293914,
  [SMALL_STATE(6172)] = 293931,
  [SMALL_STATE(6173)] = 293948,
  [SMALL_STATE(6174)] = 293965,
  [SMALL_STATE(6175)] = 293982,
  [SMALL_STATE(6176)] = 293999,
  [SMALL_STATE(6177)] = 294020,
  [SMALL_STATE(6178)] = 294039,
  [SMALL_STATE(6179)] = 294056,
  [SMALL_STATE(6180)] = 294073,
  [SMALL_STATE(6181)] = 294090,
  [SMALL_STATE(6182)] = 294107,
  [SMALL_STATE(6183)] = 294124,
  [SMALL_STATE(6184)] = 294141,
  [SMALL_STATE(6185)] = 294158,
  [SMALL_STATE(6186)] = 294179,
  [SMALL_STATE(6187)] = 294197,
  [SMALL_STATE(6188)] = 294215,
  [SMALL_STATE(6189)] = 294233,
  [SMALL_STATE(6190)] = 294251,
  [SMALL_STATE(6191)] = 294269,
  [SMALL_STATE(6192)] = 294287,
  [SMALL_STATE(6193)] = 294305,
  [SMALL_STATE(6194)] = 294323,
  [SMALL_STATE(6195)] = 294341,
  [SMALL_STATE(6196)] = 294359,
  [SMALL_STATE(6197)] = 294377,
  [SMALL_STATE(6198)] = 294395,
  [SMALL_STATE(6199)] = 294413,
  [SMALL_STATE(6200)] = 294431,
  [SMALL_STATE(6201)] = 294449,
  [SMALL_STATE(6202)] = 294463,
  [SMALL_STATE(6203)] = 294481,
  [SMALL_STATE(6204)] = 294499,
  [SMALL_STATE(6205)] = 294513,
  [SMALL_STATE(6206)] = 294529,
  [SMALL_STATE(6207)] = 294547,
  [SMALL_STATE(6208)] = 294565,
  [SMALL_STATE(6209)] = 294583,
  [SMALL_STATE(6210)] = 294601,
  [SMALL_STATE(6211)] = 294619,
  [SMALL_STATE(6212)] = 294637,
  [SMALL_STATE(6213)] = 294655,
  [SMALL_STATE(6214)] = 294673,
  [SMALL_STATE(6215)] = 294691,
  [SMALL_STATE(6216)] = 294709,
  [SMALL_STATE(6217)] = 294727,
  [SMALL_STATE(6218)] = 294745,
  [SMALL_STATE(6219)] = 294763,
  [SMALL_STATE(6220)] = 294781,
  [SMALL_STATE(6221)] = 294799,
  [SMALL_STATE(6222)] = 294817,
  [SMALL_STATE(6223)] = 294835,
  [SMALL_STATE(6224)] = 294853,
  [SMALL_STATE(6225)] = 294871,
  [SMALL_STATE(6226)] = 294889,
  [SMALL_STATE(6227)] = 294907,
  [SMALL_STATE(6228)] = 294925,
  [SMALL_STATE(6229)] = 294943,
  [SMALL_STATE(6230)] = 294961,
  [SMALL_STATE(6231)] = 294979,
  [SMALL_STATE(6232)] = 294997,
  [SMALL_STATE(6233)] = 295015,
  [SMALL_STATE(6234)] = 295033,
  [SMALL_STATE(6235)] = 295051,
  [SMALL_STATE(6236)] = 295069,
  [SMALL_STATE(6237)] = 295087,
  [SMALL_STATE(6238)] = 295105,
  [SMALL_STATE(6239)] = 295123,
  [SMALL_STATE(6240)] = 295141,
  [SMALL_STATE(6241)] = 295159,
  [SMALL_STATE(6242)] = 295177,
  [SMALL_STATE(6243)] = 295195,
  [SMALL_STATE(6244)] = 295213,
  [SMALL_STATE(6245)] = 295231,
  [SMALL_STATE(6246)] = 295249,
  [SMALL_STATE(6247)] = 295267,
  [SMALL_STATE(6248)] = 295285,
  [SMALL_STATE(6249)] = 295303,
  [SMALL_STATE(6250)] = 295321,
  [SMALL_STATE(6251)] = 295339,
  [SMALL_STATE(6252)] = 295357,
  [SMALL_STATE(6253)] = 295375,
  [SMALL_STATE(6254)] = 295393,
  [SMALL_STATE(6255)] = 295411,
  [SMALL_STATE(6256)] = 295429,
  [SMALL_STATE(6257)] = 295447,
  [SMALL_STATE(6258)] = 295465,
  [SMALL_STATE(6259)] = 295483,
  [SMALL_STATE(6260)] = 295501,
  [SMALL_STATE(6261)] = 295519,
  [SMALL_STATE(6262)] = 295537,
  [SMALL_STATE(6263)] = 295555,
  [SMALL_STATE(6264)] = 295573,
  [SMALL_STATE(6265)] = 295591,
  [SMALL_STATE(6266)] = 295609,
  [SMALL_STATE(6267)] = 295627,
  [SMALL_STATE(6268)] = 295645,
  [SMALL_STATE(6269)] = 295663,
  [SMALL_STATE(6270)] = 295681,
  [SMALL_STATE(6271)] = 295699,
  [SMALL_STATE(6272)] = 295717,
  [SMALL_STATE(6273)] = 295735,
  [SMALL_STATE(6274)] = 295753,
  [SMALL_STATE(6275)] = 295771,
  [SMALL_STATE(6276)] = 295789,
  [SMALL_STATE(6277)] = 295807,
  [SMALL_STATE(6278)] = 295825,
  [SMALL_STATE(6279)] = 295843,
  [SMALL_STATE(6280)] = 295861,
  [SMALL_STATE(6281)] = 295879,
  [SMALL_STATE(6282)] = 295897,
  [SMALL_STATE(6283)] = 295915,
  [SMALL_STATE(6284)] = 295933,
  [SMALL_STATE(6285)] = 295951,
  [SMALL_STATE(6286)] = 295969,
  [SMALL_STATE(6287)] = 295987,
  [SMALL_STATE(6288)] = 296005,
  [SMALL_STATE(6289)] = 296023,
  [SMALL_STATE(6290)] = 296041,
  [SMALL_STATE(6291)] = 296059,
  [SMALL_STATE(6292)] = 296077,
  [SMALL_STATE(6293)] = 296095,
  [SMALL_STATE(6294)] = 296113,
  [SMALL_STATE(6295)] = 296131,
  [SMALL_STATE(6296)] = 296149,
  [SMALL_STATE(6297)] = 296167,
  [SMALL_STATE(6298)] = 296185,
  [SMALL_STATE(6299)] = 296203,
  [SMALL_STATE(6300)] = 296221,
  [SMALL_STATE(6301)] = 296239,
  [SMALL_STATE(6302)] = 296257,
  [SMALL_STATE(6303)] = 296275,
  [SMALL_STATE(6304)] = 296293,
  [SMALL_STATE(6305)] = 296311,
  [SMALL_STATE(6306)] = 296329,
  [SMALL_STATE(6307)] = 296347,
  [SMALL_STATE(6308)] = 296365,
  [SMALL_STATE(6309)] = 296383,
  [SMALL_STATE(6310)] = 296401,
  [SMALL_STATE(6311)] = 296419,
  [SMALL_STATE(6312)] = 296437,
  [SMALL_STATE(6313)] = 296455,
  [SMALL_STATE(6314)] = 296473,
  [SMALL_STATE(6315)] = 296491,
  [SMALL_STATE(6316)] = 296509,
  [SMALL_STATE(6317)] = 296527,
  [SMALL_STATE(6318)] = 296545,
  [SMALL_STATE(6319)] = 296563,
  [SMALL_STATE(6320)] = 296581,
  [SMALL_STATE(6321)] = 296599,
  [SMALL_STATE(6322)] = 296617,
  [SMALL_STATE(6323)] = 296635,
  [SMALL_STATE(6324)] = 296653,
  [SMALL_STATE(6325)] = 296671,
  [SMALL_STATE(6326)] = 296689,
  [SMALL_STATE(6327)] = 296707,
  [SMALL_STATE(6328)] = 296725,
  [SMALL_STATE(6329)] = 296743,
  [SMALL_STATE(6330)] = 296761,
  [SMALL_STATE(6331)] = 296779,
  [SMALL_STATE(6332)] = 296797,
  [SMALL_STATE(6333)] = 296815,
  [SMALL_STATE(6334)] = 296833,
  [SMALL_STATE(6335)] = 296851,
  [SMALL_STATE(6336)] = 296869,
  [SMALL_STATE(6337)] = 296887,
  [SMALL_STATE(6338)] = 296905,
  [SMALL_STATE(6339)] = 296923,
  [SMALL_STATE(6340)] = 296941,
  [SMALL_STATE(6341)] = 296959,
  [SMALL_STATE(6342)] = 296977,
  [SMALL_STATE(6343)] = 296995,
  [SMALL_STATE(6344)] = 297013,
  [SMALL_STATE(6345)] = 297031,
  [SMALL_STATE(6346)] = 297049,
  [SMALL_STATE(6347)] = 297067,
  [SMALL_STATE(6348)] = 297085,
  [SMALL_STATE(6349)] = 297103,
  [SMALL_STATE(6350)] = 297121,
  [SMALL_STATE(6351)] = 297139,
  [SMALL_STATE(6352)] = 297157,
  [SMALL_STATE(6353)] = 297175,
  [SMALL_STATE(6354)] = 297193,
  [SMALL_STATE(6355)] = 297211,
  [SMALL_STATE(6356)] = 297229,
  [SMALL_STATE(6357)] = 297247,
  [SMALL_STATE(6358)] = 297265,
  [SMALL_STATE(6359)] = 297283,
  [SMALL_STATE(6360)] = 297301,
  [SMALL_STATE(6361)] = 297319,
  [SMALL_STATE(6362)] = 297337,
  [SMALL_STATE(6363)] = 297355,
  [SMALL_STATE(6364)] = 297373,
  [SMALL_STATE(6365)] = 297391,
  [SMALL_STATE(6366)] = 297409,
  [SMALL_STATE(6367)] = 297427,
  [SMALL_STATE(6368)] = 297445,
  [SMALL_STATE(6369)] = 297463,
  [SMALL_STATE(6370)] = 297481,
  [SMALL_STATE(6371)] = 297499,
  [SMALL_STATE(6372)] = 297517,
  [SMALL_STATE(6373)] = 297535,
  [SMALL_STATE(6374)] = 297553,
  [SMALL_STATE(6375)] = 297571,
  [SMALL_STATE(6376)] = 297589,
  [SMALL_STATE(6377)] = 297607,
  [SMALL_STATE(6378)] = 297625,
  [SMALL_STATE(6379)] = 297643,
  [SMALL_STATE(6380)] = 297661,
  [SMALL_STATE(6381)] = 297679,
  [SMALL_STATE(6382)] = 297697,
  [SMALL_STATE(6383)] = 297715,
  [SMALL_STATE(6384)] = 297733,
  [SMALL_STATE(6385)] = 297751,
  [SMALL_STATE(6386)] = 297769,
  [SMALL_STATE(6387)] = 297787,
  [SMALL_STATE(6388)] = 297805,
  [SMALL_STATE(6389)] = 297823,
  [SMALL_STATE(6390)] = 297841,
  [SMALL_STATE(6391)] = 297859,
  [SMALL_STATE(6392)] = 297877,
  [SMALL_STATE(6393)] = 297895,
  [SMALL_STATE(6394)] = 297913,
  [SMALL_STATE(6395)] = 297931,
  [SMALL_STATE(6396)] = 297949,
  [SMALL_STATE(6397)] = 297967,
  [SMALL_STATE(6398)] = 297985,
  [SMALL_STATE(6399)] = 298003,
  [SMALL_STATE(6400)] = 298021,
  [SMALL_STATE(6401)] = 298039,
  [SMALL_STATE(6402)] = 298057,
  [SMALL_STATE(6403)] = 298075,
  [SMALL_STATE(6404)] = 298093,
  [SMALL_STATE(6405)] = 298111,
  [SMALL_STATE(6406)] = 298129,
  [SMALL_STATE(6407)] = 298147,
  [SMALL_STATE(6408)] = 298165,
  [SMALL_STATE(6409)] = 298183,
  [SMALL_STATE(6410)] = 298197,
  [SMALL_STATE(6411)] = 298211,
  [SMALL_STATE(6412)] = 298229,
  [SMALL_STATE(6413)] = 298247,
  [SMALL_STATE(6414)] = 298265,
  [SMALL_STATE(6415)] = 298283,
  [SMALL_STATE(6416)] = 298301,
  [SMALL_STATE(6417)] = 298319,
  [SMALL_STATE(6418)] = 298337,
  [SMALL_STATE(6419)] = 298355,
  [SMALL_STATE(6420)] = 298373,
  [SMALL_STATE(6421)] = 298391,
  [SMALL_STATE(6422)] = 298409,
  [SMALL_STATE(6423)] = 298427,
  [SMALL_STATE(6424)] = 298445,
  [SMALL_STATE(6425)] = 298463,
  [SMALL_STATE(6426)] = 298481,
  [SMALL_STATE(6427)] = 298499,
  [SMALL_STATE(6428)] = 298517,
  [SMALL_STATE(6429)] = 298535,
  [SMALL_STATE(6430)] = 298553,
  [SMALL_STATE(6431)] = 298571,
  [SMALL_STATE(6432)] = 298589,
  [SMALL_STATE(6433)] = 298607,
  [SMALL_STATE(6434)] = 298625,
  [SMALL_STATE(6435)] = 298643,
  [SMALL_STATE(6436)] = 298661,
  [SMALL_STATE(6437)] = 298679,
  [SMALL_STATE(6438)] = 298697,
  [SMALL_STATE(6439)] = 298715,
  [SMALL_STATE(6440)] = 298733,
  [SMALL_STATE(6441)] = 298751,
  [SMALL_STATE(6442)] = 298769,
  [SMALL_STATE(6443)] = 298787,
  [SMALL_STATE(6444)] = 298805,
  [SMALL_STATE(6445)] = 298823,
  [SMALL_STATE(6446)] = 298841,
  [SMALL_STATE(6447)] = 298859,
  [SMALL_STATE(6448)] = 298877,
  [SMALL_STATE(6449)] = 298895,
  [SMALL_STATE(6450)] = 298913,
  [SMALL_STATE(6451)] = 298931,
  [SMALL_STATE(6452)] = 298949,
  [SMALL_STATE(6453)] = 298967,
  [SMALL_STATE(6454)] = 298985,
  [SMALL_STATE(6455)] = 299003,
  [SMALL_STATE(6456)] = 299021,
  [SMALL_STATE(6457)] = 299039,
  [SMALL_STATE(6458)] = 299057,
  [SMALL_STATE(6459)] = 299075,
  [SMALL_STATE(6460)] = 299093,
  [SMALL_STATE(6461)] = 299111,
  [SMALL_STATE(6462)] = 299129,
  [SMALL_STATE(6463)] = 299147,
  [SMALL_STATE(6464)] = 299165,
  [SMALL_STATE(6465)] = 299183,
  [SMALL_STATE(6466)] = 299201,
  [SMALL_STATE(6467)] = 299219,
  [SMALL_STATE(6468)] = 299237,
  [SMALL_STATE(6469)] = 299255,
  [SMALL_STATE(6470)] = 299273,
  [SMALL_STATE(6471)] = 299291,
  [SMALL_STATE(6472)] = 299309,
  [SMALL_STATE(6473)] = 299327,
  [SMALL_STATE(6474)] = 299345,
  [SMALL_STATE(6475)] = 299363,
  [SMALL_STATE(6476)] = 299381,
  [SMALL_STATE(6477)] = 299399,
  [SMALL_STATE(6478)] = 299417,
  [SMALL_STATE(6479)] = 299435,
  [SMALL_STATE(6480)] = 299453,
  [SMALL_STATE(6481)] = 299471,
  [SMALL_STATE(6482)] = 299489,
  [SMALL_STATE(6483)] = 299507,
  [SMALL_STATE(6484)] = 299525,
  [SMALL_STATE(6485)] = 299543,
  [SMALL_STATE(6486)] = 299561,
  [SMALL_STATE(6487)] = 299579,
  [SMALL_STATE(6488)] = 299597,
  [SMALL_STATE(6489)] = 299615,
  [SMALL_STATE(6490)] = 299633,
  [SMALL_STATE(6491)] = 299651,
  [SMALL_STATE(6492)] = 299669,
  [SMALL_STATE(6493)] = 299687,
  [SMALL_STATE(6494)] = 299705,
  [SMALL_STATE(6495)] = 299723,
  [SMALL_STATE(6496)] = 299741,
  [SMALL_STATE(6497)] = 299759,
  [SMALL_STATE(6498)] = 299777,
  [SMALL_STATE(6499)] = 299795,
  [SMALL_STATE(6500)] = 299813,
  [SMALL_STATE(6501)] = 299831,
  [SMALL_STATE(6502)] = 299849,
  [SMALL_STATE(6503)] = 299867,
  [SMALL_STATE(6504)] = 299885,
  [SMALL_STATE(6505)] = 299903,
  [SMALL_STATE(6506)] = 299921,
  [SMALL_STATE(6507)] = 299939,
  [SMALL_STATE(6508)] = 299957,
  [SMALL_STATE(6509)] = 299975,
  [SMALL_STATE(6510)] = 299993,
  [SMALL_STATE(6511)] = 300011,
  [SMALL_STATE(6512)] = 300029,
  [SMALL_STATE(6513)] = 300047,
  [SMALL_STATE(6514)] = 300065,
  [SMALL_STATE(6515)] = 300083,
  [SMALL_STATE(6516)] = 300101,
  [SMALL_STATE(6517)] = 300119,
  [SMALL_STATE(6518)] = 300137,
  [SMALL_STATE(6519)] = 300155,
  [SMALL_STATE(6520)] = 300173,
  [SMALL_STATE(6521)] = 300191,
  [SMALL_STATE(6522)] = 300209,
  [SMALL_STATE(6523)] = 300227,
  [SMALL_STATE(6524)] = 300245,
  [SMALL_STATE(6525)] = 300263,
  [SMALL_STATE(6526)] = 300281,
  [SMALL_STATE(6527)] = 300299,
  [SMALL_STATE(6528)] = 300317,
  [SMALL_STATE(6529)] = 300335,
  [SMALL_STATE(6530)] = 300353,
  [SMALL_STATE(6531)] = 300371,
  [SMALL_STATE(6532)] = 300389,
  [SMALL_STATE(6533)] = 300407,
  [SMALL_STATE(6534)] = 300425,
  [SMALL_STATE(6535)] = 300443,
  [SMALL_STATE(6536)] = 300461,
  [SMALL_STATE(6537)] = 300479,
  [SMALL_STATE(6538)] = 300497,
  [SMALL_STATE(6539)] = 300515,
  [SMALL_STATE(6540)] = 300529,
  [SMALL_STATE(6541)] = 300545,
  [SMALL_STATE(6542)] = 300563,
  [SMALL_STATE(6543)] = 300581,
  [SMALL_STATE(6544)] = 300599,
  [SMALL_STATE(6545)] = 300617,
  [SMALL_STATE(6546)] = 300635,
  [SMALL_STATE(6547)] = 300653,
  [SMALL_STATE(6548)] = 300671,
  [SMALL_STATE(6549)] = 300689,
  [SMALL_STATE(6550)] = 300707,
  [SMALL_STATE(6551)] = 300725,
  [SMALL_STATE(6552)] = 300743,
  [SMALL_STATE(6553)] = 300761,
  [SMALL_STATE(6554)] = 300779,
  [SMALL_STATE(6555)] = 300797,
  [SMALL_STATE(6556)] = 300815,
  [SMALL_STATE(6557)] = 300833,
  [SMALL_STATE(6558)] = 300851,
  [SMALL_STATE(6559)] = 300869,
  [SMALL_STATE(6560)] = 300887,
  [SMALL_STATE(6561)] = 300905,
  [SMALL_STATE(6562)] = 300923,
  [SMALL_STATE(6563)] = 300941,
  [SMALL_STATE(6564)] = 300955,
  [SMALL_STATE(6565)] = 300973,
  [SMALL_STATE(6566)] = 300991,
  [SMALL_STATE(6567)] = 301009,
  [SMALL_STATE(6568)] = 301027,
  [SMALL_STATE(6569)] = 301045,
  [SMALL_STATE(6570)] = 301063,
  [SMALL_STATE(6571)] = 301081,
  [SMALL_STATE(6572)] = 301099,
  [SMALL_STATE(6573)] = 301117,
  [SMALL_STATE(6574)] = 301133,
  [SMALL_STATE(6575)] = 301151,
  [SMALL_STATE(6576)] = 301169,
  [SMALL_STATE(6577)] = 301187,
  [SMALL_STATE(6578)] = 301205,
  [SMALL_STATE(6579)] = 301223,
  [SMALL_STATE(6580)] = 301241,
  [SMALL_STATE(6581)] = 301259,
  [SMALL_STATE(6582)] = 301277,
  [SMALL_STATE(6583)] = 301295,
  [SMALL_STATE(6584)] = 301313,
  [SMALL_STATE(6585)] = 301331,
  [SMALL_STATE(6586)] = 301349,
  [SMALL_STATE(6587)] = 301367,
  [SMALL_STATE(6588)] = 301385,
  [SMALL_STATE(6589)] = 301403,
  [SMALL_STATE(6590)] = 301421,
  [SMALL_STATE(6591)] = 301435,
  [SMALL_STATE(6592)] = 301453,
  [SMALL_STATE(6593)] = 301471,
  [SMALL_STATE(6594)] = 301489,
  [SMALL_STATE(6595)] = 301507,
  [SMALL_STATE(6596)] = 301525,
  [SMALL_STATE(6597)] = 301543,
  [SMALL_STATE(6598)] = 301561,
  [SMALL_STATE(6599)] = 301579,
  [SMALL_STATE(6600)] = 301597,
  [SMALL_STATE(6601)] = 301615,
  [SMALL_STATE(6602)] = 301633,
  [SMALL_STATE(6603)] = 301651,
  [SMALL_STATE(6604)] = 301669,
  [SMALL_STATE(6605)] = 301687,
  [SMALL_STATE(6606)] = 301705,
  [SMALL_STATE(6607)] = 301723,
  [SMALL_STATE(6608)] = 301741,
  [SMALL_STATE(6609)] = 301759,
  [SMALL_STATE(6610)] = 301777,
  [SMALL_STATE(6611)] = 301795,
  [SMALL_STATE(6612)] = 301813,
  [SMALL_STATE(6613)] = 301831,
  [SMALL_STATE(6614)] = 301849,
  [SMALL_STATE(6615)] = 301867,
  [SMALL_STATE(6616)] = 301885,
  [SMALL_STATE(6617)] = 301903,
  [SMALL_STATE(6618)] = 301921,
  [SMALL_STATE(6619)] = 301939,
  [SMALL_STATE(6620)] = 301957,
  [SMALL_STATE(6621)] = 301975,
  [SMALL_STATE(6622)] = 301993,
  [SMALL_STATE(6623)] = 302011,
  [SMALL_STATE(6624)] = 302029,
  [SMALL_STATE(6625)] = 302047,
  [SMALL_STATE(6626)] = 302065,
  [SMALL_STATE(6627)] = 302083,
  [SMALL_STATE(6628)] = 302101,
  [SMALL_STATE(6629)] = 302119,
  [SMALL_STATE(6630)] = 302137,
  [SMALL_STATE(6631)] = 302155,
  [SMALL_STATE(6632)] = 302173,
  [SMALL_STATE(6633)] = 302191,
  [SMALL_STATE(6634)] = 302209,
  [SMALL_STATE(6635)] = 302227,
  [SMALL_STATE(6636)] = 302245,
  [SMALL_STATE(6637)] = 302263,
  [SMALL_STATE(6638)] = 302281,
  [SMALL_STATE(6639)] = 302299,
  [SMALL_STATE(6640)] = 302317,
  [SMALL_STATE(6641)] = 302335,
  [SMALL_STATE(6642)] = 302353,
  [SMALL_STATE(6643)] = 302371,
  [SMALL_STATE(6644)] = 302389,
  [SMALL_STATE(6645)] = 302407,
  [SMALL_STATE(6646)] = 302423,
  [SMALL_STATE(6647)] = 302441,
  [SMALL_STATE(6648)] = 302459,
  [SMALL_STATE(6649)] = 302477,
  [SMALL_STATE(6650)] = 302495,
  [SMALL_STATE(6651)] = 302513,
  [SMALL_STATE(6652)] = 302531,
  [SMALL_STATE(6653)] = 302549,
  [SMALL_STATE(6654)] = 302567,
  [SMALL_STATE(6655)] = 302585,
  [SMALL_STATE(6656)] = 302603,
  [SMALL_STATE(6657)] = 302621,
  [SMALL_STATE(6658)] = 302639,
  [SMALL_STATE(6659)] = 302657,
  [SMALL_STATE(6660)] = 302671,
  [SMALL_STATE(6661)] = 302685,
  [SMALL_STATE(6662)] = 302703,
  [SMALL_STATE(6663)] = 302717,
  [SMALL_STATE(6664)] = 302735,
  [SMALL_STATE(6665)] = 302753,
  [SMALL_STATE(6666)] = 302771,
  [SMALL_STATE(6667)] = 302789,
  [SMALL_STATE(6668)] = 302807,
  [SMALL_STATE(6669)] = 302825,
  [SMALL_STATE(6670)] = 302843,
  [SMALL_STATE(6671)] = 302861,
  [SMALL_STATE(6672)] = 302879,
  [SMALL_STATE(6673)] = 302897,
  [SMALL_STATE(6674)] = 302915,
  [SMALL_STATE(6675)] = 302933,
  [SMALL_STATE(6676)] = 302951,
  [SMALL_STATE(6677)] = 302969,
  [SMALL_STATE(6678)] = 302987,
  [SMALL_STATE(6679)] = 303005,
  [SMALL_STATE(6680)] = 303023,
  [SMALL_STATE(6681)] = 303041,
  [SMALL_STATE(6682)] = 303059,
  [SMALL_STATE(6683)] = 303077,
  [SMALL_STATE(6684)] = 303095,
  [SMALL_STATE(6685)] = 303113,
  [SMALL_STATE(6686)] = 303131,
  [SMALL_STATE(6687)] = 303149,
  [SMALL_STATE(6688)] = 303167,
  [SMALL_STATE(6689)] = 303185,
  [SMALL_STATE(6690)] = 303203,
  [SMALL_STATE(6691)] = 303221,
  [SMALL_STATE(6692)] = 303239,
  [SMALL_STATE(6693)] = 303257,
  [SMALL_STATE(6694)] = 303275,
  [SMALL_STATE(6695)] = 303293,
  [SMALL_STATE(6696)] = 303311,
  [SMALL_STATE(6697)] = 303329,
  [SMALL_STATE(6698)] = 303347,
  [SMALL_STATE(6699)] = 303365,
  [SMALL_STATE(6700)] = 303383,
  [SMALL_STATE(6701)] = 303401,
  [SMALL_STATE(6702)] = 303419,
  [SMALL_STATE(6703)] = 303437,
  [SMALL_STATE(6704)] = 303455,
  [SMALL_STATE(6705)] = 303473,
  [SMALL_STATE(6706)] = 303491,
  [SMALL_STATE(6707)] = 303509,
  [SMALL_STATE(6708)] = 303527,
  [SMALL_STATE(6709)] = 303545,
  [SMALL_STATE(6710)] = 303563,
  [SMALL_STATE(6711)] = 303579,
  [SMALL_STATE(6712)] = 303597,
  [SMALL_STATE(6713)] = 303615,
  [SMALL_STATE(6714)] = 303633,
  [SMALL_STATE(6715)] = 303651,
  [SMALL_STATE(6716)] = 303667,
  [SMALL_STATE(6717)] = 303685,
  [SMALL_STATE(6718)] = 303703,
  [SMALL_STATE(6719)] = 303721,
  [SMALL_STATE(6720)] = 303737,
  [SMALL_STATE(6721)] = 303755,
  [SMALL_STATE(6722)] = 303773,
  [SMALL_STATE(6723)] = 303789,
  [SMALL_STATE(6724)] = 303807,
  [SMALL_STATE(6725)] = 303825,
  [SMALL_STATE(6726)] = 303841,
  [SMALL_STATE(6727)] = 303859,
  [SMALL_STATE(6728)] = 303877,
  [SMALL_STATE(6729)] = 303895,
  [SMALL_STATE(6730)] = 303911,
  [SMALL_STATE(6731)] = 303929,
  [SMALL_STATE(6732)] = 303947,
  [SMALL_STATE(6733)] = 303965,
  [SMALL_STATE(6734)] = 303983,
  [SMALL_STATE(6735)] = 304001,
  [SMALL_STATE(6736)] = 304019,
  [SMALL_STATE(6737)] = 304035,
  [SMALL_STATE(6738)] = 304053,
  [SMALL_STATE(6739)] = 304071,
  [SMALL_STATE(6740)] = 304087,
  [SMALL_STATE(6741)] = 304105,
  [SMALL_STATE(6742)] = 304123,
  [SMALL_STATE(6743)] = 304141,
  [SMALL_STATE(6744)] = 304159,
  [SMALL_STATE(6745)] = 304177,
  [SMALL_STATE(6746)] = 304195,
  [SMALL_STATE(6747)] = 304213,
  [SMALL_STATE(6748)] = 304231,
  [SMALL_STATE(6749)] = 304249,
  [SMALL_STATE(6750)] = 304267,
  [SMALL_STATE(6751)] = 304285,
  [SMALL_STATE(6752)] = 304303,
  [SMALL_STATE(6753)] = 304321,
  [SMALL_STATE(6754)] = 304339,
  [SMALL_STATE(6755)] = 304357,
  [SMALL_STATE(6756)] = 304375,
  [SMALL_STATE(6757)] = 304393,
  [SMALL_STATE(6758)] = 304411,
  [SMALL_STATE(6759)] = 304429,
  [SMALL_STATE(6760)] = 304447,
  [SMALL_STATE(6761)] = 304465,
  [SMALL_STATE(6762)] = 304479,
  [SMALL_STATE(6763)] = 304497,
  [SMALL_STATE(6764)] = 304515,
  [SMALL_STATE(6765)] = 304529,
  [SMALL_STATE(6766)] = 304547,
  [SMALL_STATE(6767)] = 304565,
  [SMALL_STATE(6768)] = 304579,
  [SMALL_STATE(6769)] = 304597,
  [SMALL_STATE(6770)] = 304615,
  [SMALL_STATE(6771)] = 304633,
  [SMALL_STATE(6772)] = 304648,
  [SMALL_STATE(6773)] = 304661,
  [SMALL_STATE(6774)] = 304676,
  [SMALL_STATE(6775)] = 304691,
  [SMALL_STATE(6776)] = 304704,
  [SMALL_STATE(6777)] = 304717,
  [SMALL_STATE(6778)] = 304732,
  [SMALL_STATE(6779)] = 304745,
  [SMALL_STATE(6780)] = 304760,
  [SMALL_STATE(6781)] = 304775,
  [SMALL_STATE(6782)] = 304788,
  [SMALL_STATE(6783)] = 304801,
  [SMALL_STATE(6784)] = 304816,
  [SMALL_STATE(6785)] = 304831,
  [SMALL_STATE(6786)] = 304846,
  [SMALL_STATE(6787)] = 304861,
  [SMALL_STATE(6788)] = 304876,
  [SMALL_STATE(6789)] = 304889,
  [SMALL_STATE(6790)] = 304904,
  [SMALL_STATE(6791)] = 304919,
  [SMALL_STATE(6792)] = 304934,
  [SMALL_STATE(6793)] = 304949,
  [SMALL_STATE(6794)] = 304964,
  [SMALL_STATE(6795)] = 304979,
  [SMALL_STATE(6796)] = 304994,
  [SMALL_STATE(6797)] = 305009,
  [SMALL_STATE(6798)] = 305024,
  [SMALL_STATE(6799)] = 305037,
  [SMALL_STATE(6800)] = 305050,
  [SMALL_STATE(6801)] = 305063,
  [SMALL_STATE(6802)] = 305078,
  [SMALL_STATE(6803)] = 305093,
  [SMALL_STATE(6804)] = 305108,
  [SMALL_STATE(6805)] = 305123,
  [SMALL_STATE(6806)] = 305138,
  [SMALL_STATE(6807)] = 305153,
  [SMALL_STATE(6808)] = 305166,
  [SMALL_STATE(6809)] = 305181,
  [SMALL_STATE(6810)] = 305196,
  [SMALL_STATE(6811)] = 305211,
  [SMALL_STATE(6812)] = 305226,
  [SMALL_STATE(6813)] = 305241,
  [SMALL_STATE(6814)] = 305256,
  [SMALL_STATE(6815)] = 305271,
  [SMALL_STATE(6816)] = 305286,
  [SMALL_STATE(6817)] = 305301,
  [SMALL_STATE(6818)] = 305316,
  [SMALL_STATE(6819)] = 305331,
  [SMALL_STATE(6820)] = 305346,
  [SMALL_STATE(6821)] = 305359,
  [SMALL_STATE(6822)] = 305374,
  [SMALL_STATE(6823)] = 305389,
  [SMALL_STATE(6824)] = 305404,
  [SMALL_STATE(6825)] = 305419,
  [SMALL_STATE(6826)] = 305432,
  [SMALL_STATE(6827)] = 305447,
  [SMALL_STATE(6828)] = 305462,
  [SMALL_STATE(6829)] = 305475,
  [SMALL_STATE(6830)] = 305490,
  [SMALL_STATE(6831)] = 305505,
  [SMALL_STATE(6832)] = 305520,
  [SMALL_STATE(6833)] = 305535,
  [SMALL_STATE(6834)] = 305550,
  [SMALL_STATE(6835)] = 305565,
  [SMALL_STATE(6836)] = 305578,
  [SMALL_STATE(6837)] = 305591,
  [SMALL_STATE(6838)] = 305606,
  [SMALL_STATE(6839)] = 305621,
  [SMALL_STATE(6840)] = 305636,
  [SMALL_STATE(6841)] = 305651,
  [SMALL_STATE(6842)] = 305666,
  [SMALL_STATE(6843)] = 305681,
  [SMALL_STATE(6844)] = 305696,
  [SMALL_STATE(6845)] = 305711,
  [SMALL_STATE(6846)] = 305726,
  [SMALL_STATE(6847)] = 305741,
  [SMALL_STATE(6848)] = 305756,
  [SMALL_STATE(6849)] = 305771,
  [SMALL_STATE(6850)] = 305784,
  [SMALL_STATE(6851)] = 305799,
  [SMALL_STATE(6852)] = 305814,
  [SMALL_STATE(6853)] = 305829,
  [SMALL_STATE(6854)] = 305844,
  [SMALL_STATE(6855)] = 305859,
  [SMALL_STATE(6856)] = 305874,
  [SMALL_STATE(6857)] = 305889,
  [SMALL_STATE(6858)] = 305904,
  [SMALL_STATE(6859)] = 305919,
  [SMALL_STATE(6860)] = 305934,
  [SMALL_STATE(6861)] = 305949,
  [SMALL_STATE(6862)] = 305964,
  [SMALL_STATE(6863)] = 305979,
  [SMALL_STATE(6864)] = 305994,
  [SMALL_STATE(6865)] = 306009,
  [SMALL_STATE(6866)] = 306024,
  [SMALL_STATE(6867)] = 306039,
  [SMALL_STATE(6868)] = 306054,
  [SMALL_STATE(6869)] = 306069,
  [SMALL_STATE(6870)] = 306081,
  [SMALL_STATE(6871)] = 306093,
  [SMALL_STATE(6872)] = 306105,
  [SMALL_STATE(6873)] = 306117,
  [SMALL_STATE(6874)] = 306129,
  [SMALL_STATE(6875)] = 306141,
  [SMALL_STATE(6876)] = 306153,
  [SMALL_STATE(6877)] = 306165,
  [SMALL_STATE(6878)] = 306177,
  [SMALL_STATE(6879)] = 306189,
  [SMALL_STATE(6880)] = 306201,
  [SMALL_STATE(6881)] = 306213,
  [SMALL_STATE(6882)] = 306225,
  [SMALL_STATE(6883)] = 306237,
  [SMALL_STATE(6884)] = 306249,
  [SMALL_STATE(6885)] = 306261,
  [SMALL_STATE(6886)] = 306273,
  [SMALL_STATE(6887)] = 306285,
  [SMALL_STATE(6888)] = 306297,
  [SMALL_STATE(6889)] = 306309,
  [SMALL_STATE(6890)] = 306321,
  [SMALL_STATE(6891)] = 306333,
  [SMALL_STATE(6892)] = 306345,
  [SMALL_STATE(6893)] = 306357,
  [SMALL_STATE(6894)] = 306369,
  [SMALL_STATE(6895)] = 306381,
  [SMALL_STATE(6896)] = 306393,
  [SMALL_STATE(6897)] = 306405,
  [SMALL_STATE(6898)] = 306417,
  [SMALL_STATE(6899)] = 306429,
  [SMALL_STATE(6900)] = 306441,
  [SMALL_STATE(6901)] = 306453,
  [SMALL_STATE(6902)] = 306465,
  [SMALL_STATE(6903)] = 306477,
  [SMALL_STATE(6904)] = 306489,
  [SMALL_STATE(6905)] = 306501,
  [SMALL_STATE(6906)] = 306513,
  [SMALL_STATE(6907)] = 306525,
  [SMALL_STATE(6908)] = 306537,
  [SMALL_STATE(6909)] = 306549,
  [SMALL_STATE(6910)] = 306561,
  [SMALL_STATE(6911)] = 306573,
  [SMALL_STATE(6912)] = 306585,
  [SMALL_STATE(6913)] = 306597,
  [SMALL_STATE(6914)] = 306609,
  [SMALL_STATE(6915)] = 306621,
  [SMALL_STATE(6916)] = 306633,
  [SMALL_STATE(6917)] = 306645,
  [SMALL_STATE(6918)] = 306657,
  [SMALL_STATE(6919)] = 306669,
  [SMALL_STATE(6920)] = 306681,
  [SMALL_STATE(6921)] = 306693,
  [SMALL_STATE(6922)] = 306705,
  [SMALL_STATE(6923)] = 306717,
  [SMALL_STATE(6924)] = 306729,
  [SMALL_STATE(6925)] = 306741,
  [SMALL_STATE(6926)] = 306753,
  [SMALL_STATE(6927)] = 306765,
  [SMALL_STATE(6928)] = 306777,
  [SMALL_STATE(6929)] = 306789,
  [SMALL_STATE(6930)] = 306801,
  [SMALL_STATE(6931)] = 306813,
  [SMALL_STATE(6932)] = 306825,
  [SMALL_STATE(6933)] = 306837,
  [SMALL_STATE(6934)] = 306849,
  [SMALL_STATE(6935)] = 306861,
  [SMALL_STATE(6936)] = 306873,
  [SMALL_STATE(6937)] = 306885,
  [SMALL_STATE(6938)] = 306897,
  [SMALL_STATE(6939)] = 306909,
  [SMALL_STATE(6940)] = 306921,
  [SMALL_STATE(6941)] = 306933,
  [SMALL_STATE(6942)] = 306945,
  [SMALL_STATE(6943)] = 306957,
  [SMALL_STATE(6944)] = 306969,
  [SMALL_STATE(6945)] = 306981,
  [SMALL_STATE(6946)] = 306993,
  [SMALL_STATE(6947)] = 307005,
  [SMALL_STATE(6948)] = 307017,
  [SMALL_STATE(6949)] = 307029,
  [SMALL_STATE(6950)] = 307041,
  [SMALL_STATE(6951)] = 307053,
  [SMALL_STATE(6952)] = 307065,
  [SMALL_STATE(6953)] = 307077,
  [SMALL_STATE(6954)] = 307089,
  [SMALL_STATE(6955)] = 307101,
  [SMALL_STATE(6956)] = 307113,
  [SMALL_STATE(6957)] = 307125,
  [SMALL_STATE(6958)] = 307137,
  [SMALL_STATE(6959)] = 307149,
  [SMALL_STATE(6960)] = 307161,
  [SMALL_STATE(6961)] = 307173,
  [SMALL_STATE(6962)] = 307185,
  [SMALL_STATE(6963)] = 307197,
  [SMALL_STATE(6964)] = 307209,
  [SMALL_STATE(6965)] = 307221,
  [SMALL_STATE(6966)] = 307233,
  [SMALL_STATE(6967)] = 307245,
  [SMALL_STATE(6968)] = 307257,
  [SMALL_STATE(6969)] = 307269,
  [SMALL_STATE(6970)] = 307281,
  [SMALL_STATE(6971)] = 307293,
  [SMALL_STATE(6972)] = 307305,
  [SMALL_STATE(6973)] = 307317,
  [SMALL_STATE(6974)] = 307329,
  [SMALL_STATE(6975)] = 307341,
  [SMALL_STATE(6976)] = 307353,
  [SMALL_STATE(6977)] = 307365,
  [SMALL_STATE(6978)] = 307377,
  [SMALL_STATE(6979)] = 307389,
  [SMALL_STATE(6980)] = 307401,
  [SMALL_STATE(6981)] = 307413,
  [SMALL_STATE(6982)] = 307425,
  [SMALL_STATE(6983)] = 307437,
  [SMALL_STATE(6984)] = 307449,
  [SMALL_STATE(6985)] = 307461,
  [SMALL_STATE(6986)] = 307473,
  [SMALL_STATE(6987)] = 307485,
  [SMALL_STATE(6988)] = 307497,
  [SMALL_STATE(6989)] = 307509,
  [SMALL_STATE(6990)] = 307521,
  [SMALL_STATE(6991)] = 307533,
  [SMALL_STATE(6992)] = 307545,
  [SMALL_STATE(6993)] = 307557,
  [SMALL_STATE(6994)] = 307569,
  [SMALL_STATE(6995)] = 307581,
  [SMALL_STATE(6996)] = 307593,
  [SMALL_STATE(6997)] = 307605,
  [SMALL_STATE(6998)] = 307617,
  [SMALL_STATE(6999)] = 307629,
  [SMALL_STATE(7000)] = 307641,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [7] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source, 0, 0, 0),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(362),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(107),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1013),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3215),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4376),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4338),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4921),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5072),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5078),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5081),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(226),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6973),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6802),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(196),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(397),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4096),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(412),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(588),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(489),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(106),
  [53] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4815),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6973),
  [57] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6133),
  [59] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1018),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(25),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(82),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(927),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1080),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1591),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1592),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4865),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4866),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4867),
  [79] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4874),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(221),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(233),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6854),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(849),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(239),
  [91] = {.entry = {.count = 1, .reusable = false}}, SHIFT(445),
  [93] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1612),
  [95] = {.entry = {.count = 1, .reusable = false}}, SHIFT(417),
  [97] = {.entry = {.count = 1, .reusable = false}}, SHIFT(656),
  [99] = {.entry = {.count = 1, .reusable = false}}, SHIFT(657),
  [101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(176),
  [103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(34),
  [105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(37),
  [107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(40),
  [109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2803),
  [111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(92),
  [113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(39),
  [115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4802),
  [117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5987),
  [119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(24),
  [121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1069),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1425),
  [125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(22),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1065),
  [129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4044),
  [131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(20),
  [133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1081),
  [135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1340),
  [137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(29),
  [139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1071),
  [141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4290),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(31),
  [145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1072),
  [147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2114),
  [149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(27),
  [151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1075),
  [153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3598),
  [155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(30),
  [157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1085),
  [159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1584),
  [161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(26),
  [163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1076),
  [165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2497),
  [167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(28),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1070),
  [171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1102),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(32),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1073),
  [177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3287),
  [179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(33),
  [181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1078),
  [183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2634),
  [185] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_call_without_parentheses, 1, 0, 4),
  [187] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_call_without_parentheses, 1, 0, 4),
  [189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(108),
  [191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(914),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1423),
  [195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1385),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1387),
  [199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4945),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4946),
  [203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4933),
  [205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4934),
  [207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(244),
  [209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(245),
  [211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6859),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(895),
  [215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(246),
  [217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(388),
  [219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(441),
  [221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(761),
  [223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(762),
  [225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(109),
  [227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4822),
  [231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5962),
  [233] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [235] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(85),
  [239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1250),
  [241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1234),
  [243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1235),
  [245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4857),
  [247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4858),
  [249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4859),
  [251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4864),
  [253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(186),
  [255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(187),
  [257] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(6973),
  [260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6858),
  [262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(891),
  [264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(188),
  [266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(451),
  [268] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(1237),
  [271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(415),
  [273] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(740),
  [276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(740),
  [278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(741),
  [280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(90),
  [282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4821),
  [286] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(6973),
  [289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5986),
  [291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(9),
  [293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6780),
  [295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3),
  [297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6827),
  [299] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(1388),
  [302] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(761),
  [305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1082),
  [307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1190),
  [309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1063),
  [311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3531),
  [313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1064),
  [315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1445),
  [317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1084),
  [319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2843),
  [321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1083),
  [323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2565),
  [325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1077),
  [327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3682),
  [329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1067),
  [331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1115),
  [333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1066),
  [335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4380),
  [337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1086),
  [339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1615),
  [341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1068),
  [343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1968),
  [345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1074),
  [347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3311),
  [349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1079),
  [351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2638),
  [353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(67),
  [355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1126),
  [357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_after_block, 1, 0, 0),
  [359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(94),
  [361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(931),
  [363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2244),
  [365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2347),
  [367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2348),
  [369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4910),
  [371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4911),
  [373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4912),
  [375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4913),
  [377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(202),
  [379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(203),
  [381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6847),
  [383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(848),
  [385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(204),
  [387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(452),
  [389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(394),
  [391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(567),
  [393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(568),
  [395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4818),
  [401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6044),
  [403] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(2349),
  [406] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(567),
  [409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(68),
  [411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1094),
  [413] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_block, 1, 0, 0),
  [415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2),
  [417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6789),
  [419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(70),
  [421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1106),
  [423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rescue_block, 1, 0, 0),
  [425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(69),
  [427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1101),
  [429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_block, 1, 0, 0),
  [431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(78),
  [433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(988),
  [435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2483),
  [437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3378),
  [439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3379),
  [441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4977),
  [443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4978),
  [445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4991),
  [447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5074),
  [449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(183),
  [451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(189),
  [453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6842),
  [455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(843),
  [457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(197),
  [459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(449),
  [461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(418),
  [463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(504),
  [465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(505),
  [467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(81),
  [469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4836),
  [473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6041),
  [475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(954),
  [477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2354),
  [479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6846),
  [481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(845),
  [483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(439),
  [485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(547),
  [487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(548),
  [489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4838),
  [491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2701),
  [493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6855),
  [495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(880),
  [497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(363),
  [499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(676),
  [501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(677),
  [503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4797),
  [505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(547),
  [508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(968),
  [510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2527),
  [512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6856),
  [514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(887),
  [516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(379),
  [518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(697),
  [520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(698),
  [522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4813),
  [524] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(676),
  [527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4),
  [529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6844),
  [531] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(697),
  [534] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(3380),
  [537] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(504),
  [540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(99),
  [542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1001),
  [544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3200),
  [546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3650),
  [548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3651),
  [550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4860),
  [552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4861),
  [554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4862),
  [556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4863),
  [558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(216),
  [560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(217),
  [562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6848),
  [564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(855),
  [566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(218),
  [568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(453),
  [570] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(3652),
  [573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(413),
  [575] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(589),
  [578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(589),
  [580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(590),
  [582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(100),
  [584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4814),
  [588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6103),
  [590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(104),
  [592] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2665),
  [594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3349),
  [596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3350),
  [598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4887),
  [600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4888),
  [602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4889),
  [604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4890),
  [606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(230),
  [608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(231),
  [610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6852),
  [612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(868),
  [614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(232),
  [616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(454),
  [618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(382),
  [620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(632),
  [622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(633),
  [624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8),
  [626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(105),
  [628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6830),
  [632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4798),
  [634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6145),
  [636] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(3351),
  [639] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(632),
  [642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6),
  [644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6795),
  [646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3460),
  [648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(885),
  [650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(395),
  [652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(821),
  [654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(822),
  [656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4810),
  [658] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(202),
  [661] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(821),
  [664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1129),
  [666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_after_block, 2, 0, 0),
  [668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1120),
  [670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_block, 2, 0, 0),
  [672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1099),
  [674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_block, 2, 0, 0),
  [676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1116),
  [678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rescue_block, 2, 0, 0),
  [680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [682] = {.entry = {.count = 1, .reusable = false}}, SHIFT(117),
  [684] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2173),
  [686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1007),
  [688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2079),
  [690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6812),
  [692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(385),
  [694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(635),
  [696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(636),
  [698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(327),
  [700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4837),
  [702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(115),
  [704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3198),
  [706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2035),
  [708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(128),
  [710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1623),
  [712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1994),
  [714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(129),
  [718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(101),
  [720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(987),
  [722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4297),
  [724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4097),
  [726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4098),
  [728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4963),
  [730] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4964),
  [732] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4965),
  [734] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4972),
  [736] = {.entry = {.count = 1, .reusable = false}}, SHIFT(247),
  [738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(248),
  [740] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6839),
  [742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(250),
  [744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(448),
  [746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4099),
  [748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(411),
  [750] = {.entry = {.count = 1, .reusable = false}}, SHIFT(482),
  [752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(483),
  [754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(303),
  [756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4149),
  [758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(76),
  [760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4799),
  [762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6147),
  [764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(127),
  [766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4391),
  [768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2006),
  [770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1971),
  [772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(124),
  [774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1926),
  [776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2050),
  [778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2023),
  [780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(116),
  [782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3485),
  [784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1651),
  [786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1785),
  [788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(120),
  [790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3072),
  [792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1905),
  [794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(133),
  [796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1186),
  [798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2067),
  [800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2047),
  [802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(126),
  [804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1912),
  [806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1913),
  [808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(136),
  [810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4449),
  [812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1980),
  [814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(134),
  [816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1311),
  [818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(123),
  [820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4152),
  [822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2080),
  [824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(119),
  [826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1589),
  [828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(122),
  [830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3665),
  [832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2085),
  [834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2077),
  [836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(118),
  [838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2186),
  [840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1955),
  [842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(114),
  [844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4557),
  [846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(130),
  [848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3364),
  [850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2101),
  [852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2082),
  [854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(125),
  [856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3681),
  [858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4519),
  [860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1640),
  [862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(121),
  [864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3017),
  [866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(135),
  [868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1444),
  [870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2111),
  [872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2097),
  [874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(113),
  [876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3376),
  [878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(132),
  [880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4427),
  [882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2018),
  [884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2105),
  [886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(131),
  [888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1467),
  [890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1950),
  [892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2059),
  [894] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct, 1, 0, 0),
  [896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3395),
  [898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4525),
  [900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2936),
  [902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2032),
  [904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2957),
  [906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2263),
  [908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2078),
  [910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2311),
  [912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1682),
  [914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3060),
  [916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1954),
  [918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3075),
  [920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3706),
  [922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2084),
  [924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4385),
  [926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2017),
  [928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2024),
  [930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2049),
  [932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3718),
  [934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2034),
  [936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4188),
  [938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2003),
  [940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1661),
  [942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1911),
  [944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4230),
  [946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3388),
  [948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2100),
  [950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1537),
  [952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4115),
  [954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1192),
  [956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2064),
  [958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1200),
  [960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1525),
  [962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2110),
  [964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4436),
  [966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1977),
  [968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(80),
  [970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2487),
  [972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1967),
  [974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1969),
  [976] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4908),
  [978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4909),
  [980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4914),
  [982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4915),
  [984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(240),
  [986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(241),
  [988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6857),
  [990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(242),
  [992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(450),
  [994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2051),
  [996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(399),
  [998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(719),
  [1000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(720),
  [1002] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3346),
  [1004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(86),
  [1006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4817),
  [1008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6141),
  [1010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3856),
  [1012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3879),
  [1014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3880),
  [1016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1454),
  [1018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1459),
  [1020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1460),
  [1022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2608),
  [1024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2636),
  [1026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2648),
  [1028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1619),
  [1030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1620),
  [1032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1621),
  [1034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1110),
  [1036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1111),
  [1038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1289),
  [1040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2707),
  [1042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2692),
  [1044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2693),
  [1046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1991),
  [1048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2011),
  [1050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2012),
  [1052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3329),
  [1054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3347),
  [1056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4425),
  [1058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4092),
  [1060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4093),
  [1062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2642),
  [1064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2643),
  [1066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2644),
  [1068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3520),
  [1070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3789),
  [1072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3884),
  [1074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1276),
  [1076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1285),
  [1078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1113),
  [1080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [1082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(260),
  [1084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1683),
  [1086] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_operator_identifier, 1, 0, 0), SHIFT(6973),
  [1089] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_stab_clause, 1, 0, 6),
  [1091] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_stab_clause, 2, 0, 11),
  [1093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(83),
  [1095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3862),
  [1097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2928),
  [1099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2930),
  [1101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4976),
  [1103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4845),
  [1105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4846),
  [1107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4847),
  [1109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(243),
  [1111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2310),
  [1113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(251),
  [1115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6832),
  [1117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(792),
  [1119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(255),
  [1121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(446),
  [1123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2931),
  [1125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(365),
  [1127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(852),
  [1129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(853),
  [1131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(102),
  [1133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4806),
  [1135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6014),
  [1137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1836),
  [1139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2714),
  [1141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4291),
  [1143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1117),
  [1145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3521),
  [1147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3477),
  [1149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3074),
  [1151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2755),
  [1153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1299),
  [1155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1305),
  [1157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(110),
  [1159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1017),
  [1161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4360),
  [1163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4488),
  [1165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4489),
  [1167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4950),
  [1169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4951),
  [1171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4958),
  [1173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5075),
  [1175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(208),
  [1177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(211),
  [1179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6837),
  [1181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(841),
  [1183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(215),
  [1185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1307),
  [1187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(447),
  [1189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4490),
  [1191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(400),
  [1193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(458),
  [1195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(459),
  [1197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(97),
  [1199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4823),
  [1201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6176),
  [1203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3478),
  [1205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2093),
  [1207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1199),
  [1209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1127),
  [1211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3144),
  [1213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4255),
  [1215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1281),
  [1217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4359),
  [1219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3479),
  [1221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1740),
  [1223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3830),
  [1225] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__items_with_trailing_separator, 3, 0, 0),
  [1227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2956),
  [1229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2052),
  [1231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2179),
  [1233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2180),
  [1235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2181),
  [1237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4374),
  [1239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4276),
  [1241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1302),
  [1243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4547),
  [1245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2670),
  [1247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2273),
  [1249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4548),
  [1251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2501),
  [1253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3041),
  [1255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4431),
  [1257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4549),
  [1259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3673),
  [1261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3674),
  [1263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3675),
  [1265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4229),
  [1267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3717),
  [1269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1578),
  [1271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3742),
  [1273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3802),
  [1275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3208),
  [1277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1681),
  [1279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4236),
  [1281] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__items_with_trailing_separator, 2, 0, 0),
  [1283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1480),
  [1285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4474),
  [1287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3368),
  [1289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3369),
  [1291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3370),
  [1293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1579),
  [1295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2702),
  [1297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3394),
  [1299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4293),
  [1301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1272),
  [1303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3433),
  [1305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1580),
  [1307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1902),
  [1309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1903),
  [1311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1904),
  [1313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3001),
  [1315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1449),
  [1317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1457),
  [1319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1458),
  [1321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4141),
  [1323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4142),
  [1325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1536),
  [1327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4143),
  [1329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3002),
  [1331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1632),
  [1333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1406),
  [1335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2033),
  [1337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3003),
  [1339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3135),
  [1341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4461),
  [1343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_body, 3, 0, 0),
  [1345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_body, 3, 0, 0),
  [1347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2587),
  [1349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_body, 2, 0, 0),
  [1351] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_body, 2, 0, 0),
  [1353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_body, 1, 0, 0),
  [1355] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_body, 1, 0, 0),
  [1357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1687),
  [1359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_body, 4, 0, 0),
  [1361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_body, 4, 0, 0),
  [1363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4442),
  [1365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(111),
  [1367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4445),
  [1369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2571),
  [1371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1237),
  [1373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3538),
  [1375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3652),
  [1377] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_block, 3, 0, 0),
  [1379] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rescue_block, 3, 0, 0),
  [1381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2647),
  [1383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2349),
  [1385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2653),
  [1387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4280),
  [1389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(721),
  [1391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3721),
  [1393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3727),
  [1395] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_after_block, 4, 0, 0),
  [1397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_block, 4, 0, 0),
  [1399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3147),
  [1401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3351),
  [1403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1005),
  [1405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4301),
  [1407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6851),
  [1409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(894),
  [1411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(408),
  [1413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(611),
  [1415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(612),
  [1417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4826),
  [1419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4307),
  [1421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_block, 4, 0, 0),
  [1423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rescue_block, 4, 0, 0),
  [1425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2963),
  [1427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2969),
  [1429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_after_block, 5, 0, 0),
  [1431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(75),
  [1433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1948),
  [1435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(896),
  [1437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1956),
  [1439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_block, 5, 0, 0),
  [1441] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_block, 5, 0, 0),
  [1443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3207),
  [1445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1388),
  [1447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3115),
  [1449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3120),
  [1451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rescue_block, 5, 0, 0),
  [1453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3166),
  [1455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3171),
  [1457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4160),
  [1459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2191),
  [1461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(897),
  [1463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2249),
  [1465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4029),
  [1467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1032),
  [1469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(348),
  [1471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3265),
  [1473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1516),
  [1475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1522),
  [1477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4024),
  [1479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(882),
  [1481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1597),
  [1483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2562),
  [1485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4040),
  [1487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3289),
  [1489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3097),
  [1491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3380),
  [1493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4171),
  [1495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3952),
  [1497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3968),
  [1499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2478),
  [1501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1380),
  [1503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4451),
  [1505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(893),
  [1507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(390),
  [1509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(800),
  [1511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(801),
  [1513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4816),
  [1515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4479),
  [1517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3762),
  [1519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3767),
  [1521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1688),
  [1523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [1525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(352),
  [1527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3212),
  [1529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6843),
  [1531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(431),
  [1533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(524),
  [1535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(525),
  [1537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4828),
  [1539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4468),
  [1541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4497),
  [1543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4235),
  [1545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4248),
  [1547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2716),
  [1549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4452),
  [1551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2960),
  [1553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2984),
  [1555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4067),
  [1557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(892),
  [1559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4072),
  [1561] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_after_block, 3, 0, 0),
  [1563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_block, 3, 0, 0),
  [1565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1602),
  [1567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4446),
  [1569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4485),
  [1571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3269),
  [1573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3108),
  [1575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3813),
  [1577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4444),
  [1579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(73),
  [1581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2519),
  [1583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2246),
  [1585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(91),
  [1587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4119),
  [1589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4113),
  [1591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4397),
  [1593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1293),
  [1595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4583),
  [1597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6149),
  [1599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(93),
  [1601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3677),
  [1603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3702),
  [1605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1298),
  [1607] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source, 1, 0, 0),
  [1609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3210),
  [1611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(103),
  [1613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3493),
  [1615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1464),
  [1617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4505),
  [1619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(96),
  [1621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3969),
  [1623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3004),
  [1625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3052),
  [1627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4587),
  [1629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6184),
  [1631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1508),
  [1633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3056),
  [1635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3014),
  [1637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4585),
  [1639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6039),
  [1641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4173),
  [1643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4575),
  [1645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6179),
  [1647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1705),
  [1649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3065),
  [1651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2329),
  [1653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4244),
  [1655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4562),
  [1657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3300),
  [1659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2308),
  [1661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1758),
  [1663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(98),
  [1665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3372),
  [1667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3371),
  [1669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2355),
  [1671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(79),
  [1673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4105),
  [1675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2020),
  [1677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(87),
  [1679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3987),
  [1681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1586),
  [1683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1766),
  [1685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source, 4, 0, 0),
  [1687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4447),
  [1689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4574),
  [1691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4594),
  [1693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(249),
  [1695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4283),
  [1697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4496),
  [1699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4588),
  [1701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5984),
  [1703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2182),
  [1705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4032),
  [1707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(236),
  [1709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3748),
  [1711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2291),
  [1713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1906),
  [1715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(89),
  [1717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4553),
  [1719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4550),
  [1721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4581),
  [1723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6148),
  [1725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4518),
  [1727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source, 3, 0, 0),
  [1729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3384),
  [1731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3822),
  [1733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1690),
  [1735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4541),
  [1737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1802),
  [1739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4570),
  [1741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6085),
  [1743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(77),
  [1745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4145),
  [1747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4144),
  [1749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4379),
  [1751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3676),
  [1753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4179),
  [1755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(84),
  [1757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1383),
  [1759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1308),
  [1761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3094),
  [1763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1917),
  [1765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(74),
  [1767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3481),
  [1769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3480),
  [1771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3409),
  [1773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1849),
  [1775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3154),
  [1777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3446),
  [1779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1650),
  [1781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3450),
  [1783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4593),
  [1785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6181),
  [1787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3161),
  [1789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1853),
  [1791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2932),
  [1793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4577),
  [1795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6174),
  [1797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1187),
  [1799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4377),
  [1801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1218),
  [1803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4151),
  [1805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4319),
  [1807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4572),
  [1809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6182),
  [1811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1564),
  [1813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2708),
  [1815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1502),
  [1817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1411),
  [1819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source, 2, 0, 0),
  [1821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(225),
  [1823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(184),
  [1825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(257),
  [1827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(219),
  [1829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(200),
  [1831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(254),
  [1833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(191),
  [1835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(178),
  [1837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(220),
  [1839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(235),
  [1841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4158),
  [1843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4195),
  [1845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4555),
  [1847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4556),
  [1849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4199),
  [1851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4504),
  [1853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4435),
  [1855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4529),
  [1857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4540),
  [1859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4569),
  [1861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4480),
  [1863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4481),
  [1865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4484),
  [1867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4495),
  [1869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4432),
  [1871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4508),
  [1873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4511),
  [1875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4512),
  [1877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4513),
  [1879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4520),
  [1881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4522),
  [1883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4523),
  [1885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4524),
  [1887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4530),
  [1889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4286),
  [1891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4147),
  [1893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4148),
  [1895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4289),
  [1897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4181),
  [1899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4187),
  [1901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4231),
  [1903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4234),
  [1905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4215),
  [1907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4238),
  [1909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4239),
  [1911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4242),
  [1913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4246),
  [1915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4259),
  [1917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4262),
  [1919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4263),
  [1921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4266),
  [1923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4267),
  [1925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4270),
  [1927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4271),
  [1929] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4272),
  [1931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4273),
  [1933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4274),
  [1935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3483),
  [1937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3484),
  [1939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2934),
  [1941] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2935),
  [1943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2958),
  [1945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2959),
  [1947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2970),
  [1949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2981),
  [1951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2982),
  [1953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2983),
  [1955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2986),
  [1957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2987),
  [1959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2988),
  [1961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2989),
  [1963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2990),
  [1965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2991),
  [1967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2992),
  [1969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2993),
  [1971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2994),
  [1973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2995),
  [1975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4396),
  [1977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1908),
  [1979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4161),
  [1981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2022),
  [1983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4065),
  [1985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4066),
  [1987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4068),
  [1989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4069),
  [1991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4070),
  [1993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4071),
  [1995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4073),
  [1997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4074),
  [1999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4075),
  [2001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4076),
  [2003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4077),
  [2005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4078),
  [2007] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4079),
  [2009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4103),
  [2011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4104),
  [2013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4107),
  [2015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2710),
  [2017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1310),
  [2019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2479),
  [2021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1189),
  [2023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2560),
  [2025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2561),
  [2027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2567),
  [2029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2568),
  [2031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2569),
  [2033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2570),
  [2035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2573),
  [2037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2574),
  [2039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2575),
  [2041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2576),
  [2043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2577),
  [2045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2578),
  [2047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2579),
  [2049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2580),
  [2051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2581),
  [2053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2582),
  [2055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2184),
  [2057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2185),
  [2059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2257),
  [2061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2259),
  [2063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2640),
  [2065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2646),
  [2067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2649),
  [2069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2650),
  [2071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2651),
  [2073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2652),
  [2075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2655),
  [2077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2656),
  [2079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2657),
  [2081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2658),
  [2083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2659),
  [2085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2660),
  [2087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2661),
  [2089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2662),
  [2091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2663),
  [2093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2314),
  [2095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4381),
  [2097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3679),
  [2099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3680),
  [2101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4164),
  [2103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3704),
  [2105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3705),
  [2107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3719),
  [2109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3720),
  [2111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3723),
  [2113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3724),
  [2115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3725),
  [2117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3726),
  [2119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3729),
  [2121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3730),
  [2123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3731),
  [2125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3732),
  [2127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3733),
  [2129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3734),
  [2131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3735),
  [2133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3736),
  [2135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3737),
  [2137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3738),
  [2139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4279),
  [2141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4392),
  [2143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4579),
  [2145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4393),
  [2147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4299),
  [2149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4300),
  [2151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4303),
  [2153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4304),
  [2155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4305),
  [2157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4306),
  [2159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4309),
  [2161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4310),
  [2163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4311),
  [2165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4312),
  [2167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4313),
  [2169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4314),
  [2171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4315),
  [2173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4316),
  [2175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4317),
  [2177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4394),
  [2179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3374),
  [2181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3375),
  [2183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4580),
  [2185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4031),
  [2187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1588),
  [2189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3386),
  [2191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3387),
  [2193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2961),
  [2195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2962),
  [2197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4124),
  [2199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2965),
  [2201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2966),
  [2203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2967),
  [2205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2968),
  [2207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2971),
  [2209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2972),
  [2211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2973),
  [2213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2974),
  [2215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2975),
  [2217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2976),
  [2219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2977),
  [2221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2978),
  [2223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2979),
  [2225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3398),
  [2227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2117),
  [2229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2118),
  [2231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1654),
  [2233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1946),
  [2235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1947),
  [2237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1949),
  [2239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1951),
  [2241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1952),
  [2243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1953),
  [2245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1957),
  [2247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1958),
  [2249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1959),
  [2251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1960),
  [2253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1961),
  [2255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1962),
  [2257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1963),
  [2259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1964),
  [2261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1965),
  [2263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2119),
  [2265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3495),
  [2267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1466),
  [2269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4127),
  [2271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2939),
  [2273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1523),
  [2275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3113),
  [2277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3114),
  [2279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3116),
  [2281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3117),
  [2283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3118),
  [2285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3119),
  [2287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3121),
  [2289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3122),
  [2291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3123),
  [2293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3124),
  [2295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3125),
  [2297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3126),
  [2299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3127),
  [2301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3128),
  [2303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3129),
  [2305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2980),
  [2307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3469),
  [2309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3851),
  [2311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3470),
  [2313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3164),
  [2315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4192),
  [2317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3167),
  [2319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3168),
  [2321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3169),
  [2323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3170),
  [2325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3172),
  [2327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3173),
  [2329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3174),
  [2331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3175),
  [2333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3176),
  [2335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3177),
  [2337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3178),
  [2339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3179),
  [2341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3180),
  [2343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3471),
  [2345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2335),
  [2347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4139),
  [2349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2336),
  [2351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2187),
  [2353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2190),
  [2355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2192),
  [2357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2196),
  [2359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2226),
  [2361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2141),
  [2363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2275),
  [2365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2276),
  [2367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2280),
  [2369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2281),
  [2371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2357),
  [2373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2155),
  [2375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2167),
  [2377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2169),
  [2379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2171),
  [2381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2337),
  [2383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1554),
  [2385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1555),
  [2387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1514),
  [2389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1515),
  [2391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1517),
  [2393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1518),
  [2395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1519),
  [2397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1520),
  [2399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4159),
  [2401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1526),
  [2403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1527),
  [2405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1528),
  [2407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1532),
  [2409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1533),
  [2411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1534),
  [2413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1535),
  [2415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1539),
  [2417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1541),
  [2419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1556),
  [2421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1795),
  [2423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1796),
  [2425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1595),
  [2427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1596),
  [2429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4022),
  [2431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1598),
  [2433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1599),
  [2435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1600),
  [2437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1601),
  [2439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4023),
  [2441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1603),
  [2443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1604),
  [2445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1605),
  [2447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1606),
  [2449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1607),
  [2451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1608),
  [2453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1609),
  [2455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1610),
  [2457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1611),
  [2459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1797),
  [2461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4026),
  [2463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4027),
  [2465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4028),
  [2467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4038),
  [2469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4042),
  [2471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4043),
  [2473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4050),
  [2475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4051),
  [2477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4052),
  [2479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3859),
  [2481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4053),
  [2483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4054),
  [2485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4055),
  [2487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4056),
  [2489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4057),
  [2491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3011),
  [2493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3016),
  [2495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3058),
  [2497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3059),
  [2499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4448),
  [2501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4450),
  [2503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4453),
  [2505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4456),
  [2507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4457),
  [2509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4458),
  [2511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4514),
  [2513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4534),
  [2515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4536),
  [2517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4537),
  [2519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4538),
  [2521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4539),
  [2523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4542),
  [2525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4433),
  [2527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4438),
  [2529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3080),
  [2531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3760),
  [2533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3761),
  [2535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3763),
  [2537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3764),
  [2539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3765),
  [2541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3766),
  [2543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3768),
  [2545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3769),
  [2547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3770),
  [2549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3771),
  [2551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3772),
  [2553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3773),
  [2555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3774),
  [2557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3775),
  [2559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3776),
  [2561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4535),
  [2563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3206),
  [2565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2684),
  [2567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2719),
  [2569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4133),
  [2571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4178),
  [2573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4182),
  [2575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3544),
  [2577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4183),
  [2579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4184),
  [2581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3927),
  [2583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3928),
  [2585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4185),
  [2587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3954),
  [2589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3955),
  [2591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3956),
  [2593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3957),
  [2595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3199),
  [2597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3970),
  [2599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3971),
  [2601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3972),
  [2603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3973),
  [2605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3974),
  [2607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3975),
  [2609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3976),
  [2611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3977),
  [2613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3978),
  [2615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3264),
  [2617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3895),
  [2619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4186),
  [2621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4030),
  [2623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3299),
  [2625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4190),
  [2627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1381),
  [2629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4157),
  [2631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4493),
  [2633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4189),
  [2635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1689),
  [2637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2076),
  [2639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2307),
  [2641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4589),
  [2643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4571),
  [2645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4573),
  [2647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4576),
  [2649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4578),
  [2651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4582),
  [2653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4584),
  [2655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4586),
  [2657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4590),
  [2659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4591),
  [2661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4592),
  [2663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3165),
  [2665] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_single, 3, 0, 18),
  [2667] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_single, 3, 0, 18),
  [2669] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_double, 4, 0, 24),
  [2671] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_double, 4, 0, 24),
  [2673] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_single, 4, 0, 24),
  [2675] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_single, 4, 0, 24),
  [2677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_dot, 3, 0, 22),
  [2679] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_dot, 3, 0, 22),
  [2681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_identifier, 1, 0, 0),
  [2683] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_identifier, 1, 0, 0),
  [2685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_operator_identifier, 1, 0, 0),
  [2687] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_operator_identifier, 1, 0, 0),
  [2689] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_dot, 3, 0, 21),
  [2691] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_dot, 3, 0, 21),
  [2693] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_dot, 3, 0, 20),
  [2695] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_dot, 3, 0, 20),
  [2697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_dot, 3, 0, 23),
  [2699] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_dot, 3, 0, 23),
  [2701] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_double, 2, 0, 8),
  [2703] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_double, 2, 0, 8),
  [2705] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_single, 2, 0, 8),
  [2707] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_single, 2, 0, 8),
  [2709] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_double, 3, 0, 18),
  [2711] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_double, 3, 0, 18),
  [2713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1019),
  [2715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1023),
  [2717] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__terminator, 1, 0, 0),
  [2719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__terminator, 1, 0, 0),
  [2721] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminator_repeat1, 2, 0, 0), SHIFT_REPEAT(1019),
  [2724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__terminator_repeat1, 2, 0, 0),
  [2726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__terminator_repeat1, 2, 0, 0),
  [2728] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminator_repeat1, 2, 0, 0), SHIFT_REPEAT(1020),
  [2731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1020),
  [2733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1024),
  [2735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1025),
  [2737] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__terminator, 2, 0, 0),
  [2739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__terminator, 2, 0, 0),
  [2741] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminator_repeat1, 2, 0, 0), SHIFT_REPEAT(1026),
  [2744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [2746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1040),
  [2748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminator_repeat1, 2, 0, 0), SHIFT_REPEAT(1028),
  [2751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1026),
  [2753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1037),
  [2755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [2757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1042),
  [2759] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminator_repeat1, 2, 0, 0), SHIFT_REPEAT(1031),
  [2762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [2764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1039),
  [2766] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminator_repeat1, 2, 0, 0), SHIFT_REPEAT(1033),
  [2769] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__terminator_repeat1, 2, 0, 0), SHIFT_REPEAT(1034),
  [2772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [2774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1041),
  [2776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1038),
  [2778] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_keyword, 2, 0, 2),
  [2780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_keyword, 2, 0, 2),
  [2782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__anonymous_dot, 2, 0, 11),
  [2784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4531),
  [2786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1000),
  [2788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5010),
  [2790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5011),
  [2792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1009),
  [2794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1002),
  [2796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(998),
  [2798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [2800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1004),
  [2802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1011),
  [2804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [2806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3081),
  [2808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(938),
  [2810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5001),
  [2812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5002),
  [2814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(930),
  [2816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(928),
  [2818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(939),
  [2820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(940),
  [2822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(941),
  [2824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(942),
  [2826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(930),
  [2828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1542),
  [2830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(920),
  [2832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4844),
  [2834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5053),
  [2836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(919),
  [2838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(921),
  [2840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(922),
  [2842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(923),
  [2844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(915),
  [2846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(916),
  [2848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(919),
  [2850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4275),
  [2852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2998),
  [2854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(955),
  [2856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4917),
  [2858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4918),
  [2860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(969),
  [2862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(996),
  [2864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(970),
  [2866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(971),
  [2868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(972),
  [2870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(973),
  [2872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [2874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2038),
  [2876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(974),
  [2878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4959),
  [2880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4960),
  [2882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(986),
  [2884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(975),
  [2886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(976),
  [2888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(977),
  [2890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(978),
  [2892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(979),
  [2894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [2896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1203),
  [2898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4212),
  [2900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3739),
  [2902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2315),
  [2904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1686),
  [2906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(967),
  [2908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5042),
  [2910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5043),
  [2912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(953),
  [2914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(956),
  [2916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(957),
  [2918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(946),
  [2920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(948),
  [2922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(949),
  [2924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(953),
  [2926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3399),
  [2928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1022),
  [2930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(171),
  [2932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(661),
  [2934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(287),
  [2936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(662),
  [2938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(264),
  [2940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(663),
  [2942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(664),
  [2944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(665),
  [2946] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__stab_clause_arguments_without_parentheses, 1, 0, 0), SHIFT(288),
  [2949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(666),
  [2951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(667),
  [2953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(668),
  [2955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(669),
  [2957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(670),
  [2959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(671),
  [2961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(672),
  [2963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(660),
  [2965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(673),
  [2967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(674),
  [2969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(675),
  [2971] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_without_parentheses, 1, 0, 0),
  [2973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1058),
  [2975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [2977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [2979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(145),
  [2981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(170),
  [2983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(165),
  [2985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(154),
  [2987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(160),
  [2989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(144),
  [2991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(153),
  [2993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(164),
  [2995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(159),
  [2997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(162),
  [2999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(140),
  [3001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(141),
  [3003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(147),
  [3005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(142),
  [3007] = {.entry = {.count = 1, .reusable = false}}, SHIFT(167),
  [3009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(168),
  [3011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(156),
  [3013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(173),
  [3015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(174),
  [3017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(148),
  [3019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(157),
  [3021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [3023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(151),
  [3025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_call_with_parentheses, 2, 0, 12),
  [3027] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_call_with_parentheses, 2, 0, 12),
  [3029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_with_parentheses, 1, 0, 3),
  [3031] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_with_parentheses, 1, 0, 3),
  [3033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(192),
  [3035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_call_with_parentheses, 2, 0, 4),
  [3037] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_call_with_parentheses, 2, 0, 4),
  [3039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6822),
  [3041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6840),
  [3043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(344),
  [3045] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_call_with_parentheses, 4, 0, 4),
  [3047] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_call_with_parentheses, 4, 0, 4),
  [3049] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__double_call, 2, 0, 13),
  [3051] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__double_call, 2, 0, 13),
  [3053] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_call_without_parentheses, 2, 0, 12),
  [3055] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_call_without_parentheses, 2, 0, 12),
  [3057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6841),
  [3059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(282),
  [3061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(268),
  [3063] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_block, 2, 0, 0),
  [3065] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_block, 2, 0, 0),
  [3067] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_call_with_parentheses, 3, 0, 12),
  [3069] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_call_with_parentheses, 3, 0, 12),
  [3071] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_call, 2, 0, 14),
  [3073] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__anonymous_call, 2, 0, 14),
  [3075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 1, 0, 2),
  [3077] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 1, 0, 2),
  [3079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [3081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(269),
  [3083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6867),
  [3085] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_block, 4, 0, 0),
  [3087] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_block, 4, 0, 0),
  [3089] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_block, 5, 0, 0),
  [3091] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_block, 5, 0, 0),
  [3093] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_call_with_parentheses, 3, 0, 4),
  [3095] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_call_with_parentheses, 3, 0, 4),
  [3097] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_block, 6, 0, 0),
  [3099] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_block, 6, 0, 0),
  [3101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_block, 3, 0, 0),
  [3103] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_block, 3, 0, 0),
  [3105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(283),
  [3107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_arguments_with_parentheses_immediate, 2, 0, 0),
  [3109] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_with_parentheses_immediate, 2, 0, 0),
  [3111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6803),
  [3113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(277),
  [3115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6819),
  [3117] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_call_without_parentheses, 2, 0, 4),
  [3119] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_call_without_parentheses, 2, 0, 4),
  [3121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6823),
  [3123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_charlist, 1, 0, 2),
  [3125] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_charlist, 1, 0, 2),
  [3127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1044),
  [3129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(343),
  [3131] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_arguments_with_parentheses, 2, 0, 0),
  [3133] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_with_parentheses, 2, 0, 0),
  [3135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6863),
  [3137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(276),
  [3139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6774),
  [3141] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_arguments_with_parentheses, 3, 0, 0),
  [3143] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_with_parentheses, 3, 0, 0),
  [3145] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_arguments_with_parentheses_immediate, 3, 0, 0),
  [3147] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_with_parentheses_immediate, 3, 0, 0),
  [3149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(237),
  [3151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_call_with_parentheses, 4, 0, 12),
  [3153] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_call_with_parentheses, 4, 0, 12),
  [3155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_block, 7, 0, 0),
  [3157] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_block, 7, 0, 0),
  [3159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 3, 0, 0),
  [3161] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 3, 0, 0),
  [3163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_slash, 2, 0, 8),
  [3165] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_slash, 2, 0, 8),
  [3167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_parenthesis, 2, 0, 8),
  [3169] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_parenthesis, 2, 0, 8),
  [3171] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_double, 2, 0, 8),
  [3173] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_double, 2, 0, 8),
  [3175] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_single, 2, 0, 8),
  [3177] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_single, 2, 0, 8),
  [3179] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_heredoc_single, 2, 0, 8),
  [3181] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_heredoc_single, 2, 0, 8),
  [3183] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_heredoc_double, 2, 0, 8),
  [3185] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_heredoc_double, 2, 0, 8),
  [3187] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_curly, 2, 0, 8),
  [3189] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_curly, 2, 0, 8),
  [3191] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_square, 2, 0, 8),
  [3193] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_square, 2, 0, 8),
  [3195] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_angle, 2, 0, 8),
  [3197] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_angle, 2, 0, 8),
  [3199] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_bar, 2, 0, 8),
  [3201] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_bar, 2, 0, 8),
  [3203] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_slash, 2, 0, 8),
  [3205] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_slash, 2, 0, 8),
  [3207] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_parenthesis, 3, 0, 18),
  [3209] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_parenthesis, 3, 0, 18),
  [3211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_curly, 3, 0, 18),
  [3213] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_curly, 3, 0, 18),
  [3215] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_square, 3, 0, 18),
  [3217] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_square, 3, 0, 18),
  [3219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_angle, 3, 0, 18),
  [3221] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_angle, 3, 0, 18),
  [3223] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_bar, 3, 0, 18),
  [3225] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_bar, 3, 0, 18),
  [3227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_slash, 3, 0, 18),
  [3229] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_slash, 3, 0, 18),
  [3231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_parenthesis, 3, 0, 18),
  [3233] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_parenthesis, 3, 0, 18),
  [3235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_double, 3, 0, 18),
  [3237] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_double, 3, 0, 18),
  [3239] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_single, 3, 0, 18),
  [3241] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_single, 3, 0, 18),
  [3243] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_heredoc_single, 3, 0, 18),
  [3245] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_heredoc_single, 3, 0, 18),
  [3247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_heredoc_double, 3, 0, 18),
  [3249] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_heredoc_double, 3, 0, 18),
  [3251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_curly, 3, 0, 18),
  [3253] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_curly, 3, 0, 18),
  [3255] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_angle, 3, 0, 18),
  [3257] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_angle, 3, 0, 18),
  [3259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_bar, 3, 0, 18),
  [3261] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_bar, 3, 0, 18),
  [3263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_slash, 3, 0, 18),
  [3265] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_slash, 3, 0, 18),
  [3267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_parenthesis, 4, 0, 24),
  [3269] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_parenthesis, 4, 0, 24),
  [3271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_curly, 4, 0, 24),
  [3273] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_curly, 4, 0, 24),
  [3275] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_square, 4, 0, 24),
  [3277] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_square, 4, 0, 24),
  [3279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_angle, 4, 0, 24),
  [3281] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_angle, 4, 0, 24),
  [3283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_bar, 4, 0, 24),
  [3285] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_bar, 4, 0, 24),
  [3287] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_slash, 4, 0, 24),
  [3289] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_slash, 4, 0, 24),
  [3291] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_parenthesis, 4, 0, 24),
  [3293] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_parenthesis, 4, 0, 24),
  [3295] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_double, 4, 0, 24),
  [3297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_double, 4, 0, 24),
  [3299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_single, 4, 0, 24),
  [3301] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_single, 4, 0, 24),
  [3303] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_heredoc_single, 4, 0, 24),
  [3305] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_heredoc_single, 4, 0, 24),
  [3307] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_heredoc_double, 4, 0, 24),
  [3309] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_heredoc_double, 4, 0, 24),
  [3311] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_curly, 4, 0, 24),
  [3313] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_curly, 4, 0, 24),
  [3315] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_square, 4, 0, 24),
  [3317] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_square, 4, 0, 24),
  [3319] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_angle, 4, 0, 24),
  [3321] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_angle, 4, 0, 24),
  [3323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_bar, 4, 0, 24),
  [3325] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_bar, 4, 0, 24),
  [3327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_slash, 4, 0, 24),
  [3329] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_slash, 4, 0, 24),
  [3331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 2, 0, 0),
  [3333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 2, 0, 0),
  [3335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_operator, 3, -1, 15),
  [3337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_operator, 3, -1, 15),
  [3339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 3, 0, 0),
  [3341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 3, 0, 0),
  [3343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_heredoc_single, 3, 0, 18),
  [3345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_heredoc_single, 3, 0, 18),
  [3347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_heredoc_double, 3, 0, 18),
  [3349] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_heredoc_double, 3, 0, 18),
  [3351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 3, 0, 0),
  [3353] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 3, 0, 0),
  [3355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bitstring, 3, 0, 0),
  [3357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bitstring, 3, 0, 0),
  [3359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map, 3, 0, 0),
  [3361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map, 3, 0, 0),
  [3363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 3, 0, 0),
  [3365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 3, 0, 0),
  [3367] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_operator, 3, 0, 20),
  [3369] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_operator, 3, 0, 20),
  [3371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dot, 3, 0, 20),
  [3373] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dot, 3, 0, 20),
  [3375] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_call_without_parentheses, 3, 0, 12),
  [3377] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_call_without_parentheses, 3, 0, 12),
  [3379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__double_call, 3, 0, 13),
  [3381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__double_call, 3, 0, 13),
  [3383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_call_without_parentheses, 3, 0, 4),
  [3385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_call_without_parentheses, 3, 0, 4),
  [3387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_heredoc_single, 2, 0, 8),
  [3389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_heredoc_single, 2, 0, 8),
  [3391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_heredoc_double, 2, 0, 8),
  [3393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_heredoc_double, 2, 0, 8),
  [3395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_heredoc_single, 4, 0, 24),
  [3397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_heredoc_single, 4, 0, 24),
  [3399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_heredoc_double, 4, 0, 24),
  [3401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_heredoc_double, 4, 0, 24),
  [3403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 4, 0, 0),
  [3405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 4, 0, 0),
  [3407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6804),
  [3409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6805),
  [3411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6806),
  [3413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6821),
  [3415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6824),
  [3417] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [3419] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean, 1, 0, 0),
  [3421] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nil, 1, 0, 0),
  [3423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nil, 1, 0, 0),
  [3425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sigil, 4, 0, 19),
  [3427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sigil, 4, 0, 19),
  [3429] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__nullary_operator, 1, 0, 1),
  [3431] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__nullary_operator, 1, 0, 1),
  [3433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_arguments_without_parentheses, 1, -1, 0),
  [3435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_without_parentheses, 1, -1, 0),
  [3437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(745),
  [3439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(304),
  [3441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(746),
  [3443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(319),
  [3445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(747),
  [3447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(748),
  [3449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(749),
  [3451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(305),
  [3453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(751),
  [3455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(752),
  [3457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(753),
  [3459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(754),
  [3461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(755),
  [3463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(756),
  [3465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(757),
  [3467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(744),
  [3469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(758),
  [3471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(759),
  [3473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(760),
  [3475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1062),
  [3477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [3479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [3481] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call, 1, 0, 3),
  [3483] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call, 1, 0, 3),
  [3485] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_without_parentheses, 1, 0, 3),
  [3487] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_without_parentheses, 1, 0, 3),
  [3489] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_atom, 2, 0, 5),
  [3491] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_atom, 2, 0, 5),
  [3493] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map, 4, 0, 25),
  [3495] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map, 4, 0, 25),
  [3497] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map, 4, 0, 0),
  [3499] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map, 4, 0, 0),
  [3501] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__capture_expression, 3, 0, 0),
  [3503] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__capture_expression, 3, 0, 0),
  [3505] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 4, 0, 0),
  [3507] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 4, 0, 0),
  [3509] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_access_call, 4, 0, 26),
  [3511] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_access_call, 4, 0, 26),
  [3513] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_call_without_parentheses, 4, 0, 12),
  [3515] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_call_without_parentheses, 4, 0, 12),
  [3517] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__double_call, 4, 0, 13),
  [3519] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__double_call, 4, 0, 13),
  [3521] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__remote_call_without_parentheses, 4, 0, 4),
  [3523] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__remote_call_without_parentheses, 4, 0, 4),
  [3525] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 5, 0, 0),
  [3527] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 5, 0, 0),
  [3529] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map, 5, 0, 27),
  [3531] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map, 5, 0, 27),
  [3533] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 5, 0, 0),
  [3535] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 5, 0, 0),
  [3537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 6, 0, 0),
  [3539] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 6, 0, 0),
  [3541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 2, 0, 0),
  [3543] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 2, 0, 0),
  [3545] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_arguments_without_parentheses, 3, -1, 0),
  [3547] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_without_parentheses, 3, -1, 0),
  [3549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_arguments_without_parentheses, 4, -1, 0),
  [3551] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_without_parentheses, 4, -1, 0),
  [3553] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 2, 0, 0),
  [3555] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 2, 0, 0),
  [3557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_call_just_do_block, 2, 0, 12),
  [3559] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_call_just_do_block, 2, 0, 12),
  [3561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bitstring, 2, 0, 0),
  [3563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bitstring, 2, 0, 0),
  [3565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_operator, 2, -1, 10),
  [3567] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_operator, 2, -1, 10),
  [3569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 2, 0, 0),
  [3571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 2, 0, 0),
  [3573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sigil, 3, 0, 19),
  [3575] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sigil, 3, 0, 19),
  [3577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1236),
  [3579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1238),
  [3581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1240),
  [3583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1241),
  [3585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1242),
  [3587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1243),
  [3589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1244),
  [3591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1245),
  [3593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1246),
  [3595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1247),
  [3597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1248),
  [3599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1249),
  [3601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1251),
  [3603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1252),
  [3605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1268),
  [3607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1269),
  [3609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1271),
  [3611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1274),
  [3613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1275),
  [3615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1279),
  [3617] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0),
  [3619] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0),
  [3621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_parenthesis, 2, 0, 8),
  [3623] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_parenthesis, 2, 0, 8),
  [3625] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_curly, 2, 0, 8),
  [3627] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_curly, 2, 0, 8),
  [3629] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_square, 2, 0, 8),
  [3631] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_square, 2, 0, 8),
  [3633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_angle, 2, 0, 8),
  [3635] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_angle, 2, 0, 8),
  [3637] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__call_arguments_without_parentheses, 2, -1, 0),
  [3639] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_without_parentheses, 2, -1, 0),
  [3641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(320),
  [3643] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0),
  [3645] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0),
  [3647] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(886),
  [3650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_i_bar, 2, 0, 8),
  [3652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_i_bar, 2, 0, 8),
  [3654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keywords, 1, 0, 0),
  [3656] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keywords, 1, 0, 0),
  [3658] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4830),
  [3660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keywords, 2, 0, 0),
  [3662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_keywords, 2, 0, 0),
  [3664] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4830),
  [3667] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_square, 3, 0, 18),
  [3669] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__quoted_square, 3, 0, 18),
  [3671] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pair, 2, 0, 17),
  [3673] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pair, 2, 0, 17),
  [3675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__capture_expression, 1, 0, 0),
  [3677] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__capture_expression, 1, 0, 0),
  [3679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1566),
  [3681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1567),
  [3683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1568),
  [3685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1569),
  [3687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(766),
  [3689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(307),
  [3691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(768),
  [3693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(326),
  [3695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(769),
  [3697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(770),
  [3699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(771),
  [3701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(345),
  [3703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(773),
  [3705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(774),
  [3707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(775),
  [3709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(776),
  [3711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(777),
  [3713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(778),
  [3715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(779),
  [3717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(765),
  [3719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(780),
  [3721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(781),
  [3723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(782),
  [3725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1047),
  [3727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(908),
  [3729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(765),
  [3731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1389),
  [3733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1391),
  [3735] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4834),
  [3738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1424),
  [3740] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1426),
  [3742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1429),
  [3744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1432),
  [3746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1437),
  [3748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1440),
  [3750] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1441),
  [3752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1442),
  [3754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1453),
  [3756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1455),
  [3758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1471),
  [3760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1574),
  [3762] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1476),
  [3764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1477),
  [3766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4834),
  [3768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(328),
  [3770] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(890),
  [3773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1030),
  [3775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(435),
  [3777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4543),
  [3779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(772),
  [3781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(306),
  [3783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(783),
  [3785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(314),
  [3787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(784),
  [3789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(785),
  [3791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(786),
  [3793] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__stab_clause_arguments_without_parentheses, 1, 0, 0), SHIFT(309),
  [3796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(787),
  [3798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(788),
  [3800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(789),
  [3802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(790),
  [3804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(791),
  [3806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(793),
  [3808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(794),
  [3810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(767),
  [3812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(795),
  [3814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(796),
  [3816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(797),
  [3818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1055),
  [3820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(767),
  [3822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(201),
  [3824] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_with_parentheses, 2, 0, 0),
  [3826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1714),
  [3828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1715),
  [3830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1716),
  [3832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1575),
  [3834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1718),
  [3836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1719),
  [3838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1720),
  [3840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1722),
  [3842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1723),
  [3844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1724),
  [3846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1726),
  [3848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1727),
  [3850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1728),
  [3852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1729),
  [3854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1730),
  [3856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1731),
  [3858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1733),
  [3860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1734),
  [3862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1736),
  [3864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1737),
  [3866] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_body, 1, 0, 0), SHIFT(1021),
  [3869] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_body, 1, 0, 0), SHIFT(259),
  [3872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(725),
  [3874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(299),
  [3876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(726),
  [3878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(727),
  [3880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(728),
  [3882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(729),
  [3884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(300),
  [3886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(730),
  [3888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(731),
  [3890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(732),
  [3892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(733),
  [3894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(734),
  [3896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(735),
  [3898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(736),
  [3900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(724),
  [3902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(737),
  [3904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(738),
  [3906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(739),
  [3908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1061),
  [3910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [3912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [3914] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_body, 2, 0, 0), SHIFT(1021),
  [3917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_body, 2, 0, 0), SHIFT(258),
  [3920] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4833),
  [3923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(374),
  [3925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1791),
  [3927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6838),
  [3929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6813),
  [3931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_with_parentheses, 3, 0, 0),
  [3933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4833),
  [3935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(416),
  [3937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(381),
  [3939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1741),
  [3941] = {.entry = {.count = 1, .reusable = false}}, SHIFT(288),
  [3943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1823),
  [3945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1824),
  [3947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1825),
  [3949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1826),
  [3951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1827),
  [3953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(364),
  [3955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(421),
  [3957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3145),
  [3959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(370),
  [3961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1822),
  [3963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(402),
  [3965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4475),
  [3967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(357),
  [3969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4375),
  [3971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1828),
  [3973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(372),
  [3975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(389),
  [3977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(409),
  [3979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1829),
  [3981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1830),
  [3983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(366),
  [3985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3042),
  [3987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1831),
  [3989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1834),
  [3991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(432),
  [3993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(420),
  [3995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1837),
  [3997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(358),
  [3999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1282),
  [4001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6862),
  [4003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6826),
  [4005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6834),
  [4007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6845),
  [4009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6784),
  [4011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(376),
  [4013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(380),
  [4015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2277),
  [4017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4284),
  [4019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(398),
  [4021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(354),
  [4023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3803),
  [4025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2350),
  [4027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2174),
  [4029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2175),
  [4031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2189),
  [4033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2194),
  [4035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2201),
  [4037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2208),
  [4039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2211),
  [4041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2214),
  [4043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2220),
  [4045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(419),
  [4047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2234),
  [4049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2241),
  [4051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(422),
  [4053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3434),
  [4055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2242),
  [4057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2252),
  [4059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2253),
  [4061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(437),
  [4063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2255),
  [4065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2363),
  [4067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2262),
  [4069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(440),
  [4071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1386),
  [4073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2264),
  [4075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2266),
  [4077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4820),
  [4079] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4820),
  [4082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1805),
  [4084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1806),
  [4086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1815),
  [4088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1816),
  [4090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1817),
  [4092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1818),
  [4094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1819),
  [4096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1820),
  [4098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1821),
  [4100] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4827),
  [4103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4827),
  [4105] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(846),
  [4108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6809),
  [4110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6864),
  [4112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1054),
  [4114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [4116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6797),
  [4118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(209),
  [4120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4804),
  [4122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(572),
  [4124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(270),
  [4126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(574),
  [4128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(335),
  [4130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(575),
  [4132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(576),
  [4134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(577),
  [4136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(271),
  [4138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(578),
  [4140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(579),
  [4142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(580),
  [4144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(581),
  [4146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(582),
  [4148] = {.entry = {.count = 1, .reusable = false}}, SHIFT(583),
  [4150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(584),
  [4152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(571),
  [4154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(585),
  [4156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(586),
  [4158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(587),
  [4160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [4162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(185),
  [4164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(337),
  [4166] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4804),
  [4169] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(844),
  [4172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(321),
  [4174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(552),
  [4176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(308),
  [4178] = {.entry = {.count = 1, .reusable = false}}, SHIFT(553),
  [4180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(318),
  [4182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(554),
  [4184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(555),
  [4186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(556),
  [4188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(265),
  [4190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(557),
  [4192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(558),
  [4194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(559),
  [4196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(560),
  [4198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(561),
  [4200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(562),
  [4202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [4204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(551),
  [4206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(564),
  [4208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(565),
  [4210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(566),
  [4212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1051),
  [4214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [4216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6818),
  [4218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3029),
  [4220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3030),
  [4222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3031),
  [4224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3032),
  [4226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4825),
  [4228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3033),
  [4230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3034),
  [4232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3035),
  [4234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3036),
  [4236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3037),
  [4238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3038),
  [4240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3039),
  [4242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4807),
  [4244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(509),
  [4246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(339),
  [4248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(510),
  [4250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(313),
  [4252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(511),
  [4254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(512),
  [4256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(513),
  [4258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(340),
  [4260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(514),
  [4262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(515),
  [4264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(516),
  [4266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(517),
  [4268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(518),
  [4270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(519),
  [4272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(520),
  [4274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(508),
  [4276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(521),
  [4278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(522),
  [4280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(523),
  [4282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1049),
  [4284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [4286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [4288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0),
  [4290] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0),
  [4292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(292),
  [4294] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4800),
  [4297] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(860),
  [4300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(909),
  [4302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(296),
  [4304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(705),
  [4306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(311),
  [4308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(706),
  [4310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(707),
  [4312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(708),
  [4314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(297),
  [4316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(709),
  [4318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(710),
  [4320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(711),
  [4322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(712),
  [4324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(713),
  [4326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(714),
  [4328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(715),
  [4330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(703),
  [4332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(716),
  [4334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(717),
  [4336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(718),
  [4338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1060),
  [4340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [4342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(256),
  [4344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6829),
  [4346] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(879),
  [4349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_repeat1, 2, 0, 0),
  [4351] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_source_repeat1, 2, 0, 0),
  [4353] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4807),
  [4356] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4825),
  [4359] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(842),
  [4362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6771),
  [4364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(640),
  [4366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(284),
  [4368] = {.entry = {.count = 1, .reusable = false}}, SHIFT(642),
  [4370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(278),
  [4372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(643),
  [4374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(644),
  [4376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(645),
  [4378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(285),
  [4380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(646),
  [4382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(647),
  [4384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(648),
  [4386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(649),
  [4388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(650),
  [4390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(651),
  [4392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(652),
  [4394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(639),
  [4396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(653),
  [4398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(654),
  [4400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(655),
  [4402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1057),
  [4404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [4406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [4408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6810),
  [4410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6793),
  [4412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4800),
  [4414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6790),
  [4416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6796),
  [4418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6860),
  [4420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6801),
  [4422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6783),
  [4424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6794),
  [4426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(682),
  [4428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(293),
  [4430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(683),
  [4432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(291),
  [4434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(684),
  [4436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(685),
  [4438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(686),
  [4440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(294),
  [4442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(687),
  [4444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(688),
  [4446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(689),
  [4448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(690),
  [4450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(691),
  [4452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(692),
  [4454] = {.entry = {.count = 1, .reusable = false}}, SHIFT(693),
  [4456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(681),
  [4458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(694),
  [4460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(695),
  [4462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(696),
  [4464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1059),
  [4466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [4468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(280),
  [4470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(336),
  [4472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3020),
  [4474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3098),
  [4476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3099),
  [4478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3100),
  [4480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3101),
  [4482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3102),
  [4484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3103),
  [4486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3104),
  [4488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3105),
  [4490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3106),
  [4492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3107),
  [4494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3132),
  [4496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3133),
  [4498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3134),
  [4500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3136),
  [4502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3137),
  [4504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3138),
  [4506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3139),
  [4508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3140),
  [4510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3141),
  [4512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3142),
  [4514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6861),
  [4516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3021),
  [4518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3022),
  [4520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3412),
  [4522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3413),
  [4524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3414),
  [4526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3415),
  [4528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3416),
  [4530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3417),
  [4532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3418),
  [4534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3419),
  [4536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3420),
  [4538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3421),
  [4540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3422),
  [4542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3423),
  [4544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3424),
  [4546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3425),
  [4548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3426),
  [4550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3427),
  [4552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3428),
  [4554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3429),
  [4556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3430),
  [4558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(312),
  [4560] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(883),
  [4563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3431),
  [4565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3023),
  [4567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3024),
  [4569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3025),
  [4571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3026),
  [4573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3027),
  [4575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3028),
  [4577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4832),
  [4579] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4832),
  [4582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6808),
  [4584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1046),
  [4586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [4588] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_body, 2, 0, 0), SHIFT(1036),
  [4591] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_body, 2, 0, 0), SHIFT(347),
  [4594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(530),
  [4596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(341),
  [4598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(532),
  [4600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(533),
  [4602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(534),
  [4604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(535),
  [4606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(342),
  [4608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(537),
  [4610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(538),
  [4612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(539),
  [4614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(540),
  [4616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(541),
  [4618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(542),
  [4620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(543),
  [4622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(529),
  [4624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(544),
  [4626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(545),
  [4628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(546),
  [4630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1050),
  [4632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [4634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(595),
  [4636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [4638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(596),
  [4640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(266),
  [4642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(597),
  [4644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(598),
  [4646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(599),
  [4648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(275),
  [4650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(600),
  [4652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(601),
  [4654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(602),
  [4656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(603),
  [4658] = {.entry = {.count = 1, .reusable = false}}, SHIFT(604),
  [4660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(605),
  [4662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(606),
  [4664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(594),
  [4666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(607),
  [4668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(608),
  [4670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(609),
  [4672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1053),
  [4674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [4676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [4678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6831),
  [4680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6773),
  [4682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6785),
  [4684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6868),
  [4686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6833),
  [4688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(267),
  [4690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(403),
  [4692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(610),
  [4694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(272),
  [4696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(641),
  [4698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(678),
  [4700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(456),
  [4702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(750),
  [4704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(298),
  [4706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(851),
  [4708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(854),
  [4710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(856),
  [4712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(857),
  [4714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(863),
  [4716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(884),
  [4718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(889),
  [4720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(591),
  [4722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(704),
  [4724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(457),
  [4726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(460),
  [4728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1052),
  [4730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [4732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [4734] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_body, 1, 0, 0), SHIFT(1036),
  [4737] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_body, 1, 0, 0), SHIFT(351),
  [4740] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4819),
  [4742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(443),
  [4744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4811),
  [4746] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4819),
  [4749] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4811),
  [4752] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(850),
  [4755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4824),
  [4757] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_body, 1, 0, 0), SHIFT(1032),
  [4760] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_body, 1, 0, 0), SHIFT(349),
  [4763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(617),
  [4765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(279),
  [4767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(618),
  [4769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(619),
  [4771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(620),
  [4773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(621),
  [4775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(281),
  [4777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(622),
  [4779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(623),
  [4781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(624),
  [4783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(625),
  [4785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(626),
  [4787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(627),
  [4789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(628),
  [4791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(616),
  [4793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(629),
  [4795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(630),
  [4797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(631),
  [4799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1056),
  [4801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [4803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3757),
  [4805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3758),
  [4807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3759),
  [4809] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_body, 2, 0, 0), SHIFT(1032),
  [4812] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_body, 2, 0, 0), SHIFT(350),
  [4815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3778),
  [4817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3779),
  [4819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3780),
  [4821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3781),
  [4823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3782),
  [4825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3783),
  [4827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3784),
  [4829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3791),
  [4831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3792),
  [4833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3793),
  [4835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3794),
  [4837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3795),
  [4839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3796),
  [4841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3797),
  [4843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3798),
  [4845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3799),
  [4847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3800),
  [4849] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4824),
  [4852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(826),
  [4854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(324),
  [4856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(827),
  [4858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(301),
  [4860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(828),
  [4862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(829),
  [4864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(830),
  [4866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(325),
  [4868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(831),
  [4870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(832),
  [4872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(833),
  [4874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(834),
  [4876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(835),
  [4878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(836),
  [4880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(837),
  [4882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(825),
  [4884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(838),
  [4886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(839),
  [4888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(840),
  [4890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [4892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(302),
  [4894] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(881),
  [4897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4831),
  [4899] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4831),
  [4902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4254),
  [4904] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4805),
  [4907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4252),
  [4909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4805),
  [4911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4808),
  [4913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4346),
  [4915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4347),
  [4917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4329),
  [4919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4330),
  [4921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4331),
  [4923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4332),
  [4925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4333),
  [4927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4334),
  [4929] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4335),
  [4931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4336),
  [4933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4337),
  [4935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4340),
  [4937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4350),
  [4939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4351),
  [4941] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4354),
  [4943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4355),
  [4945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4356),
  [4947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4357),
  [4949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4368),
  [4951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4370),
  [4953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4371),
  [4955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4372),
  [4957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4426),
  [4959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4251),
  [4961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4277),
  [4963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4809),
  [4965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4256),
  [4967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4320),
  [4969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4378),
  [4971] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4808),
  [4974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4101),
  [4976] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4136),
  [4978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4153),
  [4980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(862),
  [4982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(315),
  [4984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(864),
  [4986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(865),
  [4988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(866),
  [4990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(867),
  [4992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(316),
  [4994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(869),
  [4996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(870),
  [4998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(871),
  [5000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(872),
  [5002] = {.entry = {.count = 1, .reusable = false}}, SHIFT(873),
  [5004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(874),
  [5006] = {.entry = {.count = 1, .reusable = false}}, SHIFT(875),
  [5008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(861),
  [5010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(876),
  [5012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(877),
  [5014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(878),
  [5016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [5018] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__items_with_trailing_separator, 1, 0, 0),
  [5020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(227),
  [5022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4162),
  [5024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(309),
  [5026] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4809),
  [5029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4257),
  [5031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4253),
  [5033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4839),
  [5035] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4839),
  [5038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4243),
  [5040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4247),
  [5042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4249),
  [5044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4250),
  [5046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(488),
  [5048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(333),
  [5050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(490),
  [5052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(491),
  [5054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(492),
  [5056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(493),
  [5058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(334),
  [5060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(494),
  [5062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(495),
  [5064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(496),
  [5066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(497),
  [5068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(498),
  [5070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(499),
  [5072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(500),
  [5074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(487),
  [5076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(501),
  [5078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(502),
  [5080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(503),
  [5082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1048),
  [5084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [5086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [5088] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__stab_clause_arguments_with_parentheses_repeat1, 2, 0, 0),
  [5090] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__stab_clause_arguments_with_parentheses_repeat1, 2, 0, 0), REDUCE(aux_sym__stab_clause_arguments_without_parentheses_repeat1, 2, 0, 0),
  [5093] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__stab_clause_arguments_without_parentheses_repeat1, 2, 0, 0), SHIFT(334),
  [5096] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__stab_clause_arguments_without_parentheses_repeat1, 2, 0, 0),
  [5098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4544),
  [5100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4552),
  [5102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4560),
  [5104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4561),
  [5106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4563),
  [5108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4564),
  [5110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4565),
  [5112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4566),
  [5114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4507),
  [5116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4434),
  [5118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4455),
  [5120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4459),
  [5122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4463),
  [5124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4464),
  [5126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4465),
  [5128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4466),
  [5130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4469),
  [5132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4470),
  [5134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4471),
  [5136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4472),
  [5138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4803),
  [5140] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_with_trailing_separator, 1, 0, 0),
  [5142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(332),
  [5144] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__stab_clause_arguments_without_parentheses, 1, 0, 0), SHIFT(334),
  [5147] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4803),
  [5150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(465),
  [5152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(330),
  [5154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(467),
  [5156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(263),
  [5158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(468),
  [5160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(469),
  [5162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(470),
  [5164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(331),
  [5166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(471),
  [5168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(472),
  [5170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(473),
  [5172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(474),
  [5174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(475),
  [5176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(476),
  [5178] = {.entry = {.count = 1, .reusable = false}}, SHIFT(477),
  [5180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(464),
  [5182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(478),
  [5184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(479),
  [5186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(480),
  [5188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1045),
  [5190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [5192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [5194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4812),
  [5196] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4812),
  [5199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(806),
  [5201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(807),
  [5203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(808),
  [5205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(818),
  [5207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(819),
  [5209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(817),
  [5211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(816),
  [5213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(804),
  [5215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [5217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(805),
  [5219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(322),
  [5221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(811),
  [5223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(812),
  [5225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(813),
  [5227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(814),
  [5229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(815),
  [5231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(323),
  [5233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(810),
  [5235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(809),
  [5237] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct, 1, 0, 9),
  [5239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4287),
  [5241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5989),
  [5243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3148),
  [5245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6171),
  [5247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4477),
  [5249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5988),
  [5251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4384),
  [5253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5961),
  [5255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3044),
  [5257] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_with_parentheses_with_guard, 3, 0, 20),
  [5259] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_without_parentheses_with_guard, 3, 1, 20),
  [5261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6183),
  [5263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1839),
  [5265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6173),
  [5267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1284),
  [5269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5991),
  [5271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2294),
  [5273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6049),
  [5275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6062),
  [5277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1743),
  [5279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3805),
  [5281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3436),
  [5283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1414),
  [5285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6118),
  [5287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4900),
  [5289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4987),
  [5291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4988),
  [5293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4947),
  [5295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4948),
  [5297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4901),
  [5299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4902),
  [5301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4903),
  [5303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4904),
  [5305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4905),
  [5307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4879),
  [5309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5026),
  [5311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5027),
  [5313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5008),
  [5315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5009),
  [5317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4880),
  [5319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4881),
  [5321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4882),
  [5323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4883),
  [5325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4884),
  [5327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6003),
  [5329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6004),
  [5331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6005),
  [5333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6006),
  [5335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6007),
  [5337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6008),
  [5339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6009),
  [5341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6010),
  [5343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6011),
  [5345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6012),
  [5347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5963),
  [5349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5965),
  [5351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5966),
  [5353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5977),
  [5355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5978),
  [5357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5979),
  [5359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5980),
  [5361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5981),
  [5363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5982),
  [5365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5983),
  [5367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4848),
  [5369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5006),
  [5371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5007),
  [5373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4961),
  [5375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4962),
  [5377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4849),
  [5379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4850),
  [5381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4875),
  [5383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4892),
  [5385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4893),
  [5387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4979),
  [5389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4931),
  [5391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4932),
  [5393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4919),
  [5395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4920),
  [5397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4986),
  [5399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5005),
  [5401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5064),
  [5403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5066),
  [5405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5071),
  [5407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6104),
  [5409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6105),
  [5411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6106),
  [5413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6107),
  [5415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6108),
  [5417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6109),
  [5419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6110),
  [5421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6111),
  [5423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6112),
  [5425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6113),
  [5427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4930),
  [5429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5030),
  [5431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5031),
  [5433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5012),
  [5435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5013),
  [5437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4935),
  [5439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4936),
  [5441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4937),
  [5443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4944),
  [5445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4949),
  [5447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6063),
  [5449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6064),
  [5451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6065),
  [5453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6066),
  [5455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6067),
  [5457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6068),
  [5459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6069),
  [5461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6070),
  [5463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6071),
  [5465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6072),
  [5467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6025),
  [5469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6026),
  [5471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6027),
  [5473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6038),
  [5475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6040),
  [5477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6042),
  [5479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6043),
  [5481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6045),
  [5483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6046),
  [5485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6047),
  [5487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4851),
  [5489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5018),
  [5491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5019),
  [5493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4989),
  [5495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4990),
  [5497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4852),
  [5499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4853),
  [5501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4854),
  [5503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4855),
  [5505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4856),
  [5507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6083),
  [5509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6084),
  [5511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6185),
  [5513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6086),
  [5515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6087),
  [5517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6088),
  [5519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6089),
  [5521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6090),
  [5523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6091),
  [5525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6092),
  [5527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5077),
  [5529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5034),
  [5531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5035),
  [5533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5016),
  [5535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5017),
  [5537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5079),
  [5539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5080),
  [5541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5082),
  [5543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5083),
  [5545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5084),
  [5547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6123),
  [5549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6124),
  [5551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6125),
  [5553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6126),
  [5555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6127),
  [5557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6128),
  [5559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6129),
  [5561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6130),
  [5563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6131),
  [5565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6132),
  [5567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6150),
  [5569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6151),
  [5571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6162),
  [5573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6163),
  [5575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6164),
  [5577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6165),
  [5579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6166),
  [5581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6167),
  [5583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6168),
  [5585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6169),
  [5587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6115),
  [5589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6117),
  [5591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6119),
  [5593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6134),
  [5595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6135),
  [5597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6137),
  [5599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6138),
  [5601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6139),
  [5603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6142),
  [5605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6144),
  [5607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4876),
  [5609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5014),
  [5611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5015),
  [5613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4975),
  [5615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5085),
  [5617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4877),
  [5619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4878),
  [5621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4885),
  [5623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4886),
  [5625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4891),
  [5627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4868),
  [5629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5038),
  [5631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5039),
  [5633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5020),
  [5635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5021),
  [5637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4869),
  [5639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4870),
  [5641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4871),
  [5643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4872),
  [5645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4873),
  [5647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6152),
  [5649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6153),
  [5651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6154),
  [5653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6155),
  [5655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6156),
  [5657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6157),
  [5659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6158),
  [5661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6159),
  [5663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6160),
  [5665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6161),
  [5667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4906),
  [5669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5067),
  [5671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5068),
  [5673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5069),
  [5675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5070),
  [5677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4923),
  [5679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5076),
  [5681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4916),
  [5683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4907),
  [5685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4922),
  [5687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4894),
  [5689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5046),
  [5691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5047),
  [5693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5024),
  [5695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5025),
  [5697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4895),
  [5699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4896),
  [5701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4897),
  [5703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4898),
  [5705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4899),
  [5707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5967),
  [5709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5968),
  [5711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5969),
  [5713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5970),
  [5715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5971),
  [5717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5972),
  [5719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5973),
  [5721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5974),
  [5723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5975),
  [5725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5976),
  [5727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4992),
  [5729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5022),
  [5731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5023),
  [5733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5003),
  [5735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5004),
  [5737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4993),
  [5739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5000),
  [5741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5052),
  [5743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5065),
  [5745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5073),
  [5747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4924),
  [5749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5050),
  [5751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5051),
  [5753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5028),
  [5755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5029),
  [5757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4925),
  [5759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4926),
  [5761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4927),
  [5763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4928),
  [5765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4929),
  [5767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5993),
  [5769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5994),
  [5771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5995),
  [5773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5996),
  [5775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5997),
  [5777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5998),
  [5779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5999),
  [5781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6000),
  [5783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6001),
  [5785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6002),
  [5787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4938),
  [5789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5054),
  [5791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5055),
  [5793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5032),
  [5795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5033),
  [5797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4939),
  [5799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4940),
  [5801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4941),
  [5803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4942),
  [5805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4943),
  [5807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6015),
  [5809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6016),
  [5811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6017),
  [5813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6018),
  [5815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6019),
  [5817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6020),
  [5819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6021),
  [5821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6022),
  [5823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6023),
  [5825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6024),
  [5827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4952),
  [5829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5056),
  [5831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5057),
  [5833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5036),
  [5835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5037),
  [5837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4953),
  [5839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4954),
  [5841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4955),
  [5843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4956),
  [5845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4957),
  [5847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6028),
  [5849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6029),
  [5851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6030),
  [5853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6031),
  [5855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6032),
  [5857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6033),
  [5859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6034),
  [5861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6035),
  [5863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6036),
  [5865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6037),
  [5867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4966),
  [5869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5058),
  [5871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5059),
  [5873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5040),
  [5875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5041),
  [5877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4967),
  [5879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4968),
  [5881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4969),
  [5883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4970),
  [5885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4971),
  [5887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6050),
  [5889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6051),
  [5891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6052),
  [5893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6053),
  [5895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6054),
  [5897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6055),
  [5899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6056),
  [5901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6057),
  [5903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6058),
  [5905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6059),
  [5907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4980),
  [5909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5060),
  [5911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5061),
  [5913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5044),
  [5915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5045),
  [5917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4981),
  [5919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4982),
  [5921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4983),
  [5923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4984),
  [5925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4985),
  [5927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6073),
  [5929] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6074),
  [5931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6075),
  [5933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6076),
  [5935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6077),
  [5937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6078),
  [5939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6079),
  [5941] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6080),
  [5943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6081),
  [5945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6082),
  [5947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4994),
  [5949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5062),
  [5951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5063),
  [5953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5048),
  [5955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5049),
  [5957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4995),
  [5959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4996),
  [5961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4997),
  [5963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4998),
  [5965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4999),
  [5967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6093),
  [5969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6094),
  [5971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6095),
  [5973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6096),
  [5975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6097),
  [5977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6098),
  [5979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6099),
  [5981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6100),
  [5983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6101),
  [5985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6102),
  [5987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6048),
  [5989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6060),
  [5991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6116),
  [5993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6120),
  [5995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6121),
  [5997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6122),
  [5999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6136),
  [6001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6140),
  [6003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6143),
  [6005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6170),
  [6007] = {.entry = {.count = 1, .reusable = false}}, SHIFT(138),
  [6009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(143),
  [6011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(172),
  [6013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(158),
  [6015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(149),
  [6017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(152),
  [6019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(175),
  [6021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(161),
  [6023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(166),
  [6025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(163),
  [6027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(169),
  [6029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [6031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(146),
  [6033] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__keywords_with_trailing_separator, 3, 0, 0),
  [6035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4973),
  [6037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4974),
  [6039] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__keywords_with_trailing_separator, 2, 0, 0),
  [6041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1141),
  [6043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3881),
  [6045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2013),
  [6047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2654),
  [6049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3348),
  [6051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2694),
  [6053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2645),
  [6055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4094),
  [6057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3885),
  [6059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1295),
  [6061] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_do_block_repeat1, 2, 0, 0), SHIFT_REPEAT(34),
  [6064] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_do_block_repeat1, 2, 0, 0), SHIFT_REPEAT(37),
  [6067] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_do_block_repeat1, 2, 0, 0), SHIFT_REPEAT(40),
  [6070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_do_block_repeat1, 2, 0, 0),
  [6072] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_do_block_repeat1, 2, 0, 0), SHIFT_REPEAT(39),
  [6075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1622),
  [6077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1461),
  [6079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(295),
  [6081] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_body, 3, 0, 0), SHIFT(1021),
  [6084] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_body, 3, 0, 0), SHIFT(261),
  [6087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(286),
  [6089] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_repeat1, 2, 0, 0), SHIFT_REPEAT(1030),
  [6092] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_repeat1, 2, 0, 0), SHIFT_REPEAT(820),
  [6095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(289),
  [6097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(290),
  [6099] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(1030),
  [6102] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(798),
  [6105] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1029),
  [6108] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(138),
  [6111] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0),
  [6113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(368),
  [6115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(679),
  [6117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(680),
  [6119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(404),
  [6121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(637),
  [6123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(638),
  [6125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(414),
  [6127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(485),
  [6129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(486),
  [6131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(438),
  [6133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(658),
  [6135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(659),
  [6137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(369),
  [6139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(858),
  [6141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(859),
  [6143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(396),
  [6145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(823),
  [6147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(824),
  [6149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(392),
  [6151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(701),
  [6153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(702),
  [6155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(360),
  [6157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(592),
  [6159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(593),
  [6161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(356),
  [6163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(481),
  [6165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(484),
  [6167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(391),
  [6169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(802),
  [6171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(803),
  [6173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(455),
  [6175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(722),
  [6177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [6179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(355),
  [6181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(569),
  [6183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(570),
  [6185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(430),
  [6187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(742),
  [6189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(743),
  [6191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(406),
  [6193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(763),
  [6195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(764),
  [6197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(407),
  [6199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(462),
  [6201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(463),
  [6203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(384),
  [6205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(614),
  [6207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(615),
  [6209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(434),
  [6211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(527),
  [6213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(528),
  [6215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(428),
  [6217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(506),
  [6219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(507),
  [6221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(423),
  [6223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(699),
  [6225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [6227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(444),
  [6229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(549),
  [6231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(550),
  [6233] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0),
  [6235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_stab_clause, 2, 0, 16),
  [6237] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_stab_clause, 2, 0, 16),
  [6239] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_stab_clause, 3, 0, 20),
  [6241] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_stab_clause, 3, 0, 20),
  [6243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(924),
  [6245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(425),
  [6247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6114),
  [6249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5874),
  [6251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2713),
  [6253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(367),
  [6255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6175),
  [6257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5424),
  [6259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2999),
  [6261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(410),
  [6263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6178),
  [6265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5672),
  [6267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3000),
  [6269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(436),
  [6271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6180),
  [6273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5768),
  [6275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4341),
  [6277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(373),
  [6279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6146),
  [6281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5380),
  [6283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4342),
  [6285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(429),
  [6287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6172),
  [6289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5382),
  [6291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4343),
  [6293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(401),
  [6295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5964),
  [6297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5385),
  [6299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2507),
  [6301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5314),
  [6303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2508),
  [6305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5322),
  [6307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2511),
  [6309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5324),
  [6311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2512),
  [6313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(359),
  [6315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5985),
  [6317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5326),
  [6319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2513),
  [6321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(371),
  [6323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5990),
  [6325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5328),
  [6327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2514),
  [6329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [6331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5992),
  [6333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5330),
  [6335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1131),
  [6337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5689),
  [6339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1137),
  [6341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5691),
  [6343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1292),
  [6345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5705),
  [6347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3366),
  [6349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5253),
  [6351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3367),
  [6353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5255),
  [6355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3860),
  [6357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5257),
  [6359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3861),
  [6361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5259),
  [6363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1296),
  [6365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5707),
  [6367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1402),
  [6369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5818),
  [6371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1410),
  [6373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5835),
  [6375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1576),
  [6377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5920),
  [6379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3942),
  [6381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5297),
  [6383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3943),
  [6385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5299),
  [6387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3944),
  [6389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5301),
  [6391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3945),
  [6393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5303),
  [6395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3946),
  [6397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5305),
  [6399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3947),
  [6401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5307),
  [6403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1577),
  [6405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5924),
  [6407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4344),
  [6409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5394),
  [6411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3622),
  [6413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5958),
  [6415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3623),
  [6417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5088),
  [6419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3624),
  [6421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5100),
  [6423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2157),
  [6425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5850),
  [6427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2159),
  [6429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5864),
  [6431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2160),
  [6433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5872),
  [6435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2161),
  [6437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5880),
  [6439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2163),
  [6441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5888),
  [6443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2166),
  [6445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5896),
  [6447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3625),
  [6449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5107),
  [6451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3626),
  [6453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5112),
  [6455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2676),
  [6457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5333),
  [6459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2677),
  [6461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5335),
  [6463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3487),
  [6465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5337),
  [6467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3488),
  [6469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5339),
  [6471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3627),
  [6473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5117),
  [6475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4345),
  [6477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5396),
  [6479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4348),
  [6481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5400),
  [6483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2469),
  [6485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5367),
  [6487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2470),
  [6489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5369),
  [6491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2471),
  [6493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5371),
  [6495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2472),
  [6497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5373),
  [6499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2473),
  [6501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5375),
  [6503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2474),
  [6505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5377),
  [6507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2818),
  [6509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5504),
  [6511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2820),
  [6513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5511),
  [6515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2821),
  [6517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5515),
  [6519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2823),
  [6521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5584),
  [6523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2825),
  [6525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5596),
  [6527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2827),
  [6529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5649),
  [6531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3831),
  [6533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5397),
  [6535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3857),
  [6537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5687),
  [6539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1630),
  [6541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5404),
  [6543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1631),
  [6545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5414),
  [6547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1928),
  [6549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5091),
  [6551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1929),
  [6553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5093),
  [6555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2177),
  [6557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5095),
  [6559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2178),
  [6561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5097),
  [6563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1894),
  [6565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5420),
  [6567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1895),
  [6569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5422),
  [6571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3855),
  [6573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5628),
  [6575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(958),
  [6577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5405),
  [6579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(947),
  [6581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5407),
  [6583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3061),
  [6585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5409),
  [6587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3062),
  [6589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5411),
  [6591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3542),
  [6593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5101),
  [6595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3858),
  [6597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5737),
  [6599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3833),
  [6601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5460),
  [6603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1691),
  [6605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5426),
  [6607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1692),
  [6609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5428),
  [6611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1693),
  [6613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5430),
  [6615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1695),
  [6617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5432),
  [6619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1800),
  [6621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5434),
  [6623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1701),
  [6625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5436),
  [6627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2122),
  [6629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5134),
  [6631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3109),
  [6633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5449),
  [6635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3110),
  [6637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5451),
  [6639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1415),
  [6641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5453),
  [6643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1416),
  [6645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5455),
  [6647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2126),
  [6649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5136),
  [6651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2127),
  [6653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5138),
  [6655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2128),
  [6657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5140),
  [6659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2748),
  [6661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5471),
  [6663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2749),
  [6665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5473),
  [6667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2751),
  [6669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5475),
  [6671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2752),
  [6673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5477),
  [6675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2753),
  [6677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5479),
  [6679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2754),
  [6681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5481),
  [6683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2129),
  [6685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5142),
  [6687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1230),
  [6689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5495),
  [6691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1231),
  [6693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5497),
  [6695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2861),
  [6697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5499),
  [6699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2862),
  [6701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5501),
  [6703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2130),
  [6705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5144),
  [6707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4421),
  [6709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5169),
  [6711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4422),
  [6713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5171),
  [6715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2380),
  [6717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5517),
  [6719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2381),
  [6721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5519),
  [6723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2390),
  [6725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5521),
  [6727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2391),
  [6729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5523),
  [6731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2392),
  [6733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5525),
  [6735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2393),
  [6737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5527),
  [6739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4545),
  [6741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5173),
  [6743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(990),
  [6745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5541),
  [6747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(991),
  [6749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5543),
  [6751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4423),
  [6753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5545),
  [6755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4424),
  [6757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5547),
  [6759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3535),
  [6761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5602),
  [6763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3536),
  [6765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5630),
  [6767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4137),
  [6769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5640),
  [6771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1807),
  [6773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5563),
  [6775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1808),
  [6777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5565),
  [6779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1809),
  [6781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5567),
  [6783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1810),
  [6785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5569),
  [6787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1811),
  [6789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5571),
  [6791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1812),
  [6793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5573),
  [6795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4140),
  [6797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5646),
  [6799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6991),
  [6801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5587),
  [6803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6979),
  [6805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5589),
  [6807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3522),
  [6809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5591),
  [6811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2711),
  [6813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5342),
  [6815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2671),
  [6817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5198),
  [6819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2672),
  [6821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5201),
  [6823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3216),
  [6825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5266),
  [6827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1339),
  [6829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5608),
  [6831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1341),
  [6833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5610),
  [6835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1343),
  [6837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5612),
  [6839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1344),
  [6841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5614),
  [6843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1348),
  [6845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5616),
  [6847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1143),
  [6849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5618),
  [6851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3219),
  [6853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5268),
  [6855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2696),
  [6857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5631),
  [6859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2697),
  [6861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5633),
  [6863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2698),
  [6865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5635),
  [6867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2699),
  [6869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5637),
  [6871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3475),
  [6873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5203),
  [6875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3823),
  [6877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5538),
  [6879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3824),
  [6881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5549),
  [6883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1504),
  [6885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5652),
  [6887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1505),
  [6889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5654),
  [6891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1506),
  [6893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5656),
  [6895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1507),
  [6897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5658),
  [6899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1509),
  [6901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5660),
  [6903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1510),
  [6905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5662),
  [6907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3825),
  [6909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5552),
  [6911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(932),
  [6913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5675),
  [6915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(933),
  [6917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5677),
  [6919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3573),
  [6921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5679),
  [6923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3574),
  [6925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5681),
  [6927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3221),
  [6929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5271),
  [6931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4260),
  [6933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5692),
  [6935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4261),
  [6937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5694),
  [6939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2195),
  [6941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5696),
  [6943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2227),
  [6945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5698),
  [6947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1003),
  [6949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5708),
  [6951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1006),
  [6953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5710),
  [6955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1858),
  [6957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5712),
  [6959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1859),
  [6961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5714),
  [6963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3587),
  [6965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5724),
  [6967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3588),
  [6969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5726),
  [6971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3241),
  [6973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5728),
  [6975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3242),
  [6977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5730),
  [6979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2901),
  [6981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5740),
  [6983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2903),
  [6985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5742),
  [6987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3599),
  [6989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5744),
  [6991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3600),
  [6993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5746),
  [6995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3608),
  [6997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5756),
  [6999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3609),
  [7001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5758),
  [7003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2911),
  [7005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5760),
  [7007] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2913),
  [7009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5762),
  [7011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2342),
  [7013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5772),
  [7015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2343),
  [7017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5774),
  [7019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1624),
  [7021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5776),
  [7023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1625),
  [7025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5778),
  [7027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1871),
  [7029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5788),
  [7031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1872),
  [7033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5790),
  [7035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2374),
  [7037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5792),
  [7039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2375),
  [7041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5794),
  [7043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3249),
  [7045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5804),
  [7047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3250),
  [7049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5806),
  [7051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2384),
  [7053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5808),
  [7055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2385),
  [7057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5810),
  [7059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3637),
  [7061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5820),
  [7063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3638),
  [7065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5822),
  [7067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1877),
  [7069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5824),
  [7071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1878),
  [7073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5826),
  [7075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(961),
  [7077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5836),
  [7079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(962),
  [7081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5838),
  [7083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1210),
  [7085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5840),
  [7087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1211),
  [7089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5842),
  [7091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2408),
  [7093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5852),
  [7095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2409),
  [7097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5854),
  [7099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1456),
  [7101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5856),
  [7103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1468),
  [7105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5858),
  [7107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1633),
  [7109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5866),
  [7111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1634),
  [7113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5868),
  [7115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3826),
  [7117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5558),
  [7119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(925),
  [7121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5876),
  [7123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2432),
  [7125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5882),
  [7127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2433),
  [7129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5884),
  [7131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2438),
  [7133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5890),
  [7135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2439),
  [7137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5892),
  [7139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1896),
  [7141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5898),
  [7143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1897),
  [7145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5900),
  [7147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1222),
  [7149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5906),
  [7151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1223),
  [7153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5908),
  [7155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1495),
  [7157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5914),
  [7159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1496),
  [7161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5916),
  [7163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3222),
  [7165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5286),
  [7167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3827),
  [7169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5560),
  [7171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3223),
  [7173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5291),
  [7175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3669),
  [7177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5177),
  [7179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3670),
  [7181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5179),
  [7183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3671),
  [7185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5181),
  [7187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3672),
  [7189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5183),
  [7191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3224),
  [7193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5310),
  [7195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3832),
  [7197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5930),
  [7199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3828),
  [7201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5562),
  [7203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3476),
  [7205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5208),
  [7207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4546),
  [7209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5175),
  [7211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3849),
  [7213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5583),
  [7215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3293),
  [7217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5215),
  [7219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4232),
  [7221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5940),
  [7223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3294),
  [7225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5217),
  [7227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3295),
  [7229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5219),
  [7231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4233),
  [7233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5947),
  [7235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3296),
  [7237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5221),
  [7239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3297),
  [7241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5223),
  [7243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3298),
  [7245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5225),
  [7247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3523),
  [7249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5593),
  [7251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3083),
  [7253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3688),
  [7255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3689),
  [7257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(139),
  [7259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(137),
  [7261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2081),
  [7263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2083),
  [7265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2268),
  [7267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2269),
  [7269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3690),
  [7271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3547),
  [7273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3691),
  [7275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2120),
  [7277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3692),
  [7279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2121),
  [7281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2333),
  [7283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2334),
  [7285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3697),
  [7287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4241),
  [7289] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_repeat1, 2, 0, 0), SHIFT_REPEAT(847),
  [7292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3787),
  [7294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(433),
  [7296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2293),
  [7298] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_square_repeat1, 2, 0, 0),
  [7300] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_square_repeat1, 2, 0, 0), SHIFT_REPEAT(401),
  [7303] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_square_repeat1, 2, 0, 0), SHIFT_REPEAT(5964),
  [7306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1433),
  [7308] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_repeat1, 2, 0, 0), SHIFT_REPEAT(461),
  [7311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(383),
  [7313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1973),
  [7315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1974),
  [7317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1975),
  [7319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2140),
  [7321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1978),
  [7323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1979),
  [7325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2353),
  [7327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1435),
  [7329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3863),
  [7331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1710),
  [7333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3864),
  [7335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3865),
  [7337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1711),
  [7339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3866),
  [7341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1992),
  [7343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1993),
  [7345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1995),
  [7347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1996),
  [7349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1997),
  [7351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1998),
  [7353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3867),
  [7355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3868),
  [7357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2814),
  [7359] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_heredoc_single_repeat1, 2, 0, 0),
  [7361] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_heredoc_single_repeat1, 2, 0, 0), SHIFT_REPEAT(410),
  [7364] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_heredoc_single_repeat1, 2, 0, 0), SHIFT_REPEAT(6178),
  [7367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2815),
  [7369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3095),
  [7371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4166),
  [7373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4167),
  [7375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4454),
  [7377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4462),
  [7379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3709),
  [7381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3710),
  [7383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3711),
  [7385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3712),
  [7387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3096),
  [7389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1742),
  [7391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4703),
  [7393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3750),
  [7395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3751),
  [7397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3752),
  [7399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3753),
  [7401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2720),
  [7403] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(526),
  [7406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2746),
  [7408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2938),
  [7410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3804),
  [7412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2940),
  [7414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(405),
  [7416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3314),
  [7418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3315),
  [7420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3316),
  [7422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3317),
  [7424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3318),
  [7426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3319),
  [7428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3821),
  [7430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4209),
  [7432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3330),
  [7434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3331),
  [7436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3332),
  [7438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3333),
  [7440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3334),
  [7442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3335),
  [7444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2504),
  [7446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2505),
  [7448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3018),
  [7450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3019),
  [7452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4327),
  [7454] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_body, 3, 0, 0), SHIFT(1036),
  [7457] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_body, 3, 0, 0), SHIFT(353),
  [7460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4328),
  [7462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4526),
  [7464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(386),
  [7466] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3389),
  [7468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3390),
  [7470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3898),
  [7472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3899),
  [7474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4528),
  [7476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3785),
  [7478] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4829),
  [7481] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4835),
  [7484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3382),
  [7486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3383),
  [7488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3400),
  [7490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_parenthesis_repeat1, 2, 0, 0),
  [7492] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_parenthesis_repeat1, 2, 0, 0), SHIFT_REPEAT(373),
  [7495] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_parenthesis_repeat1, 2, 0, 0), SHIFT_REPEAT(6146),
  [7498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3410),
  [7500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3411),
  [7502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3940),
  [7504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3941),
  [7506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3043),
  [7508] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(137),
  [7511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3953),
  [7513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3146),
  [7515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3401),
  [7517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3435),
  [7519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3452),
  [7521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4476),
  [7523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(424),
  [7525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3988),
  [7527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3989),
  [7529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3990),
  [7531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3991),
  [7533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3992),
  [7535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3993),
  [7537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3448),
  [7539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3455),
  [7541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(375),
  [7543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2591),
  [7545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4006),
  [7547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4007),
  [7549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4008),
  [7551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3509),
  [7553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4010),
  [7555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4011),
  [7557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2592),
  [7559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2593),
  [7561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2594),
  [7563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2595),
  [7565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2596),
  [7567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2848),
  [7569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2849),
  [7571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2944),
  [7573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2945),
  [7575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3064),
  [7577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2759),
  [7579] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_angle_repeat1, 2, 0, 0),
  [7581] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_angle_repeat1, 2, 0, 0), SHIFT_REPEAT(359),
  [7584] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_angle_repeat1, 2, 0, 0), SHIFT_REPEAT(5985),
  [7587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2467),
  [7589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2468),
  [7591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2996),
  [7593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2997),
  [7595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2610),
  [7597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2611),
  [7599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2612),
  [7601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2613),
  [7603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1412),
  [7605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2615),
  [7607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1761),
  [7609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2616),
  [7611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(378),
  [7613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(442),
  [7615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2521),
  [7617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2522),
  [7619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2523),
  [7621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2524),
  [7623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2525),
  [7625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2526),
  [7627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1384),
  [7629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4401),
  [7631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4402),
  [7633] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(536),
  [7636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4403),
  [7638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2542),
  [7640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2544),
  [7642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2545),
  [7644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2546),
  [7646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2547),
  [7648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2548),
  [7650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4285),
  [7652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4404),
  [7654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4407),
  [7656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3579),
  [7658] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_body, 3, 0, 0), SHIFT(1032),
  [7661] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_body, 3, 0, 0), SHIFT(346),
  [7664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4408),
  [7666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4559),
  [7668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1582),
  [7670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(945),
  [7672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(959),
  [7674] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3070),
  [7676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3071),
  [7678] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1583),
  [7680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(960),
  [7682] = {.entry = {.count = 1, .reusable = false}}, SHIFT(989),
  [7684] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3082),
  [7686] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_single_repeat1, 2, 0, 0),
  [7688] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_single_repeat1, 2, 0, 0), SHIFT_REPEAT(367),
  [7691] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_single_repeat1, 2, 0, 0), SHIFT_REPEAT(6175),
  [7694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2027),
  [7696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2028),
  [7698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2761),
  [7700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3980),
  [7702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1732),
  [7704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1735),
  [7706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1739),
  [7708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1744),
  [7710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1745),
  [7712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1746),
  [7714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1765),
  [7716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1767),
  [7718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1768),
  [7720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1769),
  [7722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1770),
  [7724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1771),
  [7726] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_bar_repeat1, 2, 0, 0),
  [7728] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_bar_repeat1, 2, 0, 0), SHIFT_REPEAT(371),
  [7731] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_bar_repeat1, 2, 0, 0), SHIFT_REPEAT(5990),
  [7734] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4428),
  [7736] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3111),
  [7738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3112),
  [7740] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1420),
  [7742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1428),
  [7744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4429),
  [7746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3486),
  [7748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3580),
  [7750] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3130),
  [7752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3131),
  [7754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1431),
  [7756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1446),
  [7758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4430),
  [7760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1593),
  [7762] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4061),
  [7764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1594),
  [7766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2792),
  [7768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2793),
  [7770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2794),
  [7772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2795),
  [7774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2796),
  [7776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2797),
  [7778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2811),
  [7780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2812),
  [7782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2813),
  [7784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2816),
  [7786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2817),
  [7788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2819),
  [7790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1803),
  [7792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1804),
  [7794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4064),
  [7796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4080),
  [7798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(426),
  [7800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3489),
  [7802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1254),
  [7804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1256),
  [7806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2888),
  [7808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2889),
  [7810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2854),
  [7812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3490),
  [7814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1264),
  [7816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1267),
  [7818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2376),
  [7820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2377),
  [7822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2856),
  [7824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1838),
  [7826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2857),
  [7828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2404),
  [7830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2405),
  [7832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2406),
  [7834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2407),
  [7836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2410),
  [7838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2411),
  [7840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2426),
  [7842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2427),
  [7844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2428),
  [7846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2429),
  [7848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2430),
  [7850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2431),
  [7852] = {.entry = {.count = 1, .reusable = false}}, SHIFT(427),
  [7854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3887),
  [7856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3491),
  [7858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(992),
  [7860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(993),
  [7862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4128),
  [7864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4163),
  [7866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3888),
  [7868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3492),
  [7870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3889),
  [7872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(994),
  [7874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(995),
  [7876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4210),
  [7878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4213),
  [7880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3890),
  [7882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3891),
  [7884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3892),
  [7886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1855),
  [7888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1856),
  [7890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1857),
  [7892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1866),
  [7894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1867),
  [7896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1801),
  [7898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1892),
  [7900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1893),
  [7902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1909),
  [7904] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__keywords_with_trailing_separator, 1, 0, 0),
  [7906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4704),
  [7908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1914),
  [7910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1915),
  [7912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1851),
  [7914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3581),
  [7916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2871),
  [7918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3494),
  [7920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6919),
  [7922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6924),
  [7924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3524),
  [7926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3525),
  [7928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4835),
  [7930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2883),
  [7932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6968),
  [7934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6961),
  [7936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3526),
  [7938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3527),
  [7940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3551),
  [7942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3929),
  [7944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3930),
  [7946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3931),
  [7948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3932),
  [7950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1154),
  [7952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1155),
  [7954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1156),
  [7956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1157),
  [7958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1158),
  [7960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1159),
  [7962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1169),
  [7964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1170),
  [7966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1171),
  [7968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1172),
  [7970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1173),
  [7972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1174),
  [7974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3933),
  [7976] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3934),
  [7978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3582),
  [7980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3552),
  [7982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2721),
  [7984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2722),
  [7986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2723),
  [7988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2725),
  [7990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4197),
  [7992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2783),
  [7994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2787),
  [7996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2789),
  [7998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2790),
  [8000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4198),
  [8002] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2896),
  [8004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1565),
  [8006] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1570),
  [8008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1571),
  [8010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1572),
  [8012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1472),
  [8014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1473),
  [8016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1365),
  [8018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1366),
  [8020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1367),
  [8022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1368),
  [8024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1369),
  [8026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1370),
  [8028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3157),
  [8030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3066),
  [8032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(934),
  [8034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(935),
  [8036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3575),
  [8038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3576),
  [8040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(936),
  [8042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(937),
  [8044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3577),
  [8046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3578),
  [8048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3583),
  [8050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1114),
  [8052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1098),
  [8054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4264),
  [8056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4265),
  [8058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2235),
  [8060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2236),
  [8062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4268),
  [8064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4269),
  [8066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2237),
  [8068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2239),
  [8070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1195),
  [8072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1196),
  [8074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1014),
  [8076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1010),
  [8078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1861),
  [8080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1862),
  [8082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1016),
  [8084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(997),
  [8086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1864),
  [8088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1865),
  [8090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3981),
  [8092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3591),
  [8094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3592),
  [8096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3243),
  [8098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3244),
  [8100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3595),
  [8102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3596),
  [8104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3245),
  [8106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3246),
  [8108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3584),
  [8110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2905),
  [8112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2906),
  [8114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3603),
  [8116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3604),
  [8118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2907),
  [8120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2909),
  [8122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3606),
  [8124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3607),
  [8126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1118),
  [8128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3611),
  [8130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3612),
  [8132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2914),
  [8134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2915),
  [8136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3613),
  [8138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3614),
  [8140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2916),
  [8142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2920),
  [8144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3067),
  [8146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1122),
  [8148] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1232),
  [8150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1233),
  [8152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2346),
  [8154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2351),
  [8156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1626),
  [8158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1627),
  [8160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2361),
  [8162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2260),
  [8164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1628),
  [8166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1629),
  [8168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3618),
  [8170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3619),
  [8172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1873),
  [8174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1874),
  [8176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2378),
  [8178] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2379),
  [8180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1875),
  [8182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1876),
  [8184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2382),
  [8186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2383),
  [8188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4321),
  [8190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4322),
  [8192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1283),
  [8194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3251),
  [8196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3252),
  [8198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2386),
  [8200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2387),
  [8202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3253),
  [8204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3254),
  [8206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2388),
  [8208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2389),
  [8210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_heredoc_double_repeat1, 2, 0, 0),
  [8212] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_heredoc_double_repeat1, 2, 0, 0), SHIFT_REPEAT(436),
  [8215] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_heredoc_double_repeat1, 2, 0, 0), SHIFT_REPEAT(6180),
  [8218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1407),
  [8220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3641),
  [8222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3642),
  [8224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1879),
  [8226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1880),
  [8228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3645),
  [8230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3646),
  [8232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1881),
  [8234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1882),
  [8236] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_slash_repeat1, 2, 0, 0),
  [8238] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_slash_repeat1, 2, 0, 0), SHIFT_REPEAT(393),
  [8241] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_slash_repeat1, 2, 0, 0), SHIFT_REPEAT(5992),
  [8244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2475),
  [8246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1408),
  [8248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(963),
  [8250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(964),
  [8252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1212),
  [8254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1213),
  [8256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(965),
  [8258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(966),
  [8260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1215),
  [8262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1216),
  [8264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(361),
  [8266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2197),
  [8268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2412),
  [8270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2413),
  [8272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1469),
  [8274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1470),
  [8276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2416),
  [8278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2417),
  [8280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1353),
  [8282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1474),
  [8284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2200),
  [8286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1635),
  [8288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1636),
  [8290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1637),
  [8292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1638),
  [8294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2219),
  [8296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(926),
  [8298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(910),
  [8300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(911),
  [8302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(912),
  [8304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2221),
  [8306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2434),
  [8308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2435),
  [8310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2436),
  [8312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2437),
  [8314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2222),
  [8316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2440),
  [8318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2441),
  [8320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2442),
  [8322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2443),
  [8324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2223),
  [8326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1898),
  [8328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1899),
  [8330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1900),
  [8332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1901),
  [8334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1297),
  [8336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1224),
  [8338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1225),
  [8340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1226),
  [8342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1227),
  [8344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2476),
  [8346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2477),
  [8348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1497),
  [8350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1498),
  [8352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1499),
  [8354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1501),
  [8356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1676),
  [8358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2480),
  [8360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2481),
  [8362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1677),
  [8364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2289),
  [8366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2290),
  [8368] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_repeat1, 2, 0, 0), SHIFT_REPEAT(888),
  [8371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3548),
  [8373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(387),
  [8375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2295),
  [8377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2296),
  [8379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2297),
  [8381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2302),
  [8383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4383),
  [8385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4176),
  [8387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2484),
  [8389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4208),
  [8391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3786),
  [8393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4177),
  [8395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4406),
  [8397] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_anonymous_function_repeat1, 2, 0, 0), SHIFT_REPEAT(1029),
  [8400] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_anonymous_function_repeat1, 2, 0, 0), SHIFT_REPEAT(139),
  [8403] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_anonymous_function_repeat1, 2, 0, 0),
  [8405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3749),
  [8407] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_curly_repeat1, 2, 0, 0),
  [8409] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_curly_repeat1, 2, 0, 0), SHIFT_REPEAT(429),
  [8412] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_curly_repeat1, 2, 0, 0), SHIFT_REPEAT(6172),
  [8415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(377),
  [8417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_double_repeat1, 2, 0, 0),
  [8419] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_double_repeat1, 2, 0, 0), SHIFT_REPEAT(425),
  [8422] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_i_double_repeat1, 2, 0, 0), SHIFT_REPEAT(6114),
  [8425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3829),
  [8427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1910),
  [8429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpolation, 3, 0, 0),
  [8431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolation, 3, 0, 0),
  [8433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3632),
  [8435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6573),
  [8437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6566),
  [8439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_square_repeat1, 1, 0, 0),
  [8441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6410),
  [8443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3635),
  [8445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6645),
  [8447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6612),
  [8449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3636),
  [8451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6540),
  [8453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6634),
  [8455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2485),
  [8457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6338),
  [8459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2486),
  [8461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6348),
  [8463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2488),
  [8465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6350),
  [8467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2489),
  [8469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6715),
  [8471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6352),
  [8473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2490),
  [8475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6719),
  [8477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6354),
  [8479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2491),
  [8481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6722),
  [8483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6356),
  [8485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2492),
  [8487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6725),
  [8489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6358),
  [8491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2493),
  [8493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6729),
  [8495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6390),
  [8497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2494),
  [8499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6736),
  [8501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6392),
  [8503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2495),
  [8505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6739),
  [8507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6394),
  [8509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3639),
  [8511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6636),
  [8513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3640),
  [8515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6640),
  [8517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3643),
  [8519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6647),
  [8521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3644),
  [8523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6191),
  [8525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3647),
  [8527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6194),
  [8529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3648),
  [8531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6196),
  [8533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3649),
  [8535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6198),
  [8537] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpolation, 2, 0, 0),
  [8539] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolation, 2, 0, 0),
  [8541] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_angle_repeat1, 1, 0, 0),
  [8543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6660),
  [8545] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_bar_repeat1, 1, 0, 0),
  [8547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6409),
  [8549] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_slash_repeat1, 1, 0, 0),
  [8551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6563),
  [8553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1704),
  [8555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6666),
  [8557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1706),
  [8559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6668),
  [8561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1707),
  [8563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6670),
  [8565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1708),
  [8567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6672),
  [8569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1709),
  [8571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6674),
  [8573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1712),
  [8575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6676),
  [8577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1713),
  [8579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6678),
  [8581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1717),
  [8583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6680),
  [8585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1721),
  [8587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6682),
  [8589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1725),
  [8591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6684),
  [8593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2282),
  [8595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6216),
  [8597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2283),
  [8599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6218),
  [8601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2286),
  [8603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6220),
  [8605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2287),
  [8607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6222),
  [8609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2344),
  [8611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6224),
  [8613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2345),
  [8615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6246),
  [8617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2143),
  [8619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6248),
  [8621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2151),
  [8623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6250),
  [8625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2152),
  [8627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6262),
  [8629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2154),
  [8631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6264),
  [8633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(199),
  [8635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2756),
  [8637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6226),
  [8639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2757),
  [8641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6228),
  [8643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2758),
  [8645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6230),
  [8647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2760),
  [8649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6232),
  [8651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2763),
  [8653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6234),
  [8655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2765),
  [8657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6236),
  [8659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2768),
  [8661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6238),
  [8663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2772),
  [8665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6240),
  [8667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2782),
  [8669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6242),
  [8671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2791),
  [8673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6244),
  [8675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4352),
  [8677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6473),
  [8679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4353),
  [8681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6475),
  [8683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4358),
  [8685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6477),
  [8687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2394),
  [8689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6306),
  [8691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2395),
  [8693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6308),
  [8695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2396),
  [8697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6310),
  [8699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2397),
  [8701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6312),
  [8703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2398),
  [8705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6314),
  [8707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2399),
  [8709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6316),
  [8711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2400),
  [8713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6318),
  [8715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2401),
  [8717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6320),
  [8719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2402),
  [8721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6322),
  [8723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2403),
  [8725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6324),
  [8727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4361),
  [8729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6480),
  [8731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4362),
  [8733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6484),
  [8735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4364),
  [8737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6486),
  [8739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4365),
  [8741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6517),
  [8743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4366),
  [8745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6519),
  [8747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4367),
  [8749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6531),
  [8751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4369),
  [8753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6533),
  [8755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4004),
  [8757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6553),
  [8759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1813),
  [8761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6359),
  [8763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1814),
  [8765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6361),
  [8767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1832),
  [8769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6363),
  [8771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1833),
  [8773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6365),
  [8775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1840),
  [8777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6367),
  [8779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1841),
  [8781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6369),
  [8783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1842),
  [8785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6371),
  [8787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1846),
  [8789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6373),
  [8791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1852),
  [8793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6375),
  [8795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1854),
  [8797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6377),
  [8799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4005),
  [8801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6615),
  [8803] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__stab_clause_arguments_without_parentheses_repeat1, 2, 0, 0), SHIFT_REPEAT(531),
  [8806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1936),
  [8808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6421),
  [8810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1937),
  [8812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6443),
  [8814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1938),
  [8816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6445),
  [8818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1939),
  [8820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6447),
  [8822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1940),
  [8824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6459),
  [8826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1941),
  [8828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6461),
  [8830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1942),
  [8832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6464),
  [8834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1943),
  [8836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6466),
  [8838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1944),
  [8840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6468),
  [8842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1945),
  [8844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6470),
  [8846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1144),
  [8848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6422),
  [8850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1145),
  [8852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6424),
  [8854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1146),
  [8856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6426),
  [8858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1147),
  [8860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6428),
  [8862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1148),
  [8864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6430),
  [8866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1149),
  [8868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6432),
  [8870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1150),
  [8872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6434),
  [8874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1151),
  [8876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6436),
  [8878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1152),
  [8880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6438),
  [8882] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1153),
  [8884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6440),
  [8886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2515),
  [8888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6536),
  [8890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2516),
  [8892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6538),
  [8894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2518),
  [8896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6544),
  [8898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2520),
  [8900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6546),
  [8902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2538),
  [8904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6548),
  [8906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2539),
  [8908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6550),
  [8910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2540),
  [8912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6552),
  [8914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2541),
  [8916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6555),
  [8918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2559),
  [8920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6557),
  [8922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1513),
  [8924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6495),
  [8926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1524),
  [8928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6497),
  [8930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1543),
  [8932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6499),
  [8934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1544),
  [8936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6501),
  [8938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1545),
  [8940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6503),
  [8942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1546),
  [8944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6505),
  [8946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1547),
  [8948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6507),
  [8950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1549),
  [8952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6186),
  [8954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1553),
  [8956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6510),
  [8958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1561),
  [8960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6512),
  [8962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3228),
  [8964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6732),
  [8966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3229),
  [8968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6760),
  [8970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3230),
  [8972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6741),
  [8974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3231),
  [8976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6206),
  [8978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3232),
  [8980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6208),
  [8982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3233),
  [8984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6210),
  [8986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3234),
  [8988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6559),
  [8990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3235),
  [8992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6568),
  [8994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3236),
  [8996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6571),
  [8998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3237),
  [9000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6586),
  [9002] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_double_repeat1, 1, 0, 0),
  [9004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6761),
  [9006] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2828),
  [9008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6214),
  [9010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4009),
  [9012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6408),
  [9014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2830),
  [9016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6266),
  [9018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2831),
  [9020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6278),
  [9022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4025),
  [9024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6192),
  [9026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4041),
  [9028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6718),
  [9030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3510),
  [9032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6721),
  [9034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3301),
  [9036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6592),
  [9038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3302),
  [9040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6594),
  [9042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3303),
  [9044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6596),
  [9046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3304),
  [9048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6598),
  [9050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3305),
  [9052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6600),
  [9054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3306),
  [9056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6602),
  [9058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3307),
  [9060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6604),
  [9062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3308),
  [9064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6606),
  [9066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3309),
  [9068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6608),
  [9070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3310),
  [9072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6610),
  [9074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2832),
  [9076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6341),
  [9078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2833),
  [9080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6343),
  [9082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3511),
  [9084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6724),
  [9086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2834),
  [9088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6346),
  [9090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2835),
  [9092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6398),
  [9094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2836),
  [9096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6401),
  [9098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3512),
  [9100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6727),
  [9102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2837),
  [9104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6404),
  [9106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3513),
  [9108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6734),
  [9110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2840),
  [9112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6406),
  [9114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_parenthesis_repeat1, 1, 0, 0),
  [9116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6539),
  [9118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3834),
  [9120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6502),
  [9122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3835),
  [9124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6189),
  [9126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3958),
  [9128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6687),
  [9130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3959),
  [9132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6689),
  [9134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3960),
  [9136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6691),
  [9138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3961),
  [9140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6693),
  [9142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3962),
  [9144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6695),
  [9146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3963),
  [9148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6697),
  [9150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3964),
  [9152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6702),
  [9154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3965),
  [9156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6705),
  [9158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3966),
  [9160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6712),
  [9162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3967),
  [9164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6714),
  [9166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3836),
  [9168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6202),
  [9170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3838),
  [9172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6279),
  [9174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3839),
  [9176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6288),
  [9178] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3840),
  [9180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6291),
  [9182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3845),
  [9184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6326),
  [9186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3846),
  [9188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6411),
  [9190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3847),
  [9192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6565),
  [9194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3848),
  [9196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6587),
  [9198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3514),
  [9200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6738),
  [9202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_curly_repeat1, 1, 0, 0),
  [9204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6662),
  [9206] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_single_repeat1, 1, 0, 0),
  [9208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6767),
  [9210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(317),
  [9212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_without_parentheses, 2, 0, 0),
  [9214] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_heredoc_single_repeat1, 1, 0, 0),
  [9216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6201),
  [9218] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_i_heredoc_double_repeat1, 1, 0, 0),
  [9220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6204),
  [9222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2517),
  [9224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6542),
  [9226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1362),
  [9228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1782),
  [9230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1783),
  [9232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3894),
  [9234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3850),
  [9236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3602),
  [9238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3852),
  [9240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3853),
  [9242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3854),
  [9244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3777),
  [9246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3896),
  [9248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3459),
  [9250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6776),
  [9252] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_without_parentheses, 3, 0, 0),
  [9254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3462),
  [9256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3464),
  [9258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3465),
  [9260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3474),
  [9262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2898),
  [9264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2224),
  [9266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2225),
  [9268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2228),
  [9270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2229),
  [9272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2230),
  [9274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2798),
  [9276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2799),
  [9278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2800),
  [9280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2801),
  [9282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2802),
  [9284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2804),
  [9286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2805),
  [9288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2807),
  [9290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2809),
  [9292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2810),
  [9294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2232),
  [9296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2250),
  [9298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2254),
  [9300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2822),
  [9302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2824),
  [9304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2826),
  [9306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2829),
  [9308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2838),
  [9310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2839),
  [9312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2841),
  [9314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2842),
  [9316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2844),
  [9318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2845),
  [9320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2265),
  [9322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2270),
  [9324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2900),
  [9326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3869),
  [9328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3870),
  [9330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3871),
  [9332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3872),
  [9334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3873),
  [9336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3874),
  [9338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3875),
  [9340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3876),
  [9342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3877),
  [9344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3878),
  [9346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2902),
  [9348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3897),
  [9350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2303),
  [9352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2316),
  [9354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2317),
  [9356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2323),
  [9358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2325),
  [9360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2326),
  [9362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2327),
  [9364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3900),
  [9366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3586),
  [9368] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3901),
  [9370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3935),
  [9372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3936),
  [9374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3937),
  [9376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3938),
  [9378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2331),
  [9380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2332),
  [9382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2338),
  [9384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3939),
  [9386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3948),
  [9388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3949),
  [9390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3950),
  [9392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3951),
  [9394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3979),
  [9396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2414),
  [9398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2415),
  [9400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2418),
  [9402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2419),
  [9404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2420),
  [9406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2421),
  [9408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2422),
  [9410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2423),
  [9412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2424),
  [9414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2425),
  [9416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3922),
  [9418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2444),
  [9420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2445),
  [9422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2446),
  [9424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2447),
  [9426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2448),
  [9428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2449),
  [9430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2450),
  [9432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2451),
  [9434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2452),
  [9436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2453),
  [9438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2528),
  [9440] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_heredoc_double_repeat1, 2, 0, 0),
  [9442] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_heredoc_double_repeat1, 2, 0, 0), SHIFT_REPEAT(6719),
  [9445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2904),
  [9447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2364),
  [9449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3926),
  [9451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2908),
  [9453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2529),
  [9455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2530),
  [9457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2531),
  [9459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2532),
  [9461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2533),
  [9463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2534),
  [9465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1868),
  [9467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1869),
  [9469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1870),
  [9471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1885),
  [9473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1886),
  [9475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1887),
  [9477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1888),
  [9479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1889),
  [9481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1890),
  [9483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1891),
  [9485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1916),
  [9487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1918),
  [9489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1919),
  [9491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1920),
  [9493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1921),
  [9495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1922),
  [9497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1923),
  [9499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1924),
  [9501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1925),
  [9503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1927),
  [9505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2535),
  [9507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2536),
  [9509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2537),
  [9511] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_curly_repeat1, 2, 0, 0),
  [9513] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_curly_repeat1, 2, 0, 0), SHIFT_REPEAT(6722),
  [9516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2910),
  [9518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3496),
  [9520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2912),
  [9522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4829),
  [9524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2680),
  [9526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2700),
  [9528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3597),
  [9530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3923),
  [9532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2549),
  [9534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2550),
  [9536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2551),
  [9538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2552),
  [9540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2553),
  [9542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2554),
  [9544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2555),
  [9546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2556),
  [9548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1981),
  [9550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1160),
  [9552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1161),
  [9554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1162),
  [9556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1163),
  [9558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1164),
  [9560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1165),
  [9562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1352),
  [9564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1166),
  [9566] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1167),
  [9568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1168),
  [9570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1982),
  [9572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1983),
  [9574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1984),
  [9576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1175),
  [9578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1176),
  [9580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1177),
  [9582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1178),
  [9584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1179),
  [9586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1180),
  [9588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1181),
  [9590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1182),
  [9592] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1183),
  [9594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1184),
  [9596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1985),
  [9598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1986),
  [9600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3497),
  [9602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1987),
  [9604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1988),
  [9606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1989),
  [9608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1990),
  [9610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3498),
  [9612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4409),
  [9614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4410),
  [9616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4411),
  [9618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(262),
  [9620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4412),
  [9622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2557),
  [9624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2558),
  [9626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4413),
  [9628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4414),
  [9630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1999),
  [9632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2000),
  [9634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2001),
  [9636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2002),
  [9638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2004),
  [9640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2005),
  [9642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2007),
  [9644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2008),
  [9646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1355),
  [9648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1356),
  [9650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1357),
  [9652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1358),
  [9654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3893),
  [9656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1359),
  [9658] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1360),
  [9660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1361),
  [9662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1363),
  [9664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1364),
  [9666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2009),
  [9668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2010),
  [9670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4417),
  [9672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4418),
  [9674] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1371),
  [9676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1372),
  [9678] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1373),
  [9680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1374),
  [9682] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1375),
  [9684] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1376),
  [9686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1377),
  [9688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1378),
  [9690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1379),
  [9692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1382),
  [9694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4419),
  [9696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4420),
  [9698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3502),
  [9700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2597),
  [9702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2598),
  [9704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_single_repeat1, 1, 0, 0),
  [9706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6820),
  [9708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2600),
  [9710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2601),
  [9712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2602),
  [9714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2603),
  [9716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2604),
  [9718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2605),
  [9720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3585),
  [9722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2606),
  [9724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2607),
  [9726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3468),
  [9728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3533),
  [9730] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_heredoc_single_repeat1, 2, 0, 0),
  [9732] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_heredoc_single_repeat1, 2, 0, 0), SHIFT_REPEAT(6715),
  [9735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3924),
  [9737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3815),
  [9739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3472),
  [9741] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_square_repeat1, 2, 0, 0),
  [9743] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_square_repeat1, 2, 0, 0), SHIFT_REPEAT(6725),
  [9746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3473),
  [9748] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_parenthesis_repeat1, 1, 0, 0),
  [9750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6799),
  [9752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2617),
  [9754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2618),
  [9756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2619),
  [9758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2620),
  [9760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2621),
  [9762] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2623),
  [9764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2624),
  [9766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2625),
  [9768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2626),
  [9770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2630),
  [9772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3925),
  [9774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4081),
  [9776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4082),
  [9778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2927),
  [9780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3320),
  [9782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3321),
  [9784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3322),
  [9786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3323),
  [9788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3324),
  [9790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3325),
  [9792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3326),
  [9794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3327),
  [9796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3328),
  [9798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3837),
  [9800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4084),
  [9802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4085),
  [9804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4086),
  [9806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4087),
  [9808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4088),
  [9810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4089),
  [9812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4090),
  [9814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4091),
  [9816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3539),
  [9818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3336),
  [9820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3337),
  [9822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3338),
  [9824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3339),
  [9826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3340),
  [9828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3341),
  [9830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3342),
  [9832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3343),
  [9834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3344),
  [9836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3345),
  [9838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3841),
  [9840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3842),
  [9842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2496),
  [9844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2498),
  [9846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3843),
  [9848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2572),
  [9850] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_angle_repeat1, 2, 0, 0),
  [9852] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_angle_repeat1, 2, 0, 0), SHIFT_REPEAT(6729),
  [9855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2666),
  [9857] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_double_repeat1, 1, 0, 0),
  [9859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6836),
  [9861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2675),
  [9863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3844),
  [9865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2681),
  [9867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2682),
  [9869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2686),
  [9871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3540),
  [9873] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_bar_repeat1, 2, 0, 0),
  [9875] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_bar_repeat1, 2, 0, 0), SHIFT_REPEAT(6736),
  [9878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2687),
  [9880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3541),
  [9882] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_slash_repeat1, 2, 0, 0),
  [9884] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_slash_repeat1, 2, 0, 0), SHIFT_REPEAT(6739),
  [9887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2691),
  [9889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6849),
  [9891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(338),
  [9893] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__items_with_trailing_separator, 4, 0, 0),
  [9895] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__stab_clause_arguments_with_parentheses_repeat1, 2, 0, 0), SHIFT_REPEAT(573),
  [9898] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_with_trailing_separator, 2, 0, 0),
  [9900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(273),
  [9902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3984),
  [9904] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_single_repeat1, 2, 0, 0),
  [9906] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_single_repeat1, 2, 0, 0), SHIFT_REPEAT(6540),
  [9909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1747),
  [9911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1748),
  [9913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1749),
  [9915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1753),
  [9917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1754),
  [9919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1757),
  [9921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1760),
  [9923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1762),
  [9925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1763),
  [9927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1764),
  [9929] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_parenthesis_repeat1, 2, 0, 0),
  [9931] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_parenthesis_repeat1, 2, 0, 0), SHIFT_REPEAT(6573),
  [9934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3994),
  [9936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3995),
  [9938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3996),
  [9940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3997),
  [9942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3998),
  [9944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3999),
  [9946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3503),
  [9948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3504),
  [9950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4000),
  [9952] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__items_with_trailing_separator_repeat1, 2, 0, 0), SHIFT_REPEAT(799),
  [9955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4001),
  [9957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3505),
  [9959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3506),
  [9961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3507),
  [9963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3255),
  [9965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6798),
  [9967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4002),
  [9969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4003),
  [9971] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_heredoc_single_repeat1, 1, 0, 0),
  [9973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6781),
  [9975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3610),
  [9977] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_heredoc_double_repeat1, 1, 0, 0),
  [9979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6775),
  [9981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3620),
  [9983] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_curly_repeat1, 1, 0, 0),
  [9985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6828),
  [9987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3683),
  [9989] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_square_repeat1, 1, 0, 0),
  [9991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6807),
  [9993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3686),
  [9995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3756),
  [9997] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_angle_repeat1, 1, 0, 0),
  [9999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6772),
  [10001] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_keywords_repeat1, 2, 0, 0), SHIFT_REPEAT(4801),
  [10004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3515),
  [10006] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3456),
  [10008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3882),
  [10010] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_bar_repeat1, 1, 0, 0),
  [10012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6778),
  [10014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3883),
  [10016] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_slash_repeat1, 1, 0, 0),
  [10018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6782),
  [10020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1772),
  [10022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1773),
  [10024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4012),
  [10026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4013),
  [10028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4014),
  [10030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4015),
  [10032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4016),
  [10034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4017),
  [10036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1774),
  [10038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4018),
  [10040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4019),
  [10042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4020),
  [10044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4790),
  [10046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4792),
  [10048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4021),
  [10050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1775),
  [10052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1776),
  [10054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3457),
  [10056] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__quoted_double_repeat1, 2, 0, 0),
  [10058] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__quoted_double_repeat1, 2, 0, 0), SHIFT_REPEAT(6645),
  [10061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1777),
  [10063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3754),
  [10065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1780),
  [10067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1781),
  [10069] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_with_parentheses, 5, 0, 0),
  [10071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(222),
  [10073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(212),
  [10075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(194),
  [10077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(228),
  [10079] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_with_parentheses, 6, 0, 0),
  [10081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4614),
  [10083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4630),
  [10085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(613),
  [10087] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_left, 1, 0, 0),
  [10089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4600),
  [10091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4601),
  [10093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(634),
  [10095] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_without_parentheses, 4, 0, 0),
  [10097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4595),
  [10099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4610),
  [10101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4599),
  [10103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4604),
  [10105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4611),
  [10107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4598),
  [10109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4605),
  [10111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4606),
  [10113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4617),
  [10115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4609),
  [10117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4596),
  [10119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4597),
  [10121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4602),
  [10123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4603),
  [10125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4607),
  [10127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4608),
  [10129] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_arguments_with_parentheses, 4, 0, 0),
  [10131] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_anonymous_function_repeat1, 2, 0, 0),
  [10133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4612),
  [10135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4613),
  [10137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4615),
  [10139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4616),
  [10141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(182),
  [10143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4618),
  [10145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4619),
  [10147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4620),
  [10149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4621),
  [10151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4622),
  [10153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4623),
  [10155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4624),
  [10157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4625),
  [10159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4626),
  [10161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4627),
  [10163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4628),
  [10165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4629),
  [10167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(252),
  [10169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1198),
  [10171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3808),
  [10173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6882),
  [10175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2029),
  [10177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(198),
  [10179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1530),
  [10181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4200),
  [10183] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__stab_clause_left, 1, 0, 7),
  [10185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3040),
  [10187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4201),
  [10189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1678),
  [10191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1696),
  [10193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2030),
  [10195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1679),
  [10197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1850),
  [10199] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_with_trailing_separator, 3, 0, 0),
  [10201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(310),
  [10203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4240),
  [10205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4349),
  [10207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1403),
  [10209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3006),
  [10211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3447),
  [10213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4460),
  [10215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3713),
  [10217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3714),
  [10219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2272),
  [10221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6800),
  [10223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3063),
  [10225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1303),
  [10227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2274),
  [10229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2031),
  [10231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3069),
  [10233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4373),
  [10235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2309),
  [10237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3213),
  [10239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2717),
  [10241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2683),
  [10243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(214),
  [10245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2459),
  [10247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4292),
  [10249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2609),
  [10251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2058),
  [10253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3432),
  [10255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1206),
  [10257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4102),
  [10259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6931),
  [10261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(193),
  [10263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3820),
  [10265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1639),
  [10267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1133),
  [10269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4473),
  [10271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4440),
  [10273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1759),
  [10275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6916),
  [10277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1294),
  [10279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3391),
  [10281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2109),
  [10283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3088),
  [10285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3392),
  [10287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3716),
  [10289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6925),
  [10291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(229),
  [10293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1280),
  [10295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1197),
  [10297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1835),
  [10299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6911),
  [10301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(205),
  [10303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1482),
  [10305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4228),
  [10307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3601),
  [10309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6892),
  [10311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(213),
  [10313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1503),
  [10315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3393),
  [10317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2042),
  [10319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6948),
  [10321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(179),
  [10323] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [10325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1142),
  [10327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1529),
  [10329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6915),
  [10331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(195),
  [10333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4405),
  [10335] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__call_arguments_with_trailing_separator, 4, 0, 0),
  [10337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2267),
  [10339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6986),
  [10341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(210),
  [10343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2955),
  [10345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4308),
  [10347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6992),
  [10349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(223),
  [10351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2352),
  [10353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1132),
  [10355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6972),
  [10357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(238),
  [10359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4558),
  [10361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3403),
  [10363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6980),
  [10365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(253),
  [10367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3068),
  [10369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3155),
  [10371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1531),
  [10373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1550),
  [10375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3801),
  [10377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1738),
  [10379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3143),
  [10381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2586),
  [10383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4416),
  [10385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2319),
  [10387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4281),
  [10389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4467),
  [10391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(177),
  [10393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4439),
  [10395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3741),
  [10397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2941),
  [10399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3156),
  [10401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1680),
  [10403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2942),
  [10405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(329),
  [10407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4415),
  [10409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3073),
  [10411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1273),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__quoted_content_i_single = 0,
  ts_external_token__quoted_content_i_double = 1,
  ts_external_token__quoted_content_i_heredoc_single = 2,
  ts_external_token__quoted_content_i_heredoc_double = 3,
  ts_external_token__quoted_content_i_parenthesis = 4,
  ts_external_token__quoted_content_i_curly = 5,
  ts_external_token__quoted_content_i_square = 6,
  ts_external_token__quoted_content_i_angle = 7,
  ts_external_token__quoted_content_i_bar = 8,
  ts_external_token__quoted_content_i_slash = 9,
  ts_external_token__quoted_content_single = 10,
  ts_external_token__quoted_content_double = 11,
  ts_external_token__quoted_content_heredoc_single = 12,
  ts_external_token__quoted_content_heredoc_double = 13,
  ts_external_token__quoted_content_parenthesis = 14,
  ts_external_token__quoted_content_curly = 15,
  ts_external_token__quoted_content_square = 16,
  ts_external_token__quoted_content_angle = 17,
  ts_external_token__quoted_content_bar = 18,
  ts_external_token__quoted_content_slash = 19,
  ts_external_token__newline_before_do = 20,
  ts_external_token__newline_before_binary_operator = 21,
  ts_external_token__newline_before_comment = 22,
  ts_external_token__before_unary_op = 23,
  ts_external_token__not_in = 24,
  ts_external_token__quoted_atom_start = 25,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__quoted_content_i_single] = sym__quoted_content_i_single,
  [ts_external_token__quoted_content_i_double] = sym__quoted_content_i_double,
  [ts_external_token__quoted_content_i_heredoc_single] = sym__quoted_content_i_heredoc_single,
  [ts_external_token__quoted_content_i_heredoc_double] = sym__quoted_content_i_heredoc_double,
  [ts_external_token__quoted_content_i_parenthesis] = sym__quoted_content_i_parenthesis,
  [ts_external_token__quoted_content_i_curly] = sym__quoted_content_i_curly,
  [ts_external_token__quoted_content_i_square] = sym__quoted_content_i_square,
  [ts_external_token__quoted_content_i_angle] = sym__quoted_content_i_angle,
  [ts_external_token__quoted_content_i_bar] = sym__quoted_content_i_bar,
  [ts_external_token__quoted_content_i_slash] = sym__quoted_content_i_slash,
  [ts_external_token__quoted_content_single] = sym__quoted_content_single,
  [ts_external_token__quoted_content_double] = sym__quoted_content_double,
  [ts_external_token__quoted_content_heredoc_single] = sym__quoted_content_heredoc_single,
  [ts_external_token__quoted_content_heredoc_double] = sym__quoted_content_heredoc_double,
  [ts_external_token__quoted_content_parenthesis] = sym__quoted_content_parenthesis,
  [ts_external_token__quoted_content_curly] = sym__quoted_content_curly,
  [ts_external_token__quoted_content_square] = sym__quoted_content_square,
  [ts_external_token__quoted_content_angle] = sym__quoted_content_angle,
  [ts_external_token__quoted_content_bar] = sym__quoted_content_bar,
  [ts_external_token__quoted_content_slash] = sym__quoted_content_slash,
  [ts_external_token__newline_before_do] = sym__newline_before_do,
  [ts_external_token__newline_before_binary_operator] = sym__newline_before_binary_operator,
  [ts_external_token__newline_before_comment] = sym__newline_before_comment,
  [ts_external_token__before_unary_op] = sym__before_unary_op,
  [ts_external_token__not_in] = sym__not_in,
  [ts_external_token__quoted_atom_start] = sym__quoted_atom_start,
};

static const bool ts_external_scanner_states[27][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__quoted_content_i_single] = true,
    [ts_external_token__quoted_content_i_double] = true,
    [ts_external_token__quoted_content_i_heredoc_single] = true,
    [ts_external_token__quoted_content_i_heredoc_double] = true,
    [ts_external_token__quoted_content_i_parenthesis] = true,
    [ts_external_token__quoted_content_i_curly] = true,
    [ts_external_token__quoted_content_i_square] = true,
    [ts_external_token__quoted_content_i_angle] = true,
    [ts_external_token__quoted_content_i_bar] = true,
    [ts_external_token__quoted_content_i_slash] = true,
    [ts_external_token__quoted_content_single] = true,
    [ts_external_token__quoted_content_double] = true,
    [ts_external_token__quoted_content_heredoc_single] = true,
    [ts_external_token__quoted_content_heredoc_double] = true,
    [ts_external_token__quoted_content_parenthesis] = true,
    [ts_external_token__quoted_content_curly] = true,
    [ts_external_token__quoted_content_square] = true,
    [ts_external_token__quoted_content_angle] = true,
    [ts_external_token__quoted_content_bar] = true,
    [ts_external_token__quoted_content_slash] = true,
    [ts_external_token__newline_before_do] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
    [ts_external_token__before_unary_op] = true,
    [ts_external_token__not_in] = true,
    [ts_external_token__quoted_atom_start] = true,
  },
  [2] = {
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
    [ts_external_token__before_unary_op] = true,
    [ts_external_token__not_in] = true,
    [ts_external_token__quoted_atom_start] = true,
  },
  [3] = {
    [ts_external_token__newline_before_do] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
    [ts_external_token__before_unary_op] = true,
    [ts_external_token__not_in] = true,
    [ts_external_token__quoted_atom_start] = true,
  },
  [4] = {
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
    [ts_external_token__not_in] = true,
  },
  [5] = {
    [ts_external_token__newline_before_do] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
    [ts_external_token__not_in] = true,
  },
  [6] = {
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [7] = {
    [ts_external_token__quoted_content_i_double] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [8] = {
    [ts_external_token__quoted_content_i_single] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [9] = {
    [ts_external_token__quoted_content_i_heredoc_single] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [10] = {
    [ts_external_token__quoted_content_i_heredoc_double] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [11] = {
    [ts_external_token__quoted_content_i_parenthesis] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [12] = {
    [ts_external_token__quoted_content_i_curly] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [13] = {
    [ts_external_token__quoted_content_i_square] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [14] = {
    [ts_external_token__quoted_content_i_angle] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [15] = {
    [ts_external_token__quoted_content_i_bar] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [16] = {
    [ts_external_token__quoted_content_i_slash] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [17] = {
    [ts_external_token__quoted_content_parenthesis] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [18] = {
    [ts_external_token__quoted_content_double] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [19] = {
    [ts_external_token__quoted_content_single] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [20] = {
    [ts_external_token__quoted_content_heredoc_single] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [21] = {
    [ts_external_token__quoted_content_heredoc_double] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [22] = {
    [ts_external_token__quoted_content_curly] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [23] = {
    [ts_external_token__quoted_content_square] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [24] = {
    [ts_external_token__quoted_content_angle] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [25] = {
    [ts_external_token__quoted_content_bar] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
  [26] = {
    [ts_external_token__quoted_content_slash] = true,
    [ts_external_token__newline_before_binary_operator] = true,
    [ts_external_token__newline_before_comment] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_elixir_external_scanner_create(void);
void tree_sitter_elixir_external_scanner_destroy(void *);
bool tree_sitter_elixir_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_elixir_external_scanner_serialize(void *, char *);
void tree_sitter_elixir_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_elixir(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_elixir_external_scanner_create,
      tree_sitter_elixir_external_scanner_destroy,
      tree_sitter_elixir_external_scanner_scan,
      tree_sitter_elixir_external_scanner_serialize,
      tree_sitter_elixir_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "elixir",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 0,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
